//-----------------------------------------------------------------------
//  This file is part of the UC Merced Microsoft Robotics Developer Studio 
// Project Code Base.
//
//  Created by Benjamin Balaguer under the supervision of Stefano Carpin.
//
//  $File: ComStation2009Controller.cs $
//-----------------------------------------------------------------------

using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml;
using W3C.Soap;
using Microsoft.Dss.Core.DsspHttp;
using System.Threading;
using submgr = Microsoft.Dss.Services.SubscriptionManager;
using roboCupGUI = Microsoft.Robotics.Services.RoboCup2009GUI.Proxy;
using System.Net;

namespace Robotics.ComStation2009Controller
{
    /// <summary>
    /// Implementation class for ComStation2009Controller
    /// </summary>
    [DisplayName("ComStation2009Controller")]
    [Description("The ComStation2009Controller Service")]
    [Contract(Contract.Identifier)]
    public class ComStation2009ControllerService : DsspServiceBase
    {
        /// <summary>
        /// _state
        /// </summary>
        private ComStation2009ControllerState _state = new ComStation2009ControllerState();
        private GameBot USARSim;
        private WSS wss;
        private System.Object cmd = new System.Object();
        private string ownIP;

        // Configuration Variables
        private int wssGUIPort = 0;
        private int ownWSSPort;
        private string USARSimIP;
        private int USARSimPort;
        private int ImgSrvPort;
        private int wssPort;
        private string MOASTIP;
        private int MOASTPort;
        private string AerialRobotName;
        private string GroundRobotName;
        private string ComStationName;
        private float WSSCutOff;
        
        /// <summary>
        /// _main Port
        /// </summary>
        [ServicePort("/comstation2009controller", AllowMultipleInstances=false)]
        private ComStation2009ControllerOperations _mainPort = new ComStation2009ControllerOperations();

        // Partner
        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        private submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();
        
        [Partner("Clock", Contract = roboCupGUI.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        roboCupGUI.RoboCup2009GUIOperations _clockPort = new roboCupGUI.RoboCup2009GUIOperations();
        roboCupGUI.RoboCup2009GUIOperations _clockNotify = new roboCupGUI.RoboCup2009GUIOperations();
        

        
        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public ComStation2009ControllerService(DsspServiceCreationPort creationPort) : 
                base(creationPort)
        {
            wss = new WSS(_mainPort);
            USARSim = new GameBot(_mainPort, wss);
        }
        
        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
			base.Start();

            // Read Configuration File
            ReadConfigFile("samples/Config/RoboCupGUI.xml");

            // Get own IP address
            IPAddress[] aryLocalAddr = null;
            String strHostName = "";
            strHostName = Dns.GetHostName();
            IPHostEntry ipEntry = Dns.GetHostEntry(strHostName);
            aryLocalAddr = ipEntry.AddressList;
            ownIP = Convert.ToString(aryLocalAddr[0]);

            Activate(Arbiter.Interleave(
                new TeardownReceiverGroup(),
                new ExclusiveReceiverGroup
                (
                    // Messages that are sent to the GUI
                    Arbiter.ReceiveWithIterator<CONNECTED_MSG>(true, _mainPort, CONNECTED_MSGHandler),
                    Arbiter.ReceiveWithIterator<DISCONNECTED_MSG>(true, _mainPort, DISCONNECTED_MSGHandler),
                    Arbiter.ReceiveWithIterator<START_POSES_MSG>(true, _mainPort, START_POSES_MSGHandler),
                    Arbiter.ReceiveWithIterator<INITIALIZED_MSG>(true, _mainPort, INITIALIZED_MSGHandler),
                    // Messages that are received from the GUI
                    Arbiter.Receive<roboCupGUI.CONNECT_CMD>(true, _clockNotify, NotifyCONNECT_CMDHandler),
                    Arbiter.Receive<roboCupGUI.DISCONNECT_CMD>(true, _clockNotify, NotifyDISCONNECT_CMDHandler),
                    Arbiter.Receive<roboCupGUI.GET_POSES_CMD>(true, _clockNotify, NotifyGET_POSES_CMDHandler),
                    Arbiter.Receive<roboCupGUI.INIT_CMD>(true, _clockNotify, NotifyINIT_CMDHandler),
                    Arbiter.Receive<roboCupGUI.CHVIEW_CMD>(true, _clockNotify, NotifyCHVIEW_CMDHandler),
                    Arbiter.ReceiveWithIterator<GET_CONF_CMD>(true, _mainPort, NotifyGET_CONF_CMDHandler),
                    Arbiter.ReceiveWithIterator<DRIVE_AERIAL_CMD>(true, _mainPort, NotifyDRIVE_AERIAL_CMDHandler),
                    Arbiter.ReceiveWithIterator<MIS_PKG_CMD>(true, _mainPort, NotifyMIS_PKG_CMDHandler),
                    Arbiter.ReceiveWithIterator<IMG_SVR_CAM_CMD>(true, _mainPort, NotifyIMG_SVR_CAM_CMDHandler),
                    Arbiter.ReceiveWithIterator<IMG_SVR_OK_CMD>(true, _mainPort, NotifyIMG_SVR_OK_CMDHandler)
                ),
                new ConcurrentReceiverGroup()
            ));

            _clockPort.Subscribe(_clockNotify);
        }

        private void ReadConfigFile(string FileToRead)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(FileToRead);

            // Read the configuration XML file and store results in various variables
            USARSimIP = ((XmlElement)doc.GetElementsByTagName("USARSimIP")[0]).InnerText;
            USARSimPort = Convert.ToInt32(((XmlElement)doc.GetElementsByTagName("USARSimPort")[0]).InnerText);
            ImgSrvPort = Convert.ToInt32(((XmlElement)doc.GetElementsByTagName("ImgSrvPort")[0]).InnerText);
            wssPort = Convert.ToInt32(((XmlElement)doc.GetElementsByTagName("WSSPort")[0]).InnerText);
            ownWSSPort = Convert.ToInt32(((XmlElement)doc.GetElementsByTagName("AirRobotWSSPort")[0]).InnerText);
            MOASTIP = ((XmlElement)doc.GetElementsByTagName("MOASTIP")[0]).InnerText;
            MOASTPort = Convert.ToInt32(((XmlElement)doc.GetElementsByTagName("MOASTPort")[0]).InnerText);
            AerialRobotName = ((XmlElement)doc.GetElementsByTagName("AerialRobotName")[0]).InnerText;
            GroundRobotName = ((XmlElement)doc.GetElementsByTagName("GroundRobotName")[0]).InnerText;
            ComStationName = ((XmlElement)doc.GetElementsByTagName("ComStationName")[0]).InnerText;
            WSSCutOff = Convert.ToSingle(((XmlElement)doc.GetElementsByTagName("WSSCutOff")[0]).InnerText);
        }

        #region DSS handlers
        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Http Get Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> HttpGetHandler(HttpGet httpGet)
        {
            httpGet.ResponsePort.Post(new HttpResponseType(_state));
            yield break;
        }

        /// <summary>
        /// Subscription Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            SubscribeRequestType request = subscribe.Body;
            System.Console.WriteLine("CONTROLLER: Subscribe request from: " + request.Subscriber);

            SubscribeHelper(_submgrPort, request, subscribe.ResponsePort);

            yield break;
        }
        #endregion

        #region Sending Notifications to GUI (Messages) Handlers
        IEnumerator<ITask> CONNECTED_MSGHandler(CONNECTED_MSG connectedMSG)
        {
            base.SendNotification(_submgrPort, connectedMSG);
            connectedMSG.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        IEnumerator<ITask> DISCONNECTED_MSGHandler(DISCONNECTED_MSG disconnectedMSG)
        {
            base.SendNotification(_submgrPort, disconnectedMSG);
            disconnectedMSG.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }
        
        IEnumerator<ITask> START_POSES_MSGHandler(START_POSES_MSG startPosesMSG)
        {
            base.SendNotification(_submgrPort, startPosesMSG);
            startPosesMSG.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        IEnumerator<ITask> INITIALIZED_MSGHandler(INITIALIZED_MSG initializedMSG)
        {
            base.SendNotification(_submgrPort, initializedMSG);
            initializedMSG.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }
        #endregion

        #region Receiving Notifications from GUI (Commands) Handlers
        
        private void NotifyCONNECT_CMDHandler(roboCupGUI.CONNECT_CMD connectCMD)
        {
            USARSim.Connect(connectCMD.Body.IP, connectCMD.Body.gameBotPort);

            // Tell the GUI that we are now connected
            _mainPort.Post(new CONNECTED_MSG());
        }

        private void NotifyDISCONNECT_CMDHandler(roboCupGUI.DISCONNECT_CMD disconnectCMD)
        {
            lock (cmd)
            {
                USARSim.Disconnect();

                // Tell the GUI that we are now disconnected
                _mainPort.Post(new DISCONNECTED_MSG());
            }
        }

        private void NotifyGET_POSES_CMDHandler(roboCupGUI.GET_POSES_CMD getPosesCMD)
        {
            // Ask for the locations of valid start poses from USARSim
            USARSim.WriteLine("GETSTARTPOSES");
        }

        private void NotifyINIT_CMDHandler(roboCupGUI.INIT_CMD initCMD)
        {
            lock (cmd)
            {
                USARSim.WriteLine(USARSim.Cmd.INIT("ComStation", initCMD.Body.Name, initCMD.Body.Location, initCMD.Body.Rotation));

                _mainPort.Post(new INITIALIZED_MSG());
            }
        }

        private void NotifyCHVIEW_CMDHandler(roboCupGUI.CHVIEW_CMD chViewCMD)
        {
            lock (cmd)
            {
                //System.Console.WriteLine("CONTROLLER: Received Change View Request");
                //_mainPort.Post(new IMG_SVR_CAM_CMD(chViewCMD.Body.RobotName, chViewCMD.Body.CameraName, chViewCMD.Body.IP));
                USARSim.WriteLine(USARSim.Cmd.SET(chViewCMD.Body.RobotName, chViewCMD.Body.CameraName, chViewCMD.Body.IP));
                //System.Console.WriteLine("CONTROLLER: Sending Change View Request");
            }
        }

        IEnumerator<ITask> NotifyGET_CONF_CMDHandler(GET_CONF_CMD getConfCMD)
        {
            // Get the robot configuration from USARSim
            USARSim.WriteLine(USARSim.Cmd.GETCONF(getConfCMD.Body.Type, getConfCMD.Body.Name));
            yield break;
        }

        IEnumerator<ITask> NotifyDRIVE_AERIAL_CMDHandler(DRIVE_AERIAL_CMD driveAerialCMD)
        {
            // Send a drive command to USARSim
            USARSim.WriteLine(USARSim.Cmd.DRIVEAerial(driveAerialCMD.Body.AltitudeVelocity, driveAerialCMD.Body.LinearVelocity,
                                                      driveAerialCMD.Body.LateralVelocity, driveAerialCMD.Body.RotationalVelocity,
                                                      driveAerialCMD.Body.Normalized));
            yield break;
        }

        IEnumerator<ITask> NotifyMIS_PKG_CMDHandler(MIS_PKG_CMD misPkgCMD)
        {
            // Send a mission package command (usually a pan/tilt) to USARSim
            USARSim.WriteLine(USARSim.Cmd.MISPKG(misPkgCMD.Body.Name, misPkgCMD.Body.Link)); 
            yield break;
        }

        IEnumerator<ITask> NotifyIMG_SVR_CAM_CMDHandler(IMG_SVR_CAM_CMD imgSvrCamCMD)
        {
            // Send a command to switch the camera view to a different robot in USARSim
            USARSim.WriteLine(USARSim.Cmd.SET(imgSvrCamCMD.Body.RobotName, imgSvrCamCMD.Body.CamName, imgSvrCamCMD.Body.IP));
            yield break;
        }

        IEnumerator<ITask> NotifyIMG_SVR_OK_CMDHandler(IMG_SVR_OK_CMD imgSvrOKCMD)
        {
            // Send an "OK" command to USARSim, to get another image
            USARSim.SendOKImgSvr();
            yield break;
        }
        #endregion
    }
}
