//-----------------------------------------------------------------------
//  This file is part of the UC Merced Microsoft Robotics Developer Studio 
// Project Code Base.
//
//  Created by Benjamin Balaguer under the supervision of Stefano Carpin.
//
//  $File: WSS.cs $
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Drawing;
using System.Threading;
using System.Collections;
using System.Windows.Media.Imaging;

namespace Robotics.ComStation2009Controller
{
    #region WSS Communication Interface
    class WSS
    {
        #region Variables
        // To Talk to the WSS
        private static TcpClient wssClient;
        private static BufferedStream wssStream;

        // To Talk to the GUI
        private static TcpClient wssGUIClient;
        private static BufferedStream wssGUIStream;

        // To Listen to the GUI
        private static TcpListener wssAirRobotListener;
        private static TcpClient wssAirRobotListenerClient;
        private static BufferedStream wssAirRobotListenerStream;
        private static Thread wssAirRobotListenThread;

        private ComStation2009ControllerOperations _mainPort;

        public WSSParser Parser;
        #endregion

        #region Constructor
        public WSS(ComStation2009ControllerOperations mainPort) // Constructor
        {
            _mainPort = mainPort;
            Parser = new WSSParser(_mainPort);
        }
        #endregion

        #region Connect/Disconnect
        public void Connect(string Type, string IP, int Port)
        {
            if (Type.CompareTo("WSS") == 0)
            {
                wssClient = new TcpClient();
                wssClient.Connect(new IPEndPoint(IPAddress.Parse(IP), Port));
                wssStream = new BufferedStream(wssClient.GetStream(), 10000);
            }
            else if (Type.CompareTo("GUI") == 0)
            {
                wssGUIClient = new TcpClient();
                wssGUIClient.Connect(new IPEndPoint(IPAddress.Parse(IP), Port));
                wssGUIStream = new BufferedStream(wssGUIClient.GetStream(), 10000);
            }
            else if (Type.CompareTo("AirRobotListen") == 0)
            {
                wssAirRobotListener = new TcpListener(new IPEndPoint(IPAddress.Parse(IP), Port));
                wssAirRobotListenThread = new Thread(new ParameterizedThreadStart(HandleWSSAirRobotListen));
                wssAirRobotListenThread.Start();
            }
        }

        public void Disconnect(string Type)
        {
            if (Type.CompareTo("WSS") == 0)
            {
                wssStream.Close();
                wssClient.Close();
            }
            else if (Type.CompareTo("GUI") == 0)
            {
                wssGUIStream.Close();
                wssGUIClient.Close();
            }
            else if (Type.CompareTo("AirRobotListen") == 0)
            {
                wssAirRobotListenThread.Abort();
                wssAirRobotListenerStream.Close();
                wssAirRobotListenerClient.Close();
                wssAirRobotListener.Stop();
            }
        }
        #endregion

        #region WSS Listen Threads
        private void HandleWSSAirRobotListen(object sender)
        {
            wssAirRobotListener.Start();
            wssAirRobotListenerClient = wssAirRobotListener.AcceptTcpClient();
            wssAirRobotListenerStream = new BufferedStream(wssAirRobotListenerClient.GetStream(), 10000);
            
            while (true)
            {
                try
                {
                    Parser.Parse(ReadLine("AirRobotListen"));
                }
                catch (Exception e)
                {
                    System.Console.WriteLine("Caught ReadLine Exception: " + e.ToString());
                }
            }
        }
        #endregion

        #region Read/Write to WSS
        public string ReadLine(string Type)
        {
            char currentChar;
            bool done = false;
            string result = "";
            BufferedStream stream;

            if (Type.CompareTo("AirRobotListen") == 0)
                stream = wssAirRobotListenerStream;
            else
                stream = wssStream;

            for (int i = 0; !done; i++)
            {
                currentChar = (char)stream.ReadByte();

                // If we have found a carriage return
                if ((int)currentChar == 13)
                {
                    stream.ReadByte(); // Discard the next character (line feed)
                    done = true;       // Set the condition to drop out of the loop
                }
                else
                {
                    // Add the character to the resulting string
                    result = result + currentChar;
                }
            }

            return result;
        }

        public void WriteLine(string Type, string cmd)
        {
            ASCIIEncoding encoder = new ASCIIEncoding();
            byte[] buffer = encoder.GetBytes(cmd + (char)13 + (char)10);
            BufferedStream stream;

            if (Type.CompareTo("GUI") == 0)
                stream = wssGUIStream;
            else
                stream = wssStream;

            try
            {
                stream.Write(buffer, 0, buffer.Length);
                stream.Flush();
            }
            catch (Exception)
            {
                //System.Console.WriteLine("Caught an Exception while Writing: " + e.ToString());
            }

        }

        public void WriteBinary(string Type, ref byte[] buffer)
        {
            ASCIIEncoding encoder = new ASCIIEncoding();
            //byte[] buffer = encoder.GetBytes(cmd + (char)13 + (char)10);
            BufferedStream stream;

            if (Type.CompareTo("GUI") == 0)
                stream = wssGUIStream;
            else
                stream = wssStream;

            try
            {
                stream.Write(buffer, 0, buffer.Length);
                stream.Flush();
            }
            catch (Exception)
            {
                //System.Console.WriteLine("Caught an Exception while Writing: " + e.ToString());
            }

        }
        #endregion

        #region WSS CMDs Functions
        public bool SendInit(string RobotName, int Port)
        {
            bool result = false;
            WriteLine("WSS", "INIT {Robot " + RobotName + "} {Port " + Convert.ToString(Port) + "}");

            if (ReadLine("WSS").CompareTo(INITResponse()) == 0)
                result = true;

            return result;
        }

        public float GetSignalStrength(string RobotName)
        {
            WriteLine("WSS", "GETSS {Robot " + RobotName + "}");

            return SignalStrengthResponse(ReadLine("WSS"));
        }

        public int GetRobotPort(string RobotName)
        {
            WriteLine("WSS", "DNS {Robot " + RobotName + "}");

            return DNSResponse(ReadLine("WSS"));
        }
        #endregion

        #region WSS Responses
        public string INITResponse()
        {
            return "INITREPLY {Status OK}";
        }

        public float SignalStrengthResponse(string msg)
        {
            string tmp = msg.Substring(msg.IndexOf("{Strength ") + 10, msg.Length - 1 - msg.IndexOf("{Strength ") - 10);

            if (tmp.CompareTo("NaN") == 0)
                return -1f;
            else if (tmp.CompareTo("INF") == 0)
                return -2f;
            else
                return Convert.ToSingle(tmp);
        }

        public int DNSResponse(string msg)
        {
            string tmp = msg.Substring(msg.IndexOf("{Port ") + 6, msg.Length - 1 - msg.IndexOf("{Port ") - 6);

            return Convert.ToInt32(tmp);
        }
        #endregion
    }
    #endregion

    #region WSS String Parser
    class WSSParser
    {
        private ComStation2009ControllerOperations _mainPort;

        #region Constructor
        public WSSParser(ComStation2009ControllerOperations mainPort)
        {
            _mainPort = mainPort;
        }
        #endregion

        #region Main Parsing Function
        public void Parse(string toParse)
        {
            try
            {
                if (toParse.IndexOf("CON", 0) == 0)
                {
                    ParseCON(DeleteTag(toParse, "CON"));
                }
                else if (toParse.IndexOf("GSP", 0) == 0)
                {
                    _mainPort.Post(new GET_POSES_CMD());
                }
                else if(toParse.IndexOf("INI", 0) == 0)
                {
                    ParseINI(DeleteTag(toParse, "INI"));
                }
                else if (toParse.IndexOf("GET", 0) == 0)
                {
                    ParseGET(DeleteTag(toParse, "GET"));
                }
                else if (toParse.IndexOf("ADR", 0) == 0)
                {
                    ParseADR(DeleteTag(toParse, "ADR"));
                }
                else if (toParse.IndexOf("MIS", 0) == 0)
                {
                    ParseMIS(DeleteTag(toParse, "MIS"));
                }
                else if (toParse.IndexOf("CAC", 0) == 0)
                {
                    ParseCAC(DeleteTag(toParse, "CAC"));
                }
                else if((toParse.IndexOf("IOK", 0) == 0))
                {
                    _mainPort.Post(new IMG_SVR_OK_CMD());
                }
            }
            catch (Exception)
            {
            }
        }
        #endregion

        #region CON Message Parsing
        private void ParseCON(string CON)
        {
            string tmp;

            // Get the IP address
            string IP = GetParameter(CON, " ");
            CON = DeleteTag(CON, IP);

            // Get the GameBot Port
            tmp = GetParameter(CON, " ");
            int gbPort = Convert.ToInt32(tmp);
            CON = DeleteTag(CON, tmp);

            // Get the Image Server Port
            tmp = GetParameter(CON, " ");
            int imgsrvPort = Convert.ToInt32(tmp);

            // Post result to appropriate handler
            _mainPort.Post(new CONNECT_CMD(IP, gbPort, imgsrvPort));
        }
        #endregion

        #region INI Message Parsing
        private void ParseINI(string INI)
        {
            // Get the robot type
            string Type = GetParameter(INI, " ");
            INI = DeleteTag(INI, Type);

            // Get the robot name
            string Name = GetParameter(INI, " ");
            INI = DeleteTag(INI, Name);

            // Get the Location
            string Loc = GetParameter(INI, " ");
            INI = DeleteTag(INI, Loc);

            string Rot = GetParameter(INI, " ");
            INI = DeleteTag(INI, Rot);

            // Post result to appropriate handler
            _mainPort.Post(new INIT_CMD(Type, Name, Loc, Rot));
            
        }
        #endregion

        #region GET Message Parsing
        private void ParseGET(string GET)
        {
            // Get the GET type
            string Type = GetParameter(GET, " ");
            GET = DeleteTag(GET, Type);

            if (Type.CompareTo("MisPkg") == 0)
            {
                ParseGETMisPkg(GET, Type);
            }

        }

        private void ParseGETMisPkg(string GET, string Type)
        {
            // If no name was given
            if (GET.CompareTo("") == 0)
            {
                _mainPort.Post(new GET_CONF_CMD(Type, ""));
            }
            else
            {
                string Name = GetParameter(GET, " ");
                GET = DeleteTag(GET, Name);
                _mainPort.Post(new GET_CONF_CMD(Type, Name));
            }
        }
        #endregion

        #region ADR Message Parsing
        private void ParseADR(string ADR)
        {
            // Get the Altitude Velocity
            string AltitudeVelocity = GetParameter(ADR, " ");
            ADR = DeleteTag(ADR, AltitudeVelocity);

            // Get the Linear Velocity
            string LinearVelocity = GetParameter(ADR, " ");
            ADR = DeleteTag(ADR, LinearVelocity);

            // Get the Lateral Velocity
            string LateralVelocity = GetParameter(ADR, " ");
            ADR = DeleteTag(ADR, LateralVelocity);

            // Get the Rotational Velocity
            string RotationalVelocity = GetParameter(ADR, " ");
            ADR = DeleteTag(ADR, RotationalVelocity);

            // Get the Normalized factor
            string tmp = GetParameter(ADR, " ");
            bool Normalized = Convert.ToBoolean(tmp);
            ADR = DeleteTag(ADR, tmp);

            _mainPort.Post(new DRIVE_AERIAL_CMD(AltitudeVelocity, LinearVelocity, LateralVelocity, RotationalVelocity, Normalized));
        }
        #endregion

        #region MIS Message Parsing
        private void ParseMIS(string MIS)
        {
            // Get the Mission Package Name
            string Name = GetParameter(MIS, " ");
            MIS = DeleteTag(MIS, Name);

            // Get the Number of Links
            string tmp = GetParameter(MIS, " ");
            int links = Convert.ToInt32(tmp);
            MIS = DeleteTag(MIS, tmp);

            List<LinkProperty> link = new List<LinkProperty>();

            for (int i = 0; i < links; i++)
            {
                LinkProperty aLink = new LinkProperty();

                // Get the Link ID
                tmp = GetParameter(MIS, " ");
                aLink.ID = Convert.ToInt32(tmp);
                MIS = DeleteTag(MIS, tmp);

                // Get the Link Desired Value
                tmp = GetParameter(MIS, " ");
                aLink.DesiredVal = Convert.ToSingle(tmp);
                MIS = DeleteTag(MIS, tmp);

                // Get the Link Desired Order
                tmp = GetParameter(MIS, " ");
                aLink.DesiredOrder = Convert.ToInt32(tmp);
                MIS = DeleteTag(MIS, tmp);

                // Add link to list
                link.Add(aLink);
            }

            _mainPort.Post(new MIS_PKG_CMD(Name, link));
        }
        #endregion

        #region CAC Message Parsing
        private void ParseCAC(string CAC)
        {
            //string msg = "CAC ZeFrenchie Camera " + txtMachine.Text + " ";

            // Get the Robot Name
            string Name = GetParameter(CAC, " ");
            CAC = DeleteTag(CAC, Name);

            // Get the Type
            string CamName = GetParameter(CAC, " ");
            CAC = DeleteTag(CAC, CamName);

            // Get the IP
            string IP = GetParameter(CAC, " ");
            CAC = DeleteTag(CAC, IP);

            _mainPort.Post(new IMG_SVR_CAM_CMD(Name, CamName, IP));
        }
        #endregion

        #region String Operations
        // Deletes the front tag PLUS 1 (the delimiter)
        private string DeleteTag(string toParse, string Tag)
        {
            return toParse.Substring(Tag.Length + 1, toParse.Length - Tag.Length - 1);
        }

        // Gets the parameter from the front of the string to the delimeter
        private string GetParameter(string toParse, string delimiter)
        {
            return toParse.Substring(0, toParse.IndexOf(delimiter));
        }
        #endregion
    }
    #endregion
}
