//-----------------------------------------------------------------------
//  This file is part of the UC Merced Microsoft Robotics Developer Studio 
// Project Code Base.
//
//  Created by Benjamin Balaguer under the supervision of Stefano Carpin.
//  Based on SimpleDashboard.cs from UX\SimpleDashboard.
//
//  $File: RoboCup2009GUI.cs $
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using W3C.Soap;
using Microsoft.Ccr.Core;
using Microsoft.Ccr.Adapters.WinForms;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using Microsoft.Dss.Services.Serializer;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using cs = Microsoft.Dss.Services.Constructor;
using ds = Microsoft.Dss.Services.Directory;
using game = Microsoft.Robotics.Services.GameController.Proxy;
using submgr = Microsoft.Dss.Services.SubscriptionManager;
using System.ComponentModel;
using System.Threading;
using Microsoft.Dss.Core.DsspHttp;
using airRobotController = Robotics.ComStation2009Controller.Proxy;
using System.Drawing;

namespace Microsoft.Robotics.Services.RoboCup2009GUI
{

    /// <summary>
    /// RoboCup GUI Service
    /// </summary>
    [DisplayName("RoboCup GUI")]
    [Description("Provides access to a simple Windows UI for interacting with sensor and actuator services.\n(Partner with the 'Game Controller' service.)")]
    [Contract(Contract.Identifier)]
    class RoboCup2009GUIService : DsspServiceBase
    {
        // shared access to state is protected by the interleave pattern
        // when we activate the handlers
        RoboCup2009GUIState _state = new RoboCup2009GUIState();

        [ServicePort("/robocup2009gui", AllowMultipleInstances = false)]
        RoboCup2009GUIOperations _mainPort = new RoboCup2009GUIOperations();

        // Partner
        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        private submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        [Partner("Clock", Contract = airRobotController.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        airRobotController.ComStation2009ControllerOperations _clockPort = new airRobotController.ComStation2009ControllerOperations();
        airRobotController.ComStation2009ControllerOperations _clockNotify = new airRobotController.ComStation2009ControllerOperations();

        [Partner("GameController", Contract = game.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        game.GameControllerOperations _gameControllerPort = new game.GameControllerOperations();
        game.GameControllerOperations _gameControllerNotify = new game.GameControllerOperations();
        Port<Shutdown> _gameControllerShutdown = new Port<Shutdown>();

        DriveControl _driveControl;
        DriveControlEvents _eventsPort = new DriveControlEvents();

        private bool Connected = false;

        #region Startup
        /// <summary>
        /// RoboCup2009GUIService Default DSS Constuctor
        /// </summary>
        /// <param name="pCreate"></param>
        public RoboCup2009GUIService(DsspServiceCreationPort pCreate) : base(pCreate)
        {
        }

        /// <summary>
        /// Entry Point for the RoboCup2009GUI Service
        /// </summary>
        protected override void Start()
        {
            Activate(Arbiter.Interleave(
                new TeardownReceiverGroup
                (
                    Arbiter.Receive<DsspDefaultDrop>(false, _mainPort, DropHandler)
                ),
                new ExclusiveReceiverGroup
                (
                    Arbiter.ReceiveWithIterator<OnLoad>(true, _eventsPort, OnLoadHandler),
                    Arbiter.Receive<OnClosed>(true, _eventsPort, OnClosedHandler),
                    Arbiter.ReceiveWithIterator<OnChangeJoystick>(true, _eventsPort, OnChangeJoystickHandler),
                    // Commands that are sent to the AirRobot Controller
                    Arbiter.ReceiveWithIterator<CONNECT_CMD>(true, _mainPort, CONNECT_CMDHandler),
                    Arbiter.ReceiveWithIterator<GET_POSES_CMD>(true, _mainPort, GET_POSES_CMDHandler),
                    Arbiter.ReceiveWithIterator<INIT_CMD>(true, _mainPort, INIT_CMDHandler),
                    Arbiter.ReceiveWithIterator<CHVIEW_CMD>(true, _mainPort, CHVIEW_CMDHandler),
                    Arbiter.ReceiveWithIterator<WAYPOINT_CMD>(true, _mainPort, WAYPOINT_CMDHandler),
                    // Messages that are received from the AirRobot Controller
                    Arbiter.Receive<airRobotController.CONNECTED_MSG>(true, _clockNotify, NotifyCONNECTED_MSGHandler),
                    Arbiter.Receive<airRobotController.DISCONNECTED_MSG>(true, _clockNotify, NotifyDISCONNECTED_MSGHandler),
                    Arbiter.Receive<airRobotController.START_POSES_MSG>(true, _clockNotify, NotifySTART_POSES_MSGHandler),
                    Arbiter.Receive<airRobotController.INITIALIZED_MSG>(true, _clockNotify, NotifyINITIALIZED_MSGHandler),
                    Arbiter.ReceiveWithIterator<IMAGE_MSG>(true, _mainPort, NotifyIMAGE_MSGHandler),
                    Arbiter.ReceiveWithIterator<RTYPE_MSG>(true, _mainPort, RTYPE_MSGHandler),
                    Arbiter.ReceiveWithIterator<RSS_MSG>(true, _mainPort, RSS_MSGHandler),
                    Arbiter.ReceiveWithIterator<MAP_UPDATE_MSG>(true, _mainPort, MAP_UPDATE_MSGHandler)
                ),
                new ConcurrentReceiverGroup
                (
                    Arbiter.Receive<DsspDefaultLookup>(true,_mainPort,DefaultLookupHandler),
                    Arbiter.ReceiveWithIterator<Get>(true, _mainPort, GetHandler),
                    Arbiter.ReceiveWithIterator<HttpGet>(true, _mainPort, HttpGetHandler),
                    Arbiter.ReceiveWithIterator<Subscribe>(true, _mainPort, SubscribeHandler),
                    Arbiter.ReceiveWithIterator<game.Replace>(true, _gameControllerNotify, JoystickReplaceHandler),
                    Arbiter.ReceiveWithIterator<game.UpdateAxes>(true, _gameControllerNotify, JoystickUpdateAxesHandler),
                    Arbiter.ReceiveWithIterator<game.UpdateButtons>(true, _gameControllerNotify, JoystickUpdateButtonsHandler)
                )
            ));

            _clockPort.Subscribe(_clockNotify);

            DirectoryInsert();

            WinFormsServicePort.Post(new RunForm(CreateForm));
        }

        #endregion

        #region WinForms interaction

        System.Windows.Forms.Form CreateForm()
        {
            return new DriveControl(_eventsPort, _mainPort);
        }       

        #endregion

        #region Receiving Notifications (Messages) Handlers
        private void NotifyCONNECTED_MSGHandler(airRobotController.CONNECTED_MSG connectedMSG)
        {
            if (!Connected)
            {
                Connected = true;
                _driveControl.ProcessConnectedMSG();
            }
        }

        private void NotifyDISCONNECTED_MSGHandler(airRobotController.DISCONNECTED_MSG disconnectedMSG)
        {
            if (Connected)
            {
                Connected = false;
                _driveControl.ProcessDisconnect();
            }
        }

        private void NotifyINITIALIZED_MSGHandler(airRobotController.INITIALIZED_MSG initializedMSG)
        {
            if (Connected)
            {
                _driveControl.ProcessInitializedMSG();
            }
        }

        private void NotifySTART_POSES_MSGHandler(airRobotController.START_POSES_MSG startPosesMSG)
        {
            if (Connected)
            {
                _driveControl.ProcessStartPosesMSG(startPosesMSG.Body.startPoses);
            }
        }

        IEnumerator<ITask> NotifyIMAGE_MSGHandler(IMAGE_MSG imageMSG)
        {
            if (Connected)
            {
                if (imageMSG.Body.Valid)
                {
                    _driveControl.ProcessImageMSG(imageMSG.Body.ImgBuffer);
                }
            }

            yield break;
        }

        IEnumerator<ITask> RTYPE_MSGHandler(RTYPE_MSG rTypeMSG)
        {
            if (Connected)
            {
                _driveControl.ProcessRTypeMSG(rTypeMSG.Body.RobotNumber, rTypeMSG.Body.RobotType);
            }
            yield break;
        }

        IEnumerator<ITask> RSS_MSGHandler(RSS_MSG rssMSG)
        {
            if (Connected)
            {
                _driveControl.ProcessRssMSG(rssMSG.Body.RobotNumber, rssMSG.Body.SignalStrength, rssMSG.Body.BatteryLife);
            }
            yield break;
        }

        IEnumerator<ITask> MAP_UPDATE_MSGHandler(MAP_UPDATE_MSG mapMSG)
        {
            if (Connected)
            {
                //System.Console.WriteLine("MAP: " + mapMSG.Body.xLow + "," + mapMSG.Body.yLow + "," + mapMSG.Body.resolution + "," + mapMSG.Body.xPos + "," + mapMSG.Body.yPos + "," + mapMSG.Body.mapInfo.Count);
                //_driveControl.ProcessRssMSG(rssMSG.Body.RobotNumber, rssMSG.Body.SignalStrength);
                _driveControl.ProcessMapMSG(mapMSG.Body.robotNumber, mapMSG.Body.mapInfo, mapMSG.Body.xLow, mapMSG.Body.yLow, mapMSG.Body.resolution, mapMSG.Body.xPos, mapMSG.Body.yPos);
                //mapMSG.Body.
            }
            yield break;
        }
        #endregion

        #region Sending Notifications (Commands) Handlers
        IEnumerator<ITask> CONNECT_CMDHandler(CONNECT_CMD connectCMD)
        {
            if (!Connected)
            {
                //System.Console.WriteLine("GUI: Sending CONNECT_CMD...");
                base.SendNotification(_submgrPort, connectCMD);
                connectCMD.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                //System.Console.WriteLine("GUI: DONE Sending CONNECT_CMD...");
                yield break;
            }
        }
        
        IEnumerator<ITask> GET_POSES_CMDHandler(GET_POSES_CMD getPosesCMD)
        {
            if (Connected)
            {
                //System.Console.WriteLine("GUI: Sending GET_POSES_CMD...");
                base.SendNotification(_submgrPort, getPosesCMD);
                getPosesCMD.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            }
            yield break;
        }
        
        IEnumerator<ITask> INIT_CMDHandler(INIT_CMD initCMD)
        {
            if (Connected)
            {
                base.SendNotification(_submgrPort, initCMD);
                initCMD.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                yield break;
            }
        }

        IEnumerator<ITask> CHVIEW_CMDHandler(CHVIEW_CMD chViewCMD)
        {
            if (Connected)
            {
                base.SendNotification(_submgrPort, chViewCMD);
                chViewCMD.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                yield break;
            }
        }

        IEnumerator<ITask> WAYPOINT_CMDHandler(WAYPOINT_CMD waypointCMD)
        {
            if (Connected)
            {
                //System.Console.WriteLine("CREATING A NEW WAYPOINT COMMAND!!!");
                _state = waypointCMD.Body.GPS;
                //System.Console.WriteLine("GUI: Sending IMG_SVR_OK_CMD...");
                //base.SendNotification(_submgrPort, imgSvrOKCMD);
                //imgSvrOKCMD.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                yield break;
            }
        }
        #endregion

        #region DSS Handlers

        /// <summary>
        /// Get Handler returns RoboCup2009GUI State.
        /// </summary>
        /// <remarks>
        /// We declare this handler as an iterator so we can easily do
        /// sequential, logically blocking receives, without the need
        /// of nested Activate calls
        /// </remarks>
        /// <param name="get"></param>
        IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Http Get Handler
        /// </summary>
        IEnumerator<ITask> HttpGetHandler(HttpGet httpGet)
        {
            httpGet.ResponsePort.Post(new HttpResponseType(_state));
            yield break;
        }

        /// <summary>
        /// Subscription Handler
        /// </summary>
        IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            SubscribeRequestType request = subscribe.Body;
            System.Console.WriteLine("GUI: Subscribe request from: " + request.Subscriber);
            
            SubscribeHelper(_submgrPort, request, subscribe.ResponsePort);

            yield break;
        }

        /// <summary>
        /// Drop Handler shuts down RoboCup2009GUI
        /// </summary>
        /// <param name="drop"></param>
        void DropHandler(DsspDefaultDrop drop)
        {
            //PerformShutdown(ref _motorShutdown);
            if (_driveControl != null)
            {
                DriveControl drive = _driveControl;
                _driveControl = null;

                WinFormsServicePort.FormInvoke(
                    delegate()
                    {
                        if (!drive.IsDisposed)
                        {
                            drive.Dispose();
                        }
                    }
                );
            }

            base.DefaultDropHandler(drop);
        }
        #endregion

        #region Joystick Operations

        Choice EnumerateJoysticks()
        {
            return Arbiter.Choice(
                _gameControllerPort.GetControllers(new game.GetControllersRequest()),
                delegate(game.GetControllersResponse response)
                {
                    WinFormsServicePort.FormInvoke(
                        delegate()
                        {
                            _driveControl.ReplaceJoystickList(response.Controllers);
                        }
                    );
                },
                delegate(Fault fault)
                {
                    LogError(fault);
                }
            );
        }

        Choice SubscribeToJoystick()
        {
            return Arbiter.Choice(
                _gameControllerPort.Subscribe(_gameControllerNotify),
                delegate(SubscribeResponseType response) { },
                delegate(Fault fault) { LogError(fault); }
            );

        }

        IEnumerator<ITask> JoystickReplaceHandler(game.Replace replace)
        {
            if (_driveControl != null)
            {
                WinFormsServicePort.FormInvoke(
                    delegate()
                    {
                        _driveControl.UpdateJoystickButtons(replace.Body.Buttons);
                        _driveControl.UpdateJoystickAxes(replace.Body.Axes);
                    }
                );
            }

            yield break;
        }

        IEnumerator<ITask> JoystickUpdateAxesHandler(game.UpdateAxes update)
        {
            if (_driveControl != null)
            {
                WinFormsServicePort.FormInvoke(
                    delegate()
                    {
                        _driveControl.UpdateJoystickAxes(update.Body);
                    }
                );
            }
            yield break;
        }

        IEnumerator<ITask> JoystickUpdateButtonsHandler(game.UpdateButtons update)
        {
            if (_driveControl != null)
            {
                WinFormsServicePort.FormInvoke(
                    delegate()
                    {
                        _driveControl.UpdateJoystickButtons(update.Body);
                    }
                );
            }
            yield break;
        }

        IEnumerator<ITask> OnChangeJoystickHandler(OnChangeJoystick onChangeJoystick)
        {
            if (onChangeJoystick.DriveControl == _driveControl)
            {
                Activate(Arbiter.Choice(
                    _gameControllerPort.ChangeController(onChangeJoystick.Joystick),
                    delegate(DefaultUpdateResponseType response)
                    {
                        LogInfo("Changed Joystick");
                    },
                    delegate(Fault f)
                    {
                        LogError(null, "Unable to change Joystick", f);
                    })
                );
            }
            yield break;
        }

        #endregion

        #region Drive Control Event Handlers

        IEnumerator<ITask> OnLoadHandler(OnLoad onLoad)
        {
            _driveControl = onLoad.DriveControl;

            LogInfo("Loaded Form");

            yield return EnumerateJoysticks();

            yield return SubscribeToJoystick();
        }


        void OnClosedHandler(OnClosed onClosed)
        {
            if (onClosed.DriveControl == _driveControl)
            {
                _driveControl.CloseForm();
                //LogInfo("Form Closed");

                _mainPort.Post(new DsspDefaultDrop(DropRequestType.Instance));
            }
        }
        #endregion
    }

}
