//-----------------------------------------------------------------------
//  This file is part of the UC Merced Microsoft Robotics Developer Studio 
// Project Code Base.
//
//  Created by Benjamin Balaguer under the supervision of Stefano Carpin.
//
//  $File: WSS.cs $
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Drawing;
using System.Threading;
using System.Collections;
using System.Windows.Media.Imaging;
using Microsoft.Dss.Core.Attributes;

namespace Microsoft.Robotics.Services.RoboCup2009GUI
{
    #region RobotInformation data structure
    // To Talk to the Robots
    public class RobotInformation
    {
        string _name;
        int _port;
        string _ip;
        bool _connectionDisabled;
        TcpClient _wssClient;
        NetworkStream _wssStream;
        Thread _wssReadThread;
        public System.Object connectionDisabledLock = new System.Object();

        [DataMember]
        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }

        [DataMember]
        public int Port
        {
            get { return _port; }
            set { _port = value; }
        }

        [DataMember]
        public TcpClient wssClient
        {
            get { return _wssClient; }
            set { _wssClient = value; }
        }

        [DataMember]
        public NetworkStream wssStream
        {
            get { return _wssStream; }
            set { _wssStream = value; }
        }

        [DataMember]
        public Thread wssReadThread
        {
            get { return _wssReadThread; }
            set { _wssReadThread = value; }
        }

        [DataMember]
        public string IP
        {
            get { return _ip; }
            set { _ip = value; }
        }

        [DataMember]
        public bool ConnectionDisabled
        {
            get { return _connectionDisabled; }
            set { _connectionDisabled = value; }
        }
    }
    #endregion

    #region Class Structure for Robot Listen Threads
    class RobotConnector
    {
        private RoboCup2009GUIOperations _mainPort;
        private int RobotNumber;
        private NetworkStream stream;
        private List<RobotsInformation> RobotNames = new List<RobotsInformation>();
        private WSS wss;
        static private List<Dictionary<string, int>> relayInfo = new List<Dictionary<string, int>>();
        static private System.Object relayInfoLock = new System.Object();

        private float WSSCutOff;

        // Messages
        private IMAGE_MSG ImgMsg;
        private RTYPE_MSG rTypeMsg;
        private RSS_MSG rssMsg;
        private MAP_UPDATE_MSG mapMsg;

        public RobotConnector(RoboCup2009GUIOperations aPort, int rNum, List<RobotsInformation> rNames, NetworkStream aStream, WSS _wss, float wssCut)
        {
            _mainPort = aPort;
            RobotNumber = rNum;
            RobotNames = rNames;
            stream = aStream;
            wss = _wss;
            WSSCutOff = wssCut;

            ImgMsg = new IMAGE_MSG();
            rTypeMsg = new RTYPE_MSG();
            rssMsg = new RSS_MSG();
            mapMsg = new MAP_UPDATE_MSG();

            // Set up relay data structure
            for(int i=0; i<RobotNames.Count+1; i++)
            {
                Dictionary<string, int> aDict = new Dictionary<string, int>();
                lock (relayInfoLock)
                {
                    relayInfo.Insert(i, aDict);
                }
            }
        }

        public void ThreadHandler()
        {
            byte[] buffer;
            bool opcodeProcessed = false;
            char[] OpCode = new char[3];
            int msgRobotNumber = -1;

            //System.Console.WriteLine("Thread started for listening on robot " + RobotNumber);

            while (true)
            {
                // Read OpCode, unless it has already been processed by a RELAY message
                if (!opcodeProcessed)
                {
                    OpCode[0] = (char)stream.ReadByte();
                    //System.Console.Write(OpCode[0]);
                    OpCode[1] = (char)stream.ReadByte();
                    //System.Console.Write(OpCode[1]);
                    OpCode[2] = (char)stream.ReadByte();
                    //System.Console.Write(OpCode[2] + "\n");

                    msgRobotNumber = RobotNumber;
                }
                else
                {
                    //System.Console.WriteLine("OPCODE HAS ALREADY BEEN PROCESSED: " + OpCode[0] + OpCode[1] + OpCode[2]);
                    opcodeProcessed = false;
                }

                //System.Console.WriteLine(OpCode[0] + OpCode[1] + OpCode[2]);

                // Parse incoming data
                if (OpCode[0] == 'I' && OpCode[1] == 'M' && OpCode[2] == 'G') // IMG message
                {
                    //System.Console.WriteLine("Received an IMG message from Robot " + RobotNumber + ", relayed for Robot " + msgRobotNumber);

                    // Read the size
                    int imgSize = (int)(stream.ReadByte()) + (int)(stream.ReadByte()) * 256 + (int)(stream.ReadByte()) * 65536 + (int)(stream.ReadByte()) * 16777216;

                    // Read the image
                    buffer = new byte[imgSize];
                    int pos = 0;
                    while (pos < imgSize)
                    {
                        pos += stream.Read(buffer, pos, imgSize - pos);
                    }

                    // Set up message to be posted
                    ImgMsg.Body.ImgBuffer = buffer;
                    ImgMsg.Body.Valid = true;

                    // Post message
                    _mainPort.Post(ImgMsg);

                    // Free allocated memory
                    buffer = null;
                    //System.GC.Collect();
                }
                else if (OpCode[0] == 'R' && OpCode[1] == 'T' && OpCode[2] == 'P') // RTP message
                {
                    //System.Console.WriteLine("Received an RTP message from Robot " + RobotNumber + ", relayed for Robot " + msgRobotNumber);

                    // Get the robot type
                    char rType = (char)stream.ReadByte();
                    string RobotType;

                    if (rType == 'M')
                        RobotType = "Matilda";
                    else if (rType == 'A')
                        RobotType = "AirRobot";
                    else if (rType == 'P')
                        RobotType = "P2AT";
                    else if (rType == 'T')
                        RobotType = "Talon";
                    else
                        RobotType = "-";

                    // Set up message to be posted
                    rTypeMsg.Body.RobotNumber = msgRobotNumber;
                    rTypeMsg.Body.RobotType = RobotType;

                    // Post message
                    _mainPort.Post(rTypeMsg);
                }
                else if (OpCode[0] == 'R' && OpCode[1] == 'S' && OpCode[2] == 'S') // RSS message
                {
                    //System.Console.WriteLine("Received an RSS message from Robot " + RobotNumber + ", relayed for Robot " + msgRobotNumber);

                    float[] rssValues = new float[RobotNames.Count];
                    byte[] tmp = new byte[4]; // To store the floats

                    for (int i = 0; i < RobotNames.Count; i++)
                    {
                        tmp[0] = (byte)stream.ReadByte();
                        tmp[1] = (byte)stream.ReadByte();
                        tmp[2] = (byte)stream.ReadByte();
                        tmp[3] = (byte)stream.ReadByte();

                        rssValues[i] = BitConverter.ToSingle(tmp, 0);
                    }

                    // Read Battery Life
                    tmp[0] = (byte)stream.ReadByte();
                    tmp[1] = (byte)stream.ReadByte();
                    tmp[2] = (byte)stream.ReadByte();
                    tmp[3] = (byte)stream.ReadByte();          

                    // Set up message to be posted
                    rssMsg.Body.RobotNumber = msgRobotNumber;
                    rssMsg.Body.SignalStrength = rssValues;
                    rssMsg.Body.BatteryLife = BitConverter.ToInt32(tmp, 0);

                    // Post message
                    _mainPort.Post(rssMsg);

                    // Free allocated memory
                    rssValues = null;
                    tmp = null;
                    //System.GC.Collect();
                }
                else if (OpCode[0] == 'M' && OpCode[1] == 'A' && OpCode[2] == 'P') // MAP message
                {
                    //System.Console.WriteLine("Received an MAP message from Robot " + RobotNumber + ", relayed for Robot " + msgRobotNumber);

                    // Variable declaration
                    int numUpdates;
                    byte[] tmp4 = new byte[4]; // To store the floats or ints
                    byte[] tmp8 = new byte[8]; // To store the doubles
                    List<MOASTMapInfo> mapInfo = new List<MOASTMapInfo>();

                    mapMsg.Body.robotNumber = msgRobotNumber;

                    #region Read the xLow parameter
                    for(int i=0; i<8; i++)
                        tmp8[i] = (byte)stream.ReadByte();
                    mapMsg.Body.xLow = BitConverter.ToDouble(tmp8, 0);
                    #endregion

                    #region Read the yLow parameter
                    for (int i = 0; i < 8; i++)
                        tmp8[i] = (byte)stream.ReadByte();
                    mapMsg.Body.yLow = BitConverter.ToDouble(tmp8, 0);
                    #endregion

                    #region Read the resolution parameter
                    for (int i = 0; i < 8; i++)
                        tmp8[i] = (byte)stream.ReadByte();
                    mapMsg.Body.resolution = BitConverter.ToDouble(tmp8, 0);
                    #endregion

                    #region Read the xPos parameter
                    for (int i = 0; i < 4; i++)
                        tmp4[i] = (byte)stream.ReadByte();
                    mapMsg.Body.xPos = BitConverter.ToSingle(tmp4, 0);
                    #endregion

                    #region Read the yPos parameter
                    for (int i = 0; i < 4; i++)
                        tmp4[i] = (byte)stream.ReadByte();
                    mapMsg.Body.yPos = BitConverter.ToSingle(tmp4, 0);
                    #endregion

                    #region Read the numUpdates parameter
                    for (int i = 0; i < 4; i++)
                        tmp4[i] = (byte)stream.ReadByte();
                    numUpdates = BitConverter.ToInt32(tmp4, 0);
                    #endregion

                    #region Read Map Updates Data
                    for (int j = 0; j < numUpdates; j++)
                    {
                        MOASTMapInfo someInfo = new MOASTMapInfo();

                        #region Read the probability parameter
                        for (int i = 0; i < 4; i++)
                            tmp4[i] = (byte)stream.ReadByte();
                        someInfo.Prob = BitConverter.ToSingle(tmp4, 0);
                        #endregion

                        #region Read the xPos parameter
                        for (int i = 0; i < 8; i++)
                            tmp8[i] = (byte)stream.ReadByte();
                        someInfo.xPos = BitConverter.ToDouble(tmp8, 0);
                        #endregion

                        #region Read the xPos parameter
                        for (int i = 0; i < 8; i++)
                            tmp8[i] = (byte)stream.ReadByte();
                        someInfo.yPos = BitConverter.ToDouble(tmp8, 0);

                        mapInfo.Insert(j, someInfo);
                        #endregion
                    }
                    #endregion

                    mapMsg.Body.mapInfo = mapInfo;

                    // Post Message
                    _mainPort.Post(mapMsg);

                    // Free allocated memory
                    mapInfo = null;
                    tmp4 = null;
                    tmp8 = null;
                    //System.GC.Collect();
                }
                else if (OpCode[0] == 'R' && OpCode[1] == 'E' && OpCode[2] == 'L') // REL message
                {
                    #region Variable Declaration
                    char[] relayOpcodes = new char[RobotNames.Count + 1];
                    char[] relayMsgOpcode = new char[3];
                    string msgOpcode = "";
                    byte[] tmp = new byte[4];
                    int serial;
                    int payloadSize;
                    int goalRobot = -1;
                    int initialRobot = -1;
                    List<int> sentRobots = new List<int>();
                    List<int> unsentRobots = new List<int>();
                    #endregion

                    #region Read the relay Opcodes
                    for (int i = 0; i < RobotNames.Count + 1; i++)
                    {
                        relayOpcodes[i] = (char)stream.ReadByte();
                    }
                    #endregion

                    #region Read the serial parameter
                    for (int i = 0; i < 4; i++)
                        tmp[i] = (byte)stream.ReadByte();
                    serial = BitConverter.ToInt32(tmp, 0);
                    #endregion

                    #region Read the Payload Size parameter
                    for (int i = 0; i < 4; i++)
                        tmp[i] = (byte)stream.ReadByte();
                    payloadSize = BitConverter.ToInt32(tmp, 0);
                    #endregion

                    #region Read the relay message's Opcode
                    for (int i = 0; i < 3; i++)
                    {
                        relayMsgOpcode[i] = (char)stream.ReadByte();
                        msgOpcode += relayMsgOpcode[i];
                    }
                    #endregion

                    #region Find the robot that should get the message
                    for (int i = 0; i < RobotNames.Count + 1; i++)
                    {
                        if (relayOpcodes[i] == 'I')
                            initialRobot = i;
                        else if (relayOpcodes[i] == 'G')
                            goalRobot = i;
                        else if (relayOpcodes[i] == 'S')
                            sentRobots.Add(i);
                        else
                            unsentRobots.Add(i);
                    }
                    #endregion

                    //System.Console.WriteLine("Received an REL message from Robot " + RobotNumber + ", relayed for Robot " + initialRobot + " with msg: " + msgOpcode);


                    #region Check for the serial
                    bool serialIsGood = false;
                    lock (relayInfoLock)
                    {
                        if (relayInfo[initialRobot].ContainsKey(msgOpcode))
                        {
                            // If we have already processed this message
                            if (serial <= relayInfo[initialRobot][msgOpcode])
                            {
                                //System.Console.WriteLine("Droping REL message, " + msgOpcode + ", with serial: " + serial);
                                // Drop Message
                                for (int i = 0; i < payloadSize-3; i++)
                                    stream.ReadByte();
                            }
                            else
                            {
                                //System.Console.WriteLine("Keeping REL message, " + msgOpcode + " with serial: " + serial);
                                relayInfo[initialRobot][msgOpcode] = serial;
                                serialIsGood = true;
                            }
                        }
                        else
                        {
                            //System.Console.WriteLine("Keeping REL message, " + msgOpcode + " with serial: " + serial);
                            relayInfo[initialRobot].Add(msgOpcode, serial);
                            serialIsGood = true;
                        }
                    }
                    #endregion

                    // If the ComStation is the goal robot
                    if (goalRobot == 0)
                    {
                        if (serialIsGood)
                        {
                            OpCode[0] = relayMsgOpcode[0];
                            OpCode[1] = relayMsgOpcode[1];
                            OpCode[2] = relayMsgOpcode[2];

                            msgRobotNumber = initialRobot;
                            opcodeProcessed = true;
                        }
                    }
                    else
                    {
                        if (serialIsGood)
                        {
                            // Set up buffer
                            byte[] sendBuf = new byte[3 + RobotNames.Count + 1 + 4 + 4 + payloadSize];
                            sendBuf[0] = (byte)'R';
                            sendBuf[1] = (byte)'E';
                            sendBuf[2] = (byte)'L';

                            // Set the buffer for the initial, goal, sent, and unsent robots
                            sendBuf[3 + initialRobot] = (byte)'I';
                            sendBuf[3 + goalRobot] = (byte)'G';
                            for (int i = 0; i < sentRobots.Count; i++)
                                sendBuf[3 + sentRobots[i]] = (byte)'S';
                            for (int i = 0; i < unsentRobots.Count; i++)
                                sendBuf[3 + unsentRobots[i]] = (byte)'N';

                            // Set the buffer for the serial and payload size
                            tmp = BitConverter.GetBytes(serial);
                            Array.Copy(tmp, 0, sendBuf, 3 + RobotNames.Count + 1, 4);
                            tmp = BitConverter.GetBytes(payloadSize);
                            Array.Copy(tmp, 0, sendBuf, 3 + RobotNames.Count + 1 + 4, 4);

                            // Read the rest of the message and put it in the buffer
                            int pos = 0;
                            while (pos < payloadSize)
                            {
                                pos += stream.Read(sendBuf, pos + 3 + RobotNames.Count + 1 + 4 + 4, payloadSize - pos);
                            }

                            float ss = 0;
                            lock (wss.getSSLock)
                            {
                                ss = (wss.GetSignalStrength(RobotNames[goalRobot - 1].Name));
                            }

                            bool isDropped;
                            lock (wss.droppedRobotsLock)
                            {
                                isDropped = wss.DroppedRobots.Contains(RobotNumber);
                            }

                            // If the goal robot is in range
                            if (ss > (WSSCutOff * 0.95) && !isDropped) // CHANGED: goalRobot to GoalRobot-1
                            {
                                wss.WriteLine(goalRobot, sendBuf);
                            }
                            // If it has not been sent to at least one robot
                            else if (unsentRobots.Count >= 1)
                            {
                                List<int> sendRelayToRobot = new List<int>();
                                for (int i = 0; i < unsentRobots.Count; i++)
                                {
                                    lock (wss.droppedRobotsLock)
                                    {
                                        isDropped = wss.DroppedRobots.Contains(unsentRobots[i]);
                                    }

                                    lock (wss.getSSLock)
                                    {
                                        if ((wss.GetSignalStrength(RobotNames[unsentRobots[i] - 1].Name)) > (WSSCutOff * 0.95) && !isDropped) // CHANGED: unsentRobots[i] to unsentRobots[i]-1
                                        {
                                            sendBuf[3 + unsentRobots[i]] = (byte)'S';
                                            sendRelayToRobot.Add(unsentRobots[i]);
                                        }
                                    }
                                }

                                for (int i = 0; i < sendRelayToRobot.Count; i++)
                                {
                                    wss.WriteLine(sendRelayToRobot[i], sendBuf);
                                }
                            }
                        }
                     }
                }
                else
                {
                    System.Console.WriteLine("Last OpCode: " + OpCode[0] + OpCode[1] + OpCode[2]);
                    break;
                }
            }
        }
    }

    #endregion

    #region WSS Communication Interface
    class WSS
    {
        #region Variables
        // To Talk to the WSS
        private static TcpClient wssClient;
        private static NetworkStream wssStream;
        private List<RobotInformation> RobotInfo = new List<RobotInformation>();
        private List<RobotsInformation> RobotNames = new List<RobotsInformation>();
        public List<int> DroppedRobots = new List<int>();

        // Locking Objects
        public System.Object getSSLock = new System.Object();
        public System.Object droppedRobotsLock = new System.Object();


        // To Listen to the Robots
        private int listenPort;
        private string listenIP;

        public RoboCup2009GUIOperations _mainPort;

        private float WSSCutOff;
        private Dictionary<string, int> msgSerial = new Dictionary<string, int>();

        #endregion

        #region Constructor
        public WSS(RoboCup2009GUIOperations mainPort, List<RobotsInformation> rNames, string aIP, int aPort, float wssCut) // Constructor
        {
            _mainPort = mainPort;
            RobotNames = rNames;
            WSSCutOff = wssCut;

            for (int i = 0; i < rNames.Count; i++)
            {
                RobotInformation someInfo = new RobotInformation();
                someInfo.Name = rNames[i].Name;
                someInfo.Port = -1;
                someInfo.wssClient = new TcpClient();
                RobotInfo.Insert(i, someInfo);
            }
            listenIP = aIP;
            listenPort = aPort;
        }
        #endregion

        #region Connect/Disconnect
        public void Connect(string IP, int Port)
        {                        
            // Connect to the WSS
            wssClient = new TcpClient();
            wssClient.Connect(new IPEndPoint(IPAddress.Parse(IP), Port));
            wssStream = wssClient.GetStream();
            
            // Initialize ComStation
            if (!SendInit("CommStation", listenPort))
            {
                System.Console.WriteLine("GUI: WSS Init Command DID NOT work");
            }

            for (int i = 0; i < RobotInfo.Count; i++)
            {
                RobotInfo[i].IP = IP;
                RobotInfo[i].ConnectionDisabled = false;

                if ((RobotInfo[i].Port = GetRobotPort(RobotInfo[i].Name)) == -1)
                    System.Console.WriteLine("Error Getting Robot WSS Port");

                float ss = 0;
                lock (getSSLock)
                {
                    ss = (GetSignalStrength(RobotNames[i].Name));
                }

                bool isDropped;
                // If the connection has been droped
                if (ss <= (WSSCutOff * 0.99))
                {
                    lock (droppedRobotsLock)
                    {
                        isDropped = DroppedRobots.Contains(i);
                    }

                    if (!isDropped)
                    {
                        System.Console.WriteLine("Dropping Robot #" + (i+1));

                        lock (droppedRobotsLock)
                        {
                            DroppedRobots.Add((i+1));
                        }
                    }
                }

                // Connect to the robot
                RobotInfo[i].wssClient.Connect(new IPEndPoint(IPAddress.Parse(IP), RobotInfo[i].Port));
                RobotInfo[i].wssStream = RobotInfo[i].wssClient.GetStream();

                // Start the read thread
                RobotConnector aConnector = new RobotConnector(_mainPort, i + 1, RobotNames, RobotInfo[i].wssStream, this, WSSCutOff);
                RobotInfo[i].wssReadThread = new Thread(new ThreadStart(aConnector.ThreadHandler));
                RobotInfo[i].wssReadThread.Start();
            }
        }

        public void Disconnect()
        {
            // Close all the robot connections
            for (int i = 0; i < RobotInfo.Count; i++)
            {
                RobotInfo[i].wssReadThread.Abort();
                RobotInfo[i].wssStream.Close();
                RobotInfo[i].wssClient.Close();
            }

            // Close the WSS connection
            wssStream.Close();
            wssClient.Close();
        }
        #endregion

        #region Read/Write to WSS
        public string ReadLine(int Type)
        {
            char currentChar;
            bool done = false;
            string result = "";
            NetworkStream stream;

            // If we are reading a line directly from WSS (e.g. a response)
            if (Type == -1)
            {
                stream = wssStream;

                for (int i = 0; !done; i++)
                {
                    currentChar = (char)stream.ReadByte();

                    // If we have found a carriage return
                    if ((int)currentChar == 13)
                    {
                        stream.ReadByte(); // Discard the next character (line feed)
                        done = true;       // Set the condition to drop out of the loop
                    }
                    else
                    {
                        // Add the character to the resulting string
                        result = result + currentChar;
                    }
                }
            }

            // Deallocate memory
            stream = null;
            //System.GC.Collect(); 

            return result;
        }

        public void WriteLine(int RobotNumber, byte[] buffer)
        {  
            NetworkStream stream;

            if (RobotNumber == -1) // If we are trying to send a message to the WSS
            {
                stream = wssStream;

                try
                {
                    stream.Write(buffer, 0, buffer.Length);
                    stream.Flush();
                }
                catch (Exception)
                {
                    //System.Console.WriteLine("Caught an Exception while Writing: " + e.ToString());
                }
            }
            else
            {
                float ss = 0;
                lock (getSSLock)
                {
                    ss = (GetSignalStrength(RobotNames[RobotNumber - 1].Name)); 
                }

                bool isDropped;
                // If the connection has been droped
                if (ss <= (WSSCutOff * 0.99))
                {
                    lock (droppedRobotsLock)
                    {
                        isDropped = DroppedRobots.Contains(RobotNumber);
                    }

                    if (!isDropped)
                    {
                        System.Console.WriteLine("Dropping Robot #" + RobotNumber);

                        lock (droppedRobotsLock)
                        {
                            DroppedRobots.Add(RobotNumber);
                        }
                    }
                }

                lock (droppedRobotsLock)
                {
                    isDropped = DroppedRobots.Contains(RobotNumber);
                }
                // Check to make sure the robot is in range
                if (ss > (WSSCutOff * 0.95) && !isDropped)
                {
                    stream = RobotInfo[RobotNumber - 1].wssStream;
                    try
                    {
                        stream.Write(buffer, 0, buffer.Length);
                        stream.Flush();
                    }
                    catch (Exception)
                    {
                        //System.Console.WriteLine("Caught an Exception while Writing: " + e.ToString());
                    }
                }
                else // If the robot is not in range, we spam a relay message to all the robots in range
                {
                    // Set up buffer
                    byte[] tmp = new byte[4];
                    byte[] sendBuf = new byte[3 + RobotNames.Count + 1 + 4 + 4 + buffer.Length];
                    sendBuf[0] = (byte)'R';
                    sendBuf[1] = (byte)'E';
                    sendBuf[2] = (byte)'L';

                    // Set the buffer for the initial, goal, sent, and unsent robots
                    for(int i=0; i<RobotNames.Count+1;i++)
                        sendBuf[3+i] = (byte)'N'; // Set all robots as 'not sent'
                    sendBuf[3] = (byte)'I'; // We are the initial robot
                    sendBuf[3 + RobotNumber] = (byte)'G'; // Set the goal robot

                    // Find serial number, or create a new one
                    string msgOpcode = "" + (char)(buffer[0]) + (char)(buffer[1]) + (char)(buffer[2]);
                    if (!msgSerial.ContainsKey(msgOpcode))
                        msgSerial.Add(msgOpcode, 1);
                    else
                        msgSerial[msgOpcode] = msgSerial[msgOpcode] + 1;
                    int serial = msgSerial[msgOpcode];

                    // Set the buffer for the serial and payload size
                    tmp = BitConverter.GetBytes(serial);
                    Array.Copy(tmp, 0, sendBuf, 3 + RobotNames.Count + 1, 4);
                    tmp = BitConverter.GetBytes(buffer.Length);
                    Array.Copy(tmp, 0, sendBuf, 3 + RobotNames.Count + 1 + 4, 4);

                    // Read the rest of the message and put it in the buffer
                    Array.Copy(buffer, 0, sendBuf, 3 + RobotNames.Count + 1 + 4 + 4, buffer.Length);

                    // Get all the robots that are in range
                    List<int> robotsToSend = new List<int>();
                    for(int i=0; i<RobotNames.Count; i++)
                    {
                        if (i + 1 != RobotNumber)
                        {
                            lock (droppedRobotsLock)
                            {
                                isDropped = DroppedRobots.Contains(i+1);
                            }

                            lock (getSSLock)
                            {
                                if ((GetSignalStrength(RobotNames[i].Name)) > (WSSCutOff * 0.95) && !isDropped)
                                {
                                    robotsToSend.Add(i + 1);
                                    sendBuf[3 + i + 1] = (byte)'S';
                                }
                            }
                        }
                    }

                    // Send relay messages to all the robots
                    for (int i = 0; i < robotsToSend.Count; i++)
                    {
                        //System.Console.Write("Sending " + (char)buffer[0] + (char)buffer[1] + (char)buffer[2] + " to Robot " + robotsToSend[i] + ": " + (char)sendBuf[0] + (char)sendBuf[1] + (char)sendBuf[2] + " " + (char)sendBuf[3] + (char)sendBuf[4] + (char)sendBuf[5] + " ");
                        //Array.Copy(sendBuf, 3 + RobotNames.Count + 1, tmp, 0, 4);
                        //System.Console.Write(BitConverter.ToInt32(tmp, 0) + " ");
                        //Array.Copy(sendBuf, 3 + RobotNames.Count + 1 + 4, tmp, 0, 4);
                        //System.Console.Write(BitConverter.ToInt32(tmp, 0) + " " + (char)sendBuf[3 + RobotNames.Count + 1 + 4 + 4] + (char)sendBuf[3 + RobotNames.Count + 1 + 4 + 5] + (char)sendBuf[3 + RobotNames.Count + 1 + 4 + 6] + "\n");
                        
                        
                        stream = RobotInfo[robotsToSend[i] - 1].wssStream;
                        try
                        {
                            stream.Write(sendBuf, 0, sendBuf.Length);
                            stream.Flush();
                        }
                        catch (Exception)
                        {
                            //System.Console.WriteLine("Caught an Exception while Writing: " + e.ToString());
                        }
                    }

                    // Deallocate memory
                    sendBuf = null;
                    tmp = null; 
                }
            }

            // Deallocate memory
            buffer = null;
            stream = null;    
        }
        #endregion

        #region WSS CMDs Functions
        public bool SendInit(string RobotName, int Port)
        {
            bool result = false;
            string cmd = "INIT {Robot " + RobotName + "} {Port " + Convert.ToString(Port) + "}" + (char)13 + (char)10;
            ASCIIEncoding encoder = new ASCIIEncoding();
            byte[] buffer = encoder.GetBytes(cmd);
            
            WriteLine(-1, buffer);

            if (ReadLine(-1).CompareTo(INITResponse()) == 0)
                result = true;

            // Deallocate memory
            encoder = null;
            buffer = null;
            //System.GC.Collect(); 

            return result;
        }

        public float GetSignalStrength(string RobotName)
        {
            string cmd = "GETSS {Robot " + RobotName + "}" + (char)13 + (char)10;
            ASCIIEncoding encoder = new ASCIIEncoding();
            byte[] buffer = encoder.GetBytes(cmd);

            WriteLine(-1, buffer);

            // Deallocate memory
            encoder = null;
            buffer = null;
            //System.GC.Collect();

            return SignalStrengthResponse(ReadLine(-1));
        }

        public int GetRobotPort(string RobotName)
        {
            string cmd = "DNS {Robot " + RobotName + "}" + (char)13 + (char)10;
            ASCIIEncoding encoder = new ASCIIEncoding();
            byte[] buffer = encoder.GetBytes(cmd);

            WriteLine(-1, buffer);

            // Deallocate memory
            encoder = null;
            buffer = null;
            //System.GC.Collect();

            return DNSResponse(ReadLine(-1));
        }
        #endregion

        #region WSS Responses
        public string INITResponse()
        {
            return "INITREPLY {Status OK}";
        }

        public float SignalStrengthResponse(string msg)
        {
            try
            {
                string tmp = msg.Substring(msg.IndexOf("{Strength ") + 10, msg.IndexOf("}", msg.IndexOf("{Strength ") + 10) - msg.IndexOf("{Strength ") - 10);

                if (tmp.CompareTo("NaN") == 0)
                    return 100f;
                else if (tmp.CompareTo("INF") == 0)
                    return 101f;
                else
                    return Convert.ToSingle(tmp);
            }
            catch (Exception)
            {
                return 102f;
            }
        }

        public int DNSResponse(string msg)
        {
            string tmp = msg.Substring(msg.IndexOf("{Port ") + 6, msg.Length - 1 - msg.IndexOf("{Port ") - 6);

            return Convert.ToInt32(tmp);
        }
        #endregion
    }
    #endregion
}
