//-----------------------------------------------------------------------
//  This file is part of the UC Merced Microsoft Robotics Developer Studio 
// Project Code Base.
//
//  Created by Benjamin Balaguer under the supervision of Stefano Carpin.
//  Based on SimpleDashboard.cs from UX\SimpleDashboard.
//
//  $File: GroupingGUI.cs $
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using W3C.Soap;
using Microsoft.Ccr.Core;
using Microsoft.Ccr.Adapters.WinForms;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using Microsoft.Dss.Services.Serializer;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using cs = Microsoft.Dss.Services.Constructor;
using ds = Microsoft.Dss.Services.Directory;
using System.ComponentModel;
using System.Threading;
using Microsoft.Dss.Core.DsspHttp;
using System.Drawing;

namespace Microsoft.Robotics.Services.GroupingGUI
{

    /// <summary>
    /// Grouping GUI Service
    /// </summary>
    [DisplayName("grouping GUI")]
    [Description("Provides access to a simple Windows UI for adding groups to a map.")]
    [Contract(Contract.Identifier)]
    class GroupingGUIService : DsspServiceBase
    {
        // shared access to state is protected by the interleave pattern
        // when we activate the handlers
        GroupingGUIState _state = new GroupingGUIState();

        [ServicePort("/groupinggui", AllowMultipleInstances = false)]
        GroupingGUIOperations _mainPort = new GroupingGUIOperations();

        DriveControl _driveControl;
        DriveControlEvents _eventsPort = new DriveControlEvents();

        #region Startup
        /// <summary>
        /// GroupingGUIService Default DSS Constuctor
        /// </summary>
        /// <param name="pCreate"></param>
        public GroupingGUIService(DsspServiceCreationPort pCreate)
            : base(pCreate)
        {
        }

        /// <summary>
        /// Entry Point for the GroupingGUI Service
        /// </summary>
        protected override void Start()
        {
            Activate(Arbiter.Interleave(
                new TeardownReceiverGroup
                (
                    Arbiter.Receive<DsspDefaultDrop>(false, _mainPort, DropHandler)
                ),
                new ExclusiveReceiverGroup
                (
                    Arbiter.ReceiveWithIterator<OnLoad>(true, _eventsPort, OnLoadHandler),
                    Arbiter.Receive<OnClosed>(true, _eventsPort, OnClosedHandler)
                ),
                new ConcurrentReceiverGroup
                (
                    Arbiter.Receive<DsspDefaultLookup>(true,_mainPort,DefaultLookupHandler),
                    Arbiter.ReceiveWithIterator<Get>(true, _mainPort, GetHandler),
                    Arbiter.ReceiveWithIterator<HttpGet>(true, _mainPort, HttpGetHandler)
                )
            ));

            WinFormsServicePort.Post(new RunForm(CreateForm));
        }

        #endregion

        #region WinForms interaction

        System.Windows.Forms.Form CreateForm()
        {
            return new DriveControl(_eventsPort, _mainPort);
        }       

        #endregion

        #region DSS Handlers
        /// <summary>
        /// Get Handler returns GroupingGUI State.
        /// </summary>
        /// <remarks>
        /// We declare this handler as an iterator so we can easily do
        /// sequential, logically blocking receives, without the need
        /// of nested Activate calls
        /// </remarks>
        /// <param name="get"></param>
        IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Http Get Handler
        /// </summary>
        IEnumerator<ITask> HttpGetHandler(HttpGet httpGet)
        {
            httpGet.ResponsePort.Post(new HttpResponseType(_state));
            yield break;
        }

        /// <summary>
        /// Drop Handler shuts down GroupingGUI
        /// </summary>
        /// <param name="drop"></param>
        void DropHandler(DsspDefaultDrop drop)
        {
            //PerformShutdown(ref _motorShutdown);
            if (_driveControl != null)
            {
                DriveControl drive = _driveControl;
                _driveControl = null;

                WinFormsServicePort.FormInvoke(
                    delegate()
                    {
                        if (!drive.IsDisposed)
                        {
                            drive.Dispose();
                        }
                    }
                );
            }

            base.DefaultDropHandler(drop);
        }
        #endregion

        #region Drive Control Event Handlers

        IEnumerator<ITask> OnLoadHandler(OnLoad onLoad)
        {
            _driveControl = onLoad.DriveControl;
            yield break;
        }


        void OnClosedHandler(OnClosed onClosed)
        {
            if (onClosed.DriveControl == _driveControl)
            {
                _driveControl.CloseForm();

                _mainPort.Post(new DsspDefaultDrop(DropRequestType.Instance));
            }
        }
        #endregion
    }

}
