/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file aStarHeap.hh
    \brief defines classes used by aStarHeap graph searcher with multiple goals

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.1 $
  $Date: 2006/11/16 14:01:19 $
  \endcode

  \author Tom Kramer

This is used by aStarHeap.cc.

The aStarHeap.hh and aStarHeap.cc files define an abstract graph
class (AStarHeapGraph) and an A-star graph search class
(AStarHeapSearcher) with the following characteristics:

1. The graph may be built or changed dynamically during the search
   in several ways using the addKids, changeGoalCosts, removeKids,
   and/or useKids functions.
2. The user can add information to graph nodes by making a
   child class of nodeInfo.
3. The open nodes list is managed using the heap method.
4. The searcher will search simultaneously for multiple (N) goals.
5. Arcs with a negative cost between nodes are allowed and treated
   as though there were no arc. This is so that graphs can be handled in
   which every node has the same apparent number of children, but some
   are not actually children. Such graphs are often convenient to use.

The start must not be one of the goals. It is up to the caller to
check this.

Two different drivers (exampleAStarHeapA.cc and exampleAStarHeapB.cc)
for making a stand-alone executable are in src/tools.  Test files for
testing the executables exist but are not currently part of MOAST.

*/

#ifndef A_STAR_HEAP_HH
#define A_STAR_HEAP_HH

#include "nodeInfo.hh"

class AStarHeapSearcher;
class AStarHeapGraph;
class AStarHeapNode;

/***********************************************************************/

/*! \class AStarHeapSearcher
    \brief provides a searcher class used in aStar search with 1 goal and heap

The AStarHeapSearcher uses a heap of nodes as the list of open
nodes. The cheapest node is always heap[1]. The heap is not totally
ordered.  As a result the heap is relatively cheap to maintain.

cost[0] is set by the Searcher to the cost of reaching the goal.
status[0] is set to zero when the goal has not been reached, to 1 when
the goal has been reached but not optimally, and to 2 when the goal has
been reached optimally.

The heuristic of Euclidean distance to the goal times cheapest cost
per unit distance is a reasonable heuristic to use. Any admissible
heuristic will work.

*/

class AStarHeapSearcher
{
public:
  AStarHeapSearcher(){ heapSize = 0; }
  ~AStarHeapSearcher(){}

  int closeSearch(); /*! closes the searcher */

  int cycleSearch(); /*! pops top node from the heap, processes it */

  int getNextNode(); /*! returns the index of cheapest open node remaining */

  int initSearch(    /*! initializes the searcher */
   AStarHeapGraph * aGraph, /*!< the graph to search in                     */
   int start,               /*!< index in nodes of aGraph of start node     */
   int goalIn,              /*!< index in nodes of aGraph of goal node      */
   int * opensOut,          /*!< 0=open list empty, non-zero=list not empty */
   int * statusOut,         /*!< status of reaching goal                    */
   double * costOut);       /*!< cost of reaching goal                      */

  int reportSearch( /*! writes the path to a node into an array           */
   int maxLength,   /*!< maximum length of path (capacity of path array)  */
   int * path);     /*!< array of index numbers giving the path, set here */

private:
  double * cost;          /* array (size 1) for cost of reaching goal        */
  int goal;               /* indexes in nodes of goal node                   */
  int * heap;             /* heap of node indexes, cost ordered              */
  int heapCapacity;       /* capacity of heap (i.e. total array size)        */
  int heapSize;           /* size of heap (i.e. number of array places used) */
  int * heapSizeOut;      /* size of heap for caller                         */
  AStarHeapNode * nodes;  /* array of nodes from theGraph                    */
  int * status;           /* array (size 1) for status flag                  */
  AStarHeapGraph * theGraph; /* the graph to be searched                     */

  void heapDown();
  void heapUp(int index);
  void increaseHeap(int newCapacity);
  void processNode(int numNode, int from, double cost);
  void removeCheap();
};

/***********************************************************************/

/*! \class AStarHeapNode
    \brief provides a node class used in aStar search with heap

See aStarHeap.cc for descriptions of what the functions do.

The destructor (~AStarHeapNode) does not delete kidsIndexes or kidsCosts.
This is so that copying a node can be done faster; kidsIndexes and
kidsCosts are reused by the copy.  If it is desired to delete
kidsIndexes or kidsCosts in order to free memory, that must be done
explicitly.

Note that everything required of a AStarHeapNode (other than NodeInfo)
is here. There is no graphNode base class -- not worth the trouble. 

*/

class AStarHeapNode
{
public:
  enum status {UNOPENED, OPEN, CLOSED};
  AStarHeapNode() { kidsSize= 0; open = UNOPENED; }
  ~AStarHeapNode();

  int copyFrom            /*!< returns 0 if from is 0, otherwise returns 1 */
  (AStarHeapNode * from); /*!< node to copy from                           */

  NodeInfo * moreInfo; //!< additional information about node
  int kidsSize;        //!< number of kids of node
  int * kidsIndexes;   //!< array of kid indexes of node, built at read time
  double * kidsCosts;  //!< array of kid costs of node, built at read time
  status open;         //!< flag indicating UNOPENED, OPEN, CLOSED
  int arrivedFrom;     //!< index in nodes of node from which arrived
  double visitCost;    //!< current total cost to visit node
  double goalCost;     //!< fixed lower bound on actual cost of reaching goal
  double totalCost;    //!< visitCost plus goalCost
  int heapIndex;       //!< index in heap of node
};

/***********************************************************************/

/* \class AStarHeapGraph
   \brief provides a graph class used in aStar search with heap

Four public methods are provided to allow an application to change the
graph being searched during the search.

In addition, the application may change the graph during the search
without using any of these methods.

In the following discussions, keep in mind that an AStarHeapGraph has
both a "capacity", which is the number of nodes currently available,
and a "size", which is the number of available nodes currently being
used.

************************
THE addKids FUNCTION
************************

AddKids may activate additional nodes of the graph. If it does, it also
adds arcs from dad to those new nodes. This may also add arcs from dad
to existing nodes.  The memory for dad's old kidsIndexes and kidsCosts
is freed. Memory for dad's revised kidsIndexes and kidsCosts is
allocated.

If there are not enough nodes in the graph to accommodate the number
to be activated, addKids calls increaseCapacity to make more nodes.

The new kids are built in nodes whose indexes are assigned in
numerical order, starting with the current size of the graph. The size
is increased as the new kids are added. The index in the graph should
already be recorded as the value of "index" in the nodeInfo. It is
checked that this number is correct.

The additional kids in oldAdded must not already be kids of dad.
AddKids is not currently checking for that. A check could be added.

********************************
THE changeKidsGoalCosts FUNCTION
********************************

ChangeKidsGoalCosts changes the goalCosts of some or all the kids of dad.

The numberChanged argument is the number of (existing) kids
of dad that are to be assigned new goal costs (corresponding to kids
indexes 0 through (numberNew - 1)). The new goal costs are given
in the newGoalCosts array.

The numberChanged should not be larger than the number of kids of dad.

********************
THE useKids FUNCTION
********************

The useKids function removes the current kids of dad (if any) and
replaces them with the ones given by the arguments. The arrays
specified in the arguments are used as is, so they must be created
with "new" in the implementing system and not changed or deleted by
the implementing system.

The kids themselves must already exist and are not affected by useKids.

***********************
THE removeKids FUNCTION
***********************

The removeKids function removes all the kids of the dad node.

**********************
OTHER MEMBER FUNCTIONS
**********************

The increaseCapacity function makes a new and larger array of nodes
and copies the old nodes into it. Since all the references in the
Searcher to nodes in the graph are done by index number, not by
pointer, the Searcher does not notice the switch.

The destructor deletes the contents of the nodes and the nodes.


The graph may be built using nodes created with "new" or with nodes
created some other way (e.g. a global array). The "deleat" data member
should always be set to false if the nodes were not created with
"new", and may be set to false if the nodes were created with "new"
but they should survive the destruction of the graph. The default
value of deleat is true.

*/

class AStarHeapGraph
{
public:
  AStarHeapGraph(){deleat = true;}
  ~AStarHeapGraph();

  int addKids             /*! adds kids to dad and builds new nodes          */
  (int dad,               /*!< index in nodes of node for which to add kids  */
   int numberAdded,       /*!< total number of kids added                    */
   int numberNew,         /*!< number of new kids added                      */
   NodeInfo ** newInfos,  /*!< array of pointers to NodeInfos for new kids   */
   double * newArcCosts,  /*!< array of arc costs from dad for new kids      */
   double * newGoalCosts, /*!< array of goal costs for new kids              */
   int * oldAdded,        /*!< number of existing kids added                 */
   double * oldArcCosts); /*!< array of arc costs from dad for existing kids */

  int changeKidsGoalCosts  /*! changes goalCosts of kids of dad           */
  (int dad,                /*!< index in nodes of node to change kids for */
   int numberChanged,      /*!< number of kids to change                  */
   double * newGoalCosts); /*!< array of new goalCosts of kids            */
  
  int removeKids  /*! removes kidIndexes & kidCosts from dad, not kids   */
  (int dad);      /*!< index in nodes of node from which to remove kids  */

  int useKids               /*! specifies new kids of dad                  */
  (int dad,                 /*!< index in nodes of node to change kids for */
   int kidsSize,            /*!< new number of kids                        */
   int * kidsIndexes,       /*!< new array of kids indexes in nodes        */
   double * kidsCosts);     /*!< new array of kids costs                   */

  int capacity;            //!< total number of nodes in nodes array
  int size;                //!< number of active nodes in nodes array
  AStarHeapNode * nodes;   //!< array of nodes
  bool deleat;             //!< delete nodes in destructor? default=true

private:

  int increaseCapacity  /*! makes larger array, copies old nodes into it   */
  (int numberAdded);    /*!< minimum number by which to increase capacity  */

};

/***********************************************************************/

#endif // A_STAR_HEAP_HH
