/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file   acmRandom.h

  \brief  A 32-bit random number generator that implements the minimal
  standard that ports to all systems where the maximum integer size is
  2^32 -1 or larger.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.2 $
  $Date: 2007/02/01 18:20:21 $
  \endcode

  \ref acmRandom "Details..."

  \author Fred Proctor
*/

#ifndef ACM_RANDOM_H
#define ACM_RANDOM_H

#ifdef __cplusplus
extern "C" {
#if 0
} /* matches brace above, to un-confuse autoformatting text editors */
#endif /* if 0 */
#endif /* ifdef __cplusplus */

/*!
  AcmRandomStruct holds the state information of the random number
  generator.
 */
typedef struct {
  int which;
  long int A;
  long int Q;
  long int R;
  long int seed;
} AcmRandomStruct;

/*!
  Call acmRandomInit() before using an of the acm* functions. 
  This sets the seed to a reasonable initial value but you will want
  to call acmRandomSeed() next with your seed if you don't want
  the same series each time.
*/
extern int acmRandomInit(AcmRandomStruct * acm);

/*!
  Seeds the random number generator with your number 's', which can
  be obtained from the environment, e.g., time of day, so that each
  time you run your program the series will be different.
*/
extern int acmRandomSeed(AcmRandomStruct * acm, long int s);

/*!
  There are three possible multipliers, the original:

  A = 16807, Q = 127773, R = 2836

  and two that reportedly do better:

  A = 48271, Q = 44488, R = 3399
  A = 69621, Q = 30845, R = 23902

  Use this function to pick which one of the three you want. You would
  want to have two in order to build non-uniform variates, such as a
  normal distribution, that need two independent random number
  streams.

  The value of \a which will be normalized to 0, 1 or 2, so set
  which to be 0, 1 or 2 to pick one of the three.
*/
extern int acmRandomSelect(AcmRandomStruct * acm, int which);

/*!
  Returns the total possible random numbers generated
*/
extern long int acmRandomTotal(void);

/*!
  Returns a random number in the range [0 .. acmRandomTotal()-1],
  inclusive
*/
extern long int acmRandom(AcmRandomStruct * acm);

/*!
  Returns a double in the range [0.000000 .. 0.99999X], which is
  simply acmRandom() / acmRandomTotal(). You can multiply this
  by an integer M to get random numbers in the range [0..M-1],
  although they won't be perfectly balanced unless 
  acmRandomTotal() is an exact multiple of M. To get balanced numbers
  in a range [0..M-1] inclusive, call acmRandomRange(, M).
*/
extern double acmRandomReal(AcmRandomStruct * acm);

/*!
  Returns a random number in the range [0 .. range-1] inclusive,
  balanced so that over the long term all numbers are equally probable.
  This is accomplished by throwing some random generates away, perhaps
  up to almost half depending on 'range'. 
*/
extern long int acmRandomRange(AcmRandomStruct * acm, long int range);

/*
  General techniques for generating arbitrary variates give uniform
  variates are in Averill M. Law and W. David Kelton, _Simulation
  Modeling and Analysis_, Second Edition, McGraw-Hill, 1991,
  pp. 462-521.
*/

/*
  The normal variate uses two uniform distributions and composes them
  according to the polar method originally described in G. Marsaglia
  and T. A. Bray, "A Convenient Method for Generating Normal
  Variables," SIAM Rev., No. 6, 1964, 260-264.
*/

typedef struct {
  int hasLast;
  double last;
  AcmRandomStruct u1;
  AcmRandomStruct u2;
} AcmNormalStruct;

extern int acmNormalInit(AcmNormalStruct * nrm);
extern int acmNormalSeed(AcmNormalStruct * nrm, long int s1, long int s2);
extern double acmNormal(AcmNormalStruct * nrm, double mu, double sigma);

/*
  The exponential variate uses one uniform distribution U and inverts
  the CDF F(x) = 1 - exp(-x/mu) to get X = -mu*log(U) as per Law and
  Kelton.
*/

typedef struct {
  AcmRandomStruct u;
} AcmExponentialStruct;

extern int acmExponentialInit(AcmExponentialStruct * ex);
extern int acmExponentialSeed(AcmExponentialStruct * ex, long int s);
extern double acmExponential(AcmExponentialStruct * ex, double mu);

#ifdef __cplusplus
#if 0
{    /* matches brace above, to un-confuse autoformatting text editors */
#endif /* if 0 */
} /* matches extern "C" */
#endif /* ifdef __cplusplus */

#endif /* ACM_RANDOM_H */

/*! \defgroup acmRandom
  This random number generator is the one described in "Random Number
  Generators: Good Ones are Hard to Find," Stephen K. Park and Keith
  W. Miller, Communications of the ACM, Volume 31, Number 10, October
  1988.
  
  This generator is proposed as the "minimal standard" that will port
  to all systems where the maximum integer size is 2^32 - 1 or larger.

  The generating function is z(n+1) = a*z mod m, where m is the
  modulus 2^32 - 1. It generates all numbers in the range 1..m-1,
  or 1 to 2147483646. 0 and the modulus 2^32-1 are never generated, so
  two numbers out of the 2^32 possible in the four bytes of a long int
  are not generated, and 2^32-2 are generated.

  The function is shifted so that numbers in the range 0..2^32-3 are
  returned by acmRandom() for use. A normalized function
  acmRandomReal() is provided that returns a double, the shifted
  integer divided by 2^32-2, resulting in a floating point range
  [0.0000 .. 0.99999X], where 0.99999X is (2^32-3)/(2^32-2). This
  can be multiplied by an integer M and rounded to produce random
  integers in the range 0..M-1 with almost balanced distribution.

  With a random number generator that generates N possible numbers,
  to pick M numbers from that gives an imbalance if N is not a multiple
  of M.

  (N mod M) get (N div M) + 1 of the total N, and the remainder,
  M - (N mod M), get one less, (N div M).

  Consider the random generation of a playing card, for example, using
  a random byte generator. Here N = 256 and M = 52.  256 / 52 = 4
  remainder 48, and 48 of the cards are selected 5 times for a
  probability of 5/256 each, and 4 are selected 4 times for a
  probability of 4/256 each. Each of those unlucky 4 cards is 20% less
  likely to come up than each of the lucky 48 cards.

  For a 32-bit random number generator and 52 cards, N = 2^32 and
  M = 52. 2^32 / 52 = 82595524 remainder 48. The lucky 48 are selected 
  82595525 times and the unlucky 4 are selected 82595524 times,
  for probabilities 0.01923076925 and 0.01923076902, respectively.
  These are very close but not perfectly balanced.

  For this generator, N = 2^32-2 and M = 52. 2^32-2 / 52 = 82595524
  remainder 46. The lucky 46 are selected 82595525 times and the
  unlucky 6 are selected 82595524 times, for probabilities
  0.01923076926 and 0.01923076902, respectively.  These are very
  close but not perfectly balanced. 
*/

/*
  $Log: acmRandom.h,v $
  Revision 1.2  2007/02/01 18:20:21  proctor
  Added normal and exponential variates to random number generator; updated
  simpleSimWare to use ground vehicle bot.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.3  2005/06/10 13:24:17  proctor
  Changed doxygen comment

  Revision 1.2  2005/05/05 19:03:29  proctor
  Added doxygen-pleasing comments.

  Revision 1.1  2005/05/05 15:56:11  proctor
  Changed naming convention from e.g. NAV_DATA_EXT to NavDataExt

  Revision 1.1.1.1  2005/05/04 21:13:34  proctor
  Initial checkin of devel/ directory in moast.

*/
