/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   amSPCell.hh
  \brief  Contains class declarations for AmSPCell
  
  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.2 $
  $Date: 2006/09/28 20:31:02 $
  \endcode
  \par CVS Log:
  \ref amSPCell_hh "More..."

  \author Tom Kramer
  \date   2006-09-13
*/

#ifndef __amSPCell_hh
#define __amSPCell_hh

#include "scrollingMapCell.hh"
#define CELL_PROBINIT       0.2

/***********************************************************************/

/*! The AmSPCell class contains information about a cell.

Currently, the information is used to decide if we think there is an
obstacle on cell. 

Imagine there being a rectangular block whose cross section is the
same shape as the cell, and the block has been inserted into the
cell. The base of the block is somewhere down in the bowels of
the earth. We say we could see the block if we are within its range,
have an unobstructed view of a point on the block, and are looking
in the direction of that point.

1. height

The height is the distance we think the block extends above
the local XY plane. The height can be negative, implying we think the
top of the block is below the XY plane.

Whenever we see a point on the block, we reset its height to be the
negative of the z value of the point (negative since z is positive
downwards).  This is not in general a good way of setting the height,
since we are looking at only one point on the block, and we are not
taking into account whether we are looking up or down, but it is OK if
we are on a flat and level surface and always looking on sight lines
parallel to the XY plane.

Whenever we could see the block and do not see it, we set its height
to be effectively zero (actually the negative of the z value of the
position of the vehicle). This is also not, in general, a good way to
do it, but is OK under assumptions we are making for this version of
MOAST that (1) all objects are tall (2) the ground is flat, (3) the
ground is level and (4) our line of sight is always parallel to the
ground.

2. range

If a block has ever been seen and we have not failed to see it three
times in a row when we could have seen it, the range represents a
distance to the point within which we expect to be able to see
it. Some obstacles are small and can only be seen when we get close to
them.  Each time we see a point, if the range from which we saw it is
greater than 1.11 times the currently set range, we set the range to
0.9 times the range from which we saw it. If we fail to see the block
three times in a row when we could have seen it, the range is set
to zero on the assumption that whatever was there has gone.

If a block has never been seen, the value of range has no meaning. This
might be changed so that the range of an unseen block is the closest
we ever got to it and looked in its direction without seeing it. But
that data has no use, currently.

3. hits

This is an unsigned char representing what happened the last seven
times we could have seen the block. It is the index in a 255-element
array of all allowed seven-place base three numbers arranged in
increasing numerical order. That array does not appear in the source
code, but the same index is used for three arrays (hitsYes, hitsNo,
and probs) used for dealing with hits.

A seven-place base three number is allowed if any zeros in the number
appear at the beginning. For example, 0012112 is allowed, but
0012012 is not.

The far right digit in the number represents the most recent hit
and the far left digit represents the oldest hit. The meaning of the
base 3 digits is:
0 = have not been where we could see it
1 = did see it
2 = did not see it
For example, if hits is 43, that is index of the number 0012211. The
number means
(1) We could have seen it five times,
(2) The first, fourth and fifth times we could have seen it we did see it.
(3) The second and third times we could have seen it we did not see it. 
Hits is initialized to 0.

Conceptually, each time we could have seen the cell, we remove the
leftmost digit, shift the remaining base 3 digits one place to the
left, and put either a 1 or a 2 in the rightmost digit. This operation
is never actually performed. Instead, all possibilities have already
been computed and stored in the hitsYes and hitsNo arrays. If we did
see the cell, we look up the index of the new number in the hitsYes
array, and if we did not see it, we look up the number in the hitsNo
array. The current value of hits is used as an index into those
arrays. For example, if hits is 43 and we see the cell, the new
value of hits should be hitsYes[43], which is 87, representing 0122111.

4. probability

This is a value between 0 and 1 representing the likelihood that the
block is an obstacle. 

The enum elt is a list of constants that can be used to refer to any
of the elements within a Cell.

*/

class AmSPCell : public ScrollingMapCell
{
public:
  AmSPCell() : ScrollingMapCell()
    {height=0; range=0; probability = CELL_PROBINIT; hits = 0;}
  ~AmSPCell(){}

  double height;
  double range;
  unsigned char hits;
  double probability;

/***********************************************************************/

//! Method init reinitializes a cell.

  void init(){height=0; range=0; probability = CELL_PROBINIT; hits = 0;}

/***********************************************************************/

/*! Method setCellData sets the data in a cell, using values
    given in the arguments. */

  bool setCellData(double heightIn, double rangeIn, bool seen);

/***********************************************************************/

/*! Method setCellDataCopy sets the data in a cell by
    copying the data from another cell. */

  void setCellDataCopy(AmSPCell * from){*this = *from;}

/***********************************************************************/

};

#endif

/*! \defgroup amSPMap_hh CVS Log Information
 * $Log: amSPCell.hh,v $
 * Revision 1.2  2006/09/28 20:31:02  tomrkramer
 * Made AmSPCell be a derived class of ScrollingMapCell
 *
 * Revision 1.1  2006/09/18 18:54:36  tomrkramer
 * Initial entry. This file is used by scrolling map code.
 *

 */
