/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   amSPMap.hh
  \brief  Contains class declarations for PointInfo and AmSPMap
  
   The user can also display the internal map using in a couple different
  ways.  The functions printMapXXXX() let the user print height, range,
  count, probability, and seen (ie: printMapHeight(), printMapRange(),
  etc.).  There is also a display utility that pipes the map to a ppm
  file which ppmsDisplay puts on the screen.  First call createDisplay()
  and then, every time you want to put a new image, call showDisplay().

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.10 $
  $Date: 2008/02/11 23:40:17 $
  \endcode
  \par CVS Log:
  \ref amSPMap_hh "More..."

  \author Adam Beytin
  \date   2005-06-23
*/

#ifndef __amSPMap_hh
#define __amSPMap_hh
#include "moastTypes.hh"
#include "posemath.h"
#include "navDataExt.hh"
#include "sensorData.hh"
#define CELL_NAME AmSPCell
#include "amSPCell.hh"
#include "scrollingMap.hh"
#include <list>
#include <utility>

using namespace std;

class MapInfo
{
  public:
    double row;
    double col;
    float prob;

    MapInfo()
    {
      row=0;
      col=0;
      prob=0.0;
    }


    inline bool operator == (const MapInfo &b) const
    {
      return (row == b.row && col == b.col);
    }

    inline bool operator < (const MapInfo &b) const
    {
      return ((row < b.row) || ((row == b.row) && (col < b.col)));
    }
};

/***********************************************************************/

/*! The AmSPMap class contains the data and methods for the AM SP
    map. It is derived from the ScrollingMap class.
    The "small map" is small square centered inside the window. */

class AmSPMap : public ScrollingMap
{

/***PUBLIC**************************************************************/

public:
  list<MapInfo> MapChanges;
  double xLow;
  double yLow;
  double resolution;

  /*! AmSPMap constructor taking no arguments, does nothing except call
    the ScrollingMap constructor with no arguments. */

  AmSPMap() : ScrollingMap(){}

/***********************************************************************/

  /*! Constructor taking four arguments creates a map based on an initial
    range, resolution, output map range, and map image file name. */

  AmSPMap(int rangeIn, double resolutionIn, int outRangeIn,
	  char * mapImageFileNameIn);

/***********************************************************************/

  //! Destructor, deletes allocated memory

  ~AmSPMap();

/***********************************************************************/

  /*! Method createDisplay always deletes memory allocated to the "buf"
    buffer if the memory exists, closes the outFile if it is open,
    removes the map image file if it exists, and sets displayLevel
    to the level argument. In addition, if the level argument is
    1 or 2, it allocates memory for the "buf" buffer and starts
    imageview (which displays the picture). It would be nice if it
    would also stop imageview when level is zero, but it is not
    doing that. */

  bool createDisplay(int level);

/***********************************************************************/

  /*! Method getSmallMapCell gets a cell of the small map, given its
    row and column indexes on the small map. */

    AmSPCell * getSmallMapCell(int row, int col)
    {return getCellAtIndexes((getRange() + row - outRange),
                             (getRange() + col - outRange));}

/***********************************************************************/

  /*! Method setMapData sets the map cell having the given window grid
    row and col indexes by using data given in the arguments. */

  bool setMapData(int row, int col, double height, double range, bool seen);

/***********************************************************************/

  /*! Method setMapDataClear sets the cell whose cartesian X and Y
    coordinates are given, if that cell is in the window. The
    method also clears all the cells in the window on a line from
    the point (xVeh, yVeh) to the point (xIn, yIn).
    We make the assumption that, because we saw through the cells
    on that line, no object is in those cells. */

  bool setMapDataClear(double xVeh, double yVeh,
		       double xIn, double yIn, double height, bool seen);

/***********************************************************************/

  /*! Method setMapDataCopy sets the map cell having the given window grid
    row and col indexes by copying data from another cell into it. */

  bool setMapDataCopy(AmSPCell * from, int row, int col);
  
/***********************************************************************/

/*! Method setVehLoc sets the vehicle location at the center of the
  window, causing the window to scroll. */

  void setVehLoc(PM_CARTESIAN vehLocIn)
  {
    int xDiff;
    int yDiff;

    vehLoc = vehLocIn;
    setCenter(vehLoc.x, vehLoc.y, &xDiff, &yDiff);
  }

/***********************************************************************/

  /*! Method showDisplay shows no view, or a height view, or an
    obstacle view of the map. */

  void showDisplay();

/***********************************************************************/

  //! Method showHeightDisplay shows a height view of the map.

  void showHeightDisplay();

/***********************************************************************/

  //! Method showObstacleDisplay shows an obstacle view of the map.

  void showObstacleDisplay();

/***********************************************************************/

  //! Method showVehicle puts the vehicle in the picture.

  void showVehicle(char * buf);

/***********************************************************************/

  /*! Method writeSmallMap populates a SensorData2D, which is the
    contents of the small map that will be exported. For each cell,
    only probability data is written. */

  void writeSmallMap(SensorData2D * data);

/***********************************************************************/

  /*! Method writeSmallMapData writes a text version of the small map to
    a file.  For each cell, all data is written. */

  void writeSmallMapData(char * fileName);

/** PRIVATE*************************************************************/

private:

  char * mapImageFileName; //! map image file name used for displaying map
  int outRange;            //! range of the output map (a number of cells)
  int outSide;             //! number of cells on a side of the output map
  int displayLevel;        //! 0=no display, 1=show height, 2=show obstacles
  FILE * outFile;          //! file descriptor for the ppm graphics file
  char * buf;              //! buffer for graphics data
  PM_CARTESIAN vehLoc;     //! vehicle location
};

#endif
/*! \defgroup amSPMap_hh CVS Log Information
 * $Log: amSPMap.hh,v $
 * Revision 1.10  2008/02/11 23:40:17  tomrkramer
 * Since there are now two types of dump (for showing height or obstacles),
 * and the (new) obstacle dump is the same as showSmallDisplay was except
 * that it covers a larger area, added showObstacleDisplay and removed
 * showSmallDisplay. Also renamed showDisplay to showHeightDisplay and
 * added a new showDisplay which chooses among no display, height display
 * and obstacle display. Simplified showVehicle, since it no longer
 * serves different size displays.
 *
 * Revision 1.9  2006/10/13 16:16:33  tomrkramer
 * Changed a call to the setCenter method of scrollingMap, to take the right
 * number of arguments.
 *
 * Revision 1.8  2006/09/25 19:02:40  dr_steveb
 * Moved location of am map probabilitys to avoid multiple declarations of "CELL_NAME"
 *
 * Revision 1.7  2006/09/18 18:48:50  tomrkramer
 * Significantly changed to work with scrollingMap.hh
 *
 *
 */
