/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file linkMonCtrl.hh

  \brief NML declarations for the vehicle link monitor program. 
  and settings.


  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.2 $
  $Date: 2007/09/18 20:02:32 $
  \endcode

  \author Stephen Balakirsky
*/

#include <stdio.h>
#include <rcs.hh>
#include <posemath.h>
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"

#ifndef LINK_MON_CTRLL_HH
#define LINK_MON_CTRLL_HH

#define LINK_MON_CMD_NAME "linkMonCmd"
#define LINK_MON_STAT_NAME "linkMonStat"

#define LINK_MON_CMD_BASE        (LINK_MON_BASE)
#define LINK_MON_STAT_BASE       (LINK_MON_BASE + 100)
#define LINK_MON_CFG_BASE        (LINK_MON_BASE + 200)
#define LINK_MON_SET_BASE        (LINK_MON_BASE + 300)

#define LINK_MON_CMD_INIT_TYPE          (LINK_MON_CMD_BASE + 1)
#define LINK_MON_CMD_ABORT_TYPE         (LINK_MON_CMD_BASE + 2)
#define LINK_MON_CMD_HALT_TYPE          (LINK_MON_CMD_BASE + 3)
#define LINK_MON_CMD_SHUTDOWN_TYPE      (LINK_MON_CMD_BASE + 4)
#define LINK_MON_CMD_NOP_TYPE           (LINK_MON_CMD_BASE + 5)
#define LINK_MON_CMD_MONITOR_TYPE       (LINK_MON_CMD_BASE + 6)
#define LINK_MON_CMD_WRITE_TYPE         (LINK_MON_CMD_BASE + 7)

#define LINK_MON_STAT_TYPE              (LINK_MON_STAT_BASE + 1)

#define LINK_MON_CFG_CYCLE_TIME_TYPE    (LINK_MON_CFG_BASE + 1)
#define LINK_MON_CFG_DEBUG_TYPE         (LINK_MON_CFG_BASE + 2)
#define LINK_MON_CFG_NOP_TYPE           (LINK_MON_CFG_BASE + 3)

#define LINK_MON_SET_TYPE               (LINK_MON_SET_BASE + 1)

/////////////////////////////////////////////////////////////////
// supporting structures

/////////////////////////////////////////////////////////////////
// commands
class LinkMonCmdInit:public RCS_CMD_MSG {
public:
  LinkMonCmdInit():RCS_CMD_MSG
    (LINK_MON_CMD_INIT_TYPE, sizeof(LinkMonCmdInit)) {};
  void update(CMS *);
};

class LinkMonCmdHalt:public RCS_CMD_MSG {
public:
  LinkMonCmdHalt():RCS_CMD_MSG
    (LINK_MON_CMD_HALT_TYPE, sizeof(LinkMonCmdHalt)) {};
  void update(CMS *);
};

class LinkMonCmdAbort:public RCS_CMD_MSG {
public:
  LinkMonCmdAbort():RCS_CMD_MSG
    (LINK_MON_CMD_ABORT_TYPE, sizeof(LinkMonCmdAbort)) {};
  void update(CMS *);
};

class LinkMonCmdShutdown:public RCS_CMD_MSG {
public:
  LinkMonCmdShutdown():RCS_CMD_MSG
    (LINK_MON_CMD_SHUTDOWN_TYPE, sizeof(LinkMonCmdShutdown)) {};
  void update(CMS *);
};

class LinkMonCmdNop:public RCS_CMD_MSG {
public:
  LinkMonCmdNop():RCS_CMD_MSG
    (LINK_MON_CMD_NOP_TYPE, sizeof(LinkMonCmdNop)) {};
  void update(CMS *);
};

class LinkMonCmdMonitor:public RCS_CMD_MSG {
public:
  LinkMonCmdMonitor():RCS_CMD_MSG
    (LINK_MON_CMD_MONITOR_TYPE, sizeof(LinkMonCmdMonitor)) {};
  void update(CMS *);
};


class LinkMonCmdWrite:public RCS_CMD_MSG {
public:
  LinkMonCmdWrite():RCS_CMD_MSG
    (LINK_MON_CMD_WRITE_TYPE, sizeof(LinkMonCmdWrite)) {};
  void update(CMS *);
  char robotToWrite[MOAST_WORLD_NAME_LEN];
  char msg[255];
};

/*!
  Status message
*/
typedef struct{
  char name[MAX_TRAFFIC_VEHICLES][80];
  float cost;
  int length;
}RadioRoute;

class LinkMonStat:public RCS_STAT_MSG_V2 {
public:
  LinkMonStat():RCS_STAT_MSG_V2(LINK_MON_STAT_TYPE, sizeof(LinkMonStat)) {};
  void update(CMS *);
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(RadioRoute, route, MAX_TRAFFIC_VEHICLES);
  // route_length autogenerated
};

/*!
  Commands that will come into the configuration channel
*/
class linkMonCfgCycleTime:public RCS_CMD_MSG {
public:
  linkMonCfgCycleTime():RCS_CMD_MSG
    (LINK_MON_CFG_CYCLE_TIME_TYPE, sizeof(linkMonCfgCycleTime)) {};
  void update(CMS *);

  double cycleTime;
};

class linkMonCfgNop:public RCS_CMD_MSG {
public:
  linkMonCfgNop():RCS_CMD_MSG
    (LINK_MON_CFG_NOP_TYPE, sizeof(linkMonCfgNop)) {};
  void update(CMS *);
};

class linkMonCfgDebug:public RCS_CMD_MSG {
public:
  linkMonCfgDebug():RCS_CMD_MSG
    (LINK_MON_CFG_DEBUG_TYPE, sizeof(linkMonCfgDebug)) {};
  void update(CMS *);

  int debug;			// debug mask
};

/*!
  linkMonSet is the class that holds the quasi-static  settings,
  those that change infrequently and in response to  configuration
  messages.
 */
class linkMonSet:public RCS_STAT_MSG_V2 {
public:
  linkMonSet():RCS_STAT_MSG_V2(LINK_MON_SET_TYPE, sizeof(linkMonSet)) {};
  void update(CMS *);
};

extern int linkMonCtrl_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * linkMonCtrl_symbol_lookup(long type);

#endif // LINK_MON_HH

/*
  Modification history:

  $Log: linkMonCtrl.hh,v $
  Revision 1.2  2007/09/18 20:02:32  dr_steveb
  Added write message to wss comm server for USARSim

  Revision 1.1  2007/09/11 20:27:43  dr_steveb
  Changes for link monitoring

*/
