/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primMobJA.hh

  \brief NML declarations for vehicle prim level command, status, config
  and settings.

  \code CVS Status:
  $Author: ghenson $
  $Revision: 1.18 $
  $Date: 2008/10/13 18:26:39 $
  \endcode

  \author Fred Proctor
*/

#ifndef PRIM_MOB_JA_HH
#define PRIM_MOB_JA_HH

#include <stdio.h>
#include "moastTypes.hh"	// MAX_WAYPOINTS
#include "moastNmlOffsets.hh"

#define PRIM_MOB_JA_CMD_NAME "primMobJACmd"
#define PRIM_MOB_JA_STAT_NAME "primMobJAStat"
#define PRIM_MOB_JA_CFG_NAME "primMobJACfg"
#define PRIM_MOB_JA_SET_NAME "primMobJASet"

#define PRIM_MOB_JA_CMD_BASE        (PRIM_MOB_JA_BASE + 0)
#define PRIM_MOB_JA_STAT_BASE       (PRIM_MOB_JA_BASE + 100)
#define PRIM_MOB_JA_CFG_BASE        (PRIM_MOB_JA_BASE + 200)
#define PRIM_MOB_JA_SET_BASE        (PRIM_MOB_JA_BASE + 300)
#define PRIM_MOB_JA_ERROR_BASE      (PRIM_MOB_JA_BASE + 400)

#define PRIM_MOB_JA_CMD_NOP_TYPE              (PRIM_MOB_JA_CMD_BASE + 1)
#define PRIM_MOB_JA_CMD_INIT_TYPE             (PRIM_MOB_JA_CMD_BASE + 2)
#define PRIM_MOB_JA_CMD_ABORT_TYPE            (PRIM_MOB_JA_CMD_BASE + 3)
#define PRIM_MOB_JA_CMD_HALT_TYPE             (PRIM_MOB_JA_CMD_BASE + 4)
#define PRIM_MOB_JA_CMD_SHUTDOWN_TYPE         (PRIM_MOB_JA_CMD_BASE + 5)
#define PRIM_MOB_JA_CMD_MOVE_ARC_SEGMENT_TYPE (PRIM_MOB_JA_CMD_BASE + 6)
#define PRIM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE    (PRIM_MOB_JA_CMD_BASE + 7)
#define PRIM_MOB_JA_CMD_ROTATE_TYPE           (PRIM_MOB_JA_CMD_BASE + 8)
#define PRIM_MOB_JA_CMD_VEL_TYPE              (PRIM_MOB_JA_CMD_BASE + 9)
#define PRIM_MOB_JA_CMD_STOP_TYPE             (PRIM_MOB_JA_CMD_BASE + 10)
#define PRIM_MOB_JA_CMD_DOCK_TYPE             (PRIM_MOB_JA_CMD_BASE + 11)

#define PRIM_MOB_JA_STAT_TYPE       (PRIM_MOB_JA_STAT_BASE + 1)

#define PRIM_MOB_JA_CFG_NOP_TYPE             (PRIM_MOB_JA_CFG_BASE + 1)
#define PRIM_MOB_JA_CFG_CYCLE_TIME_TYPE      (PRIM_MOB_JA_CFG_BASE + 2)
#define PRIM_MOB_JA_CFG_MOVE_TYPE            (PRIM_MOB_JA_CFG_BASE + 3)
#define PRIM_MOB_JA_CFG_DEBUG_TYPE           (PRIM_MOB_JA_CFG_BASE + 4)

#define PRIM_MOB_JA_SET_TYPE        (PRIM_MOB_JA_SET_BASE + 1)

enum {PRIM_MOB_JA_ERROR_BAD_CFG = PRIM_MOB_JA_ERROR_BASE + 1};


class PrimMobJACmdNop:public RCS_CMD_MSG {
public:
  PrimMobJACmdNop():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_NOP_TYPE,
     sizeof(PrimMobJACmdNop)) {};
  void update(CMS *);
};

class PrimMobJACmdInit:public RCS_CMD_MSG {
public:
  PrimMobJACmdInit():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_INIT_TYPE,
     sizeof(PrimMobJACmdInit)) {};
  void update(CMS *);
};

class PrimMobJACmdHalt:public RCS_CMD_MSG {
public:
  PrimMobJACmdHalt():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_HALT_TYPE,
     sizeof(PrimMobJACmdHalt)) {};
  void update(CMS *);
};

class PrimMobJACmdAbort:public RCS_CMD_MSG {
public:
  PrimMobJACmdAbort():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_ABORT_TYPE,
     sizeof(PrimMobJACmdAbort)) {};
  void update(CMS *);
};

class PrimMobJACmdShutdown:public RCS_CMD_MSG {
public:
  PrimMobJACmdShutdown():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_SHUTDOWN_TYPE,
     sizeof(PrimMobJACmdShutdown)) {};
  void update(CMS *);
};

/*!
  The PrimMobJACmdMoveArc NML message contains an array of ArcSegments,
  up to MAX_ARC_SEGMENTS, in relative coordinates.
*/

class PrimMobJACmdMoveArcSegment:public RCS_CMD_MSG {
public:
  PrimMobJACmdMoveArcSegment():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_MOVE_ARC_SEGMENT_TYPE,
     sizeof(PrimMobJACmdMoveArcSegment)) {};
  void update(CMS *);

  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(ArcSegment, arcList, MAX_ARC_SEGMENTS);
  // will now get 'int arcList_length'
};

/*!
  The PrimMobJACmdMoveWaypoint NML message contains an array of waypoints,
  up to MAX_WAYPOINTS, in relative coordinates.
*/

class PrimMobJACmdMoveWaypoint:public RCS_CMD_MSG {
public:
  PrimMobJACmdMoveWaypoint():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE,
     sizeof(PrimMobJACmdMoveWaypoint)) {};
  void update(CMS *);
  bool prim_path;

  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(Waypoint, waypoint, MAX_WAYPOINTS);
  // will now get 'int waypoint_length'
};

class PrimMobJACmdRotate:public RCS_CMD_MSG {
public:
  PrimMobJACmdRotate():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_ROTATE_TYPE,
     sizeof(PrimMobJACmdRotate)) {};
  void update(CMS *);

  double theta;			// yaw angle, relative coords, [rad]
  double thetaTol;		// tolerance, [rad]
  /*!
    \a direction can take a value from \a PRIM_MOB_JA_CMD_ROTATE_CW,
    \a PRIM_MOB_JA_CMD_ROTATE_CCW or \a PRIM_MOB_JA_CMD_ROTATE_SHORTEST,
    where the \a CW suffix means clockwise, the \a CCW suffix means
    counterclockwise and the \a SHORTEST suffix means take the direction
    with the shortest rotation.

    Note that clockwise would result in north-to-east rotation and
    counterclockwise would result in north-to-west rotation, consistent
    with an overhead view of the vehicle. Since the vehicle's Z axis
    points down, the yaw values increase for clockwise rotation and
    decrease with counterclockwise rotation.
  */
  int direction;		
};

class PrimMobJACmdVel:public RCS_CMD_MSG {
public:
  PrimMobJACmdVel():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_VEL_TYPE,
     sizeof(PrimMobJACmdVel)) {};
  void update(CMS *);

  double v;			// trans speed
  double w;			// rot speed
};

class PrimMobJACmdStop:public RCS_CMD_MSG {
public:
  PrimMobJACmdStop():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_STOP_TYPE,
     sizeof(PrimMobJACmdStop)) {};
  void update(CMS *);
};

/*! Command to dock to a station. The command must contain a rectangle of 
free space, a rectangle that represents the docking station (obstacle), and the 
goal point for the dock. Note that the goal point is the location on the station
to dock with, not the location where the center of the vehicle will be when docked.
*/
class PrimMobJACmdDock:public RCS_CMD_MSG {
public:
  PrimMobJACmdDock():RCS_CMD_MSG
    (PRIM_MOB_JA_CMD_DOCK_TYPE,
     sizeof(PrimMobJACmdDock)) {};
  void update(CMS *);

  PM_CARTESIAN freeSpaceUpperRight;
  PM_CARTESIAN freeSpaceLowerLeft;

  PM_CARTESIAN stationUpperRight;
  PM_CARTESIAN stationLowerLeft;

  PM_CARTESIAN dockPoint;
};


/*!
  PrimMobJAStat is the class that holds the dynamic vehicle status,
  things that change frequently and in response to vehicle command
  messages.
 */

class PrimMobJAStat:public RCS_STAT_MSG_V2 {
public:
  PrimMobJAStat() : RCS_STAT_MSG_V2(PRIM_MOB_JA_STAT_TYPE, sizeof(PrimMobJAStat)) {};
  void update(CMS *);

  int command_type_sent;	// the last one sent down to servo
  int serial_number_sent;	// the last one sent down to servo
  int pathIndex;		// which one we're on, 0 = first one
};

class PrimMobJACfgCycleTime:public RCS_CMD_MSG {
public:
  PrimMobJACfgCycleTime():RCS_CMD_MSG
    (PRIM_MOB_JA_CFG_CYCLE_TIME_TYPE,
     sizeof(PrimMobJACfgCycleTime)) {};
  void update(CMS *);

  double cycleTime;
};

class PrimMobJACfgNop:public RCS_CMD_MSG {
public:
  PrimMobJACfgNop():RCS_CMD_MSG
    (PRIM_MOB_JA_CFG_NOP_TYPE,
     sizeof(PrimMobJACfgNop)) {};
  void update(CMS *);
};

class PrimMobJACfgMove:public RCS_CMD_MSG {
public:
  PrimMobJACfgMove():RCS_CMD_MSG
    (PRIM_MOB_JA_CFG_MOVE_TYPE,
     sizeof(PrimMobJACfgMove)) {};
  void update(CMS *);

  double vmax;
  double amax;
  double wmax;
  double alphamax;
  double vCutoffAngle;
  double wCutoffAngle;
};

#define PRIM_MOB_JA_DEBUG_MOVE     0x00000001
#define PRIM_MOB_JA_DEBUG_WAYPOINT 0x00000002
#define PRIM_MOB_JA_DEBUG_ARCS     0x00000004
#define PRIM_MOB_JA_DEBUG_ICR      0x00000008
#define PRIM_MOB_JA_DEBUG_ALL      0xFFFFFFFF

class PrimMobJACfgDebug:public RCS_CMD_MSG {
public:
  PrimMobJACfgDebug():RCS_CMD_MSG
    (PRIM_MOB_JA_CFG_DEBUG_TYPE,
     sizeof(PrimMobJACfgDebug)) {};
  void update(CMS *);

  int debug;			// debug mask
};

/*!
  PrimMobJASet is the class that holds the quasi-static vehicle settings,
  those that change infrequently and in response to vehicle configuration
  messages.
 */

class PrimMobJASet:public RCS_STAT_MSG {
public:
  PrimMobJASet():RCS_STAT_MSG(PRIM_MOB_JA_SET_TYPE, sizeof(PrimMobJASet)) {};
  void update(CMS *);

  double cycleTime;
  int debug;			// debug mask
  double vmax;			// max translational speed, [m/s]
  double amax;			// max translational acceleration, [m/s/s]
  double wmax;			// max rotational speed, [rad/s]
  double alphamax;		// max rotational acceleration, [rad/s/s]
  // value of angular deviation above which speed is set to zero,
  // leading to only rotation
  double vCutoffAngle;		// [rad]
  // value of angular deviation above which angular speed is clamped
  // to wmax
  double wCutoffAngle;		// [rad]
  PM_POSE controlPoint;	// (inverse of) pose of control point wrt nav data frame
};

extern int primMobJA_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * primMobJA_symbol_lookup(long type);

#endif // PRIM_MOB_JA_HH

/*
  Modification history:

  $Log: primMobJA.hh,v $
  Revision 1.18  2008/10/13 18:26:39  ghenson
  Update to prim. Only sent the robot into reverse if prim generated the path

  Revision 1.17  2008/05/22 18:49:46  ghenson
  Added debug flag for logging current ICR

  Revision 1.16  2008/04/15 17:05:23  dr_steveb
  Sample docking command in prim

  Revision 1.15  2008/04/15 15:06:32  dr_steveb
  Added new docking command

  Revision 1.14  2008/02/06 14:57:17  dr_steveb
  Increased size of arrays

  Revision 1.13  2008/01/08 20:18:15  proctor
  Added control point transform that lets the prim mobility controller act
  on the rotational center of the vehicle, which in general won't coincide
  with the nav data frame

  Revision 1.12  2007/12/19 14:48:39  proctor
  Added printDebugMsg to primMobEngine, and pulled the waypoint debug printing
  into this. Set the debug level to PRIM_MOB_JA_DEBUG_WAYPOINT, 0x02, to
  enable debug printing of waypoint information. The debug level is set to 0
  in moast.ini by default, so you will see no debug information.

  Revision 1.11  2007/05/16 15:16:21  dr_steveb
  Fixed location of state printing routines to remove compile warnings.

  Revision 1.10  2007/05/10 13:15:31  dr_steveb
  Moved rotation and arcSegment declaration from primMobJA to moastTypes so that multiple apps can use them.

  Revision 1.9  2006/10/02 14:15:10  proctor
  Added comment documentation to 'direction' attribute of PrimMobJACmdRotate.

  Revision 1.8  2006/10/02 13:59:09  proctor
  Changed PrimMobJACmdRotate to take a 'direction' value, either
  PRIM_MOB_JA_CMD_ROTATE_CW,CCW,SHORTEST. Changed primShell accordingly.

  Revision 1.7  2006/09/20 21:28:07  proctor
  Added PrimMobJACmdStop and related code for sending it from the prim shell
  and handling it in primMobMain.

  Revision 1.6  2006/09/20 12:47:22  dr_steveb
  Added clockwise flag to rotate command

  Revision 1.5  2006/06/21 19:38:13  proctor
  Added support for Java, which involved including some left-out headers
  like moastNmlOffsets.hh and moastTypes.hh

  Revision 1.4  2006/04/25 20:30:17  proctor
  Added support for moastError to nmlPrint, and added error printing for
  bad configurations to primMobJA

  Revision 1.3  2006/04/13 14:30:07  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.2  2006/04/05 01:58:31  dr_steveb
  Added NOP command

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.8  2005/10/03 13:28:37  kramer
  moved setting MAX_ARC_SEGMENTS to moastTypes.hh

  Revision 1.7  2005/08/10 16:36:01  proctor
  Changed cutoffAngle to vCutoffAngle, wCutoffAngle

  Revision 1.6  2005/07/20 21:52:31  proctor
  servoSP,Mis,MobJA.hh: added debug cfg
  primMobJA.hh: added 'isArc' flag to ArcSegment

  Revision 1.5  2005/07/15 15:10:24  proctor
  Changed SERVO_MOB_JA_DEBUG_ALL to the correct PRIM_MOB_JA_DEBUG_ALL

  Revision 1.4  2005/07/15 15:01:16  proctor
  primMobJA.hh: added PRIM_MOB_JA_DEBUG_ALL
  servoMobJA.hh: added SERVO_MOB_JA_DEBUG_ALL

  Revision 1.3  2005/06/27 14:27:08  proctor
  Added arc segment move command

  Revision 1.2  2005/06/08 14:29:54  proctor
  Added vehMobPLMain name change

  Revision 1.1  2005/06/08 12:47:40  proctor
  Name change per Mis,MobJA naming convention

  Revision 1.7  2005/06/01 13:20:45  proctor
  Moved Waypoint structure into moastTypes.hh

  Revision 1.6  2005/05/18 12:34:13  proctor
  Took out #include "moastNmlOffsets.hh", since code gen bug is fixed

  Revision 1.5  2005/05/17 21:28:01  proctor
  Worked around code gen bug by #include "moastNmlOffsets.hh"

  Revision 1.4  2005/05/13 19:02:33  proctor
  Added some comments to primMobJA.hh; added shutdown NML message to servoPL.hh

  Revision 1.3  2005/05/12 13:19:47  proctor
  Got rid of Waypoint structure, and put fields directly into NML message

  Revision 1.2  2005/05/12 12:40:22  proctor
  Added a constructor to Waypoint

  Revision 1.1  2005/05/12 12:36:03  proctor
  Initial checkin of primMobJA.hh

*/
