/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file sectMobPL.hh

  \brief NML declarations for secticle sect level command, status, config
  and settings.

  Synopsis:

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.12 $
  $Date: 2009/02/25 21:06:52 $
  \endcode

  \author Stephen Balakirsky
*/

#ifndef SECT_MOB_PL_HH
#define SECT_MOB_PL_HH

#include <rcs.hh>
#include <posemath.h>
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"

#define SECT_MOB_PL_CMD_NAME "sectMobPLCmd"
#define SECT_MOB_PL_STAT_NAME "sectMobPLStat"
#define SECT_MOB_PL_CFG_NAME "sectMobPLCfg"
#define SECT_MOB_PL_SET_NAME "sectMobPLSet"

#define SECT_MOB_PL_CMD_BASE        (SECT_MOB_PL_BASE)
#define SECT_MOB_PL_STAT_BASE       (SECT_MOB_PL_BASE + 100)
#define SECT_MOB_PL_CFG_BASE        (SECT_MOB_PL_BASE + 200)
#define SECT_MOB_PL_SET_BASE        (SECT_MOB_PL_BASE + 300)

#define SECT_MOB_PL_CMD_INIT_TYPE          (SECT_MOB_PL_CMD_BASE + 1)
#define SECT_MOB_PL_CMD_ABORT_TYPE         (SECT_MOB_PL_CMD_BASE + 2)
#define SECT_MOB_PL_CMD_HALT_TYPE          (SECT_MOB_PL_CMD_BASE + 3)
#define SECT_MOB_PL_CMD_SHUTDOWN_TYPE      (SECT_MOB_PL_CMD_BASE + 4)
#define SECT_MOB_PL_CMD_NOP_TYPE           (SECT_MOB_PL_CMD_BASE + 5)
#define SECT_MOB_PL_CMD_EXPLORE_TYPE       (SECT_MOB_PL_CMD_BASE + 6)
#define SECT_MOB_PL_CMD_CONNECT_TYPE       (SECT_MOB_PL_CMD_BASE + 7)
#define SECT_MOB_PL_CMD_DISCONNECT_TYPE    (SECT_MOB_PL_CMD_BASE + 8)
#define SECT_MOB_PL_CMD_CRASH_TYPE         (SECT_MOB_PL_CMD_BASE + 9)
//! the cmd load command tells the vehicles to load a world model
#define SECT_MOB_PL_CMD_LOAD_WM_TYPE       (SECT_MOB_PL_CMD_BASE +10)
#define SECT_MOB_PL_CMD_DELIVER_TYPE       (SECT_MOB_PL_CMD_BASE +11)
#define SECT_MOB_PL_CMD_WAYPOINT_TYPE      (SECT_MOB_PL_CMD_BASE +12)
#define SECT_MOB_PL_CMD_FILE_TYPE          (SECT_MOB_PL_CMD_BASE +13)

#define SECT_MOB_PL_STAT_TYPE              (SECT_MOB_PL_STAT_BASE + 1)

#define SECT_MOB_PL_CFG_CYCLE_TIME_TYPE    (SECT_MOB_PL_CFG_BASE + 1)
#define SECT_MOB_PL_CFG_DEBUG_TYPE         (SECT_MOB_PL_CFG_BASE + 2)
#define SECT_MOB_PL_CFG_NOP_TYPE           (SECT_MOB_PL_CFG_BASE + 3)
#define SECT_MOB_PL_CFG_CLEAR_WM_TYPE      (SECT_MOB_PL_CFG_BASE + 4)
#define SECT_MOB_PL_CFG_STORE_WM_TYPE      (SECT_MOB_PL_CFG_BASE + 5)
#define SECT_MOB_PL_CFG_DUMP_WM_TYPE       (SECT_MOB_PL_CFG_BASE + 6)
//! the config load command loads a world model into the section
#define SECT_MOB_PL_CFG_LOAD_WM_TYPE       (SECT_MOB_PL_CFG_BASE + 7)

#define SECT_MOB_PL_SET_TYPE        (SECT_MOB_PL_SET_BASE + 1)

#define MAX_SECTION_VEHICLES 4

/*!
 define who the subordinate of the section echelon is
*/
enum SectSubordinate {
  SERVO,
  PRIM,
  AM,
  VEHICLE
};

/////////////////////////////////////////////////////////////////
// supporting structures
typedef struct {
  Waypoint waypoint;
  int vehID;
}SectWaypoint;

/////////////////////////////////////////////////////////////////
// commands
class SectMobPLCmdInit:public RCS_CMD_MSG {
public:
  SectMobPLCmdInit():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_INIT_TYPE,
     sizeof(SectMobPLCmdInit)) {};
  // define who our subordinate is
  SectSubordinate  subordinate;
  /*!
     Variable length array of vehicle ids to control. The name of the array is
     "vehList", whose maximum length is defined by MAX_SECTION_VEHICLES.
     NML creates a variable named "vehList_length" that contains the number
     of vehicles in the current message.
   */
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(int, vehList,
				   MAX_SECTION_VEHICLES);

  void update(CMS *);
};

class SectMobPLCmdHalt:public RCS_CMD_MSG {
public:
  SectMobPLCmdHalt():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_HALT_TYPE,
     sizeof(SectMobPLCmdHalt)) {};
  void update(CMS *);
};

class SectMobPLCmdAbort:public RCS_CMD_MSG {
public:
  SectMobPLCmdAbort():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_ABORT_TYPE,
     sizeof(SectMobPLCmdAbort)) {};
  void update(CMS *);
};

class SectMobPLCmdShutdown:public RCS_CMD_MSG {
public:
  SectMobPLCmdShutdown():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_SHUTDOWN_TYPE,
     sizeof(SectMobPLCmdShutdown)) {};
  void update(CMS *);
};

class SectMobPLCmdNop:public RCS_CMD_MSG {
public:
  SectMobPLCmdNop():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_NOP_TYPE,
     sizeof(SectMobPLCmdNop)) {};
  void update(CMS *);
};

/*!
  The SectMobPLCmdExplore NML message is used to explore an
  unknown region and develop a map.
*/
class SectMobPLCmdExplore:public RCS_CMD_MSG {
public:
  SectMobPLCmdExplore():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_EXPLORE_TYPE,
     sizeof(SectMobPLCmdExplore)) {};
  void update(CMS *);

  //! max distance to explore in each leg of exploration
  double exploreDist;
  //! vehicle level plan should not exceed this value
  double maxCost;
};

/*!
  The SectMobPLCmdConnect NML message is used to connect to
  subordinate vehicles. The init flag is used to tell if we
  should initialize the subordinates
*/
class SectMobPLCmdConnect:public RCS_CMD_MSG {
public:
  SectMobPLCmdConnect():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_CONNECT_TYPE,
     sizeof(SectMobPLCmdConnect)) {};
  void update(CMS *);

  //! IDs of vehicles that should be connected to
    /*!
     Variable length array of vehicles to command. The name of the array is
     "vehID", the type is int, and the maximum length is defined by 
     MAX_SECTION_VEHICLES.
     NML creates a variable named "vehID_length" that contains the number
     of vehicle IDs in the current message.
   */
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(int, vehID, MAX_SECTION_VEHICLES);
  //! should an init command be sent to subordinates (true is yes)
  bool doSendInit;
};

/*!
  The SectMobPLCmdDisconnect NML message is used to disconnect
  subordinate vehicles. The halt and shutdown flags are used to tell if we
  should halt and/or shutdown the vehicles
*/
class SectMobPLCmdDisconnect:public RCS_CMD_MSG {
public:
  SectMobPLCmdDisconnect():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_DISCONNECT_TYPE,
     sizeof(SectMobPLCmdDisconnect)) {};
  void update(CMS *);

  //! should a halt command be sent out? True is do send
  bool doSendHalt;
  //! should a shutdown command be sent out? True is do send
  bool doSendShutdown;
};

/*!
  The SectMobPLCmdCrash NML message is used to test 
  the crash avoidance system by crashing two vehicles
  into each other while applying a crash warning system
*/
class SectMobPLCmdCrash:public RCS_CMD_MSG {
public:
  SectMobPLCmdCrash():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_CRASH_TYPE,
     sizeof(SectMobPLCmdCrash)) {};
  void update(CMS *);

  //! simulated driver reaction time in seconds
  float driverReact;
  //! simulated warning distance in meters
  float warnDist;
};

/*!
  The SectMobPLCmdFile NML message is used to
  send a file that contains actions to the vehicles.
  Please see the file sectWaypoint.dat in the bin
  directory for a list of accepted commands
*/
class SectMobPLCmdFile:public RCS_CMD_MSG {
public:
  SectMobPLCmdFile():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_FILE_TYPE,
     sizeof(SectMobPLCmdFile)) {};
  void update(CMS *);

  //! file name of waypoint file
  char fileName[MOAST_FILE_NAME_LEN];
  //! should the file be repeated forever
  bool repeat;
};

/*!
  The SectMobPLCmdFile NML message is used to
  send waypoints for multiple vehicle
*/
class SectMobPLCmdWaypoint:public RCS_CMD_MSG {
public:
  SectMobPLCmdWaypoint():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_WAYPOINT_TYPE,
     sizeof(SectMobPLCmdWaypoint)) {};
  void update(CMS *);

  // variable vehPoints_length automatically created to tell the number of points
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(SectWaypoint, vehPoints, MAX_SECTION_VEHICLES);
};

class SectMobPLCmdLoadWM:public RCS_CMD_MSG {
public:
  SectMobPLCmdLoadWM():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_LOAD_WM_TYPE,
     sizeof(SectMobPLCmdLoadWM)) {};
  void update(CMS *);
  char fileName[MOAST_FILE_NAME_LEN];
};

class SectMobPLCmdDeliver:public RCS_CMD_MSG {
public:
  SectMobPLCmdDeliver():RCS_CMD_MSG
    (SECT_MOB_PL_CMD_DELIVER_TYPE,
     sizeof(SectMobPLCmdDeliver)) {};
  void update(CMS *);
};


/*!
  Status message
*/
class SectMobPLStat:public RCS_STAT_MSG_V2 {
public:
  SectMobPLStat():RCS_STAT_MSG_V2(SECT_MOB_PL_STAT_TYPE, sizeof(SectMobPLStat)) {};
  void update(CMS *);

  //! system heartbeat
  int heartbeat;
  //! number of vehicles section is controlling
  int numVehicles;
  //! what level are we connected to?
  SectSubordinate  subordinate;
  //! message out of system
  char message[80];
};

/*!
  Commands that will come into the configuration channel
*/
class SectMobPLCfgCycleTime:public RCS_CMD_MSG {
public:
  SectMobPLCfgCycleTime():RCS_CMD_MSG
    (SECT_MOB_PL_CFG_CYCLE_TIME_TYPE,
     sizeof(SectMobPLCfgCycleTime)) {};
  void update(CMS *);

  double cycleTime;
};

class SectMobPLCfgNop:public RCS_CMD_MSG {
public:
  SectMobPLCfgNop():RCS_CMD_MSG
    (SECT_MOB_PL_CFG_NOP_TYPE,
     sizeof(SectMobPLCfgNop)) {};
  void update(CMS *);
};

#define SECT_MOB_PL_DEBUG_EXPLORE 0x00000001

class SectMobPLCfgDebug:public RCS_CMD_MSG {
public:
  SectMobPLCfgDebug():RCS_CMD_MSG
    (SECT_MOB_PL_CFG_DEBUG_TYPE,
     sizeof(SectMobPLCfgDebug)) {};
  void update(CMS *);

  int debug;			// debug mask
};

class SectMobPLCfgClearWM:public RCS_CMD_MSG {
public:
  SectMobPLCfgClearWM():RCS_CMD_MSG
    (SECT_MOB_PL_CFG_CLEAR_WM_TYPE,
     sizeof(SectMobPLCfgClearWM)) {};
  void update(CMS *);
};

class SectMobPLCfgStoreWM:public RCS_CMD_MSG {
public:
  SectMobPLCfgStoreWM():RCS_CMD_MSG
    (SECT_MOB_PL_CFG_STORE_WM_TYPE,
     sizeof(SectMobPLCfgStoreWM)) {};
  void update(CMS *);
  char fileName[MOAST_FILE_NAME_LEN];
};

class SectMobPLCfgDumpWM:public RCS_CMD_MSG {
public:
  SectMobPLCfgDumpWM():RCS_CMD_MSG
    (SECT_MOB_PL_CFG_DUMP_WM_TYPE,
     sizeof(SectMobPLCfgDumpWM)) {};
  void update(CMS *);
  char fileName[MOAST_FILE_NAME_LEN];
  //! dump the world model continuously
  bool dumpContinuous;
  //! skip 'skip' cycles between continuous dumps
  int skip;
  //! turn on or off output (true on, false off)
  bool turnOn;
};

class SectMobPLCfgLoadWM:public RCS_CMD_MSG {
public:
  SectMobPLCfgLoadWM():RCS_CMD_MSG
    (SECT_MOB_PL_CFG_LOAD_WM_TYPE,
     sizeof(SectMobPLCfgLoadWM)) {};
  void update(CMS *);
  char fileName[MOAST_FILE_NAME_LEN];
};

/*!
  SectMobPLSet is the class that holds the quasi-static secticle settings,
  those that change infrequently and in response to secticle configuration
  messages.
 */
class SectMobPLSet:public RCS_STAT_MSG_V2 {
public:
  SectMobPLSet():RCS_STAT_MSG_V2(SECT_MOB_PL_SET_TYPE, sizeof(SectMobPLSet)) {};
  void update(CMS *);
  
  //! the desired cycleTime of the planner
  double cycleTime;
  //! debug mask
  int debug;			
  //! file name that world model will be read/written from/to
  char wmFileName[MOAST_FILE_NAME_LEN];
  //! when dumping a file, dump every xth world model snap shot
  int fileSkip;
  //! Resolution of world model
  double worldResolution;
  //! Height of world
  int worldHeight;
  //! Width of world
  int worldWidth;
};

extern int sectMobPL_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * sectMobPL_symbol_lookup(long type);

#endif // SECT_MOB_PL_HH

/*
  Modification history:

  $Log: sectMobPL.hh,v $
  Revision 1.12  2009/02/25 21:06:52  dr_steveb
  Added VMAC commands to section echelon and made the maximum number of vehicles a #define

  Revision 1.11  2008/02/15 20:20:34  dr_steveb
  Adde waypoint command to the section echelon

  Revision 1.10  2007/12/20 02:33:13  dr_steveb
  Added new commands for manufacturing

  Revision 1.9  2007/12/17 21:37:02  dr_steveb
  Added section level manufacturing application

  Revision 1.8  2006/09/18 17:00:18  dr_steveb
  Added crash behavior

  Revision 1.7  2006/04/06 21:24:52  dr_steveb
  There are now four vehicles supported

  Revision 1.6  2006/04/05 20:48:09  dr_steveb
  Fixed case sensitive issue with file

  Revision 1.5  2006/04/05 20:24:20  dr_steveb
  Start of new section algorithm for multi-vehicle control

  Revision 1.4  2005/11/14 18:12:13  dr_steveb
  Added flag to dump command and moved isInitialized into settings channel.

  Revision 1.3  2005/10/21 14:38:13  dr_steveb
  Added world dimensions to settings message.

  Revision 1.2  2005/10/21 12:45:24  dr_steveb
  Added fields to status messages.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.1  2005/10/09 14:37:31  stephen
  Initial check-in.


*/
