/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file sectPL.hh

  \brief NML declarations for section echelon command, status, config
  and settings.

  Synopsis:
  #include <rcs.hh>
  #include <posemath.h>
  #include "moastNmlOffsets.hh"
  #include "moastTypes.hh"

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Stephen Balakirsky
*/

#ifndef SECT_PL_HH
#define SECT_PL_HH

#define SECT_PL_CMD_NAME "sectPLCmd"
#define SECT_PL_STAT_NAME "sectPLStat"
#define SECT_PL_CFG_NAME "sectPLCfg"
#define SECT_PL_SET_NAME "sectPLSet"

#define SECT_PL_CMD_BASE        (SECT_PL_BASE)
#define SECT_PL_STAT_BASE       (SECT_PL_BASE + 100)
#define SECT_PL_CFG_BASE        (SECT_PL_BASE + 200)
#define SECT_PL_SET_BASE        (SECT_PL_BASE + 300)

#define SECT_PL_CMD_INIT_TYPE           (SECT_PL_CMD_BASE + 1)
#define SECT_PL_CMD_ABORT_TYPE          (SECT_PL_CMD_BASE + 2)
#define SECT_PL_CMD_HALT_TYPE           (SECT_PL_CMD_BASE + 3)
#define SECT_PL_CMD_SHUTDOWN_TYPE       (SECT_PL_CMD_BASE + 4)
#define SECT_PL_CMD_NEUTRALIZE_IED_TYPE (SECT_PL_CMD_BASE + 8)

#define SECT_PL_STAT_TYPE              (SECT_PL_STAT_BASE + 1)

#define SECT_PL_CFG_CYCLE_TIME_TYPE    (SECT_PL_CFG_BASE + 1)
#define SECT_PL_CFG_DEBUG_TYPE         (SECT_PL_CFG_BASE + 2)

#define SECT_PL_SET_TYPE        (SECT_PL_SET_BASE + 1)

/////////////////////////////////////////////////////////////////
// supporting structures

/////////////////////////////////////////////////////////////////
// commands
class SectPLCmdInit:public RCS_CMD_MSG {
public:
  SectPLCmdInit():RCS_CMD_MSG
    (SECT_PL_CMD_INIT_TYPE,
     sizeof(SectPLCmdInit)) {};
  void update(CMS *);

  //! name of sql host
  char server[MOAST_FILE_NAME_LEN];
  //! name of sql user
  char user[MOAST_FILE_NAME_LEN];
  //! name of sql database
  char database[MOAST_FILE_NAME_LEN];
  //! password for sql user
  char password[MOAST_FILE_NAME_LEN];
};

class SectPLCmdHalt:public RCS_CMD_MSG {
public:
  SectPLCmdHalt():RCS_CMD_MSG
    (SECT_PL_CMD_HALT_TYPE,
     sizeof(SectPLCmdHalt)) {};
  void update(CMS *);
};

class SectPLCmdAbort:public RCS_CMD_MSG {
public:
  SectPLCmdAbort():RCS_CMD_MSG
    (SECT_PL_CMD_ABORT_TYPE,
     sizeof(SectPLCmdAbort)) {};
  void update(CMS *);
};

class SectPLCmdShutdown:public RCS_CMD_MSG {
public:
  SectPLCmdShutdown():RCS_CMD_MSG
    (SECT_PL_CMD_SHUTDOWN_TYPE,
     sizeof(SectPLCmdShutdown)) {};
  void update(CMS *);
};

//! command to drive to a potential IED, examine it, and deal with it
/*! The exact behavior that this command fires is based on a FSM
 */
class SectPLCmdNeutralizeIED:public RCS_CMD_MSG {
public:
  SectPLCmdNeutralizeIED():RCS_CMD_MSG
    (SECT_PL_CMD_NEUTRALIZE_IED_TYPE,
     sizeof(SectPLCmdNeutralizeIED)) {};
  void update(CMS *);
  //! Time to start movement (in seconds since the Epoch (00:00:00 UTC, January 1, 1970)
  long startTime;
  //! the id of the IED to be neutralized
  int IEDID;
};


/*!
  Status message
*/
class SectPLStat:public RCS_STAT_MSG {
public:
  SectPLStat():RCS_STAT_MSG(SECT_PL_STAT_TYPE, sizeof(SectPLStat)) {};
  void update(CMS *);

  //! Lets us know that the system is alive
  int heartbeat;
  //! the ied that is being neutralized
  int IEDID;
  //! actual cycle time of planner
  double cycleTime;
};

/*!
  Commands that will come into the configuration channel
*/
class SectPLCfgCycleTime:public RCS_CMD_MSG {
public:
  SectPLCfgCycleTime():RCS_CMD_MSG
    (SECT_PL_CFG_CYCLE_TIME_TYPE,
     sizeof(SectPLCfgCycleTime)) {};
  void update(CMS *);

  double cycleTime;
};

#define SECT_PL_DEBUG_IED 0x00000001

class SectPLCfgDebug:public RCS_CMD_MSG {
public:
  SectPLCfgDebug():RCS_CMD_MSG
    (SECT_PL_CFG_DEBUG_TYPE,
     sizeof(SectPLCfgDebug)) {};
  void update(CMS *);

  int debug;			// debug mask
};


/*!
  SectPLSet is the class that holds the quasi-static vehicle settings,
  those that change infrequently and in response to vehicle configuration
  messages.
 */
class SectPLSet:public RCS_STAT_MSG {
public:
  SectPLSet():RCS_STAT_MSG(SECT_PL_SET_TYPE, sizeof(SectPLSet)) {};
  void update(CMS *);
  
  double cycleTime;
  int debug;			// debug mask
  bool isInitialized;
  char message[80];
};

extern int sectPL_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * sectPL_symbol_lookup(long type);

#endif // SECT_PL_HH

/*
  Modification history:

  $Log: sectPL.hh,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.1  2005/06/01 12:46:27  stephen
  Initial add of code.

*/
