
/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file servoSP.hh

  \brief NML declaration for the servo sensor processing level.  Contains
         servo level sensor command, status, set, and data messages.  

  \code CVS Status:
  $Author: cj_scrapper $
  $Revision: 1.10 $
  $Date: 2007/04/10 19:10:18 $
  \endcode

  \author Chris Scrapper
*/

#ifndef SERVO_SP_HH
#define SERVO_SP_HH

#include "moastTypes.hh"	// MOAST_WORLD_NAME_LEN
#include "moastNmlOffsets.hh"

#define SERVO_SP_CMD_NAME          "servoSPCmd"
#define SERVO_SP_STAT_NAME         "servoSPStat"
#define SERVO_SP_CFG_NAME          "servoSPCfg"
#define SERVO_SP_SET_NAME          "servoSPSet"
#define SERVO_SP_DATA_NAME         "servoSPData"

#define SERVO_SP_CMD_BASE            (SERVO_SP_BASE)
#define SERVO_SP_STAT_BASE           (SERVO_SP_BASE + 100)
#define SERVO_SP_CFG_BASE            (SERVO_SP_BASE + 200)
#define SERVO_SP_SET_BASE            (SERVO_SP_BASE + 300)
#define SERVO_SP_DATA_BASE           (SERVO_SP_BASE + 400)

#define SERVO_SP_CMD_INIT_TYPE       (SERVO_SP_CMD_BASE + 1)
#define SERVO_SP_CMD_ABORT_TYPE      (SERVO_SP_CMD_BASE + 2)
#define SERVO_SP_CMD_HALT_TYPE       (SERVO_SP_CMD_BASE + 3)
#define SERVO_SP_CMD_SHUTDOWN_TYPE   (SERVO_SP_CMD_BASE + 4)
#define SERVO_SP_CMD_GO_TYPE         (SERVO_SP_CMD_BASE + 5)
#define SERVO_SP_CMD_RESET_TYPE      (SERVO_SP_CMD_BASE + 6)

#define SERVO_SP_STAT_TYPE           (SERVO_SP_STAT_BASE + 1)

#define SERVO_SP_CFG_CYCLE_TIME_TYPE (SERVO_SP_CFG_BASE + 1)
#define SERVO_SP_CFG_DEBUG_TYPE      (SERVO_SP_CFG_BASE + 2)

#define SERVO_SP_SET_TYPE            (SERVO_SP_SET_BASE + 1)

typedef enum SENSOR_TYPE{
  SERVO_SP_LINESCANNER_SENSOR,
  SERVO_SP_SONAR_SENSOR,
  SERVO_SP_TOUCH_SENSOR,
  SERVO_SP_RFID_SENSOR,
  SERVO_SP_VICTIM_SENSOR,
  SERVO_SP_INU_SENSOR,
  SERVO_SP_INS_SENSOR,
  SERVO_SP_GPS_SENSOR,
  SERVO_SP_ODOMETRY_SENSOR,
  SERVO_SP_ENCODER_SENSOR,
  SERVO_SP_GRD_TRUTH_SENSOR,
  SERVO_SP_CAMERA_SENSOR
} ServoSensorType;

 
//! maximum number of sensor on entire vehicle
#define SERVO_SP_SENSOR_MAX 35

// Maximum number of sensor parameters
#define SERVO_SP_SENSOR_PARAMETER_MAX 6

//! Linescanner parameters
#define  SERVO_SP_LINESCAN_MAX_RANGE   0
#define  SERVO_SP_LINESCAN_MIN_RANGE   1
#define  SERVO_SP_LINESCAN_RESOLUTION  2
#define  SERVO_SP_LINESCAN_FOV 3
#define  SERVO_SP_LINESCAN_BEAM_DISPERSION 4

//! Sonar Parameters
#define  SERVO_SP_SONAR_MAX_RANGE  0
#define  SERVO_SP_SONAR_MIN_RANGE  1
#define  SERVO_SP_SONAR_BEAM_DISPERSION 2
//! Camera Parameters
#define  SERVO_SP_CAMERA_MIN_FOV     0
#define  SERVO_SP_CAMERA_MAX_FOV     1
#define  SERVO_SP_CAMERA_DEFAULT_FOV 2

//! RFID parameters
#define  SERVO_SP_RFID_MAX_RANGE  0

//! Victim RFID parameters
#define  SERVO_SP_VICTIM_MAX_RANGE    0
#define  SERVO_SP_VICTIM_VERT_FOV    1
#define  SERVO_SP_VICTIM_HORZ_FOV    2

//! Touch sensor parameters
#define  SERVO_SP_TOUCH_DIAMETER 0
 
//! Ground Truth sensor parameters
#define SERVO_SP_GRD_TRUTH_UTM_OFFSET_NORTHING 0
#define SERVO_SP_GRD_TRUTH_UTM_OFFSET_EASTING 1
#define SERVO_SP_GRD_TRUTH_UTM_OFFSET_Z_VALUE 2
#define SERVO_SP_GRD_TRUTH_UTM_ZONE 3
#define SERVO_SP_GRD_TRUTH_UTM_LETTER 4



class SensorSetElem {
public:
  //! Indicates the type of sensor
  ServoSensorType sType;
  char senDataBufName[MOAST_NML_BUFFER_NAME_LEN];
  int sensorID;
  /*! Sensor mounting location
    if 0 then the sensor is hard mounted on vehicle 
    if not zero then the sensor is mounted on mission package
    To obtain the name of the mission package buffer that in which the sensor is 
    mounted, just concatenate SERVO_MIS_JA_<cmd type> + <vehicle id> + '_' + <mountID>
   */
  int mountID;
  int linkID;
  //! Sensor to Mount transform
  PM_POSE sensorToMount;
  //! The desired operating period of the sensor (a report every 'x' sec)
  double desiredPeriod;
  //! Parameters for array.
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(double, parameters, SERVO_SP_SENSOR_PARAMETER_MAX);
};

//SERVO_SP_CMD_XXX
/////////////////////////////////////////////////

class ServoSPCmdInit:public RCS_CMD_MSG {
public:
  ServoSPCmdInit():RCS_CMD_MSG(SERVO_SP_CMD_INIT_TYPE,
				   sizeof(ServoSPCmdInit)) {};
  void update(CMS *);
};

class ServoSPCmdAbort:public RCS_CMD_MSG {
public:
  ServoSPCmdAbort():RCS_CMD_MSG(SERVO_SP_CMD_ABORT_TYPE,
				    sizeof(ServoSPCmdAbort)) {};
  void update(CMS *);
};

class ServoSPCmdHalt:public RCS_CMD_MSG {
public:
  ServoSPCmdHalt():RCS_CMD_MSG(SERVO_SP_CMD_HALT_TYPE,
				   sizeof(ServoSPCmdHalt)) {};
  void update(CMS *);
};

class ServoSPCmdShutdown:public RCS_CMD_MSG {
public:
  ServoSPCmdShutdown():RCS_CMD_MSG(SERVO_SP_CMD_SHUTDOWN_TYPE,
				   sizeof(ServoSPCmdShutdown)) {};
  void update(CMS *);
};

class ServoSPCmdGo:public RCS_CMD_MSG {
public:
  ServoSPCmdGo():RCS_CMD_MSG(SERVO_SP_CMD_GO_TYPE,
				   sizeof(ServoSPCmdGo)) {};
  void update(CMS *);
};

class ServoSPCmdReset:public RCS_CMD_MSG {
  public:
    ServoSPCmdReset():RCS_CMD_MSG(SERVO_SP_CMD_RESET_TYPE,
    sizeof(ServoSPCmdReset)) {};
    void update(CMS *);
};

// SERVO_SP_STAT
//////////////////////////////////////////////

class ServoSPStat : public RCS_STAT_MSG_V2 {
public:
  ServoSPStat() : RCS_STAT_MSG_V2(SERVO_SP_STAT_TYPE, sizeof(ServoSPStat)) {};
  void update(CMS *);
  int error;			// was in stat_msg_v2

  //! incremented once each cycle
  int heartbeat;
  //! Indicates if sensor data is being transmitted
  bool transmitData;
};

/*!
  ServoSPCfgCycleTime sets the controller's cycle time
*/
class ServoSPCfgCycleTime:public RCS_CMD_MSG {
public:
  ServoSPCfgCycleTime():RCS_CMD_MSG(SERVO_SP_CFG_CYCLE_TIME_TYPE,
			       sizeof(ServoSPCfgCycleTime)) {};
  void update(CMS *);

  //! period in seconds
  double cycleTime;
};

/*!
  ServoSPCfgDebug sets the controller's debug level
*/
class ServoSPCfgDebug:public RCS_CMD_MSG {
public:
  ServoSPCfgDebug():RCS_CMD_MSG(SERVO_SP_CFG_DEBUG_TYPE,
			       sizeof(ServoSPCfgDebug)) {};
  void update(CMS *);

  //! debug level
  int debug;
};

// SERVO_SP_SET
//////////////////////////////////////////////////////

/* ServoSPSet is the class that holds the quasi-static sensor processing
   settings,those that change infrequently and in response to sensor 
   processing configuration messages.
 */
class ServoSPSet : public RCS_STAT_MSG {
public:
  ServoSPSet():RCS_STAT_MSG(SERVO_SP_SET_TYPE, sizeof(ServoSPSet)) {};
  void update(CMS *);
  //! cycle time of control loop
  double cycleTime;
  //! debug level
  int debug;
  //! vehicle id
  int vehicleID;
  //! Name of the world
  char worldName[MOAST_WORLD_NAME_LEN];
  PM_POSE startPose;
  /*! 'sConf' is opaque array, indexed using e.g. SERVO_LINESCAN_MAX_RANGE
   *  'sSetElem_length is auto-generated to indicate length of array
   */
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(SensorSetElem, sSetElem, SERVO_SP_SENSOR_MAX);
};

extern int servoSP_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * servoSP_symbol_lookup(long type);
extern const char *servoSP_enum_ServoSensorType_symbol_lookup(long v);

#endif // SERVO_SP_HH

/*
  Modification history:

  $Log: servoSP.hh,v $
  Revision 1.10  2007/04/10 19:10:18  cj_scrapper
  Added victim sensor

  Revision 1.9  2007/03/13 20:01:16  cj_scrapper
  Added Camera to ServoSP settings.  Sensors now report the mission package ID and link ID that it is mounted to.

  Revision 1.8  2007/03/12 20:12:58  cj_scrapper
  Added INS sensor

  Revision 1.7  2007/01/29 15:33:16  proctor
  Added simpleSimInfSkin support, including GPS sensor in the SimWare KB

  Revision 1.6  2007/01/16 17:05:40  cj_scrapper
  *** empty log message ***

  Revision 1.5  2006/12/12 19:20:38  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.4  2006/12/12 18:12:33  cj_scrapper
  New buffers


*/
