/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file trafficCtrl.hh

  \brief NML declarations for traffic control channel

  This channel provides an array of waypoint control structures to be used to
  control multiple traffic vehicles.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.6 $
  $Date: 2008/01/08 20:18:15 $
  \endcode

  \author Stephen Balakirsky
*/

#ifndef trafficCtrl_HH
#define trafficCtrl_HH

#include <rcs.hh>
#include "moastTypes.hh"	// waypoints
#include "moastNmlOffsets.hh"

#define TRAFFIC_CTRL_CMD_NAME "trafficCtrlCmd"
#define TRAFFIC_CTRL_STAT_NAME "trafficCtrlStat"
#define TRAFFIC_CTRL_CFG_NAME "trafficCtrlCfg"
#define TRAFFIC_CTRL_SET_NAME "trafficCtrlSet"

#define TRAFFIC_CTRL_CMD_BASE        (TRAFFIC_CTRL_BASE + 0)
#define TRAFFIC_CTRL_STAT_BASE       (TRAFFIC_CTRL_BASE + 100)
#define TRAFFIC_CTRL_CFG_BASE        (TRAFFIC_CTRL_BASE + 200)
#define TRAFFIC_CTRL_SET_BASE        (TRAFFIC_CTRL_BASE + 300)
#define TRAFFIC_CTRL_ERROR_BASE      (TRAFFIC_CTRL_BASE + 400)

#define TRAFFIC_CTRL_CMD_NOP_TYPE                 (TRAFFIC_CTRL_CMD_BASE + 1)
#define TRAFFIC_CTRL_CMD_INIT_TYPE                (TRAFFIC_CTRL_CMD_BASE + 2)
#define TRAFFIC_CTRL_CMD_ABORT_TYPE               (TRAFFIC_CTRL_CMD_BASE + 3)
#define TRAFFIC_CTRL_CMD_HALT_TYPE                (TRAFFIC_CTRL_CMD_BASE + 4)
#define TRAFFIC_CTRL_CMD_SHUTDOWN_TYPE            (TRAFFIC_CTRL_CMD_BASE + 5)
#define TRAFFIC_CTRL_CMD_MOVE_WAYPOINT_TYPE       (TRAFFIC_CTRL_CMD_BASE + 6)
#define TRAFFIC_CTRL_CMD_STOP_TYPE                (TRAFFIC_CTRL_CMD_BASE + 7)
#define TRAFFIC_CTRL_CMD_MOVE_ARC_SEGMENT_TYPE    (TRAFFIC_CTRL_CMD_BASE + 8)
#define TRAFFIC_CTRL_CMD_ACKERMAN_TYPE            (TRAFFIC_CTRL_CMD_BASE + 9)
#define TRAFFIC_CTRL_CMD_SKID_TYPE                (TRAFFIC_CTRL_CMD_BASE + 10)
#define TRAFFIC_CTRL_CMD_FLIP_SKID_TYPE           (TRAFFIC_CTRL_CMD_BASE + 11)
#define TRAFFIC_CTRL_CMD_OMNI_TYPE                (TRAFFIC_CTRL_CMD_BASE + 12)
#define TRAFFIC_CTRL_CMD_SUBMARINE_TYPE           (TRAFFIC_CTRL_CMD_BASE + 13)
#define TRAFFIC_CTRL_CMD_ROTARY_WING_TYPE         (TRAFFIC_CTRL_CMD_BASE + 14)

#define TRAFFIC_CTRL_STAT_TYPE       (TRAFFIC_CTRL_STAT_BASE + 1)

#define TRAFFIC_CTRL_CFG_NOP_TYPE             (TRAFFIC_CTRL_CFG_BASE + 1)
#define TRAFFIC_CTRL_CFG_CYCLE_TIME_TYPE      (TRAFFIC_CTRL_CFG_BASE + 2)
#define TRAFFIC_CTRL_CFG_DEBUG_TYPE           (TRAFFIC_CTRL_CFG_BASE + 3)
#define TRAFFIC_CTRL_CFG_MOVE_TYPE            (TRAFFIC_CTRL_CFG_BASE + 4)

#define TRAFFIC_CTRL_SET_TYPE        (TRAFFIC_CTRL_SET_BASE + 1)


class TrafficCtrlCmdNop:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdNop():RCS_CMD_MSG
    (TRAFFIC_CTRL_CMD_NOP_TYPE,
     sizeof(TrafficCtrlCmdNop)) {};
  void update(CMS *);
};

class TrafficCtrlCmdInit:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdInit():RCS_CMD_MSG
    (TRAFFIC_CTRL_CMD_INIT_TYPE,
     sizeof(TrafficCtrlCmdInit)) {};
  void update(CMS *);
};

class TrafficCtrlCmdHalt:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdHalt():RCS_CMD_MSG
    (TRAFFIC_CTRL_CMD_HALT_TYPE,
     sizeof(TrafficCtrlCmdHalt)) {};
  void update(CMS *);
};

class TrafficCtrlCmdAbort:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdAbort():RCS_CMD_MSG
    (TRAFFIC_CTRL_CMD_ABORT_TYPE,
     sizeof(TrafficCtrlCmdAbort)) {};
  void update(CMS *);
};

class TrafficCtrlCmdShutdown:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdShutdown():RCS_CMD_MSG
    (TRAFFIC_CTRL_CMD_SHUTDOWN_TYPE,
     sizeof(TrafficCtrlCmdShutdown)) {};
  void update(CMS *);
};


/*!
  The TrafficCtrlCmdMoveWaypoint NML message contains an array of waypoints,
  up to MAX_TRAFFIC_VEHICLES, in relative coordinates.
*/

class TrafficCtrlCmdMoveWaypoint:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdMoveWaypoint():RCS_CMD_MSG
    (TRAFFIC_CTRL_CMD_MOVE_WAYPOINT_TYPE,
     sizeof(TrafficCtrlCmdMoveWaypoint)) {};
  void update(CMS *);
  Waypoint waypoint[MAX_TRAFFIC_VEHICLES][MAX_WAYPOINTS];
  int waypoint_length[MAX_TRAFFIC_VEHICLES];
  int numVehicles;
  
  //  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(Waypoint, vehCmd, MAX_TRAFFIC_VEHICLES);
  // will now get 'int waypoint_length'
};

/*!
  TrafficCtrlCmdMoveArcSegment NML message
*/
class TrafficCtrlCmdMoveArcSegment:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdMoveArcSegment():RCS_CMD_MSG
  (TRAFFIC_CTRL_CMD_MOVE_ARC_SEGMENT_TYPE,
   sizeof(TrafficCtrlCmdMoveArcSegment)) {};
  void update(CMS *);
  ArcSegment arcsegment[MAX_TRAFFIC_VEHICLES][MAX_ARC_SEGMENTS];
  int arcsegment_length[MAX_TRAFFIC_VEHICLES];
  int numVehicles;
};

typedef struct {
  double wLeft;			// left wheel speed, [rad/s]
  double wRight;		// right wheel speed, [rad/s]
  int vehID;
}TrafficCtrlSkid;

class TrafficCtrlCmdSkid:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdSkid():RCS_CMD_MSG(TRAFFIC_CTRL_CMD_SKID_TYPE,
				   sizeof(TrafficCtrlCmdSkid)) {};
  void update(CMS *);
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficCtrlSkid, cmd, MAX_TRAFFIC_VEHICLES);
};

#define TRAFFIC_CTRL_FLIPPER_TYPE_MAX 4
typedef struct TRAFFIC_CTRL_FLIPPER_MSG
{
  int vehID;
  moastFlipperType fType;
  double flipperAngle;
}TrafficCtrlFlipperMsg;

class TrafficCtrlCmdFlipSkid:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdFlipSkid():RCS_CMD_MSG(TRAFFIC_CTRL_CMD_FLIP_SKID_TYPE,
				   sizeof(TrafficCtrlCmdFlipSkid)) {};
  void update(CMS *); 
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficCtrlSkid, cmd, MAX_TRAFFIC_VEHICLES);
  //! Cmds to control the angle of flippers on platform
  //! fCmd_length == the number of flipper messages in array
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficCtrlFlipperMsg, fCmd, TRAFFIC_CTRL_FLIPPER_TYPE_MAX*MAX_TRAFFIC_VEHICLES);
};

typedef struct {
  double propeller;
  double rudder;
  double sternPlane;
  int vehID;
}TrafficCtrlSubmarine;

class TrafficCtrlCmdSubmarine:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdSubmarine():RCS_CMD_MSG(TRAFFIC_CTRL_CMD_SUBMARINE_TYPE,
				   sizeof(TrafficCtrlCmdSubmarine)) {};
  void update(CMS *);
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficCtrlSubmarine, cmd, MAX_TRAFFIC_VEHICLES);
};

typedef struct {
  double velocity;	// linear velocity  [rads/s]
  double steerAngle;    // steering angle of front tire [rads]
  double crabAngle;     // steering angle of rear tire [rads]
  int vehID;
}TrafficCtrlAckerman;

class TrafficCtrlCmdAckerman:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdAckerman():RCS_CMD_MSG(TRAFFIC_CTRL_CMD_ACKERMAN_TYPE,
				   sizeof(TrafficCtrlCmdAckerman)) {};
  void update(CMS *);
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficCtrlAckerman, cmd, MAX_TRAFFIC_VEHICLES);

};

typedef struct {
  double vLinear;     //linear velocity [rads/s]
  double vLateral;    //lateral velocity [rads/s]
  double rot;           //rotational velocity [rads/s]
  int vehID;
}TrafficCtrlOmni;

class TrafficCtrlCmdOmni:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdOmni():RCS_CMD_MSG(TRAFFIC_CTRL_CMD_OMNI_TYPE,
				   sizeof(TrafficCtrlCmdOmni)) {};
  void update(CMS *);
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficCtrlOmni, cmd, MAX_TRAFFIC_VEHICLES);
};

typedef struct {
  double vLinear;     //linear velocity
  double vLateral;    //lateral velocity 
  double vVertical;   //Vertical Velocity
  double vRotation;   //Rotational Velocity
  int vehID;
}TrafficCtrlRotary;

class TrafficCtrlCmdRotaryWing:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdRotaryWing():RCS_CMD_MSG(TRAFFIC_CTRL_CMD_ROTARY_WING_TYPE,
				   sizeof(TrafficCtrlCmdRotaryWing)) {};
  void update(CMS *);
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficCtrlRotary, cmd, MAX_TRAFFIC_VEHICLES);
};

class TrafficCtrlCmdStop:public RCS_CMD_MSG {
public:
  TrafficCtrlCmdStop():RCS_CMD_MSG
    (TRAFFIC_CTRL_CMD_STOP_TYPE,
     sizeof(TrafficCtrlCmdStop)) {};
  void update(CMS *);
};

/*!
  TrafficCtrlStat is the class that holds the dynamic vehicle status,
  things that change frequently and in response to vehicle command
  messages.
 */

class TrafficCtrlStat:public RCS_STAT_MSG_V2 {
public:
  TrafficCtrlStat() : RCS_STAT_MSG_V2(TRAFFIC_CTRL_STAT_TYPE, sizeof(TrafficCtrlStat)) {};
  int error;
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(int, pathIndex, MAX_TRAFFIC_VEHICLES);
  // will now get 'int pathIndex_length'
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(int, primState, MAX_TRAFFIC_VEHICLES);
  // will now get 'int primState_length'
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(int, primStatus, MAX_TRAFFIC_VEHICLES);
  // will now get 'int primStatus_length'

  void update(CMS *);
};

class TrafficCtrlCfgCycleTime:public RCS_CMD_MSG {
public:
  TrafficCtrlCfgCycleTime():RCS_CMD_MSG
    (TRAFFIC_CTRL_CFG_CYCLE_TIME_TYPE,
     sizeof(TrafficCtrlCfgCycleTime)) {};
  void update(CMS *);

  double cycleTime;
};

class TrafficCtrlCfgNop:public RCS_CMD_MSG {
public:
  TrafficCtrlCfgNop():RCS_CMD_MSG
    (TRAFFIC_CTRL_CFG_NOP_TYPE,
     sizeof(TrafficCtrlCfgNop)) {};
  void update(CMS *);
};

class TrafficCtrlCfgMove:public RCS_CMD_MSG {
public:
  TrafficCtrlCfgMove():RCS_CMD_MSG
    (TRAFFIC_CTRL_CFG_MOVE_TYPE,
     sizeof(TrafficCtrlCfgMove)) {};
  void update(CMS *);

  double vmax;
  double amax;
  double wmax;
  double alphamax;
  double vCutoffAngle;
  double wCutoffAngle;
};

#define TRAFFIC_CTRL_DEBUG_MOVE 0x00000001
#define TRAFFIC_CTRL_DEBUG_ALL 0xFFFFFFFF

class TrafficCtrlCfgDebug:public RCS_CMD_MSG {
public:
  TrafficCtrlCfgDebug():RCS_CMD_MSG
    (TRAFFIC_CTRL_CFG_DEBUG_TYPE,
     sizeof(TrafficCtrlCfgDebug)) {};
  void update(CMS *);

  int debug;			// debug mask
};


typedef struct{
  int steerType;
  char robotName[MOAST_WORLD_NAME_LEN];
}RobotInfo;

/*!
  TrafficCtrlSet is the class that holds the quasi-static vehicle settings,
  those that change infrequently and in response to vehicle configuration
  messages.
 */
class TrafficCtrlSet:public RCS_STAT_MSG {
public:
  TrafficCtrlSet():RCS_STAT_MSG(TRAFFIC_CTRL_SET_TYPE, sizeof(TrafficCtrlSet)) {};
  void update(CMS *);

  double cycleTime;
  int debug;			// debug mask
  double vmax;			// max translational speed, [m/s]
  double amax;			// max translational acceleration, [m/s/s]
  double wmax;			// max rotational speed, [rad/s]
  double alphamax;		// max rotational acceleration, [rad/s/s]
  // value of angular deviation above which speed is set to zero,
  // leading to only rotation
  double vCutoffAngle;		// [rad]
  // value of angular deviation above which angular speed is clamped
  // to wmax
  double wCutoffAngle;		// [rad]
  PM_POSE controlPoint;		// pose of control point wrt nav data
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(RobotInfo, robotInfo, MAX_TRAFFIC_VEHICLES);
  // will now get 'int robotInfo_length'
};
extern int trafficCtrl_format(NMLTYPE type, void *buf, CMS * cms);
extern const char *trafficCtrl_symbol_lookup(long type);
#endif // TRAFFIC_CTRL_HH

/*
  Modification history:

  $Log: trafficCtrl.hh,v $
  Revision 1.6  2008/01/08 20:18:15  proctor
  Added control point transform that lets the prim mobility controller act
  on the rotational center of the vehicle, which in general won't coincide
  with the nav data frame

  Revision 1.5  2007/09/11 20:27:43  dr_steveb
  Changes for link monitoring

  Revision 1.4  2007/08/29 18:26:25  dr_steveb
  Added joystick control commands to the traffic cmd buffer and moved some constants to mostTypes

  Revision 1.3  2007/08/23 13:11:55  dr_steveb
  Added arc segment control of traffic (per zeid)

  Revision 1.2  2007/05/16 15:16:21  dr_steveb
  Fixed location of state printing routines to remove compile warnings.

  Revision 1.1  2007/04/25 15:08:57  dr_steveb
  Added traffic simulation files

*/
