/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   visGraph.hh
  \brief  Include file for visibility graph generation

          Algorithm based on paper from Overmars and Welzl (RUU-CS-88-7, March 1988).

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.8 $
  $Date: 2009/01/23 13:38:23 $
  \endcode
  \par CVS Log:
  \ref visGraph_hh "More..."

  \author Stephen Balakirsky
  \date   06/29/05
*/
#ifndef VIS_GRAPH_HH
#define VIS_GRAPH_HH

#include <string.h>
#include "moastTypes.hh"
#include "moastGeom.hh"
#include "nodeInfo.hh"
#include "dijkHeapN.hh"
#include "CImg.h"
#include "amMobJA.hh"

using namespace std;
#include <map>

class PolyProps {
public:
  //! constructor
  PolyProps(){
    id = -1;
    numVerts = 0;
    vertUsed = false;};
  int id;
  int numVerts;
  bool vertUsed;
};

class VisGraphNode:public NodeInfo {
public:
  VisGraphNode(int index_in):NodeInfo(index_in){};
  Location2D nodeLoc;
};

struct ltstr {
  bool operator()(const char * s1, const char * s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

class VisGraphWM {
public:
  //! constructor
  VisGraphWM( int initSize );
  //! destructor
  ~VisGraphWM();
  //! dump an image of the world model
  bool dump(cimg_library::CImg<unsigned char> *imageIn);
  //! return a line from the world model, if fromOriginal is true, the from original array
  //  otherwise from the expanded array
  LineSegment getLine( int id, bool fromOriginal );
  //! get a named point from the hash table
  Location2D getNamedPt( char *name );
  //! return the number of lines in the world model
  int getSize(bool getScaled);
  //! load predefined models into the world model
  int load(char *name, double wmGrow);
  bool isScaledSet();
  bool setLineScaled( int id, LineSegment &lineIn );
  void detectPtsInPoly();
  bool isInPoly( const Location2D &pointToCheck, int myID );
  Location2D getWMSize(){ return wmSize; };
  Location2D getWMLL(){ return wmLL; };
  void setResolution(double resIn){resolution=resIn;};
  void setVehLoc(const Location2D &vehLocIn);
  Location2D getVehLoc(){return *vehLoc;};
  double getResolution(){return resolution;};
  bool ditherPoint( Location2D *point );
  //! computed path
  int *path;
  //! number of points in the computed path
  int pathLength;
  //! the length of the computed path in meters
  double pathDistance;
private:
  //! add a line to the world model (placed in "original" array)
  int add(LineSegment &line, PolyProps properties );
  //! compare the given value to a max and min value
  void compareValues(Location2D &minWM, Location2D &maxWM, const Location2D &pointIn);
  //! compute the expanded array
  int scaleWM(double distance, bool addOrig);
  //! the current number of lines in the world model
  int currentSize;
  //! the current number of scaled lines in the world model
  int currentScaledSize;
  //! a table for storing named goal locations
  map<const char *, void *, ltstr> goalHash;
  //! the maximum number of lines in the world model
  int maxSize;
  //! storage of the line segments
  LineSegment *storageOriginal;
  LineSegment *storageScaled;
  Location2D wmSize;
  Location2D wmLL;
  //! The resolution of the output image in meters/pixal
  double resolution;
  //! The location of the vehicle
  Location2D *vehLoc;
};

class VisGraph {
public:
  //! constructor
  VisGraph( int initWMSize, int retPtSize );
  //! destructor
  ~VisGraph();
  //! add teh am perimeter nodes to the graph
  bool addGraphPoints( PM_CARTESIAN *points, int pointLength, double endTime );
  //! grow the return storage array
  void growRetStorage();
  //! build the entire visibility graph
  bool buildGraph(int numPtsIn, FILE *fp=stdout);
  //! build the graph to a point
  bool buildGraphPoint(Location2D &point, int size, int &numPtsRet, bool force=false );
  //! dump a image of the world model and visibility graph
  bool dump(char *fileName, unsigned int dumpWhat);
  //! get the peripheral node we are going to
  Location2D getAMGoal();
  Location2D getNamedPt(char *name){ return wm->getNamedPt(name);};
  void printAnswer( int pathLength, double cost, int *path);
  void setGoal( Location2D goalIn );
  void setVehLoc( const Location2D vehLocIn ){ wm->setVehLoc(vehLocIn);};
  Location2D getVehLoc(){return wm->getVehLoc();};
  bool getPathPoint(Location2D *pathPoint, int pointNum);
  double getPathDistance(){return wm->pathDistance;};
  /*
  //! get world model extents
  Location2D getExtents(){ return wm->getWMSize();};
  //! get world model lower left
  Location2D getLowerLeft(){return wm->getWMLL();};
  */
  //! retrieve a point from storage
  Location2D getStoredPt( int id );
  //! initialize the search graph
  bool initGraph(double endTime);
  //! load a world model file
  int load( char *name, double wmGrow) { return(wm->load(name, wmGrow)); };
  //! print world model
  void print(FILE *fp=stdout);
  void setOutImage( double resolutionIn );
  int search( bool newSearch, double endTime );
  void setPlanHorizon( double horizonIn ){ planHorizon = horizonIn;};
  double getPlanHorizon(){ return planHorizon; };

private:
  bool computeCosts(int node);
  double computeCost(int node1, int node2);
  //! planning horizon from AM
  double planHorizon;
  VisGraphWM *wm;
  //! storage for visibility points to be returned
  Location2D *retStorage;
  //! storage for the ids of lines that the points came from (came from starting point)
  int *retStorageID;
  int retStorageSize;
  //! the search mechanism
  DijkHeapNSearcher *theSearcher;
  //! the dijk graph structure
  DijkHeapNGraph searcherGraph;
  //! CImg library image structure
  cimg_library::CImg<unsigned char> *cImg;
};

class RotTreeNode {
private:
  //! id of this node
  unsigned int id;
  //! location of node
  PM_CARTESIAN loc;
  //! pointer to left brother
  RotTreeNode *rightBrother;
  //! pointer to left brother
  RotTreeNode *leftBrother;
  //! pointer to father
  RotTreeNode *father;
  //! pointer to rightmost son
  RotTreeNode *rightmostSon;

public:
  //! constructor
  RotTreeNode(PM_CARTESIAN &locIn);
  RotTreeNode();
  //! destructor
  ~RotTreeNode();
  //! remove node from tree
  void remove();
  //! Set the location
  void setLocation( PM_CARTESIAN &locIn );
  //! Adds p as reightmost son
  void setRightMost( RotTreeNode *p );
  //! adds p as left brother
  void setLeftOf( RotTreeNode *p );
  //! returns the right brother
  RotTreeNode *getRightBrother();
  //! returns left brother
  RotTreeNode *getLeftBrother();
  //! returns the father
  RotTreeNode *getFather();
  //! returns the rightmost son
  RotTreeNode *getRightmostSon();
  //! returns location
  PM_CARTESIAN getLocation();
};

class RotStack{
private:
  //! storage space for stack
  RotTreeNode **nodes;
  //! current size of stack
  int stackSize;
public:
  RotStack( int size );
  ~RotStack();
  //! return element off of stack
  RotTreeNode *pop();
  //! push an element on to the stack
  void push( RotTreeNode *in );
  //! is the stack empty?
  bool isEmpty();
  //! return the top of the stack without poping
  RotTreeNode *top();
};
  
class RotTree{
  //private:
public:
  RotTree();
  //! load initial values into tree values
  void load();
  //! create tree
  void createTree();
  //! sort the tree
  void sort();
  //! print the sorted list of nodes
  void printSort();
  //! print the tree
  void printTree();
  //! returns true if r lies to the left of the directed line pq
  bool leftTurn( RotTreeNode *p, RotTreeNode *q, PM_CARTESIAN r, bool special );
  //! node storage
  RotTreeNode *nodes;
  //! the number of nodes in the tree
  int numNodes;
  //! stack space
  RotStack *stack;
  //! has the tree been sorted
  bool isSorted;
};
#endif // VIS_GRAPH_HH

/*! \defgroup visGraph_hh visGraph.hh Log Information
 * $Log: visGraph.hh,v $
 * Revision 1.8  2009/01/23 13:38:23  proctor
 * Fixed comparison function
 *
 * Revision 1.7  2008/12/05 15:58:24  proctor
 * Added copies of the Ulapi interface files; fixed visGraph to use the
 * Standard Template Libraries instead of hcreate etc.;
 *
 * Revision 1.6  2008/01/07 21:49:53  dr_steveb
 * Vehicle controller for the manufacturing competition
 *
 * Revision 1.5  2007/12/20 02:33:13  dr_steveb
 * Added new commands for manufacturing
 *
 * Revision 1.4  2007/12/13 13:17:26  dr_steveb
 * Changes for vehicle level visibility graph planning.
 *
 * Revision 1.3  2007/11/08 13:20:59  dr_steveb
 * Checked in to allow for Tom to compile
 *
 * Revision 1.2  2007/11/07 21:13:50  dr_steveb
 * Checked in to allow for compliation
 *
 * Revision 1.1.1.1  2005/10/12 20:50:32  root
 * initial import
 *
 * Revision 1.4  2005/07/21 18:59:11  stephen
 * Fixed merged conflicts and added visibility graph stuff.
 *
 * Revision 1.3  2005/07/05 12:04:22  stephen
 * Tree now working? in visGraph.
 *
 * Revision 1.2  2005/07/01 19:53:01  stephen
 * Rotation tree now working
 *
 * Revision 1.1  2005/06/30 20:20:23  stephen
 * Initial check-in
 *
 *
 */
