/*
DISCLAIMER:
This software was produced by the National Institute of Standards
and Technology (NIST), an agency of the U.S. government, and by statute is
not subject to copyright in the United States.  Recipients of this software
assume all responsibility associated with its operation, modification,
maintenance, and subsequent redistribution.

See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
\file amMisMain.cc

\brief Vehicle autonomous mobility level mission controller.

\code CVS Status:
$Author: root $
$Revision: 1.1.1.1 $
$Date: 2005/10/12 20:50:32 $
\endcode

\author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>
#include "moastTypes.hh"
#include "moastNmlOffsets.hh"
#include "primMisJA.hh"
#include "amMisJA.hh"

static RCS_CMD_CHANNEL *primMisJACmdBuf = NULL;
static RCS_STAT_CHANNEL *primMisJAStatBuf = NULL;

static RCS_CMD_CHANNEL *amMisJACmdBuf = NULL;
static RCS_STAT_CHANNEL *amMisJAStatBuf = NULL;
static RCS_CMD_CHANNEL *amMisJACfgBuf = NULL;
static RCS_STAT_CHANNEL *amMisJASetBuf = NULL;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "amMisJA"

#define DEFAULT_CYCLE_TIME 0.100

static int cleanupPrimMisJANmlBuffers(void)
{
	if (NULL != primMisJACmdBuf)
		delete primMisJACmdBuf;
	if (NULL != primMisJAStatBuf)
		delete primMisJAStatBuf;

	primMisJACmdBuf = NULL;
	primMisJAStatBuf = NULL;

	return 0;
}

static int initPrimMisJANmlBuffers(char *config_file, int bufnum)
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	(void) cleanupPrimMisJANmlBuffers();

	sprintf(chanName, "%s%d", PRIM_MIS_JA_CMD_NAME, bufnum);
	primMisJACmdBuf =
		new RCS_CMD_CHANNEL(primMisJA_format, chanName,
		THIS_PROCESS, config_file);
	if (!primMisJACmdBuf->valid()) {
		delete primMisJACmdBuf;
		primMisJACmdBuf = 0;
		return 1;
	}

	sprintf(chanName, "%s%d", PRIM_MIS_JA_STAT_NAME, bufnum);
	primMisJAStatBuf =
		new RCS_STAT_CHANNEL(primMisJA_format, chanName,
		THIS_PROCESS, config_file);
	if (!primMisJAStatBuf->valid()) {
		delete primMisJAStatBuf;
		primMisJAStatBuf = 0;
		return 1;
	}

	return 0;
}

static int cleanupAmMisJANmlBuffers(void)
{
	if (NULL != amMisJACmdBuf)
		delete amMisJACmdBuf;
	if (NULL != amMisJAStatBuf)
		delete amMisJAStatBuf;
	if (NULL != amMisJACfgBuf)
		delete amMisJACfgBuf;
	if (NULL != amMisJASetBuf)
		delete amMisJASetBuf;

	amMisJACmdBuf = NULL;
	amMisJAStatBuf = NULL;
	amMisJACfgBuf = NULL;
	amMisJASetBuf = NULL;

	return 0;
}

static int initAmMisJANmlBuffers(char *config_file, int bufnum)
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	(void) cleanupAmMisJANmlBuffers();

	sprintf(chanName, "%s%d", AM_MIS_JA_CMD_NAME, bufnum);
	amMisJACmdBuf =
		new RCS_CMD_CHANNEL(amMisJA_format, chanName,
		THIS_PROCESS, config_file);
	if (!amMisJACmdBuf->valid()) {
		delete amMisJACmdBuf;
		amMisJACmdBuf = 0;
		return 1;
	}

	sprintf(chanName, "%s%d", AM_MIS_JA_STAT_NAME, bufnum);
	amMisJAStatBuf =
		new RCS_STAT_CHANNEL(amMisJA_format, chanName,
		THIS_PROCESS, config_file);
	if (!amMisJAStatBuf->valid()) {
		delete amMisJAStatBuf;
		amMisJAStatBuf = 0;
		return 1;
	}

	sprintf(chanName, "%s%d", AM_MIS_JA_CFG_NAME, bufnum);
	amMisJACfgBuf =
		new RCS_CMD_CHANNEL(amMisJA_format, chanName,
		THIS_PROCESS, config_file);
	if (!amMisJACfgBuf->valid()) {
		delete amMisJACfgBuf;
		amMisJACfgBuf = 0;
		return 1;
	}

	sprintf(chanName, "%s%d", AM_MIS_JA_SET_NAME, bufnum);
	amMisJASetBuf =
		new RCS_STAT_CHANNEL(amMisJA_format, chanName,
		THIS_PROCESS, config_file);
	if (!amMisJASetBuf->valid()) {
		delete amMisJASetBuf;
		amMisJASetBuf = 0;
		return 1;
	}

	return 0;
}

void DO_CMD_INIT(AmMisJAStat * stat,
				 RCS_CMD_CHANNEL * sub_cmd_buf,
				 RCS_STAT_CHANNEL * sub_stat_buf)
{
	PrimMisJACmdInit initMsg;
	PrimMisJAStat *sub_stat_ptr;

	sub_stat_ptr = (PrimMisJAStat *) (sub_stat_buf->get_address());

	if (state_match(stat, NEW_COMMAND)) {
		state_new(stat);
		// send a command to subordinate
		sub_stat_buf->read();
		initMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
		sub_cmd_buf->write(&initMsg);
		status_next(stat, RCS_DONE);
		state_next(stat, S0);
	} else if (state_match(stat, S0)) {
		// wait for done
		sub_stat_buf->read();
		if (sub_stat_ptr->status != RCS_EXEC) {
			// it's either done or error, so make this our status
			status_next(stat, sub_stat_ptr->status);
			state_next(stat, S1);
		}
		// else wait here
	} else {			// S1
		state_default(stat);
	}
}

void DO_CMD_HALT(AmMisJAStat * stat,
				 RCS_CMD_CHANNEL * sub_cmd_buf,
				 RCS_STAT_CHANNEL * sub_stat_buf)
{
	PrimMisJACmdHalt haltMsg;
	PrimMisJAStat *sub_stat_ptr;

	sub_stat_ptr = (PrimMisJAStat *) (sub_stat_buf->get_address());

	if (state_match(stat, NEW_COMMAND)) {
		state_new(stat);
		// send a command to subordinate
		sub_stat_buf->read();
		haltMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
		sub_cmd_buf->write(&haltMsg);
		status_next(stat, RCS_DONE);
		state_next(stat, S0);
	} else if (state_match(stat, S0)) {
		// wait for done
		sub_stat_buf->read();
		if (sub_stat_ptr->status != RCS_EXEC) {
			// it's either done or error, so make this our status
			status_next(stat, sub_stat_ptr->status);
			state_next(stat, S1);
		}
		// else wait here
	} else {			// S1
		state_default(stat);
	}
}

void DO_CMD_ABORT(AmMisJAStat * stat,
				  RCS_CMD_CHANNEL * sub_cmd_buf,
				  RCS_STAT_CHANNEL * sub_stat_buf)
{
	PrimMisJACmdAbort abortMsg;
	PrimMisJAStat *sub_stat_ptr;

	sub_stat_ptr = (PrimMisJAStat *) (sub_stat_buf->get_address());

	if (state_match(stat, NEW_COMMAND)) {
		state_new(stat);
		// send a command to subordinate
		sub_stat_buf->read();
		abortMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
		sub_cmd_buf->write(&abortMsg);
		status_next(stat, RCS_DONE);
		state_next(stat, S0);
	} else if (state_match(stat, S0)) {
		// wait for done
		sub_stat_buf->read();
		if (sub_stat_ptr->status != RCS_EXEC) {
			// it's either done or error, so make this our status
			status_next(stat, sub_stat_ptr->status);
			state_next(stat, S1);
		}
		// else wait here
	} else {			// S1
		state_default(stat);
	}
}

void DO_CMD_SHUTDOWN(AmMisJAStat * stat,
					 RCS_CMD_CHANNEL * sub_cmd_buf,
					 RCS_STAT_CHANNEL * sub_stat_buf)
{
	PrimMisJACmdShutdown shutdownMsg;
	PrimMisJAStat *sub_stat_ptr;

	sub_stat_ptr = (PrimMisJAStat *) (sub_stat_buf->get_address());

	if (state_match(stat, NEW_COMMAND)) {
		state_new(stat);
		// send a command to subordinate
		sub_stat_buf->read();
		shutdownMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
		sub_cmd_buf->write(&shutdownMsg);
		status_next(stat, RCS_DONE);
		state_next(stat, S0);
	} else if (state_match(stat, S0)) {
		// wait for done
		sub_stat_buf->read();
		if (sub_stat_ptr->status != RCS_EXEC) {
			// it's either done or error, so make this our status
			status_next(stat, sub_stat_ptr->status);
			state_next(stat, S1);
		}
		// else wait here
	} else {			// S1
		state_default(stat);
	}
}

void DO_CFG_CYCLE_TIME(AmMisJACfgCycleTime * cfg,
					   AmMisJASet * set)
{
	if (state_match(set, NEW_COMMAND)) {
		state_new(set);
		if (cfg->cycleTime <= 0.0) {
			status_next(set, RCS_ERROR);
		} else {
			set->cycleTime = cfg->cycleTime;
			status_next(set, RCS_DONE);
		}
		state_next(set, S0);
	} else {
		state_default(set);
	}
}

void DO_CFG_DEBUG(AmMisJACfgDebug * cfg,
				  AmMisJASet * set)
{
	if (state_match(set, NEW_COMMAND)) {
		state_new(set);
		set->debug = cfg->debug;
		status_next(set, RCS_DONE);
		state_next(set, S0);
	} else {
		state_default(set);
	}
}

static int done = 0;
static void quit(int sig)
{
	done = 1;
}

int main(int argc, char *argv[])
{
	char nml_file_default[] = DEFAULT_NML_FILE;
	char *nml_file_env;
	char *NML_FILE = NULL;
	char ini_file_default[] = DEFAULT_INI_FILE;
	char *ini_file_env;
	char *INI_FILE = NULL;
	int bufnum = 1;
	int retval;
	NMLTYPE cmd_type, cfg_type;
	int serial_number;
	AmMisJAStat amMisJAStat;
	AmMisJASet amMisJASet;
	AmMisJACmdInit initCmd;
	AmMisJACfgCycleTime cycleTimeCfg;

	while (1) {
		char option = getopt(argc, argv, "n:i:d:");
		if (option == -1)
			break;

		switch (option) {
	case 'n':
		NML_FILE = optarg;
		rcs_print("%s: using specified NML file %s\n", argv[0], NML_FILE);
		break;
	case 'i':
		INI_FILE = optarg;
		rcs_print("%s: using specified INI file %s\n", argv[0], INI_FILE);
		break;
	case 'd':
		if (1 != sscanf(optarg, "%i", &bufnum)) {
			rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
			return 1;
		}
		rcs_print("%s: using specified buffer suffix %d\n", argv[0], bufnum);
		break;
	default:
		rcs_print_error("%s: bad argument\n", argv[0]);
		return 1;
		break;
		}
	}

	// get the config file name from arg, if provided, next environment,
	// if there, else default
	if (NULL == NML_FILE) {
		nml_file_env = getenv("CONFIG_NML");
		if (NULL != nml_file_env) {
			NML_FILE = nml_file_env;
			rcs_print("%s: using environment NML file %s\n", argv[0], NML_FILE);
		} else {
			NML_FILE = nml_file_default;
			rcs_print("%s: using default NML file %s\n", argv[0], NML_FILE);
		}
	}

	// ditto for the .ini file
	if (NULL == INI_FILE) {
		ini_file_env = getenv("CONFIG_INI");
		if (NULL != ini_file_env) {
			INI_FILE = ini_file_env;
			rcs_print("%s: using environment INI file %s\n", argv[0], INI_FILE);
		} else {
			INI_FILE = ini_file_default;
			rcs_print("%s: using default INI file %s\n", argv[0], INI_FILE);
		}
	}

	// connect to the vehicle prim level buffers
	retval = initPrimMisJANmlBuffers(NML_FILE, bufnum);
	if (retval) {
		rcs_print_error("%s: can't allocate primMisJA NML buffers\n", argv[0]);
		return 1;
	}

	// connect to my buffers
	retval = initAmMisJANmlBuffers(NML_FILE, bufnum);
	if (retval) {
		rcs_print_error("%s: can't allocate amMisJA NML buffers\n", argv[0]);
		(void) cleanupPrimMisJANmlBuffers();
		return 1;
	}

	// initialize and set the status up so that init command goes through
	amMisJAStat.echo_serial_number = 0;
	amMisJAStat.heartbeat = 0;

	// initialize and set the settings up so that the cycle time goes through
	amMisJASet.echo_serial_number = 0;
	// others will come from .ini file shortly...

	// stuff an init command
	initCmd.serial_number = 1;
	amMisJACmdBuf->write(&initCmd);

	// stuff a cycle time config
	cycleTimeCfg.serial_number = 1;
	cycleTimeCfg.cycleTime = amMisJASet.cycleTime;
	amMisJACfgBuf->write(&cycleTimeCfg);

	done = 0;
	signal(SIGINT, quit);
	while (!done) {
		// some shortcut pointers to various commands
		AmMisJACfgCycleTime * cfgCycleTimePtr;
		AmMisJACfgDebug * cfgDebugPtr;

		// read command
		cmd_type = amMisJACmdBuf->read();
		serial_number = amMisJACmdBuf->get_address()->serial_number;
		switch (cmd_type) {

	case 0:
		// no new command
		break;

	case -1:
		// comm error
		rcs_print_error("%s: NML command read error\n", argv[0]);
		break;

	case AM_MIS_JA_CMD_INIT_TYPE:
	case AM_MIS_JA_CMD_HALT_TYPE:
	case AM_MIS_JA_CMD_ABORT_TYPE:
	case AM_MIS_JA_CMD_SHUTDOWN_TYPE:
		amMisJAStat.command_type = cmd_type;
		if (serial_number != amMisJAStat.echo_serial_number) {
			amMisJAStat.echo_serial_number = serial_number;
			amMisJAStat.state = NEW_COMMAND;
			amMisJAStat.status = RCS_EXEC;
		}
		break;

	default:
		rcs_print_error("%s: unknown command: %d\n", argv[0], (int) cmd_type);
		break;
		}				// switch (cmd_type)

		// read config
		cfg_type = amMisJACfgBuf->read();
		serial_number = amMisJACfgBuf->get_address()->serial_number;
		switch (cfg_type) {

	case 0:
		// no new command
		break;

	case -1:
		// comm error
		rcs_print_error("%s: NML config read error\n", argv[0]);
		break;

	case AM_MIS_JA_CFG_CYCLE_TIME_TYPE:
		amMisJASet.command_type = cfg_type;
		if (serial_number != amMisJASet.echo_serial_number) {
			amMisJASet.echo_serial_number = serial_number;
			amMisJASet.state = NEW_COMMAND;
			amMisJASet.status = RCS_EXEC;
		}
		cfgCycleTimePtr =
			(AmMisJACfgCycleTime *) amMisJACfgBuf->
			get_address();
		break;

	case AM_MIS_JA_CFG_DEBUG_TYPE:
		amMisJASet.command_type = cfg_type;
		if (serial_number != amMisJASet.echo_serial_number) {
			amMisJASet.echo_serial_number = serial_number;
			amMisJASet.state = NEW_COMMAND;
			amMisJASet.status = RCS_EXEC;
		}
		cfgDebugPtr =
			(AmMisJACfgDebug *) amMisJACfgBuf->get_address();
		break;

	default:
		rcs_print_error("%s: unknown config: %d\n", argv[0], (int) cfg_type);
		break;
		}				// switch (cfg_type)

		// handle command state tables
		switch (amMisJAStat.command_type) {

	case AM_MIS_JA_CMD_INIT_TYPE:
		DO_CMD_INIT(&amMisJAStat,
			primMisJACmdBuf, primMisJAStatBuf);
		break;

	case AM_MIS_JA_CMD_HALT_TYPE:
		DO_CMD_HALT(&amMisJAStat,
			primMisJACmdBuf, primMisJAStatBuf);
		break;

	case AM_MIS_JA_CMD_ABORT_TYPE:
		DO_CMD_ABORT(&amMisJAStat,
			primMisJACmdBuf, primMisJAStatBuf);
		break;

	case AM_MIS_JA_CMD_SHUTDOWN_TYPE:
		DO_CMD_SHUTDOWN(&amMisJAStat,
			primMisJACmdBuf, primMisJAStatBuf);
		break;

	default:
		// no command to handle
		break;
		}				// switch (cmd)

		// handle config state tables
		switch (amMisJASet.command_type) {

	case AM_MIS_JA_CFG_CYCLE_TIME_TYPE:
		DO_CFG_CYCLE_TIME(cfgCycleTimePtr, &amMisJASet);
		break;

	case AM_MIS_JA_CFG_DEBUG_TYPE:
		DO_CFG_DEBUG(cfgDebugPtr, &amMisJASet);
		break;

	default:
		// no config to handle
		break;
		}				// switch (cfg)

		amMisJAStat.heartbeat++;

		// update status and settings
		amMisJAStatBuf->write(&amMisJAStat);
		amMisJASetBuf->write(&amMisJASet);

		esleep(amMisJASet.cycleTime);
	}
	while (!done);

	(void) cleanupAmMisJANmlBuffers();
	(void) cleanupPrimMisJANmlBuffers();

	rcs_print("amMisJA done\n");

	return 0;
}

/*
Modification history:

$Log: amMisMain.cc,v $
Revision 1.1.1.1  2005/10/12 20:50:32  root
initial import

Revision 1.4  2005/09/30 18:55:38  proctor
Cleaned up unused variables, forgotten return values

Revision 1.3  2005/06/08 19:28:44  proctor
Moved state_match, etc. macros into moastTypes.hh

Revision 1.2  2005/06/08 14:30:41  proctor
Added MisJA name changes

Revision 1.1  2005/06/07 21:45:24  proctor
Initial checkin

*/

