/*
DISCLAIMER:
This software was produced by the National Institute of Standards
and Technology (NIST), an agency of the U.S. government, and by statute is
not subject to copyright in the United States.  Recipients of this software
assume all responsibility associated with its operation, modification,
maintenance, and subsequent redistribution.

See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
\file amSPMain.cc

\brief Vehicle amitive level sensor processing.

\code CVS Status:
$Author: dr_steveb $
$Revision: 1.19 $
$Date: 2009/04/27 16:09:01 $
\endcode

\author Tom Kramer
\date   2005-10-05
The original author was Adam Beytin
*/

/*

The intended method of terminating this process is to use "kill -INT"
to send it SIGINT, which (because of the "signal(SIGINT, quit);" line
in the source code) will cause the "quit" function to run, setting
the "done" global variable to 1. That will be caught by one of the
two "while (!done)" loops and result in an orderly exit.

This file does not use any of the map's internal variables; they are
used only in amSPMap.cc

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#include <math.h>               // sqrt()
//#include <fcntl.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>
#include <posemath.h>
#include <string>
#include <sstream>
#include "moastTypes.hh"
#include "moastNmlOffsets.hh"
#include "primSP.hh"
#include "servoSP.hh"
#include "amSPMap.hh"
#include "amSP.hh"
#include "navDataExt.hh"
#include "sensorData.hh"
#include "xml_robot_info.hh"
#include <iostream>
#include <fstream>
#include "robocupmap.h"

/***********************************************************************/


#define DEFAULT_CYCLE_TIME 0.100
#define DEFAULT_INI_FILE "../etc/moast.ini"
#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_MAP_IMAGE_FILE_NAME "AmSPMapImage"
#define DEFAULT_MAP_RANGE 500
#define DEFAULT_MAP_RES 0.100
#define DEFAULT_OUT_RANGE 50
#define UNSET_MAP_RANGE 0
#define UNSET_MAP_RES 0.0
#define UNSET_OUT_RANGE 0
#define THIS_PROCESS "amSP"
#define DEG(x) (57.295779513082323 * (x))
#define ABS(x)  (((x) < 0) ? -(x) : (x))

/***********************************************************************/

//! INS VARIABLES
fstream INSOutputFile;
static const float INSSLOPE_RANGE = 2.0; //+ or - from the ground reading
static const int INSSLOPE_RESETCOUNT = 3; //reset the slope counter after a # of non slope readings
static const int INSSLOPE_GROUNDCHANGE = 12; //change the ground reading after a # of similar readings

static float groundReading = 0.0f;
static float potentialGround = 0.0f;
static int sameReadingCount = 0;
static int slopeCounter = 0;

// NML Channels

//! prim SP channels
static RCS_CMD_CHANNEL **primSPCmdBuf = NULL;
static RCS_STAT_CHANNEL **primSPStatBuf = NULL;

//! am SP channels
static RCS_CMD_CHANNEL *amSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *amSPStatBuf = NULL;
static RCS_CMD_CHANNEL *amSPCfgBuf = NULL;
static RCS_STAT_CHANNEL *amSPSetBuf = NULL;


//! navigation data and sensor data channels
static NML *navDataExtBuf = NULL;
static NML **sensorDataListBuf = NULL;
static NML *sensorData2DBuf = NULL;
static NML *servoSPNavBuf = NULL;
/***********************************************************************/

// Other global variables

/*

processName is the actual name of the process, while THIS_PROCESS is
the name used in the NML config file and, hence, required by NML.

*/
static int bufnum = 1;                      // my default buffernumber
static int numPrimSPChan = 0;               // the number of prim sensors to connect to
static int *primSPBufNums = NULL;           // array of buffer numbers to connect to
static double argMapRes = UNSET_MAP_RES;    // map res in command args
static int argMapRange = UNSET_MAP_RANGE;   // map range in command args
static int argOutRange = UNSET_OUT_RANGE;   // outRange in command args
static SensorData2D ** data = NULL;          // For outgoing data
static int done = 0;                        // Done flag, set to 1 when done
static char * iniFileName = NULL;           // name of ini file
static AmSPMap ** map = NULL;               // internal map of the world, one map per prim channel
// name of ppm image file or pipe
static char * mapImageFileName = DEFAULT_MAP_IMAGE_FILE_NAME;
static int *mapRange = NULL;                // Range for internal map
//static int mapRange = DEFAULT_MAP_RANGE;    // Range for internal map
static double mapRes = DEFAULT_MAP_RES;     // Resolution for maps
static int newCycleTime = 0;                // 1=cycle time is new, 0=not
static char * nmlFile = NULL;               // name of NML file
static int outRange = DEFAULT_OUT_RANGE;    // num cells on side of output map
static int primCmdNumber=18;                   // number of latest command to prim
static char * processName;                  // name of this process
static AmSPSet set;                         // config command status message
static AmSPStat stat;                       // task command status message
static RCS_TIMER * timer = NULL;            // timer for cycle time
static int verbose = 1;                     // 1=print copious messages, 0=not
static int MapChangesCounter = 0;
static float vehXPos = 0;
static float vehYPos = 0;
static float vehZPos = 0;

/***********************************************************************/

static int checkMapValues();
static void cleanupAmSPNmlBuffers();
static void cleanupNmlBuffers();
static void cleanupPrimSPNmlBuffers();
static void readCfg();
static void doCfg();
static void doCfgCycleTime();
static void doCfgDebug();
static void doCfgDumpWM();
static void doCfgNop();
static void doCmd();
static void doCmdAbort(RCS_CMD_CHANNEL ** subCmdBuf,
					   RCS_STAT_CHANNEL ** subStatBuf);
static void doCmdGo(RCS_CMD_CHANNEL ** subCmdBuf,
					RCS_STAT_CHANNEL ** subStatBuf);
static void doCmdHalt(RCS_CMD_CHANNEL ** subCmdBuf,
					  RCS_STAT_CHANNEL ** subStatBuf);
static void doCmdInit(RCS_CMD_CHANNEL ** subCmdBuf,
					  RCS_STAT_CHANNEL ** subStatBuf);
static void doCmdNop();
static void doCmdShutdown(RCS_CMD_CHANNEL ** subCmdBuf,
						  RCS_STAT_CHANNEL ** subStatBuf);
static void doPublishXML(NML &nml);
static void getFileNames();
static int getOptions(int argc, char * argv[]);
static void getSensorData();
static int iniLoad();
static void iniLoadOneDouble(
							 INIFILE * iniFile, const char * parameter, const char * heading,
							 double * setMe, int * retval);
static void iniLoadOneInt(
						  INIFILE * iniFile, const char * parameter, const char * heading,
						  int * setMe, int * retval);
static void iniLoadOneString(
							 INIFILE * iniFile, const char * parameter, const char * heading,
							 char ** setMe, int * retval);
static int initAmSPNmlBuffers(char *config_file);
static int initBuffers();
static int initData();
static int initMap();
static void initParameters();
static int initPrimSPNmlBuffers(char *config_file);
static int initServoSPNmlBuffers(char *config_file);
static void quit(int sig);
static void readCmd();
static void readPrimSPStat();
static void writeOutputs();

/***********************************************************************/
/***********************************************************************/

//! fuseData fuses the data from all of the primSP buffers for a single output
/*
Returned Value: SensorData2D structure of fused data

Called By:
writeOutputs
*/
#define mymax(a,b) ((a) > (b) ? (a) : (b))
SensorData2D *fuseData()
{
	static SensorData2D *retValue = NULL;
	int primCount, count;

	if( retValue == NULL )
	{
		retValue = new SensorData2D();

		// check for consistancy of data 
		count = data[0]->featArray_length;
		for( primCount=1; primCount<numPrimSPChan; primCount++ )
		{
			if( data[primCount]->featArray_length != count )
				printf( "Error on size of output map\n" );
			if( data[primCount]->lowerLeft != data[0]->lowerLeft )
				printf( "Error on lower left of output map\n" );
		}
	}
	retValue->isGlobal = data[0]->isGlobal;
	retValue->lowerLeft = data[0]->lowerLeft;
	retValue->incrementX = data[0]->incrementX;
	retValue->incrementY = data[0]->incrementY;
	retValue->x = data[0]->x;
	retValue->y = data[0]->y;
	retValue->featArray_length = data[0]->featArray_length;

	for( count=0; count<retValue->featArray_length; count++ ) // init data
		retValue->featArray[count] = data[0]->featArray[count];

	if( numPrimSPChan > 1 )
	{
		for( count=0; count<retValue->featArray_length; count++ )
		{
			for( primCount=1; primCount<numPrimSPChan; primCount++ )
				if( (retValue->featArray[count]==AMSP_NOT_OBSTACLE) && 
					(data[primCount]->featArray[count]==AMSP_UNSEEN));
				else
					retValue->featArray[count] = mymax( retValue->featArray[count], data[primCount]->featArray[count]); 
		}
	}
	return retValue;
}


/***********************************************************************/

//! checkMapValues checks that map parameters have reasonable values
/*
Returned Value: int
If mapRes, mapRange, or outRange is not positive or if outRange is
greater than mapRange, this returns 1.
Otherwise, it returns 0.

Called By: main

*/

static int checkMapValues() /* NO ARGUMENTS */
{
	int maxRange = -1;

	if (mapRes <= 0)
	{
		rcs_print_error("%s: mapRes = %f but must be positive\n",
			processName, mapRes);
		return 1;
	}
	for( int count=0; count<numPrimSPChan; count++ )
	{
		if (mapRange[count] <= 0)
		{
			rcs_print_error("%s: mapRange = %d but must be positive\n",
				processName, mapRange[count]);
			return 1;
		}
		if( mapRange[count] > maxRange )
			maxRange = mapRange[count];
	}
	if (outRange <= 0)
	{
		rcs_print_error("%s: outRange = %d but must be positive\n",
			processName, outRange);
		return 1;
	}
	if (outRange > maxRange)
	{
		rcs_print_error("%s: output map range (%d) is greater than "
			"mapRange (%d) but must not be\n",
			processName, outRange, maxRange);
		return 1;
	}
	return 0;
}

/***********************************************************************/

/*! cleanupAmSPNmlBuffers cleans up buffers
(1) between AM SP and its superior (AM Mob or vehicle SP or shell)
(2) for input data
(3) for output data
(4) for nav data */
/*
Returned Value: 

Called By:
cleanupNmlBuffers
initAmSPNmlBuffers

*/

static void cleanupAmSPNmlBuffers() /* NO ARGUMENTS */
{
	if (NULL != amSPCmdBuf)
		delete amSPCmdBuf;
	if (NULL != amSPStatBuf)
		delete amSPStatBuf;
	if (NULL != amSPCfgBuf)
		delete amSPCfgBuf;
	if (NULL != amSPSetBuf)
		delete amSPSetBuf;
	if (NULL != navDataExtBuf)
		delete navDataExtBuf;
	if (NULL != sensorData2DBuf)
		delete sensorData2DBuf;
	for( int count=0; count<numPrimSPChan; count++ )
	{
		if( data[count] != NULL )
			delete data[count];
	}
	if (NULL != data)
		delete [] data;

	amSPCmdBuf = NULL;
	amSPStatBuf = NULL;
	amSPCfgBuf = NULL;
	amSPSetBuf = NULL;
	navDataExtBuf = NULL;
	sensorData2DBuf = NULL;
	data = NULL;
}

/***********************************************************************/

//! cleanupNmlBuffers cleans up all the NML buffers
/*
Returned Value: none

Called By:
initBuffers
main 

*/

static void cleanupNmlBuffers() /* NO ARGUMENTS */
{
	cleanupAmSPNmlBuffers();
	cleanupPrimSPNmlBuffers();
}

/***********************************************************************/

//! cleanupPrimSPNmlBuffers cleans up buffers between AM SP and Prim SP
/*
Returned Value: none

Called By:
cleanupNmlBuffers
initBuffers
initPrimSPNmlBuffers

*/

static void cleanupPrimSPNmlBuffers() /* NO ARGUMENTS */
{
	for( int i=0; i<numPrimSPChan; i++ )
	{
		if( primSPCmdBuf != NULL )
			if (NULL != primSPCmdBuf[i])
				delete primSPCmdBuf[i];
		if( primSPStatBuf != NULL )
			if (NULL != primSPStatBuf[i])
				delete primSPStatBuf[i];
		if( sensorDataListBuf != NULL )
			if (NULL != sensorDataListBuf[i])
				delete sensorDataListBuf[i];
	}
	if (NULL != primSPCmdBuf )
		delete [] primSPCmdBuf;
	if (NULL != primSPStatBuf )
		delete [] primSPStatBuf;
	if (NULL != sensorDataListBuf)
		delete [] sensorDataListBuf;
	primSPCmdBuf = NULL;
	primSPStatBuf = NULL;
	sensorDataListBuf = NULL;
}

/***********************************************************************/

//! doCfg executes config commands
/*
Returned Value: none

Called By: main

*/

static void doCfg() /* NO ARGUMENTS */
{
	if (stat.admin_state == ADMIN_SHUT_DOWN)
	{ // do nothing if shut down
		return;
	}
	switch (set.command_type)
	{  
	case AM_SP_CFG_CYCLE_TIME_TYPE:
		doCfgCycleTime();
		break;

	case AM_SP_CFG_DEBUG_TYPE:
		doCfgDebug();
		break;

	case AM_SP_CFG_DUMP_WM_TYPE:
		doCfgDumpWM();
		break;

	case AM_SP_CFG_NOP_TYPE:
		doCfgNop();
		break;

	default:
		// no config to handle
		break;
	}  // switch (cfg)
}

/***********************************************************************/

//! doCfgCycleTime sets the configuration cycle time and the timer
/*
Returned Value: none

Called By: doCfg

The old timer is deleted, and a new one is made.

*/

static void doCfgCycleTime() /* NO ARGUMENTS */
{
	AmSPCfgCycleTime * cfg;

	cfg = (AmSPCfgCycleTime *)(amSPCfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		if (cfg->cycleTime <= 0.0)
		{
			status_next(&set, RCS_ERROR);
			rcs_print_error("%s: cycleTime must be positive\n", processName);
		}
		else
		{
			set.cycleTime = cfg->cycleTime;
			if (timer)
				delete timer;
			timer = new RCS_TIMER(set.cycleTime);
			status_next(&set, RCS_DONE);
			newCycleTime = 1;
			if (verbose)
				rcs_print("cycleTime set to %f \n", set.cycleTime);
		}
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCfgDebug sets debugging
/*
Returned Value: none

Called By:  doCfg

*/

static void doCfgDebug() /* NO ARGUMENTS */
{
	AmSPCfgDebug * cfg;

	cfg = (AmSPCfgDebug *)(amSPCfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		set.debug = cfg->debug;
		status_next(&set, RCS_DONE);
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCfgDumpWM sets dumping of world model
/*
Returned Value: none

Called By:  doCfg

If AM SP is not initialized, map will be a NULL pointer and
this will set status to RCS_ERROR.

*/

static void doCfgDumpWM() /* NO ARGUMENTS */
{
	AmSPCfgDumpWM * cfg;

	cfg = (AmSPCfgDumpWM *)(amSPCfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		state_next(&set, S0);
		for( int count=0; count<numPrimSPChan; count++ )
		{
			if (map[count])
			{
				map[count]->createDisplay(cfg->level);
			}
			else
			{
				status_next(&set, RCS_ERROR);
				rcs_print_error("%s: map is NULL; cannot change dumping\n",
					processName);
				return;
			}
		}
		set.dumpLevel = cfg->level;
		status_next(&set, RCS_DONE);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCfgNop does nothing
/*
Returned Value: none

Called By:  doCfg

*/

static void doCfgNop() /* NO ARGUMENTS */
{
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		status_next(&set, RCS_DONE);
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCmd executes task commands
/*
Returned Value: none

Called By: main

*/

static void doCmd() /* NO ARGUMENTS */
{
	if (stat.admin_state == ADMIN_SHUT_DOWN)
	{ // do nothing if shut down
		return;
	}
	switch (stat.command_type)
	{  
	case AM_SP_CMD_INIT_TYPE:
		doCmdInit(primSPCmdBuf, primSPStatBuf);
		break;

	case AM_SP_CMD_HALT_TYPE:
		doCmdHalt(primSPCmdBuf, primSPStatBuf);
		break;

	case AM_SP_CMD_ABORT_TYPE:
		doCmdAbort(primSPCmdBuf, primSPStatBuf);
		break;

	case AM_SP_CMD_SHUTDOWN_TYPE:
		doCmdShutdown(primSPCmdBuf, primSPStatBuf);
		break;

	case AM_SP_CMD_GO_TYPE:
		doCmdGo(primSPCmdBuf, primSPStatBuf);
		break;

	case AM_SP_CMD_NOP_TYPE:
		doCmdNop();
		break;

	default:
		// no command to handle
		break;
	}  // switch (cmd)
}

/***********************************************************************/

//! doCmdAbort executes an abort command
/*
Returned Value: none

Called By: doCmd

When the command is new, this send an abort command to Prim SP.

If the command is not new and prim is still executing the abort command
it was sent, this waits.

When the prim abort command is finished executing successfully, this
sets stat.admin_state to UNINITIALIZED and reports the AM abort
command as RCS_DONE.

This changes stts.admin to UNINITIALIZED, even if abort fails.

*/

static void doCmdAbort(         /* ARGUMENTS                             */
					   RCS_CMD_CHANNEL ** subCmdBuf,   /*!< subordinate's command buffer        */
					   RCS_STAT_CHANNEL ** subStatBuf) /*!< subordinate's command status buffer */
{
	PrimSPCmdAbort abortMsg;
	PrimSPStat *subStatPtr;
	int primCount;

	if (state_match(&stat, NEW_COMMAND))
	{ // command is new: send abort command to sub, go to state S0, status EXEC
		state_new(&stat);
		primCmdNumber++;
		abortMsg.serial_number = primCmdNumber;
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
			subCmdBuf[primCount]->write(&abortMsg);
		stat.admin_state = ADMIN_UNINITIALIZED;
		status_next(&stat, RCS_EXEC);
		state_next(&stat, S0);
	}
	else if (state_match(&stat, S0))
	{ // command is not new
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
		{
			subStatPtr = (PrimSPStat *)(subStatBuf[primCount]->get_address());
			if (primCmdNumber != subStatPtr->echo_serial_number)
				return; // stay in state S0
			else if (subStatPtr->status == RCS_EXEC)
				return; // stay in state S0
			else if( subStatPtr->status != RCS_DONE ) // if (subStatPtr->status is RCS_ERROR or unknown)
			{
				state_next(&stat, S1);
				stat.admin_state = ADMIN_UNINITIALIZED;
				status_next(&stat, RCS_ERROR);
				rcs_print_error("%s: prim SP reported %s\n", processName,
					((subStatPtr->status == RCS_ERROR) ?
					"abort error, am SP not aborted" :
				"unknown status, am SP not aborted"));
				return;
			}
		}
		// must be done!
		state_next(&stat, S1);
		stat.admin_state = ADMIN_UNINITIALIZED;
		status_next(&stat, RCS_DONE);
		if (verbose)
			rcs_print("am SP aborted\n");
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stat);
	}
}

/***********************************************************************/

//! doCmdGo executes a go command
/*
Returned Value: none

Called By: doCmd

When the go command is new, this sends a go message to prim SP and
waits until prim SP reports that it is going.

Once prim SP is going, this always calls getSensor Data and writeOutputs.
If the cycleTime has been changed recently, this sends a new go message
to prim SP so that prim SP will send data at the same rate as AM SP
is cycling.

*/

static void doCmdGo(            /* ARGUMENTS                             */
					RCS_CMD_CHANNEL ** subCmdBuf,   /*!< subordinate's command buffer        */
					RCS_STAT_CHANNEL ** subStatBuf) /*!< subordinate's command status buffer */
{
	PrimSPCmdGo goMsg;
	PrimSPStat *subStatPtr;
	int primCount;

	if (stat.admin_state != ADMIN_INITIALIZED && stat.status!= RCS_ERROR)
	{
		status_next(&stat, RCS_ERROR);
		state_next(&stat, S1);
		rcs_print_error("%s: AM SP not initialized, cannot execute command\n",
			processName);
	}
	else if (state_match(&stat, NEW_COMMAND))
	{ // command is new; tell primSP to go
		primCmdNumber++;
		state_new(&stat);
		goMsg.outputCycleTime = set.cycleTime;
		goMsg.serial_number = primCmdNumber;
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
			subCmdBuf[primCount]->write(&goMsg);
		newCycleTime = 0;
		status_next(&stat, RCS_EXEC);
		state_next(&stat, S0);
	}
	else if (state_match(&stat, S0))
	{ // command not new
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
		{
			subStatPtr = (PrimSPStat *)(subStatBuf[primCount]->get_address());
			if (primCmdNumber != subStatPtr->echo_serial_number)
				return; // stay in state S0
			else if (subStatPtr->status ==  RCS_DONE)
			{
				state_next(&stat, S1);
				status_next(&stat, RCS_DONE);
				if (verbose)
					rcs_print("AM SP not going\n");
				return;
			}
			else if (subStatPtr->status != RCS_EXEC)// if (subStatPtr->status is RCS_ERROR or unknown)
			{
				state_next(&stat, S1);
				status_next(&stat, RCS_ERROR);
				rcs_print_error("%s: prim SP reported %s\n", processName,
					((subStatPtr->status == RCS_ERROR) ?
					"go error, AM SP not going" :
				"unknown status, AM SP not going"));
			}
		}
		// Get data and process only if prim is executing
		getSensorData();
		writeOutputs();
		if (newCycleTime)
		{
			primCmdNumber++;
			newCycleTime = 0;
			goMsg.outputCycleTime = set.cycleTime;
			goMsg.serial_number = primCmdNumber;
			for( primCount=0; primCount<numPrimSPChan; primCount++ )
				subCmdBuf[primCount]->write(&goMsg);
		}
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stat);
	}
}

/***********************************************************************/

//! doCmdHalt handles a halt command
/*
Returned Value: none

Called By: doCmd

When the command is new, this sends a halt command to Prim SP and
sets stat.admin to UNINITIALIZED.

If the command is not new and prim is still executing the halt command
it was sent, this waits.

When the prim halt command is finished executing successfully, this sets
stat.admin_state to UNINITIALIZED and reports the AM halt command as RCS_DONE.

This changes stts.admin to UNINITIALIZED, even if halt fails.

*/

static void doCmdHalt(          /* ARGUMENTS                             */
					  RCS_CMD_CHANNEL ** subCmdBuf,   /*!< subordinate's command buffer        */
					  RCS_STAT_CHANNEL ** subStatBuf) /*!< subordinate's command status buffer */
{
	PrimSPCmdHalt haltMsg;
	PrimSPStat *subStatPtr;
	int primCount;

	if (state_match(&stat, NEW_COMMAND))
	{ // command is new: send halt command to sub, go to state S0, status EXEC
		primCmdNumber++;
		state_new(&stat);
		haltMsg.serial_number = primCmdNumber;
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
			subCmdBuf[primCount]->write(&haltMsg);
		stat.admin_state = ADMIN_UNINITIALIZED;
		status_next(&stat, RCS_EXEC);
		state_next(&stat, S0);
	}
	else if (state_match(&stat, S0))
	{ // command is not new
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
		{
			subStatPtr = (PrimSPStat *)(subStatBuf[primCount]->get_address());
			if (primCmdNumber != subStatPtr->echo_serial_number)
				return; // stay in state S0
			else if (subStatPtr->status == RCS_EXEC)
				return; // stay in state S0
			else if (subStatPtr->status != RCS_DONE)
			{
				state_next(&stat, S1);
				stat.admin_state = ADMIN_UNINITIALIZED;
				status_next(&stat, RCS_ERROR);
				rcs_print_error("%s: prim SP reported %s\n", processName,
					((subStatPtr->status == RCS_ERROR) ?
					"halt error, am SP not halted" :
				"unknown status, am SP not halted"));
				return;
			}
		}
		state_next(&stat, S1);
		stat.admin_state = ADMIN_UNINITIALIZED;
		status_next(&stat, RCS_DONE);
		if (verbose)
			rcs_print("am SP halted\n");
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stat);
	}
}

/***********************************************************************/

//! doCmdInit handles an Init command
/*
Returned Value: none

Called By: doCmd

When the command is new, this calls iniLoad, calls initMap, and sends
a prim level init command.

This does not report that the AM SP init command is RCS_DONE until the
status returned from prim SP says that the prim level init command is
RCS_DONE.

If iniLoad and initMap execute without error and prim initializes
without error, this sets stat.admin_state to INITIALIZED. Otherwise,
it sets stat.admin_state to UNINITIALIZED.

The primSPStatBuf was read in main so does not need to be read here.

*/

static void doCmdInit(          /* ARGUMENTS                             */
					  RCS_CMD_CHANNEL ** subCmdBuf,   /*!< subordinate's command buffer        */
					  RCS_STAT_CHANNEL ** subStatBuf) /*!< subordinate's command status buffer */
{
	PrimSPCmdInit initMsg;
	PrimSPStat * subStatPtr;
	int primCount;

	if (state_match(&stat, NEW_COMMAND))
	{
		if( stat.admin_state == ADMIN_INITIALIZED )
		{
			state_next(&stat, S2 );
			return;
		}
		if (iniLoad() == INI_INVALID)
		{
			stat.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stat, RCS_ERROR);
			state_next(&stat, S2);
			rcs_print_error("%s: iniLoad failed, amSP not initialized\n",
				processName);
		}
		else if (initMap())
		{
			stat.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stat, RCS_ERROR);
			state_next(&stat, S2);
			rcs_print_error("%s: initMap failed, amSP not initialized\n",
				processName);
		}
		else
		{
			if (initData())
			{
				stat.admin_state = ADMIN_UNINITIALIZED;
				status_next(&stat, RCS_ERROR);
				state_next(&stat, S2);
				rcs_print_error("%s: initData failed, amSP not initialized\n",
					processName);
			}
			if(initPrimSPNmlBuffers(nmlFile))
			{
				state_next(&stat, S2);
				stat.admin_state = ADMIN_UNINITIALIZED;
				status_next(&stat, RCS_ERROR);
				rcs_print_error("%s: can't allocate primSP NML buffers\n", processName);
				cleanupPrimSPNmlBuffers();
			}
			initServoSPNmlBuffers(nmlFile);
			state_new(&stat);
			primCmdNumber++;
			initMsg.serial_number = primCmdNumber;
			for( primCount=0; primCount<numPrimSPChan; primCount++ )
				primSPCmdBuf[primCount]->write(&initMsg);      // send init command to sub
			stat.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stat, RCS_EXEC);    // go to status EXEC
			state_next(&stat, S0);           // go to state S0
		}
	}
	else if (state_match(&stat, S0))
	{ // command is not new
		for( primCount=0; primCount<numPrimSPChan; primCount++  )
		{
			subStatPtr = (PrimSPStat *)(primSPStatBuf[primCount]->get_address());
			if (primCmdNumber != subStatPtr->echo_serial_number)
				return; // stay in state S0
			else if (subStatPtr->status == RCS_EXEC)
				return; // stay in state S0
			else if (subStatPtr->status != RCS_DONE) // error or unknown
			{
				state_next(&stat, S2);
				stat.admin_state = ADMIN_UNINITIALIZED;
				status_next(&stat, RCS_ERROR);
				rcs_print_error("%s: primSP reported %s\n", processName,
					((subStatPtr->status == RCS_ERROR) ?
					"error initializing, amSP not initialized" :
				"unknown status, amSP not initialized"));
			}
		}
		state_next(&stat, S1);
		stat.admin_state = ADMIN_INITIALIZED;
		status_next(&stat, RCS_DONE);
		rcs_print("amSP initialized\n");
	}
	else
	{ // command is finished executing: stay in state S2
		state_default(&stat);
	}
}

/***********************************************************************/

//! doCmdNop does nothing
/*
Returned Value: none

Called By: doCmd

*/

static void doCmdNop() /* NO ARGUMENTS */
{
	if (state_match(&stat, NEW_COMMAND))
	{
		state_new(&stat);
		status_next(&stat, RCS_DONE);
		state_next(&stat, S0);
	}
	else
	{
		state_default(&stat);
	}
}

/***********************************************************************/

//! doCmdShutdown handles a shutdown command
/*
Returned Value:

Called By: doCmd

When the command is new:
If the administrative state is not UNINITIALIZED this reports an
error and does not change the administrative state.
Otherwise, this send a shutdown command to Prim.

If the command is not new and execution is not finished.
If the administrative state is not UNINITIALIZED this reports an
error, does not change the administrative state, and sets
the command execution state to finished.

Otherwise, if prim is finished executing successfully the shutdown
command it was sent, this sets the administrative state to SHUT_DOWN,
reports the AM shutdown command as RCS_DONE, and sets
the command execution state to finished.

Otherwise, if prim has reported an error or unknown status, this reports
an error, sets the administrative state to UNINITIALIZED, and sets
the command execution state to finished.

Otherwise, if prim is still executing the shutdown command it was
sent, this waits.

*/

static void doCmdShutdown(      /* ARGUMENTS                             */
						  RCS_CMD_CHANNEL ** subCmdBuf,   /*!< subordinate's command buffer        */
						  RCS_STAT_CHANNEL ** subStatBuf) /*!< subordinate's command status buffer */
{
	PrimSPCmdShutdown shutdownMsg;
	PrimSPStat *subStatPtr;
	int primCount;

	if (stat.admin_state != ADMIN_UNINITIALIZED)
	{
		status_next(&stat, RCS_ERROR);
		state_next(&stat, S1);
		rcs_print_error("%s: am SP not uninitialized, "
			"cannot execute shutdown command\n", processName);
	}
	else if (state_match(&stat, NEW_COMMAND))
	{// new command: send shutdown cmd to sub, go to state S0, stat EXEC
		state_new(&stat);
		primCmdNumber++;
		shutdownMsg.serial_number = primCmdNumber;
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
			subCmdBuf[primCount]->write(&shutdownMsg);
		status_next(&stat, RCS_EXEC);
		state_next(&stat, S0);
	}
	else if (state_match(&stat, S0))
	{ // command is not new
		for( primCount=0; primCount<numPrimSPChan; primCount++ )
		{
			subStatPtr = (PrimSPStat *)(subStatBuf[primCount]->get_address());
			if (primCmdNumber != subStatPtr->echo_serial_number)
				return; // stay in state S0
			else if (subStatPtr->status == RCS_EXEC)
				return; // stay in state S0
			else if (subStatPtr->status != RCS_DONE) // error or unknown
			{
				state_next(&stat, S1);
				stat.admin_state = ADMIN_UNINITIALIZED;
				status_next(&stat, RCS_ERROR);
				rcs_print_error("%s: prim SP reported %s\n", processName,
					((subStatPtr->status == RCS_ERROR) ?
					"shutdown error, am SP not shut down" :
				"unknown status, am SP not shut down"));
			}
		}
		state_next(&stat, S1);
		stat.admin_state = ADMIN_SHUT_DOWN;
		status_next(&stat, RCS_DONE);
		if (verbose)
			rcs_print("am SP shut down\n");
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stat);
	}
}

//static int xmlCounter = 0;

static void doPublishXML(NML &nml)
{
	//printf("\n\n\nInside doPblishXML\n\n\n");
	static int count = 0;
	int changesCount = 0;
	char pos[255];
	//cms_force_http_port=80;

	//printf("\n\nNumber of Changes: %d\n\n", map->MapChanges.size());

	if(map==NULL || (vehXPos==0 && vehYPos==0 && vehZPos==0))
	{
		//printf("Could not publish XML file, map is NULL or vehicle pose not set\n");
		return;
	}
	printf("\n\nINSIDE THE PUBLISH XML FUNCTIONS #%d\n\n", count++);

	//map->MapChanges.clear();
	//signal(SIGINT,clock_control_c_handler);
	//return;

	GROUND_ROBOT_INFO clk;
	A *changeArray = NULL;
	clk.XPosition = vehXPos;
	clk.YPosition = vehYPos;
	clk.ZPosition = vehZPos;
	clk.xLow = map[0]->xLow;
	clk.yLow = map[0]->yLow;
	clk.resolution = map[0]->resolution;

	// Sort the list in ascending order
	map[0]->MapChanges.sort();
	list<MapInfo>::iterator curr;
	list<MapInfo>::iterator next;
	curr = map[0]->MapChanges.begin();
	next = map[0]->MapChanges.begin();
	advance(next,1);

	while(curr!=map[0]->MapChanges.end())
	{
		// Make sure we get the latest map change
		while((next!=map[0]->MapChanges.end()) && (*curr == *next))
		{
			curr++;
			next++;
		}

		// Process New Change
		changesCount++;
		changeArray = (A*) realloc(changeArray, changesCount*sizeof(A));
		// Set the XML attribute (row and column, as a string)
		sprintf(pos, "%lf,%lf", (*curr).row , (*curr).col);
		//printf("pos: %s\n\n", pos);
		SET_NML_UNBOUNDED_STRING(changeArray[changesCount-1].pos, pos);
		changeArray[changesCount-1].val = (*curr).prob;
		clk.map[changesCount-1] = changeArray[changesCount-1];

		//if(changesCount==10)
		//break;
		curr++;
		next++;
	}

	//printf("Cols: %d, Rows: %d\n", map->getCols(), map->getRows());

	//SET_NML_UNBOUNDED_STRING(test[0].pos, "1,1");
	//test[0].val = 0.1;

	//SET_NML_UNBOUNDED_STRING(test[1].pos, "2,2");
	//test[1].val = 0.33;

	//clk.map[0] = test[0];
	//clk.map[1] = test[1];
	clk.map_length = changesCount;

	//nml_start();

	//while(!quit_clock)
	//{
	//get_current_cms_time(&clk.now);
	//if(xmlCounter > 1)
	//{
	//xmlCounter = 4;
	nml.write(clk);
	//}

	//xmlCounter++;
	//sleep(1);
	//}


	//nml_cleanup();
	if(changeArray!=NULL)
		free(changeArray);
	map[0]->MapChanges.clear();
}

/***********************************************************************/

//! getFileNames gets the names of the nml file and the ini file
/*
Returned Value: 

Called By: main

If the nmlFile (a global variable) is NULL (implying it was not in the
command arguments), this gets the nmlFile name (1) from the
environment if it is there or (2) from the default if not.

This handles the iniFileName (also a global variable) similarly.

*/

static void getFileNames() /* NO ARGUMENTS */
{
	char * nmlFileEnv;
	char * iniFileEnv;

	if (NULL == nmlFile)
	{
		nmlFileEnv = getenv("CONFIG_NML");
		if (NULL != nmlFileEnv)
		{
			nmlFile = nmlFileEnv;
		}
		else
		{
			nmlFile = DEFAULT_NML_FILE;
		}
	}
	if (NULL == iniFileName)
	{
		iniFileEnv = getenv("CONFIG_INI");
		if (NULL != iniFileEnv)
		{
			iniFileName = iniFileEnv;
		}
		else
		{
			iniFileName = DEFAULT_INI_FILE;
		}
	}
}

/***********************************************************************/

//! getOptions gets the command line options for starting the process
/*
Returned Value: int
If any arguments representing numbers cannot be read, or if any number
required to be positive is not, this returns 1.
Otherwise, it returns 0.

Called By: main

Six command line options are recognized:
1. -n <NML file name>    (a string)
2. -i <ini file name>    (a string)
3. -d <buffer number>    (an int used as a suffix for using different buffers)
4. -r <map range>        (an int used as the internal map range)
5. -o <output map range> (an int used as the output map range)
6. -z <meters>           (a double used as the size of the side of a cell)

The first three get used only at startup. The last three are saved and
reused whenever an init command is received.

This uses the global (?) variable optarg.

*/

static int getOptions( /* ARGUMENTS                                      */
					  int argc,             /*!< number of command line arguments             */
					  char * argv[])        /*!< array of command and arguments               */
{
	char option;

	while (1)
	{
		option = getopt(argc, argv, "n:i:d:r:o:z:");
		if (option == -1)
			break;

		switch (option)
		{
		case 'n':
			nmlFile = optarg;
			if (verbose)
				rcs_print("%s: using specified NML file %s\n",
				processName, nmlFile);
			break;
		case 'i':
			iniFileName = optarg;
			if (verbose)
				rcs_print("%s: using specified INI file %s\n",
				processName, iniFileName);
			break;
		case 'd':
			if (1 != sscanf(optarg, "%d", &bufnum))
			{
				rcs_print_error("%s: bad value given for buffer number: %s\n",
					processName, optarg);
				return 1;
			}
			else if (bufnum < 1)
			{
				rcs_print_error("%s: buffer number is given as %d but "
					"must be positive\n", processName, bufnum);
				return 1;
			}
			else if (verbose)
				rcs_print("%s: using specified buffer suffix %d\n",
				processName, bufnum);
			break;
		case 'r':
			if (1 != sscanf(optarg, "%d", &argMapRange))
			{
				rcs_print_error("%s: bad value given for internal map range: "
					"%s\n", processName, optarg);
				return 1;
			}
			else if (argMapRange < 1)
			{
				rcs_print_error("%s: internal map range is given as %d but "
					"must be positive\n", processName, argMapRange);
				return 1;
			}
			break;
		case 'o':
			if (1 != sscanf(optarg, "%d", &argOutRange))
			{
				rcs_print_error("%s: bad value given for output map range: %s\n",
					processName, optarg);
				return 1;
			}
			else if (argOutRange < 1)
			{
				rcs_print_error("%s: output map range is given as %d but "
					"must be positive\n", processName, argOutRange);
				return 1;
			}
			break;
		case 'z':
			if (1 != sscanf(optarg, "%lf", &argMapRes))
			{
				rcs_print_error("%s: bad value given for map resolution: %s\n",
					processName, optarg);
				return 1;
			}
			else if (argMapRes <= 0)
			{
				rcs_print_error("%s: map resolution is given as %lf but "
					"must be positive\n", processName, argMapRes);
				return 1;
			}
			break;
		default:
			rcs_print_error("%s: bad argument\n", processName);
			return 1;
			break;
		}
	}
	return 0;
}

/***********************************************************************/

//! getSensorData
/*
Returned Value: none

Called By: doCmdGo

This adds points to the map, which is a topographic map in local
coordinates. The input point positions are in local coordinates.

This is handling the range of an input point as follows:
1. If the input point range is larger than input->maxRange, this 
is resetting all the map cells on a line between the sensor and
the input point that are not farther from the sensor than
input->maxRange.
2. If the input point range is larger than 0.95 times input->maxRange,
we do not trust that a point was actually seen, and this is resetting
all the map cells on the line between the sensor and the input point
and assuming nothing was seen at the input point.
3. Otherwise (the point is within 0.95 times input->maxRange), this is
resetting all the map cells on the line between the sensor and the
input point and assuming the input point was seen.

Note that it is OK if the input point is off the map or farther from
the sensor than the range of the map. Points off the map are
handled in AmSPMap::setMapDataClear.

Note: since z is positive into the ground, height is set to -z,
not to z.

The following code was used for data collection and could be put back in.

// insert after the local variable declarations
static int exes[40]; // X values of points -2.0 to +1.9, 0.1 increments
int index;      // index for exes
static int printed = 0;
int i;

// insert after the line "y = item->dataLoc.y;"
index = (int)((x + 2.0) / 0.1);
if ((index > -1) && (index < 39) && (item->sensorLoc.y > 1))
exes[index]++;
if (input->maxRange == 0)
exes[39]++;
if ((y > 10) && (!printed))
{
printf("from amSPMain.cc\n");
printf("input range was 0 %d times\n", exes[39]);
for (i = 0; i < 39; i++)
printf("%.1f %d\n", ((i * 0.1) - 2.0), exes[i]);
printed = 1;
}

*/

static void getSensorData() /* NO ARGUMENTS */
{
	double x, y, z;  // coordinates of data point
	double range;    // distance sensor point to data point
	double xSen;     // X coordinate of sensor
	double ySen;     // Y coordinate of sensor
	double zSen;     // Z coordinate of sensor
	int step;        // index for input->featList
	SensorDataList *input;
	int go;
	SensorPoint * item;



	bool onSlope = false;
	bool gotINSDATA = false;
	int status = servoSPNavBuf->read();
	if(status == SENSOR_DATA_ARRAY_TYPE){
		SensorDataArray * servoSPNav = (SensorDataArray *)servoSPNavBuf->get_address();
		for(int i=0;i<servoSPNav->arrayElem_length;i++)
		{
			switch(servoSPNav->arrayElem[i].type)
			{
			case SERVO_SP_INS_SENSOR:    
				gotINSDATA = true;
				if(INSOutputFile) {
					INSOutputFile << servoSPNav->arrayElem[i].dataTuple[0] << " "
						<< servoSPNav->arrayElem[i].dataTuple[1] << " "
						<< servoSPNav->arrayElem[i].dataTuple[2] << " "
						<< servoSPNav->arrayElem[i].dataTuple[3] << " "
						<< servoSPNav->arrayElem[i].dataTuple[4] << " "
						<< servoSPNav->arrayElem[i].dataTuple[5] << " "
						<< DEG(servoSPNav->arrayElem[i].dataTuple[3]) << " "
						<< DEG(servoSPNav->arrayElem[i].dataTuple[4]) << " "
						<< DEG(servoSPNav->arrayElem[i].dataTuple[5]) << endl;
				}
				float INSReading = DEG(servoSPNav->arrayElem[i].dataTuple[4]);
				if(ABS(potentialGround - INSReading) < 0.1){
					if(sameReadingCount >= INSSLOPE_GROUNDCHANGE){
						groundReading = INSReading;
						potentialGround = groundReading;
						printf("*** Set the ground value to <<< %f >>>\n", groundReading);
					}else sameReadingCount++;
				}
				else{
					potentialGround = INSReading;
					sameReadingCount = 0;
				}
				float lowerBound = groundReading + INSSLOPE_RANGE;
				float upperBound = groundReading - INSSLOPE_RANGE;
				if(lowerBound < 0) lowerBound += 360;
				else if(lowerBound > 360) lowerBound -= 360;
				if(upperBound < 0) upperBound += 360;
				else if(upperBound > 360) upperBound -= 360;	
				if( INSReading > lowerBound && INSReading < upperBound){
					printf("========================================*\n");
					printf("AMSP - Slope Detected\n");
					printf("========================================\n");
					onSlope = true;
					slopeCounter = INSSLOPE_RESETCOUNT;
				}
				else if(slopeCounter >0) slopeCounter--;

				break;
			}
		}
	}
	navDataExtBuf->read(); // need to read buffer even if nav data not used here
	if(gotINSDATA){
		for( int primCount=0; primCount<numPrimSPChan; primCount++ )
		{
			go = sensorDataListBuf[primCount]->read();
			// Don't do anything if we didn't real buffer data

			if (slopeCounter == 0 && SENSOR_DATA_LIST_TYPE == go)
			{ // we got real buffer data

				input = (SensorDataList *)(sensorDataListBuf[primCount]->get_address());
				for (step = 0; step < input->featList_length; step++)
				{ // Look at each element in the incoming buffer
					item = (input->featList + step);
					x = item->dataLoc.x;
					y = item->dataLoc.y;
					z = item->dataLoc.z;
					xSen = item->sensorLoc.x;
					ySen = item->sensorLoc.y;
					zSen = item->sensorLoc.z;
					range = sqrt(((xSen - x) * (xSen - x)) +
						((ySen - y) * (ySen - y)) +
						((zSen - z) * (zSen - z)));
					if (range > input->maxRange)
					{ // Clear an empty path as far as the input->maxRange
						x =  (xSen + ((input->maxRange / range) * (x - xSen)));
						y =  (ySen + ((input->maxRange / range) * (y - ySen)));
						map[primCount]->setMapDataClear(xSen, ySen, x, y, 0, false);
					} // The 0 on the preceding line is not used, just a placeholder
					else if (range > (0.95 * input->maxRange))
					{ // Clear an empty path as far as the range
						map[primCount]->setMapDataClear(xSen, ySen, x, y, 0, false);
					}
					else if (range < input->minRange)
					{ // Range is less than input minimum range
						rcs_print("WARNING: Point range is less than minimum allowed\n");
					}
					else // if (range > 0)
					{ // Clear empty path to the point and put the point on the map
						map[primCount]->setMapDataClear(xSen, ySen, x, y, -z, true);
					}
				}
			}

			vehXPos = xSen;
			vehYPos = ySen;
			vehZPos = zSen;
		}
	}
}

/***********************************************************************/

//! iniLoad loads parameters from the named inifile
/*
Returned Value: int
If the inifile can't be opened or if any given value is invalid, this 
returns INI_INVALID. If an expected value is missing, so that the default
value is used, it returns INI_DEFAULT. Otherwise, it returns INI_OK.

Throughout MOAST, the top-down order of precedence for the source of
most values is:
1. command line argument
2. iniFile
3. default value

For the mapImageFileName, the order is:
1. iniFile
2. default value

For the iniFile name and the NML file name, the order is:
1. command line argument
2. environment variable (CONFIG_INI or CONFIG_NML)
3. default value

Called By: doCmdInit

Values loaded from iniFile:

[AM_MOB_JA]
CELL_RESOLUTION <positive number> meters
CYCLE_TIME      <positive number> seconds
MAP_RANGE       <positive int>    range of AM MOB input & AM SP output maps

[AM_SP]
MAP_IMAGE_FILE_NAME  <char *> name of file into which to print map image
AUX_SENSOR_x_CHAN    <char *> x is sequential integer value starting with 1, the 
prefix of the auxillary prim system is stored here
AUX_SENSOR_x_CHAN_RANGE <positive int> range of internal AM SP map for sensor x

The AM_MOB_JA portion of the iniFile is used for three items that should
be the same in AM mobility and AM SP.

This first calls initParameters, which resets "stat" and "set" and gives
other global variables their default values. Then it loads the
inifile, which results in some values being reset.

Note that iniLoadOneDouble and iniLoadOneInt require all parameters to be
positive, so that is not rechecked here.

*/

static int iniLoad() /* NO ARGUMENTS */
{
	INIFILE iniFile;
	int mobMapSideLength;
	int retval;
	char keyName[80];
	const char * iniString;

	retval = INI_OK;
	if (iniFile.open(iniFileName))
	{
		rcs_print_error("%s: can't open INI file %s\n",
			processName, iniFileName);
		return INI_INVALID;
	}

	if (argMapRes == UNSET_MAP_RES)
	{
		iniLoadOneDouble(&iniFile, "CELL_RESOLUTION", "AM_MOB_JA",
			&mapRes, &retval);
	}
	iniLoadOneDouble(&iniFile, "CYCLE_TIME", "AM_MOB_JA",
		&(set.cycleTime), &retval);
	if (argOutRange == UNSET_OUT_RANGE)
	{
		mobMapSideLength = 0;
		iniLoadOneInt(&iniFile, "MAP_SIDE_LENGTH", "AM_MOB_JA",
			&mobMapSideLength, &retval);
		if (mobMapSideLength && (mobMapSideLength > 0))
		{
			outRange = (((int)((mobMapSideLength / mapRes) + 0.1) + 1) / 2);
		}
	}
	iniLoadOneString(&iniFile, "MAP_IMAGE_FILE_NAME", "AM_SP",
		&mapImageFileName, &retval);
	// first see how many aux sensor we have
	numPrimSPChan = 1;
	strcpy(keyName, "AUX_SENSOR_1_CHAN" );
	while( iniFile.find( keyName, "AM_SP") )
	{
		numPrimSPChan++;
		sprintf( keyName, "AUX_SENSOR_%d_CHAN", numPrimSPChan );
	}
	fprintf( stderr, "Found %d prim subsystems from ini file\n", numPrimSPChan );
	map = new AmSPMap*[numPrimSPChan];
	mapRange = new int[numPrimSPChan];
	primSPBufNums = new int[numPrimSPChan];
	data = new SensorData2D*[numPrimSPChan];
	// now fill in data
	primSPBufNums[0] = bufnum;
	strcpy(keyName, "AUX_SENSOR_1_CHAN" );
	numPrimSPChan = 1;
	iniString = iniFile.find( keyName, "AM_SP");
	while( iniString )
	{
		primSPBufNums[numPrimSPChan] = atoi(iniString);
		sprintf( keyName, "AUX_SENSOR_%d_CHAN_RANGE", numPrimSPChan );
		iniString = iniFile.find( keyName, "AM_SP");
		if( iniString )
			mapRange[numPrimSPChan] = atoi(iniString);
		else
			mapRange[numPrimSPChan] = DEFAULT_MAP_RANGE;
		numPrimSPChan++;
		sprintf( keyName, "AUX_SENSOR_%d_CHAN", numPrimSPChan );
		iniString = iniFile.find( keyName, "AM_SP");
	}
	iniFile.close();
	initParameters();
	if (checkMapValues())
	{
		return 1;
	}
	return retval;
}

/***********************************************************************/

//! iniLoadOneDouble sets one double parameter value from the iniFile.
/*
Returned Value: none

Called By: iniLoad

1. If the parameter is found in the iniFile:
1A. If it has an invalid value, retval is set to INI_INVALID.
1B. If it has a valid value, setMe is set to that value, and retval keeps
its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

Note that this requires all parameters to be positive.

*/

static void iniLoadOneDouble( /*  ARGUMENTS                              */
							 INIFILE * iniFile,           /*! INIFILE class instance to use          */
							 const char * parameter,      /*! name of parameter in ini file          */
							 const char * heading,        /*! heading under which to find parameter  */
							 double * setMe,              /*! number to set                          */
							 int * retval)                /*! iniLoad return value, maybe reset here */
{
	const char * iniString;
	double d;

	iniString = iniFile->find(parameter, heading);
	if (iniString)
	{
		if ((1 == sscanf(iniString, "%lf", &d)) && (d > 0.0))
			*setMe = d;
		else
		{ // found, but invalid
			rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
				processName, iniString, heading, parameter,
				iniFileName);
			*retval = INI_INVALID;
		}
	}
	else
	{ // not found at all
		rcs_print("can't find [%s] %s in '%s'. using %f\n",
			heading, parameter, iniFileName, *setMe);
		*retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
	}
}

/***********************************************************************/

//! iniLoadOneInt sets one integer parameter value from the iniFile.
/*
Returned Value: none

Called By: iniLoad

1. If the parameter is found in the iniFile:
1A. If it has an invalid value, retval is set to INI_INVALID.
1B. If it has a valid value, setMe is set to that value, and retval keeps
its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

Note that this requires all parameters to be positive.

*/

static void iniLoadOneInt( /*  ARGUMENTS                              */
						  INIFILE * iniFile,        /*! INIFILE class instance to use          */
						  const char * parameter,   /*! name of parameter in ini file          */
						  const char * heading,     /*! heading under which to find parameter  */
						  int * setMe,              /*! number to set                          */
						  int * retval)             /*! iniLoad return value, maybe reset here */
{
	const char * iniString;
	int d;

	iniString = iniFile->find(parameter, heading);
	if (iniString)
	{
		if ((1 == sscanf(iniString, "%d", &d)) && (d > 0))
			*setMe = d;
		else
		{ // found, but invalid
			rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
				processName, iniString, heading, parameter,
				iniFileName);
			*retval = INI_INVALID;
		}
	}
	else
	{ // not found at all
		rcs_print("can't find [%s] %s in '%s'. using %d\n",
			heading, parameter, iniFileName, *setMe);
		*retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
	}
}

/***********************************************************************/

//! iniLoadOneString sets one char * parameter value from the iniFile.
/*
Returned Value: none

Called By: iniLoad

1. If the parameter is found in the iniFile, setMe is set to that value,
and retval keeps its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

*/

static void iniLoadOneString( /*  ARGUMENTS                               */
							 INIFILE * iniFile,           /*! INIFILE class instance to use           */
							 const char * parameter,      /*! name of parameter in ini file           */
							 const char * heading,        /*! heading under which to find parameter   */
							 char ** setMe,               /*! string to set                           */
							 int * retval)                /*! iniLoad return value, maybe reset here  */
{
	const char * iniString;

	iniString = iniFile->find(parameter, heading);
	if (iniString)
	{ // found
		*setMe = strdup(iniString);
	}
	else
	{ // not found
		rcs_print("can't find [%s] %s in '%s', using %s\n",
			heading, parameter, iniFileName, *setMe);
		*retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
	}
}

/***********************************************************************/

//! initAmSPNmlBuffers initializes the AM level SP NML buffers
/*
Returned Value: int
If there is an error, this returns 1. Otherwise it returns 0.

Called By: initBuffers and main

*/

static int initAmSPNmlBuffers( /* ARGUMENTS                              */
							  char * nmlFile)               /*!< name of nml configuration file       */
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	cleanupAmSPNmlBuffers();
	sprintf(chanName, "%s%d", AM_SP_CMD_NAME, bufnum);
	amSPCmdBuf =
		new RCS_CMD_CHANNEL(amSP_format, chanName, THIS_PROCESS, nmlFile);
	if ((amSPCmdBuf == NULL) || (!amSPCmdBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", AM_SP_STAT_NAME, bufnum);
	amSPStatBuf =
		new RCS_STAT_CHANNEL(amSP_format, chanName, THIS_PROCESS, nmlFile);
	if ((amSPStatBuf == NULL) || (!amSPStatBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", AM_SP_CFG_NAME, bufnum);
	amSPCfgBuf =
		new RCS_CMD_CHANNEL(amSP_format, chanName, THIS_PROCESS, nmlFile);
	if ((amSPCfgBuf == NULL) || (!amSPCfgBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", AM_SP_SET_NAME, bufnum);
	amSPSetBuf =
		new RCS_STAT_CHANNEL(amSP_format, chanName, THIS_PROCESS, nmlFile);
	if ((amSPSetBuf == NULL) || (!amSPSetBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
	navDataExtBuf =
		new NML(navDataExt_format, chanName, THIS_PROCESS, nmlFile);
	if ((navDataExtBuf == NULL) || (!navDataExtBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", AM_SP_MAP_NAME, bufnum);
	sensorData2DBuf =
		new NML(sensorData_format, chanName, THIS_PROCESS, nmlFile);
	if ((sensorData2DBuf == NULL) || (!sensorData2DBuf->valid()))
		return 1;

	return 0;
}

/***********************************************************************/

//! initBuffers initializes all buffers
/*
Returned Value: int
If there is an error, this returns 1. Otherwise, it returns 0.

Called By: main

*/

static int initBuffers()
{
	int retval;

	retval = initPrimSPNmlBuffers(nmlFile);
	if (retval)
	{
		rcs_print_error("%s: can't allocate primSP NML buffers\n", processName);
		cleanupPrimSPNmlBuffers();
		return 1;
	}
	retval = initAmSPNmlBuffers(nmlFile);
	if (retval)
	{
		rcs_print_error("%s: can't allocate amSP NML buffers\n", processName);
		cleanupNmlBuffers();
		return 1;
	}
	return 0;
}

/***********************************************************************/

//! initData sets some variables on sensorData2D
/*
Returned Value: if fails, return 1
otherwise returns 0

Called By: doCmdInit

*/

static int initData() /* NO ARGUMENTS */
{
	for( int count=0; count<numPrimSPChan; count++)
	{
		data[count] = new SensorData2D();
		if (data[count] == NULL)
			return 1;
		data[count]->isGlobal = false;
		data[count]->incrementX = mapRes;
		data[count]->incrementY = mapRes;
		data[count]->x = map[count]->getRows();
		data[count]->y = map[count]->getCols();
	}
	return 0;
}

/***********************************************************************/

//! initMap makes the map and creates the map display
/*
Returned Value: int
If the map cannot be made, this returns 1.
Otherwise, it returns 0.

Called By: doCmdInit

This initializes the map (a global variable) using the desired
internal range, output range, resolution, and map image file
name. These have been stored as global variables which are set by
default, by command argument or from a configuration command. The map
uses newly allocated memory.

This is testing to see if map (which is initialized to 0) already exists.
If so, it is deleted before making a new one. Thus, AM SP should not
be told to initialize if it is already initialized unless it is OK to
wipe out the existing map.

Before this is called, checkMapValues has checked that mapRange,
outRange, and mapRes are all positive, and mapRange >= outRange.

*/

static int initMap() /* NO ARGUMENTS */
{
	char *mapImageNameToUse;

	for( int count=0; count<numPrimSPChan; count++ )
	{
		if (map[count])
		{
			delete map[count];
			map[count] = NULL;
		}
		// need new memory area for each name
		mapImageNameToUse = new char[strlen(mapImageFileName)+10];
		sprintf( mapImageNameToUse,"%s-%d-%d.ppm", mapImageFileName,bufnum, count );
		printf( "Using %s as the map output name\n", mapImageNameToUse );
		map[count] = new AmSPMap(mapRange[count], mapRes, outRange, mapImageNameToUse);
		if (!map[count])
		{
			rcs_print_error("%s: initMap could not create map\n", processName);
			delete mapImageNameToUse;
			return 1;
		}
		else if (verbose)
			rcs_print("map: range = %d, output range = %d, "
			"resolution = %.3f\n\n", mapRange[count], outRange, mapRes);
	}

	return 0;
}

/***********************************************************************/

//! initParameters puts data in key system parameters and config status.
/*
Returned Value: none

Called By:
iniLoad
main

Configuration status (set) and other global variables are given
initial values here.

Global variables that get values from command line arguments
(iniFileName and nmlFile) are not reset here. Command status is
not reset here.

*/

static void initParameters() /* NO ARGUMENTS */
{
	// initialize configuration command status
	set.cycleTime = DEFAULT_CYCLE_TIME;
	set.debug = 0;
	set.dumpLevel = 0;

	// initialize other global variables
	if (argMapRes != UNSET_MAP_RES)
		mapRes = argMapRes;
	else
		mapRes = DEFAULT_MAP_RES;
	if (argOutRange != UNSET_OUT_RANGE)
		outRange = argOutRange;
	else
		outRange = DEFAULT_OUT_RANGE;
	if (argMapRange != UNSET_MAP_RANGE)
		mapRange[0] = argMapRange;
	else
		mapRange[0] = DEFAULT_MAP_RANGE;
	if (timer)
		delete timer;
	timer = new RCS_TIMER(set.cycleTime);
}

/***********************************************************************/
static int initServoSPNmlBuffers( /* ARGUMENTS                              */
								 char *nmlFile)                  /*!< name of NML file to use              */
{
	char bufName[100];
	sprintf(bufName,"%s%d_1",SERVO_SP_NAV_BUF_NAME,bufnum);
	servoSPNavBuf = new NML(sensorData_format, 
		bufName,
		THIS_PROCESS,
		nmlFile);

	if (!servoSPNavBuf->valid()) 
	{
		printf("WHY WAS THE BUFFER INVALID WHYYYY\n");
		delete servoSPNavBuf;
		servoSPNavBuf = 0;
		return false;
	}else{
		printf("BUFFER ON YUSSSSS\n");
	}
}

//! initPrimSPNmlBuffers opens NML channels to Prim SP
/*
Returned Value: int
If an error occurs, this returns 1. Otherwise, it returns 0.

Called By: initBuffers and doCmdInit

*/

static int initPrimSPNmlBuffers( /* ARGUMENTS                              */
								char *nmlFile)                  /*!< name of NML file to use              */
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	cleanupPrimSPNmlBuffers();
	primSPCmdBuf = new RCS_CMD_CHANNEL*[numPrimSPChan];
	primSPStatBuf = new RCS_STAT_CHANNEL*[numPrimSPChan];
	sensorDataListBuf = new NML*[numPrimSPChan];

	for( int primCount=0; primCount<numPrimSPChan; primCount++)
	{
		sprintf(chanName, "%s%d", PRIM_SP_CMD_NAME, primSPBufNums[primCount]);
		primSPCmdBuf[primCount] = new RCS_CMD_CHANNEL(primSP_format, chanName,
			THIS_PROCESS, nmlFile);
		if ((primSPCmdBuf[primCount] == 0) || (!primSPCmdBuf[primCount]->valid()))
			return 1;

		sprintf(chanName, "%s%d", PRIM_SP_STAT_NAME, primSPBufNums[primCount]);
		primSPStatBuf[primCount] = new RCS_STAT_CHANNEL(primSP_format, chanName,
			THIS_PROCESS, nmlFile);
		if ((primSPStatBuf[primCount] == 0) || (!primSPStatBuf[primCount]->valid()))
			return 1;

		sprintf(chanName, "%s%d", PRIM_SP_LINESCAN_NAME, primSPBufNums[primCount]);
		sensorDataListBuf[primCount] =
			new NML(sensorData_format, chanName, THIS_PROCESS, nmlFile);
		if ((sensorDataListBuf[primCount] == NULL) || (!sensorDataListBuf[primCount]->valid()))
			return 1;
	}
	return 0;
}

/***********************************************************************/

//! main is the main routine for AM SP
/*
Returned Value: int
If there is an error, this returns 1. Otherwise, it returns 0.

Called By: not called

When main starts up, it stuffs a nop command in its own
command input buffer, and it stuffs a nop config command in
its own config input buffer.

In order to get started, an init command and then a go command
must be sent to this AM SP process. This is normally done by
the vehicle level, but it may also be done by the AM shell.

*/

int main(      /* ARGUMENTS                               */
		 int argc,     /*!< one more than the number of arguments */
		 char *argv[]) /*!< array of process name and arguments   */
{
	AmSPCmdNop nopCmd;
	AmSPCfgNop nopCfg;

	processName = argv[0];
	if (getOptions(argc, argv))
	{
		return 1;
	}
	getFileNames();

	if(initAmSPNmlBuffers(nmlFile))
	{
		rcs_print_error("%s: can't allocate amSP NML buffers\n", processName);
		cleanupNmlBuffers();
		return 1;
	}
	stat.echo_serial_number = 0;
	reset_time_tracker(&stat.tt);
	//  initParameters();
	timer = new RCS_TIMER(set.cycleTime);
	nopCmd.serial_number = 1;
	amSPCmdBuf->write(&nopCmd);    // stuff a no op in own command buffer
	nopCfg.serial_number = 1;
	amSPCfgBuf->write(&nopCfg);    // stuff a no op in own config buffer
	done = 0;
	signal(SIGINT, quit);



	//cms_force_http_port=8090;
	//NML nml(xml_robot_info_format,"clock_buf","clock_proc","/usr/local/src/moast/devel/src/am/clock.nml");
	nml_start();

	while (!done)
	{
		readCmd();
		readCfg();
		readPrimSPStat();
		doCmd();
		doCfg();

		/*
		if(MapChangesCounter==100)
		{
		doPublishXML(nml);
		MapChangesCounter=0;
		}
		MapChangesCounter++;
		*/
		cycle_time_tracker(&stat.tt);
		amSPStatBuf->write(&stat);
		amSPSetBuf->write(&set);
		timer->wait();
	}

	printf("\n\n\nDone with amSPMain\n\n\n");

	nml_cleanup();

	while (!done);
	cleanupNmlBuffers();
	if (map)
		delete map;
	rcs_print("amSP done\n");
	return 0;
}

/***********************************************************************/

//!  quit sets done to 1 so that the while loop in main will terminate
/*
Returned Value: none

Called By: pointed at (not called) by signal.

*/

static void quit( /* ARGUMENTS    */
				 int sig)         /* signal value */
{
	done = 1;
}

/***********************************************************************/

//! readCfg reads the config command and modifies the config status buffer.
/*
Returned Value: none

Called By: main

*/

static void readCfg() /* NO ARGUMENTS */
{
	NMLTYPE cfgType;
	int serialNumber;

	cfgType = amSPCfgBuf->read();
	serialNumber = amSPCfgBuf->get_address()->serial_number;
	switch (cfgType)
	{
	case 0:
		// no new command
		break;

	case -1:
		// comm error
		rcs_print_error("%s: NML config read error\n", processName);
		break;

	case AM_SP_CFG_CYCLE_TIME_TYPE:
	case AM_SP_CFG_DEBUG_TYPE:
	case AM_SP_CFG_DUMP_WM_TYPE:
	case AM_SP_CFG_NOP_TYPE:
		set.command_type = cfgType;
		if (serialNumber != set.echo_serial_number)
		{
			set.echo_serial_number = serialNumber;
			set.state = NEW_COMMAND;
			set.status = RCS_EXEC;
		}
		break;

	default:
		rcs_print_error("%s: unknown config: %d\n",
			processName, (int) cfgType);
		break;
	}  // switch (cfgType)
}

/***********************************************************************/

//!readCmd reads in a task command and modifies the command status buffer
/*
Returned Value: none

Called By: main

*/

static void readCmd() /* NO ARGUMENTS */
{
	NMLTYPE cmdType;
	int serialNumber;

	cmdType = amSPCmdBuf->read();
	serialNumber = amSPCmdBuf->get_address()->serial_number;
	switch (cmdType)
	{
	case 0:
		// no new command
		break;

	case -1:
		// comm error
		rcs_print_error("%s: NML command read error\n", processName);
		break;

	case AM_SP_CMD_ABORT_TYPE:
	case AM_SP_CMD_GO_TYPE:
	case AM_SP_CMD_HALT_TYPE:
	case AM_SP_CMD_INIT_TYPE:
	case AM_SP_CMD_NOP_TYPE:
	case AM_SP_CMD_SHUTDOWN_TYPE:
		stat.command_type = cmdType;
		if (serialNumber != stat.echo_serial_number)
		{
			stat.echo_serial_number = serialNumber;
			stat.state = NEW_COMMAND;
			stat.status = RCS_EXEC;
		}
		break;

	default:
		rcs_print_error("%s: unknown command: %d\n", processName, (int) cmdType);
		break;
	}	  // switch (cmdType)
}

/***********************************************************************/

//! readPrimSPStat reads the prim SP status buffer
/*
Returned Value: none

Called By: main

The primCmdNumber, a global variable, is the number of the most recent
command sent to prim SP. primCmdNumber is set to the usual
echo_serial_number+1 when preparing to send a command to prim. In the
functions that give a prim command and then wait for a response from
prim, it is checked that the echoed command number equals
primCmdNumber. If they are not equal, those functions keep waiting.

*/

static void readPrimSPStat()  /* NO ARGUMENTS */
{
	NMLTYPE statType;

	for( int primCount=0; primCount<numPrimSPChan; primCount++ )
	{
		statType = primSPStatBuf[primCount]->read();
		switch (statType)
		{
		case 0:
			// no new status
			break;

		case -1:
			// comm error
			rcs_print_error("%s: NML prim SP status read error\n", processName);
			return;

		case PRIM_SP_STAT_TYPE:
			// new status message
			break;

		default:
			rcs_print_error("%s: unknown prim SP status: %d\n",
				processName, (int)statType);
			return;
		}
	}
}

/***********************************************************************/

//! writeOutputs writes the AM SP data outputs
/*
Returned Value: none

Called By: doCmdGo

Every time this is called, it:
1. calls the map's setVehLoc method
2. calls the map's writeSmallMap method to populate the output data
3. writes the output data into the sensorData2DBuf.

Every tenth time this is called, it calls the map's showDisplay
method.

This uses globalCount (a static local variable) to keep track of how
many times it has been called.

The printSmallMap variable is intended to be used when amSPMain
is run inside gdb. printSmallMap may be set to 1 during a gdb break.
When printSmallMap is non-zero, the small map with all data will be
printed to the file amSPdataOut (for the same area as writeShiftedMap).
The function sets printSmallMap back to 0 so that the map is only
written once.

*/

static void writeOutputs() /* NO ARGUMENTS */
{
	static int globalCount = 0;
	static int printSmallMap = 0;
	NavDataExt *nav = (NavDataExt *)(navDataExtBuf->get_address());
	char fileName[100];
	SensorData2D *dataOut;

	globalCount++;
	for( int primCount=0; primCount<numPrimSPChan; primCount++ )
	{
		map[primCount]->setVehLoc(nav->tranRel);
		map[primCount]->writeSmallMap(data[primCount]);
		if (printSmallMap)
		{
			sprintf( fileName, "amSPdataOut%d", primCount );
			map[primCount]->writeSmallMapData(fileName);
			printSmallMap = 0;
		}
		if (globalCount >= 10)
			map[primCount]->showDisplay();
	}
	if( globalCount >= 10 )
		globalCount = 0;
	dataOut = fuseData();
	sensorData2DBuf->write(dataOut);
}

/***********************************************************************/

/*
Modification history:

$Log: amSPMain.cc,v $
Revision 1.19  2009/04/27 16:09:01  dr_steveb
Made init return if the system is already in an initialized state

Revision 1.18  2008/08/18 18:12:02  dr_steveb
Changed SP to be able to accept multiple prim maps

Revision 1.17  2008/02/12 00:23:00  tomrkramer
Changed handling of updating cycle time so that a go command with a new
outputCycleTime is sent to primSP when the amSP cycle time is changed.
Updated documentation. Removed possibility of calling map->showSmallDisplay,
which no longer exists. Made minor simplifications.

Revision 1.16  2007/12/14 23:45:18  tomrkramer
Made minor fix in documentation.

Revision 1.15  2007/12/03 23:35:30  tomrkramer
Commented out code that attempted to mark vehicle's location as clear
but put an obstacle at the vehicle's location instead.

Revision 1.14  2006/10/05 20:50:16  tomrkramer
Updated to use MAP_SIDE_LENGTH from [AM_MOB_JA] rather than MAP_RANGE.

Revision 1.13  2006/09/18 18:39:37  tomrkramer
Significantly rewritten to use new amSPMap, which uses scrolling map tool.

Revision 1.12  2006/04/20 12:51:08  tomrkramer
Removed heartbeat and replaced it with functions that work equivalently
by updating stat.tt.count (reset_time_tracker(&stat.tt) and
cycle_time_tracker(&stat.tt)).

Revision 1.11  2006/04/12 22:39:08  tomrkramer
Made minor improvements in fault tolerance and documentation.

Revision 1.10  2006/04/10 21:21:08  tomrkramer
Minor changes to implementation of administrative states.

Revision 1.9  2006/04/07 16:55:01  tomrkramer
Fixed possible delete bug. Improved validity checks for buffers.

Revision 1.8  2006/04/05 13:29:23  tomrkramer
Extensive revisions. Re-ordered, reformatted, implemented NOP, stopped
starting itself, implemented admin_state and its values, etc.

Revision 1.7  2006/02/23 15:31:52  tomrkramer
increased data output rate to once every cycle (was once in 5 cycles)

Revision 1.6  2005/11/16 13:44:21  tomrkramer
Improved handling of hits and probabilities

Revision 1.5  2005/10/28 18:01:44  tomrkramer
Changed termination condition in setMapDataClear. Added commented
out code for possible later use.

Revision 1.4  2005/10/24 21:01:18  dr_steveb
Minor bug fix as made by Tom.

Revision 1.2  2005/10/23 13:49:44  dr_steveb
Check-in of changes that Tom K. made to fix missing detections of west facing obstacles.

Revision 1.19  2005/10/06 22:58:44  kramer
minor fixes

Revision 1.18  2005/10/06 22:13:50  kramer
finished changing to rcs_print

Revision 1.17  2005/10/06 21:23:37  kramer
changed output timing, fixed bug

Revision 1.16  2005/10/06 20:50:57  kramer
fixed two bitty bugs

Revision 1.15  2005/10/06 20:48:26  kramer
rearranged, added documentation, minor improvements

Revision 1.14  2005/09/30 18:55:38  proctor
Cleaned up unused variables, forgotten return values

Revision 1.13  2005/09/29 21:46:59  kramer
wholesale changes

Revision 1.12  2005/09/29 16:57:19  stephen
We now write the am data channel.

Revision 1.11  2005/09/29 13:34:11  stephen
Moved some static decs to be defines

Revision 1.10  2005/09/29 12:56:20  stephen
Merged amSPMain.hh with amSPMap.hh

Revision 1.9  2005/09/29 12:08:30  stephen
Added include of math.h for sqrt

Revision 1.8  2005/09/27 21:37:12  kramer
changed to use new SensorPoint, plus improvements

Revision 1.7  2005/09/26 21:32:40  kramer
major changes throughout

Revision 1.6  2005/08/08 19:06:54  beytin
Now running at 10 hz and outputting probably at 0.5 hz (prim outputs at 5)

Revision 1.5  2005/07/25 18:11:52  beytin
Added variable sizing to output map.  User must use -o[range] argument

Revision 1.4  2005/07/22 22:36:22  beytin
Added output to amSPMap1 channel, size of output map is same as internal,
so it needs to be small so it fits in the channel

Revision 1.3  2005/07/18 21:16:32  proctor
Replaced sensorData with servoSPLinescan, ...

Revision 1.2  2005/07/01 16:33:55  beytin
AmSPMap class implemented

Revision 1.1  2005/06/08 19:52:52  proctor
Initial checkin

*/

