/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   scrollingBox.hh
  \brief  Contains class method definitions for ScrollingBox
  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.1 $
  $Date: 2008/08/18 18:50:40 $
  \endcode
  \par CVS Log:
  \ref scrollingBox_cc "More..."

  \author Tom Kramer
  \date   2007-08-09

*/

#include "scrollingBox.hh"
#define SBOX_TINY 0.00001
#define ABS(x) (((x) > 0) ? (x) : -(x))

/***********************************************************************/

/* SEE scrollingBox.hh FOR AN EXPLANATION OF HOW TO USE THIS FILE. */
/* IT IS NOT INTENDED THAT THIS FILE BE EDITED.                    */

/***********************************************************************/


/* ScrollingBox constructor

This is only place the id of a voxel is set.

*/

ScrollingBox::ScrollingBox( /*   ARGUMENTS                                */
 int xRangeIn,              /*!< num voxels between center and edge X dir */
 int yRangeIn,              /*!< num voxels between center and edge Y dir */
 int zRangeIn,              /*!< num voxels between center and edge Z dir */
 double resIn)              /*!< length of the edge of a voxel            */
{
  int n;

  res = resIn;
  xRange = xRangeIn;
  yRange = yRangeIn;
  zRange = zRangeIn;
  xSide = ((xRange * 2) + 1);
  ySide = ((yRange * 2) + 1);
  zSide = ((zRange * 2) + 1);
  numberOfVoxels = (xSide * ySide * zSide);
  centerX = 0.0;
  centerY = 0.0;
  centerZ = 0.0;
  xLowIndex = 0;
  yLowIndex = 0;
  zLowIndex = 0;
  xHi = (res * (xRange + 0.5));
  yHi = (res * (yRange + 0.5));
  zHi = (res * (zRange + 0.5));
  xLow = -xHi;
  yLow = -yHi;
  zLow = -zHi;
  voxels = new VOXEL_NAME[numberOfVoxels];
  for (n = 0; n < numberOfVoxels; n++)
    voxels[n].id = n;
}

/***********************************************************************/

/* ScrollingBox destructor

This deletes the voxels if they were created.

*/

ScrollingBox::~ScrollingBox() /* NO ARGUMENTS */
{
  if (voxels)
    delete[] voxels;
}

/***********************************************************************/

/*! Public method checkEnd takes the coordinates of a start point in
  the box and pointers to the coordinates of an end point that may or
  may not be in the box. If the end point is not in the box, it is
  moved to the point on a face of the box where the line enters the box. */
/*

The approach is to find where the line from end to start first
hits a plane containing one of the faces of the box. If that point is
inside the face, the job is done, but if it is outside the face, the
process is repeated (once or twice) for another face.

*/

void ScrollingBox::checkEnd( /*   ARGUMENTS                   */
 double startX,              /*!< X coordinate of start point */
 double startY,              /*!< Y coordinate of start point */
 double startZ,              /*!< Z coordinate of start point */
 double * endX,              /*!< X coordinate of end point   */
 double * endY,              /*!< Y coordinate of end point   */
 double * endZ)              /*!< Z coordinate of end point   */
{
  static double factor;

  if (*endZ <= zLow)
    { // end point in -Z direction from box. Find point on -Z side of box
      if (*endZ < zLow)
	{
	  factor = ((zLow - *endZ) / (startZ - *endZ));
	  *endX = (*endX + ((startX - *endX) * factor));
	  *endY = (*endY + ((startY - *endY) * factor));
	}
      *endZ = (zLow + SBOX_TINY);
      if (*endY <= yLow)
	{ // end point in -Y direction from box. Find point on -Y side of box
	  if (*endY < yLow)
	    {
	      factor = ((yLow - *endY) / (startY - *endY));
	      *endX = (*endX + ((startX - *endX) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	    }
	  *endY = (yLow + SBOX_TINY);
	  if (*endX < xLow)
	    { // end pt in -X direction from box. Find pt on -X side of box
	      factor = ((xLow - *endX) / (startX - *endX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xLow + SBOX_TINY);
	    }
	  else if (*endX > xHi)
	    { // end pt in +X direction from box. Find pt on +X side of box
	      factor = ((*endX - xHi) / (*endX - startX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xHi - SBOX_TINY);
	    }
	  else if (*endX == xLow)
	    *endX = (xLow + SBOX_TINY);
	  else if (*endX == xHi)
	    *endX = (xHi - SBOX_TINY);
	}
      else if (*endY >= yHi)
	{ // end point in +Y direction from box. Find point on +Y side of box
	  if (*endY > yHi)
	    {
	      factor = ((*endY - yHi) / (*endY - startY));
	      *endX = (*endX + ((startX - *endX) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	    }
	  *endY = (yHi - SBOX_TINY);
	  if (*endX < xLow)
	    { // end pt in -X direction from box. Find pt on -X side of box
	      factor = ((xLow - *endX) / (startX - *endX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xLow + SBOX_TINY);
	    }
	  else if (*endX > xHi)
	    { // end pt in +X direction from box. Find pt on +X side of box
	      factor = ((*endX - xHi) / (*endX - startX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xHi - SBOX_TINY);
	    }
	  else if (*endX == xLow)
	    *endX = (xLow + SBOX_TINY);
	  else if (*endX == xHi)
	    *endX = (xHi - SBOX_TINY);
	}
      else if (*endX < xLow)
	{ // end pt in -X direction from box. Find pt on -X side of box
	  factor = ((xLow - *endX) / (startX - *endX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xLow + SBOX_TINY);
	}
      else if (*endX > xHi)
	{ // end pt in +X direction from box. Find pt on +X side of box
	  factor = ((*endX - xHi) / (*endX - startX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xHi - SBOX_TINY);
	}
      else if (*endX == xLow)
	*endX = (xLow + SBOX_TINY);
      else if (*endX == xHi)
	*endX = (xHi - SBOX_TINY);
      endInside = false;
    }
  else if (*endZ >= zHi)
    { // end point in +Z direction from box. Find point on +Z side of box
      if (*endZ > zHi)
	{
	  factor = ((*endZ - zHi) / (*endZ - startZ));
	  *endX = (*endX + ((startX - *endX) * factor));
	  *endY = (*endY + ((startY - *endY) * factor));
	}
      *endZ = (zHi - SBOX_TINY);
      if (*endY <= yLow)
	{ // end point in -Y direction from box. Find point on -Y side of box
	  if (*endY < yLow)
	    {
	      factor = ((yLow - *endY) / (startY - *endY));
	      *endX = (*endX + ((startX - *endX) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	    }
	  *endY = (yLow + SBOX_TINY);
	  if (*endX < xLow)
	    { // end pt in -X direction from box. Find pt on -X side of box
	      factor = ((xLow - *endX) / (startX - *endX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xLow + SBOX_TINY);
	    }
	  else if (*endX > xHi)
	    { // end pt in +X direction from box. Find pt on +X side of box
	      factor = ((*endX - xHi) / (*endX - startX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xHi - SBOX_TINY);
	    }
	  else if (*endX == xLow)
	    *endX = (xLow + SBOX_TINY);
	  else if (*endX == xHi)
	    *endX = (xHi - SBOX_TINY);
	}
      else if (*endY >= yHi)
	{ // end point in +Y direction from box. Find point on +Y side of box
	  if (*endY > yHi)
	    {
	      factor = ((*endY - yHi) / (*endY - startY));
	      *endX = (*endX + ((startX - *endX) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	    }
	  *endY = (yHi - SBOX_TINY);
	  if (*endX < xLow)
	    { // end pt in -X direction from box. Find pt on -X side of box
	      factor = ((xLow - *endX) / (startX - *endX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xLow + SBOX_TINY);
	    }
	  else if (*endX > xHi)
	    { // end pt in +X direction from box. Find pt on +X side of box
	      factor = ((*endX - xHi) / (*endX - startX));
	      *endY = (*endY + ((startY - *endY) * factor));
	      *endZ = (*endZ + ((startZ - *endZ) * factor));
	      *endX = (xHi - SBOX_TINY);
	    }
	  else if (*endX == xLow)
	    *endX = (xLow + SBOX_TINY);
	  else if (*endX == xHi)
	    *endX = (xHi - SBOX_TINY);
	}
      else if (*endX < xLow)
	{ // end pt in -X direction from box. Find pt on -X side of box
	  factor = ((xLow - *endX) / (startX - *endX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xLow + SBOX_TINY);
	}
      else if (*endX > xHi)
	{ // end pt in +X direction from box. Find pt on +X side of box
	  factor = ((*endX - xHi) / (*endX - startX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xHi - SBOX_TINY);
	}
      else if (*endX == xLow)
	*endX = (xLow + SBOX_TINY);
      else if (*endX == xHi)
	*endX = (xHi - SBOX_TINY);
      endInside = false;
    }
  else if (*endY <= yLow)
    { // end point in -Y direction from box. Find point on -Y side of box
      if (*endY < yLow)
	{
	  factor = ((yLow - *endY) / (startY - *endY));
	  *endX = (*endX + ((startX - *endX) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	}
      *endY = (yLow + SBOX_TINY);
      if (*endX < xLow)
	{ // end pt in -X direction from box. Find pt on -X side of box
	  factor = ((xLow - *endX) / (startX - *endX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xLow + SBOX_TINY);
	}
      else if (*endX > xHi)
	{ // end pt in +X direction from box. Find pt on +X side of box
	  factor = ((*endX - xHi)/(*endX - startX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xHi - SBOX_TINY);
	}
      else if (*endX == xLow)
	*endX = (xLow + SBOX_TINY);
      else if (*endX == xHi)
	*endX = (xHi - SBOX_TINY);
      endInside = false;
    }
  else if (*endY >= yHi)
    { // end point in +Y direction from box. Find point on +Y side of box
      if (*endY > yHi)
	{
	  factor = ((*endY - yHi) / (*endY - startY));
	  *endX = (*endX + ((startX - *endX) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	}
      *endY = (yHi - SBOX_TINY);
      if (*endX < xLow)
	{ // end pt in -X direction from box. Find pt on -X side of box
	  factor = ((xLow - *endX) / (startX - *endX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xLow + SBOX_TINY);
	}
      else if (*endX > xHi)
	{ // end pt in +X direction from box. Find pt on +X side of box
	  factor = ((*endX - xHi) / (*endX - startX));
	  *endY = (*endY + ((startY - *endY) * factor));
	  *endZ = (*endZ + ((startZ - *endZ) * factor));
	  *endX = (xHi - SBOX_TINY);
	}
      else if (*endX == xLow)
	*endX = (xLow + SBOX_TINY);
      else if (*endX == xHi)
	*endX = (xHi - SBOX_TINY);
       endInside = false;
   }
  else if (*endX > xHi)
    {
      factor = ((*endX - xHi) / (*endX - startX));
      *endY = (*endY + ((startY - *endY) * factor));
      *endZ = (*endZ + ((startZ - *endZ) * factor));
      *endX = (xHi - SBOX_TINY);
      endInside = false;
    }
  else if (*endX < xLow)
    {
      factor = ((xLow - *endX) / (startX - *endX));
      *endY = (*endY + ((startY - *endY) * factor));
      *endZ = (*endZ + ((startZ - *endZ) * factor));
      *endX = (xLow + SBOX_TINY);
      endInside = false;
    }
  else if (*endX == xLow)
    {
      *endX = (xLow + SBOX_TINY);
      endInside = true;
    }
  else if (*endX == xHi)
    {
      *endX = (xHi - SBOX_TINY);
      endInside = true;
    }
  else
    endInside = true;
}


/***********************************************************************/

/*! Public method getVoxelAtCoords returns a pointer to the voxel
  containing the point whose cartesian X and Y coordinates are x and
  y. If the point is not in the box, this returns NULL. */
/*

This does not require knowing the origin because of way scrolling
works (as though space were packed with copies of the box).

*/

VOXEL_NAME * ScrollingBox::getVoxelAtCoords( /*   ARGUMENTS             */
 double x,                                   /*!< X coordinate of point */
 double y,                                   /*!< Y coordinate of point */
 double z)                                   /*!< Z coordinate of point */
{
  int xIndex; // actual grid X index of point
  int yIndex; // actual grid Y index of point
  int zIndex; // actual grid Z index of point

  if ((x < xLow) || (x > xHi) ||
      (y < yLow) || (y > yHi) ||
      (z < zLow) || (z > zHi))
    return 0;
  xIndex = ((int)floor((x / res) + xRange + 0.5) % xSide);
  if (xIndex < 0)
    xIndex = (xIndex + xSide);
  yIndex = ((int)floor((y / res) + yRange + 0.5) % ySide);
  if (yIndex < 0)
    yIndex = (yIndex + ySide);
  zIndex = ((int)floor((z / res) + zRange + 0.5) % zSide);
  if (zIndex < 0)
    zIndex = (zIndex + zSide);
  return getVoxel3D(xIndex, yIndex, zIndex);
}

/***********************************************************************/

/*! Public method getVoxelAtIndexes returns a pointer to the voxel whose
  X, Y, and Z indexes in the box grid are xIndex and col. */
/* 

For speed, this is not checking index bounds. The indexes should satisfy
(0 <= xBoxIndex < xSide), (0 <= yBoxIndex < ySide), and
(0 <= zBoxIndex < zSide).

*/

VOXEL_NAME * ScrollingBox::getVoxelAtIndexes( /*   ARGUMENTS            */
 int xBoxIndex,                               /*!> box X index of voxel */
 int yBoxIndex,                               /*!> box Y index of voxel */
 int zBoxIndex)                               /*!> box Z index of voxel */
{
  int xActIndex;
  int yActIndex;
  int zActIndex;

  xActIndex = (((xLowIndex + xBoxIndex) >= xSide) ?
	       (xLowIndex + xBoxIndex - xSide) : (xLowIndex + xBoxIndex));
  yActIndex = (((yLowIndex + yBoxIndex) >= ySide) ?
	       (yLowIndex + yBoxIndex - ySide) : (yLowIndex + yBoxIndex));
  zActIndex = (((zLowIndex + zBoxIndex) >= zSide) ?
	       (zLowIndex + zBoxIndex - zSide) : (zLowIndex + zBoxIndex));
  return getVoxel3D(xActIndex, yActIndex, zActIndex);
}

/***********************************************************************/

/*! Public method getNextVoxel returns a pointer to the next voxel lying
  on the line whose start point and end point were given in the most
  recent call to initLine. If the end of the line was reached
  previously, this returns a NULL pointer. */
/*

Much of the code here is the same as in initLine.

The general approach is to move one voxel width in the principal
direction each time this is called. However, if an earlier call to
either initLine or getNextVoxel resulted in next2 only being set or next2
and next3 being set and one of those has not yet been returned, the
function returns one of them and does not move across a voxel.  When
the function moves one voxel in the principal direction, it may also
pass through one voxel in one or both of the secondary directions.
If so, the function sets next2 only or next2 and next3.

Many of the variables used by this function are set in initLine. See
the documentation of that function and the documentation of
scrollingBox.hh regarding what the variables mean.

*/

VOXEL_NAME * ScrollingBox::getNextVoxel() /* NO ARGUMENTS */
{
  if (!inLine)
    next1 = 0;
  else if (next3)
    {
      next1 = next2;
      next2 = next3;
      next3 = 0;
    }
  else if (next2)
    {
      next1 = next2;
      next2 = 0;
      if ((dI >= totalI) && (dJ >= totalJ) && (dK >= totalK))
	inLine = 0;
    }
  else if ((dI >= totalI) && (dJ >= totalJ) && (dK >= totalK))
    {
      inLine = 0;
      next1 = 0;
    }
  else if (principalDir < Y_PLUS_DIR)
    { //step in X direction
      next1 = getVoxelAtIndexes(i, j, k);
      if ((step1 > lack1) && (dJ < totalJ))
	{ // line passes through another voxel in the Y direction
	  if ((step2 > lack2) && (dK < totalK))
	    { // line also passes through another voxel in Z direction
	      if ((lack1 / slope1) <= (lack2 / slope2))
		{ // line passes first through voxel in the Y direction
		  j = (j + dj);
		  dJ++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		  k = (k + dk);
		  dK++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		}
	      else
		{ // line passes first through voxel in the Z direction
		  k = (k + dk);
		  dK++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		  j = (j + dj);
		  dJ++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		}
	    }
	  else
	    { // line does not pass through another voxel in Z direction
	      j = (j + dj);
	      dJ++;
	      next2 = getVoxelAtIndexes(i, j, k);
	      next3 = 0;
	      lack1 = (res - (step1 - lack1));
	      lack2 = (lack2 - step2);
	    }
	}
      else if ((step2 > lack2) && (dK < totalK))
	{ // line passes through one other voxel in Z direction
	  k = (k + dk);
	  dK++;
	  next2 = getVoxelAtIndexes(i, j, k);
	  lack1 = (lack1 - step1);
	  lack2 = (res - (step2 - lack2));
	  next3 = 0;
	}
      else
	{ // line does not pass through another voxel
	  next2 = 0;
	  next3 = 0;
	  lack1 = (lack1 - step1);
	  lack2 = (lack2 - step2);
	}
      i = (i + di);
      dI++;
    }
  else if (principalDir < Z_PLUS_DIR)
    { //step in Y direction
      next1 = getVoxelAtIndexes(i, j, k);
      if ((step1 > lack1) && (dK < totalK))
	{ // line passes through another voxel in the Z direction
	  if ((step2 > lack2) && (dI < totalI))
	    { // line also passes through another voxel in X direction
	      if ((lack1 / slope1) <= (lack2 / slope2))
		{ // line passes first through voxel in the Z direction
		  k = (k + dk);
		  dK++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		  i = (i + di);
		  dI++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		}
	      else
		{ // line passes first through voxel in the X direction
		  i = (i + di);
		  dI++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		  k = (k + dk);
		  dK++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		}
	    }
	  else
	    { // line does not pass through another voxel in X direction
	      k = (k + dk);
	      dK++;
	      next2 = getVoxelAtIndexes(i, j, k);
	      next3 = 0;
	      lack1 = (res - (step1 - lack1));
	      lack2 = (lack2 - step2);
	    }
	}
      else if ((step2 > lack2) && (dI < totalI))
	{ // line passes through one other voxel in X direction
	  i = (i + di);
	  dI++;
	  next2 = getVoxelAtIndexes(i, j, k);
	  lack1 = (lack1 - step1);
	  lack2 = (res - (step2 - lack2));
	  next3 = 0;
	}
      else
	{ // line does not pass through another voxel
	  next2 = 0;
	  next3 = 0;
	  lack1 = (lack1 - step1);
	  lack2 = (lack2 - step2);
	}
      j = (j + dj);
      dJ++;
    }
  else // if ((principalDir == Z_PLUS_DIR) || (principalDir == Z_MINUS_DIR))
    { //step in Z direction
      next1 = getVoxelAtIndexes(i, j, k);
      if ((step1 > lack1) && (dI < totalI))
	{ // line passes through another voxel in the X direction
	  if ((step2 > lack2) && (dJ < totalJ))
	    { // line also passes through another voxel in Y direction
	      if ((lack1 / slope1) <= (lack2 / slope2))
		{ // line passes first through voxel in the X direction
		  i = (i + di);
		  dI++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		  j = (j + dj);
		  dJ++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		}
	      else
		{ // line passes first through voxel in the Y direction
		  j = (j + dj);
		  dJ++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		  i = (i + di);
		  dI++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		}
	    }
	  else
	    { // line does not pass through another voxel in Y direction
	      i = (i + di);
	      dI++;
	      next2 = getVoxelAtIndexes(i, j, k);
	      next3 = 0;
	      lack1 = (res - (step1 - lack1));
	      lack2 = (lack2 - step2);
	    }
	}
      else if ((step2 > lack2) && (dJ < totalJ))
	{ // line passes through one other voxel in Y direction
	  j = (j + dj);
	  dJ++;
	  next2 = getVoxelAtIndexes(i, j, k);
	  lack1 = (lack1 - step1);
	  lack2 = (res - (step2 - lack2));
	  next3 = 0;
	}
      else
	{ // line does not pass through another voxel
	  next2 = 0;
	  next3 = 0;
	  lack1 = (lack1 - step1);
	  lack2 = (lack2 - step2);
	}
      k = (k + dk);
      dK++;
    }
  return next1;
}

/***********************************************************************/

/*! Public method getXCoordAtId returns the X coordinate of the center
  of the voxel with the given id. If the id is not the id of a voxel,
  this returns xLow. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */
  
double ScrollingBox::getXCoordAtId( /*   ARGUMENTS                      */
 int id)                            /*!> id of voxel to get X coord for */
{
  int ix; // number of voxels to the +X side of the box min X

  if ((id < 0) || (id >= numberOfVoxels))
    return xLow;
  ix = (((id % (xSide * ySide)) % xSide) - xLowIndex);
  if (ix < 0)
    ix = (ix + xSide);
  return (xLow + ((ix + 0.5) * res));
}

/***********************************************************************/

/*! Public method getXCoordAtVoxel returns the X coordinate of the
  center of the given voxel. If the voxel is NULL or its id is not the
  id of a voxel, this returns xLow. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */

double ScrollingBox::getXCoordAtVoxel( /*   ARGUMENTS                      */
 VOXEL_NAME * aVoxel)                  /*!> voxel for which to get X coord */
{
  int ix; // number of voxels to the +X side of the box min X

  if ((!aVoxel) || (aVoxel->id < 0) || (aVoxel->id >= numberOfVoxels))
    return xLow;
  ix = (((aVoxel->id % (xSide * ySide)) % xSide) - xLowIndex);
  if (ix < 0)
    ix = (ix + xSide);
  return (xLow + ((ix + 0.5) * res));
}

/***********************************************************************/

/*! Public method getXIndexAtVoxel returns the box grid X index of
  the given voxel. If the voxel is NULL or its id is not the id of a
  voxel, this returns -1. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */
  
int ScrollingBox::getXIndexAtVoxel( /*   ARGUMENTS                      */
 VOXEL_NAME * aVoxel)               /*!> voxel for which to get X index */
{
  int ix; // number of voxels to the +X side of the box min X

  if ((!aVoxel) || (aVoxel->id < 0) || (aVoxel->id >= numberOfVoxels))
    return -1;
  ix = (((aVoxel->id % (xSide * ySide)) % xSide) - xLowIndex);
  return ((ix < 0) ? (ix + xSide) : ix);
}

/***********************************************************************/

/*! Public method getXIndexAtCoord returns the box grid X index of
  the X coordinate value x, if x is inside the box. Otherwise, it
  returns -1. */

int ScrollingBox::getXIndexAtCoord( /*   ARGUMENTS    */
 double x)                          /*!< X coordinate */
{
  if ((x < xLow) || (x > xHi))
    return -1;
  return (int)(floor((x - xLow) / res));
}

/***********************************************************************/

/*! Public method getYCoordAtId returns the Y coordinate of the center
  of the voxel with the given id. If the id is not the id of a voxel,
  this returns yLow. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */
  
double ScrollingBox::getYCoordAtId( /*   ARGUMENTS                      */
 int id)                            /*!> id of voxel to get Y coord for */
{
  int iy; // number of voxels to the +Y side of the box min Y

  if ((id < 0) || (id >= numberOfVoxels))
    return yLow;
  iy = (((id % (xSide * ySide)) / xSide) - yLowIndex);
  if (iy < 0)
    iy = (iy + ySide);
  return (yLow + ((iy + 0.5) * res));
}

/***********************************************************************/

/*! Public method getYCoordAtVoxel returns the Y coordinate of the
  center of the given voxel. If the voxel is NULL or its id is not the
  id of a voxel, this returns yLow. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */

double ScrollingBox::getYCoordAtVoxel( /*   ARGUMENTS                      */
 VOXEL_NAME * aVoxel)                  /*!> voxel for which to get Y coord */
{
  int iy; // number of voxels to the +Y side of the box min Y

  if ((!aVoxel) || (aVoxel->id < 0) || (aVoxel->id >= numberOfVoxels))
    return yLow;
  iy = (((aVoxel->id % (xSide * ySide)) / xSide) - yLowIndex);
  if (iy < 0)
    iy = (iy + ySide);
  return (yLow + ((iy + 0.5) * res));
}

/***********************************************************************/

/*! Public method getYIndexAtVoxel returns the box grid Y index of
  the given voxel. If the voxel is NULL or its id is not the id of a
  voxel, this returns -1. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */
  
int ScrollingBox::getYIndexAtVoxel( /*   ARGUMENTS                      */
 VOXEL_NAME * aVoxel)               /*!> voxel for which to get Y index */
{
  int iy; // number of voxels to the +Y side of the box min Y

  if ((!aVoxel) || (aVoxel->id < 0) || (aVoxel->id >= numberOfVoxels))
    return -1;
  iy = (((aVoxel->id % (xSide * ySide)) / xSide) - yLowIndex);
  return ((iy < 0) ? (iy + ySide) : iy);
}

/***********************************************************************/

/*! Public method getYIndexAtCoord returns the box grid Y index of
  the Y coordinate value y, if y is inside the box. Otherwise, it
  returns -1. */

int ScrollingBox::getYIndexAtCoord( /*   ARGUMENTS    */
 double y)                          /*!> Y coordinate */
{
  if ((y < yLow) || (y > yHi))
    return -1;
  return (int)(floor((y - yLow) / res));
}

/***********************************************************************/

/*! Public method getZCoordAtId returns the Z coordinate of the center
  of the voxel with the given id. If the id is not the id of a voxel,
  this returns zLow. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */
  
double ScrollingBox::getZCoordAtId( /*   ARGUMENTS                      */
 int id)                            /*!> id of voxel to get Z coord for */
{
  int iz; // number of voxels to the +Z side of the box min Z

  if ((id < 0) || (id >= numberOfVoxels))
    return zLow;
  iz = ((id / (xSide * ySide)) - zLowIndex);
  if (iz < 0)
    iz = (iz + zSide);
  return (zLow + ((iz + 0.5) * res));
}

/***********************************************************************/

/*! Public method getZCoordAtVoxel returns the Z coordinate of the
  center of the given voxel. If the voxel is NULL or its id is not the
  id of a voxel, this returns zLow. */

/* This works because the id (which is assigned by the ScrollingBox
   constructor and cannot be changed) is the index in the 1D array of
   the voxel. */

double ScrollingBox::getZCoordAtVoxel( /*   ARGUMENTS                      */
 VOXEL_NAME * aVoxel)                  /*!> voxel for which to get Z coord */
{
  int iz; // number of voxels to the +Z side of the box min Z

  if ((!aVoxel) || (aVoxel->id < 0) || (aVoxel->id >= numberOfVoxels))
    return zLow;
  iz = ((aVoxel->id / (xSide * ySide)) - zLowIndex);
  if (iz < 0)
    iz = (iz + zSide);
  return (zLow + ((iz + 0.5) * res));
}

/***********************************************************************/

/*! Public method getZIndexAtVoxel returns the box grid Z index of
  the given voxel. If the voxel is NULL or its id is not the id of a
  voxel, this returns -1. */
/* 

This works because the id (which is assigned by the ScrollingBox
constructor and cannot be changed) is the index in the 1D array of
the voxel.

*/
  
int ScrollingBox::getZIndexAtVoxel( /*   ARGUMENTS                      */
 VOXEL_NAME * aVoxel)               /*!> voxel for which to get Z index */
{
  int iz; // number of voxels to the +Z side of the box min Z

  if ((!aVoxel) || (aVoxel->id < 0) || (aVoxel->id >= numberOfVoxels))
    return -1;
  iz = ((aVoxel->id / (xSide * ySide)) - zLowIndex);
  return ((iz < 0) ? (iz + zSide) : iz);
}

/***********************************************************************/

/*! Public method getZIndexAtCoord returns the box grid Z index of
  the Z coordinate value z, if z is inside the box. Otherwise, it
  returns -1. */

int ScrollingBox::getZIndexAtCoord( /*   ARGUMENTS    */
 double z)                          /*!> Z coordinate */
{
  if ((z < zLow) || (z > zHi))
    return -1;
  return (int)(floor((z - zLow) / res));
}

/***********************************************************************/

/*! Public method initLine returns a pointer to the voxel containing
   the point in the box whose coordinates are startX, startY, and
   startZ.  If the start point is outside the box, this returns a NULL
   pointer (so don't think outside the box). If the start point is in
   the box, this also sets data to support the work done by
   findNextVoxel, which finds the next voxel on the line from the
   start point to the end point. */
/*

If (endX, endY, endZ) is outside the box, this resets (endX, endY,
endZ) to the point where the line from the end to the start enters the
box. The approach is to find where the line from end to start first
hits a plane containing one of the faces of the box. If that point is
inside the face, the job is done, but if it is outside the face, the
process is repeated (once or twice) for another face.

Step0 is the distance that must be moved in the principal direction
from the start point to a face of a voxel perpendicular to the
principal direction.  It is possible that the line passes into one or
two other voxels (in addition to the one containing the start point)
before it has gone step0 in the principal direction. If there are two
such voxels, it is necessary to determine which is entered first.

When the principal direction is the X direction:
lack1 and step1 are in the Y direction,
slope1 is the change in Y per unit change in X,
lack2 and step2 are in the Z direction,
slope2 is the change in Z per unit change in X.

When the principal direction is the Y direction:
lack1 and step1 are in the Z direction,
slope1 is the change in Z per unit change in Y,
lack2 and step2 are in the X direction,
slope2 is the change in X per unit change in Y.

When the principal direction is the Z direction:
lack1 and step1 are in the X direction,
slope1 is the change in X per unit change in Z,
lack2 and step2 are in the Y direction,
slope2 is the change in Y per unit change in Z.

*/

VOXEL_NAME * ScrollingBox::initLine( /*   ARGUMENTS                   */
 double startX,                      /*!< X coordinate of start point */
 double startY,                      /*!< Y coordinate of start point */
 double startZ,                      /*!< Z coordinate of start point */
 double endX,                        /*!< X coordinate of end point   */
 double endY,                        /*!< Y coordinate of end point   */
 double endZ)                        /*!< Z coordinate of end point   */
{
  double step0;   // size of first step in primary direction (always positive)

  if ((startX < xLow) || (startX > xHi) ||
      (startY < yLow) || (startY > yHi) ||
      (startZ < zLow) || (startZ > zHi))
    {
      inLine = false;
      return 0;
    }
  i = getXIndexAtCoord(startX);
  j = getYIndexAtCoord(startY);
  k = getZIndexAtCoord(startZ);
  if ((startX == endX) && (startY == endY) && (startZ == endZ))
    { // FIX TEST ABOVE- DO THIS IF START AND END ARE IN SAME VOXEL
      inLine = false;
      return getVoxelAtIndexes(i, j, k);
    }
  inLine = true;
  checkEnd(startX, startY, startZ, &endX, &endY, &endZ);
  dI = 1;
  dJ = 1;
  dK = 1;
  deltaX = (endX - startX);
  deltaY = (endY - startY);
  deltaZ = (endZ - startZ);
  totalI = (1 + ABS(getXIndexAtCoord(endX) - getXIndexAtCoord(startX)));
  totalJ = (1 + ABS(getYIndexAtCoord(endY) - getYIndexAtCoord(startY)));
  totalK = (1 + ABS(getZIndexAtCoord(endZ) - getZIndexAtCoord(startZ)));
  di = ((deltaX >= 0) ? 1 : -1);
  dj = ((deltaY >= 0) ? 1 : -1);
  dk = ((deltaZ >= 0) ? 1 : -1);
  next1 = getVoxelAtIndexes(i, j, k);
  if ((fabs(deltaX) >= fabs(deltaY)) && (fabs(deltaX) >= fabs(deltaZ)))
    { //step0 in X direction
      principalDir = ((di == 1) ? X_PLUS_DIR : X_MINUS_DIR);
      slope1 = fabs(deltaY / deltaX);
      slope2 = fabs(deltaZ / deltaX);
      step0 = ((deltaX > 0) ?
	       (xLow + ((getXIndexAtCoord(startX) + 1) * res) - startX) :
	       (startX - (xLow + (getXIndexAtCoord(startX) * res))));
      step1 = (step0 * slope1);
      step2 = (step0 * slope2);
      lack1 = ((deltaY >= 0) ?
	       (yLow + ((getYIndexAtCoord(startY) + 1) * res) - startY) :
	       (startY - (yLow + (getYIndexAtCoord(startY) * res))));
      lack2 = ((deltaZ >= 0) ?
	       (zLow + ((getZIndexAtCoord(startZ) + 1) * res) - startZ) :
	       (startZ - (zLow + (getZIndexAtCoord(startZ) * res))));
      if (step1 > lack1)
	{ // line passes through another voxel in the Y direction
	  if (step2 > lack2)
	    { // line also passes through another voxel in Z direction
	      if ((lack1 / slope1) <= (lack2 / slope2))
		{ // line passes first through voxel in the Y direction
		  j = (j + dj);
		  dJ++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		  k = (k + dk);
		  dK++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		}
	      else
		{ // line passes first through voxel in the Z direction
		  k = (k + dk);
		  dK++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		  j = (j + dj);
		  dJ++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		}
	    }
	  else
	    { // line does not pass through another voxel in Z direction
	      j = (j + dj);
	      dJ++;
	      next2 = getVoxelAtIndexes(i, j, k);
	      next3 = 0;
	      lack1 = (res - (step1 - lack1));
	      lack2 = (lack2 - step2);
	    }
	}
      else if (step2 > lack2)
	{ // line passes through one other voxel in Z direction
	  k = (k + dk);
	  dK++;
	  next2 = getVoxelAtIndexes(i, j, k);
	  lack1 = (lack1 - step1);
	  lack2 = (res - (step2 - lack2));
	  next3 = 0;
	}
      else
	{ // line does not pass through another voxel
	  next2 = 0;
	  next3 = 0;
	  lack1 = (lack1 - step1);
	  lack2 = (lack2 - step2);
	}
      i = (i + di);
    }
  else if ((fabs(deltaY) >= fabs(deltaX)) && (fabs(deltaY) >= fabs(deltaZ)))
    { //step0 in Y direction
      principalDir = ((dj == 1) ? Y_PLUS_DIR : Y_MINUS_DIR);
      slope1 = fabs(deltaZ / deltaY);
      slope2 = fabs(deltaX / deltaY);
      step0 = ((deltaY > 0) ?
	       (yLow + ((getYIndexAtCoord(startY) + 1) * res) - startY) :
	       (startY - (yLow + (getYIndexAtCoord(startY) * res))));
      step1 = (step0 * slope1);
      step2 = (step0 * slope2);
      lack1 = ((deltaZ >= 0) ?
	       (zLow + ((getZIndexAtCoord(startZ) + 1) * res) - startZ) :
	       (startZ - (zLow + (getZIndexAtCoord(startZ) * res))));
      lack2 = ((deltaX >= 0) ?
	       (xLow + ((getXIndexAtCoord(startX) + 1) * res) - startX) :
	       (startX - (xLow + (getXIndexAtCoord(startX) * res))));
      if (step1 > lack1)
	{ // line passes through another voxel in the Z direction
	  if (step2 > lack2)
	    { // line also passes through another voxel in X direction
	      if ((lack1 / slope1) <= (lack2 / slope2))
		{ // line passes first through voxel in the Z direction
		  k = (k + dk);
		  dK++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		  i = (i + di);
		  dI++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		}
	      else
		{ // line passes first through voxel in the X direction
		  i = (i + di);
		  dI++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		  k = (k + dk);
		  dK++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		}
	    }
	  else
	    { // line does not pass through another voxel in X direction
	      k = (k + dk);
	      dK++;
	      next2 = getVoxelAtIndexes(i, j, k);
	      next3 = 0;
	      lack1 = (res - (step1 - lack1));
	      lack2 = (lack2 - step2);
	    }
	}
      else if (step2 > lack2)
	{ // line passes through one other voxel in X direction
	  i = (i + di);
	  dI++;
	  next2 = getVoxelAtIndexes(i, j, k);
	  lack1 = (lack1 - step1);
	  lack2 = (res - (step2 - lack2));
	  next3 = 0;
	}
      else
	{ // line does not pass through another voxel
	  next2 = 0;
	  next3 = 0;
	  lack1 = (lack1 - step1);
	  lack2 = (lack2 - step2);
	}
      j = (j + dj);
    }
  else //if ((fabs(deltaZ) >= fabs(deltaX)) && (fabs(deltaZ) >= fabs(deltaY)))
    { //step0 in Z direction
      principalDir = ((dk == 1) ? Z_PLUS_DIR : Z_MINUS_DIR);
      slope1 = fabs(deltaX / deltaZ);
      slope2 = fabs(deltaY / deltaZ);
      step0 = ((deltaZ > 0) ?
	       (zLow + ((getZIndexAtCoord(startZ) + 1) * res) - startZ) :
	       (startZ - (zLow + (getZIndexAtCoord(startZ) * res))));
      step1 = (step0 * slope1);
      step2 = (step0 * slope2);
      lack1 = ((deltaX >= 0) ?
	       (xLow + ((getXIndexAtCoord(startX) + 1) * res) - startX) :
	       (startX - (xLow + (getXIndexAtCoord(startX) * res))));
      lack2 = ((deltaY >= 0) ?
	       (yLow + ((getYIndexAtCoord(startY) + 1) * res) - startY) :
	       (startY - (yLow + (getYIndexAtCoord(startY) * res))));
      if (step1 > lack1)
	{ // line passes through another voxel in the X direction
	  if (step2 > lack2)
	    { // line also passes through another voxel in Y direction
	      if ((lack1 / slope1) <= (lack2 / slope2))
		{ // line passes first through voxel in the X direction
		  i = (i + di);
		  dI++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		  j = (j + dj);
		  dJ++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		}
	      else
		{ // line passes first through voxel in the Y direction
		  j = (j + dj);
		  dJ++;
		  next2 = getVoxelAtIndexes(i, j, k);
		  lack2 = (res - (step2 - lack2));
		  i = (i + di);
		  dI++;
		  next3 = getVoxelAtIndexes(i, j, k);
		  lack1 = (res - (step1 - lack1));
		}
	    }
	  else
	    { // line does not pass through another voxel in Y direction
	      i = (i + di);
	      dI++;
	      next2 = getVoxelAtIndexes(i, j, k);
	      next3 = 0;
	      lack1 = (res - (step1 - lack1));
	      lack2 = (lack2 - step2);
	    }
	}
      else if (step2 > lack2)
	{ // line passes through one other voxel in Y direction
	  j = (j + dj);
	  dJ++;
	  next2 = getVoxelAtIndexes(i, j, k);
	  lack1 = (lack1 - step1);
	  lack2 = (res - (step2 - lack2));
	  next3 = 0;
	}
      else
	{ // line does not pass through another voxel
	  next2 = 0;
	  next3 = 0;
	  lack1 = (lack1 - step1);
	  lack2 = (lack2 - step2);
	}
      k = (k + dk);
    }
  step1 = (res * slope1);
  step2 = (res * slope2);
  return next1;
}

/***********************************************************************/

/*! Public method setCenter causes the box to scroll so that the
  cartesian point (x, y, z) lies inside the voxel at the center of the
  box grid.  The xLowIndex, yLowIndex, zLowIndex, xLow, yLow, zLow,
  xHi, yHi, zHi, centerX, centerY, and centerZ of the box change when
  the center moves. Also, the voxels of the actual grid that represent
  the voxels entering (and leaving) the box grid are reinitialized.
  xDiff, yDiff, and zDiff are set to the number of voxels the center
  has moved in the X, Y, and Z directions; the values may be positive,
  zero, or negative. */
/*

If the center point has made a jump larger than xSide, ySide, or zSide
in the X, Y, or Z directions, respectively, this reinitializes all the
voxels in the actual grid. Otherwise, it reinitializes the voxels in a
portion of the actual grid consisting of three slabs, (1) one parallel
to the YZ plane bounded by the old xLowIndex and the new xLowIndex,
(2) one parallel to the ZX plane bounded by the old yLowIndex and the
new yLowIndex, and (3) one parallel to the XY plane bounded by the old
zLowIndex and the new zLowIndex. If xLowIndex, yLowIndex, or zLowIndex
wraps around to get to its new value, the corresponding slab divides
into two slabs.  If xLowIndex, yLowIndex, or zLowIndex does not
change, there is no corresponding rectangle.

Where the slabs intersect, this is initializing the same voxels two
or three times. It might be worth the trouble to change the code to
avoid doing that, but the code would then be messy.

The alternative calculation xLow = ((xRaw - (xSide - 0.5)) * res);
is correct and will avoid error buildup, but is obscure. The
calculation being used, xLow = (xLow + (*xDiff * res)); is clear
but will have error buildup (but that should be so small as to never
be significant). Similarly for yLow and zLow.

*/

void ScrollingBox::setCenter( /*   ARGUMENTS                                 */
 double x,                    /*!< X coordinate of new center point          */
 double y,                    /*!< Y coordinate of new center point          */
 double z,                    /*!< Y coordinate of new center point          */
 int * xDiff,                 /*!< no. of voxels center moved in X, set here */
 int * yDiff,                 /*!< no. of voxels center moved in Y, set here */
 int * zDiff)                 /*!< no. of voxels center moved in Z, set here */
{
  int n;
  int index;        // 1D index of a voxel in voxels array
  int xIndex;       // 3D actual X index of a voxel
  int yIndex;       // 3D actual Y index of a voxel
  int zIndex;       // 3D actual Z index of a voxel
  int newXLowIndex;
  int newYLowIndex;
  int newZLowIndex;
  int stop;
  int xRaw;         // unbounded actual X index of a voxel
  int yRaw;         // unbounded actual Y index of a voxel
  int zRaw;         // unbounded actual Z index of a voxel

  xRaw = getActualRawXIndex(x);
  yRaw = getActualRawYIndex(y);
  zRaw = getActualRawZIndex(z);
  *xDiff = (xRaw - getActualRawXIndex(centerX));
  *yDiff = (yRaw - getActualRawYIndex(centerY));
  *zDiff = (zRaw - getActualRawZIndex(centerZ));
  newXLowIndex = ((xLowIndex + *xDiff) % xSide);
  if (newXLowIndex < 0)
    newXLowIndex = (newXLowIndex + xSide);
  newYLowIndex = ((yLowIndex + *yDiff) % ySide);
  if (newYLowIndex < 0)
    newYLowIndex = (newYLowIndex + ySide);
  newZLowIndex = ((zLowIndex + *zDiff) % zSide);
  if (newZLowIndex < 0)
    newZLowIndex = (newZLowIndex + zSide);
  if ((*xDiff >= xSide) || (*xDiff <= -xSide) ||
      (*yDiff >= ySide) || (*yDiff <= -ySide) ||
      (*zDiff >= zSide) || (*zDiff <= -zSide))
    { // clear the entire box
      for (index = 0; index < numberOfVoxels; index++)
	{
	  voxels[index].init();
	}
    }
  else
    {
      if (*xDiff > 0)
	{ // center voxel moved in the positive x direction
	  stop = *xDiff;
	  xIndex = xLowIndex;
	}
      else
	{ // center voxel moved in the negative x direction or not at all
	  stop = -*xDiff;
	  xIndex = newXLowIndex;
	}
      for (n = 0; n < stop; n++) // this does nothing if *xDiff is zero
	{ // initialize voxels in slab parallel to YZ plane
	  for (yIndex = 0; yIndex < ySide; yIndex++)
	    {
	      for (zIndex = 0; zIndex < zSide; zIndex++)
		{
		  getVoxel3D(xIndex, yIndex, zIndex)->init();
		}
	    }
	  xIndex++;
	  if (xIndex >= xSide)
	    xIndex = (xIndex - xSide);
	}
      if (*yDiff > 0)
	{ // center voxel moved in the positive y direction
	  stop = *yDiff;
	  yIndex = yLowIndex;
	}
      else
	{ // center voxel moved in the negative y direction or not at all
	  stop = -*yDiff;
	  yIndex = newYLowIndex;
	}
      for (n = 0; n < stop; n++) // this does nothing if *yDiff is zero
	{ // initialize voxels in slab parallel to ZX plane
	  for (xIndex = 0; xIndex < xSide; xIndex++)
	    {
	      for (zIndex = 0; zIndex < zSide; zIndex++)
		{
		  getVoxel3D(xIndex, yIndex, zIndex)->init();
		}
	    }
	  yIndex++;
	  if (yIndex >= ySide)
	    yIndex = (yIndex - ySide);
	}
      if (*zDiff > 0)
	{ // center voxel moved in the positive z direction
	  stop = *zDiff;
	  zIndex = zLowIndex;
	}
      else
	{ // center voxel moved in the negative z direction or not at all
	  stop = -*zDiff;
	  zIndex = newZLowIndex;
	}
      for (n = 0; n < stop; n++) // this does nothing if *zDiff is zero
	{ // initialize voxels in slab parallel to XY plane
	  for (xIndex = 0; xIndex < xSide; xIndex++)
	    {
	      for (yIndex = 0; yIndex < ySide; yIndex++)
		{
		  getVoxel3D(xIndex, yIndex, zIndex)->init();
		}
	    }
	  zIndex++;
	  if (zIndex >= zSide)
	    zIndex = (zIndex - zSide);
	}
    }
  centerX = x;
  centerY = y;
  centerZ = z;
  xLowIndex = newXLowIndex;
  yLowIndex = newYLowIndex;
  zLowIndex = newZLowIndex;
  xLow = (xLow + (*xDiff * res));  //= ((xRaw - (xSide - 0.5)) * res);
  yLow = (yLow + (*yDiff * res));
  zLow = (zLow + (*zDiff * res));
  xHi  = (xLow + (xSide * res));
  yHi  = (yLow + (ySide * res));
  zHi  = (zLow + (zSide * res));
}

/***********************************************************************/

/*! \defgroup amSPMap_hh CVS Log Information
 * $Log: MOASTScrollingBox.hh,v $
 * Revision 1.1  2008/08/18 18:50:40  dr_steveb
 * Fixed files for distribution
 *
 * Revision 1.1  2008/01/24 13:49:55  tomrkramer
 * initial entry
 *
 */
