/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file aStarHeap.cc
  \brief implements aStar search with multiple goals using heap for open nodes

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.1 $
  $Date: 2006/11/16 13:57:22 $
  \endcode

  \author Tom Kramer

The aStarHeap.hh and aStarHeap.cc files define an abstract graph
class (AStarHeapGraph) and an A-star graph search class
(AStarHeapSearcher) with the following characteristics:

1. The graph may be built or changed dynamically during the search
   in several ways using the addKids, changeGoalCosts, removeKids,
   and/or useKids functions.
2. The user can add information to graph nodes by making a
   child class of nodeInfo.
3. The open nodes list is managed using the heap method.
4. The searcher will search simultaneously for multiple (N) goals.
5. Arcs with a negative cost between nodes are allowed and treated
   as though there were no arc. This is so that graphs can be handled in
   which every node has the same apparent number of children, but some
   are not actually children. Such graphs are often convenient to use.

Two different drivers (exampleAStarHeapA.cc and exampleAStarHeapB.cc)
for making a stand-alone executable are in src/tools.  Test files for
testing the executables exist but are not currently part of MOAST.

*/

#include "aStarHeap.hh"
#include <stdio.h>
#include <math.h>

/***********************************************************************/
/***********************************************************************/

/* AStarHeapGraph functions */

/***********************************************************************/

AStarHeapGraph::~AStarHeapGraph()
{
  int n;

  if (deleat == true)
    {
      for (n = 0; n < capacity; n++)
	{
	  if (nodes[n].kidsSize > 0)
	    {
	      delete [] nodes[n].kidsIndexes;
	      delete [] nodes[n].kidsCosts;
	    }
	}
      delete [] nodes;
    }
}

/***********************************************************************/

/* AStarHeapGraph::addKids

Returned Value: If an error occurs, this returns 0. Otherwise, it
  returns 1. An error occurs if:
  1. Any of numberAdded, numberNew, and numberOld is negative.
  2. In any of the newInfos, the index is incorrect.

Called By:
  external functions (such as main in exampleAStarHeapA.cc)

This may activate additional nodes of the graph. If it does, it also
adds arcs from dad to those new nodes. This may also add arcs from dad
to existing nodes. The memory for dad's old kidsIndexes and kidsCosts
is freed. Memory for dad's revised kidsIndexes and kidsCosts is
allocated.

If there are not enough nodes in the graph to accomodate the number
to be activated, increaseCapacity is called to make more nodes.

The new kids are built in nodes whose indexes are assigned in
numerical order, starting with the current size of the graph. The size
is increased in the next-to-last "for" loop. This number should
already be recorded as the value of "index" in the NodeInfo. It is
checked that this number is correct.

The additional kids in oldAdded must not already be kids of dad.
This is not currently checking for that. A check could be added.

*/

int AStarHeapGraph::addKids( /* ARGUMENTS                               */
 int dad,               /* index in nodes of node for which to add kids  */
 int numberAdded,       /* total number of kids added                    */
 int numberNew,         /* number of new kids added                      */
 NodeInfo ** newInfos,  /* array of pointers to NodeInfos for new kids   */
 double * newArcCosts,  /* array of arc costs from dad for new kids      */
 double * newGoalCosts, /* array of goal costs for new kids              */
 int * oldAdded,        /* number of existing kids added                 */
 double * oldArcCosts)  /* array of arc costs from dad for existing kids */
{
  int numberOld;      // number of existing nodes to be added to kids of dad
  int saveKidsSize;   // number of kids before additions
  int * saveIndexes;  // array of kids indexes before additions
  double * saveCosts; // array of kids costs before additions
  int n;
  int k;

  numberOld = (numberAdded - numberNew);
  if ((numberAdded <= 0) || (numberNew < 0) || (numberOld < 0))
    {
      fprintf(stderr, "negative number of nodes found in addKids\n");
      return 1;
    }
  if ((size + numberNew) > capacity)
    increaseCapacity(numberNew);
  if ((saveKidsSize = nodes[dad].kidsSize)) // set and test for nonzero
    {
      saveIndexes = nodes[dad].kidsIndexes;
      saveCosts = nodes[dad].kidsCosts;
    }
  for (n = 0; n < numberNew; n++)
    {
      if (newInfos[n]->index != (size + n))
	{
	  fprintf(stderr, "incorrect node index found in addKids\n");
	  return 1;
	}
      nodes[size + n].moreInfo = newInfos[n];
      nodes[size + n].open = AStarHeapNode::UNOPENED;
    }
  nodes[dad].kidsSize = (saveKidsSize + numberAdded);
  nodes[dad].kidsIndexes = new int[nodes[dad].kidsSize];
  nodes[dad].kidsCosts = new double[nodes[dad].kidsSize];
  for (n = 0; n < saveKidsSize; n++)
    {
      nodes[dad].kidsIndexes[n] = saveIndexes[n];
      nodes[dad].kidsCosts[n] = saveCosts[n];
    }
  for (k = 0; k < numberNew; k++, n++)
    {
      nodes[size].goalCost = newGoalCosts[n];
      nodes[dad].kidsIndexes[n] = size++;
      nodes[dad].kidsCosts[n] = newArcCosts[k];
    }
  for (k = 0; k < numberOld; k++, n++)
    {
      nodes[dad].kidsIndexes[n] = oldAdded[k];
      nodes[dad].kidsCosts[n] = oldArcCosts[k];
    }
  if (saveKidsSize)
    {
      delete [] saveIndexes;
      delete [] saveCosts;
    }
  return 0;
}

/***********************************************************************/

/* AStarHeapGraph::changeKidsGoalCosts

Returned Value: int (0)

Called By: external functions

ChangeKidsGoalCosts changes the goalCosts of some or all the kids of dad.

The numberChanged argument is the number of (existing) kids
of dad that are to be assigned new goal costs (corresponding to kids
indexes 0 through (numberNew - 1)). The new goal costs are given
in the newGoalCosts array.

The numberChanged should not be larger than the number of kids of dad.
This is not being checked.

*/

int AStarHeapGraph::changeKidsGoalCosts( /* ARGUMENTS                   */
 int dad,                  /* index in nodes of node to change kids for */
 int numberChanged,        /* number of kids to change                  */
 double * newGoalCosts)    /* array of new goalCosts of kids            */
{
  int n;
  
  for (n = 0; n < numberChanged; n++)
    nodes[nodes[dad].kidsIndexes[n]].goalCost = newGoalCosts[n];
  return 0;
}

/***********************************************************************/

/* AStarHeapGraph::increaseCapacity

Returned Value: int (0)

Called By:
  AStarHeapGraph::addKids

This function makes a new and larger array of nodes and copies the old
nodes into the front part of it. Since all array referencing is done
by index rather than by pointer, it does not matter that new memory is
being used.

The increase in capacity is the maximum of (1) the numberAdded and
(2) the minimum of the old capacity and 100000. The idea is to double
in capacity (or more, if necessary) any time an increase is needed,
until there are 100000 nodes, after which only 100000 nodes are added
each time (so that a significant amount of memory is not wasted). The
reason for doubling is to minimize the number of increases in capacity.

*/

int AStarHeapGraph::increaseCapacity( /* ARGUMENTS                     */
 int numberAdded)      /* minimum number by which to increase capacity */
{
  int increase;
  AStarHeapNode * oldNodes;
  int n;

  oldNodes = nodes;
  increase = ((capacity > 100000) ? 100000 : capacity);
  if (numberAdded > increase)
    increase = numberAdded;
  capacity = (capacity + increase);
  nodes = new AStarHeapNode[capacity];
  for (n = 0; n < size; n++)
    nodes[n].copyFrom(oldNodes + n);
  delete [] oldNodes;
  return 0;
}

/***********************************************************************/

/* AStarHeapGraph::removeKids

Returned Value: int (1)

Called By: external functions

This function removes all arcs from nodes[dad] by setting kidsSize to
0 and deleting kidsIndexes and kidsCosts.

kidsIndexes and kidsCosts must have been made using "new", since they
are deleted here.

The kids themselves are not affected by this function.

*/

int AStarHeapGraph::removeKids( /* ARGUMENTS                          */
 int dad)         /* index in nodes of node from which to remove kids */
{
  nodes[dad].kidsSize = 0;
  delete [] nodes[dad].kidsIndexes;
  delete [] nodes[dad].kidsCosts;
  nodes[dad].kidsIndexes = 0;
  nodes[dad].kidsCosts = 0;
  return 1;
}

/***********************************************************************/

/* AStarHeapGraph::useKids

Returned Value: int (1)

Called By: external functions

This removes the current kids of dad (if any) and replaces them with
the ones given by the arguments. The arrays specified in the arguments
are used as is, so they must be created with "new" in the implementing
system and not changed or deleted by the implementing system.

The kids themselves must already exist and are not affected by this
function.

*/

int AStarHeapGraph::useKids( /* ARGUMENTS                                 */
 int dad,                    /* index in nodes of node to change kids for */
 int kidsSize,               /* new number of kids                        */
 int * kidsIndexes,          /* new array of kids indexes in nodes        */
 double * kidsCosts)         /* new array of kids costs                   */
{
  if (nodes[dad].kidsSize > 0)
    {
      delete [] nodes[dad].kidsIndexes;
      delete [] nodes[dad].kidsCosts;
    }
  nodes[dad].kidsSize = kidsSize;
  nodes[dad].kidsIndexes = kidsIndexes;
  nodes[dad].kidsCosts = kidsCosts;
  return 1;
}

/***********************************************************************/
/***********************************************************************/

/* AStarHeapNode functions */

/***********************************************************************/

/* AStarHeapNode::~AStarHeapNode

This destructor (which does nothing in addition to default destruction)
is defined here only because objectcenter thought it was undefined
when it was in aStarHeap.hh. I

*/

AStarHeapNode::~AStarHeapNode() {}

/***********************************************************************/

/* AStarHeapNode::copyFrom

Returned Value: int
  If the "from" argument is 0, this returns 0. Otherwise, it returns 1.

Called By:
  AStarHeapGraph::increaseCapacity

This copies all of the data of the "from" node into "this" node.  The
kidsIndexes and kidsCosts arrays are used as given in the "from" node,
not made with "new".

*/

int AStarHeapNode::copyFrom( /* ARGUMENTS         */
 AStarHeapNode * from)       /* node to copy from */
{
  if (from != 0)
    {
      moreInfo = from->moreInfo;
      kidsSize = from->kidsSize;
      kidsIndexes = from->kidsIndexes;
      kidsCosts = from->kidsCosts;
      open = from->open;
      arrivedFrom = from->arrivedFrom;
      visitCost = from->visitCost;
      goalCost = from->goalCost;
      totalCost = from->totalCost;
      heapIndex = from->heapIndex;
    }
  return ((from == 0) ? 0 : 1);
}

/***********************************************************************/
/***********************************************************************/

/* AStarHeapSearcher functions */

/***********************************************************************/

/* AStarHeapSearcher::closeSearch

Returned Value: int (0)

Called By: external functions (e.g., main in exampleAStarHeapA.cc)

This closes the searcher. All it does is set the heap size to zero,
and even this is not really necessary.

*/

int AStarHeapSearcher::closeSearch() /* NO ARGUMENTS */
{
  heapSize = 0;
  return 0;
}

/***********************************************************************/

/* AStarHeapSearcher::cycleSearch

Returned Value: int (0)

Called By: external functions (e.g., main in exampleAStarHeapA.cc)

This removes the first (index 1) node on the heap (which is the
cheapest one) and processes the children of that node.

1. If a path to the goal known to be the cheapest path has been found,
this sets status[0] to 2 and sets cost[0].

2. Otherwise, if a path to the goal not known to be the cheapest has been
found:
  a. If status[0] is 0, this sets status[0] to 1 and sets cost[0].
  b. Otherwise, if status[0] is 1, this sets cost[0], which may
     or may not be less than previously (but will not be more).

3. Otherwise, this does not change the output data.

Once the cheapest path to the goal has been found or the heap is
empty, this does nothing.

*/

int AStarHeapSearcher::cycleSearch() /* NO ARGUMENTS */
{
  int i;           // index for kids of oldCheap
  double kidCost;  // total cost to visit i'th kid of oldCheap
  int kidNum;      // index in nodes of i'th kid of oldCheap
  int numKids;     // number of kids of oldCheap
  int oldCheap;    // index of cheap node when this starts

  if (heap[1] == goal)
    status[0] = 2;  
  else if (heapSize != 0)
    {
      if (heapCapacity < theGraph->capacity)
	{
	  nodes = theGraph->nodes;         // has been changed by addKids
	  increaseHeap(theGraph->capacity);
        }
      oldCheap = heap[1];
      removeCheap();
      nodes[oldCheap].open = AStarHeapNode::CLOSED; // old cheap node closed
      numKids = nodes[oldCheap].kidsSize;
      for (i = 0; i < numKids; i++)
	{
	  if ((nodes[oldCheap].kidsCosts)[i] < 0)
	    continue; // skip disconnected kids
	  kidNum = (nodes[oldCheap].kidsIndexes)[i];
	  kidCost = (nodes[oldCheap].visitCost +
		     (nodes[oldCheap].kidsCosts)[i]);
	  processNode(kidNum, oldCheap, kidCost);
	  if (kidNum == goal)
	    {
	      cost[0] = nodes[kidNum].visitCost;
	      status[0] = 1; // status not 2 or would not get here
	    }
	}
    }
  return 0;
}

/***********************************************************************/

/* AStarHeapSearcher::getNextNode

Returned Value: int
This returns the index in nodes of the cheapest open node remaining.
If there are no open nodes remaining, this returns -1.

Called By: external functions

*/

int AStarHeapSearcher::getNextNode()
{
  return ((heapSize == 0) ? -1 : heap[1]);
}

/***********************************************************************/

//! AStarHeapSearcher::heapDown causes misplaced node to sink to correct place
/*
Returned Value: None

Called By:  AStarHeapSearcher::removeCheap

Conceptually, the heap consists of nodes partially ordered by cost and
linked in a binary tree. In the (standard) representation used here,
however, the nodes are kept in an array called Nodes, and the heap is
a linear array of indexes in Nodes. The binary tree is implicitly
linked by having the kids of the node pointed to by heap[N] be pointed
to by heap[2N] and heap[2N+1].

The conceptual idea of the algorithm is simple: the node that was at
the top of the heap sinks by changing positions with the cheaper of
its kids repeatedly until neither kid is cheaper than it is. Because
the representation differs greatly from the concept, understanding
how the implementation works is tricky.

In order to find a node in the heap easily, the heap index which
stores a node's index is recorded in the heapIndex attribute of the
node. This must be updated whenever the node's position in the heap
changes.

The three loop variables are heap indexes:
hex (a heap index currently being examined) - conceptually a dad node
kid1 (2*hex) - conceptually one kid node
kid2 (2*hex+1) - conceptually the other kid node

The initial value of hex is 1, the heap index of the top of the heap.

Each time around the loop:

1. If the cost of the original top node is larger than the smaller of
the costs of the two kids (let kidC be the heap index of the cheaper
kid):
  A. The value in heap[hex] is changed to the value in heap[kidC].
     This represents the kid node moving to where the dad was.
  B. The heapIndex recorded in nodes[heap[kidC]] is changed to hex.
     This updates the node's knowledge of where it is in the heap.
  C. hex is changed to kidC to get ready for the next go around the loop.
2. Otherwise, the original top node has stopped sinking, so:
  A. The value in heap[hex] is set to the Nodes index of the original top node.
  B. The heapIndex recorded in the original top node is set to hex.
  C. The function returns.

The loop terminates if there are no kids or only one kid.  After the
loop exits, a check is made for only one kid existing and a swap is
made if appropriate.

Although the representation of a kid's position is changed each time a
dad sinks, the representation of the dad's new position is not changed
each time. We know the dad will be where the kid was, and we know the
value recorded in the kid's former position would be the Nodes index
of the node that started at the top of the heap, so we don't have to
record it. When the dad stops sinking, its final position is recorded.
Not recording the intermediate positions saves time but makes the code
a little more complex.

The heap is ordered by totalCost (visitCost + goalCost).

*/

void AStarHeapSearcher::heapDown() /* NO ARGUMENTS */
{
  int hex;  // index in heap of upper node (changed in loop)
  int kid1; // (2*hex) = heap index of left lower node (changed in loop)
  int kid2; // (kid1+1) = heap index of right lower node (changed in loop)
  double cost0; // totalCost of node at top of heap (fixed)
  double cost1; // totalCost of kid1 (changed in loop)
  double cost2; // totalCost of kid2 (changed in loop)
  int topNodeIndex; // Nodes index of node at top of heap (fixed)

  topNodeIndex = heap[1];
  cost0 = nodes[topNodeIndex].totalCost;

  for (hex = 1; ((kid1 = (2 * hex)) < heapSize); ) // always 2 kids
    {
      kid2 = (kid1 + 1);
      cost1 = nodes[heap[kid1]].totalCost;
      cost2 = nodes[heap[kid2]].totalCost;
      if (cost1 < cost2) // kid1 cheaper
        {
          if (cost0 > cost1)
            { // original top node must sink more
              heap[hex] = heap[kid1];
              nodes[heap[kid1]].heapIndex = hex;
              hex = kid1;
            }
          else
            { // original top node goes in hex position
              heap[hex] = topNodeIndex;
              nodes[topNodeIndex].heapIndex = hex;
              return;
            }
        }
      else // if (cost2 <= cost1) kid2 cheaper
        {
          if (cost0 > cost2)
            { // original top node must sink more
              heap[hex] = heap[kid2];
              nodes[heap[kid2]].heapIndex = hex;
              hex = kid2;
            }
          else
            { // original top node goes in hex position
              heap[hex] = topNodeIndex;
              nodes[topNodeIndex].heapIndex = hex;
              return;
            }
        }
    }
  if (kid1 == heapSize)    // only one kid
    {
      if (cost0 > nodes[heap[kid1]].totalCost) // top falls to bottom
        { // swap hex and kid1
          nodes[heap[kid1]].heapIndex = hex;
          nodes[topNodeIndex].heapIndex = kid1;
          heap[hex] = heap[kid1];
          heap[kid1] = topNodeIndex;
        }
      else // no swap, but record dad position
        {
          heap[hex] = topNodeIndex;
          nodes[topNodeIndex].heapIndex = hex;
        }
    }
  else // no kids, top has fallen to bottom, record dad position
    {
      heap[hex] = topNodeIndex;
      nodes[topNodeIndex].heapIndex = hex;
    }
}

/***********************************************************************/

//! AStarHeapSearcher::heapUp causes misplaced node to rise to correct place
/*
Returned Value:  None

Called By:  AStarHeapSearcher::processNode

See the documentation of heapDown. This function is simpler.
Conceptually, the kid0 node keeps changing places with its dad node
until the kid0 node is cheaper than the dad node. As in heapDown, the
position of the target node index (heap[kid0] in this case) is not
recorded until it stops moving.

The heap is ordered by totalCost (visitCost + goalCost).

*/

void AStarHeapSearcher::heapUp( /* ARGUMENTS                           */
 int kid0)                      /*!< heap index of element being upped */
{
  int kid;           // heap index of lower node (changed in loop)
  int dad;           // heap index of upper node (changed in loop)
  int kid0NodeIndex; // Nodes index of original kid (fixed)
  double cost0;      // total cost of kid0 (fixed)
  
  kid0NodeIndex = heap[kid0];
  cost0 = nodes[kid0NodeIndex].totalCost;
  for (kid = kid0; ((dad = (kid / 2)) > 0); kid = dad)
    {
      if (cost0 < nodes[heap[dad]].totalCost)
        { // dad sinks down
          heap[kid] = heap[dad];
          nodes[heap[dad]].heapIndex = kid;
        }
      else 
        { // kid has stopped rising, record position
          heap[kid] = kid0NodeIndex;
          nodes[kid0NodeIndex].heapIndex = kid;
          return;
        }
    }
  // if this point is reached, the node being upped is at the top
  heap[1] = kid0NodeIndex;
  nodes[kid0NodeIndex].heapIndex = 1;
}

/***********************************************************************/

//! AStarHeapSearcher::increaseHeap increases the capacity of the heap
/*
Returned Value: None

Called By: AStarHeapSearcher::cycleSearch

This makes a new heap with the given newCapacity and copies the old
heap into it. 

newCapacity must be larger than current capacity. This function is not
checking that.

This leaves garbage at the end of the new heap; that does not matter.

It will never happen that every node is on the heap. If it could
happen, the capacity of the heap would need to be one greater than the
number of nodes, since heap[0] is a dummy.

*/

void AStarHeapSearcher::increaseHeap( /* ARGUMENTS              */
 int newCapacity)                     /*!< new capacity of heap */
{
  int * oldHeap;
  int n;

  oldHeap = heap;
  heapCapacity = newCapacity;
  heap = new int[heapCapacity];
  for (n = 0; n <= heapSize; n++)
    heap[n] = oldHeap[n];
  delete [] oldHeap;
}

/***********************************************************************/

/* AStarHeapSearcher::initSearch initializes the searcher

Returned Value: int
If initialization occurs without error, this returns 0;
This prints an error message and returns 1 if:
1. The nodes of aGraph is NULL, or
2. the size of the kids of any node is negative, or
3. the size of aGraph is greater than its capacity.

Called By: external functions (e.g., main in exampleAStarHeapA.cc)

The arrivedFrom of the start node is set to -1 to serve as
a stop marker for reportSearch.

statusOut and costOut are both arrays of size 1.

This:
1. Saves (as the values of data members of the Searcher) the locations
   where the user wants to have search data recorded. These are the
   four arguments: statusesOut, costsOut, goalsFoundOut, optimaFoundOut.
2. Sets the value of the other data members of the Searcher that need setting.
3. Initializes the start node and places it on the open nodes list.

Note that the Searcher uses the caller's memory for status and cost.
This is safe because the Searcher writes their values but does not use
them otherwise. The Searcher does not use the caller's memory for
heapSize.

*/

int AStarHeapSearcher::initSearch( /* ARGUMENTS                         */
 AStarHeapGraph * aGraph, /* the graph to search in                     */
 int start,               /* index in nodes of aGraph of start node     */
 int goalIn,              /* index in nodes of aGraph of goal node      */
 int * opensOut,          /* 0=open list empty, non-zero=list not empty */
 int * statusOut,         /* status of reaching goal                    */
 double * costOut)        /* cost of reaching goal                      */
{
  int n;

  goal = goalIn;
  heapCapacity = aGraph->capacity;
  heap = new int[heapCapacity];
  heap[0] = -1;
  heapSize = 0;
  heapSizeOut = opensOut;
  *heapSizeOut = heapSize;
  status = statusOut;
  status[0] = 0;
  cost = costOut;
  cost[0] = 0;
  theGraph = aGraph;
  nodes = theGraph->nodes;
  if (nodes == 0)
    {
      fprintf(stderr, "nodes array is null");
      return 1;
    }
  if (theGraph->size > theGraph->capacity)
    {
      fprintf(stderr, "size greater than capacity\n");
      return 1;
    }
  for (n = 0; n < theGraph->size; n++)
    {
      if (nodes[n].kidsSize < 0)
	{
	  fprintf(stderr, "bad kids size\n");
	  return 1;
	}
      nodes[n].open = AStarHeapNode::UNOPENED;
    }
  processNode(start, -1, 0);
  return 0;
}

/***********************************************************************/

//! AStarHeapSearcher::processNode processes one node
/*
Returned Value: None

Called By:
  AStarHeapSearcher::cycleSearch
  AStarHeapSearcher::initSearch

This function processes the node (call it Noe) whose index in the
nodes being searched is numNode. Outline numbers here match comment
numbers below.

A. If Noe is not unopened (must be open or closed):
  1. If Noe's currently recorded cost is greater than cost:
     a. If Noe is open, do c through f.
     b. Otherwise (must be closed), print an error message.
     c. Set Noe's visitCost to cost.
     d. Set Noe's totalCost to (cost + goalCost)
     e. Set Noe's arrivedFrom to from.
     f. HeapUp from Noe in the open nodes heap.
  2. Otherwise, do nothing with Noe.
B. Else if Noe has not yet been opened:
  1. Mark Noe open.
  2. Set Noe's visitCost to cost.
  3. Set Noe's totalCost to (cost + goalCost).
  4. Set Noe's arrivedFrom to from.
  5. Insert Noe at the end of the open nodes heap.
  6. HeapUp from Noe in the open nodes heap.

Item A1b, encountering a closed node at a cost less than previously,
should never happen with the Euclidean distance heuristic, as follows.
Let E(XY) denote the Euclidean distance cost from X to Y.
Suppose, T is the target, node A is closed (as a result of having become
the least totalCost node) and a less expensive path to A is found from
node B when B becomes the least totalCost node. Then:
cost(B) < cost(A) - E(BA)
cost(B) + E(BA) + E(AT) < cost(A) - E(BA) + E(BA) + E(AT)
cost(B) + E(BA) + E(AT) < cost(A) + E(AT)
but the right side is the total cost of A, and by the triangle inequality:
E(BA) + E(AT) < E(BT), so
cost(B) + E(BT) < cost(B) + E(BA) + E(AT) = totalCost(A)
but cost(B) + E(BT) is the totalCost of B, so
totalCost(B) < totalCost(A)
Thus, B would have been the least cost node before A, a contradiction to
assumption that A was the least cost node before B.

It is not clear whether an example can be constructed with an
admissible heuristic in which a closed node can be reached at lower
cost than previously.

With the Euclidean distance heuristic, rounding error may cause a
closed node to be reached at infinitesimally lower cost. Therefore,
an error message is not printed in such cases.

*/

void AStarHeapSearcher::processNode( /* ARGUMENTS                           */
 int numNode,            /*!< index in nodes of Noe                         */
 int from,               /*!< index in nodes of node from which Noe visited */
 double cost)            /*!< cost of getting to Noe via current path       */
{
  double totalCost;
  if (nodes[numNode].open > AStarHeapNode::UNOPENED)             // A
    {
      if (nodes[numNode].visitCost > cost)                       // A1
	{
	  totalCost = (cost + nodes[numNode].goalCost);
	  if (nodes[numNode].open == AStarHeapNode::OPEN)        // A1a
	    {
	      nodes[numNode].visitCost = cost;                   // A1c
	      nodes[numNode].totalCost = totalCost;              // A1d
	      nodes[numNode].arrivedFrom = from;                 // A1e
	      heapUp(nodes[numNode].heapIndex);                  // A1f
	    }
	  else if (nodes[numNode].visitCost > (cost * 1.000001)) // A1b
	    fprintf(stderr, "Closed node found at lower cost\n");
	}
    }
  else /* (nodes[numNode].open == AStarHeapNode::UNOPENED) */    // B
    {
      totalCost = (cost + nodes[numNode].goalCost);
      nodes[numNode].open = AStarHeapNode::OPEN;                 // B1
      nodes[numNode].visitCost = cost;                           // B2
      nodes[numNode].totalCost = totalCost;                      // B3
      nodes[numNode].arrivedFrom = from;                         // B4
      heapSize++;                                                // B5
      *heapSizeOut = heapSize;
      nodes[numNode].heapIndex = heapSize;
      heap[heapSize] = numNode;
      heapUp(heapSize);                                          // B6
    }
}

/***********************************************************************/

//! AStarHeapSearcher::removeCheap removes cheapest member of the open nodes
/*
Returned Value: None

Called By:
  AStarHeapSearcher::cycleSearch

This is used to remove the cheapest member of the open nodes heap,
which is at the top of the heap (index 1).

The removal is done by putting the current last element of the heap
into the first place, reducing the size of the heap by one, and calling
heapDown.

*/

void AStarHeapSearcher::removeCheap() /* NO ARGUMENTS */
{
  heap[1] = heap[heapSize];
  heapSize--;
  *heapSizeOut = heapSize;
  heapDown();
}

/***********************************************************************/

/* AStarHeapSearcher::reportSearch

Returned Value: int
If the length of the path is less than maxLength, this returns the
length of the path. Otherwise, it returns -1.

Called By: external functions (e.g., main in exampleAStarHeapA.cc)

The arrivedFrom of the start node is set to -1 in initSearch, so
testing for -1 tests for the start node.

This is not reporting the cost because the caller will have direct
access to it.

*/

int AStarHeapSearcher::reportSearch( /* ARGUMENTS                      */
 int maxLength,    /* maximum length of path (capacity of path array)  */
 int * path)       /* array of index numbers giving the path, set here */
{
  int i;
  int n;
  int length;

  length = 1;
  for (i = goal; nodes[i].arrivedFrom != -1; i = nodes[i].arrivedFrom)
    (length)++;
  i = goal;
  if (length >= maxLength)
    return -1;
  for (n = (length - 1); n > -1; n--) 
    {
      (path)[n] = i;
      i = nodes[i].arrivedFrom;
    }
  return length;
}

/***********************************************************************/

/*
  $Log: aStarHeap.cc,v $
  Revision 1.1  2006/11/16 13:57:22  tomrkramer
  Initial version.

*/
