/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file   acmRandom.c

  \brief  A 32-bit random number generator that implements the minimal
  standard that ports to all systems where the maximum integer size is
  2^32 -1 or larger.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.2 $
  $Date: 2007/02/01 18:20:21 $
  \endcode

  \author Fred Proctor
*/

#include <math.h>		/* sqrt, log */
#include "acmRandom.h"

#ifdef __cplusplus
extern "C" {
#if 0
} /* matches brace above, to un-confuse autoformatting text editors */
#endif /* if 0 */
#endif /* ifdef __cplusplus */

#define ACM_MODULUS 2147483647
#define ACM_N (ACM_MODULUS - 1)
#define ACM_INITIAL_SEED 65531

static double ACM_N_INV = 1.0 / ((double) ACM_N);

#define NUM_ACMS 3	      /* how many different sets we support */
static AcmRandomStruct acms[NUM_ACMS] = 
  {{0, 16807, 12773, 2836, 993551},
   {1, 48271, 44488, 3399, 366914},
   {2, 69621, 30845, 23902, 650696}};

/*
  call this before using acm. This picks the first generator and sets
  the seed to a reasonable initial value but you will want to call
  acmRandomSeed() next with your seed if you don't want the same
  series each time.
*/
int acmRandomInit(AcmRandomStruct * acm)
{
  *acm = acms[0];

  return 0;
}

/*
  seeds the random number generator with your number 's', which can
  be obtained from the environment, e.g., time of day, so that each
  time you run your program the series will be different.
*/
int acmRandomSeed(AcmRandomStruct * acm, long int s)
{
  if (s <= 0) s = 1;
  acm->seed = s % ACM_MODULUS;

  return 0;
}

/*
  switches between each of the three random number generators
 */
int acmRandomSelect(AcmRandomStruct * acm, int which)
{
  *acm = acms[which % NUM_ACMS];

  return 0;
}

/*
  returns the total possible random numbers generated
*/
long int acmRandomTotal(void)
{
  return ACM_N;
}

/*
  returns a random number in the range [0 .. acmRandomTotal()-1],
  inclusive
*/
long int acmRandom(AcmRandomStruct * acm)
{
  long int lo, hi, test;

  hi = acm->seed / acm->Q;
  lo = acm->seed % acm->Q;
  test = acm->A * lo - acm->R * hi;
  
  if (test > 0)
    acm->seed = test;
  else
    acm->seed = test + ACM_MODULUS;

  return acm->seed - 1;
}

/*
  returns a double in the range [0.000000 .. 0.99999X], which is
  simply acmRandom() / acmRandomTotal(). You can multiply this
  by an integer M to get random numbers in the range [0..M-1],
  although they won't be perfectly balanced unless 
  acmRandomTotal() is an exact multiple of M. To get balanced numbers
  in a range [0..M-1] inclusive, call acmRandomRange(, M).
*/
double acmRandomReal(AcmRandomStruct * acm)
{
  return ((double) acmRandom(acm)) * ACM_N_INV;
}

/*
  returns a random number in the range [0 .. range-1] inclusive,
  balanced so that over the long term all numbers are equally probable.
  This is accomplished by throwing some random generates away, perhaps
  up to almost half depending on 'range'. 
*/
long int acmRandomRange(AcmRandomStruct * acm, long int range)
{
  long int m, n, g;

  if (range < 2) return 0;

  m = ACM_N / range;		/* how many ranges in N, e.g., 4 in
				   byte/card example */
  n = m * range;		/* actual number that are filled, e.g.,
				   208 in byte/card example */

  /* pick a random number that lies in the filled section */
  do {
    g = acmRandom(acm);
  } while (g >= n);
  /* now g is 0..n-1, uniformly, e.g., 0..207 */

  return g / m;	   /* returned val is 0..range-1, balanced */
}

/* some variates */

int acmNormalInit(AcmNormalStruct * nrm)
{
  int retval;

  nrm->hasLast = 0;

  retval = acmRandomInit(&nrm->u1);
  if (0 != retval) return retval;
  retval = acmRandomSelect(&nrm->u1, 1);
  if (0 != retval) return retval;

  retval = acmRandomInit(&nrm->u2);
  if (0 != retval) return retval;
  retval = acmRandomSelect(&nrm->u2, 2);
  if (0 != retval) return retval;

  return 0;
}

int acmNormalSeed(AcmNormalStruct * nrm, long int s1, long int s2)
{
  int retval;

  retval = acmRandomSeed(&nrm->u1, s1);
  if (0 != retval) return retval;

  retval = acmRandomSeed(&nrm->u2, s2);
  if (0 != retval) return retval;

  return 0;
}

double acmNormal(AcmNormalStruct * nrm, double mu, double sigma)
{
  double u1, u2;
  double v1, v2;
  double w, y;
  double x1, x2;

  if (nrm->hasLast) {
    nrm->hasLast = 0;
    return mu + sigma * nrm->last;
  }

  do {
    u1 = acmRandomReal(&nrm->u1);
    u2 = acmRandomReal(&nrm->u2);
    v1 = 2.0 * u1 - 1.0;
    v2 = 2.0 * u2 - 1.0;
    w = v1 * v1 + v2 * v2;
  } while (w > 1.0 || w <= 0.0);

  y = sqrt(-2.0 * log(w) / w);
  x1 = v1 * y;
  x2 = v2 * y;

  nrm->hasLast = 1;
  nrm->last = x2;

  return mu + sigma * x1;
}

int acmExponentialInit(AcmExponentialStruct * ex)
{
  int retval;

  retval = acmRandomInit(&ex->u);
  if (0 != retval) return retval;
  retval = acmRandomSelect(&ex->u, 1);
  if (0 != retval) return retval;

  return 0;
  
}

int acmExponentialSeed(AcmExponentialStruct * ex, long int s)
{
  return acmRandomSeed(&ex->u, s);
}

double acmExponential(AcmExponentialStruct * ex, double mu)
{
  double x;

  do {
    x = acmRandomReal(&ex->u);
  } while (x <= 0.0);

  return -mu * log(x);
}

#ifdef __cplusplus
#if 0
{    /* matches brace above, to un-confuse autoformatting text editors */
#endif /* if 0 */
} /* matches extern "C" */
#endif /* ifdef __cplusplus */

/*
  This random number generator is the one described in "Random Number
  Generators: Good Ones are Hard to Find," Stephen K. Park and Keith
  W. Miller, Communications of the ACM, Volume 31, Number 10, October
  1988.
  
  This generator is proposed as the "minimal standard" that will port
  to all systems where the maximum integer size is 2^32 - 1 or larger.

  The generating function is z(n+1) = a*z mod m, where m is the
  modulus 2^32 - 1. It generates all numbers in the range 1..m-1,
  or 1 to 2147483646. 0 and the modulus 2^32-1 are never generated, so
  two numbers out of the 2^32 possible in the four bytes of a long int
  are not generated, and 2^32-2 are generated.

  The function is shifted so that numbers in the range 0..2^32-3 are
  returned by acmRandom() for use. A normalized function
  acmRandomReal() is provided that returns a double, the shifted
  integer divided by 2^32-2, resulting in a floating point range
  [0.0000 .. 0.99999X], where 0.99999X is (2^32-3)/(2^32-2). This
  can be multiplied by an integer M and rounded to produce random
  integers in the range 0..M-1 with almost balanced distribution.

  With a random number generator that generates N possible numbers,
  to pick M numbers from that gives an imbalance if N is not a multiple
  of M.

  (N mod M) get (N div M) + 1 of the total N, and the remainder,
  M - (N mod M), get one less, (N div M).

  Consider the random generation of a playing card, for example, using
  a random byte generator. Here N = 256 and M = 52.  256 / 52 = 4
  remainder 48, and 48 of the cards are selected 5 times for a
  probability of 5/256 each, and 4 are selected 4 times for a
  probability of 4/256 each. Each of those unlucky 4 cards is 20% less
  likely to come up than each of the lucky 48 cards.

  For a 32-bit random number generator and 52 cards, N = 2^32 and
  M = 52. 2^32 / 52 = 82595524 remainder 48. The lucky 48 are selected 
  82595525 times and the unlucky 4 are selected 82595524 times,
  for probabilities 0.01923076925 and 0.01923076902, respectively.
  These are very close but not perfectly balanced.

  For this generator, N = 2^32-2 and M = 52. 2^32-2 / 52 = 82595524
  remainder 46. The lucky 46 are selected 82595525 times and the
  unlucky 6 are selected 82595524 times, for probabilities
  0.01923076926 and 0.01923076902, respectively.  These are very
  close but not perfectly balanced. 
*/

/*
  Modification history:

  $Log: acmRandom.c,v $
  Revision 1.2  2007/02/01 18:20:21  proctor
  Added normal and exponential variates to random number generator; updated
  simpleSimWare to use ground vehicle bot.

*/

