/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*!
  \file dijkTabs.cc
  \brief implements Dijkstra graph search with open list managed by tabs

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Tom Kramer

See the documentation of dijkTabs.hh for an overview of this search routine.

*/

#include "dijkTabs.hh"
#include <stdio.h>
#include <math.h>      // sqrt

/***********************************************************************/
/***********************************************************************/

/* DijkTabsGraph functions */

/***********************************************************************/

/* DijkTabsGraph::~DijkTabsGraph

This destructor deletes the arrays of kidIndexes and kidsCosts.

*/

DijkTabsGraph::~DijkTabsGraph()
{
  int n;
  for (n = 0; n < capacity; n++)
    {
      if (nodes[n].kidsSize > 0)
	{
	  delete [] nodes[n].kidsIndexes;
	  delete [] nodes[n].kidsCosts;
	}
    }
  delete [] nodes;
}

/***********************************************************************/

//! DijkTabsGraph::addKids adds kids to the dad node
/*
Returned Value: If an error occurs, this returns 0, otherwise, it
  returns 1. An error occurs if:
  1. Any of numberAdded, numberNew, and numberOld is negative.
  2. In any of the newInfos, the index is incorrect.

Called By:
  DijkTabsSearcher::cycleSearch()

This may activate additional nodes of the graph. If it does, it also
adds arcs from dad to those new nodes. This may also add arcs from dad
to existing nodes.  The memory for dad's old kidsIndexes and kidsCosts
is freed. Memory for dad's revised kidsIndexes and kidsCosts is
allocated.

If there are not enough nodes in the graph to accomodate the number
to be activated, increaseCapacity is called to make more nodes.

The new kids are built in nodes whose indexes are assigned in
numerical order, starting with the current size of the graph. The size
is increased in the next-to-last "for" loop. This number should
already be recorded as the value of "index" in the NodeInfo. It is
checked that this number is correct.

The additional kids in oldAdded must not already be kids of dad.
This is not currently checking for that. A check could be added.

*/

int DijkTabsGraph::addKids( /* ARGUMENTS                                 */
 int dad,               /* index in nodes of node to which to add kids   */
 int numberAdded,       /* total number of kids added                    */
 int numberNew,         /* number of new kids added                      */
 NodeInfo ** newInfos,  /* array of pointers to NodeInfos for new kids   */
 double * newCosts,     /* array of arc costs from dad for new kids      */
 int * oldAdded,        /* number of existing kids added                 */
 double * oldCosts)     /* array of arc costs from dad for existing kids */
{
  int numberOld;
  int saveKidsSize;
  int * saveIndexes;
  double * saveCosts;
  int n;
  int k;

  numberOld = (numberAdded - numberNew);
  if ((numberAdded <= 0) || (numberNew < 0) || (numberOld < 0))
    {
      fprintf(stderr, "negative number of nodes found in addKids\n");
      return 1;
    }
  if ((size + numberNew) > capacity)
    increaseCapacity(numberNew);
  if ((saveKidsSize = nodes[dad].kidsSize)) // set and test if non-zero
    {
      saveIndexes = nodes[dad].kidsIndexes;
      saveCosts = nodes[dad].kidsCosts;
    }
  for (n = 0; n < numberNew; n++)
    {
      if (newInfos[n]->index != (size + n))
	{
	  fprintf(stderr, "incorrect node index found in addKids\n");
	  return 1;
	}
      nodes[size + n].moreInfo = newInfos[n];
      nodes[size + n].open = DijkTabsNode::UNOPENED;
    }
  nodes[dad].kidsSize = (saveKidsSize + numberAdded);
  nodes[dad].kidsIndexes = new int[nodes[dad].kidsSize];
  nodes[dad].kidsCosts = new double[nodes[dad].kidsSize];
  for (n = 0; n < saveKidsSize; n++)
    {
      nodes[dad].kidsIndexes[n] = saveIndexes[n];
      nodes[dad].kidsCosts[n] = saveCosts[n];
    }
  for (k = 0; k < numberNew; k++, n++)
    {
      nodes[dad].kidsIndexes[n] = size++;
      nodes[dad].kidsCosts[n] = newCosts[k];
    }
  for (k = 0; k < numberOld; k++, n++)
    {
      nodes[dad].kidsIndexes[n] = oldAdded[k];
      nodes[dad].kidsCosts[n] = oldCosts[k];
    }
  if (saveKidsSize)
    {
      delete [] saveIndexes;
      delete [] saveCosts;
    }
  return 0;
}

/***********************************************************************/

//! DijkTabsGraph::increaseCapacity increases capacity of a graph's nodes array
/*
Returned Value: int (1)

Called By:
  DijkTabsGraph::addKids

This function makes a new and larger array of nodes and copies the old
nodes into the front part of it. Since all array referencing is done
by index rather than by pointer, it does not matter that new memory is
being used.

The increase in capacity is the maximum of (1) the numberAdded and
(2) the minimum of the old capacity and 100000. The idea is to double
in capacity (or more, if necessary) any time an increase is needed,
until there are 100000 nodes, after which only 100000 nodes are added
each time (so that a significant amount of memory is not wasted). The
reason for doubling is to minimize the number of increases in capacity.

*/

int DijkTabsGraph::increaseCapacity( /* ARGUMENTS                      */
 int numberAdded)    /*!< minimum number by which to increase capacity */
{
  int increase;
  DijkTabsNode * oldNodes;
  int n;

  oldNodes = nodes;
  increase = ((capacity > 100000) ? 100000 : capacity);
  if (numberAdded > increase)
    increase = numberAdded;
  capacity = (capacity + increase);
  nodes = new DijkTabsNode[capacity];
  for (n = 0; n < size; n++)
    nodes[n].copyFrom(oldNodes + n);
  delete [] oldNodes;
  return 1;
}

/***********************************************************************/

//! DijkTabsGraph::removeKids removes all kids from the dad node
/*
Returned Value: int (1)

Called By: external functions

This function removes all arcs from nodes[dad] by setting kidsSize to
0 and deleting kidsIndexes and kidsCosts.

kidsIndexes and kidsCosts must have been made using "new", since they
are deleted here.

The kids themselves are not affected by this function.

*/

int DijkTabsGraph::removeKids( /* ARGUMENTS                                 */
 int dad)             /*!< index in nodes of node from which to remove kids */
{
  nodes[dad].kidsSize = 0;
  delete [] nodes[dad].kidsIndexes;
  delete [] nodes[dad].kidsCosts;
  nodes[dad].kidsIndexes = 0;
  nodes[dad].kidsCosts = 0;
  return 1;
}

/***********************************************************************/

/* DijkTabsGraph::useKids

Returned Value: int (1)

Called By: external functions

This removes the current kids of dad (if any) and replaces them with
the ones given by the arguments. The arrays specified in the arguments
are used as is, so they must be created with "new" in the implementing
system and not changed or deleted by the implementing system.

The kids themselves must already exist and are not affected by this
function.

*/

int DijkTabsGraph::useKids( /* ARGUMENTS                                  */
 int dad,             /* index in nodes of node from which to remove kids */
 int kidsSize,        /* new number of kids                               */
 int * kidsIndexes,   /* new array of kids indexes in nodes               */
 double * kidsCosts)  /* new array of kids costs                          */
{
  if (nodes[dad].kidsSize > 0)
    {
      delete [] nodes[dad].kidsIndexes;
      delete [] nodes[dad].kidsCosts;
    }
  nodes[dad].kidsSize = kidsSize;
  nodes[dad].kidsIndexes = kidsIndexes;
  nodes[dad].kidsCosts = kidsCosts;
  return 1;
}

/***********************************************************************/
/***********************************************************************/

/* DijkTabsNode functions */

/***********************************************************************/

/* DijkTabsNode::~DijkTabsNode

This destructor (which does nothing in addition to default destruction)
is defined here only because objectcenter thought it was undefined
when it was in dijkTabs.hh. I

*/

DijkTabsNode::~DijkTabsNode() {}

/***********************************************************************/

//! DijkTabsNode::copyFrom copies another node into this one
/*
Returned Value: int
  If the "from" argument is 0, this returns 0. Otherwise, it returns 1.

Called By:
  DijkTabsGraph::increaseCapacity

This copies all of the data of the "from" node into "this" node.  The
kidsIndexes and kidsCosts arrays are used as given in the "from" node,
not made with "new".

*/

int DijkTabsNode::copyFrom( /* ARGUMENTS           */
 DijkTabsNode * from)       /*!< node to copy from */
{
  if (from != 0)
    {
      moreInfo = from->moreInfo;
      kidsSize = from->kidsSize;
      kidsIndexes = from->kidsIndexes;
      kidsCosts = from->kidsCosts;
      open = from->open;
      arrivedFrom = from->arrivedFrom;
      visitCost = from->visitCost;
      less = from->less;
      more = from->more;
      myTab = from->myTab;
    }
  return ((from == 0) ? 0 : 1);
}

/***********************************************************************/
/***********************************************************************/

/* DijkHeapSearcher functions */

/***********************************************************************/

//! DijkTabsSearcher::addTab adds a tab, gives it some of the next tab's nodes
/*
Returned Value: None

Called By:
  DijkTabsSearcher::insertNode1
  DijkTabsSearcher::insertNode2

This adds a tab before nextTab and gives it bestSize of nextTab's nodes.

Initially, index is set to cheap if nextTab is cheapTab. Otherwise it is
set to the [first node after the myNode of the tab before nextTab],
which is the first node belonging to nextTab.

*/

void DijkTabsSearcher::addTab( /* ARGUMENTS                         */
 DijkTab * nextTab)            /*!< tab before which to add new tab */
{
  DijkTab * newTab;
  int index;
  int n;

  newTab = new DijkTab;
  index = ((nextTab == cheapTab) ? cheap : nodes[nextTab->down->myNode].more);
  for (n = 0; n < bestSize; n++)
    {
      nodes[index].myTab = newTab;
      index = nodes[index].more;
    }
  newTab->myNode = nodes[index].less;
  newTab->down = nextTab->down; // may be 0
  newTab->up = nextTab;
  newTab->visitCost = nodes[newTab->myNode].visitCost;
  newTab->numBefore = bestSize;
  nextTab->down = newTab;
  nextTab->numBefore = (nextTab->numBefore - bestSize);
  if (newTab->down)
    newTab->down->up = newTab;
  else 
    cheapTab = newTab;
}

/***********************************************************************/

/* DijkTabsSearcher::closeSearch

Returned Value: int (0)

Called By: external functions (e.g., main in exampleDijkTabs.cc)

The only memory the DijkTabsSearcher uses is for the tabs. This function
deletes all the tabs.

*/

int DijkTabsSearcher::closeSearch() /* NO ARGUMENTS */
{
  DijkTab * aTab;
  DijkTab * nextTab;

  for (aTab = cheapTab; aTab; aTab = nextTab)
    {
      nextTab = aTab->up;
      delete aTab;
    }
  return 0;
}

/***********************************************************************/

/* DijkTabsSearcher::cycleSearch

Returned Value: int (0)

Called By: external functions (e.g., main in exampleDijkTabs.cc)

The cheap end of the list is kept track of by keeping the "cheap"
attribute of the DijkTabsSearcher set to it. The expensive end is kept track
of by keeping the "expensive" attribute of the DijkTabsSearcher set to it.

Each time a node is removed and processed, this sets bestSize to the
maximum of 10 or the largest integer smaller than the square root of
numOpen (the length of the open list). The floor of 10 is there to
prevent spending time on tab overhead when the tabs are not helping
that much anyway. Processing a node may increase numOpen, since
children of the node being removed may be added to the open list.

If a cheapest path has been found, this does nothing except set
status[0] to 2.

Otherwise, if the open nodes list is empty, this does nothing.

Otherwise:
A. This removes the first (index 1) node on the heap (which is the
   cheapest one) and processes the children of that node.
B. If a path has been found, this sets status[0] to 1 and sets
   cost[0]. These values may be the same after being reset as they
   were before, but (1) it is not worth the processing time to check and
   and reset only if they are different and (2) it is not safe to check
   the values since the caller could have reset them.

*/

int DijkTabsSearcher::cycleSearch() /* NO ARGUMENTS */
{
  int i;          // index for kids of oldCheap
  double kidCost; // total cost to visit i'th kid of oldCheap
  int kidNum;     // index in nodes of i'th kid of oldCheap
  int numKids;    // number of kids of oldCheap
  int oldCheap;   // index of cheap node when this starts
  
  if (cheap == goal)
    status[0] = 2;
  else if (cheap != -1)
    {
      nodes = theGraph->nodes;     // may have been changed by addKids
      oldCheap = cheap;
      removeCheap();
      nodes[oldCheap].open = DijkTabsNode::CLOSED;  // old cheap node closed
      numOpen--;
      *numOpenOut = numOpen;
      bestSize = ((numOpen < 100) ? 10 : (int)sqrt((double)numOpen));
      numKids = nodes[oldCheap].kidsSize;
      for (i = 0; i < numKids; i++)
	{
	  if ((nodes[oldCheap].kidsCosts)[i] < 0)
	    continue;
	  kidNum = (nodes[oldCheap].kidsIndexes)[i];
	  kidCost = (nodes[oldCheap].visitCost +
		     (nodes[oldCheap].kidsCosts)[i]);
	  processNode(kidNum, oldCheap, kidCost);
	  if (kidNum == goal)
	    {
	      cost[0] = nodes[kidNum].visitCost;
	      status[0] = 1; // status not 2 or would not get here
	    }
	}
    }
  return 0;
}

/***********************************************************************/

/* DijkTabsSearcher::getNextNode

Returned Value: int
This returns the index in nodes of the cheapest open node remaining.
If there are no open nodes remaining, this returns -1.

Called By: external functions

*/

int DijkTabsSearcher::getNextNode()
{
  return cheap;
}

/***********************************************************************/

//! DijkTabsSearcher::deleTabDown deletes tab, gives its nodes to previous tab
/*
Returned Value: None

Called By:
  DijkTabsSearcher::moveOrDeleTab

This is called only when aTab is the last tab. This deletes aTab and
gives its nodes to downTab. Since aTab is the last tab, the only tab
list link that needs changing is downTab->up, which is set to 0.

*/

void DijkTabsSearcher::deleTabDown( /* ARGUMENTS                      */
 DijkTab * aTab,                    /*!< Tab to delete                */
 DijkTab * downTab)                 /*!< next cheaper tab before aTab */
{
  int index;

  for (index = nodes[downTab->myNode].more;
       index != aTab->myNode;
       index = nodes[index].more)
    {
      nodes[index].myTab = downTab;
    }
  nodes[index].myTab = downTab;
  downTab->numBefore = (aTab->numBefore + downTab->numBefore);
  downTab->up = 0;
  downTab->myNode = aTab->myNode;
  downTab->visitCost = aTab->visitCost;
  delete aTab;
}

/***********************************************************************/

//! DijkTabsSearcher::deleTabUp deletes a tab, gives its nodes to the next tab
/*
Returned Value: None

Called By:
  DijkTabsSearcher::moveOrDeleTab

This deletes aTab and gives its nodes to upTab. It also relinks the
tab list around the gap created when aTab is removed.

*/

void DijkTabsSearcher::deleTabUp( /* ARGUMENTS                    */
 DijkTab * aTab,      /*!< tab to delete                          */
 DijkTab * upTab,     /*!< next more expensive tab after aTab     */
 DijkTab * downTab)   /*!< next cheaper tab before aTab, may be 0 */
{
  int index;
  for (index = (downTab ? nodes[downTab->myNode].more : cheap);
       index != aTab->myNode;
       index = nodes[index].more)
    {
      nodes[index].myTab = upTab;
    }
  nodes[index].myTab = upTab;
  upTab->numBefore = (aTab->numBefore + upTab->numBefore);
  upTab->down = downTab;
  delete aTab;
  if (downTab)
    downTab->up = upTab;
  else
    cheapTab = upTab;
}

/***********************************************************************/

//! DijkTabsSearcher::initSearch initializes the searcher
/*
Returned Value: int
If initialization occurs without error, this returns 0;
This prints an error message and returns 1 if:
1. The nodes of aGraph is NULL, or
2. the size of the kids of any node is negative, or
3. the size of aGraph is greater than its capacity.

Called By: external functions (e.g., main in exampleDijkTabs.cc)

The arrivedFrom of the start node is set to -1 to serve as
a stop marker for reportSearch.

statusOut and costOut are both arrays of size 1.

This:
1. Saves (as the values of data members of the Searcher) the locations
   where the user wants to have search data recorded. These are the
   three arguments: opensOut, statusOut, and costOut.
2. Sets the value of the other data members of the Searcher that need setting.
3. Initializes the start node and places it on the open nodes list.

Note that the Searcher uses the caller's memory for status, and cost.
This is safe because the Searcher writes their values but does not use
them otherwise. The Searcher does not use the caller's memory for
heapSize.

*/

int DijkTabsSearcher::initSearch( /* ARGUMENTS                         */
 DijkTabsGraph * aGraph, /* the graph to search in                     */
 int start,              /* index in nodes of aGraph of start node     */
 int finish,             /* index in nodes of aGraph of goal node      */
 int * opensOut,         /* 0=open list empty, non-zero=list not empty */
 int * statusOut,        /* status of reaching goal                    */
 double * costOut)       /* cost of reaching goal                      */
{
  int n;

  cheap = -1;
  expensive = -1;
  goal = finish;
  cheapTab = 0;
  numOpen = 0;
  numOpenOut = opensOut;
  *numOpenOut = numOpen;
  status = statusOut;
  status[0] = 0;
  cost = costOut;
  cost[0] = 0;
  bestSize = 10;
  theGraph = aGraph;
  nodes = theGraph->nodes;
  if (nodes == 0)
    {
      fprintf(stderr, "nodes array is null");
      return 1;
    }
  if (theGraph->size > theGraph->capacity)
    {
      fprintf(stderr, "size greater than capacity\n");
      return 1;
    }
  for (n = 0; n < theGraph->size; n++)
    {
      if (nodes[n].kidsSize < 0)
	{
	  fprintf(stderr, "bad kids size\n");
	  return 1;
	}
      nodes[n].open = DijkTabsNode::UNOPENED;
    }
  processNode(start, -1, 0);
  return 0;
}

/***********************************************************************/

//! DijkTabsSearcher::insertNode inserts a node in the open list in cost order
/*
Returned Value: None

Called By:
  DijkTabsSearcher::processNode

a. If cheapTab is 0 (which implies the open list is empty), call insertNode0.

b. Otherwise, if there is only 1 tab, call insertNode1.

c. Otherwise, (there are 2 or more tabs), call insertNode2.

By putting each inserted node before other nodes of the same cost (as
well as before others of higher cost), the goal node will be found
as soon as it becomes a minimal cost node, even if several other nodes
have the same minimal cost.

The node that is inserted is not marked open, since that is done
in processNode.

*/

void DijkTabsSearcher::insertNode( /* ARGUMENTS                     */
 int inNode,     /*!< index in nodes of node to insert              */
 double cost)    /*!< total cost of getting to Noe via current path */
{
  if (cheapTab == 0)
    insertNode0(inNode, cost);
  else if (cheapTab->up == 0)
    insertNode1(inNode, cost);
  else
    insertNode2(inNode, cost);
}

/***********************************************************************/

//! DijkTabsSearcher::insertNode0 inserts a node in the open list when empty
/*
Returned Value: None

Called By:
  DijkTabsSearcher::insertNode

This is called if cheapTab is NULL, in which case, the open list is empty.

Let Noe be the node whose index in nodes is inNode.

Make Noe be the only member of the open list by setting cheap and
expensive to inNode. Make cheapTab be a new tab, set its myNode to
inNode, set its up and down to NULL, set its cost to the given cost,
and set the number of nodes it has to 1.

*/

void DijkTabsSearcher::insertNode0( /* ARGUMENTS                   */
 int inNode,    /*!< index in nodes of node to insert              */
 double cost)   /*!< total cost of getting to Noe via current path */
{
  cheap = inNode;
  expensive = inNode;
  cheapTab = new DijkTab;
  nodes[inNode].myTab = cheapTab;
  cheapTab->myNode = inNode;
  cheapTab->down = 0;
  cheapTab->up = 0;
  cheapTab->visitCost = cost;
  cheapTab->numBefore = 1;
}

/***********************************************************************/

//! DijkTabsSearcher::insertNode1 inserts a node in the open list when 1 tab
/*
Returned Value: None

Called By:
  DijkTabsSearcher::insertNode

This is called when there is only one tab.

Let Noe be the node whose index in nodes is inNode.

If Noe is more expensive than any on the open list, put Noe at the
end of the open list and move the tab.

Otherwise, loop up the open list (from cheap towards expensive)
comparing the cost of Noe to the cost of each element of the
list. Stop when either Noe is no longer cheaper than nodes[index];
there must be such an index. If Noe is minimally (not necessarily
least) costly, add Noe at cheap end of the list.  Otherwise, add Noe
between nodes[index] and the preceding node.

Make cheapTab be the myTab of Noe.

Add 1 to the numBefore of cheapTab. If the numBefore is now larger than
twice bestSize, insert a new tab.

*/

void DijkTabsSearcher::insertNode1( /* ARGUMENTS                  */
 int inNode,   /*!< index in nodes of node to insert              */
 double cost)  /*!< total cost of getting to Noe via current path */
{
  int index;
  int previous;

  if (nodes[expensive].visitCost < cost)
    { // inNode most expensive; put inNode at end, move tab
      nodes[inNode].less = expensive;
      nodes[expensive].more = inNode;
      expensive = inNode;
      cheapTab->myNode = inNode;
      cheapTab->visitCost = cost;
    }
  else
    { // inNode not most expensive; find first one not cheaper than inNode
      for (index = cheap;
	   (nodes[index].visitCost < cost);
	   index = nodes[index].more);
      if (index == cheap) // put inNode at beginning
	{
	  nodes[inNode].more = cheap;
	  nodes[cheap].less = inNode;
	  cheap = inNode;
	}
      else // put inNode between previous and index
	{
	  previous = nodes[index].less;
	  nodes[inNode].more = index;
	  nodes[inNode].less = previous;
	  nodes[index].less = inNode;
	  nodes[previous].more = inNode;
	}
    }
  nodes[inNode].myTab = cheapTab;
  (cheapTab->numBefore)++;
  if (cheapTab->numBefore > (2 * bestSize))
    addTab(cheapTab);
}

/***********************************************************************/

//! DijkTabsSearcher::insertNode2 inserts a node in the open list when 2+ tabs
/*
Returned Value: None

Called By:
  DijkTabsSearcher::insertNode

This is called when there are 2 or more tabs.

Let Noe be the node whose index in nodes is inNode.

Loop down the tab list to find the first tab whose cost is not cheaper
than the cost of Noe. If there is no such tab, put Noe at the end of
the open list. If there is such a tab, go to the first node belonging
to this tab (which either is the node in the open list just after the
preceding tab (if there is a preceding tab), or is the first node (if
there is no preceding tab). Loop down the open list to find the first
node not cheaper than Noe (the node belonging to the tab is such a
node, so there is sure to be one). Put Noe into the list just before
that. This location may be at the beginning of the open list, so check
for that.

Add 1 to the numBefore of aTab. If the numBefore is now larger than
twice bestSize, insert a new tab.

If might be useful to have a marker at the expensive end of the tabs
list so that only one termination test would be necessary in the first
for loop.

*/

void DijkTabsSearcher::insertNode2( /* ARGUMENTS                  */
 int inNode,   /*!< index in nodes of node to insert              */
 double cost)  /*!< total cost of getting to Noe via current path */
{
  int index;
  int previous;
  DijkTab * aTab;

  for (aTab = cheapTab;
       (aTab->up && (aTab->visitCost < cost));
       aTab = aTab->up);
  if ((aTab->up == 0) && (aTab->visitCost < cost))
    { // aTab last tab & inNode cost most; insert node at end, move aTab
      nodes[inNode].less = aTab->myNode;
      nodes[aTab->myNode].more = inNode;
      expensive = inNode;
      aTab->visitCost = cost;
      aTab->myNode = inNode;
    }
  else
    { // inNode cost not most
      for (index = ((aTab == cheapTab) ?
		    cheap : nodes[aTab->down->myNode].more);
	   (nodes[index].visitCost < cost);
	   index = nodes[index].more);
      if (index == cheap) // put inNode at beginning
	{
	  nodes[inNode].more = cheap;
	  nodes[cheap].less = inNode;
	  cheap = inNode;
	}
      else // put inNode between previous and index
	{
	  previous = nodes[index].less;
	  nodes[inNode].more = index;
	  nodes[inNode].less = previous;
	  nodes[index].less = inNode;
	  nodes[previous].more = inNode;
	}
    }
  nodes[inNode].myTab = aTab;
  (aTab->numBefore)++;
  if (aTab->numBefore > (2 * bestSize))
    addTab(aTab);
}

/***********************************************************************/

//! DijkTabsSearcher::moveOrDeleTab moves or deletes a tab (for balancing)
/*
Returned Value: None

Called By:
  DijkTabsSearcher::removeCheap
  DijkTabsSearcher::removeNode

This is called if numBefore of aTab is too small (less than half
bestSize). aTab is either moved, in which case aTab takes some nodes
from a neighboring tab, or deleted, in which case a neighboring tab
gets all of aTab's nodes. Details follow.

A. If there is a tab after aTab (call it bTab):

A1. If the sum of the numBefore's of aTab and bTab is less than twice
bestSize, then deleTabUp is called to delete aTab and give bTab all of
aTab's nodes.

A2. If the sum is not less than twice bestSize, moveTabUp is
called to move aTab forward towards bTab, taking enough of bTab's
nodes so that aTab ends up with its numBefore equal to bestSize.

B. Otherwise, there must be a tab before aTab (call it bTab) because
this function is not called unless there are at least two tabs.

B1. If the sum of the numBefore's of aTab and bTab is less than twice
bestSize, then deleTabDown is called to delete aTab and give bTab
all of aTab's nodes.

B2. If the sum is not less than twice bestSize, moveTabDown is called
to move aTab back towards bTab, taking enough of bTab's nodes so that
aTab ends up with its numBefore equal to bestSize.

*/

void DijkTabsSearcher::moveOrDeleTab( /* ARGUMENTS                        */
 DijkTab * aTab)                      /*!< the tab to be moved or deleted */
{
  if (aTab->up)
    {
      if ((aTab->numBefore + aTab->up->numBefore) < (2 * bestSize))
	deleTabUp(aTab, aTab->up, aTab->down);
      else
	moveTabUp(aTab, aTab->up);
    }
  else
    {
      if ((aTab->numBefore + aTab->down->numBefore) < (2 * bestSize))
	deleTabDown(aTab, aTab->down);
      else
	moveTabDown(aTab, aTab->down);
    }
}

/***********************************************************************/

//! DijkTabsSearcher::moveTabDown moves a tab down (for balancing)
/*
Returned Value: None

Called By:
  DijkTabsSearcher::moveOrDeleTab

This moves aTab down towards downTab so that numBefore of aTab is bestSize.
This is not called unless the two tabs have at least twice bestSize
nodes between them, so downTab will still have at least bestSize
nodes when this is finished working.

*/

void DijkTabsSearcher::moveTabDown( /* ARGUMENTS             */
 DijkTab * aTab,                    /*!< the tab to be moved */
 DijkTab * downTab)                 /*!< the tab before aTab */
{
  int index;
  int n;

  index = downTab->myNode;
  for (n = aTab->numBefore; n < bestSize; n++)
    {
      nodes[index].myTab = aTab;
      index = nodes[index].less;
    }
  downTab->myNode = index;
  downTab->visitCost = nodes[index].visitCost;
  downTab->numBefore = (downTab->numBefore - (bestSize - aTab->numBefore));
  aTab->numBefore = bestSize;
}

/***********************************************************************/

//! DijkTabsSearcher::moveTabUp moves a tab up (for balancing)
/*
Returned Value: None

Called By:
  DijkTabsSearcher::moveOrDeleTab

This moves aTab up towards upTab so that numBefore of aTab is bestSize.
This is not called unless the two tabs have at least twice bestSize
nodes between them, so upTab will still have at least bestSize
nodes when this is finished working.

*/

void DijkTabsSearcher::moveTabUp( /* ARGUMENTS             */
 DijkTab * aTab,                  /*!< the tab to be moved */
 DijkTab * upTab)                 /*!< the tab after aTab  */
{
  int index;
  int n;

  index = aTab->myNode;
  for (n = aTab->numBefore; n < bestSize; n++)
    {
      index = nodes[index].more;
      nodes[index].myTab = aTab;
    }
  upTab->numBefore = (upTab->numBefore - (bestSize - aTab->numBefore));
  aTab->myNode = index;
  aTab->visitCost = nodes[index].visitCost;
  aTab->numBefore = bestSize;
}

/***********************************************************************/

//! DijkTabsSearcher::processNode processes a node (popped off the open list)
/*
Returned Value: None

Called By:
  DijkTabsSearcher::cycleSearch
  DijkTabsSearcher::initSearch

This function processes the node (call it Noe) whose index in the
nodes being searched is numNode. Outline numbers here match comment
numbers below.

A. If Noe is open:
  1. If Noe's currently recorded cost is greater than cost:
     a. Remove Noe from the open nodes list.
     b. Set Noe's visitCost to cost.
     c. Set Noe's arrivedFrom to from.
     d. Reinsert Noe in the open nodes list.
B. Else if Noe has not yet been opened:
  1. Mark Noe open.
  2. Set Noe's visitCost to cost.
  3. Set Noe's arrivedFrom to from.
  4. Insert Noe in the open nodes list.
  5. Add one to the number of open nodes.

If Noe is closed (so that nodes[numNode].open == DijkTabsNode::CLOSED),
nothing is done.

*/

void DijkTabsSearcher::processNode( /* ARGUMENTS                  */
 int numNode,  /*!< index in nodes of Noe                         */
 int from,     /*!< index in nodes of node from which Noe visited */
 double cost)  /*!< total cost of getting to Noe via current path */
{
  if (nodes[numNode].open == DijkTabsNode::OPEN)          // A
    {
      if (nodes[numNode].visitCost > cost)                // A1
	{
	  removeNode(numNode);                            // A1a
	  nodes[numNode].visitCost = cost;                // A1b
	  nodes[numNode].arrivedFrom = from;              // A1c
	  insertNode(numNode, cost);                      // A1d
	}
    }
  else if (nodes[numNode].open == DijkTabsNode::UNOPENED) // B
    {
      nodes[numNode].open = DijkTabsNode::OPEN;           // B1
      nodes[numNode].visitCost = cost;                    // B2
      nodes[numNode].arrivedFrom = from;                  // B3
      insertNode(numNode, cost);                          // B4
      numOpen++;                                          // B5
      *numOpenOut = numOpen;
    }
}

/***********************************************************************/

//! DijkTabsSearcher::removeCheap removes the cheapest node from the open list
/*
Returned Value: None

Called By:
  DijkTabsSearcher::cycleSearch
  DijkTabsSearcher::removeNode

This is used to remove the first member of the open nodes list.
The tab for the first open node is always cheapTab.

A. If there is only one member of the open nodes list (indicated
   by cheap being the same as expensive):

A1. To excise the cheap node from open node list, cheap and expensive
    are set to -1.

A2. To excise cheapTab from tab list, cheapTab is deleted
    and cheapTab is set to NULL.

B. Otherwise (there are two or more members of the open nodes list):

B1. cheap is reset to the successor of the original cheap.

B2. numBefore of cheapTab is decreased by one, and if there are two or
    more tabs and the numBefore of cheapTab is now less than half of
    bestSize, cheapTab is moved or deleted.

When there are two or more members of the open nodes list, the first
tab is never on the first node, so it is not necessary to check for
that.

This does not decrement numOpen because dijk does that after calling
this function.

This function does not mark cheap closed because dijk does that after
calling this function, and removeNode does not want to do that.

*/

void DijkTabsSearcher::removeCheap() /* NO ARGUMENTS */
{
  if (cheap == expensive)          // A
    {
      cheap = -1;                  // A1
      expensive = -1;
      delete cheapTab;             // A2
      cheapTab = 0;
    }
  else                             // B
    {
      cheap = nodes[cheap].more;   // B1
      (cheapTab->numBefore)--;     // B2
      if (cheapTab->up && (cheapTab->numBefore < (bestSize / 2)))
	moveOrDeleTab(cheapTab);
    }
}

/***********************************************************************/

//! DijkTabsSearcher::removeNode removes the given node from the open list
/*
Returned Value: None

Called By:
  DijkTabsSearcher::processNode

Let Noe be the node whose index in nodes is outNode. This removes
Noe from the open list as follows.

A. If Noe is at the cheap end of the open list, call removeCheap to
   remove Noe from the list.

B. Otherwise (Noe is not at the cheap end of the open list):

B1. If Noe is not at the expensive end of the open list, excise Noe
    from the list by linking Noe's predecessor and successor together.

B2. Otherwise (Noe is at the expensive end of the list), excise
    Noe by resetting expensive to the predecessor of Noe.

B3. In either case, if there is a tab on the node to remove, move
    the tab to the preceding node.

B4. Also in either case, subtract one from the numBefore of Noe's
    myTab. If there are two or more tabs and the numBefore of aTab
    is now less than half of bestSize, move aTab or delete it.

*/

void DijkTabsSearcher::removeNode( /* ARGUMENTS                          */
 int outNode)                      /*!< index in nodes of node to remove */
{
  DijkTab * aTab;  /* the tab for outNode  */

  if (outNode == cheap)                                  // A
    removeCheap();
  else
    {
      if (expensive != outNode)                          // B1
	{
	  nodes[nodes[outNode].more].less = nodes[outNode].less;
	  nodes[nodes[outNode].less].more = nodes[outNode].more;
	}
      else                                               // B2
	expensive = nodes[outNode].less;
      aTab = nodes[outNode].myTab;
      if (outNode == aTab->myNode)                       // B3
	{
	  aTab->myNode = nodes[outNode].less;
	  aTab->visitCost = nodes[aTab->myNode].visitCost;
	}
      (aTab->numBefore)--;                               // B4
      if ((aTab->up || aTab->down) && (aTab->numBefore < (bestSize / 2)))
	moveOrDeleTab(aTab);
    }
}

/***********************************************************************/

/* DijkTabsSearcher::reportSearch

Returned Value: int
If the length of the path is less than maxLength, this returns the
length of the path. Otherwise, it returns -1.

Called By: external functions (e.g., main in exampleDijkTabs.cc)

The arrivedFrom of the start node is set to -1 in initSearch, so
testing for -1 tests for the start node.

This is not reporting the cost because the caller will have direct
access to it.

*/

int DijkTabsSearcher::reportSearch( /* ARGUMENTS                     */
 int maxLength,  /* maximum length of the path                       */
 int * path)     /* array of index numbers giving the path, set here */
{
  int i;
  int n;
  int length;

  length = 1;
  for (i = goal; nodes[i].arrivedFrom != -1; i = nodes[i].arrivedFrom)
    length++;
  i = goal;
  if( length >= maxLength )
    return -1;
  for (n = (length - 1); n > -1; n--) 
    {
      path[n] = i;
      i = nodes[i].arrivedFrom;
    }
  return length;
}

/***********************************************************************/

/*
  $Log: dijkTabs.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.9  2005/07/15 21:26:49  kramer
  changed interface, improved documentation

  Revision 1.8  2005/06/29 21:08:11  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.7  2005/06/09 12:17:29  kramer
  added CVS logging
*/
