/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*
  unix_ulapi.c

  Implementation of user-level API to Unix simulations of the
  realtime part, plus some non-realtime resources like threads
  for portability.

  This assumes Unix user <-> Unix realtime.
*/

#include <stdio.h>		/* perror */
#include <stddef.h>		/* NULL */
#include <stdlib.h>		/* malloc */
#include <sys/ipc.h>		/* IPC_CREAT */
#include <sys/shm.h>		/* shmget() */
#include "ulapi.h"

ulapi_result unix_ulapi_init(ulapi_integer sel)
{
  return UL_USE_UNIX == sel ? ULAPI_OK : 
    UL_USE_DEFAULT == sel ? ULAPI_OK : ULAPI_IMPL_ERROR;
}

ulapi_result unix_ulapi_exit(void)
{
  return ULAPI_OK;
}

typedef struct {
  ulapi_id key;
  ulapi_integer size;
  ulapi_id id;
  void * addr;
} unix_shm_struct;

void * unix_ulapi_shm_new(ulapi_id key, ulapi_integer size)
{
  unix_shm_struct * shm;

  shm = malloc(sizeof(unix_shm_struct));
  if (NULL == (void *) shm) return NULL;

  shm->id = shmget((key_t) key, (int) size, IPC_CREAT | 0666);
  if (-1 == shm->id) {
    perror("shmget");
    free(shm);
    return NULL;
  }

  shm->addr = shmat(shm->id, NULL, 0);
  if ((void *) -1 == shm->addr) {
    perror("shmat");
    free(shm);
    return NULL;
  }

  return (void *) shm;
}

void * unix_ulapi_shm_addr(void * shm)
{
  return ((unix_shm_struct *) shm)->addr;
}

ulapi_result unix_ulapi_shm_delete(void * shm)
{
  struct shmid_ds d;
  int r1, r2;

  if (NULL == shm) return ULAPI_OK;

  r1 = shmdt(((unix_shm_struct *) shm)->addr);
  r2 = shmctl(IPC_RMID, ((unix_shm_struct *) shm)->id, &d);

  free(shm);

  return (r1 || r2 ? ULAPI_ERROR : ULAPI_OK);
}

ulapi_result unix_ulapi_fifo_new(ulapi_integer key, ulapi_integer *fd, ulapi_integer size)
{
  return ULAPI_ERROR;
}

ulapi_result unix_ulapi_fifo_delete(ulapi_integer key, ulapi_integer fd, ulapi_integer size)
{
  return ULAPI_ERROR;
}

ulapi_integer unix_ulapi_fifo_write(ulapi_integer fd, char *buf, ulapi_integer size)
{
  return 0;
}

ulapi_integer unix_ulapi_fifo_read(ulapi_integer fd, char *buf, ulapi_integer size)
{
  return 0;
}

/*
  Modification history:

  $Log: unix_ulapi.c,v $
  Revision 1.4  2009/04/16 19:17:38  proctor
  Updated the NML config files to reflect the simware changes; updated ulapi
  to be consistent with the master files in gomotion; added targets for the
  new simware

  Revision 1.10  2008/02/22 18:39:42  proctor
  Reworked serial interface to allocate a pointer to the id

  Revision 1.9  2007/09/21 19:28:33  proctor
  Pulled the SDL joystick code stubs into gotk.c, and reworked ulapi
  so that task/thread functions were available.

  Revision 1.8  2007/05/16 12:36:12  proctor
  Fixed genserkins iterative inverse, adding timing printouts to genkintest.

  Revision 1.7  2007/02/08 18:50:58  proctor
  Added getopt-style arg handling, and a select interface to ulapi

  Revision 1.6  2006/12/13 18:42:31  proctor
  Changed etime,esleep to ulapi_time,ulapi_sleep to avoid conflicts with rcslib

  Revision 1.5  2006/12/11 21:30:17  proctor
  Fleshed out win32 apis

  Revision 1.4  2006/12/06 18:13:40  proctor
  Added stub files for rt/ulapi, and initial win32 versions; fixed return
  values in unix pthread code

  Revision 1.3  2006/12/06 18:10:34  proctor
  Reworked shared memory interface to use a malloc'ed structure

  Revision 1.2  2006/09/27 15:29:47  proctor
  Changed rtapi_count to rtapi_int

  Revision 1.1  2006/05/19 13:04:21  proctor
  Initial checkin

  Revision 1.1.1.1  2006-02-10 21:15:15  proctor
  Initial import

*/

