/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   factoryCtrl.cc
  \brief  Provides an interface to the factory controller in USARSim

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.9 $
  $Date: 2009/05/01 18:02:05 $
  \endcode

  \author Stephen Balakirsky
  \date   02/14/2009
*/
#include <string>
#include <sstream>
#include <vector>
#include <stdio.h>              // printf
#include <stddef.h>		// sizeof(), NULL
#include <string.h>		// strcpy()
#include <stdlib.h>		// getenv()
#include <signal.h>		// signal(), SIGINT
#include <unistd.h>		// sleep()
#include <math.h>               // fabs()
#include <rcs.hh>
#include <ulapi.h>		// ulapi_XXX
#include <factCtrl.hh>
#include "skin.h"
#include "simware.h"

using namespace std;

static FILE *fp = NULL;
static skin_struct parseSkin;
static RCS_CMD_CHANNEL * factCtrlCmdBuf= NULL;
static RCS_STAT_CHANNEL * factCtrlStatBuf = NULL;
static RCS_CMD_CHANNEL * factCtrlCfgBuf= NULL;
static RCS_STAT_CHANNEL * factCtrlSetBuf = NULL;

static FactCtrlStat factCtrlStat;
static FactCtrlSet factCtrlSet;

static bool settingsAreFresh = true;
RCS_TIMER *timer; // cycle timer

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define THIS_PROCESS "factCtrlApp" // NML process name for the config file

/*******************************************
structures
*******************************************/
typedef struct {
  char parseName[80];
  int (*ask)(int (*)(simware_struct *sw));
  int (*tell)(simware_struct *sw);
}ParseThreadArgs;

/*******************************************
end of structures
*******************************************/

int factoryTell(simware_struct *sw)
{
  int count;
  static bool printedOnce = false;

  switch(sw->type)
    {
    case SW_COMMAND_RESPONSE:
      if( sw->data.response.status == -1)
	{
	  status_next(&factCtrlStat, RCS_ERROR );
	  strncpy( factCtrlStat.message, sw->data.response.errorString, 80 );
	  factCtrlStat.message[79] = '\0';
	  factCtrlStat.message_length = strlen( factCtrlStat.message );
	}
      else
	{
	  status_next(&factCtrlStat, RCS_DONE );
	  factCtrlStat.message[0] = '\0';
	  factCtrlStat.message_length = 0;
	}
      break;
    case SW_SENSOR_CARGO:
      if( !printedOnce )
	{
	  printedOnce = true;
	  printf( "Cargo_id Cargo_mem x y z roll pitch yaw\n");
	}
      for( count=0; count<sw->data.cargo.numCargo; count++ )
	{
	  printf( "%d %d %f %f %f %f %f %f\n",
		  sw->data.cargo.id[count],
		  sw->data.cargo.memory[count],
		  sw->data.cargo.x[count],
		  sw->data.cargo.y[count],
		  sw->data.cargo.z[count],
		  sw->data.cargo.roll[count],
		  sw->data.cargo.pitch[count],
		  sw->data.cargo.yaw[count] );
	}
      break;
    default:
      printf("factoryCtrl tell with unsupported type %d\n", sw->type);
      break;
    }
  cycle_time_tracker( &factCtrlStat.tt );
  factCtrlStatBuf->write(&factCtrlStat);
  return 0;
}

void parseThread( void *argIn)
{
  int retval;
  ParseThreadArgs *args = (ParseThreadArgs*)argIn;

  retval = skin_load(&parseSkin, args->parseName);
  if (0 != retval) {
    fprintf(stderr, "can't load inferior skin ``%s''\n", args->parseName);
    return;
  }
  if( args != NULL )
    delete args;
  args = NULL;
  while(1)
    {
      parseSkin.ask(factoryTell);
    }
}

static int initNmlBuffers(char * config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  sprintf(chanName, "%s%d", FACT_CTRL_CMD_NAME, bufnum);
  if (NULL != factCtrlCmdBuf) delete factCtrlCmdBuf;
  factCtrlCmdBuf = new RCS_CMD_CHANNEL(factCtrl_format, 
				       chanName, THIS_PROCESS,
				       config_file);
  if (! factCtrlCmdBuf->valid()) 
    {
      delete factCtrlCmdBuf;
      factCtrlCmdBuf = 0;
      return 1;
    }

  sprintf(chanName, "%s%d", FACT_CTRL_STAT_NAME, bufnum);
  if (NULL != factCtrlStatBuf) delete factCtrlStatBuf;
  factCtrlStatBuf = new RCS_STAT_CHANNEL(factCtrl_format, 
					 chanName, THIS_PROCESS,
					 config_file);
  if (! factCtrlStatBuf->valid()) 
    {
      delete factCtrlStatBuf;
      factCtrlStatBuf = 0;
      return 1;
    }

  sprintf(chanName, "%s%d", FACT_CTRL_CFG_NAME, bufnum);
  if (NULL != factCtrlCfgBuf) delete factCtrlCfgBuf;
  factCtrlCfgBuf = new RCS_CMD_CHANNEL(factCtrl_format, 
				       chanName, THIS_PROCESS, 
				       config_file);
  if (! factCtrlCfgBuf->valid()) 
    {
      delete factCtrlCfgBuf;
      factCtrlCfgBuf = 0;
      return 1;
    }

  sprintf(chanName, "%s%d", FACT_CTRL_SET_NAME, bufnum);
  if (NULL != factCtrlSetBuf) delete factCtrlSetBuf;
  factCtrlSetBuf = new RCS_STAT_CHANNEL(factCtrl_format, 
					chanName, THIS_PROCESS, 
					config_file);
  if (! factCtrlSetBuf->valid()) 
    {
      delete factCtrlSetBuf;
      factCtrlSetBuf = 0;
      return 1;
    }
  

  nml_start();
  return 0;
}

int cleanupNmlBuffers(void)
{
  nml_cleanup();
  factCtrlCmdBuf = NULL;
  factCtrlStatBuf = NULL;
  factCtrlCfgBuf = NULL;
  factCtrlSetBuf = NULL;

  return 0;
}

bool doCmdNop(FactCtrlStat *stat)
{
  bool commandDone = true;

  if (state_match(stat, NEW_COMMAND)){
    state_new(stat);
    state_next(stat, S0);
    status_next(stat, RCS_DONE);
  }
  return commandDone;
}

bool doCmdInit(FactCtrlStat * stat)
{
  bool commandDone = false;
  void *parseTask;
  ParseThreadArgs *args = new ParseThreadArgs();
  simware_struct initOut;

  if (state_match(stat, NEW_COMMAND)) 
    {
      if( stat->admin_state == ADMIN_INITIALIZED )
	{
	  state_next(stat, S2);
	  status_next(stat, RCS_DONE);
	  commandDone = true;
	  return commandDone;
	}
      if( fp != NULL )
	{
	  fclose(fp);
	  fp = NULL;
	}
      state_new(stat);
      if( factCtrlSet.debug > 1 )
	printf( "In Init command\n" );

      srand( 12345 );

      /* put information for VMAC1 world into settings and status buffers. 
	 This is a temporary fix
	 until we have a parser to read this information from USARSim */
      settingsAreFresh = true;
      factCtrlSet.conveyorSet_length = 3;
      factCtrlStat.conveyorStat_length = 3;

      strcpy( factCtrlSet.conveyorSet[0].name, "Loading" );
      factCtrlSet.conveyorSet[0].minSpeed = -6;
      factCtrlSet.conveyorSet[0].maxSpeed = 6;
      factCtrlSet.conveyorSet[0].location.x = 11.7120;
      factCtrlSet.conveyorSet[0].location.y = 28.6079;
      factCtrlSet.conveyorSet[0].location.z = 9.6319;
      factCtrlSet.conveyorSet[0].orientation.r = 0;
      factCtrlSet.conveyorSet[0].orientation.p = 0;
      factCtrlSet.conveyorSet[0].orientation.y = -3.1415;

      strcpy( factCtrlSet.conveyorSet[1].name, "Unload1" );
      factCtrlSet.conveyorSet[1].minSpeed = -6;
      factCtrlSet.conveyorSet[1].maxSpeed = 6;
      factCtrlSet.conveyorSet[1].location.x = 3.7249;
      factCtrlSet.conveyorSet[1].location.y = -27.2164;
      factCtrlSet.conveyorSet[1].location.z = 9.6319;
      factCtrlSet.conveyorSet[1].orientation.r = 0;
      factCtrlSet.conveyorSet[1].orientation.p = 0;
      factCtrlSet.conveyorSet[1].orientation.y = -6.2831;

      strcpy( factCtrlSet.conveyorSet[2].name, "Unload2" );
      factCtrlSet.conveyorSet[2].minSpeed = -6;
      factCtrlSet.conveyorSet[2].maxSpeed = 6;
      factCtrlSet.conveyorSet[2].location.x = 13.5007;
      factCtrlSet.conveyorSet[2].location.y = -27.2268;
      factCtrlSet.conveyorSet[2].location.z = 9.6319;
      factCtrlSet.conveyorSet[2].orientation.r = 0;
      factCtrlSet.conveyorSet[2].orientation.p = 0;
      factCtrlSet.conveyorSet[2].orientation.y = -6.2831;

      factCtrlStat.conveyorStat_length = 3;

      strcpy( factCtrlStat.conveyorStat[0].name, "Loading" );
      factCtrlStat.conveyorStat[0].speed = 0;
      strcpy( factCtrlStat.conveyorStat[1].name, "Unload1" );
      factCtrlStat.conveyorStat[1].speed = 0;
      strcpy( factCtrlStat.conveyorStat[2].name, "Unload2" );
      factCtrlStat.conveyorStat[2].speed = 0;

      state_next(stat, S0);
      status_next(stat, RCS_EXEC);
    } 
  else if(state_match(stat, S0))  // start inferior skin
    {
      parseTask = ulapi_task_new();
      args->tell = factoryTell;
      strcpy(args->parseName, "./factory.so" );
      ulapi_task_start( parseTask,
			parseThread,
			args,
			ulapi_prio_lowest(),
			1);
      state_next(stat, S1);
    }
  else if(state_match(stat, S1)) // spawn factory controller
    {
      initOut.type = SW_COMMAND_INIT;
      sprintf( initOut.data.initRobot.robot_type, "USARBot.FactoryControlBot" );
      sprintf( initOut.data.initRobot.name, "VMAC_FactoryCtrl" );
      initOut.data.initRobot.north = 1;
      initOut.data.initRobot.east = 1;
      initOut.data.initRobot.elevation = 1;
      parseSkin.tell(&initOut);

      stat->admin_state = ADMIN_INITIALIZED;
      state_next(stat, S2);
      status_next(stat, RCS_DONE);
      commandDone = true;
    }
  else 
    {
      state_default(stat);
      commandDone = true;
    }
  return commandDone;
}

bool doCmdAbort(FactCtrlStat * stat)
{
  bool commandDone = true;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if( factCtrlSet.debug > 1 )
      printf( "In Abort command\n" );
    stat->admin_state = ADMIN_UNINITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
  return commandDone;
}

bool doCmdHalt(FactCtrlStat * stat)
{
  bool commandDone = true;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if( factCtrlSet.debug > 1 )
      printf( "In Halt command\n" );
    stat->admin_state = ADMIN_UNINITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }

  return commandDone;
}

bool doCmdShutdown(FactCtrlStat * stat)
{
  bool commandDone = false;

  if (state_match(stat, NEW_COMMAND)) 
    {
      state_new(stat);
      //    if( factCtrlSet.debug > 1 )
      printf( "In Shutdown command\n" );
      stat->admin_state = ADMIN_SHUT_DOWN;
      state_next(stat, S0);
    } 
  else 
    {
    cleanupNmlBuffers();
    ulapi_exit();
    commandDone = true;
    exit(1);
    }
  return commandDone;
}

bool doCmdPackage( FactCtrlStat * stat, 
		   FactCtrlCmdDropPackage *packageCmd )
{
  bool commandDone = true;
  static simware_struct boxChute;
  int rValue;

  if (state_match(stat, NEW_COMMAND)) 
    {
      state_new(stat);
      if( factCtrlSet.debug > 1 )
	printf( "In Package Drop command\n" );
      if( stat->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(stat, RCS_ERROR );
	  state_next(stat, SE1 );
	  return commandDone;
	}
      if( fp != NULL )
	{
	  if( fscanf( fp, "%lf", &boxChute.data.boxchute.RFID_value  ) != 1 )
	    {
	      printf( "no more rfid values, using random numbers\n" );
	      fclose(fp);
	      fp = NULL;
	    }
	  //	  else
	  //	    printf( "Read rfid value of %f\n", boxChute.data.boxchute.RFID_value );
	}
      if( fp == NULL )
	{
	  rValue = rand();
	  if( rValue > RAND_MAX/2. )
	    boxChute.data.boxchute.RFID_value = 1.;
	  else
	    boxChute.data.boxchute.RFID_value = 2.;
	}
      //      printf( "Random value: %d rounded to %f\n", rValue, boxChute.data.boxchute.RFID_value );
      strncpy( boxChute.name, packageCmd->chuteName, SW_NAME_MAX );
      boxChute.type = SW_COMMAND_BOXCHUTE_RELEASE;
      printf( "Spawing box at chute %s with RFID %f\n", 
	      boxChute.name, boxChute.data.boxchute.RFID_value );
      parseSkin.tell(&boxChute);
      
      state_next(stat, S0);
    }    
  return commandDone;
}

bool doCmdSetConveyor( FactCtrlStat * stat, 
		       FactCtrlCmdSetConveyor *setConveyorCmd )
{
  bool commandDone = true;
  simware_struct setConveyor;

  if (state_match(stat, NEW_COMMAND)) 
    {
      state_new(stat);
      if( factCtrlSet.debug > 1 )
	printf( "In set conveyor command\n" );
      if( stat->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(stat, RCS_ERROR );
	  state_next(stat, SE1 );
	  return true;
	}
      setConveyor.type = SW_COMMAND_CONVEYOR_SPEED;
      strncpy( setConveyor.name, setConveyorCmd->conveyorName, SW_NAME_MAX );
      setConveyor.data.conveyor.speed = setConveyorCmd->speed;
      parseSkin.tell(&setConveyor);
      state_next(stat, S0);

      // set the conveyor speed in the status here. This is temporary until we are reading and
      // parsing usarSim messages
      for( int count=0; count<factCtrlStat.conveyorStat_length; count++ )
	if( !strcmp( factCtrlStat.conveyorStat[count].name, setConveyorCmd->conveyorName ) )
	  {
	    factCtrlStat.conveyorStat[count].speed = setConveyorCmd->speed;
	    break;
	  }
    }    
  return commandDone;
}

bool doCmdClearCargo( FactCtrlStat * stat )
{
  bool commandDone = true;
  simware_struct clearCargo;

  if (state_match(stat, NEW_COMMAND)) 
    {
      state_new(stat);
      if( factCtrlSet.debug > 1 )
	printf( "In clear cargo command\n" );
      if( stat->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(stat, RCS_ERROR );
	  state_next(stat, SE1 );
	  return true;
	}
      clearCargo.type = SW_COMMAND_CLEAR_CARGO;
      parseSkin.tell(&clearCargo);
      state_next(stat, S0);
    }    
  return commandDone;
}

#define LENGTH_MAX 10.0
#define LENGTH_MIN 0.001

#define WIDTH_MAX 10.0
#define WIDTH_MIN 0.001
void doCfgNop(FactCtrlSet *set)
{
  if (state_match(set, NEW_COMMAND)){
    state_new(set);
    state_next(set, S0);
    status_next(set, RCS_DONE);
  }
}

void doCfgCycleTime(FactCtrlSet * set, double cycleTime)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->cycleTime = cycleTime;
    if( factCtrlSet.debug > 1 )
      printf( "Setting CycleTime to %f\n", cycleTime );
    delete timer;
    timer = new RCS_TIMER(set->cycleTime);
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

void doCfgSetPackFile( FactCtrlSet *set, char *fileName )
{
  if( state_match(set, NEW_COMMAND))
    {
      state_new(set);
      if( fp != NULL )
	{
	  fclose(fp);
	  fp = NULL;
	}
      state_next(set, S0);
    }
  if( state_match(set, S0))
    {
      fp = fopen( fileName, "r" );
      if( fp == NULL )
	{
	  state_next(set, SE1);
	  status_next(set, RCS_ERROR);
	}
      else
	{
	  state_next(set, S1);
	  status_next(set, RCS_DONE);
	}
    }
}

void doCfgDebug(FactCtrlSet * set, int debug)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->debug = debug;
    if( factCtrlSet.debug > 1 )
      printf( "In Debug config\n" );
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

void printHelp(char *name)
{
  printf( "Usage: %s -n<nmlFile> -d<id> -h<hostname> -p<port>\n", name );
  printf( "\t nmlFile is the name of the nml configuration file\n" );
  printf( "\t id is the number postfix for the nml channels\n" );
  printf( "\t hostname is the hostname of the USARSim server\n" );
  printf( "\t port is the port number for the USARSim socket\n" );
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NMLFile = NULL;
  int bufnum = 1;
  int retval;
  NMLTYPE cmdType, cfgType;
  int serialNumber;
  FactCtrlCmdNop factCtrlCmdNop;
  FactCtrlCfgNop factCtrlCfgNop;
  bool commandDone;

  /****************************************************************** 
   * INITIALIZE - The initialization opens any communication buffers, 
   *              places the syste in a safe known state, and 
   *		  initializes any control parameters.
   */
  
  while (1)
    {
      char option = getopt(argc, argv, "n:d:");
      if (option == -1)
	break;

      switch (option) 
	{
	case 'd':
	  if (1 != sscanf(optarg, "%i", &bufnum)) 
	    {
	      rcs_print_error("%s: bad value for buffer number: %s\n", 
			      argv[0], optarg);
	      printHelp(argv[0]);
	      return 1;
	    }
	  break;
	case 'n':
	  NMLFile = optarg;
	  printf("%s: using specified config file %s\n", argv[0], NMLFile);
	  break;
	default:
	  rcs_print_error("%s: bad argument\n", argv[0]);
	  printHelp(argv[0]);
	  return 1;
	  break;
	}
    }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NMLFile) 
    {
      nml_file_env = getenv("CONFIG_NML");
      if (NULL != nml_file_env) 
	{
	  NMLFile = nml_file_env;
	  printf("%s: using environment config file %s\n", argv[0], NMLFile);

	} 
      else 
	{
	  NMLFile = nml_file_default;
	  printf("%s: using default config file %s\n", argv[0], NMLFile);
	}
    }
  // initialize the ulapi code
  if( ulapi_init(UL_USE_DEFAULT) != ULAPI_OK )
    {
      printf( "Unable to initialize the ulapi interface\n" );
      return -1;
    }

  // allocate all the buffers
  retval = initNmlBuffers(NMLFile, bufnum);

  if (retval) {
    fprintf(stderr,"%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }

  // stuff a nop command
  factCtrlCmdNop.serial_number = 1;
  factCtrlCmdBuf->write(&factCtrlCmdNop);

  // set the status up so that init command goes through
  factCtrlStat.echo_serial_number = 0;
  factCtrlStat.conveyorStat_length = 0;
  // stuff a nop config
  factCtrlCfgNop.serial_number = 1;
  factCtrlCfgBuf->write(&factCtrlCfgNop);

  // set the settings up so that the length goes through
  factCtrlSet.cycleTime = 0.1;
  timer = new RCS_TIMER(factCtrlSet.cycleTime);

  factCtrlSet.echo_serial_number = 0;

  // initialize the time tracker
  factCtrlStat.tt.count = 0;

  done = 0;
  signal(SIGINT, quit);

  while (! done) 
    {
      /******************************
       * House keeping: set time tracker to increment
       */
      cycle_time_tracker( &factCtrlStat.tt );

      /****************************************************************** 
       * READ COMMAND - Command information is received, and the system
       *                is prepared to execute the command. When a new
       *		    command is received, the old command is immediately
       *		    replaced by this command. If the command is not 
       *		    able to be interrupted, error flags are set and
       *		    will be reported during the writng of status
       */

      // read data if there is any to read

      // read command, block unless we are not done with the command and we need to cycle
      if( commandDone != true )
	{
	  timer->wait();
	  cmdType = factCtrlCmdBuf->read();
	}
      else
	cmdType = factCtrlCmdBuf->blocking_read(-1);
      serialNumber = factCtrlCmdBuf->get_address()->serial_number;
      switch (cmdType) {
      case 0:
	// no new command
	break;
      case -1:
	// comm error
	fprintf(stderr, "%s: NML command read error\n", argv[0]);
	break;
      case FACT_CTRL_CMD_NOP_TYPE:
      case FACT_CTRL_CMD_INIT_TYPE:
      case FACT_CTRL_CMD_HALT_TYPE:
      case FACT_CTRL_CMD_SHUTDOWN_TYPE:
      case FACT_CTRL_CMD_ABORT_TYPE:
      case FACT_CTRL_CMD_DROP_PACKAGE_TYPE:
      case FACT_CTRL_CMD_SET_CONVEYOR_TYPE:
      case FACT_CTRL_CMD_CLEAR_CARGO_TYPE:
	if (serialNumber != factCtrlStat.echo_serial_number) 
	  {
	    if( factCtrlSet.debug > 0 )
	      printf( "New %s command received\n", 
		      factCtrl_symbol_lookup(cmdType) );
	    factCtrlStat.command_type = cmdType;
	    factCtrlStat.echo_serial_number = serialNumber;
	    factCtrlStat.state = NEW_COMMAND;
	    factCtrlStat.status = RCS_EXEC;
	  }
	break;
      default:
	fprintf(stderr,"%s: unknown command: %d\n", argv[0], (int) cmdType);
	break;
      } // switch (cmdType)

      /****************************************************************** 
       * READ CONFIG - Configuration information is received, and the 
       *               system is prepared to change its setting. A separate 
       *               configuration channel is provided to allow for control 
       *               parameters to be changed without interrupting the 
       *               current controller. For example, a user may want to 
       *               change a systems cycle time without interrupting a 
       *               complex control function. The configuration channel is 
       *               used to change the cycle time parameters without 
       *               aborting the control algorithm.
       */

      // read config
      cfgType = factCtrlCfgBuf->read();
      serialNumber = factCtrlCfgBuf->get_address()->serial_number;
      switch (cfgType) {
      case 0:
	// no new command
	break;
      case -1:
	// comm error
	fprintf(stderr,"%s: NML config read error\n", argv[0]);
	break;
      case FACT_CTRL_CFG_NOP_TYPE:
      case FACT_CTRL_CFG_DEBUG_TYPE:
      case FACT_CTRL_CFG_CYCLE_TIME_TYPE:
      case FACT_CTRL_CFG_SET_PACK_FILE_TYPE:
	if (serialNumber != factCtrlSet.echo_serial_number) 
	  {
	    if( factCtrlSet.debug > 0 )
	      printf( "New %s config received\n", factCtrl_symbol_lookup(cfgType) );
	    factCtrlSet.command_type = cfgType;
	    factCtrlSet.echo_serial_number = serialNumber;
	    factCtrlSet.state = NEW_COMMAND;
	    factCtrlSet.status = RCS_EXEC;
	  }
	break;

      default:
	fprintf(stderr,"%s: unknown config: %d\n", argv[0], (int) cfgType);
	break;
      } // switch (cfgType)

      /****************************************************************** 
       * COMMAND EXECUTION - Command execution is carried out. 
       *                     All of the modules run on a fixed cycle time. 
       *                     Therefore, command functions must either 
       *                     guarantee that they finish in under the cycle time, 
       *                     or provide for being reentrant. Although this 
       *                     example refers to the command execution as being 
       *                     finite state machine (FSM) based, search based 
       *                     planning systems have also been implemented under 
       *                     this framework. The searcher simply stores its 
       *                     current location in the search when its time 
       *                     expires and then resumes searching on the next cycle.
       */
      // handle command state tables
      switch (factCtrlStat.command_type) {
      case FACT_CTRL_CMD_NOP_TYPE:
	commandDone = doCmdNop(&factCtrlStat);
	break;
      case FACT_CTRL_CMD_INIT_TYPE:
	commandDone = doCmdInit(&factCtrlStat);
	break;
      case FACT_CTRL_CMD_HALT_TYPE:
	commandDone = doCmdHalt(&factCtrlStat);
	break;
      case FACT_CTRL_CMD_ABORT_TYPE:
	commandDone = doCmdAbort(&factCtrlStat);
	break;
      case FACT_CTRL_CMD_SHUTDOWN_TYPE:
	commandDone = doCmdShutdown(&factCtrlStat);
	break;
      case FACT_CTRL_CMD_DROP_PACKAGE_TYPE:
	commandDone = doCmdPackage(&factCtrlStat,
				   (FactCtrlCmdDropPackage*)factCtrlCmdBuf->get_address());
	break;
      case FACT_CTRL_CMD_SET_CONVEYOR_TYPE:
	commandDone = doCmdSetConveyor(&factCtrlStat, 
				       (FactCtrlCmdSetConveyor*)factCtrlCmdBuf->get_address());
	break;
      case FACT_CTRL_CMD_CLEAR_CARGO_TYPE:
	commandDone = doCmdClearCargo(&factCtrlStat );
	break;
      default:
	// no command to handle
	commandDone = true;
	break;
      } // switch (cmd)

      /****************************************************************** 
       * CONFIGURE EXECUTION - Configuration is carried out
       */
      // handle config state tables
      switch (factCtrlSet.command_type) {
      case FACT_CTRL_CFG_NOP_TYPE:
	doCfgNop(&factCtrlSet);
	break;
      case FACT_CTRL_CFG_DEBUG_TYPE:
	doCfgDebug(&factCtrlSet,
		   ((FactCtrlCfgDebug *) factCtrlCfgBuf->get_address())->debug);
	break;
      case FACT_CTRL_CFG_CYCLE_TIME_TYPE:
	doCfgCycleTime(&factCtrlSet,
		       ((FactCtrlCfgCycleTime *) factCtrlCfgBuf->get_address())->cycleTime);
	break;
	case FACT_CTRL_CFG_SET_PACK_FILE_TYPE:
	doCfgSetPackFile(&factCtrlSet, 
			    ((FactCtrlCfgSetPackFile *) factCtrlCfgBuf->get_address())->fileName);
	break;
      default:
	// no config to handle
	break;
      } // switch (cfg)

      /****************************************************************** 
       * UPDATE STATUS AND SETTING - The current module status and the 
       *                             configuration's settings are sent
       *                             out over communications buffers.
       */
      // update status and settings
      factCtrlStatBuf->write(&factCtrlStat);
      if( settingsAreFresh )
	{
	  factCtrlSetBuf->write(&factCtrlSet);
	  settingsAreFresh = false;
	}

      //      timer->wait();
    }
  return cleanupNmlBuffers();
}
/*
  Modification history:

  $Log: factoryCtrl.cc,v $
  Revision 1.9  2009/05/01 18:02:05  proctor
  Added missing dependency of factoryCtrl to simware.h in Makefile.am and
  included it in factoryCtrl.cc -- this is so that the header gets added
  to the dist .tar.gz

  Revision 1.8  2009/04/24 20:25:03  dr_steveb
  Fixed nml clean-up that was causing a core dump

  Revision 1.7  2009/04/24 15:39:00  dr_steveb
  Changed shared library to be factory.so

  Revision 1.6  2009/04/24 15:13:51  dr_steveb
  Added parput support for real operation in the shops.

  Revision 1.5  2009/03/30 18:24:58  dr_steveb
  Added ask/tell structure

  Revision 1.4  2009/02/21 18:29:39  dr_steveb
  Several minor bug fixes and enhanced readability

  Revision 1.3  2009/02/19 15:26:43  dr_steveb
  Added temporary settings and status messages.

  Revision 1.2  2009/02/18 20:14:48  dr_steveb
  Fixed ability to be on remote machine.

  Revision 1.1  2009/02/17 18:42:18  dr_steveb
  Initial add of factory controller app and control shell

*/
