/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file mfd.cc

  \brief MOAST For Dummies class. Designed to ease the learning curve for using MOAST.

  This class is designed to provide a simple to use interface that allows novice programmers to create a MOAST application with '5 lines of code or less'. It has interfaces into each echelon of MOAST and can do some sample behaviors at each echelon. The calling sequence is to initialize the appropriate system with the constructor, and then to simply issue appropriate commands


  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.3 $
  $Date: 2009/05/07 19:01:07 $
  \endcode

  \author Stephen Balakirsky
  \date April 7, 2009
*/

#include <stdio.h>
#include <mfd.hh>

/////////////////////////////////////////////////////////////////////////////
/*! 
 Constructor
\param nmlFileIn The name of the nml configuraiton file to open. This is defaulted in mfd.hh
\param nmlNameIn The name of this process in the configuration file. This is defaulted in mfd.hh
*/
Mfd::Mfd(char *nmlFileIn, char *nmlNameIn )
{
  myEchelon = MFD_NO_ECHELON;
  currentError = MFD_ERROR_NONE;
  mySerialNumber = 18;

  nmlFile = new char[strlen(nmlFileIn)];
  strcpy( nmlFile, nmlFileIn );

  nmlName = new char[strlen(nmlNameIn)];
  strcpy( nmlName, nmlNameIn );
  
  servoMob = NULL;
  servoSp = NULL;
  servoEff = NULL;
  primMob = NULL;
  primSp = NULL;
  amMob = NULL;
  amSp = NULL;
}

/////////////////////////////////////////////////////////////////////////////
/*! 
 Destructor
*/
Mfd::~Mfd()
{
  delete [] nmlFile;
  delete [] nmlName;
}

/////////////////////////////////////////////////////////////////////////////
/*! Returns the error number from the last call
\return currentError
*/
MfdError Mfd::getErrorNumber()
{
  return( currentError );
}

/////////////////////////////////////////////////////////////////////////////
/*!  Initialize a new echelon.
     This includes opening all of the necessary channels and sending out 
     initialization commands
\param echelonIn The echelon of control that is desired for the instance. For example MFD_SERVO.
\param vehIDIn   The id of this vehicle. This number is appended to all of the NML buffers.
\return True on success
\return False on failure
*/
bool Mfd::init(MfdEchelon echelonIn, int vehIDIn )
{
  ServoMobJACmdInit servoMobJACmdInit;
  ServoSPCmdInit servoSPCmdInit;
  ServoSPCmdGo servoSPCmdGo;
  ServoEffJACmdInit servoEffJACmdInit;

  PrimMobJACmdInit primMobJACmdInit;
  PrimSPCmdInit primSPCmdInit;
  PrimSPCmdGo primSPCmdGo;

  AmMobJACmdInit amMobJACmdInit;
  AmSPCmdInit amSPCmdInit;
  AmSPCmdGo amSPCmdGo;

  FactCtrlCmdInit factCtrlCmdInit;
  int setType;

  currentError = MFD_ERROR_NONE;

  switch( echelonIn )
    {
    case MFD_NO_ECHELON:
      break;
    case MFD_SERVO:
      //
      // init mobility
      if( servoMob != NULL )
	delete servoMob;
      servoMob = new NmlCluster();
      // open servo mob data channels, for now there are none
      servoMob->dataBuf = NULL;
      if( !openNML( servoMob, nmlFile, SERVO_MOB_JA_CMD_NAME, SERVO_MOB_JA_STAT_NAME,
		    SERVO_MOB_JA_CFG_NAME, SERVO_MOB_JA_SET_NAME,
		    nmlName, servoMobJA_format, vehIDIn, 0) )
	{
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      servoMobJACmdInit.serial_number = ++mySerialNumber;
      servoMob->cmdBuf->write(servoMobJACmdInit);
      if( !waitForCmdDone( .1, mySerialNumber, servoMob ) )
	{
	  currentError = MFD_ERROR_SERVO_ECHELON_CMD;
	  return false;
	}
      setType =servoMob->setBuf->read();
      if( setType!=SERVO_MOB_JA_SET_TYPE &&
	  setType!= SERVO_MOB_JA_SET_GRD_VEH_TYPE &&
	  setType!=SERVO_MOB_JA_SET_AIR_BOT_TYPE)
	{
	  printf( "Error from reading settings buffer %d\n", setType );
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      servoMob->setMsg = (RCS_STAT_MSG*)servoMob->setBuf->get_address();
      printf( "finished servoMobInit\n" );
      //
      // init sp if necessary
      if( ((ServoMobJASet *)servoMob->setMsg)->senCount > 0 )
	{
	  servoSp = new NmlCluster();
	  // open servo sp data channels, for now there are none
	  servoSp->dataBuf = NULL;
	  if( !openNML( servoSp, nmlFile, SERVO_SP_CMD_NAME, SERVO_SP_STAT_NAME,
			SERVO_SP_CFG_NAME, SERVO_SP_SET_NAME,
			nmlName, servoSP_format, vehIDIn, 0) )
	    {
	      currentError = MFD_ERROR_NML_CONNECT;
	      return false;
	    }
	  servoSPCmdInit.serial_number = ++mySerialNumber;
	  servoSp->cmdBuf->write(servoSPCmdInit);
	  if( !waitForCmdDone( .1, mySerialNumber, servoSp ) )
	    {
	      currentError = MFD_ERROR_SERVO_ECHELON_CMD;
	      return false;
	    }
	  printf( "finished servoSPInit\n" );

	  servoSPCmdGo.serial_number = ++mySerialNumber;
	  servoSp->cmdBuf->write(servoSPCmdGo);
	  printf( "finished servoSPGo\n" );
	}
      else
	{
	  servoSp->dataBuf = NULL;
	  servoSp->cmdBuf = NULL;
	  servoSp->statBuf = NULL;
	  servoSp->setBuf = NULL;
	  servoSp->cfgBuf = NULL;
	}
      if( ((ServoMobJASet *)servoMob->setMsg)->effCount > 0 )
	{
	  servoEff = new NmlCluster();
	  // open servo eff data channels, for now there are none
	  servoEff->dataBuf = NULL;
	  if( !openNML( servoEff, nmlFile, SERVO_EFF_JA_CMD_NAME, SERVO_EFF_JA_STAT_NAME,
			SERVO_EFF_JA_CFG_NAME, SERVO_EFF_JA_SET_NAME,
			nmlName, servoEffJA_format, vehIDIn, 1) )
	    {
	      currentError = MFD_ERROR_NML_CONNECT;
	      return false;
	    }
	  servoEffJACmdInit.serial_number = ++mySerialNumber;
	  servoEff->cmdBuf->write(servoEffJACmdInit);
	  if( !waitForCmdDone( .1, mySerialNumber, servoEff ) )
	    {
	      currentError = MFD_ERROR_SERVO_ECHELON_CMD;
	      return false;
	    }
	  printf( "finished servoEffInit\n" );

	}
      myEchelon = MfdEchelon((int)myEchelon | MFD_SERVO);
      break;
    case MFD_PRIM:
      if( primMob != NULL )
	delete primMob;
      //
      // prim Mob Buffers
      //
      primMob = new NmlCluster();
      // open prim mob data channels, for now there are none
      primMob->dataBuf = NULL;
      if( !openNML( primMob, nmlFile, PRIM_MOB_JA_CMD_NAME, PRIM_MOB_JA_STAT_NAME,
		    PRIM_MOB_JA_CFG_NAME, PRIM_MOB_JA_SET_NAME,
		    nmlName, primMobJA_format, vehIDIn, 0) )
	{
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      //
      //prim SP buffers
      //
      primSp = new NmlCluster();
      /*
      if( !openPrimDataNML( primSp ) )
	{
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      */
      // connect to navData and line scan buffers
      //      primSp->dataBuf = new NML*[PRIM_SP_MAX_BUFFERS];
      primSp->dataBuf = new NML* [PRIM_SP_MAX_BUFFERS];
      if( !openNMLData( &primSp->dataBuf[PRIM_LINE_SCAN_1], nmlFile,
			PRIM_SP_LINESCAN_NAME, nmlName, sensorData_format, vehIDIn, 0 ) )
	fprintf( stderr, "Non-fatal error connecting to %s: %s\n",
		 PRIM_SP_LINESCAN_NAME, mfdErrorStr());
      if( !openNMLData( &primSp->dataBuf[PRIM_POSE], nmlFile,
			NAV_DATA_EXT_NAME, nmlName, navDataExt_format, vehIDIn, 0 ) )
	fprintf( stderr, "Non-fatal error connecting to %s: %s\n",
		 NAV_DATA_EXT_NAME, mfdErrorStr());

      if( !openNML( primSp, nmlFile, PRIM_SP_CMD_NAME, PRIM_SP_STAT_NAME,
		    PRIM_SP_CFG_NAME, PRIM_SP_SET_NAME,
		    nmlName, primSP_format, vehIDIn, 0) )
	{
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      primMobJACmdInit.serial_number = ++mySerialNumber;
      primMob->cmdBuf->write(primMobJACmdInit);
      if( !waitForCmdDone( .1, mySerialNumber, primMob ) )
	{
	  currentError = MFD_ERROR_PRIM_ECHELON_CMD;
	  return false;
	}
      printf( "finished primMobInit\n" );

      primSPCmdInit.serial_number = ++mySerialNumber;
      primSp->cmdBuf->write(primSPCmdInit);
      if( !waitForCmdDone( .1, mySerialNumber, primSp ) )
	{
	  currentError = MFD_ERROR_PRIM_ECHELON_CMD;
	  return false;
	}
      printf( "finished primSPInit\n" );

      primSPCmdGo.serial_number = ++mySerialNumber;
      primSp->cmdBuf->write(primSPCmdGo);
      printf( "finished primSPGo\n" );
      myEchelon = MfdEchelon((int)myEchelon | MFD_PRIM);
      break;
    case MFD_AM:
      if( amMob != NULL )
	delete amMob;
      amMob = new NmlCluster();
      // open am mob data channels, for now there are none
      amMob->dataBuf = NULL;
      if( !openNML( amMob, nmlFile, AM_MOB_JA_CMD_NAME, AM_MOB_JA_STAT_NAME,
		    AM_MOB_JA_CFG_NAME, AM_MOB_JA_SET_NAME,
		    nmlName, amMobJA_format, vehIDIn, 0) )
	{
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      amSp = new NmlCluster();
      // open am sp data channels, for now there are none
      amSp->dataBuf = NULL;
      if( !openNML( amSp, nmlFile, AM_SP_CMD_NAME, AM_SP_STAT_NAME,
		    AM_SP_CFG_NAME, AM_SP_SET_NAME,
		    nmlName, amSP_format, vehIDIn, 0) )
	{
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      amMobJACmdInit.serial_number = ++mySerialNumber;
      amMob->cmdBuf->write(amMobJACmdInit);
      if( !waitForCmdDone( .1, mySerialNumber, amMob ) )
	{
	  currentError = MFD_ERROR_AM_ECHELON_CMD;
	  return false;
	}
      printf( "finished amMobInit\n" );

      amSPCmdInit.serial_number = ++mySerialNumber;
      amSp->cmdBuf->write(amSPCmdInit);
      if( !waitForCmdDone( .1, mySerialNumber, amSp ) )
	{
	  currentError = MFD_ERROR_AM_ECHELON_CMD;
	  return false;
	}
      printf( "finished amSPInit\n" );

      amSPCmdGo.serial_number = ++mySerialNumber;
      amSp->cmdBuf->write(amSPCmdGo);
      printf( "finished amSPGo\n" );

      myEchelon = MfdEchelon((int)myEchelon | MFD_AM);
      break;
    case MFD_FACTORY:
      if( factoryCtrl != NULL )
	delete factoryCtrl;
      factoryCtrl = new NmlCluster();
      // open factory data channels, for now there are none
      factoryCtrl->dataBuf = NULL;
      if( !openNML( factoryCtrl, nmlFile, FACT_CTRL_CMD_NAME, FACT_CTRL_STAT_NAME,
		    FACT_CTRL_CFG_NAME, FACT_CTRL_SET_NAME,
		    nmlName, factCtrl_format, vehIDIn, 0) )
	{
	  currentError = MFD_ERROR_NML_CONNECT;
	  return false;
	}
      factCtrlCmdInit.serial_number = ++mySerialNumber;
      factoryCtrl->cmdBuf->write(factCtrlCmdInit);
      if( !waitForCmdDone( .1, mySerialNumber, factoryCtrl ) )
	{
	  currentError = MFD_ERROR_FACTORY_ECHELON_CMD;
	  return false;
	}
      printf( "finished factory Init\n" );
      myEchelon = MfdEchelon((int)myEchelon | MFD_FACTORY);
      break;
    default:
      printf( "Echelon %s not yet supported\n", mfdEchelonStr(echelonIn) );
      currentError = MFD_ERROR_UNSUPPORTED_FEATURE;
      return false;
      break;
    }
  
  return true;
}


/////////////////////////////////////////////////////////////////////////////
/*! Returns the echelons that are initialized
\return myEchelon
*/
MfdEchelon Mfd::isInitialized()
{
  return( myEchelon );
}


/////////////////////////////////////////////////////////////////////////////
/*!  Translate an mfdEchelon into a string
\param echelonIn The echelon to be translated
\return String on success
\return Null on failure
*/
const char *Mfd::mfdEchelonStr( MfdEchelon echelonIn )
{
  switch(echelonIn)
    {
    case MFD_NO_ECHELON:
      return "MFD_NO_ECHELON";
      break;
    case MFD_SERVO:
      return "MFD_SERVO";
      break;
    case MFD_PRIM:
      return "MFD_PRIM";
      break;
    case MFD_AM:
      return "MFD_AM";
      break;
    case MFD_VEH:
      return "MFD_VEH";
      break;
    case MFD_SECT:
      return "MFD_SECT";
      break;
    default:
      printf( "Unknown echelon to be translated %d\n", echelonIn );
      break;
    }
  return NULL;
}


/////////////////////////////////////////////////////////////////////////////
/*!  Translate the current mfdError into a string
\return String on success
\return Null on failure
*/
const char *Mfd::mfdErrorStr()
{
  switch(currentError)
    {
    case MFD_ERROR_NONE:
      return "MFD:No error";
      break;
    case MFD_ERROR_NML_CONNECT:
      return "MFD:Error from connecting to nml buffer";
      break;
    case MFD_ERROR_NML_DATA_CONNECT:
      return "MFD:Error from connecting to data buffer";
      break;
    case MFD_ERROR_NML_NO_MESSAGE:
      return "MFD:Error no new message available";
      break;
    case MFD_ERROR_NML_FAULT:
      return "MFD:Error general NML fault";
      break;
    case MFD_ERROR_UNSUPPORTED_FEATURE:
      return "MFD:A call was made to an unsupported feature";
      break;
    case MFD_ERROR_SERVO_ECHELON_NOT_INITIALIZED:
      return "MFD: A servo echelon call was made, and servo is not initialized";
      break;
    case MFD_ERROR_PRIM_ECHELON_NOT_INITIALIZED:
      return "MFD: A prim echelon call was made, and prim is not initialized";
      break;
    case MFD_ERROR_AM_ECHELON_NOT_INITIALIZED:
      return "MFD: An AM echelon call was made, and am is not initialized";
      break;
    case MFD_ERROR_FACTORY_ECHELON_NOT_INITIALIZED:
      return "MFD: A Factory echelon call was made, and factory is not initialized";
      break;
    case MFD_ERROR_SERVO_ECHELON_CMD:
      return "MFD: A servo echelon command failed";
      break;
    case MFD_ERROR_PRIM_ECHELON_CMD:
      return "MFD: A prim echelon command failed";
      break;
    case MFD_ERROR_AM_ECHELON_CMD:
      return "MFD: An am echelon command failed";
      break;
    case MFD_ERROR_FACTORY_ECHELON_CMD:
      return "MFD: A factory echelon command failed";
      break;
    case MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED:
      return "MFD: An unitialized subsystem was sent a command";
      break;
    default:
      printf( "Unknown error to be translated %d\n", currentError );
      break;
    }
  return NULL;
}

/////////////////////////////////////////////////////////////////////////////
/*!  Open an echelons NML buffers
\param chanIn   The sturcture to hold the channel pointers
\param nmlFile  The nml configuration file name
\param cmdName  The name of the echelon's command channel
\param statName The name of the echelon's status channel
\param cfgName  The name of the echelon's configuraiton channel
\param setName  The name of the echelon's settings channel
\param myName   The nml name of this process
\param fPtr     A pointer to the NML format function
\param botNum   The id of this robot
\param compID   The component or subsystem id for this set of buffers
\return true on success
\return false on failure
*/
bool Mfd::openNML(NmlCluster *chanIn, char *nmlFile, char *cmdName, char *statName, 
		  char *cfgName, char *setName,
		  char *myName, NML_FORMAT_PTR fPtr, int botNum, int compID )
{
  char bufferName[MOAST_NML_BUFFER_NAME_LEN];
  char suffix[10];
  
  // Check to see if there is a component id, this changes naming convention
  if(compID==0)
    sprintf(suffix,"%d",botNum);
  else
    sprintf(suffix,"%d_%d",botNum,compID);
  
  sprintf( bufferName, "%s%s", cmdName, suffix );
  
  if(1)
    printf( "Opening NML CMD Buffer  : %s\n", bufferName );
  chanIn->cmdBuf = new RCS_CMD_CHANNEL(fPtr, 
				       bufferName,
				       myName,
				       nmlFile);

  if(!chanIn->cmdBuf->valid()) 
    {
      delete chanIn->cmdBuf;
      chanIn->cmdBuf = NULL;
      return false;
    }

  sprintf( bufferName, "%s%s", cfgName, suffix );  

  if(1)
    printf( "Openning NML CFG Buffer  : %s\n", bufferName );
  
  chanIn->cfgBuf = new RCS_CMD_CHANNEL(fPtr, 
				       bufferName,
				       myName,
				       nmlFile);
  if(!chanIn->cfgBuf->valid()) 
    {
      delete chanIn->cfgBuf;
      chanIn->cfgBuf = NULL;
      return false;
    }

  sprintf( bufferName, "%s%s", statName, suffix );  
  
  if(1)
    printf("Openning NML STAT Buffer  : %s\n", bufferName );
  
  chanIn->statBuf = new RCS_STAT_CHANNEL(fPtr, 
					 bufferName,
					 myName,
					 nmlFile);
  if(!chanIn->statBuf->valid()) 
    {
      delete chanIn->statBuf;
      chanIn->statBuf = NULL;
      return false;
    }

  sprintf( bufferName, "%s%s", setName, suffix );  
  
  if(1)
    printf("Openning NML SET Buffer  : %s\n", bufferName );
  
  chanIn->setBuf = new RCS_STAT_CHANNEL(fPtr, 
					bufferName,
					myName,
					nmlFile );
  if(!chanIn->setBuf->valid()) 
    {
      delete chanIn->setBuf;
      chanIn->setBuf = NULL;
      return false;
    }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/*!  Open an NML data buffer
\param chanIn   The channel pointer
\param nmlFile  The nml configuration file name
\param name     The name of the echelon's command channel
\param myName   The nml name of this process
\param fPtr     A pointer to the NML format function
\param botNum   The id of this robot
\param compID   The component or subsystem id for this set of buffers
\return true on success
\return false on failure
*/
bool Mfd::openNMLData(NML **chanIn, char *nmlFile, char *name,
		  char *myName, NML_FORMAT_PTR fPtr, int botNum, int compID )
{
  char bufferName[MOAST_NML_BUFFER_NAME_LEN];
  char suffix[10];
  
  // Check to see if there is a component id, this changes naming convention
  if(compID==0)
    sprintf(suffix,"%d",botNum);
  else
    sprintf(suffix,"%d_%d",botNum,compID);
  
  sprintf( bufferName, "%s%s", name, suffix );
  
  if(1)
    printf( "Opening NML CMD Buffer  : %s\n", bufferName );
  *chanIn = new NML(fPtr, 
		   bufferName,
		   myName,
		   nmlFile);

  if(!(*chanIn)->valid()) 
    {
      delete *chanIn;
      *chanIn = NULL;
      currentError = MFD_ERROR_NML_DATA_CONNECT;
      return false;
    }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/* ! wait for the current command as indicated by cmdId to finish
\param interval How often to read status
\param cmdID    The id of the command to wait for
\param checkMe  The cluster containing the status channel pointer and status structure
\return true on success
\return false on failure
*/
bool Mfd::waitForCmdDone( double interval, int cmdId, NmlCluster *checkMe )
{
  RCS_TIMER timer(interval);
  int type;

  for(;;)
    {
      type = checkMe->statBuf->read();
      switch( type )
	{
	case 0: // no new data
	  break;
	case SERVO_MOB_JA_STAT_TYPE:
	case SERVO_MOB_JA_STAT_GRD_VEH_TYPE:
	case SERVO_MOB_JA_STAT_AIR_BOT_TYPE:
	case SERVO_EFF_JA_STAT_TYPE:
	case SERVO_SP_STAT_TYPE:
	case PRIM_MOB_JA_STAT_TYPE:
	case PRIM_SP_STAT_TYPE:
	case AM_MOB_JA_STAT_TYPE:
	case AM_SP_STAT_TYPE:
	case FACT_CTRL_STAT_TYPE:
	  checkMe->statMsg = (RCS_STAT_MSG_V2*)checkMe->statBuf->get_address();
	  if( checkMe->statMsg->echo_serial_number != cmdId )
	    break;
	  if( checkMe->statMsg->status == RCS_DONE )
	    return(true);
	  else if( checkMe->statMsg->status == RCS_ERROR )
	    return(false);
	  break;
	default:
	  printf( "Unknown status received %d\n", type );
	  return false;
	  break;
	}
      timer.wait();
    }
}

/////////////////////////////////////////////////////////////////////////////
//// Servo echelon functions
/////////////////////////////////////////////////////////////////////////////
/*!  generate a skid steering command to the servo
\param effID  The id of the effector to control
\param opcode The opcode to send from servoEffJA.hh
\param param  The parameter to send
\return true on success
\return false on failure
*/
bool Mfd::servoEffOpcode( int effID, ServoEffJAOpcodeType opcode, double param )
{
  ServoEffJACmdOpcode servoEffJACmdOpcode;

  if( (myEchelon & MFD_SERVO) == 0 )
    {
      currentError = MFD_ERROR_SERVO_ECHELON_NOT_INITIALIZED;
      return false;
    }
  else
    currentError = MFD_ERROR_NONE;
  if( servoEff == NULL )
    {
      currentError = MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED;
      return false;
    }

  servoEffJACmdOpcode.serial_number = ++mySerialNumber;
  servoEffJACmdOpcode.effID = effID;
  servoEffJACmdOpcode.opcode = opcode;
  servoEffJACmdOpcode.param = param;
  servoEff->cmdBuf->write(servoEffJACmdOpcode);
  if( !waitForCmdDone( .1, mySerialNumber, servoEff ) )
    {
      currentError = MFD_ERROR_SERVO_ECHELON_CMD;
      return false;
    }
  
  return true;
}

/*!  generate a skid steering command to the servo
\param left  The left track/wheel velocity in radians/sec
\param right The right track/wheel velocity in radians/sec
\return true on success
\return false on failure
*/
bool Mfd::servoSkidSteer( double left, double right )
{
  ServoMobJACmdSkid servoMobJACmdSkid;
  if( (myEchelon & MFD_SERVO) == 0 )
    {
      currentError = MFD_ERROR_SERVO_ECHELON_NOT_INITIALIZED;
      return false;
    }
  else
    currentError = MFD_ERROR_NONE;
  if( servoMob == NULL )
    {
      currentError = MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED;
      return false;
    }

  servoMobJACmdSkid.wLeft = left;
  servoMobJACmdSkid.wRight = right;
  servoMobJACmdSkid.serial_number = ++mySerialNumber;
  servoMob->cmdBuf->write(servoMobJACmdSkid);
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//// Prim echelon functions
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
/*! get pose information from the NavDataExt channel. See navDataExt.hh for
    a descrption of the members of this class

\param poseInfo The pose information. Note that this pointer is only valid
                until the next call to this function.
\return true on success
\return false on failure
 */
bool Mfd::primGetPose( NavDataExt **poseInfo )
{
  int type;

  if( (myEchelon & MFD_PRIM) == 0 )
    {
      currentError = MFD_ERROR_PRIM_ECHELON_NOT_INITIALIZED;
      *poseInfo = NULL;
      return false;
    }
  if( primSp->dataBuf[PRIM_POSE] == NULL )
    {
      currentError = MFD_ERROR_NML_DATA_CONNECT;
      *poseInfo = NULL;
      return false;
    }
  currentError = MFD_ERROR_NONE;

  type = primSp->dataBuf[PRIM_POSE]->read();
  switch( type )
    {
    case 0: // no new data
      currentError = MFD_ERROR_NML_NO_MESSAGE;
      break;
    case -1: // communications error
      primSp->dataBuf[PRIM_POSE] = NULL;
      *poseInfo = NULL;
      currentError = MFD_ERROR_NML_FAULT;
      return false;
      break;
    case NAV_DATA_EXT_TYPE:
      *poseInfo = (NavDataExt*)primSp->dataBuf[PRIM_POSE]->get_address();
      break;
    default:
      *poseInfo = NULL;
      currentError = MFD_ERROR_NML_FAULT;
      break;
    }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/*! get raw laser data from the sensor data channel servoSPLinescan.
    The structure SensorData1D is filled in with the data. If a
    sensorID is provided, then data from that id is read. If not
    sensorID is given, then the routine defaults to the first laser.

\param sensorData The data from the laser. Note that this pointer is only valid
                  until the next call to this function.
\return true on success
\return false on failure
 */
bool Mfd::primGetLaser( SensorData1D **sensorData, int sensorID )
{
  int type;
  int dataBufID;

  if( (myEchelon & MFD_PRIM) == 0 )
    {
      currentError = MFD_ERROR_PRIM_ECHELON_NOT_INITIALIZED;
      *sensorData = NULL;
      return false;
    }
  switch( sensorID )
    {
    case 1:
      dataBufID = PRIM_LINE_SCAN_1;
      break;
    case 2:
      dataBufID = PRIM_LINE_SCAN_2;
      break;
    case 3:
      dataBufID = PRIM_LINE_SCAN_3;
      break;
    }
  if( primSp->dataBuf[dataBufID] == NULL )
    {
      currentError = MFD_ERROR_NML_DATA_CONNECT;
      *sensorData = NULL;
      return false;
    }
  currentError = MFD_ERROR_NONE;

  type = primSp->dataBuf[dataBufID]->read();
  switch( type )
    {
    case 0: // no new data
      currentError = MFD_ERROR_NML_NO_MESSAGE;
      break;
    case -1: // communications error
      primSp->dataBuf[dataBufID] = NULL;
      *sensorData = NULL;
      currentError = MFD_ERROR_NML_FAULT;
      return false;
      break;
    case SENSOR_DATA_1D_TYPE:
      *sensorData = (SensorData1D*)primSp->dataBuf[dataBufID]->get_address();
      break;
    default:
      *sensorData = NULL;
      currentError = MFD_ERROR_NML_FAULT;
      break;
    }
  return true;
}


/////////////////////////////////////////////////////////////////////////////
/*!  generate a waypoint command to prim
\param waypoints An array of waypoints to drive to 
\param numPts    The number of waypoints in the array
\return true on success
\return false on failure
*/
bool Mfd::primMoveWaypoint( Waypoint *waypoints, int numPts )
{
  PrimMobJACmdMoveWaypoint primMobJACmdMoveWaypoint;

  if( (myEchelon & MFD_PRIM) == 0 )
    {
      currentError = MFD_ERROR_PRIM_ECHELON_NOT_INITIALIZED;
      return false;
    }
  else
    currentError = MFD_ERROR_NONE;
  if( primMob == NULL )
    {
      currentError = MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED;
      return false;
    }
  if( numPts >= MAX_WAYPOINTS )
    {
      currentError = MFD_ERROR_PRIM_ECHELON_CMD;
      return false;
    }
  for( int count=0; count<numPts; count++ )
    primMobJACmdMoveWaypoint.waypoint[count] = waypoints[count];
  primMobJACmdMoveWaypoint.waypoint_length = numPts;
  primMobJACmdMoveWaypoint.serial_number = ++mySerialNumber;
  primMob->cmdBuf->write(primMobJACmdMoveWaypoint);
  if( !waitForCmdDone( .1, mySerialNumber, primMob ) )
    {
      currentError = MFD_ERROR_PRIM_ECHELON_CMD;
      return false;
    }
  return true;
}

/*!  rotate a skid steered vehicle
\param position  The position in radians to rotate to
\param tol       How close does the rotation need to be? Defaults to .1
\param direction Which direction should it spin? Defaults to MOAST_ROTATE_SHORTEST. 
                 May also be MOAST_ROTATE_CW or MOAST_ROTATE_CCW
\return true on success
\return false on failure
*/
bool Mfd::primRotate( double position, double tol, int direction )
{
  PrimMobJACmdRotate primMobJACmdRotate;

  if( (myEchelon & MFD_PRIM) == 0 )
    {
      currentError = MFD_ERROR_PRIM_ECHELON_NOT_INITIALIZED;
      return false;
    }
  else
    currentError = MFD_ERROR_NONE;
  if( primMob == NULL )
    {
      currentError = MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED;
      return false;
    }

  primMobJACmdRotate.direction = direction;
  primMobJACmdRotate.theta = position;
  primMobJACmdRotate.thetaTol = tol;

  primMobJACmdRotate.serial_number = ++mySerialNumber;
  primMob->cmdBuf->write(primMobJACmdRotate);
  if( !waitForCmdDone( .1, mySerialNumber, primMob ) )
    {
      currentError = MFD_ERROR_PRIM_ECHELON_CMD;
      return false;
    }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//// Am echelon functions
/////////////////////////////////////////////////////////////////////////////
/*!  generate a waypoint command to am
\param waypoints An array of waypoints to drive to 
\param numPts    The number of waypoints in the array
\return true on success
\return false on failure
*/
bool Mfd::amMoveWaypoint( Waypoint *waypoints, int numPts )
{
  AmMobJACmdMoveWaypoint amMobJACmdMoveWaypoint;

  if( (myEchelon & MFD_AM) == 0 )
    {
      currentError = MFD_ERROR_AM_ECHELON_NOT_INITIALIZED;
      return false;
    }
  else
    currentError = MFD_ERROR_NONE;
  if( numPts >= MAX_WAYPOINTS )
    {
      currentError = MFD_ERROR_AM_ECHELON_CMD;
      return false;
    }
  if( amMob == NULL )
    {
      currentError = MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED;
      return false;
    }
  for( int count=0; count<numPts; count++ )
    amMobJACmdMoveWaypoint.waypoint[count] = waypoints[count];
  amMobJACmdMoveWaypoint.waypoint_length = numPts;
  amMobJACmdMoveWaypoint.serial_number = ++mySerialNumber;
  amMob->cmdBuf->write(amMobJACmdMoveWaypoint);
  if( !waitForCmdDone( .1, mySerialNumber, amMob ) )
    {
      currentError = MFD_ERROR_AM_ECHELON_CMD;
      return false;
    }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//// Factory echelon functions
/////////////////////////////////////////////////////////////////////////////
/*!  drop a package from a box chute
\param chuteName The name of the box chute to operate
\return true on success
\return false on failure
*/
bool Mfd::factoryDropPackage( char *chuteName )
{
  FactCtrlCmdDropPackage factCtrlCmdDropPackage;

  if( (myEchelon & MFD_FACTORY) == 0 )
    {
      currentError = MFD_ERROR_FACTORY_ECHELON_NOT_INITIALIZED;
      return false;
    }
  else
    currentError = MFD_ERROR_NONE;
  if( factoryCtrl == NULL )
    {
      currentError = MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED;
      return false;
    }

  if( strlen(chuteName) >= MAX_FACTORY_NAME )
    {
      currentError = MFD_ERROR_FACTORY_ECHELON_CMD;
      return false;
    }
  strcpy( factCtrlCmdDropPackage.chuteName, chuteName );
  factCtrlCmdDropPackage.serial_number = ++mySerialNumber;
  factoryCtrl->cmdBuf->write(factCtrlCmdDropPackage);
  if( !waitForCmdDone( .1, mySerialNumber, factoryCtrl ) )
    {
      currentError = MFD_ERROR_FACTORY_ECHELON_CMD;
      return false;
    }
  return true;
}

/*!  set the speed of a conveyor
\param conveyorName The name of the conveyor
\param speed        The speed of the conveyor
\return true on success
\return false on failure
*/
bool Mfd::factorySetConveyor(char *conveyorName, float speed)

{
  FactCtrlCmdSetConveyor factCtrlCmdSetConveyor;

  if( (myEchelon & MFD_FACTORY) == 0 )
    {
      currentError = MFD_ERROR_FACTORY_ECHELON_NOT_INITIALIZED;
      return false;
    }
  else
    currentError = MFD_ERROR_NONE;
  if( factoryCtrl == NULL )
    {
      currentError = MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED;
      return false;
    }

  if( strlen(conveyorName) >= MAX_FACTORY_NAME )
    {
      currentError = MFD_ERROR_FACTORY_ECHELON_CMD;
      return false;
    }
  strcpy( factCtrlCmdSetConveyor.conveyorName, conveyorName );
  factCtrlCmdSetConveyor.speed = speed;
  factCtrlCmdSetConveyor.serial_number = ++mySerialNumber;
  factoryCtrl->cmdBuf->write(factCtrlCmdSetConveyor);
  if( !waitForCmdDone( .1, mySerialNumber, factoryCtrl ) )
    {
      currentError = MFD_ERROR_FACTORY_ECHELON_CMD;
      return false;
    }
  return true;
}
/*
  Modification history:

  $Log: mfd.cc,v $
  Revision 1.3  2009/05/07 19:01:07  dr_steveb
  Check-in of code for travel

  Revision 1.2  2009/04/27 16:09:39  dr_steveb
  Getting ready for demo...

  Revision 1.1  2009/04/24 18:45:17  dr_steveb
  Initial check-in of moast for dummies class

*/
