/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primMisMain.cc
  \brief Prim Mission Package interface to Go Motion controller

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.6 $
  $Date: 2008/12/02 16:27:47 $
  \endcode
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		// printf, sscanf

#define BN "primMis"

#ifndef HAVE_GOMOTION

int main(void)
{
  fprintf(stderr, "%s: unimplemented\n", BN);
  return 1;
}

#else

#include <string.h>		// strlen
#include <stdlib.h>		// exit
#include <signal.h>		// signal, SIGINT
#include <rcs.hh>		// NEW_COMMAND, RCS_DONE, ...
#include <inifile.h>
#include <ulapi.h>
#include <go.h>
#include <trajintf.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include "primMisJA.hh"

static int TRAJ_SHM_KEY = 201;	// the usual key; should get this from .ini

/*
  go_pose_to_pm converts gomotion's go_pose types to posemath's PM_POSE
  types. These are element-by-element the same, but may differ by type, 
  maybe float for go_poses. 
*/
static PM_POSE go_pose_to_pm(go_pose pose)
{
  PM_POSE pm;

  pm.tran.x = (double) pose.tran.x;
  pm.tran.y = (double) pose.tran.y;
  pm.tran.z = (double) pose.tran.z;
  pm.rot.s = (double) pose.rot.s;
  pm.rot.x = (double) pose.rot.x;
  pm.rot.y = (double) pose.rot.y;
  pm.rot.z = (double) pose.rot.z;

  return pm;
}

/*
  pm_pose_to_go converts posemath's PM_POSE types to gomotion's go_pose
  types. These are element-by-element the same, but may differ by type, 
  maybe float for go_poses. 
*/
static go_pose pm_pose_to_go(PM_POSE pose)
{
  go_pose gp;

  gp.tran.x = (go_real) pose.tran.x;
  gp.tran.y = (go_real) pose.tran.y;
  gp.tran.z = (go_real) pose.tran.z;
  gp.rot.s = (go_real) pose.rot.s;
  gp.rot.x = (go_real) pose.rot.x;
  gp.rot.y = (go_real) pose.rot.y;
  gp.rot.z = (go_real) pose.rot.z;

  return gp;
}

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

static void doCmdNop(PrimMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {			// S0
    state_default(stat);
  }
}

static void doCmdInit(PrimMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    stat->admin_state = ADMIN_INITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {			// S0
    state_default(stat);
  }
}

static void doCmdAbort(PrimMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    stat->admin_state = ADMIN_UNINITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {			// S0
    state_default(stat);
  }
}

static void doCmdHalt(PrimMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    stat->admin_state = ADMIN_UNINITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {			// S0
    state_default(stat);
  }
}

static void doCmdShutdown(PrimMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state == ADMIN_UNINITIALIZED) {
      done = 1;
      status_next(stat, RCS_DONE);
    } else {
      status_next(stat, RCS_ERROR);
    }
    state_next(stat, S0);
  } else {			// S0
    state_default(stat);
  }
}

static void doCmdStop(PrimMisJAStat * stat, traj_cmd_struct * traj_cmd, traj_stat_struct * traj_stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state != ADMIN_INITIALIZED) {
      status_next(stat, RCS_ERROR);
      state_next(stat, S0);
    } else {
      traj_cmd->head++;
      // send a stop command to traj
      traj_cmd->type = TRAJ_CMD_STOP_TYPE;
      traj_cmd->serial_number++;
      traj_cmd->tail = traj_cmd->head;
      state_next(stat, S1);	// check for arrival
      status_next(stat, RCS_EXEC);
    }
  } else if (state_match(stat, S1)) {
    if (traj_cmd->serial_number == traj_stat->echo_serial_number) {
      state_next(stat, S2);	// check for inpos
    }
    // don't run a delta clock
  } else if (state_match(stat, S2)) {
    if (traj_stat->inpos) {
      state_next(stat, S0);
      if (traj_stat->status == GO_RCS_STATUS_ERROR) {
	status_next(stat, RCS_ERROR);
      } else {
	status_next(stat, RCS_DONE);
      }
    }
  } else {			// S0
    state_default(stat);
  }
}

static void doCmdMoveJoint(PrimMisJACmdMoveJoint * cmd, PrimMisJAStat * stat, PrimMisJASet * set, traj_cmd_struct * traj_cmd, traj_stat_struct * traj_stat, traj_set_struct * traj_set)
{
  static double dclock;
  traj_cmd_move_joint move;
  int i;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state != ADMIN_INITIALIZED) {
      // need to be initialized
      state_next(stat, S0);
      status_next(stat, RCS_ERROR);
    } else {
      move.id = (go_integer) cmd->serial_number;
      move.time = cmd->time;
      for (i = 0; i < cmd->joints_length; i++) {
	move.d[i] = cmd->joints[i];
	// FIXME-- get the v a j from servo settings
	move.v[i] = move.a[i] = move.j[1] = 1;
      }
      // write it out
      traj_cmd->head++;
      traj_cmd->type = TRAJ_CMD_MOVE_JOINT_TYPE;
      traj_cmd->serial_number++;
      traj_cmd->u.move_joint = move;
      traj_cmd->tail = traj_cmd->head;
      // start delta clock for command arrival
      dclock = 1.0;		// 1 second
      state_next(stat, S1);
      status_next(stat, RCS_EXEC);
    }
  } else if (state_match(stat, S1)) {
    if (traj_cmd->serial_number == traj_stat->echo_serial_number) {
      state_next(stat, S2);
    } else {
      dclock -= set->cycleTime;
      if (dclock <= 0.0) {
	state_next(stat, S0);
	status_next(stat, RCS_ERROR);
      }
    }
  } else if (state_match(stat, S2)) {
    if (traj_stat->inpos) {
      state_next(stat, S0);
      if (traj_stat->status == GO_RCS_STATUS_ERROR) {
	status_next(stat, RCS_ERROR);
      } else {
	status_next(stat, RCS_DONE);
      }
    }
  } else {			// S0
    state_default(stat);
  }
}

static void doCmdMoveWorld(PrimMisJACmdMoveWorld * cmd, PrimMisJAStat * stat, PrimMisJASet * set, traj_cmd_struct * traj_cmd, traj_stat_struct * traj_stat, traj_set_struct * traj_set)
{
  static double dclock;
  traj_cmd_move_world move;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state != ADMIN_INITIALIZED) {
      // need to be initialized
      state_next(stat, S0);
      status_next(stat, RCS_ERROR);
    } else {
      move.id = (go_integer) cmd->serial_number;
      move.type = GO_MOTION_LINEAR;
      // we're doing time-based moves, so set the params to be their maxes
      move.tv = traj_set->max_tvel;
      move.ta = traj_set->max_tacc;
      move.tj = traj_set->max_tjerk;
      move.rv = traj_set->max_rvel;
      move.ra = traj_set->max_racc;
      move.rj = traj_set->max_rjerk;
      move.end = pm_pose_to_go(cmd->end);
      move.time = cmd->time;
      // write it out
      traj_cmd->head++;
      traj_cmd->type = TRAJ_CMD_MOVE_WORLD_TYPE;
      traj_cmd->serial_number++;
      traj_cmd->u.move_world = move;
      traj_cmd->tail = traj_cmd->head;
      // start delta clock for command arrival
      dclock = 1.0;		// 1 second
      state_next(stat, S1);
      status_next(stat, RCS_EXEC);
    }
  } else if (state_match(stat, S1)) {
    if (traj_cmd->serial_number == traj_stat->echo_serial_number) {
      state_next(stat, S2);
    } else {
      dclock -= set->cycleTime;
      if (dclock <= 0.0) {
	state_next(stat, S0);
	status_next(stat, RCS_ERROR);
      }
    }
  } else if (state_match(stat, S2)) {
    if (traj_stat->inpos) {
      state_next(stat, S0);
      if (traj_stat->status == GO_RCS_STATUS_ERROR) {
	status_next(stat, RCS_ERROR);
      } else {
	status_next(stat, RCS_DONE);
      }
    }
  } else {			// S0
    state_default(stat);
  }
}

static void doCfgNop(PrimMisJASet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

static void doCfgCycleTime(PrimMisJACfgCycleTime * cfg, PrimMisJASet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    if (cfg->cycleTime <= 0.0) {
      status_next(set, RCS_ERROR);
    } else {
      set->cycleTime = cfg->cycleTime;
      status_next(set, RCS_DONE);
    }
    state_next(set, S0);
  } else {			// S0
    state_default(set);
  }
}

static void doCfgDebug(PrimMisJACfgDebug * cfg, PrimMisJASet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->debug = cfg->debug;
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

static traj_cmd_struct * traj_cmd_ptr;
static traj_stat_struct pp_traj_stat[2], * traj_stat_ptr, * traj_stat_test;
static traj_set_struct pp_traj_set[2], * traj_set_ptr, * traj_set_test;
static traj_comm_struct * traj_comm_ptr = NULL;
static void * traj_shm = NULL;

static int
get_traj_comm_buffers(void)
{
  double end;
  int got_it;

  // get traj shared memory buffers
  traj_shm = ulapi_shm_new(TRAJ_SHM_KEY, sizeof(traj_comm_struct));
  if (NULL == traj_shm) {
    fprintf(stderr, "can't get traj comm shm\n");
    return 1;
  }
  traj_comm_ptr = (traj_comm_struct *) ulapi_shm_addr(traj_shm);

  // no need to ping-pong our traj command buffer
  traj_cmd_ptr = &traj_comm_ptr->traj_cmd;

  // set up traj status ping-pong buffers
  traj_stat_ptr = &pp_traj_stat[0];
  traj_stat_test = &pp_traj_stat[1];
  // read in traj status, ping-pong style
  for (got_it = 0, end = etime() + 10; etime() < end;) {
    *traj_stat_ptr = traj_comm_ptr->traj_stat;
    if (traj_stat_ptr->head == traj_stat_ptr->tail &&
	traj_stat_ptr->type == TRAJ_STAT_TYPE) {
      got_it = 1;
      break;
    }
    esleep(1);
  }
  if (!got_it) {
    return 1;			// can't read anything, probably not there
  }
  // now that we have it, increment its echo serial number into our command
  traj_cmd_ptr->serial_number = traj_stat_ptr->echo_serial_number + 1;

  // set up traj settings ping-pong buffers
  traj_set_ptr = &pp_traj_set[0];
  traj_set_test = &pp_traj_set[1];
  // and read it in ping-pong style
  for (got_it = 0, end = etime() + 10; etime() < end;) {
    *traj_set_ptr = traj_comm_ptr->traj_set;
    if (traj_set_ptr->head == traj_set_ptr->tail &&
	traj_set_ptr->type == TRAJ_SET_TYPE) {
      got_it = 1;
      break;
    }
    esleep(1);
  }
  if (!got_it) {
    return 1;
  }
  // we don't send any settings here, so don't bother reading serial number

  return 0;
}

static void
free_traj_comm_buffers(void)
{
  if (NULL != traj_shm) {
    ulapi_shm_delete(traj_shm);
    traj_shm = NULL;
  }
  traj_comm_ptr = NULL;
}

static int
update_traj_comm_buffers(void)
{
  traj_stat_struct * stat;
  traj_set_struct * set;

  /* read in traj status and settings, ping-pong style */
  *traj_stat_test = traj_comm_ptr->traj_stat;
  if (traj_stat_test->head == traj_stat_test->tail) {
    stat = traj_stat_ptr;
    traj_stat_ptr = traj_stat_test;
    traj_stat_test = stat;
  }
  /*  */
  *traj_set_test = traj_comm_ptr->traj_set;
  if (traj_set_test->head == traj_set_test->tail) {
    set = traj_set_ptr;
    traj_set_ptr = traj_set_test;
    traj_set_test = set;
  }

  return 0;
}

static int iniLoad(char * prog,
		   char * filename,
		   double * cycletime,
		   int * debug)
{
  INIFILE inifile;
  const char * inistring;
  char * key;
  char * section;
  double d1;
  int i1;
  int retval = 0;

  if (0 != inifile.open(filename)) {
    rcs_print_error("%s: can't open ini file %s\n", prog, filename);
    return 1;
  }

  section = "PRIM_MIS_JA";

  key = "CYCLE_TIME";

  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 != sscanf(inistring, "%lf", &d1) || d1 <= 0.0) {
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
		      prog, inistring, section, key, filename);
      retval = -1;
    } else {
      *cycletime = d1;
    }
  } else {
    rcs_print_error("%s: can't find [%s] %s in '%s'\n",
		    prog, section, key, filename);
    retval = -1;
  }

  key = "DEBUG";

  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 != sscanf(inistring, "%i", &i1)) {
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
		      prog, inistring, section, key, filename);
      retval = -1;
    } else {
      *debug = i1;
    }
  } else {
    rcs_print_error("%s: can't find [%s] %s in '%s'\n",
		    prog, section, key, filename);
    retval = -1;
  }

  inifile.close();

  return retval;
}

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define DEFAULT_CYCLE_TIME 0.1

int main(int argc, char *argv[])
{
  char nmlFile[MOAST_FILE_NAME_LEN] = DEFAULT_NML_FILE;
  char iniFile[MOAST_FILE_NAME_LEN] = DEFAULT_INI_FILE;
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  int retval;
  int bufnum;
  RCS_CMD_CHANNEL * primMisJACmdBuf;
  RCS_STAT_CHANNEL * primMisJAStatBuf;
  RCS_CMD_CHANNEL * primMisJACfgBuf;
  RCS_STAT_CHANNEL * primMisJASetBuf;
  PrimMisJACmdMoveJoint * primMisJACmdMoveJointPtr;
  PrimMisJACmdMoveWorld * primMisJACmdMoveWorldPtr;
  PrimMisJACfgCycleTime * primMisJACfgCycleTimePtr;
  PrimMisJACfgDebug * primMisJACfgDebugPtr;
  PrimMisJACmdNop primMisJACmdNop;
  PrimMisJACfgNop primMisJACfgNop;
  PrimMisJAStat primMisJAStat;
  PrimMisJASet primMisJASet;

  bufnum = 1;
  retval = 0;
  opterr = 0;
  while (1) {
    int i1;
    char option = getopt(argc, argv, ":n:i:d:");
    if (option == -1)
      break;
    switch (option) {
    case 'n':
      strncpy(nmlFile, optarg, sizeof(nmlFile));
      nmlFile[sizeof(nmlFile)-1] = 0;
      break;
    case 'i':
      strncpy(iniFile, optarg, sizeof(iniFile));
      iniFile[sizeof(iniFile)-1] = 0;
      break;
    case 'd':
      if (1 == sscanf(optarg, "%i", &i1)) {
	if (i1 < 1) bufnum = 1;
	else if (i1 > 2) bufnum = 2;
	else bufnum = i1;
      } else {
	fprintf(stderr, "%s: bad buffer number\n", BN);
	retval = 1;
      }
      break;
    case ':':
      fprintf(stderr, "%s: missing parameter to %c\n", BN, (char) optopt);
      retval = 1;
      break;
    default:			// '?'
      fprintf(stderr, "%s: unrecognized option %c\n", BN, (char) optopt);
      retval = 1;
      break;
    }
  }
  if (retval) return retval;

  rcs_print("PrimMisMain%d started\n", bufnum);

  primMisJASet.cycleTime = DEFAULT_CYCLE_TIME;
  primMisJASet.debug = 0;

  if (0 != iniLoad(BN, iniFile, &primMisJASet.cycleTime, &primMisJASet.debug)) {
    return 1;
  }

  if (ULAPI_OK != ulapi_init(UL_USE_UNIX)) {
    return 1;
  } 

  // get our subordinate gomotion comm buffers
  retval = get_traj_comm_buffers();
  if (0 != retval) {
    fprintf(stderr, "%s: can't connect to traj\n", BN);
    return 1;
  }

  // get our NML buffers

  sprintf(chanName, "primMisJACmd%d", bufnum);
  primMisJACmdBuf = new RCS_CMD_CHANNEL(primMisJA_format, chanName, BN, nmlFile);
  if (! primMisJACmdBuf->valid()) {
    fprintf(stderr, "%s: can't open %s buffer\n", BN, chanName);
    return 1;
  }
  primMisJACmdMoveJointPtr = (PrimMisJACmdMoveJoint *) (primMisJACmdBuf->get_address());
  primMisJACmdMoveWorldPtr = (PrimMisJACmdMoveWorld *) (primMisJACmdBuf->get_address());

  sprintf(chanName, "primMisJAStat%d", bufnum);
  primMisJAStatBuf = new RCS_STAT_CHANNEL(primMisJA_format, chanName, BN, nmlFile);
  if (! primMisJAStatBuf->valid()) {
    fprintf(stderr, "%s: can't open %s buffer\n", BN, chanName);
    return 1;
  }

  sprintf(chanName, "primMisJACfg%d", bufnum);
  primMisJACfgBuf = new RCS_CMD_CHANNEL(primMisJA_format, chanName, BN, nmlFile);
  if (! primMisJACfgBuf->valid()) {
    fprintf(stderr, "%s: can't open %s buffer\n", BN, chanName);
    return 1;
  }
  primMisJACfgCycleTimePtr = (PrimMisJACfgCycleTime *) (primMisJACfgBuf->get_address());
  primMisJACfgDebugPtr = (PrimMisJACfgDebug *) (primMisJACfgBuf->get_address());

  sprintf(chanName, "primMisJASet%d", bufnum);
  primMisJASetBuf = new RCS_STAT_CHANNEL(primMisJA_format, chanName, BN, nmlFile);
  if (! primMisJASetBuf->valid()) {
    fprintf(stderr, "%s: can't open %s buffer\n", BN, chanName);
    return 1;
  }

  primMisJACmdBuf->write(&primMisJACmdNop);
  primMisJACfgBuf->write(&primMisJACfgNop);

  reset_time_tracker(&primMisJAStat.tt);

  done = 0;
  signal(SIGINT, quit);

  while (! done) {
    NMLTYPE cmdType, cfgType;
    int cmdSerialNumber, cfgSerialNumber;

    // read command
    cmdType = primMisJACmdBuf->read();
    cmdSerialNumber = primMisJACmdBuf->get_address()->serial_number;

    switch (cmdType) {
    case 0:
      // no new command
      break;

    case -1:
      // comm error
      fprintf(stderr, "%s: NML read error\n", BN);
      break;

    case PRIM_MIS_JA_CMD_NOP_TYPE:
    case PRIM_MIS_JA_CMD_INIT_TYPE:
    case PRIM_MIS_JA_CMD_HALT_TYPE:
    case PRIM_MIS_JA_CMD_ABORT_TYPE:
    case PRIM_MIS_JA_CMD_SHUTDOWN_TYPE:
    case PRIM_MIS_JA_CMD_MOVE_JOINT_TYPE:
    case PRIM_MIS_JA_CMD_MOVE_WORLD_TYPE:
    case PRIM_MIS_JA_CMD_STOP_TYPE:
      primMisJAStat.command_type = cmdType;
      if (cmdSerialNumber != primMisJAStat.echo_serial_number) {
	primMisJAStat.echo_serial_number = cmdSerialNumber;
	primMisJAStat.state = NEW_COMMAND;
      }
      break;

    default:
      fprintf(stderr, "%s: unknown command %s\n", 
	      BN, primMisJA_symbol_lookup(cmdType));
      break;
    } // switch (cmdType)

    // read config
    cfgType = primMisJACfgBuf->read();
    cfgSerialNumber = primMisJACfgBuf->get_address()->serial_number;

    switch (cfgType) {
    case 0:
      // no new command
      break;

    case -1:
      // comm error
      fprintf(stderr, "%s: NML read error\n", BN);
      break;

    case PRIM_MIS_JA_CFG_NOP_TYPE:
    case PRIM_MIS_JA_CFG_CYCLE_TIME_TYPE:
    case PRIM_MIS_JA_CFG_DEBUG_TYPE:
      primMisJASet.command_type = cfgType;
      if (cfgSerialNumber != primMisJASet.echo_serial_number) {
	primMisJASet.echo_serial_number = cfgSerialNumber;
	primMisJASet.state = NEW_COMMAND;
      }
      break;

    default:
      fprintf(stderr, "%s: unknown config %s\n", 
	      BN, primMisJA_symbol_lookup(cfgType));
      break;
    } // switch (cfgType)

    // read subordinate status and settings
    (void) update_traj_comm_buffers();

    // handle command state tables
    switch (primMisJAStat.command_type) {
    case PRIM_MIS_JA_CMD_NOP_TYPE:
      doCmdNop(&primMisJAStat);
      break;

    case PRIM_MIS_JA_CMD_INIT_TYPE:
      doCmdInit(&primMisJAStat);
      break;

    case PRIM_MIS_JA_CMD_ABORT_TYPE:
      doCmdAbort(&primMisJAStat);
      break;

    case PRIM_MIS_JA_CMD_HALT_TYPE:
      doCmdHalt(&primMisJAStat);
      break;

    case PRIM_MIS_JA_CMD_SHUTDOWN_TYPE:
      doCmdShutdown(&primMisJAStat);
      break;

    case PRIM_MIS_JA_CMD_MOVE_JOINT_TYPE:
      doCmdMoveJoint(primMisJACmdMoveJointPtr, &primMisJAStat, &primMisJASet, traj_cmd_ptr, traj_stat_ptr, traj_set_ptr);
      break;

    case PRIM_MIS_JA_CMD_MOVE_WORLD_TYPE:
      doCmdMoveWorld(primMisJACmdMoveWorldPtr, &primMisJAStat, &primMisJASet, traj_cmd_ptr, traj_stat_ptr, traj_set_ptr);
      break;

    case PRIM_MIS_JA_CMD_STOP_TYPE:
      doCmdStop(&primMisJAStat, traj_cmd_ptr, traj_stat_ptr);
      break;

    default:
      // no command to handle
      break;
    } // switch (primMisJAStat.command_type)

    // handle config state tables
    switch (primMisJASet.command_type) {
    case PRIM_MIS_JA_CFG_NOP_TYPE:
      doCfgNop(&primMisJASet);
      break;

    case PRIM_MIS_JA_CFG_CYCLE_TIME_TYPE:
      doCfgCycleTime(primMisJACfgCycleTimePtr, &primMisJASet);
      break;

    case PRIM_MIS_JA_CFG_DEBUG_TYPE:
      doCfgDebug(primMisJACfgDebugPtr, &primMisJASet);
      break;

    default:
      // no config to handle
      break;
    } // switch (primMisJASet.command_type)

    // update status and settings
    primMisJAStat.pose = go_pose_to_pm(traj_stat_ptr->ecp_act);
    cycle_time_tracker(&primMisJAStat.tt);
    primMisJAStatBuf->write(&primMisJAStat);
    primMisJASetBuf->write(&primMisJASet);

    esleep(primMisJASet.cycleTime);
  } // while (! done)

  (void) free_traj_comm_buffers();

  delete primMisJACmdBuf;
  delete primMisJAStatBuf;
  delete primMisJACfgBuf;
  delete primMisJASetBuf;

  rcs_print("PrimMisMain%d done\n", bufnum);

  return 0;
}

#endif	// HAVE_GOMOTION

/*
  Modification history:

  $Log: primMisMain.cc,v $
  Revision 1.6  2008/12/02 16:27:47  proctor
  Added ulapi to servoShell and the gocfg, wrap stuff for go motion at the
  mission servo level

  Revision 1.5  2008/09/29 15:42:59  proctor
  Changed the name of prim mission controller from primMisMain to primMis, to
  match the .nml file entry

  Revision 1.4  2008/09/25 20:49:35  proctor
  Completely rewrote primMisMain,the prim mission package, to interface to
  a Go Motion controller for doing all the real work. Autotools was changed
  to look for Go Motion. The prim mission interface was changed to eliminate
  the pan and tilt commands in favor of joint and world moves.

  Revision 1.3  2006/07/25 13:45:17  dr_steveb
  Fixed to work with new USARSim standards.

  Revision 1.2  2006/04/13 14:30:09  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.5  2005/09/30 18:55:46  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.4  2005/07/15 15:01:39  proctor
  Added handling of cycle time, debug settings

  Revision 1.3  2005/06/08 19:29:08  proctor
  Moved state_match, etc. macros into moastTypes.hh

  Revision 1.2  2005/06/08 14:30:55  proctor
  Added VehMobJA name changes

  Revision 1.1  2005/06/07 17:59:35  proctor
  Initial checkin
*/

