/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primMobMain.cc

  \brief Vehicle primitive level driver

  \code CVS Status:
  $Author: ghenson $
  $Revision: 1.40 $
  $Date: 2008/10/13 18:26:39 $
  \endcode

  \author Fred Proctor
*/
#include "primMobEngine.hh"
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "moastError.hh"	// moastError_format()
#include "navDataExt.hh"	// navDataExt_format()
#include "servoMobJA.hh"	// servoMobJA_format()
#include "servoSP.hh"		// servoSP_format()
#include "primMobJA.hh"

/*
  This macro is used in the state table functions below, and presumes that
  each function calls its status "stat". The macro records the type of the
  command in the status' "command_type_sent" field, increments the status'
  "serial_number_sent" field, and sets the command's serial number to this.
*/
#define SET_AND_RECORD(msg) stat->command_type_sent = (msg).type; (msg).serial_number = ++stat->serial_number_sent;

/*
  This macro is used in the stat table functions below, and presumes
  that each function calls its status "stat" and its subordinate's status
  "sub_stat_ptr". It returns non-zero if the subordinate's command and
  serial number match what was sent out.
*/
#define GOT_IT (sub_stat_ptr->command_type == stat->command_type_sent && sub_stat_ptr->echo_serial_number == stat->serial_number_sent)

static NML *moastErrorBuf = NULL;

static NML *navDataExtBuf = NULL;

static RCS_CMD_CHANNEL *servoMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *servoMobJAStatBuf = NULL;
// no need to connect to servo mob cfg buf
static RCS_STAT_CHANNEL *servoMobJASetBuf = NULL;

static RCS_CMD_CHANNEL *primMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *primMobJAStatBuf = NULL;
static RCS_CMD_CHANNEL *primMobJACfgBuf = NULL;
static RCS_STAT_CHANNEL *primMobJASetBuf = NULL;

static PrimMobJAStat primMobJAStat;
static PrimMobJASet primMobJASet;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "primMobJA"


/////////////////////////////////////////////////////////////////////////////////////////
//// state tables
/////////////////////////////////////////////////////////////////////////////////////////
void doCmdNop(PrimMobJAStat * stat, PrimMobEngine *engine)
{
  //  ServoMobJACmdNop nopMsg;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if(engine->doCmdNop())
      status_next(stat, RCS_DONE);
    else
      status_next(stat, RCS_ERROR);
    state_next(stat, S0);
  } else {			// SO
    state_default(stat);
  }
}

void doCmdInit(PrimMobJAStat * stat,
	       RCS_CMD_CHANNEL * sub_cmd_buf,
	       RCS_STAT_CHANNEL * sub_stat_buf,
	       PrimMobEngine *engine)
{
  ServoMobJACmdInit initMsg;
  ServoMobJAStat *sub_stat_ptr;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    SET_AND_RECORD(initMsg);
    if( !engine->doCmdInit() ){
      status_next(stat, RCS_ERROR);
      state_next(stat, S1);
    }else{
      sub_cmd_buf->write(&initMsg);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
    }
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (GOT_IT && sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      stat->admin_state = ADMIN_INITIALIZED;
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void doCmdHalt(PrimMobJAStat * stat,
	       RCS_CMD_CHANNEL * sub_cmd_buf,
	       RCS_STAT_CHANNEL * sub_stat_buf,
	       PrimMobEngine *engine)
{
  ServoMobJACmdHalt haltMsg;
  ServoMobJAStat *sub_stat_ptr;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    SET_AND_RECORD(haltMsg);
    if( engine->doCmdHalt() ){
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
    }else{
      status_next(stat, RCS_ERROR);
      state_next(stat, S1);
    }
    sub_cmd_buf->write(&haltMsg);
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (GOT_IT && sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      stat->admin_state = ADMIN_UNINITIALIZED;
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void doCmdAbort(PrimMobJAStat * stat,
		RCS_CMD_CHANNEL * sub_cmd_buf,
		RCS_STAT_CHANNEL * sub_stat_buf,
		PrimMobEngine *engine)
{
  ServoMobJACmdAbort abortMsg;
  ServoMobJAStat *sub_stat_ptr;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    SET_AND_RECORD(abortMsg);
    if( engine->doCmdAbort() ){
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
    }else{
      status_next(stat, RCS_ERROR);
      state_next(stat, S1);
    }
    sub_cmd_buf->write(&abortMsg);
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (GOT_IT && sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      stat->admin_state = ADMIN_UNINITIALIZED;
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void doCmdShutdown(PrimMobJAStat * stat,
		   RCS_CMD_CHANNEL * sub_cmd_buf,
		   RCS_STAT_CHANNEL * sub_stat_buf,
		   PrimMobEngine *engine)
{
  ServoMobJACmdShutdown shutdownMsg;
  ServoMobJAStat *sub_stat_ptr;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    SET_AND_RECORD(shutdownMsg);
    if( engine->doCmdShutdown() ){
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
    }else{
      status_next(stat, RCS_ERROR);
      state_next(stat, S1);
    }
    sub_cmd_buf->write(&shutdownMsg);
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (GOT_IT && sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      stat->admin_state = ADMIN_SHUT_DOWN;
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void doCmdMoveWaypoint( PrimMobJAStat * stat,
			PrimMobJASet * set,
			RCS_CMD_CHANNEL * sub_cmd_buf,
			RCS_STAT_CHANNEL * sub_stat_buf,
			NML * nav_data_ext_buf,
			PrimMobEngine *engine)
{
  ServoMobJACmdAckerman ackerMsg;
  ServoMobJACmdSubmarine submarineMsg;
  ServoMobJACmdSkid skidMsg;
  ServoMobJAStat *sub_stat_ptr;
  NavDataExt *nav_data_ext_ptr;
  PM_CARTESIAN here;
  PM_CARTESIAN target;
  PM_CARTESIAN here_to_target;
  PM_CARTESIAN target_to_next;
  double trash;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());
  nav_data_ext_ptr = (NavDataExt *) (nav_data_ext_buf->get_address());

  nav_data_ext_buf->read();
  sub_stat_buf->read();

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state != ADMIN_INITIALIZED) {
      status_next(stat, RCS_ERROR);
      state_next(stat, S2);
    } else {
      stat->pathIndex = 0;
      stat->status = engine->doCmdMoveWaypoint(NEW_COMMAND, 
					       nav_data_ext_ptr->tranRelRate,
					       nav_data_ext_ptr->rpyRelRate,
					       nav_data_ext_ptr->tranRel,
					       nav_data_ext_ptr->rpyRel,
					       stat->pathIndex, 
					       stat->state, 
					       trash, trash, trash);
    }
  }
  // and fall through to here if S0...
  if (state_match(stat, S0)) {
    if (engine->isAckerman()) {
      stat->status = engine->doCmdMoveWaypoint(S0, 
					       nav_data_ext_ptr->tranRelRate,
					       nav_data_ext_ptr->rpyRelRate,
					       nav_data_ext_ptr->tranRel,
					       nav_data_ext_ptr->rpyRel,
					       stat->pathIndex, 
					       stat->state, 
					       ackerMsg.velocity,
					       ackerMsg.steerAngle,
					       trash);
      ackerMsg.crabAngle = -ackerMsg.steerAngle;
      SET_AND_RECORD(ackerMsg);
      sub_cmd_buf->write(&ackerMsg);
    } else if (engine->isNautic()) {
      stat->status = engine->doCmdMoveWaypoint(S0, 
					       nav_data_ext_ptr->tranRelRate,
					       nav_data_ext_ptr->rpyRelRate,
					       nav_data_ext_ptr->tranRel,
					       nav_data_ext_ptr->rpyRel,
					       stat->pathIndex, 
					       stat->state, 
					       submarineMsg.propeller,
					       submarineMsg.rudder,
					       submarineMsg.sternPlane);
      SET_AND_RECORD(submarineMsg);
      sub_cmd_buf->write(&submarineMsg);
    } else {
      stat->status = engine->doCmdMoveWaypoint(S0, 
					       nav_data_ext_ptr->tranRelRate,
					       nav_data_ext_ptr->rpyRelRate,
					       nav_data_ext_ptr->tranRel,
					       nav_data_ext_ptr->rpyRel,
					       stat->pathIndex, 
					       stat->state, 
					       skidMsg.wLeft,
					       skidMsg.wRight,
					       trash);
      SET_AND_RECORD(skidMsg);
      sub_cmd_buf->write(&skidMsg);
    }
  } else if (state_match(stat, S1)) {			// S1 = stopping
    // done
    if (engine->isAckerman()) {
      stat->status = engine->doCmdMoveWaypoint(S1, 
					       nav_data_ext_ptr->tranRelRate,
					       nav_data_ext_ptr->rpyRelRate,
					       nav_data_ext_ptr->tranRel,
					       nav_data_ext_ptr->rpyRel,
					       stat->pathIndex, 
					       stat->state, 
					       ackerMsg.velocity,
					       ackerMsg.steerAngle,
					       trash);
      ackerMsg.crabAngle = -ackerMsg.steerAngle;
      SET_AND_RECORD(ackerMsg);
      sub_cmd_buf->write(&ackerMsg);
    } else if (engine->isNautic()) {
      stat->status = engine->doCmdMoveWaypoint(S1, 
					       nav_data_ext_ptr->tranRelRate,
					       nav_data_ext_ptr->rpyRelRate,
					       nav_data_ext_ptr->tranRel,
					       nav_data_ext_ptr->rpyRel,
					       stat->pathIndex, 
					       stat->state, 
					       submarineMsg.propeller,
					       submarineMsg.rudder,
					       submarineMsg.sternPlane);
      SET_AND_RECORD(submarineMsg);
      sub_cmd_buf->write(&submarineMsg);
    } else {
      stat->status = engine->doCmdMoveWaypoint(S1, 
					       nav_data_ext_ptr->tranRelRate,
					       nav_data_ext_ptr->rpyRelRate,
					       nav_data_ext_ptr->tranRel,
					       nav_data_ext_ptr->rpyRel,
					       stat->pathIndex, 
					       stat->state, 
					       skidMsg.wLeft,
					       skidMsg.wRight,
					       trash);
      SET_AND_RECORD(skidMsg);
      sub_cmd_buf->write(&skidMsg);
    }
  } else {
    state_default(stat);
  }
  return;
}


// FIXME-- the rotate command isn't valid for Ackerman vehicles.
void doCmdRotate(PrimMobJACmdRotate * cmd,
		 PrimMobJAStat * stat,
		 PrimMobJASet * set,
		 RCS_CMD_CHANNEL * sub_cmd_buf,
		 RCS_STAT_CHANNEL * sub_stat_buf,
		 NML * nav_data_ext_buf,
		 PrimMobEngine *engine)
{
  ServoMobJACmdSkid skidMsg;
  ServoMobJACmdAckerman ackerMsg;
  ServoMobJAStat *sub_stat_ptr;
  NavDataExt *nav_data_ext_ptr;
  double trash;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());
  nav_data_ext_ptr = (NavDataExt *) (nav_data_ext_buf->get_address());

  nav_data_ext_buf->read();
  sub_stat_buf->read();

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state != ADMIN_INITIALIZED) {
      status_next(stat, RCS_ERROR);
      state_next(stat, S1);
    } else {
      stat->status = engine->doCmdRotate( NEW_COMMAND, cmd->theta, cmd->thetaTol, cmd->direction,
					  nav_data_ext_ptr->rpyRel, 
					  stat->pathIndex,
					  stat->state, trash, trash );
    }
  }
  // and fall through to here if S0...
  if (state_match(stat, S0)) {
    if (engine->isAckerman()) {
      stat->status = engine->doCmdRotate( S0, cmd->theta, cmd->thetaTol, cmd->direction,
					  nav_data_ext_ptr->rpyRel, 
					  stat->pathIndex,
					  stat->state, ackerMsg.velocity,
					  ackerMsg.steerAngle );
      ackerMsg.crabAngle = -ackerMsg.steerAngle;
      SET_AND_RECORD(ackerMsg);
      sub_cmd_buf->write(&ackerMsg);
    } else {
      stat->status = engine->doCmdRotate( S0, cmd->theta, cmd->thetaTol, cmd->direction,
					  nav_data_ext_ptr->rpyRel, 
					  stat->pathIndex,
					  stat->state, skidMsg.wLeft, skidMsg.wRight );
      SET_AND_RECORD(skidMsg);
      sub_cmd_buf->write(&skidMsg);
    }
  }else if (state_match(stat, S1)) { // S1 = stopping
    if (engine->isAckerman()) {
      stat->status = engine->doCmdRotate( S0, cmd->theta, cmd->thetaTol, cmd->direction,
					  nav_data_ext_ptr->rpyRel, 
					  stat->pathIndex,
					  stat->state, ackerMsg.velocity,
					  ackerMsg.steerAngle );
      ackerMsg.crabAngle = -ackerMsg.steerAngle;
      SET_AND_RECORD(ackerMsg);
      sub_cmd_buf->write(&ackerMsg);
    } else {
      stat->status = engine->doCmdRotate( S1, cmd->theta, cmd->thetaTol, cmd->direction,
					nav_data_ext_ptr->rpyRel, 
					stat->pathIndex,
					stat->state, skidMsg.wLeft, skidMsg.wRight );
      SET_AND_RECORD(skidMsg);
      sub_cmd_buf->write(&skidMsg);
    }
    status_next(stat, RCS_DONE);
    state_next(stat, S2);
  } else {			// S2 = done
    // done
    state_default(stat);
  }
  return;
}

void doCmdVel(PrimMobJACmdVel * cmd,
	      PrimMobJAStat * stat,
	      RCS_CMD_CHANNEL * sub_cmd_buf,
	      RCS_STAT_CHANNEL * sub_stat_buf,
	      PrimMobEngine *engine)
{
  ServoMobJACmdAckerman ackerMsg;
  ServoMobJACmdSubmarine submarineMsg;
  ServoMobJACmdSkid skidMsg;
  ServoMobJAStat *sub_stat_ptr;
  double trash;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());

  sub_stat_buf->read();

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state != ADMIN_INITIALIZED) {
      status_next(stat, RCS_ERROR);
      state_next(stat, S1);
    } else {
      stat->status = engine->doCmdVel(NEW_COMMAND, cmd->v, cmd->w,
				      stat->state,
				      trash, trash, trash );
    }
  }
  // drop through
  if (state_match(stat, S0)) {
    if (engine->isAckerman()) {
      stat->status = engine->doCmdVel(S0, cmd->v, cmd->w,
				      stat->state,
				      ackerMsg.velocity, ackerMsg.steerAngle, trash );
      ackerMsg.crabAngle = -ackerMsg.steerAngle;
      SET_AND_RECORD(ackerMsg);
      sub_cmd_buf->write(&ackerMsg);
    } else if (engine->isNautic()) {
      stat->status = engine->doCmdVel(S0, cmd->v, cmd->w,
				      stat->state,
				      submarineMsg.propeller,
				      submarineMsg.rudder,
				      submarineMsg.sternPlane);
      SET_AND_RECORD(submarineMsg);
      sub_cmd_buf->write(&submarineMsg);
    } else {
      stat->status = engine->doCmdVel(S0, cmd->v, cmd->w,
				      stat->state,
				      skidMsg.wLeft, skidMsg.wRight, trash );
      SET_AND_RECORD(skidMsg);
      sub_cmd_buf->write(&skidMsg);
    }
  } else {			// S1
    state_default(stat);
  }

  return;
}

void doCmdStop(PrimMobJAStat * stat,
	       RCS_CMD_CHANNEL * sub_cmd_buf,
	       RCS_STAT_CHANNEL * sub_stat_buf, PrimMobEngine *engine)
{
  ServoMobJACmdAckerman ackerMsg;
  ServoMobJACmdSubmarine submarineMsg;
  ServoMobJACmdSkid skidMsg;
  ServoMobJAStat *sub_stat_ptr;
  double trash;

  sub_stat_ptr = (ServoMobJAStat *) (sub_stat_buf->get_address());

  sub_stat_buf->read();

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if (stat->admin_state != ADMIN_INITIALIZED) {
      status_next(stat, RCS_ERROR);
      state_next(stat, S0);
    } else {
      stat->status = engine->doCmdStop(NEW_COMMAND, stat->state, trash, trash, trash );
    }
  }
  // drop through

  if (state_match(stat, S1)) {
    if (engine->isAckerman()) {
      stat->status = engine->doCmdStop(S1, stat->state, ackerMsg.velocity,
				       ackerMsg.steerAngle, ackerMsg.crabAngle );
      SET_AND_RECORD(ackerMsg);
      sub_cmd_buf->write(&ackerMsg);
    } else if (engine->isNautic()) {
      stat->status = engine->doCmdStop(S1, stat->state, 
				       submarineMsg.propeller, submarineMsg.rudder, submarineMsg.sternPlane );
      SET_AND_RECORD(submarineMsg);
      sub_cmd_buf->write(&submarineMsg);
    } else {
            stat->status = engine->doCmdStop(S1, stat->state, 
					     skidMsg.wLeft, skidMsg.wRight, trash );
      SET_AND_RECORD(skidMsg);
      sub_cmd_buf->write(&skidMsg);
    }
  } else if (state_match(stat, S2)) {
    // wait for command received and executing, then we're done
    sub_stat_buf->read();
    if (GOT_IT) {
      // servo got it
      if (sub_stat_ptr->status == RCS_ERROR) {
	status_next(stat, RCS_ERROR);
      } else {
	// done or exec, we're done
	stat->status = engine->doCmdStop(S2, stat->state, trash, trash, trash );
      }
    }
    // else wait here
  } else {			// S0
    state_default(stat);
  }
  return;
}

void doCmdDock(PrimMobJAStat * stat,
	       RCS_CMD_CHANNEL * sub_cmd_buf,
	       RCS_STAT_CHANNEL * sub_stat_buf, PrimMobEngine *engine)
{
  // your docking code here
}

void doCfgNop(PrimMobJASet * set, PrimMobEngine *engine)
{
  if (state_match(set, NEW_COMMAND)) {
    set->status = engine->doCfgNop( NEW_COMMAND, set->state );
  } else {
    state_default(set);		// SO
  }
}

void doCfgCycleTime(PrimMobJACfgCycleTime * cfg,
		    PrimMobJASet * set,
		    NML * errbuf, PrimMobEngine *engine)
{
  MoastError errMsg;

  if (state_match(set, NEW_COMMAND)) {
    set->status = engine->doCfgCycleTime( NEW_COMMAND, set->state, cfg->cycleTime, set->cycleTime );
    if( set->status == RCS_ERROR ){
      errMsg.error = PRIM_MOB_JA_ERROR_BAD_CFG;
      strncpy(errMsg.message, "invalid cycle time", MOAST_ERROR_LEN);
      errMsg.message[MOAST_ERROR_LEN - 1] = 0;
      errbuf->write(&errMsg);
    }
  } else {
    state_default(set);
  }
}

void doCfgMove(PrimMobJACfgMove * cfg,
	       PrimMobJASet * set, PrimMobEngine *engine)
{
  if (state_match(set, NEW_COMMAND)) {
    set->status = engine->doCfgMove( NEW_COMMAND, set->state, 
				     cfg->vmax, cfg->amax, cfg->wmax, cfg->alphamax, cfg->vCutoffAngle, cfg->wCutoffAngle,
				     set->vmax, set->amax, set->wmax, set->alphamax, set->vCutoffAngle, set->wCutoffAngle);
  } else {
    state_default(set);
  }
}

void doCfgDebug(PrimMobJACfgDebug * cfg,
		PrimMobJASet * set, PrimMobEngine *engine)
{
  if (state_match(set, NEW_COMMAND)) {
    engine->doCfgDebug(NEW_COMMAND, set->state, cfg->debug, set->debug );
  } else {
    state_default(set);
  }
}

/////////////////////////////////////////////////////////////////////////////////////////
//// end of state tables
/////////////////////////////////////////////////////////////////////////////////////////
static int cleanupServoMobJANmlBuffers(void)
{
  if (NULL != moastErrorBuf)
    delete moastErrorBuf;
  if (NULL != navDataExtBuf)
    delete navDataExtBuf;
  if (NULL != servoMobJACmdBuf)
    delete servoMobJACmdBuf;
  if (NULL != servoMobJAStatBuf)
    delete servoMobJAStatBuf;
  if (NULL != servoMobJASetBuf)
    delete servoMobJASetBuf;

  moastErrorBuf = NULL;
  navDataExtBuf = NULL;
  servoMobJAStatBuf = NULL;
  servoMobJASetBuf = NULL;
  servoMobJACmdBuf = NULL;
  return 0;
}

static int initServoMobJANmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupServoMobJANmlBuffers();

  moastErrorBuf =
    new NML(moastError_format, "moastError", THIS_PROCESS,
	    config_file);
  if (!moastErrorBuf->valid()) {
    delete moastErrorBuf;
    moastErrorBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
  navDataExtBuf =
    new NML(navDataExt_format, chanName, THIS_PROCESS,
	    config_file);
  if (!navDataExtBuf->valid()) {
    delete navDataExtBuf;
    navDataExtBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, bufnum);
  servoMobJACmdBuf =
    new RCS_CMD_CHANNEL(servoMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMobJACmdBuf->valid()) {
    delete servoMobJACmdBuf;
    servoMobJACmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, bufnum);
  servoMobJAStatBuf =
    new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMobJAStatBuf->valid()) {
    delete servoMobJAStatBuf;
    servoMobJAStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_SET_NAME, bufnum);
  servoMobJASetBuf =
    new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMobJASetBuf->valid()) {
    delete servoMobJASetBuf;
    servoMobJASetBuf = 0;
    return 1;
  }

  return 0;
}

static int cleanupPrimMobJANmlBuffers(void)
{
  if (NULL != primMobJACmdBuf)
    delete primMobJACmdBuf;
  if (NULL != primMobJAStatBuf)
    delete primMobJAStatBuf;
  if (NULL != primMobJACfgBuf)
    delete primMobJACfgBuf;
  if (NULL != primMobJASetBuf)
    delete primMobJASetBuf;

  primMobJACmdBuf = NULL;
  primMobJAStatBuf = NULL;
  primMobJACfgBuf = NULL;
  primMobJASetBuf = NULL;

  return 0;
}

static int initPrimMobJANmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupPrimMobJANmlBuffers();

  sprintf(chanName, "%s%d", PRIM_MOB_JA_CMD_NAME, bufnum);
  primMobJACmdBuf =
    new RCS_CMD_CHANNEL(primMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!primMobJACmdBuf->valid()) {
    delete primMobJACmdBuf;
    primMobJACmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MOB_JA_STAT_NAME, bufnum);
  primMobJAStatBuf =
    new RCS_STAT_CHANNEL(primMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primMobJAStatBuf->valid()) {
    delete primMobJAStatBuf;
    primMobJAStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MOB_JA_CFG_NAME, bufnum);
  primMobJACfgBuf =
    new RCS_CMD_CHANNEL(primMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!primMobJACfgBuf->valid()) {
    delete primMobJACfgBuf;
    primMobJACfgBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MOB_JA_SET_NAME, bufnum);
  primMobJASetBuf =
    new RCS_STAT_CHANNEL(primMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primMobJASetBuf->valid()) {
    delete primMobJASetBuf;
    primMobJASetBuf = 0;
    return 1;
  }

  return 0;
}

/*
  iniLoad() will load INI file params from 'filename', returning 0
  if everything is fine, 1 if the file can't be opened or there are
  errors in the values.

  Note: angular units in the INI file are DEGREES, since it's human-readable.
  Internal units are always SI, radians. 

  Values loaded:

  [PRIM_MOB_JA]
  CYCLE_TIME = <positive number, [s]>
  DEBUG = <debug mask>

  [PLATFORM.<plaform name>], else [PLATFORM]
  MAX_TRAN_VEL = <positive number, [m/s]>
  MAX_TRAN_ACC = <positive number, [m/s/s]>
  MAX_ROT_VEL = <positive number, [deg/s]>
  MAX_ROT_ACC = <positive number, [deg/s/s]>
  V_CUTOFF_ANGLE = <positive number, [deg]>
  W_CUTOFF_ANGLE = <positive number, [deg]>
*/

#if ! defined(INI_OK)
#define INI_OK      0
#define INI_DEFAULT 1
#define INI_INVALID 2
#endif

static int iniLoad(const char * prog,
		   const char * filename,
		   PrimMobJASet * set, 
		   ServoMobJASet * servoMobJASetPtr)
{
  INIFILE inifile;
  const char * inistring;
  double d, d2, d3, d4, d5, d6;
  int i;
  PM_RPY rpy;
  int retval = INI_OK;
  char section[INIFILE_MAX_LINELEN];
  char key[INIFILE_MAX_LINELEN];

  if (0 != inifile.open(filename)) {
    rcs_print_error("%s: can't open INI file %s, using defaults for PrimMobJASet\n", prog, filename);
    set->cycleTime = DEFAULT_CYCLE_TIME;
    set->vmax = DEFAULT_MAX_TRAN_VEL;
    set->amax = DEFAULT_MAX_TRAN_ACC;
    set->wmax = DEFAULT_MAX_ROT_VEL;
    set->alphamax = DEFAULT_MAX_ROT_ACC;
    set->vCutoffAngle = DEFAULT_V_CUTOFF_ANGLE;
    set->wCutoffAngle = DEFAULT_W_CUTOFF_ANGLE;
    set->controlPoint = PM_POSE(PM_CARTESIAN(0,0,0),PM_QUATERNION(1,0,0,0));
    return INI_INVALID;
  }

  strcpy(section, "PRIM_MOB_JA");

  sprintf( key, "%s", "CYCLE_TIME" );
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->cycleTime = d;
    }
    else {
      // found, but invalid
      set->cycleTime = DEFAULT_CYCLE_TIME;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key, filename);
    set->cycleTime = DEFAULT_CYCLE_TIME;
    rcs_print_error("%s: using %f\n", prog, set->cycleTime);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  sprintf(key, "%s", "DEBUG" );
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%i", &i)) {
      // found, and valid
      set->debug = i;
    }
    else {
      // found, but invalid
      set->debug = DEFAULT_DEBUG;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key, filename);
    set->debug = DEFAULT_DEBUG;
    rcs_print_error("%s: using %d\n", prog, set->debug);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  /*
    Generate a section name like [PLATFORM.UnitLoader] if the platform
    name is not empty, otherwise generate [PLATFORM] as the default
  */
  if (0 == servoMobJASetPtr->platformName[0]) {
    strcpy(section, "PLATFORM");
  } else {
    sprintf(section, "PLATFORM.%s", servoMobJASetPtr->platformName);
  }

  sprintf(key, "%s", "MAX_TRAN_VEL");
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->vmax = d;
    }
    else {
      // found, but invalid
      set->vmax = DEFAULT_MAX_TRAN_VEL;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key, filename);
    set->vmax = DEFAULT_MAX_TRAN_VEL;
    rcs_print_error("%s: using %f\n", prog, set->vmax);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }
  
  sprintf(key, "%s", "MAX_TRAN_ACC" );
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->amax = d;
    }
    else {
      // found, but invalid
      set->amax = DEFAULT_MAX_TRAN_ACC;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key, filename);
    set->amax = DEFAULT_MAX_TRAN_ACC;
    rcs_print_error("%s: using %f\n", prog, set->amax);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  sprintf(key, "%s", "MAX_ROT_VEL" );
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->wmax = RAD(d);
    }
    else {
      // found, but invalid
      set->wmax = DEFAULT_MAX_ROT_VEL;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [PRIM_MOB_JA] MAX_ROT_VEL in '%s'\n", prog, filename);
    set->wmax = DEFAULT_MAX_ROT_VEL;
    rcs_print_error("%s: using %f\n", prog, set->wmax);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  sprintf(key, "%s", "MAX_ROT_ACC");
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->alphamax = RAD(d);
    }
    else {
      // found, but invalid
      set->alphamax = DEFAULT_MAX_ROT_ACC;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key, filename);
    set->alphamax = DEFAULT_MAX_ROT_ACC;
    rcs_print_error("%s: using %f\n", prog, set->alphamax);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  sprintf(key, "%s", "V_CUTOFF_ANGLE" );
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->vCutoffAngle = RAD(d);
    }
    else {
      // found, but invalid
      set->vCutoffAngle = DEFAULT_V_CUTOFF_ANGLE;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key, filename);
    set->vCutoffAngle = DEFAULT_V_CUTOFF_ANGLE;
    rcs_print_error("%s: using %f\n", prog, set->vCutoffAngle);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  sprintf(key, "%s", "W_CUTOFF_ANGLE" );
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->wCutoffAngle = RAD(d);
    }
    else {
      // found, but invalid
      set->wCutoffAngle = DEFAULT_W_CUTOFF_ANGLE;
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key, filename);
    set->wCutoffAngle = DEFAULT_W_CUTOFF_ANGLE;
    rcs_print_error("%s: using %f\n", prog, set->wCutoffAngle);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  sprintf(key, "%s", "CONTROL_POINT" );
  if (NULL != (inistring = inifile.find(key, section))) {
    if (6 == sscanf(inistring, "%lf %lf %lf %lf %lf %lf", 
		    &d, &d2, &d3, &d4, &d5, &d6)) {
      // found, and valid
      set->controlPoint.tran.x = d;
      set->controlPoint.tran.y = d2;
      set->controlPoint.tran.z = d3;
      rpy.r = RAD(d4);
      rpy.p = RAD(d5);
      rpy.y = RAD(d6);
      set->controlPoint.rot = rpy;
    }
    else {
      // found, but invalid
      set->controlPoint = PM_POSE(PM_CARTESIAN(0,0,0),PM_QUATERNION(1,0,0,0));
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog, inistring, section, key, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all-- ignore, leaving control pose as identity pose
    set->controlPoint = PM_POSE(PM_CARTESIAN(0,0,0),PM_QUATERNION(1,0,0,0));
  }

  inifile.close();

  return retval;
}

static int done = 0;

void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
  int retval;
  NMLTYPE cmd_type, cfg_type, set_type;
  int serial_number;
  PrimMobJACmdNop nopCmd;
  PrimMobJACfgNop nopCfg;
  ServoMobJASet * servoMobJASetPtr;
  PrimMobEngine *engine = new PrimMobEngine();
  int error;

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      break;
    case 'i':
      INI_FILE = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // connect to the vehicle servo level buffers
  retval = initServoMobJANmlBuffers(NML_FILE, bufnum);
  if (retval) {
    rcs_print_error("%s: can't allocate servoMobJA NML buffers\n", argv[0]);
    return 1;
  }
  servoMobJASetPtr = (ServoMobJASet *) servoMobJASetBuf->get_address();

  // read in servo level vehicle settings
  int readOK = 0;
  int readBadOnce = 0;
  double readTimer = 10.0;	// seconds that we should keep trying
  do {
    set_type = servoMobJASetBuf->read();
    switch (set_type) {
      // ATTN Might need fixing
    case SERVO_MOB_JA_SET_TYPE:
      servoMobJASetPtr = (ServoMobJASet *) servoMobJASetBuf->get_address();
      // do something if needed
      break;
    case SERVO_MOB_JA_SET_GRD_VEH_TYPE:
      servoMobJASetPtr = (ServoMobJASet *) servoMobJASetBuf->get_address();
      // check for valid settings 
      engine->setWheelSettings(((ServoMobJASetGrdVeh *)servoMobJASetPtr)->wheelRadius,
			       ((ServoMobJASetGrdVeh *)servoMobJASetPtr)->wheelSeparation);

      // set up steering type
      switch (servoMobJASetPtr->steerType) {
      case MOAST_ACKERMAN_STEER_TYPE:
	engine->setAckermanWheelbase(((ServoMobJASetGrdVeh *)servoMobJASetPtr)->wheelBase);
      break;
      case MOAST_SUBMARINE_STEER_TYPE: // not sure what length to use!
	engine->setNauticLength(1);
	break;
      case MOAST_SKID_STEER_TYPE:
	break;
      }
      readOK = 1;
      break;

    default:
      // can't read out settings
      if (! readBadOnce) {
	rcs_print("%s: waiting for servo mob settings...\n", argv[0]);
	readBadOnce = 1;
      }
      readTimer -= primMobJASet.cycleTime;
      esleep(primMobJASet.cycleTime);
    }
  } while (! readOK && readTimer > 0.0);
  if (readOK) {
    if (readBadOnce) {
      rcs_print("%s: got it\n", argv[0]);
    }
  } else {
    rcs_print_error("%s: can't read servo mob settings, giving up\n", argv[0]);
    (void) cleanupServoMobJANmlBuffers();
    (void) cleanupPrimMobJANmlBuffers();
    return 1;
  }

  // connect to my buffers
  retval = initPrimMobJANmlBuffers(NML_FILE, bufnum);
  if (retval) {
    rcs_print_error("%s: can't allocate primMobJA NML buffers\n", argv[0]);
    (void) cleanupServoMobJANmlBuffers();
    return 1;
  }

  // initialize and set the status up so that init command goes through
  primMobJAStat.echo_serial_number = 0;
  reset_time_tracker(&primMobJAStat.tt);
  primMobJAStat.command_type_sent = 0;
  primMobJAStat.serial_number_sent = 0;
  primMobJAStat.pathIndex = 0;

  // initialize and set the settings up so that the length goes through
  primMobJASet.echo_serial_number = 0;
  // others will come from .ini file shortly...

  // read in the INI file params
  if (INI_INVALID == iniLoad(argv[0], INI_FILE, &primMobJASet, servoMobJASetPtr)) {
    rcs_print_error("%s: error loading INI file '%s'\n", argv[0], INI_FILE);
    return 1;
  }
  printf("%s: loaded parameters for %s\n", argv[0], servoMobJASetPtr->platformName);

  engine->setSettings(primMobJASet.cycleTime,
		      primMobJASet.vmax,
		      primMobJASet.amax,
		      primMobJASet.wmax,
		      primMobJASet.alphamax,
		      primMobJASet.vCutoffAngle,
		      primMobJASet.wCutoffAngle,
		      primMobJASet.controlPoint,
		      primMobJASet.debug);

  // stuff a NOP command
  nopCmd.serial_number = 1;
  primMobJACmdBuf->write(&nopCmd);

  // stuff a NOP config
  nopCfg.serial_number = 1;
  primMobJACfgBuf->write(&nopCfg);

  done = 0;
  signal(SIGINT, quit);
  while (!done) {
    // some shortcut pointers to various commands
    PrimMobJACmdMoveArcSegment * cmdMoveArcSegmentPtr;
    PrimMobJACmdMoveWaypoint * cmdMoveWaypointPtr;
    PrimMobJACmdRotate * cmdRotatePtr;
    PrimMobJACmdVel * cmdVelPtr;
    PrimMobJACfgCycleTime * cfgCycleTimePtr;
    PrimMobJACfgMove * cfgMovePtr;
    PrimMobJACfgDebug * cfgDebugPtr;

    // read command
    cmd_type = primMobJACmdBuf->read();
    serial_number = primMobJACmdBuf->get_address()->serial_number;
    switch (cmd_type) {

    case 0:
      // no new command
      break;

    case -1:
      // comm error
      rcs_print_error("%s: NML command read error\n", argv[0]);
      break;

    case PRIM_MOB_JA_CMD_NOP_TYPE:
    case PRIM_MOB_JA_CMD_INIT_TYPE:
    case PRIM_MOB_JA_CMD_HALT_TYPE:
    case PRIM_MOB_JA_CMD_ABORT_TYPE:
    case PRIM_MOB_JA_CMD_SHUTDOWN_TYPE:
    case PRIM_MOB_JA_CMD_STOP_TYPE:
    case PRIM_MOB_JA_CMD_DOCK_TYPE:
      primMobJAStat.command_type = cmd_type;
      if (serial_number != primMobJAStat.echo_serial_number) {
	primMobJAStat.echo_serial_number = serial_number;
	primMobJAStat.state = NEW_COMMAND;
	primMobJAStat.status = RCS_EXEC;
      }
      break;

    case PRIM_MOB_JA_CMD_MOVE_ARC_SEGMENT_TYPE:
      primMobJAStat.command_type = cmd_type;
      if (serial_number != primMobJAStat.echo_serial_number) {
	primMobJAStat.echo_serial_number = serial_number;
	primMobJAStat.state = NEW_COMMAND;
	primMobJAStat.status = RCS_EXEC;
      }
      cmdMoveArcSegmentPtr = (PrimMobJACmdMoveArcSegment *) primMobJACmdBuf->get_address();
      // rebuild the waypoint array with this fresh data
      error = engine->arc_segments_to_waypoints(cmdMoveArcSegmentPtr->arcList,
						cmdMoveArcSegmentPtr->arcList_length);
      if (error < 0) {
	rcs_print_error("%s: can't generate waypoint array\n", argv[0]);
      }
      engine->resetWaypointIndex();
      break;

    case PRIM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
      primMobJAStat.command_type = cmd_type;
      if (serial_number != primMobJAStat.echo_serial_number) {
	primMobJAStat.echo_serial_number = serial_number;
	primMobJAStat.state = NEW_COMMAND;
	primMobJAStat.status = RCS_EXEC;
      }
      cmdMoveWaypointPtr = (PrimMobJACmdMoveWaypoint *) primMobJACmdBuf->get_address();
      engine->setPrimPath(cmdMoveWaypointPtr->prim_path);
      engine->setWaypoints( cmdMoveWaypointPtr->waypoint, cmdMoveWaypointPtr->waypoint_length);
      break;

    case PRIM_MOB_JA_CMD_ROTATE_TYPE:
      primMobJAStat.command_type = cmd_type;
      if (serial_number != primMobJAStat.echo_serial_number) {
	primMobJAStat.echo_serial_number = serial_number;
	primMobJAStat.state = NEW_COMMAND;
	primMobJAStat.status = RCS_EXEC;
      }
      cmdRotatePtr = (PrimMobJACmdRotate *) primMobJACmdBuf->get_address();
      break;

    case PRIM_MOB_JA_CMD_VEL_TYPE:
      primMobJAStat.command_type = cmd_type;
      if (serial_number != primMobJAStat.echo_serial_number) {
	primMobJAStat.echo_serial_number = serial_number;
	primMobJAStat.state = NEW_COMMAND;
	primMobJAStat.status = RCS_EXEC;
      }
      cmdVelPtr = (PrimMobJACmdVel *) primMobJACmdBuf->get_address();
      break;

    default:
      rcs_print_error("%s: unknown command: %d\n", argv[0], (int) cmd_type);
      break;
    }				// switch (cmd_type)

    // read config
    cfg_type = primMobJACfgBuf->read();
    serial_number = primMobJACfgBuf->get_address()->serial_number;
    switch (cfg_type) {

    case 0:
      // no new command
      break;

    case -1:
      // comm error
      rcs_print_error("%s: NML config read error\n", argv[0]);
      break;

    case PRIM_MOB_JA_CFG_NOP_TYPE:
      primMobJASet.command_type = cfg_type;
      if (serial_number != primMobJASet.echo_serial_number) {
	primMobJASet.echo_serial_number = serial_number;
	primMobJASet.state = NEW_COMMAND;
	primMobJASet.status = RCS_EXEC;
      }
      break;

    case PRIM_MOB_JA_CFG_CYCLE_TIME_TYPE:
      primMobJASet.command_type = cfg_type;
      if (serial_number != primMobJASet.echo_serial_number) {
	primMobJASet.echo_serial_number = serial_number;
	primMobJASet.state = NEW_COMMAND;
	primMobJASet.status = RCS_EXEC;
      }
      cfgCycleTimePtr =
	(PrimMobJACfgCycleTime *) primMobJACfgBuf->
	get_address();
      break;

    case PRIM_MOB_JA_CFG_MOVE_TYPE:
      primMobJASet.command_type = cfg_type;
      if (serial_number != primMobJASet.echo_serial_number) {
	primMobJASet.echo_serial_number = serial_number;
	primMobJASet.state = NEW_COMMAND;
	primMobJASet.status = RCS_EXEC;
      }
      cfgMovePtr =
	(PrimMobJACfgMove *) primMobJACfgBuf->
	get_address();
      break;

    case PRIM_MOB_JA_CFG_DEBUG_TYPE:
      primMobJASet.command_type = cfg_type;
      if (serial_number != primMobJASet.echo_serial_number) {
	primMobJASet.echo_serial_number = serial_number;
	primMobJASet.state = NEW_COMMAND;
	primMobJASet.status = RCS_EXEC;
      }
      cfgDebugPtr =
	(PrimMobJACfgDebug *) primMobJACfgBuf->get_address();
      break;

    default:
      rcs_print_error("%s: unknown config: %d\n", argv[0], (int) cfg_type);
      break;
    }				// switch (cfg_type)

    // handle command state tables
    switch (primMobJAStat.command_type) {

    case PRIM_MOB_JA_CMD_NOP_TYPE:
      doCmdNop(&primMobJAStat, engine);
      break;

    case PRIM_MOB_JA_CMD_INIT_TYPE:
      doCmdInit(&primMobJAStat,
		servoMobJACmdBuf, servoMobJAStatBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_HALT_TYPE:
      doCmdHalt(&primMobJAStat,
		servoMobJACmdBuf, servoMobJAStatBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_ABORT_TYPE:
      doCmdAbort(&primMobJAStat,
		 servoMobJACmdBuf, servoMobJAStatBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_SHUTDOWN_TYPE:
      doCmdShutdown(&primMobJAStat,
		    servoMobJACmdBuf, servoMobJAStatBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
    case PRIM_MOB_JA_CMD_MOVE_ARC_SEGMENT_TYPE:
      doCmdMoveWaypoint(&primMobJAStat,
			&primMobJASet,
			servoMobJACmdBuf, servoMobJAStatBuf, navDataExtBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_ROTATE_TYPE:
      doCmdRotate(cmdRotatePtr,
		  &primMobJAStat,
		  &primMobJASet,
		  servoMobJACmdBuf,
		  servoMobJAStatBuf, navDataExtBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_VEL_TYPE:
      doCmdVel(cmdVelPtr,
	       &primMobJAStat,
	       servoMobJACmdBuf,
	       servoMobJAStatBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_STOP_TYPE:
      doCmdStop(&primMobJAStat,
		servoMobJACmdBuf, servoMobJAStatBuf, engine);
      break;

    case PRIM_MOB_JA_CMD_DOCK_TYPE:
      doCmdDock(&primMobJAStat,
		servoMobJACmdBuf, servoMobJAStatBuf, engine);
      break;

    default:
      // no command to handle
      break;
    }				// switch (cmd)

    // handle config state tables
    switch (primMobJASet.command_type) {

    case PRIM_MOB_JA_CFG_NOP_TYPE:
      doCfgNop(&primMobJASet, engine);
      break;

    case PRIM_MOB_JA_CFG_CYCLE_TIME_TYPE:
      doCfgCycleTime(cfgCycleTimePtr, &primMobJASet, moastErrorBuf, engine);
      break;

    case PRIM_MOB_JA_CFG_MOVE_TYPE:
      doCfgMove(cfgMovePtr, &primMobJASet, engine);
      break;

    case PRIM_MOB_JA_CFG_DEBUG_TYPE:
      doCfgDebug(cfgDebugPtr, &primMobJASet, engine);
      break;

    default:
      // no config to handle
      break;
    }				// switch (cfg)

    cycle_time_tracker(&primMobJAStat.tt);

    // update status and settings
    primMobJAStatBuf->write(&primMobJAStat);
    primMobJASetBuf->write(&primMobJASet);

    esleep(primMobJASet.cycleTime);
  }
  while (!done);

  (void) cleanupPrimMobJANmlBuffers();
  (void) cleanupServoMobJANmlBuffers();

  rcs_print("primPl done\n");

  return 0;
}

/*
  Modification history:

  $Log: primMobMain.cc,v $
  Revision 1.40  2008/10/13 18:26:39  ghenson
  Update to prim. Only sent the robot into reverse if prim generated the path

  Revision 1.39  2008/05/22 18:37:42  ghenson
  Q&D code to make rotate work for Ackerman steer

  Revision 1.38  2008/04/15 17:05:22  dr_steveb
  Sample docking command in prim

  Revision 1.37  2008/04/15 15:10:39  dr_steveb
  Added stub for docking command

  Revision 1.36  2008/01/08 20:18:16  proctor
  Added control point transform that lets the prim mobility controller act
  on the rotational center of the vehicle, which in general won't coincide
  with the nav data frame

  Revision 1.35  2007/12/18 21:34:46  proctor
  In prim mob, read out the platform name from servo mob and used it to
  get platform-specific parameters V_CUTOFF_ANGLE et al; changed moast.ini
  accordingly; fixed a return value error in the segment sweep code, where
  anyIntersections doesn't really return anything

  Revision 1.34  2007/12/17 22:14:57  dr_steveb
  Got rid of the SET_ACKERMAN environment variable. Now autodetected from the servo channels

  Revision 1.33  2007/11/19 18:08:50  proctor
  Fixed S0-S1 typo in call to engine

  Revision 1.32  2007/10/17 21:35:55  dr_steveb
  Added reading of ackerman and nautic specific parameters from .ini file

  Revision 1.31  2007/10/17 19:15:20  dr_steveb
  Added ability to drive rear-wheel steer vehicles

  Revision 1.30  2007/05/10 13:18:29  dr_steveb
  Separated the prim mobility planning engine from the interface wrappers. This will allow a controller to be created (the traffic controller) that controls a large number of vehicles without the overhead of individual NML channels for each vehicle. No changes in the algorithms were made.

  Revision 1.29  2007/05/02 03:03:47  tkharris
  more reasonable wheel geometry defaults

  Revision 1.28  2007/04/25 15:28:48  dr_steveb
  Backed out accidental mods.

  Revision 1.26  2007/01/31 18:36:11  cj_scrapper
  - Added robot types to SIMware knowledge base (ground vehicles, nautic vehicles, etc. )
  - Added robot type specific stat and set messages to ServoMobJA.hh
  - Update USARSim infskin to work with USARSim version 3.01
  - Modified PrimMobMain, joystick, and joySDL to comply with new ServoMobJA messages

  Revision 1.25  2007/01/25 17:44:50  dr_steveb
  Changed state to be explicit in the waypoint movement command

  Revision 1.24  2007/01/25 15:12:11  proctor
  Added handling of command type and serial number using the _sent values
  of the status buffer, via the SET_AND_RECORD and GOT_IT macros

  Revision 1.23  2006/12/14 19:47:27  dr_steveb
  Fixed some sub driving stuff

  Revision 1.22  2006/12/12 20:32:00  proctor
  Set defaults for wheel settings if not found

  Revision 1.21  2006/12/12 20:15:58  proctor
  Changed bad flag in prim shell to not bomb when wheel parameters aren't set

  Revision 1.20  2006/12/12 19:20:40  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.19  2006/11/30 17:32:56  proctor
  Tweaked the prim mobility controller for Nautic driving; stern plane gain
  now looks good and it tracks depths nicely

  Revision 1.18  2006/11/20 20:22:21  tomrkramer
  Added initialization of v when command is new, plus a comment about setting v

  Revision 1.17  2006/10/13 19:31:13  proctor
  Tweaked Nautic driving

  Revision 1.16  2006/10/02 13:59:10  proctor
  Changed PrimMobJACmdRotate to take a 'direction' value, either
  PRIM_MOB_JA_CMD_ROTATE_CW,CCW,SHORTEST. Changed primShell accordingly.

  Revision 1.15  2006/09/28 20:59:49  proctor
  Added Nautic steering

  Revision 1.14  2006/09/28 19:56:36  proctor
  Added Z field to waypoints and handled them backward-compatible in primShell.
  Computed pitch correction in doCmdMoveWaypoint, but didn't do anything with
  them yet pending print debugging.

  Revision 1.13  2006/09/20 21:28:07  proctor
  Added PrimMobJACmdStop and related code for sending it from the prim shell
  and handling it in primMobMain.

  Revision 1.12  2006/09/20 12:47:47  dr_steveb
  Added clockwise flag to rotate command

  Revision 1.11  2006/09/01 19:55:57  proctor
  Fixed bug where the get_address() wasn't called after an NML re-init

  Revision 1.10  2006/09/01 18:29:41  proctor
  Pulled out the actual wheelbase value from USE_ACKERMAN

  Revision 1.9  2006/09/01 18:19:34  proctor
  Forwent reading of USE_ACKERMAN value; it's enough that it's simply defined

  Revision 1.8  2006/09/01 18:16:40  proctor
  Checked for null getenv value before setting useAckerman

  Revision 1.7  2006/09/01 18:12:41  proctor
  Fixed up unreachable code error when building moast.jar; added Ackerman
  steering to primMobMain.cc

  Revision 1.6  2006/08/10 15:10:46  proctor
  primMobMain.cc -- took out ML driving in preparation for Ackerman steering;
  in src/java, changed the package name to a single 'moast'

  Revision 1.5  2006/04/25 20:30:18  proctor
  Added support for moastError to nmlPrint, and added error printing for
  bad configurations to primMobJA

  Revision 1.4  2006/04/13 14:30:09  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.3  2005/12/16 21:56:36  proctor
  Set wnew.

  Revision 1.2  2005/12/13 19:02:31  tomrkramer
  added map from waypoints to arcs

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.21  2005/09/30 18:55:46  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.20  2005/08/10 16:37:28  proctor
  primMobMain.cc: handled vCutoffAngle, wCutoffAngle in iniLoad, waypoint move.
  primShell.cc: handled vCutoffAngle, wCutoffAngle; incremented serial number
  for "wp" command each time.

  Revision 1.19  2005/08/02 20:40:32  proctor
  Handled speed element of Waypoint type

  Revision 1.18  2005/08/02 19:21:24  proctor
  Added check for 'isArc' flag in arc_segments_to_waypoints and handled
  lines properly.

  Revision 1.17  2005/07/28 20:08:45  proctor
  Used a global serial number "servoMobSerialNumber" instead of reading it
  out continually from the subordinate. This is intended to fix race conditions.

  Revision 1.16  2005/07/28 19:35:02  proctor
  Took out serial_incr in all the state tables-- we increment the serial number
  by one each time

  Revision 1.15  2005/07/28 19:25:42  proctor
  Incremented low-level driving command serial number by one each time,
  instead of leaving it the same for subsequent commands

  Revision 1.14  2005/07/28 19:08:51  proctor
  Fixed error with premature RCS_DONE reporting in doCmdInit,Halt,...

  Revision 1.13  2005/07/28 15:47:54  proctor
  Added checking for subordinate command, serial number and status instead
  of just status, when determining when a command is done, in all the doCmd...

  Revision 1.12  2005/07/26 16:02:58  stephen
  Changed debug message

  Revision 1.11  2005/07/20 21:53:39  proctor
  primMobMain.cc: added 10-second timeout for reading in servo mob settings.
  primShell.cc: added handling of lines via isArc in loadPath().

  Revision 1.10  2005/07/01 16:21:01  proctor
  Added getting DEBUG from .ini file

  Revision 1.9  2005/07/01 16:01:23  proctor
  Took out annoying .nml and .ini config printing

  Revision 1.8  2005/06/27 14:27:20  proctor
  Added arc segment move command

  Revision 1.7  2005/06/22 13:26:30  proctor
  Added setting of defaults in iniLoad(), for invalid entries.

  Revision 1.6  2005/06/21 20:58:29  kramer
  put reporting DONE into DO_CMD_MOVE_WAYPOINT

  Revision 1.5  2005/06/21 18:08:41  proctor
  Saved "worst" error, INI_DEFAULT or INI_INVALID, and returned that in iniLoad.

  Revision 1.4  2005/06/21 18:03:20  proctor
  Tweaked iniLoad() so that it always sets defaults, even for INI_INVALID

  Revision 1.3  2005/06/21 13:29:47  proctor
  Reworked iniLoad() to use INI_OK,DEFAULT,INVALID and exited upon INI_INVALID.

  Revision 1.2  2005/06/08 19:29:08  proctor
  Moved state_match, etc. macros into moastTypes.hh

  Revision 1.1  2005/06/08 14:30:55  proctor
  Added VehMobJA name changes

  Revision 1.9  2005/06/07 21:45:42  proctor
  Changed servoPL to servoMobJA

  Revision 1.8  2005/05/20 15:18:45  proctor
  In primPLMain.cc, fixed a bug where the command number wasn't set properly
  for new commands; changed INI file angle units to degrees; set the pathIndex
  in status. In primShell.cc, changed angle units to degrees.

  Revision 1.7  2005/05/17 21:28:44  proctor
  Referenced complete Waypoint struct, instead of individual fields that
  were used as an interim workaround for the code gen bug.

  Revision 1.6  2005/05/16 21:15:02  proctor
  First time consistently applying suffix numbers to buffers

  Revision 1.5  2005/05/13 20:19:38  proctor
  Added [PRIM_PL] MAX_TRAN_VEL; more shell features

  Revision 1.4  2005/05/13 19:02:55  proctor
  Added waypoint driving

  Revision 1.3  2005/05/12 13:24:32  proctor
  Added RCS_STAT_SOURCE_FILE_LEN

  Revision 1.2  2005/05/12 13:20:23  proctor
  Fixed Waypoint struct workaround

  Revision 1.1  2005/05/11 20:03:24  proctor
  Changed primPL.cc to primPLMain.cc, to avoid conflict with autogenerated
  NML source code.

*/

