/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   rcsManfactEngine.cc
  \brief Provide an RCS engine template

  Contains all of the state table and nml methods

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.7 $
  $Date: 2009/02/25 21:08:30 $
  \endcode

  \author Stephen Balakirsky
  \date   03/14/2006
*/

#include <unistd.h>		// close(), write()
#include <math.h>
#include <sys/types.h>          // open()
#include <sys/stat.h>           // open()
#include <fcntl.h>              // open()
#include <stdlib.h>		// system()
#include "rcsSectManfactEngine.hh"
#include "vehMobPL.hh"
#include "../vehicle/vehMobCell.hh"
#include <search.h>

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}

/***********************************************************************/
rcsEngine::rcsEngine(double cycleTime, double resolution)
{
  int count;

  debugFlags = 0;
  timer = new RCS_TIMER( cycleTime );
  nmlStatus = new SectMobPLStat();
  nmlSettings = new SectMobPLSet();
  goalHash = 0;
  vehStates = 0;
  vehicleSerialNumber = 0;

  nmlStatus->heartbeat = 0;
  nmlStatus->admin_state = ADMIN_SHUT_DOWN;
  nmlStatus->command_type = SECT_MOB_PL_CMD_NOP_TYPE;
  nmlStatus->status = RCS_DONE;
  nmlStatus->echo_serial_number = 0;
  reset_time_tracker( &nmlStatus->tt );
  // user status initialization
  nmlStatus->numVehicles = 0;

  nmlSettings->echo_serial_number = 0;
  nmlSettings->command_type = SECT_MOB_PL_CFG_NOP_TYPE;
  nmlSettings->status = RCS_DONE;
  nmlSettings->echo_serial_number = 0;
  nmlSettings->cycleTime = cycleTime;
  strcpy( nmlStatus->message, "" );

  // default cmd, stat, cfg, set buffers
  sectMobPLCmdNml = NULL;
  sectMobPLCfgNml = NULL;
  sectMobPLSetNml = NULL;
  sectMobPLStatNml = NULL;

  // user included buffers and command pointers
  for( count=0; count<MAX_SECTION_VEHICLES; count++ )
    {
      vehicleStatNml[count] = NULL;
      vehicleCmdNml[count] = NULL;
      vehicleCfgNml[count] = NULL;
      vehicleSetNml[count] = NULL;
    }
  nmlSettings->worldResolution = resolution;
  myNmlFileName = NULL;
  mySystemNmlName = NULL;
}

/***********************************************************************/
rcsEngine::~rcsEngine()
{
  if( myNmlFileName != NULL )
    delete [] myNmlFileName;
  if( mySystemNmlName != NULL )
    delete [] mySystemNmlName;
}

/***********************************************************************/
bool rcsEngine::runStateMachine(const char *systemNMLName)
{
  NMLTYPE cmdType, cfgType;
  int cmdSerialNumber, cfgSerialNumber;
  //  char newText[100];
  int count;
  bool writeSettings;

  /* status housekeeping */
  timer->wait(); // wait for cycle start
  nmlStatus->heartbeat++;
  cycle_time_tracker( &nmlStatus->tt );

  // read command
  if( sectMobPLCmdNml != NULL )
    cmdType = sectMobPLCmdNml->read();
  else
    cmdType = 0;
  if( cmdType > 0 )
    cmdSerialNumber = sectMobPLCmdNml->get_address()->serial_number;

  // read configuration
  if( sectMobPLCfgNml != NULL )
    cfgType = sectMobPLCfgNml->read();
  else
    cfgType = 0;
  if( cfgType > 0 )
    cfgSerialNumber = sectMobPLCfgNml->get_address()->serial_number;

  // read data and status buffers
  for( count=0; count<nmlStatus->numVehicles; count++ )
    {
      if( vehicleStatNml[count] != NULL )
	checkVehCmdStatus(count);
    }

  // perform housekeeping on command and configuration
  switch(cmdType)
    {
    case 0:
      // no new command
      break;
    case -1:
      // comm error
      disconnect(1);
      break;
    case SECT_MOB_PL_CMD_INIT_TYPE:
    case SECT_MOB_PL_CMD_ABORT_TYPE:
    case SECT_MOB_PL_CMD_HALT_TYPE:
    case SECT_MOB_PL_CMD_SHUTDOWN_TYPE:
    case SECT_MOB_PL_CMD_NOP_TYPE:
    case SECT_MOB_PL_CMD_CONNECT_TYPE:
    case SECT_MOB_PL_CMD_DISCONNECT_TYPE:
    case SECT_MOB_PL_CMD_EXPLORE_TYPE:
    case SECT_MOB_PL_CMD_CRASH_TYPE:
    case SECT_MOB_PL_CMD_LOAD_WM_TYPE:
    case SECT_MOB_PL_CMD_DELIVER_TYPE:
    case SECT_MOB_PL_CMD_WAYPOINT_TYPE:
      nmlStatus->command_type = cmdType;
      if(cmdSerialNumber != nmlStatus->echo_serial_number) 
	{
	  nmlStatus->echo_serial_number = cmdSerialNumber;
	  nmlStatus->state = NEW_COMMAND;
	  nmlStatus->status = RCS_EXEC;
	}
      // clear out status message
      sprintf(nmlStatus->message, "%s", "All monitored functions OK\n" );
      break;
    default:
      fprintf(stderr, "%s: unknown command %s\n", 
	      systemNMLName, sectMobPL_symbol_lookup(cmdType));
      break;
    } // switch (cmdType)

  switch(cfgType)
    {
    case 0:
      // no new command
      break;
    case -1:
      // comm error
      disconnect(1);
      break;
    case SECT_MOB_PL_CFG_CYCLE_TIME_TYPE:
    case SECT_MOB_PL_CFG_DEBUG_TYPE:
    case SECT_MOB_PL_CFG_NOP_TYPE:
    case SECT_MOB_PL_CFG_DUMP_WM_TYPE:
      nmlSettings->command_type = cfgType;
      if(cfgSerialNumber != nmlSettings->echo_serial_number) 
	{
	  nmlSettings->echo_serial_number = cfgSerialNumber;
	  nmlSettings->state = NEW_COMMAND;
	  nmlSettings->status = RCS_EXEC;
	}
      break;
    default:
      fprintf(stderr, "%s: unknown configuration %s\n", 
	      systemNMLName, sectMobPL_symbol_lookup(cfgType));
      break;
    }

  // perform command
  switch( nmlStatus->command_type )
    {
    case SECT_MOB_PL_CMD_INIT_TYPE:
      doCmdInit();
      break;
    case SECT_MOB_PL_CMD_ABORT_TYPE:
      doCmdAbort();
      break;
    case SECT_MOB_PL_CMD_HALT_TYPE:
      doCmdHalt();
      break;
    case SECT_MOB_PL_CMD_SHUTDOWN_TYPE:
      doCmdShutdown();
      break;
    case SECT_MOB_PL_CMD_NOP_TYPE:
      doCmdNop();
      break;
    case SECT_MOB_PL_CMD_CONNECT_TYPE:
      doCmdConnect();
      break;
    case SECT_MOB_PL_CMD_DISCONNECT_TYPE:
      doCmdDisconnect();
      break;
    case SECT_MOB_PL_CMD_EXPLORE_TYPE:
      doCmdExplore();
      break;
    case SECT_MOB_PL_CMD_CRASH_TYPE:
      doCmdCrash();
      break;
    case SECT_MOB_PL_CMD_LOAD_WM_TYPE:
      doCmdLoad();
      break;
    case SECT_MOB_PL_CMD_DELIVER_TYPE:
      doCmdDeliver();
      break;
    case SECT_MOB_PL_CMD_WAYPOINT_TYPE:
      doCmdWaypoint();
    }

  // perform configuration
  switch( nmlSettings->command_type )
    {
    case SECT_MOB_PL_CFG_CYCLE_TIME_TYPE:
      writeSettings |= doCfgCycleTime();
      break;
    case SECT_MOB_PL_CFG_DEBUG_TYPE:
      writeSettings |= doCfgDebug();
      break;
    case SECT_MOB_PL_CFG_NOP_TYPE:
      writeSettings |= doCfgNop();
      break;
    case SECT_MOB_PL_CFG_DUMP_WM_TYPE:
      writeSettings |= doCfgDumpWM();
      break;
    default:
      break;
    }

  // write out buffers
  if( sectMobPLStatNml != NULL )
    {
      sectMobPLStatNml->write(nmlStatus);
    }
  if( sectMobPLSetNml != NULL && writeSettings )
    {
      sectMobPLSetNml->write(nmlSettings);
      writeSettings = false;
    }
  return true;
}

/***********************************************************************/
RCS_ADMIN_STATE rcsEngine::getAdminState()
{
  return( nmlStatus->admin_state );
}

/***********************************************************************/
RCS_STATUS rcsEngine::getStatus()
{
  return( nmlStatus->status );
}

/***********************************************************************/
bool rcsEngine::connect(const char *nmlFileName,const char *systemNMLName,
			     const char *nmlPostfix )
{
  SectMobPLCmdNop *nopCmd;
  SectMobPLCfgNop *nopCfg;
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  if( myNmlFileName != NULL )
    delete [] myNmlFileName;
  if( mySystemNmlName != NULL )
    delete [] mySystemNmlName;
  
  myNmlFileName = new char[strlen(nmlFileName)+1];
  mySystemNmlName = new char[strlen(systemNMLName)+1];
  strcpy( myNmlFileName, nmlFileName );
  strcpy(mySystemNmlName, systemNMLName);

  sprintf( chanName, "%s%s", SECT_MOB_PL_CMD_NAME, nmlPostfix );
  sectMobPLCmdNml = new RCS_CMD_CHANNEL( sectMobPL_format, chanName,
					  (char*)systemNMLName, 
					  (char*)nmlFileName );
  if( NULL == sectMobPLCmdNml || !sectMobPLCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  sprintf( chanName, "%s%s",SECT_MOB_PL_STAT_NAME, nmlPostfix );
  sectMobPLStatNml = new RCS_STAT_CHANNEL( sectMobPL_format, chanName,
					    (char*)systemNMLName,
					    (char*)nmlFileName );
  if( NULL == sectMobPLStatNml || !sectMobPLStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  sprintf( chanName, "%s%s", SECT_MOB_PL_SET_NAME, nmlPostfix );
  sectMobPLSetNml = new RCS_STAT_CHANNEL( sectMobPL_format, chanName,
					   (char*)systemNMLName, 
					   (char*)nmlFileName );
  if( NULL == sectMobPLSetNml || !sectMobPLSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  sprintf( chanName, "%s%s", SECT_MOB_PL_CFG_NAME, nmlPostfix );
  sectMobPLCfgNml = new RCS_CMD_CHANNEL( sectMobPL_format, chanName,
					  (char*)systemNMLName, 
					  (char*)nmlFileName );
  if( NULL == sectMobPLCfgNml || !sectMobPLCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  nmlStatus->admin_state = ADMIN_UNINITIALIZED;
  // stuff a nop cmd
  nopCmd = new SectMobPLCmdNop();
  nopCfg = new SectMobPLCfgNop();

  if( nopCmd <= 0 || nopCfg <= 0 )
    {
      printf( "Allocation error!!!\n");
      return false;
    }

  nopCmd->serial_number = nmlStatus->echo_serial_number+1;
  nopCfg->serial_number = nmlSettings->echo_serial_number+1;

  sectMobPLCmdNml->write(nopCmd);
  sectMobPLCfgNml->write(nopCfg);
  delete nopCmd;
  delete nopCfg;

  return true;
}

/***********************************************************************/
bool rcsEngine::connectSubordinates(int *nmlPostfix,
				    int numSubordinates,
				    SectSubordinate subordinateClass)
{
  int count;
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  if( subordinateClass == VEHICLE )
    {
      for( count=0; count<numSubordinates; count++ )
	{
	  sprintf( chanName, "%s%d", VEH_MOB_PL_CMD_NAME, nmlPostfix[count] );
	  vehicleCmdNml[count] = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
						      mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleCmdNml[count] || !vehicleCmdNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", VEH_MOB_PL_STAT_NAME, nmlPostfix[count] );
	  vehicleStatNml[count] = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
							mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleStatNml[count] || !vehicleStatNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", VEH_MOB_PL_SET_NAME, nmlPostfix[count] );
	  vehicleSetNml[count] = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
						       mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleSetNml[count] || !vehicleSetNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", VEH_MOB_PL_CFG_NAME, nmlPostfix[count] );
	  vehicleCfgNml[count] = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
						      mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleCfgNml[count] || !vehicleCfgNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }
	}
    }
  nmlStatus->numVehicles = numSubordinates;
  return true;
}

/***********************************************************************/
void rcsEngine::disconnectSubordinates()
{
  int count;

  // disconnect from buffers
  for( count=0; count<nmlStatus->numVehicles; count++ )
    {
      if( vehicleCmdNml[count] != NULL )
	{
	  delete vehicleCmdNml[count];
	  vehicleCmdNml[count] = NULL;
	}
      if( vehicleStatNml[count] != NULL )
	{
	  delete vehicleStatNml[count];
	  vehicleStatNml[count] = NULL;
	}
      if( vehicleCfgNml[count] != NULL )
	{
	  delete vehicleCfgNml[count];
	  vehicleCfgNml[count] = NULL;
	}
      if( vehicleSetNml[count] != NULL )
	{
	  delete vehicleSetNml[count];
	  vehicleSetNml[count] = NULL;
	}
    }
}

/***********************************************************************/
void rcsEngine::disconnect(int error)
{
  // disconnect from subordinates
  disconnectSubordinates();

  nmlStatus->admin_state = ADMIN_SHUT_DOWN;
  if( error )
    nmlStatus->status = RCS_ERROR;
  else
    nmlStatus->status = RCS_DONE;
  nmlStatus->command_type = SECT_MOB_PL_CMD_NOP_TYPE;
  // send last status message out 
  if( sectMobPLCmdNml != NULL )
    sectMobPLCmdNml->write(nmlStatus);

  if( sectMobPLCmdNml != NULL )
    delete sectMobPLCmdNml;
  sectMobPLCmdNml = NULL;

  if( sectMobPLCfgNml != NULL )
    delete sectMobPLCfgNml;
  sectMobPLCfgNml = NULL;

  if( sectMobPLStatNml != NULL )
    delete sectMobPLStatNml;
  sectMobPLStatNml = NULL;

  if( sectMobPLSetNml != NULL )
    delete sectMobPLSetNml;
  sectMobPLSetNml = NULL;
}

/***********************************************************************/
const char *rcsEngine::getCommandText()
{
  return sectMobPL_symbol_lookup(nmlStatus->command_type);
}


/***********************************************************************/
/*!
  Perform connect state machine
*/
void rcsEngine::doCmdConnect()
{

}

/***********************************************************************/
/*!
  Perform disconnect state machine
*/
void rcsEngine::doCmdDisconnect()
{

}

/***********************************************************************/
/*!
  Perform init state machine

  \return true finished with initialization
  \return false still working
*/
void rcsEngine::doCmdInit()
{
  VehMobPLCmdInit  *initCmd;
  SectMobPLCmdInit *sectInitCmd;

  int count;


  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      sectInitCmd = (SectMobPLCmdInit*)sectMobPLCmdNml->get_address();
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state == ADMIN_SHUT_DOWN )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
      nmlSettings->worldHeight = 0;
      nmlSettings->worldWidth = 0;
      if(vehStates!= 0)
	delete vehStates;
      if(vehicleSerialNumber != 0 )
	delete vehicleSerialNumber;
      vehStates = new RCS_STATE[sectInitCmd->vehList_length];
      vehicleSerialNumber = new int[sectInitCmd->vehList_length];
      for( count=0; count<sectInitCmd->vehList_length; count++ )
	vehicleSerialNumber[count] = 0;
      if( nmlStatus->admin_state == ADMIN_INITIALIZED )
	{
	  printf( "Section already initialized\n" );
	  state_next(nmlStatus, S10);
	}
    }
  if( state_match(nmlStatus, S0) ) // connect to subordinates
    {
      sectInitCmd = (SectMobPLCmdInit*)sectMobPLCmdNml->get_address();
      nmlStatus->subordinate = sectInitCmd->subordinate;
      connectSubordinates(sectInitCmd->vehList, sectInitCmd->vehList_length, 
			  sectInitCmd->subordinate);
      if( sectInitCmd->subordinate == VEHICLE )
	state_next(nmlStatus, S1);
      else
	{
	  status_next(nmlStatus, RCS_ERROR);
	  state_next(nmlStatus,S10 );
	}
    }
  if( state_match(nmlStatus, S1 ))  // send vehicle init commands 
    {
      initCmd = new VehMobPLCmdInit();
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  initCmd->serial_number = ++vehicleSerialNumber[count];
	  vehicleCmdNml[count]->write(initCmd);
	}
      delete initCmd;
      state_next(nmlStatus, S2);
      return;
    }
  if( state_match(nmlStatus, S2) ) //wait for vehicle done
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber[count] )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S10);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S9);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S9) )
    {
      nmlStatus->admin_state = ADMIN_INITIALIZED;
      state_next(nmlStatus, S10);
    }
}

/***********************************************************************/
/*!
  Perform abort command state machine

  \param stat status message
  \return true need to update and send set message
  \return false no message needs to be sent
*/
void rcsEngine::doCmdAbort()
{
  VehMobPLCmdAbort *abortCmd;
  int count;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
    }
  if( state_match(nmlStatus, S0) )
    {
      abortCmd = new VehMobPLCmdAbort();
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  abortCmd->serial_number = ++vehicleSerialNumber[count];
	  vehicleCmdNml[count]->write(abortCmd);
	}
      delete abortCmd;
      state_next(nmlStatus, S1);
    }
  else if( state_match(nmlStatus, S1) )
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber[count] )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S2);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S2);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S2) )
    {
      nmlStatus->admin_state = ADMIN_UNINITIALIZED;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
/*!
  Perform  halt command state machine

  \param stat status message
  \return true done with command
  \return false still working
*/
void rcsEngine::doCmdHalt()
{
  VehMobPLCmdHalt *haltCmd;
  int count;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
    }
  if( state_match(nmlStatus, S0) )
    {
      haltCmd = new VehMobPLCmdHalt();
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  haltCmd->serial_number = ++vehicleSerialNumber[count];
	  vehicleCmdNml[count]->write(haltCmd);
	}
      delete haltCmd;
      state_next(nmlStatus, S1);
    }
  else if( state_match(nmlStatus, S1) )
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber[count] )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S2);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S2);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S2) )
    {
      nmlStatus->admin_state = ADMIN_UNINITIALIZED;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
/*!
  Perform shutdown command to subordinate planning system

  \param stat vehicle status
  \return true done with shutdown
  \return false not done yet
*/
void rcsEngine::doCmdShutdown()
{
  VehMobPLCmdShutdown *shutdownCmd;
  int count;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state == ADMIN_INITIALIZED )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
    }
  if( state_match(nmlStatus, S0) )
    {
      shutdownCmd = new VehMobPLCmdShutdown();
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  shutdownCmd->serial_number = ++vehicleSerialNumber[count];
	  vehicleCmdNml[count]->write(shutdownCmd);
	}
      delete shutdownCmd;
      state_next(nmlStatus, S1);
    }
  else if( state_match(nmlStatus, S1) )
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber[count] )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S2);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S2);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S2) )
    {
      disconnect(0);
      nmlStatus->admin_state = ADMIN_SHUT_DOWN;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
void rcsEngine::doCmdNop()
{
  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S1);
      status_next(nmlStatus, RCS_DONE);
    }
}

/***********************************************************************/
void rcsEngine::doCmdCrash()
{
  // crash command not supported in manufacturing application
  if( state_match(nmlStatus, NEW_COMMAND) ) // check for init state and vehicles
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      sprintf(nmlStatus->message, "%s", "crash command not supported in manufacturing application\n" );
      status_next(nmlStatus, RCS_ERROR);
      return;
    }
}

/***********************************************************************/
void rcsEngine::doCmdExplore()
{
  VehMobPLCmdMoveWaypoint vehicleMoveCmd;

  // command not supported in manufacturing application
  if( state_match(nmlStatus, NEW_COMMAND) ) 
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      sprintf(nmlStatus->message, "%s", "command not supported in manufacturing application\n" );
      state_next(nmlStatus, S1);
    }
}

/***********************************************************************/
/*!
  Perform actions from waypoint file
*/
#define BUFFERSIZE 255
void rcsEngine::doCmdWaypoint()
{
  static VehMobPLCmdGoNamedPt     *vehMobPLCmdGoNamedPt = NULL;
  SectMobPLCmdFile     *sectWaypointCmd;
  char buffer[BUFFERSIZE];
  char *ptr;
  int count;
  static int waitTimer;
  static FILE *wpFile = NULL;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  state_next(nmlStatus, S8);
	}
      if( wpFile != NULL )
	fclose(wpFile);
      wpFile = NULL;
    }
  if( state_match(nmlStatus, S0) )
    {
      state_next(nmlStatus, S1);
      sectWaypointCmd = (SectMobPLCmdFile*)sectMobPLCmdNml->get_address();
      wpFile = fopen( sectWaypointCmd->fileName, "r" );
      if( wpFile == NULL )
	{
	  printf( "Error opening input file %s from sect waypointCmd\n", sectWaypointCmd->fileName );
	  state_next(nmlStatus, S8);
	}
    }
  if( state_match(nmlStatus, S1) )
    {
      if( fgets(buffer, BUFFERSIZE, wpFile) == NULL )
	{
	state_next(nmlStatus, S2);
	return;
	}
      ptr = buffer;
      while (isspace(*ptr))
	ptr++;			// skip leading white space
      if (0 == *ptr) 		// blank line
	{
	  if( debugFlags )
	    printf( "Blank line\n" );
	}
      else if( ptr[0]=='#' || ptr[0]==';' || (ptr[0]=='/'&&ptr[1]=='/') ) // comment
	{
	  if( debugFlags )
	    printf( "Comment line\n" );
	}
      else if( !strwcmp(ptr, "goto") )
	{
	  if( vehMobPLCmdGoNamedPt == NULL )
	    vehMobPLCmdGoNamedPt = new VehMobPLCmdGoNamedPt();
	  if( 2!=sscanf(ptr, "%*s %lf %[0-9a-zA-Z_ -]", &vehMobPLCmdGoNamedPt->neighborhood, 
			vehMobPLCmdGoNamedPt->namedPt))
	    {
	      sprintf(nmlStatus->message, "Need neighborhood and location to go to\n");	      
	      printf( "Section error: %s\n", nmlStatus->message );
	      state_next(nmlStatus, S8);
	    }
	  else
	    {
	      if( debugFlags )
		printf( "Read Goto destination: %s\n", vehMobPLCmdGoNamedPt->namedPt );
	      for( count=0; count<nmlStatus->numVehicles; count++ )
		{
		  vehMobPLCmdGoNamedPt->serial_number = ++vehicleSerialNumber[count];
		  vehicleCmdNml[count]->write(vehMobPLCmdGoNamedPt);
		  state_next(nmlStatus, S4);
		}
	    }
	}
      else if( !strwcmp(ptr, "table") )
	{
	  if( 2 != sscanf(ptr, "%*s %d %d", &count, &waitTimer) )
	    {
	      sprintf(nmlStatus->message, "Need direction and time to wait in table statement in waypoint file\n" );
	      printf( "Section error: %s\n", nmlStatus->message );
	      state_next(nmlStatus, S8);
	    }
	  else
	    {
	      if( debugFlags )
		printf( "Table direction %d for %d cycles\n", count, waitTimer );
	      state_next(nmlStatus, S3);
	    }
	}
      else if( !strwcmp(ptr, "wait") )
	{
	  if( 1 != sscanf(ptr, "%*s %d", &waitTimer) )
	    {
	      sprintf(nmlStatus->message, "Need time to wait in wait statement in waypoint file\n" );
	      printf( "Section error: %s\n", nmlStatus->message );
	      state_next(nmlStatus, S8);
	    }
	  else
	    {
	      if( debugFlags )
		printf( "Read wait of %d cycles\n", waitTimer );
	      state_next(nmlStatus, S3);
	    }
	}
      else
	{
	  sprintf(nmlStatus->message, "Unknown keyword: %s", ptr );
	  printf( "Section error: %s\n", nmlStatus->message );
	  state_next(nmlStatus, S8);
	}
      return;
    }
  if( state_match(nmlStatus, S2) )
    {
      sectWaypointCmd = (SectMobPLCmdFile*)sectMobPLCmdNml->get_address();
      if( sectWaypointCmd->repeat )
	{
	  rewind( wpFile );
	  state_next( nmlStatus, S1 );
	}
      else
	status_next( nmlStatus, RCS_DONE );
    }
  if( state_match(nmlStatus, S3) )
    {
      if( waitTimer-- <= 0 )
	{
	  state_next(nmlStatus, S1);
	  system( "echo Stop table\n" );
	}
    }
  if( state_match(nmlStatus, S4) ) //wait for vehicle done
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber[count] )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S8);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S1);
    }
  if( state_match(nmlStatus, S8) )
    {
      status_next(nmlStatus, RCS_ERROR );
      return;
    }
}

/***********************************************************************/
/*!
  Perform load world model command state machine

*/
void rcsEngine::doCmdLoad()
{
  VehMobPLCmdLoadWM *loadCmd;
  SectMobPLCmdLoadWM *sectLoadCmd;
  int count;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  state_next(nmlStatus, S8);
	}
      if( goalHash!=0 )
	hdestroy();
      goalHash = hcreate(30);
    }
  if( state_match(nmlStatus, S0) )
    {
      sectLoadCmd = (SectMobPLCmdLoadWM*)sectMobPLCmdNml->get_address();
      loadCmd = new VehMobPLCmdLoadWM();
      strcpy( loadCmd->fileName, sectLoadCmd->fileName );
      strcpy(nmlSettings->wmFileName, loadCmd->fileName);
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  loadCmd->serial_number = ++vehicleSerialNumber[count];
	  vehicleCmdNml[count]->write(loadCmd);
	}
      delete loadCmd;
      state_next(nmlStatus, S1);
    }
  if( state_match(nmlStatus, S1) )
    {
      if( loadWM(sectLoadCmd->fileName) )
	{
	  state_next(nmlStatus, S2);
	  status_next(nmlStatus, RCS_DONE);
	}
      else
	{
	  state_next(nmlStatus, S8);
	}
    }
  else if( state_match(nmlStatus, S2) )
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber[count] )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      state_next(nmlStatus, S8);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S3);
    }
  if( state_match(nmlStatus, S8) )
    {
      status_next(nmlStatus, RCS_ERROR);
      //      nmlStatus->admin_state = ADMIN_UNINITIALIZED;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
/*!
  Perform delivery of packages around a factory. This uses a simple
  traffic management scheme for two vehicles. If both want to go to the same place
  one will be sent there and the other will be sent to a staging area
  The vehicles have substates to the section state that tell where each individual
  is in their delivery process. These substates are set in the vehStates array and are
  as follows:
  S0 : Unused
  S1 : Waiting for done moving to Loading area
  S2 : Waiting for done moving to Loading Staging area
  S3 : Waiting for docking done;
  S4 : Waiting for package load done
  S5 : Waiting for undock done
  S6 : Waiting for done moving to package destination
  S7 : Waiting for done moving to package destination staging area
  S8 : Waiting for docking done at destination
  S9 : Waiting for unload done

*/
void rcsEngine::doCmdDeliver()
{
  int count;

  VehMobPLCmdGoNamedPt     *vehCmdGoNamedPt;
  VehMobPLCmdDock          *vehCmdDock;
  VehMobPLCmdUndock        *vehCmdUndock;
  VehMobPLCmdLoadPkg       *vehCmdLoadPkg;
  VehMobPLCmdUnloadPkg     *vehCmdUnloadPkg;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      status_next(nmlStatus, RCS_EXEC);
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  state_next(nmlStatus, S8);
	}
    }
  if( state_match(nmlStatus, S0) ) // send vehicles to pick-up packages
    {
      vehCmdGoNamedPt = new VehMobPLCmdGoNamedPt();
      vehCmdGoNamedPt->neighborhood = 2;
      if( nmlStatus->numVehicles >= 1 ) // send out to first vehicle
	{
	  vehCmdGoNamedPt->serial_number = ++vehicleSerialNumber[0];
	  strcpy(vehCmdGoNamedPt->namedPt, "Load");
	  if( !setHashValue(vehCmdGoNamedPt->namedPt, 1 )  )
	    return;
	  vehicleCmdNml[0]->write(vehCmdGoNamedPt);
	  vehStates[0] = S1;
	}
      if( nmlStatus->numVehicles >= 2 ) // send out to holding area
	{
	  vehCmdGoNamedPt->serial_number = ++vehicleSerialNumber[1];
	  strcpy(vehCmdGoNamedPt->namedPt, "Stage Load");
	  if( !setHashValue(vehCmdGoNamedPt->namedPt, 2 )  )
	    return;
	  vehicleCmdNml[1]->write(vehCmdGoNamedPt);
	  vehStates[1] = S2;
	}
      state_next(nmlStatus, S1);
      delete vehCmdGoNamedPt;
    }
  if( state_match(nmlStatus, S1) )
    {
      //figure out what to do for each vehicle
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  switch( vehStates[count] )
	    {
	    case S1: // waiting for movement to Load
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdDock = new VehMobPLCmdDock();
	      vehCmdDock->serial_number = ++vehicleSerialNumber[count];
	      vehicleCmdNml[count]->write(vehCmdDock);
	      delete vehCmdDock;
	      vehStates[count] = S3;
	      break;
	    case S2: /// waiting for movement to Stage Load and Load to clear
	      if( stopVehStateMachine(count) )
		break;
	      if( getHashValue("Load")!=0 )
		break;
	      vehCmdGoNamedPt = new VehMobPLCmdGoNamedPt();
	      vehCmdGoNamedPt->serial_number = ++vehicleSerialNumber[count];
	      vehCmdGoNamedPt->neighborhood = 2;
	      strcpy(vehCmdGoNamedPt->namedPt, "Load");
	      setHashValue("Load", count+1);
	      vehicleCmdNml[count]->write(vehCmdGoNamedPt);
	      delete vehCmdGoNamedPt;
	      setHashValue("Stage Load", 0);
	      vehStates[count] = S1;
	      break;
	    case S3: // Waiting for dock
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdLoadPkg = new VehMobPLCmdLoadPkg();
	      vehCmdLoadPkg->serial_number = ++vehicleSerialNumber[count];
	      vehicleCmdNml[count]->write(vehCmdLoadPkg);
	      delete vehCmdLoadPkg;
	      vehStates[count] = S4;
	      break; 
	    case S4: // Waiting for package load
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdUndock = new VehMobPLCmdUndock();
	      vehCmdUndock->serial_number = ++vehicleSerialNumber[count];
	      vehicleCmdNml[count]->write(vehCmdUndock);
	      delete vehCmdUndock;
	      vehStates[count] = S5;
	      break;
	    case S5: // Waiting for undock
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdGoNamedPt = new VehMobPLCmdGoNamedPt();
	      vehCmdGoNamedPt->serial_number = ++vehicleSerialNumber[count];
	      vehCmdGoNamedPt->neighborhood = 2;
	      // figure out where to go
	      if( getHashValue(vehiclePackageDest[count])!=0 ) // need to go to alternate location
		{
		  sprintf(vehCmdGoNamedPt->namedPt, "Stage %s", vehiclePackageDest[count]);
		  setHashValue(vehCmdGoNamedPt->namedPt, count+1);
		  vehStates[count] = S7;
		}
	      else
		{
		  strcpy(vehCmdGoNamedPt->namedPt, vehiclePackageDest[count]);
		  setHashValue(vehCmdGoNamedPt->namedPt, count+1);
		  strcpy( vehicleCachedDest[count], vehiclePackageDest[count] );
		  vehStates[count] = S6;
		}
	      vehicleCmdNml[count]->write(vehCmdGoNamedPt);
	      delete vehCmdGoNamedPt;
	      // clear out destination lock
	      setHashValue("Load", 0 );
	      break;
	    case S6: // waiting for movement to package destination
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdDock = new VehMobPLCmdDock();
	      vehCmdDock->serial_number = ++vehicleSerialNumber[count];
	      vehicleCmdNml[count]->write(vehCmdDock);
	      delete vehCmdDock;
	      vehStates[count] = S8;
	      break;
	    case S7: // waiting for movement to package destination Load and destination to clear
	      if( stopVehStateMachine(count) )
		break;
	      if( getHashValue(vehiclePackageDest[count])!=0 )
		break;
	      vehCmdGoNamedPt = new VehMobPLCmdGoNamedPt();
	      vehCmdGoNamedPt->serial_number = ++vehicleSerialNumber[count];
	      vehCmdGoNamedPt->neighborhood = 2;
	      // clear out stage destination 
	      sprintf(vehCmdGoNamedPt->namedPt, "Stage %s", vehiclePackageDest[count]);
	      setHashValue(vehCmdGoNamedPt->namedPt, 0);
	      strcpy(vehCmdGoNamedPt->namedPt, vehiclePackageDest[count]);
	      // set new hash lock
	      setHashValue(vehiclePackageDest[count], count+1);
	      strcpy( vehicleCachedDest[count], vehiclePackageDest[count] );
	      vehicleCmdNml[count]->write(vehCmdGoNamedPt);
	      delete vehCmdGoNamedPt;
	      vehStates[count] = S6;
	      break;
	    case S8: // Waiting for dock
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdUnloadPkg = new VehMobPLCmdUnloadPkg();
	      vehCmdUnloadPkg->serial_number = ++vehicleSerialNumber[count];
	      vehicleCmdNml[count]->write(vehCmdUnloadPkg);
	      delete vehCmdUnloadPkg;
	      vehStates[count] = S9;
	      break; 
	    case S9: // Waiting for package unload
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdUndock = new VehMobPLCmdUndock();
	      vehCmdUndock->serial_number = ++vehicleSerialNumber[count];
	      vehicleCmdNml[count]->write(vehCmdUndock);
	      delete vehCmdUndock;
	      vehStates[count] = S10;
	      break;
	    case S10: // Waiting for undock
	      if( stopVehStateMachine(count) )
		break;
	      vehCmdGoNamedPt = new VehMobPLCmdGoNamedPt();
	      vehCmdGoNamedPt->serial_number = ++vehicleSerialNumber[count];
	      vehCmdGoNamedPt->neighborhood = 2;
	      // clear hash value
	      setHashValue(vehicleCachedDest[count], 0 );
	      // figure out where to go
	      if( getHashValue("Load")!=0 ) // need to go to alternate location
		{
		  strcpy(vehCmdGoNamedPt->namedPt, "Stage Load" );
		  setHashValue(vehCmdGoNamedPt->namedPt, count+1);
		  vehStates[count] = S2;
		}
	      else
		{
		  strcpy(vehCmdGoNamedPt->namedPt, "Load" );
		  setHashValue(vehCmdGoNamedPt->namedPt, count+1);
		  vehStates[count] = S1;
		}
	      vehicleCmdNml[count]->write(vehCmdGoNamedPt);
	      delete vehCmdGoNamedPt;
	      break;
	    default:
	      break;
	    }
	}
      return;
    }
  if( state_match(nmlStatus, S8) )
    {
      status_next(nmlStatus, RCS_ERROR);
      nmlStatus->admin_state = ADMIN_UNINITIALIZED;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
/*!
  set the value pointed to by the key in the hash table
  \param key Key in hash table
  \value Value to set it to
  \return 1 success, 0 failure
*/
bool rcsEngine::setHashValue( char *key, int value )
{
  ENTRY e, *ep;

  e.key = key;
  ep = hsearch(e, FIND);
  if( ep==NULL )
    {
      sprintf(nmlStatus->message, "Unable to find: %s", key );
      nmlStatus->message_length = strlen(nmlStatus->message)+1;
      status_next(nmlStatus, RCS_ERROR);
      state_next(nmlStatus, S10);
      return false;
    }
  else
    *(int*)ep->data = value;
  return true;
}

/***********************************************************************/
/*!
  get the value pointed to by the key in the hash table
  \param key Key in hash table
  \return value
*/
int rcsEngine::getHashValue( char *key )
{
  ENTRY e, *ep;

  e.key = key;
  ep = hsearch(e, FIND);
  if( ep==NULL )
    {
      sprintf(nmlStatus->message, "Unable to find: %s", key );
      nmlStatus->message_length = strlen(nmlStatus->message)+1;
      status_next(nmlStatus, RCS_ERROR);
      state_next(nmlStatus, S10);
      return -1;
    }
  else
    return(*(int*)ep->data);
}

/***********************************************************************/
/*!
  check the vehicle status message to see if we should continue with this state
\return true to stop, do not go to next state
\return false to continue
*/
bool rcsEngine::stopVehStateMachine(int vehID)
{
  if( vehicleCmdEchoSerialNumber[vehID] != vehicleSerialNumber[vehID] )
    return 1;
  if( vehicleCmdStatus[vehID] == RCS_ERROR )
    {
      status_next(nmlStatus, RCS_ERROR);
      printf( "Error reported from vehicle stopVehStateMachine\n");
      state_next(nmlStatus, S10);
      return 1;
    }
    if( vehicleCmdStatus[vehID] == RCS_EXEC)
      return 1;
    return 0;
}

/***********************************************************************/
/*!
  Perform clear wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgNop()
{
  bool retValue = false;

  if( state_match(nmlSettings, NEW_COMMAND) )
    {
      state_new(nmlSettings);
      status_next(nmlSettings, RCS_DONE);
      state_next(nmlSettings, S0);
      retValue = true;
    } // fall through to checking the status
  return retValue;
}

/***********************************************************************/
/*!
  Perform cfgcycletime  state machine
*/
bool rcsEngine::doCfgCycleTime()
{
  SectMobPLCfgCycleTime *sectCfgCycleTime;

  sectCfgCycleTime = (SectMobPLCfgCycleTime*)sectMobPLCfgNml->get_address();
  delete timer;
  timer = new RCS_TIMER( sectCfgCycleTime->cycleTime );
  return false;
}

/***********************************************************************/
/*!
  Perform cfgDebug  state machine
*/
bool rcsEngine::doCfgDebug()
{
  SectMobPLCfgDebug *sectMobPLCfgDebug;
  sectMobPLCfgDebug = (SectMobPLCfgDebug*)sectMobPLCfgNml->get_address();

  debugFlags = sectMobPLCfgDebug->debug;
  return false;
}

/***********************************************************************/
/*!
  Perform clear wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgClearWM()
{
  bool retValue = false;

  if( state_match(nmlSettings, NEW_COMMAND) )
    {
      state_new(nmlSettings);
      status_next(nmlSettings, RCS_EXEC);
      state_next(nmlSettings, S0);
      retValue = true;
    } 

  if( state_match(nmlSettings, S0) )
    {
      status_next(nmlSettings, RCS_DONE);
      state_next(nmlSettings, S1);
      retValue = true;
    }
  return retValue;
}


/***********************************************************************/
/*!
  Perform load wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgLoadWM()
{
  bool retValue = false;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      status_next(nmlStatus, RCS_EXEC);
      state_next(nmlStatus, S0);
      retValue = true;
    } // fall through to loading the file
      
  if( state_match(nmlStatus, S1) )
    {
      status_next(nmlStatus, RCS_DONE);
      state_next(nmlStatus, S1);
      retValue = true;
    }
  return retValue;
}

/***********************************************************************/
/*!
  Perform store wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgStoreWM()
{
  bool retValue = false;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      status_next(nmlStatus, RCS_EXEC);
      state_next(nmlStatus, S0);
      retValue = true;
    } // fall through to storing the file

  if( state_match(nmlStatus, S0) )
    {
      status_next(nmlStatus, RCS_DONE);
      state_next(nmlStatus, S1);
      retValue = true;
    }
  return retValue;
}


/***********************************************************************/
/*!
  Perform dump wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgDumpWM()
{
  bool retValue = false;

  if( state_match(nmlSettings, NEW_COMMAND) )
    {
      state_new(nmlSettings);
      status_next(nmlSettings, RCS_DONE);
      state_next(nmlSettings, S4);
      retValue = true;
    }
  return retValue;
}

/***********************************************************************/
/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool rcsEngine::checkVehCmdStatus(int which)
{
  int type;
  VehManufactStat *manufactStat;

  type = vehicleStatNml[which]->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_STAT_TYPE:
    case VEH_MANUFACT_STAT_TYPE:
      manufactStat = (VehManufactStat*)vehicleStatNml[which]->get_address();
      vehicleCmdStatus[which] = manufactStat->status;
      vehicleCmdEchoSerialNumber[which] = manufactStat->echo_serial_number;
      strncpy( vehiclePackageDest[which], 
	       manufactStat->packageDestination, 
	       MAX_WAYPOINT_NAME_LENGTH );
      return(true);
    default:
      break;
    }
  return false;
}


/***********************************************************************/
/*!
  Load the world model to get the possible destinations
  
  \param name The name of the file to load

  \return true on success, false on failure
*/
#define INPUTBUFSIZE 256
bool rcsEngine::loadWM( char *name )
{
  FILE *mifFp, *midFp;
  char inputBuf[INPUTBUFSIZE];
  char featureName[INPUTBUFSIZE];
  ENTRY e, *ep;
  int *hashValue;
  char *hashKey;

  sprintf( inputBuf, "%s.mif", name );
  mifFp = fopen( inputBuf, "r" );
  if( mifFp == NULL )
    {
      printf( "Error opening input file %s from sect. loadWM\n", inputBuf );
      return 0;
    }
  sprintf( inputBuf, "%s.mid", name );
  midFp = fopen( inputBuf, "r" );
  if( midFp == NULL )
    {
      printf( "Error opening input file %s from sect. loadWM\n", inputBuf );
      return 0;
    }

  //open file and read data
  while( fgets(inputBuf, INPUTBUFSIZE, mifFp) != NULL )
    {
      // use fancy formatting to get at string with white space
      if( !strncasecmp( "Region", inputBuf, 6 ) )
	fscanf( midFp, "%*[^\"]%*[\"]%[^\"]%*s", featureName );
      else if( !strncasecmp( "Point", inputBuf, 5 ) )
	{
	  fscanf( midFp, "%*[^\"]%*[\"]%[^\"]%*s", featureName );
	  hashValue = new int();
	  *hashValue = 0;
	  hashKey = new char[strlen(featureName)+1];
	  strcpy( hashKey, featureName );
	  e.key = hashKey;
	  e.data = (void*)hashValue;
	  ep = hsearch(e, ENTER);
	  /* there should be no failures */
	  if( ep == NULL )
	    {
	      fprintf(stderr, "Section Loading named points failed\n");
	      return false;
	    }
	  printf( "Section Found point of interest: named: %s\n",
		  hashKey );
	}
    }
  return true;
}
/*
  Modification history:
  $Log: rcsSectManfactEngine.cc,v $
  Revision 1.7  2009/02/25 21:08:30  dr_steveb
  Fixed files to work with renamed section commands.

  Revision 1.6  2009/01/17 16:45:12  ghenson
  included stdlib.h for system() and rand()
  Could just be a gcc 4.3.x thing

  Revision 1.5  2008/02/15 20:20:35  dr_steveb
  Adde waypoint command to the section echelon

  Revision 1.4  2008/01/09 21:42:47  dr_steveb
  Minor changes before tarball release

  Revision 1.3  2008/01/07 21:49:00  dr_steveb
  Version of section controller for manufacturing competition. Added the ability to have two vehicles pick-up and deliver packages.

  Revision 1.2  2007/12/20 02:32:45  dr_steveb
  Now have two vehicles performing start of manufacturing job. More to come...

  Revision 1.1  2007/12/17 21:36:03  dr_steveb
  Initial check-in of section level manufacturing application

*/

