/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   rcsEngine.hh
  \brief Provide an RCS engine for the section mob applicaiton

  Contains all of the state tables and nml methods

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.3 $
  $Date: 2006/09/18 16:59:27 $
  \endcode

  \author Stephen Balakirsky
  \date   03/14/2006
*/
#ifndef RCS__ENGINE
#define RCS__ENGINE
#include <iostream>
#include "sensorData.hh"
#include "primSP.hh"
#include "primMobJA.hh"
#include "sectMobPL.hh"
#include "navDataExt.hh"

class rcsEngine
{
public:
  rcsEngine(double cycleTime, double resolution);
  ~rcsEngine();
  bool connect(const char *, const char*, const char*); // connect nml buffers
  void disconnect(int error); // disconnect nml buffers
  void disconnectSubordinates(); // disconnect nml buffers
  bool connectSubordinates(int *nmlPostfix,
			   int numSubordinates,
			   SectSubordinate subordinateClass );
  RCS_ADMIN_STATE getAdminState(); // return the admin state
  RCS_STATUS getStatus(); // return the module status
  const char *getCommandText(); // return the text for a given command
  bool runStateMachine(const char *systemNmlName); // runs the modules state machine

private:
  void doCmdInit();
  void doCmdAbort();
  void doCmdHalt();
  void doCmdShutdown();
  void doCmdNop();
  void doCmdConnect();
  void doCmdDisconnect();
  void doCmdExplore();
  void doCmdCrash();

  bool doCfgClearWM();
  bool doCfgCycleTime();
  bool doCfgDebug();
  bool doCfgDumpWM();
  bool doCfgLoadWM();
  bool doCfgNop();
  bool doCfgStoreWM();

  // user added routines
  bool checkVehCmdStatus(int which);
  bool checkPrimMobCmdStatus(int which);
  bool checkPrimSPCmdStatus(int which);
  bool clearMap();
  bool computeDestination( int whichVeh, double exploreDist );
  void dump(int fileID);
  bool growWorld(PM_CARTESIAN &includeLoc);
  bool growWorld(int newWorldWidth, int newWorldHeight, 
		  PM_CARTESIAN &newOffset);
  int loadWaypoints( char *filename, int vehicleID, Waypoint *waypoint, int *waypoint_length );
  bool loadWM(char *fileName);
  int mapToWorldModel(PM_CARTESIAN site);
  int mapXToWmNorth(double mapX);
  int mapYToWmEast(double mapY);
  bool readVehicleMap( int which );
  bool storeWM(char *fileName);
  PM_CARTESIAN transformPoint(int which, PM_CARTESIAN &pointIn,
			      bool transformGlobalToLocal);
  bool updatePosition(int which);
  bool updateWorld(SensorPoint *updates, int length);
  bool updateWorld(SensorData2D *updates);
  double worldModelToMapX(int index);
  double worldModelToMapY(int index);

  // nml message types
  SectMobPLStat *nmlStatus; // the modules status buffer contents
  SectMobPLSet *nmlSettings; // the modules settings buffer
  NavDataExt vehicleNav[MAX_SECTION_VEHICLES];  // nav data from the boat
  RCS_STATUS vehicleCmdStatus[MAX_SECTION_VEHICLES];
  RCS_STATUS primSPCmdStatus[MAX_SECTION_VEHICLES];
  RCS_STATUS primMobCmdStatus[MAX_SECTION_VEHICLES];
  int vehicleCmdEchoSerialNumber[MAX_SECTION_VEHICLES];
  int primSPCmdEchoSerialNumber[MAX_SECTION_VEHICLES];
  int primMobCmdEchoSerialNumber[MAX_SECTION_VEHICLES];

  //! Transforms from local to global and back
  PM_HOMOGENEOUS homLocal[MAX_SECTION_VEHICLES], homGlobal[MAX_SECTION_VEHICLES];

  // nml channels
  //! section planning channels
  RCS_CMD_CHANNEL  *sectMobPLCmdNml;
  RCS_CMD_CHANNEL  *sectMobPLCfgNml;
  RCS_STAT_CHANNEL *sectMobPLStatNml;
  RCS_STAT_CHANNEL *sectMobPLSetNml;
  //! am sensor data channel
  NML              *sensorDataNml[MAX_SECTION_VEHICLES];
  //! navigation data channel
  NML              *navDataExtNml[MAX_SECTION_VEHICLES];
  //! veh planning channels
  RCS_CMD_CHANNEL  *vehicleCmdNml[MAX_SECTION_VEHICLES];
  RCS_STAT_CHANNEL *vehicleStatNml[MAX_SECTION_VEHICLES];
  RCS_CMD_CHANNEL  *vehicleCfgNml[MAX_SECTION_VEHICLES];
  RCS_STAT_CHANNEL *vehicleSetNml[MAX_SECTION_VEHICLES];
  //! prim channels
  RCS_CMD_CHANNEL  *primSPCmdNml[MAX_SECTION_VEHICLES];
  RCS_CMD_CHANNEL  *primMobCmdNml[MAX_SECTION_VEHICLES];
  RCS_STAT_CHANNEL *primSPStatNml[MAX_SECTION_VEHICLES];
  RCS_STAT_CHANNEL *primMobStatNml[MAX_SECTION_VEHICLES];

  // cycle timer
  RCS_TIMER *timer;
  // command helpers
  PM_CARTESIAN vehicleDestination[MAX_SECTION_VEHICLES];
  bool vehicleSuccess[MAX_SECTION_VEHICLES];
  //! serial number for commands sent to vehicle
  int vehicleSerialNumber;
  // world model variables
  //! data array for writing output map
  char *dataPt;
  //! World  model 0-unexplored, 1-explored
  char **worldModel;
  //! offset of the world
  PM_CARTESIAN worldOffset;
  //! distance from center of cell to edge in straight linen
  float straightDist;
  //! distance from center of cell to edge by diag
  float diagDist;
  //! my verbose level
  bool myVerbose;
  //! the nml file that is being used
  char *myNmlFileName;
  //! the name of this process
  char *mySystemNmlName;
};
#endif

/*
  Modification history:
  $Log: rcsSectMobPlEngine.hh,v $
  Revision 1.3  2006/09/18 16:59:27  dr_steveb
  Added crash behavior

  Revision 1.2  2006/04/07 18:33:57  dr_steveb
  Check-in so compiles work for Fred. Still needs more work.

  Revision 1.1  2006/04/05 20:25:43  dr_steveb
  Generic rcs implementation for section level

*/
