/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file ext_servo.c

  \brief Implementation of external interface from Go Motion to
  SimWare.  The ext_XXX functions all have C linkage, which is what Go
  Motion expects.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.3 $
  $Date: 2009/04/20 13:05:53 $
  \endcode

  \author Fred Proctor
*/

/*
  To get this external interface linked in to the gomotion controller,
  first install gomotion, then build this, then reinstall gomotion
  like this:

  cd path/to/gomotion
  ./configure --with-extintf="/usr/local/proj/moast/devel/lib/libextservo.a -L/usr/local/proj/moast/devel/lib -lmoast -L/usr/local/rcslib/lib -lrcs"
  make

  Edit gomotion's ini file so that the EXT_INIT_STRING is the path to
  the NML file. Set the [SERVO_X] TYPE = PASS so that Go doesn't do
  servo control, but passes through setpoints to the external
  interface.

  In gomotion, run as:

  bin/gorun -i etc/gomoast.ini -m extintf
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <gotypes.h>
#include <extintf.h>
#include <rtapi.h>

#define NUM_JOINTS 8

/* these are the functions with C linkage in the C++ wrapper code */
extern int c_init(char * init_string);
extern void c_quit(void);
extern int c_write_pos(int joint, double pos);
extern double c_read_pos(int joint);

go_result ext_init(char * init_string)
{
  /* the init_string is the path to the NML file */
  if (0 != c_init(init_string)) {
    rtapi_print("can't init external servo interface\n");
    return GO_RESULT_ERROR;
  }

  return GO_RESULT_OK;
}
 
go_result ext_quit(void)
{
  c_quit();

  return GO_RESULT_OK;
}

go_result ext_joint_init(go_integer joint, go_real cycle_time)
{
  if (joint < 0 || joint >= NUM_JOINTS) return GO_RESULT_ERROR;

  return GO_RESULT_OK;
}

go_result ext_joint_enable(go_integer joint)
{
  return GO_RESULT_OK;
}

go_result ext_joint_disable(go_integer joint)
{
  return GO_RESULT_OK;
}

go_result ext_joint_quit(go_integer joint)
{
  return GO_RESULT_OK;
}

go_result ext_read_pos(go_integer joint, go_real * pos)
{
  if (joint < 0 || joint >= NUM_JOINTS) return GO_RESULT_ERROR;

  *pos = c_read_pos(joint);

  return GO_RESULT_OK;
}

go_result ext_write_pos(go_integer joint, go_real pos)
{
  if (0 != c_write_pos((int) joint, (double) pos)) return GO_RESULT_ERROR;

  return GO_RESULT_OK;
}

go_result ext_write_vel(go_integer joint, go_real vel)
{
  if (joint < 0 || joint >= NUM_JOINTS) return GO_RESULT_ERROR;

  return GO_RESULT_OK;
}

go_result ext_joint_home(go_integer joint)
{
  if (joint < 0 || joint >= NUM_JOINTS) return GO_RESULT_ERROR;
  
  return GO_RESULT_OK;
}

go_flag ext_joint_is_home(go_integer joint)
{
  if (joint < 0 || joint >= NUM_JOINTS) return 1; /* always homed */

  return 1;			/* always homed */
}

go_result ext_joint_home_latch(go_integer joint, go_real * pos)
{
  if (joint < 0 || joint >= NUM_JOINTS) return GO_RESULT_ERROR;

  *pos = 0.0;

  return GO_RESULT_OK;
}

enum {AIN_NUM = 0};
enum {AOUT_NUM = 0};
enum {DIN_NUM = 0};
enum {DOUT_NUM = 0};

go_integer ext_num_ain(void)
{
  return AIN_NUM;
}

go_integer ext_num_aout(void)
{
  return AOUT_NUM;
}

go_integer ext_num_din(void)
{
  return DIN_NUM;
}

go_integer ext_num_dout(void)
{
  return DOUT_NUM;
}

go_result ext_trigger_in(void)
{
  return GO_RESULT_OK;
}

go_result ext_read_ain(go_integer index, go_real * val)
{
  *val = 0.0;

  return GO_RESULT_OK;
}

go_result ext_write_aout(go_integer index, go_real val)
{
  return GO_RESULT_OK;
}

go_result ext_read_din(go_integer index, go_flag * val)
{
  *val = 0;

  return GO_RESULT_OK;
}

go_result ext_write_dout(go_integer index, go_flag val)
{
  return GO_RESULT_OK;
}

/*
  Modification history:

  $Log: ext_servo.c,v $
  Revision 1.3  2009/04/20 13:05:53  proctor
  Took out CXX from pthread configuration, since we use these in both C and
  C++ code and the CXX is misleading; tested gomotion control of a mission
  package with the new simware and updated a comment in ext_servo.c

  Revision 1.2  2008/10/27 18:32:09  proctor
  Fixed the handling of link params between MOAST and Go Motion

  Revision 1.1  2008/10/16 19:09:34  proctor
  Initial checkin of files that wrap Go Motion with MOAST NML

*/
