/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file matildaclient.c

  \brief Client test program for sending socket commands to Matilda server.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.1 $
  $Date: 2008/03/07 17:45:41 $
  \endcode

  \author Fred Proctor
*/

#include <stdio.h>		/* fprintf(), stderr, perror() */
#include <stddef.h>		/* sizeof(), ptrdiff_t */
#include <stdlib.h>		/* malloc */
#include <string.h>		/* memset */
#include <unistd.h>		/* read(), select(), STDOUT_FILENO */
#include <pthread.h>		/* pthread_xxx */
#include "sockutils.h"		/* get_server_fd */
#include "matilda.h"		/* USARSIM_MSG_DELIMITER */

typedef struct {
  int server_fd;
} read_thread_arg;

static void * read_thread_code(void * arg)
{
  int server_fd;
  int nchars;
  enum {BUFFERLEN = USARSIM_MSG_LEN};
  char buffer[BUFFERLEN];
  char * buffer_ptr;
  char * buffer_end;
  int BUILDLEN = BUFFERLEN;
  char * build;
  char * build_ptr;
  char * build_end;
  ptrdiff_t offset;

  server_fd = ((read_thread_arg *) arg)->server_fd;
  free(arg);

  build = (char *) malloc(BUILDLEN * sizeof(char));
  build_ptr = build;
  build_end = build + BUILDLEN;

  for (;;) {
    nchars = read(server_fd, buffer, sizeof(buffer));
    if (nchars > 0) {
      buffer_ptr = buffer;
      buffer_end = buffer + nchars;
      while (buffer_ptr != buffer_end) {
	if (build_ptr == build_end) {
	  offset = build_ptr - build;
	  BUILDLEN *= 2;
	  build = (char *) realloc(build, BUILDLEN * sizeof(char));
	  build_ptr = build + offset;
	  build_end = build + BUILDLEN;
	}
	*build_ptr++ = *buffer_ptr;
	if (*buffer_ptr++ == USARSIM_MSG_DELIMITER) {
	  offset = build_ptr - build;
	  build_ptr = build;
	  write(STDOUT_FILENO, build, offset);
	}
      }
    } else {
      break;
    }
  }

  free(build);
  close(server_fd);
  printf("read thread done\n");

  return NULL;
}

/*
  Syntax: matildaclient <TCP port number> <host name>
*/

int main(int argc, char * argv[])
{
  pthread_t thread;
  pthread_attr_t attr;
  char * hostname;
  read_thread_arg * read_thread_arg_ptr;
  int port;
  int server_fd;
  enum {BUFFERLEN = USARSIM_MSG_LEN};
  char buffer[BUFFERLEN];

  if (argc != 3) {
    fprintf(stderr, "need <TCP port number> <host name>\n");
    return 1;
  }
  if (1 != sscanf(argv[1], "%i", &port)) {
    fprintf(stderr, "bad port number: %s\n", argv[1]);
    return 1;
  }
  hostname = argv[2];

  if (-1 == (server_fd = get_server_fd(port, hostname))) {
    return 1;
  }

  /* spawn read thread */
  read_thread_arg_ptr = malloc(sizeof(*read_thread_arg_ptr));
  if (NULL == read_thread_arg_ptr) {
    fprintf(stderr, "can't allocate args for read thread\n");
    return 1;
  }
  read_thread_arg_ptr->server_fd = server_fd;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  if (0 != pthread_create(&thread, &attr, read_thread_code, (void *) read_thread_arg_ptr)) {
    fprintf(stderr, "can't spawn read thread");
    return 1;
  }

  /* ^D preferably or ^C if you want will quit */
  for (;;) {
    if (NULL == fgets(buffer, BUFFERLEN, stdin)) break;
    write(server_fd, buffer, strlen(buffer));
  }

  close(server_fd);
  printf("client done\n");

  return 0;
}

/*
  Modification history:

  $Log: matildaclient.c,v $
  Revision 1.1  2008/03/07 17:45:41  proctor
  Added matilda code in servo, and socket, tty utils in tools

*/

