/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file matildaserver.c

  \brief Socket server interface to serial Smart Motors on MATILDA robot

  Principles of operation:

  This program serves a socket for USARsim messages, acting just like
  a real USARsim on behalf of a MATILDA robot running serial Smart
  Motors.  Rather than connecting to USARsim, MOAST connects to this
  application and can then drive the MATILDA around.

  One thread for each motor, the "motor thread", is responsible for
  writing commands and reading replies between the SM on the serial port.

  One thread for each socket connection, the "message thread", is
  responsible for parsing USARsim messages and sending replies. 

  The motor- and message threads share a global data structure. The
  motor threads continually update it with motor position, and consult
  it for the commanded motor speeds. The message threads write new
  motor speeds as needed, and read out motor position as
  requested. The global data structure is mutexed.

  The Smart Motor (SM) interface uses the following commands:

  "MV" sent out once to the SM to put it in velocity mode;

  "A<accel> sent out once to set up some crisp acceleration; and

  "V<vel>" sent out continually to set the motor speed in response
  to USARsim messages.

  "RP" is sent out continually to request position, and the SM responds
  with an integer number of counts.

  The main thread runs motor threads for each serial port, then creates a socket server port and listens for
  connections. For each connection, it spawns a message thread and
  returns to await more socket connections from clients. Only one
  socket connection is expected, but others could happen.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.3 $
  $Date: 2008/03/17 18:36:20 $
  \endcode

  \author Fred Proctor & Stephen Balakirsky
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		/* f,printf, stderr, perror */
#include <stddef.h>		/* sizeof, ptrdiff_t */
#include <stdlib.h>		/* malloc, free, atoi */
#include <string.h>		/* memset, strerror */
#include <stdarg.h>		/* va_list */
#include <math.h>		/* M_PI */
#include <errno.h>		/* errno */
#include <unistd.h>		/* read, write */
#include <time.h>               /* struct timespec, nanosleep */
#include <termios.h>		/* struct termios */
#include <fcntl.h>		/* open, O_RDWR */
#include <sys/time.h>           /* gettimeofday, struct timeval */
#include <pthread.h>		/* pthread_attr_int, ... */
#include "sockutils.h"		/* get_socket,client_fd */
#include "ttycfg.h"		/* tty_raw */
#include "matilda.h"		/* USARSIM,SMART_MOTOR_MSG_LEN */
#include "inifile.h"

/*
  'dbprintf' is debug printf. Set 'debug' and you get some diagnostics
  on stderr that show what's going on. Fatal errors are printed
  regardless.
*/
static int debug = 1;
static void dbprintf(const char * fmt, ...)
{
  va_list ap;

  if (debug) {
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
  }
}

/* get the time in seconds */
static double etime(void)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);

  return ((double) tv.tv_sec) + ((double) tv.tv_usec) * 1.0e-6;
}

static void esleep(double secs)
{
  struct timespec ts;

  ts.tv_sec = (int) secs;
  ts.tv_nsec = (int) ((secs - ts.tv_sec) * 1.0e9);

  (void) nanosleep(&ts, NULL);

  return;
}

#define ROUND(x) (x) < 0.0 ? (int) ((x) - 0.5) : (int) ((x) + 0.5)

typedef struct {
  /*
    'counts_per_rad' is how many motor counts per radian.
    'rads_per_count' is its reciprocal. We have both so that we avoid
    division in both conversion directions.
  */
  double counts_per_rad;
  double rads_per_count;

  /* 'position' is updated by the motor threads as fast as practical,
     and read by the message threads when responding to position
     queries. Units are radians. */
  double position;

  /* 'speed' is the commanded speed of the motor. It is written by the
     message thread when it receives a command, and read by the motor
     thread as fast as practical for sending to the motors. Units are
     in rads/sec. */
  double speed;

  /* 'mutex' is locked/unlocked by the motor- and message threads
     when accessing the data structure. */
  pthread_mutex_t mutex;
} smart_motor_struct;

/* the global Smart Motor data structure */
static smart_motor_struct smart_motor_data[SMART_MOTOR_NUM];

typedef struct {
  /* 'serial_fd' is the file descriptor to the serial port. It is
     opened by the main thread before the motor thread runs. */
  int serial_fd;
  /* 'index' is which index into the smart_motor_data[] array the
     motor thread is working on. */
  int index;
} motor_thread_arg;

static void * motor_thread_code(void * arg)
{
  double delta_t;
  int serial_fd;
  int index;
  int ival;
  int nchars;
  enum {BUFFERLEN = SMART_MOTOR_MSG_LEN};
  char buffer[BUFFERLEN];

#define LOCK pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK pthread_mutex_unlock(&smart_motor_data[index].mutex)

  /* copy out our heap args and free them */
  serial_fd = ((motor_thread_arg *) arg)->serial_fd;
  index = ((motor_thread_arg *) arg)->index;
  free(arg);

  dbprintf("running motor thread %d on serial fd %d\n", index, serial_fd);

  /* clamp index so we won't seg fault if it's bad */
  if (index < 0) index = 0;
  else if (index >= SMART_MOTOR_NUM) index = SMART_MOTOR_NUM - 1;

  /* set up our time period if we're simulating */
  delta_t = 0.1;

  /* set up our motor for velocity mode with nice accel */
  LOCK;
  ival = ROUND(SMART_MOTOR_ACC * ACCEL_ADJUST * smart_motor_data[index].counts_per_rad);
  UNLOCK;
  sprintf(buffer, "MV A=%d G\r", ival);
  write(serial_fd, buffer, strlen(buffer));

  for (;;) {
    if (serial_fd >= 0) {
      LOCK;
      ival = ROUND(smart_motor_data[index].speed * SPEED_ADJUST * smart_motor_data[index].counts_per_rad);
      UNLOCK;
      sprintf(buffer, "V=%d RP G\r", ival);
      write(serial_fd, buffer, strlen(buffer));
      nchars = read(serial_fd, buffer, sizeof(buffer));
      if (nchars > 0) {
	LOCK;
	smart_motor_data[index].position = atoi(buffer) * smart_motor_data[index].rads_per_count; /* could do a better parsing job than 'atoi' */
	UNLOCK;
      }
    } else {
      /* clock our simulated position by vee delta tee */
      LOCK;
      smart_motor_data[index].position += smart_motor_data[index].speed * delta_t;
      UNLOCK;
      esleep(delta_t);
    }
  } /* for (;;) */

  /* if we ever break out of the loop above, we'll get here */
  close(serial_fd);
  dbprintf("read thread %d done\n", index);

  return NULL;

#undef LOCK
#undef UNLOCK
}

/*
  'handle_usarsim_message' takes a null-terminated message 'msg',
  parses it and handles it by acting on the global 'smart_motor_data'
  array, either writing speeds into it or reading position out of it
  and sending out the socket.
*/
static int handle_usarsim_message(char * msg, int client_fd)
{
  double thleft;
  double thright;
  double wleft;
  double wright;
  int retval;
  enum {BUFFERLEN = USARSIM_MSG_LEN};
  char buffer[BUFFERLEN];	/* messages back to client go here */

#define LOCK(index) pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK(index) pthread_mutex_unlock(&smart_motor_data[index].mutex)

  retval = 0;		      /* default is OK, set to -1 if needed */

  /*
    Parse out the USARsim message and handle it. Some messages
    will require setting motor speeds by writing into the global
    data structure. Some will require querying position from the
    global data structure.
  */

  /*
    Poor man's messages:

    "V <wleft> <wright>" for setting wheel speeds
    "P" for querying wheel positions
    "LEFT=<left> RIGHT=<right>" for reporting wheel positions
  */

  printf( "received message: %s\n", msg );
  if (msg[0] == 'V') {
    /*
      You just parsed a speed command and got some values for the
      left and right wheel angular speeds.  Now you stick them in
      the global data.
    */
    if (2 == sscanf(msg, "%*s %lf %lf", &wleft, &wright)) {
      LOCK(LEFT_MOTOR);
      smart_motor_data[LEFT_MOTOR].speed = wleft;
      UNLOCK(LEFT_MOTOR);
      LOCK(RIGHT_MOTOR);
      smart_motor_data[RIGHT_MOTOR].speed = wright;
      UNLOCK(RIGHT_MOTOR);
    } else {
      dbprintf("bad speed message: %s\n", msg);
      retval = -1;
    }
  } else if (msg[0] == 'P') {
    /*
      You just parsed out a position request. Now you get them from
      global data and write them to the client.
    */
    LOCK(LEFT_MOTOR);
    thleft = smart_motor_data[LEFT_MOTOR].position;
    UNLOCK(LEFT_MOTOR);
    LOCK(RIGHT_MOTOR);
    thright = smart_motor_data[RIGHT_MOTOR].position;
    UNLOCK(RIGHT_MOTOR);
    sprintf(buffer, "LEFT=%f RIGHT=%f\n", thleft, thright);
    write(client_fd, buffer, strlen(buffer));
    dbprintf(buffer);
  } else {
    retval = -1;
    dbprintf("unknown message: %s\n", msg);
  }

  return retval;

#undef LOCK
#undef UNLOCK
}

typedef struct {
  /* 'client_fd' is the file descriptor to the client socket. It is
     opened by the main thread before the message thread runs. */
  int client_fd;
} message_thread_arg;

static void * message_thread_code(void * arg)
{
  int client_fd;
  int nchars;
  enum {BUFFERLEN = USARSIM_MSG_LEN};
  char buffer[BUFFERLEN];
  char * buffer_ptr;
  char * buffer_end;
  int BUILDLEN = BUFFERLEN;
  char * build;
  char * build_ptr;
  char * build_end;
  ptrdiff_t offset;

#define LOCK(index) pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK(index) pthread_mutex_unlock(&smart_motor_data[index].mutex)

  /* copy out our heap args and free them */
  client_fd = ((message_thread_arg *) arg)->client_fd;
  free(arg);

  build = (char *) malloc(BUILDLEN * sizeof(char));
  build_ptr = build;
  build_end = build + BUILDLEN;

  for (;;) {
    nchars = read(client_fd, buffer, sizeof(buffer));
    if (nchars > 0) {
      /* compose a valid message, since we may have read a fragment */
      buffer_ptr = buffer;
      buffer_end = buffer + nchars;
      while (buffer_ptr != buffer_end) {
	if (build_ptr == build_end) {
	  /* grow build buffer to hold more data */
	  offset = build_ptr - build;
	  BUILDLEN *= 2;
	  build = (char *) realloc(build, BUILDLEN * sizeof(char));
	  build_ptr = build + offset;
	  build_end = build + BUILDLEN;
	}
	*build_ptr++ = *buffer_ptr;
	if (*buffer_ptr++ == USARSIM_MSG_DELIMITER) {
	  offset = build_ptr - build;
	  build[offset] = 0;
	  build_ptr = build;
	  /* now a full message is in 'build', with length 'offset',
	     including the delimiter kept and null-terminated for your
	     pleasure */
	  dbprintf("got ``%s''\n", build);
	  (void) handle_usarsim_message(build, client_fd);
	} /* if (*buffer_ptr++ == USARSIM_MSG_DELIMITER */
      } /* while (buffer_ptr != buffer_end) */
    } else {
      break;
    }
  } /* for (;;) */

  /* if we ever break out of the loop above, we'll get here */
  free(build);
  close(client_fd);
  dbprintf("message thread %d done\n", client_fd);

  return NULL;

#undef LOCK
#undef UNLOCK
}

/* send an init message (by USARSim format) to the client
   return 1 on success, 0 on failure
 */
int sendInit( int client_fd,   FILE *inifile )
{
  char iniString[255];
  const char *inifileText;

  inifileText = iniFind( inifile, "LEVEL", "USARSIM" );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key LEVEL in section USARSIM of inifile\n");
      return 0;
    }
  
  sprintf( iniString, "NFO {Gametype RealRobot}{Level %s}{TimeLimit 0};\r\n",
	   inifileText );

  write(client_fd, iniString, strlen(iniString));
  dbprintf( "Wrote %s to client\n", iniString );
  return 1;
}

/*
  Syntax: matildaserver <TCP port number> {<serial ports ...>}
*/
int main(int argc, char * argv[])
{
  pthread_t thread;
  pthread_attr_t attr;
  struct termios save_termios;
  motor_thread_arg * motor_thread_arg_ptr;
  message_thread_arg * message_thread_arg_ptr;
  char * sports[SMART_MOTOR_NUM];
  int port;
  int index;
  int socket_fd;
  int client_fd;
  FILE *inifile;
  char *inifileName = "../etc/robots.cfg";
  const char *inifileText;

  if (argc == 2) 
    {
      dbprintf( "%s using %s as ini file\n",
		argv[0], argv[1]);
      inifileName = argv[1];
    }

  if( (inifile = fopen(inifileName, "r")) == 0 )
    {
      fprintf(stderr, "%s: Unable to open configuration file %s\n",
	      argv[0], inifileName );
      return 1;
    }

  inifileText = iniFind( inifile, "PORT_NUMBER", "USARSIM" );
  if( inifileText == NULL )
    {
      fprintf( stderr, "%s unable to find key PORT_NUMBER in section USARSIM of inifile %s\n",
	       argv[0], inifileText );
    }
  port = atoi(inifileText);
#define BEFOREARGS 2		/* how many args before serial port args */
  for (index = 0; index < SMART_MOTOR_NUM; index++) {
    if (argc <= index + BEFOREARGS) sports[index] = NULL;
    else sports[index] = argv[index + BEFOREARGS];
  }

  /* get our socket server file descriptor */
  socket_fd = get_socket_fd(port);
  if (socket_fd < 0) {
    fprintf(stderr, "can't serve socket port %d\n", port);
    return 1;
  }

  /* initialize the Smart Motor data and start the read threads */
  for (index = 0; index < SMART_MOTOR_NUM; index++) {
    pthread_mutex_init(&smart_motor_data[index].mutex, NULL);
    /* insert actual values here, or inifile this */
    smart_motor_data[index].counts_per_rad = COUNTS_PER_RAD;
    smart_motor_data[index].rads_per_count = 1.0 / smart_motor_data[index].counts_per_rad;
    /* in case we're simulating, start somewhere interesting */
    smart_motor_data[index].position = index * M_PI;
    /* but be nicely stopped */
    smart_motor_data[index].speed = 0.0;
    motor_thread_arg_ptr = (motor_thread_arg*)malloc(sizeof(*motor_thread_arg_ptr));
    if (NULL == motor_thread_arg_ptr) {
      fprintf(stderr, "can't allocate args for read thread %d\n", index);
      return 1;
    }
    if (sports[index] == NULL) {
      motor_thread_arg_ptr->serial_fd = -1;
    } else {
      motor_thread_arg_ptr->serial_fd = open(sports[index], O_RDWR);
      if (motor_thread_arg_ptr->serial_fd < 0) {
	fprintf(stderr, "can't open serial port %s\n", sports[index]);
	return 1;
      }
      tty_raw(motor_thread_arg_ptr->serial_fd, SMART_MOTOR_BAUD, &save_termios);
    }
    motor_thread_arg_ptr->index = index;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if (0 != pthread_create(&thread, &attr, motor_thread_code, (void *) motor_thread_arg_ptr)) {
      fprintf(stderr, "can't spawn motor thread %d\n", index);
      return 1;
    }
  }

  for (;;) {
    dbprintf("waiting for client connection...\n");
    client_fd = get_client_fd(socket_fd);
    if (client_fd < 0) {
      break;
    }
    dbprintf("got one on fd %d\n", client_fd);

    if( !sendInit( client_fd, inifile ) )
      {
	fprintf(stderr, "%s: unable to send init string to client\n", argv[0] );
      }
    message_thread_arg_ptr = (message_thread_arg*)malloc(sizeof(*message_thread_arg_ptr));
    if (NULL == message_thread_arg_ptr) {
      fprintf(stderr, "can't allocate args for message thread %d\n", index);
      return 1;
    }
    message_thread_arg_ptr->client_fd = client_fd;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if (0 != pthread_create(&thread, &attr, message_thread_code, (void *) message_thread_arg_ptr)) {
      fprintf(stderr, "can't spawn message thread %d\n", client_fd);
      return 1;
    }
  } /* while (! done) */

  dbprintf("server done\n");

  return 0;
}

/*
  Modification history:

  $Log: matildaserver.c,v $
  Revision 1.3  2008/03/17 18:36:20  dr_steveb
  Created server that will interface the matilda platform with moast thru simWare

  Revision 1.2  2008/03/07 19:22:47  proctor
  Fixed bug in debug print statement

  Revision 1.1  2008/03/07 17:45:42  proctor
  Added matilda code in servo, and socket, tty utils in tools

*/

