/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file servoGui.cc

  \brief Gtk+ GUI for servo controller

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.3 $
  $Date: 2006/12/12 20:22:58 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <math.h>		// sqrt()
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <gtk/gtk.h>
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "navDataExt.hh"	// navDataExt_format()
#include "servoMobJA.hh"		// servoMobJA_format()

#define WINDOW_WIDTH 400
#define WINDOW_HEIGHT 400

#define DRAWING_WIDTH 1500
#define DRAWING_HEIGHT 1500

static NML *navDataExtBuf = NULL;
static RCS_CMD_CHANNEL *servoMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *servoMobJAStatBuf = NULL;
static RCS_STAT_CHANNEL *servoMobJASetBuf = NULL;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "servoShell"

// these are the first x,y positions reported, so we can center our vehicle
static double xOff = 0.0;
static double yOff = 0.0;

#define rcs_status_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

#define SQ(x) ((x)*(x))
#define DEG(x) (57.295779513082323 * (x))

static int cleanupServoMobJANmlBuffers(void)
{
  if (NULL != navDataExtBuf)
    delete navDataExtBuf;
  if (NULL != servoMobJACmdBuf)
    delete servoMobJACmdBuf;
  if (NULL != servoMobJAStatBuf)
    delete servoMobJAStatBuf;
  if (NULL != servoMobJASetBuf)
    delete servoMobJASetBuf;

  navDataExtBuf = NULL;
  servoMobJAStatBuf = NULL;
  servoMobJASetBuf = NULL;
  servoMobJACmdBuf = NULL;

  return 0;
}

static int initServoMobJANmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupServoMobJANmlBuffers();

  sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
  navDataExtBuf =
    new NML(navDataExt_format, chanName, THIS_PROCESS,
			 config_file);
  if (!navDataExtBuf->valid()) {
    delete navDataExtBuf;
    navDataExtBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, bufnum);
  servoMobJACmdBuf =
    new RCS_CMD_CHANNEL(servoMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMobJACmdBuf->valid()) {
    delete servoMobJACmdBuf;
    servoMobJACmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, bufnum);
  servoMobJAStatBuf =
    new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMobJAStatBuf->valid()) {
    delete servoMobJAStatBuf;
    servoMobJAStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_SET_NAME, bufnum);
  servoMobJASetBuf =
    new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMobJASetBuf->valid()) {
    delete servoMobJASetBuf;
    servoMobJASetBuf = 0;
    return 1;
  }

  return 0;
}

/*
  The timeout callback will update the labels, so they need to be global.
  Only the labels need to be global, but we'll keep them all together
  so we can see what's related to what.
*/

static GtkWidget *window;
static GtkWidget *hbox;

static GtkWidget *vbox1;
static GtkWidget *frame_rel;
static GtkWidget *vbox_rel;
static GtkWidget *label_xrel;
static GtkWidget *label_yrel;
static GtkWidget *label_wrel;
static GtkWidget *frame_abs;
static GtkWidget *vbox_abs;
static GtkWidget *label_xabs;
static GtkWidget *label_yabs;
static GtkWidget *label_wabs;
static GtkWidget *frame_true;
static GtkWidget *vbox_true;
static GtkWidget *label_xtrue;
static GtkWidget *label_ytrue;
static GtkWidget *label_wtrue;

static GtkWidget *vbox2;
static GtkWidget *frame_plot;
static GtkWidget *scrolled_window;
static GtkWidget *drawing_area;

static GdkPixmap *pixmap = NULL;

static GtkWidget *button;

static gboolean expose_event(GtkWidget * widget, GdkEventExpose * event)
{
  gdk_draw_drawable(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		    pixmap,
		    event->area.x, event->area.y,
		    event->area.x, event->area.y,
		    event->area.width, event->area.height);

  return FALSE;
}

#define min(a,b) ((a) < (b) ? (a) : (b))

static gboolean configure_event(GtkWidget * widget, GdkEventConfigure * event)
{
  GdkPixmap *old;
  gint old_width;
  gint old_height;

  old = pixmap;

  pixmap = gdk_pixmap_new(widget->window,
			  widget->allocation.width,
			  widget->allocation.height, -1);
  gdk_draw_rectangle(pixmap,
		     widget->style->white_gc,
		     TRUE,
		     0, 0,
		     widget->allocation.width, widget->allocation.height);

  if (NULL != old) {
    gdk_drawable_get_size(old, &old_width, &old_height);
    gdk_draw_drawable(pixmap, widget->style->white_gc, old,	/* src */
		      0, 0,	/* xsrc, ysrc */
		      0, 0,	/* xdst, ydst */
		      min(old_width, widget->allocation.width),
		      min(old_height, widget->allocation.height));
    g_object_unref(old);
  }

  return TRUE;
}

static void draw_point(GtkWidget * widget, gint x, gint y)
{
  GdkRectangle update_rect;

  update_rect.x = x - 1;
  update_rect.y = y - 1;
  update_rect.width = 2;
  update_rect.height = 2;
  gdk_draw_rectangle(pixmap,
		     widget->style->black_gc,
		     TRUE,
		     update_rect.x, update_rect.y,
		     update_rect.width, update_rect.height);
  gtk_widget_queue_draw_area(widget,
			     update_rect.x, update_rect.y,
			     update_rect.width, update_rect.height);
}

static void clear_rectangle(void)
{
  gdk_draw_rectangle(pixmap,
		     drawing_area->style->white_gc,
		     TRUE,
		     0, 0,
		     drawing_area->allocation.width, 
		     drawing_area->allocation.height);

  gtk_widget_queue_draw_area(drawing_area,
			     0, 0,
			     drawing_area->allocation.width,
			     drawing_area->allocation.height);
}

#define DBL_MAX_DIGITS 512	// larger than the typical 309 DBL_MAX digits
static char label_string[DBL_MAX_DIGITS];

static gint timeout_callback(gpointer data)
{
  NavDataExt * nav = (NavDataExt *) navDataExtBuf->get_address();
  PM_ROTATION_VECTOR rotvec;

  navDataExtBuf->read();

  sprintf(label_string, "%.3f", nav->tranRel.x);
  gtk_label_set_text(GTK_LABEL(label_xrel), label_string);

  sprintf(label_string, "%.3f", nav->tranRel.y);
  gtk_label_set_text(GTK_LABEL(label_yrel), label_string);

  sprintf(label_string, "%.3f", DEG(nav->rpyRel.y));
  gtk_label_set_text(GTK_LABEL(label_wrel), label_string);

  sprintf(label_string, "%.3f", nav->tranAbs.x);
  gtk_label_set_text(GTK_LABEL(label_xabs), label_string);

  sprintf(label_string, "%.3f", nav->tranAbs.y);
  gtk_label_set_text(GTK_LABEL(label_yabs), label_string);

  sprintf(label_string, "%.3f", DEG(nav->rpyAbs.y));
  gtk_label_set_text(GTK_LABEL(label_wabs), label_string);

  sprintf(label_string, "%.3f", nav->tranTrue.x);
  gtk_label_set_text(GTK_LABEL(label_xtrue), label_string);

  sprintf(label_string, "%.3f", nav->tranTrue.y);
  gtk_label_set_text(GTK_LABEL(label_ytrue), label_string);

  sprintf(label_string, "%.3f", DEG(nav->rpyTrue.y));
  gtk_label_set_text(GTK_LABEL(label_wtrue), label_string);

  draw_point(drawing_area,
	     (DRAWING_WIDTH / 2) + (gint) (10.0 * (nav->tranRel.x - xOff)),
	     (DRAWING_HEIGHT / 2) + (gint) (10.0 * (nav->tranRel.y - yOff)));

  return TRUE;			/* FALSE means stop timer */
}

/* Another callback */
static void destroy_callback(GtkWidget * widget, gpointer data)
{
  (void) cleanupServoMobJANmlBuffers();

  gtk_main_quit();
}

#define BORDER 5
#define SPACING 5
#define PADDING 5

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
  int retval;

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      rcs_print("%s: using specified NML file %s\n", argv[0], NML_FILE);
      break;
    case 'i':
      INI_FILE = optarg;
      rcs_print("%s: using specified INI file %s\n", argv[0], INI_FILE);
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      rcs_print("%s: using specified buffer suffix %d\n", argv[0], bufnum);
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // initial allocate all the buffers
  retval = initServoMobJANmlBuffers(NML_FILE, bufnum);

  if (retval) {
    fprintf(stderr, "%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }

  navDataExtBuf->read();
  xOff = ((NavDataExt *) navDataExtBuf->get_address())->tranRel.x;
  yOff = ((NavDataExt *) navDataExtBuf->get_address())->tranRel.y;

  gtk_init(&argc, &argv);

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER(window), BORDER);

  hbox = gtk_hbox_new(FALSE,	/* homogeneous child width */
		      SPACING);	/* spacing between objects */
  gtk_container_add(GTK_CONTAINER(window),	/* parent */
		    hbox);	/* child */

  vbox1 = gtk_vbox_new(FALSE,	/* homogeneous child height */
		       SPACING);	/* spacing */
  gtk_box_pack_start(GTK_BOX(hbox),	/* parent */
		     vbox1,	/* child */
		     FALSE,	/* expands */
		     FALSE,	/* widgets fill */
		     PADDING);	/* padding on each side of object */

  frame_rel = gtk_frame_new("Relative");
  gtk_box_pack_start(GTK_BOX(vbox1), frame_rel, FALSE, FALSE, PADDING);
  vbox_rel = gtk_vbox_new(FALSE, SPACING);
  gtk_container_add(GTK_CONTAINER(frame_rel), vbox_rel);
  label_xrel = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_rel), label_xrel, FALSE, FALSE, PADDING);
  label_yrel = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_rel), label_yrel, FALSE, FALSE, PADDING);
  label_wrel = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_rel), label_wrel, FALSE, FALSE, PADDING);

  frame_abs = gtk_frame_new("Absolute");
  gtk_box_pack_start(GTK_BOX(vbox1), frame_abs, FALSE, FALSE, PADDING);
  vbox_abs = gtk_vbox_new(FALSE, SPACING);
  gtk_container_add(GTK_CONTAINER(frame_abs), vbox_abs);
  label_xabs = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_abs), label_xabs, FALSE, FALSE, PADDING);
  label_yabs = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_abs), label_yabs, FALSE, FALSE, PADDING);
  label_wabs = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_abs), label_wabs, FALSE, FALSE, PADDING);

  frame_true = gtk_frame_new("Truth");
  gtk_box_pack_start(GTK_BOX(vbox1), frame_true, FALSE, FALSE, PADDING);
  vbox_true = gtk_vbox_new(FALSE, SPACING);
  gtk_container_add(GTK_CONTAINER(frame_true), vbox_true);
  label_xtrue = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_true), label_xtrue, FALSE, FALSE, PADDING);
  label_ytrue = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_true), label_ytrue, FALSE, FALSE, PADDING);
  label_wtrue = gtk_label_new("blank");
  gtk_box_pack_start(GTK_BOX(vbox_true), label_wtrue, FALSE, FALSE, PADDING);

  button = gtk_button_new_with_label ("Clear");
  g_signal_connect (G_OBJECT(button), "clicked",
		    G_CALLBACK(clear_rectangle), NULL);
  gtk_box_pack_start(GTK_BOX(vbox1), button, FALSE, FALSE, PADDING);

  vbox2 = gtk_vbox_new(TRUE, SPACING);
  gtk_box_pack_start(GTK_BOX(hbox), vbox2, TRUE, TRUE, PADDING);

  frame_plot = gtk_frame_new("Plot");
  gtk_box_pack_start(GTK_BOX(vbox2), frame_plot, TRUE, TRUE, PADDING);

  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_set_size_request(GTK_WIDGET(scrolled_window), 
			      WINDOW_WIDTH,
			      WINDOW_HEIGHT);
  gtk_container_set_border_width(GTK_CONTAINER(scrolled_window), 10);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), 
				 GTK_POLICY_ALWAYS,	// horizontal
				 GTK_POLICY_ALWAYS);	// vertical
  gtk_container_add(GTK_CONTAINER(frame_plot), scrolled_window);

  drawing_area = gtk_drawing_area_new();
  gtk_widget_set_size_request(GTK_WIDGET(drawing_area),
			      DRAWING_WIDTH,
			      DRAWING_HEIGHT);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),
					drawing_area);

  gtk_widget_show_all(window);

  pixmap = gdk_pixmap_new(drawing_area->window,
			  drawing_area->allocation.width,
			  drawing_area->allocation.height, -1);

  gdk_draw_rectangle(pixmap,
		     drawing_area->style->white_gc,
		     TRUE,
		     0, 0,
		     drawing_area->allocation.width,
		     drawing_area->allocation.height);

  /* put the scrollbars in the middle */
  GtkAdjustment * adj = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  adj->value = 0.5 * (adj->lower + adj->upper - adj->page_size);
  gtk_scrolled_window_set_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window), adj);
  gtk_adjustment_value_changed(adj);
  adj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  adj->value = 0.5 * (adj->lower + adj->upper - adj->page_size);
  gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window), adj);
  gtk_adjustment_value_changed(adj);

  /* start the timer */
  (void) g_timeout_add(100,	/* millisecs */
		       timeout_callback, 0);

  g_signal_connect(G_OBJECT(drawing_area), "expose_event",
		   G_CALLBACK(expose_event), NULL);

  g_signal_connect(G_OBJECT(drawing_area), "configure_event",
		   G_CALLBACK(configure_event), NULL);

  g_signal_connect(G_OBJECT(window), "delete_event",
		   G_CALLBACK(destroy_callback), NULL);

  gtk_main();

  return 0;
}

/*
  Modification history:

  $Log: servoGui.cc,v $
  Revision 1.3  2006/12/12 20:22:58  proctor
  Took out bumpers from servo gui-- no longer in status

  Revision 1.2  2006/10/03 15:30:20  proctor
  Changed sign on y when calling draw_point, so that XY correspond to North-East

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.9  2005/07/26 21:15:08  proctor
  Set up initial xOff, yOff offsets

  Revision 1.8  2005/07/25 22:40:47  proctor
  Plotted relative coords, not true

  Revision 1.7  2005/07/19 22:29:10  proctor
  Added a clear button

  Revision 1.6  2005/07/18 21:16:49  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.5  2005/07/12 21:43:13  proctor
  Drew tranTrue values

  Revision 1.4  2005/06/09 16:43:37  stephen
  Made extents larger.

  Revision 1.3  2005/06/07 21:45:45  proctor
  Changed servoPL to servoMobJA

  Revision 1.2  2005/06/06 20:31:58  proctor
  Made drawing area bigger, and put scrollbars in the middle on startup.

  Revision 1.1  2005/05/18 21:12:04  proctor
  Initial checkin

*/
