/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file servoShell.cc

  \brief Command line shell to vehicle servo level NML buffers.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.43 $
  $Date: 2009/04/23 19:59:53 $
  \endcode

  \author Chris Scrapper
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include <string>
#include <iostream>
#include <vector>
#include <sstream>
using namespace std;

#include <rcs.hh>
#include "ulapi.h"		// ulapi_XXX
#include "servoMobJA.hh"	// servoMobJA_format()
#include "servoMisJA.hh"	// servoMisJA_format()
#include "servoSP.hh"		// servoSP_format()
#include "servoEffJA.hh"	// servoEffJA_format()
#include "joystickData.hh"
//#include "swKBUtils.hh"		// parsing utilities

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"
#define PROCESS_NAME          "servoShell"



class ServoNodeNmlBuffers
{
public:
  RCS_CMD_CHANNEL  *cmdBuf;
  RCS_STAT_CHANNEL *statBuf;
  RCS_CMD_CHANNEL  *cfgBuf;
  RCS_STAT_CHANNEL *setBuf;
  RCS_CMD_MSG      *cmdMsg;
  RCS_STAT_MSG_V2  *statMsg;
  RCS_STAT_MSG     *setMsg;
  ServoNodeNmlBuffers();
  bool openNML(string nmlCfgFile, 
	       string cmdName, 
	       string statName, 
	       string cfgName,
	       string setName,
	       NML_FORMAT_PTR fPtr,
	       int botNum,
	       int compID=0);
  void cleanUpNMLBuf();
};

ServoNodeNmlBuffers * mob = NULL;
ServoNodeNmlBuffers * sp = NULL;
ServoNodeNmlBuffers * eff = NULL;
vector<ServoNodeNmlBuffers> mis;

int mob_serial_number=0;
ServoMobJACmdInit mobCmdInit;
ServoMobJACmdHalt mobCmdHalt;
ServoMobJACmdAbort mobCmdAbort;
ServoMobJACmdShutdown mobCmdShutdown;
ServoMobJACmdAckerman mobCmdAckerman;
ServoMobJACmdOmni mobCmdOmni;
ServoMobJACmdSkid mobCmdSkid;
ServoMobJACmdFlipSkid mobCmdFlipSkid;
ServoMobJACmdSubmarine mobCmdSubmarine;
ServoMobJACmdRotaryWing mobCmdRotaryWing;

int sp_serial_number=0;
ServoSPCmdInit spCmdInit;
ServoSPCmdHalt spCmdHalt;
ServoSPCmdAbort spCmdAbort;
ServoSPCmdShutdown spCmdShutdown;
ServoSPCmdGo spCmdGo;


int eff_serial_number=0;
ServoEffJACmdInit effCmdInit;
ServoEffJACmdHalt effCmdHalt;
ServoEffJACmdAbort effCmdAbort;
ServoEffJACmdShutdown effCmdShutdown;
ServoEffJACmdOpcode effCmdOpcode;


int mis_serial_number=0;
ServoMisJACmdInit misCmdInit;
ServoMisJACmdHalt misCmdHalt;
ServoMisJACmdAbort misCmdAbort;
ServoMisJACmdShutdown misCmdShutdown;
ServoMisJACmdMove misCmdMove;

static NML * joystickDataBuf = NULL;
static JoystickData * joystickDataPtr = NULL;

void * scanThr = NULL;

#define rcs_admin_state_to_string(s) (s) == ADMIN_INITIALIZED ? "INITIALIZED" : (s) == ADMIN_UNINITIALIZED ? "UNINITIALIZED" : (s) == ADMIN_SHUT_DOWN ? "SHUTDOWN" : "?"
#define rcs_status_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

void Tokenize(const string& str,
	      vector<string>& tokens,
	      const string& delimiters = " ")
{
  // Skip delimiters at beginning.
  string::size_type lastPos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  string::size_type pos     = str.find_first_of(delimiters, lastPos);
  
  while (string::npos != pos || string::npos != lastPos)
    {
      // Found a token, add it to the vector.
      tokens.push_back(str.substr(lastPos, pos - lastPos));
      // Skip delimiters.  Note the "not_of"
      lastPos = str.find_first_not_of(delimiters, pos);
      // Find next "non-delimiter"
      pos = str.find_first_of(delimiters, lastPos);
    }
}

void dumpTokens (const vector < string > &toks)
{
  printf ("Dump Tokens : size : %d | cap : %d\n",
	  (int)toks.size (), (int)toks.capacity ());
  for (unsigned i = 0; i < toks.size (); i++)
    printf ("\t%s\n", toks[i].c_str ());
}

int nocase_cmp (const string & s1, const string & s2)
{
  string::const_iterator it1 = s1.begin ();
  string::const_iterator it2 = s2.begin ();

  //has the end of at least one of the strings been reached?
  while ((it1 != s1.end ()) && (it2 != s2.end ()))
    {
      if (::toupper (*it1) !=::toupper (*it2))	//letters differ?
	// return -1 to indicate 'smaller than', 1 otherwise
	return (::toupper (*it1) < ::toupper (*it2)) ? -1 : 1;
      //proceed to the next character in each string
      ++it1;
      ++it2;
    }
  size_t size1 = s1.size (), size2 = s2.size ();	// cache lengths
  //return -1,0 or 1 according to strings' lengths
  if (size1 == size2)
    return 0;
  return (size1 < size2) ? -1 : 1;
}
 


///////////////////////////////////////////////////////////////////////////////////
/////       MOBILITY                                                       ////////
///////////////////////////////////////////////////////////////////////////////////
void mobHelpView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** Mobilty Help Menu     *****\n");
  printf("=================================\n");
  printf("quit                         : exit\n");
  printf("? | help                     : print mobility help\n");
  printf("admin                        : switch to admin mode\n");
  printf("eff                          : switch to effector mode\n");
  printf("mob                          : switch to mobility mode\n");
  printf("mis                          : switch to mission package mode\n");
  printf("sp                           : switch to sensor processing mode\n");
  printf("init                         : initialize mobility\n");
  printf("abort                        : powerdown mobility\n");
  printf("halt                         : powerdown mobility\n");
  printf("shutdown                     : powerdown mobility\n");  
  printf("stat | <enter>               : view mobility status\n");
  printf("settings                     : view mobililty settings\n");
  printf("cmd                          : view mobililty command\n");
  printf("skid <left> <right>          : skid drive\n");
  printf("ack2 <vel> <sAng>            : ackerman drive\n");
  printf("ack3 <vel> <sAng> <cAng>     : ackerman drive\n");
  printf("omni <lin> <lat> <rot>       : omni drive\n");
  printf("sub <prop> <rud> <stern>     : submarine drive\n");
  printf("rot <vert><lin><lat><rot>    : rotary wing drive\n");
  printf("flipskid <lt><rt><flip fAng> : skid drive and flipper control\n");
  printf("\t flip = flipper type [fr, fl, rr, rl]\n");
  printf("\t fAng = flipper Angle\n");
}


void mobCmdView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("*****  Mobilty Cmd View     *****\n");
  printf("=================================\n");
  printf("Mobility Cmd Type            : %s\n",
  	 servoMobJA_symbol_lookup( mob->cmdMsg->type ) );
  printf("Mobility Cmd Serial Number   : %d\n", mob->cmdMsg->serial_number );
  
  switch( mob->cmdMsg->type )
    {
    
    case SERVO_MOB_JA_CMD_INIT_TYPE:
    case SERVO_MOB_JA_CMD_ABORT_TYPE:
    case SERVO_MOB_JA_CMD_HALT_TYPE:
    case SERVO_MOB_JA_CMD_SHUTDOWN_TYPE:
      printf("Mobility Admin Command\n");
      break;
      
    case SERVO_MOB_JA_CMD_ACKERMAN_TYPE:
      printf("Steering Angle               : %0.2f [rads]\n",((ServoMobJACmdAckerman *) mob->cmdMsg)->steerAngle);
      printf("Crab Angle                   : %0.2f [rads]\n",((ServoMobJACmdAckerman *) mob->cmdMsg)->crabAngle);
      printf("Velocity [rads/s]            : %0.2f [rads/s]\n",((ServoMobJACmdAckerman *) mob->cmdMsg)->velocity);
      break;
    case SERVO_MOB_JA_CMD_SKID_TYPE:
      printf("Left Wheel Velocity          : %0.2f [rads/s]\n",((ServoMobJACmdSkid *) mob->cmdMsg)->wLeft);
      printf("Right Wheel Velocity         : %0.2f [rads/s]\n",((ServoMobJACmdSkid *) mob->cmdMsg)->wRight);
      break;
    case SERVO_MOB_JA_CMD_OMNI_TYPE:
      printf("Linear Velocity              : %0.2f [rads/s]\n",((ServoMobJACmdOmni *) mob->cmdMsg)->vLinear);
      printf("Lateral Velocity             : %0.2f [rads/s]\n",((ServoMobJACmdOmni *) mob->cmdMsg)->vLateral);
      printf("Anglular Velocity            : %0.2f [rads/s]\n",((ServoMobJACmdOmni *) mob->cmdMsg)->rot);
      break;
    case SERVO_MOB_JA_CMD_SUBMARINE_TYPE:
      printf("Stern Plane                  : %0.2f [rads]\n",((ServoMobJACmdSubmarine *) mob->cmdMsg)->sternPlane);
      printf("Rudder                       : %0.2f [rads]\n",((ServoMobJACmdSubmarine *) mob->cmdMsg)->rudder);
      printf("Propeller                    : %0.2f [rads/s]\n",((ServoMobJACmdSubmarine *) mob->cmdMsg)->propeller);
      break;
    case SERVO_MOB_JA_CMD_ROTARY_WING_TYPE:
      printf("Vertical Velocity            : %0.2f\n",((ServoMobJACmdRotaryWing *) mob->cmdMsg)->vVertical);
      printf("Linear  Velocity             : %0.2f\n",((ServoMobJACmdRotaryWing *) mob->cmdMsg)->vLinear);
      printf("Lateral Velocity             : %0.2f\n",((ServoMobJACmdRotaryWing *) mob->cmdMsg)->vLateral);
      printf("Rotational Velocity            : %0.2f\n",((ServoMobJACmdRotaryWing *) mob->cmdMsg)->vRotation);
      break;
    default:
      printf("Mobility Command Type Not Recognize : %d\n",(int)mob->cmdMsg->type);
      break;
    }
}


void mobSetView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** Mobilty Settings View *****\n");
  printf("=================================\n");
  printf("Settings Message Type        : %s\n", servoMobJA_symbol_lookup( mob->setMsg->type ));
  printf("Cycle Time                   : %0.4f\n",((ServoMobJASet *)mob->setMsg)->cycleTime);
  printf("Debug                        : %d\n",((ServoMobJASet *)mob->setMsg)->debug);
  printf("World Name                   : %s\n",((ServoMobJASet *)mob->setMsg)->worldName);
  printf("Platform Name                : %s\n",((ServoMobJASet *)mob->setMsg)->platformName);
  printf("Length                       : %0.4f\n",((ServoMobJASet *)mob->setMsg)->length);
  printf("Height                       : %0.4f\n",((ServoMobJASet *)mob->setMsg)->height);
  printf("Width                        : %0.4f\n",((ServoMobJASet *)mob->setMsg)->width);
  printf("Mass                         : %0.4f\n",((ServoMobJASet *)mob->setMsg)->mass);
  printf("Center of Gravity            : (%0.4f, %0.4f, %0.4f)\n",
	 ((ServoMobJASet *)mob->setMsg)->cg.x,
	 ((ServoMobJASet *)mob->setMsg)->cg.y,
	 ((ServoMobJASet *)mob->setMsg)->cg.z);
  printf("Robot Type                   : %s\n",
	 servoMobJA_enum_moastRobotType_symbol_lookup( ((ServoMobJASet *)mob->setMsg)->botType ));
  printf("Steering Type                : %s\n",
	 servoMobJA_enum_moastSteerType_symbol_lookup( ((ServoMobJASet *)mob->setMsg)->steerType ));
  printf("Number of Effectors          : %d\n",((ServoMobJASet *)mob->setMsg)->effCount);
  printf("Number of Sensors            : %d\n",((ServoMobJASet *)mob->setMsg)->senCount);
  printf("Number of Mission Packages   : %d\n",((ServoMobJASet *)mob->setMsg)->misCount);
  switch(((ServoMobJASet *)mob->setMsg)->botType )
    {
    case MOAST_GROUND_VEHICLE_TYPE:
      printf("Max Front Steer Angle        : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->maxSteerAngle);
      printf("Max Rear Steer Angle         : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->maxCrabAngle);
      printf("Max Wheel Rotation Velocity  : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->maxWheelRot);
      printf("Max Torque                   : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->maxTorque);
      printf("Chassis Length               : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->wheelBase);
      printf("Wheel Separation             : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->wheelSeparation);
      printf("Wheel Radius                 : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->wheelRadius);
      printf("Turning Radius               : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->minTurningRadius);

      for(int i=0;i<((ServoMobJASetGrdVeh *)mob->setMsg)->fSet_length;i++)
	{
	  printf("%s\n",servoMobJA_enum_moastFlipperType_symbol_lookup(((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].fType));
	  printf("   Flipper Length            : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].length);
	  printf("   Flipper Width             : %0.4f\n",((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].width); 
	  printf("   Flipper Range             : [%0.4f, %0.4lf]\n",((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].minAngle, ((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].maxAngle);
	  printf("   Flipper Location          : (%0.4f,%0.4f, %0.4lf)\n",((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].location.x,
		 ((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].location.y,
		 ((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].location.z);

	  printf("   Flipper Orientation       : (%0.4f,%0.4f, %0.4lf)\n",((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].orientation.r,
		 ((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].orientation.p,
		 ((ServoMobJASetGrdVeh *)mob->setMsg)->fSet[i].orientation.y);
	  
	}

      break;
    case  MOAST_AERIAL_ROBOT_TYPE:
      printf("Max Lateral Velocity         : %0.4f\n",((ServoMobJASetAirBot *)mob->setMsg)->maxLatVel);
      printf("Max Linear Velocity          : %0.4f\n",((ServoMobJASetAirBot *)mob->setMsg)->maxLinVel);
      printf("Max Rotational Velocity      : %0.4f\n",((ServoMobJASetAirBot *)mob->setMsg)->maxRotVel);
      printf("Max Vertical Velocity        : %0.4f\n",((ServoMobJASetAirBot *)mob->setMsg)->maxVertVel);
    break;
    default:
      break;
    }
}
void mobStatView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** Mobilty State Menu    *****\n");
  printf("=================================\n");
  printf("   Admin State           : %s\n", rcs_admin_state_to_string(((ServoMobJAStat *)mob->statMsg)->admin_state));
  printf("   command_type          : %s\n", servoMobJA_symbol_lookup(((ServoMobJAStat *)mob->statMsg)->command_type));
  printf("   echo_serial_num       : %d\n", ((ServoMobJAStat *)mob->statMsg)->echo_serial_number);
  printf("   Status                : %s\n", rcs_status_to_string(((ServoMobJAStat *)mob->statMsg)->status));
  printf("   State:                : %d\n", ((ServoMobJAStat *)mob->statMsg)->state);
  if( ((ServoMobJAStat *)mob->statMsg)->error )
    printf("   Error Message         : (%d) %s\n", 
	   ((ServoMobJAStat *)mob->statMsg)->error,
	   ((ServoMobJAStat *)mob->statMsg)->message);
  printf("   Line:                 : %d\n", ((ServoMobJAStat *)mob->statMsg)->line);
  printf("   Source Line:          : %d\n", ((ServoMobJAStat *)mob->statMsg)->source_line);
  printf("   Source File:          : %s\n", ((ServoMobJAStat *)mob->statMsg)->source_file);
  printf("   HeartBeat             : %d\n", ((ServoMobJAStat *)mob->statMsg)->heartbeat);
  printf("   Battery Voltage       : %f\n", ((ServoMobJAStat *)mob->statMsg)->batteryVolt);
  printf("   Fuel Level            : %f\n", ((ServoMobJAStat *)mob->statMsg)->fuelLevel);
  printf("   HeadLights            : %s\n", ((ServoMobJAStat *)mob->statMsg)->headLights ? "On" : "Off");
  switch(((ServoMobJASet *)mob->setMsg)->botType )
    {
    case MOAST_GROUND_VEHICLE_TYPE:
      printf("   Steering Angle        : %0.4f\n",((ServoMobJAStatGrdVeh *)mob->statMsg)->steerAngle);
      printf("   Crab Angle            : %0.4f\n",((ServoMobJAStatGrdVeh *)mob->statMsg)->crabAngle);
      printf("   Flipper Count         : %d\n",((ServoMobJAStatGrdVeh *)mob->statMsg)->fStat_length);
      for(int i=0;i<((ServoMobJAStatGrdVeh *)mob->statMsg)->fStat_length;i++)
	printf("   %s : %0.4lf\n",
	       servoMobJA_enum_moastFlipperType_symbol_lookup(((ServoMobJAStatGrdVeh *)mob->statMsg)->fStat[i].fType),
	       ((ServoMobJAStatGrdVeh *)mob->statMsg)->fStat[i].flipperAngle);

      break;
    case  MOAST_AERIAL_ROBOT_TYPE:
      printf("   Lateral Velocity    : %0.4f\n",((ServoMobJAStatAirBot *)mob->statMsg)->vLateral);
      printf("   Linear Velocity     : %0.4f\n",((ServoMobJAStatAirBot *)mob->statMsg)->vLinear);
      printf("   Rotational Velocity : %0.4f\n",((ServoMobJAStatAirBot *)mob->statMsg)->vRotation);
      printf("   Vertical Velocity   : %0.4f\n",((ServoMobJAStatAirBot *)mob->statMsg)->vVertical);
    break;
    default:
      break;
    }




}
void doMobView(const vector<string> & toks)
{
  if(!toks.empty())
    {
      if (!nocase_cmp (toks[0], "init"))
	{
	  mobCmdInit.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdInit);
	}
      else if (!nocase_cmp (toks[0], "halt"))	
	{
	  mobCmdHalt.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdHalt);
	}
      else if (!nocase_cmp (toks[0], "abort"))
	{
	  mobCmdAbort.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdAbort);
	}
      else if (!nocase_cmp (toks[0], "shutdown"))
 	{
	  mobCmdShutdown.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdShutdown);
	}
      else if (!nocase_cmp (toks[0], "ack2"))
	{
	  if(toks.size() != 3 )
	    {
	      mobHelpView();
	      return;
	    }
	  
	  mobCmdAckerman.velocity=atof(toks[1].c_str());
	  mobCmdAckerman.steerAngle=atof(toks[2].c_str());
	  mobCmdAckerman.crabAngle=0;
	  mobCmdAckerman.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdAckerman);
	}
      else if (!nocase_cmp (toks[0], "ack3"))
	{
	  if(toks.size() != 4 )
	    {
	      mobHelpView();
	      return;
	    }
	  
	  mobCmdAckerman.velocity=atof(toks[1].c_str());
	  mobCmdAckerman.steerAngle=atof(toks[2].c_str());
	  mobCmdAckerman.crabAngle=atof(toks[3].c_str());
	  mobCmdAckerman.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdAckerman);
	}
      else if (!nocase_cmp (toks[0], "omni"))
	{
	  if(toks.size() != 4 )
	    {
	      mobHelpView();
	      return;
	    }
	  mobCmdOmni.vLinear=atof(toks[1].c_str());
	  mobCmdOmni.vLateral=atof(toks[2].c_str());
	  mobCmdOmni.rot=atof(toks[3].c_str());
	  mobCmdOmni.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdOmni);
	  
	}
      else if (!nocase_cmp (toks[0], "sub"))
	{
	  if(toks.size() != 4 )
	    {
	      mobHelpView();
	      return;
	    }
	  mobCmdSubmarine.propeller=atof(toks[1].c_str());
	  mobCmdSubmarine.rudder=atof(toks[2].c_str());
	  mobCmdSubmarine.sternPlane=atof(toks[3].c_str());
	  mobCmdSubmarine.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdSubmarine);
	}
      else if (!nocase_cmp (toks[0], "skid"))
	{
	   if(toks.size() != 3 )
	    {
	      mobHelpView();
	      return;
	    }
	  mobCmdSkid.wLeft=atof(toks[1].c_str());
	  mobCmdSkid.wRight=atof(toks[2].c_str());
	  mobCmdSkid.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdSkid);
	}

      else if (!nocase_cmp (toks[0], "flipskid"))
	{
	   if(toks.size() < 3 )
	    {
	      mobHelpView();
	      return;
	    }
	  mobCmdFlipSkid.wLeft=atof(toks[1].c_str());
	  mobCmdFlipSkid.wRight=atof(toks[2].c_str());
	  
	  
	  for(unsigned int i=3;i<toks.size();i++)
	    {
	      if(!nocase_cmp (toks[i], "fr"))
		mobCmdFlipSkid.fCmd[mobCmdFlipSkid.fCmd_length].fType=MOAST_FR_FLIPPER_TYPE;
	      
	      else if(!nocase_cmp (toks[i], "fl"))
		mobCmdFlipSkid.fCmd[mobCmdFlipSkid.fCmd_length].fType=MOAST_FL_FLIPPER_TYPE;

	      else if(!nocase_cmp (toks[i], "rr"))
		mobCmdFlipSkid.fCmd[mobCmdFlipSkid.fCmd_length].fType=MOAST_RR_FLIPPER_TYPE;
	      
	      else if(!nocase_cmp (toks[i], "rl"))
		mobCmdFlipSkid.fCmd[mobCmdFlipSkid.fCmd_length].fType=MOAST_RL_FLIPPER_TYPE;
	      
	      else
		{
		  mobHelpView();
		  return;
		}
	      mobCmdFlipSkid.fCmd[mobCmdFlipSkid.fCmd_length].flipperAngle=atof(toks[++i].c_str());
	      mobCmdFlipSkid.fCmd_length++;
	    }
	      
	  mobCmdFlipSkid.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdFlipSkid);
	  
	  mobCmdFlipSkid.fCmd_length=0;
	}
      else if (!nocase_cmp (toks[0], "rot"))
	{
	   if(toks.size() != 5 )
	    {
	      mobHelpView();
	      return;
	    }
	  mobCmdRotaryWing.vVertical=atof(toks[1].c_str());
	  mobCmdRotaryWing.vLinear=atof(toks[2].c_str());
	  mobCmdRotaryWing.vLateral=atof(toks[3].c_str());
	  mobCmdRotaryWing.vRotation=atof(toks[4].c_str());
	  mobCmdRotaryWing.serial_number=mob_serial_number++;
	  mob->cmdBuf->write(mobCmdRotaryWing);
	}
      else if (!nocase_cmp (toks[0], "settings"))
	{
	  mobSetView();
	  return;
	}
      else if (!nocase_cmp (toks[0], "stat"))
	{}
      else if (!nocase_cmp (toks[0], "cmd"))
	{
	  mobCmdView();
	  return;
	}
      else if (!nocase_cmp (toks[0], "help"))
	{
	  mobHelpView();
	  return;
	}
      else if (toks[0][0] == '?')
	{
	  mobHelpView();
	  return;
	}
      else
	{
	  printf("Command %s not recognized\n",toks[0].c_str());
	  mobHelpView();
	  return;
	}
    }
  mobStatView();
  

}


///////////////////////////////////////////////////////////////////////////////////
/////       MISSION PACKAGE                                                ////////
///////////////////////////////////////////////////////////////////////////////////
void misLinkStatView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("**** MisPkg Link State Menu *****\n");
  printf("=================================\n");
  for(unsigned int i=0;i<mis.size();i++)
    {
     printf("-------------------------------------------\n");
     printf("   MisPkg ID       : %d\n", i+1);
      printf("   Time            : %f\n", 
	     ((ServoMisJAStat *)mis[i].statMsg)->time);
      printf("   Link Count      : %d\n",  
	     ((ServoMisJAStat *)mis[i].statMsg)->linkStat_length);
      
      for(int j=0;j<((ServoMisJAStat *)mis[i].statMsg)->linkStat_length;j++)
      	{
	  printf("   Link %2d         : (jointVal) %6.2f (torque) %6.2f\n", j+1, 
		 ((ServoMisJAStat *)mis[i].statMsg)->linkStat[j].jointVal,
		 ((ServoMisJAStat *)mis[i].statMsg)->linkStat[j].torque);
	}
      printf("\n");
    }
}


void misStatView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("*****  MisPkg State Menu    *****\n");
  printf("=================================\n");
  
  for(unsigned int i=0;i<mis.size();i++)
    {
      printf("-------------------------------------------\n");
      printf("   MisPkg ID       : %d\n", i+1);
      printf("   Admin State     : %s\n", rcs_admin_state_to_string(((ServoMisJAStat *)mis[i].statMsg)->admin_state));
      printf("   command_type    : %s\n", servoMisJA_symbol_lookup(((ServoMisJAStat *)mis[i].statMsg)->command_type));
      printf("   echo_serial_num : %d\n", ((ServoMisJAStat *)mis[i].statMsg)->echo_serial_number);
      printf("   Status          : %s\n", rcs_status_to_string(((ServoMisJAStat *)mis[i].statMsg)->status));
      printf("   State:          : %d\n", ((ServoMisJAStat *)mis[i].statMsg)->state);  
      if( ((ServoMisJAStat *)mis[i].statMsg)->error )
	printf("   Error Message   : (%d) %s\n", 
	       ((ServoMisJAStat *)mis[i].statMsg)->error,
	       ((ServoMisJAStat *)mis[i].statMsg)->message);
      printf("   Line:           : %d\n", ((ServoMisJAStat *)mis[i].statMsg)->line);
      printf("   Source Line:    : %d\n", ((ServoMisJAStat *)mis[i].statMsg)->source_line);
      printf("   Source File:    : %s\n", ((ServoMisJAStat *)mis[i].statMsg)->source_file);
      printf("   HeartBeat       : %d\n\n", ((ServoMisJAStat *)mis[i].statMsg)->heartbeat);

    }
}

void misSetView()
{
  PM_RPY rpy;
  printf("\033[2J"); 
  printf("=================================\n");
  printf("***** MisPkg Settings Menu  *****\n");
  printf("=================================\n");
  for(unsigned int i=0;i<mis.size();i++)
    {
      printf("*******************************************\n");
      printf("Mission Package       : %2d\n", i+1);      
      for( int j=0; j < ((ServoMisJASet *)mis[i].setMsg)->linkSet_length; j++)
	{
	  printf("-------------------------------------------\n");
	  printf("Link ID               : %2d\n",((ServoMisJASet *)mis[i].setMsg)->linkSet[j].linkID);
	  printf("Parent Link ID        : %2d\n",((ServoMisJASet *)mis[i].setMsg)->linkSet[j].parentLinkID);
	  printf("Joint Type            : %s\n" ,
		 servoMisJA_enum_ServoMisJAJointType_symbol_lookup(((ServoMisJASet *)mis[i].setMsg)->linkSet[j].jType));
	  printf("Max Velocity          : %0.2f\n",((ServoMisJASet *)mis[i].setMsg)->linkSet[j].maxVelocity);
	  printf("Max Torque            : %0.2f\n",((ServoMisJASet *)mis[i].setMsg)->linkSet[j].maxTorque);
	  printf("Range                 : %0.2f %0.2f [min,max]\n",
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].minRange,
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].maxRange);
	  printf("Location              : %0.2f %0.2f %0.2f [x,y,z]\n",
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].loc.x,
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].loc.y,
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].loc.z);
	  printf("Orientation           : %0.2f %0.2f %0.2f [r,p,y]\n",
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].rot.r,
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].rot.p,
		 ((ServoMisJASet *)mis[i].setMsg)->linkSet[j].rot.y);
     
	}
    }
}

void misCmdView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("*****  MisPkg  Cmd View     *****\n");
  printf("=================================\n");
  for(unsigned int i=0; i < mis.size(); i++)
    {
      printf("-------------------------------------------\n");
      printf("Mission Package Cmd                 : %d\n",i+1);
      printf("Mission Package Cmd Type            : %s\n",
	     servoMisJA_symbol_lookup( mis[i].cmdMsg->type ) );
      printf("Mission Package Cmd Serial Number   : %d\n", mis[i].cmdMsg->serial_number );
      switch(mis[i].cmdMsg->type)
	{
	case SERVO_MIS_JA_CMD_MOVE_TYPE:
	  for(int j=0;j<((ServoMisJACmdMove *)mis[i].cmdMsg)->linkCmd_length;j++)
	    {
	      printf("   Link %2d   Value %6.2f MoveType %s\n", 
		     ((ServoMisJACmdMove *)mis[i].cmdMsg)->linkCmd[j].linkID,
		     ((ServoMisJACmdMove *)mis[i].cmdMsg)->linkCmd[j].value,
		     servoMisJA_enum_ServoMisJALinkCmdType_symbol_lookup
		     (((ServoMisJACmdMove *)mis[i].cmdMsg)->linkCmd[j].type)
		     );
	    }
	}
    }
}

void misHelpView()
{
   printf("\033[2J");
  printf("=================================\n");
  printf("*****  MisPkg Help Menu     *****\n");
  printf("=================================\n");
  printf("quit                         : exit\n");
  printf("? | help                     : print mission package help\n");
  printf("admin                        : switch to admin mode\n");
  printf("eff                          : switch to effector mode\n");
  printf("mob                          : switch to mobility mode\n");
  printf("mis                          : switch to mission package mode\n");
  printf("sp                           : switch to sensor processing mode\n");
  printf("init                         : initialize mission package\n");
  printf("abort                        : powerdown mission package\n");
  printf("halt                         : powerdown mission package\n");
  printf("shutdown                     : powerdown mission package\n");  
  printf("settings                     : view mission package settings\n");
  printf("stat | <enter>               : view mission package status\n");
  printf("link                         : view status of each link\n");
  printf("cmd                          : view mission package command\n");
  printf("move <int><int><float><int>  : control joint of mission package\n");
  printf("\t* Description of Move Command\n"); 
  printf("\t<int>   : Mission package ID\n");
  printf("\t<int>   : Link ID\n");
  printf("\t<float> : Value for link command\n");
  printf("\t<int>   : Move Type,  Abs Angle(0) Vel(1) Torque(2)\n");
  printf("scan <start|1><misID><linkID><max><min><step><time>\n");
  printf("\t* Scan command <step>s mission package\n\tbetween <max> and <min> at rate <time>\n");
  printf("scan <stop|0>                : Stops scan cmd\n");
  
}

struct ScanInfo
{
  int misID;
  int linkID;
  double step;
  double time;
  double upper;
  double lower;
  bool run;
};

void doMisScan(void * args)
{
  ScanInfo * info = (ScanInfo *)args;
  ServoMisJACmdMove moveCmd;
  int sign=1;
  moveCmd.serial_number=0;
  moveCmd.linkCmd_length=1;
  while(info->run)
    {
      moveCmd.serial_number++;    
      if( moveCmd.linkCmd[0].value >= info->upper)
	sign=-1;
      if( moveCmd.linkCmd[0].value <= info->lower)
	sign=1;
      moveCmd.linkCmd[0].value = moveCmd.linkCmd[0].value + (sign * info->step);
      moveCmd.linkCmd[0].linkID=info->linkID;  
      mis[info->misID-1].cmdBuf->write(moveCmd);
      esleep(info->time);
    }
  moveCmd.serial_number++;    
  moveCmd.linkCmd[0].value = 0;
  mis[info->misID-1].cmdBuf->write(moveCmd);
  ulapi_task_exit();

  return;
}

void doMisView(const vector<string> & toks)
{
  static ScanInfo * info=NULL;
  
  if(!toks.empty())
    {
      if (!nocase_cmp (toks[0], "init"))
	{
	  for(unsigned int i=0;i<mis.size();i++)
	    {
	      misCmdInit.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdInit);
	    }
	}

      else if (!nocase_cmp (toks[0], "halt"))
	{
	  for(unsigned int i=0;i<mis.size();i++)
	    {
	      misCmdHalt.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdHalt);
	    }
	}
      else if (!nocase_cmp (toks[0], "abort"))
	{
	  for(unsigned int i=0;i<mis.size();i++)
	    {
	      misCmdAbort.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdAbort);
	    }
	}
      else if (!nocase_cmp (toks[0], "shutdown"))
	{
	  for(unsigned int i=0;i<mis.size();i++)
	    {
	      misCmdShutdown.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdShutdown);
	    }
	}
       else if (!nocase_cmp (toks[0], "cmd"))
	{
	  misCmdView();
	  return;
	} 
      else if (!nocase_cmp (toks[0], "link"))
	{
	  misLinkStatView();
	  return;
	} 
      else if (!nocase_cmp (toks[0], "settings"))
	{
	  misSetView();
	  return;
	}
      else if (!nocase_cmp (toks[0], "scan"))
	{
	  // stops scan and returns link to home location
	  if ((!nocase_cmp (toks[1], "stop")) || (!nocase_cmp (toks[1], "0")))
	    {
	      if( info != NULL)
		{ 
		  info->run=false;
		  (void) ulapi_task_join(scanThr);
		  delete info;
		}
	      info=NULL;
	    }
	  else if ((!nocase_cmp (toks[1], "start")) || (!nocase_cmp (toks[1], "1")))
	    {
	      
	      if( info == NULL)
		info = new ScanInfo;
	      
	      if(toks.size() != 8 )
		{
		  misHelpView();
		  return;
		}
	      info->misID = atoi(toks[2].c_str());
	      info->linkID = atoi(toks[3].c_str());
	      info->upper = atof(toks[4].c_str());
	      info->lower = atof(toks[5].c_str());
	      info->step = atof(toks[6].c_str());
	      info->time = atof(toks[7].c_str());
	      info->run = true;
	      
	      // check to see if mission package exists and that the upper is greater than the lower
	      if( (info->misID <= (int)mis.size())  &&   (info->upper > info->lower) )
		{
		  // check the range of sensor for scan
		  if( ((ServoMisJASet *)mis[info->misID-1].setMsg)->linkSet[info->linkID-1].maxRange > 
		      ((ServoMisJASet *)mis[info->misID-1].setMsg)->linkSet[info->linkID-1].minRange)
		    {
		      if( ((ServoMisJASet *)mis[info->misID-1].setMsg)->linkSet[info->linkID-1].maxRange < info->upper)
			info->upper= ((ServoMisJASet *)mis[info->misID-1].setMsg)->linkSet[info->linkID-1].maxRange;
		      if( ((ServoMisJASet *)mis[info->misID-1].setMsg)->linkSet[info->linkID-1].minRange > info->lower)
			info->lower= ((ServoMisJASet *)mis[info->misID-1].setMsg)->linkSet[info->linkID-1].minRange;
		    }
		  
		  // start scanning thread
#define SECS_TO_NSECS(secs) ((ulapi_integer) ((secs) * 1.0e9 + 0.5))
		  scanThr = ulapi_task_new();
		  if (NULL == scanThr || 
		      ULAPI_OK != ulapi_task_start(scanThr, doMisScan, info, ulapi_prio_highest(), SECS_TO_NSECS(info->time))) {
		    printf("**ERROR** <ServoShelfailed::doMisView> ulapi_task_start failed\n");
		    return;
		  }
		}
	    }
	  else 
	    {
	      misHelpView();
	      return;
	    }
	}
      else if (!nocase_cmp (toks[0], "move"))
	{
	  if(toks.size() != 5 )
	    {
	      misHelpView();
	      return;
	    }
	  if( atoi(toks[1].c_str()) <= 0 )
	    {
	      printf( "servoShell.cc: mission package id must be greater or equal to 1\n" );
	      misHelpView();
	      return;
	    }
	  misCmdMove.serial_number=mis_serial_number++;
	  misCmdMove.linkCmd_length=1;
	  misCmdMove.linkCmd[0].linkID=atoi(toks[2].c_str());
	  misCmdMove.linkCmd[0].value=atof(toks[3].c_str());
	  switch(atoi(toks[4].c_str()))
	    {
	    case 1:
	      misCmdMove.linkCmd[0].type=SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE;
	      break;
	    case 2:
	      misCmdMove.linkCmd[0].type=SERVO_MIS_JA_LINK_CMD_TORQUE_TYPE;
	      break;
	    case 0:
	    default:
	      misCmdMove.linkCmd[0].type=SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;
	      break;
	    }
	  mis[atoi(toks[1].c_str())-1].cmdBuf->write(misCmdMove);
	}
      else if (!nocase_cmp (toks[0], "stat"))
	{}
      else
	{
	  misHelpView();
	  return;
	}
	
    }
  misStatView();
  
}

///////////////////////////////////////////////////////////////////////////////////
/////       SENSOR PROCESSING                                              ////////
///////////////////////////////////////////////////////////////////////////////////
void spHelp()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** SP Help Menu          *****\n");
  printf("=================================\n");
  printf("admin                        : switch to admin mode\n");
  printf("eff                          : switch to effector mode\n");
  printf("mob                          : switch to mobility mode\n");
  printf("mis                          : switch to mission package mode\n");
  printf("sp                           : switch to sensor processing mode\n");
  printf("quit                         : exit\n");
  printf("? | help                     : print sensor processing help\n");
  printf("admin                        : switch to admin mode\n");
  printf("mob                          : switch to mobility mode\n");
  printf("mis                          : switch to mission package mode\n");
  printf("sp                           : switch to sensor processing mode\n");

  printf("init                         : initialize sensor processing  system\n");
  printf("abort                        : powerdown sensor processing system\n");
  printf("halt                         : powerdown sensor processing system\n");
  printf("shutdown                     : powerdown sensor processing system\n");
  printf("stat | <enter>               : view sensor status\n");
  printf("settings                     : view sensor settings\n");
  printf("go                           : start sensor feeds\n");
}

void spSetView()
{
  PM_RPY rpy;
  printf("\033[2J");
  printf("=================================\n");
  printf("***** SP Settings Menu      *****\n");
  printf("=================================\n");
  printf("Sensor Processing Settings Menu\n");
  printf("Cycle Time                   : %0.4f\n",((ServoSPSet *)sp->setMsg)->cycleTime);
  rpy = ((ServoSPSet *)sp->setMsg)->startPose.rot;
  printf("Start Location               : [%0.4lf,  %0.4lf,  %0.4lf]\n",
	 ((ServoSPSet *)sp->setMsg)->startPose.tran.x, 
	 ((ServoSPSet *)sp->setMsg)->startPose.tran.y, 
	 ((ServoSPSet *)sp->setMsg)->startPose.tran.z );
  printf("Start Rotation               : [%0.4lf,  %0.4lf,  %0.4lf]\n",
	 rpy.r, rpy.p, rpy.y);
  printf("Debug                        : %d\n",((ServoSPSet *)sp->setMsg)->debug);
  printf("World Name                   : %s\n",((ServoSPSet *)sp->setMsg)->worldName);
  printf("Vehicle ID                   : %d\n",((ServoSPSet *)sp->setMsg)->vehicleID);
  printf("Number of Sensors            : %d\n",((ServoSPSet *)sp->setMsg)->sSetElem_length);
  for(int i=0;i<((ServoSPSet *)sp->setMsg)->sSetElem_length;i++)
    {
      printf("-------------------------------------------\n");
      printf("   Sensor ID                    : %d\n",((ServoSPSet *)sp->setMsg)->sSetElem[i].sensorID);
      printf("   Sensor Type                  : %s\n",
	     servoSP_enum_ServoSensorType_symbol_lookup( ((ServoSPSet *)sp->setMsg)->sSetElem[i].sType) );
      printf("   Sensor Data Buffer           : %s\n",((ServoSPSet *)sp->setMsg)->sSetElem[i].senDataBufName);
      printf("   Mount ID                     : %d\n",((ServoSPSet *)sp->setMsg)->sSetElem[i].mountID);
      printf("   Link  ID                     : %d\n",((ServoSPSet *)sp->setMsg)->sSetElem[i].linkID);
      rpy= ((ServoSPSet *)sp->setMsg)->sSetElem[i].sensorToMount.rot;
      //printf("   Test                    : %d\n",((ServoSPSet *)sp->setMsg)->sSetElem[i].sensorID);
      printf("   Sensor to Mount              : (%0.4f,%0.4f,%0.4f,%0.4f,%0.4f,%0.4f)\n",
	     ((ServoSPSet *)sp->setMsg)->sSetElem[i].sensorToMount.tran.x,
	     ((ServoSPSet *)sp->setMsg)->sSetElem[i].sensorToMount.tran.y,
	     ((ServoSPSet *)sp->setMsg)->sSetElem[i].sensorToMount.tran.z,
	     rpy.r,
	     rpy.p,
	     rpy.y );
      printf("   Desired Period               : %0.4f\n",((ServoSPSet *)sp->setMsg)->sSetElem[i].desiredPeriod);
      switch(((ServoSPSet *)sp->setMsg)->sSetElem[i].sType )
	{
	case SERVO_SP_LINESCANNER_SENSOR:
	  printf("   Max Range                    : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_LINESCAN_MAX_RANGE]);
	  printf("   Min Range                    : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_LINESCAN_MIN_RANGE]);
	  printf("   Resolution                   : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_LINESCAN_RESOLUTION]);
	  printf("   Field of View                : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_LINESCAN_FOV]);
	  printf("   Beam Dispersion              : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_LINESCAN_BEAM_DISPERSION]);
	  break;
	case SERVO_SP_SONAR_SENSOR:
	  printf("   Max Range                    : %0.4f\n",  
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_SONAR_MAX_RANGE]);
	  printf("   Min Range                    : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_SONAR_MIN_RANGE]);
	  printf("   Beam Dispersion              : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_SONAR_BEAM_DISPERSION]);
	  break;
	case SERVO_SP_RFID_SENSOR:
	  printf("   Max Range                    : %0.4f\n",  
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_RFID_MAX_RANGE]);
	  break;
	case SERVO_SP_VICTIM_SENSOR:
	  printf("   Max Range                    : %0.4f\n",  
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_VICTIM_MAX_RANGE]);
	  printf("   Vertical FOV                 : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_VICTIM_VERT_FOV]);
	  printf("   Horizontal FOV               : %0.4f\n", 
		 ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_VICTIM_HORZ_FOV]);
	  break;
	case SERVO_SP_GRD_TRUTH_SENSOR:
	   printf("   Utm Offset                   : (%0.4f, %0.4f, %0.4f)\n", 
		  ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_NORTHING],
		  ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_EASTING],
		  ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_Z_VALUE]);
	   printf("   Utm Zone | Letter            : %d | %c\n", 
		  (int)((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_ZONE],
		  (char)((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_LETTER]);
	   break;
	case SERVO_SP_CAMERA_SENSOR:
	   printf("   FOV (min,max,default)        : (%0.4f, %0.4f, %0.4f)\n", 
		  ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_CAMERA_MIN_FOV],
		  ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_CAMERA_MAX_FOV],
		  ((ServoSPSet *)sp->setMsg)->sSetElem[i].parameters[SERVO_SP_CAMERA_DEFAULT_FOV]);
	   break;
	default:
	  break;
	}
      
    }
}

void spStatView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** SP State Menu         *****\n");
  printf("=================================\n");
  printf("   Admin State            : %s\n", rcs_admin_state_to_string(((ServoSPStat *)sp->statMsg)->admin_state));
  printf("   command_type           : %s\n", servoSP_symbol_lookup(((ServoSPStat *)sp->statMsg)->command_type));
  printf("   Status                 : %s\n", rcs_status_to_string(((ServoSPStat *)sp->statMsg)->status));
  printf("   State                  : %d\n", ((ServoSPStat *)sp->statMsg)->state);
  if( ((ServoSPStat *)sp->statMsg)->error )
    printf("   Error Message          : (%d) %s\n", 
	   ((ServoSPStat *)sp->statMsg)->error,
	   ((ServoSPStat *)sp->statMsg)->message);
  printf("   Line                   : %d\n", ((ServoSPStat *)sp->statMsg)->line);
  printf("   Source Line            : %d\n", ((ServoSPStat *)sp->statMsg)->source_line);
  printf("   Source File            : %s\n", ((ServoSPStat *)sp->statMsg)->source_file);
  printf("   Heart Beat             : %d\n", ((ServoSPStat *)sp->statMsg)->heartbeat);
  printf("   Transmitting Data      : %s\n", ((ServoSPStat *)sp->statMsg)->transmitData ? "Yes" : "No");
}

void doSPView(const vector<string> & toks)
{
  if(!toks.empty())
    {
      if (!nocase_cmp (toks[0], "init"))
	{
	  spCmdInit.serial_number=sp_serial_number++;
	  sp->cmdBuf->write(spCmdInit);
	}
      else if (!nocase_cmp (toks[0], "halt"))
	{
	  spCmdHalt.serial_number=sp_serial_number++;
	  sp->cmdBuf->write(spCmdHalt);
	}
      else if (!nocase_cmp (toks[0], "abort"))
	{
	  spCmdAbort.serial_number=sp_serial_number++;
	  sp->cmdBuf->write(spCmdAbort);
	}	
      else if (!nocase_cmp (toks[0], "shutdown"))
		{
	  spCmdShutdown.serial_number=sp_serial_number++;
	  sp->cmdBuf->write(spCmdShutdown);
	}
      else if (!nocase_cmp (toks[0], "go"))
	{
	  spCmdGo.serial_number=sp_serial_number++;
	  sp->cmdBuf->write(spCmdGo);
	}
      else if (!nocase_cmp (toks[0], "settings"))
	{
	  spSetView();
	  return;
	}
      else if (!nocase_cmp (toks[0], "stat"))
	{}
      else if (!nocase_cmp (toks[0], "help"))
	{
	  spHelp();
	  return;
	}
      else if (toks[0][0] == '?')
	{
	  spHelp();
	  return;
	}
      else
	{
	  printf("Command %s not recognized\n",toks[0].c_str());
	  spHelp();
	  return;
	} 
    }
  spStatView();
}


///////////////////////////////////////////////////////////////////////////////////
/////       EFFECTER SECTION                                               ////////
///////////////////////////////////////////////////////////////////////////////////

ServoEffJAOpcodeType getStringToOpcode(string name)
{
  if (!nocase_cmp (name, "animate"))
    return SERVO_EFF_JA_ANIMATE_OPCODE_TYPE;
  
  else if (!nocase_cmp (name, "activate"))
    return SERVO_EFF_JA_ACTIVATE_OPCODE_TYPE;
  
  else if (!nocase_cmp (name, "fire"))
    return SERVO_EFF_JA_FIRE_OPCODE_TYPE;
  
  else if (!nocase_cmp (name, "release"))
    return SERVO_EFF_JA_RELEASE_OPCODE_TYPE;
  
  else if (!nocase_cmp (name, "reset"))
    return SERVO_EFF_JA_RESET_OPCODE_TYPE;
  
  else if (!nocase_cmp (name, "nop"))
    return SERVO_EFF_JA_NOP_OPCODE_TYPE;
  
  else
    return SERVO_EFF_JA_INVALID_OPCODE_TYPE;
}


void effHelpView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** Effecter Help Menu    *****\n");
  printf("=================================\n");
  printf("admin                        : switch to admin mode\n");
  printf("eff                          : switch to effector mode\n");
  printf("mob                          : switch to mobility mode\n");
  printf("mis                          : switch to mission package mode\n");
  printf("sp                           : switch to sensor processing mode\n");
  printf("quit                            : exit\n");
  printf("? | help                        : print sensor processing help\n");
  printf("init                            : initialize effecters\n");
  printf("abort                           : powerdown effecters\n");
  printf("halt                            : powerdown effecters\n");
  printf("shutdown                        : powerdown effecters\n");
  printf("stat | <enter>                  : view effecter's status\n");
  printf("settings                        : view effecter's settings\n");
  printf("opcode <effID> <opcode> <param> : send opcode command to effecter\n");
  printf("* valid opcodes: animate, activate, fire, release, reset, nop \n");
}


void effSetView()
{
  printf("\033[2J");
  printf("==================================\n");
  printf("***** Effecter Settings Menu *****\n");
  printf("==================================\n");
  printf("Effecter Settings Menu\n");
  printf("Cycle Time                   : %0.4f\n",((ServoEffJASet *)eff->setMsg)->cycleTime);
  printf("Debug                        : %d\n",((ServoEffJASet *)eff->setMsg)->debug);
  printf("Vehicle ID                   : %d\n",((ServoEffJASet *)eff->setMsg)->vehicleID);
  printf("Number of Effector           : %d\n",((ServoEffJASet *)eff->setMsg)->effector_length);
  for(int i=0;i < ((ServoEffJASet *)eff->setMsg)->effector_length; i++)
    {
       printf("-------------------------------------------\n");
       printf("Effector ID                  : %d\n",((ServoEffJASet *)eff->setMsg)->effector[i].effID);
       printf("Number Opcodes               : %d\n",((ServoEffJASet *)eff->setMsg)->effector[i].opConf_length);
       for(int j=0;j<((ServoEffJASet *)eff->setMsg)->effector[i].opConf_length;j++)
	 {
	    printf("Opcode Type                  : %s \n",
	   	  servoEffJA_enum_ServoEffJAOpcodeType_symbol_lookup(((ServoEffJASet *)eff->setMsg)->effector[i].opConf[j].opcode));
	 
 	   printf("Opcode Parameter Range       : [%8.4lf %8.4lf]\n",
 		  ((ServoEffJASet *)eff->setMsg)->effector[i].opConf[j].max,
 		  ((ServoEffJASet *)eff->setMsg)->effector[i].opConf[j].min);
 	 }
    }
}


void effStatView()
{
  printf("\033[2J");  
  printf("=================================\n");
  printf("***** Effector State Menu   *****\n");
  printf("=================================\n");
  printf("   Admin State            : %s\n", rcs_admin_state_to_string(((ServoEffJAStat *)eff->statMsg)->admin_state));
  printf("   command_type           : %s\n", servoEffJA_symbol_lookup(((ServoEffJAStat *)eff->statMsg)->command_type));
  if( ((ServoEffJAStat *)eff->statMsg)->command_type == SERVO_EFF_JA_CMD_OPCODE_TYPE )
    printf("   opcode_type            : %s\n", servoEffJA_enum_ServoEffJAOpcodeType_symbol_lookup(effCmdOpcode.opcode));
  printf("   Status                 : %s\n", rcs_status_to_string(((ServoEffJAStat *)eff->statMsg)->status));
  printf("   State                  : %d\n", ((ServoEffJAStat *)eff->statMsg)->state);
  if( ((ServoEffJAStat *)eff->statMsg)->error )
    printf("   Error Message          : (%d) %s\n", 
	   ((ServoEffJAStat *)eff->statMsg)->error,
	   ((ServoEffJAStat *)eff->statMsg)->message);
  printf("   Line                   : %d\n", ((ServoEffJAStat *)eff->statMsg)->line);
  printf("   Source Line            : %d\n", ((ServoEffJAStat *)eff->statMsg)->source_line);
  printf("   Source File            : %s\n", ((ServoEffJAStat *)eff->statMsg)->source_file);
  printf("   Heart Beat             : %d\n", ((ServoEffJAStat *)eff->statMsg)->heartbeat);

}

void doEffView(const vector<string> & toks)
{
  if(!toks.empty())
    {

      if (!nocase_cmp (toks[0], "init"))
	{
	  effCmdInit.serial_number=eff_serial_number++;
	  eff->cmdBuf->write(effCmdInit);
	}
      else if (!nocase_cmp (toks[0], "halt"))
	{
	  effCmdHalt.serial_number=eff_serial_number++;
	  eff->cmdBuf->write(effCmdHalt);
	}
      else if (!nocase_cmp (toks[0], "abort"))
	{
	  effCmdAbort.serial_number=eff_serial_number++;
	  eff->cmdBuf->write(effCmdAbort);
	}	
      else if (!nocase_cmp (toks[0], "shutdown"))
		{
	  effCmdShutdown.serial_number=eff_serial_number++;
	  eff->cmdBuf->write(effCmdShutdown);
	}
      
      else if (!nocase_cmp (toks[0], "opcode"))
	{
	  dumpTokens(toks);
	  if(toks.size() != 4 )
	    {
	      effHelpView();
	      return;
	    }
	  effCmdOpcode.serial_number=eff_serial_number++;
	  effCmdOpcode.effID=atoi(toks[1].c_str());
	  effCmdOpcode.opcode=getStringToOpcode(toks[2]);
	  effCmdOpcode.param=atof(toks[3].c_str());
	  //printf("%d %d %lf\n",   effCmdOpcode.effID,effCmdOpcode.opcode,effCmdOpcode.param);
	  if ( effCmdOpcode.opcode == SERVO_EFF_JA_INVALID_OPCODE_TYPE)  
	    {
	      effHelpView();
	      return;
	    }
	  eff->cmdBuf->write(effCmdOpcode);
	}
      
      else if (!nocase_cmp (toks[0], "settings"))
	{
	  effSetView();
	  return;
	}
      else if (!nocase_cmp (toks[0], "stat"))
	{}
      else if (!nocase_cmp (toks[0], "help"))
	{
	  effHelpView();
	  return;
	}
      else if (toks[0][0] == '?')
	{
	  effHelpView();
	  return;
	}
      else
	{
	  printf("Command %s not recognized\n",toks[0].c_str());
	  effHelpView();
	  return;
	} 
    }
  effStatView();
}

///////////////////////////////////////////////////////////////////////////////////
/////       ADMINISTRATION                                                 ////////
///////////////////////////////////////////////////////////////////////////////////
void adminHelp()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** Admin Help Menu       *****\n");
  printf("=================================\n");
  printf("quit                         : exit\n");
  printf("? | help                     : print this help\n");
  printf("admin                        : switch to admin mode\n");
  printf("eff                          : switch to effector mode\n");
  printf("mob                          : switch to mobility mode\n");
  printf("mis                          : switch to mission package mode\n");
  printf("sp                           : switch to sensor processing mode\n");
  printf("init                         : initialize entire system\n");
  printf("abort                        : powerdown entire system\n");
  printf("halt                         : powerdown entire system\n");
}
void adminStatView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("***** Admin Help Menu       *****\n");
  printf("=================================\n");
  printf("Mobility\n");
  if(mob != NULL )
    {
      printf("   Admin State  : %s\n", rcs_admin_state_to_string(((ServoMobJAStat *)mob->statMsg)->admin_state));
      printf("   command_type : %s\n", servoMobJA_symbol_lookup(((ServoMobJAStat *)mob->statMsg)->command_type));
      printf("   Status       : %s\n", rcs_status_to_string(((ServoMobJAStat *)mob->statMsg)->status));
      printf("   HeartBeat    : %d\n", ((ServoMobJAStat *)mob->statMsg)->heartbeat);
    }
 
  printf("Sensor Processing\n");
  if(sp != NULL)
    {
      printf("   Admin State  : %s\n", rcs_admin_state_to_string(((ServoSPStat *)sp->statMsg)->admin_state));
      printf("   command_type : %s\n", servoSP_symbol_lookup(((ServoSPStat *)sp->statMsg)->command_type));
      printf("   Status       : %s\n", rcs_status_to_string(((ServoSPStat *)sp->statMsg)->status));
      printf("   HeartBeat    : %d\n", ((ServoSPStat *)sp->statMsg)->heartbeat);
    }

  printf("Effector\n");
  if(eff != NULL )
    {
      printf("   Admin State  : %s\n", rcs_admin_state_to_string(((ServoEffJAStat *)eff->statMsg)->admin_state));
      printf("   command_type : %s\n", servoEffJA_symbol_lookup(((ServoEffJAStat *)eff->statMsg)->command_type));
      printf("   Status       : %s\n", rcs_status_to_string(((ServoEffJAStat *)eff->statMsg)->status));
      printf("   HeartBeat    : %d\n", ((ServoEffJAStat *)eff->statMsg)->heartbeat);
    }
  
  printf("Mission Package\n");
  for(unsigned int i=0;i<mis.size();i++)
    {
      printf("   MisPkg ID    : %d\n", i+1);
      printf("   Admin State  : %s\n", rcs_admin_state_to_string(((ServoMisJAStat *)mis[i].statMsg)->admin_state));
      printf("   command_type : %s\n", servoMisJA_symbol_lookup(((ServoMisJAStat *)mis[i].statMsg)->command_type));
      printf("   Status       : %s\n", rcs_status_to_string(((ServoMisJAStat *)mis[i].statMsg)->status));
      printf("   HeartBeat    : %d\n", ((ServoMisJAStat *)mis[i].statMsg)->heartbeat);
    }
  if (joystickDataBuf != NULL && joystickDataPtr != NULL) {
    joystickDataBuf->read();
    printf("Joystick: %d\n", joystickDataPtr->mode);
  }
}
void doAdminView(const vector<string> & toks)
{ 
  unsigned int i=0;
  if(!toks.empty())
    {
      if (!nocase_cmp (toks[0], "init"))
	{
	  if(mob!=NULL)
	    {
	      mobCmdInit.serial_number=mob_serial_number++;
	      mob->cmdBuf->write(mobCmdInit);
	    }
	  if(sp != NULL)
	    {
	      spCmdInit.serial_number=sp_serial_number++;
	      sp->cmdBuf->write(spCmdInit);
	    }
	  if(eff != NULL)
	    {
	      effCmdInit.serial_number=eff_serial_number++;
	      eff->cmdBuf->write(effCmdInit);
	    }
	  for(i=0;i<mis.size();i++)
	    {
	      misCmdInit.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdInit);
	    }
	}
      else if (!nocase_cmp (toks[0], "halt"))
	{
	  
	  if(mob!=NULL)
	    {
	      mobCmdHalt.serial_number=mob_serial_number++;
	      mob->cmdBuf->write(mobCmdHalt);
	    }
	  if(sp != NULL)
	    {
	      spCmdHalt.serial_number=sp_serial_number++;
	      sp->cmdBuf->write(spCmdHalt);
	    }
	  if(eff != NULL)
	    {
	      effCmdHalt.serial_number=eff_serial_number++;
	      eff->cmdBuf->write(effCmdHalt);
	    }
	  for(i=0;i<mis.size();i++)
	    {
	      misCmdHalt.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdHalt);
	    }
	}
      else if (!nocase_cmp (toks[0], "abort"))
	{
	  if(mob!=NULL)
	    {
	      mobCmdAbort.serial_number=mob_serial_number++;
	      mob->cmdBuf->write(mobCmdAbort);
	    }
	  if(sp != NULL)
	    {
	      spCmdAbort.serial_number=sp_serial_number++;
	      sp->cmdBuf->write(spCmdAbort);
	    }
	  if(eff != NULL)
	    {
	      effCmdAbort.serial_number=eff_serial_number++;
	      eff->cmdBuf->write(effCmdAbort);
	    }

	  for(i=0;i<mis.size();i++)
	    {
	      misCmdAbort.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdAbort);
	    }
	}
      else if (!nocase_cmp (toks[0], "shutdown"))
	{
	  if(mob!=NULL)
	    {
	      mobCmdShutdown.serial_number=mob_serial_number++;
	      mob->cmdBuf->write(mobCmdShutdown);
	    }
	  if(sp != NULL)
	    {
	      spCmdShutdown.serial_number=sp_serial_number++;
	      sp->cmdBuf->write(spCmdShutdown);
	    }
	  if(eff != NULL)
	    {
	      effCmdShutdown.serial_number=eff_serial_number++;
	      eff->cmdBuf->write(effCmdShutdown);
	    }
	  for(i=0;i<mis.size();i++)
	    {
	      misCmdShutdown.serial_number=mis_serial_number++;
	      mis[i].cmdBuf->write(misCmdShutdown);
	    }
	} 
      else if (!nocase_cmp (toks[0], "help"))
	{
	  adminHelp();
	  return;
	}
      else if (toks[0][0] == '?')
	{
	  adminHelp();
	  return;
	}
      else
	{
	  printf("Command %s not recognized\n",toks[0].c_str());
	  adminHelp();
	  return;
	} 
    }
  adminStatView();
}

void cleanUp()
{
  if( mob != NULL )
    delete mob;
  if( sp != NULL ) 
    delete sp;
  if( eff != NULL)
    delete eff;
  mis.clear();
}


void runServoShell() 
{
  char * adminPrompt = "servoAdmin> ";
  char * mobPrompt = "servoMob> ";
  char * misPrompt = "servoMis> ";
  char * spPrompt = "servoSP> ";
  char * effPrompt = "servoEff> ";
  char * prompt = adminPrompt;
  string line;
  vector<string> toks;
  unsigned int i;
  void (*viewFunctionPtr)(const vector<string>&)=&doAdminView;
  
  while (!feof(stdin)) {
    
#ifndef HAVE_READLINE_READLINE_H
    char s[128];
    printf(prompt);
    fflush(stdout);
    cin.getline(s,128,'\n');
    line=s;
#else
    line = readline(prompt);
    if (!line.empty()) {
      add_history(line.c_str());
    }
#endif
    
    toks.clear();
    Tokenize(line,toks);
    
    if(toks.empty())
      {}
    else if (!nocase_cmp (toks[0], "quit"))
      {
	cleanUp();
	exit(1);
      }
    else if (!nocase_cmp (toks[0], "q"))
      {
	cleanUp();
	exit(1);
      }
    else if (!nocase_cmp (toks[0], "admin"))
      {
	prompt = adminPrompt;
	viewFunctionPtr=&doAdminView;
	toks.erase(toks.begin());
      }
    else if (!nocase_cmp (toks[0], "mob"))
      {
	if(mob == NULL)
	  continue;
	prompt = mobPrompt;
	viewFunctionPtr=&doMobView;
	toks.erase(toks.begin());
      }
    else if (!nocase_cmp (toks[0], "mis"))
      {
	if( mis.size() == 0 )
	  continue;
	prompt = misPrompt;
	viewFunctionPtr=&doMisView;
	toks.erase(toks.begin());
      }
    else if (!nocase_cmp (toks[0], "sp"))
      {
	if( sp == NULL)
	  continue;
	prompt = spPrompt;
	viewFunctionPtr=&doSPView;
	toks.erase(toks.begin());
      }  
    else if (!nocase_cmp (toks[0], "eff"))
      {
	if(eff == NULL)
	  continue;
	prompt = effPrompt;
	viewFunctionPtr=&doEffView;
	toks.erase(toks.begin());
      }
    else
      {}

    if(mob != NULL )
      {
	mob->cmdBuf->read();
	mob->statBuf->read();
	mob->setBuf->read();
      }
    if(sp != NULL)
      {
	sp->statBuf->read();
	sp->setBuf->read();
      }
    if(eff != NULL)
      {
	eff->statBuf->read();
	eff->setBuf->read();
      }
    for(i=0;i<mis.size();i++)
      {
	mis[i].cmdBuf->read();
	mis[i].statBuf->read();	
	mis[i].setBuf->read();
      }
    (*viewFunctionPtr)(toks);  
  }
  
}

int main(int argc, char *argv[])
{
 
  string nmlFile=DEFAULT_NML_FILE;
  string iniFile=DEFAULT_INI_FILE;
  int botNum=1;
  
  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      nmlFile = optarg;
      break;
    case 'i':
      iniFile = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &botNum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  (void) ulapi_init(UL_USE_DEFAULT);
  
  // open mobility channels
  mob =  new ServoNodeNmlBuffers;
  if( ! mob->openNML(nmlFile,
		    SERVO_MOB_JA_CMD_NAME,SERVO_MOB_JA_STAT_NAME,
		    SERVO_MOB_JA_CFG_NAME,SERVO_MOB_JA_SET_NAME,  
		    servoMobJA_format,botNum))
    {
      printf("** ERROR **  Error Initializing MOB NML Buffers\n");  
      return false;
    }
  
  // Read settings to see what channels are valid
  mob->setBuf->read();
  if( ((ServoMobJASet *)mob->setMsg)->senCount > 0 )
    {
      sp = new ServoNodeNmlBuffers;
      if( ! sp->openNML(nmlFile,
			SERVO_SP_CMD_NAME,SERVO_SP_STAT_NAME,
			SERVO_SP_CFG_NAME,SERVO_SP_SET_NAME,  
			servoSP_format,botNum))
	{
	  printf("** ERROR **  Error Initializing SP NML Buffers\n");  
	  return false;
	}
    }
  
  
  
  if( ((ServoMobJASet *)mob->setMsg)->effCount > 0 )
    {
      eff = new ServoNodeNmlBuffers;
      if( ! eff->openNML(nmlFile,
			 SERVO_EFF_JA_CMD_NAME,SERVO_EFF_JA_STAT_NAME,
			 SERVO_EFF_JA_CFG_NAME,SERVO_EFF_JA_SET_NAME,  
			 servoEffJA_format,botNum,
			 /*
			   Since we've changed the effector NML
			   buffers so that we can have more than
			   one per vehicle, we need to use the
			   1_1 suffix conventions. However, we'll
			   still only use one of them in servoShell.
			  */
			 1))
	{
	  printf("** ERROR **  Error Initializing EFF NML Buffers\n");  
	}
    }
  
  
  if( ((ServoMobJASet *)mob->setMsg)->misCount > 0 )
    for(int i=0;i<((ServoMobJASet *)mob->setMsg)->misCount;i++)
      {
	mis.push_back(*new ServoNodeNmlBuffers);
	if( ! mis[mis.size()-1].openNML(nmlFile,
					SERVO_MIS_JA_CMD_NAME,SERVO_MIS_JA_STAT_NAME,
					SERVO_MIS_JA_CFG_NAME,SERVO_MIS_JA_SET_NAME,  
					servoMisJA_format,botNum,i+1))
	  {
	    printf("** ERROR **  Error Initializing MIS NML Buffers\n");  
	    mis.pop_back();
	    break;
	  }
      }
  set_rcs_print_destination(RCS_PRINT_TO_NULL);
  joystickDataBuf = new NML(joystickData_format, 
			    "joystickData",
			    PROCESS_NAME,
			    nmlFile.c_str());
  if (! joystickDataBuf->valid()) {
    printf("** ERROR **  Error Initializing Joystick Data Buffer\n");  
    delete joystickDataBuf;
    joystickDataBuf = NULL;
    joystickDataPtr = NULL;
  } else {
    joystickDataPtr = (JoystickData *) (joystickDataBuf->get_address());
  }

  set_rcs_print_destination(RCS_PRINT_TO_STDOUT);

  runServoShell();

  if (NULL != joystickDataBuf) {
    delete joystickDataBuf;
  }
  joystickDataBuf = NULL;
  joystickDataPtr = NULL;

  return 0;
}

ServoNodeNmlBuffers::ServoNodeNmlBuffers() 
{
  cmdBuf=NULL;
  statBuf=NULL;
  cfgBuf=NULL;
  setBuf=NULL;
}
bool ServoNodeNmlBuffers::openNML(string nmlFile,
				  string cmdName, 
				  string statName, 
				  string cfgName,
				  string setName,
				  NML_FORMAT_PTR fPtr,
				  int botNum,
				  int compID) {
  stringstream str; 
  char suffix[10];
  
  // Check to see if there is a component id, this changes naming convention
  if(compID==0)
    sprintf(suffix,"%d",botNum);
  else
    sprintf(suffix,"%d_%d",botNum,compID);
  
  str << cmdName << suffix;
  
  if(1)
    printf("Opening NML CMD Buffer  : %s\n", str.str().c_str());
  
  cmdBuf = new RCS_CMD_CHANNEL(fPtr, 
			       str.str().c_str(), 
			       PROCESS_NAME, 
			       nmlFile.c_str());

  if (!cmdBuf->valid()) 
    {
      delete cmdBuf;
      cmdBuf = NULL;
      return false;
    }
  
  cmdMsg= cmdBuf->get_address();
  str.str("");
  
  
  str << cfgName << suffix;

  if(1)
    printf("Openning NML CFG Buffer  : %s\n", str.str().c_str());
  
  cfgBuf = new RCS_CMD_CHANNEL(fPtr, 
			       str.str().c_str(), 
			       PROCESS_NAME,
			       nmlFile.c_str());
  if (!cfgBuf->valid()) 
    {
      delete cfgBuf;
      cfgBuf = NULL;
      return false;
    }
  str.str("");
  
  str << statName << suffix;
  
  if(1)
    printf("Openning NML STAT Buffer  : %s\n", str.str().c_str());
  
  statBuf = new RCS_STAT_CHANNEL(fPtr, 
				 str.str().c_str(), 
				 PROCESS_NAME,
				 nmlFile.c_str());
  if (!statBuf->valid()) 
    {
      delete statBuf;
      statBuf = NULL;
      return false;
    }
  statMsg= (RCS_STAT_MSG_V2 *)statBuf->get_address();
  str.str("");
  
  str << setName << suffix;
  
  if(1)
    printf("Openning NML SET Buffer  : %s\n", str.str().c_str());
  
  setBuf = new RCS_STAT_CHANNEL(fPtr, 
				str.str().c_str(), 
				PROCESS_NAME,
				nmlFile.c_str());
  if (!setBuf->valid()) 
    {
      delete setBuf;
      setBuf = NULL;
      return false;
    }
  setMsg= setBuf->get_address();
  str.str("");

 
  
  
  return true;
}

void ServoNodeNmlBuffers::cleanUpNMLBuf()
{
  if (NULL != cmdBuf)
    delete cmdBuf;
  cmdBuf=NULL;
  
  if (NULL != statBuf)
    delete statBuf;
  statBuf=NULL;
  
  if (NULL != setBuf)
    delete setBuf;
  setBuf=NULL;
  
  if (NULL != cfgBuf)
    delete cfgBuf;
  cfgBuf=NULL;
  
} 

