/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file servoShell.cc

  \brief Command line shell to vehicle servo level NML buffers.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.15 $
  $Date: 2007/08/23 13:07:06 $
  \endcode

  \author Chris Scrapper
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv(), atoi(), strtol()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <math.h>
#include <string>
#include <iostream>
#include <vector>
#include <sstream>
using namespace std;

#include <rcs.hh>
#include "sensorData.hh"
#include "navDataExt.hh"
#include "servoSP.hh"	

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"
#define PROCESS_NAME          "slamStub"
#define DEG(x) (57.295779513082323 * (x))

class ServoNodeNmlBuffers
{
public:
  RCS_CMD_CHANNEL  *cmdBuf;
  RCS_STAT_CHANNEL *statBuf;
  RCS_CMD_CHANNEL  *cfgBuf;
  RCS_STAT_CHANNEL *setBuf;
  RCS_STAT_MSG_V2 * statMsg;
  RCS_STAT_MSG * setMsg;
  ServoNodeNmlBuffers();
  bool openNML(string nmlCfgFile, 
	       string cmdName, 
	       string statName, 
	       string cfgName,
	       string setName,
	       NML_FORMAT_PTR fPtr,
	       int botNum,
	       int compID=0);
  void cleanUpNMLBuf();
};

ServoNodeNmlBuffers sp;
NML *navDataExtBuf = NULL;
NML *servoSPNavBuf = NULL;
NavDataExt navDataExt;
SensorDataArray * servoSPNav;
int openNML(string nmlFile, int botNum)
{
  stringstream str; 
 
  if( ! sp.openNML(nmlFile,
		   SERVO_SP_CMD_NAME,SERVO_SP_STAT_NAME,
		   SERVO_SP_CFG_NAME,SERVO_SP_SET_NAME,  
		   servoSP_format,botNum))
    {
      printf("** ERROR **  Error Initializing SP NML Buffers\n");  
      return false;
    }
  
  str << NAV_DATA_EXT_NAME <<  botNum;
  if(0)
    printf("Opening NML Buffer : %s\n", str.str().c_str());
  navDataExtBuf = new NML(navDataExt_format, 
			  str.str().c_str(),
			  PROCESS_NAME,
			  nmlFile.c_str());
  if (!navDataExtBuf->valid()) 
    {
      delete navDataExtBuf;
      navDataExtBuf = 0;
      return false;
    }
  str.str("");
  

  str << SERVO_SP_NAV_BUF_NAME <<  botNum << "_1";;
  if(0)
    printf("Opening NML Buffer : %s\n", str.str().c_str());
  servoSPNavBuf = new NML(sensorData_format, 
			str.str().c_str(),
			PROCESS_NAME,
			nmlFile.c_str());
  
  if (!servoSPNavBuf->valid()) 
    {
      delete servoSPNavBuf;
      servoSPNavBuf = 0;
      return false;
    }
  servoSPNav=(SensorDataArray *)servoSPNavBuf->get_address();
  str.str("");
  return true;
}

void closeNML()
{
  sp.cleanUpNMLBuf();
  if(navDataExtBuf!=NULL)
    delete navDataExtBuf;
  navDataExtBuf=NULL;
  if(servoSPNavBuf!=NULL)
    delete servoSPNavBuf;
  servoSPNavBuf=NULL;
}

void doNavView()
{
  printf("\033[2J");
  printf("=================================\n");
  printf("*****    NAV DATA EXT       *****\n");
  printf("=================================\n");
  printf("time (sec):         %0.2lf\n", navDataExt.time);
  printf("UTM:                %d %c\n", (int) navDataExt.utmZone, navDataExt.utmLetter);
  printf("tranAbs:            %f %f %f\n", (double) navDataExt.tranAbs.x,
	 (double) navDataExt.tranAbs.y, (double) navDataExt.tranAbs.z);
  printf("rpyAbs:             %f %f %f\n", (double) DEG(navDataExt.rpyAbs.r),
	 (double) DEG(navDataExt.rpyAbs.p), (double) DEG(navDataExt.rpyAbs.y));
  printf("tranRel:            %f %f %f\n", (double) navDataExt.tranRel.x,
	 (double) navDataExt.tranRel.y, (double) navDataExt.tranRel.z);
  printf("rpyRel:             %f %f %f\n", (double) DEG(navDataExt.rpyRel.r),
	 (double) DEG(navDataExt.rpyRel.p), (double) DEG(navDataExt.rpyRel.y));
  printf("tranTrue:           %f %f %f\n", (double) navDataExt.tranTrue.x,
	 (double) navDataExt.tranTrue.y, (double) navDataExt.tranTrue.z);
  printf("rpyTrue:            %f %f %f\n", (double) DEG(navDataExt.rpyTrue.r),
	 (double) DEG(navDataExt.rpyTrue.p), (double) DEG(navDataExt.rpyTrue.y));
  printf("tranAbsRate:        %f %f %f\n", (double) navDataExt.tranAbsRate.x,
	 (double) navDataExt.tranAbsRate.y, (double) navDataExt.tranAbsRate.z);
  printf("tranRelRate:        %f %f %f\n", (double) navDataExt.tranRelRate.x,
	 (double) navDataExt.tranRelRate.y, (double) navDataExt.tranRelRate.z);
  printf("rpyAbsRate:         %f %f %f\n", (double) navDataExt.rpyAbsRate.r,
	(double) navDataExt.rpyAbsRate.p, (double) navDataExt.rpyAbsRate.y);
  printf("rpyRelRate:         %f %f %f\n", (double) navDataExt.rpyRelRate.r,
	 (double) navDataExt.rpyRelRate.p, (double) navDataExt.rpyRelRate.y);

  printf("speed:              %f\n", (double) navDataExt.speed);
}

void run(int debug)
{
  RCS_TIMER timer(0.02);
  int type;
  PM_CARTESIAN utmOffset;
  double delta_t;
  PM_CARTESIAN xyz;
  PM_RPY rpy;

  double t,x,y,theta;

  while(1)
    {
      sp.statBuf->read();
      type=sp.setBuf->read();
      switch (type)
	{
	case 0:
	  break;
	case -1:
	  rcs_print_error("NML cmd read error\n");
	  break;
	case SERVO_SP_SET_TYPE:
	  for(int i=0;i<((ServoSPSet *)sp.setMsg)->sSetElem_length;i++)
	    {
	      switch(((ServoSPSet *)sp.setMsg)->sSetElem[i].sType)
		{
		case SERVO_SP_GRD_TRUTH_SENSOR:
		  navDataExt.utmZone=(int)((ServoSPSet *)sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_ZONE];
		  navDataExt.utmLetter=(char)((ServoSPSet *)sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_LETTER];
		  utmOffset=PM_CARTESIAN(((ServoSPSet *)sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_NORTHING],
					 ((ServoSPSet *)sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_EASTING],
					 ((ServoSPSet *)sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_Z_VALUE]);
		  break;
		case SERVO_SP_ODOMETRY_SENSOR:
		  break;
		default:
		  break;
		}
	    }
	  break;
	default:
	  break;
	}
      
      type = servoSPNavBuf->read();
      switch (type)
	{
	case 0:
	  break;
	case -1:
	  rcs_print_error("NML cmd read error\n");
	  break;
	case SENSOR_DATA_ARRAY_TYPE:
	  for(int i=0;i<servoSPNav->arrayElem_length;i++)
	    {
	      switch(servoSPNav->arrayElem[i].type)
		{
		case SERVO_SP_GRD_TRUTH_SENSOR:
		  delta_t=servoSPNav->arrayElem[i].collectionTime-navDataExt.time;
		  navDataExt.time=servoSPNav->arrayElem[i].collectionTime;
		  
		  xyz=PM_CARTESIAN(servoSPNav->arrayElem[i].dataTuple[0],
				   servoSPNav->arrayElem[i].dataTuple[1],
				   servoSPNav->arrayElem[i].dataTuple[2]);
		  rpy=PM_RPY(servoSPNav->arrayElem[i].dataTuple[3],
			     servoSPNav->arrayElem[i].dataTuple[4],
			     servoSPNav->arrayElem[i].dataTuple[5]);
		  
		  navDataExt.tranRelRate=(xyz - navDataExt.tranRel)/delta_t;
		  navDataExt.rpyRelRate=(rpy - navDataExt.rpyRel)/delta_t;
		  navDataExt.tranRel=xyz;
		  navDataExt.rpyRel=rpy;
		  
		  navDataExt.tranAbs=utmOffset + xyz;
		  navDataExt.rpyAbs=rpy;
		  
		  navDataExt.tranAbsRate=navDataExt.tranRelRate;
		  navDataExt.rpyAbsRate=navDataExt.rpyRelRate;
		  
		  navDataExt.tranTrue=navDataExt.tranAbs;
		  navDataExt.rpyTrue=navDataExt.rpyAbs;
		  navDataExt.tranTrueRate=navDataExt.tranAbsRate;
		  navDataExt.rpyTrueRate=navDataExt.rpyAbsRate;
		  navDataExt.absToRel = PM_POSE(navDataExt.tranAbs,navDataExt.rpyAbs) * inv(PM_POSE(navDataExt.tranRel,navDataExt.rpyRel));
		  
		  navDataExt.speed=sqrt(navDataExt.tranRelRate.x*navDataExt.tranRelRate.x +
					navDataExt.tranRelRate.y*navDataExt.tranRelRate.y +
					navDataExt.tranRelRate.z*navDataExt.tranRelRate.z );
		  
		  navDataExt.steeringCurvature=0;
		  navDataExt.steeringCrabAngle=0;
		  /* write navData Here for now */
		  navDataExtBuf->write(navDataExt); 
		  break;
		case SERVO_SP_ODOMETRY_SENSOR:
		  t=servoSPNav->arrayElem[i].collectionTime;
		  x=servoSPNav->arrayElem[i].dataTuple[0];
		  y=servoSPNav->arrayElem[i].dataTuple[1];
		  theta=servoSPNav->arrayElem[i].dataTuple[2];
		  break;
		}
	    }
	  

	  if (debug & 1) {
	    // print nav data
#if 0	 
	  printf("%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf \n", 
		 navDataExt.time, 
		 delta_t, 
		 navDataExt.tranRel.x, 
		 navDataExt.tranRel.y,
		 navDataExt.rpyRel.y,
		 navDataExt.rpyRelRate.y,
		 navDataExt.speed,
		 t,
		 x,
		 y,
		 theta);
	  //#else
	    printf("%lf %lf %lf %lf %lf %lf \n", 
		   navDataExt.time, 
		   navDataExt.tranRel.x, 
		   navDataExt.tranRel.y,
		   t,
		   x,
		   y);
#endif
	  }

	  if (debug == ~0) {
	    // only if verbose
	    doNavView();
	  }
	  
	  break;
	default:
	  break;
	}
      timer.wait();
    }
}


int main(int argc, char *argv[])
{
 
  string nmlFile=DEFAULT_NML_FILE;
  string iniFile=DEFAULT_INI_FILE;
  int botNum=1;
  long int debug=1;

  while (1) {
    char option = getopt(argc, argv, "i:g:v");
    if (option == -1)
      break;

    switch (option) {
    case 'i':
      botNum = atoi(optarg);
      break;
    case 'g':
      // set some debug bits, see 'debug' lines below for bit meanings
      debug = strtol(optarg, NULL, 0);
      break;
    case 'v':
      // verbose means set all debug bits
      debug = ~0;
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }
  
  if(! openNML(nmlFile,botNum) )
    {
     printf("** ERROR ** <slamStub::main> error while opening NML buffers\n");
     exit(-1);
    }
  
  nml_start();
  run(debug);
  closeNML();
  return 0;
}

ServoNodeNmlBuffers::ServoNodeNmlBuffers() 
{
  cmdBuf=NULL;
  statBuf=NULL;
  cfgBuf=NULL;
  setBuf=NULL;
}
bool ServoNodeNmlBuffers::openNML(string nmlFile,
				  string cmdName, 
				  string statName, 
				  string cfgName,
				  string setName,
				  NML_FORMAT_PTR fPtr,
				  int botNum,
				  int compID) {
  stringstream str; 
  char suffix[10];
  
  // Check to see if there is a component id, this changes naming convention
  if(compID==0)
    sprintf(suffix,"%d",botNum);
  else
    sprintf(suffix,"%d_%d",botNum,compID);
  
  str << cmdName << suffix;
  
  if(0)
    printf("Opening NML CMD Buffer  : %s\n", str.str().c_str());
  
  cmdBuf = new RCS_CMD_CHANNEL(fPtr, 
			       str.str().c_str(), 
			       PROCESS_NAME, 
			       nmlFile.c_str());
  if (!cmdBuf->valid()) 
    {
      delete cmdBuf;
      cmdBuf = NULL;
      return false;
    }
  str.str("");
  
  
  str << cfgName << suffix;

  if(0)
    printf("Opening NML CFG Buffer  : %s\n", str.str().c_str());
  
  cfgBuf = new RCS_CMD_CHANNEL(fPtr, 
			       str.str().c_str(), 
			       PROCESS_NAME,
			       nmlFile.c_str());
  if (!cfgBuf->valid()) 
    {
      delete cfgBuf;
      cfgBuf = NULL;
      return false;
    }
  str.str("");
  
  str << statName << suffix;
  
  if(0)
    printf("Opening NML STAT Buffer  : %s\n", str.str().c_str());
  
  statBuf = new RCS_STAT_CHANNEL(fPtr, 
				 str.str().c_str(), 
				 PROCESS_NAME,
				 nmlFile.c_str());
  if (!statBuf->valid()) 
    {
      delete statBuf;
      statBuf = NULL;
      return false;
    }
  statMsg= (RCS_STAT_MSG_V2 *)statBuf->get_address();
  str.str("");
  
  str << setName << suffix;
  
  if(0)
    printf("Opening NML SET Buffer  : %s\n", str.str().c_str());
  
  setBuf = new RCS_STAT_CHANNEL(fPtr, 
				str.str().c_str(), 
				PROCESS_NAME,
				nmlFile.c_str());
  if (!setBuf->valid()) 
    {
      delete setBuf;
      setBuf = NULL;
      return false;
    }
  setMsg= setBuf->get_address();
  str.str("");
  return true;
}

void ServoNodeNmlBuffers::cleanUpNMLBuf()
{
  if (NULL != cmdBuf)
    delete cmdBuf;
  cmdBuf=NULL;
  
  if (NULL != statBuf)
    delete statBuf;
  statBuf=NULL;
  
  if (NULL != setBuf)
    delete setBuf;
  setBuf=NULL;
  
  if (NULL != cfgBuf)
    delete cfgBuf;
  cfgBuf=NULL;
} 

