/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I.
*****************************************************************************/

import java.io.*;
import skel.*;
//import rcs.*;
import rcs.nml.*;


/**
 *
 *  Brief example of writing an NML buffer
 *
 *  @author Dr. Stephen Balakisky
 *  @author Steve Tran
 *
 *  @version 1.1   2009/01/10 04:59:52 Added java examples of nmlRead and
 *      nmlWrite. Fixed makefile to make these examples and added run scripts
 *      to run them.
 *  @version 1.2   2009/01/13 15:49:55  Got rid of nmlprint in favor of printf
 *  @version 1.3   2009/01/18 13:42:49  Fixed nmlRead and nmlWrite
 *  @version 1.5   2009/01/18 14:04:15  Fixed the parse of double (yet again).
 *      Don't know why it did not stick the last time.
 *  @version 2.0   2009/01/25 20:00:00  Refactored code to follow Java Coding
 *      Standards.  Fixed indentation and spacing of code to improve
 *      readability.  Reduced the amount of logic in the main method by
 *      creating several helper methods to make the code more modular.
 */
public class nmlWrite {

    private static String DEFAULT_NML_FILE = "./skel.nml";
    private static String SKEL_CMD_NAME = "skelCmd"; // NML cmd buffer name
    private static String THIS_PROCESS = "nmlWrite"; // NML server process name
    private static NMLConnection skel_cmd_buf;


    /*
     *  Initialization of the NML buffers.  The code in here was provided
     *   to me.  The only modification I made was I changed its signature to
     *   return a boolean instead or 1 or 0.
     *
     *  @param config_file The specific configuration file to use
     *  @return true if successful, false if errored
     */
    private static boolean init_nml_buffers(String config_file) {

        boolean success = false;

        try {
            skel_cmd_buf = new NMLConnection(
                new skel_MsgDict(), SKEL_CMD_NAME, THIS_PROCESS, config_file );
            success = true;
        } catch(NMLException nml_except) {
            System.err.println("Cannot open nml buffer.");
        }

        return success;
    }


    /*
     * Writes the shutdown command to the buffer, then exits the program.
     *
     */
    private static void shutdown() {
        skel_cmd_buf.write(new SkelCmdShutdown());
        System.exit(0);
    }


    /*
     * Sets up the configurations for the application.  If the user specified
     *  their own NML file, then use that if and only if there is one parameter
     *  on the command line.  Else, use the default file.
     *
     * If the NML file does not exist, then an exit code of 1 is returned.
     */
    private static void config(String[] params) {

        String NML_FILE = DEFAULT_NML_FILE;

        if (params.length == 1) {
            NML_FILE = params[0];
            System.out.println("using specified config file " + NML_FILE);
        }
        else {
            System.out.println("using default config file " + NML_FILE);
        }


        if(!new File(NML_FILE).exists()) {
            System.out.println("file " + NML_FILE + " does not exist");
            System.exit(1);
        }


        if (!init_nml_buffers(NML_FILE)) {
            System.out.println("can't allocate NML buffers");
            return;
        }
    }


    /*
     * Handles the actual processing of input.  The user must enter a G for
     *  GO commands 
     *
     * Most of this code was broken out from the main method from the original
     *  code provided.  I broke it out into its own method to make it more
     *  modular in design.  Also added more defensive programming techniques
     *  which may negatively impact performance, but increases reliability.
     *
     * @throws IOException Whenever an I/O Exception occurs, it's up to the
     *      caller to handle it because the original code didn't care.
     */
    private static void process() throws IOException {

        String curLine;
            BufferedReader in = new BufferedReader(
                new InputStreamReader(System.in));

        for(;;) {

            // Ask the user for what type of command they want,
            // won't continue unless a 'G' or 'S' is typed in
            System.out.println("GO Command (g) ?");
            curLine = in.readLine();
            if(curLine == null) {
                curLine = "";
            }


            if("g".equalsIgnoreCase(curLine.trim())) {

                SkelCmdGo goCmd = new SkelCmdGo();
                System.out.print("GO Command\nWhere: ");

                try {
                    goCmd.where = (curLine = in.readLine()) == null ?
                        -1.0 : Double.parseDouble(curLine);
                } catch (NumberFormatException nfe) {
                    System.out.println("Invalid double : using default of 0");
                    goCmd.where = 0;
                }

                // if a negative number is typed in, quit the program
                if( goCmd.where < 0 ) {
                    shutdown();
                }

                skel_cmd_buf.write(goCmd);

            }

        } //end for

    }


    /**
     *  Main Driver for the Write program.  Broken up into pieces to make the
     *   code more readable.
     *
     *  @param args The arguments passed in through the command line.
     */
    public static void main(String[] args)  {

        System.out.println("nmlWrite Java Version!");
        System.out.println("usage: please run skelSvr before running this program.");
        System.out.println("Next, run this program (nmlWrite) and nmlRead.");
        System.out.println("Type a value <CR> and the value should appear at the nmlRead application.");
        System.out.println("Type a negative value to exit.\n\n");

        config(args);   //sets up all our buffers, won't return if errored


        try {
            process();
        } catch (IOException ioe) {
            ioe.printStackTrace();
            System.exit(1);
        }

    }

}//end class

/*
  Modification history:

  $Log: nmlWrite.java,v $
  Revision 1.7  2009/02/24 19:40:31  dr_steveb
  Java code should now work!

  Revision 1.6  2009/02/13 16:47:45  wshackle
  uses the rcs_states in rcs.jar rcs.utils package instead of the local one if the rcslib version is greater than 2009.01.22, uses time_tracker.cycle() and time_tracker.reset() which have been available for a long time instead of timetracker

  Revision 1.5  2009/01/18 14:04:15  dr_steveb
  Fixed the parse of double (yet again). Don't know why it did not stick the last time.

  Revision 1.3  2009/01/18 13:42:49  dr_steveb
  Fixed nmlRead and nmlWrite

  Revision 1.2  2009/01/13 15:49:55  dr_steveb
  Got rid of nmlprint in favor of printf

  Revision 1.1  2009/01/10 04:59:52  dr_steveb
  Added java examples of nmlRead and nmlWrite. Fixed makefile to make these examples and added run scripts to run them.

*/
