/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   exampleScrollBox.cc
  \brief  Contains an example of using ScrollingBox
*/

#include "exampleScrollBox.hh"
#include "../common/MOASTScrollingBox.hh"

#include <stdio.h>

/***********************************************************************/

/*

This file contains an example of how to use the ScrollingBox class.
Applications should use the first two #include lines shown above, in
the order shown, with exampleScrollBox.hh changed to the name of a
header file for the application. See exampleScrollBox.hh,
scrollingBox.hh, and scrollingBox.cc for more documentation.

This example:
1. Builds a 33 by 33 by 33 scrolling box.
2. Prints a text representation of the box showing three standard
   2D views (XY, YZ to the left of XY, XZ below XY). Each view shows
   the voxels passed through by a line from the start point to the
   end point given by the command line arguments.
3. Prints the ids of the voxels passed through and how many there are.
4. Scrolls the box, so that the center is moved from (0,0) to (12,9).
5. Repeats steps 2 and 3.

The example also prints two files (voxel_picture and floor_picture)
that may be read and displayed by the draw3D executable. These represent
the original voxels and the line

The start point must be on the box. The end point may be off the box.

Note that the first voxel on a line is obtained by a call to initLine,
while the rest of the voxels on the line are obtained by calls to
getNextVoxel. When getNextVoxel returns a NULL pointer, the end of
the line has been reached.

*/

/***********************************************************************/

/*! printBoxWithLine prints a text picture representation of three
    orthographic views (top, front, and right) of the
    box and the voxels on it through which a line passes. */

void printBoxWithLine( /*   ARGUMENTS                           */
 ScrollingBox * box,   /*!< the box to use                      */
 double startX,        /*!< X coordinate of start point of line */
 double startY,        /*!< Y coordinate of start point of line */
 double startZ,        /*!< Z coordinate of start point of line */
 double endX,          /*!< X coordinate of end point of line   */
 double endY,          /*!< Y coordinate of end point of line   */
 double endZ)          /*!< Z coordinate of end point of line   */
{
  char top[33][33];
  char side[33][33];
  char front[33][33];
  int xIndex;
  int yIndex;
  int zIndex;
  int n;
  exampleVoxel * aVoxel;

  aVoxel = box->initLine(startX, startY, startZ, endX, endY, endZ);
  if (!aVoxel)
    {
      printf("Start point is outside box\n");
      return;
    }
  aVoxel->item = 'X';
  for (; ;)
    {
      aVoxel = box->getNextVoxel();
      if (aVoxel == 0)
	break;
      aVoxel->item = 'X';
    }
  for (xIndex = 0; xIndex < 33; xIndex++) // prepare top view
    {
      for (yIndex = 0; yIndex< 33; yIndex++)
	{
	  for (zIndex = 0; zIndex < 33; zIndex++ )
	    {
	      if (box->getVoxelAtIndexes(xIndex, yIndex, zIndex)->item == 'X')
		break;
	    }
	  top[xIndex][yIndex] = ((zIndex == 33) ? '.' : 'X');
	}
    }
  for (xIndex = 0; xIndex < 33; xIndex++) // prepare front view
    {
      for (zIndex = 0; zIndex < 33; zIndex++)
	{
	  for (yIndex = 0; yIndex < 33; yIndex++ )
	    {
	      if (box->getVoxelAtIndexes(xIndex, yIndex, zIndex)->item == 'X')
		break;
	    }
	  front[xIndex][zIndex] = ((yIndex == 33) ? '.' : 'X');
	}
    }
  for (yIndex = 0; yIndex < 33; yIndex++) // prepare side view
    {
      for (zIndex = 0; zIndex < 33; zIndex++)
	{
	  for (xIndex = 0; xIndex < 33; xIndex++ )
	    {
	      if (box->getVoxelAtIndexes(xIndex, yIndex, zIndex)->item == 'X')
		break;
	    }
	  side[yIndex][zIndex] = ((xIndex == 33) ? '.' : 'X');
	}
    }
  
  printf("\n");
  for (n = 32; n > -1; n--) // print top and side views
    {
      for (xIndex = 0; xIndex < 33; xIndex++)
	{
	  printf("%c ", top[xIndex][n]);
	}
      printf("   ");
      for (yIndex = 0; yIndex < 33; yIndex++)
	{
	  printf("%c ", side[yIndex][n]);
	}
      printf("\n");
    }
  printf("\n");
  for (zIndex = 32; zIndex > -1; zIndex--) // print front view
    {
      for (xIndex = 0; xIndex < 33; xIndex++)
	{
	  printf("%c ", front[xIndex][zIndex]);
	}
      printf("\n");
    }
  printf("\n");
}

/***********************************************************************/

void printVoxel3D (
 FILE * out,
 double cx,
 double cy,
 double cz,
 double res)
{
  fprintf(out, "polygon\n");
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy, cz+res);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy, cz+res);
  fprintf(out, "polygon\n");
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy+res, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy+res, cz+res);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy, cz+res);
  fprintf(out, "polygon\n");
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy+res, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy+res, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy+res, cz+res);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy+res, cz+res);
  fprintf(out, "polygon\n");
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy+res, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy, cz+res);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy+res, cz+res);
  fprintf(out, "polygon\n");
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy+res, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy+res, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy, cz);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy, cz);
  fprintf(out, "polygon\n");
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy, cz+res);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy, cz+res);
  fprintf(out, "%.4f %.4f %.4f\n", cx+res, cy+res, cz+res);
  fprintf(out, "%.4f %.4f %.4f\n", cx, cy+res, cz+res);
}

/***********************************************************************/

void printBoxAndLine(  /*   ARGUMENTS                           */
 ScrollingBox * box,   /*!< the box to use                      */
 double startX,        /*!< X coordinate of start point of line */
 double startY,        /*!< Y coordinate of start point of line */
 double startZ,        /*!< Z coordinate of start point of line */
 double endX,          /*!< X coordinate of end point of line   */
 double endY,          /*!< Y coordinate of end point of line   */
 double endZ)          /*!< Z coordinate of end point of line   */
{
  FILE * out;
  exampleVoxel * aVoxel;
  double cx;             // X coordinate of corner of voxel
  double cy;             // Y coordinate of corner of voxel
  double cz;             // Z coordinate of corner of voxel
  double res;            // resolution of box

  out = fopen("pics3D/voxel_picture", "w");
  if (out == 0)
    {
      printf("could not open file pics3D/voxel_picture for writing\n");
      return;
    }
  aVoxel = box->initLine(startX, startY, startZ, endX, endY, endZ);
  if (!aVoxel)
    {
      printf("Start point is outside box\n");
      return;
    }
  fprintf(out, "voxels on a line\n\n");
  fprintf(out, "data\n");
  fprintf(out, "line\n");
  fprintf(out, "%.4f %.4f %.4f\n", startX, startY, startZ);
  fprintf(out, "%.4f %.4f %.4f\n", endX, endY, endZ);
  res = box->getResolution();
  cx = (box->getXCoordAtVoxel(aVoxel) - (res * 0.5));
  cy = (box->getYCoordAtVoxel(aVoxel) - (res * 0.5));
  cz = (box->getZCoordAtVoxel(aVoxel) - (res * 0.5));
  printVoxel3D(out, cx, cy, cz, res);
  for (aVoxel = box->getNextVoxel(); aVoxel; aVoxel = box->getNextVoxel())
    {
      cx = (box->getXCoordAtVoxel(aVoxel) - (res * 0.5));
      cy = (box->getYCoordAtVoxel(aVoxel) - (res * 0.5));
      cz = (box->getZCoordAtVoxel(aVoxel) - (res * 0.5));
      printVoxel3D(out, cx, cy, cz, res);
    }
  fprintf(out, "end\n");
  fclose(out);
}

/***********************************************************************/

/*! printLineData prints the ids of the voxels through which a line passes
    and prints how many of those voxels there are. */

void printLineData(  /*   ARGUMENTS                           */
 ScrollingBox * box, /*!< the box to use                      */
 double startX,      /*!< X coordinate of start point of line */
 double startY,      /*!< Y coordinate of start point of line */
 double startZ,      /*!< Z coordinate of start point of line */
 double endX,        /*!< X coordinate of end point of line   */
 double endY,        /*!< Y coordinate of end point of line   */
 double endZ)        /*!< Z coordinate of end point of line   */
{
  int n;
  exampleVoxel * aVoxel;

  aVoxel = box->initLine(startX, startY, startZ, endX, endY, endZ);
  if (!aVoxel)
    {
      printf("Start point is outside box\n");
      return;
    }
  printf("%d ", aVoxel->getId());
  for (n = 1; ; n++)
    {
      aVoxel = box->getNextVoxel();
      if (aVoxel == 0)
	break;
      printf("%d ", aVoxel->getId());
    }
  printf("\n%d voxels\n", n);
}

/***********************************************************************/

int main(       /*   ARGUMENTS                        */
 int argc,      /*!< number of command line arguments */
 char * argv[]) /*!< text of command line arguments   */
{
  ScrollingBox * box;
  double startX;
  double startY;
  double startZ;
  double endX;
  double endY;
  double endZ;
  int xDiff;
  int yDiff;
  int zDiff;

  if (argc != 7)
    {
      printf("Usage: %s <startX> <startY> <startZ> <endX> <endY> <endZ>\n",
	     argv[0]);
      return 1;
    }
  if (1 != sscanf(argv[1], "%lf", &startX))
    {
      printf("Bad first argument %s\n", argv[1]);
      return 1;
    }
  if (1 != sscanf(argv[2], "%lf", &startY))
    {
      printf("Bad second argument %s\n", argv[2]);
      return 1;
    }
  if (1 != sscanf(argv[3], "%lf", &startZ))
    {
      printf("Bad third argument %s\n", argv[3]);
      return 1;
    }
  if (1 != sscanf(argv[4], "%lf", &endX))
    {
      printf("Bad fourth argument %s\n", argv[4]);
      return 1;
    }
  if (1 != sscanf(argv[5], "%lf", &endY))
    {
      printf("Bad fifth argument %s\n", argv[5]);
      return 1;
    }
  if (1 != sscanf(argv[6], "%lf", &endZ))
    {
      printf("Bad sixth argument %s\n", argv[6]);
      return 1;
    }
  printf("\nDraw and print voxel ids for line from "
	 "(%.2lf, %.2lf, %.2lf) to (%.2lf, %.2lf, %.2lf)\n",
	 startX, startY, startZ, endX, endY, endZ);
  box = new ScrollingBox(16, 16, 16, 1.0);
  printBoxWithLine(box, startX, startY, startZ, endX, endY, endZ);
  printBoxAndLine(box, startX, startY, startZ, endX, endY, endZ);
  printLineData(box, startX, startY, startZ, endX, endY, endZ);
  box->setCenter(9.0, 3.0, 1.0, &xDiff, &yDiff, &zDiff);
  printf("\nScroll the center from (0,0,0) to (9,3,1) and do the same line\n");
  printBoxWithLine(box, startX, startY, startZ, endX, endY, endZ);
  printLineData(box, startX, startY, startZ, endX, endY, endZ);

  return 0;
}

/***********************************************************************/

