/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file linkMonShell.cc

  \brief Command line shell to traffic simulator

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.2 $
  $Date: 2007/09/18 20:03:49 $
  \endcode

  \author Stephen Balakirsky
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "trafficCtrl.hh"
#include "trafficData.hh"
#include "linkMonCtrl.hh"

static RCS_CMD_CHANNEL *linkMonCmdNml = NULL;
static RCS_STAT_CHANNEL *linkMonStatNml = NULL;
static RCS_STAT_CHANNEL *trafficCtrlSetNml = NULL;
static NML *trafficDataNml = NULL;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "linkMonShell"

static int cleanupNmlBuffers(void)
{
  if (NULL != linkMonCmdNml)
    delete linkMonCmdNml;
  if (NULL != linkMonStatNml)
    delete linkMonStatNml;
  if (NULL != trafficDataNml)
    delete trafficDataNml;
  if (NULL != trafficCtrlSetNml)
    delete trafficCtrlSetNml;

  linkMonCmdNml = NULL;
  linkMonStatNml = NULL;
  trafficDataNml = NULL;
  trafficCtrlSetNml = NULL;

  return 0;
}

static int initNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupNmlBuffers();

  sprintf(chanName, "%s%d", LINK_MON_CMD_NAME, bufnum);
  linkMonCmdNml = new RCS_CMD_CHANNEL(linkMonCtrl_format, chanName,
				      THIS_PROCESS, config_file);
  if(!linkMonCmdNml->valid())
    {
      delete linkMonCmdNml;
      linkMonCmdNml = 0;
      return 1;
    }

  sprintf(chanName, "%s%d", LINK_MON_STAT_NAME, bufnum);
  linkMonStatNml = new RCS_STAT_CHANNEL(linkMonCtrl_format, chanName,
					THIS_PROCESS, config_file);
  if(!linkMonStatNml->valid()) 
    {
      delete linkMonStatNml;
      linkMonStatNml = 0;
      return 1;
    }

  sprintf(chanName, "%s%d", TRAFFIC_DATA_NAME, bufnum);
  trafficDataNml = 
    new NML(trafficData_format, chanName,
	    THIS_PROCESS, config_file);
  if (!trafficDataNml->valid()) {
    delete trafficDataNml;
    trafficDataNml = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", TRAFFIC_CTRL_SET_NAME, bufnum);
  trafficCtrlSetNml =
    new RCS_STAT_CHANNEL(trafficCtrl_format, chanName,
			 THIS_PROCESS, config_file);
  if (!trafficCtrlSetNml->valid()) {
    delete trafficCtrlSetNml;
    trafficCtrlSetNml = 0;
    return 1;
  }

  return 0;
}

#define rcs_stat_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

static void printStat(void)
{
  int count, count2;
  LinkMonStat * stat = (LinkMonStat *) linkMonStatNml->get_address();
  TrafficData *data = (TrafficData *) trafficDataNml->get_address();
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();

  printf("command_type:       %s\n",
	 linkMonCtrl_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  //  printf("state:              %s\n", rcsVehStateToString(stat->state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("heartbeat:          %d\n", (int) stat->tt.count);
  printf("actual cycleTime:   %lf\n",stat->tt.avg);
  printf("time:               %f\n", data->time);
  printf("data for %d vehicles\n",   data->vehicles_length );
  printf("xLoc:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].tranTrue.x );
    }
  printf("\nyLoc:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].tranTrue.y );
    }
  printf("\nzLoc:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].tranTrue.z );
    }
  printf("\nbotName:");
  for( count=0; count<set->robotInfo_length; count++ )
    {
      printf( "\t%s", set->robotInfo[count].robotName );
    }
  printf( "\n" );
  for( count=0; count<stat->route_length; count++ )
    {
      printf( "route: " );
      for( count2=0; count2<stat->route[count].length; count2++ )
	{
	  printf( "-> %s ", stat->route[count].name[count2] );
	}
      printf( "with cost: %f\n", stat->route[count].cost );
    }
  /*
  printf("\nroll:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].rpyTrue.r );
    }
  printf("\npitch:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].rpyTrue.p );
    }
  printf("\nyall:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].rpyTrue.y );
    }
  printf("\nspeed:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].speed );
    }
  printf("\nsteer:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].steeringCurvature );
    }
  printf("\npIndex:");
  for( count=0; count<stat->pathIndex_length; count++ )
    {
      printf( "\t%d", stat->pathIndex[count] );
    }
  printf("\npState:");
  for( count=0; count<stat->primState_length; count++ )
    {
      printf( "\t%s", rcsPrimStateToString(stat->primState[count]) );
    }
  printf("\npStatus:");
  for( count=0; count<stat->primStatus_length; count++ )
    {
      printf( "\t%s", rcs_stat_to_string(stat->primStatus[count]) );
    }
  */
  printf( "\n" );
}

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}

static void printHelp(void)
{
  printf("q | quit            -- quit\n");
  printf("? | help            -- print this help\n");
  printf("init                -- initialize\n");
  printf("halt                -- halt\n");
  printf("abort               -- abort\n");
  printf("shutdown            -- shutdown\n");
  printf("monitor             -- monitor commands\n");
  printf("write <robot> <msg> -- write the message <msg> to robot <robot> over the wss\n");
  /*
  printf("ct <secs>         -- set cycle time to <secs>\n");
  printf("mvl <x> <y>       -- move to <x> <y> in local coordinates\n");
  printf("mvg <x> <y>       -- move to <x> <y> in global coordinates\n");
  printf("exp <x> <y>       -- explore while moving to local waypoint\n");
  printf("dump              -- dump world model and display with imageview\n");
  printf("debug <0, 1, ...> -- configure debug level\n");
  */
}

int main(int argc, char *argv[])
{
  int serialNumber = 1;
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char *line;
  char *ptr;
  int retval;
  LinkMonCmdNop linkMonCmdNop;
  LinkMonCmdInit linkMonCmdInit;
  LinkMonCmdAbort linkMonCmdAbort;
  LinkMonCmdHalt linkMonCmdHalt;
  LinkMonCmdShutdown linkMonCmdShutdown;
  LinkMonCmdMonitor linkMonCmdWaypoint;
  LinkMonCmdWrite linkMonCmdWrite;

#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      break;
    case 'i':
      INI_FILE = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // initial allocate all the buffers
  retval = initNmlBuffers(NML_FILE, bufnum);

  if (retval) {
    fprintf(stderr, "%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }

  while (!feof(stdin)) {
#ifndef HAVE_READLINE_READLINE_H
    printf("> ");
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) {
      break;
    }
    line = buffer;
#else
    line = readline("> ");
    if (NULL == line) {
      break;
    }
    if (*line) {
      add_history(line);
    }
#endif

    linkMonStatNml->read();
    //    linkMonSetNml->read();
    trafficDataNml->read();
    trafficCtrlSetNml->read();

    ptr = line;
    while (isspace(*ptr))
      ptr++;			// skip leading white space
    if (0 == *ptr) {		// blank line
      printStat();
    } else if (ptr[0] == 'q' || !strcmp(ptr, "quit")) {
      break;
    } else if (ptr[0] == '?' || !strcmp(ptr, "help")) {
      printHelp();
    } else if (!strwcmp(ptr, "init")) {
      linkMonCmdInit.serial_number = serialNumber;
      serialNumber+=2;
      linkMonCmdNml->write(&linkMonCmdInit);
    } else if (!strwcmp(ptr, "nop")) {
      linkMonCmdNop.serial_number = serialNumber;
      serialNumber+=2;
      linkMonCmdNml->write(&linkMonCmdNop);
    } else if (!strwcmp(ptr, "halt")) {
      linkMonCmdHalt.serial_number = serialNumber;
      serialNumber+=2;
      linkMonCmdNml->write(&linkMonCmdHalt);
    } else if (!strwcmp(ptr, "abort")) {
      linkMonCmdAbort.serial_number = serialNumber;
      serialNumber+=2;
      linkMonCmdNml->write(&linkMonCmdAbort);
    } else if (!strwcmp(ptr, "shutdown")) {
      linkMonCmdShutdown.serial_number = serialNumber;
      serialNumber+=2;
      linkMonCmdNml->write(&linkMonCmdShutdown);
    } else if (!strwcmp(ptr, "monitor")) {
      linkMonCmdWaypoint.serial_number = serialNumber;
      serialNumber+=2;
      linkMonCmdNml->write(&linkMonCmdWaypoint);
    } else if (!strwcmp(ptr, "write")) {
      if( 2 !=
	  sscanf(ptr, "%*s %s %s",
		 linkMonCmdWrite.robotToWrite, linkMonCmdWrite.msg ))
	printf( "need a robot name and message argument\n" );
      else
	{
	  linkMonCmdWrite.serial_number = serialNumber;
	  serialNumber+=2;
	  linkMonCmdNml->write(&linkMonCmdWrite);
	}

      /*
    } else if (!strwcmp(ptr, "wp")) {
      if (1 != sscanf(ptr, "%*s %i",
		      &linkMonCmdWaypoint.numVehicles)) {
	printf("need number of vehicles and file\n");
      }else{
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;                        // skip over space
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over num vehicles
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need a data file\n");
	  continue;
	}
	retVal = loadWaypoints(ptr,
			       &linkMonCmdWaypoint.waypoint[0][0],
			       &linkMonCmdWaypoint.waypoint_length[0]);
	if( retVal != 0 ){
	  printf("can't load ``%s''\n", ptr);
	}else{
	  for( count=1; count<linkMonCmdWaypoint.numVehicles; count++ )
	    {
	      for( count2=0;count2<linkMonCmdWaypoint.waypoint_length[0]; count2++ )
		{
		  linkMonCmdWaypoint.waypoint[count][count2] =
		    linkMonCmdWaypoint.waypoint[0][count2];
		}
	      linkMonCmdWaypoint.waypoint_length[count] = 
		linkMonCmdWaypoint.waypoint_length[0];
	    }

	  linkMonCmdNml->write(&linkMonCmdWaypoint);
	}
      }
      */
    } else {
      printf("?\n");
    }

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  return cleanupNmlBuffers();
}

/*
  Modification history:

  $Log: linkMonShell.cc,v $
  Revision 1.2  2007/09/18 20:03:49  dr_steveb
  Added write command

  Revision 1.1  2007/09/11 20:26:53  dr_steveb
  Initial check-in


*/

