/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file moastErrorLog.cc
  \brief Prints contents of queued moastError buffer

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.1 $
  $Date: 2006/10/19 21:01:18 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>		// printf, sscanf, feof, stdin
#include <signal.h>		// signal, SIGINT
#include <unistd.h>		// getopt, optarg,opt,err
#include "moastError.hh"

#define DEFAULT_NML_FILE "../etc/moast.nml"

#define BN "moastErrorLog"

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  char * nmlFile = DEFAULT_NML_FILE;
  NML * moastErrorBuf = NULL;
  MoastError * moastErrorPtr = NULL;
  int retval;

  retval = 0;
  opterr = 0;
  while (1) {
    char option = getopt(argc, argv, ":n:");
    if (option == -1)
      break;
    switch (option) {
    case 'n':
      nmlFile = optarg;
      break;
    case ':':
      fprintf(stderr, "%s: missing parameter to %c\n", BN, (char) optopt);
      retval = 1;
      break;
    default:			// '?'
      fprintf(stderr, "%s: unrecognized option %c\n", BN, (char) optopt);
      retval = 1;
      break;
    }
  }
  if (retval) return retval;

  moastErrorBuf = new NML(moastError_format, "moastError", BN, nmlFile);
  if (! moastErrorBuf->valid()) {
    fprintf(stderr, "%s: can't open moastError buffer\n", BN);
    return 1;
  }
  moastErrorPtr = (MoastError *) (moastErrorBuf->get_address());

  done = 0;
  signal(SIGINT, quit);

  while (! done) {
    NMLTYPE type;

    type = moastErrorBuf->read();

    switch (type) {
    case 0:
      // no new data
      break;
    case -1:
      // read error, probably buffer gone, so quit
      done = 1;
      break;
    case MOAST_ERROR_TYPE:
      // our data
      printf("%s\n", moastErrorPtr->message);
      break;
    default:
      // something else
      fprintf(stderr, "%s: unknown error message type: %d\n", BN, (int) type);
      break;
    }
  }

  delete moastErrorBuf;

  return 0;
}

/*
  Modification history:

  $Log: moastErrorLog.cc,v $
  Revision 1.1  2006/10/19 21:01:18  proctor
  Added moastErrorLog.cc and made related changes to NML and autotools files.

  Revision 1.1  2006/10/18 21:27:37  proctor
  Initial checkin

*/

