/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*!
  \file nmlPrint.cc
  \brief Reads and prints the specified NML buffer

  Synopsis: nmlPrint {--nml \<file\>} \n
                     {--period \<secs\>} \n
		     {--block} \n
		     --buffer,-b \<buffer name\> \n
		     --process,-p \<process name\>
  
  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.1 $
  $Date: 2005/12/07 21:36:40 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		// sscanf()
#include <string.h>		// strcmp()
#include <stddef.h>		// NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#include <sys/time.h>               // gettimeofday()
#ifdef HAVE_GETOPT_H
#include <getopt.h>		// getopt_long()
#else
#include "getopt.h"
#endif
#include <rcs.hh>		// NMTYPE, CMS
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_EXT_BASE, ...
#include "navDataExt.hh"	// navDataExt_format(), NavDataExt, ...
#include "sensorData.hh"	// sensorData_format(), SensorData1D, ...
#include "servoMobJA.hh"	// servoMobJA_format(), ServoMobJAStat, ...
#include "servoMisJA.hh"        // servoMisJA_format(), ServoMisJAStat, ...
#include "amMobJA.hh"           // amMobJA_format(), amMobJACmd, ...
#include "primMobJA.hh"         // primMobJA_format(), primMobJACmd, ..

#define DEFAULT_NML_FILE "../etc/moast.nml"

static int all_format(NMLTYPE type, void *buffer, CMS *cms)
{
  if (navDataExt_format(type, buffer, cms)) return 1;
  if (sensorData_format(type, buffer, cms)) return 1;
  if (servoMobJA_format(type, buffer, cms)) return 1;
  if (servoMisJA_format(type, buffer, cms)) return 1;
  if (amMobJA_format(type, buffer, cms)) return 1;
  if (primMobJA_format(type, buffer, cms)) return 1;
  return 0;
}

static void printHelp(void)
{
  printf("options:\n{--nml <file>}\n{--period <secs>}\n{--block}\n--buffer,-b <buffer name>\n--process,-p <process name>\n{--number,-n <buffer number>}\n{--help,-h}\n");
}

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  int c;
  int this_option_optind;
  int option_index;
  struct option long_options[] = {
    /* --option, needs_arg, 0, 'shortcut char' */
    {"nml", required_argument, 0, 0},
    {"period", required_argument, 0, 0},
    {"block", no_argument, 0, 0},
    {"buffer", required_argument, 0, 'b'},
    {"process", required_argument, 0, 'p'},
    {"number", required_argument, 0, 'n'},
    {"help", no_argument, 0, 'h'},
    {"notime", no_argument, 0, 0},
    {0, 0, 0, 0}
  };
  NML * nmlBuf = 0;
  NML * nmlBuf2 = 0;
  NMLTYPE type, type2;
  char * envptr = 0;
  char nmlFile[MOAST_FILE_NAME_LEN] = {0};
  char bufferName[MOAST_NML_BUFFER_NAME_LEN] = {0};
  char bufferName2[MOAST_NML_BUFFER_NAME_LEN] = {0};
  char processName[MOAST_NML_BUFFER_NAME_LEN] = {0};
  char bufNumString[MOAST_NML_BUFFER_NAME_LEN] = {0};
  int bufNum = 1;
  double period = -1.0;		// negative = just go once
  int doBlock = 0;
  int printTimestamp = 1;
  int t, u;
  double th;
  timeval tv;
  double timeOut;
  
  opterr = 0;			// inhibit getopt_long's error printing
  while (1) {
    this_option_optind = optind ? optind : 1;
    option_index = 0;

    // The leading colon means generate a ':' for missing parameters.
    // Colons mean a required argument for the preceding option.
    c = getopt_long(argc, argv, ":b:p:n:h",
		    long_options, &option_index);
    if (c == -1)
      break;

    switch (c) {
      // this handles --long args
    case 0:		
      if (!strcmp("nml", long_options[option_index].name)) {
	strncpy(nmlFile, optarg, MOAST_FILE_NAME_LEN);
	nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
      } else if (!strcmp("period", long_options[option_index].name)) {
	if (1 != sscanf(optarg, "%lf", &period)) {
	  fprintf(stderr, "%s: bad paramter for --period: `%s'\n",
		  argv[0], optarg);
	}
      } else if (!strcmp("block", long_options[option_index].name)) {
	doBlock = 1;
      } else if (!strcmp("notime", long_options[option_index].name)) {
	printTimestamp = 0;
      } else {
	// should never get here, since 0 above means OK
	fprintf(stderr, "%s: unknown option `%s'",
		long_options[option_index].name);
	if (optarg)
	  fprintf(stderr, " with arg %s", optarg);
	fprintf(stderr, "\n");
      }
      break;

      // the rest handle -short args

    case 'b':
      strncpy(bufferName, optarg, MOAST_NML_BUFFER_NAME_LEN - 1);
      break;

    case 'p':
      strncpy(processName, optarg, MOAST_NML_BUFFER_NAME_LEN - 1);
      break;

    case 'n':
      if (1 != sscanf(optarg, "%i", &bufNum)) {
	printf("%s: bad number parameter: ``%s''\n", argv[0],  optarg);
	return 1;
      }
      break;

    case 'h':
      printHelp();
      return 0;
      break;

    case '?':
      fprintf(stderr, "%s: unknown option `%s'\n",
	      argv[0], argv[this_option_optind]);
      return 1;
      break;

    case ':':
      fprintf(stderr, "%s: missing parameter to `%s'\n",
	      argv[0], argv[this_option_optind]);
      return 1;
      break;

    default:
      fprintf(stderr, "%s: bad argument `%s'\n",
	      argv[0], argv[this_option_optind]);
      return 1;
      break;
    }
  }

  strncpy(processName, "primSP", MOAST_NML_BUFFER_NAME_LEN - 1);
  strncpy(bufferName, "servoSPLinescan", MOAST_NML_BUFFER_NAME_LEN - 1);
  strncpy(bufferName2, "navDataExt", MOAST_NML_BUFFER_NAME_LEN - 1);


  if (optind < argc) {
    fprintf(stderr, "%s: extra arguments provided:", argv[0]);
    while (optind < argc)
      fprintf(stderr, " %s", argv[optind++]);
    fprintf(stderr, "\n");
    return 1;
  }
  if (bufferName[0] == 0) {
    fprintf(stderr, "%s: missing the --buffer,-b option\n", argv[0]);
    return 1;
  }
  if (processName[0] == 0) {
    fprintf(stderr, "%s: missing the --process,-p option\n", argv[0]);
    return 1;
  }

  // OK, we made it through the arguments

  if (0 == nmlFile[0]) {
    if (NULL == (envptr = getenv("CONFIG_NML"))) {
      strncpy(nmlFile, DEFAULT_NML_FILE, MOAST_FILE_NAME_LEN);
      nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
    } else {
      strncpy(nmlFile, envptr, MOAST_FILE_NAME_LEN);
      nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
    }
  }

  sprintf(bufNumString, "%d", bufNum);
  strcat(bufferName, bufNumString);
  strcat(bufferName2, bufNumString);

  nmlBuf = new NML(all_format, bufferName, processName, nmlFile);
  if (!nmlBuf->valid()) {
    delete nmlBuf;
    fprintf(stderr, "can't connect to NML buffer\n");
    return 1;
  }

  nmlBuf2 = new NML(all_format, bufferName2, processName, nmlFile);
  if (!nmlBuf2->valid()) {
    delete nmlBuf2;
    fprintf(stderr, "can't connect to NML buffer\n");
    return 1;
  }

  // set up some shortcut pointers
  NavDataExt *navDataExt = (NavDataExt *) nmlBuf2->get_address();
  SensorData1D *sensorData1D = (SensorData1D *) nmlBuf->get_address();
  SensorData2D *sensorData2D = (SensorData2D *) nmlBuf->get_address();
  SensorDataList *sensorDataList = (SensorDataList *) nmlBuf->get_address();
  ServoMobJAStat *servoMobJAStat = (ServoMobJAStat *) nmlBuf->get_address();
  ServoMisJAStat *servoMisJAStat = (ServoMisJAStat *) nmlBuf->get_address();
  AmMobJACmdMoveWaypoint *amMoveWp = 
    (AmMobJACmdMoveWaypoint *)nmlBuf->get_address();
  PrimMobJACmdMoveWaypoint *primMoveWp = 
    (PrimMobJACmdMoveWaypoint *)nmlBuf->get_address();

  done = 0;
  signal(SIGINT, quit);
  while (!done) {
    if (doBlock) {
      type = nmlBuf->blocking_read(1.0);
      type2 = nmlBuf2->blocking_read(1.0);
    } else {
      type = nmlBuf->read();
      type2 = nmlBuf2->read();
    }
    gettimeofday( &tv, NULL );
    timeOut = (double)tv.tv_sec + (double)tv.tv_usec/1000000; 
    if (printTimestamp) printf( "Timestamp: %lf\n", timeOut );
    for( int ii=0; ii<2; ii++ )
      {
	if( ii == 1 )
	  type = type2;
    switch (type) {
    case 0:
      break;

    case NAV_DATA_EXT_TYPE:
      /*
      printf("%d %f %f %f %f %f %f ",
	     navDataExt->time,
	     navDataExt->tranAbs.x, 
	     navDataExt->tranAbs.y,
	     navDataExt->tranAbs.z,
	     navDataExt->tranRel.x,
	     navDataExt->tranRel.y,
	     navDataExt->tranRel.z);
      printf("%f %f %f %f %f %f ",
	     navDataExt->rpyAbs.r,
	     navDataExt->rpyAbs.p,
	     navDataExt->rpyAbs.y,
	     navDataExt->rpyRel.r,
	     navDataExt->rpyRel.p,
	     navDataExt->rpyRel.y);
      printf("%f %f %f\n",
	     navDataExt->tranRelRate.x, 
	     navDataExt->tranRelRate.y,
	     navDataExt->tranRelRate.z);
      */
      printf("%d %f %f %f %f %f %f\n",
	     navDataExt->time,
	     navDataExt->tranAbs.x, 
	     navDataExt->tranAbs.y,
	     navDataExt->tranAbs.z,
	     navDataExt->rpyAbs.r,
	     navDataExt->rpyAbs.p,
	     navDataExt->rpyAbs.y);
      break;

    case SENSOR_DATA_1D_TYPE:
      printf( "MOAST SCAN %d\n", sensorData1D->featArray_length );
      for (t = 0, th = sensorData1D->start;
	   t < sensorData1D->featArray_length;
	   t++, th += sensorData1D->increment) {
	printf("%f %f\n", th, sensorData1D->featArray[t]);
      }
      break;
      
    case SENSOR_DATA_LIST_TYPE:
      printf("%d %lf %lf %d\n", 
	     sensorDataList->isGlobal,
	     sensorDataList->maxRange,
	     sensorDataList->minRange,
	     sensorDataList->featList_length);
      for (t = 0; t < sensorDataList->featList_length; t++ )
	printf("%lf %lf %lf %lf %lf %lf %lf\n", 
	       sensorDataList->featList[t].dataLoc.x,
	       sensorDataList->featList[t].dataLoc.y,
	       sensorDataList->featList[t].dataLoc.z,
	       sensorDataList->featList[t].sensorLoc.x,
	       sensorDataList->featList[t].sensorLoc.y,
	       sensorDataList->featList[t].sensorLoc.z,
	       sensorDataList->featList[t].param1);

      break;

    case SENSOR_DATA_2D_TYPE:
      printf( "P3\n#%lf %lf %lf %lf\n%d %d\n255\n", sensorData2D->lowerLeft.x, 
	      sensorData2D->lowerLeft.y,
	      sensorData2D->incrementX, sensorData2D->incrementY,
	      sensorData2D->x, sensorData2D->y);
      for (t=0; t<sensorData2D->x; t++ )
	{
	  for (u=0; u<sensorData2D->y; u++ )
	    {
	      /*
	      printf("%lf ", sensorData2D->featArray[t*sensorData2D->x+u]);
	      */
	      if( sensorData2D->featArray[t*sensorData2D->x+u] == 0 )
		printf( "0 0 0 " );
	      else if( sensorData2D->featArray[t*sensorData2D->x+u] <= .21 )
		printf( "128 128 128" );
	      else if( sensorData2D->featArray[t*sensorData2D->x+u] <= .41 )
		printf( "254 254 0" );
	      else 
		printf( "254 0 0" );
	      printf( " " );
	    }
	  printf( "\n" );
	}
      fflush(stdout);
      break;

    case SERVO_MOB_JA_STAT_TYPE:
      printf("%d\n", servoMobJAStat->echo_serial_number);
      printf("%d\n", servoMobJAStat->heartbeat);
      break;

    case SERVO_MIS_JA_STAT_TYPE:
      printf("%lf %lf %lf ", 
	     servoMisJAStat->mountToVeh.tran.x,
	     servoMisJAStat->mountToVeh.tran.y,
	     servoMisJAStat->mountToVeh.tran.z);
      printf("%lf %lf %lf %lf\n", 
	     servoMisJAStat->mountToVeh.rot.s,
	     servoMisJAStat->mountToVeh.rot.x,
	     servoMisJAStat->mountToVeh.rot.y,
	     servoMisJAStat->mountToVeh.rot.z);
      break;

    case AM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
      printf( "x y z neighborhood speed quality %d waypoints\n",
	      amMoveWp->waypoint_length);
      for (t=0; t<amMoveWp->waypoint_length; t++ )
	{
	  printf("%lf %lf %lf %lf %lf %d\n", 
		 amMoveWp->waypoint[t].p.x,
		 amMoveWp->waypoint[t].p.y,
		 amMoveWp->waypoint[t].p.z,
		 amMoveWp->waypoint[t].neighborhood,
		 amMoveWp->waypoint[t].speed,
		 amMoveWp->waypoint[t].quality);
	}
      break;
    case PRIM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
      printf( "x y z neighborhood speed quality %d waypoints\n",
	      primMoveWp->waypoint_length);
      for (t=0; t<primMoveWp->waypoint_length; t++ )
	{
	  printf("%lf %lf %lf %lf %lf %d\n", 
		 primMoveWp->waypoint[t].p.x,
		 primMoveWp->waypoint[t].p.y,
		 primMoveWp->waypoint[t].p.z,
		 primMoveWp->waypoint[t].neighborhood,
		 primMoveWp->waypoint[t].speed,
		 primMoveWp->waypoint[t].quality);
	}
      break;
    default:
      break;
    }
      }
    if (period < 0.0) {
      break;
    } else {
      esleep(period);
    }
  }

  delete nmlBuf;
  delete nmlBuf2;

  return 0;
}

/*
  Modification history:

  $Log: nmlPrintRome.cc,v $
  Revision 1.1  2005/12/07 21:36:40  dr_steveb
  Modified version of nmlPrint to print multiple buffer's information into a single file for use in visualization of data.

  Revision 1.2  2005/10/20 20:59:14  dr_steveb
  Changed sensor2D buffer to output ppm file.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.13  2005/10/03 15:05:58  stephen
  Added printing of prim waypoint command

  Revision 1.12  2005/09/30 19:42:42  stephen
  Added printing of amMobJACmd

  Revision 1.11  2005/09/30 18:55:56  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.10  2005/09/30 14:31:19  stephen
  Added additional fields to sensor_2D printing

  Revision 1.9  2005/09/30 13:58:13  proctor
  Added --notime flag to inhibit printing of timestamp

  Revision 1.8  2005/09/29 13:16:37  stephen
  Made changes to work with sensorPoint changes.

  Revision 1.7  2005/09/21 00:15:20  stephen
  Added sensor2D printing.

  Revision 1.6  2005/09/15 21:17:16  stephen
  Fixed gettimeofday to have NULL parameter.

  Revision 1.5  2005/09/14 14:10:03  stephen
  Added additional channels for printing

  Revision 1.4  2005/07/26 21:15:55  proctor
  Changed some printing formats

  Revision 1.3  2005/07/25 22:41:19  proctor
  Printed tranRelRate

  Revision 1.2  2005/07/07 19:04:41  proctor
  Removed navPrint.cc, it's now nmlPrint.cc

  Revision 1.1  2005/07/06 22:27:36  proctor
  Initial checkin

*/
