/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file pathGen.cc

  \brief File for creating prim arcs and lines

   Lines are output as <x> <y> <tol> <speed>
   Arcs are output as <center-x> <center-y> <end-x> <end-y> <theta> <tol> <speed>

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.1 $
  $Date: 2008/04/17 20:33:40 $
  \endcode

  \author Stephen Balakirsky
  \date   2008-04-01
*/

/*! \ingroup Tools 

*/

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <stdio.h>   /* Standard input/output definitions */
#include <moastTypes.hh>
#include <moastGeom.hh> // lineSegment

#define LINE_LENGTH 10

typedef enum LINE_DIRECTION {
  NORTH,
  EAST,
  SOUTH,
  WEST
}LineDirection;

/* writes out a line and returns the new end point */
Location2D drawLine( FILE *fpOut, const Location2D &current, LineDirection lineDirection, double tol, double speed )
{
  Location2D retValue = current;

  switch( lineDirection )
    {
    case NORTH:
      retValue.north += LINE_LENGTH;
      break;
    case EAST:
      retValue.east += LINE_LENGTH;
      break;
    case SOUTH:
      retValue.north -= LINE_LENGTH;
      break;
    case WEST:
      retValue.east -= LINE_LENGTH;
      break;
    }
  fprintf( fpOut, "%6.2f %6.2f %5.2f %5.2f\n", retValue.north, retValue.east, tol, speed ); 
  return retValue;
}

/*
  endpoint of startLine will contain offset arc center.
 */
Location2D drawArc( FILE *fpOut, const Location2D &current, double radianOfArc, 
		    LineSegment &startLine, double offset, double curvature, int directionToMove,
		    double tol, double speed )
{
  LineSegment newCenter;
  Location2D end;
  double theta;

  newCenter = startLine.moveLineParallel( offset, directionToMove );
  theta = atan2( (startLine.end.east-newCenter.end.east), (startLine.end.north-newCenter.end.north) );
  theta += radianOfArc;
  end.north = curvature * cos(theta) + newCenter.end.north;
  end.east = curvature * sin(theta) + newCenter.end.east;
  fprintf( fpOut, "%6.2f %6.2f %6.2f %6.2f %5.2f %5.2f %5.2f\n", 
	   newCenter.end.north, newCenter.end.east,
	   end.north, end.east,
	   radianOfArc,
	   tol, speed );
  return end;
}

Location2D drawPattern( FILE *fpOut, const Location2D &startLoc, double tol, double speed )
{
  Location2D newEnd = startLoc;
  LineSegment newCenter;
  LineSegment temp;
  double curvature;

  newCenter.start = newEnd;
  newEnd = drawLine(fpOut, newEnd, NORTH, tol, speed );

  newCenter.end = newEnd;
  curvature = 2;
  newEnd = drawArc( fpOut, newEnd, M_PI_2, newCenter, curvature, curvature, -1, tol, speed );

  newCenter.start = newEnd;
  newEnd = drawLine(fpOut, newEnd, EAST, tol, speed );

  newCenter.end = newEnd;
  curvature = 1;
  newEnd = drawArc( fpOut, newEnd, -M_PI_2, newCenter, curvature, curvature, 1, tol, speed );

  temp = newCenter.moveLineParallel( curvature, 1 );
  newCenter.start = newCenter.end;
  newCenter.end = temp.end;
  newEnd = drawArc( fpOut, newEnd, M_PI_2, newCenter, 2*curvature, curvature, -1, tol, speed );

  newCenter.start = newEnd;
  newEnd = drawLine(fpOut, newEnd, EAST, tol, speed );

  newCenter.end = newEnd;
  curvature = .5;
  newEnd = drawArc( fpOut, newEnd, -M_PI, newCenter, curvature, curvature, 1, tol, speed );

  curvature = .25;
  newEnd = drawArc( fpOut, newEnd, M_PI_2, newCenter, 1+curvature, curvature, 1, tol, speed );

  return newEnd;
}

int main(int argc, char *argv[])
{
  char *fileOut = "out.dat";
  FILE *fpOut;
  char c;
  Location2D startLoc;
  double tol;
  double vehWidth = 0; //1.1;

  
  while((c=getopt(argc,argv,"o:"))!=EOF)
    {
      switch(c)
	{
	case 'o':
	  fileOut = optarg;
	  break;
	}
    }
  if (NULL == (fpOut = fopen(fileOut, "w")))
    {
      printf( "unable to open file: out.dat \n" );
      return 1;
    }

  // output header stuff
  fprintf( fpOut, "# Lines are lines of <x> <y> <tol> <speed>\n# Arcs are lines of <center-x> <center-y> <end-x> <end-y> <theta> <tol> <speed>\n# Length units are meters\n# Angle units (for <theta>) are in radians\n\n# absolute coordinates\nabs\n" );



  startLoc.north = -5;
  startLoc.east = 5;
  tol = 2. + vehWidth;
  startLoc = drawLine( fpOut, startLoc, NORTH, tol, .8 );
  startLoc = drawPattern( fpOut, startLoc, tol, .8 );
  tol = 1. + vehWidth;
  startLoc = drawPattern( fpOut, startLoc, tol, .8 );
  tol = .5 + vehWidth;
  startLoc = drawPattern( fpOut, startLoc, tol, .8 );
  fclose( fpOut );
}
