/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   cluster.cc
  \brief  provides clusters of points from a world model map

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.7 $
  $Date: 2007/11/08 13:22:30 $
  \endcode

  \author Stephen Balakirsky
  \date   02/12/2007
*/
#include "cluster.hh"
#include "CImg.h"
///////////////////////////////////////////////////////////////////////////////
/*!
  compute the distance between two points
  \param pt1 Point 1 of the two points
  \param pt2 Point 2 of the two points
  \return The distance between the points
 */
double myDist(PM_CARTESIAN &pt1, PM_CARTESIAN &pt2 )
{
  return( sqrt((pt1.x-pt2.x)*(pt1.x-pt2.x)+(pt1.y-pt2.y)*(pt1.y-pt2.y)) );
}

///////////////////////////////////////////////////////////////////////////////
//! comparison function used by qsort
/*! This function will sort by  x-coordinate
    If the x-coordinates are equal, then sort by
    y-coordinates. 
    \param vN1 item 1 to compare
    \param vN2 item 2 to compare
    \return -1 if vN1<vN2, 0 if they are equal, and 1 if vN1>vN2
*/
int ClusterMemberCompare( const void *vN1, const void *vN2 )
{
  ClusterMember *n1 = (ClusterMember*)vN1;
  ClusterMember *n2 = (ClusterMember*)vN2;

  if( n1->xLoc > n2->xLoc )
    return 1;
  if( n1->xLoc < n2->xLoc )
    return -1;
  if( n1->yLoc > n2->yLoc )
    return 1;
  if( n1->yLoc < n2->yLoc )
    return -1;
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
//! Implements the less than operator
/*!
  This function first compares the x-coordinate.
  If the x-coordinates are equal, then it compares the y-coordinate
  \param comp The item to be compared to
  \return Returns true if the item being compared is less than the instance
 */
bool ClusterMember::operator<(const ClusterMember *comp)
{
  if( comp == NULL )
    return false;
  if( comp->xLoc < this->xLoc )
    return true;
  if( comp->xLoc > this->xLoc )
    return false;
  if( comp->yLoc < this->yLoc )
    return true;
  return false;
}

///////////////////////////////////////////////////////////////////////////////
//! Constructor
/*!
  Sets the maxNumClusters to an arbitary number. Each cluster will eventually become
  a polygon
 */
ClusterHead::ClusterHead()
{
  maxNumClusters = 250;
  numClusters = 0;
  clusterEnds = new ClusterMember*[maxNumClusters];
}

///////////////////////////////////////////////////////////////////////////////
//! Destructor
ClusterHead::~ClusterHead()
{
  //  delete [] clusterEnds;
}

///////////////////////////////////////////////////////////////////////////////
//! Check to see if we can increase the number of reported clusters
/*!
  Really does nothing except try to increas the numClusters variable
  \return True on success
          False on failure
 */
bool ClusterHead::incNumClusters()
{
  numClusters++;
  if( numClusters == maxNumClusters )
    {
      numClusters--;
      return false;
    }
  return true;
}

///////////////////////////////////////////////////////////////////////////////
//! Check for polygon intersection
void ClusterHead::checkIntersection( int cluster1, int cluster2 )
{
  Sweep sweep;
  LineSegment l1;
  ClusterMember *currentMember;
  LineSegment working; // working storage
  Location2D savePoint; // need to save the first segment for use 
  int count;

  for( count=0; count<2; count++ )
    {
      if( count== 0 )
	working.properties = cluster1;
      else
	working.properties = cluster2;
      // set up values and save first line segment
      currentMember = clusterEnds[working.properties];
      working.end.north = currentMember->xLoc;
      working.end.east = currentMember->yLoc;
      if( currentMember->previous==NULL )
	return;
      currentMember = currentMember->previous;
      savePoint = working.end;
      while( currentMember->previous!=NULL )
	{
	  working.start = working.end;
	  working.end.north = currentMember->xLoc;
	  working.end.east = currentMember->yLoc;
	  currentMember = currentMember->previous;
	  sweep.segments.push_back(working);
	}
      working.start = working.end;
      working.end.north = currentMember->xLoc;
      working.end.east = currentMember->yLoc;
      sweep.segments.push_back(working);
      working.start = working.end;
      working.end = savePoint;
      sweep.segments.push_back(working);
    }
  printf( "Lines into intersect algorithm:\n" );
  vector<LineSegment>::iterator k;
  for( k=sweep.segments.begin(); k!=sweep.segments.end(); k++ )
    {
      (*k).print();
    }
  sweep.anyIntersections();

}

///////////////////////////////////////////////////////////////////////////////
//! Expand a single polygon
/*!
  expands the given polygon the the amout specified. For now, polygons must
  be defined in a clockwise fashion
  \param clusterID The cluster to expand
  \param dist The distance to move the points
 */
void ClusterHead::expandPolygon( int clusterID, double dist )
{
  ClusterMember *currentMember;
  ClusterMember *changePointer; // will change the element that this points to
  LineSegment saveSegment; // need to save the first segment for use 
                           // in determining the last segment
  LineSegment working1, working2; // working storage
  Location2D tempPoint;
  int direction;
  //  double startDist, endDist;
  
  currentMember = clusterEnds[clusterID];
  changePointer = clusterEnds[clusterID];
  

  /* let's not do this yet... problems with using the centroid
  // compute the centroid
  tempPoint.north = 0;
  tempPoint.east = 0;
  direction = 0;
  while( currentMember->previous!=NULL )
    {
      tempPoint.north += currentMember->xLoc;
      tempPoint.east += currentMember->yLoc;
      direction++;
      currentMember = currentMember->previous;
    }
  tempPoint.north += currentMember->xLoc;
  tempPoint.east += currentMember->yLoc;
  direction++;
  tempPoint.north /= direction;
  tempPoint.east /= direction;
// end of unused code
  */

  // set up values and save first line segment
  currentMember = clusterEnds[clusterID];
  working2.start.north = currentMember->xLoc;
  working2.start.east = currentMember->yLoc;
  if( currentMember->previous==NULL )
    return;
  currentMember = currentMember->previous;
  changePointer = changePointer->previous;
  working2.end.north = currentMember->xLoc;
  working2.end.east= currentMember->yLoc;
  
  saveSegment = working2;


  // first figure out the direction to move to expand the polygon
  /* don't use yet
  startDist = working2.distToPoint( tempPoint );
  */
  // shift the first line  
  direction = 1;
  working2.moveLineParallel(dist, direction);
  /* don't use yet
  endDist = working2.distToPoint( tempPoint );
  //if the direction is correct, then we should be farther from the centroid
  if( endDist < startDist ) // wrong way
    {
      direction = -1;
      working2 = saveSegment;
      working2.moveLineParallel(dist, direction);
    }
  */
  while( currentMember->previous!=NULL )
    {
      working1 = working2;
      working2.start.north = currentMember->xLoc;
      working2.start.east = currentMember->yLoc;
      currentMember = currentMember->previous;
      working2.end.north = currentMember->xLoc;
      working2.end.east = currentMember->yLoc;
      // shift the new line  
      working2.moveLineParallel(dist, direction);

      // find the intersection of working1 and working2
      tempPoint = working1.intersect(working2);
      changePointer->xLoc = tempPoint.north;
      changePointer->yLoc = tempPoint.east;
      changePointer = changePointer->previous;

    }
  // now find last line and last two vertices
  working1 = working2;
  working2.start.north = changePointer->xLoc;
  working2.start.east = changePointer->yLoc;
  working2.end = saveSegment.start;
  // shift the new line 
  working2.moveLineParallel(dist, direction);
  // find the intersection of working1 and working2
  tempPoint = working1.intersect(working2);
  changePointer->xLoc = tempPoint.north;
  changePointer->yLoc = tempPoint.east;
  // now find the last vertix
  changePointer = clusterEnds[clusterID];
  working1 = working2;
  working2 = saveSegment;
  // shift the new line 
  working2.moveLineParallel(dist, direction);
  // find the intersection of working1 and working2
  tempPoint = working1.intersect(working2);
  changePointer->xLoc = tempPoint.north;
  changePointer->yLoc = tempPoint.east;
}

///////////////////////////////////////////////////////////////////////////////
//! Print a single cluster
/*!
  Note that this function is overloaded
  Prints a single cluster's members. Each point is printed out as: <y x 0>.
 */
void ClusterHead::print(int clusterID, FILE *filePt)
{
  ClusterMember *currentMember;
  int howMany;

  currentMember = clusterEnds[clusterID];
  howMany = 1;
  // figure out how many elements
  while( currentMember->previous!=NULL )
    {
      howMany++;
      currentMember = currentMember->previous;
    }
  // print out header
  fprintf( filePt, "Region 1\n  %d\n", howMany );
  currentMember = clusterEnds[clusterID];
  // print out vertices
  while( currentMember->previous!=NULL )
    {
      fprintf( filePt, "%lf %lf\n", currentMember->xLoc, currentMember->yLoc );
      currentMember = currentMember->previous;
    }
  fprintf( filePt, "%lf %lf\n", currentMember->xLoc, currentMember->yLoc );
  // print out footer
  fprintf( filePt, "    Pen(11, 2, 8421504)\n    Brush(2, 12632256, 8421504)\n" );
}

///////////////////////////////////////////////////////////////////////////////
//! print all of the clusters
/*!
  Note that this function is overloaded
  Prints the first point of all of the clusters in the form <x y>. 
 */
void ClusterHead::print(FILE *filePt)
{
  for( int count=0; count<numClusters; count++ )
    if( clusterEnds[count] != NULL )
      fprintf(filePt, "%lf %lf\n", clusterEnds[count]->xLoc, clusterEnds[count]->yLoc);
}

///////////////////////////////////////////////////////////////////////////////
//! Constructor for a cluster
/*!
  Initializes the memory space for a cluster. Initially set to 250 elements.
  If needed, this will automatically be increased.
  \param gapSizeIn A gap of larger than this will cause a new cluster to be formed
 */
Cluster::Cluster(int gapSizeIn)
{
  xRawSize = 0;
  yRawSize = 0;
  rawData = NULL;
  hullDataIn = new PM_CARTESIAN [250];
  hullDataOut = new PM_CARTESIAN [250];
  hullDataTemp = new PM_CARTESIAN [250];
  hullDataSize = 250;
  gapSize = gapSizeIn;
}

///////////////////////////////////////////////////////////////////////////////
//! Destructor
Cluster::~Cluster()
{
  deleteRaw();
  //  delete [] hullDataIn;
  //  delete [] hullDataOut;
  hullDataSize = 0;
}

///////////////////////////////////////////////////////////////////////////////
void Cluster::buildHulls()
{
  int numPts, printCount;
  FILE *clusterLines, *clusterPoints;

  clusterLines = fopen( "clusterLines.dat", "w" );
  clusterPoints = fopen( "clusterPoints.dat", "w" );

  for(int count=0; count<clusterHead.getNumClusters(); count++)
    {
      if( clusterHead.clusterEnds[count] != NULL )
	{
	  clusterHead.print(count, clusterPoints);
	  //	  printf( "building hull for cluster %d\n", count );
	  numPts = buildHull(clusterHead.clusterEnds[count], gapSize);
	  for( printCount=0; printCount<numPts; printCount++ )
	    {
	      fprintf( clusterLines, "%f %f\n", hullDataOut[printCount].y, 
		       hullDataOut[printCount].x );
	    }
	  fprintf( clusterLines, "%f %f\n\n", hullDataOut[0].y, hullDataOut[0].x );
	}
    }
  fclose( clusterLines );
  fclose(clusterPoints);
}

///////////////////////////////////////////////////////////////////////////////
/* returns the number of points in the hull
 */
int Cluster::buildHull(ClusterMember *start, int overShoot )
{
  int numConvexHullPts;
  int concaveHullPts;
  int count, count2;
  PM_CARTESIAN tempPoint, *tempPt;
  double totalDist;

  numHullDataIn = 0;
  while( start!=NULL )
    {
      hullDataIn[numHullDataIn].x = start->xLoc;
      hullDataIn[numHullDataIn].y = start->yLoc;
      hullDataIn[numHullDataIn].z = 0; // not used in hull
      numHullDataIn++;
      if(numHullDataIn==hullDataSize)
	expandHullData();
      start = start->previous;
    }
  numConvexHullPts = chainHull_2D( hullDataIn, numHullDataIn, hullDataOut );

  /*  printf( "hulldata\n");
  for( count=0; count<numHullDataIn; count++ )
    printf( "%d %d\n", (int)hullDataIn[count].y, (int)hullDataIn[count].x );
  printf( "\n" );
*/
  //  printf( "original hull\n" );

  for( count=0; count<numConvexHullPts; count++ )
    printf( "%d %d\n", (int)hullDataOut[count].y, (int)hullDataOut[count].x);
  printf( "\n" );


  // move around the convex hull and find the closest point on the perpendicular
  concaveHullPts = 0;
  for( count=0; count<numConvexHullPts-1; count++ )
    {
      hullDataTemp[concaveHullPts] = hullDataOut[count];
      if( ++concaveHullPts == hullDataSize )
	expandHullData();
      // compute how far we need to move along each edge
      totalDist = myDist(hullDataOut[count], hullDataOut[count+1]);
      for( count2=0; count2<totalDist; count2+=overShoot )
	{
	  hullDataTemp[concaveHullPts] = 
	    findClosestPerp(hullDataOut[count], 
			    hullDataOut[count+1],
			    count2, overShoot);
	  if( (int)hullDataTemp[concaveHullPts].x!=(int)hullDataTemp[concaveHullPts-1].x &&
	      (int)hullDataTemp[concaveHullPts].y!=(int)hullDataTemp[concaveHullPts-1].y )
	    if( ++concaveHullPts == hullDataSize )
	      expandHullData();
	}
    }
  /*
  // mark all points used in hull
  for( count=0; count<numConvexHullPts; count++ )
    markPt( hullDataOut[count], numHullDataIn, overShoot );
  // now let's try to find the concave hull
  concaveHullPts = 0;
  for( count=0; count<numConvexHullPts-1; count++ )
    {
      hullDataTemp[concaveHullPts] = hullDataOut[count];
      tempPoint = hullDataTemp[concaveHullPts++];
      while( myDist(tempPoint, hullDataOut[count+1]) > overShoot )
	{
	  tempPoint = movePoint( tempPoint, hullDataOut[count+1], overShoot );
	  if( findNearestClusterPt(tempPoint, hullDataTemp[concaveHullPts], 
				   numHullDataIn, overShoot))
	    {
	      tempPoint = hullDataTemp[concaveHullPts];
	      concaveHullPts++;
	    }
	}
    }
  */
  tempPt = hullDataOut;
  hullDataOut = hullDataTemp;
  hullDataTemp = tempPt;
  return concaveHullPts;
}

///////////////////////////////////////////////////////////////////////////////
void Cluster::checkIntersections()
{
  clusterHead.checkIntersection(0,1);
}



///////////////////////////////////////////////////////////////////////////////
/*!
  expand all of the polygons by the value given
  \param dist The amount to expand each side of a polygon
 */
void Cluster::expandPolygons( double dist )
{
  int count;

  for( count=0; count<clusterHead.getNumClusters(); count++ )
    clusterHead.expandPolygon(count, dist);
}

///////////////////////////////////////////////////////////////////////////////
/*! 
 * find the closest point along the perpindicular to the line
 */
PM_CARTESIAN Cluster::findClosestPerp(PM_CARTESIAN &linePtA, 
				      PM_CARTESIAN &linePtB,
				      int distAlongLine,
				      int neighborhood)
{
  PM_CARTESIAN perpPt, pointToCheck, returnData;
  int distance;
  int count;
  double minTest, currentMin = 100000;

  perpPt = movePoint(linePtA, linePtB, distAlongLine );
  pointToCheck = perpPt;
  
  distance = 0;
  //  while(distance < 100)
  for(;;)
    {
      for(count=0; count<numHullDataIn; count++ )
	{
	  if( (minTest=myDist(hullDataIn[count], pointToCheck)) < 
	      currentMin )
	    {
	      currentMin = minTest;
	      returnData = hullDataIn[count];
	    }
	}
      //      if( currentMin < neighborhood )
      if( currentMin < 2 )
	{
	  //	  printf( "dist: %d min: %f\n", distance, currentMin );
	  return returnData;
	}
      distance += 1;
      pointToCheck = movePointPerp(linePtA, linePtB, perpPt, distance);
      //      printf( "checking dist: %d\n", distance );
    }
  //  printf( "Exiting due to lack of data\n");
  //  exit(0);
}

///////////////////////////////////////////////////////////////////////////////
/*! routine used to test various things in the system, for debug only
 */
void Cluster::test(int testNo)
{
  PM_CARTESIAN ptA, ptB, ptC, perpPt;
  int count;
  switch(testNo)
    {
    case 1:
      ptA.x = 5;
      ptA.y = 54;
      ptB.x = 13;
      ptB.y = 22;
      ptC = movePoint( ptA, ptB, 5 );
      printf( "%f %f\n %f %f\n%f %f\n\n",
	      ptA.x, ptA.y, 
	      ptC.x, ptC.y,
	      ptB.x, ptB.y);
      for( count=0; count<10; count++ )
	{
	  perpPt = movePointPerp( ptA, ptB, ptC, count );
	  printf( "%f %f\n", perpPt.x, perpPt.y );
	}
      break;
    }
}

///////////////////////////////////////////////////////////////////////////////
/*! Move a point along the line ptA-ptB
 *  The point will be moved "distance" along the line
 */
PM_CARTESIAN Cluster::movePoint( PM_CARTESIAN &ptA, PM_CARTESIAN &ptB, double distance)
{
  PM_CARTESIAN ptC;
  double distAB, distCB;

  distAB = myDist( ptA, ptB );
  distCB = distAB-distance;

  ptC.x = distCB/distAB*(ptA.x-ptB.x) + ptB.x;
  ptC.y = distCB/distAB*(ptA.y-ptB.y) + ptB.y;
  return ptC;
}

///////////////////////////////////////////////////////////////////////////////
/*! Move a point along the perpendicular to line ptA-ptB
 *  The point will be moved "distance" along the perpendicular
 */
PM_CARTESIAN Cluster::movePointPerp( PM_CARTESIAN &linePtA, 
				     PM_CARTESIAN &linePtB, 
				     PM_CARTESIAN &ptC, double distance)
{
  PM_CARTESIAN perpPt;
  double theta;
  
  theta = atan2( linePtB.y-linePtA.y, linePtB.x-linePtA.x) + M_PI_2;
  perpPt.x = ptC.x + distance*cos(theta);
  perpPt.y = ptC.y + distance*sin(theta);
  return perpPt;
  /*
  double distAB;
  distAB = myDist( linePtA, linePtB );
  perpPt.x = ptC.x - distance/distAB*(linePtA.y-linePtB.y);
  perpPt.y = ptC.y + distance/distAB*(linePtA.x-linePtB.x);
  return perpPt;
  */
}

///////////////////////////////////////////////////////////////////////////////
bool Cluster::findNearestClusterPt(PM_CARTESIAN pointIn, 
				   PM_CARTESIAN &pointReturn,
				   int numPoints, int threshold)
{
  double myMin, temp, emergencyMin;
  int mark=-1, emergencyMark=-1;

  myMin = 100000;
  emergencyMin = 100000;
  /*
  myMin = myDist( pointIn, hullDataIn[0] );
  mark = 0;
  pointReturn = hullDataIn[0];
  if( myMin < threshold )
    {
      pointReturn = hullDataIn[0];
      return false;
    }
  */
  for( int count = 0; count<numPoints; count++ )
    {
      if( (temp=myDist(pointIn, hullDataIn[count]))<myMin )
	{
	  if( temp < threshold )
	    {
	      pointReturn = hullDataIn[count];
	      return false;
	    }
	  if( hullDataIn[count].z != 1 )
	    {
	      myMin = temp;
	      mark = count;
	      pointReturn = hullDataIn[count];
	    }
	  else if( temp < emergencyMin )
	    {
	      emergencyMin = temp;
	      emergencyMark = count;
	    }
	}
    }
  if( mark < 0 ) // no unmarked found
    {
      pointReturn = hullDataIn[emergencyMark];
      /*
      printf( "used emergency redundant point <%d %d>\n",
	      (int)pointReturn.x, (int)pointReturn.y );
      */
      return false;
    }
  if( hullDataIn[mark].z != 1 )
    {
      hullDataIn[mark].z = 1;
      markPt( hullDataIn[mark], numPoints, threshold );
      /*
      printf("distance from <%d %d> to <%d %d>: %f\n", 
	     (int)pointIn.x, (int)pointIn.y,
	     (int)pointReturn.x, (int)pointReturn.y, myMin );
      */
      return true;
    }
  return false;
}

///////////////////////////////////////////////////////////////////////////////
void Cluster::markPt( PM_CARTESIAN &ptIn, int numPoints, int range )
{
  int count;
  static int counter = 0;
  for( count=0; count<numPoints; count++ )
    {
      if( hullDataIn[count].z != 1 )
	{
	  if( myDist(ptIn, hullDataIn[count]) < range )
	    {
	      hullDataIn[count].z = 1;
	      counter++;
	    }
	}
    }
  //    printf( "set %d of %d points as seen\n", counter, numPoints );
}

///////////////////////////////////////////////////////////////////////////////
void Cluster::deleteRaw()
{
  delete [] rawData;
  xRawSize = 0;
  yRawSize = 0;
  rawData = NULL;
}

///////////////////////////////////////////////////////////////////////////////
void Cluster::expandHullData()
{
  PM_CARTESIAN *tempHullDataIn = new PM_CARTESIAN [hullDataSize*2];
  PM_CARTESIAN *tempHullDataOut = new PM_CARTESIAN [hullDataSize*2];
  PM_CARTESIAN *tempHullDataTemp = new PM_CARTESIAN [hullDataSize*2];

  //  printf( "expanding hulldata\n" );
  for( int count=0; count<hullDataSize; count++ )
    {
      tempHullDataIn[count] = hullDataIn[count];
      tempHullDataOut[count] = hullDataOut[count];
      tempHullDataTemp[count] = hullDataTemp[count];
    }

  delete [] hullDataIn;
  delete [] hullDataOut;
  delete [] hullDataTemp;

  hullDataSize *= 2;
  hullDataIn = tempHullDataIn;
  hullDataOut = tempHullDataOut;
  hullDataTemp = tempHullDataTemp;
}

///////////////////////////////////////////////////////////////////////////////
//! read in the data from a file
/*!
  Reads multiple formats set by "mode"
  \param name Name of file to open
  \param mode The mode of file access:
              0 - ASCII file that contains xsize ysize followed by data
              1 - ASCII file that contains only data
              2 - Tiff file that contains obstacle information in red channel (255 for obstacle)
              3 - Mif vector file where points and polygon meanings are assigned in the .mid file
  \return trun on success, false on failure
*/
#define INPUTBUFSIZE 256
bool Cluster::readFile( char *name, int mode )
{
  FILE *fp;
  int newXSize, newYSize;
  int counter;
  int tempX, tempY;
  cimg_library::CImg<unsigned char> wmImg;
  int count1, count2;
  char inputBuf[INPUTBUFSIZE];

  if( mode==0 || mode==1 )
    {
      fp = fopen( name, "r" );
      if( fp == NULL )
	{
	  printf( "Error opening input file from cluster\n" );
	  return false;
	}
      if( mode == 0 )
	{
	  fscanf( fp, "%d %d", &newXSize, &newYSize );
	}
      else
	{
	  newXSize = 101;
	  newYSize = 101;
	}
    }
  else if( mode==2)
    {
      wmImg = wmImg.get_load(name);
      if( wmImg.dimy()==0 || wmImg.dimx()==0 )
	return false;
      newXSize = wmImg.dimx();
      newYSize = wmImg.dimy();
    }
  else if( mode==3 ) // read mid, mif files
    {
      fp = fopen( name, "r" );
      if( fp == NULL )
	{
	  printf( "Error opening input file from cluster\n" );
	  return false;
	}
      newXSize = 1000; // allocate a one-dimensional array to store vector data
      newYSize = 1;
    }
  // allocate memory and initialize
  if( newXSize!=(int)xRawSize || newYSize!=(int)yRawSize )
    {
      deleteRaw();
      xRawSize = (unsigned int)newXSize;
      yRawSize = (unsigned int)newYSize;
      rawData = new ClusterMember[xRawSize*yRawSize];
      if( rawData == NULL )
	{
	  printf( "cluster unable to allocate memory\n" );
	  xRawSize = yRawSize = 0;
	  rawData = NULL;
	  return false;
	}
      counter=0;
      for( unsigned int xCount=0; xCount<xRawSize; xCount++ )
	for( unsigned int yCount=0; yCount<yRawSize; yCount++ )
	  {
	    rawData[counter].xLoc = xCount;
	    rawData[counter].yLoc = yCount;
	    rawData[counter].data = 0;
	    counter++;
	  }
    }
  // now that we have space, let's read the data
  //  printf( "reading %dx%d array\n", xRawSize, yRawSize );
  if( mode==0 )
    {
      for( counter=0; counter<(int)xRawSize*(int)yRawSize; counter++ )
	{
	  fscanf(fp, "%f", &rawData[counter].data );
	}
    }
  else if( mode==1 )
    {
      while( fscanf( fp, "%d %d", &tempX, &tempY ) == 2 )
	{
	  rawData[tempX*(int)yRawSize+(int)tempY].data = 1;
	}
    }
  else if( mode==2 )
    {
      wmImg.display();
      counter = 0;
      for( count1=0; count1<wmImg.dimx(); count1++ )
	{
	  for( count2=0; count2<wmImg.dimy(); count2++ )
	    {
	      if( wmImg(count1, count2, 0, 0)== 255 )
		{
		  //		  printf( "%d %d\n", count1, count2);
		  rawData[counter].data = 1;
		}
	      counter++;
	    }
	}
    }
  else if( mode==3 )
    {
      counter = 0;
      while( fgets(inputBuf, INPUTBUFSIZE, fp) != NULL )
	{
	  if( !strncasecmp( "Region", inputBuf, 6 ) )
	    {
	      fscanf( fp, "%d", &count1 );
	      // read first element to initialize the cluster
	      fscanf( fp, "%lf %lf", &rawData[counter].xLoc,
		      &rawData[counter].yLoc );
	      rawData[counter].data = 1;
	      joinCluster( counter, -1 );
	      counter++;
	      // read the rest of the polygon
	      for( count2=1; count2<count1; count2++ )
		{
		  fscanf( fp, "%lf %lf", &rawData[counter].xLoc,
			  &rawData[counter].yLoc );
		  rawData[counter].data = 1;
		  joinCluster( counter, counter-1 );
		  counter++;
		}
	    }
	}
    }
  return true;
}

///////////////////////////////////////////////////////////////////////////////
bool Cluster::readNML( NML *chanIn )
{
  int type;
  RCS_TIMER *timer = new RCS_TIMER(.5);
  SensorData2D *obs;
  int counter;
  
  while( (type = chanIn->read()) != SENSOR_DATA_2D_TYPE )
    {
      timer->wait();
    }
  obs = (SensorData2D*)chanIn->get_address();
  //  printf( "increment between cells: %f\n", obs->incrementX );
  if( obs->x!=(int)xRawSize || obs->y!=(int)yRawSize )
    {
      deleteRaw();
      xRawSize = (unsigned int)obs->x;
      yRawSize = (unsigned int)obs->y;
      rawData = new ClusterMember[xRawSize*yRawSize];
      if( rawData == NULL )
	{
	  printf( "cluster unable to allocate memory\n" );
	  xRawSize = yRawSize = 0;
	  rawData = NULL;
	  return false;
	}
      counter = 0;
      for( unsigned int xCount=0; xCount<xRawSize; xCount++ )
	for( unsigned int yCount=0; yCount<yRawSize; yCount++ )
	  {
	    rawData[counter].xLoc = xCount;
	    rawData[counter].yLoc = yCount;
	    counter++;
	  }
    }
      for( counter=0; counter<obs->featArray_length; counter++ )
	{
	  rawData[counter].data = obs->featArray[counter];
	}
      return true;
}

///////////////////////////////////////////////////////////////////////////////
/* builds the clusters. Returns the number of clusters constructed
 */
int Cluster::build()
{
  int counter;

  for( counter=0; counter< (int)xRawSize*(int)yRawSize; counter++ )
    {
      checkCluster(counter);
    }
  return clusterHead.getNumClusters();
}


///////////////////////////////////////////////////////////////////////////////
/*! checks if an element is above the obstacle threshold, if so
 * then look to see if there is a cluster to join. Checks current row (y) back
 * as far as "gap size" to see if there is a cluster. If so, find out if only
 * cluster for this pixel. If so, join it. If not, merge the two together. Then
 * Check the rest of the area (previous x) and do the same.
 * 
 */
void Cluster::checkCluster( int toCheck )
{
  double xStart, xEnd, yStart, yEnd, xLoc, yLoc;
  int elementNum;

  rawData[toCheck].clusterID = -1;
  // do we need to check this element?
  if( rawData[toCheck].data < .41 )
    {
      return;
    }
  // first check my row
  yStart = rawData[toCheck].yLoc-1;
  if( yStart >=0 ) // need to check the rest of my row
    {
      yEnd = rawData[toCheck].yLoc-gapSize;
      if( yEnd < 0 )
	yEnd = 0;
      for( yLoc=yStart; yLoc>=yEnd; yLoc-- )
	{
	  elementNum = (int)(rawData[toCheck].xLoc*yRawSize+yLoc);
	  if( rawData[elementNum].clusterID != -1 ) // join this cluster
	    {
	      if( rawData[toCheck].clusterID == rawData[elementNum].clusterID )
		continue;
	      else if( rawData[toCheck].clusterID != -1 )
		merge( rawData[toCheck].clusterID, 
		       rawData[elementNum].clusterID );
	      else
		joinCluster(toCheck, elementNum);
	    }
	}
    }
  // check rest of area
  xStart = rawData[toCheck].xLoc-1;
  if( xStart<0 )
    {
      if( rawData[toCheck].clusterID==-1 )// no other clusters to join so create new onw
	joinCluster(toCheck, -1 );
      // else already in cluster so just return
      return;
    }
  xEnd = rawData[toCheck].xLoc-gapSize;
  if( xEnd < 0 )
    xEnd = 0;

  yStart = rawData[toCheck].yLoc+gapSize;
  yEnd = rawData[toCheck].yLoc-gapSize;
  if( yStart >= (int)yRawSize ) // was >
    yStart = yRawSize-1;
  if( yEnd < 0 )
    yEnd = 0;

  for( xLoc=xStart; xLoc>=xEnd; xLoc-- )
    {
      for( yLoc=yStart; yLoc>=yEnd; yLoc-- )
	{
	  elementNum = (int)((xLoc*yRawSize)+yLoc);
	  if( rawData[elementNum].clusterID != -1 ) // join this cluster
	    {
	      if( rawData[toCheck].clusterID == rawData[elementNum].clusterID )
		continue;
	      else if( rawData[toCheck].clusterID != -1 )
		merge( rawData[toCheck].clusterID, 
		       rawData[elementNum].clusterID );
	      else
		joinCluster(toCheck, elementNum);
	    }

	}
    }
  // no clusters so start new one
  if( rawData[toCheck].clusterID==-1)
    {
      //      printf( "no other clusters (2) for %d %d\n", rawData[toCheck].xLoc, rawData[toCheck].yLoc );
      joinCluster(toCheck, -1);
    }
}

///////////////////////////////////////////////////////////////////////////////
/* keeping the most recently generated cluster, so it should have the largest value
 * need to merge the "clusterMerge" values into this cluster in decending order
 */
void Cluster::merge( int clusterKeep, int clusterMerge )
{
  ClusterMember *newMember, *currentMember, *tempMember, *tempMember2;

  /*
  printf( "merging: keep %d merge %d\n", clusterKeep, clusterMerge );
  clusterHead.print(clusterKeep);
  clusterHead.print(clusterMerge);
  */
  currentMember = clusterHead.clusterEnds[clusterKeep];
  newMember = clusterHead.clusterEnds[clusterMerge];
  while( newMember!=NULL ) // find a place for all of these
    {
      newMember->clusterID = currentMember->clusterID;
      while( newMember < currentMember->previous )
	currentMember = currentMember->previous;
      tempMember = currentMember->previous;
      tempMember2 = newMember->previous;
      currentMember->previous = newMember;
      newMember->previous = tempMember;
      currentMember = newMember;
      newMember = tempMember2;
    }
  clusterHead.clusterEnds[clusterMerge] = NULL;
  /*
  printf( "merged...\n");
  clusterHead.print(clusterKeep);
  */
}

///////////////////////////////////////////////////////////////////////////////
//! adds an element to a cluster
/*!
  Adds an element to a cluster and adjusts the rawData members to correspond. 
  Sets the clusterID and navigation pointers to where they should be.
  \param newMember The element number of the rawData array that will be added to the cluster that
                   currentMember is a member of.
  \param currentMember Any member of the cluster that the newMember will join. If -1, then create
         a new cluster.
 */
void Cluster::joinCluster( int newMember, int currentMember )
{
  /*
  if( currentMember == -1 )
    printf( "Creating new cluster %d\n", clusterHead.getNumClusters() );
  else
    printf( "adding member %d to cluster %d\n", newMember, rawData[currentMember].clusterID);
  */
  if( currentMember == -1 ) // new cluster
    {
      clusterHead.clusterEnds[clusterHead.getNumClusters()] = &rawData[newMember];
      rawData[newMember].clusterID = clusterHead.getNumClusters();
      if( !clusterHead.incNumClusters())
	printf( "\nRan out of clusters\n\n\n;" );
/*
      printf( "Added <%d %d> to cluster %d\n", rawData[newMember].xLoc, rawData[newMember].yLoc,
	  rawData[newMember].clusterID );
*/
      return;
    }
  rawData[newMember].clusterID = rawData[currentMember].clusterID;
  rawData[newMember].previous = clusterHead.clusterEnds[rawData[currentMember].clusterID];
  clusterHead.clusterEnds[rawData[currentMember].clusterID] = &rawData[newMember];
  /*
  printf( "Added <%d %d> to cluster %d\n", rawData[newMember].xLoc, rawData[newMember].yLoc,
	  rawData[newMember].clusterID );
  */
}

///////////////////////////////////////////////////////////////////////////////
//! Create a visibility graph of the polygons contained in the cluster
/*!
    This is a very ineficient routine. It should really be written to use a sweep line.
 */
/*
void Cluster:createVisGraph()
{
  int fromID, toID;
  ClusterMember *fromMember, *toMember, *polyCheck;
  LineSegment toCheck, polyLine;

  // for now we will just write the vis graph out to a file in mif format
  FILE *fp;

  fp = fopen( "visGraph.mif", "w" );
  // not sure of the header information; just copying one from an existing file
  fprintf( fp, "Version 450\nCharset \"Neutral\"\nDelimiter \",\"\n" );
  fprintf( fp, "CoordSys NonEarth Units \"m\" Bounds (5.26511741156923, 1.483181780225416) (84.6429652083507, 45.56448243241343)\n");
  fprintf( fp, "Columns 0\nData\n\n" );

  for( fromID=0; fromID<clusterHead.getNumClusters()-1; fromID++ )
    {
      // connect a line from each vertice in polygon[count] to 
      // every vertice in every polygon that comes after that
      fromMember = clusterHead.clusterEnds[fromID];
      while( fromMember!=NULL )
	{
	  toCheck.start.north = fromMember->xLoc;
	  toCheck.start.east = fromMember->yLoc;
	  for( toID=fromID; toID<clusterHead.getNumClusters(); toID++ )
	    {
	      toMember = clusterHead.clusterEnds[toID];
	      while( toMember!= NULL )
		{
		  toCheck.end.north = toMember->xLoc;
		  toCheck.end.east = toMember->yLoc;
		  // now we have the line to check, so check it against all of the polygons
		  for( checkID=0; checkID<clusterHead.getNumClusters(); checkID++ )
		    {
		      if( checkID == 
		      polyCheck = 
		    }
		  toMember = toMember->previous;
		}
	    }
	  currentMember = currentMember->previous;
	}
    }
}
*/

///////////////////////////////////////////////////////////////////////////////
//! Print out all of the linked cluster information in the form of a MIF GIS file
/*! This routine will print out the header for the MIF file and then call the
    clusterHead print routine to print out the actual polygons
*/
void Cluster::print(FILE *fp)
{
  int count;

  // not sure of the header information; just copying one from an existing file
  fprintf( fp, "Version 450\nCharset \"Neutral\"\nDelimiter \",\"\n" );
  fprintf( fp, "CoordSys NonEarth Units \"m\" Bounds (5.26511741156923, 1.483181780225416) (84.6429652083507, 45.56448243241343)\n");
  fprintf( fp, "Columns 0\nData\n\n" );

  for( count=0; count<clusterHead.getNumClusters(); count++ )
    {
	clusterHead.print(count, fp);
	fprintf( fp, "\n" );
    }
}

///////////////////////////////////////////////////////////////////////////////
void Cluster::print()
{
  printf( "P3\n%d %d\n255\n", xRawSize, yRawSize );
  printf( "#Found %d clusters\n", clusterHead.getNumClusters() );
  for( int count=0; count<(int)xRawSize*(int)yRawSize; count++ )
    {
      if( rawData[count].clusterID == -1 )
	printf( "254 254 254 " );
      else
	printf( "%d %d %d ", rawData[count].clusterID,rawData[count].clusterID,rawData[count].clusterID );
    }
  printf( "\n" );
  fflush(stdout);
  clusterHead.print();
}

/*
  Modification history:

  $Log: cluster.cc,v $
  Revision 1.7  2007/11/08 13:22:30  dr_steveb
  Not quite ready, but premature checking-in of other files means that these need to go to!

  Revision 1.6  2007/08/03 18:24:33  dr_steveb
  Joystick no longer waits on mission package.

  Revision 1.5  2007/05/16 15:22:11  dr_steveb
  Finished traffic control interfaces

  Revision 1.4  2007/04/25 15:09:04  dr_steveb
  Added traffic simulation files

  Revision 1.3  2007/03/21 20:31:52  dr_steveb
  Fixed initialization of "emergency" location and added printing of convex hulls.

  Revision 1.2  2007/03/21 19:29:49  dr_steveb
  Fixed bug that caused the world to wrap

  Revision 1.1  2007/02/27 14:51:04  dr_steveb
  New tool for calc. the concave hull

*/
