/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vectorWM.cc
  \brief  provides class to hold vertices of vector world model

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.1 $
  $Date: 2007/04/25 15:41:37 $
  \endcode

  \author Stephen Balakirsky
  \date   03/21/2007
*/

#include "vectorWM.hh"

///////////////////////////////////////////////////////////////////////////////
// global functions
///////////////////////////////////////////////////////////////////////////////
int ClusterMemberCompare( const void *vN1, const void *vN2 );

///////////////////////////////////////////////////////////////////////////////
double myDist2(ClusterMember *pt1, ClusterMember &pt2 )
{
  return( sqrt((pt1->xLoc-pt2.xLoc)*(pt1->xLoc-pt2.xLoc)+
	       (pt1->yLoc-pt2.yLoc)*(pt1->yLoc-pt2.yLoc)) );
}

///////////////////////////////////////////////////////////////////////////////
// VectorWM class
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
VectorWM::VectorWM(int initialSize, int extentsIn, int exclusionIn )
{
  wmData = new ClusterMember[initialSize];
  numVertices = 0;
  maxNumVertices = initialSize;
  extents = extentsIn;
  exclusion = exclusionIn;
}

///////////////////////////////////////////////////////////////////////////////
VectorWM::~VectorWM()
{
  delete [] wmData;
}

///////////////////////////////////////////////////////////////////////////////
/*! 
 * Adds vertices that are within the world model extents.
 * If useExclusion is set, then vertices in the exclusion zone,
 * the center of the map, are also eliminated
 */
void VectorWM::addVerts( ClusterMember *toadd, ClusterMember &mapCenter,
			 bool useExclusion)
{
  double distance;
  while( toadd!=NULL )
    {
      distance = myDist2(toadd, mapCenter);
      if( distance<extents && (!useExclusion || distance>exclusion) )
	{
	  if( numVertices==maxNumVertices )
	    expandDataArray();
	  wmData[numVertices].xLoc = toadd->xLoc;
	  wmData[numVertices].yLoc = toadd->yLoc;
	  wmData[numVertices].clusterID = -1;
	  numVertices++;
	}
      toadd = toadd->previous;
    }
}

///////////////////////////////////////////////////////////////////////////////
void VectorWM::expandDataArray()
{
  ClusterMember *temp;
  int count;

  maxNumVertices += 250;
  temp = new ClusterMember[maxNumVertices];
  for(count=0; count<numVertices; count++)
    temp[count] = wmData[count];
  delete [] wmData;
  wmData = temp;
}

///////////////////////////////////////////////////////////////////////////////
void VectorWM::sort()
{
  qsort( wmData, numVertices, sizeof(ClusterMember), &ClusterMemberCompare );
}

///////////////////////////////////////////////////////////////////////////////
/*!
 * we need to cluster the vertices that are contained in the data set. For each new vertex,
 * we only check the preceeding vertices. Since the data is sorted, we only need to check until
 * the ?-coordinate is greator then the cluster distance.
 */
/*
  Modification history:

  $Log: vectorWM.cc,v $
  Revision 1.1  2007/04/25 15:41:37  dr_steveb
  First check-in of vector world model test.


*/
