/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehDrone.cc
  \brief Provides a very basic controller that drives a vehicle repeatately from point 'a' to point 'b'

  This program has the following state table:
  NEW_COMMAND: open nml buffers
  S0: Initialization (init prim and get ready to drive)
  S1: Spawn vehicle (teleport/spawn vehicle to location 'a')
  S2: Drive vehicle (command vehicle to drive to location 'b')
  S3: Check goal (Check for prim done)
  S4: AtGoal (Remove vehicle, go to S1 )

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Stephen Balakirsky
  \date   07/18/2005
*/
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <rcs.hh>
#include <unistd.h>
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "primMobJA.hh"
#include "servoMobJA.hh"
#include "navDataExt.hh"
#include <signal.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif

#define MY_NAME_BASE "vehDrone"

/////////////////////////////////////////////
///////////////  classes  ///////////////////
/////////////////////////////////////////////


/////////////////////////////////////////////
//////////////   globals   //////////////////
/////////////////////////////////////////////
//! Transforms from local to global and back
PM_HOMOGENEOUS homLocal, homGlobal;
//! where are we anyway?
PM_CARTESIAN vehLocation;

//! Serial number to prim level
int myPrimSerial;
//! Serial number to servo level
int myServoSerial;
// NML Channels
//! prim planning channels
RCS_CMD_CHANNEL  *primMobJACmdNml;
RCS_STAT_CHANNEL *primMobJAStatNml;
RCS_CMD_CHANNEL *primMobJACfgNml;
RCS_STAT_CHANNEL  *primMobJASetNml;
//! servo Mob channels
RCS_CMD_CHANNEL  *servoMobJACmdNml;
RCS_STAT_CHANNEL *servoMobJAStatNml;
RCS_CMD_CHANNEL *servoMobJACfgNml;
RCS_STAT_CHANNEL  *servoMobJASetNml;
//! navigation data channel
NML              *navDataExtNml;


/*!
  Transform a point from/to global coordinates to/from local coordinates

\param pointIn The point to be converted
\param transformGlobalToLocal True- transform to local, false- transform to global
\return Transformed point
*/
PM_CARTESIAN transformPoint(Location2D &point2DIn, bool transformGlobalToLocal)
{
  PM_HOMOGENEOUS homTransform;
  PM_HOMOGENEOUS homIn, homOut;
  PM_CARTESIAN pointReturn;
  PM_CARTESIAN pointIn;

  pointIn.x = point2DIn.north;
  pointIn.y = point2DIn.east;
  pointIn.z = 0;
  homIn.tran.x = pointIn.x;
  homIn.tran.y = pointIn.y;
  homIn.tran.z = pointIn.z;
  homIn.rot = PM_RPY(0, 0, 0);

  if( transformGlobalToLocal )
    {
      homTransform = homLocal * inv(homGlobal);
    }
  else
    {
      homTransform = homGlobal * inv(homLocal);
    }
  homOut = homTransform * homIn;
  pointReturn.x = homOut.tran.x;
  pointReturn.y = homOut.tran.y;
  pointReturn.z = homOut.tran.z;

  return(pointReturn);
}

/*!
  Read in new vehicle position

  \return true on success
  \return false on failure
*/
bool updatePosition()
{
  int type;
  NavDataExt *navData;
  bool mapIsGlobal = true;

  type = navDataExtNml->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case NAV_DATA_EXT_TYPE:
      navData = (NavDataExt*)navDataExtNml->get_address();
      if( mapIsGlobal )
	{
	  vehLocation.x = navData->tranAbs.x;
	  vehLocation.y = navData->tranAbs.y;
	}
      else
	{
	  vehLocation.x = navData->tranRel.x;
	  vehLocation.y = navData->tranRel.y;
	}
      homLocal.tran = navData->tranRel;
      homLocal.rot = navData->rpyRel;
      homGlobal.tran = navData->tranAbs;
      homGlobal.rot = navData->rpyAbs;
      return(true);
    default:
      return(false);
    }
  return(false);
}

/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool checkPrimCmdStatus( RCS_STATUS &stat, int &echo_serial_number )
{
  int type;

  type = primMobJAStatNml->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case PRIM_MOB_JA_STAT_TYPE:
      stat = primMobJAStatNml->get_address()->status;
      echo_serial_number = primMobJAStatNml->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool checkServoSetStatus( RCS_STATUS &stat, int &echo_serial_number )
{
  int type;

  type = servoMobJASetNml->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case SERVO_MOB_JA_SET_TYPE:
      stat = servoMobJASetNml->get_address()->status;
      echo_serial_number = servoMobJASetNml->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool checkServoCmdStatus( RCS_STATUS &stat, int &echo_serial_number )
{
  int type;

  type = servoMobJAStatNml->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case SERVO_MOB_JA_STAT_TYPE:
      stat = servoMobJAStatNml->get_address()->status;
      echo_serial_number = servoMobJAStatNml->get_address()->echo_serial_number;
      return(true);
    default:
      fprintf( stderr, "Unknown cmd type received: %s %d\n", servoMobJA_symbol_lookup(type), type );
      break;
    }
  return false;
}

/*!
  Open all of the NML channels
  
  \param vehID The number to add to all of the channel names
  \return true on success
  \return false on failure
*/
bool openNML( int vehID )
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];

  sprintf( myName, "%s%d", MY_NAME_BASE, vehID );

  if(!configNml)
    {
      fprintf(stderr,"Please set CONFIG_NML environment variable.\n");
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_CMD_NAME, vehID );
  primMobJACmdNml = new RCS_CMD_CHANNEL( primMobJA_format, chanName,
				      myName, configNml );
  if( NULL == primMobJACmdNml || !primMobJACmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_STAT_NAME, vehID );
  primMobJAStatNml = new RCS_STAT_CHANNEL( primMobJA_format, chanName,
					myName, configNml );
  if( NULL == primMobJAStatNml || !primMobJAStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_SET_NAME, vehID );
  primMobJASetNml = new RCS_STAT_CHANNEL( primMobJA_format, chanName,
				      myName, configNml );
  if( NULL == primMobJASetNml || !primMobJASetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_CFG_NAME, vehID );
  primMobJACfgNml = new RCS_CMD_CHANNEL( primMobJA_format, chanName,
				       myName, configNml );
  if( NULL == primMobJACfgNml || !primMobJACfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, vehID );
  servoMobJACmdNml = new RCS_CMD_CHANNEL( servoMobJA_format, chanName,
				       myName, configNml );
  if( NULL == servoMobJACmdNml || !servoMobJACmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, vehID );
  servoMobJAStatNml = new RCS_STAT_CHANNEL( servoMobJA_format, chanName,
					 myName, configNml );
  if( NULL == servoMobJAStatNml || !servoMobJAStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_SET_NAME, vehID );
  servoMobJASetNml = new RCS_STAT_CHANNEL( servoMobJA_format, chanName,
				       myName, configNml );
  if( NULL == servoMobJASetNml || !servoMobJASetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_CFG_NAME, vehID );
  servoMobJACfgNml = new RCS_CMD_CHANNEL( servoMobJA_format, chanName,
					myName, configNml );
  if( NULL == servoMobJACfgNml || !servoMobJACfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", NAV_DATA_EXT_NAME, vehID );
  navDataExtNml = new NML( navDataExt_format, chanName,
			   myName, configNml );
  if( NULL == navDataExtNml || !navDataExtNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, chanName );
      return(false);
    }

  return(true);
}

/*!
  free all memory and prepare to exit
*/
void closeAll()
{
  PrimMobJACmdShutdown shutdown;
  
  shutdown.serial_number = myPrimSerial++;
  primMobJACmdNml->write(shutdown);

  if( primMobJACmdNml != NULL )
    delete primMobJACmdNml;
  primMobJACmdNml = NULL;
  if( primMobJAStatNml != NULL )
    delete primMobJAStatNml;
  primMobJAStatNml = NULL;
  if( primMobJASetNml != NULL )
    delete primMobJASetNml;
  primMobJASetNml = NULL;
  if( primMobJACfgNml != NULL )
    delete primMobJACfgNml;
  primMobJACfgNml = NULL;
  if( servoMobJACmdNml != NULL )
    delete servoMobJACmdNml;
  servoMobJACmdNml = NULL;
  if( servoMobJAStatNml != NULL )
    delete servoMobJAStatNml;
  servoMobJAStatNml = NULL;
  if( servoMobJASetNml != NULL )
    delete servoMobJASetNml;
  servoMobJASetNml = NULL;
  if( servoMobJACfgNml != NULL )
    delete servoMobJACfgNml;
  servoMobJACfgNml = NULL;
  if( navDataExtNml != NULL )
    delete navDataExtNml;
  navDataExtNml = NULL;
}


/*!
  Signal handler
  \param signum Signal that has been caught
*/
void sigHandler( int signum )
{
  switch( signum )
    {
    case SIGINT:
      fprintf( stderr, "Shutting down vehMobPL program\n" );
      closeAll();
      exit(1);
      break;
    default:
      fprintf( stderr, "Unknown signal caught\n" );
      break;
    }
}

int main(int argc, char **argv)
{
  bool useSimple = false;
  bool myVerbose = false;
  Location2D a, b;
  PM_CARTESIAN tempCart;
  RCS_TIMER *timer;
  float cycleTime = .25;
  RCS_STAT_MSG *status = new RCS_STAT_MSG( 100, sizeof(RCS_STAT_MSG) );
  char c;
  char cmdLine[80];
  ServoMobJACfgTeleport teleport;
  ServoMobJACmdInit servoInit;
  ServoMobJACmdAbort servoAbort;
  PrimMobJACmdMoveWaypoint waypoint;
  PrimMobJACmdInit primInit;
  int echo_serial_number;
  RCS_STATUS primStatus;
  int delayCount;
  bool isPrimInitialized = false;
  bool isServoInitialized=false;
  PM_RPY aRot;
  
  aRot.r = 0;
  aRot.p = 0;
  aRot.y = 0;

  int vehID = 2;
  a.north = 4334130.5;
  //  a.east = 308306.99;
  a.east = 308123;
  b.north = 4334130.;
  b.east = 308250.;
  myPrimSerial = 18;
  myServoSerial = 18;
  while((c=getopt(argc,argv,"d:c:v"))!=EOF) 
    {
      switch( c )
	{
	case 'c':
	  // set cycle time
	  cycleTime = atof(optarg);
	  break;
	case 'd':
	  // set vehicle id
	  vehID = atoi(optarg);
	  break;
	case 'v':
	  // turn on verbose printouts
	  myVerbose = true;
	  break;
	}
    }

  // initialize run
  status->state = NEW_COMMAND;
  signal( SIGINT, sigHandler );

  for(;;)
    {
      // read input buffers
      if( isServoInitialized )
	updatePosition();
      if( state_match(status, NEW_COMMAND) ) // NEW_COMMAND: open nml buffers
	{
	  if( myVerbose )
	    fprintf(stderr, "In state NEW_COMMAND:\r" );
	  state_new(status);
	  status_next(status, RCS_EXEC);
	  state_next(status, S3);
	  if(!openNML( vehID ))
	    {
	      fprintf( stderr, "Error opening nml buffers\n" );
	      return(-1);
	    }
	  timer = new RCS_TIMER(cycleTime);
	  delayCount = 0;
	}
      //////////////////////// not done
      else if( state_match(status, S0 ) ) // S0: Initialization: spawn simplesim or usarSim
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S0:\r" );
	  status_next(status, RCS_EXEC);
	  state_next(status, S1 );
	  if( useSimple ) 	  // spawn simplesim
	    sprintf( cmdLine, "./simpleSim -d%d&", vehID );
	  else // spawn userSim
	    fprintf( stderr, "Spawing of userSim disabled\n" );
	    //	    sprintf( cmdLine, "./usarSim bugs-220.mxnet.mel.nist.gov& " );
	  system(cmdLine);
	  delayCount = 0;
	}

      ///////////////////////// not done
      else if( state_match( status, S1) ) // S1: Initialization: init usarSim (spawn vehicle)
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S1:\r" );
	  status_next(status, RCS_EXEC);
	  if( delayCount < 4  )
	    delayCount++;
	  else
	    {
	      if( useSimple )
		{
		  isServoInitialized = true;
		  state_next( status, S2 );
		}
	      else if( !isServoInitialized )
		{
		  servoInit.serial_number = myServoSerial++;
		  servoMobJACmdNml->write( servoInit );
		  if( myVerbose )
		    fprintf( stderr, "Initialized servo vehicle\n" );
		  isServoInitialized = true;
		}
	      else
		{
		  checkServoCmdStatus( primStatus, echo_serial_number );
		  if( primStatus==RCS_DONE && echo_serial_number==myServoSerial-1 )
		    state_next(status, S2 );
		}
	    }
	}
      //////////////////////////// not done
      else if( state_match( status, S2) ) // S2: Initialization: spawn prim
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S2:\r" );
	  delayCount = 0;
	  status_next(status, RCS_EXEC);
	  state_next(status, S3 );
	  // spawn prim
	  fprintf( stderr, "\n\nSpawn primMobMain now!!\n" );
	  sprintf( cmdLine, "./primMobMain -d%d&", vehID );
	  //	  system(cmdLine);
	}
      else if( state_match( status, S3) ) // S3: Initialization: init prim
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S3:\r" );
	  status_next(status, RCS_EXEC);
	  if( delayCount < 4 )
	    delayCount++;
	  else
	    {
	      if( !isPrimInitialized )
		{
		  primInit.serial_number = myPrimSerial++;
		  primMobJACmdNml->write(primInit);
		  isPrimInitialized = true;
		  isServoInitialized = true; // gets initialized by prim
		}
	      else
		{
		  checkPrimCmdStatus( primStatus, echo_serial_number );
		  if( primStatus==RCS_DONE && echo_serial_number==myPrimSerial-1 )
		    {
		      state_next(status, S6 );
		      sleep(2);
		    }
		}
	    }
	}
      else if( state_match(status, S4 ) )  // S4: Maintanance: teleport vehicle to location 'a')
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S4:\r" );
	  status_next(status, RCS_EXEC);

	  if( !isServoInitialized )
	    {
	      // teleport the vehicle
	      teleport.goal.tran.x = a.north;
	      teleport.goal.tran.y = a.east;
	      teleport.goal.tran.z = 0;
	      teleport.goal.rot = aRot;
	      teleport.serial_number = myServoSerial++;
	      servoMobJACfgNml->write(teleport);
	      if( myVerbose )
		{
		  waypoint.waypoint[0].p = transformPoint( a, true );
		  fprintf(stderr, "Teleporting vehicle to: global:<%lf %lf>, local:<%lf %lf>\n", 
			  a.north,
			  a.east,
			  waypoint.waypoint[0].p.x,
			  waypoint.waypoint[0].p.y );
		}
	      isServoInitialized = true;
	    }
	  else
	    {
	      checkServoSetStatus( primStatus, echo_serial_number );
	      fprintf( stderr, "Status: %d Serial: %d Need: %d\n", primStatus, 
		       echo_serial_number, myServoSerial-1 );
	      if( primStatus==RCS_DONE && echo_serial_number==myServoSerial-1 )
		state_next(status, S6);
	    }
	}
      /////////////////////// not done
      else if( state_match(status, S5 ) ) // S5: check for status of teleport or init
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S5:\r" );
	  status_next(status, RCS_EXEC);
	}
      else if( state_match(status, S6 ) ) // S6:Drive vehicle (command vehicle to drive to location 'b')
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S6:\r" );
	  status_next(status, RCS_EXEC);
	  state_next(status, S7 );
	  // start vehicle driving
	  waypoint.waypoint[0].p = transformPoint( b, true );
	  waypoint.waypoint[0].neighborhood = 10.;
	  waypoint.waypoint[0].quality = 1;
	  waypoint.waypoint[0].speed = 100;
	  waypoint.waypoint_length = 1;
	  waypoint.serial_number = myPrimSerial++;
	  primMobJACmdNml->write(waypoint);
	  if( myVerbose )
	    fprintf(stderr, "\nSending vehicle to: global:<%lf %lf>, local:<%lf %lf>\n", 
		    b.north,
		    b.east,
		    waypoint.waypoint[0].p.x,
		    waypoint.waypoint[0].p.y );
	}
      else if( state_match(status, S7 ) ) // Check goal (Check for prim done)
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S7:\r" );
	  status_next(status, RCS_EXEC);
	  // monitor driving
	  checkPrimCmdStatus( primStatus, echo_serial_number );
	  if( primStatus==RCS_DONE && echo_serial_number==myPrimSerial-1 )
	    state_next(status, S8);	
	}
      else if( state_match( status, S8) ) // AtGoal (Remove vehicle, go to S1 )
	{
	  status_next(status, RCS_EXEC);
	  if( myVerbose )
	    fprintf(stderr, "In state S8:\r" );
	  /*
	  if( !useSimple )
	    {
	      servoAbort.serial_number = myPrimSerial++;
	      servoMobJACmdNml->write(servoAbort);
	      state_next(status, S9);
	    }
	  else
	  */
	  isServoInitialized = false;
	  state_next(status, S4);
	}
      else if( state_match( status, S9 ) ) // check for usarSim status
	{
	  status_next(status, RCS_EXEC);
	  checkPrimCmdStatus( primStatus, echo_serial_number );
	  if( primStatus==RCS_DONE && echo_serial_number==myPrimSerial-1 )
	    {
	      state_next(status, S4);	
	      isServoInitialized = false;
	    }
	}

      timer->wait();
    }

  
}

/*
  Modification history:

  $Log: vehDrone.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.13  2005/09/30 18:55:56  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.12  2005/08/10 18:38:54  stephen
  Something changed.

  Revision 1.11  2005/08/09 21:08:30  stephen
  Small changes to make work better in demo

  Revision 1.10  2005/07/28 20:55:35  stephen
  Fixed serial number on settings channel for teleport

  Revision 1.9  2005/07/28 15:17:58  stephen
  Removed init parameters.

  Revision 1.8  2005/07/27 21:00:13  stephen
  Working on state machine.

  Revision 1.7  2005/07/26 20:33:02  stephen
  Worked on state machine.

  Revision 1.6  2005/07/21 19:05:13  stephen
  Work in progress.

  Revision 1.5  2005/07/21 18:53:15  stephen
  Work in progress.

  Revision 1.4  2005/07/21 17:18:51  stephen
  Debug in process.

  Revision 1.3  2005/07/21 13:57:58  stephen
  State machine fixes.

  Revision 1.2  2005/07/19 17:35:27  stephen
  First working version (works with simpleSim)

  Revision 1.1  2005/07/18 20:54:24  stephen
  New file check-in


*/
