/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehMobPLEngine.hh
  \brief  Include file for the VehMobPLEngine class


  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.13 $
  $Date: 2007/05/16 20:57:21 $
  \endcode

  \author Stephen Balakirsky
  \date   05/17/2005
*/
//#define USE_SPQR


#ifndef VEH_MOB_PL_ENGINE_HH
#define VEH_MOB_PL_ENGINE_HH

#include "moastTypes.hh"
#include "dijkHeapN.hh"
#include "sensorData.hh"
#include "amSP.hh"
#include "vehMobPL.hh"
#include "vehMobMap.hh"
#include "nmlMap.hh"
#include "CImg.h"
#include <vector>

////////////////////////////////

using namespace std;
#define MAX_DIJK_WAYPOINTS 10000

//////////////////////////////////////////////////////////
// structures
/////////////
typedef struct
{
  int id;
  char name[SENSOR_DATA_TXT_BUFFER_MAX];
  char injury[SENSOR_DATA_TXT_BUFFER_MAX];
  PM_CARTESIAN location;
  PM_CARTESIAN goalLocation;
}VictInfo;  

//////////////////////////////////////////////////////////
// functions
/////////////
double myDisp( PM_CARTESIAN &pt1, PM_CARTESIAN &pt2 );

class VehMobPLEngine
{
public:
  //functions
  VehMobPLEngine( int northCells, int eastCells, float cellResolution, int obsGrowIn,
	       int verbose=0 );
  ~VehMobPLEngine();
  bool addGraphPoints( PM_CARTESIAN *points, int pointLength );
  bool blockClosestWay( PM_CARTESIAN location[], int numLocations, unsigned int untilWhen, 
			goalPoint *currentGoal, VehCellMark markIn,
			bool force);
  bool buildGraph();
  void clearMap();
  void clearWorld(unsigned int oldestTime);
  void dump(char *fileName,  PM_CARTESIAN &vehLocation);
  void mapOut( NML *nmlOutChan );
  void mapOutSpqr( NML *nmlOutChan,  PM_CARTESIAN &globalVehLoc );
  int search( PM_CARTESIAN &goal, 
	      double planStartHorizon, bool newSearch, double endTime );
  void printAns( double cost, double planStartHorizon );
  int updateWorld( SensorPoint *updates, int Length, unsigned int timeSeen, VehCellMark markIn );
  int updateWorld( SensorData2D *updates, unsigned int timeSeen,  PM_CARTESIAN &vehLocation );
  bool growWorld( int newGraphWidth, int newGraphHeight );
  bool growWorld( PM_CARTESIAN &includeLoc );
  int getGraphHeight(){return(graphHeight);};
  int getGraphWidth(){return(graphWidth);};
  void init(  PM_POSE &vehLocIn );
  double maxCost( int depth );
  void discount( bool doTurnOn );
  bool setCostLookup();
  bool loadWM( char *file );
  bool storeWM( char *file );
  void setVehLoc(  PM_POSE &vehLocIn );
  double getCost(){
    return(costsOut);
  }
  int getPathLength(){
    return(pathLength);
  }
  //  int worldToImageConvert( double x, double y );
  PM_CARTESIAN worldToImageConvert( double x, double y );
  bool isObserved( PM_CARTESIAN &start, double direction, double range, int observeTime);

  //variables
  int pathLength;
  PM_CARTESIAN *pathOut;
  //! victim information
  vector <VictInfo> victInfo;

private:
  //functions
  int modifyArcs( double north, double east );
  double setCost( double graphDist, double north1, double east1, double north2, double east2 );
  void updateGraph( double minx, double miny, double maxx, double maxy );
  bool checkGraph();
  // variables
  //! number of seed nodes from the AM echelon
  int numAmSeedNodes;
  //! the location of the seed nodes
  PM_CARTESIAN *amSeedNodes;
  //! should verbose output be turned on?
  int myVerbose;
  //! The planned path is returned from the searcher here
  int *path;
  //! The distance traveled when going from the center of a cell to a straight exit
  float straightDist;
  //! The distance traveled when going from the center of a cell to a diagnol exit
  float diagDist;
  //! The inverse of the cell resolution
  //  float multFactor;
  //! The graphWidth is associated with y
  int graphWidth;
  //! The graphHeight is associated with x
  int graphHeight;
  //! The graph structure
  DijkHeapNGraph *searcherGraph;
  //! The searcher structure
  DijkHeapNSearcher *theSearcher;
  //! world location in meters of (0,0) of graph
  //  PM_CARTESIAN graphOffset; 
  //! The world model structure as read from the sensors
  VehMobMap *worldMap;
  /*
  unsigned char **worldModel;
  //! The obstacle grow map as maintained by the planner
  char **obsGrowMap;
  //! Map of hazards to be avoided
  unsigned char **hazardMap;
  */
  //! CImg library image structure
  cimg_library::CImg<unsigned char> *cImg;
  //! How big in cells should obstacles be grown (0-no growth)?
  int obsGrow;
  //! Storage for discounted path that must be restored
  int *discountPath;
  //! The length of the discounted path
  int discountPathLength;
  //! Lookup table between world model input value and cost.
  double *costLookup;
  //! cost of path
  double costsOut;
  //! map output message
  SensorData2D *vehMapOut;
  //! spqr output message
  MapMsg *spqrMapOut;
};

#endif // VEH_MOB_PL_ENGINE_HH

/*
  Modification history:

  $Log: vehMobPLEngine.hh,v $
  Revision 1.13  2007/05/16 20:57:21  dr_steveb
  Fixed graph wrapping!

  Revision 1.12  2007/02/27 19:12:19  dr_steveb
  Fixed drawing of no-planning region for vehicle and added mobility world

  Revision 1.11  2007/02/27 14:46:27  dr_steveb
  New interface between AM and VEH

  Revision 1.10  2007/01/30 21:29:02  dr_steveb
  Added new AM-VEH interface

  Revision 1.9  2006/10/31 12:49:48  dr_steveb
  Fixes to interface to SPQR

  Revision 1.8  2006/10/31 09:30:43  dr_steveb
  Changes for SPQR integration

  Revision 1.7  2006/10/24 13:09:49  dr_steveb
  New vehicle level victim locator and output on moastError. MoastError is now read by the notifier program. Use notifier -bmoastError

  Revision 1.6  2006/10/18 00:18:21  dr_steveb
  Added victim channel to vehicle level. It now hunts for victims. Also added CImg library from sourceforge to perform annotation. Modified make files to include this library.

  Revision 1.5  2006/10/10 13:30:19  dr_steveb
  New vehicle level exploration based on camera FOV implemented.

  Revision 1.4  2006/10/03 20:24:25  dr_steveb
  Exploration command added to vehicle level

  Revision 1.3  2006/09/30 16:04:22  dr_steveb
  Changed to use Tom's new world model and added am error checking.

  Revision 1.2  2005/10/17 18:35:13  dr_steveb
  Fixed map growing, added planLength to status message. Added getPlanLength to vehMobPLEngine. Discovered problem with use of ppmsdisplay

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.9  2005/10/11 15:26:34  stephen
  Added plan cost to status message and status printing. This required a new class variable (planCost).

  Revision 1.8  2005/10/03 15:06:31  stephen
  Vehicle level now works with either prim or am by setting a compile flag (USEPRIMMOB)

  Revision 1.7  2005/09/29 13:18:10  stephen
  Beginning of changes to work with Tom's new prim and am sp.

  Revision 1.6  2005/09/21 00:46:07  stephen
  Modifications to work with Adam's am and prim.

  Revision 1.5  2005/07/18 19:56:12  stephen
  Changed interface to confrom to dijkHeapN

  Revision 1.4  2005/06/29 21:08:17  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.3  2005/06/14 19:19:13  stephen
  Changed to use SensorPoint in sensor data message.

  Revision 1.2  2005/06/09 16:43:46  stephen
  Fixed initialization problem.

  Revision 1.1  2005/06/08 14:31:15  proctor
  Added MobJA name changes

  Revision 1.3  2005/05/31 21:19:33  proctor
  Changed tabsGraph, searcher to DijkTabsGraph, DijkTabsSearcher, etc. to
  reflect Tom's changes to the naming convention.

  Revision 1.2  2005/05/24 17:05:07  stephen
  Changed to Tom's new Dijk algorithm.

  Revision 1.1  2005/05/23 18:07:06  stephen
  First add of file.
*/
