/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehMobPLMain.cc
  \brief Provides a layer planning system for use in the vehicle echelon

  This planner operates with several different knowledge layers to plan vehicle
  level trajectories for the system. These trajectories are not dynamically correct.
  The layers included in this program include:
  -  a visibility graph of obstacles. This layer plans paths around
  polygonal objects in the world. The cost function applies one cost for
  traversal outside of polygons, and another for traversal inside of polygons.

  A modified A* search is used in this planner.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.41 $
  $Date: 2007/12/17 21:35:05 $
  \endcode

  \author Stephen Balakirsky
  \date   05/25/2005
*/
//#define USEPRIMMOB

#include <vector>
#include <stack>
#include <stdio.h>
#include <strings.h>
#include <signal.h>    // signal()
#include <unistd.h>    // close()
#include <sys/types.h> // open()
#include <sys/stat.h>  // open()
#include <fcntl.h> // open()
#include <stdlib.h>    // getenv()
#include <rcs.hh>
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "primMobJA.hh"
#include "primSP.hh"
#include "amMobJA.hh"
#include "amSP.hh"
#include "sensorData.hh"
#include "navDataExt.hh"
#include "moastError.hh"
#include "vehMobPL.hh"
#include "vehMobPLEngine.hh"
#include "nmlMap.hh"
#include "config.h"
#if HAVE_MYSQL
#include <mysql.h>
#include "moastMySQL.hh"
#endif

//#define SENSOR_RANGE 3.5
//! Time that the vehicle has to move STUCK_DISTANCE before it is deemed stuck.
#define STUCK_TIME 30.
//! Distance that the vehicle must move to not be stuck
//#define STUCK_DISTANCE .2
#define STUCK_DISTANCE 0

#define MY_NAME_BASE "vehMobPL"
#define DEG(x) (57.295779513082323 * (x))
#define RAD(x)  (0.017453292519943 * (x))

using namespace std;

////////////////////////////
// globals
/////////////////////
//! my verbose level
int myVerbose;
#if HAVE_MYSQL
//! connection to mySql. Only exists if we have  HAVE_MYSQL defined
MoastMySQL sqlConnect;
#endif
//! password to use for sql access
char sqlPassword[MOAST_FILE_NAME_LEN];

//! Should the map be output
bool drawMap;
//! How many planning cycles should be skipped between output cycles
int drawSkip;
//! is the goal global or local?
bool goalIsGlobal;
//! is the map global or local?
bool mapIsGlobal;
//! Transforms from local to global and back
PM_POSE poseToLocal, poseToGlobal;
//! Serial number to am level
int myAmSerial;
// NML Channels
//! am planning channels
RCS_CMD_CHANNEL  *amMobJACmdNml;
RCS_STAT_CHANNEL *amMobJAStatNml;
RCS_STAT_CHANNEL *amMobJASetNml;
RCS_CMD_CHANNEL  *amMobJACfgNml;
//! am SP channels
RCS_CMD_CHANNEL  *amSPCmdNml;
RCS_STAT_CHANNEL *amSPStatNml;
RCS_STAT_CHANNEL *amSPSetNml;
RCS_CMD_CHANNEL  *amSPCfgNml;
//! am sensor data channel
NML              *sensorDataNml;
//! navigation data channel
NML              *navDataExtNml;
//! victim RFID
NML              *victRFIDNml;
//! Error channel, where we send victim information
NML              *moastErrorNml;
//! SPQR Mappipng channel
NML              *spqrMapInNml;
NML              *spqrMapOutNml;
//! veh planning channels
RCS_CMD_CHANNEL  *vehMobPLCmdNml;
RCS_STAT_CHANNEL *vehMobPLStatNml;
RCS_CMD_CHANNEL  *vehMobPLCfgNml;
RCS_STAT_CHANNEL *vehMobPLSetNml;
NML              *vehMapOutNml;

// global commands received or sent
#ifdef USEPRIMMOB
PrimMobJACmdMoveWaypoint *moveWaypointCmd;
PrimMobJAStat *amStatus;
#else
AmMobJACmdMoveWaypoint *moveWaypointCmd;
AmMobJAStat *amStatus;
#endif
AmMobJASet amMobJASettings;

//! where are we anyway?
// NOTE TO ME! ADD TIME IN HERE TO DEAL WITH TIME SHIFT
double tempTime;
PM_POSE vehLocation;
//! goal stack
stack<goalPoint> goalStack;

/*!
  Have we found this victim already? Check for name equal FalsePositive and injury not equal to None
*/
bool victFound( goalPoint goalIn, vector<VictInfo>&victInfo )
{
  vector<VictInfo>::iterator it;
  bool foundVict = false;

  for( it=victInfo.begin(); it!=victInfo.end(); it++ )
    {
      if( myDisp( it->goalLocation, goalIn.location ) < 1 ) // same point
	{
	  foundVict = true;
	  break;
	}
    }
  if( foundVict == false )
    return false;
  if( !strcmp(it->name, "FalsePositive") )
    return true;
  if( strcmp(it->injury, "None") )
    return true;
  return false;
}

/*!
  Transform a point from/to global coordinates to/from local coordinates

\param pointIn The point to be converted
\param transformGlobalToLocal True- transform to local, false- transform to global
\return Transformed point
*/
PM_CARTESIAN transformPoint(PM_CARTESIAN &pointIn, bool transformGlobalToLocal, PM_POSE *transformIn = NULL)
{
  PM_CARTESIAN pointReturn;
  PM_POSE *poseToUse;

  if( transformIn != NULL )
    poseToUse = transformIn;
  else if( transformGlobalToLocal )
    poseToUse = &poseToLocal;
  else
    poseToUse = &poseToGlobal;

  pointReturn = poseToUse->rot * pointIn + poseToUse->tran;
  return pointReturn;
}

/*!
  Compute a simple cofidence measure on the the maximum cost of the path.
  If path goes through an obstacle, the quality is low. 
  If path goes near an obstacle, the quality is ok.
  If path avoids all obstacles, the quality is great

\param maxCost The maximum cost of the path
\param obsCost The cost of an obstacle
\param proxCost The cost of being near an obstacle

\return 0 - low quality
\return 5 - ok quality
\return 10 - great quality
*/
int computeConfidence( double maxCost, double obsCost, double proxCost )
{
  //  fprintf( stderr, "Max Cost is: %lf\n", maxCost );
  if( maxCost >= obsCost )
    return(0);
  else if( maxCost >= proxCost )
    return(5);
  else
    return(10);
}

/*!
  Check the settings back from subordinate planning system
  
  \param settings The settings that were returned

  \return true on good read
  \return false on no new settings
*/

bool checkAmMobSettings( AmMobJASet &amSettings, RCS_STAT_CHANNEL *channel )
{
  int type;

  type = channel->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case AM_MOB_JA_SET_TYPE:
      amSettings = *(AmMobJASet*)channel->get_address();
      return(true);
    default:
      break;
    }
  return false;
}


/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/

#ifdef USEPRIMMOB
bool checkAmMobCmdStatus( PrimMobJAStat **stat, int &echo_serial_number, RCS_STAT_CHANNEL *channel )
#else
bool checkAmMobCmdStatus( AmMobJAStat **stat, int &echo_serial_number, RCS_STAT_CHANNEL *channel )
#endif
{
  int type;

  type = channel->read();
  switch( type )
    {
    case 0: // no new data
      break;
#ifdef USEPRIMMOB
    case PRIM_MOB_JA_STAT_TYPE:
      *stat = (PrimMobJAStat*)channel->get_address();
#else
    case AM_MOB_JA_STAT_TYPE:
      *stat = (AmMobJAStat*)channel->get_address();
#endif
      echo_serial_number = channel->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

#ifdef USEPRIMMOB
bool checkAmSPCmdStatus( PrimSPStat **stat, int &echo_serial_number, RCS_STAT_CHANNEL *channel )
#else
bool checkAmSPCmdStatus( AmSPStat **stat, int &echo_serial_number, RCS_STAT_CHANNEL *channel )
#endif
{
  int type;

  type = channel->read();
  switch( type )
    {
    case 0: // no new data
      break;
#ifdef USEPRIMMOB
    case PRIM_SP_STAT_TYPE:
      *stat = (PrimSPtat*)channel->get_address();
#else
    case AM_SP_STAT_TYPE:
      *stat = (AmSPStat*)channel->get_address();
#endif
      echo_serial_number = channel->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

/*!
  Read in new vehicle position

  \param statIn Channel to read from

  \return true on success
  \return false on failure
*/
bool updatePosition( NML *statIn )
{
  int type;
  NavDataExt *navData;
  //  PM_CARTESIAN trialPt;

  type = statIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case NAV_DATA_EXT_TYPE:
      navData = (NavDataExt*)statIn->get_address();
      tempTime = navData->time;
      if( mapIsGlobal )
	{
	  vehLocation.tran = navData->tranAbs;
	  vehLocation.rot = navData->rpyAbs;
	}
      else
	{
	  vehLocation.tran = navData->tranRel;
	  vehLocation.rot = navData->rpyRel;
	}
      
      // new settings for rotations
      poseToLocal.tran = navData->tranRel - navData->tranAbs;
      poseToGlobal.tran = navData->tranAbs - navData->tranRel; 
      poseToLocal.rot = navData->rpyRel - navData->rpyAbs;
      poseToGlobal.rot = navData->rpyAbs - navData->rpyRel;


      // lets try it
      /*
      trialPt = navData->tranRel;
      trialPt = transformPoint( trialPt, false );
      printf( "Transforming local <%f %f> to global <%f %f>\n",
	      navData->tranRel.x, navData->tranRel.y, trialPt.x, trialPt.y );
      trialPt = navData->tranAbs;
      trialPt = transformPoint( trialPt, true );
      printf( "Transforming global <%f %f> to local <%f %f>\n",
	      navData->tranAbs.x, navData->tranAbs.y, trialPt.x, trialPt.y );
      */
      /* old values
      homGlobal.tran = navData->tranAbs;
      homLocal.tran = navData->tranRel;
      homLocal.rot = navData->rpyRel;
      homGlobal.rot = navData->rpyAbs;
      */
      return(true);
    default:
      return(false);
    }
  return(false);
}

/*!
  Send the planned path to the subordinate

  \param confidence Confidence in path
  \param path       Array of path points
  \param length     Number of path points

  \return true on success
  \return false on failure
*/
bool sendPath( unsigned char confidence, PM_CARTESIAN *path, int length )
{
  int i;

  if( amMobJACmdNml==NULL )
    return false;
  if( length>MAX_WAYPOINTS )
    length = MAX_WAYPOINTS;

  moveWaypointCmd->waypoint_length = length;
  for(i=0; i<length; i++ )
    {
      if( mapIsGlobal )
	{
	  moveWaypointCmd->waypoint[i].p = transformPoint( path[i], true );
	}
      else
	{
	  moveWaypointCmd->waypoint[i].p.x = path[i].x;
	  moveWaypointCmd->waypoint[i].p.y = path[i].y;
	  moveWaypointCmd->waypoint[i].p.z = path[i].z;
	}
      moveWaypointCmd->waypoint[i].neighborhood = .2; // was .5
      moveWaypointCmd->waypoint[i].quality = confidence;
      moveWaypointCmd->waypoint[i].speed = .5; // was 2
      //      printf( "Sending waypoint %d at <%f %f>\n", i, path[i].x, path[i].y );
    }
  moveWaypointCmd->serial_number = ++myAmSerial;

  /////////////////////////////////////
  // for debug only send one point
  moveWaypointCmd->waypoint_length = 1;
  ////////////////////////////////////

  amMobJACmdNml->write(moveWaypointCmd);

  // debug!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //sleep(2);
  //  exit(1);

  /////////////////////////////////////////////////////////////////////////////


  return true;
}

/*!
  Perform abort command state machine

  \param stat status message
*/
void doCmdAbort(VehMobPLStat *stat)
{
#ifdef USEPRIMMOB
  PrimMobJACmdAbort *abortMobCmd;
  PrimMobJAStat *amMobStatus;
  PrimSPStat *amSPStatus;
#else
  AmMobJACmdAbort *abortMobCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
#endif
  AmSPCmdAbort *abortSPCmd;
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
#ifdef USEPRIMMOB
      abortMobCmd = new PrimMobJACmdAbort();
#else
      abortMobCmd = new AmMobJACmdAbort();
#endif
      abortSPCmd = new AmSPCmdAbort;
      abortMobCmd->serial_number = ++myAmSerial;
      abortSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(abortMobCmd);
      amSPCmdNml->write(abortSPCmd);
      delete abortMobCmd;
      delete abortSPCmd;
    } // fall through to checking the status of the mobility
  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    }
}

/*!
  Perform  halt command state machine

  \param stat status message
  \return true need to update and send set message
  \return false no message needs to be sent
*/
bool doCmdHalt(VehMobPLStat *stat)
{
  bool retValue = false;
#ifdef USEPRIMMOB
  PrimMobJACmdHalt *haltMobCmd;
  PrimMobJAStat *amMobStatus;
  PrimSPStat *amSPStatus;
#else
  AmMobJACmdHalt *haltMobCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
#endif
  AmSPCmdHalt *haltSPCmd;
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      retValue = true;
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
#ifdef USEPRIMMOB
      haltMobCmd = new PrimMobJACmdHalt();
#else
      haltMobCmd = new AmMobJACmdHalt();
#endif
      haltSPCmd = new AmSPCmdHalt();
      haltMobCmd->serial_number = ++myAmSerial;
      haltSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(haltMobCmd);
      amSPCmdNml->write(haltSPCmd);
      delete haltMobCmd;
      delete haltSPCmd;
    } // fall through to checking the satus
  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  retValue = true;
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    }
  return retValue;
}

/*!
  Perform init state machine

  \param stat status message
  \return true when settings and status needs updating
  \return false no message needs to be sent
*/
bool doCmdInit(VehMobPLStat *stat, VehMobPLSet *settings, VehMobPLEngine *engine)
{
  bool retValue = false;
#ifdef USEPRIMMOB
  PrimMobJACmdInit *initMobCmd;
  PrimSPCmdGo *goSPCmd;
  PrimSPCmdInit *initSPCmd;
  PrimMobJAStat *amMobStatus;
  PrimSPStat *amSPStatus;
#else
  AmMobJACmdInit *initMobCmd;
  AmSPCmdGo *goSPCmd;
  AmSPCmdInit *initSPCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
#endif
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      retValue = true;
      state_new(stat);
      if( settings->isInitialized == true )
	{
	  status_next(stat, RCS_DONE);
	  state_next(stat, S3);
	}
      else
	{
	  status_next(stat, RCS_EXEC);
	  state_next(stat, S0);
	}
#ifdef USE_SPQR // goto reading navigation data since it does not accept an init
      state_next(stat, S4);
      return retValue;
#endif
#ifdef USEPRIMMOB
      initMobCmd = new PrimMobJACmdInit();
      initSPCmd = new PrimSPCmdInit();
#else
      initMobCmd = new AmMobJACmdInit();
      initSPCmd = new AmSPCmdInit();
#endif
      initMobCmd->serial_number = ++myAmSerial;
      initSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(initMobCmd);
      amSPCmdNml->write(initSPCmd);
      delete initMobCmd;
      delete initSPCmd;
    } // fall through to checking the status 
  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  /*
	  printf( "In doCmdInit state 0, serial  %d %d", 
		  echo_serial_number, myAmSerial );
	  printf( " status: %d\n", amMobStatus->status );
	  */
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		{
		  state_next(stat, S1 );
		  retValue = true; // update status since switching state
		}
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S6);
		  retValue = true;
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  /*
	  printf( "In doCmdInit state 1, serial  %d %d", 
		  echo_serial_number, myAmSerial );
	  printf( " status: %d\n", amSPStatus->status );
	  */
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  retValue = true; // update status since switching state
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S6);
		}
	    }
	}
    } // fall through to sending "go" command
  if( state_match(stat, S2) )
    {
      //      printf( "In doCmdInit state 2\n" );
#ifdef USEPRIMMOB
      goSPCmd = new PrimSPCmdGo();
#else
      goSPCmd = new AmSPCmdGo();
#endif
      goSPCmd->serial_number = ++myAmSerial;
      amSPCmdNml->write(goSPCmd);
      retValue = true;
      state_next(stat, S3);
      delete goSPCmd;
      retValue = true; // update status since switching state
    } // fall through to checking status
  if( state_match(stat, S3) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  /*
	  printf( "In doCmdInit state 3, serial  %d %d", 
		  echo_serial_number, myAmSerial );
	  printf( " status: %d\n", amSPStatus->status );
	  */
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE || amSPStatus->status == RCS_EXEC )
		{
		  retValue = true;
		  status_next(stat, RCS_EXEC);
		  state_next(stat, S4 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  retValue = true;
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S6);
		}
	    }
	}
    } // fall through to reading navigation data
  if( state_match(stat, S4) )
    {
      status_next(stat, RCS_EXEC);
      if( updatePosition(navDataExtNml) )
	{
	  retValue = true;
	  state_next(stat, S5);
	  if( stat->status != RCS_ERROR )
	    sprintf( stat->message, "\n" );
	}
      else
	sprintf( stat->message, "Waiting for navData\n");
    }// fall througgh to setting the graph offsets
  if( state_match(stat, S5) )
    {
      engine->init( vehLocation );
      engine->buildGraph();
      retValue = true;
      settings->isInitialized = true;
      status_next(stat, RCS_DONE);
      state_next(stat, S6);
    }
  return retValue;
}

/*!
  Perform shutdown command to subordinate planning system

  \param stat vehicle status
  \return true done with shutdown
  \return false not done yet
*/
bool doCmdShutdown(VehMobPLStat *stat)
{
  bool retValue = false;
#ifdef USEPRIMMOB
  PrimMobJACmdShutdown *shutdownMobCmd;
  PrimMobJAStat *amMobStatus;
  PrimSPStat *amSPStatus;
#else
  AmMobJACmdShutdown *shutdownMobCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
#endif
  AmSPCmdShutdown *shutdownSPCmd;
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      retValue = true;
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);

#ifdef USEPRIMMOB
      shutdownMobCmd = new PrimMobJACmdShutdown();
#else
      shutdownMobCmd = new AmMobJACmdShutdown();
#endif
      shutdownSPCmd = new AmSPCmdShutdown();
      shutdownMobCmd->serial_number = ++myAmSerial;
      shutdownSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(shutdownMobCmd);
      amSPCmdNml->write(shutdownSPCmd);
      delete shutdownMobCmd;
      delete shutdownSPCmd;
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  retValue = true;
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    }
  return retValue;
}

/*!
  Perform noop command to subordinate planning system

  \param stat vehicle status
*/
void doCmdNop(VehMobPLStat *stat)
{
  /*
#ifdef USEPRIMMOB
  PrimMobJACmdNop *nopMobCmd;
#else
  AmMobJACmdNop *nopMobCmd;
#endif
  AmSPCmdNop *nopSPCmd;
  */
#ifdef USEPRIMMOB
  PrimMobJAStat *amMobStatus;
  PrimSPStat *amSPStatus;
#else
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
#endif

  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_DONE);
      state_next(stat, S2);
      /*
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);

#ifdef USEPRIMMOB
      nopMobCmd = new PrimMobJACmdNop();
#else
      nopMobCmd = new AmMobJACmdNop();
#endif
      nopSPCmd = new AmSPCmdNop();
      nopMobCmd->serial_number = ++myAmSerial;
      nopSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(nopMobCmd);
      amSPCmdNml->write(nopSPCmd);
      delete nopMobCmd;
      delete nopSPCmd;
      */
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    }
}

/*! 
  Perform exploration planning state machine

  \param stat vehicle status
*/
// see vehMobPL.hh for state names and values
void doCmdMovement(VehMobPLStat *stat, VehMobPLSet *settings, VehMobPLEngine *engine, int movementType, 
		   vector<VictInfo>&victInfo)
{
  vector<VictInfo>::iterator it;
  double i, j;
  goalPoint goalToSend;
  PM_CARTESIAN obsPt;
  PM_RPY vehRpy;
  int confidence, searchReturn;
  int echo_serial_number;
  AmMobJACmdSpin amSpinCmd;
  AmMobJACmdPrepareMove amPrepareMoveCmd;
  static double observeAngle;
  static int cycleCount, stuckCount; // , planCount=0;
  static bool newSearch = true;
  double endTime = etime() + settings->cycleTime; // time the cycle must end
  /*! The next variable block is used to assure that the vehicle is moving 
      and not stuck. The vehicle must move a STUCK_DISTANCE of Ecludian distance
      in STUCK_TIME to not be stuck
  */
  //! location that the STUCK_DISTANCE is computed from
  static PM_CARTESIAN lastLocation;
  //! time that the STUCK_TIME is computed from
  static double lastMoveTime;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      /*
      if( planCount >= stat->heartbeat + 10 )
	return;
      planCount = stat->heartbeat;
      */
      lastLocation = vehLocation.tran;
      lastMoveTime = etime();
      stuckCount = 0;
      if( movementType == VEH_MOB_PL_CMD_EXPLORE_TYPE )
	{
	  // add all victims to list
	  for( it=victInfo.begin(); it!=victInfo.end(); it++ )
	    {
	      goalToSend.location = it->goalLocation;
	      goalToSend.neighborhood = 1;
	      goalToSend.isGlobal = mapIsGlobal;
	      goalToSend.goalType = VICTIM;
	      goalStack.push(goalToSend);
	    }
	  observeAngle = -RFIDHalfFOV;
	  state_next( stat, S_START_OBS );
	}
      else
	state_next(stat, S_PREPARE_MOVE);
    }

  if( state_match(stat, S_PREPARE_MOVE) ) // prepare am for movement
    {
      /* by setting lastMoveTime to -1, we will start planning after 
         preparing to move */
      lastMoveTime = -1;
      amPrepareMoveCmd.serial_number = ++myAmSerial;
      amMobJACmdNml->write(amPrepareMoveCmd);
      state_next(stat, S_CHECK_AM_STAT_FROM_PREPARE);
      return;
    }

  if( state_match(stat, S_CHECK_STUCK) )
    {
      //      observeAngle = -RAD(45);
      observeAngle = -RFIDHalfFOV;
      cycleCount = 0;
      if( myDisp(vehLocation.tran, lastLocation) < STUCK_DISTANCE )
	{
	  //	  printf( "%f seconds until spin!\n", STUCK_TIME - (etime()-lastMoveTime) );
	  if( etime()-lastMoveTime>STUCK_TIME )
	    {
	      lastMoveTime = -1;
	      state_next( stat, S_STUCK );
	      status_next(stat, RCS_EXEC );
	      return;
	    }
	}
      else
	{
	  lastLocation = vehLocation.tran;
	  lastMoveTime = etime();
	  stuckCount = 0;
	  //	  printf( "Reseting last move time...\n" );
	}
      if( movementType == VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE )
	{
	  state_next( stat, S_BLOCK_CENTER );
	  status_next(stat, RCS_EXEC);
	}
      else if( movementType == VEH_MOB_PL_CMD_EXPLORE_TYPE )
	{
	  state_next( stat, S_START_OBS );
	  status_next(stat, RCS_EXEC);
	}
      else
	{
	  status_next(stat, RCS_ERROR );
	  sprintf(stat->message, "Illegal movement type %d\n", movementType );
	  state_next(stat, S_MOVEMENT_ERR);
	}
    }
  if( state_match(stat, S_START_OBS) ) // start observations
    {
      observeAngle += RFIDHalfFOV*2;
      if( observeAngle > 2*M_PI )
	{
	  state_next(stat, S_PREPARE_MOVE);
	  return;
	}

      //      printf( "checking to observe at %f %f\n", observeAngle, DEG(observeAngle) );
      // below was SENSOR_RANGE
      if( !engine->isObserved( vehLocation.tran, observeAngle, RFIDRange*.8, stat->tt.count ) )
	{
	  //	  if( myVerbose )
	  printf( "Observing %f (%f degrees)\n", observeAngle, DEG(observeAngle) );
	  state_next(stat, S_CHECK_AM_STAT_FROM_SPIN);
	  amSpinCmd.serial_number = ++myAmSerial;
	  amSpinCmd.direction = MOAST_ROTATE_SHORTEST;
	  amSpinCmd.absAngle = observeAngle;
	  vehRpy = vehLocation.rot;
	  /* commented out fudge factor
	  if( observeAngle > vehRpy.y )
	    observeAngle += RAD(5);
	  else
	    observeAngle -= RAD(5);
	  */
	  amSpinCmd.tolerance = RAD(5);
	  amMobJACmdNml->write(amSpinCmd);
	  return;
	}
    }
  if( state_match(stat, S_CHECK_AM_STAT_FROM_SPIN) ) // check AM status
    {
      if( checkAmMobCmdStatus( &amStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amStatus->status == RCS_ERROR )
		{
		  if( amStatus->amMobJAError == AM_MOB_JA_ERROR_PLAN_HORIZON )
		    {
		      state_next(stat, S_PREPARE_MOVE);
		      sprintf( stat->message, "AM Err: AM_MOB_JA_ERROR_PLAN_HORIZON" );
		      fprintf( stderr, "%s\n", stat->message );
		    }
		  else if( amStatus->amMobJAError ==  AM_MOB_JA_ERROR_POSITION )
		    {
		      if( lastMoveTime == -1 )
			{
			  state_next(stat, S_PREPARE_MOVE);
			  sprintf( stat->message, "AM Err: AM_MOB_JA_ERROR_POSITION; preparing..." );
			  fprintf( stderr, "%s\n", stat->message );
			  lastLocation = vehLocation.tran;
			  lastMoveTime = etime();
			}
		      else
			{
			  sprintf( stat->message, "AM Err: AM_MOB_JA_ERROR_POSITION; starting obs..." );
			  fprintf( stderr, "%s\n", stat->message );
			  observeAngle -= RFIDHalfFOV*2;
			  state_next(stat, S_START_OBS);
			}
		    }
		  else
		    {
		      status_next(stat, RCS_ERROR);
		      sprintf( stat->message, "AM Err reported: %d", amStatus->amMobJAError );
		      fprintf( stderr, "%s\n", stat->message );
		      state_next(stat, S_MOVEMENT_ERR);
		    }
		}
	      else if( amStatus->status == RCS_DONE )
		{
		  if( lastMoveTime == -1 )
		    state_next(stat, S_PREPARE_MOVE);
		  else
		    state_next(stat, S_START_OBS);
		  lastLocation = vehLocation.tran;
		  lastMoveTime = etime();
		  //		  stuckCount = 0;

		}
	    }
	}
    } // fall through to planning
  if( state_match(stat, S_CHECK_AM_STAT_FROM_PREPARE) ) // check AM status
    {
      if( checkAmMobCmdStatus( &amStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amStatus->status == RCS_ERROR )
		{
		  if( amStatus->amMobJAError == AM_MOB_JA_ERROR_PLAN_HORIZON )
		    {
		      state_next(stat, S_PREPARE_MOVE);
		      sprintf( stat->message, "AM Err: AM_MOB_JA_ERROR_PLAN_HORIZON" );
		      fprintf( stderr, "%s\n", stat->message );
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_POSITION )
		    {
		      state_next(stat, S_PREPARE_MOVE);
		      sprintf( stat->message, "AM Err: AM_MOB_JA_ERROR_POSITION" );
		      fprintf( stderr, "%s\n", stat->message );
		    }
		  else
		    {
		      status_next(stat, RCS_ERROR);
		      sprintf( stat->message, "AM Err reported: %d", amStatus->amMobJAError );
		      fprintf( stderr, "%s\n", stat->message );
		      state_next(stat, S_MOVEMENT_ERR);
		    }
		}
	      else if( amStatus->status == RCS_DONE )
		{
		  state_next(stat, S_BLOCK_CENTER);
		  lastLocation = vehLocation.tran;
		  lastMoveTime = etime();
		  return;
		}
	    }
	}
    } 

  if( state_match(stat, S_BLOCK_CENTER) )
    {
      PM_CARTESIAN planCenter[500];
      int planCenterLength = 0;
      state_next(stat, S_PLAN_ROUTE);

      //      /*  comment out blocking of center area for now
      //top
      j=vehLocation.tran.y-(stat->planStartHorizon*.8); 
      for( i=vehLocation.tran.x-(stat->planStartHorizon*.8); 
	   i<vehLocation.tran.x+(stat->planStartHorizon*.8);
	   i+= settings->cellResolution )
	{
	  planCenter[planCenterLength].x = i;
	  planCenter[planCenterLength].y = j;
	  planCenterLength++;
	}
      //bottom
      j=vehLocation.tran.y+(stat->planStartHorizon*.8); 
      for( i=vehLocation.tran.x-(stat->planStartHorizon*.8); 
	   i<vehLocation.tran.x+(stat->planStartHorizon*.8);
	   i+= settings->cellResolution )
	{
	  planCenter[planCenterLength].x = i;
	  planCenter[planCenterLength].y = j;
	  planCenterLength++;
	}
      //left
      i=vehLocation.tran.x-(stat->planStartHorizon*.8); 
      for( j=vehLocation.tran.y-(stat->planStartHorizon*.8); 
	   j<vehLocation.tran.y+(stat->planStartHorizon*.8);
	   j+= settings->cellResolution )
	{
	  planCenter[planCenterLength].x = i;
	  planCenter[planCenterLength].y = j;
	  planCenterLength++;
	}
      //right
      i=vehLocation.tran.x+(stat->planStartHorizon*.8); 
      for( j=vehLocation.tran.y-(stat->planStartHorizon*.8); 
	   j<vehLocation.tran.y+(stat->planStartHorizon*.8);
	   j+= settings->cellResolution )
	{
	  planCenter[planCenterLength].x = i;
	  planCenter[planCenterLength].y = j;
	  planCenterLength++;
	}
      engine->blockClosestWay( planCenter, planCenterLength, stat->tt.count+1, NULL, 
      KEEP_OUT, false );
    }// fall through to planning

  if( state_match(stat, S_PLAN_ROUTE) )
    {
      // find goal
      for(;;)
	{
	  if( goalStack.size() == 0 ) // all done
	    {
	      status_next(stat, RCS_DONE);
	      state_next(stat, S_DONE_MOVEMENT);
	      return;
	    }
	  goalToSend = goalStack.top();
	  //	  printf("Size of goal stack: %d current: <%f %f>\n", goalStack.size(),
	  //		 goalToSend.location.x, goalToSend.location.y);
	  if( goalToSend.goalType == WAYPOINT )
	    break;
	  // find a victim that we don't know about or a waypoint to go to
	  if( victFound(goalToSend, engine->victInfo) || movementType==VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE)
	    {
	      goalStack.pop();
	      if(goalStack.size() == 0)
		{
		  state_next(stat, S_DONE_MOVEMENT);
		  status_next(stat, RCS_DONE );
		  sprintf( stat->message, "no waypoint\n" );
		  break;
		}
	    }
	  else
	    break;
	}
      
      if( goalToSend.isGlobal && !mapIsGlobal )
	{
	  goalToSend.location = transformPoint(goalToSend.location, true);
	  goalToSend.isGlobal = false;
	}
      else if( !goalToSend.isGlobal && mapIsGlobal )
	{
	  goalToSend.location = transformPoint(goalToSend.location, false);
	  goalToSend.isGlobal = true;
	}
      // are we close enough to the goal to only send goal point?
      if( goalToSend.goalType != WAYPOINT )
	printf( "Vict at: <%f %f> vehicle at: <%f %f>\n",
		goalToSend.location.x,
		goalToSend.location.y,
		vehLocation.tran.x,
		vehLocation.tran.y );
      if( myDisp(goalToSend.location, vehLocation.tran) <= stat->planStartHorizon )
	{
	  printf( "Sending single goal point to vehicle\n" );
	  sendPath( 10, &goalToSend.location, 1 );
	  state_next(stat, S_CHECK_AM_ERR);
	}
      else
	{
	  if( settings->usePathDiscount && newSearch)
	    engine->discount(true);

	  // seed graph with AM points
	  if( newSearch )
	    {
	      // last seed is vehicle location
	    engine->addGraphPoints( amStatus->seed, amStatus->seed_length-1 ); 
	    }

	  searchReturn = engine->search( goalToSend.location, 
					 stat->planStartHorizon, 
					 newSearch, endTime ); 
	  if( searchReturn==1 || searchReturn==2 )
	    {
	      newSearch=true;
	      status_next(stat, RCS_EXEC);
	      confidence = computeConfidence( engine->maxCost((int)(settings->confDistance/settings->cellResolution)), 
					      settings->zeroConf, settings->fiveConf );
	      stat->planCost = engine->getCost();
	      stat->planLength = engine->getPathLength();
	      sendPath( confidence, engine->pathOut, engine->getPathLength() );
	      /* not correcto technique for computing done
		 if( engine->getPathLength() <= 2 )
		 {
		 state_next(stat, S_MOVEMENT_ERR);
		 status_next(stat, RCS_DONE);
		 }
	      */
	      if( settings->usePathDiscount )
		engine->discount(false);
	      state_next(stat, S_CHECK_AM_ERR);
	    }
	  else if( searchReturn==0 ) // not done
	    {
	      newSearch = false;
	      state_next(stat, S_PLAN_ROUTE );
	    }
	  else
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S_MOVEMENT_ERR);
	      //	  sendAbort();
	    }
	}
      return;
    } // check status
  if( state_match(stat, S_CHECK_AM_ERR) )
    {
      stat->planStartHorizon = amMobJASettings.planHorizon; // reset the start horizon
      if( checkAmMobCmdStatus( &amStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  sprintf(stat->message, "No reports" );
	  if( echo_serial_number == myAmSerial )
	    {
	      if( ++cycleCount > 5 )
		state_next(stat, S_CHECK_STUCK );
	      else
		state_next(stat, S_BLOCK_CENTER);
	      if( amStatus->status == RCS_ERROR )
		{
		  state_next(stat, S_PREPARE_MOVE);
		  if( amStatus->amMobJAError == AM_MOB_JA_ERROR_WAYPOINT )
		    {
		      if(amStatus->seed_length == 0)
			sprintf(stat->message, "too many waypoints");
		      else if(amStatus->seed_length == 1)
			{
			  sprintf(stat->message, "bad waypoint location (%.2lf, %.2lf)",
				 amStatus->seed[0].x, amStatus->seed[0].y);
			  // mark the cell as an obstacle for the next 250 cycles
			  goalToSend = goalStack.top();
			  if( goalToSend.isGlobal && !mapIsGlobal )
			    {
			      goalToSend.location = transformPoint(goalToSend.location, true);
			      goalToSend.isGlobal = false;
			    }
			  else if( !goalToSend.isGlobal && mapIsGlobal )
			    {
			      goalToSend.location = transformPoint(goalToSend.location, false);
			      goalToSend.isGlobal = true;
			    }

			  //print out problem path
			  printf( "Problem with path node: %lf %lf\n", amStatus->seed[0].x, amStatus->seed[0].y );
			  for( int jj=0; jj<engine->getPathLength(); jj++ )
			    {
			      printf( "%d %lf %lf\n", jj, 
				      engine->pathOut[jj].x,
				      engine->pathOut[jj].y );
			    }


			  if( engine->blockClosestWay( &amStatus->seed[0], 1, stat->tt.count +250, 
						       &goalToSend, AM_POINT, true ) )
			    {
			      // need to move goal
			      goalStack.pop();
			      obsPt.x = goalToSend.location.x - (goalToSend.location.x-vehLocation.tran.x)/
				myDisp(vehLocation.tran, goalToSend.location) * settings->cellResolution;
			      obsPt.y = goalToSend.location.y - (goalToSend.location.y-vehLocation.tran.y)/
				myDisp(vehLocation.tran, goalToSend.location) * settings->cellResolution;
			      if( myDisp(vehLocation.tran, obsPt) > 2*settings->cellResolution )
				{
				  printf( "GOAL BLOCKED: moving from <%f %f> to <%f %f>\n",
					  goalToSend.location.x,
					  goalToSend.location.y,
					  obsPt.x,
					  obsPt.y );
				  goalToSend.location = obsPt;
				  goalStack.push(goalToSend);
				}
			      else // can't move goal with current movement stragey
				printf( "GOAL UNREACHABLE <%f %f> moving on...\n", 
					goalToSend.location.x,
					goalToSend.location.y );
			    }
			}
		      else if(amStatus->seed_length == 3)
			{
			  sprintf(stat->message, "waypoints (%.2lf, %.2lf) and (%.2lf, %.2lf) too "
				  "close", amStatus->seed[0].x, amStatus->seed[0].y,
				  amStatus->seed[1].x, amStatus->seed[1].y);
			}
		      else
			sprintf(stat->message, "unspecified am waypoint problem");
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_MOVE )
		    {
		      sprintf(stat->message, "am reports: AM_MOB_JA_ERROR_MOVE, trying to prepare again..." );
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_PLAN_PRIM )
		    {
		      sprintf(stat->message, "am reports: AM_MOB_JA_ERROR_PRIM, trying to spin..." );
		      lastMoveTime = -1;
		      //		      state_next( stat, S_STUCK );
		      state_next( stat, S_PLAN_ROUTE );
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_POSITION )
		    {
		      sprintf(stat->message, "am reports: AM_MOB_JA_ERROR_POSITION, preparing again..." );
		      state_next( stat, S_PREPARE_MOVE );
		    }
		  else
		    {
		      sprintf(stat->message, "Unspecified am error: %d", amStatus->amMobJAError );
		    }
		  fprintf( stderr, "%s\n", stat->message );
		}
	      else if( amStatus->status == RCS_DONE )
		{
		  // we are done with this move
		  goalToSend = goalStack.top();
		  if( goalToSend.goalType == WAYPOINT )
		    {
		      if( hypot( goalToSend.location.x-vehLocation.tran.x, 
				 goalToSend.location.y-vehLocation.tran.y ) < 
			  goalToSend.neighborhood )
			goalStack.pop();
		      state_next(stat, S_CHECK_STUCK);
		    }
		  else
		    {
		      state_next(stat, S_STUCK);
		    }
		    
		}
	    }
	}
    }
  if( state_match(stat, S_STUCK) ) // we have been sitting, so start observations
    {
      // first clear out any obstacles placed due to the amPlanner
      engine->clearWorld(stat->tt.count);
      vehRpy = vehLocation.rot;
      observeAngle = vehRpy.y - RAD(10);
      if( observeAngle < 0 )
	observeAngle += 2*M_PI;
      printf( "Stuck with count %d, so observing %f (%f degrees)\n", ++stuckCount, 
	      observeAngle, DEG(observeAngle) );
      amSpinCmd.serial_number = ++myAmSerial;
      amSpinCmd.direction = MOAST_ROTATE_CW;
      amSpinCmd.absAngle = observeAngle;
      amSpinCmd.tolerance = RAD(5);
      amMobJACmdNml->write(amSpinCmd);
      state_next(stat, S_CHECK_AM_STAT_FROM_SPIN);
      return;
    }
}

/*!
  Perform clear wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgClearWM(VehMobPLSet *stat, VehMobPLEngine *engine)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      retValue = true;
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      engine->clearMap();
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}

/*!
  Perform nop

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgNop(VehMobPLSet *stat)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      retValue = true;
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}

/*!
  Perform load wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCmdLoadWM(VehMobPLStat *stat, VehMobPLSet *settings, VehMobPLEngine *engine)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      if( !mapIsGlobal )
	{
	  fprintf( stderr, "\n\n\n" );	
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "Map storage/retrieval not supported in local maps\n" );
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "\n\n\n" );	
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S1);
	}
      else
	{
	  status_next(stat, RCS_EXEC);
	  state_next(stat, S0);
	}
      retValue = true;
    } // fall through to loading the file
      
  if( state_match(stat, S0) )
    {
      engine->loadWM( settings->wmFileName );
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}

/*!
  Perform store wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgStoreWM(VehMobPLSet *stat, VehMobPLSet *settings, VehMobPLEngine *engine)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      if( !mapIsGlobal )
	{
	  fprintf( stderr, "\n\n\n" );	
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "Map storage/retrieval not supported in local maps\n" );
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "\n\n\n" );	
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S1);
	}
      else
	{
	  status_next(stat, RCS_EXEC);
	  state_next(stat, S0);
	}
      retValue = true;
    } // fall through to storing the file

  if( state_match(stat, S0) )
    {
      engine->storeWM( settings->wmFileName );
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}


/*!
  Perform dump wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgDumpWM(VehMobPLSet *settings, VehMobPLEngine *engine)
{
  bool retValue = false;
  static int skip = 0;
  char sysStr[200];
  int mag;

  if( state_match(settings, NEW_COMMAND) )
    {
      state_new(settings);
      sprintf( sysStr, "rm -rf %s", settings->wmFileName );
      if( system(sysStr) == -1)
	{
	  status_next(settings, RCS_ERROR);
	  state_next(settings, S4);
	}
      else
	{
	  if( settings->cellResolution <= .1 ) mag = 1;
	  else if( settings->cellResolution <= .2 ) mag = 2;
	  else mag = 4;
	  sprintf(sysStr,"./imageview -f%s -s%d&\n", 
		  settings->wmFileName, mag );
	  if( system(sysStr) == -1 )
	    {
	      status_next(settings, RCS_ERROR);
	      state_next(settings, S4);
	    }
	  else
	    {
	      skip = settings->fileSkip;
	      status_next(settings, RCS_EXEC);
	      state_next(settings, S0);
	    }
	}
	  retValue = true;
    } // if no error, fall through to seeing if file should be written

  if( state_match(settings, S0) )
    {
      if( skip++ >= settings->fileSkip )
	{
	  status_next(settings, RCS_EXEC);
	  state_next(settings, S1);
	}
      // no set settings have changed, so do not write
    } // possibly fall through to writing file
  if( state_match(settings, S1) )
    {
      engine->dump( settings->wmFileName, vehLocation.tran );
      if( settings->fileSkip == -1 ) // done
	{
	  status_next(settings, RCS_DONE);
	  state_next(settings, S4);
	  retValue = true;
	}
      else
	{
	  skip = 0;
	  status_next(settings, RCS_EXEC);
	  state_next(settings, S0);
	}
    }
  return retValue;
}

/*!
  Perform sql initialization

  \param stat    vehicle status
  \param sqlParams mysql parameter block

  \return        true if set channel needs to be updated
*/
bool doCfgSQL(VehMobPLSet *stat )
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      state_next(stat, S1);

#ifdef HAVE_MYSQL
      if( sqlConnect.init( stat->sqlServer, stat->sqlUser,
			   sqlPassword,
			   stat->sqlDatabase ) != 1 )
	{
	  stat->sqlConnected = false;
	  status_next(stat, RCS_ERROR);
	}
      else
	{
	  stat->sqlConnected = true;
	  status_next(stat, RCS_DONE);
	}
      retValue = true;
#else
      // no sql compiled in, so command is an error
      fprintf( stderr, "mysql not supported\n" );
      status_next(stat, RCS_ERROR);
      retValue = true;
#endif
    }
  return retValue;
}


/*!
  Read in command and act on it

\param cmdIn         The handle of the channel for commands to be read from
\param stat          The status return

*/
RCS_STATUS readCfg( RCS_CMD_CHANNEL *cmdIn, VehMobPLSet *currentStat )
{
  VehMobPLCfgStoreWM      *storeWMCmd;
  VehMobPLCfgDumpWM       *dumpWMCmd;
  VehMobPLCfgSQL          *sqlCmd;
  int type;
  int serial_number;

  type = cmdIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_CFG_NOP_TYPE:
    case VEH_MOB_PL_CFG_CYCLE_TIME_TYPE:
    case VEH_MOB_PL_CFG_DEBUG_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CFG_CLEAR_WM_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CFG_STORE_WM_TYPE:
      storeWMCmd = (VehMobPLCfgStoreWM*)cmdIn->get_address();
      serial_number = storeWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  strncpy( currentStat->wmFileName, storeWMCmd->fileName, MOAST_FILE_NAME_LEN );
	}
      break;
    case VEH_MOB_PL_CFG_DUMP_WM_TYPE:
      dumpWMCmd = (VehMobPLCfgDumpWM*)cmdIn->get_address();
      serial_number = dumpWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  if( !strcmp(dumpWMCmd->fileName, "uninitialized")  )
	    strncpy( currentStat->wmFileName, "VehMobPLDispOutput.png", MOAST_FILE_NAME_LEN );
	  else
	    strncpy( currentStat->wmFileName, dumpWMCmd->fileName, MOAST_FILE_NAME_LEN );
	  if( dumpWMCmd->dumpContinuous )
	    currentStat->fileSkip = dumpWMCmd->skip;
	  else
	    currentStat->fileSkip = -1;	    
	}
      break;
    case VEH_MOB_PL_CFG_SQL_TYPE:
      sqlCmd = (VehMobPLCfgSQL*)cmdIn->get_address();
      serial_number = sqlCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;

	  strncpy( currentStat->sqlServer, sqlCmd->server, MOAST_FILE_NAME_LEN );
	  strncpy( currentStat->sqlUser, sqlCmd->user, MOAST_FILE_NAME_LEN );
	  strncpy( currentStat->sqlDatabase, sqlCmd->database, MOAST_FILE_NAME_LEN );
	  strncpy( sqlPassword, sqlCmd->password, MOAST_FILE_NAME_LEN );
	}
      break;
    default:
      fprintf( stdout, "Command %s not yet supported\n", vehMobPL_symbol_lookup(type) );
      currentStat->command_type = type;
      return RCS_ERROR;
    }

  return RCS_DONE;
}

/*!
  Read in command and act on it

\param cmdIn         The handle of the channel for commands to be read from
\\param currentStat   The current status of the module

*/
void readCmd( RCS_CMD_CHANNEL *cmdIn, VehMobPLStat *currentStat, VehMobPLSet *currentSet )
{
  VehMobPLCmdMoveWaypoint *waypointCmd;
  VehMobPLCmdExplore *exploreCmd;
  int type, serial_number;
  static int  noPrint = 0;
  goalPoint *newGoalPoint;
  VehMobPLCmdLoadWM       *loadWMCmd;


  type = cmdIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_CMD_ABORT_TYPE:
    case VEH_MOB_PL_CMD_HALT_TYPE:
    case VEH_MOB_PL_CMD_INIT_TYPE:
    case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
    case VEH_MOB_PL_CMD_NOP_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
      waypointCmd = (VehMobPLCmdMoveWaypoint*)cmdIn->get_address();
      serial_number = waypointCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  while( !goalStack.empty() ) // empty out stack
	    {
	      goalStack.pop();
	    }
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  newGoalPoint = new goalPoint();
	  newGoalPoint->location = waypointCmd->p;
	  newGoalPoint->isGlobal = waypointCmd->isGlobal;
	  newGoalPoint->goalType = WAYPOINT;
	  newGoalPoint->neighborhood = waypointCmd->neighborhood;
	  goalStack.push(*newGoalPoint);
	  delete newGoalPoint;
	}
      break;
    case VEH_MOB_PL_CMD_EXPLORE_TYPE:
      exploreCmd = (VehMobPLCmdExplore*)cmdIn->get_address();
      serial_number = exploreCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  while( !goalStack.empty() ) // empty out stack
	    {
	      goalStack.pop();
	    }
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  newGoalPoint = new goalPoint();
	  newGoalPoint->location = exploreCmd->p;
	  newGoalPoint->isGlobal = exploreCmd->isGlobal;
	  newGoalPoint->goalType = WAYPOINT;
	  newGoalPoint->neighborhood = exploreCmd->neighborhood;
	  goalStack.push(*newGoalPoint);
	  delete newGoalPoint;
	}
      break;
    case VEH_MOB_PL_CMD_LOAD_WM_TYPE:
      loadWMCmd = (VehMobPLCmdLoadWM*)cmdIn->get_address();
      serial_number = loadWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  strncpy( currentSet->wmFileName, loadWMCmd->fileName, MOAST_FILE_NAME_LEN );
	}
      break;


    default:
      if( noPrint < 10 )
	{
	  fprintf( stdout, "Command %s not yet supported\n", vehMobPL_symbol_lookup(type) );
	  noPrint++;
	}
      break;
    }
  return;
}

/*!
  Read in victims and populate world model

\param victChan The channel to read from

\return True if new victim or new information
        False if no new information
*/
bool readVictims( NML *victChan, vector<VictInfo>&victInfo, int cmdType )
{
  static int id = 0;
  char *sPtr, *token;
  int t;
  SensorDataText *sensorDataTxt;
  int type;
  PM_CARTESIAN transformPt;
  vector<VictInfo>::iterator it;
  bool foundVict;
  VictInfo newInfo;
  bool retValue = false;
  goalPoint *newGoalPoint;
  MoastError moastError;

  // victims have changed so much that we should ignore them for now
  return false;
  if( victChan == NULL )
    return false;
  type = victChan->read();
  switch( type )
    {
    case 0: // no new data
      return false;
    case SENSOR_DATA_TXT_TYPE:
      sensorDataTxt = (SensorDataText*)victChan->get_address();
      for( t=0;t<sensorDataTxt->strPtr_length;t++) 
	{
	  transformPt.x = sensorDataTxt->dVal[t*4];
	  transformPt.y = sensorDataTxt->dVal[t*4+1];
	  transformPt.z = sensorDataTxt->dVal[t*4+2];
	  //	  printf( "Time of navData:vict %lf %lf\n", tempTime, sensorDataTxt->dVal[t*4+3]);
	  //time  = sensorDataTxt->dVal[t*4+3];
	  transformPt = transformPoint( transformPt, mapIsGlobal, &vehLocation );
	  
	  foundVict = false;
	  for( it=victInfo.begin(); it!=victInfo.end(); it++ )
	    {
	      if( myDisp( it->location, transformPt ) < 1 ) // same point
		{
		  foundVict = true;
		  break;
		}
	    }
	  sPtr=&sensorDataTxt->txtBuffer[sensorDataTxt->strPtr[t]];
	  if( !foundVict ) // need to add victim
	    {
	      // compute point 1m from victim for location of goal point
	      newInfo.id = ++id;
	      newInfo.goalLocation.x = transformPt.x - (transformPt.x-vehLocation.tran.x)/
		myDisp(vehLocation.tran, transformPt);
	      newInfo.goalLocation.y = transformPt.y - (transformPt.y-vehLocation.tran.y)/
		myDisp(vehLocation.tran, transformPt);
	      newInfo.location = transformPt;
	      //	      printf( "Vict msg(n): %s\n", sPtr );
	      token = strsep( &sPtr, " " );
	      if( token != NULL )
		sprintf( newInfo.name, "%s", token );
	      else
		sprintf( newInfo.name, "unknown" );
	      token = strsep( &sPtr, " " );
	      if( token != NULL )
		sprintf( newInfo.injury, "%s", token );
	      else
		sprintf( newInfo.injury, "unknown" );
	      victInfo.push_back(newInfo);
	      if( cmdType == VEH_MOB_PL_CMD_EXPLORE_TYPE )
		{
		  newGoalPoint = new goalPoint();
		  newGoalPoint->location = newInfo.goalLocation;
		  newGoalPoint->isGlobal = mapIsGlobal;
		  newGoalPoint->goalType = VICTIM;
		  newGoalPoint->neighborhood = 1;
		  goalStack.push(*newGoalPoint);
		  delete newGoalPoint;
		}
	      sprintf( moastError.message,
		       "New Victim found: id: %d Name: %s Injury: %s at map: (%f, %f)",
		       newInfo.id,
		       newInfo.name,
		       newInfo.injury,
		       transformPt.x,
		       transformPt.y);
	      printf( "%s, goal: <%f %f> veh: <%f %f>\n", moastError.message,
		      newInfo.goalLocation.x,
		      newInfo.goalLocation.y,
		      vehLocation.tran.x,
		      vehLocation.tran.y);

	      if( moastErrorNml != NULL )
		{
		  moastError.message_length = strlen(moastError.message);
		  moastErrorNml->write(moastError);
		}
	      retValue = true;
	    }
	  else // check to see if data has changed
	    {
	      //	      printf( "Vict msg(c): %s\n", sPtr );
	      token = strtok( sPtr, " " );
	      if( token != NULL )
		sprintf( newInfo.name, "%s", token );
	      else
		sprintf( newInfo.name, "unknown" );
	      token = strtok( NULL, " " );
	      if( token != NULL )
		sprintf( newInfo.injury, "%s", token );
	      else
		sprintf( newInfo.injury, "unknown" );
	      if( strcmp(newInfo.name, "None") && strcmp(newInfo.name, it->name) )
		{
		  printf( "Updated victim name from %s to %s\n", 
			  it->name, newInfo.name );
		  strcpy( it->name, newInfo.name );
		  if( moastErrorNml != NULL )
		    {
		      sprintf( moastError.message,
			       "Update Victim: id: %d Name: %s Injury: %s",
			       it->id,
			       it->name,
			       it->injury);
		      moastError.message_length = strlen(moastError.message);
		      moastErrorNml->write(moastError);
		    }
		  retValue = true;
		}
	      if( strcmp(newInfo.injury, "None") && strcmp(newInfo.injury, it->injury) )
		{
		  printf( "Updated victim injury from %s to %s\n", 
			  it->injury, newInfo.injury );
		  strcpy( it->injury, newInfo.injury );
		  if( moastErrorNml != NULL )
		    {
		      sprintf( moastError.message,
			       "Update Victim: id: %d Name: %s Injury: %s",
			       it->id,
			       it->name,
			       it->injury);
		      moastError.message_length = strlen(moastError.message);
		      moastErrorNml->write(moastError);
		    }
		  retValue = true;
		}
	    }
	}
      break;
    default:
      fprintf( stderr, "victim channel error\n" );
      fprintf( stdout, "Command %s not yet supported\n", sensorData_symbol_lookup(type) );
      return false;
    }
  return retValue;
}

/*!
  Read in spqr obstacles and populate world model

\param obsChan The channel to read from

\return Obstacle message on success 
        NULL on failure
*/
MapMsg *readSPQRMap( NML *obsChan )
{
  int type;

  type = obsChan->read();
  switch( type )
    {
    case 0: // no new data
      return(NULL);
    case MAP_MSG_TYPE:
      return((MapMsg*)obsChan->get_address());
    default:
      fprintf( stderr, "SPQR obstacle channel error\n" );
      fprintf( stdout, "Command %s not yet supported\n", sensorData_symbol_lookup(type) );
      return(NULL);
    }
}

/*!
  Read in obstacles and populate world model

\param obsChan The channel to read from

\return Obstacle message on success 
        NULL on failure
*/
NMLmsg *readObstacles( NML *obsChan, int &type )
{
  type = obsChan->read();
  switch( type )
    {
    case 0: // no new data
      return(NULL);
    case SENSOR_DATA_LIST_TYPE:
      return(obsChan->get_address());
    case SENSOR_DATA_2D_TYPE:
      return(obsChan->get_address());
    default:
      fprintf( stderr, "obstacle channel error\n" );
      fprintf( stdout, "Command %s not yet supported\n", sensorData_symbol_lookup(type) );
      return(NULL);
    }
}

/*!
  free all memory and prepare to exit
*/
void closeAll()
{
  if( amMobJACmdNml != NULL )
    delete amMobJACmdNml;
  amMobJACmdNml = NULL;
  if( amMobJAStatNml != NULL )
    delete amMobJAStatNml;
  amMobJAStatNml = NULL;
  if( amMobJASetNml != NULL )
    delete amMobJASetNml;
  amMobJASetNml = NULL;
  if( amMobJACfgNml != NULL )
    delete amMobJACfgNml;
  amMobJACfgNml = NULL;
  if( amSPCmdNml != NULL )
    delete amSPCmdNml;
  amSPCmdNml = NULL;
  if( amSPStatNml != NULL )
    delete amSPStatNml;
  amSPStatNml = NULL;
  if( amSPSetNml != NULL )
    delete amSPSetNml;
  amSPSetNml = NULL;
  if( amSPCfgNml != NULL )
    delete amSPCfgNml;
  amSPCfgNml = NULL;
  if( vehMobPLCmdNml != NULL )
    delete vehMobPLCmdNml;
  vehMobPLCmdNml = NULL;
  if( vehMobPLStatNml != NULL )
    delete vehMobPLStatNml;
  vehMobPLStatNml = NULL;
  if( vehMobPLSetNml != NULL )
    delete vehMobPLSetNml;
  vehMobPLSetNml = NULL;
  if( vehMobPLCfgNml != NULL )
    delete vehMobPLCfgNml;
  vehMobPLCfgNml = NULL;
  if( sensorDataNml != NULL )
    delete sensorDataNml;
  sensorDataNml = NULL;
  if( spqrMapInNml != NULL )
    delete spqrMapInNml;
  spqrMapInNml = NULL;
  if( spqrMapOutNml != NULL )
    delete spqrMapOutNml;
  spqrMapOutNml = NULL;
  if( navDataExtNml != NULL )
    delete navDataExtNml;
  navDataExtNml = NULL;
  if( victRFIDNml != NULL )
    delete victRFIDNml;
  if( moastErrorNml != NULL )
    delete moastErrorNml;
  if( vehMapOutNml != NULL )
    delete vehMapOutNml;
  if( moveWaypointCmd != NULL )
    delete moveWaypointCmd;
}

/*!
  Open all of the NML channels
  
  \param vehID The number to add to all of the channel names
\param openFlags Tell what special channels to open. 1- open spqrMap channel
  \return true on success
  \return false on failure
*/
bool openNML( int vehID, unsigned int openFlags )
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];
  VehMobPLCmdNop *nopCmd;
  VehMobPLCfgNop *nopCfg;
  MoastError moastError;

  sprintf( myName, "%s", MY_NAME_BASE );

  if(!configNml)
    {
      fprintf(stderr,"Please set CONFIG_NML environment variable.\n");
      return(false);
    }
  sprintf( chanName, "%s", MOAST_ERROR_NAME );
  moastErrorNml = new NML( moastError_format, chanName,
			   myName, configNml );
  if( NULL == moastErrorNml || !moastErrorNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       myName, chanName );
      moastErrorNml = NULL;
      // minor channel, so don't return on error
    }

#ifdef USEPRIMMOB
  sprintf( chanName, "%s%d", PRIM_MOB_JA_CMD_NAME, vehID );
  amMobJACmdNml = new RCS_CMD_CHANNEL( primMobJA_format, chanName,
				      myName, configNml );
  if( NULL == amMobJACmdNml || !amMobJACmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_STAT_NAME, vehID );
  amMobJAStatNml = new RCS_STAT_CHANNEL( primMobJA_format, chanName,
					myName, configNml );
  if( NULL == amMobJAStatNml || !amMobJAStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_SET_NAME, vehID );
  amMobJASetNml = new RCS_CMD_CHANNEL( primMobJA_format, chanName,
				      myName, configNml );
  if( NULL == amMobJASetNml || !amMobJASetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_CFG_NAME, vehID );
  amMobJACfgNml = new RCS_STAT_CHANNEL( primMobJA_format, chanName,
				       myName, configNml );
  if( NULL == amMobJACfgNml || !amMobJACfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

#else

  sprintf( chanName, "%s%d", AM_MOB_JA_CMD_NAME, vehID );
  amMobJACmdNml = new RCS_CMD_CHANNEL( amMobJA_format, chanName,
				      myName, configNml );
  if( NULL == amMobJACmdNml || !amMobJACmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_MOB_JA_STAT_NAME, vehID );
  amMobJAStatNml = new RCS_STAT_CHANNEL( amMobJA_format, chanName,
					myName, configNml );
  if( NULL == amMobJAStatNml || !amMobJAStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_MOB_JA_SET_NAME, vehID );
  amMobJASetNml = new RCS_STAT_CHANNEL( amMobJA_format, chanName,
				      myName, configNml );
  if( NULL == amMobJASetNml || !amMobJASetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_MOB_JA_CFG_NAME, vehID );
  amMobJACfgNml = new RCS_CMD_CHANNEL( amMobJA_format, chanName,
				       myName, configNml );
  if( NULL == amMobJACfgNml || !amMobJACfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

#endif // end of USEPRIMMOB

  sprintf( chanName, "%s%d", AM_SP_CMD_NAME, vehID );
  amSPCmdNml = new RCS_CMD_CHANNEL( amSP_format, chanName,
				       myName, configNml );
  if( NULL == amSPCmdNml || !amSPCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_STAT_NAME, vehID );
  amSPStatNml = new RCS_STAT_CHANNEL( amSP_format, chanName,
					 myName, configNml );
  if( NULL == amSPStatNml || !amSPStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_SET_NAME, vehID );
  amSPSetNml = new RCS_STAT_CHANNEL( amSP_format, chanName,
				       myName, configNml );
  if( NULL == amSPSetNml || !amSPSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_CFG_NAME, vehID );
  amSPCfgNml = new RCS_CMD_CHANNEL( amSP_format, chanName,
					myName, configNml );
  if( NULL == amSPCfgNml || !amSPCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CMD_NAME, vehID );
  vehMobPLCmdNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
				      myName, configNml );
  if( NULL == vehMobPLCmdNml || !vehMobPLCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_STAT_NAME, vehID );
  vehMobPLStatNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				       myName, configNml );
  if( NULL == vehMobPLStatNml || !vehMobPLStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_SET_NAME, vehID );
  vehMobPLSetNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				     myName, configNml );
  if( NULL == vehMobPLSetNml || !vehMobPLSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CFG_NAME, vehID );
  vehMobPLCfgNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
				      myName, configNml );
  if( NULL == vehMobPLCfgNml || !vehMobPLCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  //  sprintf( chanName, "%s%d", PRIM_SP_LINESCAN_NAME, vehID );
  sprintf( chanName, "%s%d", AM_SP_MAP_NAME, vehID );
  sensorDataNml = new NML( sensorData_format, chanName,
			      myName, configNml );
  if( NULL == sensorDataNml || !sensorDataNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, chanName );
      return(false);
    }
  if( (openFlags&&0x1) )
    {
      spqrMapInNml = new NML( nmlMap_format, "MapBufP2DX",
			    myName, configNml );
      if( NULL == spqrMapInNml || !spqrMapInNml->valid() )
	{
	  rcs_print_error( "%s: Error - Can't open %s buffer\n", 
			   myName, "MapBufP2DX" );
	  // not a critical problem, so keep going
	}
      spqrMapOutNml = new NML( nmlMap_format, "MapBufZerg",
			    myName, configNml );
      if( NULL == spqrMapOutNml || !spqrMapOutNml->valid() )
	{
	  rcs_print_error( "%s: Error - Can't open %s buffer\n", 
			   myName, "MapBufZerg" );
	  // not a critical problem, so keep going
	}

    }
  sprintf( chanName, "%s%d", NAV_DATA_EXT_NAME, vehID );
  navDataExtNml = new NML( navDataExt_format, chanName,
			   myName, configNml );
  if( NULL == navDataExtNml || !navDataExtNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, chanName );
      return(false);
    }

  //  sprintf( chanName, "%s%d_1", SERVO_SP_VICT_RFID_BUF_NAME, vehID );
  victRFIDNml = new NML( sensorData_format, chanName,
			   myName, configNml );
  if( NULL == victRFIDNml || !victRFIDNml->valid() )
    {
      sprintf( moastError.message,"%s: Error - Can't open %s buffer", 
	       myName, chanName );
      rcs_print_error( "%s\n", moastError.message );
      victRFIDNml = NULL;
      if( moastErrorNml != NULL )
	{
	  moastError.message_length = strlen(moastError.message);
	  moastErrorNml->write(moastError);
	}
      //      return(false);
      // minor sensor, so don't return on error
    }
  sprintf( chanName, "%s%d", VEH_MOB_MAP_NAME, vehID );
  vehMapOutNml = new NML( vehMobPL_format, chanName,
				     myName, configNml );
  if( NULL == vehMapOutNml || !vehMapOutNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }


  // stuff a nop
  nopCmd = new VehMobPLCmdNop();
  nopCfg = new VehMobPLCfgNop();
  nopCmd->serial_number = 100;
  nopCfg->serial_number = 100;

  vehMobPLCmdNml->write(nopCmd);
  vehMobPLCfgNml->write(nopCfg);
  delete nopCmd;
  delete nopCfg;
  return(true);
}

/*!
  Signal handler
  \param signum Signal that has been caught
*/
void sigHandler( int signum )
{
  switch( signum )
    {
    case SIGPIPE:
      fprintf( stderr, "Output pipe closed\n" );
      drawMap = false;
      break;
    case SIGINT:
      fprintf( stderr, "Shutting down vehMobPL program\n" );
      closeAll();
      exit(1);
      break;
    default:
      fprintf( stderr, "Unknown signal caught\n" );
      break;
    }
}

int main(int argc, char **argv)
{
  int count;
  PM_CARTESIAN obsLoc;
  char c;
  VehMobPLEngine *engine;
  NMLmsg *newObs;
  SensorDataList *newObsList;
  SensorData2D *newObs2D, *spqr2D;
  MapMsg *spqrMap = NULL;
  int obsType;
  int skip;
  RCS_TIMER *timer;
  VehMobPLStat *vehMobPLStat;
  VehMobPLSet  *vehMobPLSet;
  double obsProx = 250;
  double obsCost = 5000;
  bool done, writeSet;
  bool newVict;
  unsigned int moastLocation, spqrLocation;
  int countRow, countCol;

  // initial settings of variables that may be set with getopt
  //! The re-planning cycle time
  double cycleTime = .2;
  //! Use path discounting to avoid rapid path switching
  bool useDiscount = false;
  //! How far to look down the path for determining the confidence
  int confDist = 10;
  //! id of vehicle that we are planning for
  int vehID = 1;
  //! The initial resolution of theh planning map
  float cellResolution = .1;
  //! The number of cells to grow obstacles
  int obsGrow = 1;
  //! initial map extents
  int extents = 50; //0;
  //! should we read the SPQR map
  bool readSPQR = false;
  

  // variables that are initialized here or by getopt
  // Set drawing to off, with a skip of 5 
  drawMap = false;
  drawSkip = 5;
  // map is global by default
  mapIsGlobal = true;
  // set verbose off
  myVerbose = 0;

  myAmSerial = 1;
  goalIsGlobal = false;
#ifdef USEPRIMMOB
  moveWaypointCmd = new PrimMobJACmdMoveWaypoint();
#else
  moveWaypointCmd = new AmMobJACmdMoveWaypoint();
  moveWaypointCmd->computeHorizonCosts = true;
#endif
  skip = 0;
  
  signal( SIGPIPE, sigHandler );
  signal( SIGINT, sigHandler );
  while((c=getopt(argc,argv,"c:d:e:hil:m:r:v:g:s"))!=EOF) 
    {
      switch(c) 
	{
	case 'c':
	  // set cycle time
	  cycleTime = atof(optarg);
	  fprintf( stderr, "Cycle time set to %lf\n", cycleTime );
	  break;
	case 'd':
	  // set vehicle id
	  vehID = atoi(optarg);
	  fprintf( stderr, "Connecting to vehicle %d\n", vehID );
	  break;
	case 'e':
	  extents = atoi(optarg);
	  fprintf( stderr, "Setting initial map extents to %d\n", extents );
	  break;
	  /*
	case 'h':
	  // turn on discounting
	  useDiscount = true;
	  fprintf( stderr, "Using path discount to reduce hysterisis\n" );
	  break;
	  */
	case 'g':
	  obsGrow = atoi(optarg);
	  fprintf( stderr, "Obsgrow set to %d\n", obsGrow );
	  break;
	case 'i':
	  // use local imu solution instead of global
	  mapIsGlobal = false;
	  fprintf( stderr, "Using local imu (local coord) for planning map\n" );
	  break;
	case 'l':
	  confDist = atoi(optarg);
	  fprintf( stderr, "Confidence look-ahead is %dm\n", confDist );
	  break;
	case 'm':
	  // set map drawing on
	  drawSkip = atoi(optarg);
	  drawMap = true;
	  fprintf( stderr, "Drawing map on with skip of %d\n", drawSkip );
	  break;
	case 'r':
	  // set cell resolution
	  cellResolution = atof(optarg);
	  fprintf( stderr, "vehMobPLMain:Cell resolution set to %lf\n", cellResolution );
	  break;
	case 's':
	  readSPQR = true;
	  break;
	case 'v':
	  myVerbose = atoi(optarg);
	  break;
	default:
	  fprintf( stderr, "usage: %s -c<cycle time> -d<vehID> -h "
		   "(turn hystrisis on) -i (use local imu coord for map) "
		   "-l<confidance look ahead> -m<skip> (draw map every skip cycles)"
		   "-v (verbose) -s (use spqr mapping)\n", argv[0] );
	  exit(1);
	}
    }
  timer = new RCS_TIMER(cycleTime);

  engine = new VehMobPLEngine( extents, extents, cellResolution, obsGrow, 
			       myVerbose );
  engine->setCostLookup();
  //  engine->buildGraph();
  if( readSPQR )
    spqr2D = new SensorData2D();

  if( !openNML(vehID, readSPQR) )
    exit(1);

  // send out first settings message
  vehMobPLSet = new VehMobPLSet();
  //  vehMobPLSet->command_type = VEH_MOB_PL_CFG_DUMP_WM_TYPE;
  vehMobPLSet->command_type = VEH_MOB_PL_CFG_NOP_TYPE;
  vehMobPLSet->state = NEW_COMMAND;
  vehMobPLSet->cycleTime = cycleTime;
  vehMobPLSet->obstaclePad = 0;
  vehMobPLSet->usePathDiscount = useDiscount;
  vehMobPLSet->debug = myVerbose;
  vehMobPLSet->isInitialized = false;
  vehMobPLSet->isUsingSQL = false;
  vehMobPLSet->confDistance = confDist;
  vehMobPLSet->cellResolution = cellResolution;
  vehMobPLSet->zeroConf = obsCost;
  vehMobPLSet->fiveConf = obsProx;
  strcpy(vehMobPLSet->wmFileName, "VehMobPLDispOutput.tif" ); // was ppm
  vehMobPLSet->fileSkip = 5;
  vehMobPLSet->echo_serial_number = -1;
  vehMobPLSetNml->write(vehMobPLSet);

  // send out first status message
  vehMobPLStat = new VehMobPLStat();
  vehMobPLStat->echo_serial_number = -1;
  vehMobPLStat->state = NEW_COMMAND;
  vehMobPLStat->status = RCS_EXEC;
  vehMobPLStat->command_type = VEH_MOB_PL_CMD_NOP_TYPE;
  vehMobPLStat->planStartHorizon = 4.5;
  vehMobPLStatNml->write(vehMobPLStat);

  for(;;)
    {
      /////////////////////////////////////////////
      // update vehicle information
      /////////////////////////////////////////////
      // update vehicle position
      if( vehMobPLSet->isInitialized )
	{
	  updatePosition(navDataExtNml);
	  engine->setVehLoc(vehLocation);
	}
      // read configuration channel
      readCfg(vehMobPLCfgNml, vehMobPLSet);
      // read command channel
      readCmd(vehMobPLCmdNml, vehMobPLStat, vehMobPLSet);
      // read sensor channels
      if( vehMobPLSet->isInitialized == true )
	{
	  newObs = readObstacles(sensorDataNml, obsType );
	  newVict = readVictims(victRFIDNml, engine->victInfo, vehMobPLStat->command_type);
	  if( readSPQR )
	    spqrMap = readSPQRMap(spqrMapInNml);
	}
      else
	newObs = NULL;
      //read amSettings channel
#ifdef USEPRIMMOB
      amMobJASettings.planHorizon = 0;
#else
      checkAmMobSettings(amMobJASettings,amMobJASetNml); 
#endif

      /////////////////////////////////////////////
      // process new information
      /////////////////////////////////////////////
      // process configuration
      switch( vehMobPLSet->command_type )
	{
	case VEH_MOB_PL_CFG_NOP_TYPE:
	  writeSet = doCfgNop( vehMobPLSet );
	  break;
	case VEH_MOB_PL_CFG_CLEAR_WM_TYPE:
	  writeSet = doCfgClearWM( vehMobPLSet, engine );
	  break;
	case VEH_MOB_PL_CFG_STORE_WM_TYPE:
	  writeSet = doCfgStoreWM( vehMobPLSet, vehMobPLSet, engine );
	  break;
	case VEH_MOB_PL_CFG_DUMP_WM_TYPE:
	  writeSet = doCfgDumpWM( vehMobPLSet, engine );
	  break;
	case VEH_MOB_PL_CFG_SQL_TYPE:
	  writeSet = doCfgSQL( vehMobPLSet );
	  break;
	default:
	  writeSet = false;
	  break;
	}
      if( writeSet )
	vehMobPLSetNml->write(vehMobPLSet);
      // process obstacles
      if( newObs!=NULL && vehMobPLStat->state!=S_PLAN_ROUTE )
	{
	  if( obsType == SENSOR_DATA_LIST_TYPE )
	    {
	      newObsList = (SensorDataList*)newObs;
	      if( mapIsGlobal && !newObsList->isGlobal ) // set local coordinates to global
		{
		  for( count=0; count<newObsList->featList_length; count++ )
		    {
		      obsLoc.x = newObsList->featList[count].dataLoc.x;
		      obsLoc.y = newObsList->featList[count].dataLoc.y;
		      obsLoc.z = newObsList->featList[count].dataLoc.z;
		      obsLoc = transformPoint(obsLoc, false );
		      newObsList->featList[count].dataLoc.x = obsLoc.x;
		      newObsList->featList[count].dataLoc.y = obsLoc.y;
		      obsLoc.x = newObsList->featList[count].sensorLoc.x;
		      obsLoc.y = newObsList->featList[count].sensorLoc.y;
		      obsLoc.z = newObsList->featList[count].sensorLoc.z;
		      obsLoc = transformPoint(obsLoc, false );
		      newObsList->featList[count].sensorLoc = obsLoc;
		    }
		}
	      else if( !mapIsGlobal && newObsList->isGlobal ) // set global coordinates to local
		{
		  for( count=0; count<newObsList->featList_length; count++ )
		    {
		      obsLoc.x = newObsList->featList[count].dataLoc.x;
		      obsLoc.y = newObsList->featList[count].dataLoc.y;
		      obsLoc.z = newObsList->featList[count].dataLoc.z;
		      obsLoc = transformPoint(obsLoc, true );
		      newObsList->featList[count].dataLoc.x = obsLoc.x;
		      newObsList->featList[count].dataLoc.y = obsLoc.y;
		      obsLoc = newObsList->featList[count].sensorLoc;
		      obsLoc = transformPoint(obsLoc, true );
		      newObsList->featList[count].sensorLoc = obsLoc;
		    }
		}
	      engine->updateWorld( newObsList->featList, newObsList->featList_length, 
				   vehMobPLStat->tt.count, NO_MARK );
	    }
	  else if( obsType == SENSOR_DATA_2D_TYPE )
	    {
	      newObs2D = (SensorData2D*)newObs;
	      if( mapIsGlobal && !newObs2D->isGlobal ) // set local coordinates to global
		{
		  obsLoc = newObs2D->lowerLeft;
		  obsLoc = transformPoint(obsLoc, false);
		  newObs2D->lowerLeft = obsLoc;
		}
	      else if( !mapIsGlobal && newObs2D->isGlobal ) // set global coordinates to local
		{
		  obsLoc = newObs2D->lowerLeft;
		  obsLoc = transformPoint(obsLoc, true);
		  newObs2D->lowerLeft = obsLoc;
		}
	      engine->updateWorld( newObs2D, vehMobPLStat->tt.count, vehLocation.tran );
	    }
	}
      if( spqrMap != NULL )
	{
	  if( !mapIsGlobal ) // set global coordinates to local
	    {
	      obsLoc.x = spqrMap->mapX;
	      obsLoc.y = -spqrMap->mapY;
	      obsLoc = transformPoint(obsLoc, true);
	      spqr2D->lowerLeft = obsLoc;
	    }
	  else
	    {
	      spqr2D->lowerLeft.x = spqrMap->mapX;
	      spqr2D->lowerLeft.y = -spqrMap->mapY;
	    }
	  spqr2D->incrementX = spqrMap->mapRealWidth/spqrMap->nCols;
	  spqr2D->incrementY = spqr2D->incrementX;
	  spqr2D->x = spqrMap->nRows;
	  spqr2D->y = spqrMap->nCols;
	  spqr2D->featArray_length = spqrMap->map_length;
	  for(countRow=0; countRow<(int)spqrMap->nRows; countRow++ )
	    {
	      for(countCol=0; countCol<(int)spqrMap->nCols; countCol++ )
		{
		  spqrLocation = countRow*spqrMap->nCols + countCol;
		  moastLocation = countCol*spqrMap->nRows + countRow;
		  if( spqrMap->map[spqrLocation] == C8Black )
		    spqr2D->featArray[moastLocation] = AMSP_OBSTACLE;
		  else if( spqrMap->map[spqrLocation] == C8Blue )
		spqr2D->featArray[moastLocation] = AMSP_UNSEEN;
	      else if( spqrMap->map[spqrLocation] == C8White )
		spqr2D->featArray[moastLocation] = AMSP_NOT_OBSTACLE;
	      else // should not happen
		spqr2D->featArray[moastLocation] = AMSP_MAYBE_OBSTACLE;
		}
	    }
	  engine->updateWorld( spqr2D, vehMobPLStat->tt.count, vehLocation.tran );
	}

      // process commands
      switch( vehMobPLStat->command_type )
	{
	case VEH_MOB_PL_CMD_NOP_TYPE:
	  doCmdNop( vehMobPLStat );
	  break;
	case VEH_MOB_PL_CMD_ABORT_TYPE:
	  doCmdAbort( vehMobPLStat );
	  break;
	case VEH_MOB_PL_CMD_HALT_TYPE:
	  if( doCmdHalt( vehMobPLStat ) )
	    {
	      vehMobPLSet->isInitialized = false;
	      vehMobPLSetNml->write(vehMobPLSet);
	    }
	  break;
	case VEH_MOB_PL_CMD_INIT_TYPE:
	  if( doCmdInit( vehMobPLStat, vehMobPLSet, engine ))
	    {
	      vehMobPLSetNml->write(vehMobPLSet);
	    }
	  break;
	case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
	  done = doCmdShutdown( vehMobPLStat );
	  vehMobPLStatNml->write(vehMobPLStat);
	  return(1);
	  break;
	case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
	  if( !vehMobPLSet->isInitialized )
	    {
	      status_next(vehMobPLStat, RCS_ERROR);
	      break;
	    }
	  doCmdMovement( vehMobPLStat, vehMobPLSet, engine, VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE, engine->victInfo );
	  break;
	case VEH_MOB_PL_CMD_EXPLORE_TYPE:
	  if( !vehMobPLSet->isInitialized )
	    {
	      status_next(vehMobPLStat, RCS_ERROR);
	      break;
	    }
	  doCmdMovement( vehMobPLStat, vehMobPLSet, engine, VEH_MOB_PL_CMD_EXPLORE_TYPE, engine->victInfo );
	  break;
	case VEH_MOB_PL_CMD_LOAD_WM_TYPE:
	  writeSet = doCmdLoadWM( vehMobPLStat, vehMobPLSet, engine );
	  break;
	}
      //      vehMobPLStat->worldX = engine->getGraphHeight();
      //      vehMobPLStat->worldY = engine->getGraphWidth();
      vehMobPLStatNml->write(vehMobPLStat);
      cycle_time_tracker(&vehMobPLStat->tt);
      // write out world model
      if( vehMobPLSet->isInitialized )
	{
	  engine->mapOut(vehMapOutNml);
	  if( readSPQR ) // if we are reading their map, we should write out ours
	    {
	      obsLoc = transformPoint(vehLocation.tran, false);
	      engine->mapOutSpqr(spqrMapOutNml, obsLoc);
	    }
	}
      timer->wait();
    }
  // clock_gettime
}

/*
  Modification history:

  $Log: vehMobPLMain.cc,v $
  Revision 1.41  2007/12/17 21:35:05  dr_steveb
  Fixed end-point condition so that vehicle now gets to final goal

  Revision 1.40  2007/12/13 13:18:41  dr_steveb
  Changes for vehicle level visibility graph planning.

  Revision 1.39  2007/11/08 13:27:51  dr_steveb
  Changed location of LineSegment

  Revision 1.38  2007/05/16 20:57:28  dr_steveb
  Fixed graph wrapping!

  Revision 1.37  2007/05/16 15:19:06  dr_steveb
  Trying to fix graph wrapping. Not fixed yet, however, fixed compile issue with rotate command.

  Revision 1.36  2007/04/25 15:09:05  dr_steveb
  Added traffic simulation files

  Revision 1.35  2007/03/19 15:11:23  dr_steveb
  Changed the way that the size of the graphHeight and graphWidth are being set. Fixed calling location of buildGraph.

  Revision 1.34  2007/02/27 19:12:19  dr_steveb
  Fixed drawing of no-planning region for vehicle and added mobility world

  Revision 1.33  2007/02/27 14:46:27  dr_steveb
  New interface between AM and VEH

  Revision 1.32  2007/01/30 21:29:02  dr_steveb
  Added new AM-VEH interface

  Revision 1.31  2006/12/14 19:47:27  dr_steveb
  Fixed some sub driving stuff

  Revision 1.30  2006/12/14 16:16:53  dr_steveb
  More changes for new interface

  Revision 1.29  2006/12/12 19:20:40  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.28  2006/11/28 19:06:32  dr_steveb
  Improved "done" criteria in movement commands.

  Revision 1.27  2006/10/31 12:49:48  dr_steveb
  Fixes to interface to SPQR

  Revision 1.26  2006/10/31 09:30:43  dr_steveb
  Changes for SPQR integration

  Revision 1.25  2006/10/24 13:09:49  dr_steveb
  New vehicle level victim locator and output on moastError. MoastError is now read by the notifier program. Use notifier -bmoastError

  Revision 1.24  2006/10/18 00:18:21  dr_steveb
  Added victim channel to vehicle level. It now hunts for victims. Also added CImg library from sourceforge to perform annotation. Modified make files to include this library.

  Revision 1.23  2006/10/10 13:30:19  dr_steveb
  New vehicle level exploration based on camera FOV implemented.

  Revision 1.22  2006/10/03 20:24:25  dr_steveb
  Exploration command added to vehicle level

  Revision 1.21  2006/09/30 16:04:22  dr_steveb
  Changed to use Tom's new world model and added am error checking.

  Revision 1.20  2006/09/01 17:56:08  dr_steveb
  Reordered init state machine to perform "go" before reading navData.

  Revision 1.19  2006/05/06 11:24:22  dr_steveb
  Adde nop command to vehShell and increased tolerance on waypoints in the planner.

  Revision 1.18  2006/05/04 17:42:00  dr_steveb
  Changes to make multi-vehicle section exploration work.

  Revision 1.17  2006/04/12 13:02:16  dr_steveb
  Section level exploration now works for a single vehicle

  Revision 1.16  2006/04/06 21:26:13  dr_steveb
  Changed my process name to remove the id

  Revision 1.15  2006/04/05 17:47:51  dr_steveb
  Added sending of "go" command to sp.

  Revision 1.14  2006/04/05 02:02:43  dr_steveb
  Changed to use NOP instead of NOOP

  Revision 1.13  2006/03/22 18:06:36  dr_steveb
  Fixed interface to veh level for imageview

  Revision 1.12  2006/03/22 14:47:38  dr_steveb
  Changed to use imageview instead of old display program.

  Revision 1.11  2005/10/27 13:06:16  dr_steveb
  Minor changes

  Revision 1.10  2005/10/24 21:27:09  dr_steveb
  Added the ability to set the cell resolution from the command line.

  Revision 1.9  2005/10/23 15:41:08  dr_steveb
  Changes to add dump command to amSP and adjust for its new resolution.

  Revision 1.8  2005/10/23 15:09:46  dr_steveb
  Changed default map resolution to .2 to match AM

  Revision 1.7  2005/10/21 14:13:24  dr_steveb
  Changed init state machine so that the system will only initialize once.

  Revision 1.6  2005/10/21 13:18:32  dr_steveb
  Needed to add include of config.h in order to get HAVE_CYGWIN to work.

  Revision 1.5  2005/10/21 13:02:45  dr_steveb
  Various fixes.

  Revision 1.4  2005/10/19 12:14:51  dr_steveb
  Changed parameters sent to lower-level planner (speed).

  Revision 1.3  2005/10/17 18:35:13  dr_steveb
  Fixed map growing, added planLength to status message. Added getPlanLength to vehMobPLEngine. Discovered problem with use of ppmsdisplay

  Revision 1.2  2005/10/16 15:31:02  dr_steveb
  Various changes that will have to be redone for the new distribution.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.20  2005/10/11 15:26:34  stephen
  Added plan cost to status message and status printing. This required a new class variable (planCost).

  Revision 1.19  2005/10/09 14:47:33  stephen
  General bug fixes.

  Revision 1.18  2005/10/03 15:06:31  stephen
  Vehicle level now works with either prim or am by setting a compile flag (USEPRIMMOB)

  Revision 1.17  2005/09/30 18:55:58  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.16  2005/09/30 18:34:57  proctor
  Changed include <sys/fcntl.h> to include <fcntl.h>

  Revision 1.15  2005/09/29 21:45:32  stephen
  Added dump command and made some fixes.

  Revision 1.14  2005/09/29 13:18:10  stephen
  Beginning of changes to work with Tom's new prim and am sp.

  Revision 1.13  2005/09/21 00:46:06  stephen
  Modifications to work with Adam's am and prim.

  Revision 1.12  2005/07/26 13:25:11  stephen
  Added visibility graph planner (not yet working)

  Revision 1.11  2005/07/18 21:16:59  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.10  2005/07/18 19:56:12  stephen
  Changed interface to confrom to dijkHeapN

  Revision 1.9  2005/06/29 21:08:17  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.8  2005/06/15 13:39:53  stephen
  Added use of mysql class for opening channel to sql and fixed writing of set channel.

  Revision 1.7  2005/06/14 19:19:14  stephen
  Changed to use SensorPoint in sensor data message.

  Revision 1.6  2005/06/13 17:26:29  stephen
  Continuing work on adding commands and creating true state-machine.

  Revision 1.5  2005/06/09 16:43:46  stephen
  Fixed initialization problem.

  Revision 1.4  2005/06/09 12:32:58  stephen
  Removal of sqrtf and start of turning cfg channel into state machine.

  Revision 1.3  2005/06/08 20:49:37  stephen
  Changed occupancyData to sensorData

  Revision 1.2  2005/06/08 19:29:14  proctor
  Moved state_match, etc. macros into moastTypes.hh

  Revision 1.1  2005/06/08 14:31:15  proctor
  Added MobJA name changes

  Revision 1.10  2005/06/08 13:40:10  stephen
  Beginnings of changes to have true cfg cycle.

  Revision 1.9  2005/06/06 12:45:06  stephen
  Made true state machine.

  Revision 1.8  2005/06/02 12:56:35  stephen
  Changed waypoint to use only a single waypoint.

  Revision 1.7  2005/06/01 13:21:46  proctor
  Changed code that handles waypoints to use the Waypoint structure now
  declared in moastTypes.cc

  Revision 1.6  2005/05/31 21:19:33  proctor
  Changed tabsGraph, searcher to DijkTabsGraph, DijkTabsSearcher, etc. to
  reflect Tom's changes to the naming convention.

  Revision 1.5  2005/05/27 23:22:48  proctor
  Fixed floorf nonportability

  Revision 1.4  2005/05/25 17:48:16  stephen
  Fixed waitForStatus to wait between channel reads.

  Revision 1.3  2005/05/24 17:05:07  stephen
  Changed to Tom's new Dijk algorithm.

  Revision 1.2  2005/05/24 12:39:18  stephen
  Added signal handler for SIGINT signal to properly exit.

  Revision 1.1  2005/05/23 18:03:31  stephen
  Added vehicle level planner.
*/
