/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file vehSPMain.cc

  \brief Vehicle-level SP process.

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Fred Proctor
  \date 8 June 2005
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>
#include "moastTypes.hh"
#include "moastNmlOffsets.hh"
#include "amSP.hh"
#include "vehSP.hh"

static RCS_CMD_CHANNEL *amSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *amSPStatBuf = NULL;

static RCS_CMD_CHANNEL *vehSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *vehSPStatBuf = NULL;
static RCS_CMD_CHANNEL *vehSPCfgBuf = NULL;
static RCS_STAT_CHANNEL *vehSPSetBuf = NULL;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "vehSP"

#define DEFAULT_CYCLE_TIME 0.100

static int cleanupAmSPNmlBuffers(void)
{
  if (NULL != amSPCmdBuf)
    delete amSPCmdBuf;
  if (NULL != amSPStatBuf)
    delete amSPStatBuf;

  amSPCmdBuf = NULL;
  amSPStatBuf = NULL;

  return 0;
}

static int initAmSPNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupAmSPNmlBuffers();

  sprintf(chanName, "%s%d", AM_SP_CMD_NAME, bufnum);
  amSPCmdBuf =
    new RCS_CMD_CHANNEL(amSP_format, chanName,
			THIS_PROCESS, config_file);
  if (!amSPCmdBuf->valid()) {
    delete amSPCmdBuf;
    amSPCmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", AM_SP_STAT_NAME, bufnum);
  amSPStatBuf =
    new RCS_STAT_CHANNEL(amSP_format, chanName,
			 THIS_PROCESS, config_file);
  if (!amSPStatBuf->valid()) {
    delete amSPStatBuf;
    amSPStatBuf = 0;
    return 1;
  }

  return 0;
}

static int cleanupVehSPNmlBuffers(void)
{
  if (NULL != vehSPCmdBuf)
    delete vehSPCmdBuf;
  if (NULL != vehSPStatBuf)
    delete vehSPStatBuf;
  if (NULL != vehSPCfgBuf)
    delete vehSPCfgBuf;
  if (NULL != vehSPSetBuf)
    delete vehSPSetBuf;

  vehSPCmdBuf = NULL;
  vehSPStatBuf = NULL;
  vehSPCfgBuf = NULL;
  vehSPSetBuf = NULL;

  return 0;
}

static int initVehSPNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupVehSPNmlBuffers();

  sprintf(chanName, "%s%d", VEH_SP_CMD_NAME, bufnum);
  vehSPCmdBuf =
    new RCS_CMD_CHANNEL(vehSP_format, chanName,
			THIS_PROCESS, config_file);
  if (!vehSPCmdBuf->valid()) {
    delete vehSPCmdBuf;
    vehSPCmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", VEH_SP_STAT_NAME, bufnum);
  vehSPStatBuf =
    new RCS_STAT_CHANNEL(vehSP_format, chanName,
			 THIS_PROCESS, config_file);
  if (!vehSPStatBuf->valid()) {
    delete vehSPStatBuf;
    vehSPStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", VEH_SP_CFG_NAME, bufnum);
  vehSPCfgBuf =
    new RCS_CMD_CHANNEL(vehSP_format, chanName,
			THIS_PROCESS, config_file);
  if (!vehSPCfgBuf->valid()) {
    delete vehSPCfgBuf;
    vehSPCfgBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", VEH_SP_SET_NAME, bufnum);
  vehSPSetBuf =
    new RCS_STAT_CHANNEL(vehSP_format, chanName,
			 THIS_PROCESS, config_file);
  if (!vehSPSetBuf->valid()) {
    delete vehSPSetBuf;
    vehSPSetBuf = 0;
    return 1;
  }

  return 0;
}

void DO_CMD_INIT(VehSPStat * stat,
		 RCS_CMD_CHANNEL * sub_cmd_buf,
		 RCS_STAT_CHANNEL * sub_stat_buf)
{
  AmSPCmdInit initMsg;
  AmSPStat *sub_stat_ptr;

  sub_stat_ptr = (AmSPStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    // send a command to subordinate
    sub_stat_buf->read();
    initMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
    sub_cmd_buf->write(&initMsg);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void DO_CMD_HALT(VehSPStat * stat,
		 RCS_CMD_CHANNEL * sub_cmd_buf,
		 RCS_STAT_CHANNEL * sub_stat_buf)
{
  AmSPCmdHalt haltMsg;
  AmSPStat *sub_stat_ptr;

  sub_stat_ptr = (AmSPStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    // send a command to subordinate
    sub_stat_buf->read();
    haltMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
    sub_cmd_buf->write(&haltMsg);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void DO_CMD_ABORT(VehSPStat * stat,
		  RCS_CMD_CHANNEL * sub_cmd_buf,
		  RCS_STAT_CHANNEL * sub_stat_buf)
{
  AmSPCmdAbort abortMsg;
  AmSPStat *sub_stat_ptr;

  sub_stat_ptr = (AmSPStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    // send a command to subordinate
    sub_stat_buf->read();
    abortMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
    sub_cmd_buf->write(&abortMsg);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void DO_CMD_SHUTDOWN(VehSPStat * stat,
		  RCS_CMD_CHANNEL * sub_cmd_buf,
		  RCS_STAT_CHANNEL * sub_stat_buf)
{
  AmSPCmdShutdown shutdownMsg;
  AmSPStat *sub_stat_ptr;

  sub_stat_ptr = (AmSPStat *) (sub_stat_buf->get_address());

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    // send a command to subordinate
    sub_stat_buf->read();
    shutdownMsg.serial_number = sub_stat_ptr->echo_serial_number + 1;
    sub_cmd_buf->write(&shutdownMsg);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else if (state_match(stat, S0)) {
    // wait for done
    sub_stat_buf->read();
    if (sub_stat_ptr->status != RCS_EXEC) {
      // it's either done or error, so make this our status
      status_next(stat, sub_stat_ptr->status);
      state_next(stat, S1);
    }
    // else wait here
  } else {			// S1
    state_default(stat);
  }
}

void DO_CFG_CYCLE_TIME(VehSPCfgCycleTime * cfg,
		       VehSPSet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    if (cfg->cycleTime <= 0.0) {
      status_next(set, RCS_ERROR);
    } else {
      set->cycleTime = cfg->cycleTime;
      status_next(set, RCS_DONE);
    }
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

void DO_CFG_DEBUG(VehSPCfgDebug * cfg,
		  VehSPSet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->debug = cfg->debug;
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
  int retval;
  NMLTYPE cmd_type, cfg_type;
  int serial_number;
  VehSPStat vehSPStat;
  VehSPSet vehSPSet;
  VehSPCmdInit initCmd;
  VehSPCfgCycleTime cycleTimeCfg;

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      rcs_print("%s: using specified NML file %s\n", argv[0], NML_FILE);
      break;
    case 'i':
      INI_FILE = optarg;
      rcs_print("%s: using specified INI file %s\n", argv[0], INI_FILE);
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      rcs_print("%s: using specified buffer suffix %d\n", argv[0], bufnum);
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file nvehe from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
      rcs_print("%s: using environment NML file %s\n", argv[0], NML_FILE);
    } else {
      NML_FILE = nml_file_default;
      rcs_print("%s: using default NML file %s\n", argv[0], NML_FILE);
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
      rcs_print("%s: using environment INI file %s\n", argv[0], INI_FILE);
    } else {
      INI_FILE = ini_file_default;
      rcs_print("%s: using default INI file %s\n", argv[0], INI_FILE);
    }
  }

  // connect to the vehicle am level buffers
  retval = initAmSPNmlBuffers(NML_FILE, bufnum);
  if (retval) {
    rcs_print_error("%s: can't allocate amSP NML buffers\n", argv[0]);
    return 1;
  }

  // connect to my buffers
  retval = initVehSPNmlBuffers(NML_FILE, bufnum);
  if (retval) {
    rcs_print_error("%s: can't allocate vehSP NML buffers\n", argv[0]);
    (void) cleanupAmSPNmlBuffers();
    return 1;
  }

  // initialize and set the status up so that init command goes through
  vehSPStat.echo_serial_number = 0;
  vehSPStat.heartbeat = 0;

  // initialize and set the settings up so that the cycle time goes through
  vehSPSet.echo_serial_number = 0;
  // others will come from .ini file shortly...

  // stuff an init command
  initCmd.serial_number = 1;
  vehSPCmdBuf->write(&initCmd);

  // stuff a cycle time config
  cycleTimeCfg.serial_number = 1;
  cycleTimeCfg.cycleTime = vehSPSet.cycleTime;
  vehSPCfgBuf->write(&cycleTimeCfg);

  done = 0;
  signal(SIGINT, quit);
  while (!done) {
    // some shortcut pointers to various commands
    VehSPCfgCycleTime * cfgCycleTimePtr;
    VehSPCfgDebug * cfgDebugPtr;

    // read command
    cmd_type = vehSPCmdBuf->read();
    serial_number = vehSPCmdBuf->get_address()->serial_number;
    switch (cmd_type) {

    case 0:
      // no new command
      break;

    case -1:
      // comm error
      rcs_print_error("%s: NML command read error\n", argv[0]);
      break;

    case VEH_SP_CMD_INIT_TYPE:
    case VEH_SP_CMD_HALT_TYPE:
    case VEH_SP_CMD_ABORT_TYPE:
    case VEH_SP_CMD_SHUTDOWN_TYPE:
      vehSPStat.command_type = cmd_type;
      if (serial_number != vehSPStat.echo_serial_number) {
	vehSPStat.echo_serial_number = serial_number;
	vehSPStat.state = NEW_COMMAND;
	vehSPStat.status = RCS_EXEC;
      }
      break;

    default:
      rcs_print_error("%s: unknown command: %d\n", argv[0], (int) cmd_type);
      break;
    }				// switch (cmd_type)

    // read config
    cfg_type = vehSPCfgBuf->read();
    serial_number = vehSPCfgBuf->get_address()->serial_number;
    switch (cfg_type) {

    case 0:
      // no new command
      break;

    case -1:
      // comm error
      rcs_print_error("%s: NML config read error\n", argv[0]);
      break;

    case VEH_SP_CFG_CYCLE_TIME_TYPE:
      vehSPSet.command_type = cfg_type;
      if (serial_number != vehSPSet.echo_serial_number) {
	vehSPSet.echo_serial_number = serial_number;
	vehSPSet.state = NEW_COMMAND;
	vehSPSet.status = RCS_EXEC;
      }
      cfgCycleTimePtr =
	(VehSPCfgCycleTime *) vehSPCfgBuf->
	get_address();
      break;

    case VEH_SP_CFG_DEBUG_TYPE:
      vehSPSet.command_type = cfg_type;
      if (serial_number != vehSPSet.echo_serial_number) {
	vehSPSet.echo_serial_number = serial_number;
	vehSPSet.state = NEW_COMMAND;
	vehSPSet.status = RCS_EXEC;
      }
      cfgDebugPtr =
	(VehSPCfgDebug *) vehSPCfgBuf->get_address();
      break;

    default:
      rcs_print_error("%s: unknown config: %d\n", argv[0], (int) cfg_type);
      break;
    }				// switch (cfg_type)

    // handle command state tables
    switch (vehSPStat.command_type) {

    case VEH_SP_CMD_INIT_TYPE:
      DO_CMD_INIT(&vehSPStat,
		  amSPCmdBuf, amSPStatBuf);
      break;

    case VEH_SP_CMD_HALT_TYPE:
      DO_CMD_HALT(&vehSPStat,
		  amSPCmdBuf, amSPStatBuf);
      break;

    case VEH_SP_CMD_ABORT_TYPE:
      DO_CMD_ABORT(&vehSPStat,
		   amSPCmdBuf, amSPStatBuf);
      break;

    case VEH_SP_CMD_SHUTDOWN_TYPE:
      DO_CMD_SHUTDOWN(&vehSPStat,
		      amSPCmdBuf, amSPStatBuf);
      break;

    default:
      // no command to handle
      break;
    }				// switch (cmd)

    // handle config state tables
    switch (vehSPSet.command_type) {

    case VEH_SP_CFG_CYCLE_TIME_TYPE:
      DO_CFG_CYCLE_TIME(cfgCycleTimePtr, &vehSPSet);
      break;

    case VEH_SP_CFG_DEBUG_TYPE:
      DO_CFG_DEBUG(cfgDebugPtr, &vehSPSet);
      break;

    default:
      // no config to handle
      break;
    }				// switch (cfg)

    vehSPStat.heartbeat++;

    // update status and settings
    vehSPStatBuf->write(&vehSPStat);
    vehSPSetBuf->write(&vehSPSet);

    esleep(vehSPSet.cycleTime);
  }
  while (!done);

  (void) cleanupVehSPNmlBuffers();
  (void) cleanupAmSPNmlBuffers();

  rcs_print("vehSP done\n");

  return 0;
}

/*
  Modification history:

  $Log: vehSPMain.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.2  2005/09/30 18:55:58  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.1  2005/06/08 20:07:15  proctor
  initial checkin

*/

