/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file gomath.h

  \brief Declarations for pose math functions

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.1 $
  $Date: 2007/05/09 15:19:06 $
  \endcode

  \author Fred Proctor
*/

#ifndef GO_MATH_H
#define GO_MATH_H

#include <stddef.h>		/* sizeof */
#include <math.h>		/* M_PI */
#include <float.h>		/* FLT,DBL_MIN,MAX,EPSILON */
#include "gotypes.h"		/* go_integer,real */

#define go_sq(x) ((x)*(x))
#define go_cub(x) ((x)*(x)*(x))
#define go_qua(x) ((x)*(x)*(x)*(x))

extern void go_sincos(go_real t, go_real * s, go_real * c);

extern go_real go_cbrt(go_real t);

#ifdef M_PI
#define GO_PI M_PI
#else
#define GO_PI 3.14159265358979323846
#endif
#define GO_2_PI (2.0*GO_PI)

#ifdef M_PI_2
#define GO_PI_2 M_PI_2
#else
#define GO_PI_2 1.57079632679489661923
#endif

#ifdef M_PI_4
#define GO_PI_4 M_PI_4
#else
#define GO_PI_4 0.78539816339744830962
#endif

#define GO_TO_DEG(rad) ((rad)*57.295779513082323)
#define GO_TO_RAD(deg) ((deg)*0.0174532925199432952)

/* how close translational quantities must be to be equal */
#define GO_TRAN_CLOSE(x,y) (fabs((x)-(y)) < GO_REAL_EPSILON)
#define GO_TRAN_SMALL(x) (fabs(x) < GO_REAL_EPSILON)

/* how close rotational quantities must be to be equal */
#define GO_ROT_CLOSE(x,y) (fabs((x)-(y)) < GO_REAL_EPSILON)
#define GO_ROT_SMALL(x) (fabs(x) < GO_REAL_EPSILON)

/* how close general quantities must be to be equal */
#define GO_CLOSE(x,y) (fabs((x)-(y)) < GO_REAL_EPSILON)
#define GO_SMALL(x) (fabs(x) < GO_REAL_EPSILON)

typedef struct {
  go_real x, y, z;
} go_cart;

typedef struct {
  go_real theta, phi, r;
} go_sph;

typedef struct {
  go_real theta, r, z;
} go_cyl;

typedef struct {
  go_real x, y, z;
} go_uvec;

typedef struct {
  go_real s, x, y, z;
} go_rvec;

/*            |  m.x.x   m.y.x   m.z.x  | */
/* go_mat m = |  m.x.y   m.y.y   m.z.y  | */
/*            |  m.x.z   m.y.z   m.z.z  | */

typedef struct {
  go_cart x;			/* x unit vector */
  go_cart y;			/* y unit vector */
  go_cart z;			/* z unit vector */
} go_mat;

typedef struct {
  go_real s, x, y, z;
} go_quat;

typedef struct {
  go_real z, y, zp;
} go_zyz;

typedef struct {
  go_real z, y, x;
} go_zyx;

typedef struct {
  go_real r, p, y;
} go_rpy;

typedef struct {
  go_cart tran;
  go_quat rot;
} go_pose;

extern go_pose
go_pose_this(go_real x, go_real y, go_real z,
	     go_real rs, go_real rx, go_real ry, go_real rz);

extern go_pose
go_pose_identity(void);

typedef struct {
  go_cart tran;
  go_mat rot;
} go_hom;

typedef struct {
  go_cart tran;
  go_rpy rot;
} go_rpose;

/*
  struct arguments are passed to functions as const pointers since the
  speed is at least as fast for all but structs of one or two elements.
*/

/* translation rep conversion functions */

extern go_result go_cart_sph_convert(const go_cart *, go_sph *);
extern go_result go_cart_cyl_convert(const go_cart *, go_cyl *);
extern go_result go_sph_cart_convert(const go_sph *, go_cart *);
extern go_result go_sph_cyl_convert(const go_sph *, go_cyl *);
extern go_result go_cyl_cart_convert(const go_cyl *, go_cart *);
extern go_result go_cyl_sph_convert(const go_cyl *, go_sph *);

/* rotation vector to true vector conversions */

extern go_result go_cart_rvec_convert(const go_cart *, go_rvec *);
extern go_result go_rvec_cart_convert(const go_rvec *, go_cart *);
extern go_result go_cart_rpy_convert(const go_cart *, go_rpy *);
extern go_result go_rpy_cart_convert(const go_rpy *, go_cart *);

/* rotation rep conversion functions */

extern go_result go_rvec_quat_convert(const go_rvec *, go_quat *);
extern go_result go_rvec_mat_convert(const go_rvec *, go_mat *);
extern go_result go_rvec_zyz_convert(const go_rvec *, go_zyz *);
extern go_result go_rvec_zyx_convert(const go_rvec *, go_zyx *);
extern go_result go_rvec_rpy_convert(const go_rvec *, go_rpy *);

extern go_result go_quat_rvec_convert(const go_quat *, go_rvec *);
extern go_result go_quat_mat_convert(const go_quat *, go_mat *);
extern go_result go_quat_zyz_convert(const go_quat *, go_zyz *);
extern go_result go_quat_zyx_convert(const go_quat *, go_zyx *);
extern go_result go_quat_rpy_convert(const go_quat *, go_rpy *);

extern go_result go_mat_rvec_convert(const go_mat *, go_rvec *);
extern go_result go_mat_quat_convert(const go_mat *, go_quat *);
extern go_result go_mat_zyz_convert(const go_mat *, go_zyz *);
extern go_result go_mat_zyx_convert(const go_mat *, go_zyx *);
extern go_result go_mat_rpy_convert(const go_mat *, go_rpy *);

extern go_result go_zyz_rvec_convert(const go_zyz *, go_rvec *);
extern go_result go_zyz_quat_convert(const go_zyz *, go_quat *);
extern go_result go_zyz_mat_convert(const go_zyz *, go_mat *);
extern go_result go_zyz_zyx_convert(const go_zyz *, go_zyx *);
extern go_result go_zyz_rpy_convert(const go_zyz *, go_rpy *);

extern go_result go_zyx_rvec_convert(const go_zyx *, go_rvec *);
extern go_result go_zyx_quat_convert(const go_zyx *, go_quat *);
extern go_result go_zyx_mat_convert(const go_zyx *, go_mat *);
extern go_result go_zyx_zyz_convert(const go_zyx *, go_zyz *);
extern go_result go_zyx_rpy_convert(const go_zyx *, go_rpy *);

extern go_result go_rpy_rvec_convert(const go_rpy *, go_rvec *);
extern go_result go_rpy_quat_convert(const go_rpy *, go_quat *);
extern go_result go_rpy_mat_convert(const go_rpy *, go_mat *);
extern go_result go_rpy_zyz_convert(const go_rpy *, go_zyz *);
extern go_result go_rpy_zyx_convert(const go_rpy *, go_zyx *);

/* combined rep conversion functions */

extern go_result go_pose_hom_convert(const go_pose *, go_hom *);
extern go_result go_hom_pose_convert(const go_hom *, go_pose *);

/* translation functions, that work only with the preferred
   go_cart type. Other types must be converted to go_cart
   to use these, e.g., there's no go_sph_cyl_compare() */

extern go_flag go_cart_cart_compare(const go_cart *, const go_cart *);
extern go_result go_cart_cart_dot(const go_cart *, const go_cart *,
				  go_real *);
extern go_result go_cart_cart_cross(const go_cart *, const go_cart *,
				    go_cart *);
extern go_result go_cart_mag(const go_cart *, go_real *);
extern go_result go_cart_cart_disp(const go_cart *, const go_cart *,
				   go_real *);
extern go_result go_cart_cart_add(const go_cart *, const go_cart *,
				  go_cart *);
extern go_result go_cart_cart_sub(const go_cart *, const go_cart *,
				  go_cart *);
extern go_result go_cart_scale_mult(const go_cart *, go_real, go_cart *);
extern go_result go_cart_scale_div(const go_cart *, go_real, go_cart *);
extern go_result go_cart_neg(const go_cart *, go_cart *);
extern go_result go_cart_unit(const go_cart *, go_cart *);
extern go_result go_cart_is_norm(const go_cart *);
extern go_result go_cart_inv(const go_cart *, go_cart *);
extern go_result go_cart_cart_proj(const go_cart *, const go_cart *,
				   go_cart *);
extern go_result go_cart_plane_proj(const go_cart *, const go_cart *,
				    go_cart *);
extern go_result go_cart_cart_angle(const go_cart *, const go_cart *,
				    go_real *);
extern go_result go_cart_centroid(const go_cart * varray,
				  go_integer num,
				  go_cart * centroid);
extern go_result go_cart_centroidize(const go_cart * vinarray,
				     go_integer num,
				     go_cart * centroid,
				     go_cart * voutarray);
extern go_result go_cart_cart_pose(const go_cart *, const go_cart *,
				   go_cart *, go_cart *,
				   go_integer, go_pose *);

/*!
  Returns the Cartesian point \a p whose distances from three other points
  \a c1, \a c2 and \a c3 are \a l1, \a l2 and \a l3, respectively. In
  general there are 0, 1 or two points possible. If no point is possible,
  this returns GO_RESULT_ERROR, otherwise the points are returned in \a
  p1 and \a p2, which may be the same point.
*/

go_result go_cart_trilaterate(const go_cart * c1, 
			      const go_cart * c2,
			      const go_cart * c3,
			      go_real l1,
			      go_real l2,
			      go_real l3,
			      go_cart * p1,
			      go_cart * p2);

/* unit vector functions, which assume unit vectors as arguments. If they
   aren't the results will be unexpected */

extern go_result go_uvec_uvec_dot(const go_uvec *, const go_uvec *,
				  go_real *);
extern go_result go_uvec_uvec_angle(const go_uvec *, const go_uvec *,
				    go_real *);

/* quat functions */

extern go_flag go_quat_quat_compare(const go_quat *, const go_quat *);
extern go_result go_quat_mag(const go_quat *, go_real *);
extern go_result go_quat_unit(const go_quat *, go_quat *);
extern go_result go_quat_norm(const go_quat *, go_quat *);
extern go_result go_quat_inv(const go_quat *, go_quat *);
extern go_result go_quat_is_norm(const go_quat *);
extern go_result go_quat_scale_mult(const go_quat *, go_real, go_quat *);
extern go_result go_quat_scale_div(const go_quat *, go_real, go_quat *);
extern go_result go_quat_quat_mult(const go_quat *, const go_quat *,
				   go_quat *);
extern go_result go_quat_cart_mult(const go_quat *, const go_cart *,
				   go_cart *);

/* rotation vector functions */

extern go_flag go_rvec_rvec_compare(const go_rvec * r1, const go_rvec * r2);
extern go_result go_rvec_scale_mult(const go_rvec *, go_real, go_rvec *);
extern go_result go_rvec_scale_div(const go_rvec *, go_real, go_rvec *);
extern go_result go_rvec_is_norm(const go_rvec *);
extern go_result go_rvec_norm(const go_rvec *, go_rvec *);

/* rotation matrix functions */

/*        |  m.x.x   m.y.x   m.z.x  |   */
/*   M =  |  m.x.y   m.y.y   m.z.y  |   */
/*        |  m.x.z   m.y.z   m.z.z  |   */

extern go_result go_mat_norm(const go_mat *, go_mat *);
extern go_result go_mat_is_norm(const go_mat *);
extern go_result go_mat_inv(const go_mat *, go_mat *);
extern go_result go_mat_cart_mult(const go_mat *, const go_cart *, go_cart *);
extern go_result go_mat_mat_mult(const go_mat *, const go_mat *, go_mat *);

/* pose functions*/

extern go_flag go_pose_pose_compare(const go_pose *, const go_pose *);
extern go_result go_pose_inv(const go_pose *, go_pose *);
extern go_result go_pose_cart_mult(const go_pose *, const go_cart *,
				   go_cart *);
extern go_result go_pose_pose_mult(const go_pose *, const go_pose *,
				   go_pose *);

/* homogeneous transform functions */

extern go_result go_hom_inv(const go_hom *, go_hom *);

/* declarations for general MxN matrices */

/*!
  Declare a matrix variable \a m with \a rows rows and \a cols columns.
  Allocates \a rows X \a columns of space in \a mspace.
*/

typedef go_real go_vector;

typedef struct {
  go_integer rows;
  go_integer cols;
  go_real ** el;
  go_real ** elcpy;
  go_real * v;
  go_integer * index;
} go_matrix;

#define GO_MATRIX_DECLARE(M,Mspace,_rows,_cols) \
go_matrix M = {0, 0, 0, 0, 0, 0}; \
struct { \
  go_real * el[_rows]; \
  go_real * elcpy[_rows]; \
  go_real stg[_rows][_cols]; \
  go_real stgcpy[_rows][_cols]; \
  go_real v[_rows]; \
  go_integer index[_rows]; \
} Mspace

#define go_matrix_init(M,Mspace,_rows,_cols) \
M.el = Mspace.el;				    \
M.elcpy = Mspace.elcpy;				    \
for (M.rows = 0; M.rows < (_rows); M.rows++) { \
  M.el[M.rows] = Mspace.stg[M.rows];	    \
  M.elcpy[M.rows] = Mspace.stgcpy[M.rows];	    \
} \
M.rows = (_rows); \
M.cols = (_cols); \
M.v = Mspace.v; \
M.index = Mspace.index

extern go_result
go_set_singular_epsilon(go_real epsilon);

extern go_result
ludcmp(go_real ** a,
       go_real * scratchrow,
       go_integer n,
       go_integer * indx,
       go_real * d);

extern go_result
lubksb(go_real ** a,
       go_integer n,
       go_integer * indx,
       go_real * b);

/* MxN matrix, Mx1 vector functions */

extern go_result
go_cart_vector_convert(const go_cart * c,
		       go_vector * v);
extern go_result
go_vector_cart_convert(const go_real * v,
		       go_cart * c);

extern go_result
go_quat_matrix_convert(const go_quat * quat,
		      go_matrix * matrix);

extern go_result
go_mat_matrix_convert(const go_mat * mat,
		      go_matrix * matrix);

extern go_result
go_matrix_matrix_add(const go_matrix * a,
		     const go_matrix * b,
		     go_matrix * apb);

extern go_result
go_matrix_matrix_copy(const go_matrix * src,
		      go_matrix * dst);

extern go_result
go_matrix_matrix_mult(const go_matrix * a,
		      const go_matrix * b,
		      go_matrix * ab);

extern go_result
go_matrix_vector_mult(const go_matrix * a,
		      const go_vector * v,
		      go_vector * av);

/*!
  The matrix-vector cross product is a matrix of the same dimension,
  whose columns are the column-wise cross products of the matrix
  and the vector. The matrices must be 3xN, the vector 3x1.
*/
extern go_result
go_matrix_vector_cross(const go_matrix * a,
		       const go_vector * v,
		       go_matrix * axv);

extern go_result
go_matrix_transpose(const go_matrix * a,
		    go_matrix * at);

extern go_result
go_matrix_inv(const go_matrix * a,
	      go_matrix * ainv);

/* Square matrix functions, where matN is an NxN matrix, and vecN
   is an Nx1 vector */

/*!
  Given a 6x6 matrix \a a, computes the inverse and returns it in
  \a ainv. Leaves \a a untouched. Returns GO_RESULT_OK if there is an
  inverse, else GO_RESULT_SINGULAR if the matrix is singular.
*/
extern go_result go_mat6_inv(const go_real a[6][6],
			     go_real ainv[6][6]);

/*!
  Given two 6x6 matrices \a a and \a b, multiplies them and returns
  the result in \a axb. Leaves \a a and \a b untouched.
  Returns GO_RESULT_OK.
*/
extern go_result go_mat6_mat6_mult(const go_real a[6][6],
				   const go_real b[6][6],
				   go_real axb[6][6]);

/*!
  Given a 6x6 matrix \a a and a 6x1 vector \a v, multiplies them
  and returns the result in \a axv. Leaves \a a and \a v untouched.
  Returns GO_RESULT_OK.
*/
extern go_result go_mat6_vec6_mult(const go_real a[6][6],
				   const go_real v[6],
				   go_real axv[6]);

/* Denavit-Hartenberg to pose conversions */

/*
  The link frams is assumed to be

  | i-1
  |    T
  |   i

  that is, elements of the link frame expressed wrt the
  previous link frame.
*/

/*!
  These DH parameters follow the convention in John J. Craig,
  _Introduction to Robotics: Mechanics and Control_.
*/
typedef struct {
  go_real a;			/*< a[i-1] */
  go_real alpha;		/*< alpha[i-1] */
  go_real d;			/*< d[i] */
  go_real theta;		/*< theta[i] */
} go_dh;

/*!
  PK parameters are used for parallel kinematic mechanisms, and
  represent the Cartesian positions of the ends of the link in the
  stationary base frame and the moving platform frame.
 */
typedef struct {
  go_cart base;			/*< position of fixed end in base frame */
  go_cart platform;		/*< position of moving end in platform frame  */
} go_pk;

/*!
  PP parameters represent the pose of the link with respect to the
  previous link, with \a prismatic non-zero for prismatic links
  and zero for revolute links. Prismatic joints are assumed to
  move along the Z axis, revolute joints about the Z axis.
 */
typedef struct {
  go_pose pose;		/* e.g., from go_dh_pose_convert */
} go_pp;

/*! Types of link parameter representations  */
enum {
  GO_LINK_DH = 1,		/*< for Denavit-Hartenberg params  */
  GO_LINK_PK,			/*< for parallel kinematics  */
  GO_LINK_PP			/*< for serial kinematics */
};

/*!
  This is the generic link structure for PKM sliding/cable links and 
  serial revolute/prismatic links.
 */
typedef struct {
  union {
    go_dh dh; /*< if you have DH params and don't want to convert to PP */
    go_pk pk; /*< if you have a parallel machine, e.g., hexapod or robot crane */
    go_pp pp; /*< if you have a serial machine, e.g., an industrial robot  */
  } u;
  go_flag type;			/*< one of GO_LINK_DH,PK,PP  */
  go_flag quantity;		/*< one of GO_QUANTITY_LENGTH,ANGLE */
} go_link;

/*!
  Converts DH parameters in \a dh to their pose equivalent, stored
  in \a pose.
 */
extern go_result go_dh_pose_convert(const go_dh * dh, go_pose * pose);

/*!
  Converts \a pose to the equivalent DH parameters, stored in \a dh.
  Warning! Conversion from these DH parameters back to a pose via \a
  go_dh_pose_convert will NOT in general result in the same
  pose. Poses have 6 degrees of freedom, DH parameters have 4, and
  conversion to DH parameters loses some information. The source of
  this information loss is the convention imposed on DH parameters for
  choice of X-Y-Z axis directions. With poses, there is no such
  convention, and poses are thus freer than DH parameters.
 */
extern go_result go_pose_dh_convert(const go_pose * pose, go_dh * dh);

/*
  Fixes the pose in PP parameters \a pp to its value when the joint
  variable is \a joint, storing the result in \a ppout. \a pp
  represents the pose when the joint value is 0, its origin; \a ppout
  represents the pose at an arbitrary joint value.
*/
extern go_result go_link_joint_set(const go_link * link, go_real joint, go_link * linkout);

typedef struct {
  go_real re;
  go_real im;
} go_complex;

extern go_complex go_complex_add(go_complex z1, go_complex z2);
extern go_complex go_complex_sub(go_complex z1, go_complex z2);
extern go_complex go_complex_mult(go_complex z1, go_complex z2);
extern go_complex go_complex_div(go_complex z1, go_complex z2, go_result * result);
extern go_complex go_complex_scale(go_complex z, go_real scale);
extern go_real go_complex_mag(go_complex z);
extern go_real go_complex_arg(go_complex z);
extern void go_complex_sqrt(go_complex z, go_complex * z1, go_complex * z2);
extern void go_complex_cbrt(go_complex z, go_complex * z1, go_complex * z2, go_complex * z3);

typedef struct {
  /* leading coefficient is 1, x^2 + ax + b = 0 */
  go_real a;
  go_real b;
} go_quadratic;

typedef struct {
  /* leading coefficient is 1, x^3 + ax^2 + bx + c = 0 */
  go_real a;
  go_real b;
  go_real c;
} go_cubic;

typedef struct {
  /* leading coefficient is 1, x^4 + ax^3 + bx^2 + cx + d = 0 */
  go_real a;
  go_real b;
  go_real c;
  go_real d;
} go_quartic;

extern go_result go_quadratic_solve(const go_quadratic * quad,
				    go_complex * z1,
				    go_complex * z2);

extern go_result go_cubic_solve(const go_cubic * cub,
				go_complex * z1,
				go_complex * z2,
				go_complex * z3);

extern go_result go_quartic_solve(const go_quartic * quart,
				  go_complex * z1,
				  go_complex * z2,
				  go_complex * z3,
				  go_complex * z4);

extern go_result go_tridiag_reduce(go_real ** a,
				   go_integer n,
				   go_real * d,
				   go_real * e);

extern go_result go_tridiag_ql(go_real * d,
			       go_real * e,
			       go_integer n,
			       go_real ** z);

#endif /* GO_MATH_H */

/*
  $Log: gomath.h,v $
  Revision 1.1  2007/05/09 15:19:06  proctor
  Added the math source files from go motion, and changed the Makefile.am
  so that robojoy links these in.

  Revision 1.27  2007/05/08 21:10:20  proctor
  Reworked the handling of kinematic parameters, taking them out of the TRAJ
  section and putting them in the SERVO sections as DH,PP,PK_PARAMETERS
  and getting rid of unit handling-- now everything is converted to Go units
  before sending to Go.

  Revision 1.26  2007/05/04 15:20:24  proctor
  Comments on go_link

  Revision 1.25  2007/05/04 13:06:04  proctor
  Added comments on go_dh, go_pp and put in placeholders for go_pk and go_link

  Revision 1.24  2007/05/01 18:24:03  proctor
  Added go_cart_rpy,uvec convert functions

  Revision 1.23  2007/04/27 17:33:28  proctor
  Added go_set_singular_epsilon, and verified gojactest.c works and builds
  inverses for all three cases

  Revision 1.22  2007/04/26 21:37:19  proctor
  Added gojactest.c and conversions between go_quat and go_matrix

  Revision 1.21  2007/04/24 23:10:08  proctor
  Used new go_matrix structure

  Revision 1.20  2007/04/24 15:19:51  proctor
  Added general matrix functions

  Revision 1.19  2006/12/22 13:48:57  proctor
  Added traj_ref setup, but not handling-- still need to run the calculations
  in do_cmd_move_world, and this will wait until we get time for testing

  Revision 1.18  2006/11/22 15:36:15  proctor
  Fixed sqrt check on mat to quat convert, to fix problems in puma kins

  Revision 1.17  2006/11/20 17:39:28  proctor
  Took out RC matrix index macro in place of [][] notation

  Revision 1.16  2006/11/16 18:48:23  proctor
  Fixed unix rtapi_arg bug with string malloc

  Revision 1.15  2006/11/14 14:02:48  proctor
  Added roboch stuff

  Revision 1.14  2006/11/08 21:28:25  proctor
  Added roboch Robocrane-Horizontal code, and updates to gomath for the
  trilateration functions. Also made the comparison values coarser.

  Revision 1.13  2006/10/31 21:19:07  proctor
  Completed go_pose_cart_mult, adding go_tridiag functions

  Revision 1.12  2006/10/10 14:24:35  proctor
  Replaced the original go_traj_cj_compute with one that had the correct
  two checks for the existence of an acceleration phase, and split out the
  code into smaller easy pieces. Moved go_sq,cub,qua into gomath.h.

  Revision 1.11  2006/10/07 19:38:05  proctor
  Put go_cbrt in gomath.c,h, took out of gotraj.c

  Revision 1.10  2006/09/20 19:55:41  proctor
  Fixed bug in go_cubic_solve where values that should be complex conjugates weren't,
  and complex roots of scalars were being used.

  Revision 1.9  2006/09/18 20:17:58  proctor
  Added beginnings of absolute orientation, namely complex number handling
  and quadratic, cubic solving in gomath.c

  Revision 1.8  2006/09/07 13:50:50  proctor
  Fixed stepper motor interface, and added deadband to PID.

  Revision 1.7  2006/07/25 14:36:25  proctor
  Changed go_rot to go_rvec

  Revision 1.6  2006-07-24 21:57:55  proctor
  In gomath, added Denavit-Hartenberg/pose conversion functions;
  in rtai, added some stepper-related changes

*/

