/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primMobEngine.hh

  \brief Vehicle primitive level planning engine

  \code CVS Status:
  $Author: ghenson $
  $Revision: 1.7 $
  $Date: 2008/10/13 18:26:39 $
  \endcode

  \author Stephen Balakirsky & Fred Proctor
*/
#ifndef __primMobEngine__
#define __primMobEngine__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv(), strtol()
#include <signal.h>		// SIGINT, signal()
#include <ctype.h>
#include <string.h>		// strncpy
#include <math.h>		// sqrt(), fabs(), atan2()
#include <float.h>		// FLT_EPSILON
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>
#ifndef RCS_STAT_SOURCE_FILE_LEN
#define RCS_STAT_SOURCE_FILE_LEN 64
#endif
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN

#define DEG(x) (57.295779513082323 * (x))
#define RAD(x)  (0.017453292519943 * (x))

#define PRIM_MOB_ENGINE_LOG_FILE_PREFIX "primMobEngineLog"
#define PRIM_MOB_ENGINE_LOG_FILE_SUFFIX ".log"
// how long one can be, at least prefix + suffix + digits in an int
#define PRIM_MOB_ENGINE_LOG_FILE_NAME_LEN 80

/*
  In IEEE floating point,

  FLT_MIN = 1.175494e-38, FLT_EPSILON 1.192093e-07
  DBL_MIN = 2.225074e-308, DBL_EPSILON 2.220446e-16

  The FLT_EPSILON value makes a good fuzzy comparison for the
  equality of two reals.
*/

#define IS_SMALL(x) (fabs(x) < FLT_EPSILON)
#define IS_POSITIVE(x) ((x) >= FLT_EPSILON)
#define WAYPOINT_INITIAL_SIZE MAX_WAYPOINTS // how many waypoints we start with
#define HYPOT(x,y) sqrt((x)*(x)+(y)*(y))
// force angles to be in range [-Pi..Pi)
#define NORM(x) while ((x) >= PM_PI) (x) -= PM_2_PI; while ((x) < -PM_PI) (x) += PM_2_PI


#define DEFAULT_CYCLE_TIME 0.050
#define DEFAULT_MAX_TRAN_VEL 1.0 // m/s
#define DEFAULT_MAX_TRAN_ACC 1.0 // m/s/s
#define DEFAULT_MAX_ROT_VEL 1.0	// rad/s
#define DEFAULT_MAX_ROT_ACC 1.0 // rad/s/s
#define DEFAULT_V_CUTOFF_ANGLE 1.0 // rad
#define DEFAULT_W_CUTOFF_ANGLE 1.0 // rad
#define DEFAULT_DEBUG 0

class PrimMobEngine{
 private:
  // computationally efficient values from wheel radius, separation
  double v, w; // angular and linear velocities
  double wheel_radius;
  double wheel_radius_inverse;
  double wheel_separation;
  double wheel_separation_half;
  int waypoint_size;	// size of whole array
  int waypoint_num;	// how many currently in the array
  int waypoint_index;	// which one we're working on
  Waypoint *waypoint_array;
  int *wpsToArcs;       // maps waypoints to arcs
  bool useAckerman;
  double ackermanWheelbase;
  bool useNautic;
  double nauticLength;
  FILE * log_fp;
  // information from settings channel
  double setCycleTime;
  double setVmax;
  double setAmax;
  double setWmax;
  double setAlphamax;
  double setVCutoffAngle;
  double setWCutoffAngle;
  PM_POSE setControlPoint;
  int    setDebug;
  bool in_rev;
  bool right_rev;
  bool prim_path;
  // end of settings

  inline double magxy(PM_CARTESIAN);


 public:
  PrimMobEngine();
  bool isAckerman();
  bool isNautic();
  bool isPrimPath();
  void setPrimPath(bool primpath);
  void setWaypoints( Waypoint*, int );
  void resetWaypointIndex();
  void setAckermanWheelbase(double);
  void setNauticLength(double);
  void setWheelSettings(double, double);
  int arc_segments_to_waypoints(ArcSegment *, int);
  void setSettings(double cycleTimeIn,
		   double vmaxIn,
		   double amaxIn,
		   double wmaxIn,
		   double alphamaxIn,
		   double vCutoffAngleIn,
		   double wCutoffAngleIn,
		   PM_POSE controlPoint,
		   int debugIn);
  RCS_STATUS doCmdNop();
  RCS_STATUS doCmdInit();
  RCS_STATUS doCmdHalt();
  RCS_STATUS doCmdAbort();
  RCS_STATUS doCmdShutdown();
  RCS_STATUS doCmdMoveWaypoint( RCS_STATE state,
				const PM_CARTESIAN &tranRate, 
				const PM_RPY &rpyRate,
				const PM_CARTESIAN &tran,
				const PM_RPY &rpy,
				// the rest of the arguments are return values
				int &pathIndex,
				int &nextState,
				double &retV1, double &retV2, double &retV3);
  RCS_STATUS doCmdRotate(RCS_STATE state, double cmdTheta, double cmdThetaTol, int cmdDirection,
			 const PM_RPY &rpy,
			 // the rest of the arguments are return values
			 int &pathIndex,
			 int &nextState,
			 double &retV1, double &retV2 );
  RCS_STATUS doCmdVel(RCS_STATE state, double cmdV, double cmdW,
		       // the rest of the arguments are return values
		       int &nextState,
		       double &retV1, double &retV2, double &retV3 );

  RCS_STATUS doCmdStop(RCS_STATE state,
		       // the rest of the arguments are return values
		       int &nextState,
		       double &retV1, double &retV2, double &retV3 );
  RCS_STATUS doCfgNop(RCS_STATE state, int &nextState);
  RCS_STATUS doCfgCycleTime(RCS_STATE state, int &nextState, double cycleTimeIn, double &cycleTimeOut);

  RCS_STATUS doCfgMove(RCS_STATE state, int &nextState, 
		       double vmaxIn, double amaxIn, double wmaxIn, double alphamaxIn, 
		       double vCutoffAngleIn, double wCutoffAngleIn,
		       // returned values
		       double &vmaxOut, double &amaxOut, double &wmaxOut, double &alphamaxOut, 
		       double &vCutoffAngleOut, double &wCutoffAngleOut);

  RCS_STATUS doCfgDebug(RCS_STATE state, int &nextState, int debugIn, int &debugOut );

  void printDebugMsg(int mask, const char * fmt, ...);
};


#endif
/*
  Modification history:

  $Log: primMobEngine.hh,v $
  Revision 1.7  2008/10/13 18:26:39  ghenson
  Update to prim. Only sent the robot into reverse if prim generated the path

  Revision 1.6  2008/05/22 18:49:05  ghenson
  included state flags for ICR based reverse

  Revision 1.5  2008/01/08 20:18:15  proctor
  Added control point transform that lets the prim mobility controller act
  on the rotational center of the vehicle, which in general won't coincide
  with the nav data frame

  Revision 1.4  2007/12/19 18:36:37  proctor
  Fixed logging so that logs are created for each debug config message, and
  given the debug mask as part of the log file name.
  Changed the prim waypoint algorithm to slow down only when approaching the
  last waypoint. This should allow it to get up to speed when moving along
  intermediate waypoints.

  Revision 1.3  2007/12/19 14:48:38  proctor
  Added printDebugMsg to primMobEngine, and pulled the waypoint debug printing
  into this. Set the debug level to PRIM_MOB_JA_DEBUG_WAYPOINT, 0x02, to
  enable debug printing of waypoint information. The debug level is set to 0
  in moast.ini by default, so you will see no debug information.

  Revision 1.2  2007/05/16 15:16:21  dr_steveb
  Fixed location of state printing routines to remove compile warnings.

  Revision 1.1  2007/05/11 16:38:33  dr_steveb
  Changed the location of primMobEngine.hh to the include directory

  Revision 1.1  2007/05/10 13:18:29  dr_steveb
  Separated the prim mobility planning engine from the interface wrappers. This will allow a controller to be created (the traffic controller) that controls a large number of vehicles without the overhead of individual NML channels for each vehicle. No changes in the algorithms were made.

*/
