/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   scrollingMap.hh
  \brief  Contains class declaration for ScrollingMap

*****************************************************************************

HOW TO USE THIS FILE

To use this file in an application:
1.  The user must write an application header file (call it app.hh).
2.  The user must write an application implementation file (call it app.cc).
3.  The user must write a file (call it cell.hh) that defines a cell class.
4.  app.hh must #define CELL_NAME.
5.  app.hh must #include cell.hh.
6.  app.hh must #include "scrollingMap.hh" (AFTER cell.hh!).
7.  The name of the cell class in cell.hh must be the value of CELL_NAME.
8.  The cell class in cell.hh must be derived from the ScrollingMapCell class.
9.  cell.hh must #include "scrollingMapCell.hh"
10. cell.hh must have a public "init()" method taking no arguments.
11. app.hh may define a map class that is a derived class of the
    ScrollingMap class defined here (or it may use the ScrollingMap class).
12. app.cc must #include app.hh.
13. app.cc must #include "scrollingMap.cc".

EXAMPLE 1 (detailed)

See the files exampleScrollMap.cc, exampleScrollMap.hh, and exampleCell.hh
in the src/tools directory. Also try out the executable exampleScrollMap
in the bin directory. This is a simple example showing many of the features
of a scrolling map. This example uses the ScrollingMap class directly;
it does not derive a more specialized class from ScrollingMap.

EXAMPLE 2 (outline)

The amSPMain executable uses the scrolling map. Only a small fraction
of the code involves the scrolling map. The connections between the
executable and the scrolling map are made using the 13 rules above, as
follows.

The app is amSPMap.
The file amSPMap.hh has the following lines:

  #define CELL_NAME AmSPCell
  #include "amSPCell.hh"
  #include "scrollingMap.hh"
  ...
  class AmSPMap : public ScrollingMap

The file amSPCell.hh has the following lines:

  #include "scrollingMapCell.hh"
  ...
  class AmSPCell : public ScrollingMapCell
  ...
  void init(){height=0; range=0; probability = CELL_PROBINIT; hits = 0;}

The file amSPMap.cc has the following lines:

  #include "amSPMap.hh"
  #include "../tools/scrollingMap.cc"

*************************************************************************

Conceptually, a scrolling map is a square window that lies on an unbounded
cartesian XY plane and can move around (scroll) so that different parts of
the plane can be seen in the window. The edges of the window stay aligned
with the axes of the plane at all times. The window is divided into a grid
of square cells called the "window grid". When the window moves, it moves
an integer number of grid cells in the X and/or Y directions. The user can
tell the window to scroll so that its center cell contains a specific
cartesian point called the center point. The number of cells on a side of
the window grid is odd to make it possible to keep the center point in the
cell at the center of the window grid. Three methods dealing with the
center point are provided:
1. getCenterX (takes no arguments, returns the X coordinate used to locate
   the center cell of the window grid).
2. getCenterY (takes no arguments, returns the Y coordinate used to locate
   the center cell of the window grid).
3. setCenter (takes doubles x and y, and pointers to ints xDiff and yDiff,
   scrolls the window so that the point (x, y) is in the center cell of
   the window grid; also sets xDiff and yDiff to the number of cells
   the center cell has moved).

The constructor of a scrollingMap, takes as arguments the "resolution" (a
double representing the length of a side of a cell), and the "range" (an
int representing the number of cells between the center cell of the window
grid and the edge of the window grid). The number of cells on a side of the
window grid, called "side", is one more than twice the range. In equation
form, side = ((2 * range) + 1). No length units are specified. Units are
left to the user and may be implicit.

Several access methods are provided to get data about a ScrollingMap.
1. getCols returns the number of columns in the window grid.
2. getRange returns the map range.
3. getResolution returns the map resolution.
4. getRows returns the number of rows in the window grid.
5. getXHi returns the maximum value of X of any point in the window.
6. getXLow returns the minimum value of X of any point in the window.
7. getYHi returns the maximum value of Y of any point in the window.
8. getYLow returns the minimum value of Y of any point in the window.

Three access methods are provided to get a pointer to a cell:
1. getCellAtIndexes (which takes X and Y window grid indexes as arguments).
2. getCellAtCoords (which takes X and Y plane coordinates as arguments).
3. getCellWithId (which takes an id number as an argument).

Several access methods are provided for getting coordinates and indexes.
1. getXCoordAtCell returns the X coordinate of the center of a given cell.
2. getXCoordAtId returns the X coordinate of the center of the cell
   with a given id.
3. getYCoordAtCell returns the Y coordinate of the center of a given cell.
4. getYCoordAtId returns the Y coordinate of the center of the cell
   with a given id.
5. getXIndexAtCell returns the window grid X index of a given cell.
6. getYIndexAtCell returns the window grid Y index of a given cell.
7. getXIndexAtCoord returns the window grid X index of an X coordinate value.
8. getYIndexAtCoord returns the window grid Y index of a Y coordinate value.

Two methods are provided to make it possible to get all the cells that
are crossed by a line between two points.
1. initLine takes the coordinates of the two points and returns a pointer
   to the cell containing the first point. It also gets the system ready
   for calls to getNextCell.
2. getNextCell returns a pointer to the next cell on the line defined by
   the most recent call to initLine.

To make it dynamic and efficient, the implementation of the ScrollingMap
class differs from the conceptual view.  The implementation uses what is
called the "actual grid", a fixed grid of the same size as the window
grid. The lower left cell of the window grid (minimum X and Y) is located
on the actual grid by the "originX" and the "originY" which are indexes on
the actual grid of that cell.  The actual grid is divided into four
sections by horizontal and vertical lines through lower left corner of the
cell whose indexes are originX and originY. The four sections of the actual
grid represent four parts of the window grid.

Cells are expected to be instances of a subclass of the ScrollingMapCell
class, which is defined in scrollingMapCell.hh. The subclass should be
defined in some other file by the user. The ScrollingMapCell class has
only one data member, an integer id number. Id numbers are assigned to
cells by the ScrollingMap class constructor. There is no method of
changing them.

It is required that that an "init" method taking no arguments and returning
void be defined in the cell subclass defined by the user. The init method
is invoked by methods of the ScrollingMap class to initialize cells as
desired by the user. The ScrollingMap class knows the locations and id
numbers of the cells but does not know anything about the user's cell class
except that it has an init method.

The actual grid is implemented as a dynamic 1D array of cells called
"cells". The private method getCell1D gets a pointer to a cell given an
argument that is the cell's index in the one-dimensional array.  This is
identical to the public getCellWithId method except that it does not check
whether the id argument is in bounds.

The getCell2D private access method gets a pointer to a cell as though
the actual grid were a 2D array. That method takes the cell's indexes
on the actual grid as arguments. For the getCell2D method, the 1D
array is conceptually mapped to the 2D actual grid by chopping the 1D
array into horizontal sections "side" cells long and stacking
successive sections on top of one another vertically.

A "row" of a grid is a set of cells with the same X index. A
"column" of a grid is a set of cells with the same Y index.
Note that this differs from the usual convention.
Since the grids are square, the following are all equal:
 - the number of cells in a row
 - the number of cells in a column
 - the number of rows
 - the number of columns
 - the number of cells on any side of the grid.

To make the code clearer (despite the fact that a single term would
suffice), both "rows" and "cols" (i.e., columns) are data members of
the ScrollingMap class.

When a scrolling map is first created, the window grid and the actual grid
are the same, and the origin of the plane is in the center of the cell in
the center of the grid. The X and Y indexes of the cell at the center of
the actual grid are both equal to the range.

Conceptually, when the window scrolls, empty cells enter the window grid
from the direction in which the window moved, and cells leave the window
grid in the opposite direction.  For implementation efficiency, when
scrolling takes place, rather than copying from one place to another data
for the cells on the actual grid that represent parts of the window grid
that did not enter or leave the window grid, the data for those cells is
not touched. Rather, the data in the cells that conceptually left the
window grid is wiped out, and those same cells are reused to represent the
cells that entered the window grid. As mentioned earlier, the user must
define an init method that reinitializes the contents of a cell. That
method is invoked by the map during scrolling.

The best way to visualize the relationship between the window and
the actual grid is to imagine that the plane is tiled with copies
of the actual grid. The edges of the copies divide the window into
one, two, or four parts. The location of each of the parts on a copy
of the actual grid is where that part lies on the actual grid.

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.5 $
  $Date: 2006/10/13 16:13:25 $
  \endcode
  \par CVS Log:
  \ref amSPMap_hh "More..."

  \author Tom Kramer
  \date   2006-09-08
*/

/***********************************************************************/

#ifndef __scrollingMap_hh
#define __scrollingMap_hh

#include <math.h> // floor

/***********************************************************************/
/***********************************************************************/

class ScrollingMap {

/**PUBLIC***************************************************************/

public:

/***********************************************************************/

//! The constructor taking no arguments creates a map but sets nothing.

  ScrollingMap(){};

/***********************************************************************/

/*! The constructor taking two arguments creates a map based on a
    range and resolution. */

  ScrollingMap(int rangeIn, double resolutionIn);

/***********************************************************************/

//!  The destructor deletes allocated memory.

  ~ScrollingMap();

/***********************************************************************/

/*! Public method getCellAtCoords returns a pointer to a cell whose cartesian
  X and Y coordinates are x and y. If the cartesian point is not in 
  the window, this returns NULL. */

  CELL_NAME * getCellAtCoords(double x, double y);

/***********************************************************************/

/*! Public method getCellAtIndexes returns a pointer to the cell whose
  X and Y indexes in the window grid are row and col. */

CELL_NAME * getCellAtIndexes(int row, int col);

/***********************************************************************/

/*! Public method getCellWithId returns a pointer to a cell whose id
  is given. If the id is not the id of a cell, this returns NULL. */

  CELL_NAME * getCellWithId(int id)
  {
    if ((id > -1) && (id < numberOfCells))
      return (cells + id);
    else
      return 0;
  }

/***********************************************************************/

/*! Public method getCenterX returns the cartesian X coordinate of the
  point used to set the center of the map. */

  double getCenterX() {return centerX;}

/***********************************************************************/

/*! The getCenterY method returns the cartesian Y coordinate of the
  point used to set the center of the map. */

  double getCenterY() {return centerY;}

/***********************************************************************/

//! Public method getCols returns the number of columns in the window grid.

  int getCols() {return cols;}

/***********************************************************************/

/*! Public method getNextCell returns a pointer to the next cell on
  the line defined by the most recent call to initLine. If the end of
  the line has been reached, a null pointer is returned. If setCenter
  has been called since the most recent call to initLine, a null
  pointer is returned. */

  CELL_NAME * getNextCell();
  
/***********************************************************************/

//! Public method getRange returns the map range.

  int getRange() {return range;}

/***********************************************************************/

//! Public method getResolution returns the map resolution.

  double getResolution() {return res;}

/***********************************************************************/

//! Public method getRows returns the number of rows in the window grid.

  int getRows() {return rows;}

/***********************************************************************/

/*! Public method getXCoordAtCell returns the X coordinate of the
  center of the given cell. If the cell is NULL or its id is not the
  id of a cell, this returns xLow. */

  double getXCoordAtCell(CELL_NAME * aCell);

/***********************************************************************/

/*! Public method getXCoordAtId returns the X coordinate of the center
  of the cell with the given id. If the id is not the id of a cell,
  this returns xLow. */

  double getXCoordAtId(int id);

/***********************************************************************/

/*! Public method getXCoordAtIndex returns the X coordinate of the
  center of the cell with the given X index. If the index is out of
  bounds, this returns a value off the map. */

  double getXCoordAtIndex(int xIndex){return(xLow + ((xIndex + 0.5) * res));}

/***********************************************************************/

/*! Public method getXHi returns the cartesian X coordinate of a point
  of the window with the greatest X value (i.e., any point on the top
  edge of the window). */

  double getXHi() {return xHi;}

/***********************************************************************/

/*! Public method getXIndexAtCell returns the window grid X index of the
  given cell. If the cell is NULL or its id is not the id of a cell, this
  returns -1. */

  int getXIndexAtCell(CELL_NAME * aCell);

/***********************************************************************/

/*! Public method getXIndexAtCoord returns the window grid X index of
  the X coordinate value x, if x is inside the window. Otherwise, it
  returns -1. */

  int getXIndexAtCoord(double x);

/***********************************************************************/

/*! Public method getXLow returns the cartesian X coordinate of a
  point of the window with the least X value (i.e., any point on the
  bottom edge of the window). */

  double getXLow() {return xLow;}

/***********************************************************************/

/*! Public method getYCoordAtCell returns the Y coordinate of the
  center of the given cell. If the cell is NULL or its id is not the
  id of a cell, this returns yLow. */

  double getYCoordAtCell(CELL_NAME * aCell);

/***********************************************************************/

/*! Public method getYCoordAtId returns the Y coordinate of the center
  of the cell with the given id. If the id is not the id of a cell,
  this returns yLow. */

  double getYCoordAtId(int id);

/***********************************************************************/

/*! Public method getYCoordAtIndex returns the Y coordinate of the
  center of the cell with the given Y index. If the index is out of
  bounds, this returns a value off the map. */

  double getYCoordAtIndex(int yIndex){return(yLow + ((yIndex + 0.5) * res));}

/***********************************************************************/

/*! Public method getYHi returns the cartesian Y coordinate of a point
  of the window with the greatest Y value (i.e., any point on the
  right edge of the window). */

  double getYHi() {return yHi;}

/***********************************************************************/

/*! Public method getYIndexAtCell returns the window grid Y index of the
  given cell. If the cell is NULL or its id is not the id of a cell, this
  returns -1. */

  int getYIndexAtCell(CELL_NAME * aCell);

/***********************************************************************/

/*! Public method getYIndexAtCoord returns the window grid Y index of
  the Y coordinate value y, if y is inside the window. Otherwise, it
  returns -1. */

  int getYIndexAtCoord(double y);

/***********************************************************************/

/*! Public method getYLow returns the cartesian Y coordinate of a
  point of the window with the least Y value (i.e., any point on the
  left edge of the window). */

  double getYLow() {return yLow;}

/***********************************************************************/

/*! Public method initLine initializes following a line from a start
  point to an end point. If the start point is in the window, a
  pointer to the cell containing the start point is returned; if not,
  a null pointer is returned.  The end point does not need to be in
  the window. The line will be followed from the start point up to the
  edge of the map. After this method is called, calls may be made
  repeatedly to getNextCell to get the cells through which the line
  passes. See the documentation of getNextCell. */

  CELL_NAME * initLine(double startX, double startY, double endX, double endY);

/***********************************************************************/

/*! Public method setCenter causes the window to scroll so that the
  cartesian point (x, y) lies inside the cell at the center of the
  window grid.  The originX, originY, xLow, xHi, yLow, yHi, centerX,
  and centerY of the map change when the center moves. Also, the cells
  of the actual grid that represent the cells entering (and leaving)
  the window grid are reinitialized. xDiff and yDiff are set to the
  number of cells the center has moved in the X and Y directions; the
  values may be positive, zero, or negative. */

  void setCenter(double x, double y, int * xDiff, int * yDiff);

/***********************************************************************/
/**PRIVATE**************************************************************/

private:

/*! Method getActualRawXIndex returns the X index on the actual grid of
  an cartesian X coordinate as though the grid were unbounded. */

  int getActualRawXIndex(double x)
  {
    return (int)floor((x / res) + halfSide);
  }
  
/***********************************************************************/

/*! Method getActualRawYIndex returns the Y index on the actual grid of
  a cartesian Y coordinate as though the grid were unbounded. */

  int getActualRawYIndex(double y)
  {
    return (int)floor((y / res) + halfSide);
  }

/***********************************************************************/

/*! Method getCell1D returns a pointer to the cell of the 1D array
  with the given 1D index. */

/*

For speed, this is not checking index bounds. The index should satisfy
(0 <= index < numberOfCells).

*/

  CELL_NAME * getCell1D(int index) {return (cells + index);}

/***********************************************************************/

/*! Method getCell2D returns a pointer to the cell of the actual grid
  with the given 2D indexes. */

/* 

For speed, this is not checking index bounds. The indexes should satisfy
(0 <= row < cols) and (0 <= col < rows).

*/

  CELL_NAME * getCell2D(int row, int col)
  {
    return (cells + (row * cols) + col);
  }

/***********************************************************************/

  CELL_NAME * cells;      //! 1D array of cells
  double centerX;         //! cartesian X coordinate of center point
  double centerY;         //! cartesian Y coordinate of center point
  int cols;               //! number of columns in the grid
  double halfSide;        //! half of side
  int numberOfCells;      //! number of cells
  int originX;            //! X index on actual grid of corner of window grid
  int originY;            //! Y index on actual grid of corner of window grid
  int range;              //! number of cells between center and side of grid
  double res;             //! short term for resolution
  double resolution;      //! length of a side of a cell
  int rows;               //! number of rows in the grid
  int side;               //! number of cells on side of grid
  double xHi;             //! maximum cartesian value of X in window
  double xLow;            //! minimum cartesian value of X in window
  double yHi;             //! maximum cartesian value of Y in window
  double yLow;            //! minimum cartesian value of Y in window

  /* The following are used by initLine and getNextCell */

  int deltaI;           //! amount by which i must change
  int deltaJ;           //! amount by which j must change
  double deltaX;        //! change in cartesian X from start to end
  double deltaY;        //! change in cartesian Y from start to end
  int dI;               //! amount by which i has changed
  int dJ;               //! amount by which j has changed
  bool inLine;          //! whether a line is being followed
  int i;                //! window grid X index of cell
  int j;                //! window grid Y index of cell
  double lack;          //! distance needed in secondary direction (positive)
  CELL_NAME * nextCell; //! cached next cell
  CELL_NAME * next1;    //! immediate next cell
  double step2;         //! size of step in secondary direction (positive)

};

#endif

/*! \defgroup amSPMap_hh CVS Log Information
 * $Log: scrollingMap.hh,v $
 * Revision 1.5  2006/10/13 16:13:25  tomrkramer
 * Added methods getXCoordAtIndex() and getYCoordAtIndex().
 * Changed setCenter to set xDiff and yDiff, ints which must be passed to it.
 * Readjusted a couple paragraphs that had lines that were too long.
 *
 * Revision 1.4  2006/09/29 14:57:08  tomrkramer
 * Added methods getXIndexAtCell and getYIndexAtCell.
 *
 * Revision 1.3  2006/09/28 22:21:44  tomrkramer
 * Major changes - added id to cell, methods of getting the location of a
 * cell, and methods for getting the cells through which a line passes.
 *
 * Revision 1.2  2006/09/27 19:43:17  dr_steveb
 * Moved private to protected function
 *
 * Revision 1.1  2006/09/18 18:52:00  tomrkramer
 * Initial entry.
 *
 *
 */
