
/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file servoEffJA.hh

  \brief NML declaration for the servo effector level.  Contains
         servo level sensor command, status, set, and data messages.  

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.4 $
  $Date: 2009/04/23 19:59:53 $
  \endcode

  \author Chris Scrapper
*/

#ifndef SERVO_EFF_JA_HH
#define SERVO_EFF_JA_HH

#include "moastTypes.hh" // MOAST_WORLD_NAME_LEN
#include "moastNmlOffsets.hh"

#define SERVO_EFF_JA_CMD_NAME          "servoEffJACmd"
#define SERVO_EFF_JA_STAT_NAME         "servoEffJAStat"
#define SERVO_EFF_JA_CFG_NAME          "servoEffJACfg"
#define SERVO_EFF_JA_SET_NAME          "servoEffJASet"

#define SERVO_EFF_JA_CMD_BASE            (SERVO_EFF_JA_BASE)
#define SERVO_EFF_JA_STAT_BASE           (SERVO_EFF_JA_BASE + 100)
#define SERVO_EFF_JA_CFG_BASE            (SERVO_EFF_JA_BASE + 200)
#define SERVO_EFF_JA_SET_BASE            (SERVO_EFF_JA_BASE + 300)


#define SERVO_EFF_JA_CMD_INIT_TYPE       (SERVO_EFF_JA_CMD_BASE + 1)
#define SERVO_EFF_JA_CMD_ABORT_TYPE      (SERVO_EFF_JA_CMD_BASE + 2)
#define SERVO_EFF_JA_CMD_HALT_TYPE       (SERVO_EFF_JA_CMD_BASE + 3)
#define SERVO_EFF_JA_CMD_SHUTDOWN_TYPE   (SERVO_EFF_JA_CMD_BASE + 4)
#define SERVO_EFF_JA_CMD_OPCODE_TYPE     (SERVO_EFF_JA_CMD_BASE + 5)

#define SERVO_EFF_JA_STAT_TYPE           (SERVO_EFF_JA_STAT_BASE + 1)

#define SERVO_EFF_JA_CFG_CYCLE_TIME_TYPE (SERVO_EFF_JA_CFG_BASE + 1)
#define SERVO_EFF_JA_CFG_DEBUG_TYPE      (SERVO_EFF_JA_CFG_BASE + 2)

#define SERVO_EFF_JA_SET_TYPE            (SERVO_EFF_JA_SET_BASE + 1)

//! Maximum number of effectors permitted in messages
#define SERVO_EFF_JA_EFFECTOR_MAX 6

//! Number of opcode types, defined below
#define SERVO_EFF_JA_OPCODE_TYPE_MAX 6
//! Valid Opcodes that can be used for controlling effector
typedef enum SERVO_EFF_JA_OPCODE_TYPE
  {
    SERVO_EFF_JA_ACTIVATE_OPCODE_TYPE,
    SERVO_EFF_JA_ANIMATE_OPCODE_TYPE,
    SERVO_EFF_JA_FIRE_OPCODE_TYPE,
    SERVO_EFF_JA_RELEASE_OPCODE_TYPE,
    SERVO_EFF_JA_RESET_OPCODE_TYPE,
    SERVO_EFF_JA_NOP_OPCODE_TYPE,
    SERVO_EFF_JA_INVALID_OPCODE_TYPE
  }ServoEffJAOpcodeType;


//! Configuration information a specific opcode
typedef struct SERVO_EFF_JA_OPCODE_CONFIG
{
  //! Effector opcode 
  ServoEffJAOpcodeType opcode;
  //! Maximum value of the opcode's parameter
  double max;
  //! Minimum value of the opcode's parameter
  double min;
}ServoEffJAOpcodeConfig;


//! configuration information for an effector
typedef struct SERVO_EFF_JA_CONFIG
{ 
  int effID;
  /*! Sensor mounting location
    if 0 then the sensor is hard mounted on vehicle 
    if not zero then the sensor is mounted on mission package
    To obtain the name of the mission package buffer that in which the sensor is 
    mounted, just concatenate SERVO_MIS_JA_<cmd type> + <vehicle id> + '_' + <mountID>
  */
  int mountID;
  int linkID;
  //! Sensor to Mount transform
  PM_POSE effToMount;
 
  //! Opcode that effector implements
  //! variable `opConf_length`  is the number of opcodes effector operates
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(ServoEffJAOpcodeConfig, opConf, SERVO_EFF_JA_OPCODE_TYPE_MAX);  
}ServoEffJAConfig;


//SERVO_EFF_JA_CMD_XXX
/////////////////////////////////////////////////

class ServoEffJACmdInit:public RCS_CMD_MSG {
public:
  ServoEffJACmdInit():RCS_CMD_MSG(SERVO_EFF_JA_CMD_INIT_TYPE,
				   sizeof(ServoEffJACmdInit)) {};
  void update(CMS *);
};

class ServoEffJACmdAbort:public RCS_CMD_MSG {
public:
  ServoEffJACmdAbort():RCS_CMD_MSG(SERVO_EFF_JA_CMD_ABORT_TYPE,
				    sizeof(ServoEffJACmdAbort)) {};
  void update(CMS *);
};

class ServoEffJACmdHalt:public RCS_CMD_MSG {
public:
  ServoEffJACmdHalt():RCS_CMD_MSG(SERVO_EFF_JA_CMD_HALT_TYPE,
				   sizeof(ServoEffJACmdHalt)) {};
  void update(CMS *);
};

class ServoEffJACmdShutdown:public RCS_CMD_MSG {
public:
  ServoEffJACmdShutdown():RCS_CMD_MSG(SERVO_EFF_JA_CMD_SHUTDOWN_TYPE,
				   sizeof(ServoEffJACmdShutdown)) {};
  void update(CMS *);
};


class ServoEffJACmdOpcode:public RCS_CMD_MSG {
public:
  ServoEffJACmdOpcode():RCS_CMD_MSG(SERVO_EFF_JA_CMD_OPCODE_TYPE,
				    sizeof(ServoEffJACmdOpcode)) {};
  void update(CMS *);
  int effID;
  ServoEffJAOpcodeType opcode;
  double param;
};


// SERVO_EFF_JA_STAT
//////////////////////////////////////////////

class ServoEffJAStat : public RCS_STAT_MSG_V2 {
public:
  ServoEffJAStat() : RCS_STAT_MSG_V2(SERVO_EFF_JA_STAT_TYPE, sizeof(ServoEffJAStat)) {};
  void update(CMS *);
  //! error code for the effector
  int error; 
  //! incremented once each cycle
  int heartbeat;
};

/*!
  ServoEffJACfgCycleTime sets the controller's cycle time
*/
class ServoEffJACfgCycleTime:public RCS_CMD_MSG {
public:
  ServoEffJACfgCycleTime():RCS_CMD_MSG(SERVO_EFF_JA_CFG_CYCLE_TIME_TYPE,
			       sizeof(ServoEffJACfgCycleTime)) {};
  void update(CMS *);
  //! period in seconds
  double cycleTime;
};

/*!
  ServoEffJACfgDebug sets the controller's debug level
*/
class ServoEffJACfgDebug:public RCS_CMD_MSG {
public:
  ServoEffJACfgDebug():RCS_CMD_MSG(SERVO_EFF_JA_CFG_DEBUG_TYPE,
			       sizeof(ServoEffJACfgDebug)) {};
  void update(CMS *);
  //! debug level
  int debug;
};

// SERVO_EFF_JA_SET
//////////////////////////////////////////////////////

class ServoEffJASet : public RCS_STAT_MSG {
public:
  ServoEffJASet():RCS_STAT_MSG(SERVO_EFF_JA_SET_TYPE, sizeof(ServoEffJASet)) {};
  void update(CMS *);
  //! cycle time of control loop
  double cycleTime;
  //! debug level
  int debug;
  //! vehicle id
  int vehicleID;
  //! Contains configuration information on the effectors on the vehicle
  //! variable `effector_length`  is the number of effectors
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(ServoEffJAConfig, effector, SERVO_EFF_JA_EFFECTOR_MAX);    

};

extern int servoEffJA_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * servoEffJA_symbol_lookup(long type);
extern const char *servoEffJA_enum_ServoEffJAOpcodeType_symbol_lookup(long v);
#endif // SERVO_EFF_JA_HH
