/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file servoMisJA.hh

  \brief NML declarations for status of mission payloads.

  The mission servo level is responsible for pointing the mission package
  carrying items like sensors. It primarily runs the ServoMisJACmdMove
  command. Given a \c PM_POSE \b goal, the servo level tries to move
  the mission package immediately to the goal. The prim level is responsible
  for any accel/decel control.

  The mission level also maintains its transformation to vehicle
  coordinates. This is necessary as part of the overall transformation of
  mission package sensor coordinates to world coordinates.
  
  \code CVS Status:
  $Author: proctor $
  $Revision: 1.7 $   $
  $Date: 2008/12/02 14:58:13 $
  \endcode

  \author Chris Scrapper
*/

#ifndef SERVO_MIS_JA_HH
#define SERVO_MIS_JA_HH

#include <rcs.hh>		// RCS_CMD,STAT_MSG, CMS, PM_POSE
#include "moastNmlOffsets.hh"	// SERVO_MIS_BASE
#include "moastTypes.hh"	// MOAST_WORLD_NAME_LEN

#define SERVO_MIS_JA_CMD_NAME "servoMisJACmd"
#define SERVO_MIS_JA_STAT_NAME "servoMisJAStat"
#define SERVO_MIS_JA_CFG_NAME "servoMisJACfg"
#define SERVO_MIS_JA_SET_NAME "servoMisJASet"

#define SERVO_MIS_JA_CMD_BASE        (SERVO_MIS_JA_BASE)
#define SERVO_MIS_JA_STAT_BASE       (SERVO_MIS_JA_BASE + 100)
#define SERVO_MIS_JA_CFG_BASE        (SERVO_MIS_JA_BASE + 200)
#define SERVO_MIS_JA_SET_BASE        (SERVO_MIS_JA_BASE + 300)

#define SERVO_MIS_JA_CMD_INIT_TYPE      (SERVO_MIS_JA_CMD_BASE + 1)
#define SERVO_MIS_JA_CMD_ABORT_TYPE     (SERVO_MIS_JA_CMD_BASE + 2)
#define SERVO_MIS_JA_CMD_HALT_TYPE      (SERVO_MIS_JA_CMD_BASE + 3)
#define SERVO_MIS_JA_CMD_SHUTDOWN_TYPE  (SERVO_MIS_JA_CMD_BASE + 4)
#define SERVO_MIS_JA_CMD_MOVE_TYPE  (SERVO_MIS_JA_CMD_BASE + 5)

#define SERVO_MIS_JA_STAT_TYPE       (SERVO_MIS_JA_STAT_BASE + 1)

#define SERVO_MIS_JA_CFG_CYCLE_TIME_TYPE (SERVO_MIS_JA_CFG_BASE + 1)
#define SERVO_MIS_JA_CFG_DEBUG_TYPE      (SERVO_MIS_JA_CFG_BASE + 2)

#define SERVO_MIS_JA_SET_TYPE        (SERVO_MIS_JA_SET_BASE + 1)

#define SERVO_MIS_JA_DEBUG_ALL 0xFFFFFFFF

#define SERVO_MIS_JA_MIS_PKG_MAX 4
#define SERVO_MIS_JA_LINK_MAX 12

//////////////////////////////////////////////
//! SERVO_MIS_JA Link structures
//////////////////////////////////////////////
typedef enum SERVO_MIS_JA_JOINT_TYPE
  {
    SERVO_MIS_JA_INVALID_JOINT_TYPE = -1,
    SERVO_MIS_JA_PRISMATIC_JOINT_TYPE,
    SERVO_MIS_JA_REVOLUTE_JOINT_TYPE,
    SERVO_MIS_JA_SCISSOR_JOINT_TYPE
  }ServoMisJAJointType;  

typedef enum SERVO_MIS_JA_LINK_CMD_TYPE
  {
    SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE,
    SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE,
    SERVO_MIS_JA_LINK_CMD_TORQUE_TYPE
  }ServoMisJALinkCmdType;

typedef struct SERVO_MIS_JA_LINK_STAT
{
  // Link ID
  int linkID;
  /*! The value of the link's degree of freedom. If the joint is revolute,
    then it is an angle. If the joint is prismatic, then it is a distance. */
  double jointVal;
  //! Torque on joint
  double torque;

}ServoMisJALinkStat;

typedef struct SERVO_MIS_JA_LINK_CMD
{
  int linkID;
  double value;
  ServoMisJALinkCmdType type;
}ServoMisJALinkCmd;

typedef struct SERVO_MIS_JA_LINK_SET 
{
  // Link ID
  int linkID;
  // ID of parent link, if = -1, then vehicle
  int parentLinkID;
  // Joint Type
  ServoMisJAJointType jType;
  // Maximun angular velocity [rads/s]
  double maxVelocity;
  // Maximun torque
  double maxTorque;
  //! If minRange > MaxRange, then joint is not bounded 
  // Maximum Range 
  double maxRange;
  // Minimum Range, 
  double minRange;
  /*! Location with respect to parent, if parentLinkID == -1, 
   *  then the location with respect to vehicle */
  PM_CARTESIAN loc;
  /*! Rotation with respect to parent, if parentLinkID == -1, 
   *  then the orientation with respect to vehicle */
  PM_RPY rot;
}ServoMisJALinkSet;


//////////////////////////////////////////////
//! SERVO_MIS_JA_CMD_XXX
//////////////////////////////////////////////

class ServoMisJACmdInit:public RCS_CMD_MSG {
public:
  ServoMisJACmdInit():RCS_CMD_MSG(SERVO_MIS_JA_CMD_INIT_TYPE,
				   sizeof(ServoMisJACmdInit)) {};
  void update(CMS *);
};

class ServoMisJACmdAbort:public RCS_CMD_MSG {
public:
  ServoMisJACmdAbort():RCS_CMD_MSG(SERVO_MIS_JA_CMD_ABORT_TYPE,
				    sizeof(ServoMisJACmdAbort)) {};
  void update(CMS *);
};

class ServoMisJACmdHalt:public RCS_CMD_MSG {
public:
  ServoMisJACmdHalt():RCS_CMD_MSG(SERVO_MIS_JA_CMD_HALT_TYPE,
				   sizeof(ServoMisJACmdHalt)) {};
  void update(CMS *);
};

class ServoMisJACmdShutdown:public RCS_CMD_MSG {
public:
  ServoMisJACmdShutdown():RCS_CMD_MSG(SERVO_MIS_JA_CMD_SHUTDOWN_TYPE,
				   sizeof(ServoMisJACmdShutdown)) {};
  void update(CMS *);
};

//! Array containing the set joint commands for all the links in mission package
class ServoMisJACmdMove:public RCS_CMD_MSG {
public:
  ServoMisJACmdMove():RCS_CMD_MSG(SERVO_MIS_JA_CMD_MOVE_TYPE,
				   sizeof(ServoMisJACmdMove)) {};
  void update(CMS *);
  //! linkSetCmd_length == the number of links in mission package
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(ServoMisJALinkCmd, linkCmd,   SERVO_MIS_JA_LINK_MAX);
};

//////////////////////////////////////////////
//! SERVO_MIS_JA_STAT
//////////////////////////////////////////////


class ServoMisJAStat : public RCS_STAT_MSG_V2 {
public:
  ServoMisJAStat() : RCS_STAT_MSG_V2(SERVO_MIS_JA_STAT_TYPE, sizeof(ServoMisJAStat)) {};
  void update(CMS *);
  //! Error code, -1 if no error
  int error;
  //! Time stamp of when data was collect
  double time;
  //! Increment very cycle
  int heartbeat;
  //! Array containing the status of all the links in mission package
  //! linkStat_length == the number of links in mission package
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(ServoMisJALinkStat, linkStat,   SERVO_MIS_JA_LINK_MAX);
};

//////////////////////////////////////////////
//! SERVO_MIS_JA_CFG_XXX
//////////////////////////////////////////////

//!ServoMisJACfgCycleTime sets the controller's cycle time
class ServoMisJACfgCycleTime:public RCS_CMD_MSG {
public:
  ServoMisJACfgCycleTime():RCS_CMD_MSG(SERVO_MIS_JA_CFG_CYCLE_TIME_TYPE,
			       sizeof(ServoMisJACfgCycleTime)) {};
  void update(CMS *);

  //! period in seconds
  double cycleTime;
};

//! ServoMisJACfgDebug sets the controller's debug level
class ServoMisJACfgDebug:public RCS_CMD_MSG {
public:
  ServoMisJACfgDebug():RCS_CMD_MSG(SERVO_MIS_JA_CFG_DEBUG_TYPE,
			       sizeof(ServoMisJACfgDebug)) {};
  void update(CMS *);

  //! debug level
  int debug;
};


//////////////////////////////////////////////////////
// SERVO_MIS_JA_SET
//////////////////////////////////////////////////////

class ServoMisJASet : public RCS_STAT_MSG {
public:
  ServoMisJASet():RCS_STAT_MSG(SERVO_MIS_JA_SET_TYPE, sizeof(ServoMisJASet)) {};
  void update(CMS *);
  //! controller's cycle time period, [s]
  double cycleTime;
  //! debug level
  int debug;
  //! Name of world, if real world then worldName = "live"
  char worldName[MOAST_WORLD_NAME_LEN];
  //! Link settings for mission package 
  //! linkSet_length == number of links in mission package
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(ServoMisJALinkSet, linkSet, SERVO_MIS_JA_LINK_MAX);  
};

extern int servoMisJA_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * servoMisJA_symbol_lookup(long type);
extern const char * servoMisJA_enum_ServoMisJAJointType_symbol_lookup(long v);
extern const char *servoMisJA_enum_ServoMisJALinkCmdType_symbol_lookup(long v);


#endif // SERVO_MIS_JA_HH

/*
  Modification history:

  $Log: servoMisJA.hh,v $
  Revision 1.7  2008/12/02 14:58:13  proctor
  Made an obtuse comment more descriptive

  Revision 1.6  2007/10/18 17:53:45  cj_scrapper
  Adding Scissor Joint

  Revision 1.5  2007/10/18 17:42:43  cj_scrapper
  Adding Scissor joint

  Revision 1.4  2007/03/12 17:32:07  cj_scrapper
  Added new servo Mission package commands

  Revision 1.3  2006/06/21 19:38:13  proctor
  Added support for Java, which involved including some left-out headers
  like moastNmlOffsets.hh and moastTypes.hh

  Revision 1.2  2005/11/15 17:26:50  cj_scrapper
  Documentation Error

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.12  2005/08/10 21:05:12  proctor
  Changed "isVel" to "moveType" in ServoMisJACmdMove

  Revision 1.11  2005/08/05 16:33:03  scrapper
  Added min range variable to set msg.

  Revision 1.10  2005/08/05 14:10:30  scrapper
  Commenting

  Revision 1.9  2005/08/05 13:34:04  scrapper
  Added MOAST_WORLD_NAME_LEN to moastTypes.  Added worldName to all servo 'set' messages.

  Revision 1.8  2005/08/04 15:35:13  scrapper
  Switched data in SP and MIS.  Added maxRange to mis set.

  Revision 1.7  2005/08/04 14:47:05  proctor
  Added stat.mountToVeh

  Revision 1.6  2005/08/02 20:49:17  scrapper
  Added flag to MisCmdMov that indicates if the data in the PM_POSE is either a pose or a set of velocities.

  Revision 1.5  2005/07/20 21:52:31  proctor
  servoSP,Mis,MobJA.hh: added debug cfg
  primMobJA.hh: added 'isArc' flag to ArcSegment

  Revision 1.4  2005/07/13 19:47:38  proctor
  Added baseToVeh mission package mounting transform

  Revision 1.3  2005/07/12 21:42:42  proctor
  Added nested headers; baseToVeh in servoMisJA.hh; sensorToBase in
  servoSP.hh

  Revision 1.2  2005/06/08 14:29:54  proctor
  Added vehMobPLMain name change

  Revision 1.1  2005/06/08 12:47:41  proctor
  Name change per Mis,MobJA naming convention

  Revision 1.4  2005/06/07 18:41:55  proctor
  Added cycle time, debug config

  Revision 1.3  2005/06/07 18:01:26  proctor
  Added primMis.hh; changed servoMIS to servoMisJA

  Revision 1.2  2005/06/06 15:30:08  stephen
  Fixed mission channel defines.

  Revision 1.1  2005/06/06 15:23:06  stephen
  Moved file.

  Revision 1.3  2005/05/17 14:53:16  stephen
  Fixed parameter array to include a type.

  Revision 1.2  2005/05/15 13:47:20  stephen
  Added commands, and broke up status into status and set channels

  Revision 1.1  2005/05/10 17:34:37  proctor
  Changed vehicle to servoPL, sensor to servoSP

  Revision 1.2  2005/05/05 20:08:27  proctor
  Worked around code generator bug by using single-line bodies for the
  constructor, {}, instead of breaking the braces across a line.

  Revision 1.1  2005/05/05 18:04:28  proctor
  Changed command to cmd, status to stat, etc.

  Revision 1.1  2005/05/05 15:56:12  proctor
  Changed naming convention from e.g. NAV_DATA_EXT to NavDataExt

*/
