/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I.

*/

/*!
  \file amSPMap.cc

  \brief Contains code implementing anSPCell and AmSPMap

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.19 $
  $Date: 2008/02/12 00:06:50 $
  \endcode

  \author Tom Kramer
  \date   2005-10-05
  The original author was Adam Beytin
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>	   // round()
#include <string.h>	   // strlen()
#include <fcntl.h>	   // open()
#include <sys/stat.h>      // S_IWUSR, etc.
#include <unistd.h>        // close()
#include "config.h"        // make HAVE_CYGWIN flag work
#include "amSP.hh"         // AMSP_NOT_OBSTACLE, etc.
#include "amSPMap.hh"      // header file
#include "../common/scrollingMap.cc" // code for scrolling map using AmSPCell

/*

The probs array gives the probability represented by each 7-digit
base 3 number. It is initialized by calcProbs.

The hitsYes array gives the value that hits will be when the existing
value is the array index and the latest viewing of a cell is a hit.
This could be calculated as
  0 <= n <= 126 : ((2 * n) + 1)
127 <= n <= 190 : ((2 * n) + 1) - 128)
191 <= n <= 254 : ((2 * n) + 1) - 256)
but using a constant array is faster.

The hitsNo array gives the value that hits will be when the existing
value is the array index and the latest viewing of a cell is not a hit.
This could be calculated by adding 1 to the values in the hitsYes
array, but using a constant array is faster.

*/

static double probs[255];

/***********************************************************************/

static const char hitsYes[255] = {
/* index  value    old num    new num */
/*   0 */    1, /* 0000000 -> 0000001 */
/*   1 */    3, /* 0000001 -> 0000011 */
/*   2 */    5, /* 0000002 -> 0000021 */
/*   3 */    7, /* 0000011 -> 0000111 */
/*   4 */    9, /* 0000012 -> 0000121 */
/*   5 */   11, /* 0000021 -> 0000211 */
/*   6 */   13, /* 0000022 -> 0000221 */
/*   7 */   15, /* 0000111 -> 0001111 */
/*   8 */   17, /* 0000112 -> 0001121 */
/*   9 */   19, /* 0000121 -> 0001211 */
/*  10 */   21, /* 0000122 -> 0001221 */
/*  11 */   23, /* 0000211 -> 0002111 */
/*  12 */   25, /* 0000212 -> 0002121 */
/*  13 */   27, /* 0000221 -> 0002211 */
/*  14 */   29, /* 0000222 -> 0002221 */
/*  15 */   31, /* 0001111 -> 0011111 */
/*  16 */   33, /* 0001112 -> 0011121 */
/*  17 */   35, /* 0001121 -> 0011211 */
/*  18 */   37, /* 0001122 -> 0011221 */
/*  19 */   39, /* 0001211 -> 0012111 */
/*  20 */   41, /* 0001212 -> 0012121 */
/*  21 */   43, /* 0001221 -> 0012211 */
/*  22 */   45, /* 0001222 -> 0012221 */
/*  23 */   47, /* 0002111 -> 0021111 */
/*  24 */   49, /* 0002112 -> 0021121 */
/*  25 */   51, /* 0002121 -> 0021211 */
/*  26 */   53, /* 0002122 -> 0021221 */
/*  27 */   55, /* 0002211 -> 0022111 */
/*  28 */   57, /* 0002212 -> 0022121 */
/*  29 */   59, /* 0002221 -> 0022211 */
/*  30 */   61, /* 0002222 -> 0022221 */
/*  31 */   63, /* 0011111 -> 0111111 */
/*  32 */   65, /* 0011112 -> 0111121 */
/*  33 */   67, /* 0011121 -> 0111211 */
/*  34 */   69, /* 0011122 -> 0111221 */
/*  35 */   71, /* 0011211 -> 0112111 */
/*  36 */   73, /* 0011212 -> 0112121 */
/*  37 */   75, /* 0011221 -> 0112211 */
/*  38 */   77, /* 0011222 -> 0112221 */
/*  39 */   79, /* 0012111 -> 0121111 */
/*  40 */   81, /* 0012112 -> 0121121 */
/*  41 */   83, /* 0012121 -> 0121211 */
/*  42 */   85, /* 0012122 -> 0121221 */
/*  43 */   87, /* 0012211 -> 0122111 */
/*  44 */   89, /* 0012212 -> 0122121 */
/*  45 */   91, /* 0012221 -> 0122211 */
/*  46 */   93, /* 0012222 -> 0122221 */
/*  47 */   95, /* 0021111 -> 0211111 */
/*  48 */   97, /* 0021112 -> 0211121 */
/*  49 */   99, /* 0021121 -> 0211211 */
/*  50 */  101, /* 0021122 -> 0211221 */
/*  51 */  103, /* 0021211 -> 0212111 */
/*  52 */  105, /* 0021212 -> 0212121 */
/*  53 */  107, /* 0021221 -> 0212211 */
/*  54 */  109, /* 0021222 -> 0212221 */
/*  55 */  111, /* 0022111 -> 0221111 */
/*  56 */  113, /* 0022112 -> 0221121 */
/*  57 */  115, /* 0022121 -> 0221211 */
/*  58 */  117, /* 0022122 -> 0221221 */
/*  59 */  119, /* 0022211 -> 0222111 */
/*  60 */  121, /* 0022212 -> 0222121 */
/*  61 */  123, /* 0022221 -> 0222211 */
/*  62 */  125, /* 0022222 -> 0222221 */
/*  63 */  127, /* 0111111 -> 1111111 */
/*  64 */  129, /* 0111112 -> 1111121 */
/*  65 */  131, /* 0111121 -> 1111211 */
/*  66 */  133, /* 0111122 -> 1111221 */
/*  67 */  135, /* 0111211 -> 1112111 */
/*  68 */  137, /* 0111212 -> 1112121 */
/*  69 */  139, /* 0111221 -> 1112211 */
/*  70 */  141, /* 0111222 -> 1112221 */
/*  71 */  143, /* 0112111 -> 1121111 */
/*  72 */  145, /* 0112112 -> 1121121 */
/*  73 */  147, /* 0112121 -> 1121211 */
/*  74 */  149, /* 0112122 -> 1121221 */
/*  75 */  151, /* 0112211 -> 1122111 */
/*  76 */  153, /* 0112212 -> 1122121 */
/*  77 */  155, /* 0112221 -> 1122211 */
/*  78 */  157, /* 0112222 -> 1122221 */
/*  79 */  159, /* 0121111 -> 1211111 */
/*  80 */  161, /* 0121112 -> 1211121 */
/*  81 */  163, /* 0121121 -> 1211211 */
/*  82 */  165, /* 0121122 -> 1211221 */
/*  83 */  167, /* 0121211 -> 1212111 */
/*  84 */  169, /* 0121212 -> 1212121 */
/*  85 */  171, /* 0121221 -> 1212211 */
/*  86 */  173, /* 0121222 -> 1212221 */
/*  87 */  175, /* 0122111 -> 1221111 */
/*  88 */  177, /* 0122112 -> 1221121 */
/*  89 */  179, /* 0122121 -> 1221211 */
/*  90 */  181, /* 0122122 -> 1221221 */
/*  91 */  183, /* 0122211 -> 1222111 */
/*  92 */  185, /* 0122212 -> 1222121 */
/*  93 */  187, /* 0122221 -> 1222211 */
/*  94 */  189, /* 0122222 -> 1222221 */
/*  95 */  191, /* 0211111 -> 2111111 */
/*  96 */  193, /* 0211112 -> 2111121 */
/*  97 */  195, /* 0211121 -> 2111211 */
/*  98 */  197, /* 0211122 -> 2111221 */
/*  99 */  199, /* 0211211 -> 2112111 */
/* 100 */  201, /* 0211212 -> 2112121 */
/* 101 */  203, /* 0211221 -> 2112211 */
/* 102 */  205, /* 0211222 -> 2112221 */
/* 103 */  207, /* 0212111 -> 2121111 */
/* 104 */  209, /* 0212112 -> 2121121 */
/* 105 */  211, /* 0212121 -> 2121211 */
/* 106 */  213, /* 0212122 -> 2121221 */
/* 107 */  215, /* 0212211 -> 2122111 */
/* 108 */  217, /* 0212212 -> 2122121 */
/* 109 */  219, /* 0212221 -> 2122211 */
/* 110 */  221, /* 0212222 -> 2122221 */
/* 111 */  223, /* 0221111 -> 2211111 */
/* 112 */  225, /* 0221112 -> 2211121 */
/* 113 */  227, /* 0221121 -> 2211211 */
/* 114 */  229, /* 0221122 -> 2211221 */
/* 115 */  231, /* 0221211 -> 2212111 */
/* 116 */  233, /* 0221212 -> 2212121 */
/* 117 */  235, /* 0221221 -> 2212211 */
/* 118 */  237, /* 0221222 -> 2212221 */
/* 119 */  239, /* 0222111 -> 2221111 */
/* 120 */  241, /* 0222112 -> 2221121 */
/* 121 */  243, /* 0222121 -> 2221211 */
/* 122 */  245, /* 0222122 -> 2221221 */
/* 123 */  247, /* 0222211 -> 2222111 */
/* 124 */  249, /* 0222212 -> 2222121 */
/* 125 */  251, /* 0222221 -> 2222211 */
/* 126 */  253, /* 0222222 -> 2222221 */
/* 127 */  127, /* 1111111 -> 1111111 */
/* 128 */  129, /* 1111112 -> 1111121 */
/* 129 */  131, /* 1111121 -> 1111211 */
/* 130 */  133, /* 1111122 -> 1111221 */
/* 131 */  135, /* 1111211 -> 1112111 */
/* 132 */  137, /* 1111212 -> 1112121 */
/* 133 */  139, /* 1111221 -> 1112211 */
/* 134 */  141, /* 1111222 -> 1112221 */
/* 135 */  143, /* 1112111 -> 1121111 */
/* 136 */  145, /* 1112112 -> 1121121 */
/* 137 */  147, /* 1112121 -> 1121211 */
/* 138 */  149, /* 1112122 -> 1121221 */
/* 139 */  151, /* 1112211 -> 1122111 */
/* 140 */  153, /* 1112212 -> 1122121 */
/* 141 */  155, /* 1112221 -> 1122211 */
/* 142 */  157, /* 1112222 -> 1122221 */
/* 143 */  159, /* 1121111 -> 1211111 */
/* 144 */  161, /* 1121112 -> 1211121 */
/* 145 */  163, /* 1121121 -> 1211211 */
/* 146 */  165, /* 1121122 -> 1211221 */
/* 147 */  167, /* 1121211 -> 1212111 */
/* 148 */  169, /* 1121212 -> 1212121 */
/* 149 */  171, /* 1121221 -> 1212211 */
/* 150 */  173, /* 1121222 */
/* 151 */  175, /* 1122111 */
/* 152 */  177, /* 1122112 */
/* 153 */  179, /* 1122121 */
/* 154 */  181, /* 1122122 */
/* 155 */  183, /* 1122211 */
/* 156 */  185, /* 1122212 */
/* 157 */  187, /* 1122221 */
/* 158 */  189, /* 1122222 */
/* 159 */  191, /* 1211111 -> 2111111 */
/* 160 */  193, /* 1211112 -> 2111121 */
/* 161 */  195, /* 1211121 */
/* 162 */  197, /* 1211122 */
/* 163 */  199, /* 1211211 */
/* 164 */  201, /* 1211212 */
/* 165 */  203, /* 1211221 */
/* 166 */  205, /* 1211222 */
/* 167 */  207, /* 1212111 */
/* 168 */  209, /* 1212112 */
/* 169 */  211, /* 1212121 */
/* 170 */  213, /* 1212122 */
/* 171 */  215, /* 1212211 -> 2122111 */
/* 172 */  217, /* 1212212 */
/* 173 */  219, /* 1212221 */
/* 174 */  221, /* 1212222 */
/* 175 */  223, /* 1221111 */
/* 176 */  225, /* 1221112 */
/* 177 */  227, /* 1221121 */
/* 178 */  229, /* 1221122 */
/* 179 */  231, /* 1221211 */
/* 180 */  233, /* 1221212 */
/* 181 */  235, /* 1221221 -> 2212211 */
/* 182 */  237, /* 1221222 */
/* 183 */  239, /* 1222111 */
/* 184 */  241, /* 1222112 */
/* 185 */  243, /* 1222121 */
/* 186 */  245, /* 1222122 */
/* 187 */  247, /* 1222211 */
/* 188 */  249, /* 1222212 */
/* 189 */  251, /* 1222221 -> 2222211 */
/* 190 */  253, /* 1222222 -> 2222221 */
/* 191 */  127, /* 2111111 -> 1111111 */
/* 192 */  129, /* 2111112 -> 1111121 */
/* 193 */  131, /* 2111121 */
/* 194 */  133, /* 2111122 */
/* 195 */  135, /* 2111211 */
/* 196 */  137, /* 2111212 */
/* 197 */  139, /* 2111221 */
/* 198 */  141, /* 2111222 */
/* 199 */  143, /* 2112111 */
/* 200 */  145, /* 2112112 */
/* 201 */  147, /* 2112121 */
/* 202 */  149, /* 2112122 */
/* 203 */  151, /* 2112211 */
/* 204 */  153, /* 2112212 -> 1122121 */
/* 205 */  155, /* 2112221 -> 1122211 */
/* 206 */  157, /* 2112222 -> 1122221 */
/* 207 */  159, /* 2121111 -> 1211111 */
/* 208 */  161, /* 2121112 -> 1211121 */
/* 209 */  163, /* 2121121 -> 1211211 */
/* 210 */  165, /* 2121122 -> 1211221 */
/* 211 */  167, /* 2121211 -> 1212111 */
/* 212 */  169, /* 2121212 -> 1212121 */
/* 213 */  171, /* 2121221 */
/* 214 */  173, /* 2121222 */
/* 215 */  175, /* 2122111 */
/* 216 */  177, /* 2122112 */
/* 217 */  179, /* 2122121 */
/* 218 */  181, /* 2122122 */
/* 219 */  183, /* 2122211 */
/* 220 */  185, /* 2122212 */
/* 221 */  187, /* 2122221 */
/* 222 */  189, /* 2122222 -> 1222221 */
/* 223 */  191, /* 2211111 */
/* 224 */  193, /* 2211112 */
/* 225 */  195, /* 2211121 */
/* 226 */  197, /* 2211122 */
/* 227 */  199, /* 2211211 */
/* 228 */  201, /* 2211212 */
/* 229 */  203, /* 2211221 */
/* 230 */  205, /* 2211222 */
/* 231 */  207, /* 2212111 */
/* 232 */  209, /* 2212112 */
/* 233 */  211, /* 2212121 -> 2121211*/
/* 234 */  213, /* 2212122 */
/* 235 */  215, /* 2212211 */
/* 236 */  217, /* 2212212 */
/* 237 */  219, /* 2212221 */
/* 238 */  221, /* 2212222 */
/* 239 */  223, /* 2221111 */
/* 240 */  225, /* 2221112 */
/* 241 */  227, /* 2221121 */
/* 242 */  229, /* 2221122 */
/* 243 */  231, /* 2221211 */
/* 244 */  233, /* 2221212 */
/* 245 */  235, /* 2221221 */
/* 246 */  237, /* 2221222 -> 2212221 */
/* 247 */  239, /* 2222111 */
/* 248 */  241, /* 2222112 */
/* 249 */  243, /* 2222121 */
/* 250 */  245, /* 2222122 */
/* 251 */  247, /* 2222211 */
/* 252 */  249, /* 2222212 */
/* 253 */  251, /* 2222221 */
/* 254 */  253  /* 2222222 -> 2222221 */
};

static const char hitsNo[255] = {
/* index  value    old num    new num */
/*   0 */    2, /* 0000000 -> 0000002 */
/*   1 */    4, /* 0000001 -> 0000012 */
/*   2 */    6, /* 0000002 -> 0000022 */
/*   3 */    8, /* 0000011 */
/*   4 */   10, /* 0000012 */
/*   5 */   12, /* 0000021 */
/*   6 */   14, /* 0000022 */
/*   7 */   16, /* 0000111 */
/*   8 */   18, /* 0000112 */
/*   9 */   20, /* 0000121 */
/*  10 */   22, /* 0000122 */
/*  11 */   24, /* 0000211 */
/*  12 */   26, /* 0000212 */
/*  13 */   28, /* 0000221 */
/*  14 */   30, /* 0000222 */
/*  15 */   32, /* 0001111 */
/*  16 */   34, /* 0001112 */
/*  17 */   36, /* 0001121 */
/*  18 */   38, /* 0001122 */
/*  19 */   40, /* 0001211 -> 0012112 */
/*  20 */   42, /* 0001212 */
/*  21 */   44, /* 0001221 */
/*  22 */   46, /* 0001222 */
/*  23 */   48, /* 0002111 */
/*  24 */   50, /* 0002112 */
/*  25 */   52, /* 0002121 */
/*  26 */   54, /* 0002122 */
/*  27 */   56, /* 0002211 */
/*  28 */   58, /* 0002212 */
/*  29 */   60, /* 0002221 */
/*  30 */   62, /* 0002222 */
/*  31 */   64, /* 0011111 */
/*  32 */   66, /* 0011112 */
/*  33 */   68, /* 0011121 */
/*  34 */   70, /* 0011122 */
/*  35 */   72, /* 0011211 */
/*  36 */   74, /* 0011212 */
/*  37 */   76, /* 0011221 */
/*  38 */   78, /* 0011222 */
/*  39 */   80, /* 0012111 -> 0121112*/
/*  40 */   82, /* 0012112 */
/*  41 */   84, /* 0012121 */
/*  42 */   86, /* 0012122 */
/*  43 */   88, /* 0012211 */
/*  44 */   90, /* 0012212 */
/*  45 */   92, /* 0012221 */
/*  46 */   94, /* 0012222 */
/*  47 */   96, /* 0021111 */
/*  48 */   98, /* 0021112 */
/*  49 */  100, /* 0021121 */
/*  50 */  102, /* 0021122 */
/*  51 */  104, /* 0021211 */
/*  52 */  106, /* 0021212 */
/*  53 */  108, /* 0021221 */
/*  54 */  110, /* 0021222 */
/*  55 */  112, /* 0022111 */
/*  56 */  114, /* 0022112 */
/*  57 */  116, /* 0022121 */
/*  58 */  118, /* 0022122 */
/*  59 */  120, /* 0022211 -> 0222112 */
/*  60 */  122, /* 0022212 */
/*  61 */  124, /* 0022221 */
/*  62 */  126, /* 0022222 */
/*  63 */  128, /* 0111111 */
/*  64 */  130, /* 0111112 */
/*  65 */  132, /* 0111121 */
/*  66 */  134, /* 0111122 */
/*  67 */  136, /* 0111211 */
/*  68 */  138, /* 0111212 */
/*  69 */  140, /* 0111221 */
/*  70 */  142, /* 0111222 */
/*  71 */  144, /* 0112111 */
/*  72 */  146, /* 0112112 */
/*  73 */  148, /* 0112121 */
/*  74 */  150, /* 0112122 */
/*  75 */  152, /* 0112211 */
/*  76 */  154, /* 0112212 */
/*  77 */  156, /* 0112221 */
/*  78 */  158, /* 0112222 */
/*  79 */  160, /* 0121111 -> 1211112 */
/*  80 */  162, /* 0121112 */
/*  81 */  164, /* 0121121 */
/*  82 */  166, /* 0121122 */
/*  83 */  168, /* 0121211 */
/*  84 */  170, /* 0121212 */
/*  85 */  172, /* 0121221 */
/*  86 */  174, /* 0121222 */
/*  87 */  176, /* 0122111 */
/*  88 */  178, /* 0122112 */
/*  89 */  180, /* 0122121 */
/*  90 */  182, /* 0122122 */
/*  91 */  184, /* 0122211 */
/*  92 */  186, /* 0122212 */
/*  93 */  188, /* 0122221 */
/*  94 */  190, /* 0122222 */
/*  95 */  192, /* 0211111 */
/*  96 */  194, /* 0211112 */
/*  97 */  196, /* 0211121 */
/*  98 */  198, /* 0211122 */
/*  99 */  200, /* 0211211 -> 2112112 */
/* 100 */  202, /* 0211212 */
/* 101 */  204, /* 0211221 */
/* 102 */  206, /* 0211222 */
/* 103 */  208, /* 0212111 */
/* 104 */  210, /* 0212112 */
/* 105 */  212, /* 0212121 */
/* 106 */  214, /* 0212122 */
/* 107 */  216, /* 0212211 */
/* 108 */  218, /* 0212212 */
/* 109 */  220, /* 0212221 */
/* 110 */  222, /* 0212222 */
/* 111 */  224, /* 0221111 */
/* 112 */  226, /* 0221112 */
/* 113 */  228, /* 0221121 */
/* 114 */  230, /* 0221122 */
/* 115 */  232, /* 0221211 */
/* 116 */  234, /* 0221212 */
/* 117 */  236, /* 0221221 */
/* 118 */  238, /* 0221222 */
/* 119 */  240, /* 0222111 -> 2221112 */
/* 120 */  242, /* 0222112 */
/* 121 */  244, /* 0222121 */
/* 122 */  246, /* 0222122 */
/* 123 */  248, /* 0222211 */
/* 124 */  250, /* 0222212 */
/* 125 */  252, /* 0222221 */
/* 126 */  254, /* 0222222 -> 2222222 */
/* 127 */  128, /* 1111111 -> 1111112 */
/* 128 */  130, /* 1111112 */
/* 129 */  132, /* 1111121 */
/* 130 */  134, /* 1111122 */
/* 131 */  136, /* 1111211 */
/* 132 */  138, /* 1111212 */
/* 133 */  140, /* 1111221 */
/* 134 */  142, /* 1111222 */
/* 135 */  144, /* 1112111 */
/* 136 */  146, /* 1112112 */
/* 137 */  148, /* 1112121 */
/* 138 */  150, /* 1112122 */
/* 139 */  152, /* 1112211 -> 1122112*/
/* 140 */  154, /* 1112212 */
/* 141 */  156, /* 1112221 */
/* 142 */  158, /* 1112222 */
/* 143 */  160, /* 1121111 */
/* 144 */  162, /* 1121112 */
/* 145 */  164, /* 1121121 */
/* 146 */  166, /* 1121122 */
/* 147 */  168, /* 1121211 */
/* 148 */  170, /* 1121212 */
/* 149 */  172, /* 1121221 */
/* 150 */  174, /* 1121222 */
/* 151 */  176, /* 1122111 */
/* 152 */  178, /* 1122112 */
/* 153 */  180, /* 1122121 */
/* 154 */  182, /* 1122122 */
/* 155 */  184, /* 1122211 */
/* 156 */  186, /* 1122212 */
/* 157 */  188, /* 1122221 */
/* 158 */  190, /* 1122222 */
/* 159 */  192, /* 1211111 -> 2111112 */
/* 160 */  194, /* 1211112 */
/* 161 */  196, /* 1211121 */
/* 162 */  198, /* 1211122 */
/* 163 */  200, /* 1211211 */
/* 164 */  202, /* 1211212 */
/* 165 */  204, /* 1211221 */
/* 166 */  206, /* 1211222 */
/* 167 */  208, /* 1212111 */
/* 168 */  210, /* 1212112 */
/* 169 */  212, /* 1212121 */
/* 170 */  214, /* 1212122 */
/* 171 */  216, /* 1212211 */
/* 172 */  218, /* 1212212 */
/* 173 */  220, /* 1212221 */
/* 174 */  222, /* 1212222 */
/* 175 */  224, /* 1221111 */
/* 176 */  226, /* 1221112 */
/* 177 */  228, /* 1221121 */
/* 178 */  230, /* 1221122 */
/* 179 */  232, /* 1221211 -> 2212112 */
/* 180 */  234, /* 1221212 */
/* 181 */  236, /* 1221221 */
/* 182 */  238, /* 1221222 */
/* 183 */  240, /* 1222111 */
/* 184 */  242, /* 1222112 */
/* 185 */  244, /* 1222121 */
/* 186 */  246, /* 1222122 */
/* 187 */  248, /* 1222211 */
/* 188 */  250, /* 1222212 */
/* 189 */  252, /* 1222221 */
/* 190 */  254, /* 1222222 -> 2222222 */
/* 191 */  128, /* 2111111 -> 1111112 */
/* 192 */  130, /* 2111112 */
/* 193 */  132, /* 2111121 */
/* 194 */  134, /* 2111122 */
/* 195 */  136, /* 2111211 */
/* 196 */  138, /* 2111212 */
/* 197 */  140, /* 2111221 */
/* 198 */  142, /* 2111222 */
/* 199 */  144, /* 2112111 -> 1121112 */
/* 200 */  146, /* 2112112 */
/* 201 */  148, /* 2112121 */
/* 202 */  150, /* 2112122 */
/* 203 */  152, /* 2112211 */
/* 204 */  154, /* 2112212 */
/* 205 */  156, /* 2112221 */
/* 206 */  158, /* 2112222 */
/* 207 */  160, /* 2121111 */
/* 208 */  162, /* 2121112 */
/* 209 */  164, /* 2121121 */
/* 210 */  166, /* 2121122 */
/* 211 */  168, /* 2121211 */
/* 212 */  170, /* 2121212 */
/* 213 */  172, /* 2121221 */
/* 214 */  174, /* 2121222 */
/* 215 */  176, /* 2122111 */
/* 216 */  178, /* 2122112 */
/* 217 */  180, /* 2122121 */
/* 218 */  182, /* 2122122 */
/* 219 */  184, /* 2122211 -> 1222112 */
/* 220 */  186, /* 2122212 */
/* 221 */  188, /* 2122221 */
/* 222 */  190, /* 2122222 */
/* 223 */  192, /* 2211111 */
/* 224 */  194, /* 2211112 */
/* 225 */  196, /* 2211121 */
/* 226 */  198, /* 2211122 */
/* 227 */  200, /* 2211211 */
/* 228 */  202, /* 2211212 */
/* 229 */  204, /* 2211221 */
/* 230 */  206, /* 2211222 */
/* 231 */  208, /* 2212111 */
/* 232 */  210, /* 2212112 */
/* 233 */  212, /* 2212121 */
/* 234 */  214, /* 2212122 */
/* 235 */  216, /* 2212211 */
/* 236 */  218, /* 2212212 */
/* 237 */  220, /* 2212221 */
/* 238 */  222, /* 2212222 */
/* 239 */  224, /* 2221111 -> 2211112 */
/* 240 */  226, /* 2221112 */
/* 241 */  228, /* 2221121 */
/* 242 */  230, /* 2221122 */
/* 243 */  232, /* 2221211 */
/* 244 */  234, /* 2221212 */
/* 245 */  236, /* 2221221 */
/* 246 */  238, /* 2221222 */
/* 247 */  240, /* 2222111 */
/* 248 */  242, /* 2222112 */
/* 249 */  244, /* 2222121 */
/* 250 */  246, /* 2222122 */
/* 251 */  248, /* 2222211 */
/* 252 */  250, /* 2222212 */
/* 253 */  252, /* 2222221 */
/* 254 */  254  /* 2222222 -> 2222222*/
};

const double epsilon = 0.2;

/***********************************************************************/

//! calcProbs calculates the lookup table of probabilities
/*
Returned Value: none

Called By: AmSPMap::AmSPMap

An array of seven digits represents what happened the last seven times
we looked in the direction of a cell. A 1 means we saw an obstacle.
A 2 means we did not see an obstacle. A 0 means we have not looked
at the cell that many times yet. For example, the digits array
{1, 2, 1, 1, 0, 0, 0} means we have looked at a cell four times.
The next-to-most recent time we looked, we did not see an obstacle.
The other three times we did. Zeros may occur only at the end of the
digits array.

The increase or decrease in probability is weighted by how recently we
have seen or failed to see an obstacle.  The weighting is done by the
weights array, {8, 5, 3, 2, 1, 0, 0}.  The two zeros at the end might
be changed so that six or seven rather than five points have an
effect. Five points were used in the previous implementation, so
five are used here.

The digits are changed in the outer "for" loop to represent the
allowed values of the array.

To calculate the probability that there is an obstacle on a cell, each
entry in the weights array is multiplied by a factor associated with
the corresponding entry in the digits array, and the products are
added together. If a digit is 0, the associated factor is 0. If the
digit is 1, the associated factor is AMSP_PROBINC. If the digit is 2,
the associated factor is AMSP_PROBDEC.

For example, suppose the digits array is {1, 2, 1, 1, 0, 0, 0}.
Suppose also that AMSP_PROBINC is 0.05 and AMSP_PROBDEC is -0.04
The probability is then given by:

(AMSP_PROBINC * 8) + (AMSP_PROBDEC * 5) + (AMSP_PROBINC * 3) +
                     (AMSP_PROBINC * 2) + (0 * 1) + (0 * 0) + (0 * 0)
= 0.4 - 0.2 + 0.15 + 0.1 + 0.0 + 0.0 +0.0
= 0.45

If the result is greater than 1, it is set to 1. If the result is
negative, it is set to 0.

There is no strong theoretical basis for the values of AMSP_PROBINC,
AMSP_PROBDEC and weights. It may be desirable to change them.

For each valid value of the digits array, this calculates the
probability associated with the value, and stores the probability in
the probs array at the index representing that value.

The order in which the digits array is changed is increasing numerical
order when the digits of the array are used in reverse order to make a
number. The number may be thought of as a base 3 number.

*/

void calcProbs() /* NO ARGUMENTS */
{
  int n;
  int k;
  static const double weights[] = {8, 5, 3, 2, 1, 0, 0};
  int digits[] = {1, 0, 0, 0, 0, 0, 0}; // base 3 number in reverse order
  
  probs[0] = AMSP_PROBINIT;
  for (n = 1; n < 255; n++)
    {
      probs[n] = 0;
      for (k = 0; k < 7; k++)
	{
	  probs[n] = (probs[n] + (weights[k] * 
				  ((digits[k] == 2) ? AMSP_PROBDEC :
				   (digits[k] == 1) ? AMSP_PROBINC : 0)));
	}
      if (probs[n]  > 1.0)
	probs[n]  = 1.0;
      else if (probs[n] < 0.0)
	probs[n]  = 0.0;
      for (k = 0; k < 7; k++)
	{
	  if (digits[k] == 0)
	    {
	      digits[k] = 1;
	      break;
	    }
	  else if (digits[k] == 1)
	    {
	      digits[k] = 2;
	      break;
	    }
	  else // if (digits[k] == 2)
	    digits[k] = 1;
	}
    }
}

/***********************************************************************/

/*! AmSPCell::setCellData sets the data in a cell, using values
    given in the arguments. */
/*
Returned Value: none

Called By: AmSPMap::setMapData

If (1) the cell has a positive range, (2) the input range is larger than
that, and (3) the cell was not seen, the cell is not modified at all.
Otherwise:
1. The height of the cell is set to the given height.
2. The range of the cell may be reset.
3. The hits of the cell are changed.
4. The probability of the cell representing an obstacle is recalculated.

The weird calculation using "% 8" picks out those numbers whose
associated base 3 number ends in 222 (equivalently, the associated
digits array starts with 2, 2, 2). See the hitsYes array to
verify that the calculation does that.

*/

bool AmSPCell::setCellData(  /*   ARGUMENTS                             */
 double heightIn,            /*!< height at which a cell was viewed     */
 double rangeIn,             /*!< distance from which a cell was viewed */
 bool seen)                  /*!< whether an obstacle was seen          */
{
  bool result= false;

  if (range)
    { // range is not already zero.
      if (seen)
	{ // Point was seen. Set cell height and maybe increase cell range
	  height = heightIn;
	  if (rangeIn > (1.11111 * range))
	    { // input range is larger than cell range, so set cell range
	      range = (rangeIn * 0.9);
	    }
	  hits = hitsYes[hits];
          result = fabs(probability-probs[hits]) >= epsilon;
	  probability = probs[hits];
	}
      else
	{ // Point was not seen.
	  if (rangeIn < range)
	    { // Inside cell range
	      if (heightIn < height) // reset cell height if smaller
		height = heightIn;
	      if ((((hits + 2) % 8) == 0) && (hits > 6))
		range = 0; // if not seen 3 times: set cell range 0
	      hits = hitsNo[hits];
              result = fabs(probability-probs[hits]) >= epsilon;
	      probability = probs[hits];
	    }
	}     // Outside cell range. Do nothing.
    }
  else
    { // cell range is zero. Set cell height
      height = heightIn;
      if (seen)
	{ // Point was seen. Set cell range and set last hits digit to 1.
	  range = (rangeIn * 0.9);
	  hits = hitsYes[hits];
	}
      else // Point was not seen. Set last hits digit to 2.
	hits = hitsNo[hits];
      result = fabs(probability-probs[hits]) >= epsilon;
      probability = probs[hits];
    }

  return result;
}

/***********************************************************************/

/*! AmSPMap::AmSPMap constructor taking four arguments creates a map
  based on an initial range, resolution, output map range, and map image
  file name. */
/*
Returned Value: AmSPMap *

Called By: initMap (in amSPMain.cc)

*/

AmSPMap::AmSPMap(           /*   ARGUMENTS                    */
 int rangeIn,               /*!< internal map range           */
 double resolutionIn,       /*!< length of the side of a cell */
 int outRangeIn,            /*!< range of output map          */
 char * mapImageFileNameIn) /*!< name of map image file       */
  : ScrollingMap(rangeIn, resolutionIn)
{
  mapImageFileName = mapImageFileNameIn;
  outRange = outRangeIn;
  outSide = ((outRange * 2) + 1);
  displayLevel = 0;
  outFile = 0;
  buf = NULL;
  calcProbs();
  
  beingRead = false;
}

/***********************************************************************/

//! AmSPMap::~AmSPMap destructor. Deletes allocated memory, closes open files.
/*
Returned Value: none

Called By: (implicitly via delete)
  initMap (in amSPMain.cc)
  main (in amSPMain.cc)

*/

AmSPMap::~AmSPMap() /* NO ARGUMENTS */
{
  char sysStr[200];
  int noWarn; // just to stop compiler warnings

  if (buf)
    delete [] buf;
  if (outFile)
    fclose(outFile);
  sprintf(sysStr, "rm -rf %s\n", mapImageFileName);
  noWarn = system(sysStr);
}

/***********************************************************************/

/*! Method createDisplay always deletes memory allocated to the "buf"
    buffer if the memory exists, closes the outFile if it is open,
    removes the map image file if it exists, and sets displayLevel
    to the level argument. In addition, if the level argument is
    1 or 2, it allocates memory for the "buf" buffer and starts
    imageview (which displays the picture). It would be nice if it
    would also stop imageview when level is zero, but it is not
    doing that. */
/*

Returned Value: bool
  If the buffer cannot be made, this returns false.
  Otherwise, it returns true.

Called By: initMap (in amSPMain.cc)

The map image file name is in mapImageFileName. The file is opened and
closed each time we redraw the map. The file is in ppm format.

This creates a buffer that is used for writing the ppm output
file. First we write everything to the buffer. Then we write the
buffer to the file with one write command. This is faster than writing
repeatedly directly to the file.  Also, with the buffer, the vehicle
image can be written over the map. This is much more difficult to do
if writing directly to a file.

The imageview executable requires a ".ppm" suffix on the file name.

This could be improved by better communication with imageview, so that
imageview would shut itself off when the map display is turned off.
Currently, imageview keeps running when the map display is turned off
but stops updating the display. Communication could be established
using a lock file. The lock file would be written when the display was
turned on and deleted when the display is turned off. Imageview would
check for the existence of the lock file when deciding whether to keep
running. The lock file name would be added to the arguments to
imageview.

*/

bool AmSPMap::createDisplay( /*   ARGUMENTS                                 */
 int level)                  /*!< non-zero to turn display on, zero for off */
{
  char sysStr[200];  
  int noWarn; // just to stop compiler warnings
  
  // clean up
  if(outFile)
    {
      fclose(outFile);
      outFile = 0;
    }
  if (buf != NULL)
    delete [] buf;
  buf = NULL;
  sprintf(sysStr, "rm -rf %s\n", mapImageFileName);
  noWarn = system(sysStr);
  
  if (level)
    {
      buf = new char[3 * getRows() * getCols()];
      if (!buf)
	{
	  rcs_print_error("Could not create AmSPMap display buffer\n");
	  return false;
	}
      sprintf(sysStr, "./imageview -f %s -r 1000 -s 1 &\n", mapImageFileName);
      noWarn = system(sysStr);
    }
  displayLevel = level;
  return true;
}

/***********************************************************************/

/*! AmSPMap::setMapData sets the map cell having the given window grid
    row and col indexes by using data given in the arguments. */
/*
Returned Value: bool
  If row and col are in range, this returns true.
  Otherwise, it returns false.

Called By:
  getSensorData (amSPMain.cc) (commented out)
  setMapDataClear

This may be called by setMapDataClear with row or col out of bounds,
so the bounds check is needed.

*/

bool AmSPMap::setMapData( /*   ARGUMENTS              */
 int row,                 /*!< row index of cell      */
 int col,                 /*!< column index of cell   */
 double height,           /*!< height of point to set */
 double range,            /*!< range of point         */
 bool seen)               /*!< whether point was seen */
{
  MapInfo someInfo;
  if ((row > -1) && (row < getRows()) && (col > -1) && (col < getCols()))
    {
      if(getCellAtIndexes(row, col)->setCellData(height, range, seen))
      {
        someInfo.row = getXCoordAtCell(getCellAtIndexes(row, col));
        someInfo.col = getYCoordAtCell(getCellAtIndexes(row, col));
        someInfo.prob = getCellAtIndexes(row, col)->probability;
        //while(beingRead); // If beingRead is set, block until it is reset to false.
        MapChanges.push_back(someInfo);
        xLow = getXLow();
        yLow = getYLow();
        resolution = getResolution();
      }
      
      return true;
    }
  return false;
}

/***********************************************************************/

/*! AmSPMap::setMapDataCopy sets the map cell having the given window grid
    row and col indexes by copying data from another cell into it. */
/*
Returned Value: bool
  If row and col are in range and the "from" cell isn't NULL,
  this returns true. Otherwise, it returns false.

Called By: not currently called

Data is copied from the "from" cell into the cell at the given
row and column.

*/

bool AmSPMap::setMapDataCopy( /*   ARGUMENTS                         */
 AmSPCell * from,             /*!< cell to copy from                 */
 int row,                     /*!< row index of cell to copy into    */
 int col)                     /*!< column index of cell to copy into */
{
  if ((row > -1) && (row < getRows()) && (col > -1) && (col < getCols())
      && from)
    {
      getCellAtIndexes(row, col)->setCellDataCopy(from);
      return true;
    }
  return false;
}
 
/***********************************************************************/

/*! AmSPMap::setMapDataClear sets the cell whose cartesian X and Y
    coordinates are given, if that cell is in the window. The
    method also clears all the cells in the window on a line from
    the point (xVeh, yVeh) to the point (xIn, yIn).
    We make the assumption that, because we saw through the cells
    on that line, no object is in those cells. */
/*
Returned Value: bool
  If the vehicle is not on the map, this returns false.
  Otherwise, this returns whatever setMapData returns (either
  true or false).

Called By: getSensorData (in amSPMain.cc)

If the vehicle is not on the map, this does nothing.

If the vehicle is on the map, this resets all the unseen map cells
that lie on a line between the vehicle and the point (xIn, yIn). Then,
if (xIn, yIn) is on the map, it resets the map cell containing that
point.

If (xIn, yIn) is not on the map, this finds a point just inside the
map on the line between the vehicle and (xIn, yIn) and uses that point
instead of (xIn, yIn).

The height argument is used only if seen is true.

This is similar to Bresenham's algorithm.

It is not adequate to find endI and endJ and check for
((i == endI) && (j == endJ)). That check may fail because of round-off
error. After the full distance between the two points has been
traveled, the current point can be in the wrong cell.

To prevent walls from disappearing when a line of sight is nearly
parallel to a wall, obstacles are thickened by one cell in the
direction of the line of sight as follows. If the last cell is an
obstacle, one cell beyond the last cell (in the direction being
traveled) is also marked as an obstacle. If the last cell is a
spurious obstacle (from noise) or if it is a real obstacle that moves
away, and the ladar looks in that direction again for a while, both
the cells will cease being obstacles, so marking an extra cell as
obstacle should not degrade performance. If the last cell is a real
persistent obstacle, however, the extra obstacle cell will never be
looked through to see an obstacle beyond it, so it will persist even
if the real obstacle cell does not.

This algorithm has the drawback that where the data is noisy and there
are extensive obstacles such as walls, cells on the other side of the
extensive obstacles may be marked as obstacles as a result of the
noise. These cells will persist as obstacles until (and unless) the
vehicle gets to a position where they can be seen. Cells in direct
view are also apt to be marked as obstacles temporarily, but since
they are in view, the marking as obstacle does not persist.

Step1 is originally the distance in the primary direction of travel
between the vehicle and the side of the cell containing the vehicle.
After that distance has been traveled, step1 is set to the resolution
(i.e. the width of a cell).

The second call to setMapData in the checkCell macro may have the
column or row out of bounds. If that happens, setMapData does not
set anything.

*/

#define checkCell(tI, tJ) if ((dI >= deltaI) && (dJ >= deltaJ))\
  { dist = hypot((xIn - xVeh), (yIn - yVeh));\
    if (seen)\
      { setMapData(i,    j,    height, dist, true);\
        setMapData(i+tI, j+tJ, height, dist, true);}\
    else\
      setMapData(i, j, -vehLoc.z, dist, false);\
    return true;}\
  else\
    setMapData(i, j, -vehLoc.z, 0, false)

#define AMSP_TINY 0.00001

bool AmSPMap::setMapDataClear( /* ARGUMENTS                        */
 double xVeh,   /*!< X coordinate of vehicle when point data taken */
 double yVeh,   /*!< Y coordinate of vehicle when point data taken */
 double xIn,    /*!< X coordinate of point                         */
 double yIn,    /*!< Y coordinate of point                         */
 double height, /*!< height for the the point (xIn, yIn)           */
 bool seen)     /*!< whether point (xIn, yIn) was seen             */
{
  int cols;       // number of columns
  int deltaI;     // amount by which i must change
  int deltaJ;     // amount by which j must change
  double deltaX;  // change in cartesian X from start to end
  double deltaY;  // change in cartesian Y from start to end
  double dist;    // distance from vehicle to observed point
  int dI;         // amount by which i has changed
  int dJ;         // amount by which j has changed
  int i;          // window grid X index of cell
  int j;          // window grid Y index of cell
  double lack;    // distance needed in secondary direction (always positive)
  double res;     // resolution of map
  int rows;       // number of rows
  double slope;   // slope with respect to principal direction
  double step1;   // size of step in primary direction (always positive)
  double step2;   // size of step in secondary direction (always positive)

  if ((xVeh < getXLow()) || (xVeh > getXHi()) ||
      (yVeh < getYLow()) || (yVeh > getYHi()))
    return false;

  // If (xIn, yIn) is off the map, set (xIn, yIn) to the point where the
  // line from (xIn, yIn) to (xVeh, yVeh) enters the map, and set seen
  // to false
  if (yIn < getYLow())
    {
      seen = false;
      xIn = (xIn + ((xVeh - xIn) * ((getYLow() - yIn) / (yVeh - yIn))));
      yIn = (getYLow() + AMSP_TINY);
      if (xIn < getXLow())
	{
	  yIn = (yIn + ((yVeh - yIn) * ((getXLow() - xIn) / (xVeh - xIn))));
	  xIn = (getXLow() + AMSP_TINY);
	}
      else if (xIn > getXHi())
	{
	  yIn = (yIn + ((yVeh - yIn) * ((xIn - getXHi()) / (xIn - xVeh))));
	  xIn = (getXHi() - AMSP_TINY);
	}
      else if (xIn == getXLow())
	xIn = (getXLow() + AMSP_TINY);
      else if (xIn == getXHi())
	xIn = (getXHi() - AMSP_TINY);
    }
  else if (yIn > getYHi())
    {
      seen = false;
      xIn = (xIn + ((xVeh - xIn) * ((yIn - getYHi()) / (yIn - yVeh))));
      yIn = (getYHi() - AMSP_TINY);
      if (xIn < getXLow())
	{
	  yIn = (yIn - ((yIn - yVeh) * ((getXLow() - xIn) / (xVeh - xIn))));
	  xIn = (getXLow() + AMSP_TINY);
	}
      else if (xIn > getXHi())
	{
	  yIn = (yIn - ((yIn - yVeh) * ((xIn - getXHi()) / (xIn - xVeh))));
	  xIn = (getXHi() - AMSP_TINY);
	}
      else if (xIn == getXLow())
	xIn = (getXLow() + AMSP_TINY);
      else if (xIn == getXHi())
	xIn = (getXHi() - AMSP_TINY);
    }
  else if (xIn > getXHi())
    {
      seen = false;
      yIn = (yIn + ((yVeh - yIn) * ((xIn - getXHi()) / (xIn - xVeh))));
      xIn = (getXHi() - AMSP_TINY);
    }
  else if (xIn < getXLow())
    {
      seen = false;
      yIn = (yIn + ((yVeh - yIn) * ((getXLow() - xIn) / (xVeh - xIn))));
      xIn = (getXLow() + AMSP_TINY);
    }
  i = getXIndexAtCoord(xVeh);
  j = getYIndexAtCoord(yVeh);
  dI = 0;
  dJ = 0;
  deltaX = (xIn - xVeh);
  deltaY = (yIn - yVeh);
  cols = getCols();
  rows = getRows();
  res = getResolution();
  if (fabs(deltaX) > fabs(deltaY))
    { //step1 in X direction
      slope = (deltaY / deltaX);
      if (deltaX > 0)
	{ // step1 towards positive X
	  step1 = (getXLow() + ((getXIndexAtCoord(xVeh) + 1) * res) - xVeh);
	  deltaI = (int)((deltaX + res - step1 - AMSP_TINY) / res);
	  if (deltaY >= 0)
	    { // Y increasing, slope positive
	      step2 = (step1 * slope);
	      lack = (getYLow() + ((getYIndexAtCoord(yVeh) + 1) * res) - yVeh);
	      deltaJ = (int)((deltaY + res - lack - AMSP_TINY) / res);
	      checkCell(1,0);
	      if (step2 > lack)
		{ // line passes through cell to right
		  j++;
		  dJ++;
		  checkCell(1,0);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell to right
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * slope);
	      for (i++, dI++; ; i++, dI++)
		{ // do rest of line
		  checkCell(1,0);
		  if (step2 > lack)
		    { // line passes through cell to right
		      j++;
		      dJ++;
		      checkCell(1,0);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell to right
		      lack = (lack - step2);
		    }
		}
	    }
	  else // if (deltaY < 0)
	    { // Y decreasing, slope negative
	      step2 = (step1 * -slope);
	      lack = (yVeh - (getYLow() + (getYIndexAtCoord(yVeh) * res)));
	      deltaJ = (int)((-deltaY + res - lack - AMSP_TINY) /res);
	      checkCell(1,0);
	      if (step2 > lack)
		{ // line passes through cell to left
		  j--;
		  dJ++;
		  checkCell(1,0);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell to left
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * -slope);
	      for (i++, dI++ ; ; i++, dI++)
		{ // do rest of line
		  checkCell(1,0);
		  if (step2 > lack)
		    { // line passes through cell to left
		      j--;
		      dJ++;
		      checkCell(1,0);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell to left
		      lack = (lack - step2);
		    }
		}
	    }
	}
      else // if (deltaX < 0)
	{ // step1 towards negative X
	  step1 = (xVeh - (getXLow() + (getXIndexAtCoord(xVeh) * res)));
	  deltaI = (int)((-deltaX + res - step1 - AMSP_TINY) / res);
	  if (deltaY >= 0)
	    { // Y increasing, slope negative
	      step2 = (step1 * -slope);
	      lack = (getYLow() + ((getYIndexAtCoord(yVeh) + 1) * res) - yVeh);
	      deltaJ = (int)((deltaY + res - lack - AMSP_TINY)/ res);
	      checkCell(-1,0);
	      if (step2 > lack)
		{ // line passes through cell to right
		  j++;
		  dJ++;
		  checkCell(-1,0);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell to right
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * -slope);
	      for (i--, dI++; ; i--, dI++)
		{ // do rest of line
		  checkCell(-1,0);
		  if (step2 > lack)
		    { // line passes through cell to right
		      j++;
		      dJ++;
		      checkCell(-1,0);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell to right
		      lack = (lack - step2);
		    }
		}
	    }
	  else // if (deltaY < 0)
	    { // Y decreasing, slope positive
	      step2 = (step1 * slope);
	      lack = (yVeh - (getYLow() + (getYIndexAtCoord(yVeh) * res)));
	      deltaJ = (int)((-deltaY + res - lack - AMSP_TINY)/res);
	      checkCell(-1,0);
	      if (step2 > lack)
		{ // line passes through cell to left
		  j--;
		  dJ++;
		  checkCell(-1,0);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell to left
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * slope);
	      for (i--, dI++ ; ; i--, dI++)
		{ // do rest of line
		  checkCell(-1,0);
		  if (step2 > lack)
		    { // line passes through cell to left
		      j--;
		      dJ++;
		      checkCell(-1,0);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell to left
		      lack = (lack - step2);
		    }
		}
	    }
	}
    }
  else // if (fabs(deltaY) > fabs(deltaX))
    { // step1 in the Y direction
      slope = (deltaX / deltaY);
      if (deltaY > 0)
	{ // step1 towards positive Y
	  step1 = (getYLow() + ((getYIndexAtCoord(yVeh) + 1) * res) - yVeh);
	  deltaJ = (int)((deltaY + res - step1-AMSP_TINY) / res);
	  if (deltaX >= 0)
	    { // X increasing
	      step2 = (step1 * slope);
	      lack = (getXLow() + ((getXIndexAtCoord(xVeh) + 1) * res) - xVeh);
	      deltaI = (int)((deltaX + res - lack - AMSP_TINY) / res);
	      checkCell(0,1);
	      if (step2 > lack)
		{ // line passes through cell above
		  i++;
		  dI++;
		  checkCell(0,1);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell above
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * slope);
	      for (j++, dJ++; ; j++, dJ++)
		{ // do rest of line
		  checkCell(0,1);
		  if (step2 > lack)
		    { // line passes through cell above
		      i++;
		      dI++;
		      checkCell(0,1);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell above
		      lack = (lack - step2);
		    }
		}
	    }
	  else // if (deltaX < 0)
	    { // X decreasing
	      step2 = (step1 * -slope);
	      lack = (xVeh - (getXLow() + (getXIndexAtCoord(xVeh) * res)));
	      deltaI = (int)((-deltaX + res - lack - AMSP_TINY)/ res);
	      checkCell(0,1);
	      if (step2 > lack)
		{ // line passes through cell below
		  i--;
		  dI++;
		  checkCell(0,1);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell below
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * -slope);
	      for (j++, dJ++; ; j++, dJ++)
		{ // do rest of line
		  checkCell(0,1);
		  if (step2 > lack)
		    { // line passes through cell below
		      i--;
		      dI++;
		      checkCell(0,1);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell below
		      lack = (lack - step2);
		    }
		}
	    }
	}
      else // if (deltaY < 0)
	{ // step1 towards negative Y
	  step1 = (yVeh - (getYLow() + (getYIndexAtCoord(yVeh) * res)));
	  deltaJ = (int)((-deltaY + res - step1 - AMSP_TINY) / res);
	  if (deltaX >= 0)
	    { // X increasing
	      step2 = (step1 * -slope);
	      lack = (getXLow() + ((getXIndexAtCoord(xVeh) + 1) * res) - xVeh);
	      deltaI = (int)((deltaX + res - lack - AMSP_TINY) / res);
	      checkCell(0,-1);
	      if (step2 > lack)
		{ // line passes through cell above
		  i++;
		  dI++;
		  checkCell(0,-1);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell above
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * -slope);
	      for (j--, dJ++; ; j--, dJ++)
		{ // do rest of line
		  checkCell(0,-1);
		  if (step2 > lack)
		    { // line passes through cell above
		      i++;
		      dI++;
		      checkCell(0,-1);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell above
		      lack = (lack - step2);
		    }
		}
	    }
	  else // if (deltaX < 0)
	    { // X decreasing
	      step2 = (step1 * slope);
	      lack = (xVeh -  (getXLow() + (getXIndexAtCoord(xVeh) * res)));
	      deltaI = (int)((deltaX + res - lack - AMSP_TINY) / res);
	      checkCell(0,-1);
	      if (step2 > lack)
		{ // line passes through cell below
		  i--;
		  dI++;
		  checkCell(0,-1);
		  lack = (res - (step2 - lack));
		}
	      else
		{  // line does not pass through cell below
		  lack = (lack - step2);
		}
	      step1 = res;
	      step2 = (step1 * slope);
	      for (j--, dJ++; ; j--, dJ++)
		{ // do rest of line
		  checkCell(0,-1);
		  if (step2 > lack)
		    { // line passes through cell below
		      i--;
		      dI++;
		      checkCell(0,-1);
		      lack  = (res - (step2 - lack));
		    }
		  else
		    { // line does not pass through cell below
		      lack = (lack - step2);
		    }
		}
	    }
	}
    }
}

/***********************************************************************/

/*! AmSPMap::showDisplay shows no view, or a height view, or an
  obstacle view of the map according to whether displayLevel is
  is 0, 1, or 2. */
/*
Returned Value: none

Called By: writeOutputs (in amSPMain.cc)

*/

void AmSPMap::showDisplay() /* NO ARGUMENTS */
{
  if (displayLevel == 0)
    return;
  else if (displayLevel == 1)
    showHeightDisplay();
  else if (displayLevel == 2)
    showObstacleDisplay();
}

/***********************************************************************/

//! AmSPMap::showHeightDisplay shows a height view of the map.
/*
Returned Value: none

Called By: showDisplay

This writes a file in ppm format. The file name is in mapImageFileName.
The file will be displayed by the imageview executable.

A ppm file represents an image as follows:
1. At the top there are three lines terminated by newlines.
1a. The first line of the file is "P6".
1b. The second line of the file is of the form "c r", where c is the
   number of elements per row and r is the number of rows.
1c. The third line of the file is "255".
2. The rest of the file consists of (3 * c * r) chars taken in triples,
   with each triple representing RGB color intensity of one pixel.
The file is mapped to a picture as follows.
Each row is displayed left-to-right. Successive rows are displayed 
top-to-bottom.
The first row of the picture is the first 3c chars of the file,
The second row of the picture is the next 3c chars of the file, etc. 

For each cell of the map that has been seen, this is assigning an
intensity of cyan (aqua) to the cell according to the height given in
the cell. The intensity is scaled over the observed range of heights
so that the highest cells are as bright as possible, and the lowest
are almost black.

Each cell that has not been seen is colored yellow.

The probability given in a cell is not being used in drawing the display.

The ppm image is built downward in rows, so this starts at the last
row of the map to display and works downward.

To scale the brightness of seen cells, this has to find the greatest
and least observed heights, minHeight and maxHeight.  This begins by
finding the first cell that has been seen and using its height as the
initial value of both minHeight and maxHeight. Then it goes through
the rest of the cells and resets minHeight or maxHeight whenever one
is found that is lower or higher than the current minimum or maximum.

*/

void AmSPMap::showHeightDisplay() /* NO ARGUMENTS */
{
  int count;                   // buffer index
  int i;                       // row index
  int j;                       // column index
  double minHeight;            // least height of cell on map
  double maxHeight;            // greatest height of cell on map
  double coeff;                // scale factor for cyan
  double leastIntensity = 30;  // minimum value of intensity of cyan
  char intensity;
  size_t noWarn;               // just to stop compiler warnings
  AmSPCell * aCell;            // cell currently being examined
  
  if (displayLevel != 1)
    { // Quit if this function should not have been called
      return;
    }
  // Find the first point that has been seen
  for (i = 0, j = getCols(); ((j == getCols()) && (i < getRows())); i++)
    {
      for (j = 0; ((j < getCols()) && (!getCellAtIndexes(i,j)->hits)); j++);
    }
  if ((i == getRows()) && (j == getCols()))
    { // No cells have been seen
      minHeight = 0;
      maxHeight = 0;
    }
  else
    { // found first seen cell; set initial values of minHeight and maxHeight 
      minHeight = getCellAtIndexes(i,j)->height;
      maxHeight = minHeight;
      for (; i < getRows(); i++)
	{ // Check rest of cells to see if there are any higher or lower values
	  for (j = 0; j < getCols(); j++)
	    {
	      aCell = getCellAtIndexes(i,j);
	      if (aCell->hits)
		{
		  if (aCell->height < minHeight)
		    minHeight = aCell->height;
		  else if (aCell->height > maxHeight)
		    maxHeight = aCell->height;
		}
	    }
	}
    }
  
  // Figure out the scaling value
  if (maxHeight == minHeight)
    coeff = 0;
  else
    coeff = (255 - leastIntensity) / (maxHeight - minHeight);
  // Open the output file
  outFile = fopen(mapImageFileName, "w");
  if (outFile == 0)
    {
      rcs_print_error("Could not open file %s\n", mapImageFileName);
      return;
    }
  // Write the output file header
  fprintf(outFile, "P6\n%d %d\n255\n", getRows(), getCols());

  // Buffer the data itself
  count = 0;
  for (i = (getRows() - 1); i >= 0; i--)
    {
      for (j = 0; j < getCols(); j++)
	{
	  aCell = getCellAtIndexes(i,j);
	  intensity = (char)(leastIntensity + 
			     (int)(coeff * (aCell->height - minHeight)));
	  if (!aCell->hits)
	    { // if a point has not been seen, color it yellow
	      buf[count]   = 255;  // Red
	      buf[count+1] = 255;  // Green
	      buf[count+2] = 0;    // Blue
	    }
	  else
	    { // otherwise, show it in cyan. Brighter is higher.
	      buf[count]   = 0;           // Red
	      buf[count+1] = intensity;   // Green
	      buf[count+2] = intensity;   // Blue
	    }
	  count = (count + 3);
	}
    }
  showVehicle(buf);
  // Write the data to the output file
  noWarn = fwrite(buf, sizeof(char), count, outFile);
  fclose(outFile);
  outFile = 0;
}

/***********************************************************************/

//! AmSPMap::showObstacleDisplay shows an obstacle view of the map.
/*
Returned Value: none

Called By: showDisplay

The ppm image is built downward in rows, so this starts at the last
row of the map to display and works downward.

This is showing the obstacle probability, not the height.

*/

void AmSPMap::showObstacleDisplay() /* NO ARGUMENTS */
{
  int count;          // buffer index
  int i;              // row index
  int j;              // column index
  AmSPCell * aCell;
  size_t noWarn; // just to stop compiler warnings
  
  if (displayLevel != 2)
    { // Quit if this function should not have been called
      return;
    }
  outFile = fopen(mapImageFileName, "w");
  if (outFile == 0)
    {
      rcs_print_error("Could not open file %s\n", mapImageFileName);
      return;
    }
  // Write the output file header
  fprintf(outFile, "P6\n%d %d\n255\n", getRows(), getCols());
  // Now buffer the data itself
  count = 0;
  for (i = (getRows() - 1); i > -1; i--)
    {
      for (j = 0; j < getCols(); j++)
	{
	  aCell = getCellAtIndexes(i,j);
	  if (!aCell->hits)
	    { // if a cell has not been seen, color it yellow
	      buf[count]   = 255;  // Red
	      buf[count+1] = 255;  // Green
	      buf[count+2] = 0;    // Blue
	    }
	  else if (aCell->probability < AMSP_MAYBE_OBSTACLE)
	    { // if a cell is seen and not an obstacle, color it green
	      buf[count]   = 0;  // Red
	      buf[count+1] = 255;  // Green
	      buf[count+2] = 0;    // Blue
	    }
	  else if (aCell->probability < AMSP_OBSTACLE)
	    { // if a cell is seen and may be an obstacle, color it orange
	      buf[count]   = 255;  // Red
	      buf[count+1] = 127;  // Green
	      buf[count+2] = 0;    // Blue
	    }
	  else
	    { // if a cell is seen and is an obstacle, color it red
	      buf[count]   = 255;  // Red
	      buf[count+1] = 0;    // Green
	      buf[count+2] = 0;    // Blue
	    }
	  count = (count + 3);
	}
    }
  showVehicle(buf);
  // Write the data to the output file
  noWarn = fwrite(buf, sizeof(char), count, outFile);
  fclose(outFile);
  outFile = 0;
}

/***********************************************************************/

//! AmSPMap::showVehicle puts the vehicle in the buffer for the ppm map.
/*
Returned Value: none

Called By:
  AmSPMap::showDisplay
  AmSPMap::showSmallDisplay

This puts the vehicle where it should be according to its vehLoc.
That is always expected to be at the center of the picture but is
not arbitrarily put there. If the vehicle is not on map, this
returns without making an picture of the vehicle.

This shows the vehicle in blue.

The imageRow calculations take into account that fact that the
map rows are numbered bottom up, while the image rows are
numbered top down.  

The 3's are there because there are 3 chars in one image pixel.

Do not simplify ((vehHeight / 2) * 3) to (vehLength / 2),
since the two expressions are not equal when vehHeight is odd.

*/

void AmSPMap::showVehicle( /*   ARGUMENTS                 */
 char * buf)               /*!< buffer containing picture */
{
  int i;         // index for chars
  int j;         // index for rows of chars
  int middle;    // index in buf of the middle point of the vehicle
  int picLength; // number of chars in one row of the image
  int vehHeight; // number of image rows for vehicle
  int vehLength; // number of chars for vehicle in one image row
  int start;     // buffer index of start of a row for vehicle
  int mapRow;    // scrolled map row
  int mapCol;    // scrolled map column
  int imageRow;  // image row
  int imageCol;  // image column
  int last;      // index of last image pixel

  mapRow = getXIndexAtCoord(vehLoc.x);
  if (mapRow < 0) // off the big map
    return;
  imageRow = (getRows() - mapRow - 1);
  mapCol = getYIndexAtCoord(vehLoc.y);
  if (mapCol < 0)
    return;  // off the big map
  picLength = (3 * getCols());
  last = ((picLength * getRows()) - 1);
  imageCol = (mapCol * 3);
  vehHeight = 7;
  vehLength = (3 * vehHeight); // 3 times as many chars as image pixels
  middle = ((imageRow * picLength) + imageCol);
  start = (middle - (picLength * (vehHeight / 2)) - ((vehHeight / 2) * 3));
  if (start < 0)
    return; // don't write out of bounds
  if ((start + (vehHeight * picLength) + vehLength) > last)
    return; // don't write out of bounds
  for (j = 0; j < vehHeight; j++)
    {
      for (i = 0; i < vehLength; i = (i+3))
	{
	  buf[start + i]     = 0;   // Red
	  buf[start + i + 1] = 0;   // Green
	  buf[start + i + 2] = 255; // Blue
	}
      start = (start + picLength);
    }
}

/***********************************************************************/

/*! AmSPMap::writeSmallMap populates a SensorData2D, which is
    the contents of the small map that will be exported.  For each
    cell, only obstacle probability data is written. */
/*
Returned Value: none

Called By: writeOutputs (in amSPMain.cc)

This builds a sensorData2D representation of a square part of the
window at the center of the window and having a side length of
outSide. Only probability data is used to build the smallMap.  The
height of the cells is not being used here.

This is writing into the "data" argument using local coordinates.

In writing map cells into the featArray of the data, the following
discrete values from amSP.hh are used:
AMSP_UNSEEN 0.0
AMSP_NOT_OBSTACLE 0.2
AMSP_MAYBE_OBSTACLE 0.3
AMSP_OBSTACLE 0.6

A cell is AMSP_UNSEEN if its hits is zero.

If a map cell is not AMSP_UNSEEN, the three obstacle values are
thresholded according to the probability given in the map cell. Less
than 0.3 is AMSP_NOT_OBSTACLE; 0.3 to 0.6 is AMSP_MAYBE_OBSTACLE;
greater than 0.6 is AMSP_OBSTACLE.

If the output map will not fit in the buffer, this prints an error
message and returns.

*/

void AmSPMap::writeSmallMap( /*   ARGUMENTS            */
 SensorData2D * data)        /*!< data set to populate */
{
  int i;            // X index for output square
  int j;            // Y index for output square
  AmSPCell * aCell; // current cell

  if ((outSide * outSide) > MAX_SENSOR_DATA_2D)
    {
      rcs_print_error("output map too big for 2D sensor data\n");
      return;
    }
  data->isGlobal = false;
  data->incrementX = getResolution();
  data->incrementY = getResolution();
  data->x = outSide;
  data->y = outSide;
  data->featArray_length = outSide * outSide;
  data->lowerLeft.x = (getXLow() + (getResolution() * (getRange()- outRange)));
  data->lowerLeft.y = (getYLow() + (getResolution() * (getRange()- outRange)));
  for (i = 0; i < outSide; i++)
    {
      for (j = 0; j < outSide; j++)
	{
	  aCell = getSmallMapCell(i, j);
	  data->featArray[(i * outSide) + j] =
	    ((aCell->hits == 0) ? AMSP_UNSEEN :
	     (aCell->probability < AMSP_MAYBE_OBSTACLE) ? AMSP_NOT_OBSTACLE :
	     (aCell->probability < AMSP_OBSTACLE)       ? AMSP_MAYBE_OBSTACLE
	     : AMSP_OBSTACLE);
	}
    }
}

/***********************************************************************/

/*! AmSPMap::writeSmallMapData writes a text version of the small map
  to a file. For each cell, all data is written. */
/*
Returned Value: none

Called By: writeOutputs (in amSPMain.cc)

This writes a file (whose name is given by the "fileName" argument)
containing a text representation of a square part of the window at the
center of the window and having a side length of outSide. This is the
same area as writeSmallMap.  After the header, the file contains one
line for each cell of the small map. Each such line has all four data
items in the cell: height, range, hits, and probability.

*/

void AmSPMap::writeSmallMapData( /*   ARGUMENTS             */
 char * fileName)                /*!< name of file to write */
{
  int i;   // X index for output square
  int j;   // Y index for output square
  AmSPCell * aCell;
  FILE * outFile;

  outFile = fopen(fileName, "w");
  fprintf(outFile, "resolution = %.4f\n", getResolution());
  fprintf(outFile, "number of cells on a side = %d\n", outSide);
  fprintf(outFile, "total number of cells = %d\n", (outSide * outSide));
  fprintf(outFile, "lowerLeft.x = %f\n",
	  (getXLow() + (getResolution() * (getRange()- outRange))));
  fprintf(outFile, "lowerLeft.y = %f\n",
	  (getYLow() + (getResolution() * (getRange()- outRange))));
  for (i = 0; i < outSide; i++)
    {
      for (j = 0; j < outSide; j++)
	{
	  aCell = getSmallMapCell(i, j);
	  fprintf(outFile, "%.4f, %.4f, %d, %.4f\n", aCell->height,
		  aCell->range, aCell->hits, aCell->probability);
	}
    }
  fclose(outFile);
}

/***********************************************************************/

/*
  Modification history:

  $Log: amSPMap.cc,v $
  Revision 1.19  2008/02/12 00:06:50  tomrkramer
  Changed from using "bool readyToDisplay" to using "int displayLevel".
  Renamed showDisplay to showHeightDisplay. Wrote new showDisplay.
  Changed showSmallDisplay to showObstacleDisplay. Simplified showVehicle.

  Revision 1.18  2007/12/14 23:42:55  tomrkramer
  Improved documentation.

  Revision 1.17  2006/10/18 00:10:50  dr_steveb
  Fixed location of scrolling map include file.

  Revision 1.16  2006/10/05 20:48:19  tomrkramer
  Fixed minor bug (failing to set outFile to 0 after closing it) that was
  triggered by a "dump 0" command in amShell, causing a crash.

  Revision 1.15  2006/09/26 14:57:50  tomrkramer
  Added #include "amSP.hh" so AMSP_NOT_OBSTACLE, etc. are defined.

  Revision 1.14  2006/09/19 15:08:40  tomrkramer
  Improved out-of-bounds checking in showVehicle.

  Revision 1.13  2006/09/18 18:38:18  tomrkramer
  Significantly rewritten to use scrolling map tool. Also removed unused
  code and unused choice of display utility.

  Revision 1.12  2006/04/07 16:49:56  tomrkramer
  Fixed minor bug in constructor. Moved a check.

  Revision 1.11  2006/04/05 13:32:20  tomrkramer
  Switched to using imageview for display.

  Revision 1.10  2005/11/16 13:44:21  tomrkramer
  Improved handling of hits and probabilities

  Revision 1.9  2005/11/02 14:57:22  tomrkramer
  Thickened obstacles by one cell away from viewpoint to keep them solid

  Revision 1.6  2005/10/23 13:49:44  dr_steveb
  Check-in of changes that Tom K. made to fix missing detections of west facing obstacles.

  Revision 1.17  2005/10/06 22:13:26  kramer
  finished changing to rcs_print

  Revision 1.16  2005/10/05 23:11:34  kramer
  added documentation, made many minor changes


*/
