/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   visGraph.cc
  \brief  Generate a visibility graph

          Algorithm based on paper from Overmars and Welzl (RUU-CS-88-7, March 1988).

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.12 $
  $Date: 2009/01/23 13:47:16 $
  \endcode
  \par CVS Log:
  \ref visGraph_cc "More..."

  \author Stephen Balakirsky
  \date   06/29/05
*/
#include "posemath.h"
#include "visGraph.hh"
#include "moastGeom.hh"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

static int strncasecmp(const char * left, const char * right, int len)
{
  char ll, rl;

  while (len > 0) {
    ll = tolower(*left), rl = tolower(*right);
    if (ll < rl) return -1;
    if (ll > rl) return 1;
    left++, right++, len--;
  }

  return 0;
}

/////////////////////////////////////////////////////////////////////////////
//! constructor for visibility graph
/*!
  \param initWMSize The initialize size of the world model
  \param retPtSize  The number of points that can be returned at one time
*/
VisGraph::VisGraph( int initWmSize, int retPtSize )
{
  retStorageSize = retPtSize;
  retStorage = new Location2D[retStorageSize];
  retStorageID = new int[retStorageSize];
  cImg = NULL;
  wm = new VisGraphWM( initWmSize );
  planHorizon = 0;
}

/////////////////////////////////////////////////////////////////////////////
//! destructor for visibility graph
VisGraph::~VisGraph()
{
  delete [] retStorage;
  delete [] retStorageID;
  delete wm;
}

/////////////////////////////////////////////////////////////////////////////
//! returns the path point at the given location in the path array
/*! Return the specified path point. Return true on success
  \param pathPoint Location to store point
  \param pointNum the number of the point to store
  \return true on success, false if point does not exist
 */
bool VisGraph::getPathPoint(Location2D *pathPoint, int pointNum)
{
  VisGraphNode *visGraphNode;
  DijkHeapNNode * nodes;

  if( pointNum >= wm->pathLength )
    return false;
  nodes = searcherGraph.nodes;
  visGraphNode = (VisGraphNode*)nodes[wm->path[pointNum]].moreInfo;
  *pathPoint = visGraphNode->nodeLoc;
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//! set output image
/*!
  allocate storage for the output image using the CImg library
  \param resolution The resolution of the image. It is assumed that the model is in meters
*/
void VisGraph::setOutImage( double resolutionIn )
{
  Location2D wmSize;

  wm->setResolution(resolutionIn);
  wmSize = wm->getWMSize();
  wmSize.north /= resolutionIn;
  wmSize.east /= resolutionIn;
  if( cImg != NULL )
    delete cImg;
  //  cImg = new cimg_library::CImg<unsigned char>((int)wmSize.east,(int)wmSize.north,1,3);
  cImg = new cimg_library::CImg<unsigned char>((int)wmSize.north+1,(int)wmSize.east+1,1,3);
}

/////////////////////////////////////////////////////////////////////////////
//! dump image of world model and visibility graph
/*!
\param dumpWhat sets what should be dumped. This is a set of flags:
       0x01 - Dump the graph
       0x02 - Dump the vehicle location
       0x04 - Dump seeds
  \return true on success and false on failure
*/
bool VisGraph::dump(char *fileName, unsigned int dumpWhat)
{
  bool dumpGraph = dumpWhat & 0x01;
  bool dumpVehLoc = dumpWhat & 0x02;
  bool dumpSeeds = dumpWhat & 0x04;
  unsigned char myColor[3];
  unsigned char *red, *blue, *green;
  Location2D point1, point2;
  int count, kidCount;
  DijkHeapNNode * nodes;
  double tempEast;
  Location2D wmLL = wm->getWMLL();
  VisGraphNode *visGraphNode;
  
  point1 = wm->getWMSize();
  tempEast = point1.east/wm->getResolution();

  red =   &myColor[0];
  green = &myColor[1];
  blue =  &myColor[2];

  if( cImg == NULL )
    return false;
  *red = 255;
  *blue = 255;
  *green = 255;
  cImg->fill(*myColor);
  nodes = searcherGraph.nodes;
  if( dumpGraph )
    {
      *red = 0;
      *blue = 255;
      *green = 0;
      for( count=0; count<searcherGraph.size-1; count++ )
	{
	  visGraphNode = (VisGraphNode*)nodes[count].moreInfo;
	  point1 = visGraphNode->nodeLoc;
	  point1.east = (int)(tempEast-(point1.east-wmLL.east)/wm->getResolution());
	  point1.north = (int)((point1.north-wmLL.north)/wm->getResolution());
	  for( kidCount=0; kidCount<nodes[count].kidsSize; kidCount++ )
	    {
	      visGraphNode = (VisGraphNode*)nodes[nodes[count].kidsIndexes[kidCount]].moreInfo;
	      point2 = visGraphNode->nodeLoc;
	      point2.east = (int)(tempEast-(point2.east-wmLL.east)/ wm->getResolution());
	      point2.north = (int)((point2.north-wmLL.north)/ wm->getResolution());
	      cImg->draw_line( (int)point1.north, (int)point1.east, 
			       (int)point2.north, (int)point2.east, myColor );
	    }
	}
    }
  // dump the obstacles and vehicle location
  if( dumpVehLoc )
    wm->dump(cImg);
  // if a path exists, then draw it
  *red = 0;
  *green = 255;
  *blue = 0;
  
  for(count=0; count<wm->pathLength-1; count++)
    {
      visGraphNode = (VisGraphNode*)nodes[wm->path[count]].moreInfo;
      point1 = visGraphNode->nodeLoc;
      point1.east = (int)(tempEast-(point1.east-wmLL.east)/wm->getResolution());
      point1.north = (int)((point1.north-wmLL.north)/wm->getResolution());
      visGraphNode = (VisGraphNode*)nodes[wm->path[count+1]].moreInfo;
      point2 = visGraphNode->nodeLoc;
      point2.east = (int)(tempEast-(point2.east-wmLL.east)/ wm->getResolution());
      point2.north = (int)((point2.north-wmLL.north)/ wm->getResolution());
      cImg->draw_line( (int)point1.north, (int)point1.east, (int)point2.north, (int)point2.east, myColor );
    }

  // draw usable am seeds
  *red = 0;
  *green = 0;
  *blue = 255;
  if( dumpSeeds )
    {
      for( count=0; count<MAX_AM_SEED_NODES; count++ )
	{
	  if( nodes[searcherGraph.size-1].kidsCosts[count] != -1 )
	    {
	      visGraphNode = (VisGraphNode*)nodes[nodes[searcherGraph.size-1].kidsIndexes[count]].moreInfo;
	      point1 = visGraphNode->nodeLoc;
	      point1.east = (int)(tempEast-(point1.east-wmLL.east)/wm->getResolution());
	      point1.north = (int)((point1.north-wmLL.north)/wm->getResolution());
	      cImg->draw_point( (int)point1.north, (int)point1.east, myColor );
	    }
	}
    }
  cImg->save(fileName);
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//! grow the return storage array
void VisGraph::growRetStorage()
{
  int newSize;
  Location2D *temp;
  int *tempInt;

  newSize = retStorageSize * 2;
  temp = new Location2D[newSize];
  tempInt = new int[newSize];

  for( int count=0; count<retStorageSize; count++ )
    {
      temp[count] = retStorage[count];
      tempInt[count] = retStorageID[count];
    }
  delete [] retStorage;
  delete [] retStorageID;
  retStorage = temp;
  retStorageID = tempInt;
  retStorageSize = newSize;
  printf( "world model return storage grown to %d\n", retStorageSize );
}

/////////////////////////////////////////////////////////////////////////////
Location2D VisGraph::getStoredPt( int id )
{
  return retStorage[id];
}

/////////////////////////////////////////////////////////////////////////////
//! build the visibility graph from each vertex
/*! Build the visibility graph from the start point of each line
    in the polygons. For the polygon in question, insert its
    two original lines into the model to block interior visibility
*/
bool VisGraph::buildGraph(int numPtsIn, FILE *fp)
{
  //  int currentID = -1;
  int count, outCount;
  int numPts;
  PolyProps *properties;
  Location2D point;
  LineSegment line;
  Location2D pointIn;

  /* this was the old way where we only added two original lines
  for( count=0; count<wm->getSize(); count++ )
    {
      line = wm->getLine( count, true );
      properties = (polyProps*)line.properties;
      if( properties->id != currentID )
	{
	  currentID = properties->id;
	  line = wm->getLine(count, true);
	  line.properties = 0; // set to non-valid
	  wm->setLineScaled(wm->getSize(), line);

	  line = wm->getLine(count+properties->numVerts-1, true);
	  line.properties = 0; // set to non-valid
	  wm->setLineScaled(wm->getSize()+1, line );
	}
      else
	{
	  line = wm->getLine(count, true);
	  line.properties = 0; // set to non-valid
	  wm->setLineScaled(wm->getSize(), line);
	  line = wm->getLine(count-1, true);
	  line.properties = 0; // set to non-valid
	  wm->setLineScaled(wm->getSize()+1, line);
	}
	
  end of old way */
  for( count=0; count<numPtsIn; count++ )
    {
      if( count<wm->getSize(false) )
	{
	  line = wm->getLine(count,true);
	  properties = (PolyProps*)line.properties;
	  properties->vertUsed = true;
	}
      line =  wm->getLine(count,false);
      pointIn = line.start;
      if( line.properties != 0 )
	{
	  //	  if(!buildGraphPoint( pointIn, wm->getSize()+2, numPts ))
	  
	  if(!buildGraphPoint( pointIn, numPtsIn, numPts ))
	    return false;
	  for( outCount=0; outCount<numPts; outCount++ )
	    {
	      point = getStoredPt(outCount);
	      fprintf( fp, "Pline 2\n%f %f\n%f %f\n    Pen(1, 2, 0)\n",
		       pointIn.north,
		       pointIn.east,
		       point.north,
		       point.east );
	    }
	}
    }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//! build the visibility graph for a point
/*! This is an order n^2 operation. Please note that the properties of a line must be "1" for it to be considered
  \param point The point to build for
  \param size The number of segments to search for the graph
  \param numPtsRet The number of points that are stored in retStorage
  \param force Use points even if they have already been used by previous creation calls
  \return True on success
 */
bool VisGraph::buildGraphPoint( Location2D &point, int size, int &numPtsRet, bool force )
{
  int count, count2;
  LineSegment line;
  LineSegment wmLine;
  LineSegment intersectLine;
  bool addToGraph;
  PolyProps *properties;

  numPtsRet = 0;
  
  line.start = point;
  for( count=0; count<size; count++ )
    {
      wmLine = wm->getLine(count, false);
      line.end = wmLine.start;
      count2 = 0;
      if(wmLine.properties != 1 ) // test if an expanded polygon, if not then don't add to graph
	addToGraph = false;
      else if( !force )// test if already added to graph
	{
	  intersectLine = wm->getLine(count, true);
	  properties = (PolyProps *)intersectLine.properties;
	  if( properties->vertUsed )
	    addToGraph = false;
	  else
	    addToGraph = true;
	}
      else
	addToGraph = true;
      while( count2<size && addToGraph )
	{
	  intersectLine = wm->getLine(count2, false);
	  if( line.testIntersect( intersectLine ) )
	    {
	      addToGraph = false;
	    }
	  count2++;
	}
      if( addToGraph )
	{
	  retStorageID[numPtsRet] = count;
	  retStorage[numPtsRet++] = line.end;
	  if( numPtsRet == retStorageSize )
	    growRetStorage();
	}
    }
  return true;
}


/////////////////////////////////////////////////////////////////////////////
//! print the world model
void VisGraph::print(FILE *fp)
{
  LineSegment line, orig;
  PolyProps *properties;
  int currentID = -1;

  // not sure of the header information; just copying one from an existing file
  fprintf( fp, "Version 450\nCharset \"Neutral\"\nDelimiter \",\"\n" );
  fprintf( fp, "CoordSys NonEarth Units \"m\" Bounds (5.26511741156923, 1.483181780225416) (84.6429652083507, 45.56448243241343)\n");
  fprintf( fp, "Columns 0\nData\n\n" );
  for( int count=0; count<wm->getSize(false); count++ )
    {
      line = wm->getLine(count, true);
      properties = (PolyProps*)line.properties;
      if( currentID != properties->id )
	{
	  if( currentID != -1 )
	    fprintf( fp, "    Pen(11, 2, 8421504)\n    Brush(2, 4286945, 13882323)\n" );

	  currentID = properties->id;
	  fprintf( fp, "Region 1\n  %d\n", properties->numVerts );
	}
      fprintf( fp, "%lf %lf\n", line.start.north, line.start.east );
      /*
      fprintf( fp, "L %d %d %d %d\n",
	       (int)line.start.north, 
	       (int)line.start.east,
	       (int)line.end.north,
	       (int)line.end.east );
      */
    }
  fprintf( fp, "    Pen(11, 2, 8421504)\n    Brush(2, 4286945, 13882323)\n" );


  if( wm->isScaledSet() )
    {
      currentID = -1;
      for( int count=0; count<wm->getSize(false); count++ )
	{
	  orig = wm->getLine(count, true);
	  line = wm->getLine(count, false);
	  properties = (PolyProps*)orig.properties;
	  if( currentID != properties->id )
	    {
	      if( currentID != -1 )
		fprintf( fp, "    Pen(11, 2, 8421504)\n    Brush(2, 16711680, 13882323)\n" );
	      currentID = properties->id;
	      fprintf( fp, "Region 1\n  %d\n", properties->numVerts );
	    }
	  fprintf( fp, "%lf %lf\n", line.start.north, line.start.east );
	}
      fprintf( fp, "    Pen(11, 2, 8421504)\n    Brush(2, 16711680, 13882323)\n" );
    }
}

/////////////////////////////////////////////////////////////////////////////////////////
//// graph search routines 
/////////////////////////////////////////////////////////////////////////////////////////
//! compute the cost from node1 to node 2
/*!
  This uses a simple euclidean distance computation to get the cost from one node to another
  \param node1 The first node of the cost segment
  \param node1 The second node of the cost segment
  \return The cost of going from node1 to node2
*/
double VisGraph::computeCost(int node1, int node2 )
{
  Location2D *pt1, *pt2;
  LineSegment line;
  VisGraphNode *visGraphNode;
  DijkHeapNNode * nodes;
  double retValue;

  nodes = searcherGraph.nodes;

  // get locations of points
  visGraphNode = (VisGraphNode*)nodes[node1].moreInfo;
  pt1 = &visGraphNode->nodeLoc;
  visGraphNode = (VisGraphNode*)nodes[node2].moreInfo;
  pt2 = &visGraphNode->nodeLoc;
  retValue = sqrt((pt1->north-pt2->north)*(pt1->north-pt2->north)+
		  (pt1->east-pt2->east)*(pt1->east-pt2->east));
  retValue *= 119; //VEH_WM_MODERATE_MAX;
  return retValue;
}

/////////////////////////////////////////////////////////////////////////////
bool  VisGraph::computeCosts(int node)
{
  DijkHeapNNode * nodes;
  int i;
  float cost;

  if( node>searcherGraph.size || node<0)
    {
      printf( "illegal node number %d\n", node );
      return false;
    }
  nodes = searcherGraph.nodes;
  for( i=0; i<nodes[node].kidsSize; i++ )
    {
      if( nodes[node].kidsCosts[i] == -2 ) // need to get information
	{
	  cost = computeCost(node, nodes[node].kidsIndexes[i]);
	  nodes[node].kidsCosts[i] = cost;
	  //	  nodes[i].kidsCosts[node] = cost;
	}
    }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/*! Initialize the graph structure. Will only run throught this once. Static
    variables prevent multiple entries.
    \param endTime The system time that we need to return by
    \return true if we are finished, false if we need to be called again
*/
bool VisGraph::initGraph(double endTime)
{
  DijkHeapNNode * nodes;
  PolyProps *properties;
  LineSegment line;
  Location2D pointIn;
  int numPts;
  int count;
  VisGraphNode *visGraphNode;
  static int *indexes = NULL;
  static double *costs = NULL;
  double costFlag = -2; // flag that says we need a cost
  static int builtTo = 0;
  int debugUse;
  int maxConnections = 500;

  if( indexes == NULL )
    {
      indexes = new int[maxConnections];
      costs = new double[maxConnections];
      theSearcher = new DijkHeapNSearcher();
      // want to have the graph large enough for both the vis graph nodes and am end points
      // also add two points (start and end)
      searcherGraph.size = wm->getSize(false) + MAX_AM_SEED_NODES + 2;
      searcherGraph.capacity = searcherGraph.size;
      nodes = new DijkHeapNNode[searcherGraph.size];
      searcherGraph.nodes = nodes;
      // initialize the node structure
      builtTo = 0;
      while( builtTo < searcherGraph.size )
	{
	  nodes[builtTo].moreInfo = new VisGraphNode(builtTo);
	  nodes[builtTo].kidsSize = 0;
	  builtTo++;
	}
      builtTo = 0;
      // we need a node that is our planning start point. This node will be connected to
      // all of the am peripheral nodes. The node that is the start node is (searcherGraph.size-1)
      numPts = searcherGraph.size-1;
      nodes[numPts].kidsSize = MAX_AM_SEED_NODES;
      nodes[numPts].kidsIndexes = new int[MAX_AM_SEED_NODES];
      nodes[numPts].kidsCosts = new double[MAX_AM_SEED_NODES];
      for( count=0; count<MAX_AM_SEED_NODES; count++ )
	{
	  nodes[numPts].kidsIndexes[count] = wm->getSize(false)+count;
	  nodes[numPts].kidsCosts[count] = -1;
	}
    }

  // build the visibility graph
  while( builtTo < wm->getSize(false) )
    {
      nodes = searcherGraph.nodes;
      line = wm->getLine(builtTo,true);
      properties = (PolyProps*)line.properties;
      properties->vertUsed = true;
      line =  wm->getLine(builtTo,false);
      pointIn = line.start;
      visGraphNode = (VisGraphNode*)nodes[builtTo].moreInfo;
      visGraphNode->nodeLoc = pointIn;
      if(!buildGraphPoint( pointIn, wm->getSize(true), numPts )) // error, abort creation
	{
	  printf( "Error on constructing visibility graph\n" );
	  builtTo = wm->getSize(false);
	  delete [] indexes;
	  delete [] costs;
	  return true;
	}
      /*
      nodes[builtTo].kidsSize = numPts;
      nodes[builtTo].kidsIndexes = new int[nodes[builtTo].kidsSize];
      nodes[builtTo].kidsCosts = new double[nodes[builtTo].kidsSize];
      */
      if( numPts > maxConnections )
	{
	  printf( "Error; Too many graph connections!\n\n" );
	  delete [] indexes;
	  delete [] costs;
	  maxConnections = numPts + 1;
	  indexes = new int[maxConnections];
	  costs = new double[maxConnections];
	}
      for( count=0; count<numPts; count++ ) // set the kids up
	{
	  /*
	  nodes[builtTo].kidsIndexes[count] = retStoredID[count];
	  nodes[builtTo].kidsCosts[count] = -2; // my flag that we need a cost
	  */
	  debugUse = retStorageID[count];
	  indexes[count] = retStorageID[count];
	  costs[count] = costFlag;
	  searcherGraph.addKids(retStorageID[count], 1, 0, NULL, NULL, &builtTo, &costFlag);
	}
      if( numPts > 0 )
	searcherGraph.addKids(builtTo, numPts, 0, NULL, NULL, indexes, costs );
      
      builtTo++;
      if( etime() > endTime )
	return false;
    }
  delete [] indexes;
  delete [] costs;
  indexes = NULL;
  costs = NULL;
  return true;
}

void VisGraph::printAnswer( int pathLength, double cost, int *path)
 {
  int n;
  LineSegment line;
  VisGraphNode *visGraphNode;
  Location2D point;
  DijkHeapNNode * nodes;

  nodes = searcherGraph.nodes;
  
  printf( "Path:\n" );
  for (n = 0; n < pathLength; n++)
    {
      visGraphNode = (VisGraphNode*)nodes[path[n]].moreInfo;
      point = visGraphNode->nodeLoc;
      printf( "node %d at <%lf %lf>\n",nodes[path[n]].moreInfo->index, point.north, point.east ); 
    }
  printf("\n");
  printf("With cost: %.4f\n\n", cost);
}

/////////////////////////////////////////////////////////////////////////////
//! returns the peripheral goal that the planner is using
 Location2D VisGraph::getAMGoal()
 {
   VisGraphNode *visGraphNode;
   Location2D point;
 
   if( wm->pathLength < 2 )
     {
       point.north = 0;
       point.east = 0;
       printf( "No point to return from planner!\n" );
       return point;
     }
   visGraphNode = (VisGraphNode*)searcherGraph.nodes[wm->path[1]].moreInfo;
   return visGraphNode->nodeLoc;
 }

/////////////////////////////////////////////////////////////////////////////
//! Add nodes to get from the AM perimeter to the graph
/*!
  This routine takes the perimeter nodes from AM and computes how they
  fit into the visibility graph. This
  will allow the vehicle echelon to plan where it knows best while taking 
  advantage of the AM echelons plan. To do this, we have a dummy node
  that is connected to all of the perimeter nodes. We can then start
  the search at the dummy node and it imediately gets out to the perimeter. 
  Note that the last seed is the vehicle location and it is not used

  \param points The points along the AM plan to add
  \param pointLength the number of these points
  \param endTime The time to stop
  \return true when done, false if need to be re-entered
*/
bool VisGraph::addGraphPoints( PM_CARTESIAN *points, int pointLength, double endTime )
{
  int count2;
  int *nodesToGraph;
  double *costsToGraph;
  int numPts, nodeNum;
  Location2D loc2d;
  DijkHeapNNode * nodes;
  VisGraphNode *visGraphNode;
  bool haveGoodSeed;
  static int count = 0; // don't mess with this value!
  static int numAdded = 0;

  nodes = searcherGraph.nodes;

  // connect new points to rest of graph. These will be
  // uni-directional connections (i.e. you will not be able
  // to get to any of these nodes from anywhere else
  haveGoodSeed = false;
  while( count< pointLength-1 )
    {
      /*************************************************************
       There seems to be a problem with what AM and Veh see as north, so fix it here!
       *************************************************************/
      loc2d.north = points[count].y;
      loc2d.east = points[count].x;
      nodeNum = wm->getSize(false)+count;
      visGraphNode = (VisGraphNode*)nodes[nodeNum].moreInfo;
      visGraphNode->nodeLoc = loc2d;
      // only use a point if it is not in an obstacle and has a valid cost.
      // Also, only use the last point if we have no others
      //      if( count==pointLength-1 && !haveGoodSeed )
      //	printf( "Checking vehicle location as seed...\n" );
      if( points[count].z>-1 && wm->ditherPoint( &loc2d ) ) //&& count!=pointLength-1) || 
	//	  (count==pointLength-1 && !haveGoodSeed))
	{
	  //	  if( count==pointLength-1 )
	  //	    printf( "vehicle location OK as seed...\n" );
	  haveGoodSeed = true;
	  if(!buildGraphPoint( loc2d, wm->getSize(true), numPts, true )) // error, abort creation
	    {
	      printf( "Error on constructing visibility graph\n" );
	      count = 0;
	      numAdded = 0;
	      return false;
	    }
	  nodesToGraph = new int[numPts];
	  costsToGraph = new double[numPts];
	  for( count2=0; count2<numPts; count2++ )
	    {
	      nodesToGraph[count2] = retStorageID[count2];
	      costsToGraph[count2] = -2; // my flag that we need a cost
	    }
	  // link peripheral node to visible points
	  searcherGraph.useKids(nodeNum, numPts, nodesToGraph, costsToGraph);
	  // set cost of arc from start node to peripheral node
	  nodes[searcherGraph.size-1].kidsCosts[count] = points[count].z * 2.; 
	  //	  nodes[searcherGraph.size-1].kidsIndexes[count] = wm->getSize(false)+count;  set when graph initialized
	  numAdded++;
	}
      else
	{
	  /*
	  if( points[count].z <= 0 )
	    printf( "Point killed by z value\n" );
	  */
	  nodes[nodeNum].kidsSize = 0;
	  nodes[searcherGraph.size-1].kidsCosts[count] = -1;
	}
      count++;
      /*
      if( etime() > endTime )
	return false; // out of time, but not error
      */
    }
  // if still no good seeds we have an issue...
  if( !haveGoodSeed )
    {
      count = 0;
      return false;
    }
  // zero out rest of conections
  for( count=pointLength; count<MAX_AM_SEED_NODES; count++ )
    {
      //      nodes[searcherGraph.size-1].kidsIndexes[count] = wm->getSize(false)+count; set when graph initialized
      nodes[searcherGraph.size-1].kidsCosts[count] = -1;
    }
  count = 0;
  /* This is now done differently!
  // now link up the dummy node to the new nodes
  nodesToGraph = new int[pointLength];
  costsToGraph = new double[pointLength];
  for( count2=0; count2<pointLength; count2++ )
    {
      if( points[count].z > 0 )
	{
	  nodesToGraph[count] = wm->getSize(false)+count2;
	  costsToGraph[count] = points[count2].z; 
	  count++;
	}
    }
  searcherGraph.useKids(searcherGraph.size-1, count, nodesToGraph, costsToGraph);
  count = 0;
  */
       //  printf( "Added %d new connections nodes out of %d\n", numAdded, pointLength );
  numAdded = 0;
  return true;
}

/////////////////////////////////////////////////////////////////////////////////////////
/*! The location of the goal node is stored in the graph structure at the size -2 location
 */
void VisGraph::setGoal( Location2D goalIn )
{
  int numPts;
  int count, count2;
  int builtTo = searcherGraph.size-2; // location to store goal
  double costFlag = -2; // flag that says we need a cost
  DijkHeapNNode *goalNode, *targetNode;
  DijkHeapNNode * nodes;
  VisGraphNode *visGraphNode;
  int *nodesToGraph;
  double *costsToGraph;

  // set the location in the graph node
  nodes = searcherGraph.nodes;
  goalNode = &nodes[builtTo];
  visGraphNode = (VisGraphNode*)goalNode->moreInfo;
  visGraphNode->nodeLoc = goalIn;
  // clear out the old goal connections
  for( count=0; count<goalNode->kidsSize; count++ )
    {
      targetNode = &nodes[goalNode->kidsIndexes[count]];
      nodesToGraph = new int[targetNode->kidsSize-1];
      costsToGraph = new double[targetNode->kidsSize-1];
      numPts = 0;
      for( count2=0; count2<targetNode->kidsSize; count2++ )
	{
	  if( targetNode->kidsIndexes[count2] != builtTo )
	    {
	      nodesToGraph[numPts] = targetNode->kidsIndexes[count2];
	      costsToGraph[numPts] = targetNode->kidsCosts[count2];
	      numPts++;
	    }
	}
      searcherGraph.useKids(goalNode->kidsIndexes[count], numPts, nodesToGraph, costsToGraph);
    }
  if(!buildGraphPoint( goalIn, wm->getSize(true), numPts, true )) // error, abort creation
    {
      printf( "Error on constructing visibility graph\n" );
      return;
    }
  nodesToGraph = new int[numPts];
  costsToGraph = new double[numPts];
  for( count=0; count<numPts; count++ ) // set the kids up
    {
      searcherGraph.addKids(retStorageID[count], 1, 0, NULL, NULL, &builtTo, &costFlag);
      nodesToGraph[count] = retStorageID[count];
      costsToGraph[count] = -1;
    }
  // set the goal node to point to all nodes that point to it so that we can delete them later
  searcherGraph.useKids(builtTo, numPts, nodesToGraph, costsToGraph);
  //  printf( "Goal set to node %d loc <%f %f>\n", builtTo, goalIn.north, goalIn.east );
}

/////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform the actual search while conforming to the time constraints

\param newSearch A new search is just starting
\param endTime The time the search must end by
\return -1 error
\return The value of pathFound from the searcher (0 not reached, 1 reached, 2 optimally reached
*/
#define MAX_DIJK_WAYPOINTS 500
int VisGraph::search( bool newSearch, double endTime )
{
  int numberOfSearches = 0;
  static int startNode, finishNode;
  //  static int errCount=0;
  static int opensOut;
  static int statusesOut;
  static int goalsFoundOut=0;
  static int optimaFoundOut=0;
  static double costsOut;
  int n;
  int firstPoint;
  Location2D pathPoint;
  VisGraphNode *visGraphNode1;//, *visGraphNode2;
  bool nearGoal;
  static int numCycles=0;

  if( newSearch )
    {
      startNode = searcherGraph.size-1;
      finishNode = searcherGraph.size-2;
      if( theSearcher->initSearch( 
				  &searcherGraph, /* the graph to search in                     */
				  startNode,               /* index in nodes of aGraph of start node     */
				  1,          /* number of goals                            */
				  &finishNode,           /* indexes in nodes of aGraph of goal nodes   */
				  &opensOut,          /* 0=open list empty, non-zero=list not empty */
				  &statusesOut,       /* statuses of reaching goals                 */
				  &costsOut,       /* costs of reaching goals                    */
				  &goalsFoundOut,     /* number of goals reached                    */
				  &optimaFoundOut)    /* number of goals reached optimally          */
	  )
	return( -1 ); // init error
    }
  while( optimaFoundOut != 1 && opensOut != 0 && numberOfSearches++ < 10000)
    {
      if( numberOfSearches++ > 1000 )
	{
	  if( etime() > endTime )
	    {
	      if( goalsFoundOut )
		break;
	      printf( "OUT OF TIME FOR VEH PLAN! WAITING FOR NEXT CYCLE\n" );
	      return( statusesOut );
	    }
	  numberOfSearches = 0;
	}
      // get the next node and compute the cost of reaching its kids
      n = theSearcher->getNextNode();
      computeCosts(n);
      theSearcher->cycleSearch();
      numCycles++;
    }
  //  if( optimaFoundOut )
  if( goalsFoundOut )
    {
      wm->pathLength = theSearcher->reportSearch( finishNode, MAX_DIJK_WAYPOINTS, wm->path );
      if( wm->pathLength == -1 )
	return( -1 );
      if( !optimaFoundOut )
	printf( "Sub-optimal goal being reported\n" );
      /*
      printAnswer(wm->pathLength, costsOut, wm->path);
      printf( "Plan took %d cycles\n", numCycles );
      */
      // trim path to first point that is greater then the error allowance but still before the am horizon
      // if such a point does not exist, send horizon
      firstPoint = 1;
      n = 2;
      while( n<wm->pathLength)
	{
	  getPathPoint(&pathPoint, n );
	  wm->pathDistance = pathPoint.dist(getVehLoc());
	  if( wm->pathDistance>planHorizon )
	    break;
	  if( wm->pathDistance> 1. )
	    {
	      firstPoint = n;
	      break;
	    }
	  n++;
	}
      for( n=0; n<wm->pathLength-firstPoint; n++ )
	wm->path[n] = wm->path[n + firstPoint];
      wm->pathLength -= firstPoint;
      wm->pathDistance = 0;
      nearGoal = true;
      for( n=0; n<wm->pathLength; n++ )
	{
	  visGraphNode1 = (VisGraphNode*)searcherGraph.nodes[wm->path[n]].moreInfo;
	  if( visGraphNode1->nodeLoc.dist(getVehLoc()) > 1.5 * planHorizon )
	    {
	      nearGoal = false;
	      break;
	    }
	}
      if( nearGoal )
	{
	  //	  printf( "Almost there <%lf:%lf> path was %d... sending actual goal to AM\n", wm->pathDistance, planHorizon,
	  //  wm->pathLength);
	  //printAnswer(wm->pathLength, costsOut, wm->path);
	  wm->pathLength = 1;
	  wm->path[0] = searcherGraph.size-2;
	  return 2;
	}

      /*
// old way of computing to only send the goal point. Does not work well!
      for( n=0; n<wm->pathLength-1; n++ )
	{
	  visGraphNode1 = (VisGraphNode*)searcherGraph.nodes[wm->path[n]].moreInfo;
	  visGraphNode2 = (VisGraphNode*)searcherGraph.nodes[wm->path[n+1]].moreInfo;
	  wm->pathDistance += visGraphNode1->nodeLoc.dist(visGraphNode2->nodeLoc);
	}
      if( wm->pathDistance < planHorizon * 1.5 )
	{
	  //	  printf( "Almost there <%lf:%lf> path was %d... sending actual goal to AM\n", wm->pathDistance, planHorizon,
	  //  wm->pathLength);
	  //printAnswer(wm->pathLength, costsOut, wm->path);
	  wm->pathLength = 1;
	  wm->path[0] = searcherGraph.size-2;
	  return 2;
	}
      */
      //      printf( "Total distance remaining %lf\n", wm->pathDistance );
      numCycles = 0;
    }
  else
    {
      printf( "Plan has %d cycles so far...\r", numCycles );
      fflush(stdout);
    }
  if( goalsFoundOut || opensOut==0 )
    theSearcher->closeSearch();
  if( statusesOut==0 && opensOut==0 )
    {
      numCycles = 0;
      statusesOut = -1;
    }
  return( statusesOut );
}

/////////////////////////////////////////////////////////////////////////////////////////
//// end of graph search code
/////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//! constructor for world model
/*!
  \param initSize  The initial size of the storage array
*/
VisGraphWM::VisGraphWM( int initSize )
{
  storageOriginal = new LineSegment[initSize];
  currentSize = 0;
  maxSize = initSize;
  for( int i=0; i<maxSize; i++ )
    storageOriginal[i].properties = (long)new PolyProps();
  pathLength = 0;
  path = new int[MAX_DIJK_WAYPOINTS];
  vehLoc = NULL;
}

/////////////////////////////////////////////////////////////////////////////
//! free all allocated memory
VisGraphWM::~VisGraphWM()
{
  for( int i=0; i<maxSize; i++ )
    delete (PolyProps*)storageOriginal->properties;
  delete [] storageOriginal;
  if( storageScaled != NULL )
    {
      /* never allocated this memory
      for( int i=0; i<currentScaledSize; i++ )
	delete (PolyProps*)storageScaled->properties;
      */
      delete [] storageScaled;
    }
  delete [] path;
  if( vehLoc != NULL ) {
    delete vehLoc;
  }
}

/////////////////////////////////////////////////////////////////////////////
//! try to dither a point to move out of obstacles
/*!
 moves the point in several directions by .5m to try to move out of obstacles
 \return true if outside of polygons
*/
bool VisGraphWM::ditherPoint( Location2D *point )
{
  Location2D loc2d;
  int count, direction;

  count = 1;
  direction = 0;
  if( !isInPoly(*point, -1) ) 
    return true;
  /*  else
    return false;
  */
  while( count<10 ) 
    {
      switch( direction )
	{
	case 0: // move north
	  loc2d.north = point->north + count*.1;
	  loc2d.east = point->east;
	  break;
	case 1: // move south
	  loc2d.north = point->north - count*.1;
	  loc2d.east = point->east;
	  break;
	case 2: // move east
	  loc2d.north = point->north;
	  loc2d.east = point->east + count*.1;
	  break;
	case 3: // move west
	  loc2d.north = point->north;
	  loc2d.east = point->east - count*.1;
	  break;
	}
      if( ++direction > 3 )
	{
	  direction = 0;
	  count++;
	}
      if( !isInPoly(loc2d, -1) ) 
	{
	  *point = loc2d;
	  return true;
	}
      count++;
    }
  return false;
}

/////////////////////////////////////////////////////////////////////////////
//! Set the location of the vehicle in the world
void VisGraphWM::setVehLoc( const Location2D &vehLocIn )
{
  if( vehLoc == NULL )
    vehLoc = new Location2D();
  *vehLoc = vehLocIn;
}

/////////////////////////////////////////////////////////////////////////////
//! Add a line segment to the world model
/*!
  \param line The line segment to be added
  \return id of line on success,
          -1 on failure
*/
int VisGraphWM::add( LineSegment &line, PolyProps properties )
{
  LineSegment *temp;
  int newSize;
  PolyProps *setProps;

  if( currentSize == maxSize )
    {
      newSize = maxSize*2;
      temp = new LineSegment[newSize];
      for( int count=0; count<maxSize; count++ )
	temp[count] = storageOriginal[count];
      for( int count=maxSize; count<newSize; count++ )
	temp[count].properties = (long)new PolyProps();
      delete [] storageOriginal;
      storageOriginal = temp;
      maxSize = newSize;
      printf( "Increased wm size to %d\n", newSize );
    }
  setProps = (PolyProps*)storageOriginal[currentSize].properties;
  storageOriginal[currentSize] = line;
  *setProps = properties;
  storageOriginal[currentSize].properties = (long)setProps;
  currentSize++;
  return( currentSize - 1 );
}

/////////////////////////////////////////////////////////////////////////////
//! dump the world model to a CImg image structure
bool VisGraphWM::dump(cimg_library::CImg<unsigned char> *imageIn)
{
  unsigned char myColor[3];
  unsigned char *red, *blue, *green;
  cimg_library::CImg<int> polygonPts;
  int count, count2, currentID;
  PolyProps *properties;
  int polygonSize;
  double tempEast = wmSize.east/resolution;
  Location2D point1;

  red =   &myColor[0];
  green = &myColor[1];
  blue =  &myColor[2];

  if( imageIn == NULL )
    return false;

  currentID = -1;
  polygonSize = 0;

  *red = 255;
  *green = 0;
  *blue = 0;
  for( count=0; count<currentSize; count++ )
    {
      properties = (PolyProps*)storageOriginal[count].properties;
      if( properties->id != currentID )
	{
	  currentID = properties->id;
	  if( polygonSize > 1 )// create polygon
	    {
	      polygonPts.assign(polygonSize,3);
	      for( count2=0; count2<polygonSize; count2++ )
		{
		  polygonPts(count2,1) = (int)(tempEast-(storageScaled[count-count2-1].start.east-wmLL.east)/resolution);
		  polygonPts(count2,0) = (int)((storageScaled[count-count2-1].start.north-wmLL.north)/resolution);
		}
	      imageIn->draw_polygon( polygonPts, myColor );
	    }
	  polygonSize = 0;
	}
      polygonSize++;
    }
  // draw last polygon
  if( polygonSize > 1 )// create polygon
    {
      polygonPts.assign(polygonSize,3);
      for( count2=0; count2<polygonSize; count2++ )
	{
	  polygonPts(count2,1) = (int)(tempEast-(storageScaled[count-count2-1].start.east-wmLL.east)/resolution);
	  polygonPts(count2,0) = (int)((storageScaled[count-count2-1].start.north-wmLL.north)/resolution);
	}
      imageIn->draw_polygon( polygonPts, myColor );
    }
  // draw vehicle location if available
  if( vehLoc != NULL )
    {
      *red = 0;
      *green = 255;
      *blue = 0;
      point1.east = (int)(tempEast-(vehLoc->east-wmLL.east)/resolution);
      point1.north = (int)((vehLoc->north-wmLL.north)/resolution);
      imageIn->draw_circle( (int)point1.north, (int)point1.east, (int)(1./(resolution*2)), myColor );
      //  printf( "Drawing vehLocation at: <%f %f> %d %d\n", vehLoc->north, vehLoc->east, (int)point1.north, (int)point1.east );
    }
  return true;
}


/////////////////////////////////////////////////////////////////////////////
//! is the scaled version of the world model set?
bool VisGraphWM::isScaledSet()
{
  if( storageScaled != NULL )
    return true;
  else
    return false;
}

/////////////////////////////////////////////////////////////////////////////
//! compare the given value to min and max values
/*! used to set the world models x, y range and lower left corner
  \param minWM The temporary storage space for the min values
  \param maxWM The temporary storage space for the max values
  \param pointIn The value that will be compared
 */
void VisGraphWM::compareValues(Location2D &minWM, Location2D &maxWM, const Location2D &pointIn)
{
  if( pointIn.north > maxWM.north )
    maxWM.north = pointIn.north;
  if( pointIn.north < minWM.north )
    minWM.north = pointIn.north;
  if( pointIn.east > maxWM.east )
    maxWM.east = pointIn.east;
  if( pointIn.east < minWM.east )
    minWM.east = pointIn.east;
}

//! scales the world model polygonsby a given amount
/*! On a visibility graph, you can either leave the obstacle
    polygons alone, and plan on a non-holonomic robot or
    scale the obstacles by half the size of the robot and plan
    with a point. We plan with a point!
    \param distance The distance to move each line of the polygon
    \param addOrig Adds the original polygons to the end of the data array with the
           properties set to 0.
    \return The number of lines added to the scaled world model
*/
int VisGraphWM::scaleWM(double distance, bool addOrig)
{
  PolyProps *properties;
  int currentID, saveLineNo, count, direction=-1;
  int retValue;
  Location2D minWM, maxWM;

  // first allocate the necessary space
  currentScaledSize = currentSize + 3;
  if( addOrig )
    currentScaledSize += currentSize;
  storageScaled = new LineSegment[currentScaledSize]; // two is for extra vis graph segments
  
  saveLineNo = 0;
  properties = (PolyProps*)storageOriginal[0].properties;
  currentID = properties->id;
  // shift the first line
  storageScaled[0] = storageOriginal[0].moveLineParallel(distance, direction);
  storageScaled[0].properties = 1; // set to one to note valid edge
  minWM.north = storageScaled[0].start.north;
  minWM.east = storageScaled[0].start.east;
  maxWM.north = storageScaled[0].start.north;
  maxWM.east = storageScaled[0].start.east;
  compareValues(minWM, maxWM, storageScaled[0].start);
  for( count=1; count<currentSize; count++ )
    {
      properties = (PolyProps*)storageOriginal[count].properties;
      if( properties->id != currentID )
	{
	  storageScaled[saveLineNo].start = storageScaled[count-1].intersect(storageScaled[saveLineNo]);
	  storageScaled[count-1].end = storageScaled[saveLineNo].start;
	  saveLineNo = count;
	  currentID = properties->id;
	  // need to move first line of poly and increment
	  storageScaled[count] = storageOriginal[count].moveLineParallel(distance, direction);
	  storageScaled[count].properties = 1; // set to one to note valid edge
	  count++;
	}
      storageScaled[count] = storageOriginal[count].moveLineParallel(distance, direction);
      storageScaled[count].properties = 1; // set to one to note valid edge
      storageScaled[count].start = storageScaled[count-1].intersect(storageScaled[count]);
      storageScaled[count-1].end = storageScaled[count].start;
      compareValues(minWM, maxWM, storageScaled[count].start);
    }
  // move the last line
  storageScaled[saveLineNo].start = storageScaled[count-1].intersect(storageScaled[saveLineNo]);
  storageScaled[count-1].end = storageScaled[saveLineNo].start;
  compareValues(minWM, maxWM, storageScaled[saveLineNo].start);
  // set the world model extent values
  wmSize.north = maxWM.north - minWM.north;
  wmSize.east = maxWM.east - minWM.east;
  wmLL = minWM;
  retValue = currentSize;

  // now add the original lines if necessary
  // note that this a a quick fix to eliminate the need for polygon merging
  // we should really merge overlapping polygons at this point. It would
  // be much more efficient
  if( addOrig )
    {
      currentID = currentSize;
      for( count=0; count<currentSize; count++ )
	{
	  storageScaled[currentID] = storageOriginal[count];
	  storageScaled[currentID++].properties = 0; // set to zero to note non-valid edge
	}
      retValue = currentID;
      // if we are adding the original, then we should also mark as invalid any points that
      // lie inside of a polygon. The problem with this is that the points created by the
      // polygon intersection are not included in the graph
      detectPtsInPoly();
    }
  return retValue;
}

//! detect any points that lie inside of a polygon. These will be marked as invalid
void VisGraphWM::detectPtsInPoly()
{
  int count;
  int myID;
  Location2D checkPoint;
  PolyProps *properties;

  for(count=0; count<currentSize; count++ ) // this is the point to test
    {
      properties = (PolyProps*)storageOriginal[count].properties;
      myID = properties->id;
      checkPoint.north = storageScaled[count].start.north;
      checkPoint.east = storageScaled[count].start.east;
      if( isInPoly(checkPoint, myID) )
	storageScaled[count].properties = 0;
    }
}

//! detect if a single point lies inside of a polygon. 
/*!
  \param myID - The id of the polygon that the point is a member of. It not member of any polygon, use -1.
 */
bool VisGraphWM::isInPoly(const Location2D &pointToCheck, int myID)
{
  int count2;
  bool markInvalid;
  int currentID = -118;
  PolyProps *properties;

  markInvalid = 0;
  for( count2=0; count2<currentSize; count2++ )
    {
      properties = (PolyProps*)storageOriginal[count2].properties;
      if( properties->id != currentID )
	{
	  if( markInvalid && myID!=currentID )
	    return true;
	  currentID = properties->id;
	  markInvalid = 0;
	  //	      break;
	}
      if((((storageScaled[count2].start.east<=pointToCheck.east) && 
	   (pointToCheck.east<storageScaled[count2].end.east)) ||
	  ((storageScaled[count2].end.east<=pointToCheck.east) && 
	   (pointToCheck.east<storageScaled[count2].start.east))) &&
	 (pointToCheck.north<(storageScaled[count2].end.north-storageScaled[count2].start.north) * 
	  (pointToCheck.east-storageScaled[count2].start.east)/(storageScaled[count2].end.east-storageScaled[count2].start.east)+
	  storageScaled[count2].start.north))
	markInvalid = !markInvalid;
    }
  return (markInvalid && myID!=currentID );
}

/////////////////////////////////////////////////////////////////////////////
//! return a line segment from the world model
/*! \param id The id of the line to be returned
    \return line segment on success, unknown on failure
*/
LineSegment VisGraphWM::getLine( int id, bool fromOriginal )
{
  if( fromOriginal )
    {
      if( id > currentSize )
	return( storageOriginal[0] );
      else
	return( storageOriginal[id] );
    }
  else
    {
      if( id > currentScaledSize )
	return( storageScaled[0] );
      else
	return( storageScaled[id] );
    }
}

/////////////////////////////////////////////////////////////////////////////
/*! get the size of the world model
  \param getScaled If true return the size of the scaled model, if false
         return the unscaled size
  \return The size of the selected model
*/
int VisGraphWM::getSize( bool getScaled )
{
  if( getScaled )
    return(currentScaledSize);
  else
    return(currentSize);
}

/////////////////////////////////////////////////////////////////////////////
//! set a line in the scaled world model
/*! \param id the id of the line to set
    \param lineIn the line to copy
    \return true on success
 */
bool VisGraphWM::setLineScaled( int id, LineSegment &lineIn )
{
  if( id > currentScaledSize )
    return false;
  storageScaled[id] = lineIn;
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//! return a named point from the hash table.
/*!
  If point is not found, return (0,0)
 */
Location2D VisGraphWM::getNamedPt( char *name )
{
  map<const char *, void *, ltstr>::iterator iter;
  void * ep;
  Location2D retValue;

  iter = goalHash.find(name);
  if (iter == goalHash.end()) {
      retValue.north = 0;
      retValue.east = 0;
      return retValue;
  } else {
    ep = iter->second;
    // debug
    retValue = *((Location2D *) ep);
    printf( "returning <%f %f> for point %s\n", retValue.north, retValue.east, name );
    // end of debug
    return *((Location2D *) ep);
  }
}

/////////////////////////////////////////////////////////////////////////////
//! load the world model of the visibility graph with the input from a file
#define INPUTBUFSIZE 256
/*!
  \param name the file to load
  \param wmGrow the amount to grow obstacles by
  \return the number of lines added to the world model
 */
int VisGraphWM::load(char *name, double wmGrow)
{
  LineSegment line;
  FILE *mifFp, *midFp;
  int count1, count2;
  Location2D saveMe;
  char inputBuf[INPUTBUFSIZE];
  PolyProps properties;
  char featureName[INPUTBUFSIZE];
  Location2D *hashValue;
  char *hashKey;
  map<const char *, void *, ltstr>::iterator iter;

  sprintf( inputBuf, "%s.mif", name );
  mifFp = fopen( inputBuf, "r" );
  if( mifFp == NULL )
    {
      printf( "Error opening input file %s from veh. loadWM\n", inputBuf );
      return 0;
    }
  sprintf( inputBuf, "%s.mid", name );
  midFp = fopen( inputBuf, "r" );
  if( midFp == NULL )
    {
      printf( "Error opening input file %s from veh. loadWM\n", inputBuf );
      return 0;
    }
  // delete expanded world model and zero regular world model
  if( storageScaled != NULL )
    {
      /* never allocated this memory
      for( int i=0; i<currentScaledSize; i++ )
	delete (PolyProps*)storageScaled->properties;
      */
      delete [] storageScaled;
    }
  currentSize = 0;
  properties.id = 0;

  //open file and read data
  while( fgets(inputBuf, INPUTBUFSIZE, mifFp) != NULL )
    {
      // use fancy formatting to get at string with white space
      if( !strncasecmp( "Region", inputBuf, 6 ) )
	{
	  count2 = fscanf( midFp, "%*[^\"]%*[\"]%[^\"]%*s", featureName );
	  fscanf( mifFp, "%d", &count1 );
	  properties.numVerts = count1;
	  // read first element to initialize the line
	  fscanf( mifFp, "%lf %lf", &line.end.north,
		  &line.end.east );
	  saveMe = line.end;
	  // read the rest of the polygon
	  for( count2=1; count2<count1; count2++ )
	    {
	      line.start = line.end;
	      fscanf( mifFp, "%lf %lf", &line.end.north,
		      &line.end.east );
	      add(line, properties);
	    }
	  line.start = line.end;
	  line.end = saveMe;
	  add(line, properties);
	  properties.id++;
	}
      else if( !strncasecmp( "Point", inputBuf, 5 ) )
	{
	  count2 = fscanf( midFp, "%*[^\"]%*[\"]%[^\"]%*s", featureName );
	  hashValue = new Location2D();
	  sscanf( inputBuf, "%*s %lf %lf", &hashValue->north, &hashValue->east );
	  hashKey = new char[strlen(featureName)+1];
	  strcpy( hashKey, featureName );

	  // if the key exists, replace the value with the new one,
	  // otherwise add it as a new one
	  iter = goalHash.lower_bound(hashKey);
	  if ((goalHash.end() != iter)
	      && !(goalHash.key_comp()(hashKey, iter->first))) {
	    iter->second = hashValue;
	  } else {
	    goalHash.insert(pair<const char *, void *>(hashKey, hashValue));
	  }
	
	  printf( "Found point of interest: <%f %f> named: %s\n",
		  hashValue->north, hashValue->east, hashKey );
	}
    }
  // scale the world model
  count1 = scaleWM(wmGrow, true);
  return count1;
}

/////////////////////////////////////////////////////////////////////////////
//! comparison function used by qsort
/*! This function will sort by decreasing x-coordinate
    If the x-coordinates are equal, then sort by
    decreasing y-coordinates
*/
int rotTreeCompar( const void *vN1, const void *vN2 )
{
  RotTreeNode *n1 = (RotTreeNode*)vN1;
  RotTreeNode *n2 = (RotTreeNode*)vN2;

  PM_CARTESIAN loc1, loc2;

  loc1 = n1->getLocation();
  loc2 = n2->getLocation();

  if( loc1.x < loc2.x )
    return 1;
  else if( loc1.x > loc2.x )
    return -1;
  else if( loc1.y < loc2.y )
    return 1;
  else if( loc1.y > loc2.y )
    return -1;
  else
    return 0;
}



/////////////////////////////////////////////////////////////////////////////
RotTreeNode::RotTreeNode( PM_CARTESIAN &locIn )
{
  setLocation( locIn );
  remove();
  RotTreeNode();
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode::RotTreeNode()
{
  remove();
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode::~RotTreeNode()
{
}

/////////////////////////////////////////////////////////////////////////////
void RotTreeNode::remove()
{
  if( rightBrother != NULL )
    rightBrother->leftBrother = NULL;
  if( leftBrother != NULL )
    leftBrother->rightBrother = NULL;
  if( father != NULL )
    if( father->rightmostSon == this )
      father->rightmostSon = NULL;
  if( rightmostSon != NULL )
    rightmostSon->father = NULL;

  rightBrother = NULL;
  leftBrother  = NULL;
  father       = NULL;
  rightmostSon = NULL;
}

/////////////////////////////////////////////////////////////////////////////
void RotTreeNode::setLocation( PM_CARTESIAN &locIn )
{
  loc = locIn;
}

/////////////////////////////////////////////////////////////////////////////
void RotTreeNode::setRightMost( RotTreeNode *p )
{
  // set father of p
  p->father = this;
  if( rightmostSon != NULL )
    {
      // set new brother relationship
      p->setLeftOf(rightmostSon);
    }
  rightmostSon = p;
}

/////////////////////////////////////////////////////////////////////////////
void RotTreeNode::setLeftOf( RotTreeNode *p )
{
  // break previous links
  if( leftBrother != NULL )
    leftBrother->rightBrother = NULL;
  if( p->rightBrother != NULL )
    (p->rightBrother)->leftBrother = NULL;
  // add new links
  leftBrother = p;
  p->father = this->father;
  p->rightBrother = this;
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode *RotTreeNode::getRightBrother()
{
  return rightBrother;
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode *RotTreeNode::getLeftBrother()
{
  return leftBrother;
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode *RotTreeNode::getFather()
{
  return father;
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode *RotTreeNode::getRightmostSon()
{
  return rightmostSon;
}

/////////////////////////////////////////////////////////////////////////////
PM_CARTESIAN RotTreeNode::getLocation()
{
  return( loc );
}

/////////////////////////////////////////////////////////////////////////////
RotStack::RotStack(int size)
{
  nodes = new RotTreeNode*[size];
  stackSize = 0;
}

/////////////////////////////////////////////////////////////////////////////
RotStack::~RotStack()
{
  delete [] nodes;
  stackSize = 0;
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode *RotStack::pop()
{
  if( stackSize > 0 )
    stackSize--;
  return(nodes[stackSize]);
}

/////////////////////////////////////////////////////////////////////////////
void RotStack::push( RotTreeNode *in )
{
  nodes[stackSize++] = in;
}

/////////////////////////////////////////////////////////////////////////////
bool RotStack::isEmpty()
{
  if( stackSize > 0 )
    return(false);
  else
    return(true);
}

/////////////////////////////////////////////////////////////////////////////
RotTreeNode *RotStack::top()
{
  return( nodes[stackSize-1] );
}

/////////////////////////////////////////////////////////////////////////////
RotTree::RotTree()
{
  stack = NULL;
  isSorted = false;
}

/////////////////////////////////////////////////////////////////////////////
void RotTree::sort()
{
  qsort( nodes, numNodes, sizeof(RotTreeNode),
	 &rotTreeCompar );
  isSorted = true;
}

/////////////////////////////////////////////////////////////////////////////
void RotTree::load()
{
  PM_CARTESIAN loc;
  nodes = new RotTreeNode[11];
  isSorted = false;
  int nodeCount = 0;

  // set +- infinity nodes
  loc.x = -1000000;
  loc.y = 0;
  nodes[nodeCount++].setLocation(loc);
  loc.x = 1000000;
  nodes[nodeCount++].setLocation(loc);

  // build square
  loc.x = 10;
  loc.y = 10;
  nodes[nodeCount++].setLocation(loc);
  loc.y = 50;
  nodes[nodeCount++].setLocation(loc);
  loc.x = 50;
  nodes[nodeCount++].setLocation(loc);
  loc.y = 10;
  nodes[nodeCount++].setLocation(loc);

  // build triangle
  loc.x = 70;
  loc.y = 70;
  nodes[nodeCount++].setLocation(loc);
  loc.x = 120;
  nodes[nodeCount++].setLocation(loc);
  loc.x = 100;
  loc.y = 120;
  nodes[nodeCount++].setLocation(loc);

  // add additional line
  loc.x = 52;
  loc.y = 52;
  nodes[nodeCount++].setLocation(loc);
  loc.x = 58;
  loc.y = 58;
  nodes[nodeCount++].setLocation(loc);


  numNodes = nodeCount;
}

/////////////////////////////////////////////////////////////////////////////
void RotTree::createTree()
{
  int count;
  RotTreeNode *p, *pr, *q, *z;
  PM_CARTESIAN loc;
  bool special;

  if( !isSorted )
    sort();
  nodes[0].setRightMost(&nodes[numNodes-1]);
  for( count=1; count<numNodes-1; count++ )
    nodes[numNodes-1].setRightMost(&nodes[count]);

  if( stack != NULL )
    delete stack;
  stack = new RotStack(numNodes);
  stack->push( &nodes[1] );

  while( !stack->isEmpty() )
    {
      p = stack->pop();
      pr = p->getRightBrother();
      q = p->getFather();
      
      if( q != &nodes[numNodes-1] )
	{
	  loc = p->getLocation();
	  fprintf( stdout, "L %d %d ",
		   (int)loc.x, (int)loc.y );
	  loc = q->getLocation();
	  fprintf( stdout, "%d %d\n",
		   (int)loc.x, (int)loc.y );
	}
      z = q->getLeftBrother();
      p->remove();
      if( z == NULL )
	special = false;
      else if( z->getFather() == &nodes[0] )
	special = true;
      else
	special = false;
      if( z==NULL || !leftTurn(p, z, (z->getFather())->getLocation(), special) )
	q->setLeftOf( p );
      else
	{
	  if( z == &nodes[0] )
	    special = true;
	  else
	    special = false;
	  while( z->getRightmostSon()!=NULL && leftTurn(p, z->getRightmostSon(),z->getLocation(), special) )
	    z = z->getRightmostSon();
	  z->setRightMost(p);
	  if( z == stack->top() )
	    z = stack->pop();
	}
      if( p->getLeftBrother() == NULL && p->getFather() != &nodes[0] )
	stack->push( p );
      if( pr != NULL )
	stack->push(pr);
    }
}

/////////////////////////////////////////////////////////////////////////////
void RotTree::printSort()
{
  int count;
  PM_CARTESIAN loc;

  for( count=0; count<numNodes; count++ )
    {
      loc = nodes[count].getLocation();
      fprintf( stderr, "%lf %lf\n",
	       loc.x, loc.y );
    }
}

/////////////////////////////////////////////////////////////////////////////
void RotTree::printTree()
{
  int count;
  PM_CARTESIAN locSon, locFather;
  
  fprintf( stderr, "\n\nPrinting tree:\n" );
  for( count=1; count<numNodes-1; count++ )
    {
      if( nodes[count].getFather() != NULL )
	{
	  locSon = nodes[count].getLocation();
	  locFather = (nodes[count].getFather())->getLocation();
	  fprintf( stderr, "L %d %d %d %d\n",
		   (int)locSon.x, (int)locSon.y,
		   (int)locFather.x, (int)locFather.y );
	}
    }
}

/////////////////////////////////////////////////////////////////////////////
bool RotTree::leftTurn( RotTreeNode *p, RotTreeNode *q, PM_CARTESIAN locR, bool special )
{
  PM_CARTESIAN locP, locQ;
  double slopePQ, slopePR;

  locP = p->getLocation();
  locQ = q->getLocation();

  // special case for locR is infinity points
  if( special )
    {
      if(locP.x<locQ.x || (locP.x==locQ.x && locP.y<locQ.y))
	return true;
      else
	return false;
    }
  slopePQ = (locQ.y - locP.y)/(locQ.x - locP.x);
  slopePR = (locR.y - locP.y)/(locR.x - locP.x);

  if( slopePR > slopePQ )
    return true;
  else
    return false;
}
/*! CVS Log Information
 * $Log: visGraph.cc,v $
 * Revision 1.12  2009/01/23 13:47:16  dr_steveb
 * Added printing of destination for named point movement
 *
 * Revision 1.11  2009/01/23 13:37:49  proctor
 * Added more map stuff
 *
 * Revision 1.10  2009/01/22 22:15:55  proctor
 * Used insert and find instead of [] in goalHash map
 *
 * Revision 1.9  2008/12/05 15:58:24  proctor
 * Added copies of the Ulapi interface files; fixed visGraph to use the
 * Standard Template Libraries instead of hcreate etc.;
 *
 * Revision 1.8  2008/03/11 13:17:23  dr_steveb
 * Added code to plot prim commanded arcs in gnuplot.
 *
 * Revision 1.7  2008/02/12 17:47:50  dr_steveb
 * Minor bug fixes
 *
 * Revision 1.6  2008/01/07 21:49:28  dr_steveb
 * Manufacturing competition version of the vehicle controller
 *
 * Revision 1.5  2007/12/20 02:32:45  dr_steveb
 * Now have two vehicles performing start of manufacturing job. More to come...
 *
 * Revision 1.4  2007/12/13 13:18:40  dr_steveb
 * Changes for vehicle level visibility graph planning.
 *
 * Revision 1.3  2007/11/08 13:21:59  dr_steveb
 * Not done yet, but premature checking in of other files means these have to go to!
 *
 * Revision 1.2  2007/11/07 21:19:22  dr_steveb
 * Checked-in to allow for compliation
 *
 * Revision 1.1.1.1  2005/10/12 20:50:32  root
 * initial import
 *
 * Revision 1.4  2005/07/21 18:59:11  stephen
 * Fixed merged conflicts and added visibility graph stuff.
 *
 * Revision 1.3  2005/07/05 12:04:22  stephen
 * Tree now working? in visGraph.
 *
 * Revision 1.2  2005/07/01 19:53:04  stephen
 * Rotation tree now working.
 *
 * Revision 1.1  2005/06/30 20:20:41  stephen
 * Initial check-in
 *
 *
 */
