/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primMobEngine.cc

  \brief Vehicle primitive level planning engine
  This code used to reside in primMobMain.cc but has been pulled out to allow
  for multiple uses of the prim planner.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.12 $
  $Date: 2009/02/27 21:08:13 $
  \endcode

  \author Stephen Balakirsky & Fred Proctor
*/

#include <stdio.h>		// printf, vprintf
#include <stdarg.h>		// va_list, va_start, va_arg, va_end
#include <rcs.hh>		// etime
#include "primMobJA.hh"		// PRIM_MOB_JA_DEBUG_XXX
#include "primMobEngine.hh"

PrimMobEngine::PrimMobEngine()
{
  w = 0;
  v = 0;
  wheel_radius = 1.0;
  wheel_radius_inverse = 1.0;
  wheel_separation = 1.0;
  wheel_separation_half = 0.5;
  waypoint_size = WAYPOINT_INITIAL_SIZE;	// size of whole array
  waypoint_num = 0;	// how many currently in the array
  waypoint_index = 0;	// which one we're working on
  log_fp = NULL;
  setCycleTime = DEFAULT_CYCLE_TIME;
  setVmax = DEFAULT_MAX_TRAN_VEL;
  setAmax = DEFAULT_MAX_TRAN_ACC;
  setWmax = DEFAULT_MAX_ROT_VEL;
  setAlphamax = DEFAULT_MAX_ROT_ACC;
  setVCutoffAngle = DEFAULT_V_CUTOFF_ANGLE;
  setWCutoffAngle = DEFAULT_W_CUTOFF_ANGLE;
  setControlPoint = PM_POSE(PM_CARTESIAN(0,0,0),PM_QUATERNION(1,0,0,0));
  setDebug = 0;
  in_rev=false;

  // build the initial waypoint array, using 'malloc' since a 'realloc'
  // may be done later
  waypoint_array =
  waypoint_array = (Waypoint *) malloc(WAYPOINT_INITIAL_SIZE * sizeof(Waypoint));
  if(NULL == waypoint_array) 
    {
      rcs_print_error("primMobEngine: can't allocate waypoint array\n");
      exit(1);
    }
  // build the initial wpsToArcs array, using 'malloc' since a 'realloc'
  // may be done later
  wpsToArcs = (int *) malloc(WAYPOINT_INITIAL_SIZE * sizeof(int));
  if(NULL == wpsToArcs) 
    {
      rcs_print_error("primMobEngine: can't allocate wpsToArcs array\n");
      exit(1);
    }
  useAckerman = false;
  ackermanWheelbase = 1.0;
  useNautic = false;
  nauticLength = 1.0;

  prim_path = false;
}

// most state machines return true on success, false on error
RCS_STATUS PrimMobEngine::doCmdNop()
{
  return RCS_DONE;
}

RCS_STATUS PrimMobEngine::doCmdInit()
{
  return RCS_DONE;
}

// FIXME-- gradually stop
RCS_STATUS PrimMobEngine::doCmdHalt()
{
  return RCS_DONE;
}

RCS_STATUS PrimMobEngine::doCmdAbort()
{
  return RCS_DONE;
}

RCS_STATUS PrimMobEngine::doCmdShutdown()
{
  return RCS_DONE;
}

/* the return values retV1 and retV2 mean the following:
 * for ackerman: retV1 is velocity and retV2 is steerAngle
 * for nautic: retV1 is propeller, retV2 is rudder, retV3 is sternPlane
 * for skid: retV1 is wLeft and retV2 is wRight
 */
RCS_STATUS PrimMobEngine::doCmdMoveWaypoint( RCS_STATE state,
					     const PM_CARTESIAN &tranRate, 
					     const PM_RPY &rpyRate,
					     const PM_CARTESIAN &tranIn,
					     const PM_RPY &rpyIn,
					     // the rest of the arguments are return values
					     int &pathIndex,
					     int &nextState,
					     double &retV1, double &retV2, double &retV3)
{
  //  static double v = 0.0;
  //  static double w = 0.0;
  double pitchdot = 0.0;
  PM_CARTESIAN tran;
  PM_RPY rpy;
  PM_POSE pose;
  PM_CARTESIAN target;
  PM_CARTESIAN here_to_target;
  PM_CARTESIAN target_to_next;

  PM_CARTESIAN icr;
  PM_CARTESIAN lcenter, rcenter;
  PM_ROTATION_VECTOR rotvec;

  double target_distance;
  double distance_to_go;
  double vmax;
  double here_yaw, goal_yaw, dyaw, out_yaw;
  double here_pitch, goal_pitch, dpitch;
  double vnew, wnew, pitchdotnew;
  double vp, vc;
  double delta_t;
  bool backing_up;
  RCS_STATUS retValue = RCS_EXEC;

  delta_t = setCycleTime;
  nextState = state;

  /*
    We have nav data as
	
    . 0
    .  T
    Veh

    and want the pose of the control point as

    . 0
    .  T
    Ctl
	
    given the pose of the control point naturally wrt the vehicle frame

    Veh
    .  T
    Ctl

    so we need to do

    . 0      0    Veh
    .  T =    T *    T
    Ctl    Veh    Ctl
  */

  // set 'pose' to be nav data
  pose.tran = tranIn;
  pose.rot = rpyIn;
  // postmultiply by control pose wrt nav data to get control point
  pose = pose * setControlPoint;
  // fill 'tran' and 'rot' back in as their shifted values
  tran = pose.tran;
  rpy = pose.rot;
  // and use 'em as before

  switch (state) {
    case NEW_COMMAND:
      v = HYPOT(tranRate.x, tranRate.y);
      w = rpyRate.y;
      retValue = RCS_EXEC;
      nextState = S0;
      break;
    case S0:
      here_yaw = rpy.y;
      NORM(here_yaw);
      here_pitch = rpy.p;
      NORM(here_pitch);
      // check if we're at the end
      for (; waypoint_index < waypoint_num; waypoint_index++) {
	vmax = waypoint_array[waypoint_index].speed;
	if (vmax < 0.0) {
	  /*
	    New! If the speed to the waypoint is negative, it means we
	    are to back up. This is done by rotating our heading halfway
	    around, and negating our translational speed.
	  */
	  backing_up = true;
	  vmax = -vmax;		// make it positive
	  // and switch our head with our tail
	  here_yaw += PM_PI;
	  NORM(here_yaw);
	  // we'll still need to negate our output speed
	} else {
	  backing_up = false;
	}
	target = waypoint_array[waypoint_index].p;
	if (vmax > setVmax) vmax = setVmax;
	here_to_target = target - tran;
	target_distance = magxy(here_to_target);
	if (target_distance < waypoint_array[waypoint_index].neighborhood) {
	  // we're within the neighborhood, so increment and drop it
	  printDebugMsg(PRIM_MOB_JA_DEBUG_WAYPOINT, "did waypoint %d of %d\n",
		waypoint_index + 1, waypoint_num);
	  continue;
	}
	
	/* FIXME: wrap this with isAckerman */
	/* not close enough, can we make the point? */
	rotvec.s = here_yaw - (PM_PI/2), rotvec.x = rotvec.y = 0, rotvec.z = 1;
	/* FIXME: ICR should be dynmic in object init - we hard code it */
	icr.x=.4, icr.y=0;

	rcenter = tran + rotvec*icr;
	lcenter = tran - rotvec*icr;

	if (setDebug & PRIM_MOB_JA_DEBUG_ICR) {
	  rotvec.s = here_yaw, rotvec.x = rotvec.y = 0, rotvec.z = 1;
	  PM_CARTESIAN tol;
	  tol.x=waypoint_array[waypoint_index].neighborhood, tol.y=0;
	  PM_CARTESIAN skip;
	  skip = tran + rotvec*tol;

	  FILE *fh_icr = fopen ("icr.dat", "w");
	  fprintf (fh_icr, "%lf %lf\n%lf %lf\n%lf %lf\n%lf %lf\n",
	    rcenter.x, rcenter.y, lcenter.x, lcenter.y,
	    skip.x, skip.y, rcenter.x, rcenter.y);
	  fclose (fh_icr);
	}

	goal_yaw = atan2(here_to_target.y, here_to_target.x);
	dyaw = goal_yaw - here_yaw;
	NORM(dyaw);
	//printf ("dyaw %lf\n", dyaw);

	if (in_rev) {
	    if (.1 > dyaw && dyaw > -.1) {
	//	printf ("go forward\n");
		in_rev=false;
		break;
	    }
	    retV1 = -vmax;
	    retV2 = (right_rev) ? -setWmax : setWmax;
	    return retValue;

	}

  	if (isPrimPath()) {
	 if ((magxy(target-rcenter) < magxy(icr))) {
	  in_rev = true;
	  right_rev = true;
	  retV1 = -vmax;
	  retV2 = -setWmax;
	  return retValue;
	 }

	 if ((magxy(target-lcenter) < magxy(icr))) {
	  in_rev = true;
	  right_rev = false;
	  retV1 = -setVmax;
	  retV2 = setWmax;
	  return retValue;
	 }
	}
	/* FIXME: end wrap this with isAckerman ??? */

	break;
      }

      // now we're on the next waypoint

      /*
	Calculate the straight-line polygon distance from here to the
	last waypoint, as 'distance_to_go'. Set this to 'target_distance'
	initially, which is the distance from here to the next waypoint.
	For all the waypoints from the next waypoint to the end, add
	those distances.
       */
      distance_to_go = target_distance;
      for (int t = waypoint_index; // from next waypoint
	   t < waypoint_num - 1; // to next-to-last waypoint
	   t++) {
	// add distance between this and the next
	distance_to_go += magxy(waypoint_array[t].p - waypoint_array[t + 1].p);
      }

      if (waypoint_index >= waypoint_num) { // if true, should always be ==
	// we've arrived, set pathIndex to index of last arc
	pathIndex = wpsToArcs[waypoint_index - 1];
	nextState = S1;
      } else {
	pathIndex = wpsToArcs[waypoint_index];
	goal_yaw = atan2(here_to_target.y, here_to_target.x);
	NORM(goal_yaw);
	goal_pitch = atan2(-here_to_target.z, target_distance);
	NORM(goal_pitch);
	// is this the last waypoint?
	if (waypoint_index == waypoint_num - 1) {
	  out_yaw = 0.0;
	} else {
	  target_to_next = waypoint_array[waypoint_index + 1].p - target;
	  out_yaw = atan2(target_to_next.y, target_to_next.x) - goal_yaw;
	}
	// go to this waypoint
	dyaw = goal_yaw - here_yaw;
	NORM(dyaw);
	dpitch = goal_pitch - here_pitch;
	NORM(dpitch);

	vnew = vmax * (1.0 - fabs(dyaw) / setVCutoffAngle);
	if (vnew < 0.0) vnew = 0.0;
	vp = vmax * (1.0 - fabs(out_yaw) / setVCutoffAngle);
	if (vp < 0.0) vp = 0.0;

	/*
	  Slow down as we approach the waypoint, either only for the
	  last waypoint, or for each waypoint, depending on whether we
	  use 'target_distance' or 'distance_to_go'.

	  'vc' is the speed above which we would overshoot the
	  waypoint given our distance away and our max decel.
	  Speeds are clamped to this.
	*/
	vc = sqrt(2.0 * setAmax * 
		  distance_to_go + // target distance or distance_to_go
		  vp * vp);

	if (vnew > vc)
	  vnew = vc;
	if (vnew > v + setAmax * delta_t)
	  vnew = v + setAmax * delta_t;
	else if (vnew < v - setAmax * delta_t)
	  vnew = v - setAmax * delta_t;
	v = (vnew < 0.0 ? 0 : vnew);
	
	wnew = setWmax / setWCutoffAngle * dyaw;
	if (wnew > setWmax) wnew = setWmax;
	else if (wnew < -setWmax) wnew = -setWmax;
	if (wnew > w + setAlphamax * delta_t)
	  wnew = w + setAlphamax * delta_t;
	else if (wnew < w - setAmax * delta_t)
	  wnew = w - setAlphamax * delta_t;
	w = wnew;

	pitchdotnew = setWmax / setWCutoffAngle * dpitch;
	if (pitchdotnew > setWmax) pitchdotnew = setWmax;
	else if (pitchdotnew < -setWmax) pitchdotnew = -setWmax;
	if (pitchdotnew > pitchdot + setAlphamax * delta_t)
	  pitchdotnew = pitchdot + setAlphamax * delta_t;
	else if (pitchdotnew < pitchdot - setAmax * delta_t)
	  pitchdotnew = pitchdot - setAlphamax * delta_t;
	pitchdot = pitchdotnew;

	if (useAckerman) {
	  if (backing_up) {
	    retV1 = -v;
	    retV2 = atan2(w * ackermanWheelbase, v);
	  } else {
	    retV1 = v;
	    // minus sign necessary for vehicle conventions
	    retV2 = -atan2(w * ackermanWheelbase, v);
	  }
	} else if (useNautic) {
	  /*
	    The sub won't have angular speed without linear speed. If
	    the linear speed low, we need to boost it up to get angular
	    speed.
	  */
	  if (v < fabs(w)) v = fabs(w);
	  retV1 = v;
#define RUDDER_SCALE      1.0
#define STERN_PLANE_SCALE -10.0
	  retV2 = RUDDER_SCALE * atan2(w * nauticLength, v);
	  retV3 = STERN_PLANE_SCALE * atan2(pitchdot * nauticLength, v);
	} else {
	  if (backing_up) {
	    retV1 = -wheel_radius_inverse * (v - wheel_separation_half * w);
	    retV2 = -wheel_radius_inverse * (v + wheel_separation_half * w);
	  } else {
	    retV1 = wheel_radius_inverse * (v + wheel_separation_half * w);
	    retV2 = wheel_radius_inverse * (v - wheel_separation_half * w);
	  }
	}
      }
      break;
    case S1:		// S1 = stopping
    // done
      if (useAckerman) {
	retV1 = 0;
	retV2 = 0;
	retV3 = 0;
      } else if (useNautic) {
	retV1 = 0;
	retV2 = 0;
	retV3 = 0;
      } else {
	retV1 = 0;
	retV2 = 0;
	retV3 = 0;
      }
      nextState = S2; // done
      retValue = RCS_DONE;
      break;
    default:
      break;
    }

  printDebugMsg(PRIM_MOB_JA_DEBUG_MOVE,
		"position is %f %f %f\n", 
		(double) tran.x,
		(double) tran.y,
		(double) rpy.y);

  return retValue;
}

// FIXME-- the rotate command isn't valid for Ackerman vehicles.
/* the return values retV1 and retV2 mean the following:
 * for ackerman: retV1 is velocity and retV2 is steerAngle
 * for nautic: retV1 is propeller, retV2 is rudder, retV3 is sternPlane
 * for skid: retV1 is wLeft and retV2 is wRight
 */
RCS_STATUS PrimMobEngine::doCmdRotate(RCS_STATE state, double cmdTheta, double cmdThetaTol, int cmdDirection,
				      const PM_RPY &rpy,
				      // the rest of the arguments are return values
				      int &pathIndex,
				      int &nextState,
				      double &retV1, double &retV2 )
{
  double here_theta;
  double dtheta;
  double delta_t;
  double wnew;
  RCS_STATUS retValue = RCS_EXEC;

  delta_t = setCycleTime;
  nextState = state;
  switch( state )
    {
    case NEW_COMMAND:
      pathIndex = 0;
      nextState = S0;
      break;
    case S0:
      here_theta = rpy.y;
      dtheta = cmdTheta - here_theta;
    if (dtheta >= PM_PI) dtheta -= PM_2_PI;
    else if (dtheta <= -PM_PI) dtheta += PM_2_PI;
    // check if we're at the end
    if (fabs(dtheta) < cmdThetaTol) {
      // we've arrived
      retValue = RCS_DONE;
      nextState = S1;
    } else {
      wnew = setWmax / PM_PI * dtheta;
      if (cmdDirection == MOAST_ROTATE_CW && wnew < 0.0) {
	wnew = -wnew;
      } else if (cmdDirection == MOAST_ROTATE_CCW && wnew > 0.0) {
	wnew = -wnew;
      } 
      // else leave wnew alone, taking shortest route
      if (wnew > w + setAlphamax * delta_t) {
	wnew = w + setAlphamax * delta_t;
      } else if (wnew < w - setAmax * delta_t) {
	wnew = w - setAlphamax * delta_t;
      }
      w = wnew;
      retV1 = wheel_radius_inverse * (wheel_separation_half * w);
      retV2 = wheel_radius_inverse * (-wheel_separation_half * w);
    }
    break;
    case S1: // S1 = stopping
      retV1 = 0;
      retV2 = 0;
      retValue = RCS_DONE;
      nextState = S2;
      break;
    default: // done
      break;
    }
  return retValue;
}

/* the return values retV1 and retV2 mean the following:
 * for ackerman: retV1 is velocity and retV2 is steerAngle
 * for nautic: retV1 is propeller, retV2 is rudder, retV3 is sternPlane
 * for skid: retV1 is wLeft and retV2 is wRight
 */
RCS_STATUS PrimMobEngine::doCmdVel(RCS_STATE state, double cmdV, double cmdW,
				    // the rest of the arguments are return values
				    int &nextState,
				    double &retV1, double &retV2, double &retV3 )
{
  RCS_STATUS retValue;

  retValue = RCS_EXEC;
  nextState = state;
  switch( state )
    {
    case NEW_COMMAND:
      nextState = S0;
      break;
    case S0:
      if (useAckerman) {
	retV1 = cmdV;
	// minus sign necessary for vehicle conventions
	retV2 = -atan2(cmdW * ackermanWheelbase, cmdV);
      } else if (useNautic) {
	retV1 = cmdV;
	retV2 = atan2(cmdW * nauticLength, cmdV);
	retV3 = 0.0;
      } else {
	retV1 = wheel_radius_inverse * (cmdV + wheel_separation_half * cmdW);
	retV2 = wheel_radius_inverse * (cmdV - wheel_separation_half * cmdW);
      }
      nextState = S1;
      retValue = RCS_DONE;
      break;
    default:
      retValue = RCS_DONE;
    }
  return retValue;
}

RCS_STATUS PrimMobEngine::doCmdStop(RCS_STATE state,
				    // the rest of the arguments are return values
				    int &nextState,
				    double &retV1, double &retV2, double &retV3 )
{
  RCS_STATUS retValue = RCS_EXEC;

  nextState = state;
  switch( state )
    {
    case NEW_COMMAND:
      nextState = S1;
      break;
    case S1:
      retV1 = 0;
      retV2 = 0;
      retV3 = 0;
      nextState = S2;
      break;
    case S2:
      retValue = RCS_DONE;
      nextState = S0;
      break;
    case S0:
    default:
      retValue = RCS_DONE;
      break;
    }
  return retValue;
}

RCS_STATUS PrimMobEngine::doCfgNop(RCS_STATE state, int &nextState)
{
  RCS_STATUS retValue = RCS_DONE;

  switch( state )
    {
    case NEW_COMMAND:
      nextState = S0;
      break;
    default:
      break;
  }
  return retValue;
}

RCS_STATUS PrimMobEngine::doCfgCycleTime(RCS_STATE state, int &nextState, double cycleTimeIn, double &cycleTimeOut)
{
  RCS_STATUS retValue = RCS_DONE;

  switch( state )
    {
    case NEW_COMMAND:
      if (cycleTimeIn <= 0.0) 
	retValue = RCS_ERROR;
      else 
	{
	  setCycleTime = cycleTimeIn;
	  cycleTimeOut = cycleTimeIn;
	}
      nextState = S0;
      break;
    default:
      break;
    }
  return retValue;
}

RCS_STATUS PrimMobEngine::doCfgMove(RCS_STATE state, int &nextState,
				    double vmaxIn, double amaxIn, double wmaxIn, double alphamaxIn, 
				    double vCutoffAngleIn, double wCutoffAngleIn,
				    // returned values
				    double &vmaxOut, double &amaxOut, double &wmaxOut, double &alphamaxOut, 
				    double &vCutoffAngleOut, double &wCutoffAngleOut)
{
  RCS_STATUS retValue = RCS_DONE;

  switch( state )
    {
    case NEW_COMMAND:
      nextState = S0;
      vmaxOut = vmaxIn;
      amaxOut = amaxIn;
      wmaxOut = wmaxIn;
      alphamaxOut = alphamaxIn;
      vCutoffAngleOut = vCutoffAngleIn;
      wCutoffAngleOut = wCutoffAngleIn;
      setVmax = vmaxIn;
      setAmax = amaxIn;
      setWmax = wmaxIn;
      setAlphamax = alphamaxIn;
      setVCutoffAngle = vCutoffAngleIn;
      setWCutoffAngle = wCutoffAngleIn;
      break;
    default:
      break;
    }
  return retValue;
}

/*
  Each time a debug config message is received, any previously opened
  log file is closed, and a new one is opened. The names of log files
  are

  <PRIM_MOB_ENGINE_LOG_FILE_PREFIX>-<debug><PRIM_MOB_ENGINE_LOG_FILE_SUFFIX>

  e.g., "primMobEngineLog-3.log" for a debug mask of (decimal) 3.
 */

RCS_STATUS PrimMobEngine::doCfgDebug(RCS_STATE state, int &nextState, int debugIn, int &debugOut )
{
  char filename[PRIM_MOB_ENGINE_LOG_FILE_NAME_LEN];
  RCS_STATUS retValue = RCS_DONE;

  switch (state) {
  case NEW_COMMAND:
    setDebug = debugIn;
    debugOut = debugIn;
    if (NULL != log_fp) {
      fclose(log_fp);
      log_fp = NULL;
    }
    if (setDebug != 0) {
      sprintf(filename, "%s-%d%s", 
	      PRIM_MOB_ENGINE_LOG_FILE_PREFIX,
	      setDebug,
	      PRIM_MOB_ENGINE_LOG_FILE_SUFFIX);

      log_fp = fopen(filename, "w");
      if (NULL == log_fp) {
	retValue = RCS_ERROR;
      }
    }
    nextState = S0;
    break;

  default:			// S0
    break;
  }

  return retValue;
}

void PrimMobEngine::setAckermanWheelbase(double setTo)
{
  if (IS_POSITIVE(setTo)) 
    {
      useAckerman = true;
      useNautic = false;
      ackermanWheelbase = setTo;
      rcs_print("setAckermanWheelbase: using Ackerman steering, wheelbase %f\n", ackermanWheelbase);
    } 
  else 
    {
      rcs_print_error("setAckermanWheelbase: bad USE_ACKERMAN wheelbase %f\n", ackermanWheelbase);
      useAckerman = false;
    }
}

void PrimMobEngine::setNauticLength(double setTo)
{
  if (IS_POSITIVE(setTo)) 
    {
      useNautic = true;
      useAckerman = false;
      nauticLength = setTo;
      rcs_print("setNauticLength: using Nautic steering, length %f\n", nauticLength);
    } 
  else 
    {
      rcs_print_error("setNauticLength: bad USE_NAUTIC length %f\n", nauticLength);
      useNautic = false;
    }
 }

bool PrimMobEngine::isAckerman()
{
  return useAckerman;
}

bool PrimMobEngine::isNautic()
{
  return useNautic;
}

void PrimMobEngine::setWheelSettings( double radius, double separation )
 {
   wheel_radius = radius;
   wheel_separation = separation;
   if (! IS_SMALL(wheel_radius) &&
       ! IS_SMALL(wheel_separation)) 
     {
       wheel_radius_inverse = 1.0 / wheel_radius;
       wheel_separation_half = 0.5 * wheel_separation;
     } 
   else 
     {
       // not found, using defaults
       wheel_radius_inverse = 1.0 / 0.0825;
       wheel_separation_half = 0.5 * 0.38;
       rcs_print_error("wheel radius, separation not valid, using defaults\n");
     }
 }

void PrimMobEngine::resetWaypointIndex()
{
  waypoint_index = 0;
}

void PrimMobEngine::setWaypoints( Waypoint* waypoint, int length)
{
  for (waypoint_num = 0; waypoint_num < length; waypoint_num++) 
    {
      waypoint_array[waypoint_num] = waypoint[waypoint_num];
    }
  waypoint_index = 0;
}

/*
  We use magxy() for the Cartesian magnitude of a distance vector
  to ignore any Z difference, since we're in a plane.
*/
inline double PrimMobEngine::magxy(PM_CARTESIAN v)
{
  return sqrt(v.x*v.x + v.y*v.y);
}

/*
  arc_segments_to_waypoints() takes a nominal start position 'start',
  an array of ArcSegments 'arcs', how many they are 'arc_num',
  a pointer to a pointer to an array of waypoints 'wps', and a pointer
  to the number of slots available, 'wp_num'.

  It converts the arcs to waypoints using the annular_tol of the arc
  to determine how many points to pepper the arc with. 

  If the space allocated in 'wps' runs out (need more than 'wp_num'),
  the waypoint array is reallocated, hence the need for a ptr to a ptr.
  If this happens, 'wp_num' is updated.
  The wpsToArcs map has the same number of entries as the waypoint
  array and is updated at the same time.

  Return value: number of waypoints created, or -1 if there is some error.
*/

int PrimMobEngine::arc_segments_to_waypoints(ArcSegment * arcs, int arc_num)
{
  PM_CARTESIAN rce;
  PM_ROTATION_VECTOR rotvec;
  PM_CARTESIAN start;
  PM_CARTESIAN path, scratch;
  int arc_index = 0;
  int wp_index = 0;

  if (arc_num <= 0) {
    waypoint_num = 0;
    return waypoint_num;
  }

  rce = arcs[0].end - arcs[0].center;
  rotvec.s = -arcs[0].theta, rotvec.x = rotvec.y = 0, rotvec.z = 1;
  start = arcs[0].center + rotvec * rce;

  for (arc_index = 0; arc_index < arc_num; arc_index++) {
    ArcSegment arc = arcs[arc_index];

    if (arc.isArc) {
      PM_CARTESIAN rvec = start - arc.center;
      PM_CARTESIAN out;
      PM_ROTATION_VECTOR rot;
      double radius = mag(rvec);
      double dth = 1.0 * arc.annular_tol / radius; // FIXME-- testing
      double theta = fabs(arc.theta);
      double segs = theta / dth;
      double t;

      printDebugMsg(PRIM_MOB_JA_DEBUG_ARCS,
		    "got arc start %f %f center %f %f end %f %f theta %f tol %f \n", 
		    start.x, start.y,
		    arc.center.x, arc.center.y, arc.end.x, arc.end.y,
		    arc.theta, arc.annular_tol);

      segs = ceil(segs);
      dth = theta / segs;
      rot.x = rot.y = 0;
      rot.z = arc.theta > 0 ? 1 : -1;

      for (t = 0; t <= theta; t += dth) {
	if (wp_index == waypoint_size) {
	  waypoint_size *= 2;
	  waypoint_array = (Waypoint *) realloc(waypoint_array, waypoint_size * sizeof(Waypoint));
	  if (NULL == waypoint_array) {
	    waypoint_num = -1;
	    return waypoint_num;
	  }
	  wpsToArcs = (int *) realloc(wpsToArcs, waypoint_size * sizeof(int));
	  if (NULL == wpsToArcs) {
	    waypoint_num = -1;
	    return waypoint_num;
	  }
	}
	rot.s = t;
	out = arc.center + rot * rvec;
	waypoint_array[wp_index].p.x = out.x;
	waypoint_array[wp_index].p.y = out.y;
	waypoint_array[wp_index].neighborhood = arc.annular_tol;
	waypoint_array[wp_index].speed = arc.speed;
	waypoint_array[wp_index].quality = 1; // FIXME-- no quality in an ArcSegment
	wpsToArcs[wp_index] = arc_index;
	/*printDebugMsg(PRIM_MOB_JA_DEBUG_WAYPOINT,
		      "generated waypoint %f %f %f %f\n", 
		      (double) waypoint_array[wp_index].p.x,
		      (double) waypoint_array[wp_index].p.y,
		      (double) waypoint_array[wp_index].neighborhood,
		      (double) waypoint_array[wp_index].speed);*/
	wp_index++;
      }
    } else {
      // it's a line, just add it
      if (wp_index == waypoint_size) {
	waypoint_size *= 2;
	waypoint_array = (Waypoint *) realloc(waypoint_array, waypoint_size * sizeof(Waypoint));
	if (NULL == waypoint_array) {
	    waypoint_num = -1;
	    return waypoint_num;
	}
	wpsToArcs = (int *) realloc(wpsToArcs, waypoint_size * sizeof(int));
	if (NULL == wpsToArcs) {
	    waypoint_num = -1;
	    return waypoint_num;
	}
      }
#define SHORT 1
      //scale back the path by SHORT meter(s)
      path.x = start.x - arc.end.x;
      path.y = start.y - arc.end.y;
      if (magxy(path)>SHORT) {
        scratch = start - ((magxy(path)-SHORT)*norm(path));
        waypoint_array[wp_index].p.x = scratch.x;
        waypoint_array[wp_index].p.y = scratch.y;
      } else {
	waypoint_array[wp_index].p.x = arc.end.x;
	waypoint_array[wp_index].p.y = arc.end.y;
      }
      waypoint_array[wp_index].p.z = arc.end.z;
      waypoint_array[wp_index].neighborhood = arc.annular_tol;
      waypoint_array[wp_index].speed = arc.speed;
      waypoint_array[wp_index].quality = 1; // FIXME-- no quality in an ArcSegment
      wpsToArcs[wp_index] = arc_index;
      wp_index++;
    }

    start = arc.end;
  }

  /* This makes sure the bot starts at start for a line segment */
  waypoint_array[wp_index-1].p.x = start.x;
  waypoint_array[wp_index-1].p.y = start.y;

  waypoint_num = wp_index;

  if (setDebug & PRIM_MOB_JA_DEBUG_WAYPOINT) {
    FILE *wps = fopen ("wps.dat", "w");
    for (wp_index=0; wp_index< waypoint_num; wp_index++) {
	fprintf (wps, "%lf %lf\n",
	     waypoint_array[wp_index].p.x, waypoint_array[wp_index].p.y);
    }
    fclose (wps);
  }
  return waypoint_num;
}

void PrimMobEngine::setSettings(double cycleTimeIn,
				double vmaxIn,
				double amaxIn,
				double wmaxIn,
				double alphamaxIn,
				double vCutoffAngleIn,
				double wCutoffAngleIn,
				PM_POSE controlPoint,
				int debugIn)
{
  setCycleTime = cycleTimeIn;
  setVmax = vmaxIn;
  setAmax = amaxIn;
  setWmax = wmaxIn;
  setAlphamax = alphamaxIn;
  setVCutoffAngle = vCutoffAngleIn;
  setWCutoffAngle = wCutoffAngleIn;
  setControlPoint = controlPoint;
  setDebug = debugIn;
}

void PrimMobEngine::printDebugMsg(int mask, const char * fmt, ...)
{
  va_list ap;
  FILE * fp;

  if (mask == (mask & setDebug)) {
    fp = NULL != log_fp ? log_fp : stdout;
    fprintf(fp, "%f\t", etime());
    va_start(ap, fmt);
    vfprintf(fp, fmt, ap);
    va_end(ap);
  }
}

bool PrimMobEngine::isPrimPath()
{
  return prim_path;
}

void PrimMobEngine::setPrimPath(bool primpath)
{
  prim_path = primpath;
}

/*
  Modification history:

  $Log: primMobEngine.cc,v $
  Revision 1.12  2009/02/27 21:08:13  dr_steveb
  Fixed docking command and prim now issues "done"

  Revision 1.11  2008/10/13 18:26:39  ghenson
  Update to prim. Only sent the robot into reverse if prim generated the path

  Revision 1.10  2008/05/22 18:45:35  ghenson
  should have been two commits. Waypoints are generated slightly shorter to
  allow robot room to turn. Also, backup if you can not reach a point do to
  the ICR of the bot.

  Revision 1.9  2008/02/21 21:47:10  proctor
  Fixed driving backward with skid steering

  Revision 1.8  2008/02/21 21:19:06  proctor
  Added backup capability, with negative waypoint speed signifying reverse

  Revision 1.7  2008/02/06 14:56:45  dr_steveb
  Added support for additional debug flags and performed array boundary checks

  Revision 1.6  2008/01/08 20:18:16  proctor
  Added control point transform that lets the prim mobility controller act
  on the rotational center of the vehicle, which in general won't coincide
  with the nav data frame

  Revision 1.5  2007/12/19 18:36:38  proctor
  Fixed logging so that logs are created for each debug config message, and
  given the debug mask as part of the log file name.
  Changed the prim waypoint algorithm to slow down only when approaching the
  last waypoint. This should allow it to get up to speed when moving along
  intermediate waypoints.

  Revision 1.4  2007/12/19 14:48:40  proctor
  Added printDebugMsg to primMobEngine, and pulled the waypoint debug printing
  into this. Set the debug level to PRIM_MOB_JA_DEBUG_WAYPOINT, 0x02, to
  enable debug printing of waypoint information. The debug level is set to 0
  in moast.ini by default, so you will see no debug information.

  Revision 1.3  2007/12/18 21:34:46  proctor
  In prim mob, read out the platform name from servo mob and used it to
  get platform-specific parameters V_CUTOFF_ANGLE et al; changed moast.ini
  accordingly; fixed a return value error in the segment sweep code, where
  anyIntersections doesn't really return anything

  Revision 1.2  2007/05/16 15:20:09  dr_steveb
  Finished prim decomposition for traffic simulation

  Revision 1.1  2007/05/10 13:18:28  dr_steveb
  Separated the prim mobility planning engine from the interface wrappers. This will allow a controller to be created (the traffic controller) that controls a large number of vehicles without the overhead of individual NML channels for each vehicle. No changes in the algorithms were made.

*/

