/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primSPMain.cc

  \brief Vehicle primitive level mission package controller.

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.22 $
  $Date: 2008/02/11 23:56:36 $
  \endcode

  \author Fred Proctor

The intended method of terminating this process is to use "kill -INT"
to send it SIGINT, which (because of the "signal(SIGINT, quit);" line
in the source code) will cause the "quit" function to run, setting
the "done" global variable to 1. That will be caught by one of the
two "while (!done)" loops and result in an orderly exit.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#include <math.h>               // sin(), cos()
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>               // INIFILE, etc.
#include "moastTypes.hh"
#include "moastNmlOffsets.hh"
#include "servoSP.hh"           // ServoSPStat, ServoSPCmdAbort, etc.
#include "primSP.hh"            // PrimSPStat. etc.
#include "navDataExt.hh"        // NavDataExt
#include "sensorData.hh"        // SensorDataList, SensorData1D
#include "posemath.h"

/***********************************************************************/

// NML Channels

//! servo SP channels
static RCS_CMD_CHANNEL *servoSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *servoSPStatBuf = NULL;
static RCS_STAT_CHANNEL *servoSPSetBuf = NULL;

//! prim SP channels
static RCS_CMD_CHANNEL *primSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *primSPStatBuf = NULL;
static RCS_CMD_CHANNEL *primSPCfgBuf = NULL;
static RCS_STAT_CHANNEL *primSPSetBuf = NULL;

//! navigation data and sensor data channels
static NML *navDataExtBuf = NULL;
static NML *sensorData1DBuf = NULL;
static NML *sensorDataListBuf = NULL;


/***********************************************************************/

// Other global variables

/*

processName is the actual name of the process, while THIS_PROCESS is
the name used in the NML config file and, hence, required by NML.

*/

static int              done = 0;           // done flag, set by quit function
static char *           iniFileName = NULL; // name of ini file
static int              lastCmdGo;          // 1=go cmd is executing, 0=not
static NavDataExt *     nav;                // nav data
static char *           nmlFile = NULL;     // name of NML file
static SensorDataList   output;             // output data
static double           outputTime;  // time interval between outputs
static char *           processName; // name of this process
static PM_QUATERNION    senQuatVeh;  // sensor's rotation on vehicle
static PM_CARTESIAN     senTranVeh;  // sensor's location on vehicle
static int              servoCmdNumber; // number of latest command to servo
                                        // name for servo data buffer
static char             servoDataBufName[MOAST_NML_BUFFER_NAME_LEN];
static PrimSPSet        set;         // config command status message
static PrimSPStat       stat;        // task command status message
static RCS_TIMER *      timer = NULL;// timer for cycle time
static PM_QUATERNION    vehQuat;     // vehicle's rotation in local space
static PM_RPY           vehRpy;      // vehicle's roll-pitch-yaw in local space
static PM_CARTESIAN     vehTran;     // vehicle's location in local space
static int              verbose = 1; // 1=print copious messages, 0=not
static bool             isAirRobot = false; // Set by getoptions.

/***********************************************************************/

#define DEFAULT_CYCLE_TIME 0.010
#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"
#define THIS_PROCESS "primSP"

#ifndef INI_OK
#define INI_OK      0
#define INI_DEFAULT 1
#define INI_INVALID 2
#endif

/***********************************************************************/

static void cleanupNmlBuffers();
static void cleanupOtherNmlBuffers();
static void cleanupPrimSPNmlBuffers();
static void cleanupServoSPNmlBuffers();
static void computePoints();
static void doCfg();
static void doCfgCycleTime();
static void doCfgDebug();
static void doCfgNop();
static void doCmd();
static void doCmdAbort(RCS_CMD_CHANNEL * subCmdBuf,
		       RCS_STAT_CHANNEL * subStatBuf);
static void doCmdGo(RCS_CMD_CHANNEL * subCmdBuf,
		       RCS_STAT_CHANNEL * subStatBuf);
static void doCmdHalt(RCS_CMD_CHANNEL * subCmdBuf,
		      RCS_STAT_CHANNEL * subStatBuf);
static void doCmdInit(RCS_CMD_CHANNEL * subCmdBuf,
		      RCS_STAT_CHANNEL * subStatBuf);
static void doCmdShutdown(RCS_CMD_CHANNEL * subCmdBuf,
			  RCS_STAT_CHANNEL * subStatBuf);
static void getFileNames();
static int getOptions(int argc, char * argv[], int * bufNum);
static int initBuffers(int bufnum);
static int iniLoad();
static void iniLoadOneDouble(INIFILE * iniFile, const char * parameter,
			     const char * heading, double * setMe,
			     int * retval);
static int initOtherNmlBuffers(char *config_file, int bufnum);
static int initParameters();
static int initPrimSPNmlBuffers(char *config_file, int bufnum);
static int initServoSPNmlBuffers(char *config_file, int bufnum);
int main(int argc, char *argv[]);
static void quit(int sig);
static void readCfg();
static void readCmd();
static void readServoSPStat();
static void setPoint(SensorPoint * setMe, double range,
		     double theta,double timeDiff);
static int setSensorPosition();

/***********************************************************************/

//! cleanupNmlBuffers cleans up all the NML buffers
/*
Returned Value: none

Called By:
  initBuffers
  main 

*/

static void cleanupNmlBuffers() /* NO ARGUMENTS */
{
  cleanupServoSPNmlBuffers();
  cleanupPrimSPNmlBuffers();
  cleanupOtherNmlBuffers();
}

/***********************************************************************/

//! cleanupOtherNmlBuffers cleans up the other (not Prim or Servo) NML buffers 

/*
Returned Value: none

Called By:
  cleanupNmlBuffers
  initOtherNmlBuffers

*/

static void cleanupOtherNmlBuffers() /* NO ARGUMENTS */
{
  if (NULL != navDataExtBuf)
    delete navDataExtBuf;
  if (NULL != sensorData1DBuf)
    delete sensorData1DBuf;
  if (NULL != sensorDataListBuf)
    delete sensorDataListBuf;

  navDataExtBuf = NULL;
  sensorData1DBuf = NULL;
  sensorDataListBuf = NULL;
}

/***********************************************************************/

//! cleanupPrimNmlBuffers cleans up the Prim NML buffers

/*
Returned Value: none

Called By:
  cleanupNmlBuffers
  initPrimSPNmlBuffers

*/

static void cleanupPrimSPNmlBuffers() /* NO ARGUMENTS */
{
  if (NULL != primSPCmdBuf)
    delete primSPCmdBuf;
  if (NULL != primSPStatBuf)
    delete primSPStatBuf;
  if (NULL != primSPCfgBuf)
    delete primSPCfgBuf;
  if (NULL != primSPSetBuf)
    delete primSPSetBuf;

  primSPCmdBuf = NULL;
  primSPStatBuf = NULL;
  primSPCfgBuf = NULL;
  primSPSetBuf = NULL;
}

/***********************************************************************/

//! cleanupServoSPNmlBuffers cleans up the Servo NML buffers 

/*
Returned Value: none

Called By:
  cleanupNmlBuffers
  initBuffers
  initServoSPNmlBuffers

*/

static void cleanupServoSPNmlBuffers() /* NO ARGUMENTS */
{
  if (NULL != servoSPCmdBuf)
    delete servoSPCmdBuf;
  if (NULL != servoSPStatBuf)
    delete servoSPStatBuf;
  if (NULL != servoSPSetBuf)
    delete servoSPSetBuf;

  servoSPCmdBuf = NULL;
  servoSPStatBuf = NULL;
  servoSPSetBuf = NULL;
}

/***********************************************************************/

/*! computePoints builds the "output" SensorDataList representing a list
  of pairs of sensor location and sensed location points */
/*
Returned Value: none

Called By: doCmdGo

Sending Output
--------------

This sends data whenver there is data to send and the time interval since
the last data was sent is greater than outputTime.

Comment on Data Reduction
-------------------------

The data is not being aggregated. Rather, for every two doubles of
input data it is sending four doubles.

Order of Processing Points
--------------------------

It is desirable to store points in the featList in order of decreasing
range, because when a hit on a farther cell is made, AM SP treats any
nearer cell as being unoccupied, which tends to make the nearer cell
be not an obstacle. When looking almost parallel to wall, if the
nearer cells are handled before the farther cells, the wall may
disappear. The order of handling cells does not make much difference
when looking almost perpendicular to a wall. As currently implemented
in MOAST, when the sensor angle is 0, it is (usually) pointing
straight ahead, and the most distant points are often straight ahead.

In view of the above, when the sweep angle of the data passes through
zero, this function splits the data in two parts: from zero down and
from zero up. Doing this tends to make the more distant points be
processed first.

Adjusting Vehicle Position
--------------------------

A small error in the vehicle position (particularly the rotational
position) may result in a large error in the position of a data
point. Hence, the vehicle position is adjusted by interpolating
the position linearly from the time of the nav data to the time
of the point data, using the rotational and translational velocity
reported in the nav data. The rate of change of yaw is used as
the rotational rate.

For this to work, it is necessary that the time given in the point
data and the time given by nav data are both in seconds and have the
same origin.

--------------------------

This function returns without computing any points if no new nav data is
available when the nav data buffer is read or if no new sensor
data is available when the sensor data buffer is read.

This function checks that the purported size of the input data is not too
big for the input message. It prints an error message and returns
without computing any points if the input data is too big.

This function checks that the output data will fit in the output message.
It stops adding data to the output message and prints a warning message
if the message is full (which is indicated by featLength.

This function uses the global variables:
  navDataExtBuf
  sensorData1DBuf
  sensorDataListBuf
  output (a SensorDataList)
  outputTime
  nav
  vehRpy

*/

static void computePoints()
{
  double range;
  double theta;
  int step;
  int step1;
  int navReadType;
  int sensorReadType;
  double start;
  int inputLength;
  double inc;
  double timeDiff;
  double timeInc;
  static int featLength = 0; // length of featArray
  SensorData1D * input = (SensorData1D *)(sensorData1DBuf->get_address());
  static double lastTime = 0;

  sensorReadType = sensorData1DBuf->read(); // Get data from the level below us
  navReadType = navDataExtBuf->read(); // We need to know where the vehicle is
  if ((sensorReadType != SENSOR_DATA_1D_TYPE) ||
      (navReadType != NAV_DATA_EXT_TYPE))
    return;
  nav = (NavDataExt *)(navDataExtBuf->get_address());

  if(!nav->isSLAMon)
    return;

  vehRpy.r = nav->rpyRel.r;
  vehRpy.p = nav->rpyRel.p;
  start = input->start;
  
  inputLength = input->featArray_length;
 
  if (inputLength > MAX_SENSOR_DATA_1D)
    {
      rcs_print_error("%s: input too large for SensorData1D\n", processName);
      return;
    }
  inc = input->increment;
  timeDiff = (input->collectTimeStart - nav->time);
  
  timeInc = ((inputLength < 2) ? 0 :
	     (input->collectTimeDuration / (inputLength - 1)));
  if (((start < 0) && ((start + ((inputLength - 1) * inc)) > 0)) ||
      ((start > 0) && ((start + ((inputLength - 1) * inc)) < 0)))
    { // works for (start neg, inc pos) and (start pos, inc neg)
      step1 = (int)(-start / inc);
      theta = (start + (inc * step1));  // near zero
      timeDiff += (timeInc * step1);
//printf("inc : %lf  \t  start : %lf   timeInc : %lf \n", inc, start, timeInc);
//	printf("step 1 %d     theta %lf     timeDiff %lf\n", step1, theta, timeDiff);
      for (step = 0;
	   ((step <= step1) && (featLength < MAX_SENSOR_DATA_LIST));
	   step++)
	{ // from near zero to beginning
	  range = input->featArray[step1 - step];
	  setPoint((output.featList + featLength++), range, theta, timeDiff);
          theta -= inc;
	  timeDiff -= timeInc;

//printf("(output.featList[featLength] %lf,%lf,%lf / range %lf / theta %lf / timeDiff %lf)\n", output.featList[featLength].dataLoc.x, output.featList[featLength].dataLoc.y, output.featList[featLength].dataLoc.z, range, theta, timeDiff);
	}
  //   printf("------------------------------------------------------\n");
	
      theta = (start + (inc * step1) + inc);
      timeDiff = (input->collectTimeStart - nav->time);
      timeDiff += ((timeInc * step1) + timeInc);
      for ( ; ((step < inputLength) && (featLength < MAX_SENSOR_DATA_LIST));
	    step++)
	{ // from near zero to end
	  range = input->featArray[step];
	  setPoint((output.featList + featLength++), range, theta, timeDiff);
	  theta += inc;
	  timeDiff += timeInc;

//printf("(output.featList[featLength] %lf,%lf,%lf / range %lf / theta %lf / timeDiff %lf)\n", output.featList[featLength].dataLoc.x, output.featList[featLength].dataLoc.y, output.featList[featLength].dataLoc.z, range, theta, timeDiff);
	}
  //      printf("+++++++++++++++++++++++++++++++++++++++++++++++\n");
    }
  else
    { // sweep does not pass through zero
      theta = start;
      for (step = 0;
	   ((step < inputLength) && (featLength < MAX_SENSOR_DATA_LIST));
	    step++)
	{
	  range = input->featArray[step];
	  setPoint((output.featList + featLength++), range, theta, timeDiff);
	  theta += inc;
	  timeDiff += timeInc;
	}
    }
  if (featLength >= MAX_SENSOR_DATA_LIST)
    {
      rcs_print("%s: WARNING: output message is full\n", processName);
    }
  if (((etime() - lastTime) > outputTime) && (featLength > 0))
    {
      output.featList_length = featLength;
      output.maxRange = input->maxRange;
      output.minRange = input->minRange;
      sensorDataListBuf->write(&output);
      featLength = 0;
      lastTime = etime();
    }
}

/***********************************************************************/

//! doCfg executes config commands
/*
Returned Value: none

Called By: main

If admin_state is ADMIN_SHUT_DOWN, this returns immediately.

*/

static void doCfg() /* NO ARGUMENTS */
{
  if (stat.admin_state == ADMIN_SHUT_DOWN)
    { // do nothing if shut down
      return;
    }
  switch (set.command_type)
    {
    case PRIM_SP_CFG_CYCLE_TIME_TYPE:
      doCfgCycleTime();
      break;
      
    case PRIM_SP_CFG_DEBUG_TYPE:
      doCfgDebug();
      break;
      
    case PRIM_SP_CFG_NOP_TYPE:
      doCfgNop();
      break;
      
    default: // no config to handle
      break;
    }
}

/***********************************************************************/

//! doCfgCycleTime sets the configuration cycle time and the timer
/* 
Returned Value: none

Called By: doCfg

The old timer is deleted, and a new one is made.

*/

static void doCfgCycleTime() /* NO ARGUMENTS */
{
  PrimSPCfgCycleTime * cfg;

  cfg = (PrimSPCfgCycleTime *)(primSPCfgBuf->get_address());
  if (state_match(&set, NEW_COMMAND))
    {
      state_new(&set);
      if (cfg->cycleTime <= 0.0)
	{
	  status_next(&set, RCS_ERROR);
	  rcs_print_error("%s: cycleTime must be positive\n", processName);
	}
      else
	{
	  set.cycleTime = cfg->cycleTime;
	  if (timer)
	    delete timer;
	  timer = new RCS_TIMER(set.cycleTime);
	  status_next(&set, RCS_DONE);
	  if (verbose)
	    rcs_print("cycleTime set to %f \n", set.cycleTime);
	}
      state_next(&set, S0);
    }
  else
    {
      state_default(&set);
    }
}

/***********************************************************************/

//! doCfgDebug sets the debug level.
/*
Returned Value: none

Called By:  doCfg

*/

static void doCfgDebug() /* NO ARGUMENTS */
{
  PrimSPCfgDebug * cfg;

  cfg = (PrimSPCfgDebug *)(primSPCfgBuf->get_address());
  if (state_match(&set, NEW_COMMAND))
    {
      state_new(&set);
      set.debug = cfg->debug;
      status_next(&set, RCS_DONE);
      state_next(&set, S0);
    }
  else
    {
      state_default(&set);
    }
}

/***********************************************************************/

//! doCfgNop does nothing and is reported done on the first cycle.
/*
Returned Value: none

Called By:  doCfg

*/

static void doCfgNop() /* NO ARGUMENTS */
{
  if (state_match(&set, NEW_COMMAND))
    {
      state_new(&set);
      status_next(&set, RCS_DONE);
      state_next(&set, S0);
    }
  else
    {
      state_default(&set);
    }
}

/***********************************************************************/

//! doCmd handles task and administrative commands.
/*
Returned Value: none

Called By:  main

If admin_state is ADMIN_SHUT_DOWN, this returns immediately.

*/

static void doCmd() /* NO ARGUMENTS */
{
  if (stat.admin_state == ADMIN_SHUT_DOWN)
    { // do nothing if shut down
      return;
    }
  switch (stat.command_type)
    {
    case PRIM_SP_CMD_INIT_TYPE:
      doCmdInit(servoSPCmdBuf, servoSPStatBuf);
      break;
      
    case PRIM_SP_CMD_HALT_TYPE:
      doCmdHalt(servoSPCmdBuf, servoSPStatBuf);
      break;
      
    case PRIM_SP_CMD_ABORT_TYPE:
      doCmdAbort(servoSPCmdBuf, servoSPStatBuf);
      break;
      
    case PRIM_SP_CMD_SHUTDOWN_TYPE:
      doCmdShutdown(servoSPCmdBuf, servoSPStatBuf);
      break;
      
    case PRIM_SP_CMD_GO_TYPE:
      doCmdGo(servoSPCmdBuf, servoSPStatBuf);
      break;
      
    default:
      // no command to handle
      break;
    }
}

/***********************************************************************/

//! doCmdAbort handles an abort command.
/*
Returned Value: none

Called By:  doCmd

This sends an abort command to servo SP and waits for a done from
servo SP before reporting done.

*/

static void doCmdAbort(         /* ARGUMENTS                    */
 RCS_CMD_CHANNEL * subCmdBuf,   /*!< subordinate command buffer */
 RCS_STAT_CHANNEL * subStatBuf) /*!< subordinate status buffer  */
{
  ServoSPCmdAbort abortMsg;
  ServoSPStat * subStatPtr;

  subStatPtr = (ServoSPStat *)(subStatBuf->get_address());
  if (state_match(&stat, NEW_COMMAND))
    { // command is new: send abort command to sub, go to state S0, status EXEC
      state_new(&stat);
      servoCmdNumber = (subStatPtr->echo_serial_number + 1);
      abortMsg.serial_number = servoCmdNumber;
      subCmdBuf->write(&abortMsg);
      stat.admin_state = ADMIN_UNINITIALIZED;
      status_next(&stat, RCS_EXEC);
      state_next(&stat, S0);
    }
  else if (state_match(&stat, S0))
    {//  // command is not new
      if (servoCmdNumber != subStatPtr->echo_serial_number); //stay in state S0
      else if (subStatPtr->status == RCS_EXEC); // stay in state S0
      else if (subStatPtr->status == RCS_DONE)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_DONE);
	  if (verbose)
	    rcs_print("prim SP aborted\n");
	}
      else // if (amStatPtr->status is RCS_ERROR or unknown)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_ERROR);
	  rcs_print_error("%s: servo SP reported %s\n", processName,
			  ((subStatPtr->status == RCS_ERROR) ?
			   "abort error, prim SP not aborted" :
			   "unknown status, prim SP not aborted"));
	}

    }
  else
    { // command is finished executing: stay in state S1
      state_default(&stat);
    }
}

/***********************************************************************/

//! doCmdGo handles a go command.
/*
Returned Value: none

Called By:  doCmd

If the command is new, the function sends a go command to servo SP.

If the command is not new, this checks the status message returned by
servo SP.

Normally, both servo SP and prim SP stay in the RCS_EXEC status when a
go command is being executed. However, if servo reports done, this
reports done.

This function calls computePoints every time it is called (i.e. every
cycle).

*/

static void doCmdGo(            /* ARGUMENTS                    */
 RCS_CMD_CHANNEL * subCmdBuf,   /*!< subordinate command buffer */
 RCS_STAT_CHANNEL * subStatBuf) /*!< subordinate status buffer  */
{
  PrimSPCmdGo * goCmd;
  ServoSPCmdGo goMsg;
  ServoSPStat * subStatPtr;

  goCmd = (PrimSPCmdGo *)(primSPCmdBuf->get_address());
  subStatPtr = (ServoSPStat *)(subStatBuf->get_address());
  if (stat.admin_state != ADMIN_INITIALIZED)
    {
      status_next(&stat, RCS_ERROR);
      state_next(&stat, S1);
      rcs_print_error("%s: prim SP not initialized, cannot execute command\n",
		      processName);
    }
  else if (state_match(&stat, NEW_COMMAND))
    {
      state_new(&stat);
      outputTime = goCmd->outputCycleTime;
      if (!lastCmdGo)
	{
	  servoCmdNumber = (subStatPtr->echo_serial_number + 1);
	  goMsg.serial_number = servoCmdNumber;
	  subCmdBuf->write(&goMsg);
	}
      status_next(&stat, RCS_EXEC);
      state_next(&stat, S0);
    }
  else if (state_match(&stat, S0))
    { // command is not new
      if (servoCmdNumber != subStatPtr->echo_serial_number); //stay in state S0
      else if (subStatPtr->status == RCS_EXEC) // stay in state S0
	{  // once we start, don't stop; state and status stay the same.
	  computePoints();
	}
      else if (subStatPtr->status == RCS_DONE)
	{
	  state_next(&stat, S1);
	  status_next(&stat, RCS_DONE);
	  if (verbose)
	    rcs_print("prim SP not going\n");
	}
      else // if (subStatPtr->status is RCS_ERROR or unknown)
	{
	  state_next(&stat, S1);
	  status_next(&stat, RCS_ERROR);
	  rcs_print_error("%s: servo SP reported %s\n", processName,
			  ((subStatPtr->status == RCS_ERROR) ?
			   "go error, prim SP not going" :
			   "unknown status, prim SP not going"));
	}
    }
  else
    { // command is finished executing: stay in state S1
      state_default(&stat);
    }
}

/***********************************************************************/

//! doCmdHalt performs halt command state machine
/*
Returned Value: none

Called By: doCmd

If the command is new, this sends a halt command to servo and
sets stat.admin to UNINITIALIZED.

If the command is not new and servo is still executing the halt command
it was sent, this waits.

When the servo halt command is finished executing successfully, this sets
stat.admin to UNINITIALIZED and reports the prim halt command as RCS_DONE.

This changes stat.admin to UNINITIALIZED, even if halt fails.

*/

static void doCmdHalt(          /* ARGUMENTS                    */
 RCS_CMD_CHANNEL * subCmdBuf,   /*!< subordinate command buffer */
 RCS_STAT_CHANNEL * subStatBuf) /*!< subordinate status buffer  */
{
  ServoSPCmdHalt haltMsg;
  ServoSPStat * subStatPtr;

  subStatPtr = (ServoSPStat *)(subStatBuf->get_address());
  if (state_match(&stat, NEW_COMMAND))
    {
      state_new(&stat);
      // send a command to subordinate
      servoCmdNumber = (subStatPtr->echo_serial_number + 1);
      haltMsg.serial_number = servoCmdNumber;
      subCmdBuf->write(&haltMsg);
      stat.admin_state = ADMIN_UNINITIALIZED;
      status_next(&stat, RCS_EXEC);
      state_next(&stat, S0);
    }
  else if (state_match(&stat, S0))
    { // command is not new
      if (servoCmdNumber != subStatPtr->echo_serial_number); //stay in state S0
      else if (subStatPtr->status == RCS_EXEC); // stay in state S0
      else if (subStatPtr->status == RCS_DONE)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_DONE);
	  if (verbose)
	    rcs_print("prim SP halted\n");
	}
      else // if (subStatPtr->status is RCS_ERROR or unknown)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_ERROR);
	  rcs_print_error("%s: servo SP reported %s\n", processName,
			  ((subStatPtr->status == RCS_ERROR) ?
			   "halt error, prim SP not halted" :
			   "unknown status, prim SP not halted"));
	}
    }
  else
    { // command is finished executing: stay in state S1
      state_default(&stat);
    }
}

/***********************************************************************/

//! doCmdInit handles an Init command
/*
Returned Value: none

Called By: doCmd

If the command is new, this sends servo SP an init command and
sets stat.admin to UNINITIALIZED.

If the command is not new, this checks the status reported by servo
SP.

This does not report that the prim init command is RCS_DONE until the
status returned from servo SP says that the servo level init command is
RCS_DONE.

If servo SP initializes without error, this sets stat.admin to
INITIALIZED. Otherwise, it sets stat.admin to UNINITIALIZED.

The servoSPStatBuf was read in main so does not need to be read here.

*/

static void doCmdInit(          /* ARGUMENTS                    */
 RCS_CMD_CHANNEL * subCmdBuf,   /*!< subordinate command buffer */
 RCS_STAT_CHANNEL * subStatBuf) /*!< subordinate status buffer  */
{
  ServoSPCmdInit initMsg;
  ServoSPStat * subStatPtr;

  subStatPtr = (ServoSPStat *)(subStatBuf->get_address());
  if (state_match(&stat, NEW_COMMAND))
    { // command is new
      initParameters();
      if (iniLoad() == INI_INVALID)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_ERROR);
	  rcs_print_error("%s: iniLoad failed, Prim SP not initialized\n",
			  processName);
	}
      else
	{
	  state_new(&stat);
	  servoCmdNumber = (subStatPtr->echo_serial_number + 1);
	  initMsg.serial_number = servoCmdNumber;
	  subCmdBuf->write(&initMsg); // send init to sub
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_EXEC);
	  state_next(&stat, S0);
	}
    }
  else if (state_match(&stat, S0))
    { // command is not new
      if (servoCmdNumber != subStatPtr->echo_serial_number); //stay in state S0
      else if (subStatPtr->status == RCS_EXEC); // stay in state S0
      else if (subStatPtr->status == RCS_DONE)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_INITIALIZED;
	  status_next(&stat, RCS_DONE);
	  if (verbose)
	    rcs_print("prim SP initialized\n");
	}
      else // if (subStatPtr->status is RCS_ERROR or unknown)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_ERROR);
	  rcs_print_error("%s: servo SP reported %s\n", processName,
			  ((subStatPtr->status == RCS_ERROR) ?
			   "error initializing, prim SP not initialized" :
			   "unknown status, prim SP not initialized"));
	}
    }
  else
    { // command is finished executing: stay in state S1
      state_default(&stat);
    }
}

/***********************************************************************/

//! doCmdShutdown handles a shutdown command
/*
Returned Value: none

Called By: doCmd

If the command is new:
  If the administrative state is not UNINITIALIZED this reports an
  error and does not change the administrative state.
  Otherwise, this send a shutdown command to Servo SP.

If the command is not new and execution is not finished:
  If the administrative state is not UNINITIALIZED this reports an
  error, does not change the administrative state, and sets
  the command execution state to finished.

  Otherwise, if servo SP is finished executing successfully the shutdown
  command it was sent, this sets the administrative state to SHUT_DOWN,
  reports the prim shutdown command as RCS_DONE, and sets
  the command execution state to finished.

  Otherwise, if servo has reported an error or unknown status, this reports
  an error, sets the administrative state to UNINITIALIZED, and sets
  the command execution state to finished.
  
  Otherwise, if servo is still executing the shutdown command it was
  sent, this waits.

*/

static void doCmdShutdown(        /* ARGUMENTS                    */
 RCS_CMD_CHANNEL * subCmdBuf,   /*!< subordinate command buffer */
 RCS_STAT_CHANNEL * subStatBuf) /*!< subordinate status buffer  */
{
  ServoSPCmdShutdown shutdownMsg;
  ServoSPStat *subStatPtr;

  subStatPtr = (ServoSPStat *)(subStatBuf->get_address());
  if (stat.admin_state != ADMIN_UNINITIALIZED)
    {
      status_next(&stat, RCS_ERROR);
      state_next(&stat, S1);
      rcs_print_error("%s: prim SP not uninitialized, "
		      "cannot execute shutdown command\n", processName);
    }
  else if (state_match(&stat, NEW_COMMAND))
    {// new command: send shutdown cmd to sub, go to state S0, stat EXEC
      state_new(&stat);
      servoCmdNumber = (subStatPtr->echo_serial_number + 1);
      shutdownMsg.serial_number = servoCmdNumber;
      subCmdBuf->write(&shutdownMsg);
      status_next(&stat, RCS_DONE);
      state_next(&stat, S0);
    }
  else if (state_match(&stat, S0))
    { // command is not new
      if (servoCmdNumber != subStatPtr->echo_serial_number); //stay in state S0
      else if (subStatPtr->status == RCS_EXEC); // stay in state S0
      else if (subStatPtr->status == RCS_DONE)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_SHUT_DOWN;
	  status_next(&stat, RCS_DONE);
	  if (verbose)
	    rcs_print("prim SP shut down\n");
	}
      else // if (subStatPtr->status is RCS_ERROR or unknown)
	{
	  state_next(&stat, S1);
	  stat.admin_state = ADMIN_UNINITIALIZED;
	  status_next(&stat, RCS_ERROR);
	  rcs_print_error("%s: servo SP reported %s\n", processName,
			  ((subStatPtr->status == RCS_ERROR) ?
			   "shutdown error, prim SP not shut down" :
			   "unknown status, prim SP not shut down"));
	}
    }
  else
    { // command is finished executing: stay in state S1
      state_default(&stat);
    }
}

/***********************************************************************/

//! getFileNames gets the names of the nmlFile and the iniFile
/*
Returned Value: none

Called By: main

If the nmlFile (a global variable) is NULL (implying it was not in the
command arguments), this gets the nmlFile name (1) from the
environment if it is there or (2) from the default if not.

This handles the iniFileName (also a global variable) similarly.

*/

static void getFileNames() /* NO ARGUMENTS */
{
  char * nmlFileEnv;
  char * iniFileEnv;

  if (NULL == nmlFile)
    {
      nmlFileEnv = getenv("CONFIG_NML");
      if (NULL != nmlFileEnv)
	{
	  nmlFile = nmlFileEnv;
	}
      else
	{
	  nmlFile = DEFAULT_NML_FILE;
	}
    }
  if (NULL == iniFileName)
    {
      iniFileEnv = getenv("CONFIG_INI");
      if (NULL != iniFileEnv)
	{
	  iniFileName = iniFileEnv;
	}
      else
	{
	  iniFileName = DEFAULT_INI_FILE;
	}
    }
}

/***********************************************************************/

//! getOptions gets the command line options for starting the process
/*
Returned Value: int
If there is an error, this returns 1. Otherwise, it returns 0.

Called By: main

Three command line options are recognized:
1. -n <NML file name>  (a string)
2. -i <ini file name>  (a string)
3. -d <buffer number>  (int giving a suffix for using different buffers)

These are used only at startup.

This uses the global (?) variable optarg (which is apparently a
pointer to a persistent string)

*/

static int getOptions( /*   ARGUMENTS                                     */
 int argc,             /*!< number of command line arguments              */
 char * argv[],        /*!< array of command and arguments                */
 int * bufNum)         /*!< buffer number, set here if specified with -d  */
{
   while (1)
     {
       char option = getopt(argc, argv, "n:i:d:p:");
       if (option == -1)
	 break;
       switch (option)
	 {
	 case 'n':
	   nmlFile = optarg;
	   if (verbose)
	     rcs_print("%s: using specified NML file %s\n",
		       processName, nmlFile);
	   break;
	 case 'i':
	   iniFileName = optarg;
	   rcs_print("%s: using specified INI file %s\n",
		     processName, iniFileName);
	   break;
	 case 'd':
	   if (1 != sscanf(optarg, "%i", bufNum))
	     {
	       rcs_print_error("%s: bad value for buffer number: %s\n",
			       processName, optarg);
	       return 1;
	     }
	   if (verbose)
	     rcs_print("%s: using specified buffer suffix %d\n",
		       processName, *bufNum);
	   break;
    case 'p':
      if(strcmp(optarg,"AirRobot") == 0)
      {
        isAirRobot = true;
        printf("primSP: Platform detected as AirRobot.\n");
      }
      break;
	 default:
	   rcs_print_error("%s: bad argument\n", processName);
	   return 1;
	   break;
	 }
     }
  return 0;
}

/***********************************************************************/

//! iniLoad loads parameters from the named inifile
/*
Returned Value: int
  If the inifile can't be opened or if any given value is invalid, this 
  returns INI_INVALID. If an expected value is missing, so that the default
  value is used, it returns INI_DEFAULT. Otherwise, it returns INI_OK.

Called By: doCmdInit

Throughout MOAST, the top-down order of precedence for the source of
most values is:
1. command line argument
2. iniFile
3. default value

For the iniFile and the NML file name, the order is:
1. command line argument
2. environment variable (CONFIG_INI or CONFIG_NML)
3. default value

This function looks for the following values of Prim SP
parameters in the iniFile.

  [PRIM_SP]
  CYCLE_TIME          <positive number>  seconds


Before this is called, doCmdInit calls initParameters, which resets
"stat" and "set" and gives all other global variables their default
values. This function loads the inifile, which results in some values
being reset.

Note that iniLoadOneDouble requires all parameters to be positive, so
that is not rechecked here.

If a cycle time for Prim SP is obtained from the inifile, this resets
set.cycleTime and the timer.

This uses the following global variables:
 iniFileName
 processName
 set
 timer

*/

static int iniLoad()  /* NO ARGUMENTS */
{
  INIFILE iniFile;
  int retval;

  retval = INI_OK;
  if (iniFile.open(iniFileName))
    {
      rcs_print_error("%s: can't open INI file %s\n",
		      processName, iniFileName);
      return INI_INVALID;
    }
  iniLoadOneDouble(&iniFile, "CYCLE_TIME", "PRIM_SP",
		   &(set.cycleTime), &retval);
  iniFile.close();
  if (retval == INI_OK)
    {
      if (timer)
	delete timer;
      timer = new RCS_TIMER(set.cycleTime);
    }
  return retval;
}

/***********************************************************************/


//! iniLoadOneDouble sets one parameter value from the iniFile.
/*
Returned Value: none

Called By: iniLoad

1. If the parameter is found in the iniFile:
1A. If it has an invalid value, retval is set to INI_INVALID.
1B. If it has a valid value, setMe is set to that value, and retval keeps
    its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

Note that this requires all parameters to be positive.

*/

static void iniLoadOneDouble( /*   ARGUMENTS                              */
 INIFILE * iniFile,           /*!< INIFILE class instance to use          */
 const char * parameter,      /*!< name of parameter in ini file          */
 const char * heading,        /*!< heading under which to find parameter  */
 double * setMe,              /*!< number to set                          */
 int * retval)                /*!< iniLoad return value, maybe reset here */
{
  const char * iniString;
  double d;

  iniString = iniFile->find(parameter, heading);
  if (iniString)
    {
      if ((1 == sscanf(iniString, "%lf", &d)) && (d > 0.0))
	*setMe = d;
      else
	{ // found, but invalid
	  rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
			  processName, iniString, heading, parameter,
			  iniFileName);
	  *retval = INI_INVALID;
	}
    }
  else
    { // not found at all
      rcs_print("can't find [%s] %s in '%s'. using %f\n",
		heading, parameter, iniFileName, *setMe);
      *retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
    }
}

/***********************************************************************/

//! initBuffers initializes all buffers
/*
Returned Value: int
If there is an error, this returns 1. Otherwise, it returns 0.

Called By: main

If there is an error, those buffers that have already been initialized
are cleaned up.

*/

static int initBuffers( /* ARGUMENTS                              */
 int bufnum)            /*!< number to use as channel name suffix */
{
  int retval;

  retval = initServoSPNmlBuffers(nmlFile, bufnum);
  if (retval)
    {
      rcs_print_error("%s: can't allocate servoSP NML buffers\n",
		      processName);
      cleanupServoSPNmlBuffers();
      return 1;
    }
  retval = initPrimSPNmlBuffers(nmlFile, bufnum);
  if (retval)
    {
      rcs_print_error("%s: can't allocate primSP NML buffers\n", processName);
      cleanupNmlBuffers();
      return 1;
    }
  retval = initOtherNmlBuffers(nmlFile, bufnum);
  if (retval)
    {
      rcs_print_error("%s: can't allocate other NML buffers\n", processName);
      cleanupNmlBuffers();
      return 1;
    }
  return 0;
}

/***********************************************************************/

/*! initOtherNmlBuffers makes the nav data, input data, and output
  data buffers.*/
/*
Returned Value: int
If there is an error, this returns 1. Otherwise it returns 0.

Called By: initBuffers

If there is an error, any buffers that have been allocated will need
to be deleted. This is handled in initBuffers.

This sets the global variable servoDataBufName, which is used in setPosition.

*/

static int initOtherNmlBuffers(
 char *config_file,
 int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  cleanupOtherNmlBuffers();
   sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
  navDataExtBuf =
    new NML(navDataExt_format, chanName, THIS_PROCESS, config_file);
  if ((navDataExtBuf == NULL) || (!navDataExtBuf->valid()))
    return 1;

  if(!isAirRobot)
  {
    sprintf(servoDataBufName, "%s%d_1", SERVO_SP_LINESCAN_BUF_NAME, bufnum);
    sensorData1DBuf =
      new NML(sensorData_format, servoDataBufName, THIS_PROCESS, config_file);
    if ((sensorData1DBuf == NULL) || (!sensorData1DBuf->valid()))
      return 1;
    
    sprintf(chanName, "%s%d", PRIM_SP_LINESCAN_NAME, bufnum);
    sensorDataListBuf =
        new NML(sensorData_format, chanName, THIS_PROCESS, config_file);
    if ((sensorDataListBuf == NULL) || (!sensorDataListBuf->valid()))
      return 1;
  }
  else
  {
    printf("primSP: Skipping Linescan buffer due to AirRobot platform.\n");
  }

  return 0;
}

/***********************************************************************/

//! initParameters puts data in key system parameters and config status.
/*
Returned Value: int
  If there is an error, this returns 1. Otherwise, it returns 0.

Called By:
  main

Configuration status (set) and the global variables "output" and "timer"
are given initial values here.

This is setting the increment in the output to 0.1 meters. This is
a completely arbitrary value and should not be taken to have any
significance. The data is not being taken from a grid.

Global variables that get values from command line arguments
(iniFileName and nmlFile) are not reset here. Command status is
not reset here.

*/

static int initParameters() /* NO ARGUMENTS */
{
  // initialize configuration command status
  set.cycleTime = DEFAULT_CYCLE_TIME;
  set.debug = 0;

  // initialize other global variables
  if (timer) // probably not necessary, since timer should be NULL here.
    delete timer;
  timer = new RCS_TIMER(set.cycleTime);
  if (timer == NULL)
    return 1;

  output.isGlobal = false;
  output.increment = 0.1;  // in meters

  return 0;
}


/***********************************************************************/

//! initPrimSPNmlBuffers makes the command, config, stat, and set buffers.
/*
Returned Value: int
If there is an error, this returns 1. Otherwise it returns 0.

Called By: initBuffers

If there is an error, any buffers that have been allocated will need
to be deleted. This is handled in initBuffers.

*/

static int initPrimSPNmlBuffers(
 char *config_file,
 int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  cleanupPrimSPNmlBuffers();
  sprintf(chanName, "%s%d", PRIM_SP_CMD_NAME, bufnum);
  primSPCmdBuf =
    new RCS_CMD_CHANNEL(primSP_format, chanName, THIS_PROCESS, config_file);
  if ((primSPCmdBuf == NULL) || (!primSPCmdBuf->valid()))
    return 1;

  sprintf(chanName, "%s%d", PRIM_SP_CFG_NAME, bufnum);
  primSPCfgBuf =
    new RCS_CMD_CHANNEL(primSP_format, chanName, THIS_PROCESS, config_file);
  if ((primSPCfgBuf == NULL) || (!primSPCfgBuf->valid()))
    return 1;

  sprintf(chanName, "%s%d", PRIM_SP_STAT_NAME, bufnum);
  primSPStatBuf =
    new RCS_STAT_CHANNEL(primSP_format, chanName, THIS_PROCESS, config_file);
  if ((primSPStatBuf == NULL) || (!primSPStatBuf->valid()))
    return 1;

  sprintf(chanName, "%s%d", PRIM_SP_SET_NAME, bufnum);
  primSPSetBuf =
    new RCS_STAT_CHANNEL(primSP_format, chanName, THIS_PROCESS, config_file);
  if ((primSPSetBuf == NULL) || (!primSPSetBuf->valid()))
    return 1;

  return 0;
}

/***********************************************************************/

//! initServoSPNmlBuffers makes the servo command and stat buffers.
/*
Returned Value: int
If there is an error, this returns 1. Otherwise it returns 0.

Called By: initBuffers

If there is an error, any buffers that have been allocated will need
to be deleted. This is handled in initBuffers.

*/

static int initServoSPNmlBuffers(
 char *config_file,
 int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  cleanupServoSPNmlBuffers();
  sprintf(chanName, "%s%d", SERVO_SP_CMD_NAME, bufnum);
  servoSPCmdBuf =
    new RCS_CMD_CHANNEL(servoSP_format, chanName, THIS_PROCESS, config_file);
  if ((servoSPCmdBuf == NULL) || (!servoSPCmdBuf->valid()))
      return 1;

  sprintf(chanName, "%s%d", SERVO_SP_STAT_NAME, bufnum);
  servoSPStatBuf =
    new RCS_STAT_CHANNEL(servoSP_format, chanName, THIS_PROCESS, config_file);
  if ((servoSPStatBuf == NULL) || (!servoSPStatBuf->valid()))
    return 1;

  sprintf(chanName, "%s%d", SERVO_SP_SET_NAME, bufnum);
  servoSPSetBuf =
    new RCS_STAT_CHANNEL(servoSP_format, chanName, THIS_PROCESS, config_file);
  if ((servoSPSetBuf == NULL) || (!servoSPSetBuf->valid()))
    return 1;

  return 0;
}

/***********************************************************************/

/* main

When main starts up, it stuffs a nop command in its own command input
buffer, and it stuffs a nop config command in its own config input
buffer.

In order to get started, an init command and then a go commmand
must be sent to this Prim SP process. This is normally done by
the AM SP process, but it may also be done by the prim shell.

*/

int main(int argc, char *argv[])
{
  int bufnum = 1;
  PrimSPCmdNop nopCmd;
  PrimSPCfgNop nopCfg;

  processName = argv[0];
  if (getOptions(argc, argv, &bufnum))
    {
      return 1;
    }
  getFileNames();
  if (initBuffers(bufnum))
    {
      return 1;
    }
  stat.echo_serial_number = 0;
  reset_time_tracker(&stat.tt);
  set.echo_serial_number = 0;
  if (initParameters())
    {
      return 1;
    }
  if (setSensorPosition())
    {
      return 1;
    }
  nopCmd.serial_number = 1;
  primSPCmdBuf->write(&nopCmd); // stuff a no op in own command buffer
  nopCfg.serial_number = 1;
  primSPCfgBuf->write(&nopCfg); // stuff a no op in own config buffer
  done = 0;
  signal(SIGINT, quit);
  
  // Testing:
  while (!done)
    {
      readCmd();
      readCfg();
      readServoSPStat();
      doCmd();
      doCfg();
      cycle_time_tracker(&stat.tt);
      primSPStatBuf->write(&stat);
      primSPSetBuf->write(&set);
      timer->wait();
    }
  while (!done);
  cleanupNmlBuffers();
  rcs_print("primSP done\n");
  return 0;
}

/***********************************************************************/

//! quit sets done to 1 so that the while loop in main will terminate
/*
Returned Value: none

Called By: pointed at (not called) by signal, called if SIGINT occurs.

See documentation at the beginning of this file regarding termination.
The signal function requires a pointer to a function with one argument
which is an int.

*/

static void quit( /*   ARGUMENTS                                          */
 int sig)         /*!< argument required by the signal function, not used */
{
  done = 1;
}

/***********************************************************************/

//! readCfg reads the config command and modifies the config status buffer.
/*
Returned Value: none

Called By: main

*/

static void readCfg() /* NO ARGUMENTS */
{
  NMLTYPE cfgType;
  int serialNumber;

  cfgType = primSPCfgBuf->read();
  serialNumber = primSPCfgBuf->get_address()->serial_number;
  switch (cfgType)
    {
    case 0:
      // no new command
      break;
      
    case -1:
      // comm error
      rcs_print_error("%s: NML config read error\n", processName);
      break;
      
    case PRIM_SP_CFG_CYCLE_TIME_TYPE:
    case PRIM_SP_CFG_DEBUG_TYPE:
    case PRIM_SP_CFG_NOP_TYPE:
      set.command_type = cfgType;
      if (serialNumber != set.echo_serial_number)
	{
	  set.echo_serial_number = serialNumber;
	  set.state = NEW_COMMAND;
	  set.status = RCS_EXEC;
	}
      break;

    default:
      rcs_print_error("%s: unknown config: %d\n", processName, (int)cfgType);
      break;
    }
}

/***********************************************************************/

//!readCmd reads in a task command and modifies the command status buffer
/*
Returned Value: none

Called By: main

*/

static void readCmd() /* NO ARGUMENTS */
{
  NMLTYPE cmdType;
  int serialNumber;

  cmdType = primSPCmdBuf->read();
  serialNumber = primSPCmdBuf->get_address()->serial_number;
  switch (cmdType)
    {
    case 0:
      // no new command
      break;
      
    case -1:
      // comm error
      rcs_print_error("%s: NML command read error\n", processName);
      break;
      
    case PRIM_SP_CMD_ABORT_TYPE:
    case PRIM_SP_CMD_GO_TYPE:
    case PRIM_SP_CMD_HALT_TYPE:
    case PRIM_SP_CMD_INIT_TYPE:
    case PRIM_SP_CMD_NOP_TYPE:
    case PRIM_SP_CMD_SHUTDOWN_TYPE:
      lastCmdGo = ((stat.command_type == PRIM_SP_CMD_GO_TYPE) &&
		   (stat.status = RCS_EXEC) && (stat.state = S0));
      stat.command_type = cmdType;
      if (serialNumber != stat.echo_serial_number)
	{
	  stat.echo_serial_number = serialNumber;
	  stat.state = NEW_COMMAND;
	  stat.status = RCS_EXEC;
	}
      break;
      
    default:
      rcs_print_error("%s: unknown command: %d\n", processName, (int)cmdType);
      break;
    }
}

/***********************************************************************/

//! readServoSPStat reads the servo SP status buffer
/*
Returned Value: none

Called By: main

The servoCmdNumber, a global variable, is the number of the most
recent command sent to servo SP. servoCmdNumber is set to the usual
echo_serial_number+1 when preparing to send a command to servo. In the
functions that give a servo command and then wait for a response from
servo, it is checked that the echoed command number equals
servoCmdNumber. If they are not equal, those functions keep waiting.

*/

static void readServoSPStat()  /* NO ARGUMENTS */
{
  NMLTYPE statType;

  statType = servoSPStatBuf->read();
  switch (statType)
    {
    case 0:
      // no new status
      break;
      
    case -1:
      // comm error
      rcs_print_error("%s: NML servo SP status read error\n", processName);
      break;
      
    case SERVO_SP_STAT_TYPE:
      // new status message
      break;
      
    default:
      rcs_print_error("%s: unknown servo SP status: %d\n",
		      processName, (int)statType);
      break;
    }
}

/***********************************************************************/

//! setPoint sets a SensorPoint in local coordinates.
/*
Returned Value: none

Called By: computePoints

This takes a range and angle (theta) and sets a SensorPoint in local
coordinates that is at that range and angle, taking into account the
translation and rotation of the sensor relative to the vehicle, and
the translation and rotation of the vehicle relative to local
coordinates.

Theta is the angle between the positive X-axis of the sensor and the
projection on the XY plane of the sensor of line from the sensor to
the point (i.e. ignore the z value of the point). Theta is positive
counterclockwise when viewed from under the ground and clockwise when
being viewed from above.

The position of the vehicle is being adjusted linearly with time
from the position reported by navData, using X and Y translational
velocity and yaw rotational velocity.

This is using the global variables: nav, vehTran, vehRpy, vehQuat,
senTranVeh, senQuatVeh.

The "vehQuat = vehRpy" equates two different types. This works
because "=" has been overloaded to deal with it (yuk).

The following code (which does not make the correction) was previously
used where the equations with the same left-hand side appear.

  vehTran = nav->tranRel;  // local loc'n
  vehRpy.y = nav->rpyRel.y;

The following code was used to print data when the correction
is making a difference of at least one cell. The timeDiff's that
have been observed being printed by this code were in the range
0.05 to 0.22.

if((timeDiff * nav->rpyRelRate.y * 10) > 0.1) // error of 1 cell at 10 m
{printf("timeDiff=%.4lf yawRate=%.4lf xRate=%.4lf yRate=%.4lf\n",
	timeDiff, nav->rpyRelRate.y, nav->tranRelRate.x,nav->tranRelRate.y);}

*/

static void setPoint( /* ARGUMENTS                                      */
 SensorPoint * setMe, /*!< sensor point to set                          */
 double range,        /*!< distance to point from sensor, meters        */
 double theta,        /*!< angle of point from sensor X-axis, radians   */
 double timeDiff)     /*!< seconds since position read, may be negative */
{
  static PM_CARTESIAN senTran;
  static PM_CARTESIAN objTran;

  vehTran = (nav->tranRel + (timeDiff * nav->tranRelRate));  // local loc'n
  vehRpy.y = (nav->rpyRel.y + (timeDiff * nav->rpyRelRate.y));
  vehQuat = vehRpy;  // vehicle's rotation in local space

  //printf("timeDiff: %lf\n", timeDiff);
  //printf("nav->tranRel: %lf,%lf,%lf\n", nav->tranRel.x,nav->tranRel.y,nav->tranRel.z);
  //printf("nav->tranRelRate: %lf,%lf,%lf\n", nav->tranRelRate.x,nav->tranRelRate.y,nav->tranRelRate.z);
  //printf("vehTran: %lf,%lf,%lf\n", vehTran.x,vehTran.y,vehTran.z);

  objTran.x = (cos(theta) * range);
  objTran.y = (sin(theta) * range);
  objTran.z = 0;

  // Convert the coordinates of the sensed point to vehicle-relative
  objTran = senTranVeh + (senQuatVeh * objTran);
  // Convert the coordinates of the sensed point to local-relative
  objTran = vehTran + (vehQuat * objTran);
  // Convert the coordinates of the sensor to local-relative
  senTran = vehTran + (vehQuat * senTranVeh);
  
  setMe->sensorLoc = senTran;
  setMe->dataLoc = objTran;
  //setMe->isSensorLocAccurate = nav->isSLAMon;
}

/***********************************************************************/

//! setSensorPosition sets the position of the sensor on the vehicle.
/*
Returned Value: int
  If any of the following errors occurs, this returns 1.
  Otherwise, it returns 0.
1. There is no message in the servo SP set buffer.
2. There is an error in reading the servo SP set buffer.
3. There is a message in the servo SP set buffer that is not a servo
   SP set message.
4. A servoSP set message is read OK, but none of the sSetElem both
  (1) is of type SERVO_SP_LINESCANNER_SENSOR and
  (2) puts data in the buffer whose name is servoDataBufName
     (a global variable set in initOtherNmlBuffers).

Called By: main

This function sets senTranVeh and senQuatVeh to what is given in a
servo SP set message. The function looks for an sSetElem of type
SERVO_SP_LINESCANNER_SENSOR that puts data in the buffer whose name is
servoDataBufName. When that sSetElem is found, the data for the sensor
described in the sSetElem is transcribed into senTranVeh and
senQuatVeh.

*/

static int setSensorPosition()
{
  NMLTYPE setType;
  ServoSPSet * setMsg;
  SensorSetElem * setElem;
  int stop;
  int n;
  int allOK;

  allOK = 0;
  setType = servoSPSetBuf->read();
  switch (setType)
    {
    case 0:
      // no new set message
      rcs_print_error("%s: NML servo SP set read error\n", processName);
      break;
      
    case -1:
      // comm error
      rcs_print_error("%s: NML servo SP set read error\n", processName);
      break;
      
    case SERVO_SP_SET_TYPE:
      // new set message
      setMsg = (ServoSPSet *)servoSPSetBuf->get_address();
      stop = setMsg->sSetElem_length;
      for (n = 0; n < stop; n++)
	{
	  setElem = ((setMsg->sSetElem) + n);
	  if ((setElem->sType == SERVO_SP_LINESCANNER_SENSOR) &&
	      (strcmp(setElem->senDataBufName, servoDataBufName) == 0))
	    {
	      break;
	    }
	}
      if (n < stop)
	{ // found the sensor we want
	  senTranVeh = setElem->sensorToMount.tran;
	  //PM_CARTESIAN(0, 0, -1.509);
	  senQuatVeh = setElem->sensorToMount.rot;
	  //PM_RPY(0, 0, 0);
	  allOK = 1;
	}
      break;
      
    default:
      rcs_print_error("%s: unknown servo SP set: %d\n",
		      processName, (int)setType);
      break;
    }
  return (allOK ? 0 : 1);
}

/***********************************************************************/

/*
  Modification history:

  $Log: primSPMain.cc,v $
  Revision 1.22  2008/02/11 23:56:36  tomrkramer
  Changed default cycle time from 0.100 to 0.010. Updated documentation.
  added #defines for reading inifile values. Added iniLoad and
  iniLoadOneDouble. Added checks on length of input and output arrays.
  Changed method of deciding when to print output from counting cycles
  to using a timer set according to the value transmitted in a
  PrimSPCmdGo. Changed execution of go command so that if prim SP is
  already going, the only thing done is to change outputCycleTime.

  Revision 1.21  2008/02/06 14:56:45  dr_steveb
  Added support for additional debug flags and performed array boundary checks
 
  Revision 1.20  2007/12/04 23:08:10  tomrkramer
  Made minor improvements in format and documentation.

  Revision 1.19  2007/12/03 21:23:40  tomrkramer
  Changed method of getting ladar (linescanner) position. It was hard-coded.
  Now it is obtained from servo SP.

  Revision 1.18  2006/12/12 19:20:40  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.17  2006/08/04 13:54:11  tomrkramer
  Removed verbose diagnostic code. Added "_1" to chanName for servo linescan.

  Revision 1.16  2006/08/04 13:40:54  tomrkramer
  Changed so primSP does not wait for DONE after sending GO to servo

  Revision 1.13  2006/05/16 20:18:46  tomrkramer
  Added check in computePoints that new navData was received.

  Revision 1.12  2006/04/19 22:48:52  tomrkramer
  Removed heartbeat and replaced it with functions that work equivalently
  by updating stat.tt.count (reset_time_tracker(&stat.tt) and
  cycle_time_tracker(&stat.tt)).

  Revision 1.11  2006/04/13 14:30:09  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.10  2006/04/12 22:06:36  tomrkramer
  Made miscellaneous minor improvements in fault resistance and documentation.

  Revision 1.9  2006/04/12 14:30:04  tomrkramer
  Fixed minor bugs in getOptions and added documentation to it.

  Revision 1.8  2006/04/10 21:22:48  tomrkramer
  Minor changes to implementation of adminstrative states.

  Revision 1.7  2006/04/07 19:46:09  tomrkramer
  Completely rearranged. Documentation improved, administrative states
  implemented.

  Revision 1.6  2005/12/16 19:36:48  tomrkramer
  Improved efficiency. No change in functionality.

  Revision 1.5  2005/12/15 16:01:58  tomrkramer
  added position correction and improved code efficiency

  Revision 1.4  2005/10/28 17:50:51  tomrkramer
  changed order in which data is reported to AM SP. Was left to right;
  is now middle to left plus middle to right.

  Revision 1.3  2005/10/25 18:49:27  tomrkramer
  changed default cycle time from 0.02 to 0.1

  Revision 1.2  2005/10/14 20:23:18  dr_steveb
  Added timout when waiting for init to finish.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.11  2005/10/06 23:02:07  kramer
  minor fixes

  Revision 1.10  2005/09/30 18:55:46  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.9  2005/09/27 21:33:08  kramer
  changed definition of SensorPoint

  Revision 1.8  2005/09/26 21:26:10  kramer
  changed sensor z from 1.509 to -1.509, added comments

  Revision 1.7  2005/08/09 17:09:44  beytin
  Changed to allow USAR to convert max/min ranges instead of prim

  Revision 1.6  2005/08/08 19:06:09  beytin
  Now running at 50 hz and outputting at 5 hz

  Revision 1.5  2005/07/28 17:58:35  scrapper
  Added timer to control loop.

  Revision 1.4  2005/07/22 22:34:42  beytin
  Now takes into account simpleSim offset, but it's hardcoded in

  Revision 1.3  2005/07/18 21:16:46  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.2  2005/07/01 16:36:55  beytin
  Initial working version, only supports 2D data

  Revision 1.1  2005/06/08 19:52:34  proctor
  Initial checkin

*/

