/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file gocfg_servo.cc

  \brief Configures Go Motion kinematics parameters from settings in
  a MOAST NML buffer.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.3 $
  $Date: 2009/05/04 20:55:03 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>		// getenv, atoi
#include <stddef.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>
#include "moastTypes.hh"
#include "moastNmlOffsets.hh"
#include "servoMisJA.hh"

#define BN "gocfg_servo"
#define THIS_PROCESS "primMis"

#ifndef HAVE_GOMOTION
int main(void)
{
  fprintf(stderr, "%s: unimplemented\n", BN);
  return 1;
}
#else

#include <ulapi.h>
#include <go.h>
#include <servointf.h>

#define SERVO_SHM_KEY 101

static RCS_STAT_CHANNEL * servoMisJASetBuf = NULL;

static ServoMisJASet * setBufPtr = NULL;

static void cleanupServoMisJANmlBuffers(void)
{
  if (NULL != servoMisJASetBuf)
    delete servoMisJASetBuf;

  servoMisJASetBuf = NULL;

  setBufPtr = NULL;

  return;
}

static int initServoMisJANmlBuffers(char * config_file, char * process, int vehid, int compid)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupServoMisJANmlBuffers();

  sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_SET_NAME, vehid, compid);
  servoMisJASetBuf =
    new RCS_STAT_CHANNEL(servoMisJA_format, chanName, process, config_file);
  if (!servoMisJASetBuf->valid()) {
    delete servoMisJASetBuf;
    servoMisJASetBuf = 0;
    return -1;
  }

  setBufPtr = (ServoMisJASet *) servoMisJASetBuf->get_address();
  servoMisJASetBuf->read();

  return 0;
}

// converts MOAST link types to Go link types
#define moast_to_go_type(mt) (mt) == SERVO_MIS_JA_PRISMATIC_JOINT_TYPE ? GO_QUANTITY_LENGTH : (mt) == SERVO_MIS_JA_REVOLUTE_JOINT_TYPE ? GO_QUANTITY_ANGLE : GO_QUANTITY_LENGTH

int main(int argc, char * argv[])
{
  char nml_file[MOAST_FILE_NAME_LEN] = "moast.nml";
  char * nml_file_env;

  void * servo_shm = NULL;
  servo_comm_struct * servo_comm_ptr = NULL;
  servo_cfg_struct servo_cfg;
  go_link link_params;

  int vehid;
  int compid;
  int retval;
  int debug;

  vehid = 1;
  compid = 1;
  retval = 0;
  debug = 0;

  // override default NML file with an environment variable, if present
  nml_file_env = getenv("CONFIG_NML");
  if (NULL != nml_file_env) {
    strncpy(nml_file, nml_file_env, MOAST_FILE_NAME_LEN);
    nml_file[MOAST_FILE_NAME_LEN - 1] = 0;
  }

  opterr = 0;
  while (1) {
    char option = getopt(argc, argv, ":n:v:c:d");
    if (option == -1)
      break;
    switch (option) {
    case 'n':
      // override the default or environment variable with command line
      strncpy(nml_file, optarg, MOAST_FILE_NAME_LEN);
      nml_file[MOAST_FILE_NAME_LEN - 1] = 0;
      break;
    case 'v':
      vehid = atoi(optarg);
      break;
    case 'c':
      compid = atoi(optarg);
      break;
    case ':':
      fprintf(stderr, "%s: missing parameter to %c\n", BN, (char) optopt);
      retval = 1;
      break;
    default:			// '?'
      fprintf(stderr, "%s: unrecognized option %c\n", BN, (char) optopt);
      retval = 1;
      break;
    }
  }
  if (retval) return retval;

  if (ULAPI_OK != ulapi_init(UL_USE_DEFAULT)) {
    return 1;
  }

  if (go_init()) {
    fprintf(stderr, "%s: go_init error\n", BN);
    return 1;
  }

   servo_shm = ulapi_shm_new(SERVO_SHM_KEY, SERVO_NUM * sizeof(servo_comm_struct));
   if (NULL == servo_shm) {
     fprintf(stderr, "%s: can't get servo comm shm\n", BN);
     return 1;
   }
   servo_comm_ptr = (servo_comm_struct *) ulapi_shm_addr(servo_shm);

  if (0 != initServoMisJANmlBuffers(nml_file, THIS_PROCESS, vehid, compid)) {
    rcs_print("%s: can't initialize servo NML buffers\n", BN);
    return -1;
  }

  for (int joint = 0; joint < setBufPtr->linkSet_length; joint++) {
    int index;
    go_rpy rpy;
    go_pp pp;

    printf("Link ID               : %2d\n",setBufPtr->linkSet[joint].linkID);
    printf("Parent Link ID        : %2d\n",setBufPtr->linkSet[joint].parentLinkID);
    printf("Joint Type            : %s\n" ,
	   servoMisJA_enum_ServoMisJAJointType_symbol_lookup(setBufPtr->linkSet[joint].jType));
    printf("Max Velocity          : %0.2f\n",setBufPtr->linkSet[joint].maxVelocity);
    printf("Max Torque            : %0.2f\n",setBufPtr->linkSet[joint].maxTorque);
    printf("Range                 : %0.2f %0.2f [min,max]\n",
	   setBufPtr->linkSet[joint].minRange,
	   setBufPtr->linkSet[joint].maxRange);
    printf("Location              : %0.2f %0.2f %0.2f [x,y,z]\n",
	   setBufPtr->linkSet[joint].loc.x,
	   setBufPtr->linkSet[joint].loc.y,
	   setBufPtr->linkSet[joint].loc.z);
    printf("Orientation           : %0.2f %0.2f %0.2f [r,p,y]\n",
	   setBufPtr->linkSet[joint].rot.r,
	   setBufPtr->linkSet[joint].rot.p,
	   setBufPtr->linkSet[joint].rot.y);

#define INDEX(j) ((j) + 0)
    index = INDEX(joint);

    link_params.quantity = moast_to_go_type(setBufPtr->linkSet[index].jType);
    pp.pose.tran.x = setBufPtr->linkSet[index].loc.x;
    pp.pose.tran.y = setBufPtr->linkSet[index].loc.y;
    pp.pose.tran.z = setBufPtr->linkSet[index].loc.z;
    rpy.r = setBufPtr->linkSet[index].rot.r;
    rpy.p = setBufPtr->linkSet[index].rot.p;
    rpy.y = setBufPtr->linkSet[index].rot.y;
    go_rpy_quat_convert(&rpy, &pp.pose.rot);
    link_params.u.pp = pp;
    link_params.type = GO_LINK_PP;

    servo_cfg.type = SERVO_CFG_LINK_TYPE;
    servo_cfg.serial_number = servo_comm_ptr[joint].servo_set.echo_serial_number + 1;
    servo_cfg.u.link.link = link_params;
    servo_cfg.tail = servo_cfg.head = servo_cfg.serial_number;
    servo_comm_ptr[joint].servo_cfg = servo_cfg;
  }

  cleanupServoMisJANmlBuffers();

  (void) go_exit();
  (void) ulapi_exit();

  return retval;
}

#endif	// else defined (HAVE_GOMOTION)

/*
  Modification history:

  $Log: gocfg_servo.cc,v $
  Revision 1.3  2009/05/04 20:55:03  proctor
  Disabled shared library _so builds for Cygwin; added disclaimer and brief
  comments for servo wrapper code

*/
