#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <signal.h>
#include <stdio.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif

#include <errno.h>
#include <string>
using namespace std;

//Inferior Skin Headers
#include "usarInfSkin.hh"
#ifdef HAVE_PLAYER
#include "playerInfSkin.hh"
#endif

//Superior Skin Headers
#include "moastSupSkin.hh"

//SIMware Stuff
#include "simWareKB.hh"
#include "swKBUtils.hh"

#include <posemath.h>


// SIMware component pointers
SIMwareInfSkin * infSkin;
SIMwareSupSkin * supSkin;
SIMwareKB *kb;

// SIMware configuration parameters
string platName;
string botName; 
PM_POSE stPoseOffset;
int vehID          = 1;
string stPoseName  = "default";
string infSkinName = "usarsim";
string supSkinName = "moast";
int getWorldInfo   = 0;
int debug          = 0;
bool done          = false;

// Functions
void helpMenu();
void getArgs(int, char *[]);
void infSkinAlloc(string);
void supSkinAlloc(string);
void sigHandler (int);

static void
touch_simWare_running_file(int argc, char **argv)
{
  static const char *RUNNING_FILE="simWare.running";
  FILE *f =fopen(RUNNING_FILE,"w");
  if(!f)
    {
      fprintf(stderr,"Can't open %s for writing -- %s\n",
	      RUNNING_FILE, strerror(errno));
      return;
    }
  for(int i = 0; i < argc; i++)
    {
      fprintf(f,"%s ",argv[i]);
    }
  fprintf(f," running.\n");
  fclose(f);
}


int
main (int argc, char *argv[])
{
  if( signal (SIGINT, sigHandler) == SIG_ERR )
    {
      perror (" Could not set signal SIGINT");
      exit(1);
    }
  if( signal (SIGTERM, sigHandler)  == SIG_ERR )
    {
      perror (" Could not set signal SIGTERM");
      exit(1);
    }

  getArgs(argc, argv);
  
  kb = new SIMwareKB;
  kb->setSettingsFresh(false);
  
  if(debug) printf("<simWare::main> Opening inferior skin : %s \n", infSkinName.c_str());
  infSkinAlloc(infSkinName);
  if( infSkin == NULL )
    {
      printf("<simWare::main> ** ERROR ** allocating inferior skin\n");
      exit(-1);
    }
  
  if(debug) printf("<simWare::main> Opening superior skin : %s \n", supSkinName.c_str());
  supSkinAlloc(supSkinName);
  if( supSkin == NULL )
    {
      printf("<simWare::main> ** ERROR ** allocating superior skin\n");
      exit(-1);
    }
 
  if(debug) printf("<simWare::main> Configuring SIMware\n");
  if( !infSkin->configInfSkin(kb,supSkin) )
    {
      printf("<simWare::main> ** ERROR ** configuring inferior skin\n");
      exit(-1);
    }

  if( !supSkin->configSupSkin(kb,infSkin) )
    {
      printf("<simWare::main> ** ERROR ** configuring superior skin\n");
      exit(-1);
    }
  
  if(debug) printf("<simWare::main> opening inferior skin\n");
  if(!infSkin->openSkin())
    {
      printf("<simWare::main> ** ERROR ** opening inferior skin\n");
      exit(-1);
    }
   
  if(debug) printf("<simWare::main> load a priori world information into KB\n");
  if(! infSkin->loadWorldInfo() )
    {
      printf("<simWare::main> ** ERROR ** loading world information from inferior skin\n");
      exit(-1);
    }
  
  if( getWorldInfo )
    {
      kb->world.dumpWorldInfo();
      exit(1);
    }
  if(debug) printf("<simWare::main> initializing robot starting location\n");
  kb->world.setStartPoseOffset(stPoseOffset);
  if( -1 == (kb->world.setStartPose(stPoseName)) )
    {      
      printf("<simWare::main> ** ERROR ** start pose '%s' not recognized\n", stPoseName.c_str());
      exit(-1);
    }
   
  if(debug) printf("<simWare::main> initializing robot\n");
  if( ! infSkin->startBot(vehID,platName, botName, kb->world.getStartPose()) )
    {
      printf("<simWare::main> ** ERROR ** starting robot\n");
      exit(-1);
    }
  
  if(debug) printf("<simWare::main> opening superior Skin\n");
  if(!supSkin->openSkin())
    {
      printf("<simWare::main> ** ERROR ** opening superior skin\n");
      exit(-1);
    }
  
  touch_simWare_running_file(argc, argv);

  while(!done)
    esleep(1);
  
  printf("<simWare::main> Done main while loop. Closing/deleting skins . . . \n");
  
  if(supSkin)
    {
      supSkin->closeSkin();
    }
  if(infSkin)
    {
      infSkin->closeSkin();
    }
  
  if(supSkin) 
    {
      delete supSkin;
      supSkin=0;
    }

  if(infSkin)
    {
      delete infSkin;
      infSkin=0;
    }


  if(kb)
    {
      delete kb;
      kb=0;
    } 

  return 0;
}


void infSkinAlloc(string skinName)
{
  if(! nocase_cmp(skinName,"usarsim"))
    infSkin = new USARSimInfSkin;
#ifdef HAVE_PLAYER
  else if(!nocase_cmp(skinName, "player"))
    infSkin = new PlayerInfSkin;
#endif
  else
    {
      infSkin = NULL;
      fprintf( stderr, "simWare: inferior skin named %s not linked to application\n",
	       skinName.c_str() );
    }
   
}

void supSkinAlloc(string skinName)
{
  if(! nocase_cmp(skinName,"moast"))
    supSkin = new MoastSupSkin;
  else
    supSkin = NULL;
}


void getArgs(int argc, char *argv[])
{
  char option;
    
  while (1)
    {

      static struct option long_options[] =
	{
	  {"vehID"           , required_argument, 0, 'i'},
	  {"platform-name"   , required_argument, 0, 'p'},
	  {"robot-name"      , required_argument, 0, 'n'},
	  {"start_pose_name" , required_argument, 0, 's'},
	  {"infSkin-name"    , required_argument, 0, '1'},
	  {"supSkin-name"    , required_argument, 0, '2'},
	  {"help"            , no_argument      , 0, 'h'},
	  {"world-info-only" , no_argument      , &getWorldInfo, 1},
	  {"debug"           , no_argument      , &debug, 1},
	  {0, 0, 0, 0}    
	};
      
      int option_index = 0;
	  
      option = getopt_long (argc, argv, "x:y:z:r:i:p:n:s:h",
			    long_options, &option_index);
    
      if (option == -1)
        break;
    
      switch (option)
        {
	case 0:
	  break;

        case 'x':
	  stPoseOffset.tran.x = atof(optarg);
	  break;
	case 'y':
	  stPoseOffset.tran.y = atof(optarg);
	  break;
	case 'z':
	  stPoseOffset.tran.z = atof(optarg);
	  break;
	case 'r':
	  stPoseOffset.rot = PM_RPY(0,0,atof(optarg));
	  break;
        case 'i':
          vehID = atoi(optarg);
          break;
       	case 'p':
	  platName = optarg;
	  break;
	case 'n':
	  botName = optarg;
	  break;
	case 's':
	  stPoseName = optarg;
	  break;
        case '1':
	  infSkinName = optarg;
	  break;
	case '2':
	  supSkinName = optarg;
	  break;
	case 'h':
	  helpMenu();
	  exit(1);
	  break;
	case '?':
	  exit(1);
	  break;
	default:
	  helpMenu();
	  exit(1);
	  break;
	}
    }
}

void helpMenu()
{
  printf("\nSIMware:\n");
  printf("usage: simWare [options]... [[-p|--platform-type] NAME]\n");
  printf("  -i, --vehID INT               unique vehicle id, defaults to 1\n");
  printf("  -p, --platform-type NAME      name of the platform\n");
  printf("  -n, --robot-name NAME         name of the robot\n");
  printf("  -s, --start-pose NAME         name of absolute start pose, defaults to 'default'\n");
  printf("  -x DOUBLE                     X offset for starting pose, defaults to 0\n");
  printf("  -y DOUBLE                     Y offset for starting pose, defaults to 0\n");
  printf("  -z DOUBLE                     Z offset for starting pose, defaults to 0\n");
  printf("  -r DOUBLE                     orientation offset for starting pose, defaults=0\n");
  printf("  --infSkin NAME                name of inferior skin, defaults to 'usarsim'\n");
  printf("  --supSkin NAME                name of superior skin, defaults to 'moast'\n");
  printf("  --world-info-only             shows world information & terminates\n");
  printf("\n *Note* Start pose used by inferior skins is the 'start-pose' + the 'offset.'\n");
  printf("      There are two predefined 'start-pose':\n");
  printf("       'default' - which uses a defined pose in the world info\n");
  printf("       'offset'  - which uses just offset as absolute pose\n");

}


void sigHandler (int signum)
{ 
  printf("\n<simWare::sigHandler>       : Recv %d\n", signum);
  done = true;
}
