/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file swKBSen.hh

  \brief 
  \code CVS Status:
  $Author: cj_scrapper $
  $Revision: 1.12 $
  $Date: 2007/11/14 02:11:36 $
  \endcode

  \author Chris Scrapper
*/


#ifndef SIMWARE_KB_SEN_HH
#define SIMWARE_KB_SEN_HH

#include <string>
#include <vector>
#include <list>
using namespace std;
#include <posemath.h>

#include "swKBUtils.hh"
#include "swKBMsg.hh"
#include "swKBMis.hh"
#include "swKBTypes.hh"


class SIMwareKBSen
{
protected:
  static int senCount;
  string sName;
  int sID;
  //! Handle to mount, if NULL then mounted to vehicle
  SIMwareKBMount * mount;
  //! Transform from sensor to link mounted, or vehicle
  PM_POSE senToMount;
  //! Indicates with the sensor has new data
  bool dFlag;
  //! Steering mechanism of robotic platform
  SIMwareSensorType sType;
  //! Cycle time of sensor
  double cTime;
  //! TimeStamp
  double tStamp;
public:
  //! Constructor
  SIMwareKBSen ();
  //! Destructor
  virtual ~ SIMwareKBSen ();
  //! Sets the robotic platform name
  void setName (string);
  //! Returns robotic platform name
  string getName ();
  //! Sets the robotic platform id
  void setSenID (int);
  //! Returns robotic platform id
  int getSenID ();
  //! Sets t(double xVal)      {x=xVal;}he steering type of robotic platform
  void setSensorType (SIMwareSensorType);
  //! Gets steering typeof robotic platform
  SIMwareSensorType getSensorType ();
  //! Get cycle time of sensor
  double getCycleTime ();
  //! Set cycle time of sensor
  void setCycleTime (double);
  //! Get the time data was reported
  double getTimeStamp ();
  //! Set the time data is obtained
  void setTimeStamp (double);
  
  //! Get Sensor to Mount transform
  PM_POSE getSen2Mount ();
  
  //! Set sensor to mount transform
  //void setSen2Mount (PM_POSE, string);
  void setSen2MountXYZ(double, double, double);
  void setSen2MountRPY(double, double, double);
  void setMount(SIMwareKBMount *);
  SIMwareKBMount * getMount();
  
  void resetNewDataFlag ();
  bool hasNewData ();
  virtual void clearData () =0;
};

//////   Odometry Sensor
class SIMwareKBSenOdometry:public SIMwareKBSen
{
private:
  //! Position relative to robot's starting position
  double x;
  //! Position relative to robot's starting position
  double y;
  //! Position relative to robot's starting position
  double orient;
public:
  //! Constructor
    SIMwareKBSenOdometry ();
    SIMwareKBSenOdometry (string);
  //! Set the scalar component x of the position vector
  void setX (double);
  //! Get the scalar component x of the position vector
  double getX ();
  //! Set the scalar component y of the position vector
  void setY (double);
  //! Get the scalar component y of the position vector
  double getY ();
  //! Set the scalar component theta of the position vector
  void setOrient (double);
  //! Get the scalar component theta of the position vector
  double getOrient ();
  void setOdometry (double, double, double);
  void clearData ();
};


/*!
  The Inertial Navigation Unit (INU) sensor reports vehicle orientation,
  typically via double integration on accelerometers.
*/
class SIMwareKBSenINU:public SIMwareKBSen
{
private:
  //! Orientation reported by INU
  PM_RPY orient;
public:
  //! Constructor
  SIMwareKBSenINU ();
  SIMwareKBSenINU (string);
  //! Get the orientation reported by sensor
  PM_RPY getOrientation ();
  double getRoll();
  double getPitch();
  double getYaw();
  //! Set the orientation reported by sensor
  void setOrientation (PM_RPY);
  void setOrientation (double, double, double);
  void clearData ();
};


/*!
  The Inertial Navigation System (INU) sensor reports vehicle position
  and orientation, typically via double integration on accelerometers.
*/
class SIMwareKBSenINS:public SIMwareKBSen
{
private:
  PM_CARTESIAN xyz;
  PM_RPY rpy;
public:
  //! Constructor
  SIMwareKBSenINS ();
  SIMwareKBSenINS (string);
  PM_RPY getOrientation ();
  double getRoll();
  double getPitch();
  double getYaw();
  PM_CARTESIAN getLocation();
  double getX();
  double getY();
  double getZ();
  //! Set the orientation reported by sensor
  void setOrientation (PM_RPY);
  void setOrientation (double, double, double);
  void setLocation(PM_CARTESIAN);
  void setLocation (double, double, double);
  void clearData ();
};


/*!
  The Global Positioning System (GPS) sensor reports northing, easting
  and elevation all in [m], as per Universal Tranverse Mercator (UTM)
*/

class SIMwareKBSenGPS:public SIMwareKBSen
{
private:
  int LatitudeDegrees;
  float LatitudeMinutes;
  char LatitudeCardinality;
  int LongitudeDegrees;
  float LongitudeMinutes;
  char LongitudeCardinality;
  int Fix;
  int Satellites;

public:
  //! Constructor
  SIMwareKBSenGPS(void);
  SIMwareKBSenGPS(string);

  //! Mutators
  void setLatitudeDegrees(int);
  void setLatitudeMinutes(float);
  void setLatitudeCardinality(char);
  void setLongitudeDegrees(int);
  void setLongitudeMinutes(float);
  void setLongitudeCardinality(char);
  void setFix(int);
  void setSatellites(int);

  //! Accessors
  int getLatitudeDegrees(void);
  float getLatitudeMinutes(void);
  char getLatitudeCardinality(void);
  int getLongitudeDegrees(void);
  float getLongitudeMinutes(void);
  char getLongitudeCardinality(void);
  int getFix(void);
  int getSatellites(void);

  //! Clear the GPS data and flags
  void clearData();
};

//////   Encoder 
class SIMwareKBSenEncoder:public SIMwareKBSen
{
private:
  //! Resolution - minimum sensed spin angle
  double res;
  //! Number of ticks reported by the encoder
  int tick;
public:
  //! Constructor
    SIMwareKBSenEncoder ();
    SIMwareKBSenEncoder (string);
  //! Get the resolution of encoder
  double getResolution ();
  //! Set the resolution of sensor
  void setResolution (double);
  //! Get the number of ticks reported by encoder
  int getTick ();
  //! Set the number of ticks reported by encoder
  void setTick (int);
  void clearData ();
};


//////   Sonar Sensor
class SIMwareKBSenSonar:public SIMwareKBSen
{
private:
  //! Maximum range of sonar sensor
  double maxRange;
  //! Minimum range of sonar sensor
  double minRange;
  //! Beam angle of sonar sensor
  double bDispersion;
  //! Range reported by sonar sensor
  double rVal;
public:
  //! Constructor
    SIMwareKBSenSonar ();
    SIMwareKBSenSonar (string);
  //! Get maximum range of sonar sensor
  double getMaxRange ();
  //! Set the maximum range of sonar sensor
  void setMaxRange (double);
  //! Get the minimum range of sonar sensor
  double getMinRange ();
  //! Set minimum range of sonar sensor
  void setMinRange (double);
  //! Get the beam angle of sonar sensor
  double getBeamDispersion ();
  //! Set the beam angle of sonar sensor
  void setBeamDispersion (double);
  //! Get the range value reported by sonar sensor
  double getRangeReturn ();
  //! Set the range valuebSep=b;}
  void setRangeReturn (double);
  void clearData ();
};

//////   Line scanner 
class SIMwareKBSenLinescan:public SIMwareKBSen
{
private:
  //! Maximum range of line scanner
  double maxRange;
  //! Minimum range of line scanner
  double minRange;
  //! Beam spacing measured in rads
  double bSep;
  double bDispersion;
  //! Field of view of line scanner
  double fov;
  //! Range return values of linescanner, ret[0] is right most return
    vector < double >retVals;
public:
  //! constructor
    SIMwareKBSenLinescan ();
    SIMwareKBSenLinescan (string);
  //! Set maximum range of linescanner
  void setMaxRange (double);
  //! Get maximum range of linescanner
  double getMaxRange ();
  //! Set minimum range of linescanner
  void setMinRange (double);
  //! Get minimum range of linescanner
  double getMinRange ();
  //! Set the beam seperation of linescanner measured in rads
  void setBeamSpacing (double);
  //! Get the beam seperation of linescanner measured in rads
  double getBeamSpacing ();
  //! Set the beam dispersion for the laser on linescanner 
  void setBeamDispersion(double);
  //! Get the beam dispersion for the laser on linescanner 
  double getBeamDispersion();
  //! Set the linescanners field of view 
  void setFov (double);
  //! Get the linescanners field of view
  double getFov ();
  //! Returns linescanner returns and clears return vector
    vector < double >getReturns ();
  //! Append linescanner return
  void appendRangeReturn (double);
  //! Set element of linescanner return
  void setRangeReturn (int, double);
  void setRangeReturn (const vector < double >&);
  //! Get the number of linescanner returns 
  int getReturnCount ();
  void clearData ();
  double getRangeReturn(int);
};


//////   Camera 
class SIMwareKBSenCamera:public SIMwareKBSen
{
private:
  //! Current fov of camera
  double fov;
  //! Default fov of camera
  double defFov;
  //! Minimum field of view of camera
  double minFov;
  //! Maximum field of view of camera
  double maxFov;
public:
  //! Constructor
    SIMwareKBSenCamera ();
    SIMwareKBSenCamera (string);
  //! Set the maximum field of view of camera
  void setMaxFov (double);
  //! Get the maximum field of view for camera
  double getMaxFov ();
  //! Set the minimum field of view of camera
  void setMinFov (double);
  //! Get the minimum field of view for camera
  double getMinFov ();
  //! Set camera's current field of view
  void setFov (double);
  //! Get camera's current field of view
  double getFov ();
  //! Set camera's default field of view
  void setDefFov (double);
  //! Get camera's default field of view
  double getDefFov ();
  void clearData ();
};


//////   RFID sensor
class SIMwareKBSenRFID:public SIMwareKBSen
{
private:
  //! Maximum range of rfid sensor
  double maxRange;
  //! Data returns of the RFID sensor
  list < SIMwareRfidMsg > rfidMsgs;
public:
  //! Constructor
    SIMwareKBSenRFID ();
    SIMwareKBSenRFID (string);
  //! Set maximum range of sensor
  void setMaxRange (double);
  //! Get maximum range of sensor
  double getMaxRange ();
  //! Get the number of returns recieved from rfid sensor
  int getReturnCount ();
 
  void addRfidMsg (SIMwareRfidMsg);
  SIMwareRfidMsg popRfidMsg ();
  bool hasRfidMsg();
  void clearData ();
};

//////   VICTIM SENSOR
class SIMwareKBSenVictim:public SIMwareKBSen
{
private:
  //! Maximum range of Victim rfid sensor
  double maxRange;
  double vFov;
  double hFov;
  //! Returns from victim rfid sensor
  list < SIMwareVictimMsg > victMsgs;
public:
  //! Constructor
  SIMwareKBSenVictim ();
  SIMwareKBSenVictim (string);
  //! Set the maximum range of victim rfid sensor 
  void setMaxRange (double);
  //! get the maximum range of victim rfid sensor
  double getMaxRange ();
  void setHorizontalFOV(double);
  double getHorizontalFOV();
  void setVerticalFOV(double);
  double getVerticalFOV();

  void addVictMsg (SIMwareVictimMsg);
  SIMwareVictimMsg popVictMsg ();
  bool hasVictMsg();

  void clearData ();
};

//////   Touch 
class SIMwareKBSenTouch:public SIMwareKBSen
{
private:
  //! Diameter of the touch sensor
  double dia;
  bool touch;
public:
  //! Constructor
    SIMwareKBSenTouch ();
    SIMwareKBSenTouch (string);
  //! Get the diameter of touch sensor
  double getDiameter ();
  //! Set the diameter of touch sensor
  void setDiameter (double);
  void clearData ();
  bool getTouch();
  void setTouch(bool);
};

//////   Ground Truth Sensor
class SIMwareKBSenGrdTruth:public SIMwareKBSen
{
private:
  PM_CARTESIAN xyz;
  PM_RPY rpy;
public:
  //! Constructor
  SIMwareKBSenGrdTruth ();
  SIMwareKBSenGrdTruth (string);
  void clearData();
  
  void setPose(PM_POSE);
  void setLocation(PM_CARTESIAN);
  void setLocation(double, double, double);
  void setOrientation(PM_RPY);
  void setOrientation(double, double, double);
  
  PM_CARTESIAN getLocation();
  double getX();
  double getY();
  double getZ();

  PM_RPY getOrientation();
  double getRoll();
  double getPitch();
  double getYaw();
};


#endif
