/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simpleMobSim.cc

  \brief  Simple simulator for wheelchair-, unicycle-like robot, with
  simple GPS- and INS jump and drift.

  \ref simpleMobSim

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.7 $
  $Date: 2007/01/25 15:12:11 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <stdio.h>		// sscanf()
#include <stddef.h>		// NULL
#include <stdlib.h>		
#include <signal.h>		// signal(), SIGINT
#include <math.h>		// fabs(), sin(), cos()
#include <float.h>		// DBL_EPSILON
#include <pthread.h>
#include <rcs.hh>
#ifndef RCS_STAT_SOURCE_FILE_LEN
#define RCS_STAT_SOURCE_FILE_LEN 64
#endif
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_EXT_BASE
#include "navDataExt.hh"	// navDataExt_format()
#include "servoMobJA.hh"	// servoMobJA_format()
#include "servoSP.hh"		// servoSP_format()
#include "acmRandom.h"
#include "simpleSim.hh"

static NML *navDataExtBuf = NULL;
static RCS_CMD_CHANNEL *servoMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *servoMobJAStatBuf = NULL;
static RCS_STAT_CHANNEL *servoMobJASetBuf = NULL;
static RCS_CMD_CHANNEL *servoMobJACfgBuf = NULL;

static NavDataExt navDataExt;
static ServoMobJAStat servoMobJAStat;
static ServoMobJASet servoMobJASet;

#define THIS_PROCESS "simware"	// NML process name for the server
#define NAV_PROCESS "slamStub"	// process name that sets nav data

#define DEFAULT_CYCLE_TIME 0.010
#undef USE_REAL_TIME

static AcmRandomStruct rnd;
static double abs_dclock;
static PM_POSE simPose;
static PM_POSE relPose, absPose;
static PM_POSE simToRel, simToAbs;

#define DEFAULT_WHEEL_RADIUS 1.0
#define DEFAULT_WHEEL_SEPARATION 1.0
static double wheel_radius = DEFAULT_WHEEL_RADIUS;
static double wheel_radius_inv = 1.0 / wheel_radius;
static double wheel_separation = DEFAULT_WHEEL_SEPARATION;
static double wheel_separation_inv = 1.0 / wheel_separation;

static double delta_t = DEFAULT_CYCLE_TIME;
static double delta_t_inv = 1.0 / DEFAULT_CYCLE_TIME;

#define MIN_OMEGA   1.0e-6
#define DEFAULT_ALPHA 1.0e3

// origins for absolute position tran and rpy
static PM_CARTESIAN absOriginTran = PM_CARTESIAN(0,0,0);
static PM_RPY absOriginRpy = PM_RPY(0,0,0);

// absolute (GPS) jump params, limits on how far to jump per interval
static PM_CARTESIAN absExcursionTran = PM_CARTESIAN(0,0,0);
static PM_RPY absExcursionRpy = PM_RPY(0,0,0);
// seconds per jump, < 0 means inhibit
static double absExcursionInterval = -1.0;

// origins for relative position tran and rpy
static PM_CARTESIAN relOriginTran = PM_CARTESIAN(0,0,0);
static PM_RPY relOriginRpy = PM_RPY(0,0,0);

// relative (INS) drift params, limits on how far to drift per second
static PM_CARTESIAN relExcursionTran = PM_CARTESIAN(0,0,0);
static PM_RPY relExcursionRpy = PM_RPY(0,0,0);

#if ! HAVE_PM_RPY_OPERATORS

PM_RPY operator +(PM_RPY r1, PM_RPY r2)
{
  PM_RPY ret;

  ret.r = r1.r + r2.r;
  ret.p = r1.p + r2.p;
  ret.y = r1.y + r2.y;

  return ret;
}

PM_RPY operator -(PM_RPY r1, PM_RPY r2)
{
  PM_RPY ret;

  ret.r = r1.r - r2.r;
  ret.p = r1.p - r2.p;
  ret.y = r1.y - r2.y;

  return ret;
}

#endif // ! HAVE_PM_RPY_OPERATORS

static int invJacobian(double v, double w, double R_inv, double L,
		       double *w1, double *w2)
{
  *w1 = R_inv * (v + 0.5 * w * L);
  *w2 = R_inv * (v - 0.5 * w * L);

  return 0;
}

static int fwdJacobian(double w1, double w2, double R, double L_inv,
		       double *v, double *w)
{
  *v = 0.5 * R * (w1 + w2);
  *w = L_inv * R * (w1 - w2);

  return 0;
}

static int computeDs(double v, double w, double th, double delta_t,
		     double *dx, double *dy, double *dth)
{
  double r;
  double cth, sth;
  double dthp, dxp, dyp;

  if (fabs(w) < MIN_OMEGA) {
    dthp = 0.0;
    dxp = v * delta_t;
    dyp = 0.0;
  } else {
    r = v / w;
    dthp = w * delta_t;
    dxp = r * sin(dthp);
    dyp = r * (1.0 - cos(dthp));
  }

  cth = cos(th);
  sth = sin(th);

  *dx = cth * dxp - sth * dyp;
  *dy = sth * dxp + cth * dyp;
  *dth = dthp;

  return 0;
}

static int cleanupServoMobJANmlBuffers(void)
{
  if (NULL != navDataExtBuf)
    delete navDataExtBuf;
  if (NULL != servoMobJACmdBuf)
    delete servoMobJACmdBuf;
  if (NULL != servoMobJAStatBuf)
    delete servoMobJAStatBuf;
  if (NULL != servoMobJASetBuf)
    delete servoMobJASetBuf;
  if (NULL != servoMobJACfgBuf)
    delete servoMobJACfgBuf;

  navDataExtBuf = NULL;
  servoMobJAStatBuf = NULL;
  servoMobJASetBuf = NULL;
  servoMobJACmdBuf = NULL;
  servoMobJACfgBuf = NULL;

  return 0;
}

static int initServoMobJANmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupServoMobJANmlBuffers();

  sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
  navDataExtBuf =
    new NML(navDataExt_format, chanName, NAV_PROCESS, config_file);
  if (!navDataExtBuf->valid()) {
    delete navDataExtBuf;
    navDataExtBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, bufnum);
  servoMobJACmdBuf =
    new RCS_CMD_CHANNEL(servoMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMobJACmdBuf->valid()) {
    delete servoMobJACmdBuf;
    servoMobJACmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, bufnum);
  servoMobJAStatBuf =
    new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMobJAStatBuf->valid()) {
    delete servoMobJAStatBuf;
    servoMobJAStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_SET_NAME, bufnum);
  servoMobJASetBuf =
    new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMobJASetBuf->valid()) {
    delete servoMobJASetBuf;
    servoMobJASetBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_CFG_NAME, bufnum);
  servoMobJACfgBuf =
    new RCS_CMD_CHANNEL(servoMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMobJACfgBuf->valid()) {
    delete servoMobJACfgBuf;
    servoMobJACfgBuf = 0;
    return 1;
  }

  return 0;
}

static void doCmdInit(ServoMobJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdHalt(ServoMobJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdAbort(ServoMobJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdShutdown(ServoMobJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    (void) simpleMobSimQuit();
    state_default(stat);
  }
}

#define M_SQ(x) ((x)*(x))

static int checkObstacle(double x, double y,
			 double spacing, double radius,
			 double * xnew, double * ynew)
{
  double xmod, ymod;
  double xobst, yobst;
  double r2, R2, R_minus_r;
  PM_CARTESIAN vec;

  // translate arbitrary x,y into 0..SPACING square
  xmod = fmod(x, spacing);
  if (xmod < 0.0) xmod += spacing;
  ymod = fmod(y, spacing);
  if (ymod < 0.0) ymod += spacing;
  // R2 is the obstacle radius squared; saves a sqrt op
  R2 = M_SQ(radius);
  vec.z = 0.0;

  /*
    This snippet checks if xmod,ymod is within the obstacle radius
    from xobst,yobst. If so, it computes the unit vector from the
    obstacle center to the xmod,ymod point, stretches this by the
    radius and adds it to the obstacle center to get the point
    on the obstacle edge nearest xmod,ymod. If the point is at the
    center, it favors a point on the x axis.
   */
#define DOIT \
  r2 = M_SQ(xmod - xobst) + M_SQ(ymod - yobst); \
  if (r2 <= R2) { \
    R_minus_r = radius - sqrt(r2); \
    vec.x = xmod - xobst, vec.y = ymod - yobst; \
    if (! isNorm(vec = unit(vec))) { \
      vec.x = radius, vec.y = vec.z = 0.0; \
    } \
    vec = R_minus_r * vec; \
    *xnew = x + vec.x, *ynew = y + vec.y; \
    return 1; \
  }
  xobst = 0.0, yobst = 0.0; // obstacle 1
  DOIT;
  xobst = spacing; // obstacle 2
  DOIT;
  yobst = spacing; // obstacle 3
  DOIT;
  xobst = 0.0;			// obstacle 4
  DOIT;
#undef DOIT

  return 0;
}

static double w_left_req = 0.0;
static double w_right_req = 0.0;

static void doCmdSkid(ServoMobJAStat * stat)
{
  double alpha_dt;
  double dx, dy, dth;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_EXEC);
    state_next(stat, S0);
  }
  // fall through to steady state

  /*
     Simulation of mid-level driving:
     1. Compute the actual angular speed change we can incur, alpha_dt.
     2. Clamp the requested w_left,right_req so that they change by a
     maximum of alpha_dt.
     3. Run these clamped values through the forward jacobian to get
     simpleWorld.v, simpleWorld.w.
     4. Run simpleWorld.v, simpleWorld.w through the interpolation to get
     differential dx, dy, dth.
     5. Increment simpleWorld.x, simpleWorld.y and simpleWorld.yaw by their
     differentials.
   */

  simpleWorldLock(&simpleWorld);

  alpha_dt = DEFAULT_ALPHA * delta_t;
  if (w_left_req - simpleWorld.wLeft < -alpha_dt)
    simpleWorld.wLeft -= alpha_dt;
  else if (w_left_req - simpleWorld.wLeft > alpha_dt)
    simpleWorld.wLeft += alpha_dt;
  else
    simpleWorld.wLeft = w_left_req;
  if (w_right_req - simpleWorld.wRight < -alpha_dt)
    simpleWorld.wRight -= alpha_dt;
  else if (w_right_req - simpleWorld.wRight > alpha_dt)
    simpleWorld.wRight += alpha_dt;
  else
    simpleWorld.wRight = w_right_req;
  fwdJacobian(simpleWorld.wLeft, simpleWorld.wRight, wheel_radius,
	      wheel_separation_inv, &simpleWorld.v, &simpleWorld.w);
  computeDs(simpleWorld.v, simpleWorld.w, simpleWorld.yaw, delta_t, &dx, &dy,
	    &dth);
  simpleWorld.x += dx;
  simpleWorld.y += dy;
  simpleWorld.yaw += dth;

  simpleWorldUnlock(&simpleWorld);

  state_default(stat);
}

static void doCfgCycleTime(ServoMobJACfgCycleTime * cfg, ServoMobJASet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    if (cfg->cycleTime > 0.0) {
      set->cycleTime = cfg->cycleTime;
      delta_t = set->cycleTime;
      delta_t_inv = 1.0 / delta_t;
      status_next(set, RCS_DONE);
    } else {
      status_next(set, RCS_ERROR);
    }
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

/*
  Teleport locations are in absolute (e.g., GPS) coordinates. We need
  to update the simulated x/y/yaw position so that it transforms to the
  absolute position given.
 */
static void doCfgTeleport(ServoMobJACfgTeleport * cfg, ServoMobJASet * set)
{
  PM_POSE sim;

  if (state_match(set, NEW_COMMAND)) {
    state_new(set); 

    // compute simulated position such that it transforms to given one
    sim = inv(simToAbs) * cfg->goal;

    // set the simulated world
    simpleWorldLock(&simpleWorld);
    simpleWorld.x = sim.tran.x;
    simpleWorld.y = sim.tran.y;
    simpleWorld.yaw = PM_RPY(sim.rot).y;
    simpleWorldUnlock(&simpleWorld);

    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

/*
  iniLoad() will load INI file params from 'filename', returning 0
  if everything is fine, 1 if the file can't be opened or there are
  errors in the values.

  Values loaded:

  [SIMPLE_SIM]
  ; these are all <x> <y> <z> <roll> <pitch> <yaw>, [m] and [rad]
  ; the initial value of absolute position reported in nav data
  ABSOLUTE_ORIGIN
  ; the initial value of relative position reported in nav data
  RELATIVE_ORIGIN
  ; how far the absolute position can jump, [m,rad/jump]
  ABSOLUTE_EXCURSION
  ; how often the jump occurs, [s]
  ABSOLUTE_EXCURSION_INTERVAL = <positive number>
  ; how far the relative position can drift, [m,rad/s]
  RELATIVE_EXCURSION

  [SERVO_MOB_JA]
  CYCLE_TIME = <positive number>

  TYPE = ACKERMAN | SKID | DUAL_INDEPENDENT

  ; for ACKERMAN
  TURNING_RADIUS = <positive number>

  ; for SKID
  SKID_SEPARATION = <positive number>

  ; for DUAL_INDEPENDENT
  WHEEL_RADIUS = <positive number>
  WHEEL_SEPARATION = <positive number>
*/
static int iniLoad(const char *prog, const char *filename)
{
  INIFILE inifile;
  const char SIMPLE_SIM[] = "SIMPLE_SIM";
  const char SERVO_MOB_JA[] = "SERVO_MOB_JA";
  const char ABSOLUTE_ORIGIN[] = "ABSOLUTE_ORIGIN";
  const char RELATIVE_ORIGIN[] = "RELATIVE_ORIGIN";
  const char ABSOLUTE_EXCURSION[] = "ABSOLUTE_EXCURSION";
  const char ABSOLUTE_EXCURSION_INTERVAL[] = "ABSOLUTE_EXCURSION_INTERVAL";
  const char RELATIVE_EXCURSION[] = "RELATIVE_EXCURSION";
  const char CYCLE_TIME[] = "CYCLE_TIME";
  const char TYPE[] = "TYPE";
  const char WHEEL_RADIUS_VAL[] = "WHEEL_RADIUS";
  const char WHEEL_SEPARATION_VAL[] = "WHEEL_SEPARATION";
  const char *section;
  const char *key;
  const char *inistring;
  PM_CARTESIAN tran;
  PM_RPY rpy;
  double d;
  int retval = 0;

  if (0 != inifile.open(filename)) {
    rcs_print_error("%s: can't open INI file %s\n", prog, filename);
    return 1;
  }

  section = SIMPLE_SIM;

  key = ABSOLUTE_ORIGIN;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (6 == sscanf(inistring, "%lf %lf %lf %lf %lf %lf",
		    &tran.x, &tran.y, &tran.z, &rpy.r, &rpy.p, &rpy.y)) {
      // found, and valid
      rpy.r *= TO_RAD, rpy.p *= TO_RAD, rpy.y *= TO_RAD;
      absOriginTran = tran;
      absOriginRpy = rpy;
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      rcs_print_error("%s: using zero values\n", prog);
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    rcs_print_error("%s: using zero values\n", prog);
    retval = 1;
  }

  key = RELATIVE_ORIGIN;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (6 == sscanf(inistring, "%lf %lf %lf %lf %lf %lf",
		    &tran.x, &tran.y, &tran.z, &rpy.r, &rpy.p, &rpy.y)) {
      // found, and valid
      rpy.r *= TO_RAD, rpy.p *= TO_RAD, rpy.y *= TO_RAD;
      relOriginTran = tran;
      relOriginRpy = rpy;
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      rcs_print_error("%s: using zero values\n", prog);
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    rcs_print_error("%s: using zero values\n", prog);
    retval = 1;
  }

  key = ABSOLUTE_EXCURSION;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (6 == sscanf(inistring, "%lf %lf %lf %lf %lf %lf",
		    &tran.x, &tran.y, &tran.z, &rpy.r, &rpy.p, &rpy.y)) {
      // found, and valid
      rpy.r *= TO_RAD, rpy.p *= TO_RAD, rpy.y *= TO_RAD;
      absExcursionTran = tran;
      absExcursionRpy = rpy;
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      rcs_print_error("%s: using zero values\n", prog);
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    rcs_print_error("%s: using zero values\n", prog);
    retval = 1;
  }

  key = ABSOLUTE_EXCURSION_INTERVAL;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d)) {
      // found, and valid
      absExcursionInterval = d;
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      rcs_print_error("%s: using zero value\n", prog);
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    rcs_print_error("%s: using zero value\n", prog);
    retval = 1;
  }

  key = RELATIVE_EXCURSION;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (6 == sscanf(inistring, "%lf %lf %lf %lf %lf %lf",
		    &tran.x, &tran.y, &tran.z, &rpy.r, &rpy.p, &rpy.y)) {
      // found, and valid
      rpy.r *= TO_RAD, rpy.p *= TO_RAD, rpy.y *= TO_RAD;
      relExcursionTran = tran;
      relExcursionRpy = rpy;
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      rcs_print_error("%s: using zero values\n", prog);
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    rcs_print_error("%s: using zero values\n", prog);
    retval = 1;
  }

  section = SERVO_MOB_JA;

  key = CYCLE_TIME;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d) && d > 0.0) {
      // found, and valid
      servoMobJASet.cycleTime = d;
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      servoMobJASet.cycleTime = DEFAULT_CYCLE_TIME;
      rcs_print_error("%s: using %f\n", prog, servoMobJASet.cycleTime);
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    retval = 1;
  }

  key = TYPE;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (!strcmp(inistring, "ACKERMAN")) {
      // ignore
    } else if (!strcmp(inistring, "SKID")) {
      // ignore
    } else if (!strcmp(inistring, "DUAL_INDEPENDENT")) {
      key = WHEEL_RADIUS_VAL;
      if (NULL != (inistring = inifile.find(key, section))) {
	if (1 == sscanf(inistring, "%lf", &d) && d > 0.0) {
	  wheel_radius = d;
	} else {
	  rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
			  prog, inistring, section, key, filename);
	  wheel_radius = DEFAULT_WHEEL_RADIUS;
	  rcs_print_error("%s: using %f\n", prog, wheel_radius);
	  retval = 1;
	}
      } else {
	// not found at all
	rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section,
			key, filename);
	retval = 1;
      }
      key = WHEEL_SEPARATION_VAL;
      if (NULL != (inistring = inifile.find(key, section))) {
	if (1 == sscanf(inistring, "%lf", &d) && d > 0.0) {
	  wheel_separation = d;
	  wheel_separation_inv = 1.0 / wheel_separation;
	} else {
	  rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
			  prog, inistring, section, key, filename);
	  wheel_separation = DEFAULT_WHEEL_SEPARATION;
	  rcs_print_error("%s: using %f\n", prog, wheel_separation);
	  retval = 1;
	}
      } else {
	// not found at all
	rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section,
			key, filename);
	retval = 1;
      }
    } else {
      // something else
      rcs_print_error("%s: unknown [%s] %s '%s' in '%s'\n", prog, section,
		      key, inistring, filename);
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    retval = 1;
  }

  inifile.close();

  return retval;
}

int simpleMobSimInit(int argc, char *argv[])
{
  int retval;
  double then;
  PM_RPY rpy;
  ServoMobJACmdInit servoMobJACmdInit;
  ServoMobJACfgCycleTime servoMobJACfgCycleTime;

  // connect to the NML buffers
  retval = initServoMobJANmlBuffers(NML_FILE, bufnum);
  if (retval) {
    rcs_print_error("%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }

  // read in the INI file params
  if (0 != iniLoad(argv[0], INI_FILE)) {
    rcs_print_error("%s: error loading INI file '%s'\n", argv[0], INI_FILE);
    return 1;
  }

  // initialize and report our status
  servoMobJAStat.echo_serial_number = 0;
  servoMobJAStat.heartbeat = 0;
  servoMobJAStatBuf->write(&servoMobJAStat);

  // initialize and report our settings
  servoMobJASet.echo_serial_number = 0;
  servoMobJASet.type = MOAST_SKID_STEER_TYPE;
  servoMobJASet.parameters[WHEEL_RADIUS] = wheel_radius;
  servoMobJASet.parameters[WHEEL_SEPARATION] = wheel_separation;
  servoMobJASet.parameters_length = 2;
  servoMobJASetBuf->write(&servoMobJASet);

  // stuff an init cmd
  servoMobJACmdInit.serial_number = 1;
  servoMobJACmdBuf->write(&servoMobJACmdInit);

  // stuff a cycle time cfg
  servoMobJACfgCycleTime.serial_number = 1;
  servoMobJACfgCycleTime.cycleTime = servoMobJASet.cycleTime;
  servoMobJACfgBuf->write(&servoMobJACfgCycleTime);

  // set up nav data constants
  navDataExt.utmZone = 18;
  navDataExt.utmLetter = 'S';
  navDataExt.steeringCrabAngle = 0.0;

  // initialize our PM_POSE rep of world
  simPose.tran.x = simpleWorld.x;
  simPose.tran.y = simpleWorld.y;
  simPose.tran.z = 0;
  rpy.r = 0;
  rpy.p = 0;
  rpy.y = simpleWorld.yaw;
  simPose.rot = rpy;

  // set up our coordinate transforms
  simToRel.tran = relOriginTran;
  simToRel.rot = relOriginRpy;
  simToAbs.tran = absOriginTran;
  simToAbs.rot = absOriginRpy;

  // initialize random number generator
  acmRandomInit(&rnd);

  abs_dclock = absExcursionInterval;
  then = etime() - servoMobJASet.cycleTime; // estimate "previous" time stamp

  return 0;
}

static int done = 0;

int simpleMobSimQuit(void)
{
  done = 1;

  return 0;
}

void *simpleMobSimStep(void *arg)
{
  NMLTYPE command_type, cfg_type;
  int serial_number;
  PM_CARTESIAN cart;
  PM_RPY rpy;
  double newx, newy;
#ifdef USE_REAL_TIME
  double now, then = 0.0;
#endif

  done = 0;
  while (!done) {
    // compute actual elapsed time
#ifdef USE_REAL_TIME
    now = etime();
    delta_t = now - then;
    delta_t_inv = 1.0 / delta_t;
    then = now;
#endif

    // drift relative position
    simToRel.tran.x +=
      relExcursionTran.x * delta_t * (2.0 * acmRandomReal(&rnd) - 1.0);
    simToRel.tran.y +=
      relExcursionTran.y * delta_t * (2.0 * acmRandomReal(&rnd) - 1.0);
    simToRel.tran.z +=
      relExcursionTran.z * delta_t * (2.0 * acmRandomReal(&rnd) - 1.0);
    rpy = simToRel.rot;
    rpy.r += relExcursionRpy.r * delta_t * (2.0 * acmRandomReal(&rnd) - 1.0);
    rpy.p += relExcursionRpy.p * delta_t * (2.0 * acmRandomReal(&rnd) - 1.0);
    rpy.y += relExcursionRpy.y * delta_t * (2.0 * acmRandomReal(&rnd) - 1.0);
    simToRel.rot = rpy;

    // jump GPS every ABS_JUMP_INTERVAL seconds
    abs_dclock -= delta_t;
    if (abs_dclock <= 0.0) {
      abs_dclock = absExcursionInterval;
      simToAbs.tran.x = absOriginTran.x +
	absExcursionTran.x * (2.0 * acmRandomReal(&rnd) - 1.0);
      simToAbs.tran.y = absOriginTran.y +
	absExcursionTran.y * (2.0 * acmRandomReal(&rnd) - 1.0);
      simToAbs.tran.z = absOriginTran.z +
	absExcursionTran.z * (2.0 * acmRandomReal(&rnd) - 1.0);
      rpy = simToAbs.rot;
      rpy.r = absOriginRpy.r +
	absExcursionRpy.r * (2.0 * acmRandomReal(&rnd) - 1.0);
      rpy.p = absOriginRpy.p +
	absExcursionRpy.p * (2.0 * acmRandomReal(&rnd) - 1.0);
      rpy.y = absOriginRpy.y +
	absExcursionRpy.y * (2.0 * acmRandomReal(&rnd) - 1.0);
      simToAbs.rot = rpy;
    }

    // read cmd
    command_type = servoMobJACmdBuf->read();
    serial_number = servoMobJACmdBuf->get_address()->serial_number;

    switch (command_type) {
    case 0:
      // no new cmd
      break;

    case -1:
      // comm error
      rcs_print_error("NML cmd read error\n");
      break;

    case SERVO_MOB_JA_CMD_INIT_TYPE:
    case SERVO_MOB_JA_CMD_HALT_TYPE:
    case SERVO_MOB_JA_CMD_ABORT_TYPE:
    case SERVO_MOB_JA_CMD_SHUTDOWN_TYPE:
      servoMobJAStat.command_type = command_type;
      if (serial_number != servoMobJAStat.echo_serial_number) {
	servoMobJAStat.echo_serial_number = serial_number;
	servoMobJAStat.state = NEW_COMMAND;
	servoMobJAStat.status = RCS_EXEC;
      }
      break;

    case SERVO_MOB_JA_CMD_SKID_TYPE:
      servoMobJAStat.command_type = command_type;
      if (serial_number != servoMobJAStat.echo_serial_number) {
	servoMobJAStat.echo_serial_number = serial_number;
	servoMobJAStat.state = NEW_COMMAND;
	servoMobJAStat.status = RCS_EXEC;
      }
      w_left_req = ((ServoMobJACmdSkid *) servoMobJACmdBuf->
		    get_address())->wLeft;
      w_right_req = ((ServoMobJACmdSkid *) servoMobJACmdBuf->
		     get_address())->wRight;
      break;

    default:
      rcs_print_error("unknown cmd: %d\n", (int) command_type);
      break;
    }				// switch (command_type)

    // read cfg
    cfg_type = servoMobJACfgBuf->read();
    serial_number = servoMobJACfgBuf->get_address()->serial_number;

    switch (cfg_type) {
    case 0:
      // no new cmd
      break;

    case -1:
      // comm error
      rcs_print_error("NML cfg read error\n");
      break;

    case SERVO_MOB_JA_CFG_CYCLE_TIME_TYPE:
    case SERVO_MOB_JA_CFG_TELEPORT_TYPE:
      servoMobJASet.command_type = cfg_type;
      if (serial_number != servoMobJASet.echo_serial_number) {
	servoMobJASet.echo_serial_number = serial_number;
	servoMobJASet.state = NEW_COMMAND;
	servoMobJASet.status = RCS_EXEC;
      }
      break;
    }

    // handle cmd state tables
    switch (servoMobJAStat.command_type) {
    case SERVO_MOB_JA_CMD_INIT_TYPE:
      doCmdInit(&servoMobJAStat);
      break;

    case SERVO_MOB_JA_CMD_HALT_TYPE:
      doCmdHalt(&servoMobJAStat);
      break;

    case SERVO_MOB_JA_CMD_ABORT_TYPE:
      doCmdAbort(&servoMobJAStat);
      break;

    case SERVO_MOB_JA_CMD_SHUTDOWN_TYPE:
      doCmdShutdown(&servoMobJAStat);
      break;

    case SERVO_MOB_JA_CMD_SKID_TYPE:
      doCmdSkid(&servoMobJAStat);
      break;

    default:
      // no cmd to handle
      break;
    }				// switch (cmd)

    // handle config state tables
    switch (servoMobJASet.command_type) {
    case SERVO_MOB_JA_CFG_CYCLE_TIME_TYPE:
      doCfgCycleTime((ServoMobJACfgCycleTime *) servoMobJACfgBuf->get_address(), &servoMobJASet);
      break;

    case SERVO_MOB_JA_CFG_TELEPORT_TYPE:
      doCfgTeleport((ServoMobJACfgTeleport *) servoMobJACfgBuf->get_address(), &servoMobJASet);
      break;

    default:
      // no config to handle
      break;
    }

    // check if we have to compute obstacles
    // since we're the only one writing, we can read without locking
    if (simpleWorld.doObstacles &&
	checkObstacle(simpleWorld.x, simpleWorld.y,
		      simpleWorld.obstacleSpacing,
		      simpleWorld.obstacleRadius,
		      &newx, &newy)) {
      // now we have to write, so we need to lock it
      simpleWorldLock(&simpleWorld);
      simpleWorld.x = newx;
      simpleWorld.y = newy;
      simpleWorld.v = simpleWorld.w = 0.0;
      simpleWorld.wLeft = simpleWorld.wRight = 0.0;
      simpleWorldUnlock(&simpleWorld);
    }

    // update stat and set

    navDataExt.time = etime();

    // update our PM_POSE rep of world
    simpleWorldLock(&simpleWorld);
    simPose.tran.x = simpleWorld.x;
    simPose.tran.y = simpleWorld.y;
    rpy.y = simpleWorld.yaw;
    simpleWorldUnlock(&simpleWorld);
    simPose.tran.z = 0;
    rpy.r = 0;
    rpy.p = 0;
    simPose.rot = rpy;
    navDataExt.tranTrue = simPose.tran;
    navDataExt.rpyTrue = simPose.rot;
    navDataExt.tranTrueRate = (navDataExt.tranTrue - cart) * delta_t_inv;
    navDataExt.rpyTrueRate = (navDataExt.rpyTrue - rpy) * delta_t_inv;

    // update our relative pose, and relative rates
    relPose = simToRel * simPose;
    cart = navDataExt.tranRel;
    rpy = navDataExt.rpyRel;
    navDataExt.tranRel = relPose.tran;
    navDataExt.rpyRel = relPose.rot;
    navDataExt.tranRelRate = (navDataExt.tranRel - cart) * delta_t_inv;
    navDataExt.rpyRelRate = (navDataExt.rpyRel - rpy) * delta_t_inv;

    // ditto for absolute pose and rates
    absPose = simToAbs * simPose;
    cart = navDataExt.tranAbs;
    rpy = navDataExt.rpyAbs;
    navDataExt.tranAbs = absPose.tran;
    navDataExt.rpyAbs = absPose.rot;
    navDataExt.tranAbsRate = (navDataExt.tranAbs - cart) * delta_t_inv;
    navDataExt.rpyAbsRate = (navDataExt.rpyAbs - rpy) * delta_t_inv;

    // now for the abs to rel transform
    navDataExt.absToRel = relPose * inv(absPose);

    // and misc
    navDataExt.speed = simpleWorld.v;
    if (fabs(simpleWorld.v) < DBL_EPSILON) {
      navDataExt.steeringCurvature = DBL_MAX;
    } else {
      navDataExt.steeringCurvature = simpleWorld.w / simpleWorld.v;
    }
    servoMobJAStat.heartbeat++;

    servoMobJASetBuf->write(&servoMobJASet);
    servoMobJAStatBuf->write(&servoMobJAStat);
    navDataExtBuf->write(&navDataExt);

    esleep(servoMobJASet.cycleTime);
  }				// while (! done)

  (void) cleanupServoMobJANmlBuffers();

  pthread_exit(0);

  return (void *) 0;
}

/*!
  \defgroup simpleMobSim
  This simulation reads and writes the servoMobJA NML buffers navDataExt,
  servoMobJACmd, servoMobJAStat and servoMobJASet, and simulates
  a dual-independent type servoMobJA's response. The navDataExt fields for
  absolute and relative position are perturbed by simulated GPS jumps
  and INS drift, respectively.

  The entry point for simpleMobSim is the function of the same name,
  simpleMobSim(), which is called cyclically by the main loop in
  simpleSim.cc. That main program creates the NML buffers and sets
  the global variables declared in simpleSim.hh.
*/

/*
  Modification history:

  $Log: simpleMobSim.cc,v $
  Revision 1.7  2007/01/25 15:12:11  proctor
  Added handling of command type and serial number using the _sent values
  of the status buffer, via the SET_AND_RECORD and GOT_IT macros

  Revision 1.6  2006/12/12 19:20:40  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.5  2006/07/26 15:44:22  dr_steveb
  Changes to new buffer and process names

  Revision 1.4  2006/04/07 13:12:12  proctor
  Took out getopt() from the threads, and made NML_FILE, INI_FILE and bufnum
  globals. Apparently getopt() doesn't work from within threads.

  Revision 1.3  2005/12/16 21:59:12  proctor
  simpleMobSim.cc: set navDataExt.time to etime().
  simpleSPSim.cc: ditto, plus filled in the sensorData1D.collectTimes.

  Revision 1.2  2005/12/15 20:24:33  proctor
  Fixed bug with delta_t_inv always being 0, and changed navDataExt.time to
  be floating point seconds instead of original integer milliseconds.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.21  2005/10/11 20:11:52  proctor
  Took out extra elbow room, since it wasn't needed-- same as bigger obstacle

  Revision 1.20  2005/10/05 20:09:13  proctor
  Added 0.1 extra elbow room when moving outside the obstacle

  Revision 1.19  2005/10/05 19:26:33  proctor
  Fixed fmod bug where negative values didn't map into 0..mod region, and caused
  obstacle checking to fail.

  Revision 1.18  2005/09/30 18:55:50  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.17  2005/07/26 21:15:33  proctor
  Fixed up the .ini file handling a bit

  Revision 1.16  2005/07/20 21:54:46  proctor
  simpleMis,Mob,SPSim.cc: added quitting upon shutdown.
  simpleSPSim.cc: also added cmd and cfg interface to make it a full RCS module.

  Revision 1.15  2005/07/19 18:56:58  proctor
  Added .ini file setting of obstacle params

  Revision 1.14  2005/07/19 16:46:00  proctor
  Added iniLoad support for [SIMPLE_SIM]

  Revision 1.13  2005/07/19 14:37:48  proctor
  Changed teleport elements from cart, rpy to pose

  Revision 1.12  2005/07/19 14:02:25  proctor
  Disabled GPS and INS jump/drift; fixed teleport to accept GPS (abs) coords

  Revision 1.11  2005/07/18 21:16:51  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.10  2005/07/13 20:36:29  proctor
  Handled cycle time configuration

  Revision 1.9  2005/07/13 19:48:20  proctor
  Fixed handling of sensorHeight, sensorPitch globals

  Revision 1.8  2005/07/12 21:43:35  proctor
  Added tilted ladar simulation

  Revision 1.7  2005/07/07 19:04:16  proctor
  Fixed obstacle jumping back to origin

  Revision 1.6  2005/07/06 22:27:29  proctor
  Stuffed servoMobJAStat's new heartbeat

  Revision 1.5  2005/06/29 14:32:05  proctor
  Took out annoying .nml, .ini file printing; added obstacle collisions
  in simpleMobSim.cc

  Revision 1.4  2005/06/28 20:48:26  proctor
  Changed DO_CMD to doCmd, for consistency

  Revision 1.3  2005/06/27 21:04:33  proctor
  Took out extra "true" transform, adding SimpleWorld structure.

  Revision 1.2  2005/06/23 17:54:58  proctor
  Changed NML connect name to servoMob

  Revision 1.1  2005/06/21 20:17:59  proctor
  Initial checkin of this code, which splits out original simpleSim.cc into
  a mobility thread.

 */
