/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file moastshell.cpp

  \brief Command line shell to test Moast superior skin.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <string.h>		// strncmp()
#include <stdlib.h>		// getenv()
#include <ctype.h>		// isspace
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include <navDataExt.hh>
#include <sensorData.hh>
#include <servoMobJA.hh>
#include <servoMisJA.hh>
#include <servoEffJA.hh>
#include <servoSP.hh>
#include "moastbufs.h"
#include "ulapi.h"

static char thisProcess[] = "servoShell";

#define DEFAULT_NML_FILE "moast.nml"
static char nmlFile[MOAST_FILE_NAME_LEN] = DEFAULT_NML_FILE;

static int packageIndex = 0;
static int effectorIndex = 0;

static int missionNumber = 1;

static enum {WHICH_MOB, WHICH_MIS, WHICH_EFF, WHICH_SP, WHICH_SONAR, WHICH_RANGE, WHICH_NAV} whichSub = WHICH_MOB;
static enum {WHICH_STAT, WHICH_SET} whichLook = WHICH_STAT;

static void printServoMobJAStat(ServoMobJAStat *stat)
{
  printf("command_type:       %s\n", servoMobJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
}

static void printServoMobJAStatGrdVeh(ServoMobJAStatGrdVeh *stat)
{
  printf("                    Ground Vehicle\n");
  printf("command_type:       %s\n", servoMobJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
}

static void printServoMobJAStatAirBot(ServoMobJAStatAirBot *stat)
{
  printf("                    Air Robot\n");
  printf("command_type:       %s\n", servoMobJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
}

static void printServoMobJASet(ServoMobJASet *set)
{
  printf("command_type:       %s\n", servoMobJA_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("cycle_time:         %f\n", (double) set->cycleTime);
}

static void printServoMobJASetGrdVeh(ServoMobJASetGrdVeh *set)
{
  printf("                    Ground Vehicle\n");
  printf("command_type:       %s\n", servoMobJA_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("cycle_time:         %f\n", (double) set->cycleTime);
  printf("steerType:          %s\n", set->steerType == MOAST_SKID_STEER_TYPE ? "Skid" : set->steerType == MOAST_ACKERMAN_STEER_TYPE ? "Ackerman" : set->steerType == MOAST_OMNI_STEER_TYPE ? "Omni" : "?");
}

static void printServoMobJASetAirBot(ServoMobJASetAirBot *set)
{
  printf("                    Air Robot\n");
  printf("command_type:       %s\n", servoMobJA_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("cycle_time:         %f\n", (double) set->cycleTime);
}

static void printServoMisJAStat(ServoMisJAStat *stat)
{
  int t;

  printf("command_type:       %s\n", servoMisJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
  for (t = 0; t < stat->linkStat_length && t < SERVO_MIS_JA_LINK_MAX; t++) {
    printf("joint %d value:     %f\n", stat->linkStat[t].linkID, stat->linkStat[t].jointVal);
    printf("        torque:    %f\n", stat->linkStat[t].torque);
  }
}

static char *itoa(int i)
{
  static char str[3*sizeof(i)];
  sprintf(str, "%d", i);
  return str;
}

static void printServoMisJASet(ServoMisJASet *set)
{
  int t;

  printf("command_type:       %s\n", servoMisJA_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("cycle_time:         %f\n", (double) set->cycleTime);
  for (t = 0; t < set->linkSet_length && t < SERVO_MIS_JA_LINK_MAX; t++) {
    printf("joint %d parent:    %d\n", set->linkSet[t].linkID, set->linkSet[t].parentLinkID);
    printf("      type:         %s\n", set->linkSet[t].jType == SERVO_MIS_JA_REVOLUTE_JOINT_TYPE ? "revolute" : set->linkSet[t].jType == SERVO_MIS_JA_PRISMATIC_JOINT_TYPE ? "prismatic" : set->linkSet[t].jType == SERVO_MIS_JA_SCISSOR_JOINT_TYPE ? "scissor" : itoa(set->linkSet[t].jType));
    printf("      max speed:    %f\n", set->linkSet[t].maxVelocity);
    printf("      max torque:   %f\n", set->linkSet[t].maxTorque);
    printf("      max range:    %f\n", set->linkSet[t].maxRange);
    printf("      min range:    %f\n", set->linkSet[t].minRange);
    printf("      pose:         %f,%f,%f %f,%f,%f\n", 
	   set->linkSet[t].loc.x, set->linkSet[t].loc.y, set->linkSet[t].loc.z,
	   set->linkSet[t].rot.r, set->linkSet[t].rot.p, set->linkSet[t].rot.y);
  }
}

static void printServoEffJAStat(ServoEffJAStat *stat)
{
  printf("command_type:       %s\n", servoEffJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
}

static void printServoEffJASet(ServoEffJASet *set)
{
  printf("command_type:       %s\n", servoEffJA_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("cycle_time:         %f\n", (double) set->cycleTime);
}

static void printServoSPStat(ServoSPStat *stat)
{
  printf("command_type:       %s\n", servoSP_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
}

static void printServoSPSet(ServoSPSet *set)
{
  printf("command_type:       %s\n", servoSP_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("cycle_time:         %f\n", (double) set->cycleTime);
}

static void printSonarData(void)
{
  SensorData3D sonarData;
  NMLTYPE type;
  int t;
  int num;

  if (getSonarBuf(thisProcess, nmlFile, missionNumber)) {
    type = sonarBuf->read();
    switch (type) {
    case 0:
      printf("no new data\n");
      break;
    case -1:
      printf("read error\n");
      break;
    case SENSOR_DATA_3D_TYPE:
      sonarData = *(reinterpret_cast<SensorData3D *>(sonarBuf->get_address()));
      num = MAX_SENSOR_DATA_3D;
      if (sonarData.element_length < num) num = sonarData.element_length;
      printf("%d sonars:\n", num);
      for (t = 0; t < num; t++) {
	printf("%f : %f %f,%f,%f\n", sonarData.element[t].time, sonarData.element[t].range, sonarData.element[t].pose.tran.x, sonarData.element[t].pose.tran.y, sonarData.element[t].pose.tran.z);
      }
      break;
    }
  } else {
    printf("data not available\n");
  }
  
  return;
}

static void printRangescannerData(int number)
{
  SensorData1D range;
  NMLTYPE type;
  int index = number - 1;
  int t;
  int num;

  if (getRangescannerBuf(thisProcess, nmlFile, index, missionNumber)) {
    type = rangescannerBuf[index]->read();
    switch (type) {
    case 0:
      printf("no new data\n");
      break;
    case -1:
      printf("read error\n");
      break;
    case SENSOR_DATA_1D_TYPE:
      range = *(reinterpret_cast<SensorData1D *>(rangescannerBuf[index]->get_address()));
      num = MAX_SENSOR_DATA_1D;
      if (range.featArray_length < num) num = range.featArray_length;
      printf("%d ranges at time %f:\n", num, range.collectTimeStart);
      for (t = 0; t < num; t++) {
	printf("%f %f\n", range.start + t*range.increment, range.featArray[t]);
      }
      printf("\n");
      break;
    }
  } else {
    printf("data not available\n");
  }

  return;
}

static void printNavDataExt(void)
{
  NavDataExt nav;
  NMLTYPE type;

  if (getNavDataExtBuf(thisProcess, nmlFile, missionNumber)) {
    type = navDataExtBuf->read();
    switch (type) {
    case 0:
      printf("no new data\n");
      break;
    case -1:
      printf("read error\n");
      break;
    case NAV_DATA_EXT_TYPE:
      nav = *(reinterpret_cast<NavDataExt *>(navDataExtBuf->get_address()));
      printf("%f : %f,%f,%f %f,%f,%f\n",
	     nav.time,
	     nav.tranAbs.x,
	     nav.tranAbs.y,
	     nav.tranAbs.z,
	     nav.rpyAbs.r,
	     nav.rpyAbs.p,
	     nav.rpyAbs.y);
      break;
    }
  } else {
    printf("data not available\n");
  }
  
  return;
}

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}

int main(int argc, char *argv[])
{
  int c;
  int this_option_optind;
  int option_index;
  struct option long_options[] = {
    /* --option, needs arg, 0, 'shortcut char' or 0 */
    {"nml", required_argument, 0, 'n'},
    {"mission", required_argument, 0, 'm'},
    {0, 0, 0, 0}
  };
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  enum {PROMPTSIZE = 16};
  char prompt[PROMPTSIZE];
  char *line;
  char *ptr;
  int t;
  NMLTYPE mobStatType = 0;
  NMLTYPE mobSetType = 0;

  ulapi_init(UL_USE_DEFAULT);

  opterr = 0;
  while (1) {
    this_option_optind = optind ? optind : 1;
    option_index = 0;

    /* the leading colon means generate a ':' for missing parameters */
    /* colons mean a required argument for the preceding option */
    c = getopt_long(argc, argv, ":n:m:", long_options, &option_index);
    if (c == -1) break;

    switch (c) {
      /* this handles --long args with no shortcut characters */
    case 0:		
      if (! strcmp("nml", long_options[option_index].name)) {
	strncpy(nmlFile, optarg, MOAST_FILE_NAME_LEN);
	nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
      } else if (! strcmp("mission", long_options[option_index].name)) {
	if (1 != sscanf(optarg, "%i", &missionNumber)) {
	  fprintf(stderr, "bad value for mission number: %s\n", optarg);
	  return 1;
	}
      } else {
	/* should never get here, since 0 above means OK */
	fprintf(stderr, "unknown option `%s'", long_options[option_index].name);
	if (optarg) {
	  fprintf(stderr, " with arg %s", optarg);
	}
	fprintf(stderr, "\n");
	return 1;
      }
      break;

      /* the rest handle -short args */
    case 'n':
      strncpy(nmlFile, optarg, MOAST_FILE_NAME_LEN);
      nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
      break;

    case 'm':
      if (1 != sscanf(optarg, "%i", &missionNumber)) {
	fprintf(stderr, "bad value for mission number: %s\n", optarg);
	return 1;
      }
      break;

    case '?':
      fprintf(stderr, "unknown option `%s'\n", argv[this_option_optind]);
      return 1;
      break;

    case ':':
      fprintf(stderr, "missing parameter to `%s'\n", argv[this_option_optind]);
      return 1;
      break;

    default:
      fprintf(stderr, "bad argument `%s'\n",argv[this_option_optind]);
      return 1;
      break;
    }
  }
  if (optind < argc) {
    fprintf(stderr, "extra arguments provided:");
    while (optind < argc) {
      fprintf(stderr, " %s", argv[optind++]);
    }
    fprintf(stderr, "\n");
    return 1;
  }

  initMoastBufs();

  for (t = 0; t < MAX_SENSOR_DATA_1D_BUFFERS; t++) {
    rangescannerBuf[t] = NULL;
  }

  if (0 != getServoMobJABufs(thisProcess, nmlFile, missionNumber)) {
    cleanupNml();
    return 1;
  }
  // just get the first one now, and wait for a 'num' command
  // to get another
  if (0 != getServoMisJABufs(thisProcess, nmlFile, 1, missionNumber)) {
    cleanupNml();
    return 1;
  }
  if (0 != getServoEffJABufs(thisProcess, nmlFile, 1, missionNumber)) {
    cleanupNml();
    return 1;
  }
  if (0 != getServoSPBufs(thisProcess, nmlFile, missionNumber)) {
    cleanupNml();
    return 1;
  }

#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  while (!feof(stdin)) {
    if (WHICH_SONAR == whichSub) ulapi_snprintf(prompt, sizeof(prompt), "sonar> ");
    else if (WHICH_RANGE == whichSub) ulapi_snprintf(prompt, sizeof(prompt), "range> ");
    else if (WHICH_NAV == whichSub) ulapi_snprintf(prompt, sizeof(prompt), "nav> ");
    else ulapi_snprintf(prompt, sizeof(prompt), "%s%s> ",
			whichSub == WHICH_MIS ? "mis" :
			whichSub == WHICH_EFF ? "eff" :
			whichSub == WHICH_SP ? "sp" : "mob",
			whichLook == WHICH_SET ? "set" : "stat");
#ifndef HAVE_READLINE_READLINE_H
    printf(prompt);
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) break;
    buffer[BUFFERSIZE - 1] = 0;
    line = buffer;
#else
    line = readline(prompt);
    if (NULL == line) break;
    if (*line) add_history(line);
#endif

    servoMobJAStatBuf->read();
    mobStatType = servoMobJAStatBuf->get_address()->type;
    if (SERVO_MOB_JA_STAT_TYPE == mobStatType) {
      servoMobJAStat = *(reinterpret_cast<ServoMobJAStat *>(servoMobJAStatBuf->get_address()));
    } else if (SERVO_MOB_JA_STAT_GRD_VEH_TYPE == mobStatType) {
      servoMobJAStatGrdVeh = *(reinterpret_cast<ServoMobJAStatGrdVeh *>(servoMobJAStatBuf->get_address()));
    } else if (SERVO_MOB_JA_STAT_AIR_BOT_TYPE == mobStatType) {
      servoMobJAStatAirBot = *(reinterpret_cast<ServoMobJAStatAirBot *>(servoMobJAStatBuf->get_address()));
    }

    servoMobJASetBuf->read();
    mobSetType = servoMobJASetBuf->get_address()->type;
    if (SERVO_MOB_JA_SET_TYPE == mobSetType) {
      servoMobJASet = *(reinterpret_cast<ServoMobJASet *>(servoMobJASetBuf->get_address()));
    } else if (SERVO_MOB_JA_SET_GRD_VEH_TYPE == mobSetType) {
      servoMobJASetGrdVeh = *(reinterpret_cast<ServoMobJASetGrdVeh *>(servoMobJASetBuf->get_address()));
    } else if (SERVO_MOB_JA_SET_AIR_BOT_TYPE == mobSetType) {
      servoMobJASetAirBot = *(reinterpret_cast<ServoMobJASetAirBot *>(servoMobJASetBuf->get_address()));
    }

    servoMisJAStatBuf[packageIndex]->read();
    if (SERVO_MIS_JA_STAT_TYPE == servoMisJAStatBuf[packageIndex]->get_address()->type) {
      servoMisJAStat[packageIndex] = *(reinterpret_cast<ServoMisJAStat *>(servoMisJAStatBuf[packageIndex]->get_address()));
    }
    servoMisJASetBuf[packageIndex]->read();
    if (SERVO_MIS_JA_SET_TYPE == servoMisJASetBuf[packageIndex]->get_address()->type) {
      servoMisJASet[packageIndex] = *(reinterpret_cast<ServoMisJASet *>(servoMisJASetBuf[packageIndex]->get_address()));
    }

    servoEffJAStatBuf[effectorIndex]->read();
    if (SERVO_EFF_JA_STAT_TYPE == servoEffJAStatBuf[effectorIndex]->get_address()->type) {
      servoEffJAStat[effectorIndex] = *(reinterpret_cast<ServoEffJAStat *>(servoEffJAStatBuf[effectorIndex]->get_address()));
    }
    servoEffJASetBuf[effectorIndex]->read();
    if (SERVO_EFF_JA_SET_TYPE == servoEffJASetBuf[effectorIndex]->get_address()->type) {
      servoEffJASet[effectorIndex] = *(reinterpret_cast<ServoEffJASet *>(servoEffJASetBuf[effectorIndex]->get_address()));
    }

    servoSPStatBuf->read();
    if (SERVO_SP_STAT_TYPE == servoSPStatBuf->get_address()->type) {
      servoSPStat = *(reinterpret_cast<ServoSPStat *>(servoSPStatBuf->get_address()));
    }
    servoSPSetBuf->read();
    if (SERVO_SP_SET_TYPE == servoSPSetBuf->get_address()->type) {
      servoSPSet = *(reinterpret_cast<ServoSPSet *>(servoSPSetBuf->get_address()));
    }

    ptr = line;
    while (isspace(*ptr)) ptr++; // skip whitespace
    if (0 == *ptr) {
      // blank line
#define PRINT_IT							\
      if (WHICH_SONAR == whichSub) printSonarData();			\
      else if (WHICH_RANGE == whichSub) printRangescannerData(1);	\
      else if (WHICH_NAV == whichSub) printNavDataExt();		\
      else if (WHICH_MIS == whichSub) {					\
	if (WHICH_SET == whichLook) printServoMisJASet(&servoMisJASet[packageIndex]);	\
	else printServoMisJAStat(&servoMisJAStat[packageIndex]);			\
      } else if (WHICH_EFF == whichSub) {				\
	if (WHICH_SET == whichLook) printServoEffJASet(&servoEffJASet[effectorIndex]);	\
	else printServoEffJAStat(&servoEffJAStat[effectorIndex]);			\
      } else if (WHICH_SP == whichSub) {				\
	if (WHICH_SET == whichLook) printServoSPSet(&servoSPSet);	\
	else printServoSPStat(&servoSPStat);				\
      } else {								\
	if (WHICH_SET == whichLook) {					\
	  if (SERVO_MOB_JA_SET_TYPE == mobSetType) printServoMobJASet(&servoMobJASet); \
	  else if (SERVO_MOB_JA_SET_GRD_VEH_TYPE == mobSetType) printServoMobJASetGrdVeh(&servoMobJASetGrdVeh);	\
	  else if (SERVO_MOB_JA_SET_AIR_BOT_TYPE == mobSetType) printServoMobJASetAirBot(&servoMobJASetAirBot);	\
	} else {							\
	  if (SERVO_MOB_JA_STAT_TYPE == mobStatType) printServoMobJAStat(&servoMobJAStat); \
	  else if (SERVO_MOB_JA_STAT_GRD_VEH_TYPE == mobStatType) printServoMobJAStatGrdVeh(&servoMobJAStatGrdVeh); \
	  else if (SERVO_MOB_JA_STAT_AIR_BOT_TYPE == mobStatType) printServoMobJAStatAirBot(&servoMobJAStatAirBot); \
	}								\
      }
      PRINT_IT;
    } else if (ptr[0] == 'q' || !strwcmp(ptr, "quit")) {
      break;
    } else if (!strwcmp(ptr, "mob")) {
      whichSub = WHICH_MOB;
      PRINT_IT;
    } else if (!strwcmp(ptr, "mis")) {
      whichSub = WHICH_MIS;
      PRINT_IT;
    } else if (!strwcmp(ptr, "eff")) {
      whichSub = WHICH_EFF;
      PRINT_IT;
    } else if (!strwcmp(ptr, "sp")) {
      whichSub = WHICH_SP;
      PRINT_IT;
    } else if (!strwcmp(ptr, "stat")) {
      whichLook = WHICH_STAT;
      PRINT_IT;
    } else if (!strwcmp(ptr, "set")) {
      whichLook = WHICH_SET;
      PRINT_IT;
    } else if (!strwcmp(ptr, "init")) {
      ServoMobJACmdInit mobMsg;
      ServoMisJACmdInit misMsg;
      ServoEffJACmdInit effMsg;
      ServoSPCmdInit spMsg;
      if (WHICH_MIS == whichSub) {
	SERVO_MIS_JA_CMD_WRITE(packageIndex, misMsg);
      } else if (WHICH_EFF == whichSub) {
	SERVO_EFF_JA_CMD_WRITE(effectorIndex, effMsg);
      } else if (WHICH_SP == whichSub) {
	SERVO_SP_CMD_WRITE(spMsg);
      } else {
	SERVO_MOB_JA_CMD_WRITE(mobMsg);
      }
      whichLook = WHICH_STAT;
    } else if (!strwcmp(ptr, "num")) {
      int i;
      if (1 == sscanf(ptr, "%*s %i", &i)) {
	if (WHICH_MIS == whichSub) {
	  if (i <= 0 || i > SERVO_MIS_JA_MIS_PKG_MAX) {
	    printf("syntax: num <1 .. %d>\n", SERVO_MIS_JA_MIS_PKG_MAX);
	  } else {
	    if (0 != getServoMisJABufs(thisProcess, nmlFile, i, missionNumber)) {
	      printf("can't get servo mis bufs for %d\n", i);
	    } else {
	      packageIndex = i-1;
	    }
	  }
	} else if (WHICH_EFF == whichSub) {
	  if (i <= 0 || i > SERVO_EFF_JA_EFFECTOR_MAX) {
	    printf("syntax: num <1 .. %d>\n", SERVO_EFF_JA_EFFECTOR_MAX);
	  } else {
	    if (0 != getServoEffJABufs(thisProcess, nmlFile, i, missionNumber)) {
	      printf("can't get servo eff bufs for %d\n", i);
	    } else {
	      effectorIndex = i-1;
	    }
	  }
	}
      } else {
	printf("syntax: num <subsystem number>\n");
      }
    } else if (!strwcmp(ptr, "skid")) {
      double d1, d2;
      ServoMobJACmdSkid skidMsg;
      if (2 == sscanf(ptr, "%*s %lf %lf", &d1, &d2)) {
	skidMsg.wLeft = d1, skidMsg.wRight = d2;
	SERVO_MOB_JA_CMD_WRITE(skidMsg);
	whichSub = WHICH_MOB;
	whichLook = WHICH_STAT;
      } else {
	printf("syntax: skid <left speed> <right speed>\n");
      }
    } else if (!strwcmp(ptr, "ackerman")) {
      double d1, d2;
      ServoMobJACmdAckerman ackermanMsg;
      if (2 == sscanf(ptr, "%*s %lf %lf", &d1, &d2)) {
	ackermanMsg.velocity = d1, ackermanMsg.steerAngle = d2;
	SERVO_MOB_JA_CMD_WRITE(ackermanMsg);
	whichSub = WHICH_MOB;
	whichLook = WHICH_STAT;
      } else {
	printf("syntax: ackerman <speed> <steering angle>\n");
      }
#define MISIT(CMD,TYPE)							\
    } else if (!strwcmp(ptr, CMD)) {					\
      double d1;							\
      ServoMisJACmdMove moveMsg;					\
      for (t = 0; t < SERVO_MIS_JA_LINK_MAX; t++) {			\
	while (! isspace(*ptr) && 0 != *ptr) ptr++;			\
	while (isspace(*ptr)) ptr++;					\
	if (1 != sscanf(ptr, "%lf", &d1)) break;			\
	moveMsg.linkCmd[t].linkID = t + 1;				\
	moveMsg.linkCmd[t].value = d1;					\
	moveMsg.linkCmd[t].type = TYPE;					\
	moveMsg.linkCmd_length = moveMsg.linkCmd[t].linkID;		\
      }									\
      if (moveMsg.linkCmd_length > 0) {					\
	SERVO_MIS_JA_CMD_WRITE(packageIndex, moveMsg);			\
	whichSub = WHICH_MIS;						\
	whichLook = WHICH_STAT;						\
      } else {								\
	printf("syntax: %s <sequence of joint values>\n", CMD);		\
      }
      MISIT("mispos", SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE);
      MISIT("misvel", SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE);
      MISIT("mistorq", SERVO_MIS_JA_LINK_CMD_TORQUE_TYPE);
    } else if (!strwcmp(ptr, "ct")) {
      double d1;
      ServoMobJACfgCycleTime mobMsg;
      ServoMisJACfgCycleTime misMsg;
      ServoEffJACfgCycleTime effMsg;
      ServoSPCfgCycleTime spMsg;
      if (1 == sscanf(ptr, "%*s %lf", &d1)) {
	if (WHICH_MIS == whichSub) {
	  misMsg.cycleTime = d1;
	  SERVO_MIS_JA_CFG_WRITE(packageIndex, misMsg);
	} else if (WHICH_EFF == whichSub) {
	  effMsg.cycleTime = d1;
	  SERVO_EFF_JA_CFG_WRITE(effectorIndex, effMsg);
	} else if (WHICH_SP == whichSub) {
	  spMsg.cycleTime = d1;
	  SERVO_SP_CFG_WRITE(spMsg);
	} else {
	  mobMsg.cycleTime = d1;
	  SERVO_MOB_JA_CFG_WRITE(mobMsg);
	}
	whichLook = WHICH_SET;
      } else {
	printf("syntax: ct <cycle time>\n");
      }
    } else if (!strwcmp(ptr, "sonar")) {
      whichSub = WHICH_SONAR;
      printSonarData();
    } else if (!strwcmp(ptr, "range")) {
      whichSub = WHICH_RANGE;
      printRangescannerData(1);
    } else if (!strwcmp(ptr, "nav")) {
      whichSub = WHICH_NAV;
      printNavDataExt();
    } else {
      printf("?\n");
    }

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  cleanupNml();

  ulapi_exit();

  return 0;
}
