#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <ctype.h>
#include <stdarg.h>
#include "ulapi.h"
#include "simware.h"
#include "skin.h"

static int debug = 0;

static void dbprintf(const char * fmt, ...)
{
  va_list ap;

  if (debug) {
    va_start(ap, fmt);
    fprintf(stderr, "usarsiminf: ");
    vfprintf(stderr, fmt, ap);
    va_end(ap);
  }
}

static int verbose = 0;

static void vbprintf(const char * fmt, ...)
{
  va_list ap;

  if (verbose) {
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
  }
}

static void errprintf(const char * fmt, ...)
{
  va_list ap;

  fprintf(stderr, "usarsiminf: ");
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
}

static int no_sup_tell(sw_struct *sw)
{
  errprintf("no sup_tell function defined\n");

  return -1;
}

static int (*sup_tell)(sw_struct *sw) = no_sup_tell;

static int socket_fd = -1;

/* wraps socket writing with a verbose printf, for debugging */
static ulapi_integer usarsim_socket_write(ulapi_integer id, char * buf, ulapi_integer len)
{
  vbprintf(buf);
  return ulapi_socket_write(id, buf, len);
}

static int waiting_for_conf = 0;
static int waiting_for_geo = 0;

#define SOCKET_MUTEX_KEY 1
static void *socket_mutex = NULL;

#define BUFFERLEN 8
static int buildlen = BUFFERLEN;
static char *build = NULL;
static char *build_ptr;
static char *build_end;

#define DELIMITER 10
#define MAX_MSG_LEN 256
#define MAX_TOKEN_LEN 256

/* only works with arrays, not heap */
#define NULLTERM(s) (s)[sizeof(s)-1]=0

static char *get_head(char *msg, char *head)
{
  char *ptr = msg;

  while (isspace(*ptr)) ptr++;	/* skip over space  */
  if (*ptr == 0) return msg;	/* didn't find {  */

  /* copy everything until we hit space or {  */
  while (1) {
    if ((*ptr == 0) ||
	(*ptr == ',') ||
	(*ptr == '}')) {
      /* unexpected delimiter -- return the original pointer */
      return msg;
    }
    if (isspace(*ptr) ||
	(*ptr == '{')) {
      /* expected delimiter -- break to return current pointer */
      break;
    }
    /* else copy it */
    *head++ = *ptr++;
  }
  *head = 0;			/* null terminate  */

  return ptr;
}

static char *get_key(char *msg, char *key)
{
  char *ptr = msg;

  while ((*ptr != 0) && (*ptr != '{')) ptr++; /* find the {  */
  if (*ptr == 0) return msg;	/* didn't find the {  */
  ptr++;			/* skip over the {  */
  while (isspace(*ptr)) ptr++;	/* skip over any space  */

  /* copy everything until we hit space */
  while (1) {
    if ((*ptr == 0) ||
	(*ptr == ',') ||
	(*ptr == '{') ||
	(*ptr == '}')) {
      /* unexpected delimiter -- return the original pointer */
      return msg;
    }
    if (isspace(*ptr)) {
      /* expected delimiter -- break to return current pointer */
      break;
    }
    /* else copy it */
    *key++ = *ptr++;
  }
  *key = 0;			/* null terminate  */

  return ptr;
}

static char *get_value(char *msg, char *value)
{
  char *ptr = msg;

  while ((isspace(*ptr)) || (*ptr == ',')) ptr++; /* skip over delimiters  */

  /* copy everything until we hit a delimiter  */
  while (! isspace(*ptr)) {
    if (*ptr == 0) return msg;	/* didn't finish cleanly */
    if ((*ptr == ',') ||
	(*ptr == '}')) {
      break;
    }
    /* ok to copy  */
    *value++ = *ptr++;
  }
  *value = 0;			/* null terminate  */

  return ptr;
}

typedef struct _usarsim_class_struct {
  sw_struct sw;
  int did_conf;
  int did_geo;
  struct _usarsim_class_struct *next;
} usarsim_class_struct, *usarsim_class_list;

#define usarsim_class_init(TYPE,CS) (CS) = malloc(sizeof(*(CS))); sw_init(&(CS)->sw); sw_set_type(&(CS)->sw,TYPE); (CS)->did_conf = 0; (CS)->did_geo = 0

static int usarsim_class_find(usarsim_class_list ptr, char *name, usarsim_class_list *where)
{
  if (NULL == ptr) return -1;

  while (! sw_no_name(&ptr->sw)) {
    if (sw_match_name(&ptr->sw, name)) {
      /* found it */
      *where = ptr;
      return 0;
    }
    ptr = ptr->next;
  }

  /* a new one-- fill in the terminal empty structure... */
  sw_set_name(&ptr->sw, name);
  ptr->did_conf = 0;
  ptr->did_geo = 0;

  /* ...and get a new terminal empty structure */
  usarsim_class_init(sw_get_type(&ptr->sw), ptr->next);
  if (NULL == ptr->next) return -1;

  *where = ptr;
  return 0;
}

static void usarsim_class_print(usarsim_class_list cs)
{
  usarsim_class_struct *ptr = cs;

  if (NULL == ptr) return;

  while (! sw_no_name(&ptr->sw)) {
    printf("%s %d %d\n", sw_get_name(&ptr->sw), ptr->did_conf, ptr->did_geo);
    ptr = ptr->next;
  }

  return;
}

static usarsim_class_list encoders;
static usarsim_class_list sonars;
static usarsim_class_list rangescanners;
static usarsim_class_list touches;
static usarsim_class_list co2sensors;
static usarsim_class_list groundtruths;
static usarsim_class_list inses;
static usarsim_class_list gpses;
static usarsim_class_list odometries;
static usarsim_class_list victims;

static usarsim_class_list misstas;

static usarsim_class_list effectors;

static usarsim_class_struct robot;

/*
  SEN {Type Encoder} 

  {Name ECLeft Tick 0} {Name ECRight Tick 0} {Name ECLeftBack Tick 0} {Name ECRightBack Tick 0}
*/
static int handle_sen_encoder(char *msg)
{
#define VAR_DECL				\
  char token[MAX_TOKEN_LEN];			\
  char *ptr = msg;				\
  char *nextptr;				\
  int sawname = 0;				\
  int count = 0;				\
  double time = 0;				\
  double d;					\
  usarsim_class_struct def;			\
  usarsim_class_list where = &def
  VAR_DECL;
  int i;

  /*
    NOTE -- the 'where' pointer doesn't get set until "Name" is seen,
    but where->sw may be accessed earlier by "Time". USARsim needs to
    be fixed so that all messages begin with "Name". Since in all
    cases, the only time data precedes "Name" is with "Time", we'll
    save time locally and fill it in when 'tell' is called. Just in
    case, we start off 'where' to point to a dummy 'def' struct. That
    way, any rogue USARsim messages that don't set "Name" until later
    won't cause a null pointer dereference. Any data that precedes
    "Name" (except for "Time") will be lost.
  */

#define TELL(OP)				\
  if (sw_no_name(&where->sw)) {			\
    dbprintf("no name to tell\n");		\
  } else {					\
    sw_set_time(&where->sw, time);		\
    sw_set_op(&where->sw, OP);			\
    sup_tell(&where->sw);			\
  }

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
#define EXPECT(TOKEN)				\
      nextptr = get_value(ptr, token);		\
      if (nextptr == ptr) return -1;		\
      if (strcmp(token, TOKEN)) return -1;	\
      ptr = nextptr;
      EXPECT("Encoder");
    } else if (! strcmp(token, "Name")) {
#define GET_NAME(LIST,OP)						\
      if (sawname) {							\
	TELL(OP);							\
      }									\
      sawname = 1;							\
      nextptr = get_value(ptr, token);					\
      if (nextptr == ptr) return -1;					\
      if (0 != usarsim_class_find(LIST, token, &where)) return -1;	\
      ptr = nextptr;
      GET_NAME(encoders, SW_SEN_ENCODER_STAT);
      EXPECT("Tick");
#define GET_INTEGER(VAR)				\
      nextptr = get_value(ptr, token);			\
      if (nextptr == ptr) return -1;			\
      if (1 != sscanf(token, "%i", &i)) return -1;	\
      where->sw.data.VAR = i;				\
      count++;						\
      ptr = nextptr;
      GET_INTEGER(encoder.tick);
    } else if (! strcmp(token, "Time")) {
#define GET_TIME					\
      nextptr = get_value(ptr, token);			\
      if (nextptr == ptr) return -1;			\
      if (1 != sscanf(token, "%lf", &d)) return -1;	\
      time = d;						\
      ptr = nextptr;
      GET_TIME;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_ENCODER_STAT);

  return count;
}

/*
  SEN {Time 5608.8500} {Type Sonar}

  {Name F1 Range 2.5098} {Name F2 Range 3.3153} {Name F3 Range 4.9950} {Name F4 Range 2.0774} {Name F5 Range 2.0780} {Name F6 Range 5.0000} {Name F7 Range 5.0000} {Name F8 Range 4.9978} {Name R1 Range 4.9978} {Name R2 Range 4.0775} {Name R3 Range 2.9782} {Name R4 Range 2.6137} {Name R5 Range 2.6148} {Name R6 Range 2.9942} {Name R7 Range 3.3052} {Name R8 Range 2.5098}
*/

static int handle_sen_sonar(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Sonar");
    } else if (! strcmp(token, "Time")) {
      GET_TIME;
    } else if (! strcmp(token, "Name")) {
      GET_NAME(sonars, SW_SEN_SONAR_STAT);
      EXPECT("Range");
#define GET_REAL(VAR)					\
      nextptr = get_value(ptr, token);			\
      if (nextptr == ptr) return -1;			\
      if (1 != sscanf(token, "%lf", &d)) return -1;	\
      where->sw.data.VAR = d;				\
      count++;						\
      ptr = nextptr;
      GET_REAL(sonar.range);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_SONAR_STAT);

  return count;
}

/*
  SEN {Time 5608.85} {Type RangeScanner} 

  {Name Scanner1} {Resolution 0.0174} {FOV 3.1415} {Range 11.3059,11.3014,11.3002,...}
*/

static int handle_sen_rangescanner(char *msg)
{
  VAR_DECL;
  int number = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("RangeScanner");
    } else if (! strcmp(token, "Name")) {
      GET_NAME(rangescanners, SW_SEN_RANGESCANNER_STAT);
    } else if (! strcmp(token, "Time")) {
      GET_TIME;
    } else if (! strcmp(token, "Resolution")) {
      GET_REAL(rangescanner.resolution);
    } else if (! strcmp(token, "FOV")) {
      GET_REAL(rangescanner.fov);
    } else if (! strcmp(token, "Range")) {
      /*
	We won't use the usual GET_REAL macro to get range values,
	since there are a bunch separated by commas, and we'll also
	want to keep track of the cumulative number.
      */
      while (1) {
	nextptr = get_value(ptr, token);
	if (nextptr == ptr) {
	  if (0 == number) return -1; /* need at least one range value */
	  else break;
	}
	if (1 != sscanf(token, "%lf", &d)) return -1;
	if (number >= SW_SEN_RANGESCANNER_MAX) {
	  /* drop it */
	} else {
	  where->sw.data.rangescanner.range[number] = d;
	  number++;
	}
	ptr = nextptr;
      }
      /* all ok, so credit the count */
      count++;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  where->sw.data.rangescanner.number = number;
  TELL(SW_SEN_RANGESCANNER_STAT);

  return count;
}

/*
  SEN {Type Touch}

  {Name Touch Touch False}
*/

static int handle_sen_touch(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Touch");
    } else if (! strcmp(token, "Name")) {
      GET_NAME(touches, SW_SEN_TOUCH_STAT);
      /* FIXME -- conf and geo are not handled inside usarsim */
      where->did_conf = where->did_geo = 1;
      EXPECT("Touch");
      /* expecting "True" or "False" */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "False")) where->sw.data.touch.touched = 0;
      else if (! strcmp(token, "True")) where->sw.data.touch.touched = 1;
      else return -1;
      ptr = nextptr;
      /* all ok, so credit the count */
      count++;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_TOUCH_STAT);

  return count;
}

/*
  SEN {Type CO2Sensor}

  {Name CO2} {Gas CO2} {Density 0.00}
*/
static int handle_sen_co2sensor(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("CO2Sensor");
    } else if (! strcmp(token, "Name")) {
      GET_NAME(co2sensors, SW_SEN_CO2_STAT);
      /* FIXME -- conf and geo are not handled inside usarsim */
      where->did_conf = where->did_geo = 1;
    } else if (! strcmp(token, "Gas")) {
      EXPECT("CO2");
    } else if (! strcmp(token, "Density")) {
      GET_REAL(co2sensor.density);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_CO2_STAT);

  return count;
}

/*
  SEN {Time 5608.8500} {Type GroundTruth} {Name GroundTruth} {Location 12.61,-2.68,1.64} {Orientation 0.00,6.23,0.00}
*/

static int handle_sen_groundtruth(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GroundTruth");
    } else if (! strcmp(token, "Time")) {
      GET_TIME;
    } else if (! strcmp(token, "Name")) {
      GET_NAME(groundtruths, SW_SEN_GROUNDTRUTH_STAT);
    } else if (! strcmp(token, "Location")) {
      GET_REAL(groundtruth.position.x);
      GET_REAL(groundtruth.position.y);
      GET_REAL(groundtruth.position.z);
    } else if (! strcmp(token, "Orientation")) {
      GET_REAL(groundtruth.position.roll);
      GET_REAL(groundtruth.position.pitch);
      GET_REAL(groundtruth.position.yaw);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_GROUNDTRUTH_STAT);

  return count;
}

/*
  SEN {Type GPS}

  {Name GPS} {Latitude 39,8.0341,N} {Longitude 77,13.0001,W} {Fix 1} {Satellites 11}
*/

static int handle_sen_gps(char *msg)
{
  VAR_DECL;
  double latdeg;
  double latmin;
  double londeg;
  double lonmin;
  int i;
  int south;
  int west;

  latdeg = londeg = latmin = lonmin = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GPS");
    } else if (! strcmp(token, "Name")) {
      GET_NAME(gpses, SW_SEN_GPS_STAT);
    } else if (! strcmp(token, "Latitude")) {
      /*
	Because our GET_REAL macro prefixes the argument with 'data.',
	we can't use it for local vars. Let's have more macro fun
	with GET_LOCAL_REAL.
      */
#define GET_LOCAL_REAL(VAR)				\
      nextptr = get_value(ptr, token);			\
      if (nextptr == ptr) return -1;			\
      if (1 != sscanf(token, "%lf", &d)) return -1;	\
      VAR = d;						\
      count++;						\
      ptr = nextptr;
      GET_LOCAL_REAL(latdeg);
      GET_LOCAL_REAL(latmin);
      /* this is too specific to define convenience macro */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if ('N' == token[0]) south = 0;
      else if ('S' == token[0]) south = 1;
      else return -1;
      count++;
      ptr = nextptr;
    } else if (! strcmp(token, "Longitude")) {
      GET_LOCAL_REAL(londeg);
      GET_LOCAL_REAL(lonmin);
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if ('E' == token[0]) west = 0;
      else if ('W' == token[0]) west = 1;
      else return -1;
      count++;
      ptr = nextptr;
    } else if (! strcmp(token, "Fix")) {
      GET_INTEGER(gps.fix);
    } else if (! strcmp(token, "Satellites")) {
      GET_INTEGER(gps.satellites);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  where->sw.data.gps.latitude = latdeg + (latmin / 60.0);
  if (south) where->sw.data.gps.latitude = -where->sw.data.gps.latitude;
  where->sw.data.gps.longitude = londeg + (lonmin / 60.0);
  if (west) where->sw.data.gps.longitude = -where->sw.data.gps.longitude;

  TELL(SW_SEN_GPS_STAT);

  return count;
}

/*
  SEN {Type INS}

  {Name INS} {Location 12.61,-2.68,1.84} {Orientation 0.00,0.00,0.00}
*/

static int handle_sen_ins(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("INS");
    } else if (! strcmp(token, "Name")) {
      GET_NAME(inses, SW_SEN_INS_STAT);
    } else if (! strcmp(token, "Location")) {
      GET_REAL(ins.position.x);
      GET_REAL(ins.position.y);
      GET_REAL(ins.position.z);
    } else if (! strcmp(token, "Orientation")) {
      GET_REAL(ins.position.roll);
      GET_REAL(ins.position.pitch);
      GET_REAL(ins.position.yaw);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_INS_STAT);

  return count;
}

/*
  SEN {Type Odometry}

  {Name Odometry} {Pose 0.0000,0.0000,0.0000}

  FIXME -- this needs to have full x y z r p w info
*/

static int handle_sen_odometry(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Odometry");
    } else if (! strcmp(token, "Name")) {
      GET_NAME(odometries, SW_SEN_ODOMETRY_STAT);
    } else if (! strcmp(token, "Pose")) {
      GET_REAL(odometry.position.x);
      GET_REAL(odometry.position.y);
      GET_REAL(odometry.position.z);
      where->sw.data.odometry.position.roll = 0;
      where->sw.data.odometry.position.pitch = 0;
      where->sw.data.odometry.position.yaw = 0;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_ODOMETRY_STAT);

  return count;
}

/*
  SEN {Time 5609.4467} {Type VictSensor} {Status NoVictims}

  FIXME -- the name is required to be VictSensor in the .uc code.
*/

static int handle_sen_victim(char *msg)
{
  VAR_DECL;

  /* FIXME per comment above */
  if (0 != usarsim_class_find(victims, "VictSensor", &where)) return -1;
  sw_set_name(&where->sw, "VictSensor");

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("VictSensor");
    } else if (! strcmp(token, "Time")) {
      GET_TIME;
    } else if (! strcmp(token, "Status")) {
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "Victims")) where->sw.data.victim.victims = 1;
      else if (! strcmp(token, "NoVictims")) where->sw.data.victim.victims = 0;
      else return -1;
      ptr = nextptr;
      /* all ok, so credit the count */
      count++;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_VICTIM_STAT);

  return count;
}

static int handle_sen(char *msg)
{
  char token[MAX_TOKEN_LEN];
  char *ptr = msg;
  char *nextptr;
  int count = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    /* look for {Type <name>}, and pass the whole msg to the sensor */
    if (! strcmp(token, "Type")) {
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "Sonar")) {
	return handle_sen_sonar(msg);
      } else if (! strcmp(token, "RangeScanner")) {
	return handle_sen_rangescanner(msg);
      } else if (! strcmp(token, "Encoder")) {
	return handle_sen_encoder(msg);
      } else if (! strcmp(token, "Touch")) {
	return handle_sen_touch(msg);
      } else if (! strcmp(token, "CO2Sensor")) {
	return handle_sen_co2sensor(msg);
      } else if (! strcmp(token, "GroundTruth")) {
	return handle_sen_groundtruth(msg);
      } else if (! strcmp(token, "GPS")) {
	return handle_sen_gps(msg);
      } else if (! strcmp(token, "INS")) {
	return handle_sen_ins(msg);
      } else if (! strcmp(token, "Odometry")) {
	return handle_sen_odometry(msg);
      } else if (! strcmp(token, "VictSensor")) {
	return handle_sen_victim(msg);
      } else {
	errprintf("Unknown sensor type %s\n", token);
	/* skip it and keep going */
      }
    }
    /* else something else to be handled by sensor, probably {Time #} */
  }

  return count;
}

/*
  STA {Type GroundVehicle}

  {Time 5609.45} {FrontSteer 0.0000} {RearSteer 0.0000} {LightToggle False} {LightIntensity 0} {Battery 1199} {View -1}

  {Time 48.22} {FrontSteer 0.0000} {RearSteer 0.0000} {FRFlipper -0.0006} {FLFlipper -0.0005} {RRFlipper 0.0005} {RLFlipper 0.0006} {LightToggle False} {LightIntensity 0} {Battery 1196} {View -1}

  FIXME -- there's no name here, consequently we can only have one
  vehicle. The name of the vehicle is irrelevant.
*/

static int handle_sta_groundvehicle(char *msg)
{
  VAR_DECL;

  /* since we only have one robot, we just set it explicitly here
     without going through 'usarsim_class_find' and needing a name */
  where = &robot;
  where->sw.type = SW_ROBOT_GROUNDVEHICLE;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GroundVehicle");
    } else if (! strcmp(token, "Time")) {
      GET_TIME;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_ROBOT_STAT);

  return count;
}

/*
  STA {Type BaseMachine}

  {Time 5609.45}

  FIXME -- there's no name here.
*/
static int handle_sta_basemachine(char *msg)
{
  VAR_DECL;

  where = &robot;
  where->sw.type = SW_ROBOT_FIXED;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("BaseMachine");
    } else if (! strcmp(token, "Time")) {
      GET_TIME;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_ROBOT_STAT);

  return count;
}

static int handle_sta(char *msg)
{
  char token[MAX_TOKEN_LEN];
  char *ptr = msg;
  char *nextptr;
  int count = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    /* look for {Type <name>}, and pass the whole msg to the STA */
    if (! strcmp(token, "Type")) {
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "GroundVehicle")) {
	return handle_sta_groundvehicle(msg);
      } else if (! strcmp(token, "BaseMachine")) {
	return handle_sta_basemachine(msg);
      } else {
	errprintf("Unknown STA type %s\n", token);
	/* skip it and keep going */
      }
    }
    /* else something else, probably {Time #} */
  }

  return count;
}

/*
  MISSTA {Time 5609.45} {Name CameraPanTilt} {Link 1} {Value 0.0000} {Torque -20.00} {Link 2} {Value 0.0000} {Torque -20.00}

  MISSTA {Time 48.22} {Name TeleMaxArm} {Link 1} {Value 0.0000} {Torque -300.00} {Link 2} {Value -0.0000} {Torque -300.00} {Link 3} {Value 0.0002} {Torque -300.00} {Link 4} {Value 0.0000} {Torque -300.00} {Link 5} {Value 0.0000} {Torque -300.00} {Link 6} {Value 0.0000} {Torque -300.00} {Link 7} {Value 0.0000} {Torque -300.00} {Link 8} {Value 0.0000} {Torque -300.00} {Link 9} {Value 0.0000} {Torque -300.00} {Link 10} {Value 0.0000} {Torque -300.00} {Link 11} {Value 0.0000} {Torque -300.00}
*/

static int handle_missta(char *msg)
{
  VAR_DECL;
  int i;
  int linkindex = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Time")) {
      GET_TIME;
    } else if (! strcmp(token, "Name")) {
      GET_NAME(misstas, SW_MISPKG_STAT);
      where->sw.data.mispkg.number = 0;
    } else if (! strcmp(token, "Link")) {
      /* expecting number */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (1 != sscanf(token, "%i", &i)) return -1;
      if (i < 1) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = 1;
      } else if (i > SW_MISPKG_LINK_MAX) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = SW_MISPKG_LINK_MAX;
      }
      linkindex = i - 1;
      if (i > where->sw.data.mispkg.number) where->sw.data.mispkg.number = i;
      count++;
      ptr = nextptr;
    } else if (! strcmp(token, "Value")) {
      GET_REAL(mispkg.link[linkindex].position);
    } else if (! strcmp(token, "Torque")) {
      GET_REAL(mispkg.link[linkindex].torque);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_MISPKG_STAT);

  return count;
}

/*
  RES {Type Effecter} {Name Loading} {Status OK}
*/
static int handle_res(char *msg)
{
  static int didit = 0;

  if (! didit) {
    errprintf("not handling %s\n", msg);
    didit = 1;
  }
  return 0;
}

/*
  CONF {Type Encoder} {Name ECLeft} {Resolution 0.0174}
*/
static int handle_conf_encoder(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Encoder");
    } else if (! strcmp(token, "Name")) {
#define CONF_NAME(TYPE,OP)			\
      GET_NAME(TYPE, OP);			\
      where->did_conf = 1;
      CONF_NAME(encoders, SW_SEN_ENCODER_SET);
    } else if (! strcmp(token, "Resolution")) {
      GET_REAL(encoder.resolution);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_ENCODER_SET);

  return count;
}

/*
  CONF {Type Sonar} {Name R8} {MaxRange 5.0000} {MinRange 0.1000} {BeamAngle 0.3491}
*/

static int handle_conf_sonar(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Sonar");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(sonars, SW_SEN_SONAR_SET);
    } else if (! strcmp(token, "MaxRange")) {
      GET_REAL(sonar.maxrange);
    } else if (! strcmp(token, "MinRange")) {
      GET_REAL(sonar.minrange);
    } else if (! strcmp(token, "BeamAngle")) {
      GET_REAL(sonar.beamangle);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_SONAR_SET);

  return count;
}

/*
  CONF {Type RangeScanner} {Name Scanner1} {MaxRange 20.0000} {MinRange 0.1000} {Resolution 0.0174} {Fov 3.1415} {Paning True} {Tilting False}
*/
static int handle_conf_rangescanner(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("RangeScanner");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(rangescanners, SW_SEN_RANGESCANNER_SET);
    } else if (! strcmp(token, "MaxRange")) {
      GET_REAL(rangescanner.maxrange);
    } else if (! strcmp(token, "MinRange")) {
      GET_REAL(rangescanner.minrange);
    } else if (! strcmp(token, "Resolution")) {
      GET_REAL(rangescanner.resolution);
    } else if (! strcmp(token, "Fov")) {
      GET_REAL(rangescanner.fov);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_RANGESCANNER_SET);

  return count;
}

static int handle_conf_touch(char *msg)
{
  static int didit = 0;

  if (! didit) {
    errprintf("not handling %s\n", msg);
    didit = 1;
  }
  return 0;
}

static int handle_conf_co2sensor(char *msg)
{
  static int didit = 0;

  if (! didit) {
    errprintf("not handling %s\n", msg);
    didit = 1;
  }
  return 0;
}

/*
  CONF {Type GroundTruth} {Name GroundTruth} {ScanInterval 0.05}
*/
static int handle_conf_groundtruth(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GroundTruth");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(groundtruths, SW_SEN_GROUNDTRUTH_SET);
    } else if (! strcmp(token, "ScanInterval")) {
      GET_REAL(groundtruth.period);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_GROUNDTRUTH_SET);

  return count;
}

/*
  CONF {Type GPS} {Name GPS} {ScanInterval 0.20}
*/
static int handle_conf_gps(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GPS");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(gpses, SW_SEN_GPS_SET);
    } else if (! strcmp(token, "ScanInterval")) {
      GET_REAL(gps.period);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_GPS_SET);

  return count;
}

/*
  CONF {Type INS} {Name INS} {ScanInterval 0.20}
*/
static int handle_conf_ins(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("INS");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(inses, SW_SEN_INS_SET);
    } else if (! strcmp(token, "ScanInterval")) {
      GET_REAL(ins.period);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_INS_SET);

  return count;
}

/*
  CONF {Type Odometry} {Name Odometry} {ScanInterval 0.2000} {EncoderResolution 0.0099}
*/
static int handle_conf_odometry(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Odometry");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(odometries, SW_SEN_ODOMETRY_SET);
    } else if (! strcmp(token, "ScanInterval")) {
      GET_REAL(odometry.period);
    } else if (! strcmp(token, "EncoderResolution")) {
      GET_REAL(odometry.resolution);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_ODOMETRY_SET);

  return count;
}

/*
  CONF {Type VictSensor} {Name VictSensor} {MaxRange 6.0000} {HorizontalFOV 0.6981} {VerticalFOV 0.6981}
*/
static int handle_conf_victim(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("VictSensor");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(victims, SW_SEN_VICTIM_SET);
    } else if (! strcmp(token, "MaxRange")) {
      GET_REAL(victim.maxrange);
    } else if (! strcmp(token, "HorizontalFOV")) {
      GET_REAL(victim.hfov);
    } else if (! strcmp(token, "VerticalFOV")) {
      GET_REAL(victim.vfov);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_VICTIM_SET);

  return count;
}

/*
  CONF {Type MisPkg} {Name TeleMaxArm} {Link 1} {JointType Revolute} {MaxSpeed 0.17} {MaxTorque 300.00} {MinRange 1.00} {MaxRange 0.00} ...
*/
static int handle_conf_mispkg(char *msg)
{
  VAR_DECL;
  int i;
  int linkindex = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("MisPkg");
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(misstas, SW_MISPKG_SET);
      where->sw.data.mispkg.number = 0;
    } else if (! strcmp(token, "Link")) {
      /* expecting number */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (1 != sscanf(token, "%i", &i)) return -1;
      if (i < 1) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = 1;
      } else if (i > SW_MISPKG_LINK_MAX) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = SW_MISPKG_LINK_MAX;
      }
      linkindex = i - 1;
      ptr = nextptr;
      if (i > where->sw.data.mispkg.number) where->sw.data.mispkg.number = i;
      count++;
    } else if (! strcmp(token, "JointType")) {
      /* expecting "Revolute" or "Prismatic" */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "Prismatic")) {
	where->sw.data.mispkg.link[linkindex].type = SW_LINK_PRISMATIC;
      } else if (! strcmp(token, "Revolute")) {
	where->sw.data.mispkg.link[linkindex].type = SW_LINK_REVOLUTE;
      } else if (! strcmp(token, "Scissor")) {
	where->sw.data.mispkg.link[linkindex].type = SW_LINK_SCISSOR;
      }	else {
	errprintf("bad value for %s JointType: %s\n", where->sw.name, token);
	where->sw.data.mispkg.link[linkindex].type = SW_NONE;	
      }
      ptr = nextptr;
      /* all ok, so credit the count */
      count++;
    } else if (! strcmp(token, "MaxSpeed")) {
      GET_REAL(mispkg.link[linkindex].maxspeed);
    } else if (! strcmp(token, "MaxTorque")) {
      GET_REAL(mispkg.link[linkindex].maxtorque);
    } else if (! strcmp(token, "MinRange")) {
      GET_REAL(mispkg.link[linkindex].minrange);
    } else if (! strcmp(token, "MaxRange")) {
      GET_REAL(mispkg.link[linkindex].maxrange);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_MISPKG_SET);

  return count;
}

/*
  CONF {Type GroundVehicle}

  {Name P2AT} {SteeringType SkidSteered} {Mass 14.0000} {MaxSpeed 5.3850} {MaxTorque 60.0000} {MaxFrontSteer 0.0000} {MaxRearSteer 0.0000}
*/
static int handle_conf_groundvehicle(char *msg)
{
  VAR_DECL;

  where = &robot;		/* one robot, no need to find it */
  where->sw.type = SW_ROBOT_GROUNDVEHICLE;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GroundVehicle");
      where->did_conf = 1;
      ptr = nextptr;
    } else if (! strcmp(token, "SteeringType")) {
      /* expecting "SkidSteered", "AckermanSteered" or "OmniDrive" */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "SkidSteered")) {
	where->sw.data.groundvehicle.steertype = SW_STEER_SKID;
      } else if (! strcmp(token, "AckermanSteered")) {
	where->sw.data.groundvehicle.steertype = SW_STEER_ACKERMAN;
      } else if (! strcmp(token, "OmniDrive")) {
	where->sw.data.groundvehicle.steertype = SW_STEER_OMNI;
      }	else {
	where->sw.data.groundvehicle.steertype = SW_STEER_UNKNOWN;
	errprintf("bad value for SteeeringType: %s\n", token);
      }
      ptr = nextptr;
      /* all ok, so credit the count */
      count++;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_ROBOT_SET);

  return count;
}

/*
  CONF {Type BaseMachine} {Name FactoryControlBot}
*/
static int handle_conf_basemachine(char *msg)
{
  VAR_DECL;

  where = &robot;		/* one robot, no need to find it */
  where->sw.type = SW_ROBOT_FIXED;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("BaseMachine");
      where->did_conf = 1;
      ptr = nextptr;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_ROBOT_SET);

  return count;
}

/*
  CONF {Type Effecter} {Name Roller} {Opcode Animate} {MaxVal 6.00} {MinVal -6.00}
*/
static int handle_conf_effector(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Effecter");	/* sic */
    } else if (! strcmp(token, "Name")) {
      CONF_NAME(effectors, SW_EFFECTOR_SET);
    } else if (! strcmp(token, "MaxVal")) {
      GET_REAL(effector.maxvalue);
    } else if (! strcmp(token, "MinVal")) {
      GET_REAL(effector.minvalue);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }
  /* FIXME -- ignoring Opcode */

  TELL(SW_EFFECTOR_SET);

  return count;
}

static int handle_conf(char *msg)
{
  char token[MAX_TOKEN_LEN];
  char *ptr = msg;
  char *nextptr;
  int count = 0;

  waiting_for_conf = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    /* look for {Type <name>}, and pass the whole msg to the sensor */
    if (! strcmp(token, "Type")) {
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "Sonar")) {
	return handle_conf_sonar(msg);
      } else if (! strcmp(token, "RangeScanner")) {
	return handle_conf_rangescanner(msg);
      } else if (! strcmp(token, "Encoder")) {
	return handle_conf_encoder(msg);
      } else if (! strcmp(token, "Touch")) {
	return handle_conf_touch(msg);
      } else if (! strcmp(token, "CO2Sensor")) {
	return handle_conf_co2sensor(msg);
      } else if (! strcmp(token, "GroundTruth")) {
	return handle_conf_groundtruth(msg);
      } else if (! strcmp(token, "GPS")) {
	return handle_conf_gps(msg);
      } else if (! strcmp(token, "INS")) {
	return handle_conf_ins(msg);
      } else if (! strcmp(token, "Odometry")) {
	return handle_conf_odometry(msg);
      } else if (! strcmp(token, "VictSensor")) {
	return handle_conf_victim(msg);
      } else if (! strcmp(token, "MisPkg")) {
	return handle_conf_mispkg(msg);
      } else if (! strcmp(token, "GroundVehicle")) {
	return handle_conf_groundvehicle(msg);
      } else if (! strcmp(token, "BaseMachine")) {
	return handle_conf_basemachine(msg);
      } else if (! strcmp(token, "Effecter")) { /* sic */
	return handle_conf_effector(msg);
      } else {
	errprintf("Unknown conf type %s\n", token);
	/* skip it and keep going */
      }
    }
  }

  return count;
}

/*
  GEO {Type Encoder} {Name ECRight Location 0.0000,0.0000,0.0000 Orientation 0.0000,0.0000,1.5707 Mount RightFWheel} < {Name ...} >
*/
static int handle_geo_encoder(char *msg)
{
  VAR_DECL;

  vbprintf(msg);

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Encoder");
    } else if (! strcmp(token, "Name")) {
#define GEO_NAME(TYPE,OP)			\
      GET_NAME(TYPE,OP)				\
      where->did_geo = 1;
      GEO_NAME(encoders, SW_SEN_ENCODER_SET);
      EXPECT("Location");
      GET_REAL(encoder.mount.x);
      GET_REAL(encoder.mount.y);
      GET_REAL(encoder.mount.z);
      EXPECT("Orientation");
      GET_REAL(encoder.mount.roll);
      GET_REAL(encoder.mount.pitch);
      GET_REAL(encoder.mount.yaw);
      /* ignore MOUNT RightFWheel */
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_ENCODER_SET);

  return count;
}

/*
  GEO {Type Sonar} {Name F2 Location 0.1850,-0.1150,0.0000 Orientation 0.0000,0.0000,-0.8727 Mount HARD}

  Why didn't the original USARSim designers use the same tag convention
  as for the CONF message?
*/

static int handle_geo_sonar(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Sonar");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(sonars, SW_SEN_SONAR_SET);
      EXPECT("Location");
      GET_REAL(sonar.mount.x);
      GET_REAL(sonar.mount.y);
      GET_REAL(sonar.mount.z);
      EXPECT("Orientation");
      GET_REAL(sonar.mount.roll);
      GET_REAL(sonar.mount.pitch);
      GET_REAL(sonar.mount.yaw);
      /* ignore MOUNT HARD */
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_SONAR_SET);

  return count;
}

/*
  GEO {Type RangeScanner} {Name Scanner1 Location 0.1439,0.0000,-0.0920 Orientation 0.0000,0.0000,0.0000 Mount HARD}
*/

static int handle_geo_rangescanner(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("RangeScanner");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(rangescanners, SW_SEN_RANGESCANNER_SET);
      EXPECT("Location");
      GET_REAL(rangescanner.mount.x);
      GET_REAL(rangescanner.mount.y);
      GET_REAL(rangescanner.mount.z);
      EXPECT("Orientation");
      GET_REAL(rangescanner.mount.roll);
      GET_REAL(rangescanner.mount.pitch);
      GET_REAL(rangescanner.mount.yaw);
      /* ignore MOUNT HARD */
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_RANGESCANNER_SET);

  return count;
}

static int handle_geo_touch(char *msg)
{
  static int didit = 0;

  if (! didit) {
    errprintf("not handling %s\n", msg);
    didit = 1;
  }
  return 0;
}

static int handle_geo_co2sensor(char *msg)
{
  static int didit = 0;

  if (! didit) {
    errprintf("not handling %s\n", msg);
    didit = 1;
  }
  return 0;
}

/*
  GEO {Type GroundTruth} {Name GroundTruth Location 0.0000,0.0000,0.0000 Orientation 0.0000,0.0000,0.0000 Mount HARD}
*/
static int handle_geo_groundtruth(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GroundTruth");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(groundtruths, SW_SEN_GROUNDTRUTH_SET);
      EXPECT("Location");
      GET_REAL(groundtruth.mount.x);
      GET_REAL(groundtruth.mount.y);
      GET_REAL(groundtruth.mount.z);
      EXPECT("Orientation");
      GET_REAL(groundtruth.mount.roll);
      GET_REAL(groundtruth.mount.pitch);
      GET_REAL(groundtruth.mount.yaw);
      /* ignore MOUNT HARD */
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_GROUNDTRUTH_SET);

  return count;
}

/*
  GEO {Type GPS} {Name GPS Location 0.0000,0.0000,0.0000 Orientation 0.0000,-0.0000,0.0000 Mount HARD}
*/
static int handle_geo_gps(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GPS");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(gpses, SW_SEN_GPS_SET);
      EXPECT("Location");
      GET_REAL(gps.mount.x);
      GET_REAL(gps.mount.y);
      GET_REAL(gps.mount.z);
      EXPECT("Orientation");
      GET_REAL(gps.mount.roll);
      GET_REAL(gps.mount.pitch);
      GET_REAL(gps.mount.yaw);
      /* ignore MOUNT HARD */
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_GPS_SET);

  return count;
}

/*
  GEO {Type INS} {Name INS Location 0.0000,0.0000,0.0000 Orientation 0.0000,-0.0000,0.0000 Mount HARD}
*/
static int handle_geo_ins(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("INS");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(inses, SW_SEN_INS_SET);
      EXPECT("Location");
      GET_REAL(ins.mount.x);
      GET_REAL(ins.mount.y);
      GET_REAL(ins.mount.z);
      EXPECT("Orientation");
      GET_REAL(ins.mount.roll);
      GET_REAL(ins.mount.pitch);
      GET_REAL(ins.mount.yaw);
      /* ignore MOUNT HARD */
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_INS_SET);

  return count;
}

/*
  GEO {Type Odometry} {Name Odometry 
Location 0.0000,0.0000,0.0000 Orientation 0.0000,-0.0000,0.0000 Mount HARD}
*/
static int handle_geo_odometry(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Odometry");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(odometries, SW_SEN_ODOMETRY_SET);
      EXPECT("Location");
      GET_REAL(odometry.mount.x);
      GET_REAL(odometry.mount.y);
      GET_REAL(odometry.mount.z);
      EXPECT("Orientation");
      GET_REAL(odometry.mount.roll);
      GET_REAL(odometry.mount.pitch);
      GET_REAL(odometry.mount.yaw);
      /* ignore MOUNT HARD */
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_ODOMETRY_SET);

  return count;
}

/*
  GEO {Type VictSensor} {Name VictSensor Location 0.0600,0.0000,-0.0087 Orientation 0.0000,0.0000,0.0000 Mount CameraPanTilt_Link2}
*/
static int handle_geo_victim(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("VictSensor");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(victims, SW_SEN_VICTIM_SET);
      EXPECT("Location");
      GET_REAL(victim.mount.x);
      GET_REAL(victim.mount.y);
      GET_REAL(victim.mount.z);
      EXPECT("Orientation");
      GET_REAL(victim.mount.roll);
      GET_REAL(victim.mount.pitch);
      GET_REAL(victim.mount.yaw);
      EXPECT("Mount");
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      strncpy(where->sw.data.victim.parent, token, sizeof(where->sw.data.victim.parent));
      NULLTERM(where->sw.data.victim.parent);
      count++;
      ptr = nextptr;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_SEN_VICTIM_SET);

  return count;
}

/*
  GEO {Type MisPkg} {Name TeleMaxArm} {Link 1} {ParentLink -1} {Location 0.1789,-0.0014,-0.0905} {Orientation 3.1415,0.0000,0.0000} {Link 2} {ParentLink 1} {Location 0.0258,-0.0720,0.1566} {Orientation 1.5707,0.0000,0.0000} ...

  FIXME -- for consistency, should be {Link 1 Location x,x,x Orientation x,x,x} {ParentLink -1} ...
*/
static int handle_geo_mispkg(char *msg)
{
  VAR_DECL;
  int i;
  int linkindex = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("MisPkg");
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(misstas, SW_MISPKG_SET);
      where->sw.data.mispkg.number = 0;
    } else if (! strcmp(token, "Link")) {
      /* expecting number */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (1 != sscanf(token, "%i", &i)) return -1;
      if (i < 1) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = 1;
      } else if (i > SW_MISPKG_LINK_MAX) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = SW_MISPKG_LINK_MAX;
      }
      linkindex = i - 1;
      if (i > where->sw.data.mispkg.number) where->sw.data.mispkg.number = i;
      count++;
      ptr = nextptr;
    } else if (! strcmp(token, "ParentLink")) {
      /* expecting number */
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (1 != sscanf(token, "%i", &i)) return -1;
      if (i < -1) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = -1;
      } else if (i > SW_MISPKG_LINK_MAX) {
	errprintf("invalid link number for %s: %d\n", where->sw.name, i);
	i = SW_MISPKG_LINK_MAX;
      }
      /* change convention of base at -1 to base at 0 */
      if (i < 0) i = 0;
      where->sw.data.mispkg.link[linkindex].parent = i;
      count++;
      ptr = nextptr;
    } else if (! strcmp(token, "Location")) {
      GET_REAL(mispkg.link[linkindex].mount.x);
      GET_REAL(mispkg.link[linkindex].mount.y);
      GET_REAL(mispkg.link[linkindex].mount.z);
    } else if (! strcmp(token, "Orientation")) {
      GET_REAL(mispkg.link[linkindex].mount.roll);
      GET_REAL(mispkg.link[linkindex].mount.pitch);
      GET_REAL(mispkg.link[linkindex].mount.yaw);
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_MISPKG_SET);

  return count;
}

/*
  GEO {Type GroundVehicle}

  {Name P2AT} {Dimensions 0.5238,0.4968,0.2913} {COG 0.0000,0.0000,0.0000} {WheelRadius 0.1300} {WheelSeparation 0.4712} {WheelBase 0.2884}
*/
static int handle_geo_groundvehicle(char *msg)
{
  VAR_DECL;

  where = &robot;
  where->sw.type = SW_ROBOT_GROUNDVEHICLE;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("GroundVehicle");
      where->did_geo = 1;
      ptr = nextptr;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_ROBOT_SET);

  return count;
}

/*
  GEO {Type BaseMachine} {Name FactoryControlBot} {Dimensions 0.0000,0.0000,0.0000}
*/
static int handle_geo_basemachine(char *msg)
{
  VAR_DECL;

  where = &robot;
  where->sw.type = SW_ROBOT_FIXED;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("BaseMachine");
      where->did_geo = 1;
      ptr = nextptr;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_ROBOT_SET);

  return count;
}

/*
  GEO {Type Effecter} {Name Roller Location 0.0000,0.0000,0.0000 Orientation -0.0000,-0.0000,-0.0002 Mount UnitLoaderTop_Link1}
*/
static int handle_geo_effector(char *msg)
{
  VAR_DECL;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    if (! strcmp(token, "Type")) {
      EXPECT("Effecter");	/* sic */
    } else if (! strcmp(token, "Name")) {
      GEO_NAME(effectors, SW_EFFECTOR_SET);
      EXPECT("Location");
      GET_REAL(effector.mount.x);
      GET_REAL(effector.mount.y);
      GET_REAL(effector.mount.z);
      EXPECT("Orientation");
      GET_REAL(effector.mount.roll);
      GET_REAL(effector.mount.pitch);
      GET_REAL(effector.mount.yaw);
      EXPECT("Mount");
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      strncpy(where->sw.data.effector.parent, token, sizeof(where->sw.data.effector.parent));
      NULLTERM(where->sw.data.effector.parent);
      count++;
      ptr = nextptr;
    } else {
      /* skip unknown entry  */
      nextptr = get_value(ptr, token);
    }
  }

  TELL(SW_EFFECTOR_SET);

  return count;
}

static int handle_geo(char *msg)
{
  char token[MAX_TOKEN_LEN];
  char *ptr = msg;
  char *nextptr;
  int count = 0;

  waiting_for_geo = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    ptr = nextptr;

    /* look for {Type <name>}, and pass the whole msg to the sensor */
    if (! strcmp(token, "Type")) {
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) return -1;
      if (! strcmp(token, "Sonar")) {
	return handle_geo_sonar(msg);
      } else if (! strcmp(token, "RangeScanner")) {
	return handle_geo_rangescanner(msg);
      } else if (! strcmp(token, "Encoder")) {
	return handle_geo_encoder(msg);
      } else if (! strcmp(token, "Touch")) {
	return handle_geo_touch(msg);
      } else if (! strcmp(token, "CO2Sensor")) {
	return handle_geo_co2sensor(msg);
      } else if (! strcmp(token, "GroundTruth")) {
	return handle_geo_groundtruth(msg);
      } else if (! strcmp(token, "GPS")) {
	return handle_geo_gps(msg);
      } else if (! strcmp(token, "INS")) {
	return handle_geo_ins(msg);
      } else if (! strcmp(token, "Odometry")) {
	return handle_geo_odometry(msg);
      } else if (! strcmp(token, "VictSensor")) {
	return handle_geo_victim(msg);
      } else if (! strcmp(token, "MisPkg")) {
	return handle_geo_mispkg(msg);
      } else if (! strcmp(token, "GroundVehicle")) {
	return handle_geo_groundvehicle(msg);
      } else if (! strcmp(token, "BaseMachine")) {
	return handle_geo_basemachine(msg);
      } else if (! strcmp(token, "Effecter")) { /* sic */
	return handle_geo_effector(msg);
      } else {
	errprintf("Unknown geo type %s\n", token);
	/* skip it and keep going */
      }
    }
  }

  return count;
}

/*
  NFO {Gametype BotDeathMatch} {Level DM-Factory_250} {TimeLimit 0}

  NFO {StartPoses 3} {Name UnitLoader1 Location 10.24,3.54,10.05 Orientation 0.00,0.00,1.60} {Name UnitLoader2 Location 10.24,3.54,10.05 Orientation 0.00,0.00,1.60} {Name UnitLoader3 Location 10.24,3.54,10.05 Orientation 0.00,0.00,1.60}

*/
static int handle_nfo(char *msg)
{
  dbprintf(msg);

  return 1;
}

static int handle_em(char *msg)
{
  char token[MAX_TOKEN_LEN];
  char *ptr = msg;
  char *nextptr = ptr;
  int count = 0;

  while (1) {
    nextptr = get_key(ptr, token);
    if (nextptr == ptr) break;
    dbprintf("key %s\n", token);
    ptr = nextptr;
    nextptr = get_value(ptr, token);
    if (nextptr == ptr) break;
    dbprintf("value %s\n", token);
    ptr = nextptr;
    count++;
    while (1) {
      nextptr = get_value(ptr, token);
      if (nextptr == ptr) break;
      dbprintf("value %s\n", token);
      ptr = nextptr;
    }
  }

  return count;
}

static int do_sen_confs(usarsim_class_list where, char *type)
{
  char str[MAX_MSG_LEN];

  if (NULL == where) return -1;

  while (0 != where->sw.name[0]) {					
    if (! where->did_conf && ! waiting_for_conf) {
      ulapi_snprintf(str, sizeof(str), "GETCONF {Type %s} {Name %s}\r\n", type, where->sw.name);
      NULLTERM(str);
      ulapi_mutex_take(socket_mutex);
      usarsim_socket_write(socket_fd, str, strlen(str));
      ulapi_mutex_give(socket_mutex);		
      waiting_for_conf = 1;				
    } else if (! where->did_geo && ! waiting_for_geo) {
      ulapi_snprintf(str, sizeof(str), "GETGEO {Type %s} {Name %s}\r\n", type, where->sw.name);
      NULLTERM(str);
      ulapi_mutex_take(socket_mutex);
      usarsim_socket_write(socket_fd, str, strlen(str));
      ulapi_mutex_give(socket_mutex);		
      waiting_for_geo = 1;				
    }									
    where = where->next;						
  }									

  return 0;
}

static int do_eff_confs(usarsim_class_list where, char *type)
{
  char str[MAX_MSG_LEN];

  if (NULL == where) return -1;

  while (0 != where->sw.name[0]) {					
    if (! where->did_conf && ! waiting_for_conf) {			
      ulapi_snprintf(str, sizeof(str), "Factory {CmdType GetConf} {ItemType %s}\r\n", type);
      NULLTERM(str);
      ulapi_mutex_take(socket_mutex);					
      usarsim_socket_write(socket_fd, str, strlen(str));		
      ulapi_mutex_give(socket_mutex);					
      waiting_for_conf = 1;						
    } else if (! where->did_geo && ! waiting_for_geo) {		
      ulapi_snprintf(str, sizeof(str), "Factory {CmdType GetGeo} {ItemType %s}\r\n", type);
      NULLTERM(str);
      ulapi_mutex_take(socket_mutex);					
      usarsim_socket_write(socket_fd, str, strlen(str));		
      ulapi_mutex_give(socket_mutex);					
      waiting_for_geo = 1;						
    }									
    where = where->next;						
  }

  return 0;
}

static int do_robot_confs(usarsim_class_list where)
{
  char str[MAX_MSG_LEN];

  if (! where->did_conf && ! waiting_for_conf) {
    ulapi_snprintf(str, sizeof(str), "GETCONF {Type Robot} {Name %s}\r\n", where->sw.name);
    NULLTERM(str);
    ulapi_mutex_take(socket_mutex);
    usarsim_socket_write(socket_fd, str, strlen(str));
    ulapi_mutex_give(socket_mutex);		
    waiting_for_conf = 1;				
  } else if (! where->did_geo && ! waiting_for_geo) {
    ulapi_snprintf(str, sizeof(str), "GETGEO {Type Robot} {Name %s}\r\n", where->sw.name);
    NULLTERM(str);
    ulapi_mutex_take(socket_mutex);
    usarsim_socket_write(socket_fd, str, strlen(str));
    ulapi_mutex_give(socket_mutex);		
    waiting_for_geo = 1;				
  }									

  return 0;
}

/*!
  \return Returns -1 on error, otherwise returns a number indicating how
  many elements were handled.
*/
static int handle_msg(char *msg)
{
  char head[MAX_TOKEN_LEN];
  char *ptr = msg;
  char *nextptr;
  int count;

  vbprintf(msg);

  nextptr = get_head(ptr, head);
  if (nextptr == ptr) {
    errprintf("no head on %s\n", msg);
    return -1;
  }

  if (! strcmp(head, "NFO")) {
    count = handle_nfo(msg);
  } else if (! strcmp(head, "SEN")) {
    count = handle_sen(msg);
  } else if (! strcmp(head, "STA")) {
    count = handle_sta(msg);
  } else if (! strcmp(head, "MISSTA")) {
    count = handle_missta(msg);
  } else if (! strcmp(head, "RES")) {
    count = handle_res(msg);
  } else if (! strcmp(head, "CONF")) {
    count = handle_conf(msg);
  } else if (! strcmp(head, "GEO")) {
    count = handle_geo(msg);
  } else {
    errprintf("%s?\n", msg);
    count = handle_em(msg);
  }

  do_sen_confs(encoders, "Encoder");
  do_sen_confs(sonars, "Sonar");
  do_sen_confs(rangescanners, "RangeScanner");
  do_sen_confs(touches, "Touch");
  do_sen_confs(co2sensors, "CO2Sensor");
  do_sen_confs(groundtruths, "GroundTruth");
  do_sen_confs(inses, "INS");
  do_sen_confs(gpses, "GPS");
  do_sen_confs(odometries, "Odometry");
  do_sen_confs(victims, "VictSensor");
  do_sen_confs(misstas, "MisPkg");

  do_eff_confs(effectors, "Effectors");

  do_robot_confs(&robot);

  return count;
}

/*
  Init string options: 

  -d,--debug : turn debug printing on
  -v,--verbose : turn verbose printing on
  -h,--host <name> : set host name
  -p,--port <port> : set TCP/IP port number
  -t,--type <type name> : set robot class type name
  -n,--name <name> : set robot name
  -s,--start <x y z r p w> : set start location
 */
int skin_inf_init(char *init_string, int (*stell)(sw_struct *sw))
{
#define INIT_STRING_LEN 256
  char str[INIT_STRING_LEN];
  int argc;
  char **argv;
  char *host = "localhost";
  char *type = "";
  char *name = "";
  char *startname = NULL;
  char freeHost = 0;
  char freeType = 0;
  char freeName = 0;
  char freeStartname = 0;
  double x = 0;
  double y = 0;
  double z = 0;
  double roll = 0;
  double pitch = 0;
  double yaw = 0;
  int port = 3000;
  int t;
  int retval;

  argc = ulapi_to_argv(init_string, &argv);

  /*
    Would like to use getopt here, but the 'optreset' feature to
    reset it after being used by main isn't working, so we have
    to do it manually.
  */
  retval = 0;
  for (t = 0; t < argc; t++) {
    if (! strcmp(argv[t], "-d") ||
	! strcmp(argv[t], "--debug")) {
      debug = 1;
    } else if (! strcmp(argv[t], "-v") ||
	       ! strcmp(argv[t], "--verbose")) {
      verbose = 1;
    } else if (! strcmp(argv[t], "-h") ||
	       ! strcmp(argv[t], "--host")) {
      t++;
      if (t == argc) {
	errprintf("inf_init: missing argument to %s\n", argv[t-1]);
	retval = -1;
	break;
      }
      host = malloc(strlen(argv[t]) + 1);
      strcpy(host, argv[t]);
      freeHost = 1;
    } else if (! strcmp(argv[t], "-p") ||
	       ! strcmp(argv[t], "--port")) {
      t++;
      if (t == argc) {
	errprintf("inf_init: missing argument to %s\n", argv[t-1]);
	retval = -1;
	break;
      }
      port = atoi(argv[t]);
    } else if (! strcmp(argv[t], "-t") ||
	       ! strcmp(argv[t], "--type")) {
      t++;
      if (t == argc) {
	errprintf("inf_init: missing argument to %s\n", argv[t-1]);
	retval = -1;
	break;
      }
      type = malloc(strlen(argv[t]) + 1);
      strcpy(type, argv[t]);
      freeType = 1;
    } else if (! strcmp(argv[t], "-n") ||
	       ! strcmp(argv[t], "--name")) {
      t++;
      if (t == argc) {
	errprintf("inf_init: missing argument to %s\n", argv[t-1]);
	retval = -1;
	break;
      }
      name = malloc(strlen(argv[t]) + 1);
      strcpy(name, argv[t]);
      freeName = 1;
    } else if (! strcmp(argv[t], "-s") ||
	       ! strcmp(argv[t], "--start")) {
      t++;
      if (t == argc) {
	errprintf("inf_init: missing argument to %s\n", argv[t-1]);
	retval = -1;
	break;
      }
      if (6 != sscanf(argv[t], "%lf,%lf,%lf,%lf,%lf,%lf", &x, &y, &z, &roll, &pitch, &yaw)) {
	startname = malloc(strlen(argv[t]) + 1);
	strcpy(startname, argv[t]);
	freeStartname = 1;
      }
    } else {
      errprintf("inf_init: unrecognized option: %s\n", argv[t]);
      retval = -1;
      break;
    }
  }

#undef CLEANUP
#define CLEANUP					\
  {						\
    ulapi_free_argv(argc, argv);		\
    if (freeHost) free(host);			\
    freeHost = 0;				\
    if (freeName) free(name);			\
    freeName = 0;				\
    if (freeStartname) free(startname);		\
    freeStartname = 0;				\
  }

  if (0 != retval) {
    CLEANUP;
    return -1;
  }

  if (debug) {
    dbprintf("inf_init: using host %s, port %d, debug %s, verbose %s\n", host, port, debug ? "true" : "false", verbose ? "true" : "false");
  }

  sup_tell = stell;

  socket_fd = ulapi_socket_client(port, host);
  if (socket_fd < 0) {
    errprintf("can't open socket to %s port %d\n", host, port);
    CLEANUP;
    return -1;
  }

  socket_mutex = ulapi_mutex_new(SOCKET_MUTEX_KEY);
  if (NULL == socket_mutex) {
    ulapi_socket_close(socket_fd);
    socket_fd = -1;
    CLEANUP;
    return -1;
  }

  if (NULL != startname) {
    ulapi_snprintf(str, sizeof(str), "STARTPOSES\r\nINIT {Classname USARBot.%s} {Name %s} {Start %s}\r\n", type, name, startname);
  } else {
    ulapi_snprintf(str, sizeof(str), "STARTPOSES\r\nINIT {Classname USARBot.%s} {Name %s} {Location %f,%f,%f} {Rotation %f,%f,%f}\r\n", type, name, x, y, z, roll, pitch, yaw);
  }
  NULLTERM(str);
  ulapi_mutex_take(socket_mutex);
  usarsim_socket_write(socket_fd, str, strlen(str));
  ulapi_mutex_give(socket_mutex);

  /*
    Note:

    Some of the USARsim resources don't announce themselves.  Sensors
    do, but effectors don't. For the ones that don't, like effectors,
    we need to force a request for geo and conf information. It may
    come back empty.

    Better would be to change the USARsim resources to all announce
    themselves.
  */

  ulapi_snprintf(str, sizeof(str), "GETCONF {Type Effecter}\r\n");
  NULLTERM(str);
  ulapi_mutex_take(socket_mutex);
  usarsim_socket_write(socket_fd, str, strlen(str));
  ulapi_mutex_give(socket_mutex);

  ulapi_snprintf(str, sizeof(str), "GETGEO {Type Effecter}\r\n");
  NULLTERM(str);
  ulapi_mutex_take(socket_mutex);
  usarsim_socket_write(socket_fd, str, strlen(str));
  ulapi_mutex_give(socket_mutex);

  ulapi_snprintf(str, sizeof(str), "GETCONF {Type RFID}\r\n");
  NULLTERM(str);
  ulapi_mutex_take(socket_mutex);
  usarsim_socket_write(socket_fd, str, strlen(str));
  ulapi_mutex_give(socket_mutex);

  build = realloc(build, buildlen * sizeof(char));
  build_ptr = build;
  build_end = build + buildlen;

  usarsim_class_init(SW_SEN_ENCODER, encoders);
  usarsim_class_init(SW_SEN_SONAR, sonars);
  usarsim_class_init(SW_SEN_RANGESCANNER, rangescanners);
  usarsim_class_init(SW_SEN_TOUCH, touches);
  usarsim_class_init(SW_SEN_CO2, co2sensors);
  usarsim_class_init(SW_SEN_GROUNDTRUTH, groundtruths);
  usarsim_class_init(SW_SEN_INS, inses);
  usarsim_class_init(SW_SEN_GPS, gpses);
  usarsim_class_init(SW_SEN_ODOMETRY, odometries);
  usarsim_class_init(SW_SEN_VICTIM, victims);

  usarsim_class_init(SW_MISPKG, misstas);

  usarsim_class_init(SW_EFFECTOR, effectors);

  /*
    Note: the name of a robot appears to be ignored, and USARSim will
    set the name to be the type. It's OK to pass the name in geo and
    conf requests; since there can be only one robot, the name is
    ignored.
   */
  sw_set_name(&robot.sw, 0 != name[0] ? name : type);
  /*
    don't know or care what type, since this will be set explicitly in
    the handle_sta_xxx functions
  */
  sw_set_type(&robot.sw, 0);
  robot.did_conf = 0;
  robot.did_geo = 0;

  CLEANUP;

  return 0;
}

int skin_inf_ask(void)
{
  char buffer[BUFFERLEN];
  char *buffer_ptr;
  char *buffer_end;
  ptrdiff_t offset;
  int nchars;

  nchars = ulapi_socket_read(socket_fd, buffer, BUFFERLEN);
  if (-1 == nchars) {		/* bad read */
    return -1;
  }
  if (0 == nchars) {		/* end of file */
    return -1;
  }
  buffer_ptr = buffer;
  buffer_end = buffer + nchars;

  while (buffer_ptr != buffer_end) {
    if (build_ptr == build_end) {
      offset = build_ptr - build;
      buildlen *= 2;
      build = (char *) realloc(build, buildlen * sizeof(char));
      build_ptr = build + offset;
      build_end = build + buildlen;
    }
    *build_ptr++ = *buffer_ptr;
    if (*buffer_ptr++ == DELIMITER) {
      offset = build_ptr - build;
      build_ptr = build;
      build[offset] = 0;
      if (handle_msg(build) < 0) {
	errprintf("inf_ask: error handling %s\n", build);
      }
    }
  }

  return 0;
}

int skin_inf_tell(sw_struct *sw)
{
  char str[MAX_MSG_LEN];
  int t;

  switch (sw->type) {
  case SW_ROBOT_GROUNDVEHICLE:
    switch (sw->op) {

    case SW_ROBOT_SKID_MOVE:
      ulapi_snprintf(str, sizeof(str), "Drive {Left %f} {Right %f}\r\n", sw->data.groundvehicle.leftspeed, sw->data.groundvehicle.rightspeed);
      NULLTERM(str);
      ulapi_mutex_take(socket_mutex);
      usarsim_socket_write(socket_fd, str, strlen(str));
      ulapi_mutex_give(socket_mutex);
      break;

    case SW_ROBOT_ACKERMAN_MOVE:
      /* FIXME -- how do we know if the vehicle is front steered or
	 rear steered? */
      ulapi_snprintf(str, sizeof(str), "Drive {Speed %f} {FrontSteer %f} {RearSteer %f} {Normalized False} {Light False} {Flip False}\r\n", sw->data.groundvehicle.speed, sw->data.groundvehicle.heading, -sw->data.groundvehicle.heading);
      NULLTERM(str);
      ulapi_mutex_take(socket_mutex);
      usarsim_socket_write(socket_fd, str, strlen(str));
      ulapi_mutex_give(socket_mutex);
      break;

    default:
      errprintf("inf_tell: not handling groundvehicle op %d\n", sw->op);
      break;
    } /* switch SW_ROBOT_GROUNDVEHICLE op */
    break;

  case SW_MISPKG:
    switch (sw->op) {

    case SW_MISPKG_POSITION:
    case SW_MISPKG_SPEED:
    case SW_MISPKG_TORQUE:
      /* e.g., MISPKG {Name P2AT} {Link 1} {Value 1.23} {Order 0 | 1 | 2} */
      ulapi_mutex_take(socket_mutex);

      ulapi_snprintf(str, sizeof(str), "MISPKG {Name %s} ", sw->name);
      NULLTERM(str);
      usarsim_socket_write(socket_fd, str, strlen(str));

      for (t = 0; t < sw->data.mispkg.number; t++) {
	ulapi_snprintf(str, sizeof(str), "{Link %d} {Value %f} {Order %d} ", 
		       t + 1, 
		       SW_MISPKG_SPEED == sw->op ? sw->data.mispkg.link[t].speed : SW_MISPKG_TORQUE == sw->op ? sw->data.mispkg.link[t].torque : sw->data.mispkg.link[t].position,
		       SW_MISPKG_SPEED == sw->op ? 1 : SW_MISPKG_TORQUE == sw->op ? 2 : 0);
	NULLTERM(str);
	usarsim_socket_write(socket_fd, str, strlen(str));
      }

      usarsim_socket_write(socket_fd, "\r\n", 2);

      ulapi_mutex_give(socket_mutex);
      break;

    default:
      errprintf("inf_tell: not handling mispkg op %d\n", sw->op);
      break;
    } /* switch SW_MISPKG op */
    break;

  default:
    errprintf("inf_tell: not handling type %s\n", sw_type_to_string(sw->type));
    break;
  } /* switch (sw->type) */

  return 0;
}

int skin_inf_fini(void)
{
  dbprintf("inf_fini");

  return 0;
}
