/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   skelApp.cc
  \brief Skeleton RCS application.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.7 $
  $Date: 2009/01/13 15:49:55 $
  \endcode

  \author Fred Proctor & Stephen Balakirsky
  \date   03/2/2005
*/

#include <stdio.h>              // printf
#include <stddef.h>		// sizeof(), NULL
#include <string.h>		// strcpy()
#include <stdlib.h>		// getenv()
#include <signal.h>		// signal(), SIGINT
#include <unistd.h>		// sleep()
#include <math.h>               // fabs()
#include "rcs.hh"
#ifndef RCS_STAT_SOURCE_FILE_LEN
#define RCS_STAT_SOURCE_FILE_LEN 64
#endif
#include "skelCommon.hh"
#include "skel.hh"

static RCS_CMD_CHANNEL * skelCmdBuf= NULL;
static RCS_STAT_CHANNEL * skelStatBuf = NULL;
static RCS_CMD_CHANNEL * skelCfgBuf= NULL;
static RCS_STAT_CHANNEL * skelSetBuf = NULL;

static SkelStat skelStat;
static SkelSet skelSet;

RCS_TIMER *timer; // cycle timer

#define DEFAULT_NML_FILE "./skel.nml"
#define SKEL_CMD_NAME "skelCmd" // NML cmd buffer name
#define SKEL_STAT_NAME "skelStat" // NLM stat buffer name
#define SKEL_CFG_NAME "skelCfg" // NML stat buffer name
#define SKEL_SET_NAME "skelSet" // NML stat buffer name
#define THIS_PROCESS "skelApp" // NML process name for the server

static int initNmlBuffers(char * config_file)
{
  if (NULL != skelCmdBuf) delete skelCmdBuf;
  skelCmdBuf = new RCS_CMD_CHANNEL(skel_format, SKEL_CMD_NAME, THIS_PROCESS,
				   config_file);
  if (! skelCmdBuf->valid()) {
    delete skelCmdBuf;
    skelCmdBuf = 0;
    return 1;
  }

  if (NULL != skelStatBuf) delete skelStatBuf;
  skelStatBuf = new RCS_STAT_CHANNEL(skel_format, SKEL_STAT_NAME, THIS_PROCESS,
				     config_file);
  if (! skelStatBuf->valid()) {
    delete skelStatBuf;
    skelStatBuf = 0;
    return 1;
  }

  if (NULL != skelCfgBuf) delete skelCfgBuf;
  skelCfgBuf = new RCS_CMD_CHANNEL(skel_format, SKEL_CFG_NAME, THIS_PROCESS, 
				   config_file);
  if (! skelCfgBuf->valid()) {
    delete skelCfgBuf;
    skelCfgBuf = 0;
    return 1;
  }

  if (NULL != skelSetBuf) delete skelSetBuf;
  skelSetBuf = new RCS_STAT_CHANNEL(skel_format, SKEL_SET_NAME, THIS_PROCESS, 
				    config_file);
  if (! skelSetBuf->valid()) {
    delete skelSetBuf;
    skelSetBuf = 0;
    return 1;
  }

  return 0;
}

int cleanupNmlBuffers(void)
{
  if (NULL != skelCmdBuf) delete skelCmdBuf;
  if (NULL != skelStatBuf) delete skelStatBuf;
  if (NULL != skelCfgBuf) delete skelCfgBuf;
  if (NULL != skelSetBuf) delete skelSetBuf;

  skelCmdBuf = NULL;
  skelStatBuf = NULL;
  skelCfgBuf = NULL;
  skelSetBuf = NULL;

  return 0;
}

#define state_match(s,a) (s)->line = (s)->source_line = __LINE__, (s)->state == (a)
#define state_new(s) strncpy((s)->source_file, __FILE__, RCS_STAT_SOURCE_FILE_LEN), (s)->source_file[RCS_STAT_SOURCE_FILE_LEN-1] = 0
#define state_next(s,a) (s)->state = (a)
#define status_next(s,a) (s)->status = (a)
#define state_default(s) (s)->line = (s)->source_line = __LINE__

void doCmdNop(SkelStat *stat)
{
  if (state_match(stat, NEW_COMMAND)){
    state_new(stat);
    state_next(stat, S0);
    status_next(stat, RCS_DONE);
  }
}

void doCmdInit(SkelStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if( skelSet.debug > 1 )
      printf( "In Init command\n" );
    stat->admin_state = ADMIN_INITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

void doCmdGo(SkelStat * stat, double where)
{
  static int direction;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if( skelSet.debug > 1 )
      printf( "In Go command\n" );

    // check if we are allowed to execute this command
    if( stat->admin_state != ADMIN_INITIALIZED ){
      status_next(stat, RCS_ERROR);
      return;
    }
    stat->where = where;
    status_next(stat, RCS_EXEC);
    // determine direction to go
    if( stat->where < stat->here )
      direction = -1;
    else
      direction = 1;
    state_next(stat, S0);
  } else if (state_match(stat, S0)) {
    state_new(stat);
    // see if done, if not increment
    if( fabs(stat->where - stat->here) < .1 )
      state_next(stat, S1);
    else
      stat->here += .05 * direction;
  } else if (state_match(stat, S1)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S2);
  } else {			// this will be S3
    state_default(stat);
    /*
      We always set 'here' to 'where' regardless of whether it's a new
      command or not. This lets the supervisor change the 'where' setting
      without restarting our state table. This can save some time.
    */
    stat->here = stat->where;
  }
}

void doCmdAbort(SkelStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if( skelSet.debug > 1 )
      printf( "In Abort command\n" );
    stat->admin_state = ADMIN_UNINITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}
void doCmdHalt(SkelStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if( skelSet.debug > 1 )
      printf( "In Halt command\n" );
    stat->admin_state = ADMIN_UNINITIALIZED;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}
void doCmdShutdown(SkelStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    if( skelSet.debug > 1 )
      printf( "In Shutdown command\n" );
    stat->admin_state = ADMIN_SHUT_DOWN;
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    cleanupNmlBuffers();
    exit(1);
  }
}

#define LENGTH_MAX 10.0
#define LENGTH_MIN 0.001

#define WIDTH_MAX 10.0
#define WIDTH_MIN 0.001
void doCfgNop(SkelSet *set)
{
  if (state_match(set, NEW_COMMAND)){
    state_new(set);
    state_next(set, S0);
    status_next(set, RCS_DONE);
  }
}

void doCfgCycleTime(SkelSet * set, double cycleTime)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->cycleTime = cycleTime;
    if( skelSet.debug > 1 )
      printf( "Setting CycleTime to %f\n", cycleTime );
    delete timer;
    timer = new RCS_TIMER(set->cycleTime);
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

void doCfgDebug(SkelSet * set, int debug)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->debug = debug;
    if( skelSet.debug > 1 )
      printf( "In Debug config\n" );
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

void doCfgLength(SkelSet * set, double length)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->lengthReq = length;
    if( skelSet.debug > 1 )
      printf( "Config Length to %f\n", length );
    if (set->lengthReq > LENGTH_MAX) {
      set->length = LENGTH_MAX;
      status_next(set, RCS_ERROR);
    }
    else if (set->lengthReq < LENGTH_MIN) {
      set->length = LENGTH_MIN;
      status_next(set, RCS_ERROR);
    }
    else {
      set->length = set->lengthReq;
      status_next(set, RCS_DONE);
    }
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

void doCfgWidth(SkelSet * set, double width)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->widthReq = width;
    if( skelSet.debug > 1 )
      printf( "Config Width to %f\n", width );
    if (set->widthReq > WIDTH_MAX) {
      set->width = WIDTH_MAX;
      status_next(set, RCS_ERROR);
    }
    else if (set->widthReq < WIDTH_MIN) {
      set->width = WIDTH_MIN;
      status_next(set, RCS_ERROR);
    }
    else {
      set->width = set->widthReq;
      status_next(set, RCS_DONE);
    }
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  char configFileDefault[] = DEFAULT_NML_FILE;
  char * configFileEnv;
  char * NMLFile = configFileDefault;
  int retval;
  NMLTYPE cmdType, cfgType;
  int serialNumber;
  SkelCmdNop skelCmdNop;
  SkelCfgNop skelCfgNop;
  
  /****************************************************************** 
   * INITIALIZE - The initialization opens any communication buffers, 
   *              places the syste in a safe known state, and 
   *		  initializes any control parameters.
   */
  
  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (argc > 1) {
    NMLFile = argv[1];
    printf("%s: using specified config file %s\n", argv[0], NMLFile);
  } else {
    configFileEnv = getenv("CONFIG_NML");
    if (NULL != configFileEnv) {
      NMLFile = configFileEnv;
      printf("%s: using environment config file %s\n", argv[0], NMLFile);
    } else {
      printf("%s: using default config file %s\n", argv[0], NMLFile);
    }
  }

  // allocate all the buffers
  retval = initNmlBuffers(NMLFile);

  if (retval) {
    fprintf(stderr,"%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }

  // stuff a nop command
  skelCmdNop.serial_number = 1;
  skelCmdBuf->write(&skelCmdNop);

  // set the status up so that init command goes through
  skelStat.echo_serial_number = 0;
  skelStat.where = 0.0;
  skelStat.here = 0.0;

  // stuff a nop config
  skelCfgNop.serial_number = 1;
  skelCfgBuf->write(&skelCfgNop);

  // set the settings up so that the length goes through
  skelSet.cycleTime = 0.1;
  timer = new RCS_TIMER(skelSet.cycleTime);

  skelSet.echo_serial_number = 0;
  skelSet.length = 0.5 * (LENGTH_MAX - LENGTH_MIN);
  skelSet.width = 0.5 * (WIDTH_MAX - WIDTH_MIN);

  // initialize the time tracker
  skelStat.tt.count = 0;

  done = 0;
  signal(SIGINT, quit);
  while (! done) {
    /******************************
     * House keeping: set time tracker to increment
     */
    cycle_time_tracker( &skelStat.tt );

    /****************************************************************** 
     * READ COMMAND - Command information is received, and the system
     *                is prepared to execute the command. When a new
     *		    command is received, the old command is immediately
     *		    replaced by this command. If the command is not 
     *		    able to be interrupted, error flags are set and
     *		    will be reported during the writng of status
     */

    // read data if there is any to read

    // read command
    cmdType = skelCmdBuf->read();
    serialNumber = skelCmdBuf->get_address()->serial_number;
    switch (cmdType) {
    case 0:
      // no new command
      break;
    case -1:
      // comm error
      fprintf(stderr, "%s: NML command read error\n", argv[0]);
      break;
    case SKEL_CMD_NOP_TYPE:
    case SKEL_CMD_INIT_TYPE:
    case SKEL_CMD_HALT_TYPE:
    case SKEL_CMD_SHUTDOWN_TYPE:
    case SKEL_CMD_ABORT_TYPE:
    case SKEL_CMD_GO_TYPE:
      if (serialNumber != skelStat.echo_serial_number) {
	if( skelSet.debug > 0 )
	  printf( "New %s command received\n", skel_symbol_lookup(cmdType) );
	skelStat.command_type = cmdType;
	skelStat.echo_serial_number = serialNumber;
	skelStat.state = NEW_COMMAND;
	skelStat.status = RCS_EXEC;
      }
      break;
    default:
      fprintf(stderr,"%s: unknown command: %d\n", argv[0], (int) cmdType);
      break;
    } // switch (cmdType)


  /****************************************************************** 
   * READ CONFIG - Configuration information is received, and the 
   *               system is prepared to change its setting. A separate 
   *               configuration channel is provided to allow for control 
   *               parameters to be changed without interrupting the 
   *               current controller. For example, a user may want to 
   *               change a systems cycle time without interrupting a 
   *               complex control function. The configuration channel is 
   *               used to change the cycle time parameters without 
   *               aborting the control algorithm.
   */

    // read config
    cfgType = skelCfgBuf->read();
    serialNumber = skelCfgBuf->get_address()->serial_number;
    switch (cfgType) {
    case 0:
      // no new command
      break;
    case -1:
      // comm error
      fprintf(stderr,"%s: NML config read error\n", argv[0]);
      break;
    case SKEL_CFG_NOP_TYPE:
    case SKEL_CFG_LENGTH_TYPE:
    case SKEL_CFG_WIDTH_TYPE:
    case SKEL_CFG_DEBUG_TYPE:
    case SKEL_CFG_CYCLE_TIME_TYPE:
      if (serialNumber != skelSet.echo_serial_number) {
	if( skelSet.debug > 0 )
	  printf( "New %s config received\n", skel_symbol_lookup(cfgType) );
	skelSet.command_type = cfgType;
	skelSet.echo_serial_number = serialNumber;
	skelSet.state = NEW_COMMAND;
	skelSet.status = RCS_EXEC;
      }
      break;

    default:
      fprintf(stderr,"%s: unknown config: %d\n", argv[0], (int) cfgType);
      break;
    } // switch (cfgType)


  /****************************************************************** 
   * COMMAND EXECUTION - Command execution is carried out. 
   *                     All of the modules run on a fixed cycle time. 
   *                     Therefore, command functions must either 
   *                     guarantee that they finish in under the cycle time, 
   *                     or provide for being reentrant. Although this 
   *                     example refers to the command execution as being 
   *                     finite state machine (FSM) based, search based 
   *                     planning systems have also been implemented under 
   *                     this framework. The searcher simply stores its 
   *                     current location in the search when its time 
   *                     expires and then resumes searching on the next cycle.
   */

    // handle command state tables
    switch (skelStat.command_type) {
    case SKEL_CMD_NOP_TYPE:
      doCmdNop(&skelStat);
      break;
    case SKEL_CMD_INIT_TYPE:
      doCmdInit(&skelStat);
      break;
    case SKEL_CMD_GO_TYPE:
      doCmdGo(&skelStat,((SkelCmdGo *) skelCmdBuf->get_address())->where);
      break;
    case SKEL_CMD_HALT_TYPE:
      doCmdHalt(&skelStat);
      break;
    case SKEL_CMD_ABORT_TYPE:
      doCmdAbort(&skelStat);
      break;
    case SKEL_CMD_SHUTDOWN_TYPE:
      doCmdShutdown(&skelStat);
      break;
    default:
      // no command to handle
      break;
    } // switch (cmd)


  /****************************************************************** 
   * CONFIGURE EXECUTION - Configuration is carried out
   */
    // handle config state tables
    switch (skelSet.command_type) {
    case SKEL_CFG_NOP_TYPE:
      doCfgNop(&skelSet);
      break;
    case SKEL_CFG_DEBUG_TYPE:
      doCfgDebug(&skelSet,((SkelCfgDebug *) skelCfgBuf->get_address())->debug);
      break;
    case SKEL_CFG_CYCLE_TIME_TYPE:
      doCfgCycleTime(&skelSet,((SkelCfgCycleTime *) skelCfgBuf->get_address())->cycleTime);
      break;
    case SKEL_CFG_LENGTH_TYPE:
      doCfgLength(&skelSet,((SkelCfgLength *) skelCfgBuf->get_address())->length );
      break;
    case SKEL_CFG_WIDTH_TYPE:
      doCfgWidth(&skelSet, ((SkelCfgWidth *) skelCfgBuf->get_address())->width);
      break;
    default:
      // no config to handle
      break;
    } // switch (cfg)


  /****************************************************************** 
   * UPDATE STATUS AND SETTING - The current module status and the 
   *                             configuration's settings are sent
   *                             out over communications buffers.
   */
    // update status and settings
    skelStatBuf->write(&skelStat);
    skelSetBuf->write(&skelSet);

    timer->wait();
  }

  return cleanupNmlBuffers();
}

/*
  Modification history:

  $Log: skelApp.cc,v $
  Revision 1.7  2009/01/13 15:49:55  dr_steveb
  Got rid of nmlprint in favor of printf

  Revision 1.6  2008/09/19 15:22:52  dr_steveb
  Final changes to skelApp before tutorial

  Revision 1.5  2008/09/19 02:53:51  dr_steveb
  Made Skel app more readable and fixed makefile for cygwin

  Revision 1.4  2006/03/03 15:52:34  dr_steveb
  Added documentation to make file correspond to the moast reference manual.

  Revision 1.3  2005/10/18 21:41:25  dr_steveb
  Added debug messages so that the debug command does something.

  Revision 1.2  2005/10/18 21:20:30  dr_steveb
  Added cfgCycleTime and cfgDebug.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.1  2005/10/11 18:45:02  stephen
  Initial addition of skeleton rcs controller.

*/
