/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file exampleAStarHeapA.cc
    \brief provides a driver for testing AStarHeap search

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.1 $
  $Date: 2006/11/17 15:25:14 $
  \endcode

  \author Tom Kramer

See the documentation of aStarHeap.hh for an overview of this search routine.

This is an aStar search routine driver for testing aStar search with
incremental graph building. The driver reads the entire graph from a
file picked by the user (the static graph), but drip feeds it to the
Searcher, which builds its graph while searching. The heuristic is the
minimum cost of the kids a node, which is weak but admissible.

The indexes of the nodes in the Searcher (since they are assigned on
the fly according to how the search goes) will differ from the indexes
of the nodes in the test file read by the driver.  To implement
insertKids and to report the path correctly, a map (driveMap) from the
indexes of driver's graph to the indexes of the Searcher's graph is
needed. A map the other way (searchMap) is also needed. The maps are
simply arrays of integers declared here as global variables. If the
index of a node in the driver's nodes is n, then the index of the
corresponding node in the Searcher's nodes is driveMap[n]. The maps
have 200000 entries, so graphs with more nodes than that should not be
used.

The routines here use stdio for reading and writing.

*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "aStarHeap.hh"

#define A_STAR_MAP_SIZE 200000
#define A_STAR_PATH_SIZE 1000

AStarHeapGraph driverGraph;
AStarHeapGraph searcherGraph;
int driveMap[A_STAR_MAP_SIZE];
int searchMap[A_STAR_MAP_SIZE];

/***********************************************************************/

static void getGraphData(int argc, char * argv[], AStarHeapGraph * theGraph);
static double getMinArcCost(int driverN);
static void getStartAndEnd(int * start, int * goal, int size);
static void initSearcherGraph(int start, int goal);
static void insertKids(int searcherDad, int startIndex);
int main(int argc, char * argv[]);
static void printAnswer(int length, int * path, double cost);
static void readGraph(char * fileName, AStarHeapGraph * aGraph);

/***********************************************************************/

//! insertKids adds kids to the dad node and creates new kids
/*
Returned Value: none 

Called By: main
  
The startIndex argument is used to set the index attribute of the
nodeInfos and to update the driveMap and searchMap.

*/

static void insertKids( /* ARGUMENTS                                        */
 int searcherDad,  /*!< index in Searcher nodes of node to insert kids for  */
 int startIndex)   /*!< starting index in Searcher nodes for new added kids */
{
  AStarHeapNode * driverDad; 
  int numberAdded;             // total number of kids added
  int numberNew;               // number of new added kids
  int numberOld;               // number of old added kids
  NodeInfo ** newInfos;        // NodeInfos for new kids added
  double * newArcCosts;        // costs of new kids
  double * newGoalCosts;       // goalCosts of new kids
  int * oldAdded;              // indexes in nodes of old kids added
  double * oldArcCosts;        // costs of old kids added
  int n;                       // index for kids of driverDad
  int j;                       // index for new kids of searcherDad node
  int k;                       // index for old kids of searcherDad node
  int driverN;                 // index in driver nodes of a kid

  numberNew = 0;
  numberOld = 0;
  j = 0;
  k = 0;
  driverDad = (driverGraph.nodes + searchMap[searcherDad]);
  numberAdded = driverDad->kidsSize;
  for (n = 0; n < numberAdded; n++)
    {
      driverN = driverDad->kidsIndexes[n];
      if (driveMap[driverN] >= searcherGraph.size)
	fprintf(stderr, "bug in maintaining driveMap\n");
      else if (driveMap[driverN] >= 0)
	numberOld++;
      else
	numberNew++;
    }
  if (numberNew > 0)
    {
      newInfos = new NodeInfo * [numberNew];
      newArcCosts = new double[numberNew];
      newGoalCosts = new double[numberNew];
    }
  else
    {
      newInfos = 0;
      newArcCosts = 0;
      newGoalCosts = 0;
    }
  if (numberOld > 0)
    {
      oldAdded = new int[numberOld];
      oldArcCosts = new double[numberOld];
    }
  else
    {
      oldAdded = 0;
      oldArcCosts = 0;
    }
  for (n = 0; n < numberAdded; n++)
    {
      driverN = driverDad->kidsIndexes[n];
      if (driveMap[driverN] >= 0)
	{
	  oldAdded[k] = driveMap[driverN];
	  oldArcCosts[k] = driverDad->kidsCosts[n];
	  k++;
	}
      else
	{
	  newInfos[j] = new NodeInfo(startIndex + j);
	  newArcCosts[j] = driverDad->kidsCosts[n];
	  //newGoalCosts[j] = 0.0; // will act like Dijkstra with 0
	  newGoalCosts[j] = getMinArcCost(driverN);
	  driveMap[driverN] = (startIndex + j);
	  searchMap[startIndex + j] = driverN;
	  j++;
	}
    }
  searcherGraph.addKids(searcherDad, numberAdded, numberNew, newInfos,
			newArcCosts, newGoalCosts, oldAdded, oldArcCosts);
  if (numberNew > 0)
    {
      delete [] newInfos;
      delete [] newArcCosts;
      delete [] newGoalCosts;
    }
  if (numberOld > 0)
    {
      delete [] oldAdded;
      delete [] oldArcCosts;
    }
}

/***********************************************************************/

//! getGraphData reads the graph from a file and builds an AStarHeapGraph
/*
Returned Value: none

Called By:  main

This reads the graph whose name is argv[1].

If the number of command arguments is not 1, this prints an error
message and exits.

*/

static void getGraphData(    /* ARGUMENTS                          */
 int argc,                   /*!< number of command line arguments */
 char * argv[],              /*!< array of command line arguments  */
 AStarHeapGraph * theGraph) /*!< graph to build, built here       */
{
  if (argc == 2)
    {
      readGraph(argv[1], theGraph);
    }
  else
    {
      fprintf(stderr, "Usage: %s <fileName>\n", argv[0]);
      exit(1);
    }
}

/***********************************************************************/

//! getMinArcCost finds the minimum cost of the kids of a node
/*
Returned Value: double
  This returns the minimum cost of the kids of nodes[driverN] in
  the driverGraph. This may be used as a heuristic in aStar search,
  since it is clearly never greater than the cost to the goal.

Called By: AStarHeapGraph::getMoreKids

This assumes there is at least one kid.

*/

static double getMinArcCost( /* ARGUMENTS                                 */
 int driverN)         /*!< index in nodes of driverGraph to find cost for */
{
  AStarHeapNode * aNode;
  int m;
  double result;

  aNode = (driverGraph.nodes + driverN);
  result = 1000000.0;
  for (m = 0; m < aNode->kidsSize; m++)
    {
      if (aNode->kidsCosts[m] < result)
	result = aNode->kidsCosts[m];
    }
  return result;
}

/***********************************************************************/

//! getStartAndEnd gets the indexes of the start and goal nodes from the user
/*
Returned Value: none

Called by: main

This asks the user to give the indexes of the start and goal nodes
and sets those indexes. It is checked that the values are in range and
that start does not equal goal.

The fflush(stdin) is needed to handle entry of bad data. Without it,
the loop does not wait for the user to enter something after bad
data has been entered; it just keeps cycling.

*/

static void getStartAndEnd( /* ARGUMENTS                       */
 int * start,               /*!< index of start node, set here */
 int * goal,                /*!< index of goal node, set here  */
 int size)                  /*!< number of nodes in graph      */
{
  *start = -1;
  *goal = -1;
  while ((*start < 0) || (*start >= size) ||
	 (*goal < 0)  || (*goal >= size)  ||
	 (*start == *goal))
    {
      printf("Enter start and goal ==> ");
      fflush(stdout);
      scanf("%d %d", start, goal);
      fflush(stdin);
      if ((*start < 0) || (*start >= size) || (*goal < 0) || (*goal >= size))
	fprintf(stderr, "start and/or goal out of range\n");
      if (*start == *goal)
	fprintf(stderr, "start must differ from goal\n");
    }
}

/***********************************************************************/

//! initSearcherGraph puts the start node & goal nodes into the searcher graph
/*
Returned Value: none

Called By:
  main

This puts two nodes in the searcherGraph and fills them.  The first
one, (index 0) is equivalent to the goal node from the driverGraph.
The second one (index 1) is equivalent to the start node from the
driverGraph. The initialization of this node is done in initSearch,
not here (except that its goal cost is set to zero here -- the value
zero is irrelevant since it is not used).

It also updates the driveMap and the searchMap.

This uses the global variables searcherGraph, driveMap, and searchMap.

*/

static void initSearcherGraph( /* ARGUMENTS                               */
 int start,                    /*!< index in driver's graph of start node */
 int goal)                     /*!< index in driver's graph of goal node  */
{
  searcherGraph.nodes = new AStarHeapNode[2];
  searcherGraph.capacity = 2;
  searcherGraph.size = 2;
  searcherGraph.nodes[0].moreInfo = new NodeInfo(0);
  searcherGraph.nodes[0].goalCost = 0;
  searcherGraph.nodes[0].kidsSize = 0;
  searchMap[0] = goal;
  driveMap[goal] = 0;
  searcherGraph.nodes[1].goalCost = 0;
  searchMap[1] = start;
  driveMap[start] = 1;
}

/***********************************************************************/

//! main runs the driver
/*
This:
1. calls getGraphData to read the driverGraph.
2. calls getStartAndEnd to ask the user to pick start and goal nodes.
3. calls initSearcherGraph to build the first bit of the searcherGraph.
4. calls Searcher.initSearch to initialize the Searcher.
5. runs a loop in which:
   A. By calling getNextNode, the next node to be opened is identified
      (call it Noe).
   B. By calling insertKids, data about kids of Noe is inserted in Noe,
      and the kid nodes of Noe are built in the searcherGraph if they do
      not already exist there.
   C. CycleSearch is called.
   The loop runs until either of the following occurs:
   A. It is determined that there is no path to the goal.
   B. The cheapest path to the goal has been found.
6. prints the path from start to goal and its cost, if a path has been
   found, and prints the number of cycles (calls to cycleSearch) used
   during the search.
7. closes the Searcher.

The caller of cycleSearch in a system using this version of aStar will
have to decide what criteria are appropriate for deciding when to stop
cycling. In item 5 above, this driver seeks to find a path to the
goal regardless of how long it takes.

Note that if the goal cannot be reached and there is no bound on the
number of cycles, this will search the entire connected portion of the
graph containing the start. The aStar algorithm has no advantage over
the Dijkstra algorithm in this case; in fact Dijkstra would be better
since it has lower overhead (does not need to calculate the
heuristic).

*/

int main(       /* ARGUMENTS                                     */
 int argc,      /*!< one more than number of arguments           */
 char * argv[]) /*!< array of executable name and argument names */
{	
  int start;
  int goal;
  AStarHeapSearcher theSearcher;
  int opens;
  int status[1];
  double cost[1];
  int path[A_STAR_PATH_SIZE];
  int length;
  int cycles;
  int n;
  
  for (n = 0; n < A_STAR_MAP_SIZE; n++)
    {
      driveMap[n] = -1;
      searchMap[n] = -1;
    }
  getGraphData(argc, argv, &driverGraph);
  cycles = 0;
  getStartAndEnd(&start, &goal, driverGraph.size);
  initSearcherGraph(start, goal);
  if (theSearcher.initSearch(&searcherGraph, 1, 0, &opens, status, cost))
    return 1; // initialization error occurred
  while (opens && (status[0] != 2))
    {
      n = theSearcher.getNextNode();
      insertKids(n, searcherGraph.size);
      theSearcher.cycleSearch();
      cycles++;
    }
  if (status[0] == 2)
    {
      length = theSearcher.reportSearch(A_STAR_PATH_SIZE, path);
      printf("Path from start to goal:");
      printAnswer(length, path, cost[0]);
    }
  else
    printf("No path found from start to goal\n");
  printf("%d cycles\n", cycles);
  theSearcher.closeSearch();
  return 0;
}

/***********************************************************************/

//! printAnswer prints a search path and its cost
/*
Returned Value: none

Called By:  main

This prints a search path and its cost.

*/

static void printAnswer( /* ARGUMENTS                                */
 int length,             /*!< length of the path                     */
 int * path,             /*!< array of index numbers giving the path */
 double cost)            /*!< cost of the path                       */
{
  int n;
  
  for (n = 0; n < length; n++)
    {
      printf(" %d", searchMap[path[n]]);
    }
  printf("\n");
  printf("With cost: %.4f\n\n", cost);
}

/***********************************************************************/

//! readGraph reads the named file and builds an AStarHeapGraph
/*
Returned Value: none

Called by:
  getGraphData

This reads the file named fileName into aGraph. The size and capacity
of the graph are set to the size given on the first line of the file.

The input is expected to be a file representing a directed graph.
The input does not have to be acyclic.
The expected format of an input file is:

1. one line with the number of nodes
2. one line for each node of the form:
      index  numberOfChildren  child  child  child ... 
   where a child is of the form:
      (childIndex, connectionCost)
   connectionCost is read as a double.
   If there are more lines of nodes than specified, the extras will
     be ignored.
   If a node line has more children than specified, an error will occur.
   A node should not be its own child.

EXAMPLE file

3
0       1       (1,12)
1       2       (2,19.2) (0,14.5)
2       2       (0,20) (1,15)

In the format string for the fscanf in the last "for" loop below, the
space before the left parenthesis is needed. This seems contrary to
what C manuals say: leading white space is skipped automatically.
Perhaps leading white space is skipped only before numbers.

*/

static void readGraph(     /* ARGUMENTS                  */
 char * fileName,          /*!< name of file to read     */
 AStarHeapGraph * aGraph) /*!< graph to build from file */
{ 
  int i;
  int j;
  AStarHeapNode * nodes;
  int index;
  int kidSize;
  FILE * inport;

  inport = fopen(fileName, "r");
  if (inport == 0)
    {
      fprintf(stderr, "cannot open file for reading\n");
      exit(1);
    }
  if (fscanf(inport, "%d \n", &(aGraph->size)) != 1)
    {
      fprintf(stderr, "cannot read graph size\n");
      exit(1);
    }
  aGraph->capacity = aGraph->size;
  nodes = new AStarHeapNode[aGraph->size];
  aGraph->nodes = nodes;
  for (i = 0; i < aGraph->size; i++)
    {
      if (fscanf(inport, "%d", &index) != 1)
	{
	  fprintf(stderr, "cannot read index\n");
	  exit(1);
	}
      if (i != index)
	{
	  fprintf(stderr, "bad index\n");
	  exit(1);
	}
      nodes[i].moreInfo = new NodeInfo(i);
      if (fscanf(inport, "%d", &kidSize) != 1)
	{
	  fprintf(stderr, "cannot read kids size\n");
	  exit(1);
	}
      nodes[i].kidsSize = kidSize;
      nodes[i].kidsIndexes = new int[kidSize];
      nodes[i].kidsCosts = new double[kidSize];
      for (j = 0; j < kidSize; j++)
	{
	  if (fscanf(inport, " ( %d , %lf )", &(nodes[i].kidsIndexes[j]),
		     &(nodes[i].kidsCosts[j])) != 2)
	    {
	      fprintf(stderr, "cannot read kids data\n");
	      exit(1);
	    }
	}
      fscanf(inport, "\n");
    }
}

/***********************************************************************/

