/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file exampleAStarHeapNA.cc
    \brief provides a driver for testing AStarHeapN search

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Tom Kramer

See the documentation of aStarHeapN.hh for an overview of this search routine.

This is an aStar search routine driver for testing aStar search with
incremental graph building. The driver reads the entire graph from a
file picked by the user (the static graph), but drip feeds it to the
Searcher, which builds its graph while searching.  This routine allows
searching simultaneously for several other goals in addition to the
finish (the target). The heuristic is the minimum cost of the kids a
node, which is weak but admissible.

The indexes of the nodes in the Searcher (since they are assigned on
the fly according to how the search goes) will differ from the indexes
of the nodes in the test file read by the driver.  To implement
insertKids and to report the path correctly, a map (driveMap) from the
indexes of driver's graph to the indexes of the Searcher's graph is
needed. A map the other way (searchMap) is also needed. The maps are
simply arrays of integers declared here as global variables. If the
index of a node in the driver's nodes is n, then the index of the
corresponding node in the Searcher's nodes is driveMap[n]. The maps
have 200000 entries, so graphs with more nodes than that should not be
used.

The routines here use stdio for reading and writing.

*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "aStarHeapN.hh"

#define A_STAR_MAP_SIZE 200000
#define A_STAR_PATH_SIZE 1000
#define MAX_GOALS 100
#define MAX_CYCLES 5000

AStarHeapNGraph driverGraph;
AStarHeapNGraph searcherGraph;
int driveMap[A_STAR_MAP_SIZE];
int searchMap[A_STAR_MAP_SIZE];

/***********************************************************************/

static void getGoals(int start, int finish, int * numGoals,
		     int * goals, int size);
static void getGraphData(int argc, char * argv[], AStarHeapNGraph * theGraph);
static double getMinArcCost(int driverN);
static void getStartAndEnd(int * start, int * finish, int size);
static void initSearcherGraph(int start, int numGoals, int * goals);
static void insertKids(int searcherDad, int startIndex);
int main(int argc, char * argv[]);
static void printAnswer(int length, int * path, double cost);
static void readGraph(char * fileName, AStarHeapNGraph * aGraph);

/***********************************************************************/

//! insertKids adds kids to the dad node and creates new kids
/*
Returned Value: none 

Called By: main
  
The startIndex argument is used to set the index attribute of the
nodeInfos and to update the driveMap and searchMap.

*/

static void insertKids( /* ARGUMENTS                                        */
 int searcherDad,  /*!< index in Searcher nodes of node to insert kids for  */
 int startIndex)   /*!< starting index in Searcher nodes for new added kids */
{
  AStarHeapNNode * driverDad; 
  int numberAdded;             // total number of kids added
  int numberNew;               // number of new added kids
  int numberOld;               // number of old added kids
  NodeInfo ** newInfos;        // NodeInfos for new kids added
  double * newArcCosts;        // costs of new kids
  double * newGoalCosts;       // goalCosts of new kids
  int * oldAdded;              // indexes in nodes of old kids added
  double * oldArcCosts;        // costs of old kids added
  int n;                       // index for kids of driverDad
  int j;                       // index for new kids of searcherDad node
  int k;                       // index for old kids of searcherDad node
  int driverN;                 // index in driver nodes of a kid

  numberNew = 0;
  numberOld = 0;
  j = 0;
  k = 0;
  driverDad = (driverGraph.nodes + searchMap[searcherDad]);
  numberAdded = driverDad->kidsSize;
  for (n = 0; n < numberAdded; n++)
    {
      driverN = driverDad->kidsIndexes[n];
      if (driveMap[driverN] >= searcherGraph.size)
	fprintf(stderr, "bug in maintaining driveMap\n");
      else if (driveMap[driverN] >= 0)
	numberOld++;
      else
	numberNew++;
    }
  if (numberNew > 0)
    {
      newInfos = new NodeInfo * [numberNew];
      newArcCosts = new double[numberNew];
      newGoalCosts = new double[numberNew];
    }
  else
    {
      newInfos = 0;
      newArcCosts = 0;
      newGoalCosts = 0;
    }
  if (numberOld > 0)
    {
      oldAdded = new int[numberOld];
      oldArcCosts = new double[numberOld];
    }
  else
    {
      oldAdded = 0;
      oldArcCosts = 0;
    }
  for (n = 0; n < numberAdded; n++)
    {
      driverN = driverDad->kidsIndexes[n];
      if (driveMap[driverN] >= 0)
	{
	  oldAdded[k] = driveMap[driverN];
	  oldArcCosts[k] = driverDad->kidsCosts[n];
	  k++;
	}
      else
	{
	  newInfos[j] = new NodeInfo(startIndex + j);
	  newArcCosts[j] = driverDad->kidsCosts[n];
	  //newGoalCosts[j] = 0.0; // will act like Dijkstra with 0
	  newGoalCosts[j] = getMinArcCost(driverN);
	  driveMap[driverN] = (startIndex + j);
	  searchMap[startIndex + j] = driverN;
	  j++;
	}
    }
  searcherGraph.addKids(searcherDad, numberAdded, numberNew, newInfos,
			newArcCosts, newGoalCosts, oldAdded, oldArcCosts);
  if (numberNew > 0)
    {
      delete [] newInfos;
      delete [] newArcCosts;
      delete [] newGoalCosts;
    }
  if (numberOld > 0)
    {
      delete [] oldAdded;
      delete [] oldArcCosts;
    }
}

/***********************************************************************/

/* getGoals

This records the finish as goals[0]. Then, it repeatedly asks the user
to enter the index of a goal or to enter s to search. It records the
goals given by the user and it sets numGoals. numGoals includes the
finish, so that, for example, if the user gives 2 goals, numGoals will
be 3.

This prevents the user from entering the same goal twice, or putting a
goal at the finish, or putting a goal at the start, or setting a goal
out of range.

*/

static void getGoals( /* ARGUMENTS                        */
 int start,           /*!< index in nodes of start node   */
 int finish,          /*!< index in nodes of finish node  */
 int * numGoals,      /*!< number of goal nodes, set here */
 int * goals,         /*!< array of goals, filled here    */
 int size)            /*!< number of nodes in the graph   */
{
  int n;

  goals[0] = finish;
  for ( *numGoals = 1; *numGoals < MAX_GOALS; (*numGoals)++ )
    {
      printf("enter goal index, or s to search => ");
      fflush(stdout);
      if (scanf("%d", (goals+*numGoals)) != 1)
	{
	  fflush(stdin);
	  break;
	}
      fflush(stdin);
      if ((goals[*numGoals] < 0) || (goals[*numGoals] > size))
	{
	  fprintf(stderr, "goal out of range\n");
	  (*numGoals)--;
	}
      if (goals[*numGoals] == start)
	{
	  fprintf(stderr, "start must not be goal\n");
	  (*numGoals)--;
	}
      for (n = 0; n < *numGoals; n++)
	{
	  if (goals[n] == goals[*numGoals])
	    {
	      fprintf(stderr, "cannot use same goal twice\n");
	      (*numGoals)--;
	    }
	}
    }
}

/***********************************************************************/

//! getGraphData reads the graph from a file and builds an AStarHeapNGraph
/*
Returned Value: none

Called By:  main

This reads the graph whose name is argv[1].

If the number of command arguments is not 1, this prints an error
message and exits.

*/

static void getGraphData(    /* ARGUMENTS                          */
 int argc,                   /*!< number of command line arguments */
 char * argv[],              /*!< array of command line arguments  */
 AStarHeapNGraph * theGraph) /*!< graph to build, built here       */
{
  if (argc == 2)
    {
      readGraph(argv[1], theGraph);
    }
  else
    {
      fprintf(stderr, "Usage: %s <fileName>\n", argv[0]);
      exit(1);
    }
}

/***********************************************************************/

//! getMinArcCost finds the minimum cost of the kids of a node
/*
Returned Value: double
  This returns the minimum cost of the kids of nodes[driverN] in
  the driverGraph. This may be used as a heuristic in aStar search,
  since it is clearly never greater than the cost to the goal.

Called By: AStarHeapNGraph::getMoreKids

This assumes there is at least one kid.

*/

static double getMinArcCost( /* ARGUMENTS                                 */
 int driverN)         /*!< index in nodes of driverGraph to find cost for */
{
  AStarHeapNNode * aNode;
  int m;
  double result;

  aNode = (driverGraph.nodes + driverN);
  result = 1000000.0;
  for (m = 0; m < aNode->kidsSize; m++)
    {
      if (aNode->kidsCosts[m] < result)
	result = aNode->kidsCosts[m];
    }
  return result;
}

/***********************************************************************/

//! getStartAndEnd gets the indexes of the start and finish nodes from the user
/*
Returned Value: none

Called by: main

This asks the user to give the indexes of the start and finish nodes
and sets those indexes. It is checked that the values are in range and
that start does not equal finish.

The fflush(stdin) is needed to handle entry of bad data. Without it,
the loop does not wait for the user to enter something after bad
data has been entered; it just keeps cycling.

*/

static void getStartAndEnd( /* ARGUMENTS                        */
 int * start,               /*!< index of start node, set here  */
 int * finish,              /*!< index of finsih node, set here */
 int size)                  /*!< number of nodes in graph       */
{
  *start = -1;
  *finish = -1;
  while ((*start < 0)  || (*start >= size)  ||
	 (*finish < 0) || (*finish >= size) ||
	 (*start == *finish))
    {
      printf("Enter start and finish ==> ");
      fflush(stdout);
      scanf("%d %d", start, finish);
      fflush(stdin);
      if ((*start < 0)  || (*start >= size) ||
	  (*finish < 0) || (*finish >= size))
	fprintf(stderr, "start and/or finish out of range\n");
      if (*start == *finish)
	fprintf(stderr, "start must differ from finish\n");
    }
}

/***********************************************************************/

//! initSearcherGraph puts the start node & goal nodes into the searcher graph
/*
Returned Value: none

Called By:
  main

This puts (numGoals + 1) nodes in the searcherGraph and fills them.
The first numGoals of these are equivalent to the goals in the
driverGraph, in the same order. The first of those, (index 0) is
equivalent to the finish node from the driverGraph. The last node put
in the searcherGraph (index numGoals) is equivalent to the start node
from the driverGraph. The initialization of this node is done in
initSearch, not here (except that its goal cost is set to zero here
-- the value zero is irrelevant since it is not used).

It also updates the driveMap and the searchMap.

This uses the global variables searcherGraph, driveMap, and searchMap.

*/

static void initSearcherGraph( /* ARGUMENTS                               */
 int start,                    /*!< index in driver's graph of start node */
 int numGoals,                 /*!< size of goals array                   */
 int * driverGoals)            /*!< array of goals in the driver's graph  */
{
  int n;

  searcherGraph.nodes = new AStarHeapNNode[numGoals + 1];
  searcherGraph.capacity = (numGoals + 1);
  searcherGraph.size = (numGoals + 1);
  for (n = 0; n < numGoals; n++)
    {
      searcherGraph.nodes[n].moreInfo = new NodeInfo(n);
      searcherGraph.nodes[n].goalCost = 0;
      searcherGraph.nodes[n].kidsSize = 0;
      searchMap[n] = driverGoals[n];
      driveMap[driverGoals[n]] = n;
    }
  searcherGraph.nodes[numGoals].goalCost = 0;
  searchMap[numGoals] = start;
  driveMap[start] = numGoals;
}

/***********************************************************************/

//! main runs the driver
/*
This:
1. calls getGraphData to read the driverGraph.
2. calls getStartAndEnd to ask the user to pick start and finish nodes.
3. calls getGoals to ask the user to pick additional goal nodes.
4. calls initSearcherGraph to build the first bit of the searcherGraph.
5. calls Searcher.initSearch to initialize the Searcher.
6. runs a loop in which:
   A. By calling getNextNode, the next node to be opened is identified
      (call it Noe).
   B. By calling insertKids, data about kids of Noe is inserted in Noe,
      and the kid nodes of Noe are built in the searcherGraph if they do
      not already exist there.
   C. CycleSearch is called.
   The loop runs until any one of the following occurs:
   A. It is determined that there is no path to the finish.
   B. The cheapest path to the finish has been found and MAX_CYCLES have run.
   C. The cheapest paths to the finish and all the goals have been found.
7. prints the path from start to finish and its cost, if a path has been
   found, and prints the number of cycles (calls to cycleSearch) used
   during the search.
8. prints the path and its cost for each of the goals to which a path
   was found, and prints whether the cost is known to be optimal.
9. closes the Searcher.

The caller of cycleSearch in a system using this version of aStar will
have to decide what criteria are appropriate for deciding when to stop
cycling. In item 6 above, this driver seeks to find a path to the
finish regardless of how long it takes, but not to take too much time
if not all the goals are reached.

Note that if any of the goals cannot be reached and there is no bound
on the number of cycles, this will search the entire connected portion
of the graph containing the start. The aStar algorithm has no advantage
over the Dijkstra algorithm in this case; in fact Dijkstra would be
better since it has lower overhead (does not need to calculate the
heuristic).

*/

int main(       /* ARGUMENTS                                     */
 int argc,      /*!< one more than number of arguments           */
 char * argv[]) /*!< array of executable name and argument names */
{	
  int start;
  int finish;
  AStarHeapNSearcher theSearcher;
  int numGoals;
  int driverGoals[MAX_GOALS];
  int searcherGoals[MAX_GOALS];
  int opens;
  int statuses[MAX_GOALS];
  double costs[MAX_GOALS];
  int goalsFound;
  int optimaFound;
  int path[A_STAR_PATH_SIZE];
  int length;
  int cycles;
  int n;
  
  for (n = 0; n < A_STAR_MAP_SIZE; n++)
    {
      driveMap[n] = -1;
      searchMap[n] = -1;
    }
  getGraphData(argc, argv, &driverGraph);
  cycles = 0;
  getStartAndEnd(&start, &finish, driverGraph.size);
  getGoals(start, finish, &numGoals, driverGoals, driverGraph.size);
  initSearcherGraph(start, numGoals, driverGoals);
  for (n = 0; n < numGoals; n++)
    searcherGoals[n] = n;
  if (theSearcher.initSearch(&searcherGraph, numGoals, numGoals,
			     searcherGoals, &opens, statuses, costs,
			     &goalsFound, &optimaFound))
    return 1; // initialization error occurred
  while (opens &&
	 ((statuses[0] != 2) || (cycles < MAX_CYCLES)) &&
	 (optimaFound < numGoals))
    {
      n = theSearcher.getNextNode();
      insertKids(n, searcherGraph.size);
      theSearcher.cycleSearch();
      cycles++;
    }
  if (statuses[0] == 2)
    {
      length = theSearcher.reportSearch(0, A_STAR_PATH_SIZE, path);
      printf("Path from start to finish:");
      printAnswer(length, path, costs[0]);
    }
  else
    printf("No path found from start to finish\n");
  printf("%d cycles\n", cycles);
  for (n = 1; n < numGoals; n++)
    {
      if (statuses[n] > 0)
	{
	  length = theSearcher.reportSearch(searcherGoals[n],
					    A_STAR_PATH_SIZE, path);
	  printf("%s", ((statuses[n] == 2) ? "Optimal path " : "Path "));
	  printf("from start to node %d:", driverGoals[n]);
	  printAnswer(length, path, costs[n]);
	}
    }
  theSearcher.closeSearch();
  return 0;
}

/***********************************************************************/

//! printAnswer prints a search path and its cost
/*
Returned Value: none

Called By:  main

This prints a search path and its cost.

*/

static void printAnswer( /* ARGUMENTS                                */
 int length,             /*!< length of the path                     */
 int * path,             /*!< array of index numbers giving the path */
 double cost)            /*!< cost of the path                       */
{
  int n;
  
  for (n = 0; n < length; n++)
    {
      printf(" %d", searchMap[path[n]]);
    }
  printf("\n");
  printf("With cost: %.4f\n\n", cost);
}

/***********************************************************************/

//! readGraph reads the named file and builds an AStarHeapNGraph
/*
Returned Value: none

Called by:
  getGraphData

This reads the file named fileName into aGraph. The size and capacity
of the graph are set to the size given on the first line of the file.

The input is expected to be a file representing a directed graph.
The input does not have to be acyclic.
The expected format of an input file is:

1. one line with the number of nodes
2. one line for each node of the form:
      index  numberOfChildren  child  child  child ... 
   where a child is of the form:
      (childIndex, connectionCost)
   connectionCost is read as a double.
   If there are more lines of nodes than specified, the extras will
     be ignored.
   If a node line has more children than specified, an error will occur.
   A node should not be its own child.

EXAMPLE file

3
0       1       (1,12)
1       2       (2,19.2) (0,14.5)
2       2       (0,20) (1,15)

In the format string for the fscanf in the last "for" loop below, the
space before the left parenthesis is needed. This seems contrary to
what C manuals say: leading white space is skipped automatically.
Perhaps leading white space is skipped only before numbers.

*/

static void readGraph(     /* ARGUMENTS                  */
 char * fileName,          /*!< name of file to read     */
 AStarHeapNGraph * aGraph) /*!< graph to build from file */
{ 
  int i;
  int j;
  AStarHeapNNode * nodes;
  int index;
  int kidSize;
  FILE * inport;

  inport = fopen(fileName, "r");
  if (inport == 0)
    {
      fprintf(stderr, "cannot open file for reading\n");
      exit(1);
    }
  if (fscanf(inport, "%d \n", &(aGraph->size)) != 1)
    {
      fprintf(stderr, "cannot read graph size\n");
      exit(1);
    }
  aGraph->capacity = aGraph->size;
  nodes = new AStarHeapNNode[aGraph->size];
  aGraph->nodes = nodes;
  for (i = 0; i < aGraph->size; i++)
    {
      if (fscanf(inport, "%d", &index) != 1)
	{
	  fprintf(stderr, "cannot read index\n");
	  exit(1);
	}
      if (i != index)
	{
	  fprintf(stderr, "bad index\n");
	  exit(1);
	}
      nodes[i].moreInfo = new NodeInfo(i);
      if (fscanf(inport, "%d", &kidSize) != 1)
	{
	  fprintf(stderr, "cannot read kids size\n");
	  exit(1);
	}
      nodes[i].kidsSize = kidSize;
      nodes[i].kidsIndexes = new int[kidSize];
      nodes[i].kidsCosts = new double[kidSize];
      for (j = 0; j < kidSize; j++)
	{
	  if (fscanf(inport, " ( %d , %lf )", &(nodes[i].kidsIndexes[j]),
		     &(nodes[i].kidsCosts[j])) != 2)
	    {
	      fprintf(stderr, "cannot read kids data\n");
	      exit(1);
	    }
	}
      fscanf(inport, "\n");
    }
}

/***********************************************************************/

