/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   exampleScrollMap.cc
  \brief  Contains an example of using ScrollingMap
*/

#include "exampleScrollMap.hh"
#include "../common/scrollingMap.cc"

#include <stdio.h>

/***********************************************************************/

/*

This file contains an example of how to use the ScrollingMap class.
See scrollingMap.hh for more documentation. Applications should
use the #define line and the first three #include lines shown above,
in the order shown, with exampleCell and exampleCell.hh changed
to the name of a cell class and the name of a header file defining
the cell class.

This example:
1. Builds a 33 by 33 scrolling map.
2. Prints a text representation of the map showing the cells passed
   through by a line from the start point to the end point given by
   the command line arguments.
3. Prints the ids of the cells passed through and how many there are.
4. Scrolls the map, so that the center is moved from (0,0) to (12,9).
5. Repeats steps 2 and 3.

The start point must be on the map. The end point may be off the map.

Note that the first cell on a line is obtained by a call to initLine,
while the rest of the cells on the line are obtained by calls to
getNextCell. When getNextCell returns a NULL pointer, the end of
the line has been reached.

See the documentation in exampleCell.hh, also.

*/

/***********************************************************************/

/*! printMapWithLine prints a textual graphical representation of the
    map and the cells on it through which a line passes. */

void printMapWithLine( /*   ARGUMENTS                           */
 ScrollingMap * map,   /*!< the map to use                      */
 double startX,        /*!< X coordinate of start point of line */
 double startY,        /*!< Y coordinate of start point of line */
 double endX,          /*!< X coordinate of end point of line   */
 double endY)          /*!< Y coordinate of end point of line   */
{
  char canvas[33][33];
  int row;
  int col;
  exampleCell * aCell;

  for (row = 0; row < 33; row++)
    {
      for (col = 0; col < 33; col++)
	{
	  canvas[row][col] = '.';
	}
    }
  aCell = map->initLine(startX, startY, endX, endY);
  if (!aCell)
    {
      printf("Start point is off map\n");
      return;
    }
  row = map->getXIndexAtCell(aCell);
  col = map->getYIndexAtCell(aCell);
  canvas[row][col] = 'X';
  for (; ;)
    {
      aCell = map->getNextCell();
      if (aCell == 0)
	break;
      row = map->getXIndexAtCell(aCell);
      col = map->getYIndexAtCell(aCell);
      canvas[row][col] = 'X';
    }
  printf("\n");
  for (row = 32; row > -1; row--)
    {
      for (col = 0; col < 33; col++)
	{
	  printf("%c ", canvas[row][col]);
	}
      printf("\n");
    }
  printf("\n");
}

/***********************************************************************/

/*! printLineData prints the ids of the cells through which a line passes
    and prints how many of those cells there are. */

void printLineData(  /*   ARGUMENTS                           */
 ScrollingMap * map, /*!< the map to use                      */
 double startX,      /*!< X coordinate of start point of line */
 double startY,      /*!< Y coordinate of start point of line */
 double endX,        /*!< X coordinate of end point of line   */
 double endY)        /*!< Y coordinate of end point of line   */
{
  int n;
  exampleCell * aCell;

  aCell = map->initLine(startX, startY, endX, endY);
  if (!aCell)
    {
      printf("Start point is off map\n");
      return;
    }
  printf("%d ", aCell->getId());
  for (n = 1; ; n++)
    {
      aCell = map->getNextCell();
      if (aCell == 0)
	break;
      printf("%d ", aCell->getId());
    }
  printf("\n%d cells\n", n);
}

/***********************************************************************/

int main(       /*   ARGUMENTS                        */
 int argc,      /*!< number of command line arguments */
 char * argv[]) /*!< text of command line arguments   */
{
  ScrollingMap * map;
  double startX = 10.0;
  double startY = 5.0;
  double endX = 11.0;
  double endY = -10.0;
  int xDiff;
  int yDiff;

  if (argc != 5)
    {
      printf("Usage: %s <startX> <startY> <endX> <endY>\n", argv[0]);
      return 1;
    }
  if (1 != sscanf(argv[1], "%lf", &startX))
    {
      printf("Bad first argument %s\n", argv[1]);
      return 1;
    }
  if (1 != sscanf(argv[2], "%lf", &startY))
    {
      printf("Bad second argument %s\n", argv[2]);
      return 1;
    }
  if (1 != sscanf(argv[3], "%lf", &endX))
    {
      printf("Bad third argument %s\n", argv[3]);
      return 1;
    }
  if (1 != sscanf(argv[4], "%lf", &endY))
    {
      printf("Bad fourth argument %s\n", argv[4]);
      return 1;
    }
  printf("\nDraw and print cell ids for line from "
	 "(%.2lf, %.2lf) to (%.2lf, %.2lf)\n", startX, startY, endX, endY);
  map = new ScrollingMap(16, 1.0);
  printMapWithLine(map, startX, startY, endX, endY);
  printLineData(map, startX, startY, endX, endY);
  map->setCenter(12.0, 9.0, &xDiff, &yDiff);
  printf("\nScroll the center from (0,0) to (12,9) and do the same line\n");
  printMapWithLine(map, startX, startY, endX, endY);
  printLineData(map, startX, startY, endX, endY);

  return 0;
}

/***********************************************************************/

