/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file joysdl.cc

  \brief Joystick interface using Simple DirectMedia Layer

  \code CVS Status:
  $Author: cj_scrapper $
  $Revision: 1.12 $
  $Date: 2007/05/01 18:00:25 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>		// atoi,f
#include <string.h>		// memset
#include "SDL.h"
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "servoMobJA.hh"
#include "servoSP.hh"
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif

/*
  RCS command writers get their own number pair for commands,
  with 0 reserved to indicate nobody, and pairs being an odd
  number and the next even number.
*/
#define WHONUM 2011		// nobody else should have this

// Serial number to servo level
static int servoMobSerial = WHONUM;
static int servoSPSerial = WHONUM;

// alternates between odd and even, then back to odd, e.g., 1 2 1 2 1
//#define NEXTS(s) ((s) = ((s)&1) ? ((s)+1) : ((s)-1))
#define NEXTS(s) (s += 2)
// the above did not work. Messages were being written too fast,
// so they were being overwritten at the lower level and then
// it ignored the message since it had the same serial number.
// So let's just add two each time and stay odd

// NML Channels
static RCS_CMD_CHANNEL *servoMobJACmdNml;
static RCS_STAT_CHANNEL *servoMobJAStatNml;
static RCS_CMD_CHANNEL *servoMobJACfgNml;
static RCS_STAT_CHANNEL *servoMobJASetNml;
static RCS_CMD_CHANNEL *servoSPCmdNml;
static RCS_STAT_CHANNEL *servoSPStatNml;
static ServoMobJACmdSkid skidMsg;
static ServoMobJACmdAckerman ackerMsg;
static ServoMobJACmdSubmarine submarineMsg;
static ServoMobJACmdRotaryWing rotaryWingMsg;

static int mode;
static int vehID;
static int doInit;
static double scale[4];

#define MY_NAME_BASE "joystick"

/*!
  Open all of the NML channels
  
  \param vehID The number to add to all of the channel names
  \return true on success
  \return false on failure
*/
bool openNML(int vehID)
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];

  sprintf(myName, "%s", MY_NAME_BASE);

  if (!configNml) {
    fprintf(stderr, "Please set CONFIG_NML environment variable.\n");
    return (false);
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, vehID);
  servoMobJACmdNml = new RCS_CMD_CHANNEL(servoMobJA_format, chanName,
					 myName, configNml);
  if (NULL == servoMobJACmdNml || !servoMobJACmdNml->valid()) {
    rcs_print_error("%s: Error - Can't open %s buffer\n", myName, chanName);
    return (false);
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, vehID);
  servoMobJAStatNml = new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
					   myName, configNml);
  if (NULL == servoMobJAStatNml || !servoMobJAStatNml->valid()) {
    rcs_print_error("%s: Error - Can't open %s buffer\n", myName, chanName);
    return (false);
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_SET_NAME, vehID);
  servoMobJASetNml = new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
					  myName, configNml);
  if (NULL == servoMobJASetNml || !servoMobJASetNml->valid()) {
    rcs_print_error("%s: Error - Can't open %s buffer\n", myName, chanName);
    return (false);
  }

  sprintf(chanName, "%s%d", SERVO_MOB_JA_CFG_NAME, vehID);
  servoMobJACfgNml = new RCS_CMD_CHANNEL(servoMobJA_format, chanName,
					 myName, configNml);
  if (NULL == servoMobJACfgNml || !servoMobJACfgNml->valid()) {
    rcs_print_error("%s: Error - Can't open %s buffer\n", myName, chanName);
    return (false);
  }

  sprintf(chanName, "%s%d", SERVO_SP_CMD_NAME, vehID);
  servoSPCmdNml = new RCS_CMD_CHANNEL(servoSP_format, chanName,
				      myName, configNml);
  if (NULL == servoSPCmdNml || !servoSPCmdNml->valid()) {
    rcs_print_error("%s: Error - Can't open %s buffer\n", myName, chanName);
    return (false);
  }

  sprintf(chanName, "%s%d", SERVO_SP_STAT_NAME, vehID);
  servoSPStatNml = new RCS_STAT_CHANNEL(servoSP_format, chanName,
					myName, configNml);
  if (NULL == servoSPStatNml || !servoSPStatNml->valid()) {
    rcs_print_error("%s: Error - Can't open %s buffer\n", myName, chanName);
    return (false);
  }

  return (true);
}

/*!
  free all memory and prepare to exit
*/
void closeNML()
{
  skidMsg.wLeft = 0;
  skidMsg.wRight = 0;
  ackerMsg.velocity = 0;
  ackerMsg.crabAngle = 0;
  ackerMsg.steerAngle = 0;
  submarineMsg.propeller = 0;
  submarineMsg.rudder = 0;
  submarineMsg.sternPlane = 0;
  
  switch (mode) {
  case 0:
  case 1:
    skidMsg.serial_number = NEXTS(servoMobSerial);
    servoMobJACmdNml->write(skidMsg);
    break;
  case 2:
    ackerMsg.serial_number = NEXTS(servoMobSerial);
    servoMobJACmdNml->write(ackerMsg);
    break;
  case 3:
    submarineMsg.serial_number = NEXTS(servoMobSerial);
    servoMobJACmdNml->write(submarineMsg);
    break;
  case 4:
    rotaryWingMsg.serial_number = NEXTS(servoMobSerial);
    servoMobJACmdNml->write(rotaryWingMsg);
    break;
  }
  return;

  // deletes seem to cause seg fault?
  if (servoMobJACmdNml != NULL)
    delete servoMobJACmdNml;
  servoMobJACmdNml = NULL;
  if (servoMobJAStatNml != NULL)
    delete servoMobJAStatNml;
  servoMobJAStatNml = NULL;
  if (servoMobJASetNml != NULL)
    delete servoMobJASetNml;
  servoMobJASetNml = NULL;
  if (servoMobJACfgNml != NULL)
    delete servoMobJACfgNml;
  servoMobJACfgNml = NULL;
  if (servoSPCmdNml != NULL)
    delete servoSPCmdNml;
  servoSPCmdNml = NULL;
  if (servoSPStatNml != NULL)
    delete servoSPStatNml;
  servoSPStatNml = NULL;
}

void WatchJoystick(SDL_Joystick * joystick)
{
  int axis;
  double values[4] = {0.0};
  const char *name;
  int done;
  SDL_Event event;

  /* Print info about the joystick we are watching */
  name = SDL_JoystickName(SDL_JoystickIndex(joystick));
  printf("Watching joystick %d: (%s)\n", SDL_JoystickIndex(joystick),
	 name ? name : "Unknown Joystick");
  printf("Joystick has %d axes, %d hats, %d balls, and %d buttons\n",
	 SDL_JoystickNumAxes(joystick), SDL_JoystickNumHats(joystick),
	 SDL_JoystickNumBalls(joystick), SDL_JoystickNumButtons(joystick));

  /* Loop, getting joystick events! */
  done = 0;
  while (!done) {
    while (SDL_PollEvent(&event)) {
      switch (event.type) {
      case SDL_JOYAXISMOTION:
	/*
	  event.jaxis.which is joystick number

	  event.jaxis.axis is
	  0 = left LR, aka X
	  1 = left UD, aka Y
	  2 = right LR, aka Z
	  3 = right UD, aka W

	  event.jaxis.value is -32768 .. 32767
	  For X, negative left
	  For Y, negative up

	  We'll negate [1] and [3] implicitly to make the two up-down
	  (UD) axes natural, so scales should normally be positive
	*/
	axis = event.jaxis.axis;
	if (axis < 0) axis = 0;
	else if (axis > 3) axis = 3;
#define INV_RANGE (1.0 / 32768.0)
	values[axis] = scale[axis] * ((double) event.jaxis.value) * INV_RANGE;
	if (doInit) {
	  switch (mode) {
	  case 0:
	  case 1:
	    skidMsg.wLeft = -values[1];	// Y = left UD
	    skidMsg.wRight = -values[3]; // W = right UD
	    skidMsg.serial_number = NEXTS(servoMobSerial);
	    servoMobJACmdNml->write(skidMsg);
	    /*
	    printf( "Wrote serial #: %d with %f %f\n",
		    skidMsg.serial_number,
		    skidMsg.wLeft,
		    skidMsg.wRight );
	    */
	    break;
	  case 2:
	    ackerMsg.steerAngle = values[2]; // Z = right LR
	    ackerMsg.velocity = -values[1]; // W = left UD
	    ackerMsg.serial_number = NEXTS(servoMobSerial);
	    servoMobJACmdNml->write(ackerMsg);
	    break;
	  case 3:
	    submarineMsg.rudder = values[2]; // right LR
	    submarineMsg.sternPlane = -values[3]; // right UD
	    submarineMsg.propeller = -values[1]; // left UD
	    submarineMsg.serial_number = NEXTS(servoMobSerial);
	    servoMobJACmdNml->write(submarineMsg);
	    break;
	  case 4:
	    //rotaryWingMsg.rot.r = values[2];
	    //rotaryWingMsg.rot.y = -values[3];
	    //rotaryWingMsg.rot.p = values[0];
	    //rotaryWingMsg.propeller = -values[1];
	    //servoMobJACmdNml->write(rotaryWingMsg);
	    break;
	  }
	} else {
	  printf("%f\t%f\t%f\t%f\n", 
		 values[0], values[1], values[2], values[3]);
	}
	break;

      case SDL_JOYHATMOTION:
	printf("Joystick %d hat %d value:", event.jhat.which, event.jhat.hat);
	if (event.jhat.value == SDL_HAT_CENTERED)
	  printf(" centered");
	if (event.jhat.value & SDL_HAT_UP)
	  printf(" up");
	if (event.jhat.value & SDL_HAT_RIGHT)
	  printf(" right");
	if (event.jhat.value & SDL_HAT_DOWN)
	  printf(" down");
	if (event.jhat.value & SDL_HAT_LEFT)
	  printf(" left");
	printf("\n");
	break;
      case SDL_JOYBALLMOTION:
	printf("Joystick %d ball %d delta: (%d,%d)\n",
	       event.jball.which,
	       event.jball.ball, event.jball.xrel, event.jball.yrel);
	break;
      case SDL_JOYBUTTONDOWN:
	printf("Joystick %d button %d down\n",
	       event.jbutton.which, event.jbutton.button);
	break;
      case SDL_JOYBUTTONUP:
	printf("Joystick %d button %d up\n",
	       event.jbutton.which, event.jbutton.button);
	break;
      case SDL_KEYDOWN:
	if (event.key.keysym.sym != SDLK_ESCAPE) {
	  break;
	}
	/* Fall through to signal quit */
      case SDL_QUIT:
	done = 1;
	break;
      default:
	break;
      }
    }
  }
}

int main(int argc, char *argv[])
{
  int joystickNum;
  double end;
  int gotIt;
  ServoMobJAStat * servoMobJAStat;
  ServoSPStat * servoSPStat;
  ServoMobJACmdInit servoMobInit;
  ServoSPCmdInit servoSPInit;
  ServoSPCmdGo servoSPGo;
  char c;
  const char *name;
  int i;
  SDL_Joystick *joystick;

  joystickNum = 0;

  mode = 0;
  vehID = 0;
  doInit = 1;

  skidMsg.wLeft = 0;
  skidMsg.wRight = 0;
  ackerMsg.velocity = 0;
  ackerMsg.steerAngle = 0;
  ackerMsg.crabAngle = 0;
  submarineMsg.propeller = 0;
  submarineMsg.rudder = 0;
  submarineMsg.sternPlane = 0;

  scale[0] = scale[1] = scale[2] = scale[3] = 1.0;

  while ((c = getopt(argc, argv, "j:d:m:x:y:z:w:n")) != EOF) {
    switch (c) {
    case 'j':
      // set joystick number
      joystickNum = atoi(optarg);
      break;
    case 'd':
      // set vehicle id
      vehID = atoi(optarg);
      break;
    case 'm':
      // set operational mode
      mode = atoi(optarg);
      break;
    case 'n':
      // skip initialization
      doInit = 0;
      break;
    case 'x':
      // set scale[0]
      scale[0] = atof(optarg);
      break;
    case 'y':
      // set scale[0]
      scale[1] = atof(optarg);
      break;
    case 'z':
      // set scale[0]
      scale[2] = atof(optarg);
      break;
    case 'w':
      // set scale[0]
      scale[3] = atof(optarg);
      break;
    }
  }

  if (doInit) {

    if (!openNML(vehID)) {
      fprintf(stderr, "Error opening nml buffers\n");
      return 1;
    }
    servoMobJAStat = (ServoMobJAStat *) servoMobJAStatNml->get_address();
    servoSPStat = (ServoSPStat *) servoSPStatNml->get_address();

    // send init command to servo mob
    servoMobInit.serial_number = NEXTS(servoMobSerial);
    servoMobJACmdNml->write(servoMobInit);
    for (end = etime() + 3.0, gotIt = 0; etime() < end; esleep(0.1)) {
      servoMobJAStatNml->read();
      if ( servoMobJAStat->echo_serial_number == servoMobSerial &&
	  servoMobJAStat->status == RCS_DONE) {
	gotIt = 1;
      }
    }
    if (!gotIt) {
      fprintf(stderr, "can't initialize servo mob\n");
      closeNML();
      return 1;
    }

    // send init command to servo sp
    servoSPInit.serial_number = NEXTS(servoSPSerial);
    servoSPCmdNml->write(servoSPInit);
    for (end = etime() + 3.0, gotIt = 0; etime() < end; esleep(0.1)) {
      servoSPStatNml->read();
      if (servoSPStat->echo_serial_number == servoSPSerial &&
	  servoSPStat->status == RCS_DONE) {
	gotIt = 1;
      }
    }
    if (!gotIt) {
      fprintf(stderr, "can't initialize servo sp\n");
      closeNML();
      return 1;
    }
    // send go command to servo sp
    servoSPGo.serial_number = NEXTS(servoSPSerial);
    servoSPCmdNml->write(servoSPGo);
    for (end = etime() + 3.0, gotIt = 0; etime() < end; esleep(0.1)) {
      servoSPStatNml->read();
      if (servoSPStat->echo_serial_number == servoSPSerial &&
	  servoSPStat->status == RCS_EXEC) {
	gotIt = 1;
      }
    }
    if (!gotIt) {
      fprintf(stderr, "can't go-ify servo sp\n");
      closeNML();
      return 1;
    }
  }				// doInit

  /* Initialize SDL (Note: video is required to start event loop) */
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0) {
    fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
    exit(1);
  }

  /* Print information about the joysticks */
  printf("There are %d joysticks attached\n", SDL_NumJoysticks());
  for (i = 0; i < SDL_NumJoysticks(); ++i) {
    name = SDL_JoystickName(i);
    printf("Joystick %d: %s\n", i, name ? name : "Unknown Joystick");
  }

  joystick = SDL_JoystickOpen(joystickNum);
  if (joystick == NULL) {
    printf("Couldn't open joystick %d: %s\n", joystickNum, SDL_GetError());
  } else {
    WatchJoystick(joystick);	// the main loop
    SDL_JoystickClose(joystick);
  }

  SDL_QuitSubSystem(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK);

  return 0;
}

/*
  Modification history:

  $Log: joySDL.cc,v $
  Revision 1.12  2007/05/01 18:00:25  cj_scrapper
  Adding new rotary wing drive command to simware and updated servoMobJA rotary wing drive command

  Revision 1.11  2007/04/30 18:43:27  dr_steveb
  Added AirRobot

  Revision 1.10  2007/03/20 19:17:15  dr_steveb
  Fixed problem with serial number that prevented proper operation.

  Revision 1.9  2007/02/05 22:07:38  dr_steveb
  Fixed reading of status.

  Revision 1.8  2007/02/01 18:20:24  proctor
  Added normal and exponential variates to random number generator; updated
  simpleSimWare to use ground vehicle bot.

*/
