/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file plotArcs.cc

  \brief outputs gnuplot commands to plot a prim echelon arcs file

  This program can be used with plotArcs | gnuplot.
  Arguments: -f filename_in -o filename_out -i increment
  See reference manual for full documentation.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.3 $
  $Date: 2008/04/17 20:33:41 $
  \endcode

  \author Stephen Balakirsky
*/
/*! \ingroup Tools 

*/
//
#define BUFFERLEN 256
using namespace std;

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <ctype.h> /* isspace */
#include <stdlib.h> /* atof */
#include <vector>
#include <moastTypes.hh>
#include <moastGeom.hh> // lineSegment

class MyLine:public LineSegment{
public:
  double annular_tol;
  double speed;
  bool dontTouch;
  int id;
};

class MyArc:public LineSegment{
public:
  Location2D center;
  double theta;
  double annular_tol;
  double speed;
  double startAngle;
  double endAngle;
  double radius;
  bool dontTouch;
  int id;
};

static vector<MyLine>linesRead;
static vector<MyArc>arcsRead;

double solveArc( const double &north, const Location2D &start, 
		 const Location2D &end, const double &radius, const double &theta )
{
  double rootTerm;
  // compute the root term
  rootTerm = sqrt( start.east*start.east + radius*radius -
		   (north-start.north)*(north-start.north) - 
		   start.east*start.east);
  if( theta < 0 )
    rootTerm *= -1;
  return (start.east + rootTerm);
}

void shiftLines( double increment, int offset, int numLines, int numArcs )
{
  MyLine *lineIt;
  MyArc *arcIt;
  LineSegment temp;
  int count;

  // shift arcs
  for( count=numArcs; count<(int)arcsRead.size(); count++ )
    {
      arcIt = &arcsRead[count];
      // compute radius
      arcIt->radius = sqrt( (arcIt->center.north - arcIt->end.north)*(arcIt->center.north - arcIt->end.north) +
			 (arcIt->center.east - arcIt->end.east)*(arcIt->center.east - arcIt->end.east) ) + 
	arcIt->annular_tol * offset;
      if( arcIt->radius < 0 ) arcIt->radius = 0;
      // compute starting location
      arcIt->startAngle = atan2( arcIt->end.north - arcIt->center.north, arcIt->end.east - arcIt->center.east);
      if( arcIt->theta < 0 )
	{
	  arcIt->endAngle = arcIt->startAngle;
	  arcIt->startAngle += arcIt->theta;
	}
      else
	{
	  arcIt->endAngle = arcIt->startAngle + arcIt->theta;
	}
      arcIt->start.north = arcIt->radius * sin(arcIt->startAngle) + arcIt->center.north;
      arcIt->start.east = arcIt->radius * cos(arcIt->startAngle) + arcIt->center.east;
      arcIt->end.north = arcIt->radius * sin(arcIt->endAngle) + arcIt->center.north;
      arcIt->end.east = arcIt->radius * cos(arcIt->endAngle) + arcIt->center.east;
    }

  // shift lines
  for( count=numLines; count<(int)linesRead.size(); count++ )
    {
      lineIt = &linesRead[count];
      if( offset != 0 )
	{
	  temp = lineIt->moveLineParallel( lineIt->annular_tol, offset );
	  lineIt->start = temp.start;
	  lineIt->end = temp.end;
	}
    }
}

void buildLines( FILE *fp, bool dontTouch, int id )
{
  char buffer[BUFFERLEN];
  char *ptr;
  Location2D center, start, end, drawStart, drawEnd;
  LineSegment lastArc;
  bool startSet=false;
  int lineNum;
  MyLine line;
  MyArc arc;

  lineNum = 0;
  while( !feof(fp))
    {
      if (NULL == fgets(buffer, BUFFERLEN, fp))
	break;
      lineNum++;
      ptr = buffer;
      while (isspace(*ptr))
	ptr++;
      if (0 == *ptr)
	continue;			// blank line
      if ('#' == *ptr)
	continue;			// comment
      if (! strncmp(ptr, "abs", 3)) 
	{
	  //	  printf( "# Using abs coordinates\n" );
	  continue;
	}
      else if (! strncmp(ptr, "rel", 3)) 
	{
	  //	  printf( "# Using rel coordinates\n" );
	  continue;
	}
      if (7 == sscanf(ptr, "%lf %lf %lf %lf %lf %lf %lf",
			   &arc.center.north,
			   &arc.center.east,
			   &arc.end.north,
			   &arc.end.east,
			   &arc.theta,
			   &arc.annular_tol,
			   &arc.speed))
	{
	  if( !startSet )
	    {
	      fprintf( stderr, "First file entry must be a line!\n" );
	      exit(1);
	    }
	  arc.start = line.start;
	  arc.dontTouch = dontTouch;
	  arc.id = id;
	  arcsRead.push_back(arc);
	  line.start = arc.end;
	} 
      else if (4 == sscanf(ptr, "%lf %lf %lf %lf",
			   &line.end.north,
			   &line.end.east,
			   &line.annular_tol,
			   &line.speed)) 
	{
	  if( !startSet )
	    {
	      startSet = true;
	      line.start = line.end;
	    }
	  else
	    {
	      line.dontTouch = dontTouch;
	      line.id = id;
	      linesRead.push_back(line);
	      line.start = line.end;
	    }
	} 
      else 
	{
	  fprintf(stderr, "syntax error in input, line %d: %s\n",
		  lineNum, buffer);
	  return;
	}
    }
}

#define MINDIST .2
void trimLines(double increment)
{
  vector<MyLine>::iterator lineIt;
  vector<MyArc>::iterator arcIt;
  vector<MyLine>::iterator lineIt2;
  vector<MyArc>::iterator arcIt2;
  LineSegment lineCheck;
  Location2D intersectPt;
  double newAngle;
  bool foundStart, foundEnd;


  // check arc-line intersection
  for( lineIt=linesRead.begin(); lineIt!=linesRead.end(); lineIt++ )
    {
      for( arcIt=arcsRead.begin(); arcIt!=arcsRead.end(); arcIt++ )
	{
	  lineCheck.start.north = arcIt->radius * sin(newAngle) + arcIt->center.north;
	  lineCheck.start.east = arcIt->radius * cos(newAngle) + arcIt->center.east;
	  //	  printf( "Start: %f End: %f\n", arcIt->startAngle, arcIt->endAngle );
	  for( newAngle = arcIt->startAngle; newAngle<arcIt->endAngle; newAngle+= increment )
	    {
	      lineCheck.end.north = arcIt->radius * sin(newAngle) + arcIt->center.north;
	      lineCheck.end.east = arcIt->radius * cos(newAngle) + arcIt->center.east;
	      if( lineIt->testIntersect(lineCheck) )
		{
		  intersectPt = lineIt->intersect(lineCheck);
		  if( !lineIt->dontTouch )
		    {
		      if( intersectPt.dist(lineIt->start) > intersectPt.dist(lineIt->end) )
			lineIt->end = intersectPt;
		      else
			lineIt->start = intersectPt;
		    }
		  if( !arcIt->dontTouch )
		    {
		      if( newAngle - arcIt->startAngle < arcIt->endAngle - newAngle )
			arcIt->startAngle = newAngle;
		      else
			arcIt->endAngle = newAngle;
		    }
		  //		  break;
		  //		  printf( "Made a change to line <%f %f> <%f %f>\n",
		  //			  lineIt->start.north, lineIt->start.east,
		  //			  lineIt->end.north, lineIt->end.east);
		}
	      lineCheck.start = lineCheck.end;
	    }
	}  
    }

  // check line-line intersection
  for( lineIt=linesRead.begin(); lineIt!=linesRead.end(); lineIt++ )
    {
      for( lineIt2=linesRead.begin(); lineIt2!=linesRead.end(); lineIt2++ )
	{
	  if( lineIt == lineIt2 ) continue;
	  lineCheck.start = lineIt2->start;
	  lineCheck.end = lineIt2->end;
	  if( lineIt->testIntersect(lineCheck) )
	    {
	      intersectPt = lineIt->intersect(lineCheck);
	      if( !lineIt->dontTouch )
		{
		  if( intersectPt.dist(lineIt->start) > intersectPt.dist(lineIt->end) )
		    lineIt->end = intersectPt;
		  else
		    lineIt->start = intersectPt;
		}
	      if( !lineIt2->dontTouch )
		{
		  if( intersectPt.dist(lineIt2->start) > intersectPt.dist(lineIt2->end) )
		    lineIt2->end = intersectPt;
		  else
		    lineIt2->start = intersectPt;
		}
	    }
	}  
    }


  return;
  // now see if we have any hanging lines
  for( lineIt=linesRead.begin(); lineIt!=linesRead.end(); lineIt++ )
    {
      foundStart = false;
      foundEnd = false;
      // check against arcs
      for( arcIt=arcsRead.begin(); arcIt!=arcsRead.end(); arcIt++ )
	{
	  if( lineIt->start.dist(arcIt->start)<MINDIST || lineIt->start.dist(arcIt->end)<MINDIST )
	    foundStart = true;
	  if( lineIt->end.dist(arcIt->start)<MINDIST || lineIt->end.dist(arcIt->end)<MINDIST )
	    foundEnd = true;
	}
      // check against lines
      for( lineIt2=linesRead.begin(); lineIt2!=linesRead.end(); lineIt2++ )
	{
	  if( lineIt2 == lineIt )continue;
	  if( lineIt->start.dist(lineIt2->start)<MINDIST || lineIt->start.dist(lineIt2->end)<MINDIST )
	    foundStart = true;
	  if( lineIt->end.dist(lineIt2->start)<MINDIST || lineIt->end.dist(lineIt2->end)<MINDIST )
	    foundEnd = true;
	}
      if( !foundStart || !foundEnd )
	{
	  lineIt2 = lineIt;
	  lineIt++;
	  linesRead.erase(lineIt2);
	  printf( "Erasing line\n" );
	}
    }
  // how about hanging arcs
  for( arcIt=arcsRead.begin(); arcIt!=arcsRead.end(); arcIt++ )
    {
      foundStart = false;
      foundEnd = false;
      // check against arcs
      for( arcIt2=arcsRead.begin(); arcIt2!=arcsRead.end(); arcIt2++ )
	{
	  if( arcIt2 == arcIt )continue;
	  if( arcIt->start.dist(arcIt2->start)<MINDIST || arcIt->start.dist(arcIt2->end)<MINDIST )
	    foundStart = true;
	  if( arcIt->end.dist(arcIt2->start)<MINDIST || arcIt->end.dist(arcIt2->end)<MINDIST )
	    foundEnd = true;
	}
      // check against lines
      for( lineIt2=linesRead.begin(); lineIt2!=linesRead.end(); lineIt2++ )
	{
	  if( arcIt->start.dist(lineIt2->start)<MINDIST || arcIt->start.dist(lineIt2->end)<MINDIST )
	    foundStart = true;
	  if( arcIt->end.dist(lineIt2->start)<MINDIST || arcIt->end.dist(lineIt2->end)<MINDIST )
	    foundEnd = true;
	}
      if( !foundStart || !foundEnd )
	{
	  printf( "Erasing arc\n" );
	  arcIt2 = arcIt;
	  arcIt++;
	  arcsRead.erase(arcIt2);
	}
    }
}

int plotLine(FILE *fp,   FILE *fpOut, double increment, bool centerOnly)
{
  vector<MyLine>::iterator lineIt;
  vector<MyArc>::iterator arcIt;
  Location2D center, start, end, drawStart, drawEnd, lastPoint;
  LineSegment lastArc;
  double newAngle;
  int retval = 1;
  int numArcs, numLines;
  bool directionPos;
  FILE *fpOut1, *fpOut2, *fpPt = fpOut;

  buildLines(fp, true, 0);
  shiftLines(increment, 0, 0, 0 );
  rewind( fp );
  numLines = linesRead.size();
  numArcs = arcsRead.size();
  //  if( !centerOnly )
  //    {
      // open error output files
      if (NULL == (fpOut1 = fopen("Error1.dat", "w")))
	{
	  printf( "unable to open file %s\n", "Error1.dat" );
	  return 1;
	}

      if (NULL == (fpOut2 = fopen("Error2.dat", "w")))
	{
	  printf( "unable to open file %s\n", "Error2.dat" );
	  return 1;
	}

      // build extra lines
      buildLines(fp, false, 1);
      rewind( fp );
      shiftLines(increment, -1, numLines, numArcs);
      numLines = linesRead.size();
      numArcs = arcsRead.size();
      buildLines(fp, false, 2);
      rewind( fp );
      shiftLines(increment, 1, numLines, numArcs);
      trimLines(increment);
      //    }

  // plot the arcs
  for( arcIt=arcsRead.begin(); arcIt!=arcsRead.end(); arcIt++ )
    {
      lastPoint.north = 0;
      lastPoint.east = 0;
      directionPos = true;
      // compute which way to plot
      if( (arcIt->radius * sin(arcIt->startAngle) + arcIt->center.north) > 
	  (arcIt->radius * sin(arcIt->endAngle) + arcIt->center.north) )
	directionPos = false;

      if( !directionPos )
	{
	  for( newAngle = arcIt->endAngle; newAngle>=arcIt->startAngle; newAngle-= increment )
	    {
	      start.north = arcIt->radius * sin(newAngle) + arcIt->center.north;
	      start.east = arcIt->radius * cos(newAngle) + arcIt->center.east;
	      if( centerOnly )
		{
		  if( arcIt->id == 0 )
		    fpPt = fpOut;
		  else 
		    fpPt = fpOut1;
		  if( fabs(lastPoint.north- start.north)>=.01 || fabs(lastPoint.east != start.east)>=.01 )
		    fprintf( fpPt, "%5.2f %5.2f %5.2f %5.2f\n", start.north, start.east, arcIt->annular_tol, arcIt->speed );
		}
	      else
		if( fabs(lastPoint.north- start.north)>=.01 || fabs(lastPoint.east != start.east)>=.01 )
		  fprintf( fpPt, "%5.2f %5.2f\n", start.north, start.east );
	      lastPoint = start;
	    }
	}
      else
	{
	  for( newAngle = arcIt->startAngle; newAngle<arcIt->endAngle; newAngle+= increment )
	    {
	      start.north = arcIt->radius * sin(newAngle) + arcIt->center.north;
	      start.east = arcIt->radius * cos(newAngle) + arcIt->center.east;
	      if( centerOnly )
		{
		  if( arcIt->id == 0 )
		    fpPt = fpOut;
		  else
		    fpPt = fpOut2;
		  if( fabs(lastPoint.north- start.north)>=.01 || fabs(lastPoint.east != start.east)>=.01 )
		    fprintf( fpPt, "%5.2f %5.2f %5.2f %5.2f\n", start.north, start.east, arcIt->annular_tol, arcIt->speed );
		}
	      else
		  if( fabs(lastPoint.north- start.north)>=.01 || fabs(lastPoint.east != start.east)>=.01 )
		    fprintf( fpPt, "%5.2f %5.2f\n", start.north, start.east );
	      lastPoint = start;
	    }
	}
      fprintf( fpPt, "\n" );
    }
  // plot the lines
  if( !centerOnly )
    {
      for( lineIt=linesRead.begin(); lineIt!=linesRead.end(); lineIt++ )
	{
	  fprintf( fpPt, "%5.2f %5.2f\n", lineIt->start.north, lineIt->start.east );
	  fprintf( fpPt, "%5.2f %5.2f\n\n", lineIt->end.north, lineIt->end.east );
	}
    }
  return retval;
}


int main(int argc, char *argv[])
{
  char *fileName = NULL;
  char *fileOut = "out.dat";
  FILE *fp, *fpOut;
  char c;
  double increment = .01;
  bool centerOnly = false;

  while((c=getopt(argc,argv,"cf:i:o:"))!=EOF)
    {
      switch(c)
	{
	case 'f':
	  fileName = optarg;
	  break;
	case 'i':
	  increment = atof(optarg);
	  break;
	case 'o':
	  fileOut = optarg;
	  break;
	case 'c':
	  centerOnly = true;
	  break;
	}
    }
  if( fileName == NULL )
    {
      printf( "must include file name\n" );
      return 1;
    }
  if (NULL == (fp = fopen(fileName, "r")))
    {
      printf( "unable to open file %s\n", fileName );
      return 1;
    }

  if (NULL == (fpOut = fopen(fileOut, "w")))
    {
      printf( "unable to open file: out.dat \n" );
      return 1;
    }
  linesRead.clear();
  arcsRead.clear();
  plotLine(fp, fpOut, increment, centerOnly );
  fprintf( fpOut, "\n" );
  /*
  rewind( fp );
  linesRead.clear();
  arcsRead.clear();
  plotLine(fp, fpOut, increment, 1);
  fprintf( fpOut, "\n" );
  rewind( fp );
  plotLine(fp, fpOut, increment, -1);
  */
  fclose(fp);
  fclose(fpOut);

}

/*
  Modification history:

  $Log: plotArcs.cc,v $
  Revision 1.3  2008/04/17 20:33:41  dr_steveb
  Tools for the VMAC competition

  Revision 1.2  2008/03/11 17:24:37  dr_steveb
  Fixed some syntax errors

  Revision 1.1  2008/03/11 13:17:26  dr_steveb
  Added code to plot prim commanded arcs in gnuplot.

*/
