/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   cluster.hh
  \brief  provides clusters of points from a world model map

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.4 $
  $Date: 2007/11/08 13:22:30 $
  \endcode

  \author Stephen Balakirsky
  \date   02/12/2007
*/
#ifndef __cluster__
#define __cluster__

#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <math.h>
#include <rcs.hh>
#include <posemath.h>
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "sensorData.hh"
#include "sweep.hh"

int chainHull_2D(PM_CARTESIAN* P, int n, PM_CARTESIAN* H);
class ClusterMember
{
public:
  ClusterMember(){ previous=NULL;};
  bool operator<(const ClusterMember *comp);
  
  ClusterMember *previous;
  float data;
  //  int xLoc, yLoc;
  double xLoc, yLoc;
  int clusterID;
};

class ClusterHead
{
private:
  int maxNumClusters;
  int numClusters;
public:
  ClusterHead();
  ~ClusterHead();
  void checkIntersection( int cluster1, int cluster2 );
  void expandPolygon( int clusterID, double dist );
  int getNumClusters(){return numClusters;};
  bool incNumClusters();
  void print(FILE *filePt=stdout);
  void print(int clusterID, FILE *filePt=stdout);
  //! The first point in each cluster
  ClusterMember **clusterEnds;
};

class Cluster
{
public:
  Cluster(int gapSizeIn);
  ~Cluster();
  int build();
  void checkIntersections();
  void expandPolygons( double dist );
  void joinCluster( int newMember, int currentMember );
  void print();
  void print(FILE *fp);
  bool readFile(char *name, int mode);
  bool readNML( NML *chan );
  void buildHulls();
  // routine for general tests
  void test(int testNo);

private:
  void deleteRaw();
  int buildHull(ClusterMember *start, int overShoot);
  void expandHullData();
  void checkCluster( int toCheck );
  bool findNearestClusterPt(PM_CARTESIAN pointIn, 
			    PM_CARTESIAN &pointReturn,
			    int numPoints, int threshold);
  PM_CARTESIAN findClosestPerp(PM_CARTESIAN &linePtA, 
			       PM_CARTESIAN &linePtB,
			       int distAlongLine, int neighborhood);
  void markPt( PM_CARTESIAN &ptIn, int numPoints, int range );
  PM_CARTESIAN movePoint( PM_CARTESIAN &from, PM_CARTESIAN &end, 
			  double distance);
  PM_CARTESIAN movePointPerp( PM_CARTESIAN &linePtA, PM_CARTESIAN &linePtB, 
			      PM_CARTESIAN &perpPt, double distance);
  /*
  PM_CARTESIAN movePointPerp( Location2D &linePtA, Location2D &linePtB, 
			      Location2D &perpPt, double distance);
  */
  void merge(int clusterKeep, int clusterMerge );
  ClusterHead clusterHead;
  unsigned int xRawSize, yRawSize; // dimensions of data array
  ClusterMember *rawData;
  PM_CARTESIAN *hullDataIn, *hullDataOut, *hullDataTemp;
  int hullDataSize;
  int numHullDataIn;
  int gapSize;
};
#endif

/*
  Modification history:

  $Log: cluster.hh,v $
  Revision 1.4  2007/11/08 13:22:30  dr_steveb
  Not quite ready, but premature checking-in of other files means that these need to go to!

  Revision 1.3  2007/04/25 15:09:04  dr_steveb
  Added traffic simulation files

  Revision 1.2  2007/03/21 19:29:49  dr_steveb
  Fixed bug that caused the world to wrap

  Revision 1.1  2007/02/27 14:51:04  dr_steveb
  New tool for calc. the concave hull

*/
