/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehMobCell.hh
  \brief  scrolling map cell for vehicle planner
  This map is based on the scrolling map cell contained in scrollingMap.hh.
  Currently, the cell contains information on if there is an obstacle,
  if it lies in an obstacle grow region, or if it contains a hazard.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.6 $
  $Date: 2007/05/16 20:57:14 $
  \endcode

  \author Stephen Balakirsky
  \date   09/20/2006
*/
#ifndef __vehMobCell
#define __vehMobCell
#include "scrollingMapCell.hh"
////////////////////////////////
/*! world model default values
    first set represents break points in piecewise linear cost curve
*/
#define VEH_WM_EASY_MAX          59
#define VEH_WM_MODERATE_MAX     119
#define VEH_WM_DIFFICULT_MAX    179
#define VEH_WM_EXTREAM_MAX      239
#define VEH_WM_DANGEROUS_MAX    254
//! set unknown terrain to a value
#define VEH_WM_UNKNOWN 100
#define VEH_WM_UNOBSERVABLE 101

/*!
    next set is cost increment value (maps to slope of piecewise linear cost curve)
*/
#define VEH_WM_EASY_INCREMENT           1
#define VEH_WM_MODERATE_INCREMENT      10
#define VEH_WM_DIFFICULT_INCREMENT    100
#define VEH_WM_EXTREAM_INCREMENT     1000
#define VEH_WM_DANGEROUS_INCREMENT  10000
#define VEH_WM_NOGO_INCREMENT      100000

//! obstacles above this value should be grown
#define VEH_WM_GROW_OBSTACLE_MIN    254
//! what cost lookup index should grown obstacles use?
//#define VEH_WM_GROW_OBSTACLE_INDEX  VEH_WM_MODERATE_MAX
#define VEH_WM_GROW_OBSTACLE_INDEX  VEH_WM_DANGEROUS_MAX-1

enum VehCellMark {
  NO_MARK,
  AM_POINT,
  OBSERVE_POINT,
  KEEP_OUT
};

class VehMobCell : public ScrollingMapCell
{
public:
  //  VehMobCell() : ScrollingMapCell(){init();}
  VehMobCell() : ScrollingMapCell(){}
  ~VehMobCell(){}

  //! obstacleMax is used for isCleared calc. and image dump
  unsigned char obstacleMax;
  //! obstacleMin is used by the planner
  unsigned char obstacleMin;
  unsigned char grow;
  unsigned char hazard;
  VehCellMark mark; // used to mark special cells
  unsigned int lastSeen;
  bool isCleared; // have the cell been cleared of victims?
  int graphID; // the id of the graph node that this cells matches

  //! Method init reinitializes a cell
  void init(){
    obstacleMax=VEH_WM_UNKNOWN; 
    obstacleMin=VEH_WM_UNKNOWN;
    grow=0; 
    hazard=0;
    mark=NO_MARK;
    graphID = -1;
    isCleared = false;}

  /*! Method setCellData sets the data in a cell, using values
      given in the arguments. */
  void setCellData(unsigned char obstacleMaxIn, unsigned char obstacleMinIn,
		   unsigned char growIn, 
		   unsigned char hazardIn, unsigned int lastSeenIn,
		   VehCellMark markIn, bool isClearedIn){
    obstacleMax = obstacleMaxIn;
    obstacleMin = obstacleMinIn;
    grow = growIn;
    hazard = hazardIn;
    lastSeen = lastSeenIn;
    mark = markIn;
    isCleared = isClearedIn;}
};
#endif
/*! CVS Log Information
 * $Log: vehMobCell.hh,v $
 * Revision 1.6  2007/05/16 20:57:14  dr_steveb
 * Fixed graph wrapping!
 *
 * Revision 1.5  2007/02/27 14:46:26  dr_steveb
 * New interface between AM and VEH
 *
 * Revision 1.4  2006/10/18 00:18:21  dr_steveb
 * Added victim channel to vehicle level. It now hunts for victims. Also added CImg library from sourceforge to perform annotation. Modified make files to include this library.
 *
 * Revision 1.3  2006/10/10 13:30:17  dr_steveb
 * New vehicle level exploration based on camera FOV implemented.
 *
 * Revision 1.2  2006/10/03 20:24:25  dr_steveb
 * Exploration command added to vehicle level
 *
 * Revision 1.1  2006/09/30 16:05:30  dr_steveb
 * New world model files
 *
 * 
 * 

 */
