/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehMobPLEngine.cc
  \brief  Provides a planning engine for use by the waypoint planner

  This is the actual planning engine that maintains the world model and graph
  structures. The originial code was developed for the LAGR program.

  TODO: in routine search, need to pay attention to time constraints to end search

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.23 $
  $Date: 2007/12/13 13:18:41 $
  \endcode

  \author Stephen Balakirsky
  \date   05/17/2005
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <strings.h>
#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <time.h>
#include "vehMobPLEngine.hh"
#include "nodeInfo.hh"
#include "amSP.hh"      // AMSP_OBSTACLE, etc.
#include "amMobJA.hh"  // MAX_AM_SEED_NODES
#include "vehMobPL.hh"
//#ifdef USE_SPQR
#include "nmlMap.hh"
#include "nmlPosition.hh"
//#endif

#define DEG(x) (57.295779513082323 * (x))
#define RAD(x)  (0.017453292519943 * (x))

/////////////////////////////////////////////////////////////////////////////
double myDisp( PM_CARTESIAN &pt1, PM_CARTESIAN &pt2 )
{
  return( sqrt((pt1.x-pt2.x)*(pt1.x-pt2.x)+(pt1.y-pt2.y)*(pt1.y-pt2.y)) );
}

/////////////////////////////////////////////////////////////////////////////
/*! 
  constructor
  \param northCells The number of initial cells in the north direction (x).
  \param eastCells  The number of initial cells in the east direction (y).
  \param cellResolution The resolution in meters of each cell.
  \param obsGrowIn      How big in cells should obstacles be grown (0- no grow)?
  \param verbose        Verbose flag.
*/
VehMobPLEngine::VehMobPLEngine( int northCells, int eastCells, float cellResolution, 
			int obsGrowIn, int verbose )
{
  if( northCells != eastCells )
    {
      fprintf( stderr, "vehMobPLEnging: must have square map! Setting to be square\n" );
      if( northCells > eastCells )
	eastCells = northCells;
      else
	northCells = eastCells;
    }

  // initialize values
  numAmSeedNodes = 0;
  amSeedNodes = new PM_CARTESIAN[MAX_AM_SEED_NODES];
  costsOut = 0;
  costLookup = NULL;
  obsGrow = obsGrowIn;
  myVerbose = verbose;
  graphWidth = eastCells;
  graphHeight = northCells;
  discountPathLength = 0;
  pathLength = -1;
  straightDist = cellResolution/2.;
  //  diagDist = sqrtf( 2.*cellResolution*cellResolution)/2.;
  diagDist = (float)(cellResolution * M_SQRT1_2);

  // allocate memory
  pathOut = new PM_CARTESIAN[1];
  cImg = new cimg_library::CImg<unsigned char>(graphHeight,graphWidth,1,3);
  searcherGraph = new DijkHeapNGraph();
  searcherGraph->capacity = 0;
  theSearcher = new DijkHeapNSearcher();
  path = new int[MAX_DIJK_WAYPOINTS];
  costLookup = new double[256];
  vehMapOut = new SensorData2D();
  spqrMapOut = new MapMsg();
  /* moved to init function
  // lets make sure that we agree on the size
  graphHeight /= 2;
  worldMap = new VehMobMap( graphHeight, cellResolution );
  graphHeight = graphHeight *2 + 1;
  graphWidth = graphHeight;
  */
  fprintf( stderr, "Finished initialization\n" );
}

/////////////////////////////////////////////////////////////////////////////
/*!
  Destructor
*/
VehMobPLEngine::~VehMobPLEngine()
{
  if( pathOut != NULL )
    delete pathOut;
  if( cImg != NULL )
    delete cImg;
  delete searcherGraph;
  delete theSearcher;
  if( path != NULL )
    delete path;
  if( costLookup != NULL )
    delete [] costLookup;
  delete worldMap;
  if( discountPathLength > 0 )
    {
      delete [] discountPath;
      discountPathLength = 0;
    }
  if( vehMapOut != NULL )
    delete vehMapOut;
  if( spqrMapOut != NULL )
    delete spqrMapOut;
}

/***********************************************************************/
void VehMobPLEngine::init( PM_POSE &vehLocIn)
{
  int trash1, trash2;

  // lets make sure that we agree on the size
  graphHeight /= 2;
  worldMap = new VehMobMap( graphHeight, straightDist*2 );
  if( worldMap<=0 )
    printf( "******************************\nUnable to allocate memory for world model BAD!!!\n*************************************\n");
  graphHeight = worldMap->getRows();
  graphWidth = worldMap->getCols();
  /*
  graphHeight = graphHeight *2 + 1;
  graphWidth = graphHeight;
  */
  worldMap->setCenter( vehLocIn.tran.x, vehLocIn.tran.y, &trash1, &trash2);
  //  setVehLoc(vehLocIn);
}

/////////////////////////////////////////////////////////////////////////////
/*! run some basic checks on the graph to make sure that it is properly 
    constructed
*/
bool VehMobPLEngine::checkGraph()
{
  int nodeCount, kidCount;
  int graphSize = graphWidth*graphHeight;
  PM_CARTESIAN father, kid;
  int kidIndex;
  bool retValue = true;
  double separation, maxSep;

  maxSep = diagDist *2.1;
  printf( "Checking graph for dist %.3f...\n", maxSep );
  father.z = 0;
  kid.z = 0;
  for( nodeCount=0; nodeCount<graphSize; nodeCount++ )
    {
      for( kidCount=0; kidCount<searcherGraph->nodes[nodeCount].kidsSize; kidCount++ )
	{
	  father.x = worldMap->getXCoordAtId(searcherGraph->nodes[nodeCount].moreInfo->index);
	  father.y = worldMap->getYCoordAtId(searcherGraph->nodes[nodeCount].moreInfo->index);
	  kidIndex = searcherGraph->nodes[nodeCount].kidsIndexes[kidCount];
	  kid.x = worldMap->getXCoordAtId(searcherGraph->nodes[kidIndex].moreInfo->index);
	  kid.y = worldMap->getYCoordAtId(searcherGraph->nodes[kidIndex].moreInfo->index);
	  if( (separation = myDisp(father, kid)) > maxSep )
	    {
	      printf( "<%d:%d>:%.3f\n", nodeCount, kidCount, separation );
	      retValue = false;
	    }
	}
    }
  if( retValue )
    printf( "passed graph check\n");
  else
    printf( "failed graph check\n");
  return retValue;
}

/////////////////////////////////////////////////////////////////////////////
/*!
  Build graph that is graphWidth x graphHeight in size. This graph will be
  8-connected with edge elements having -1 for their edge values that do not
  exist.

  \return true no error
  \return false error
*/
bool VehMobPLEngine::buildGraph()
{
  int graphSize = graphWidth*graphHeight;
  int graphNodeID, childNum;
  int east, north, childCount;
  double childCost;
  bool topBad, bottomBad, leftBad, rightBad;
  float straightCost, diagCost;
  VehMobCell *cellPt;

  straightCost = straightDist * costLookup[VEH_WM_UNKNOWN]*2;
  diagCost = diagDist * costLookup[VEH_WM_UNKNOWN]*2;

  if( searcherGraph->capacity != 0 )
    delete [] searcherGraph->nodes;

  /* need to have an extra node to attach to AM echelons
     ring of pre-planned nodes. This node will be located
     at nodes[graphSize+1] and will be the starting point of the
     search. The cost of traveling from this node to the AM determined
     perimiter will be the costs returned from AM in the seed values
  */
  graphSize++; // see above comment
  searcherGraph->nodes = new DijkHeapNNode[graphSize+1];
  searcherGraph->capacity = graphSize+1;
  searcherGraph->size = graphSize;
  searcherGraph->nodes[graphSize-1].kidsSize = 0; // set no kids

  // double loop for the 2D graph structure
  for( north=0; north<graphHeight; north++ )
    {
      for( east=0; east<graphWidth; east++ )
	{
	  cellPt = worldMap->getCellAtCoords( worldMap->getXLow()+north*worldMap->getResolution() + 
					      worldMap->getResolution()/2.,
					      worldMap->getYLow()+east*worldMap->getResolution() +
					      worldMap->getResolution()/2.);
	  graphNodeID = north*graphWidth+east; // was set below print statement 10/30/07
	  cellPt->graphID = graphNodeID;
	  if( cellPt == NULL )
	    {
	      printf( "\n\n Big trouble... map does not match graph\n\n\n");
	      return false;
	    }
	  searcherGraph->nodes[graphNodeID].kidsSize = 0;
	  searcherGraph->nodes[graphNodeID].kidsIndexes = new int[8];
	  searcherGraph->nodes[graphNodeID].kidsCosts = new double[8];
	  searcherGraph->nodes[graphNodeID].moreInfo = new NodeInfo();
	  searcherGraph->nodes[graphNodeID].moreInfo->index = cellPt->getId();

	  // loop through the 8 children starting north and working clockwise
	  if( north+1 == graphHeight )
	    topBad = true;
	  else
	    topBad = false;
	  if( east+1 == graphWidth )
	    rightBad = true;
	  else
	    rightBad = false;
	  if( north == 0 )
	    bottomBad = true;
	  else
	    bottomBad = false;
	  if( east == 0 )
	    leftBad = true;
	  else
	    leftBad = false;
	  for( childCount=0; childCount<8; childCount++ )
	    {
	      switch( childCount )
		{
		case 0:
		  if( topBad )
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID + graphWidth;
		      childCost = straightCost;
		    }
		  break;
		case 1:
		  if( topBad || rightBad )
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID + graphWidth + 1;
		      childCost = diagCost;
		    }
		  break;
		case 2:
		  if( rightBad )
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID + 1;
		      childCost = straightCost;
		    }
		  break;
		case 3:
		  if( bottomBad || rightBad )
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID - graphWidth + 1;
		      childCost = diagCost;
		    }
		  break;
		case 4:
		  if( bottomBad )
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID - graphWidth;
		      childCost = straightCost;
		    }
		  break;
		case 5:
		  if( leftBad || bottomBad)
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID - graphWidth - 1;
		      childCost = diagCost;
		    }
		  break;
		case 6:
		  if( leftBad )
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID - 1;
		      childCost = straightCost;
		    }
		  break;
		case 7:
		  if( leftBad || topBad )
		    {
		      childNum = graphNodeID;
		      childCost = -1;
		    }
		  else
		    {
		      childNum = graphNodeID + graphWidth - 1;
		      childCost = diagCost;
		    }
		  break;
		default:
		  return(false);
		}
	      if( childCost >= 0 )
		{
		  searcherGraph->nodes[graphNodeID].kidsIndexes[searcherGraph->nodes[graphNodeID].kidsSize] = childNum;
		  searcherGraph->nodes[graphNodeID].kidsCosts[searcherGraph->nodes[graphNodeID].kidsSize] = childCost;
		  searcherGraph->nodes[graphNodeID].kidsSize++;
		}
	    }
	}
    }

  graphNodeID = searcherGraph->size-1;
  searcherGraph->nodes[graphNodeID].kidsSize = numAmSeedNodes;
  searcherGraph->nodes[graphNodeID].moreInfo = new NodeInfo();
  searcherGraph->nodes[graphNodeID].moreInfo->index = graphNodeID; // not used for anything in this node
  if( numAmSeedNodes == 0 )
    {
      //      return( checkGraph());
      return true;
    }

  searcherGraph->nodes[graphNodeID].kidsIndexes = new int[numAmSeedNodes];
  searcherGraph->nodes[graphNodeID].kidsCosts = new double[numAmSeedNodes];
  //  printf( "building new graph: start node %d\n", nodeID );
  for( childCount=0; childCount<numAmSeedNodes; childCount++ )
    {
      cellPt = worldMap->getCellAtCoords( amSeedNodes[childCount].x, 
					  amSeedNodes[childCount].y );
      cellPt->graphID = graphNodeID;
      searcherGraph->nodes[graphNodeID].kidsIndexes[childCount] = cellPt->getId();
      searcherGraph->nodes[graphNodeID].kidsCosts[childCount] = amSeedNodes[childCount].z;
      /*
      printf( "Connecting node %d at <%f %f> cost: %f\n",
	      childNum,
	      amSeedNodes[childCount].x,
	      amSeedNodes[childCount].y,
	      amSeedNodes[childCount].z );
      */
    }
  return(true);
  //  return( checkGraph() );
}

/////////////////////////////////////////////////////////////////////////////
//! Add costs to get to the AM perimeter to the graph
/*!
  This routine will connect the dummy node to the AM perimeter nodes. This
  will allow the vehicle echelon to plan where it knows best while taking 
  advantage of the AM echelons plan
*/
bool VehMobPLEngine::addGraphPoints( PM_CARTESIAN *points, int pointLength )
{
  VehMobCell *cellPt;
  int count, numAdded;
  static int *nodesToGraph = 0;
  static double *costsToGraph = 0;
  bool needToGrow = false;
  PM_CARTESIAN location;

  nodesToGraph = new int[MAX_AM_SEED_NODES];
  costsToGraph = new double[MAX_AM_SEED_NODES];
  numAdded = 0;
  for( count=0; count<pointLength; count++ )
    {
      if( points[count].z > 0 )
	{
	  cellPt = worldMap->getCellAtCoords( points[count].x, points[count].y );
	  if( cellPt == 0 )
	    {
	      needToGrow = true;
	      if( !growWorld(points[count]) )
		{
		  fprintf( stderr, "Error growing world for AM perimeter location\n" );
		  return false;
		}
	      cellPt = worldMap->getCellAtCoords( points[count].x, points[count].y );
	      if( cellPt == 0 ) 
		return false;
	    }
	  nodesToGraph[numAdded] = cellPt->graphID;
	  costsToGraph[numAdded] = points[count].z;
	  amSeedNodes[numAdded] = points[count];
	  if( myVerbose )
	    fprintf( stderr, "seed %d of id %d is <%f %f> with cost %f\n",
		     numAdded, nodesToGraph[numAdded],
		     points[count].x, points[count].y,
		     points[count].z );
	  numAdded++;
	  numAmSeedNodes = numAdded;
	}
    }
  if( numAdded == 0 ) // no points added so use vehicle location
    {
      printf( "Using vehicle location as start point\n" );
      location = worldMap->getVehLoc();
      cellPt = worldMap->getCellAtCoords( location.x, location.y ); // will never need to grow from veh location
      if( cellPt == 0 ) 
	return false;
      nodesToGraph[numAdded] = cellPt->graphID;
      costsToGraph[numAdded] = 0;
      location.z = 0;
      amSeedNodes[numAdded] = location;
      if( myVerbose )
	fprintf( stderr, "seed %d of id %d is <%f %f> with cost %f\n",
		 numAdded, nodesToGraph[numAdded],
		 points[count].x, points[count].y,
		 points[count].z );
      numAdded++;
      numAmSeedNodes = numAdded;
    }
  if( needToGrow ) // need to reassign nodeIDs since they changed from growth
    {
      printf( "Grew map so changing node numbering\n" );
      numAdded = 0;
      for( count=0; count<pointLength; count++ )
	{
	  if( points[count].z > 0 )
	    {
	      cellPt = worldMap->getCellAtCoords( points[count].x, points[count].y );
	      nodesToGraph[numAdded] = cellPt->graphID;
	      costsToGraph[numAdded] = points[count].z;
	      if( myVerbose )
		fprintf( stderr, "seed %d of id %d is <%f %f> with cost %f\n",
			 numAdded, nodesToGraph[numAdded],
			 points[count].x, points[count].y,
			 points[count].z );
	      numAdded++;
	    }
	}
    }
  searcherGraph->useKids(searcherGraph->size-1, numAdded, nodesToGraph, costsToGraph);
  return true;
}

/////////////////////////////////////////////////////////////////////////////
//!  Discount prior path or remove discount that was already applied
/*!
  This routine attempts to remove flopping between two similarly priced
  paths by discounting the last path that was traversed.
  \param doTurnOn true: apply discount, false: remove discount
*/
void VehMobPLEngine::discount( bool doTurnOn )
{
  /*
  int count;
  int north0, north1, east0, east1;
  int diffX, diffY, kid = 0;
  double nextDistance;

  if( doTurnOn )
    {
      if( discountPathLength < pathLength )
	{
	  if( discountPathLength != 0 )
	    delete [] discountPath;
	  discountPath = new int[pathLength];
	}
      for( count=0; count<pathLength; count++ )
	discountPath[count] = path[count];
      discountPathLength = pathLength;
    }

  for( count=0; count<discountPathLength-1; count++ )
    {
      north0 = mapXToWmNorth( graphToMapX(discountPath[count]) );
      north1 = mapXToWmNorth( graphToMapX(discountPath[count+1]) );
      east0 = mapYToWmEast( graphToMapY(discountPath[count]) );
      east1 = mapYToWmEast( graphToMapY(discountPath[count+1]) );
      diffX = north1-north0;
      diffY = east1-east0;
      if( diffX==0 || diffY == 0 )
	nextDistance = straightDist;
      else
	nextDistance = diagDist;
      if( doTurnOn )
	nextDistance *= 0.9; // discount factor of .9
      if( diffX == 1 )
	{
	  if( diffY == -1 )
	    kid = 7;
	  else if( diffY == 0 )
	    kid = 0;
	  else // diffY = 1
	    kid = 1;
	}
      else if( diffX == 0 )
	{
	  if( diffY == -1 )
	    kid = 6;
	  // 0, 0 not allowed
	  else if( diffY == 1 )
	    kid = 2;
	}
      else // diffX == -1
	{
	  if( diffY == -1 )
	    kid = 5;
	  else if( diffY == 0 )
	    kid = 4;
	  else // diff = 1
	    kid = 3;
	}
      searcherGraph->nodes[discountPath[count]].kidsCosts[kid] = setCost( nextDistance, north0, east0,
								  north1, east1 );
    }
  */
  return;
}

/***********************************************************************/
//! Routine to grow world model
/*!
  This routine will create a new larger world model and then copy the old one into it.
  \param includeLoc Location to include in new world model
  \return true on success and false on failure
*/
bool VehMobPLEngine::growWorld(  PM_CARTESIAN &includeLoc )
{
  int newGraphWidth, newGraphHeight;
  double fudge = worldMap->getResolution()/2.; // account for double rounding errors in comparisions

  // determine the new lower left corner of the world 
  if( includeLoc.x <= worldMap->getXLow() + fudge  ) // below world
    {
      newGraphHeight = (int)(ceil((worldMap->getXLow() - includeLoc.x)/worldMap->getResolution())*2.) + graphHeight + 1;
    }
  else if( includeLoc.x >= worldMap->getXHi() - fudge ) // above world
    {
      newGraphHeight = (int)(ceil((includeLoc.x - worldMap->getXHi())/worldMap->getResolution())*2.) + 
	graphHeight + 1;
    }
  else // everything OK
    {
      newGraphHeight = graphHeight;
    }

  if( includeLoc.y <= worldMap->getYLow() + fudge ) // left of world
    {
      newGraphWidth = (int)(ceil((worldMap->getYLow() - includeLoc.y)/worldMap->getResolution())*2.) + graphWidth + 1;
    }
  else if( includeLoc.y >= worldMap->getYHi() - fudge ) // right of world
    {
      newGraphWidth = (int)(ceil((includeLoc.y - worldMap->getYHi())/worldMap->getResolution())*2.) +
	graphWidth + 1;
    }
  else // everything OK
    {
      newGraphWidth = graphWidth;
    }
  if( newGraphWidth*worldMap->getResolution() > 1000 || newGraphHeight*worldMap->getResolution() > 1000 ) // map wants to grow larger then 1000 meters!
    {
      fprintf( stderr, "Map wants to grow too large! <%d %d>\n", (int) (newGraphWidth*worldMap->getResolution()),
	       (int) (newGraphHeight*worldMap->getResolution()) );
      return(false);
    }
  return( growWorld( newGraphWidth, newGraphHeight ) );
}

/***********************************************************************/
//! Routine to grow world model
/*!
  This routine will create a new larger world model and then copy the old one into it.
  \param newGraphWidth new width of world model
  \param newGraphHeight new height of world model
  \return true on success and false on failure
*/
bool VehMobPLEngine::growWorld( int newGraphWidth, int newGraphHeight )
{
  VehMobMap *newWorldMap;
  int oldWidth, oldHeight;

  if( newGraphHeight != newGraphWidth ) // only square supported now
    {
      if( newGraphHeight > newGraphWidth )
	newGraphWidth = newGraphHeight;
      else
	newGraphHeight = newGraphWidth;
    }
  
  printf( "vehMobPLEngine::growWorld: old:<%d, %d> new:<%d %d>\n",
	  graphHeight, graphWidth, newGraphHeight, newGraphWidth);
  //*********************
  // allocate and copy in new world model
  // lets make sure that we agree on the size
  newGraphHeight /= 2;
  newWorldMap = worldMap->copy(newGraphHeight);
  newGraphHeight = newWorldMap->getRows();
  newGraphWidth = newGraphHeight;
  if( newWorldMap == NULL )
    return false;

  delete worldMap;
  delete cImg;
  //******************
  // reset widths

  oldWidth = graphWidth;
  oldHeight = graphHeight;
  graphWidth = newGraphWidth;
  graphHeight = newGraphHeight;
  worldMap = newWorldMap;
  if( worldMap<=0 )
    {
      printf( "******************************\nUnable to allocate memory for world model BAD!!!\n*************************************\n");
      return false;
    }

  cImg = new cimg_library::CImg<unsigned char>(graphHeight,graphWidth,1,3);
  //  cImg.resize(graphHeight,graphWidth,1,3,0);
  if( !buildGraph() )
    return( false );

  // make sure that all the needed updates are performed
  /*
  if( heightDiff+oldHeight < graphHeight )
    oldHeight++;
  if( widthDiff+oldWidth < graphWidth )
    oldWidth++;
  if( heightDiff > 0 )
    heightDiff--;
  if( widthDiff > 0 )
    widthDiff--;
  updateGraph( heightDiff, widthDiff, heightDiff+oldHeight, widthDiff+oldWidth );
  */
  //  updateGraph( 0, 0, graphHeight, graphWidth );
  updateGraph( worldMap->getXLow(), worldMap->getYLow(),
	       worldMap->getXHi(), worldMap->getYHi() );
  return( true );
}

/***********************************************************************/
/*!
  Perform the actual search while conforming to the time constraints

\param start The start location.
\param end   The end location.
\return -1 error
\return The value of pathFound from the searcher (0 not reached, 1 reached, 2 optimally reached
*/
int VehMobPLEngine::search(  PM_CARTESIAN &goal,
			    double planStartHorizon, bool newSearch, double endTime )
{
  VehMobCell *cellPt;
  int numberOfSearches = 0;
  static int startNode, finishNode;
  //  static int errCount=0;
  static int opensOut;
  static int statusesOut;
  static int goalsFoundOut=0;
  static int optimaFoundOut=0;

  if( newSearch )
    {
      cellPt = worldMap->getCellAtCoords( goal.x, goal.y );
      if( cellPt == 0 )
	{
	  if( !growWorld(goal) )
	    {
	      fprintf( stderr, "Error growing world for finish location\n" );
	      return(-1);
	    }
	  // now get goal location
	  cellPt = worldMap->getCellAtCoords( goal.x, goal.y );
	  if( cellPt == 0 ) 
	    return -1;
	}
      finishNode = cellPt->graphID;
  
      startNode = searcherGraph->size-1;
      if( theSearcher->initSearch( 
				  searcherGraph, /* the graph to search in                     */
				  startNode,               /* index in nodes of aGraph of start node     */
				  1,          /* number of goals                            */
				  &finishNode,           /* indexes in nodes of aGraph of goal nodes   */
				  &opensOut,          /* 0=open list empty, non-zero=list not empty */
				  &statusesOut,       /* statuses of reaching goals                 */
				  &costsOut,       /* costs of reaching goals                    */
				  &goalsFoundOut,     /* number of goals reached                    */
				  &optimaFoundOut)    /* number of goals reached optimally          */
	  )
	return( -1 ); // init error
    }
  while( optimaFoundOut != 1 && opensOut != 0 && numberOfSearches++ < 10000)
    {
      if( numberOfSearches++ > 5000 )
	{
	  if( etime() > endTime )
	    {
	      if( goalsFoundOut )
		break;
	      printf( "OUT OF TIME FOR VEH PLAN! WAITING FOR NEXT CYCLE\n" );
	      return( statusesOut );
	    }
	  numberOfSearches = 0;
	}
      theSearcher->cycleSearch();
    }
  //  if( optimaFoundOut )
  if( goalsFoundOut )
    {
      pathLength = theSearcher->reportSearch( finishNode, MAX_DIJK_WAYPOINTS, path );
      if( pathLength == -1 )
	return( -1 );
      if( !optimaFoundOut )
	printf( "Sub-optimal goal being reported\n" );
      printAns(costsOut, planStartHorizon );
      //      delete [] path;
      //      path = NULL;
    }
  //  else
  //    fprintf(stderr, "No path found %d\r", errCount++ );
  if( goalsFoundOut || opensOut==0 )
    theSearcher->closeSearch();
  //  printf( "Number of searches: %d\n", numberOfSearches );
  return( statusesOut );
}

/////////////////////////////////////////////////////////////////////////////
/*! 
  Determine the maximum cost of the first "checkCells" cells in the path

  \param checkCells The number of cells to check
  \return The maximum cost
*/
double VehMobPLEngine::maxCost( int checkCells )
{
  int count;
  int pathOutLength = pathLength-1;
  double maxCost = 0;
  double currentCost;

  if( myVerbose>1 )
    fprintf( stderr, "Plan points and costs from maxCost\n" );
  if( pathOutLength < checkCells )
    checkCells = pathOutLength;
  for( count=0; count<checkCells-1; count++ )
    {
      currentCost = setCost( 1, pathOut[count].x, pathOut[count].y,
			     pathOut[count+1].x, pathOut[count+1].y );
      if( myVerbose>1 )
	fprintf( stderr, "<%f %f> cost: %f\n", pathOut[count+1].x,
		 pathOut[count+1].y, currentCost );
      if( currentCost > maxCost )
	maxCost = currentCost;
    }
  return(maxCost);
}

/***********************************************************************/
/*!
  Convert the path to map coordinates and save in the pathOut array

\param cost The total path cost. Only used for printing answer with verbose on
*/
void VehMobPLEngine::printAns( double cost, double planStartHorizon)     
{
  int n, pathOutLength;
  static int maxPathLength = 1;
  double minDistance, tempDistance;
  PM_CARTESIAN planStart;
  int closeIndex;

  //  myVerbose = 1;
  //  if( !checkGraph())
  //    printf("\n\n\n\n\nError in graph\n\n\n\n");


  if( numAmSeedNodes <= 0 )
    return;

  while ( pathLength > maxPathLength ) //get enough room for path output
    {
      delete pathOut;
      maxPathLength *= 2;
      pathOut = new PM_CARTESIAN[maxPathLength];
    }

  /* just match the first point to a seed node and
     send out the seed location
  */
  //  planStart = worldMap->graphToMap(path[1]);
  planStart = worldMap->graphToMap(searcherGraph->nodes[path[1]].moreInfo->index);
  minDistance = hypot( planStart.x-amSeedNodes[0].x, planStart.y-amSeedNodes[0].y );
  closeIndex = 0;
  for( n=2; n<numAmSeedNodes; n++ )
    {
      if( amSeedNodes[n].z > 0 )
	{
	  tempDistance = hypot( planStart.x-amSeedNodes[n].x, planStart.y-amSeedNodes[n].y );
	  if( tempDistance < minDistance )
	    {
	      closeIndex = n;
	      minDistance = tempDistance;
	    }
	}
    }
  if( myVerbose )
    printf( "Distance from seed <%f %f> to path start <%f %f>: %f\n", 
	    amSeedNodes[closeIndex].x,
	    amSeedNodes[closeIndex].y,
	    planStart.x,
	    planStart.y,
	    minDistance );
  pathOut[0] = amSeedNodes[closeIndex];

  if( myVerbose )
    fprintf(stderr, "Path from start to finish:\n");
  /* remove first planStartHorizon  meters of path */
  /*
  n = 1;
  while( n<pathLength &&  myDisp( worldMap->graphToMap(path[0]), 
				worldMap->graphToMap(path[n]) )<
	 planStartHorizon )
    {
      n++;
      if( myVerbose )
	{
	  pathOut[0] = worldMap->graphToMap(path[n]);
	fprintf( stderr, "Removing from path %d <%f, %f>\n", 
		 path[n], pathOut[0].x, pathOut[0].y);
	}
    }
  if( n==pathLength ) // don't eliminate the whole path!
    //    n--;
    n=1;
  */
  n=1;
  for (pathOutLength = 2; pathOutLength < pathLength; pathOutLength++, n++)
    {
      pathOut[n] = worldMap->graphToMap(searcherGraph->nodes[path[pathOutLength]].moreInfo->index);
      if( myVerbose )
	fprintf( stderr, "%d <%f, %f>\n", path[n], 
		 pathOut[n].x, pathOut[n].y);
    }
  if( myVerbose )
    {
      printf("\n");
      printf("With cost: %.4f\n\n", cost);
    }


  //    myVerbose =0;
}

/***********************************************************************/
/*! Clear any map entries newer then the given value
    Also resets the graph costs and grow maps
 */
void VehMobPLEngine::clearWorld( unsigned int oldestTime )
{
  VehMobCell *cellPt;
  double xLoc, yLoc;
  SensorPoint sensorPoint;
  double resolution = worldMap->getResolution();

  sensorPoint.param1 = VEH_WM_UNKNOWN;
  for( xLoc=worldMap->getXHi()+resolution/2.; xLoc>worldMap->getXLow()-resolution/2.; 
       xLoc-=resolution )
    {
      for( yLoc=worldMap->getYLow()+resolution/2.; yLoc<worldMap->getYHi(); 
	   yLoc+=resolution )
	{
	  cellPt = worldMap->getCellAtCoords(xLoc,yLoc);
	  if( cellPt != NULL )
	    {
	      if( cellPt->lastSeen > oldestTime )
		{
		  sensorPoint.dataLoc.x = xLoc;
		  sensorPoint.dataLoc.y = yLoc;
		  sensorPoint.sensorLoc = sensorPoint.dataLoc;
		  updateWorld( &sensorPoint, 1, oldestTime, NO_MARK);
		}
	    }
	}
    }
}

/***********************************************************************/
/*! Temporaily block a given waypoint
 returns a 1 if the blocked point is the current goal
 returns 0 else
 */
bool VehMobPLEngine::blockClosestWay( PM_CARTESIAN location[], int numLocations, 
				      unsigned int untilWhen, goalPoint *currentGoal,
				      VehCellMark markValue, bool force)
{
  PM_CARTESIAN pathLocation;
  SensorPoint sensorPoint;
  //  double minDist, currentDist;
  //  int which;
  VehMobCell *cellPt, *goalCell;
  int retValue = 0;
  
  for( int i=0; i<numLocations; i++ )
    {
      sensorPoint.dataLoc = location[i];
      sensorPoint.sensorLoc = sensorPoint.dataLoc;
      sensorPoint.param1 = VEH_WM_DANGEROUS_MAX;

      cellPt = worldMap->getCellAtCoords(sensorPoint.dataLoc.x, 
					 sensorPoint.dataLoc.y );
      if( cellPt == 0 )
	{
	  if( !growWorld(location[i]) )
	    {
	      fprintf( stderr, "Error growing world for blockClosestWay\n" );
	      return(0);
	    }
	  // now get goal location
	  cellPt = worldMap->getCellAtCoords( sensorPoint.dataLoc.x, 
					      sensorPoint.dataLoc.y );
	  if( cellPt == 0 ) 
	    return 0;
	}

      if( myVerbose >= 1 )
	fprintf( stdout, "Removing cell: <%f, %f> with value <%d %d> until %d",
		 sensorPoint.dataLoc.x, sensorPoint.dataLoc.y, 
		 cellPt->obstacleMin, cellPt->obstacleMax,untilWhen );

      if( !force && cellPt->obstacleMin==VEH_WM_UNKNOWN )
	continue;
      // mark the cell as an obstacle until untilWhen
      //      if( numLocations == 1 )
      //      cellPt->mark = markValue;
      updateWorld( &sensorPoint, 1, untilWhen, markValue);

      if( myVerbose >= 1 )
	fprintf( stdout, " new value <%d %d>\n",
		 cellPt->obstacleMin, cellPt->obstacleMax );
      if( currentGoal != NULL )
	{
	  goalCell = worldMap->getCellAtCoords( currentGoal->location.x,
						currentGoal->location.y );
	  if( cellPt->getId() == goalCell->getId() )
	    retValue = 1;
	}
    }
  return retValue;
}

/***********************************************************************/
/*! Update the arc costs of all cells between given values
  \param minx Minimum x value to check
  \param miny Minimum y value to check
  \param maxx Maximum x value to check
  \param maxy Maximum y value to check
*/
void VehMobPLEngine::updateGraph( double minx, double miny, double maxx, double maxy )
{
  double xCount, yCount;

  for( xCount=minx; xCount<maxx; xCount+=worldMap->getResolution() )
    {
      for( yCount=miny; yCount<maxy; yCount+=worldMap->getResolution() )
	{
	  modifyArcs( xCount, yCount );
	}
    }
}

/***********************************************************************/
/*! Update the world model the the given map updates

  This routine calls the point version for each element of the map.
  This could be made more efficient.

  \param updates The updates to be applied
  \param timeSeen When was this update seen
  \return Always returns 1 (Useful!)
*/
int VehMobPLEngine::updateWorld( SensorData2D *updates, unsigned int timeSeen,  PM_CARTESIAN &vehLocation )
{
  SensorPoint updatePoint;
  int xCount, yCount;
  PM_CARTESIAN location;
  int featLoc;

  location.x = updates->lowerLeft.x;
  for( xCount = 0; xCount < updates->x; xCount++ )
    {
      location.y = updates->lowerLeft.y;
      for( yCount = 0; yCount < updates->y; yCount++ )
	{
	  updatePoint.dataLoc = location;
	  updatePoint.sensorLoc = vehLocation;
	  featLoc = xCount*updates->y + yCount;
	  if( updates->featArray[featLoc] == AMSP_OBSTACLE )
	    updatePoint.param1 = VEH_WM_DANGEROUS_MAX;
	  else if( updates->featArray[featLoc] == AMSP_MAYBE_OBSTACLE )
	    updatePoint.param1 = VEH_WM_DIFFICULT_MAX;
	  else if( updates->featArray[featLoc] == AMSP_NOT_OBSTACLE )
	    updatePoint.param1 = VEH_WM_EASY_MAX/2;
	  else if( updates->featArray[featLoc] == AMSP_UNSEEN ) // do not process these
	    updatePoint.param1 = VEH_WM_UNKNOWN;

	  if( updatePoint.param1 != VEH_WM_UNKNOWN )
	    updateWorld( &updatePoint, 1, timeSeen, NO_MARK );
	  location.y += updates->incrementY;
	}
      location.x += updates->incrementX;
    }
  return(1);
}

/***********************************************************************/
/*! Update the world model the the given map updates

  \param updates The updates to be applied
  \param length The number of updates to apply
  \param timeSeen When was this update seen
  \return Always returns 1 (Useful!)
*/
int VehMobPLEngine::updateWorld( SensorPoint *updates, int length, unsigned int timeSeen, VehCellMark markIn )
{
  int count;
  int nCount, eCount, obsGrowInc;
  PM_CARTESIAN growLoc;
  VehMobCell *cellPt;
  double nValue, eValue;
  double angleToCell, testAngle;

  for( count=0; count<length; count++ )
    {
      if( myVerbose>2)
	fprintf( stdout, "<%f %f %f> %f\n", updates[count].dataLoc.x, 
		 updates[count].dataLoc.y,
		 updates[count].dataLoc.z,  updates[count].param1 );
      // get cell pointer for location
      cellPt = worldMap->getCellAtCoords(updates[count].dataLoc.x, 
					 updates[count].dataLoc.y );
      // if necessary, expand world to accomodate new obstacle
      if( cellPt == 0 )
	{
	  growLoc.x = updates[count].dataLoc.x;
	  growLoc.y = updates[count].dataLoc.y;
	  growLoc.z = 0;
	  if( !growWorld( growLoc ) )
	    continue;
	  cellPt = worldMap->getCellAtCoords(updates[count].dataLoc.x, 
					     updates[count].dataLoc.y );
	}
      /* do we need to do anything? 
	    If report is older, then ignore
	    If report is newer, then believe
	    If report is in same time frame, then figure out what to do
      */
      // check to see if we have to mark this as cleared
      if( !cellPt->isCleared )
	{
	  // can we see it now?
	  angleToCell = atan2( (updates[count].dataLoc.x - updates[count].sensorLoc.x), 
			       (updates[count].sensorLoc.y - updates[count].dataLoc.y) );
	  angleToCell -= M_PI_2;
	  if( angleToCell < 0 )
	    angleToCell += 2*M_PI;
	  testAngle = fabs(angleToCell-worldMap->getVehYaw());
	  // are we in the FOV of the sensor
	  if( testAngle<RFIDHalfFOV || testAngle>(2*M_PI-RFIDHalfFOV) )
	    {
	      // are we in the range of the sensor
	      if( myDisp(updates[count].dataLoc, updates[count].sensorLoc)<RFIDRange )
		cellPt->isCleared = true;
	    }
	}
      if( cellPt->lastSeen > timeSeen ) // this report is older then the cell data
	continue;
      else if( cellPt->lastSeen == timeSeen )
	{
	  if( cellPt->obstacleMax < updates[count].param1 )
	    cellPt->obstacleMax = (unsigned char)updates[count].param1;
	  if( cellPt->obstacleMin <= updates[count].param1 )
	    continue;
	}
      cellPt->lastSeen = timeSeen;
      cellPt->mark = markIn;
      // determine how obsGrowMap will be modified
      if( obsGrow == 0 )
	obsGrowInc = 0;
      else if( cellPt->obstacleMin >= VEH_WM_GROW_OBSTACLE_MIN &&
	       updates[count].param1 >= VEH_WM_GROW_OBSTACLE_MIN )
	obsGrowInc = 0;
      else if( cellPt->obstacleMin >= VEH_WM_GROW_OBSTACLE_MIN )
	obsGrowInc = -1;
      else if( updates[count].param1 >= VEH_WM_GROW_OBSTACLE_MIN )
	obsGrowInc = +1;
      else
	obsGrowInc = 0;

      cellPt->obstacleMin = (unsigned char)updates[count].param1;
      cellPt->obstacleMax = (unsigned char)updates[count].param1;
      modifyArcs( updates[count].dataLoc.x, updates[count].dataLoc.y );

      if( obsGrowInc != 0 )
	{
	  for( nCount= -obsGrow; nCount<=obsGrow; nCount++ )
	    {
	      for( eCount= -obsGrow; eCount<=obsGrow; eCount++ )
		{
		  // modify 16 arc costs
		  nValue = updates[count].dataLoc.x+nCount*worldMap->getResolution();
		  eValue = updates[count].dataLoc.y+eCount*worldMap->getResolution();
		  cellPt = worldMap->getCellAtCoords(nValue, eValue );
		  if( cellPt != 0 )
		    {
		      if( obsGrowInc<0 && cellPt->grow==0 )
			modifyArcs( nValue, eValue );
		      cellPt->grow += obsGrowInc;
		      modifyArcs( nValue, eValue );
		    }
		}
	    }
	}
      else
	{
	  // modify 16 arc costs (already done above, is this necessary?)
	  //	  modifyArcs( updates[count].dataLoc.x, updates[count].dataLoc.y );
	}
    }
  return(1);
}


/***********************************************************************/
//! Clear the world model and obsGrow map
/*!
  Clear all of the value from the obsGrow and world model maps.
  The world model and arc costs are set to default values
*/
void VehMobPLEngine::clearMap()
{
  double straightCost = costLookup[VEH_WM_UNKNOWN] * straightDist*2;
  double diagCost = costLookup[VEH_WM_UNKNOWN] * diagDist*2;
  int nodeCount, childCount;
  double xLoc, yLoc;
  VehMobCell *cellPt;

  worldMap->clearMap();

  for( nodeCount=0; nodeCount<graphWidth*graphHeight; nodeCount++ )
    {
      for( childCount=0; childCount<8; childCount+=2 )
	{
	  searcherGraph->nodes[nodeCount].kidsCosts[childCount] = straightCost;
	  searcherGraph->nodes[nodeCount].kidsCosts[childCount+1] = diagCost;
	}
    }
  // invalidate the edges on the perimiter
  // top
  xLoc = worldMap->getXHi();
  for( yLoc=worldMap->getYLow(); yLoc<=worldMap->getYHi(); yLoc+=worldMap->getResolution() )
    {
      cellPt = worldMap->getCellAtCoords(xLoc, yLoc);
      searcherGraph->nodes[cellPt->graphID].kidsCosts[7] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[0] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[1] = -1;
    }
  // bottom
  xLoc = worldMap->getXLow();
  for( yLoc=worldMap->getYLow(); yLoc<=worldMap->getYHi(); yLoc+=worldMap->getResolution() )
    {
      cellPt = worldMap->getCellAtCoords(xLoc, yLoc);
      searcherGraph->nodes[cellPt->graphID].kidsCosts[5] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[4] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[3] = -1;
    }
  // left
  yLoc = worldMap->getYLow();
  for( xLoc=worldMap->getXLow(); xLoc<=worldMap->getXHi(); xLoc+=worldMap->getResolution() )
    {
      cellPt = worldMap->getCellAtCoords(xLoc, yLoc);
      searcherGraph->nodes[cellPt->graphID].kidsCosts[7] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[6] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[5] = -1;
    }
  // right
  yLoc = worldMap->getYHi();
  for( xLoc=worldMap->getXLow(); xLoc<=worldMap->getXHi(); xLoc+=worldMap->getResolution() )
    {
      cellPt = worldMap->getCellAtCoords(xLoc, yLoc);
      searcherGraph->nodes[cellPt->graphID].kidsCosts[1] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[2] = -1;
      searcherGraph->nodes[cellPt->graphID].kidsCosts[3] = -1;
    }
}

/***********************************************************************/
//! convert vehicle coordinates to output file map coordinates
/*! This will convert from world coordinates in the scrolling map to 
    image coordinates for the ppm output file

    \param xLoc The x location to be converted
    \param yLoc The y location to be converted
    \return The index into the image array
*/
inline PM_CARTESIAN VehMobPLEngine::worldToImageConvert( double x, double y )
{
  PM_CARTESIAN retValue;

  retValue.y = (int)((worldMap->getXHi()-x) / worldMap->getResolution()); // position in x
  retValue.x = (int)((y-worldMap->getYLow()) / worldMap->getResolution());
  return retValue;
}
/*
inline int VehMobPLEngine::worldToImageConvert( double x, double y )
{
  PM_CARTESIAN pmCart;
  int index;

  pmCart = worldToImageConvert( x, y );
  index = pmCart.x * graphWidth + pmCart.y;
  return index;
}
*/
/***********************************************************************/
 /*! write out the vehicle level map to NML */
#define MAP_OUT_RADIUS 5
void VehMobPLEngine::mapOut( NML *nmlOutChan )
{
  vector<VictInfo>::iterator it;
  double xLoc, yLoc;
  PM_CARTESIAN location;
  VehMobCell *cellPt;
  double resolution;
  double mapSide;
  int arrayLength;

  resolution = worldMap->getResolution();
  location = worldMap->getVehLoc();
  mapSide = MAP_OUT_RADIUS-resolution/2.;

  arrayLength = 0;
  vehMapOut->isGlobal = true;
  vehMapOut->lowerLeft.x = location.x-mapSide;
  vehMapOut->lowerLeft.y = location.y-mapSide;
  vehMapOut->incrementX = resolution;
  vehMapOut->incrementY = resolution;
  vehMapOut->x = (int)floor(MAP_OUT_RADIUS*2/resolution);
  vehMapOut->y = (int)floor(MAP_OUT_RADIUS*2/resolution);
  for( xLoc=vehMapOut->lowerLeft.x; xLoc<location.x+mapSide; xLoc+=resolution )
    {
      for( yLoc=vehMapOut->lowerLeft.y; yLoc<location.y+mapSide; yLoc+=resolution )
	{
	  cellPt = worldMap->getCellAtCoords(xLoc,yLoc);
	  if( cellPt != 0 )
	    vehMapOut->featArray[arrayLength++] = cellPt->obstacleMax;
	  else
	    vehMapOut->featArray[arrayLength++] = VEH_WM_UNKNOWN;
	}
    }
  vehMapOut->featArray_length = arrayLength;
  nmlOutChan->write(vehMapOut);
}

void VehMobPLEngine::mapOutSpqr( NML *nmlOutChan,  PM_CARTESIAN &vehGlobalLoc )
{
  vector<VictInfo>::iterator it;
  double xLoc, yLoc;
  PM_CARTESIAN location;
  VehMobCell *cellPt;
  double resolution;
  int arrayLength;

  resolution = worldMap->getResolution();
  location = worldMap->getVehLoc();

  arrayLength = 0;
  spqrMapOut->mapX = vehGlobalLoc.x - MAP_OUT_RADIUS;
  spqrMapOut->mapY = -vehGlobalLoc.y + MAP_OUT_RADIUS;
  for( yLoc=location.y-MAP_OUT_RADIUS; yLoc<location.y+MAP_OUT_RADIUS; yLoc+=resolution )
    {
      for( xLoc=location.x-MAP_OUT_RADIUS; xLoc<location.x+MAP_OUT_RADIUS; xLoc+=resolution )
	{
	  cellPt = worldMap->getCellAtCoords(xLoc,yLoc);
	  if( cellPt != 0 )
	    {
	      if( cellPt->obstacleMax == VEH_WM_UNKNOWN )
		spqrMapOut->map[arrayLength++] = C8Blue;
	      else if( cellPt->obstacleMax >= VEH_WM_GROW_OBSTACLE_MIN )
		spqrMapOut->map[arrayLength++] = C8Black;
	      else
		spqrMapOut->map[arrayLength++] = C8White;
	    }
	  else
	    spqrMapOut->map[arrayLength++] = C8Blue;
	  //	  printf( "Map element: %d is %d\n", arrayLength-1, spqrMapOut->map[arrayLength-1] );
	}
    }
  spqrMapOut->map_length = arrayLength;
  spqrMapOut->nRows = (int) sqrt(((double) arrayLength));
  spqrMapOut->nCols = spqrMapOut->nRows;
  spqrMapOut->mapRealWidth = spqrMapOut->nRows * resolution;

  //  printf( "Sending spqr map of length: %d rows: %d cols: %d\n", arrayLength, spqrMapOut->nRows, spqrMapOut->nCols );
  nmlOutChan->write(spqrMapOut);
}

/***********************************************************************/
//!  Dump the contents of the world model to a file
/*!
  The file may be opened once and left open to permit repeated output 
  (for example to a pipe or history file). If a file is open, then changing
  the filename will not affect the location of the output.

  \param fileName The name of the file to dump to
  \param vehLocation The location of the vehicle (drawn in a different color)
  \param closeOut Should the file be closed? 1-yes
*/
void VehMobPLEngine::dump( char *fileName,  PM_CARTESIAN &vehLocation )
{
  vector<VictInfo>::iterator it;
  int trash, i;
  double xLoc, yLoc;
  PM_CARTESIAN location;
  unsigned char *red, *blue, *green;
  VehMobCell *cellPt;
  double resolution;
  unsigned char myColor[3], backgroundColor[3];
  char victString[255];
  int length;
  FILE *fp;

  red =   &myColor[0];
  green = &myColor[1];
  blue =  &myColor[2];
  if( fileName == NULL )
    return;

  resolution = worldMap->getResolution();

  /*
  *red = 255;
  *green = 255;
  *blue = 255;
  cImg(0,0, 0) = *red;
  cImg(0,0, 1) = *green;
  cImg(0,0, 2) = *blue;
  cImg.save_png( fileName );

  return;
  */

  for( xLoc=worldMap->getXHi()-resolution/2.; xLoc>worldMap->getXLow(); xLoc-=resolution )
    {
      for( yLoc=worldMap->getYLow()+resolution/2.; yLoc<worldMap->getYHi(); yLoc+=resolution )
	{
	  cellPt = worldMap->getCellAtCoords(xLoc,yLoc);
	  *red = cellPt->obstacleMax;
	  *blue = cellPt->obstacleMax;
	  *green = cellPt->obstacleMax;
	  if( cellPt->isCleared ) // mark green for no victims
	    *green = 255;
	  if( cellPt->mark == AM_POINT || cellPt->mark == OBSERVE_POINT )
	    {
	      *red = 255;
	      *green = 255;
	      *blue = 255;
	    }
	  else if( cellPt->mark == KEEP_OUT )
	    {
	      *red = 0;
	      *green = 128;
	      *blue = 0;
	    }
	  else if( cellPt->obstacleMax >= VEH_WM_EXTREAM_MAX ) // red for obstacles
	    {
	      *red = 255;
	      *green = 0;
	      *blue = 0;
	    }
	  else if( cellPt->grow > 0 )
	    {
	      trash = 16*cellPt->grow;
	      if( trash > 128 )
		trash = 128;
	      *blue = 127 + (unsigned char)trash;
	    }
	  else if( cellPt->obstacleMax == VEH_WM_UNKNOWN ) // yellow for unknown
	    {
	      *red = 255;
	      *green = 255;
	      *blue = 0;
	    }
	  else if( cellPt->obstacleMax > VEH_WM_EASY_MAX )
	    {
	      *red = 252;
	      //	      green = 171;
	      *blue = 84;
	    }
	  location = worldToImageConvert(xLoc,yLoc);
	  /*
	  if( location.x > cImg->width || location.y > cImg->height )
	    printf( "Trying to draw point that is out of range\n" );
	  */
	  cImg->draw_point( (int)location.x, (int)location.y, 0, myColor);
	  //	  (*cImg)((int)location.x, (int)location.y, 0) = *red;
	  //	  (*cImg)((int)location.x, (int)location.y, 1) = *green;
	  //	  (*cImg)((int)location.x, (int)location.y, 2) = *blue;
	}
    }
  for( i=0; i<pathLength; i++ )
    {
      location = worldMap->graphToMap(searcherGraph->nodes[path[i]].moreInfo->index);
      location = worldToImageConvert(location.x, location.y);
      *red = 124;
      *green = 62;
      *blue = 167;
      /*
      if( location.x > cImg->width || location.y > cImg->height )
	printf( "Trying to draw point that is out of range\n" );
      */
      cImg->draw_point( (int)location.x, (int)location.y, 0, myColor);
      //      (*cImg)((int)location.x, (int)location.y, 0) = *red;
      //      (*cImg)((int)location.x, (int)location.y, 1) = *green;
      //      (*cImg)((int)location.x, (int)location.y, 2) = *blue;
    }
  // draw AM echelon seed values

  *red = 249;
  *green = 168;
  *blue = 88;
  for( i=0; i<numAmSeedNodes; i++ )
    {
      if( amSeedNodes[i].z > 0 )
	{
	  location = worldToImageConvert(amSeedNodes[i].x, amSeedNodes[i].y);
	  cImg->draw_point( (int)location.x, (int)location.y, 0, myColor);
	}
    }
  // draw vehicle location in purple
  location = worldToImageConvert( vehLocation.x, vehLocation.y );
  *red   = 124;
  *green = 62;
  *blue  = 167;
  cImg->draw_circle( (int)location.x, (int)location.y, (int)(1./(resolution*2)), myColor );
  //  cImg->draw_point( (int)location.x, (int)location.y, 0, myColor);
  // (*cImg)((int)location.x, (int)location.y, 0) = *red;
  // (*cImg)((int)location.x, (int)location.y, 1) = *green;
  // (*cImg)((int)location.x, (int)location.y, 2) = *blue;


  // draw victim names
  backgroundColor[0] = 255;
  backgroundColor[1] = 255;
  backgroundColor[2] = 255;
  for( it=victInfo.begin(); it!=victInfo.end(); it++ )
    {
      location = worldToImageConvert( it->location.x, it->location.y );
      //      sprintf( victString, "Victim: %s\n        %s", it->name, it->injury );
      sprintf( victString, "%d", it->id );
      *red = 0;
      *green = 0;
      *blue = 255;
      cImg->draw_circle( (int)location.x, (int)location.y, (int)(1./(resolution*2)), myColor );
      //      *red = 128;
      //      *green = 0;
      //      *blue = 0;
      cImg->draw_text( victString, (int)location.x, (int)location.y, myColor, backgroundColor, 7 );
    }


  //  cImg.save_bmp( "test.bmp" );
  //  cImg.save_pnm( "test.pnm" );
  length = strlen(fileName);
  //  printf( "ending of string %s is %s\n", fileName, &fileName[length-3] );
  if( !strcmp("ppm", &fileName[length-3]))
    cImg->save_png( fileName );  
  else if( !strcmp("tif", &fileName[length-3]))
    {
      // get filename for tfw file
      strncpy( victString, fileName, length-3 );
      strcpy( &victString[length-3], "tfw" );
      fp = fopen( victString, "w" );
      if( fp != NULL )
	{
	  fprintf( fp, "%lf\n%lf\n%lf\n%lf\n%lf\n%lf\n",
		   worldMap->getResolution(),
		   0.,0.,
		   -worldMap->getResolution(),
		   worldMap->getYHi(),
		   worldMap->getXLow() );
	  fclose(fp);
	}
      cImg->save_tiff( fileName );
    }
}


/***********************************************************************/
/*!
  Change the value of the arcs in the given cell to reflect the current values
  in the world model

  \param north The north (x) value to be set
  \param east The east (y) value to be set
*/
int VehMobPLEngine::modifyArcs( double north, double east )
{
  double cost, resolution;
  VehMobCell *cellUL, *cellU, *cellUR;
  VehMobCell *cellLeft, *cellPt, *cellR;
  VehMobCell *cellLL, *cellDown, *cellLR;

  resolution = worldMap->getResolution();
  cellUL = worldMap->getCellAtCoords( north + resolution,
				      east - resolution );
  cellU = worldMap->getCellAtCoords( north + resolution,
				     east );
  cellUR = worldMap->getCellAtCoords( north + resolution,
				      east + resolution );
  cellLeft = worldMap->getCellAtCoords( north,
					east - resolution );
  cellPt = worldMap->getCellAtCoords( north, east );
  cellR = worldMap->getCellAtCoords( north,
				     east + resolution );
  cellLL = worldMap->getCellAtCoords( north - resolution,
				      east - resolution );
  cellDown = worldMap->getCellAtCoords( north - resolution,
					east );
  cellLR = worldMap->getCellAtCoords( north - resolution,
				      east + resolution );
  //top row
  // upper left
  if( cellUL )
    {
      cost = setCost( diagDist, north, east, 
		      north+resolution, east-resolution );
      searcherGraph->nodes[cellPt->graphID].kidsCosts[7] = cost;
      searcherGraph->nodes[cellUL->graphID].kidsCosts[3] = cost;
    }
  // straight up
  if( cellU )
    {
      cost = setCost( straightDist, north, east,
		      north+resolution, east );
      searcherGraph->nodes[cellPt->graphID].kidsCosts[0] = cost;
      searcherGraph->nodes[cellU->graphID].kidsCosts[4] = cost;
    }
  // upper right
  if( cellUR )
    {
      cost = setCost( diagDist, north, east, north+resolution, east+resolution );
      searcherGraph->nodes[cellPt->graphID].kidsCosts[1] = cost;
      searcherGraph->nodes[cellUR->graphID].kidsCosts[5] = cost;
    }
  // lower left
  if( cellLL )
    {
      cost = setCost( diagDist, north, east, north-resolution, east-resolution);
      searcherGraph->nodes[cellPt->graphID].kidsCosts[5] = cost;
      searcherGraph->nodes[cellLL->graphID].kidsCosts[1] = cost;
    }
  // straight down
  if( cellDown )
    {
      cost = setCost( straightDist, north, east, north-resolution, east );
      searcherGraph->nodes[cellPt->graphID].kidsCosts[4] = cost;
      searcherGraph->nodes[cellDown->graphID].kidsCosts[0] = cost;
    }
  // lower right
  if( cellLR )
    {
      cost = setCost( diagDist, north, east, north-resolution, east+resolution );
      searcherGraph->nodes[cellPt->graphID].kidsCosts[3] = cost;
      searcherGraph->nodes[cellLR->graphID].kidsCosts[7] = cost;
    }
  if( cellLeft ) // left
    {
      cost = setCost( straightDist, north, east, north, east-resolution );
      searcherGraph->nodes[cellPt->graphID].kidsCosts[6] = cost;
      searcherGraph->nodes[cellLeft->graphID].kidsCosts[2] = cost;
    }
  if( cellR ) // right
    {
      cost = setCost( straightDist, north, east, north, east+resolution );
      searcherGraph->nodes[cellPt->graphID].kidsCosts[2] = cost;
      searcherGraph->nodes[cellR->graphID].kidsCosts[6] = cost;
    }
  return(1);
}

/***********************************************************************/
/*!
  Set the cost lookup table values
*/
bool VehMobPLEngine::setCostLookup()
{
  int count;

  costLookup[0] = 1; // Jim says that we will never get index 0, but we know better
  if( VEH_WM_EASY_MAX >= 256 )
    return false;
  for( count=1; count<=VEH_WM_EASY_MAX; count++ )
    costLookup[count] = costLookup[count-1] + VEH_WM_EASY_INCREMENT;
  if( VEH_WM_EASY_MAX == 255 )
    return true;
  if( VEH_WM_MODERATE_MAX >= 256 )
    return false;
  for( count=VEH_WM_EASY_MAX+1; count<=VEH_WM_MODERATE_MAX; count++ )
    costLookup[count] = costLookup[count-1] + VEH_WM_MODERATE_INCREMENT;
  if( VEH_WM_MODERATE_MAX == 255 )
    return true;
  if( VEH_WM_DIFFICULT_MAX >= 256 )
    return false;
  for( count=VEH_WM_MODERATE_MAX+1; count<=VEH_WM_DIFFICULT_MAX; count++ )
    costLookup[count] = costLookup[count-1] + VEH_WM_DIFFICULT_INCREMENT;
  if( VEH_WM_DIFFICULT_MAX == 255 )
    return true;
  if( VEH_WM_EXTREAM_MAX >= 256 )
    return false;
  for( count=VEH_WM_DIFFICULT_MAX+1; count<=VEH_WM_EXTREAM_MAX; count++ )
    costLookup[count] = costLookup[count-1] + VEH_WM_EXTREAM_INCREMENT;
  if( VEH_WM_EXTREAM_MAX == 255 )
    return true;
  if( VEH_WM_DANGEROUS_MAX >= 256 )
    return false;
  for( count=VEH_WM_EXTREAM_MAX+1; count<VEH_WM_DANGEROUS_MAX; count++ )
    costLookup[count] = costLookup[count-1] + VEH_WM_DANGEROUS_INCREMENT;
  if( VEH_WM_DANGEROUS_MAX == 255 )
    return true;
  for( count=VEH_WM_DANGEROUS_MAX+1; count<256; count++ )
    costLookup[count] = costLookup[count-1] + VEH_WM_NOGO_INCREMENT;
  return true;
}

/***********************************************************************/
//!  The system cost function.
/*! 
  Returns the cost of going from world model cell 1 to cell 2
  \param graphDist The distance traveled between cells
  \param north1 The north (x) value of cell 1
  \param east1  The east (y) value of cell 1
  \param north2 The north (x) value of cell 2
  \param east2  The east (y) value of cell 2
*/
double VehMobPLEngine::setCost( double graphDist, double north1, double east1, double north2, double east2 )
{
  double cost;
  VehMobCell *cell1, *cell2;

  cell1 = worldMap->getCellAtCoords(north1, east1 );
  cell2 = worldMap->getCellAtCoords(north2, east2 );
  
  cost = graphDist * ( costLookup[cell1->obstacleMin] + 
		       costLookup[cell2->obstacleMin] +
		       costLookup[VEH_WM_GROW_OBSTACLE_INDEX] * 
		       (cell1->grow + 
			cell2->grow) );
  return(cost);
}

/***********************************************************************/
//! Store the world model to a file
/*!
  This is not really a planner function, but it is implemented here for
  convience.

\param fileName The file to store to
\return True on success
\return False on failure
*/
bool VehMobPLEngine::storeWM( char *fileName )
{
  FILE *fid;
  double xLoc, yLoc;
  VehMobCell *cellPt;

  fid = fopen( fileName, "w" );
  if( fid == NULL )
    return false;

  for( xLoc=worldMap->getXLow(); xLoc<=worldMap->getXHi(); xLoc++ )
    {
      for( yLoc=worldMap->getYLow(); yLoc<=worldMap->getYHi(); yLoc++ )
	{
	  cellPt = worldMap->getCellAtCoords(xLoc,yLoc);
	  if( cellPt->obstacleMin != VEH_WM_UNKNOWN )
	    {
	      fprintf( fid, "%f %f %d\n", xLoc, yLoc, cellPt->obstacleMin );
	    }
	}
    }
  fclose( fid );
  return true;
}

/***********************************************************************/
//! set the vehicle location
void VehMobPLEngine::setVehLoc(  PM_POSE &vehLocIn )
{
  worldMap->setVehLoc(vehLocIn);
}


/***********************************************************************/
//! Observe to the given range in the given direction
bool VehMobPLEngine::isObserved(  PM_CARTESIAN &start, double direction, double range, int observeTime )
{
  double endX, endY;
  VehMobCell *cellPt;
  SensorPoint unseen;
  
  //  startX = worldMap->getCenterX();
  //  startY = worldMap->getCenterY();
  endX = range * cos(direction) + start.x;
  endY = range * sin(direction)+ start.y;

  cellPt = worldMap->initLine( start.x, start.y, endX, endY );
  while( cellPt != NULL )
    {
      //      if( cellPt->obstacle==VEH_WM_UNKNOWN && cellPt->grow==0 )
      // base on isCleared
      if( !cellPt->isCleared )
	{
	  unseen.dataLoc = worldMap->graphToMap(cellPt->getId());
	  /*
	  unseen.sensorLoc = unseen.dataLoc;
	  unseen.param1 = VEH_WM_DANGEROUS_MAX;
	  updateWorld( &unseen, 1, observeTime+200 );
	  */
	  cellPt->mark = OBSERVE_POINT;
	  //	  printf( "Failed to see <%f %f> in direction %f\n",
	  //		  unseen.dataLoc.x, unseen.dataLoc.y, DEG(direction) );
	  return false;
	}
      else if( cellPt->obstacleMax>=VEH_WM_GROW_OBSTACLE_MIN || cellPt->grow>0)
	return true;
      cellPt = worldMap->getNextCell();
    }
  return true;
}

/***********************************************************************/
//! Load the world model to a file
/*!
  This is not really a planner function, but it is implemented here for
  convience.

\param fileName The file to load from
\return True on success
\return False on failure
*/
bool VehMobPLEngine::loadWM( char *fileName )
{
  SensorPoint *updates;
  int length;
  //  cimg_library::CImg<unsigned char> wmImg;
  FILE *fid;
  char *geoInfo, *geoPt;
  double eastOffset, northOffset;
  double northRes, eastRes;
  int north, east;
  PM_CARTESIAN includeLoc;

  /*
  // test code
  cimg_library::CImg<unsigned char> img(300,200,1,3);

  img.fill(32);
  img.noise(128);
  img.blur(2,0,0);
  const unsigned char white[] = {255,255,255};
  img.draw_text("Hello World",80,80,white,0,32);
  img.display();
  // end of test
  */
  cimg_library::CImg<unsigned char> wmImg(1000,1000,1,3);
  for( north=0; north<1000; north++ )
    for( east=0; east<1000; east++ )
      {
	wmImg(north,east,0,0) = 255;
	wmImg(north,east,0,1) = 0;
	wmImg(north,east,0,2) = 0;
      }
  //wmImg = wmImg.get_load(fileName);
  printf( "image size: %d %d\n", wmImg.dimy(), wmImg.dimx());
  if( wmImg.dimy()==0 || wmImg.dimx()==0 )
    return false;
  wmImg.display();
  geoInfo = new char[strlen(fileName)+5];
  strcpy( geoInfo, fileName );
  geoPt = strrchr( geoInfo, '.' );
  sprintf( geoPt, ".tfw" );
  printf( "goereference file to open: %s\n", geoInfo );
  fid = fopen(geoInfo, "r" );
  if( fid <= 0 )
    return false;
  fscanf( fid, "%lf", &eastRes ); // read x resolution
  fscanf( fid, "%lf", &eastOffset ); // read and throw away transloation
  if( eastOffset != 0 )
    {
      printf( "Offsets not supported when loading wm\n" );
      return false;
    }
  fscanf( fid, "%lf", &eastOffset ); // read and throw away rotation
  if( eastOffset != 0 )
    {
      printf( "Rotations not supported when loading wm\n" );
      return false;
    }
  fscanf( fid, "%lf", &northRes ); // read negative of y resolution
  northRes *= -1;
  fscanf( fid, "%lf", &eastOffset ); // read the x ground coordinate of pixel 1,1 (upper left)
  fscanf( fid, "%lf", &northOffset ); // read the y ground coordinate of pixel 1,1 (upper left)
  fclose( fid );
  
  printf(" data from georeference file: Offset: <%lf %lf> res: <%lf %lf>\n", 
	 northOffset,
	 eastOffset,
	 northRes,
	 eastRes);

  // grow world to fit new image
  /*
  includeLoc.x = ;
  includeLoc.y = ;
  growWorld(includeLoc);
  includeLoc.x = ;
  includeLoc.y = ;
  growWorld(includeLoc);
  */
  updates = new SensorPoint [100];
  length = 0;
  for( north=0; north<wmImg.dimy(); north++ ) // max of image height
    {
      for( east=0; east<wmImg.dimx(); east++ ) // max of image width
	{
	  if( wmImg(east,north,0,0) == 255 ) // obstacle
	    {
	      updates[length].dataLoc.z = 0;
	      updates[length].dataLoc.x = north*northRes + northOffset;
	      updates[length].dataLoc.y = east*eastRes + eastOffset;
	      updates[length].param1 = VEH_WM_DANGEROUS_MAX;
	      length++;
	      if( length == 100 )
		{
		  updateWorld( updates, length, 0, NO_MARK );
		  length = 0;
		}
	    }
	}
    }
  if( length > 0 )
    updateWorld( updates, length, 0, NO_MARK );
  length = 0;
  delete [] updates;
  delete geoInfo;
  return true;
}

/*
  Modification history:

  $Log: vehMobPLEngine.cc,v $
  Revision 1.23  2007/12/13 13:18:41  dr_steveb
  Changes for vehicle level visibility graph planning.

  Revision 1.22  2007/11/08 13:27:51  dr_steveb
  Changed location of LineSegment

  Revision 1.21  2007/05/16 20:57:21  dr_steveb
  Fixed graph wrapping!

  Revision 1.20  2007/05/16 15:19:05  dr_steveb
  Trying to fix graph wrapping. Not fixed yet, however, fixed compile issue with rotate command.

  Revision 1.19  2007/05/02 17:22:54  tkharris
  Bugfix: pathout array length is always 1 less than pathLength (see printAns)

  Revision 1.18  2007/04/25 15:09:05  dr_steveb
  Added traffic simulation files

  Revision 1.17  2007/03/19 15:11:23  dr_steveb
  Changed the way that the size of the graphHeight and graphWidth are being set. Fixed calling location of buildGraph.

  Revision 1.16  2007/02/27 19:12:19  dr_steveb
  Fixed drawing of no-planning region for vehicle and added mobility world

  Revision 1.15  2007/02/27 14:46:27  dr_steveb
  New interface between AM and VEH

  Revision 1.14  2007/01/30 21:29:02  dr_steveb
  Added new AM-VEH interface

  Revision 1.13  2006/12/20 15:24:03  proctor
  Added stdio.h, and double casting, to fix some gcc 3 compiler complaints.

  Revision 1.12  2006/10/31 12:49:48  dr_steveb
  Fixes to interface to SPQR

  Revision 1.11  2006/10/31 09:30:43  dr_steveb
  Changes for SPQR integration

  Revision 1.10  2006/10/24 13:09:49  dr_steveb
  New vehicle level victim locator and output on moastError. MoastError is now read by the notifier program. Use notifier -bmoastError

  Revision 1.9  2006/10/18 00:18:21  dr_steveb
  Added victim channel to vehicle level. It now hunts for victims. Also added CImg library from sourceforge to perform annotation. Modified make files to include this library.

  Revision 1.8  2006/10/10 13:30:19  dr_steveb
  New vehicle level exploration based on camera FOV implemented.

  Revision 1.7  2006/10/03 20:24:25  dr_steveb
  Exploration command added to vehicle level

  Revision 1.6  2006/09/30 16:04:22  dr_steveb
  Changed to use Tom's new world model and added am error checking.

  Revision 1.5  2006/09/26 14:54:00  tomrkramer
  Added #include "amSP.hh" so AMSP_OBSTACLE, etc. would be defined.

  Revision 1.4  2005/10/21 13:02:45  dr_steveb
  Various fixes.

  Revision 1.3  2005/10/17 18:35:13  dr_steveb
  Fixed map growing, added planLength to status message. Added getPlanLength to vehMobPLEngine. Discovered problem with use of ppmsdisplay

  Revision 1.2  2005/10/16 15:31:02  dr_steveb
  Various changes that will have to be redone for the new distribution.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.12  2005/10/11 15:26:34  stephen
  Added plan cost to status message and status printing. This required a new class variable (planCost).

  Revision 1.11  2005/10/03 15:06:30  stephen
  Vehicle level now works with either prim or am by setting a compile flag (USEPRIMMOB)

  Revision 1.10  2005/09/30 19:26:21  proctor
  Fixed warnings with %lf used with printf, it's for scanf only

  Revision 1.9  2005/09/30 18:55:58  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.8  2005/09/29 13:18:09  stephen
  Beginning of changes to work with Tom's new prim and am sp.

  Revision 1.7  2005/09/21 00:46:06  stephen
  Modifications to work with Adam's am and prim.

  Revision 1.6  2005/07/18 19:56:11  stephen
  Changed interface to confrom to dijkHeapN

  Revision 1.5  2005/06/29 21:08:17  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.4  2005/06/14 19:19:13  stephen
  Changed to use SensorPoint in sensor data message.

  Revision 1.3  2005/06/09 16:43:46  stephen
  Fixed initialization problem.

  Revision 1.2  2005/06/09 12:32:58  stephen
  Removal of sqrtf and start of turning cfg channel into state machine.

  Revision 1.1  2005/06/08 14:31:15  proctor
  Added MobJA name changes

  Revision 1.7  2005/06/06 12:45:06  stephen
  Made true state machine.

  Revision 1.6  2005/05/31 21:19:32  proctor
  Changed tabsGraph, searcher to DijkTabsGraph, DijkTabsSearcher, etc. to
  reflect Tom's changes to the naming convention.

  Revision 1.5  2005/05/28 02:03:56  proctor
  Made declaration of getMoreKids virtual, and took out default implementation;
  now vehPLEngine.cc has one implementation, and exampleDijkTabs.cc has another.

  Revision 1.4  2005/05/27 23:22:48  proctor
  Fixed floorf nonportability

  Revision 1.3  2005/05/27 21:07:48  proctor
  Took out getMoreKids() since it's not called.

  Revision 1.2  2005/05/24 17:05:06  stephen
  Changed to Tom's new Dijk algorithm.

  Revision 1.1  2005/05/23 18:03:30  stephen
  Added vehicle level planner.
*/
