/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file vehShell.cc

  \brief Command line shell to vehicle level NML buffers.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.14 $
  $Date: 2007/12/20 02:32:46 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "vehMobPL.hh"		// vehMobPL_format()

static RCS_CMD_CHANNEL *vehMobPLCmdBuf = NULL;
static RCS_STAT_CHANNEL *vehMobPLStatBuf = NULL;
static RCS_CMD_CHANNEL *vehMobPLCfgBuf = NULL;
static RCS_STAT_CHANNEL *vehMobPLSetBuf = NULL;
int statType; // the type of status read on the status channel

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "vehShell"

static char* rcsVehStateToString(int s) {
  static char str[40];
  switch (s) {
  case S0: return "S0";
  case S1: return "S1";
  case S2: return "S2";
  case S3: return "S3";
  case S4: return "S4";
  case S5: return "S5";
  case S6: return "S6";
  case S_CHECK_AM_ERR: return "S_CHECK_AM_ERR";
  case S_CHECK_AM_STAT_FROM_PREPARE: return "S_CHECK_AM_STAT_FROM_PREPARE";
  case S_CHECK_AM_STAT_FROM_SPIN: return "S_CHECK_AM_STAT_FROM_SPIN";
  case S_CHECK_STUCK: return "S_CHECK_STUCK";
  case S_PLAN_ROUTE: return "S_PLAN_ROUTE";
  case S_PREPARE_MOVE: return "S_PREPARE_MOVE";
  case S_START_OBS: return "S_START_OBS";
  case S_STUCK: return "S_STUCK";
  case S_MOVEMENT_ERR: return "S_MOVEMENT_ERR";
  case S_DONE_MOVEMENT: return "S_DONE_MOVEMENT";
  case S_BLOCK_CENTER: return "S_BLOCK_CENTER";
  case S_ADD_AM_NODES: return "S_ADD_AM_NODES";
  case S_ADD_VEH_GOAL: return "S_ADD_VEH_GOAL";
  case S_CLEAR_GRAPH: return "S_CLEAR_GRAPH";
  case S_SEND_PATH: return "S_SEND_PATH";
  default:
    sprintf(str, "S%d?", s);
  }
  return str;
}

static int cleanupVehMobPLNmlBuffers(void)
{
  if (NULL != vehMobPLCmdBuf)
    delete vehMobPLCmdBuf;
  if (NULL != vehMobPLStatBuf)
    delete vehMobPLStatBuf;
  if (NULL != vehMobPLSetBuf)
    delete vehMobPLSetBuf;
  if (NULL != vehMobPLCfgBuf)
    delete vehMobPLCfgBuf;

  vehMobPLStatBuf = NULL;
  vehMobPLSetBuf = NULL;
  vehMobPLCmdBuf = NULL;
  vehMobPLCfgBuf = NULL;

  return 0;
}

static int initVehMobPLNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupVehMobPLNmlBuffers();

  sprintf(chanName, "%s%d", VEH_MOB_PL_CMD_NAME, bufnum);
  vehMobPLCmdBuf =
    new RCS_CMD_CHANNEL(vehMobPL_format, chanName,
			THIS_PROCESS, config_file);
  if (!vehMobPLCmdBuf->valid()) {
    delete vehMobPLCmdBuf;
    vehMobPLCmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", VEH_MOB_PL_STAT_NAME, bufnum);
  vehMobPLStatBuf =
    new RCS_STAT_CHANNEL(vehMobPL_format, chanName,
			 THIS_PROCESS, config_file);
  if (!vehMobPLStatBuf->valid()) {
    delete vehMobPLStatBuf;
    vehMobPLStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", VEH_MOB_PL_CFG_NAME, bufnum);
  vehMobPLCfgBuf =
    new RCS_CMD_CHANNEL(vehMobPL_format, chanName,
			THIS_PROCESS, config_file);
  if (!vehMobPLCfgBuf->valid()) {
    delete vehMobPLCfgBuf;
    vehMobPLCfgBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", VEH_MOB_PL_SET_NAME, bufnum);
  vehMobPLSetBuf =
    new RCS_STAT_CHANNEL(vehMobPL_format, chanName,
			 THIS_PROCESS, config_file);
  if (!vehMobPLSetBuf->valid()) {
    delete vehMobPLSetBuf;
    vehMobPLSetBuf = 0;
    return 1;
  }

  return 0;
}

#define rcs_stat_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

static void printStat(void)
{
  VehMobPLStat * stat = (VehMobPLStat *) vehMobPLStatBuf->get_address();
  VehMobPLSet * set = (VehMobPLSet *) vehMobPLSetBuf->get_address();

  printf("command_type:       %s\n",
	 vehMobPL_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %s\n", rcsVehStateToString(stat->state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("heartbeat:          %d\n", stat->tt.count);
  printf("set cycleTime:      %f\n", (double) set->cycleTime);
  printf("actual cycleTime:   %lf\n",stat->tt.avg);
  printf("plan cost:          %lf\n",stat->planCost);
  printf("start horizon:      %lf\n",stat->planStartHorizon);
  printf("debug:              %X\n", set->debug);
  printf("message:            %s\n", stat->message);

  if( statType == VEH_MANUFACT_STAT_TYPE )
    {
      printf("veh destination:    %s\n", ((VehManufactStat*)stat)->vehDestination);
      printf("pkg destination:    %s\n", ((VehManufactStat*)stat)->packageDestination);
      printf("has pkg:            %d\n", ((VehManufactStat*)stat)->hasPackage);
    }
}

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}

static void printHelp(void)
{
  printf("q | quit          -- quit\n");
  printf("? | help          -- print this help\n");
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
  printf("ct <secs>         -- set cycle time to <secs>\n");
  printf("go <name>         -- go to a named point\n");
  printf("mvl <x> <y>       -- move to <x> <y> in local coordinates\n");
  printf("mvg <x> <y>       -- move to <x> <y> in global coordinates\n");
  printf("exp <x> <y>       -- explore while moving to local waypoint\n");
  printf("dump              -- dump world model and display with imageview\n");
  printf("load <file>       -- load geoTiff world model file <file>\n");
  printf("debug <0, 1, ...> -- configure debug level\n");
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char *line;
  char *ptr;
  char fileType[100];
  int retval;
  VehMobPLStat *vehMobPLStatPtr;
  VehMobPLSet *vehMobPLSetPtr;
  VehMobPLCmdInit vehMobPLCmdInit;
  VehMobPLCmdGoNamedPt vehMobPLCmdGoNamedPt;
  VehMobPLCmdHalt vehMobPLCmdHalt;
  VehMobPLCmdAbort vehMobPLCmdAbort;
  VehMobPLCmdShutdown vehMobPLCmdShutdown;
  VehMobPLCmdNop vehMobPLCmdNop;
  VehMobPLCmdExplore vehMobPLCmdExplore;
  VehMobPLCmdLoadWM vehMobPLCmdLoadWM;
  VehMobPLCfgCycleTime vehMobPLCfgCycleTime;
  VehMobPLCfgDebug vehMobPLCfgDebug;
  VehMobPLCfgDumpWM vehMobPLCfgDump;

  VehMobPLCmdMoveWaypoint vehMobPLCmdMoveWaypoint;

  bool test = false;
#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  while (1) {
    char option = getopt(argc, argv, "n:i:d:t");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      break;
    case 'i':
      INI_FILE = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    case 't':
      test = true;
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // initial allocate all the buffers
  if( test )
    retval = 0;
  else
    retval = initVehMobPLNmlBuffers(NML_FILE, bufnum);

  if (retval) {
    fprintf(stderr, "%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }
  // set up our convenient pointer to status and settings
  if( !test )
    {
      vehMobPLStatPtr = (VehMobPLStat *) vehMobPLStatBuf->get_address();
      vehMobPLSetPtr = (VehMobPLSet *) vehMobPLSetBuf->get_address();
    }
  else
    {
      vehMobPLStatPtr = new VehMobPLStat();
      vehMobPLSetPtr = new VehMobPLSet();
    }

  while (!feof(stdin)) {
#ifndef HAVE_READLINE_READLINE_H
    printf("> ");
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) {
      break;
    }
    line = buffer;
#else
    line = readline("> ");
    if (NULL == line) {
      break;
    }
    if (*line) {
      add_history(line);
    }
#endif
    if( !test )
      {
	statType = vehMobPLStatBuf->read();
	vehMobPLSetBuf->read();
      }
    ptr = line;
    while (isspace(*ptr))
      ptr++;			// skip leading white space
    if (0 == *ptr) {		// blank line
      printStat();
    } else if (ptr[0] == 'q' || !strcmp(ptr, "quit")) {
      break;
    } else if (ptr[0] == '?' || !strcmp(ptr, "help")) {
      printHelp();
    } else if (!strwcmp(ptr, "init")) {
      vehMobPLCmdInit.serial_number =
	vehMobPLStatPtr->echo_serial_number + 1;
      if( !test )
	vehMobPLCmdBuf->write(&vehMobPLCmdInit);
    } else if (!strwcmp(ptr, "nop")) {
      vehMobPLCmdNop.serial_number = 
	vehMobPLStatPtr->echo_serial_number + 1;
      if( !test )
	vehMobPLCmdBuf->write(&vehMobPLCmdNop);
    } else if (!strwcmp(ptr, "halt")) {
      vehMobPLCmdHalt.serial_number =
	vehMobPLStatPtr->echo_serial_number + 1;
      if( !test )
	vehMobPLCmdBuf->write(&vehMobPLCmdHalt);
    } else if (!strwcmp(ptr, "abort")) {
      vehMobPLCmdAbort.serial_number =
	vehMobPLStatPtr->echo_serial_number + 1;
      if( !test )
	vehMobPLCmdBuf->write(&vehMobPLCmdAbort);
    } else if (!strwcmp(ptr, "shutdown")) {
      vehMobPLCmdShutdown.serial_number =
	vehMobPLStatPtr->echo_serial_number + 1;
      if( !test )
	vehMobPLCmdBuf->write(&vehMobPLCmdShutdown);
    } else if (!strwcmp(ptr, "ct")) {
      if (1 !=
	  sscanf(ptr, "%*s %lf",
		 &vehMobPLCfgCycleTime.cycleTime)) {
	printf("need a cycleTime argument\n");
      } else {
	vehMobPLCfgCycleTime.serial_number =
	  vehMobPLSetPtr->echo_serial_number + 1;
      if( !test )
	vehMobPLCfgBuf->
	  write(&vehMobPLCfgCycleTime);
      }
    } else if (!strwcmp(ptr, "debug")) {
      if (1 != sscanf(ptr, "%*s %i",
		      &vehMobPLCfgDebug.debug)) {
	printf("need debug mask\n");
      } else {
	vehMobPLCfgDebug.serial_number =
	  vehMobPLSetPtr->echo_serial_number + 1;
	if( !test )
	  vehMobPLCfgBuf->write(&vehMobPLCfgDebug);
     }
    } else if (!strwcmp(ptr, "mvl")) {
      if (2 != sscanf(ptr, "%*s %lf %lf",
		      &vehMobPLCmdMoveWaypoint.p.x,
		      &vehMobPLCmdMoveWaypoint.p.y)) {
	printf("need location to move to\n");
      } else {
	vehMobPLCmdMoveWaypoint.serial_number =
	  vehMobPLStatPtr->echo_serial_number + 1;
	vehMobPLCmdMoveWaypoint.isGlobal = false;
	vehMobPLCmdMoveWaypoint.neighborhood = 1;
	vehMobPLCmdMoveWaypoint.startTime = 0;
	//	vehMobPLCmdMoveWaypoint.speed = 1;
	if( !test )
	  vehMobPLCmdBuf->write(&vehMobPLCmdMoveWaypoint);
      }
    } else if (!strwcmp(ptr, "mvg")) {
      if (2 != sscanf(ptr, "%*s %lf %lf",
		      &vehMobPLCmdMoveWaypoint.p.x,
		      &vehMobPLCmdMoveWaypoint.p.y)) {
	printf("need location to move to\n");
      } else {
	vehMobPLCmdMoveWaypoint.serial_number =
	  vehMobPLStatPtr->echo_serial_number + 1;
	vehMobPLCmdMoveWaypoint.isGlobal = true;
	vehMobPLCmdMoveWaypoint.neighborhood = 5;
	vehMobPLCmdMoveWaypoint.startTime = 0;
	if( !test )
	  vehMobPLCmdBuf->write(&vehMobPLCmdMoveWaypoint);
      }
    } else if (!strwcmp(ptr, "dump")) {
      if( 2 != sscanf(ptr, "%*s %d %s", &vehMobPLCfgDump.skip,
		      fileType) ){
	printf("need number of frames to skip between outputs and file type\n");
      } else {
	vehMobPLCfgDump.serial_number =
	  vehMobPLSetPtr->echo_serial_number + 1;
	vehMobPLCfgDump.dumpContinuous = true;
	sprintf(vehMobPLCfgDump.fileName, "VehMobPLDispOutput-%d.%s",
		bufnum, (!strwcmp("tif", fileType)? "tif": "ppm"));
	if( !test )
	  vehMobPLCfgBuf->write(&vehMobPLCfgDump);
      }
    } else if (!strwcmp(ptr, "load")) {
      if( 1 != sscanf(ptr, "%*s %s", vehMobPLCmdLoadWM.fileName) ){
	printf("need file name\n");
      } else {
	vehMobPLCmdLoadWM.serial_number =
	  vehMobPLStatPtr->echo_serial_number + 1;
	if( !test )
	  vehMobPLCmdBuf->write(&vehMobPLCmdLoadWM);
      }
    } else if (!strwcmp(ptr, "expl")) {
      if (2 != sscanf(ptr, "%*s %lf %lf",
		      &vehMobPLCmdExplore.p.x,
		      &vehMobPLCmdExplore.p.y)) {
	printf("need location to explore to\n");
      } else {
	vehMobPLCmdExplore.serial_number = 
	  vehMobPLStatPtr->echo_serial_number + 1;
	vehMobPLCmdExplore.isGlobal = false;
	vehMobPLCmdExplore.neighborhood = 1;
	vehMobPLCmdExplore.startTime = 0;
	if( !test )
	  vehMobPLCmdBuf->write(&vehMobPLCmdExplore);
      }
    } else if (!strwcmp(ptr, "expg")) {
      if (2 != sscanf(ptr, "%*s %lf %lf",
		      &vehMobPLCmdExplore.p.x,
		      &vehMobPLCmdExplore.p.y)) {
	printf("need location to explore to\n");
      } else {
	vehMobPLCmdExplore.serial_number = 
	  vehMobPLStatPtr->echo_serial_number + 1;
	vehMobPLCmdExplore.isGlobal = true;
	vehMobPLCmdExplore.neighborhood = 1;
	vehMobPLCmdExplore.startTime = 0;
	if( !test )
	  vehMobPLCmdBuf->write(&vehMobPLCmdExplore);
      }
    } else if (!strwcmp(ptr, "go")) {
      vehMobPLCmdGoNamedPt.serial_number = 
	vehMobPLStatPtr->echo_serial_number + 1;
      if( 2!=sscanf(ptr, "%*s %lf %[0-9a-zA-Z_ -]", &vehMobPLCmdGoNamedPt.neighborhood, fileType))
	printf( "Need neighborhood and location to go to\n");
      else if( strlen(fileType) > MAX_WAYPOINT_NAME_LENGTH )
	printf( "Name to be sent is too long!\n" );
      else
	{
	  printf( "Read destination: %s\n", fileType );
	  strcpy( vehMobPLCmdGoNamedPt.namedPt, fileType );
	  if( !test )
	    vehMobPLCmdBuf->write(&vehMobPLCmdGoNamedPt);
	}
    } else {
      printf("?\n");
    }

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  return cleanupVehMobPLNmlBuffers();
}

/*
  Modification history:

  $Log: vehShell.cc,v $
  Revision 1.14  2007/12/20 02:32:46  dr_steveb
  Now have two vehicles performing start of manufacturing job. More to come...

  Revision 1.13  2007/12/18 18:24:37  dr_steveb
  Added commands for loading and unloading packages

  Revision 1.12  2007/12/13 13:18:41  dr_steveb
  Changes for vehicle level visibility graph planning.

  Revision 1.11  2007/11/08 13:27:51  dr_steveb
  Changed location of LineSegment

  Revision 1.10  2007/05/16 15:19:06  dr_steveb
  Trying to fix graph wrapping. Not fixed yet, however, fixed compile issue with rotate command.

  Revision 1.9  2007/05/02 17:02:48  tkharris
  Append robot number to veh dump so that vehicles don't overwrite each other's
  maps.

  Revision 1.8  2007/02/27 14:46:27  dr_steveb
  New interface between AM and VEH

  Revision 1.7  2007/01/30 21:29:02  dr_steveb
  Added new AM-VEH interface

  Revision 1.6  2006/10/03 20:24:25  dr_steveb
  Exploration command added to vehicle level

  Revision 1.5  2006/09/30 16:04:22  dr_steveb
  Changed to use Tom's new world model and added am error checking.

  Revision 1.4  2006/05/06 11:24:22  dr_steveb
  Adde nop command to vehShell and increased tolerance on waypoints in the planner.

  Revision 1.3  2006/04/12 13:02:16  dr_steveb
  Section level exploration now works for a single vehicle

  Revision 1.2  2006/03/22 14:47:38  dr_steveb
  Changed to use imageview instead of old display program.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.5  2005/10/11 15:26:35  stephen
  Added plan cost to status message and status printing. This required a new class variable (planCost).

  Revision 1.4  2005/10/03 15:06:31  stephen
  Vehicle level now works with either prim or am by setting a compile flag (USEPRIMMOB)

  Revision 1.3  2005/09/29 21:45:32  stephen
  Added dump command and made some fixes.

  Revision 1.2  2005/09/21 00:46:06  stephen
  Modifications to work with Adam's am and prim.

  Revision 1.1  2005/07/01 16:12:20  proctor
  Initial checkin

*/

