/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehVisGraph.cc
  \brief A visibility graph planning module for MOAST

  This module implements a visibility based planning system. It takes a lot of
  short cuts to get a basic system up and running.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.10 $
  $Date: 2008/03/11 13:17:27 $
  \endcode

  \author Stephen Balakirsky
  \date   11/14/2007
*/

#include <vector>
#include <stack>
#include <stddef.h>		// sizeof(), NULL
#include <string.h>		// strcpy()
#include <stdlib.h>		// getenv()
#include <signal.h>		// signal(), SIGINT
#include <unistd.h>		// sleep()
#include "rcs.hh"
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "amMobJA.hh"
#include "amSP.hh"
#include "vehMobPL.hh"
#include "visGraph.hh"
#include "moastGeom.hh"
#include "acmRandom.h"

using namespace std;
////////////////////////////
// globals
/////////////////////
//! my verbose level
int myVerbose;
//! the visibility graph
VisGraph *visGraph;
//! Serial number to am level
int myAmSerial;

// NML Channels
//! am planning channels
static RCS_CMD_CHANNEL  *amMobJACmdNml = NULL;
static RCS_STAT_CHANNEL *amMobJAStatNml = NULL;
static RCS_STAT_CHANNEL *amMobJASetNml = NULL;
static RCS_CMD_CHANNEL  *amMobJACfgNml = NULL;
//! am SP channels
static RCS_CMD_CHANNEL  *amSPCmdNml = NULL;
static RCS_STAT_CHANNEL *amSPStatNml = NULL;
static RCS_STAT_CHANNEL *amSPSetNml = NULL;
static RCS_CMD_CHANNEL  *amSPCfgNml = NULL;
//! veh planning channels
static RCS_CMD_CHANNEL  *vehMobPLCmdNml = NULL;
static RCS_STAT_CHANNEL *vehMobPLStatNml = NULL;
static RCS_CMD_CHANNEL  *vehMobPLCfgNml = NULL;
static RCS_STAT_CHANNEL *vehMobPLSetNml = NULL;
static NML              *vehMapOutNml = NULL;

static AmMobJACmdMoveWaypoint *moveWaypointCmd;
static AmMobJAStat *amStatus;
static AmMobJASet amMobJASettings;

//! Should the map be output
bool drawMap;
//! How many planning cycles should be skipped between output cycles
int drawSkip;
//! goal stack
stack<goalPoint> goalStack;

RCS_TIMER *timer; // cycle timer

#define MY_NAME_BASE "vehMobPL" // the name of my process (this one)

//////////////////////////////////////////////////////////////////////////////////////////////////
//! check the command status of the am mobility subsystem
bool checkAmMobCmdStatus( AmMobJAStat **stat, int &echo_serial_number, RCS_STAT_CHANNEL *channel )
{
  int type;

  type = channel->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case AM_MOB_JA_STAT_TYPE:
      *stat = (AmMobJAStat*)channel->get_address();
      echo_serial_number = channel->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Check the settings back from subordinate planning system
  
  \param settings The settings that were returned

  \return true on good read
  \return false on no new settings
*/
bool checkAmMobSettings( AmMobJASet &amSettings, RCS_STAT_CHANNEL *channel )
{
  int type;

  type = channel->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case AM_MOB_JA_SET_TYPE:
      amSettings = *(AmMobJASet*)channel->get_address();
      return(true);
    default:
      break;
    }
  return false;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
//! check the status of the am sensor processing subsystem
bool checkAmSPCmdStatus( AmSPStat **stat, int &echo_serial_number, RCS_STAT_CHANNEL *channel )
{
  int type;

  type = channel->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case AM_SP_STAT_TYPE:
      *stat = (AmSPStat*)channel->get_address();
      echo_serial_number = channel->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//! Send the planned path to the subordinate
/*!
  This routine assumes that both the vehicle and am echelons are operating in the
  same coordinate frame. This works for the simple worlds that we are using since the
  offset between the global and local is <0,0>. In real-world applications, the 
  cordinate frame would need to be changed

  \param path       Point to go to 

  \return true on success
  \return false on failure
*/
#define MIN_SAFE_DISTANCE 2.0
bool sendPath( double horizon, bool thresholdPath, Location2D *lastPath, float neighborhoodIn=.2, float speed=5 ) // was .4 for real demo, 8 for virt
{
  Location2D pathPoint;
  static double lastNeighborhood = -1;
  double neighborhood;

  if( amMobJACmdNml==NULL )
    return false;

  if( !visGraph->getPathPoint( &pathPoint, 0 ) )
    return false;
  moveWaypointCmd->waypoint_length = 1;
  //  moveWaypointCmd->waypoint[0].p = path.makePM_CARTESIAN();
  /* there seems to be a problem with the x,y north,south mapping, so fix here
   */
  //  if( neighborhood == -1 ) // dynamic neighborhood
  //    {
      if( visGraph->getPathDistance() < 3. )
	neighborhood = neighborhoodIn;
      else
	neighborhood = 2;
      //      printf( "Path distance: %f\n", visGraph->getPathDistance() );
      //    }
  if( thresholdPath && lastPath->dist(pathPoint) < .1 && lastNeighborhood==neighborhood)
    return true;
  else
    {
      *lastPath = pathPoint;
      lastNeighborhood = neighborhood;
    }

  moveWaypointCmd->waypoint[0].p.x = pathPoint.east;
  moveWaypointCmd->waypoint[0].p.y = pathPoint.north;
  moveWaypointCmd->waypoint[0].neighborhood = neighborhood;
  //  moveWaypointCmd->waypoint[0].quality = confidence;
  moveWaypointCmd->waypoint[0].speed = speed;
  printf( "Sending AM to <%f %f> at speed: %f tol: %f\n",
	  moveWaypointCmd->waypoint[0].p.x,
	  moveWaypointCmd->waypoint[0].p.y,
	  moveWaypointCmd->waypoint[0].speed,
	  moveWaypointCmd->waypoint[0].neighborhood );
	  
  /*
  // slow down if too close to something ahead of us
  if( visGraph->getPathDistance()<MIN_SAFE_DISTANCE )
    {
      moveWaypointCmd->waypoint[0].speed *= (visGraph->getPathDistance()/MIN_SAFE_DISTANCE);
    }
  */
  moveWaypointCmd->serial_number = ++myAmSerial;
  amMobJACmdNml->write(moveWaypointCmd);
  //  printf( "Sending vehcile to <%lf %lf>\n", pathPoint.north, pathPoint.east );
  return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform abort command state machine

  \param stat status message
*/
void doCmdAbort(VehManufactStat *stat)
{
  AmMobJACmdAbort *abortMobCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
  AmSPCmdAbort *abortSPCmd;
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      abortMobCmd = new AmMobJACmdAbort();
      abortSPCmd = new AmSPCmdAbort;
      abortMobCmd->serial_number = ++myAmSerial;
      abortSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(abortMobCmd);
      amSPCmdNml->write(abortSPCmd);
      delete abortMobCmd;
      delete abortSPCmd;
    } // fall through to checking the status of the mobility
  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform  halt command state machine

  \param stat status message
  \return true need to update and send set message
  \return false no message needs to be sent
*/
bool doCmdHalt(VehManufactStat *stat)
{
  bool retValue = false;
  AmMobJACmdHalt *haltMobCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
  AmSPCmdHalt *haltSPCmd;
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      retValue = true;
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      haltMobCmd = new AmMobJACmdHalt();
      haltSPCmd = new AmSPCmdHalt();
      haltMobCmd->serial_number = ++myAmSerial;
      haltSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(haltMobCmd);
      amSPCmdNml->write(haltSPCmd);
      delete haltMobCmd;
      delete haltSPCmd;
    } // fall through to checking the satus
  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  retValue = true;
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    }
  return retValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform init state machine

  \param stat status message
  \return true when settings and status needs updating
  \return false no message needs to be sent
*/
bool doCmdInit(VehManufactStat *stat, VehMobPLSet *settings)
{
  bool retValue = false;
  AmMobJACmdInit *initMobCmd;
  AmSPCmdGo *goSPCmd;
  AmSPCmdInit *initSPCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      retValue = true;
      state_new(stat);
      if( settings->isInitialized == true )
	{
	  status_next(stat, RCS_DONE);
	  state_next(stat, S3);
	}
      else
	{
	  status_next(stat, RCS_EXEC);
	  state_next(stat, S0);
	}
      initMobCmd = new AmMobJACmdInit();
      initSPCmd = new AmSPCmdInit();
      initMobCmd->serial_number = ++myAmSerial;
      initSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(initMobCmd);
      amSPCmdNml->write(initSPCmd);
      delete initMobCmd;
      delete initSPCmd;
    } // fall through to checking the status 
  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		{
		  state_next(stat, S1 );
		  retValue = true; // update status since switching state
		}
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S6);
		  retValue = true;
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  retValue = true; // update status since switching state
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S6);
		}
	    }
	}
    } // fall through to sending "go" command
  if( state_match(stat, S2) )
    {
      goSPCmd = new AmSPCmdGo();
      goSPCmd->serial_number = ++myAmSerial;
      amSPCmdNml->write(goSPCmd);
      retValue = true;
      state_next(stat, S3);
      delete goSPCmd;
      retValue = true; // update status since switching state
    } // fall through to checking status
  if( state_match(stat, S3) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE || amSPStatus->status == RCS_EXEC )
		{
		  retValue = true;
		  status_next(stat, RCS_EXEC);
		  state_next(stat, S4 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  retValue = true;
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S6);
		}
	    }
	}
    }// fall througgh to initializing the visibility graph
  if( state_match(stat, S4) )
    {
      visGraph = new VisGraph( 20, 20 );
      retValue = true;
      settings->isInitialized = true;
      status_next(stat, RCS_DONE);
      state_next(stat, S6);
    }
  return retValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform shutdown command to subordinate planning system

  \param stat vehicle status
  \return true done with shutdown
  \return false not done yet
*/
bool doCmdShutdown(VehManufactStat *stat)
{
  bool retValue = false;
  AmMobJACmdShutdown *shutdownMobCmd;
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;
  AmSPCmdShutdown *shutdownSPCmd;
  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      retValue = true;
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);

      shutdownMobCmd = new AmMobJACmdShutdown();
      shutdownSPCmd = new AmSPCmdShutdown();
      shutdownMobCmd->serial_number = ++myAmSerial;
      shutdownSPCmd->serial_number = myAmSerial;
      amMobJACmdNml->write(shutdownMobCmd);
      amSPCmdNml->write(shutdownSPCmd);
      delete shutdownMobCmd;
      delete shutdownSPCmd;
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  retValue = true;
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  retValue = true;
		  state_next(stat, S2);
		}
	    }
	}
    }
  return retValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*! 
  Perform movement state machine

  \param stat vehicle status
*/
// see vehMobPL.hh for state names and values
void doCmdMovement(VehManufactStat *stat, VehMobPLSet *settings)
{
  goalPoint goalToSend;
  PM_CARTESIAN obsPt;
  PM_RPY vehRpy;
  int searchReturn;
  int echo_serial_number;
  AmMobJACmdSpin amSpinCmd;
  AmMobJACmdPrepareMove amPrepareMoveCmd;
  double endTime = etime() + settings->cycleTime; // time the cycle must end
  Location2D loc2d;
  static bool newSearch = true;
  static bool lastPathValid = false;
  static Location2D lastPath;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S_PREPARE_MOVE);
    }

  if( state_match(stat, S_PREPARE_MOVE) ) // prepare am for movement
    {
      amPrepareMoveCmd.serial_number = ++myAmSerial;
      lastPathValid = false;
      amMobJACmdNml->write(amPrepareMoveCmd);
      state_next(stat, S_CHECK_AM_STAT_FROM_PREPARE);
      if( etime() > endTime )
	return;
    } // fall through to checking status
  if( state_match(stat, S_CHECK_AM_STAT_FROM_PREPARE) ) // check AM status
    {
      if( checkAmMobCmdStatus( &amStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amStatus->status == RCS_ERROR )
		{
		  if( amStatus->amMobJAError == AM_MOB_JA_ERROR_PLAN_HORIZON )
		    {
		      state_next(stat, S_PREPARE_MOVE);
		      sprintf( stat->message, "AM Err: AM_MOB_JA_ERROR_PLAN_HORIZON" );
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_POSITION )
		    {
		      state_next(stat, S_PREPARE_MOVE);
		      sprintf( stat->message, "AM Err: AM_MOB_JA_ERROR_POSITION" );
		    }
		  else
		    {
		      status_next(stat, RCS_ERROR);
		      sprintf( stat->message, "AM Err reported: %d", amStatus->amMobJAError );
		      state_next(stat, S_MOVEMENT_ERR);
		    }
		  fprintf( stderr, "%s\n", stat->message );
		}
	      else if( amStatus->status == RCS_DONE )
		{
		  state_next(stat, S_ADD_VEH_GOAL);
		}
	    }
	}
      if( etime() > endTime )
	return;
    }  // fall through 
  if( state_match(stat, S_ADD_VEH_GOAL) )
    {
      // find goal
      if( goalStack.size() == 0 ) // all done
	{
	  status_next(stat, RCS_DONE);
	  state_next(stat, S_DONE_MOVEMENT);
	  strcpy( stat->vehDestination, "None" );
	  return;
	}
      goalToSend = goalStack.top();
      loc2d = goalToSend.location;
      visGraph->setGoal( loc2d );
      state_next(stat, S_ADD_AM_NODES);
      newSearch = true;
      if( etime() > endTime )
	return;
    }
  if( state_match(stat, S_ADD_AM_NODES) )
    {
      /*************************************************************
       There seems to be a problem with what AM and Veh see as north, so fix it here!
       *************************************************************/
      // last seed is vehicle location
      loc2d = amStatus->seed[amStatus->seed_length-1];
      loc2d.north = amStatus->seed[amStatus->seed_length-1].y;
      loc2d.east = amStatus->seed[amStatus->seed_length-1].x;
      visGraph->setVehLoc(loc2d);
      if( amStatus->seed_length >= 1 ) // note that last seed is vehicle location
	{
	if( visGraph->addGraphPoints(amStatus->seed, amStatus->seed_length, endTime))
	  state_next(stat, S_PLAN_ROUTE);
	else
	  {
	    sprintf( stat->message, "Error from adding graph points" );
	    fprintf( stderr, "Veh Error: %s\n", stat->message );
	    state_next(stat, S_PREPARE_MOVE);
	  }
	}
      else 
	fprintf( stderr, "vehPlan: no seeds from AM Echelon\n" );
      if( etime() > endTime )
	return;
    } // fall thorugh

  if( state_match(stat, S_PLAN_ROUTE) )
    {
      searchReturn = visGraph->search( newSearch, endTime ); 
      newSearch = false;
      if( searchReturn==1 || searchReturn==2 )
	state_next( stat, S_SEND_PATH );
      else if( searchReturn==0 ) // not done
	state_next(stat, S_PLAN_ROUTE );
      else
	{
	  status_next(stat, RCS_ERROR);
	  sprintf( stat->message, "VehErr: search graph returned error %d", searchReturn );
	  fprintf( stderr, "%s\n", stat->message );
	  fprintf( stderr, "Dumping vehicle graph to vehDump.ppm\n" );
	  visGraph->dump( "vehDump.ppm",0xF );
	  state_next(stat, S_MOVEMENT_ERR);
	  //	  sendAbort();
	}
      //	}
      if( etime() > endTime )
	return;
    } 
  if( state_match(stat, S_SEND_PATH) )
    {
      newSearch = true;
      /*
	stat->planCost = engine->getCost();
	stat->planLength = engine->getPathLength();
      */
      goalToSend = goalStack.top();
      sendPath( stat->planStartHorizon, lastPathValid, &lastPath, goalToSend.neighborhood );
      lastPathValid = true;
      state_next(stat, S_CHECK_AM_ERR);
    }
  if( state_match(stat, S_CHECK_AM_ERR) )
    {
      stat->planStartHorizon = amMobJASettings.planHorizon; // reset the start horizon
      visGraph->setPlanHorizon(amMobJASettings.planHorizon);
      //      printf( "Setting plan horizon to %f\n", amMobJASettings.planHorizon );
      if( checkAmMobCmdStatus( &amStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  sprintf(stat->message, "No reports" );
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amStatus->status == RCS_EXEC && amStatus->amMobJAError == AM_MOB_JA_ERROR_NONE )
		state_next(stat, S_ADD_AM_NODES);
	      if( amStatus->status == RCS_ERROR )
		{
		  state_next(stat, S_PREPARE_MOVE);
		  if( amStatus->amMobJAError == AM_MOB_JA_ERROR_WAYPOINT )
		    {
		      if(amStatus->seed_length == 0)
			{
			  sprintf(stat->message, "too many waypoints");
			}
		      else if(amStatus->seed_length == 1)
			{
			  sprintf(stat->message, "bad waypoint location (%.2lf, %.2lf)",
				 amStatus->seed[0].x, amStatus->seed[0].y);
			  //print out problem path
			  /*
			  for( int jj=0; jj<engine->getPathLength(); jj++ )
			    {
			      printf( "%d %lf %lf\n", jj, 
				      engine->pathOut[jj].x,
				      engine->pathOut[jj].y );
			    }
			  */
			}
		      else if(amStatus->seed_length == 3)
			{
			  sprintf(stat->message, "waypoints (%.2lf, %.2lf) and (%.2lf, %.2lf) too "
				  "close", amStatus->seed[0].x, amStatus->seed[0].y,
				  amStatus->seed[1].x, amStatus->seed[1].y);
			}
		      else
			sprintf(stat->message, "unspecified am waypoint problem");
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_MOVE )
		    {
		      sprintf(stat->message, "am reports: AM_MOB_JA_ERROR_MOVE, trying to prepare again..." );
		      state_next( stat, S_PREPARE_MOVE );
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_PLAN_PRIM )
		    {
		      sprintf(stat->message, "am reports: AM_MOB_JA_ERROR_PRIM, trying to replan..." );
		      state_next( stat, S_PREPARE_MOVE );
		    }
		  else if( amStatus->amMobJAError == AM_MOB_JA_ERROR_POSITION )
		    {
		      sprintf(stat->message, "am reports: AM_MOB_JA_ERROR_POSITION, trying to prepare again..." );
		      state_next( stat, S_PREPARE_MOVE );
		    }
		  else
		    {
		      sprintf(stat->message, "Unspecified am error: %d", amStatus->amMobJAError );
		    }
		  fprintf( stderr, "%s\n", stat->message );
		}
	      else if( amStatus->status == RCS_DONE )
		{
		  goalToSend = goalStack.top();
		  loc2d = goalToSend.location;
		  if( loc2d.dist(visGraph->getVehLoc()) < goalToSend.neighborhood ) // really done!
		    {
		      goalStack.pop();
		      state_next( stat, S_ADD_VEH_GOAL );
		    }
		  else
		    { // not really done
		      sprintf(stat->message, "am reports done when not done" );
		      fprintf( stderr, "Veh: %s\n", stat->message );
		      state_next( stat, S_PREPARE_MOVE );
		    }
		}
	    }
	}
    }
      /*
  if( state_match(stat, S_STUCK) ) // we have been sitting, so start observations
    {
      // first clear out any obstacles placed due to the amPlanner
      engine->clearWorld(stat->heartbeat);
      vehRpy = vehLocation.rot;
      observeAngle = vehRpy.y - RAD(10);
      if( observeAngle < 0 )
	observeAngle += 2*M_PI;
      printf( "Stuck with count %d, so observing %f (%f degrees)\n", ++stuckCount, 
	      observeAngle, DEG(observeAngle) );
      amSpinCmd.serial_number = ++myAmSerial;
      amSpinCmd.direction = MOAST_ROTATE_CW;
      amSpinCmd.absAngle = observeAngle;
      amSpinCmd.tolerance = RAD(5);
      amMobJACmdNml->write(amSpinCmd);
      state_next(stat, S_CHECK_AM_STAT_FROM_SPIN);
      return;
    }
  */
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*! 
  Perform dock command. This command does not really do anything yet.
*/
void doCmdDock(VehManufactStat *stat)
{
  status_next(stat, RCS_DONE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*! 
  Perform undock command. This command does not really do anything yet.
*/
void doCmdUndock(VehManufactStat *stat)
{
  status_next(stat, RCS_DONE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*! 
  Perform unload command. This command does not really do anything yet.
*/
void doCmdLoadPkg(VehManufactStat *stat)
{
  static AcmRandomStruct *acm = NULL;
  long int whereTo;

  if( acm == NULL )
    {
      acm = new AcmRandomStruct();
      acmRandomInit( acm );
    }
  whereTo = acmRandomRange( acm, 3 );
  switch( whereTo )
    {
    case 0:
      strcpy( stat->packageDestination, "Drop 1" );
      break;
    case 1:
      strcpy( stat->packageDestination, "Drop 2" );
      break;
    case 2:
      strcpy( stat->packageDestination, "Drop 3" );
      break;
    }
  stat->hasPackage = true;
  status_next(stat, RCS_DONE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*! 
  Perform load command. This command sets the package destination randomly
*/
void doCmdUnloadPkg(VehManufactStat *stat)
{
  stat->hasPackage = false;
  strcpy( stat->packageDestination, "None" );
  status_next(stat, RCS_DONE);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform noop command to subordinate planning system

  \param stat vehicle status
*/
void doCmdNop(VehManufactStat *stat)
{
  AmMobJAStat *amMobStatus;
  AmSPStat *amSPStatus;

  int echo_serial_number;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_DONE);
      state_next(stat, S2);
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      if( checkAmMobCmdStatus( &amMobStatus, echo_serial_number, amMobJAStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amMobStatus->status == RCS_DONE )
		state_next(stat, S1 );
	      else if( amMobStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    } // fall through to checking the status of the sp
  if( state_match(stat, S1) )
    {
      if( checkAmSPCmdStatus( &amSPStatus, echo_serial_number, amSPStatNml ) )
	{
	  if( echo_serial_number == myAmSerial )
	    {
	      if( amSPStatus->status == RCS_DONE )
		{
		  status_next(stat, RCS_DONE);
		  state_next(stat, S2 );
		}
	      else if( amSPStatus->status == RCS_ERROR )
		{
		  status_next(stat, RCS_ERROR);
		  state_next(stat, S2);
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform nop

  \param stat    vehicle status

  \return        true if set channel needs to be updated
*/
bool doCfgNop(VehMobPLSet *stat)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      retValue = true;
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform load wm state machine

  \param stat    vehicle status

  \return        true if set channel needs to be updated
*/
bool doCmdLoadWM(VehManufactStat *stat, VehMobPLSet *settings)
{
  bool retValue = false;
  double endTime = etime() + settings->cycleTime; // time the cycle must end

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      if( visGraph->load(settings->wmFileName, 1.) <= 0 ) // was .8
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S6);
	}
      else
	state_next(stat, S0);
      retValue = true;
    } // fall through to allocating output image
  if( state_match(stat, S0) )
    {
      visGraph->setOutImage(.1); // set to 10cm resolution
      state_next(stat, S1);
    }
  if( state_match(stat, S1) )
    {
      if( visGraph->initGraph(endTime) )
	{
	  status_next(stat, RCS_DONE);
	  state_next(stat, S6);
	}
    }
  return retValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform store wm state machine

  \param stat    vehicle status

  \return        true if set channel needs to be updated
*/
bool doCfgStoreWM(VehMobPLSet *stat, VehMobPLSet *settings)
{
  bool retValue = false;
  FILE *fp;

  if( state_match(stat, NEW_COMMAND) ) // open the file
    {
      state_new(stat);
      fp = fopen( settings->wmFileName, "w" );
      if( fp == NULL )
	{
	  state_next(stat, S6);
	  status_next(stat, RCS_ERROR);
	}
      else
	state_next(stat, S0);
      retValue = true;
    } 

  if( state_match(stat, S0) ) // write out the information
    {
      visGraph->print(fp);
      fclose(fp);
      status_next(stat, RCS_DONE);
      state_next(stat, S6);
      retValue = true;
    }
  return retValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Perform dump wm state machine. Sets up the viewer program and
  dumps the world model to a file

  \param stat    vehicle status

  \return        true if set channel needs to be updated
*/
bool doCfgDumpWM(VehMobPLSet *settings)
{
  bool retValue = false;
  static int skip = 0;
  char sysStr[200];
  float mag;

  if( state_match(settings, NEW_COMMAND) )
    {
      state_new(settings);
      sprintf( sysStr, "rm -rf %s", settings->wmFileName );
      if( system(sysStr) == -1)
	{
	  status_next(settings, RCS_ERROR);
	  state_next(settings, S4);
	}
      else
	{
	  // set magnification to .5
	  mag = .5;
	  sprintf(sysStr,"./imageview -f%s -s%lf&\n", 
		  settings->wmFileName, mag );
	  if( system(sysStr) == -1 )
	    {
	      status_next(settings, RCS_ERROR);
	      state_next(settings, S4);
	    }
	  else
	    {
	      skip = settings->fileSkip;
	      status_next(settings, RCS_EXEC);
	      state_next(settings, S0);
	    }
	}
	  retValue = true;
    } // if no error, fall through to seeing if file should be written

  if( state_match(settings, S0) )
    {
      if( skip++ >= settings->fileSkip )
	{
	  status_next(settings, RCS_EXEC);
	  state_next(settings, S1);
	  retValue = true;
	}
    } // possibly fall through to writing file
  if( state_match(settings, S1) )
    {
      //      visGraph->print();
      visGraph->dump( settings->wmFileName,0x6 );
      if( settings->fileSkip == -1 ) // done
	{
	  status_next(settings, RCS_DONE);
	  state_next(settings, S4);
	  retValue = true;
	}
      else
	{
	  skip = 0;
	  status_next(settings, RCS_EXEC);
	  state_next(settings, S0);
	  retValue = true;
	}
    }
  return retValue;
}

/*!
  Read in config and act on it

\param cmdIn         The handle of the channel for commands to be read from
\param stat          The status return

*/
RCS_STATUS readCfg( RCS_CMD_CHANNEL *cmdIn, VehMobPLSet *currentStat )
{
  VehMobPLCfgStoreWM      *storeWMCmd;
  VehMobPLCfgDumpWM       *dumpWMCmd;
  int type;
  int serial_number;

  type = cmdIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_CFG_NOP_TYPE:
    case VEH_MOB_PL_CFG_CYCLE_TIME_TYPE:
    case VEH_MOB_PL_CFG_DEBUG_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CFG_CLEAR_WM_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CFG_STORE_WM_TYPE:
      storeWMCmd = (VehMobPLCfgStoreWM*)cmdIn->get_address();
      serial_number = storeWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  strncpy( currentStat->wmFileName, storeWMCmd->fileName, MOAST_FILE_NAME_LEN );
	}
      break;
    case VEH_MOB_PL_CFG_DUMP_WM_TYPE:
      dumpWMCmd = (VehMobPLCfgDumpWM*)cmdIn->get_address();
      serial_number = dumpWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  if( !strcmp(dumpWMCmd->fileName, "uninitialized")  )
	    strncpy( currentStat->wmFileName, "VehMobPLDispOutput.png", MOAST_FILE_NAME_LEN );
	  else
	    strncpy( currentStat->wmFileName, dumpWMCmd->fileName, MOAST_FILE_NAME_LEN );
	  if( dumpWMCmd->dumpContinuous )
	    currentStat->fileSkip = dumpWMCmd->skip;
	  else
	    currentStat->fileSkip = -1;	    
	}
      break;
    default:
      fprintf( stdout, "Command %s not yet supported\n", vehMobPL_symbol_lookup(type) );
      currentStat->command_type = type;
      return RCS_ERROR;
    }
  return RCS_DONE;
}

/*!
  Read in command and act on it

\param cmdIn         The handle of the channel for commands to be read from
\\param currentStat   The current status of the module

*/
void readCmd( RCS_CMD_CHANNEL *cmdIn, VehManufactStat *currentStat, VehMobPLSet *currentSet )
{
  VehMobPLCmdMoveWaypoint *waypointCmd;
  VehMobPLCmdGoNamedPt *goNamedPtCmd;
  int type, serial_number;
  static int  noPrint = 0;
  goalPoint *newGoalPoint;
  VehMobPLCmdLoadWM *loadWMCmd;


  type = cmdIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_CMD_ABORT_TYPE:
    case VEH_MOB_PL_CMD_HALT_TYPE:
    case VEH_MOB_PL_CMD_INIT_TYPE:
    case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
    case VEH_MOB_PL_CMD_NOP_TYPE:
    case VEH_MOB_PL_CMD_DOCK_TYPE:
    case VEH_MOB_PL_CMD_UNDOCK_TYPE:
    case VEH_MOB_PL_CMD_LOAD_PKG_TYPE:
    case VEH_MOB_PL_CMD_UNLOAD_PKG_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
      // the planner supports multiple waypoints, but the command only
      // enters one
      waypointCmd = (VehMobPLCmdMoveWaypoint*)cmdIn->get_address();
      serial_number = waypointCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  while( !goalStack.empty() ) // empty out stack
	    {
	      goalStack.pop();
	    }
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  newGoalPoint = new goalPoint();
	  newGoalPoint->location = waypointCmd->p;
	  newGoalPoint->isGlobal = waypointCmd->isGlobal;
	  newGoalPoint->goalType = WAYPOINT;
	  newGoalPoint->neighborhood = waypointCmd->neighborhood;
	  goalStack.push(*newGoalPoint);
	  delete newGoalPoint;
	}
      break;
    case VEH_MOB_PL_CMD_GO_NAMED_PT_TYPE:
      // the planner supports multiple waypoints, but the command only
      // enters one
      goNamedPtCmd = (VehMobPLCmdGoNamedPt*)cmdIn->get_address();
      serial_number = goNamedPtCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  while( !goalStack.empty() ) // empty out stack
	    {
	      goalStack.pop();
	    }
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  newGoalPoint = new goalPoint();
	  newGoalPoint->location = visGraph->getNamedPt(goNamedPtCmd->namedPt).makePM_CARTESIAN();
	  strcpy( currentStat->vehDestination, goNamedPtCmd->namedPt );
	  newGoalPoint->isGlobal = true;
	  newGoalPoint->goalType = WAYPOINT;
	  newGoalPoint->neighborhood = goNamedPtCmd->neighborhood;
	  goalStack.push(*newGoalPoint);
	  delete newGoalPoint;
	}
      break;
    case VEH_MOB_PL_CMD_LOAD_WM_TYPE:
      loadWMCmd = (VehMobPLCmdLoadWM*)cmdIn->get_address();
      serial_number = loadWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  strncpy( currentSet->wmFileName, loadWMCmd->fileName, MOAST_FILE_NAME_LEN );
	}
      break;
    default:
      if( noPrint < 10 )
	{
	  fprintf( stdout, "Command %s not yet supported\n", vehMobPL_symbol_lookup(type) );
	  noPrint++;
	}
      break;
    }
  return;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  free all memory and prepare to exit
*/
int closeAll()
{
  if( amMobJACmdNml != NULL )
    delete amMobJACmdNml;
  amMobJACmdNml = NULL;
  if( amMobJAStatNml != NULL )
    delete amMobJAStatNml;
  amMobJAStatNml = NULL;
  if( amMobJASetNml != NULL )
    delete amMobJASetNml;
  amMobJASetNml = NULL;
  if( amMobJACfgNml != NULL )
    delete amMobJACfgNml;
  amMobJACfgNml = NULL;
  if( amSPCmdNml != NULL )
    delete amSPCmdNml;
  amSPCmdNml = NULL;
  if( amSPStatNml != NULL )
    delete amSPStatNml;
  amSPStatNml = NULL;
  if( amSPSetNml != NULL )
    delete amSPSetNml;
  amSPSetNml = NULL;
  if( amSPCfgNml != NULL )
    delete amSPCfgNml;
  amSPCfgNml = NULL;
  if( vehMobPLCmdNml != NULL )
    delete vehMobPLCmdNml;
  vehMobPLCmdNml = NULL;
  if( vehMobPLStatNml != NULL )
    delete vehMobPLStatNml;
  vehMobPLStatNml = NULL;
  if( vehMobPLSetNml != NULL )
    delete vehMobPLSetNml;
  vehMobPLSetNml = NULL;
  if( vehMobPLCfgNml != NULL )
    delete vehMobPLCfgNml;
  vehMobPLCfgNml = NULL;
  if( vehMapOutNml != NULL )
    delete vehMapOutNml;
  if( moveWaypointCmd != NULL )
    delete moveWaypointCmd;
  return 1;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  Open all of the NML channels
  
  \param vehID The number to add to all of the channel names
\param openFlags Tell what special channels to open. not currently used
  \return true on success
  \return false on failure
*/
bool openNML( int vehID, unsigned int openFlags )
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];
  VehMobPLCmdNop *nopCmd;
  VehMobPLCfgNop *nopCfg;

  sprintf( myName, "%s", MY_NAME_BASE );

  if(!configNml)
    {
      fprintf(stderr,"Please set CONFIG_NML environment variable.\n");
      return(false);
    }
  sprintf( chanName, "%s%d", AM_MOB_JA_CMD_NAME, vehID );
  amMobJACmdNml = new RCS_CMD_CHANNEL( amMobJA_format, chanName,
				      myName, configNml );
  if( NULL == amMobJACmdNml || !amMobJACmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_MOB_JA_STAT_NAME, vehID );
  amMobJAStatNml = new RCS_STAT_CHANNEL( amMobJA_format, chanName,
					myName, configNml );
  if( NULL == amMobJAStatNml || !amMobJAStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_MOB_JA_SET_NAME, vehID );
  amMobJASetNml = new RCS_STAT_CHANNEL( amMobJA_format, chanName,
				      myName, configNml );
  if( NULL == amMobJASetNml || !amMobJASetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_MOB_JA_CFG_NAME, vehID );
  amMobJACfgNml = new RCS_CMD_CHANNEL( amMobJA_format, chanName,
				       myName, configNml );
  if( NULL == amMobJACfgNml || !amMobJACfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_CMD_NAME, vehID );
  amSPCmdNml = new RCS_CMD_CHANNEL( amSP_format, chanName,
				       myName, configNml );
  if( NULL == amSPCmdNml || !amSPCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_STAT_NAME, vehID );
  amSPStatNml = new RCS_STAT_CHANNEL( amSP_format, chanName,
					 myName, configNml );
  if( NULL == amSPStatNml || !amSPStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_SET_NAME, vehID );
  amSPSetNml = new RCS_STAT_CHANNEL( amSP_format, chanName,
				       myName, configNml );
  if( NULL == amSPSetNml || !amSPSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_CFG_NAME, vehID );
  amSPCfgNml = new RCS_CMD_CHANNEL( amSP_format, chanName,
					myName, configNml );
  if( NULL == amSPCfgNml || !amSPCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CMD_NAME, vehID );
  vehMobPLCmdNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
				      myName, configNml );
  if( NULL == vehMobPLCmdNml || !vehMobPLCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_STAT_NAME, vehID );
  vehMobPLStatNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				       myName, configNml );
  if( NULL == vehMobPLStatNml || !vehMobPLStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_SET_NAME, vehID );
  vehMobPLSetNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				     myName, configNml );
  if( NULL == vehMobPLSetNml || !vehMobPLSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CFG_NAME, vehID );
  vehMobPLCfgNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
				      myName, configNml );
  if( NULL == vehMobPLCfgNml || !vehMobPLCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_MAP_NAME, vehID );
  vehMapOutNml = new NML( vehMobPL_format, chanName,
				     myName, configNml );
  if( NULL == vehMapOutNml || !vehMapOutNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }


  // stuff a nop
  nopCmd = new VehMobPLCmdNop();
  nopCfg = new VehMobPLCfgNop();
  nopCmd->serial_number = 100;
  nopCfg->serial_number = 100;

  vehMobPLCmdNml->write(nopCmd);
  vehMobPLCfgNml->write(nopCfg);
  delete nopCmd;
  delete nopCfg;
  return(true);
}

static bool done = false;
/*!
  Signal handler
  \param signum Signal that has been caught
*/
void sigHandler( int signum )
{
  switch( signum )
    {
    case SIGPIPE:
      fprintf( stderr, "Output pipe closed\n" );
      drawMap = false;
      break;
    case SIGINT:
      fprintf( stderr, "Shutting down vehMobPL program\n" );
      //      closeAll();
      done = true;
      break;
    default:
      fprintf( stderr, "Unknown signal caught\n" );
      break;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
  char c;
  bool writeSet;
  VehManufactStat *vehMobPLStat;
  VehMobPLSet  *vehMobPLSet;
  // initial settings of variables that may be set with getopt
  //! The re-planning cycle time
  double cycleTime;
  //! id of vehicle that we are planning for
  int vehID;
  //! The size in meters to grow the obstacles
  double obsGrow;

  /****************************************************************** 
   * INITIALIZE - The initialization opens any communication buffers, 
   *              places the syste in a safe known state, and 
   *		  initializes any control parameters.
   */
  // variables that are initialized here or by getopt
  cycleTime = .2;
  vehID = 1;
  obsGrow = .5;
  // Set drawing to off, with a skip of 5 
  drawMap = false;
  drawSkip = 5;
  // set verbose off
  myVerbose = 0;
  myAmSerial = 1;
  moveWaypointCmd = new AmMobJACmdMoveWaypoint();
  moveWaypointCmd->computeHorizonCosts = true;
  // set up signal handling
  signal( SIGPIPE, sigHandler );
  signal( SIGINT, sigHandler );

  // read the command line parameters to set additional variables
  while((c=getopt(argc,argv,"c:d:g:m:v:"))!=EOF) 
    {
      switch(c) 
	{
	case 'c':
	  // set cycle time
	  cycleTime = atof(optarg);
	  fprintf( stderr, "Cycle time set to %lf\n", cycleTime );
	  break;
	case 'd':
	  // set vehicle id
	  vehID = atoi(optarg);
	  fprintf( stderr, "Connecting to vehicle %d\n", vehID );
	  break;
	case 'g':
	  obsGrow = atoi(optarg);
	  fprintf( stderr, "Obsgrow set to %lf\n", obsGrow );
	  break;
	case 'm':
	  // set map drawing on
	  drawSkip = atoi(optarg);
	  drawMap = true;
	  fprintf( stderr, "Drawing map on with skip of %d\n", drawSkip );
	  break;
	case 'v':
	  myVerbose = atoi(optarg);
	  break;
	default:
	  fprintf( stderr, "usage: %s -c<cycle time> -d<vehID> "
		   "-g <grow obstacles by this; in m> -m<skip> (draw map every skip cycles)"
		   "-v (verbose)\n", argv[0] );
	  done = true;
	}
    }
  timer = new RCS_TIMER(cycleTime);

  // initial allocate all the buffers
  if( !openNML(vehID, 0) )
    {
      done = true;
      rcs_print_error("%s: can't allocate NML buffers\n", argv[0]);
    }

  // send out first settings message
  vehMobPLSet = new VehMobPLSet();
  //  vehMobPLSet->command_type = VEH_MOB_PL_CFG_DUMP_WM_TYPE;
  vehMobPLSet->command_type = VEH_MOB_PL_CFG_NOP_TYPE;
  vehMobPLSet->state = NEW_COMMAND;
  vehMobPLSet->cycleTime = cycleTime;
  vehMobPLSet->obstaclePad = 0;
  vehMobPLSet->usePathDiscount = 0;
  vehMobPLSet->debug = myVerbose;
  vehMobPLSet->isInitialized = false;
  vehMobPLSet->isUsingSQL = false;
  vehMobPLSet->confDistance = 0;
  vehMobPLSet->cellResolution = 0;
  vehMobPLSet->zeroConf = 0;
  vehMobPLSet->fiveConf = 0;
  strcpy(vehMobPLSet->wmFileName, "VehMobPLDispOutput.ppm" ); // ppm faster then tif
  vehMobPLSet->fileSkip = 5;
  vehMobPLSet->echo_serial_number = -1;
  vehMobPLSetNml->write(vehMobPLSet);

  // send out first status message
  vehMobPLStat = new VehManufactStat();
  reset_time_tracker(&vehMobPLStat->tt);
  vehMobPLStat->echo_serial_number = -1;
  vehMobPLStat->state = NEW_COMMAND;
  vehMobPLStat->status = RCS_EXEC;
  vehMobPLStat->command_type = VEH_MOB_PL_CMD_NOP_TYPE;
  vehMobPLStat->planStartHorizon = 4.5;
  vehMobPLStat->hasPackage = false;
  strcpy( vehMobPLStat->packageDestination, "None" );
  strcpy( vehMobPLStat->vehDestination, "None" );
  vehMobPLStatNml->write(vehMobPLStat);

  while (! done) {

  /****************************************************************** 
   * READ COMMAND - Command information is received, and the system
   *                is prepared to execute the command. When a new
   *		    command is received, the old command is immediately
   *		    replaced by this command. If the command is not 
   *		    able to be interrupted, error flags are set and
   *		    will be reported during the writng of status
   */

    // read command
    readCmd(vehMobPLCmdNml, vehMobPLStat, vehMobPLSet);

  /****************************************************************** 
   * READ CONFIG - Configuration information is received, and the 
   *               system is prepared to change its setting. A separate 
   *               configuration channel is provided to allow for control 
   *               parameters to be changed without interrupting the 
   *               current controller. For example, a user may want to 
   *               change a systems cycle time without interrupting a 
   *               complex control function. The configuration channel is 
   *               used to change the cycle time parameters without 
   *               aborting the control algorithm.
   */

    // read config
    readCfg(vehMobPLCfgNml, vehMobPLSet);
    //read amSettings channel
    checkAmMobSettings(amMobJASettings,amMobJASetNml); 

  /****************************************************************** 
   * COMMAND EXECUTION - Command execution is carried out. 
   *                     All of the modules run on a fixed cycle time. 
   *                     Therefore, command functions must either 
   *                     guarantee that they finish in under the cycle time, 
   *                     or provide for being reentrant. Although this 
   *                     example refers to the command execution as being 
   *                     finite state machine (FSM) based, search based 
   *                     planning systems have also been implemented under 
   *                     this framework. The searcher simply stores its 
   *                     current location in the search when its time 
   *                     expires and then resumes searching on the next cycle.
   */

    // handle command state tables
    switch( vehMobPLStat->command_type )
      {
      case VEH_MOB_PL_CMD_NOP_TYPE:
	doCmdNop( vehMobPLStat );
	break;
      case VEH_MOB_PL_CMD_ABORT_TYPE:
	doCmdAbort( vehMobPLStat );
	break;
      case VEH_MOB_PL_CMD_HALT_TYPE:
	if( doCmdHalt( vehMobPLStat ) )
	  {
	    vehMobPLSet->isInitialized = false;
	    vehMobPLSetNml->write(vehMobPLSet);
	  }
	break;
      case VEH_MOB_PL_CMD_INIT_TYPE:
	if( doCmdInit( vehMobPLStat, vehMobPLSet ))
	  {
	    vehMobPLSetNml->write(vehMobPLSet);
	  }
	break;
      case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
	done = doCmdShutdown( vehMobPLStat );
	vehMobPLStatNml->write(vehMobPLStat);
	return(1);
	break;
      case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
	if( !vehMobPLSet->isInitialized )
	  {
	    status_next(vehMobPLStat, RCS_ERROR);
	    break;
	  }
	doCmdMovement( vehMobPLStat, vehMobPLSet );
	break;
      case VEH_MOB_PL_CMD_GO_NAMED_PT_TYPE:
	if( !vehMobPLSet->isInitialized )
	  {
	    status_next(vehMobPLStat, RCS_ERROR);
	    break;
	  }
	doCmdMovement( vehMobPLStat, vehMobPLSet );
	break;
      case VEH_MOB_PL_CMD_LOAD_WM_TYPE:
	if( !vehMobPLSet->isInitialized )
	  {
	    status_next(vehMobPLStat, RCS_ERROR);
	    break;
	  }
	writeSet = doCmdLoadWM( vehMobPLStat, vehMobPLSet );
	break;
      case VEH_MOB_PL_CMD_DOCK_TYPE:
	doCmdDock( vehMobPLStat );
	break;
      case VEH_MOB_PL_CMD_UNDOCK_TYPE:
	doCmdUndock( vehMobPLStat );
	break;
      case VEH_MOB_PL_CMD_LOAD_PKG_TYPE:
	doCmdLoadPkg( vehMobPLStat );
	break;
      case VEH_MOB_PL_CMD_UNLOAD_PKG_TYPE:
	doCmdUnloadPkg( vehMobPLStat );
	break;
      }

  /****************************************************************** 
   * CONFIGURE EXECUTION - Configuration is carried out
   */
    // handle config state tables
    switch( vehMobPLSet->command_type )
      {
      case VEH_MOB_PL_CFG_NOP_TYPE:
	writeSet = doCfgNop( vehMobPLSet );
	break;
      case VEH_MOB_PL_CFG_STORE_WM_TYPE:
	writeSet = doCfgStoreWM( vehMobPLSet, vehMobPLSet );
	break;
      case VEH_MOB_PL_CFG_DUMP_WM_TYPE:
	writeSet = doCfgDumpWM( vehMobPLSet );
	break;
      default:
	writeSet = false;
	break;
      }

  /****************************************************************** 
   * UPDATE STATUS AND SETTING - The current module status and the 
   *                             configuration's settings are sent
   *                             out over communications buffers.
   */
    // update status and settings
    cycle_time_tracker(&vehMobPLStat->tt);
    vehMobPLStatNml->write(vehMobPLStat);
    if( writeSet )
      vehMobPLSetNml->write(vehMobPLSet);
    timer->wait();
  }

  return closeAll();
}

/*
  Modification history:

  $Log: vehVisGraph.cc,v $
  Revision 1.10  2008/03/11 13:17:27  dr_steveb
  Added code to plot prim commanded arcs in gnuplot.

  Revision 1.9  2008/02/13 21:40:46  dr_steveb
  Added change of state on error from adding am seed nodes.

  Revision 1.8  2008/02/12 17:47:50  dr_steveb
  Minor bug fixes

  Revision 1.7  2008/01/09 21:42:47  dr_steveb
  Minor changes before tarball release

  Revision 1.6  2008/01/07 21:48:23  dr_steveb
  Version of vehicle controller for manufacturing competition.

  Revision 1.5  2007/12/20 02:32:46  dr_steveb
  Now have two vehicles performing start of manufacturing job. More to come...

  Revision 1.4  2007/12/18 18:24:37  dr_steveb
  Added commands for loading and unloading packages

  Revision 1.3  2007/12/17 21:35:06  dr_steveb
  Fixed end-point condition so that vehicle now gets to final goal

  Revision 1.2  2007/12/13 22:33:31  dr_steveb
  Fixed for new AM error interface

  Revision 1.1  2007/12/13 13:39:52  dr_steveb
  New visibility-graph based planning system. Only plans on a priori data at this time

*/
