/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file aStarHeapN.hh
    \brief defines classes used by aStarHeap graph searcher with multiple goals

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.3 $
  $Date: 2006/11/30 23:45:56 $
  \endcode

  \author Tom Kramer

This is used by aStarHeapN.cc.

The aStarHeapN.hh and aStarHeapN.cc files define an abstract graph
class (AStarHeapGraph) and an A-star graph search class
(AStarHeapNSearcher) with the following characteristics:

1. The graph may be built or changed dynamically during the search
   in several ways using the addKids, changeGoalCosts, removeKids,
   and/or useKids functions.
2. The user can add information to graph nodes by making a
   child class of nodeInfo.
3. The open nodes list is managed using the heap method.
4. The searcher will search simultaneously for multiple (N) goals.
5. Arcs with a negative cost between nodes are allowed and treated
   as though there were no arc. This is so that graphs can be handled in
   which every node has the same apparent number of children, but some
   are not actually children. Such graphs are often convenient to use.

The start must not be one of the goals. It is up to the caller to
check this.

Two different drivers (exampleAStarHeapNA.cc and exampleAStarHeapNB.cc)
for making a stand-alone executable are in src/tools.  Test files for
testing the executables exist but are not currently part of MOAST.

This is for standard aStar search to a primary goal, G, plus search to
other goals. For each node other than G, a lower bound on the total
cost to reach G is found. This lower bound is called the heuristic.

In general, using aStarHeapN search is not recommended, because in an
abstract graph, aStarHeapN search does not guarantee that paths to
goals other than G are optimal. An example is shown below. We suspect
(but have not found a proof) that if the heuristic is a multiple of a
distance measure satisfying the triangle inequality, and the actual
cost of an arc is never less than that multiple of the distance
between the nodes at the ends of the arc, then the paths found to
other goals are also optimal.

In this example, aStarHeapN search finds a non-optimal path to node B.
In the figure, S is the start node, G is the goal node, and A and B are
two other nodes. The arcs in the graph are shown by arrows. The cost of
each arc is given by the number next to the arc. The heuristic for the
remaining cost to get to the goal is shown by h=4 for A and h=1 for B.
This is an admissible heuristic because in both cases it is less than
the actual costs, which are 5 and 10, respectively. Suppose the
AStarHeapN routine has been asked to find paths to G and B.


                             1
                  ---A------------->B---
                 /   ^              ^   \
                 |   |              |   |
                 |   |              |   |
              h=4|    \            /    |h=1
                 |     \1        3/     |
                5|      \        /      |10
                 |       \      /       |
                 |        \    /        |
                  \        \  /        /
                   \        \/        /
                    -------->G<-------
                             ^
                             |1
                             |
                             S

Starting at S, G will be opened first, at a cost of 1. When G is opened,
A and B will be added to the queue. B will be opened next since its total
estimated cost (1+3+1 = 5) is less than the total estimated cost of
A (1+1+4 = 6). The "arrived from" node for B will be G. The actual total
cost of getting to B from S via G will be (1+3 = 4). The routine will then
stop, since the all goals have been reached, apparently optimally. It
is clear, however, that the path SGAB goes from S to B at a cost of
(1+1+1 = 3), which is less than the cost of the path found by aStarHeapN.

*/

#ifndef A_STAR_HEAP_N_HH
#define A_STAR_HEAP_N_HH

#include "nodeInfo.hh"

class AStarHeapNSearcher;
class AStarHeapNGraph;
class AStarHeapNNode;

/***********************************************************************/

/*! \class AStarHeapNSearcher
    \brief provides a searcher class used in aStar search with N goals and heap

The AStarHeapNSearcher uses a heap of nodes as the list of open
nodes. The cheapest node is always heap[1]. The heap is not totally
ordered.  As a result the heap is relatively cheap to maintain.

The goals, costs, and statuses arrays are all of length numGoals.
costs[n] is set by the Searcher to the cost of reaching goals[n].
statuses[n] is set to zero when goals[n] has not been reached, to 1 when
goals[n] has been reached but not optimally, and to 2 when goals[n] has
been reached optimally.

The goals includes the target of the search as its first entry.
The heuristic used in the search should be based on the target
(Euclidean distance to the target is appropriate).

The goalsMap is a map from node number to index in goals. The
goalsMapSize does not change after being set by initSearcher. The
size is set to be just big enough so that the maximum of the indexes
of the goals is in bounds. That is, if the maximum of the indexes of
the goals is M, goalsMapSize is (M+1).

If the heuristic of Euclidean distance to target is used, and goals other
than the target are in the general vicinity of the cheapest path to
the target, aStar search should be very effective. If the other goals
are far from the cheapest path to the target, aStar (since it tends
to open nodes in an expanding balloon around a line from start to
target) will be slower than Dijkstra (since it tends to open nodes
in an expanding balloon around the start point and does not require
the extra work of computing a heuristic).

*/

class AStarHeapNSearcher
{
public:
  AStarHeapNSearcher(){ heapSize = 0; }
  ~AStarHeapNSearcher(){}

  int closeSearch(); /*! closes the searcher */
  int cycleSearch(); /*! pops top node from the heap, processes it */
  int getNextNode(); /*! returns the index of cheapest open node remaining */
  int initSearch(    /*! initializes the searcher */
   AStarHeapNGraph * aGraph, /*!< the graph to search in                     */
   int start,                /*!< index in nodes of aGraph of start node     */
   int numGoalsIn,           /*!< number of goals                            */
   int * goalsIn,            /*!< indexes in nodes of aGraph of goal nodes   */
   int * opensOut,           /*!< 0=open list empty, non-zero=list not empty */
   int * statusesOut,        /*!< statuses of reaching goals                 */
   double * costsOut,        /*!< costs of reaching goals                    */
   int * goalsFoundOut,      /*!< number of goals reached                    */
   int * optimaFoundOut);    /*!< number of goals reached optimally          */
  int reportSearch( /*! writes the path to a node into an array           */
   int index,       /*!< index in nodes of node to find path to           */
   int maxLength,   /*!< maximum length of path (capacity of path array)  */
   int * path);     /*!< array of index numbers giving the path, set here */

private:
  double * costs;         /* array of costs of reaching goals                */
  int * goals;            /* array of indexes in nodes of goal nodes         */
  int * goalsFound;       /* number of goals reached                         */
  int * goalsMap;         /* map from node number to index in goals (array)  */
  int goalsMapSize;       /* size of goalsMap array                          */
  int * heap;             /* heap of node indexes, cost ordered              */
  int heapCapacity;       /* capacity of heap (i.e. total array size)        */
  int heapSize;           /* size of heap (i.e. number of array places used) */
  int * heapSizeOut;      /* size of heap for caller                         */
  AStarHeapNNode * nodes; /* array of nodes from theGraph                    */
  int numGoals;           /* length of goals array                           */
  int * optimaFound;      /* number of goals reached optimally               */
  int * statuses;         /* array of status flags                           */
  AStarHeapNGraph * theGraph; /* the graph to be searched                    */

  void buildGoalsMap();
  void heapDown();
  void heapUp(int index);
  void increaseHeap(int newCapacity);
  void processNode(int numNode, int from, double cost);
  void removeCheap();
};

/***********************************************************************/

/*! \class AStarHeapNNode
    \brief provides a node class used in aStar search with heap

See aStarHeapN.cc for descriptions of what the functions do.

The destructor (~AStarHeapNNode) does not delete kidsIndexes or kidsCosts.
This is so that copying a node can be done faster; kidsIndexes and
kidsCosts are reused by the copy.  If it is desired to delete
kidsIndexes or kidsCosts in order to free memory, that must be done
explicitly.

Note that everything required of a AStarHeapNNode (other than NodeInfo)
is here. There is no graphNode base class -- not worth the trouble. 

*/

class AStarHeapNNode
{
public:
  enum status {UNOPENED, OPEN, CLOSED};
  AStarHeapNNode() { kidsSize= 0; open = UNOPENED; }
  ~AStarHeapNNode();

  int copyFrom             /*!< returns 0 if from is 0, otherwise returns 1 */
  (AStarHeapNNode * from); /*!< node to copy from                           */

  NodeInfo * moreInfo; //!< additional information about node
  int kidsSize;        //!< number of kids of node
  int * kidsIndexes;   //!< array of kid indexes of node, built at read time
  double * kidsCosts;  //!< array of kid costs of node, built at read time
  status open;         //!< flag indicating UNOPENED, OPEN, CLOSED
  int arrivedFrom;     //!< index in nodes of node from which arrived
  double visitCost;    //!< current total cost to visit node
  double goalCost;     //!< fixed lower bound on actual cost of reaching goal
  double totalCost;    //!< visitCost plus goalCost
  int heapIndex;       //!< index in heap of node
};

/***********************************************************************/

/* \class AStarHeapNGraph
   \brief provides a graph class used in aStar search with heap

Four public methods are provided to allow an application to change the
graph being searched during the search.

In addition, the application may change the graph during the search
without using any of these methods.

In the following discussions, keep in mind that an AStarHeapNGraph has
both a "capacity", which is the number of nodes currently available,
and a "size", which is the number of available nodes currently being
used.

************************
THE addKids FUNCTION
************************

AddKids may activate additional nodes of the graph. If it does, it also
adds arcs from dad to those new nodes. This may also add arcs from dad
to existing nodes.  The memory for dad's old kidsIndexes and kidsCosts
is freed. Memory for dad's revised kidsIndexes and kidsCosts is
allocated.

If there are not enough nodes in the graph to accommodate the number
to be activated, addKids calls increaseCapacity to make more nodes.

The new kids are built in nodes whose indexes are assigned in
numerical order, starting with the current size of the graph. The size
is increased as the new kids are added. The index in the graph should
already be recorded as the value of "index" in the nodeInfo. It is
checked that this number is correct.

The additional kids in oldAdded must not already be kids of dad.
AddKids is not currently checking for that. A check could be added.

********************************
THE changeKidsGoalCosts FUNCTION
********************************

ChangeKidsGoalCosts changes the goalCosts of some or all the kids of dad.

The numberChanged argument is the number of (existing) kids
of dad that are to be assigned new goal costs (corresponding to kids
indexes 0 through (numberNew - 1)). The new goal costs are given
in the newGoalCosts array.

The numberChanged should not be larger than the number of kids of dad.

************************
THE useKids FUNCTION
************************

The useKids function removes the current kids of dad (if any) and
replaces them with the ones given by the arguments. The arrays
specified in the arguments are used as is, so they must be created
with "new" in the implementing system and not changed or deleted by
the implementing system.

The kids themselves must already exist and are not affected by useKids.

**********************
OTHER MEMBER FUNCTIONS
**********************

The increaseCapacity function makes a new and larger array of nodes
and copies the old nodes into it. Since all the references in the
Searcher to nodes in the graph are done by index number, not by
pointer, the Searcher does not notice the switch.

The destructor deletes the contents of the nodes and the nodes.

The removeKids function removes all the kids of the dad node.

The graph may be built using nodes created with "new" or with nodes
created some other way (e.g. a global array). The "deleat" data member
should always be set to false if the nodes were not created with
"new", and may be set to false if the nodes were created with "new"
but they should survive the destruction of the graph. The default
value of deleat is true.

*/

class AStarHeapNGraph
{
public:
  AStarHeapNGraph(){deleat = true;}
  ~AStarHeapNGraph();

  int addKids             /*! adds kids to dad and builds new nodes          */
  (int dad,               /*!< index in nodes of node for which to add kids  */
   int numberAdded,       /*!< total number of kids added                    */
   int numberNew,         /*!< number of new kids added                      */
   NodeInfo ** newInfos,  /*!< array of pointers to NodeInfos for new kids   */
   double * newArcCosts,  /*!< array of arc costs from dad for new kids      */
   double * newGoalCosts, /*!< array of goal costs for new kids              */
   int * oldAdded,        /*!< number of existing kids added                 */
   double * oldArcCosts); /*!< array of arc costs from dad for existing kids */

  int changeKidsGoalCosts  /*! changes goalCosts of kids of dad           */
  (int dad,                /*!< index in nodes of node to change kids for */
   int numberChanged,      /*!< number of kids to change                  */
   double * newGoalCosts); /*!< array of new goalCosts of kids            */
  
  int removeKids  /*! removes kidIndexes & kidCosts from dad, not kids   */
  (int dad);      /*!< index in nodes of node from which to remove kids  */

  int useKids               /*! specifies new kids of dad                  */
  (int dad,                 /*!< index in nodes of node to change kids for */
   int kidsSize,            /*!< new number of kids                        */
   int * kidsIndexes,       /*!< new array of kids indexes in nodes        */
   double * kidsCosts);     /*!< new array of kids costs                   */

  int capacity;            //!< total number of nodes in nodes array
  int size;                //!< number of active nodes in nodes array
  AStarHeapNNode * nodes;  //!< array of nodes
  bool deleat;             //!< delete nodes in destructor? default=true

private:

  int increaseCapacity  /*! makes larger array, copies old nodes into it   */
  (int numberAdded);    /*!< minimum number by which to increase capacity  */

};

/***********************************************************************/

#endif // A_STAR_HEAP_N_HH
