/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file dijkHeap.hh
    \brief defines classes used by dijkHeap graph searcher

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Tom Kramer

This is used by dijkHeap.cc.

The dijkHeap.hh and dijkHeap.cc files define an abstract graph class
(DijkHeapGraph) and a Dijkstra graph search class (DijkHeapSearcher)
with the following characteristics:

1. The graph may be built or changed dynamically during the search
   in several ways using the addKids, removeKids, and/or useKids
   functions.
2. The user can add information to graph nodes by making a
   child class of nodeInfo.
3. The open nodes list is managed using the heap method.
4. Arcs with a negative cost between nodes are allowed and treated
   as though there were no arc. This is so that graphs can be handled in
   which every node has the same apparent number of children, but some
   are not actually children. Such graphs are often convenient to use.

This searcher does NOT implement searching for multiple goals simultaneously.

The start must not be the same as the goal. It is up to the caller to
check this.

A driver, exampleDijkHeap.cc, for making a stand-alone executable is
in src/tools.  Test files for testing the executable exist but are not
currently part of MOAST.

*/

#ifndef DIJK_HEAP_HH
#define DIJK_HEAP_HH

#include "nodeInfo.hh"

class DijkHeapSearcher;
class DijkHeapGraph;
class DijkHeapNode;

/***********************************************************************/

/*! \class DijkHeapSearcher
    \brief provides a searcher class used in Dijkstra search with heap

The DijkHeapSearcher uses a heap of nodes as the list of open
nodes. The cheapest node is always heap[1]. The heap is not totally
ordered.  As a result the heap is relatively cheap to maintain.

*/

class DijkHeapSearcher
{
public:
  DijkHeapSearcher(){ heapSize = 0; }
  ~DijkHeapSearcher(){}

  int closeSearch(); /*! closes the searcher */

  int cycleSearch(); /*! pops top node from the heap, processes it */

  int getNextNode(); /*! returns the index of cheapest open node remaining */

  int initSearch(          /*!  initializes the searcher                   */
   DijkHeapGraph * aGraph, /*!< the graph to search in                     */
   int start,              /*!< index in nodes of aGraph of start node     */
   int finish,             /*!< index in nodes of aGraph of finish node    */
   int * opensOut,         /*!< 0=open list empty, non-zero=list not empty */
   int * statusOut,        /*!< status of reaching goal                    */
   double * costOut);      /*!< cost of reaching goal                      */

  int reportSearch( /*! writes the path to the goal into an array         */
   int maxLength,   /*!< maximum length of path (capacity of path array)  */
   int * path);     /*!< array of index numbers giving the path, set here */

private:
  double * cost;            /* array (size 1) for cost of reaching goal */
  int goal;                 /* index in nodes of goal (finish) node     */
  int * heap;               /* heap of node indexes, cost ordered       */
  int heapCapacity;         /* capacity of heap (i.e. total array size) */
  int heapSize;             /* size of heap                             */
  int * heapSizeOut;        /* size of heap for caller                  */
  DijkHeapNode * nodes;     /* array of nodes from theGraph             */
  int * status;             /* array (size 1) for status flag           */
  DijkHeapGraph * theGraph; /* the graph to be searched                 */

  void heapDown();
  void heapUp(int index);
  void increaseHeap(int newCapacity);
  void processNode(int numNode, int from, double cost);
  void removeCheap();
};

/***********************************************************************/

/*! \class DijkHeapNode
    \brief provides a node class used in Dijkstra search with Heap

See dijkHeap.cc for descriptions of what the other functions do.

The destructor (~DijkHeapNode) does not delete kidsIndexes or kidsCosts.
This is so that copying a node can be done faster; kidsIndexes and
kidsCosts are reused by the copy.  If it is desired to delete
kidsIndexes or kidsCosts in order to free memory, that must be done
explicitly.

Note that everything required of a DijkHeapNode (other than NodeInfo)
is here. There is no graphNode base class -- not worth the trouble. 

Note: The open flag could be removed, since the heapIndex could do
double duty (non-negative is real index, -2 is unopened, -1 is closed).

*/

class DijkHeapNode
{
public:
  enum status {UNOPENED, OPEN, CLOSED};
  DijkHeapNode() { kidsSize= 0; open = UNOPENED; }
  ~DijkHeapNode();

  int copyFrom           /*!< returns 0 if from is 0, otherwise returns 1 */
  (DijkHeapNode * from); /*!< node to copy from                           */

  NodeInfo * moreInfo; //!< additional information about node
  int kidsSize;        //!< number of kids of node
  int * kidsIndexes;   //!< array of kid indexes of node
  double * kidsCosts;  //!< array of kid costs of node
  status open;         //!< flag for UNOPENED, OPEN, CLOSED
  int arrivedFrom;     //!< index in nodes of node from which arrived
  double visitCost;    //!< current total cost to visit node
  int heapIndex;       //!< index in heap of node
};

/***********************************************************************/

/*! \class DijkHeapGraph
    \brief provides a graph class used in Dijkstra search with heap

Three public methods are provided to allow an application to change the
graph being searched during the search.

In addition, the application may change the graph during the search
without using any of these methods.

In the following discussions, keep in mind that a DijkHeapNGraph has
both a "capacity", which is the number of nodes currently available,
and a "size", which is the number of available nodes currently being
used.

************************
THE addKids FUNCTION
************************

AddKids may activate additional nodes of the graph. If it does, it also
adds arcs from dad to those new nodes. This may also add arcs from dad
to existing nodes.  The memory for dad's old kidsIndexes and kidsCosts
is freed. Memory for dad's revised kidsIndexes and kidsCosts is
allocated.

If there are not enough nodes in the graph to accommodate the number
to be activated, addKids calls increaseCapacity to make more nodes.

The new kids are built in nodes whose indexes are assigned in
numerical order, starting with the current size of the graph. The size
is increased as the new kids are added. The index in the graph should
already be recorded as the value of "index" in the nodeInfo. It is
checked that this number is correct.

The additional kids in oldAdded must not already be kids of dad.
AddKids is not currently checking for that. A check could be added.

************************
THE useKids FUNCTION
************************

The useKids function removes the current kids of dad (if any) and
replaces them with the ones given by the arguments. The arrays
specified in the arguments are used as is, so they must be created
with "new" in the implementing system and not changed or deleted by
the implementing system.

The kids themselves must already exist and are not affected by useKids.

**********************
OTHER MEMBER FUNCTIONS
**********************

The increaseCapacity function makes a new and larger array of nodes
and copies the old nodes into it. Since all the references in the
Searcher to nodes in the graph are done by index number, not by
pointer, the Searcher does not notice the switch.

The destructor deletes the contents of the nodes and the nodes.

The removeKids function removes all the kids of the dad node.

*/

class DijkHeapGraph
{
public:
  DijkHeapGraph(){}
  ~DijkHeapGraph();

  int addKids            /*! adds kids to dad                              */
  (int dad,              /*!< index in nodes of node for which to add kids */
   int numberAdded,      /*!< total number of kids added                    */
   int numberNew,        /*!< number of new kids added                      */
   NodeInfo ** newInfos, /*!< array of pointers to NodeInfos for new kids   */
   double * newCosts,    /*!< array of arc costs from dad for new kids      */
   int * oldAdded,       /*!< number of existing kids added                 */
   double * oldCosts);   /*!< array of arc costs from dad for existing kids */

  int removeKids  /*!  removes kidIndexes & kidCosts from dad, not kids  */
  (int dad);      /*!< index in nodes of node from which to remove kids  */

  int useKids           /*! specifies new kids of dad                  */
  (int dad,             /*!< index in nodes of node to change kids for */
   int kidsSize,        /*!< new number of kids                        */
   int * kidsIndexes,   /*!< new array of kids indexes in nodes        */
   double * kidsCosts); /*!< new array of kids costs                   */

  int capacity;         //!< total number of nodes in nodes array
  int size;             //!< number of active nodes in nodes array
  DijkHeapNode * nodes; //!< array of nodes

private:

  int increaseCapacity  /*! makes larger array, copies old nodes into it */
  (int numberAdded);    /*  minimum number by which to increase capacity */

};

/***********************************************************************/

#endif // DIJK_HEAP_HH

/*
  $Log: dijkHeap.hh,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.6  2005/07/18 14:06:07  kramer
  added goal not start warning to documentation

  Revision 1.5  2005/07/15 21:23:29  kramer
  minor changes in documentation

  Revision 1.4  2005/07/15 14:01:22  kramer
  changed interface, improved documentation

  Revision 1.3  2005/06/29 21:08:05  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.2  2005/06/09 12:16:37  kramer
  added CVS logging
*/
