/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file mfd.hh

  \brief MOAST For Dummies class. Designed to ease the learning curve for using MOAST.

  This class is designed to provide a simple to use interface that allows novice programmers to create a MOAST application with '5 lines of code or less'. It has interfaces into each echelon of MOAST and can do some sample behaviors at each echelon. The calling sequence is to initialize the appropriate system with the constructor, and then to simply issue appropriate commands


  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.3 $
  $Date: 2009/05/07 19:01:07 $
  \endcode

  \author Stephen Balakirsky
  \date April 7, 2009
*/

#include <stdio.h>

// rcs headers
#include <rcs.hh>
#include <posemath.h>

// moast nml headers
#include <moastNmlOffsets.hh>
#include <moastTypes.hh>
#include <servoMobJA.hh>
#include <servoSP.hh>
#include <servoEffJA.hh>
#include <primMobJA.hh>
#include <primSP.hh>
#include <amMobJA.hh>
#include <amSP.hh>
#include <factCtrl.hh>
#include <navDataExt.hh>
#include <sensorData.hh>

#ifndef MFD_HH
#define MFD_HH

//! Enum structure for representing the selected echelon. Numbers are powers of two to allow for future 'oring' of echelons
typedef enum {
  MFD_NO_ECHELON = 0,
  MFD_SERVO =      2,
  MFD_PRIM =       4,
  MFD_AM =         8,
  MFD_VEH =       16,
  MFD_SECT =      32,
  MFD_FACTORY =   64
}MfdEchelon;

typedef enum {
  MFD_ERROR_NONE,
  MFD_ERROR_NML_CONNECT,
  MFD_ERROR_NML_DATA_CONNECT,
  MFD_ERROR_NML_NO_MESSAGE,
  MFD_ERROR_NML_FAULT,
  MFD_ERROR_SUBSYSTEM_NOT_INITIALIZED,
  MFD_ERROR_UNSUPPORTED_FEATURE,

  MFD_ERROR_SERVO_ECHELON_NOT_INITIALIZED,
  MFD_ERROR_PRIM_ECHELON_NOT_INITIALIZED,
  MFD_ERROR_AM_ECHELON_NOT_INITIALIZED,
  MFD_ERROR_FACTORY_ECHELON_NOT_INITIALIZED,

  MFD_ERROR_SERVO_ECHELON_CMD,
  MFD_ERROR_PRIM_ECHELON_CMD,
  MFD_ERROR_AM_ECHELON_CMD,
  MFD_ERROR_FACTORY_ECHELON_CMD
}MfdError;

typedef struct {
  RCS_CMD_CHANNEL  *cmdBuf;
  RCS_STAT_CHANNEL *statBuf;
  RCS_CMD_CHANNEL  *cfgBuf;
  RCS_STAT_CHANNEL *setBuf;
  NML              **dataBuf; // data buffers for cluster, may be void

  RCS_STAT_MSG_V2  *statMsg;
  RCS_STAT_MSG     *setMsg;
}NmlCluster;

//! Possible types for use in prim cluster's dataBuf. To access, use dataBuf[xx] where xx is one of the enum values.
typedef enum {
  PRIM_POSE = 0,
  PRIM_LINE_SCAN_1,
  PRIM_LINE_SCAN_2,
  PRIM_LINE_SCAN_3,
  PRIM_SP_MAX_BUFFERS
}PrimDataBuffers;

#define DEFAULT_NML_FILE "../etc/moast.nml" /*!< Default to use for nml configuration file */
#define DEFAULT_NML_NAME "mfd"              /*!< Default name to use for this process in the nml config. file */

class Mfd
{
public:
  //! variables
  
  //! functions
  Mfd(char *nmlFileIn=DEFAULT_NML_FILE, 
      char *nmlNameIn=DEFAULT_NML_NAME);           /* constructor, defaults nml file and name */
  ~Mfd();                                          /* destructor */
  MfdError getErrorNumber();                       /* returns the error number from the last mfd call */
  bool init( MfdEchelon echelonIn=MFD_SERVO, 
	     int vehIDIn=1);                       /* initialize the system, defaults echelon and vehID */
  MfdEchelon isInitialized();                      /* returns the initialized echelons */
  const char *mfdEchelonStr(MfdEchelon echelonIn); /* returns the text for a given echelon */
  const char *mfdErrorStr();                       /* prints the latest error (if any) */
  bool openNML(NmlCluster *chanIn, char *nmlFile, char *cmdName, char *statName,  /* opens nml buffers to particular echelon */
	       char *cfgName, char *setName,
	       char *myName, NML_FORMAT_PTR fPtr, int botNum, int compID );
  bool openNMLData(NML **chanIn, char *nmlFile, char *name,
		   char *myName, NML_FORMAT_PTR fPtr, int botNum, int compID );

  //! servo echelon functions
  bool servoSkidSteer(double left, double right );      /* skid-steer control for servo */
  bool servoEffOpcode( int effID, ServoEffJAOpcodeType opcode, double param ); /* send an opcode to an effector */

  //! prim echelon functions
  bool primGetPose( NavDataExt **poseInfo );
  bool primGetLaser( SensorData1D **sensorData, int sensorID=1 );
  bool primMoveWaypoint( Waypoint *waypoints, int numPts );
  bool primRotate( double position, double tol=.1, int direction=MOAST_ROTATE_SHORTEST );

  //! am echelon functions
  bool amMoveWaypoint( Waypoint *waypoints, int numPts );

  //! factory echelon functions
  bool factoryDropPackage(char *chute);
  bool factorySetConveyor(char *conveyorName, float speed);
private:
  //!variables
  MfdError currentError; /* what is the current error, if any? */
  MfdEchelon myEchelon;  /* what is my control echelon */
  char *nmlFile;         /* the nml configuration file name */
  char *nmlName;         /* the name of this process in the nml configuration file */
  int mySerialNumber;    /* serial number to send for nml commands */

  //! nml channels
  NmlCluster *servoMob;
  NmlCluster *servoSp;
  NmlCluster *servoEff;
  NmlCluster *primMob;
  NmlCluster *primSp;
  NmlCluster *amMob;
  NmlCluster *amSp;
  NmlCluster *factoryCtrl;
  
  //!functions
  bool waitForCmdDone( double interval, int cmdId, NmlCluster *checkMe ); /* wait for current command to complete */

};
#endif  
/*
  Modification history:

  $Log: mfd.hh,v $
  Revision 1.3  2009/05/07 19:01:07  dr_steveb
  Check-in of code for travel

  Revision 1.2  2009/04/27 16:09:39  dr_steveb
  Getting ready for demo...

  Revision 1.1  2009/04/24 18:45:16  dr_steveb
  Initial check-in of moast for dummies class

*/
