/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file vehMobPL.hh

  \brief NML declarations for vehicle veh level command, status, config
  and settings.


  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.18 $
  $Date: 2007/12/18 18:25:21 $
  \endcode

  \author Stephen Balakirsky
*/

#include <stdio.h>
#include <rcs.hh>
#include <posemath.h>
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"

#ifndef VEH_MOB_PL_HH
#define VEH_MOB_PL_HH

#define VEH_MOB_PL_CMD_NAME "vehMobPLCmd"
#define VEH_MOB_PL_STAT_NAME "vehMobPLStat"
#define VEH_MOB_PL_CFG_NAME "vehMobPLCfg"
#define VEH_MOB_PL_SET_NAME "vehMobPLSet"
#define VEH_MOB_MAP_NAME "vehMapOut"
#define MAX_WAYPOINT_NAME_LENGTH 40

#define VEH_MOB_PL_CMD_BASE        (VEH_MOB_PL_BASE)
#define VEH_MOB_PL_STAT_BASE       (VEH_MOB_PL_BASE + 100)
#define VEH_MOB_PL_CFG_BASE        (VEH_MOB_PL_BASE + 200)
#define VEH_MOB_PL_SET_BASE        (VEH_MOB_PL_BASE + 300)

#define VEH_MOB_PL_CMD_INIT_TYPE          (VEH_MOB_PL_CMD_BASE + 1)
#define VEH_MOB_PL_CMD_ABORT_TYPE         (VEH_MOB_PL_CMD_BASE + 2)
#define VEH_MOB_PL_CMD_HALT_TYPE          (VEH_MOB_PL_CMD_BASE + 3)
#define VEH_MOB_PL_CMD_SHUTDOWN_TYPE      (VEH_MOB_PL_CMD_BASE + 4)
#define VEH_MOB_PL_CMD_NOP_TYPE           (VEH_MOB_PL_CMD_BASE + 5)
#define VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE (VEH_MOB_PL_CMD_BASE + 6)
#define VEH_MOB_PL_CMD_MOVE_ROAD_TYPE     (VEH_MOB_PL_CMD_BASE + 7)
#define VEH_MOB_PL_CMD_MOVE_SQL_TYPE      (VEH_MOB_PL_CMD_BASE + 8)
#define VEH_MOB_PL_CMD_MOVE_HAZARD_TYPE   (VEH_MOB_PL_CMD_BASE + 9)
#define VEH_MOB_PL_CMD_EXPLORE_TYPE       (VEH_MOB_PL_CMD_BASE +10)
#define VEH_MOB_PL_CMD_LOAD_WM_TYPE       (VEH_MOB_PL_CMD_BASE +11)
#define VEH_MOB_PL_CMD_DOCK_TYPE          (VEH_MOB_PL_CMD_BASE +12)
#define VEH_MOB_PL_CMD_UNDOCK_TYPE        (VEH_MOB_PL_CMD_BASE +13)
#define VEH_MOB_PL_CMD_LOAD_PKG_TYPE      (VEH_MOB_PL_CMD_BASE +14)
#define VEH_MOB_PL_CMD_UNLOAD_PKG_TYPE    (VEH_MOB_PL_CMD_BASE +15)

// veh commands below here are not yet implemented
#define VEH_MOB_PL_CMD_VICT_LOCATE_TYPE   (VEH_MOB_PL_CMD_BASE +16)
#define VEH_MOB_PL_CMD_GO_NAMED_PT_TYPE   (VEH_MOB_PL_CMD_BASE +17)

#define VEH_MOB_PL_STAT_TYPE              (VEH_MOB_PL_STAT_BASE + 1)
#define VEH_MANUFACT_STAT_TYPE            (VEH_MOB_PL_STAT_BASE + 2)

#define VEH_MOB_PL_CFG_CYCLE_TIME_TYPE    (VEH_MOB_PL_CFG_BASE + 1)
#define VEH_MOB_PL_CFG_DEBUG_TYPE         (VEH_MOB_PL_CFG_BASE + 2)
#define VEH_MOB_PL_CFG_NOP_TYPE           (VEH_MOB_PL_CFG_BASE + 3)
#define VEH_MOB_PL_CFG_MOVE_TYPE          (VEH_MOB_PL_CFG_BASE + 4)
#define VEH_MOB_PL_CFG_CLEAR_WM_TYPE      (VEH_MOB_PL_CFG_BASE + 5)
#define VEH_MOB_PL_CFG_STORE_WM_TYPE      (VEH_MOB_PL_CFG_BASE + 6)
#define VEH_MOB_PL_CFG_DUMP_WM_TYPE       (VEH_MOB_PL_CFG_BASE + 7)
#define VEH_MOB_PL_CFG_SQL_TYPE           (VEH_MOB_PL_CFG_BASE + 8)

#define VEH_MOB_PL_SET_TYPE               (VEH_MOB_PL_SET_BASE + 1)

#define VEH_MOB_PL_MAX_WAYPOINTS 100

/* state values used in doCmdMovement state machine
 */
//state names for movement state machine
#define S_CHECK_AM_ERR               18
#define S_CHECK_AM_STAT_FROM_PREPARE 19
#define S_CHECK_AM_STAT_FROM_SPIN    20
#define S_CHECK_STUCK                21
#define S_PLAN_ROUTE                 22
#define S_PREPARE_MOVE               23
#define S_START_OBS                  24
#define S_STUCK                      25
#define S_MOVEMENT_ERR               26
#define S_DONE_MOVEMENT              27
#define S_BLOCK_CENTER               28
#define S_ADD_AM_NODES               29
#define S_ADD_VEH_GOAL               30
#define S_CLEAR_GRAPH                31
#define S_SEND_PATH                  32

//#define RFIDHalfFOV .78539 // M_PI_4
/*! variables used for exploration gives the range and FOV of the 
    RFID sensor. Should be replaced with information read from servo
*/
#define RFIDHalfFOV .26
#define RFIDRange 6

/////////////////////////////////////////////////////////////////
// supporting structures
//! Povide detailed information for a command to achieve a road location
typedef struct // roadPoint
{
  //! value of world_id from database
  int worldID;
  //! value of lane_id from database
  int laneID;
  //! constrained offset from waypoint location down roadway
  /*! Constraints:\n
     0 - must be HERE\n
     1 - must be +- tollerance\n
     2 - must be in lane segment\n
     3 - must be in lane\n
  */
  constrainedInt offset;


  //! constrained orientation of vehicle on the road
  //! positive clockwise rotation where North=0 
  /*! Constraints:\n
     0 - must be in POSE\n
     1 - must be +- tollerance\n
     2 - no orientation defined\n
  */
  constrainedFloat orientation;


  //! constrained offset from waypoint location accross roadway
  /* ! Constraints:\n
     0 - must be HERE\n
     1 - must be +- tollerance\n
     2 - must be in lane segment\n
     3 - must be in lane cluster\n
     4 - must be in road\n
  */
  constrainedInt LSID;

 //! constrained integer for the speed 
  /* Constraints:\n
     0 - must be traveling at this speed while this state\n
     1 - must be +- tollerance of speed\n
     2 - not defined
  */
  constrainedInt speed;

  //! constrained integer for the time
  /* Constraints:\n
     0 - must be at this time +- tollerance
     1 - must be earlier than this time
     2 - must be later than this time
  */
  constrainedInt time;

}roadPoint;

//! used by SQL drive to constrain the neighborhood.
enum distConstraint
  {
    EXACTLY,
    LESS_THEN,
    GREATOR_THEN
  };

enum GoalType
  {
    VICTIM,
    WAYPOINT
  };

typedef struct 
{
  PM_CARTESIAN location;
  bool isGlobal;
  GoalType goalType;
  double neighborhood;
}goalPoint;

/////////////////////////////////////////////////////////////////
// commands
class VehMobPLCmdInit:public RCS_CMD_MSG {
public:
  VehMobPLCmdInit():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_INIT_TYPE,
     sizeof(VehMobPLCmdInit)) {};
  void update(CMS *);
};

class VehMobPLCmdHalt:public RCS_CMD_MSG {
public:
  VehMobPLCmdHalt():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_HALT_TYPE,
     sizeof(VehMobPLCmdHalt)) {};
  void update(CMS *);
};

class VehMobPLCmdAbort:public RCS_CMD_MSG {
public:
  VehMobPLCmdAbort():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_ABORT_TYPE,
     sizeof(VehMobPLCmdAbort)) {};
  void update(CMS *);
};

class VehMobPLCmdShutdown:public RCS_CMD_MSG {
public:
  VehMobPLCmdShutdown():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_SHUTDOWN_TYPE,
     sizeof(VehMobPLCmdShutdown)) {};
  void update(CMS *);
};

class VehMobPLCmdNop:public RCS_CMD_MSG {
public:
  VehMobPLCmdNop():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_NOP_TYPE,
     sizeof(VehMobPLCmdNop)) {};
  void update(CMS *);
};

/*!
  The VehMobPLCmdGoNamedPt NML message contains a named point
  and associated neighborhood to drive to.
*/
class VehMobPLCmdGoNamedPt:public RCS_CMD_MSG {
public:
  VehMobPLCmdGoNamedPt():RCS_CMD_MSG
			 (VEH_MOB_PL_CMD_GO_NAMED_PT_TYPE,
			  sizeof(VehMobPLCmdGoNamedPt)) {};
  void update(CMS *);
  char namedPt[MAX_WAYPOINT_NAME_LENGTH];
  double neighborhood;
};

/*!
  The VehMobPLCmdMoveWaypoint NML message contains a waypoint
  and associated neighborhood to drive to. The location may be
  specified in global or in relative coordinates.
*/
class VehMobPLCmdMoveWaypoint:public RCS_CMD_MSG {
public:
  VehMobPLCmdMoveWaypoint():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE,
     sizeof(VehMobPLCmdMoveWaypoint)) {};
  void update(CMS *);

  //! waypoints in local coordinates, units [m]
  PM_CARTESIAN p;
  //! neighborhood within which the waypoint is deemed attained, units [m]
  double neighborhood;
  //! is the waypoint in global coordinates?
  bool isGlobal;
  //! Time to start movement (in seconds since the Epoch (00:00:00 UTC, January 1, 1970)
  long startTime;
};

//! command to drive a road driving path
/*! This class contains a command to drive a specific set of on-road 
    waypoints.
 */
class VehMobPLCmdMoveRoad: public RCS_CMD_MSG
{
public:
  //! constructor
  VehMobPLCmdMoveRoad():RCS_CMD_MSG
  ( VEH_MOB_PL_CMD_MOVE_ROAD_TYPE, 
    sizeof( VehMobPLCmdMoveRoad )) {};
  //! standard NML update function
  void update( CMS* );

  //! Time to start movement (in seconds since the Epoch (00:00:00 UTC, January 1, 1970)
  long startTime;

  //! Structure that is the road point to drive to
  roadPoint location;
};

//! command to drive to a specific UTM location while avoiding a hazard
/*! This location may be on or off road and 
    driving style will be controlled by parameters
 */
class VehMobPLCmdMoveHazard: public RCS_CMD_MSG
{
public:
  //! constructor
  VehMobPLCmdMoveHazard():RCS_CMD_MSG
  ( VEH_MOB_PL_CMD_MOVE_HAZARD_TYPE,
    sizeof( VehMobPLCmdMoveHazard )) {};
  //! standard NML update function
  void update( CMS* );

  //! Time to start movement (in seconds since the Epoch (00:00:00 UTC, January 1, 1970)
  long startTime;
  //! Location to drive to
  PM_CARTESIAN location;
  //! is the position in global coordinates?
  bool isGlobal;
  //! name of table that contains the id
  /*! it is assumed that the table contains the following fields:
      1) ID - the id of the hazard as passed into this command
      2) IsHazard - true if the hazard is valid, false if it has been neutralized
      3) Class - The class of the hazard
  */
  char hazardTable[MOAST_FILE_NAME_LEN];
  //! ID of the hazard to by driven around from table 'table'
  int hazardId;
  //! name of the table that contains the hazard class information
  /*! it is assumed that the table contains the following fields:
    1) ID - the id of the class of hazard
    2) Hazard_Radius - the distance that must be maintained from this hazard.
  */
  char hazardClassTable[MOAST_FILE_NAME_LEN];
  //! constraint on the distance
  distConstraint constraint;
};

//! command to drive a path to a specified point. 
/*! This class contains a command to drive a specific location.
    Both the location and the required neighborhood are determined
    from a sql look-up.
 */
class VehMobPLCmdMoveSQL: public RCS_CMD_MSG
{
public:
  //! constructor
  VehMobPLCmdMoveSQL():RCS_CMD_MSG
  ( VEH_MOB_PL_CMD_MOVE_SQL_TYPE,
    sizeof( VehMobPLCmdMoveSQL )) {};
  //! standard NML update function
  void update( CMS* );

  //! Time to start movement (in seconds since the Epoch (00:00:00 UTC, January 1, 1970)
  long startTime;
  //! ID of the point to by driven to from table 'table'
  int id;
  //! name of table that contains the id
  char table[80];
  //! name of the column from table 'table' that contains the distance
  char distance[80];
  //! constraint on the distance
  distConstraint constraint;
};

//! command to explore while driving a path
/*! This class will drive to a given location while
    trying to populate the world model as it drives
 */
class VehMobPLCmdExplore: public RCS_CMD_MSG
{
public:
  //! constructor
  VehMobPLCmdExplore():RCS_CMD_MSG
  ( VEH_MOB_PL_CMD_EXPLORE_TYPE,
    sizeof( VehMobPLCmdExplore )) {};
  //! standard NML update function
  void update( CMS* );

  //! waypoint in local coordinates, units [m]
  PM_CARTESIAN p;
  //! neighborhood within which the waypoint is deemed attained, units [m]
  double neighborhood;
  //! is the waypoint in global coordinates?
  bool isGlobal;
  //! Time to start movement (in seconds since the Epoch (00:00:00 UTC, January 1, 1970)
  long startTime;
};


//! command to explore and search for victims while driving a path
/*! This class will drive to a given location while
    trying to populate the world model as it drives
    if a victim is located, an attempt is made to identify the victim and
    his/her status before continuing to the original goal
 */
class VehMobPLCmdVictLocate: public RCS_CMD_MSG
{
public:
  //! constructor
  VehMobPLCmdVictLocate():RCS_CMD_MSG
  ( VEH_MOB_PL_CMD_VICT_LOCATE_TYPE,
    sizeof( VehMobPLCmdVictLocate )) {};
  //! standard NML update function
  void update( CMS* );

  //! waypoint in local coordinates, units [m]
  PM_CARTESIAN p;
  //! neighborhood within which the waypoint is deemed attained, units [m]
  double neighborhood;
  //! is the waypoint in global coordinates?
  bool isGlobal;
  //! Time to start movement (in seconds since the Epoch (00:00:00 UTC, January 1, 1970)
  long startTime;
};

class VehMobPLCmdLoadWM:public RCS_CMD_MSG {
public:
  VehMobPLCmdLoadWM():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_LOAD_WM_TYPE,
     sizeof(VehMobPLCmdLoadWM)) {};
  void update(CMS *);
  char fileName[MOAST_FILE_NAME_LEN];
};

class VehMobPLCmdDock:public RCS_CMD_MSG {
public:
  VehMobPLCmdDock():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_DOCK_TYPE,
     sizeof(VehMobPLCmdDock)) {};
  void update(CMS *);
};

class VehMobPLCmdUndock:public RCS_CMD_MSG {
public:
  VehMobPLCmdUndock():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_UNDOCK_TYPE,
     sizeof(VehMobPLCmdUndock)) {};
  void update(CMS *);
};

class VehMobPLCmdLoadPkg:public RCS_CMD_MSG {
public:
  VehMobPLCmdLoadPkg():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_LOAD_PKG_TYPE,
     sizeof(VehMobPLCmdLoadPkg)) {};
  void update(CMS *);
};

class VehMobPLCmdUnloadPkg:public RCS_CMD_MSG {
public:
  VehMobPLCmdUnloadPkg():RCS_CMD_MSG
    (VEH_MOB_PL_CMD_UNLOAD_PKG_TYPE,
     sizeof(VehMobPLCmdUnloadPkg)) {};
  void update(CMS *);
};

/*!
  Status messages
*/
//! general status class
class VehMobPLStat:public RCS_STAT_MSG_V2 {
public:
  VehMobPLStat():RCS_STAT_MSG_V2(VEH_MOB_PL_STAT_TYPE, sizeof(VehMobPLStat)) {};
  void update(CMS *);

  //! cost of last plan generated
  double planCost;
  //! length of last plan generated
  int planLength;
  //! plan start horizon; the distance from the vehicle where the plan begins
  double planStartHorizon;
  char message[80];
};

//! class for manufacturing status
class VehManufactStat:public RCS_STAT_MSG_V2 {
public:
  VehManufactStat():RCS_STAT_MSG_V2(VEH_MANUFACT_STAT_TYPE, sizeof(VehManufactStat)) {};
  void update(CMS *);
  //! where are we going?
  char vehDestination[MAX_WAYPOINT_NAME_LENGTH];
  //! where is the package going? Only valid if hasPackage is 1
  char packageDestination[MAX_WAYPOINT_NAME_LENGTH];
  //! do we have a package on board?
  int hasPackage;
  //! cost of last plan generated
  double planCost;
  //! length of last plan generated
  int planLength;
  //! plan start horizon; the distance from the vehicle where the plan begins
  double planStartHorizon;
  char message[80];
};

/*!
  Commands that will come into the configuration channel
*/
class VehMobPLCfgCycleTime:public RCS_CMD_MSG {
public:
  VehMobPLCfgCycleTime():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_CYCLE_TIME_TYPE,
     sizeof(VehMobPLCfgCycleTime)) {};
  void update(CMS *);

  double cycleTime;
};

class VehMobPLCfgNop:public RCS_CMD_MSG {
public:
  VehMobPLCfgNop():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_NOP_TYPE,
     sizeof(VehMobPLCfgNop)) {};
  void update(CMS *);
};

class VehMobPLCfgMove:public RCS_CMD_MSG {
public:
  VehMobPLCfgMove():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_MOVE_TYPE,
     sizeof(VehMobPLCfgMove)) {};
  void update(CMS *);

  //! include a buffer (in meters) around obstacles
  double obstaclePad;
  //! discount the previous path to add stability
  bool usePathDiscount;
};

#define VEH_MOB_PL_DEBUG_MOVE 0x00000001

class VehMobPLCfgDebug:public RCS_CMD_MSG {
public:
  VehMobPLCfgDebug():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_DEBUG_TYPE,
     sizeof(VehMobPLCfgDebug)) {};
  void update(CMS *);

  int debug;			// debug mask
};

class VehMobPLCfgClearWM:public RCS_CMD_MSG {
public:
  VehMobPLCfgClearWM():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_CLEAR_WM_TYPE,
     sizeof(VehMobPLCfgClearWM)) {};
  void update(CMS *);
};

class VehMobPLCfgStoreWM:public RCS_CMD_MSG {
public:
  VehMobPLCfgStoreWM():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_STORE_WM_TYPE,
     sizeof(VehMobPLCfgStoreWM)) {};
  void update(CMS *);
  char fileName[MOAST_FILE_NAME_LEN];
};

class VehMobPLCfgDumpWM:public RCS_CMD_MSG {
public:
  VehMobPLCfgDumpWM():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_DUMP_WM_TYPE,
     sizeof(VehMobPLCfgDumpWM)) {};
  void update(CMS *);
  char fileName[MOAST_FILE_NAME_LEN];
  //! dump the world model continuously
  bool dumpContinuous;
  //! skip 'skip' cycles between continuous dumps
  int skip;
};

/*! If sql support is compiled in, this command
  sets the server parameters. If sql is not used,
  this command is meaningless
  */
class VehMobPLCfgSQL:public RCS_CMD_MSG {
public:
  VehMobPLCfgSQL():RCS_CMD_MSG
    (VEH_MOB_PL_CFG_SQL_TYPE,
     sizeof(VehMobPLCfgSQL)) {};
  void update(CMS *);

  //! name of sql host
  char server[MOAST_FILE_NAME_LEN];
  //! name of sql user
  char user[MOAST_FILE_NAME_LEN];
  //! name of sql database
  char database[MOAST_FILE_NAME_LEN];
  //! password for sql user
  char password[MOAST_FILE_NAME_LEN];
};

/*!
  VehMobPLSet is the class that holds the quasi-static vehicle settings,
  those that change infrequently and in response to vehicle configuration
  messages.
 */
class VehMobPLSet:public RCS_STAT_MSG {
public:
  VehMobPLSet():RCS_STAT_MSG(VEH_MOB_PL_SET_TYPE, sizeof(VehMobPLSet)) {};
  void update(CMS *);
  
  //! the desired cycleTime of the planner
  double cycleTime;
  //! how many cells should obstacles be padded by
  double obstaclePad;
  //! should previous paths be discounted
  bool usePathDiscount;
  //! debug mask
  int debug;			
  //! has the planner been initialized
  bool isInitialized;
  //! is the planner using sql tables
  bool isUsingSQL;
  //! how far down path to use in computing path confidence
  int confDistance;
  //! the resolution of a planning cell in meters
  double cellResolution;
  //! if path cost exceeds this, then confidence is 0
  double zeroConf;
  //! if path cost exceeds this but not zeroConf, then confidence is 5
  double fiveConf;
  //! file name that world model will be read/written from/to
  char wmFileName[MOAST_FILE_NAME_LEN];
  //! when dumping a file, dump every xth world model snap shot
  int fileSkip;
  //! name of sql host
  char sqlServer[MOAST_FILE_NAME_LEN];
  //! name of sql user
  char sqlUser[MOAST_FILE_NAME_LEN];
  //! name of sql database
  char sqlDatabase[MOAST_FILE_NAME_LEN];
  //! are we using the sql database
  bool sqlConnected;
};

extern int vehMobPL_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * vehMobPL_symbol_lookup(long type);

#endif // VEH_MOB_PL_HH

/*
  Modification history:

  $Log: vehMobPL.hh,v $
  Revision 1.18  2007/12/18 18:25:21  dr_steveb
  Added package loading and unloading commands

  Revision 1.17  2007/12/17 21:37:03  dr_steveb
  Added section level manufacturing application

  Revision 1.16  2007/12/13 13:17:26  dr_steveb
  Changes for vehicle level visibility graph planning.

  Revision 1.15  2007/05/16 15:16:21  dr_steveb
  Fixed location of state printing routines to remove compile warnings.

  Revision 1.14  2007/05/02 02:59:46  tkharris
  cleaned up rcsVehStateToString

  Revision 1.13  2007/02/27 14:51:54  dr_steveb
  Changed version number, added concave hull tools

  Revision 1.12  2007/02/01 22:19:22  proctor
  Took out backslashes from rcsVehStateToString macro so that it's all on one
  line, making the RCS code generator happy.

  Revision 1.11  2007/01/30 21:29:02  dr_steveb
  Added new AM-VEH interface

  Revision 1.10  2006/12/13 20:33:46  dr_steveb
  Changed range of victRFID sensor

  Revision 1.9  2006/11/28 19:06:31  dr_steveb
  Improved "done" criteria in movement commands.

  Revision 1.8  2006/10/30 12:59:34  dr_steveb
  Files added for connection to SPQR

  Revision 1.7  2006/10/18 00:14:37  dr_steveb
  Initial add of CImg.h which comes from sourceforge.

  Revision 1.6  2006/10/03 20:24:25  dr_steveb
  Exploration command added to vehicle level

  Revision 1.5  2006/09/30 13:08:49  dr_steveb
  Changed heartbeat to unsigned int and autogenerated .cc files

  Revision 1.4  2006/04/05 01:58:31  dr_steveb
  Added NOP command

  Revision 1.3  2005/10/21 12:45:24  dr_steveb
  Added fields to status messages.

  Revision 1.2  2005/10/17 18:33:34  dr_steveb
  Added plan length to status message.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.7  2005/10/11 14:45:18  stephen
  Added plan cost to status message

  Revision 1.6  2005/06/20 19:34:09  stephen
  Incremental changes to messages.

  Revision 1.5  2005/06/15 13:38:53  stephen
  Added mysql support.

  Revision 1.4  2005/06/14 19:18:36  stephen
  Changed location and declaration of SensorPoint.

  Revision 1.3  2005/06/09 16:43:15  stephen
  Added information to the settings message.

  Revision 1.2  2005/06/08 20:49:36  stephen
  Changed occupancyData to sensorData

  Revision 1.1  2005/06/08 14:29:54  proctor
  Added vehMobPLMain name change

  Revision 1.1  2005/06/08 12:47:41  proctor
  Name change per Mis,MobJA naming convention

  Revision 1.6  2005/06/06 12:42:23  stephen
  Moved sql parameters out of init call.

  Revision 1.5  2005/06/02 12:56:07  stephen
  Made waypoint movement only use 1 waypoint.

  Revision 1.3  2005/06/01 12:45:45  stephen
  Added commands for vehicle level ied mission.

  Revision 1.2  2005/05/23 14:15:54  stephen
  Added teleport command.

  Revision 1.1  2005/05/18 12:45:44  stephen
  Initial check-in.

*/
