/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file amShell.cc

  \brief Command line shell to vehicle AM level NML buffers.

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.22 $
  $Date: 2008/02/12 00:01:39 $
  \endcode

  \author Fred Proctor
*/

/***********************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "primMobJA.hh"		// PRIM_MOB_JA_CMD_ROTATE_CW, etc.
#include "amMobJA.hh"		// amMobJA_format(), commands, etc.
#include "amSP.hh"              // amSP_format()
#include "navDataExt.hh"

/***********************************************************************/

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"
#define THIS_PROCESS "amShell"
#define MOB 1
#define SP  2
#define MIS 3
#define ADM 4

#define DEG(x) (57.295779513082323 * (x))
#define RAD(x)  (0.017453292519943 * (x))

/***********************************************************************/

static NML *navDataExtBuf = NULL;

static RCS_CMD_CHANNEL * amMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL * amMobJAStatBuf = NULL;
static RCS_CMD_CHANNEL * amMobJACfgBuf = NULL;
static RCS_STAT_CHANNEL * amMobJASetBuf = NULL;

static RCS_CMD_CHANNEL * amSPCmdBuf = NULL;
static RCS_STAT_CHANNEL * amSPStatBuf = NULL;
static RCS_CMD_CHANNEL * amSPCfgBuf = NULL;
static RCS_STAT_CHANNEL * amSPSetBuf = NULL;

static AmMobJAStat * amMobJAStat;
static AmMobJASet * amMobJASet;

static AmMobJACmdInit amMobJACmdInit;
static AmMobJACmdHalt amMobJACmdHalt;
static AmMobJACmdAbort amMobJACmdAbort;
static AmMobJACmdShutdown amMobJACmdShutdown;
static AmMobJACmdNop amMobJACmdNop;
static AmMobJACmdSpin amMobJACmdSpin;
static AmMobJACmdPrepareMove amMobJACmdPrepareMove;
static AmMobJACmdMoveWaypoint amMobJACmdMoveWaypoint;
static AmMobJACfgCellResolution amMobJACfgCellResolution;
static AmMobJACfgCycleTime amMobJACfgCycleTime;
static AmMobJACfgDebug amMobJACfgDebug;
static AmMobJACfgDumpWM amMobJACfgDumpWM;
static AmMobJACfgPlanHorizon amMobJACfgPlanHorizon;
static AmMobJACfgVehicleMinTurnRad amMobJACfgVehicleMinTurnRad;
static AmMobJACfgVehicleWidth amMobJACfgVehicleWidth;
static AmMobJACfgNop amMobJACfgNop;

static AmSPStat * amSPStat;
static AmSPSet * amSPSet;

static AmSPCmdInit amSPCmdInit;
static AmSPCmdHalt amSPCmdHalt;
static AmSPCmdAbort amSPCmdAbort;
static AmSPCmdShutdown amSPCmdShutdown;
static AmSPCmdGo amSPCmdGo;
static AmSPCmdNop amSPCmdNop;
static AmSPCfgCycleTime amSPCfgCycleTime;
static AmSPCfgDebug amSPCfgDebug;
static AmSPCfgDumpWM amSPCfgDumpWM;
static AmSPCfgNop amSPCfgNop;

static char *INI_FILE = NULL;
static char *NML_FILE = NULL;

/***********************************************************************/

static void convertAbsToRel(double x, double y, double * xx, double * yy);
static void doAdmCmd(char * ptr);
static void doCmdMove(char * ptr);
static void doCmdWaypoint(char * ptr);
static void doCmdWaypointPlan(char * ptr);
static void doMobCmd(char * ptr);
static void doSPCmd(char * ptr);
static void getIniFile();
static void getMobError(char * mess);
static void getNmlFile();
static int getOptions(int argc, char * argv[], int * bufnum);
static char * getUserInput(char * buffer, char ** line, int seeWhat, int size);
static int initAmNmlBuffers(char *config_file, int bufnum);
static int loadWaypoints(char * filename, Waypoint * waypoint,
			 int * waypoint_length);
static void printAdmHelp();
static void printAdmStat();
static void printMisStat();
static void printMobHelp();
static void printMobStat();
static void printSPHelp();
static void printSPStat();
static const char * rcs_stat_to_string(int s);
static void setOddSerial(int echoed, int * serial);
static int strwcmp(char *test, char *match);
static void writeMobCfg(RCS_CMD_MSG * cfg);
static void writeMobCmd(RCS_CMD_MSG * cmd);
static void writeSPCfg(RCS_CMD_MSG * cfg);
static void writeSPCmd(RCS_CMD_MSG * cmd);

/***********************************************************************/

//! convertAbsToRel converts from absolute to relative coordinates
/*
Returned Value: none

Called By: loadWaypoints

This sets *xx and *yy to the relative coordinates corresponding to
the absolute coordinates x and y.

*/

static void convertAbsToRel( /*   ARGUMENTS                   */
 double x,                   /*!< x value to be converted     */
 double y,                   /*!< y value to be converted     */
 double * xx,                /*!< converted x value, set here */
 double * yy)                /*!< converted y value, set here */
{
  PM_CARTESIAN tran;

  tran = ((NavDataExt *)
	  (navDataExtBuf->get_address()))->absToRel * PM_CARTESIAN(x, y, 0);
  *xx = tran.x;
  *yy = tran.y;
  return;
}

/***********************************************************************/

/*! doAdmCmd either (1) carries out a user administrative command type
    that is directed to both AM mobility and AM SP, or (2) prints
    help for administrative commands. */
/*
Returned Value: none

Called By: main

There is no leading white space in ptr when this is called.

*/

static void doAdmCmd( /*   ARGUMENTS                   */
 char * ptr)          /*!< pointer into command string */
{
  if (ptr[0] == '?' || !strwcmp(ptr, "help"))
    {
      printAdmHelp();
    }
  else if (!strwcmp(ptr, "init"))
    {
      writeMobCmd(&amMobJACmdInit);
      writeSPCmd(&amSPCmdInit);
    }
  else if (!strwcmp(ptr, "halt"))
    {
      writeMobCmd(&amMobJACmdHalt);
      writeSPCmd(&amSPCmdHalt);
    }
  else if (!strwcmp(ptr, "abort"))
    {
      writeMobCmd(&amMobJACmdAbort);
      writeSPCmd(&amSPCmdAbort);
    }
  else if (!strwcmp(ptr, "shutdown"))
    {
      writeMobCmd(&amMobJACmdShutdown);
      writeSPCmd(&amSPCmdShutdown);
    }
  else
    {
      printf("?\n");
    }
}

/***********************************************************************/

/*! doCmdMove checks a move command given by the user, builds a
    moveWaypoint command, and sends the command to AM Mob. */
/*
Returned Value: none

Called By: doCmd

This requires that the user provide at least two doubles (x and y
coordinates) as arguments. If there is a third double argument,
it is used as the neighborhood. If there is a fourth double argument,
it is used as the speed. If no neighborhood argument is provided,
0.2 is used. If no speed argument is provided, 1.0 is used.

*/

static void doCmdMove( /*   ARGUMENTS                   */
 char * ptr)           /*!< pointer into command string */
{
  double x;
  double y;
  double neighborhood;
  double speed;
  int argc;
  
  argc = sscanf(ptr, "%lf %lf %lf %lf", &x, &y, &neighborhood, &speed);
  if (argc < 2)
    {
      printf("need x and y arguments\n");
      return;
    }
  if (argc < 4)
    speed = 0.7;
  if (argc < 3)
    neighborhood = 0.5;
  if (neighborhood < 0)
    {
      printf("neighborhood must not be negative\n");
      return;
    }
  if (speed <= 0)
    {
      printf("speed must be positive\n");
      return;
    }
  amMobJACmdMoveWaypoint.waypoint[0].p.x = x;
  amMobJACmdMoveWaypoint.waypoint[0].p.y = y;
  amMobJACmdMoveWaypoint.waypoint[0].p.z = 0;
  amMobJACmdMoveWaypoint.waypoint[0].neighborhood = neighborhood;
  amMobJACmdMoveWaypoint.waypoint[0].speed = speed;
  amMobJACmdMoveWaypoint.waypoint_length = 1;
  amMobJACmdMoveWaypoint.computeHorizonCosts =  false;
  writeMobCmd(&amMobJACmdMoveWaypoint);
}

/***********************************************************************/

/*! doCmdWaypoint checks a move waypoint command given by the user
    that does not include planning, reads the waypoint file named
    in the command, builds a moveWaypoint command, and sends the
    command to AM Mob. */
/*
Returned Value: none

Called By: doCmd

This is allowing white space (excluding trailing and leading white space)
in the file name, so it is not feasible to use sscanf on the file name.

There may be a newline after the file name. If there is one, this
replaces it with a 0 so that the newline is not taken to be part of the name.

The ptr is set to the character after the command name when this starts up.

*/

static void doCmdWaypoint( /*   ARGUMENTS                   */
 char * ptr)               /*!< pointer into command string */
{
  int n;

  while (isspace(*ptr))
    ptr++;
  if (*ptr == 0)
    {
      fprintf(stderr, "need a waypoint file name\n");
      return;
    }
  for (n = 0; ptr[n] != 0; n++)
    {
      if (ptr[n] == '\n')
	{ // if there is a newline, delete it
	  ptr[n] = 0;
	  break;
	}
    }
  for (n--; ((ptr[n] == ' ') || (ptr[n] == '\t')); n--)
    { // delete trailing white space, if any
      ptr[n] = 0;
    }
  if (loadWaypoints(ptr, amMobJACmdMoveWaypoint.waypoint,
		    &amMobJACmdMoveWaypoint.waypoint_length))
    {
      printf("can't load ``%s''\n", ptr);
      return;
    }
  amMobJACmdMoveWaypoint.computeHorizonCosts =  false;
  amMobJACmdMoveWaypoint.neighborhood = 1.0; // placeholder
  amMobJACmdMoveWaypoint.speed = 1.0; // placeholder
  writeMobCmd(&amMobJACmdMoveWaypoint);
}

/***********************************************************************/

/*! doCmdWaypointPlan checks a plan and move waypoint command given by
    the user, reads the waypoint file named in the command, builds a
    moveWaypoint command, and sends the command to AM Mob. */
/*
Returned Value: none

Called By: doCmd

This is allowing white space (excluding trailing and leading white space)
in the file name, so it is not feasible to use sscanf on the file name.

There may be a newline after the file name. If there is one, this
replaces it with a 0 so that the newline is not taken to be part of the name.

The ptr is set to the character after the command name when this starts up.

*/

static void doCmdWaypointPlan( /*   ARGUMENTS                   */
 char * ptr)                   /*!< pointer into command string */
{
  int n;
  int k;
  
  if (2 != sscanf(ptr, "%lf %lf%n", &amMobJACmdMoveWaypoint.neighborhood,
		  &amMobJACmdMoveWaypoint.speed, &k))
    {
      printf("need neighborhood and speed arguments\n");
      return;
    }
  if (amMobJACmdMoveWaypoint.neighborhood < 0)
    {
      printf("neighborhood must not be negative\n");
      return;
    }
  if (amMobJACmdMoveWaypoint.speed <= 0)
    {
      printf("speed must be positive\n");
      return;
    }
  ptr = (ptr + k);
  while (isspace(*ptr))
    ptr++;
  if (*ptr == 0)
    {
      fprintf(stderr, "need a waypoint file name\n");
      return;
    }
  for (n = 0; ptr[n] != 0; n++)
    {
      if (ptr[n] == '\n')
	{ // if there is a newline, delete it
	  ptr[n] = 0;
	  break;
	}
    }
  for (n--; ((ptr[n] == ' ') || (ptr[n] == '\t')); n--)
    { // delete trailing white space, if any
      ptr[n] = 0;
    }
  if (loadWaypoints(ptr, amMobJACmdMoveWaypoint.waypoint,
		    &amMobJACmdMoveWaypoint.waypoint_length))
    {
      printf("can't load ``%s''\n", ptr);
      return;
    }
  amMobJACmdMoveWaypoint.computeHorizonCosts =  true;
  writeMobCmd(&amMobJACmdMoveWaypoint);
}

/***********************************************************************/

//! doMobCmd carries out a user command to AM mobility
/*
Returned Value: none

Called By: main

There is no leading white space in ptr when this is called.

*/

static void doMobCmd( /*   ARGUMENTS                   */
 char * ptr)          /*!< pointer into command string */
{
  int dumpVal;
  char dir[100];
  int way;

  if (ptr[0] == '?' || !strwcmp(ptr, "help"))
    {
      printMobHelp();
    }
  else if (!strwcmp(ptr, "init"))
    {
      writeMobCmd(&amMobJACmdInit);
    }
  else if (!strwcmp(ptr, "halt"))
    {
      writeMobCmd(&amMobJACmdHalt);
    }
  else if (!strwcmp(ptr, "abort"))
    {
      writeMobCmd(&amMobJACmdAbort);
    }
  else if (!strwcmp(ptr, "shutdown"))
    {
      writeMobCmd(&amMobJACmdShutdown);
    }
  else if (!strwcmp(ptr, "nop"))
    {
      writeMobCmd(&amMobJACmdNop);
    }
  else if (!strwcmp(ptr, "spin"))
    {
      if(3 != sscanf(ptr, "%*s %s %lf %lf",
		     dir, &amMobJACmdSpin.absAngle, &amMobJACmdSpin.tolerance))
	{
	  printf("need direction, angle, and tolerance arguments\n");
	}
      else
	{
	  way = ((strcmp(dir,"CW") == 0) ? MOAST_ROTATE_CW :
		 (strcmp(dir,"CCW") == 0) ? MOAST_ROTATE_CCW :
		 (strcmp(dir,"SHORT") == 0) ? MOAST_ROTATE_SHORTEST :
		 -1);
	  if (way == -1)
	    {
	      printf("direction must be CW, CCW or SHORT\n");
	    }
	  else
	    {
	      amMobJACmdSpin.direction = way;
	      amMobJACmdSpin.absAngle = RAD(amMobJACmdSpin.absAngle);
	      amMobJACmdSpin.tolerance = RAD(amMobJACmdSpin.tolerance);
	      writeMobCmd(&amMobJACmdSpin);
	    }
	}
    }
  else if (!strwcmp(ptr, "plan"))
    {
      if (2 != sscanf(ptr, "%*s %lf %lf",
		      &amMobJACmdPrepareMove.neighborhood,
		      &amMobJACmdPrepareMove.speed))
	{
	  printf("need neighborhood and speed arguments\n");
	}
      else if (amMobJACmdPrepareMove.neighborhood < 0)
	{
	  printf("neighborhood must not be negative\n");
	}
      else if (amMobJACmdPrepareMove.speed <= 0)
	{
	  printf("speed must be positive\n");
	}
      else
	{
	  writeMobCmd(&amMobJACmdPrepareMove);
	}
    }
  else if (!strwcmp(ptr, "move"))
    {
      doCmdMove(ptr + 4);
    }
  else if (!strwcmp(ptr, "wp"))
    {
      doCmdWaypoint(ptr + 2);
    }
  else if (!strwcmp(ptr, "wpl"))
    {
      doCmdWaypointPlan(ptr + 3);
    }
  else if (!strwcmp(ptr, "cellRes"))
    {
      if (1 != sscanf(ptr, "%*s %lf",
		      &amMobJACfgCellResolution.cellResolution))
	{
	  printf("need a cell resolution argument\n");
	}
      else
	{
	  writeMobCfg(&amMobJACfgCellResolution);
	}
    }
  else if (!strwcmp(ptr, "ct"))
    {
      if (1 != sscanf(ptr, "%*s %lf", &amMobJACfgCycleTime.cycleTime))
	{
	  printf("need a cycleTime argument\n");
	}
      else
	{
	  writeMobCfg(&amMobJACfgCycleTime);
	}
    }
  else if (!strwcmp(ptr, "debug"))
    {
      if (1 != sscanf(ptr, "%*s %i", &amMobJACfgDebug.debug))
	{
	  printf("need debug mask\n");
	}
      else
	{
	  writeMobCfg(&amMobJACfgDebug);
	}
    }
  else if (!strwcmp(ptr, "horizon"))
    {
      if (1 != sscanf(ptr, "%*s %lf", &amMobJACfgPlanHorizon.planHorizon))
	{
	  printf("need a plan horizon\n");
	}
      else
	{
	  writeMobCfg(&amMobJACfgPlanHorizon);
	}
    }
  else if (!strwcmp(ptr, "vehWidth"))
    {
      if (1 != sscanf(ptr, "%*s %lf", &amMobJACfgVehicleWidth.vehicleWidth))
	{
	  printf("need a vehicle width argument\n");
	}
      else
	{
	  writeMobCfg(&amMobJACfgVehicleWidth);
	} 
    }
  else if (!strwcmp(ptr, "turnRad"))
    {
      if (1 != sscanf(ptr, "%*s %lf",
		      &amMobJACfgVehicleMinTurnRad.vehicleMinTurnRad))
	{
	  printf("need a vehicle minimum turning radius argument\n");
	}
      else
	{
	  writeMobCfg(&amMobJACfgVehicleMinTurnRad);
	} 
    }
  else if (!strwcmp(ptr, "dump"))
    {
      if (1 != sscanf(ptr, "%*s %d", &dumpVal))
	{
	  printf("need 1 for on or 0 for off\n");
	}
      else if (dumpVal == 0)
	{
	  amMobJACfgDumpWM.turnOn = false;
	  writeMobCfg(&amMobJACfgDumpWM);
	}
      else if (dumpVal == 1)
	{
	  amMobJACfgDumpWM.turnOn = true;
	  writeMobCfg(&amMobJACfgDumpWM);
	}
      else
	{
	  printf("need 1 for on or 0 for off\n");
	}
    }
  else if (!strwcmp(ptr, "go"))
    {
      printf("must be in \"sp\" mode\n");
    }
  else
    {
      printf("?\n");
    }
}

/***********************************************************************/

/* doSPCmd carries out a user command to AM sensory processing

Returned Value: none

Called By: main

There is no leading white space in ptr when this is called.

*/

static void doSPCmd( /*   ARGUMENTS                   */
 char * ptr)         /*!< pointer into command string */
{
  int dumpVal;

  if (ptr[0] == '?' || !strwcmp(ptr, "help"))
    {
      printSPHelp();
    }
  else if (!strwcmp(ptr, "init"))
    {
      writeSPCmd(&amSPCmdInit);
    }
  else if (!strwcmp(ptr, "halt"))
    {
      writeSPCmd(&amSPCmdHalt);
    }
  else if (!strwcmp(ptr, "abort"))
    {
      writeSPCmd(&amSPCmdAbort);
    }
  else if (!strwcmp(ptr, "shutdown"))
    {
      writeSPCmd(&amSPCmdShutdown);
    }
  else if (!strwcmp(ptr, "nop"))
    {
      writeSPCmd(&amSPCmdNop);
    }
  else if (!strwcmp(ptr, "go"))
    {
      writeSPCmd(&amSPCmdGo);
    }
  else if (!strwcmp(ptr, "ct"))
    {
      if (1 != sscanf(ptr, "%*s %lf", &amSPCfgCycleTime.cycleTime))
	{
	  printf("need a cycleTime argument\n");
	}
      else
	{
	  writeSPCfg(&amSPCfgCycleTime);
	}
    }
  else if (!strwcmp(ptr, "debug"))
    {
      if (1 != sscanf(ptr, "%*s %i", &amSPCfgDebug.debug))
	{
	  printf("need debug mask\n");
	}
      else
	{
	  writeSPCfg(&amSPCfgDebug);
	}
    }
  else if (!strwcmp(ptr, "dump"))
    {
      if (1 != sscanf(ptr, "%*s %d", &dumpVal))
	{
	  printf("need dump value of 0 (off), 1 (height), or 2(obstacles)\n");
	}
      else if ((dumpVal < 0) || (dumpVal > 2))
	{
	  printf("dump value must be 0, 1, or 2\n");
	}
      else
	{
	  amSPCfgDumpWM.level = dumpVal;
	  writeSPCfg(&amSPCfgDumpWM);
	}
    }
  else if ((!strwcmp(ptr, "wp")) ||
	   (!strwcmp(ptr, "wpl")) ||
	   (!strwcmp(ptr, "plan")) ||
	   (!strwcmp(ptr, "cellRes")) ||
	   (!strwcmp(ptr, "horizon")) ||
	   (!strwcmp(ptr, "turnRad")) ||
	   (!strwcmp(ptr, "vehicleWidth")))
    {
      printf("must be in \"mob\" mode\n");
    }
  else
    {
      printf("?\n");
    }
}

/***********************************************************************/

//! getIniFile gets the name of the ini file
/*
Returned Value: none

Called By: main

This gets the ini file name from arg, if provided, next environment,
if there, else default. A pointer to the name is stored in the
global variable INI_FILE.

*/

static void getIniFile() /* NO ARGUMENTS */
{
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;

  if (NULL == INI_FILE)
    {
      ini_file_env = getenv("CONFIG_INI");
      if (NULL != ini_file_env)
	{
	  INI_FILE = ini_file_env;
	}
      else
	{
	  INI_FILE = ini_file_default;
	}
    }
}

/***********************************************************************/

//! getMobError writes into a string details about errors reported by mob.
/*
Returned Value: none

Called By: printMobStat

*/

static void getMobError( /* ARGUMENTS */
 char * mess)
{
  if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_MOVE)
    sprintf(mess, "failed to plan to waypoints");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_NONE)
    sprintf(mess, "no error");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_NOT_INIT)
    sprintf(mess, "not initialized");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_PLAN_HORIZON)
    sprintf(mess, "failed to plan to planning horizon");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_PLAN_PRIM)
    sprintf(mess, "failed to make plan to send to prim");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_POSITION)
    sprintf(mess, "failed to update position");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_PRIM)
    sprintf(mess, "prim reported error or sent meaningless status message");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_UNSPECIFIED)
    sprintf(mess, "unspecified error");
  else if (amMobJAStat->amMobJAError == AM_MOB_JA_ERROR_WAYPOINT)
    {
      if (amMobJAStat->seed_length == 0)
	sprintf(mess, "too many waypoints");
      else if (amMobJAStat->seed_length == 1)
	sprintf(mess, "bad waypoint location (%.2lf, %.2lf)",
		amMobJAStat->seed[0].x, amMobJAStat->seed[0].y);
      else if (amMobJAStat->seed_length == 2)
	sprintf(mess, "too few waypoints; must be at least 1");
      else if (amMobJAStat->seed_length == 3)
	sprintf(mess, "waypoints (%.2lf, %.2lf) and (%.2lf, %.2lf) too close",
		amMobJAStat->seed[0].x, amMobJAStat->seed[0].y,
		amMobJAStat->seed[1].x, amMobJAStat->seed[1].y);
      else
	sprintf(mess, "unspecified waypoint problem");
    }
  else
    sprintf(mess, "unknown error type");
}

/***********************************************************************/

//! getNmlFile gets the name of the NML file
/*
Returned Value: none

Called By: main

This gets the NML file name from arg, if provided, next environment,
if there, else default.  A pointer to the name is stored in the
global variable NML_FILE.

*/

static void getNmlFile() /* NO ARGUMENTS */
{
  char *nml_file_env;
  char nml_file_default[] = DEFAULT_NML_FILE;

  if (NULL == NML_FILE)
    {
      nml_file_env = getenv("CONFIG_NML");
      if (NULL != nml_file_env)
	{
	  NML_FILE = nml_file_env;
	}
      else
	{
	  NML_FILE = nml_file_default;
	}
    }
}

/***********************************************************************/

//! getOptions gets the options given in the command that starts amShell.
/*
Returned Value: int
  If any of the following errors occurs, this returns 1.
  Otherwise, it returns 0.
  1. A buffer number cannot be read.
  2. an unknown option is used.

Called By: main

*/

static int getOptions( /*   ARGUMENTS                                      */
 int argc,             /*!< one more than number of command line arguments */ 
 char * argv[],        /*!< array of executable name and command line args */
 int * bufnum)         /*!< buffer number, may be set here                 */
{
  while (1)
    {
      char option = getopt(argc, argv, "n:i:d:");
      if (option == -1)
	break;
      
      switch (option)
	{
	case 'n':
	  NML_FILE = optarg;
	  break;
	case 'i':
	  INI_FILE = optarg;
	  break;
	case 'd':
	  if (1 != sscanf(optarg, "%i", bufnum))
	    {
	      rcs_print_error("%s: bad value for buffer number: %s\n",
			      argv[0], optarg);
	      return 1;
	    }
	  break;
	default:
	  rcs_print_error("%s: bad argument\n", argv[0]);
	  return 1;
	  break;
	}
    }
  return 0;
}

/***********************************************************************/

/*! getUserInput waits for the user to type something on the keyboard
   and stores the string the user types in the "line" argument. */
/*
Returned Value: char * (the string the use typed)

Called By: main

Without the fflush(stdin), an extra <enter> is required to make a
command run. It is not clear why this is.

The function runs differently according to whether the OS for which
the code is compiled has "readline" capabilities.

*/

static char * getUserInput( /*   ARGUMENTS                                */
 char * buffer,             /*!< buffer for storing user's string         */
 char ** line,              /*!< string typed by user, set here           */
 int seeWhat,               /*!< MOB=mobility SP=SP MIS=mission ADM=admin */
 int size)                  /*!< buffer size (for fgets)                  */
{
#ifndef HAVE_READLINE_READLINE_H
  printf("%s> ", ((seeWhat == MOB) ? "Mob" :
		  (seeWhat == MIS) ? "Mis" :
		  (seeWhat == ADM) ? "Adm" : "SP"));
  fflush(stdout);
  if (NULL == fgets(buffer, size, stdin))
    {
      return 0;
    }
  fflush(stdin);
  *line = buffer;
  return buffer;
#else
  *line = readline((seeWhat == MOB) ? "Mob> " :
                   (seeWhat == MIS) ? "Mis> " :
		   (seeWhat == ADM) ? "Adm> " : "SP> ");
  if ((*line) && (**line))
    {
      add_history(*line);
    }
  fflush(stdin);
  return *line;
#endif
}

/***********************************************************************/

//! initAmNmlBuffers creates the NML buffers used by amShell
/*
Returned Value: int
  If any of the following errors occurs, this returns 1.
  Otherwise, it returns 0.
  1. The navDataExtBuf is not valid when created.
  2. The amMobJACmdBuf is not valid when created.
  3. The amMobJAStatBuf is not valid when created.
  4. The amMobJACfgBuf is not valid when created.
  5. The amMobJASetBuf is not valid when created.
  6. The amSPCmdBuf is not valid when created.
  7. The amSPStatBuf is not valid when created.
  8. The amSPCfgBuf is not valid when created.
  9. The amSPSetBuf is not valid when created.

Called By: main

This returns 1 to main if any buffer is not valid when created,
whereupon main quits, so it is not necessary to clean up here.

*/

static int initAmNmlBuffers( /*   ARGUMENTS                      */
 char *config_file,          /*!< name of NML config file to use */
 int bufnum)                 /*!< buffer number to use           */
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
  navDataExtBuf =
    new NML(navDataExt_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_MOB_JA_CMD_NAME, bufnum);
  amMobJACmdBuf =
    new RCS_CMD_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_MOB_JA_STAT_NAME, bufnum);
  amMobJAStatBuf =
    new RCS_STAT_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_MOB_JA_CFG_NAME, bufnum);
  amMobJACfgBuf =
    new RCS_CMD_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_MOB_JA_SET_NAME, bufnum);
  amMobJASetBuf =
    new RCS_STAT_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_SP_CMD_NAME, bufnum);
  amSPCmdBuf =
    new RCS_CMD_CHANNEL(amSP_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_SP_STAT_NAME, bufnum);
  amSPStatBuf =
    new RCS_STAT_CHANNEL(amSP_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_SP_CFG_NAME, bufnum);
  amSPCfgBuf =
    new RCS_CMD_CHANNEL(amSP_format, chanName, THIS_PROCESS, config_file);
  sprintf(chanName, "%s%d", AM_SP_SET_NAME, bufnum);
  amSPSetBuf =
    new RCS_STAT_CHANNEL(amSP_format, chanName, THIS_PROCESS, config_file);
  return (((!navDataExtBuf->valid())  ||
	   (!amMobJACmdBuf->valid())  || (!amSPCmdBuf->valid())  ||
	   (!amMobJACfgBuf->valid())  || (!amSPCfgBuf->valid())  ||
	   (!amMobJAStatBuf->valid()) || (!amSPStatBuf->valid()) ||
	   (!amMobJASetBuf->valid())  || (!amSPSetBuf->valid())) ? 1 : 0);
}

/***********************************************************************/

//! loadWaypoints load waypoints from a file into a Waypoint
/*
Returned Value: int
  If any of the following errors occur, this returns 1.
  Otherwise, it returns 0.
  1. The named file cannot be opened.
  2. There is a syntax error in the file.

Called By:
  doCmdWaypoint
  doCmdWaypointPlan

The file this parses can have five types of lines in any order
1. blank lines
2. lines starting with '#', which are comments
3. lines starting with "abs". Following lines are treated as absolute coords.
4. lines starting with "rel". Following lines are treated as relative coords.
5. lines of the form <x> <y> <neighborhood> <speed> (four doubles)
   These are saved as waypoints. Z values of waypoints are set to 0.

*/

static int loadWaypoints( /*   ARGUMENTS                        */
 char * filename,         /*!< name of the file to load from    */
 Waypoint * wptr,         /*!< Waypoint into which to load data */
 int * waypoint_length)   /*!< number of points loaded          */
{
  FILE *fp;
  enum {BUFFERLEN = 256};
  char buffer[BUFFERLEN];
  char *ptr;
  int line;
  int abs;
  int retval;

  if (NULL == (fp = fopen(filename, "r")))
    return 1;

  abs = 1;			// default is absolute coords
  *waypoint_length = 0;
  line = 0;
  retval = 0;
  navDataExtBuf->read(); // needed in case convertAbsToRel is called
  while (!feof(fp)) {
    if (NULL == fgets(buffer, BUFFERLEN, fp))
      break;
    line++;
    ptr = buffer;
    while (isspace(*ptr))
      ptr++;
    if (0 == *ptr)
      continue;			// blank line
    if ('#' == *ptr)
      continue;			// comment
    if (! strncmp(ptr, "abs", 3))
      abs = 1;
    else if (! strncmp(ptr, "rel", 3))
      abs = 0;
    else if (4 == sscanf(ptr, "%lf %lf %lf %lf",
			 &wptr->p.x,
			 &wptr->p.y,
			 &wptr->neighborhood,
			 &wptr->speed))
      {
	if (abs)
	  {
	    convertAbsToRel(wptr->p.x, wptr->p.y, &wptr->p.x, &wptr->p.y);
	  }
	wptr->p.z = 0.0;
	(*waypoint_length)++;
	wptr++;
	if (*waypoint_length >= MAX_WAYPOINTS) break;
      }
    else
      {
	fprintf(stderr, "syntax error in %s, line %d: %s\n",
		filename, line, buffer);
	retval = 1;
	break;
      }
  }

  fclose(fp);
  return retval;
}

/***********************************************************************/

//! main is the main function of the amShell
/*
Returned Value: int
  If any of the following errors occurs, this returns 1.
  Otherwise, it returns 0.
  1. getOptions fails.
  2. initAmNmlBuffers fails.

Called By: not called, runs automatically

The initial readings of status information seem to get bad data, so
when this starts up, the AM mob and AM SP status buffers are read
repeatedly until the command types are in range and the echoed command
numbers are positive.

This is using odd numbers for command serial numbers. It is expected
that the Vehicle controller (or other controller sending commands to
AM Mob and AM SP) will use even numbers. Each time a new command is
given, the new serial number is always odd and larger than both the
current value of the serial number and the most recently echoed serial
number.

Previously, there was a problem with command numbering requiring that
a timer be used so that AM mobility had a chance to send status
between cycles of amShell. Without the timer, when a command, C, that
takes time to start executing (such as move waypoint) was sent, this
would ask for new user input before getting status on C.  As a result,
the new command, D, would get the same serial_number as C. If C and D
were of the same type, AM mobility would think the command was old and
do nothing. This was waiting for 0.1 second up to ten times for
amMobility to send status. The timer is no longer used (except at
startup).

*/

int main(      /*   ARGUMENTS                                           */
 int argc,     /*!< one more than number of command line arguments      */
 char *argv[]) /*!< array of executable name and command line arguments */
{
  int bufnum = 1;
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char * line;
  char * ptr;
  int seeWhat;
  RCS_TIMER * timer;

#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif
  timer = new RCS_TIMER(0.1);
  if (getOptions(argc, argv, &bufnum))
    return 1;
  getIniFile();
  getNmlFile();

  // initial allocate all the buffers
  if (initAmNmlBuffers(NML_FILE, bufnum))
    {
      fprintf(stderr, "%s: can't allocate NML buffers\n", argv[0]);
      return 1;
    }
  // set up convenient pointers to status and settings
  amMobJAStat = (AmMobJAStat *) amMobJAStatBuf->get_address();
  amMobJASet = (AmMobJASet *) amMobJASetBuf->get_address();
  amSPStat = (AmSPStat *) amSPStatBuf->get_address();
  amSPSet = (AmSPSet *) amSPSetBuf->get_address();
  for(; ; timer->wait())
    { // check repeatedly until no garbage status is read
      amMobJAStatBuf->read();
      amMobJASetBuf->read();
      amSPStatBuf->read();
      amSPSetBuf->read();
      if ((amMobJAStat->echo_serial_number >= 1) &&
	  (amMobJAStat->command_type >= AM_MOB_JA_CMD_INIT_TYPE) &&
	  (amMobJAStat->command_type <= AM_MOB_JA_CMD_SPIN_TYPE) &&
	  (amSPStat->echo_serial_number >= 1) &&
	  (amSPStat->command_type >= AM_SP_CMD_INIT_TYPE) &&
	  (amSPStat->command_type <= AM_SP_CMD_NOP_TYPE))
	break;
    }
  seeWhat = ADM;
  while (!feof(stdin))
    {
      timer->wait();
      amMobJAStatBuf->read();
      amMobJASetBuf->read();
      amSPStatBuf->read();
      amSPSetBuf->read();
#ifndef HAVE_READLINE_READLINE_H
      if (getUserInput(buffer, &line, seeWhat, BUFFERSIZE) == NULL)
#else
      if (getUserInput(NULL, &line, seeWhat, 0) == NULL)
#endif
	break;
      ptr = line;
      while (isspace(*ptr))
	ptr++;	            // skip leading white space
      if (0 == *ptr)
	{		    // user entered no command
	  if (seeWhat == MOB)
	    {
	      printMobStat();
	    }
	  else if(seeWhat == SP)
	    {
	      printSPStat();
	    }
	  else if (seeWhat == MIS)
	    {
	      printMisStat();
	    }
	  else // if (seeWhat == ADM)
	    {
	      printAdmStat();
	    }
	  continue;
	}
      else if (ptr[0] == 'q')
	{ // any word beginning with q
	  break;
	}
      else if (!strwcmp(ptr, "mob"))
	{
	  seeWhat = MOB;
	  continue;
	}
      else if (!strwcmp(ptr, "sp"))
	{
	  seeWhat = SP;
	  continue;
	}
      else if (!strwcmp(ptr, "mis"))
	{
	  seeWhat = MIS;
	  continue;
	}
       else if (!strwcmp(ptr, "adm"))
	{
	  seeWhat = ADM;
	  continue;
	}
     if (seeWhat == MOB)
	{
	  doMobCmd(ptr);
	}
      else if (seeWhat == SP)
	{
	  doSPCmd(ptr);
	}
      else if (seeWhat == ADM)
	{
	  doAdmCmd(ptr);
	}
      else // if (seeWhat == MIS)
	{
	  // nothing yet
	}
#ifdef HAVE_READLINE_READLINE_H
      free(line);
#endif
    }
  return 0;
}

/***********************************************************************/

//! printAdmHelp prints help for admin mode commands
/*
Returned Value: none

Called By: doAdmCmd

*/

static void printAdmHelp() /* NO ARGUMENTS */
{
  printf("q | quit      -- quit MOAST\n");
  printf("? | help      -- print this AM Admin help\n");
  printf("sp            -- switch to AM SP\n");
  printf("mob           -- switch to AM mobility\n");
  printf("mis           -- switch to AM mission\n");
  printf("init          -- initialize AM Mob, Mis, and SP\n");
  printf("halt          -- halt AM Mob, Mis, and SP\n");
  printf("abort         -- abort AM Mob, Mis, and SP\n");
  printf("shutdown      -- shutdown AM Mob, Mis, and SP\n");
}

/***********************************************************************/

//! printAdmStat prints abbreviated status for mobility and SP
/*
Returned Value: none

Called By: main

*/

static void printAdmStat()
{
  AmMobJAStat * mobStat = amMobJAStat;
  AmSPStat * SPStat = amSPStat;

  printf("Mob command_type:       %s\n",
	 amMobJA_symbol_lookup(mobStat->command_type));
  printf("Mob echo_serial_number: %d\n", (int) mobStat->echo_serial_number);
  printf("Mob status:             %s\n", rcs_stat_to_string(mobStat->status));
  printf("Mob command state:      %d\n", (int) mobStat->state);
  printf("Mob admin state:        %s\n",
	 amMobJA_enum_RCS_ADMIN_STATE_symbol_lookup(mobStat->admin_state));
    printf("\n");

  printf("SP command_type:       %s\n",
	 amSP_symbol_lookup(SPStat->command_type));
  printf("SP echo_serial_number: %d\n", (int) SPStat->echo_serial_number);
  printf("SP status:             %s\n", rcs_stat_to_string(SPStat->status));
  printf("SP command state:      %d\n", (int) SPStat->state);
  printf("SP admin state:        %s\n",
	 amSP_enum_RCS_ADMIN_STATE_symbol_lookup(SPStat->admin_state));
}

/***********************************************************************/

//! printMisStat prints mission status information
/*
Returned Value: none

Called By: main

*/

static void printMisStat() /* NO ARGUMENTS */
{
  printf("Sorry. No mission status yet.\n");
}

/***********************************************************************/

//! printMobHelp prints help for mob mode commands
/*
Returned Value: none

Called By: doMobCmd

*/

static void printMobHelp() /* NO ARGUMENTS */
{
  printf("q | quit              -- quit MOAST\n");
  printf("? | help              -- print this AM Mob help\n");
  printf("sp                    -- switch to AM SP\n");
  printf("adm                   -- switch to AM admin\n");
  printf("mis                   -- switch to AM mission\n");
  printf("init                  -- initialize AM Mob\n");
  printf("halt                  -- halt AM Mob\n");
  printf("abort                 -- abort AM Mob\n");
  printf("shutdown              -- shutdown AM Mob\n");
  printf("nop                   -- AM Mob no op\n");
  printf("move <x><y><tol><spd> -- AM Mob move (tol and spd are optional)\n");
  printf("spin <dir><ang><tol>  -- AM Mob spin to angle (dir=CW,CCW,SHORT)\n");
  printf("plan <nhd> <spd>      -- AM Mob plan neighborhood speed\n");
  printf("wp <file>             -- AM Mob move (pts in <file>)\n");
  printf("wpl <nhd><spd><file>  -- AM Mob move (pts in <file>) and plan\n");
  printf("cellRes <meters>      -- set AM Mob cell resolution to <meters>\n");
  printf("ct <secs>             -- set AM Mob cycle time to <secs>\n");
  printf("debug <0, 1, ...>     -- configure AM Mob debug level\n");
  printf("dump <0, 1>           -- turn AM Mob world model display off/on\n");
  printf("horizon <meters>      -- set AM Mob planning horizon to <meters>\n");
  printf("turnRad <meters>      -- set AM Mob min turn radius to <meters>\n");
  printf("vehWidth <meters>     -- set AM Mob vehicle width to <meters>\n");  
}

/***********************************************************************/

//! printMobStat prints mobility status information
/*
Returned Value: none

Called By: main

*/

static void printMobStat() /* NO ARGUMENTS */
{
  AmMobJAStat * stat = amMobJAStat;
  AmMobJASet * set = amMobJASet;
  static char mess[100];

  getMobError(mess);
  printf("Mob command_type:   %s\n",
	 amMobJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("error message:      %s\n", mess);
  printf("command state:      %d\n", (int) stat->state);
  printf("admin state:        %s\n",
	 amMobJA_enum_RCS_ADMIN_STATE_symbol_lookup(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("heartbeat:          %d\n", (int) stat->tt.count);
  printf("cycleTime:          %f\n", set->cycleTime);
  printf("debug:              %X\n", set->debug);
  printf("dump:               %d\n", ((set->dumpWM == true) ? 1 : 0));
  printf("vehicleMinTurnRad:  %f\n", set->vehicleMinTurnRad);
  printf("vehicleWidth:       %f\n", set->vehicleWidth);
  printf("horizon:            %f\n", set->planHorizon );
  if( stat->command_type == AM_MOB_JA_CMD_PREPARE_MOVE_TYPE )
    {
      for( int i=0; i<stat->seed_length; i++ )
	printf( "seed: %d at <%lf %lf> cost: %lf\n", 
		i,
		stat->seed[i].x,
		stat->seed[i].y,
		stat->seed[i].z );
    }
}

/***********************************************************************/

//! printSPHelp prints help for SP mode commands
/*
Returned Value: none

Called By: doSPCmd

*/

static void printSPHelp() /* NO ARGUMENTS */
{
  printf("q | quit          -- quit\n");
  printf("? | help          -- print this AM SP help\n");
  printf("mob               -- switch to AM mobility\n");
  printf("adm               -- switch to AM admin\n");
  printf("mis               -- switch to AM mission\n");
  printf("init              -- initialize AM SP\n");
  printf("halt              -- halt AM SP\n");
  printf("abort             -- abort AM SP\n");
  printf("shutdown          -- shutdown AM SP\n");
  printf("nop               -- AM SP no op\n");
  printf("go                -- AM SP produce data\n");
  printf("ct <secs>         -- set AM SP cycle time to <secs>\n");
  printf("debug <0, 1, ...> -- set AM SP debug level\n");
  printf("dump <0,1,2>      -- set map pic off or to height or obstacles\n");
}

/***********************************************************************/

//! printSPStat prints SP status information
/*
Returned Value: none

Called By: main

*/

static void printSPStat(void) /* NO ARGUMENTS */
{
  AmSPStat * stat = amSPStat;
  AmSPSet * set = amSPSet;

  printf("SP command_type:    %s\n",
	 amSP_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin state:        %s\n",
	 amSP_enum_RCS_ADMIN_STATE_symbol_lookup(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("heartbeat:          %d\n", (int) stat->tt.count);
  printf("cycleTime:          %f\n", set->cycleTime);
  printf("debug:              %X\n", set->debug);
  printf("dump:               %s\n", ((set->dumpLevel == 1) ? "1 (height)" :
				      (set->dumpLevel == 2) ? "2 (obstacles)" :
				      "0 (off)"));
}

/***********************************************************************/

//! rcs_stat_to_string gives the string corresponding to a status code
/*
Returned Value: char * (the string)

Called By:
   printMobStat
   printSPStat

*/

static const char * rcs_stat_to_string( /*   ARGUMENTS   */
 int s)                                 /*!< status code */
{
  return ((s == RCS_DONE) ? "RCS_DONE" :
	  (s == RCS_EXEC) ? "RCS_EXEC" :
	  (s == RCS_ERROR) ? "RCS_ERROR" : "?");
}

/***********************************************************************/

/*! setOddSerial sets a serial number to the smallest odd number larger
  than the echoed number and the incoming value of the serial number */
/*

Returned Value: None

Called By:
  writeMobCfg
  writeMobCmd
  writeSPCfg
  writeSPCmd

*/

static void setOddSerial( /*   ARGUMENTS            */
 int echoed,              /*!< echoed serial number */
 int * serial)            /*!< serial number to set */
{
  *serial = ((echoed <= *serial) ?
	     (*serial + ((*serial % 2) ? 2 : 1)) :
	     (echoed  + ((echoed  % 2) ? 2 : 1)));
}

/***********************************************************************/

/*! strwcmp tests whether the following two conditions hold regarding a
    test string and a match string; N is the strlen of the match string.
    1. The first N characters of the test string are the same as the
       corresponding characters of the "match" string.
    2. Character N+1 of the test string is either NULL or white space. */
/*
Returned Value: int
  If both conditions are true, this returns 1. Otherwise, it returns 0.

Called By:
  doAdmCmd
  doMobCmd
  doSPCmd
  main

*/

static int strwcmp( /*   ARGUMENTS       */
 char *test,        /*!< string to test  */
 char *match)       /*!< string to match */
{
  while ((0 != *test) && (0 != *match))
    {
      if (*test != *match)
	return 1;
      test++, match++;
    }
  if ((0 == *match) && ((0 == *test) || (isspace(*test))))
    return 0;
  return 1;
}

/***********************************************************************/

/*! writeMobCfg writes an AM Mobility config command and updates the
  AM Mobility config command serial number. */
/*
Returned Value: none

Called By:
  doAdmCmd
  doCmdMove
  doCmdWaypoint
  doCmdWaypointPlan
  doMobCmd

*/

static void writeMobCfg( /*   ARGUMENTS               */
 RCS_CMD_MSG * cfg)      /*!< config command to write */
{
  static int serial = 1; // SP config serial number

  setOddSerial(amMobJASet->echo_serial_number, &serial);
  cfg->serial_number = serial;
  amMobJACfgBuf->write(cfg);
}

/***********************************************************************/

/*! writeMobCmd writes an AM mobility command and updates the AM mobility
  command serial number. */
/*
Returned Value: none

Called By:
  doAdmCmd
  doCmdMove
  doCmdWaypoint
  doCmdWaypointPlan
  doMobCmd

*/

static void writeMobCmd( /*   ARGUMENTS         */
 RCS_CMD_MSG * cmd)      /*!< command to write  */
{
  static int serial = 1; // mob command serial number

  setOddSerial(amMobJAStat->echo_serial_number, &serial);
  cmd->serial_number = serial;
  amMobJACmdBuf->write(cmd);
}

/***********************************************************************/

/*! writeSPCfg writes an AM SP config command and updates the AM SP config 
  command serial number. */
/*
Returned Value: none

Called By:
  doSPCfg

*/

static void writeSPCfg( /*   ARGUMENTS               */
 RCS_CMD_MSG * cfg)     /*!< config command to write */
{
  static int serial = 1; // SP config serial number

  setOddSerial(amSPSet->echo_serial_number, &serial);
  cfg->serial_number = serial;
  amSPCfgBuf->write(cfg);
}

/***********************************************************************/

/*! writeSPCmd writes an AM SP command and updates the AM SP command
  serial number. */
/*
Returned Value: None

Called By:
  doAdmCmd
  doCmd
  doSPCmd

*/

static void writeSPCmd( /*   ARGUMENTS        */
 RCS_CMD_MSG * cmd)     /*!< command to write */
{
  static int serial = 1; // SP command serial number

  setOddSerial(amSPStat->echo_serial_number, &serial);
  cmd->serial_number = serial;
  amSPCmdBuf->write(cmd);
}

/***********************************************************************/

/*
  Modification history:

  $Log: amShell.cc,v $
  Revision 1.22  2008/02/12 00:01:39  tomrkramer
  Changed choice of dumping from 2 choices (1=on height map, 0=off) to 3
  choices  (1=on height map, 2=on obstacle map, 0=off).

  Revision 1.21  2007/12/18 14:39:29  tomrkramer
  Fixed print command error (changed %i  to %lf) for planHorizon.

  Revision 1.20  2007/12/14 23:41:07  tomrkramer
  Added commands for setting minimum turning radius and plan horizon.
  Fixed minor typo.

  Revision 1.19  2007/05/16 14:00:45  dr_steveb
  Changed definitions of rotate.

  Revision 1.18  2007/01/26 17:21:20  tomrkramer
  Improved sending of commands to AM SP and AM Mob. Made command numbers odd.

  Revision 1.17  2007/01/24 22:15:55  tomrkramer
  Finished implementing improved messages for new error codes.

  Revision 1.16  2007/01/24 16:37:23  tomrkramer
  Added error messages.

  Revision 1.15  2007/01/24 16:08:22  dr_steveb
  Changed the way that seed values are printed

  Revision 1.14  2006/11/27 19:41:51  dr_steveb
  Added plan horizon to am settings message.

  Revision 1.13  2006/10/26 20:11:56  tomrkramer
  Added move command taking x and y coordinates.
  Changed code so that amShell can be started and stopped any time
  moast is running.
  Changed code so that it is not necessary to hit <enter> twice between
  commands.
  Fixed minor bug in printing SP status.
  Moved a few functions into alphabetical order.
  Improved documentation a little.
  Added writeMobCmd function.

  Revision 1.12  2006/10/02 21:22:18  tomrkramer
  Updated to use new form of spin command.

  Revision 1.11  2006/09/26 19:56:27  tomrkramer
  Added reading status buffers repeatedly at startup until the the data in
  them looks correct. The first readings of the buffers may get bad data.

  Revision 1.10  2006/09/25 16:51:06  tomrkramer
  Added reporting detailed mobility errors. Changed "tollerance" to
  "tolerance". Changed starting mode to admin.

  Revision 1.9  2006/09/24 21:29:52  dr_steveb
  Added spin command

  Revision 1.8  2006/04/20 12:52:55  tomrkramer
  Changed source of heartbeat data for AM Mob and AM SP to stat.tt.count.

  Revision 1.7  2006/04/07 16:45:58  tomrkramer
  Added adm (admin) mode for sending administrative commands to both Mob and SP
  simultaneously. Added go command for SP. Added nop command for Mob and SP.
  Added stub of mis (mission) mode. Added hash-defs for mode names.

  Revision 1.6  2006/04/05 13:22:02  tomrkramer
  Added admin and stub for mission. Also added admin_state in status.

  Revision 1.5  2006/02/22 23:25:50  tomrkramer
  added new user commands, added documentation, fixed bugs

  Revision 1.4  2006/02/16 21:17:59  tomrkramer
  added dump command for AM mobility and put vehicleWidth in status display

  Revision 1.3  2006/01/17 22:23:24  tomrkramer
  added functionality, divided main into smaller functions, reformatted

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.2  2005/09/30 19:41:58  stephen
  Added waypoint command.

  Revision 1.1  2005/07/01 16:01:14  proctor
  Took out annoying .nml and .ini config printing in amMobMain.cc; added
  amShell.cc

*/

