/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   moastGeom.cc
  \brief  Geometric routines that are commonly used by various MOAST modules.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.5 $
  $Date: 2008/12/02 19:51:59 $
  \endcode

  \author Stephen Balakirsky
  \date   07/05/05
*/
#include <math.h> // fabs
#include "moastTypes.hh"	// M_PI
#include "moastGeom.hh"

//! Determines if three points are collinear
/*! Based on code from Geometric ...
  \param a           First point 
  \param b           Second point
  \param c           Thrid point
  \param threshold   A tollerance on the linearity, defaults to 0.
  \return True if points are linear within threshold value
 */
bool LineSegment::collinear( Location2D &a, Location2D &b, Location2D &c, double threshold )
{
  if( fabs((b.north - a.north) * (c.east - a.east) -
      (c.north - a.north) * (b.east - a.east)) <= threshold )
    return(true);
  else
    return(false);
}

//! computes if two line segments intersect
/*! Based on code from Geometric ...
  \param l1    segment 1
  \param l2    segment 2
  \return      True if line segments intersect
*/
bool LineSegment::testIntersect(  LineSegment &l2 )
{
  if( testIntersectProp( l2 ) )
    return true;
  else if( between( this->start, this->end, l2.start ) ||
	   between( this->start, this->end, l2.end ) ||
	   between( l2.start, l2.end, this->start ) ||
	   between( l2.start, l2.end, this->end ) )
    return true;
  else
    return false;
}

//! is point c between points a & b?
/*! Based on code from Geometric ...
  \param a point a
  \param b point b
  \param c point c
  \return true if c is between a & b
*/
bool LineSegment::between(  Location2D &a, Location2D &b, Location2D &c )
{
  if( !collinear( a, b, c ) )
    return false;
  
  /* if ab not vertical, check betweenness on x; else on y */
  if( a.east != b.east )
    return( ((a.east<c.east) && (c.east<b.east)) ||
	    ((a.east>c.east) && (c.east>b.east)) );
  else
    return( ((a.north<c.north) && (c.north<b.north)) ||
	    ((a.north>c.north) && (c.north>b.north)) );
}

//! check if two line segments intersect each other in their interiors
/*! Based on code from Geometric ...
  \param l1 Line segment 1
  \param l2 Line segment 2
  \return true if they intersect
*/
bool LineSegment::testIntersectProp(  LineSegment &l2 )
{
  if( collinear( this->start, this->end, l2.start ) ||
      collinear( this->start, this->end, l2.end ) ||
      collinear( l2.start, l2.end, this->start ) ||
      collinear( l2.start, l2.end, this->end ))
    return false;
  return( (left(l2.start) ^ left(l2.end) ) &&
	  (l2.left(this->start) ^ l2.left(this->end) ) );
}

//! is the point to the left of the line?
/*! Based on code from Geometric ...
  \param line line segment to test
  \param point point to test
  \return true if point is to the left of the directed line
*/
bool LineSegment::left(  Location2D &point )
{
  if( ((this->end.east - this->start.east) * (point.north - this->start.north) -
       (point.east - this->start.east) * (this->end.north - this->start.north)) > 0 )
    return true;
  else
    return false;
}

LineSegment::LineSegment( double n1, double e1, double n2, double e2 )
{
  start.north = n1;
  start.east = e1;
  end.north = n2;
  end.east = e2;
}

LineSegment::LineSegment( Location2D startIn, Location2D endIn )
{
  start = startIn;
  end = endIn;
}

//! returns the distance from the line to the point
/*!
  from http://mathworld.wolfram.com/Point-LineDistance2-Dimensional.html
*/
double LineSegment::distToPoint( const Location2D &pointIn )
{
  double retValue;
  retValue = (fabs((this->end.north - this->start.north)*(this->start.east - pointIn.east) -
		   (this->start.north - pointIn.north)*(this->end.east - this->start.east))/
	      sqrt((this->end.north - this->start.east)*(this->end.north - this->start.east)+
		   (this->end.east - this->start.east)*(this->end.east - this->start.east)));
  return retValue;
}

//! returns the point where two lines intersect
Location2D LineSegment::intersect( const LineSegment &line2 )
{
  double denominator, slope1; //, slope2;
  Location2D retValue;

  denominator = (line2.end.east - line2.start.east) *
    (this->end.north - this->start.north) -
    (line2.end.north - line2.start.north) *
    (this->end.east - this->start.east);
  slope1 = ( (line2.end.north - line2.start.north) *
	     (this->start.east - line2.start.east) -
	     (line2.end.east - line2.start.east) *
	     (this->start.north - line2.start.north) )/denominator;
    /*
      slope2 = ( (this.end.north - this.start.north) *
      (this.start.east - line2.start.east) -
      (this.end.east - this.start.east) *
      (this.start.north - line2.start.north) );
    */

    // could use either slope1 or slope 2
  retValue.north = this->start.north + slope1 * (this->end.north - this->start.north);
  retValue.east = this->start.east + slope1 * (this->end.east - this->start.east);
  return retValue;
}
  
//! move a line parallel to itself by a distance
/*! Will move a line by a distance parallel to itself, the direction flag
  allows the direction of movement to be set
  
  \param distance The distance to move
  \param direction Move one way with 1, the other with -1
*/
LineSegment LineSegment::moveLineParallel( double distance, int direction )
{
  LineSegment retValue;
  Location2D tempPt;
  double theta;
  
  theta = atan2( this->end.east-this->start.east, this->end.north-this->start.north) - (M_PI_2 * direction);
  retValue.start.north = this->start.north + distance*cos(theta);
  retValue.start.east = this->start.east + distance*sin(theta);
  retValue.end.north = this->end.north + distance*cos(theta);
  retValue.end.east = this->end.east + distance*sin(theta);
  return retValue;
}

void LineSegment::setStart( double north, double east )
{
  start.north = north;
  start.east = east;
}
void LineSegment::setEnd( double north, double east )
{
  end.north = north;
  end.east = east;
};
void LineSegment::print(FILE *output)
{
  fprintf( output, "%f %f\n%f %f\n\n",
	   start.north, start.east,
	   end.north, end.east );
}


/*! CVS Log Information
 * $Log: moastGeom.cc,v $
 * Revision 1.5  2008/12/02 19:51:59  proctor
 * Included moastTypes.hh for M_PI
 *
 * Revision 1.4  2007/12/13 13:18:40  dr_steveb
 * Changes for vehicle level visibility graph planning.
 *
 * Revision 1.3  2007/11/08 13:21:59  dr_steveb
 * Not done yet, but premature checking in of other files means these have to go to!
 *
 * Revision 1.2  2007/11/07 19:56:30  dr_steveb
 * Moved the LineSegment class from moastTypes into moastGeom
 *
 * Revision 1.1.1.1  2005/10/12 20:50:32  root
 * initial import
 *
 * Revision 1.2  2005/07/26 20:30:18  stephen
 * Debug in progress
 *
 * Revision 1.1  2005/07/21 18:59:11  stephen
 * Fixed merged conflicts and added visibility graph stuff.
 *
 *
 */
