/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   moastMySQL.cc
  \brief Provides an interface to mySQL services for the MOAST system.

  This class provides a generic interface to the mySQL tables. The idea
  is that any changes to the table structure will be reflected here and
  will not require changes throughout the code base.

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Stephen Balakirsky
  \date   06/14/2005
*/
#include <stdio.h>
#include <string.h>
#include <mysql.h>
#include "moastMySQL.hh"

///////////////////////////////////////////////////////////////////////////////
//! constructor
/*! Initialize all variables
 */
MoastMySQL::MoastMySQL(int myVerboseIn)
{
  myVerbose = myVerboseIn;
  hostName = NULL;
  userName = NULL;
  password = NULL;
  dbName   = NULL;
  conn     = NULL;
  resSet   = NULL;
  commandBuffer = new char[MOAST_MAX_SQL_STRING_LEN];
}

///////////////////////////////////////////////////////////////////////////////
//! destructor
/*! Basically just calls close
 */
MoastMySQL::~MoastMySQL()
{
  closeDB();
}

///////////////////////////////////////////////////////////////////////////////
//! Close connection and clean up
/*! Clean up all allocated memory
 */
void MoastMySQL::closeDB()
{
  if( hostName != NULL )
    delete hostName;
  if( userName != NULL )
    delete userName;
  if( password != NULL )
    delete password;
  if( dbName != NULL )
    delete dbName;
  if( conn != NULL )
    mysql_close(conn);
  if( resSet != NULL )
    mysql_free_result (resSet); 
  if( commandBuffer != NULL )
    delete commandBuffer;
}

///////////////////////////////////////////////////////////////////////////////
//! open the database
/*!
  Allocates memory and opens the mySQL database
\param hostIn     The name of the sql host
\param userIn     The name of the sql user
\param passwordIn The password for the user account
\param dbIn       The name of the database to open
\param socketIn   Optional parameter. The name of the socket (DEFAULT NULL)
\param portIn     Optional parameter. The port number (DEFAULT 0)
\param flagsIn    Optional parameter. Additional flags (DEFAULT 0)

\return -1 error
\return  1 success
*/
int MoastMySQL::init(char *hostIn, char *userIn, char *passwordIn, 
		     char *dbIn, char *socketIn, 
		     unsigned int portIn,
		     unsigned int flagsIn )
{
  int length;

  if( hostIn == NULL || userIn == NULL ||
      passwordIn == NULL || dbIn == NULL )
    return -1;
  if( conn != NULL )
    closeDB();
  length = strlen(hostIn) + 1;
  hostName = new char [length];
  strncpy( hostName, hostIn, length );
  length = strlen(userIn) + 1;
  userName = new char [length];
  strncpy( userName, userIn, length );
  length = strlen(passwordIn) + 1;
  password = new char [length];
  strncpy( password, passwordIn, length );
  length = strlen(dbIn) + 1;
  dbName = new char [length];
  strncpy( dbName, dbIn, length );
  if( socketIn == NULL )
    socketName = NULL;
  else
    {
      length = strlen(socketIn) + 1;
      socketName = new char [length];
      strncpy( socketName, socketIn, length );
    }
  portNum = portIn;
  flags = flagsIn;

  conn = mysql_init(NULL);
  if( conn == NULL )
    {
      if( myVerbose)
	fprintf( stderr, "MoastMySQL::init() failed (probably out of memory)\n" );
      return( -1 );
    }
  if(mysql_real_connect(conn, hostName, userName, password, 
			dbName, portNum, socketName, flags) == NULL) 
    { 
      if( myVerbose )
	fprintf(stderr,"MoastMySQL::init()::mysql_real_connect() failed:\nError %u (%s)\n",
		mysql_errno(conn), mysql_error(conn)); 
      mysql_close(conn); 
      return(-1);
    }
  return(1);
}

///////////////////////////////////////////////////////////////////////////////
void MoastMySQL::showQuery()
{
}
///////////////////////////////////////////////////////////////////////////////
int MoastMySQL::submitQuery()
{
  return 1;
}
///////////////////////////////////////////////////////////////////////////////
int MoastMySQL::updateQuery()
{
  return 1;
}
///////////////////////////////////////////////////////////////////////////////
void MoastMySQL::processQuery()
{
}
///////////////////////////////////////////////////////////////////////////////
void MoastMySQL::setSqlQueryCmd()
{
}
///////////////////////////////////////////////////////////////////////////////
void MoastMySQL::freeResults()
{
}

/*
  $Log: moastMySQL.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.2  2005/06/29 21:08:11  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.1  2005/06/15 13:39:10  stephen
  Added mysql support.
*/
