/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file factoryShell.cc

  \brief Command line shell to factory NML buffers

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.4 $
  $Date: 2009/03/30 18:24:58 $
  \endcode

  \author Stephen Balakirsky
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <string.h>		// strncmp()
#include <stdlib.h>		// getenv()
#include <ctype.h>		// isspace
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "factCtrl.hh"

static RCS_CMD_CHANNEL *factCtrlCmdBuf = NULL;
static RCS_STAT_CHANNEL *factCtrlStatBuf = NULL;
static RCS_CMD_CHANNEL *factCtrlCfgBuf = NULL;
static RCS_STAT_CHANNEL *factCtrlSetBuf = NULL;

static int serialNumber = 1;	// odd for shells

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define THIS_PROCESS "factCtrlShell"

#define FACT_WRITE(msg) serialNumber += 2; (msg).serial_number = serialNumber; factCtrlCmdBuf->write(&(msg))
#define FACT_CFG_WRITE(msg) serialNumber += 2; (msg).serial_number = serialNumber; factCtrlCfgBuf->write(&(msg))

enum {SEE_NONE, SEE_ADMIN, SEE_CMD};
enum {PRINT_STAT, PRINT_SET};

static int factOK = 0;

static int cleanupNmlBuffers(void)
{
  if (NULL != factCtrlCmdBuf)
    delete factCtrlCmdBuf;
  if (NULL != factCtrlStatBuf)
    delete factCtrlStatBuf;
  if (NULL != factCtrlSetBuf)
    delete factCtrlSetBuf;
  if (NULL != factCtrlCfgBuf)
    delete factCtrlCfgBuf;

  factCtrlStatBuf = NULL;
  factCtrlSetBuf = NULL;
  factCtrlCmdBuf = NULL;
  factCtrlCfgBuf = NULL;

  factOK = 0;

  return 0;
}

static int initNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupNmlBuffers();

  sprintf(chanName, "%s%d", FACT_CTRL_CMD_NAME, bufnum);
  factCtrlCmdBuf =
    new RCS_CMD_CHANNEL(factCtrl_format, chanName,
			THIS_PROCESS, config_file);
  if (!factCtrlCmdBuf->valid()) {
    (void) cleanupNmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", FACT_CTRL_STAT_NAME, bufnum);
  factCtrlStatBuf =
    new RCS_STAT_CHANNEL(factCtrl_format, chanName,
			 THIS_PROCESS, config_file);
  if (!factCtrlStatBuf->valid()) {
    (void) cleanupNmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", FACT_CTRL_CFG_NAME, bufnum);
  factCtrlCfgBuf =
    new RCS_CMD_CHANNEL(factCtrl_format, chanName,
			THIS_PROCESS, config_file);
  if (!factCtrlCfgBuf->valid()) {
    (void) cleanupNmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", FACT_CTRL_SET_NAME, bufnum);
  factCtrlSetBuf =
    new RCS_STAT_CHANNEL(factCtrl_format, chanName,
			 THIS_PROCESS, config_file);
  if (!factCtrlSetBuf->valid()) {
    (void) cleanupNmlBuffers();
    return 1;
  }

  for (double end = etime() + 2.0; etime() < end; esleep(0.1)) {
    factCtrlStatBuf->read();
    factCtrlSetBuf->read();
    if( ((FactCtrlStat *) (factCtrlStatBuf->get_address()))->type ==
	FACT_CTRL_STAT_TYPE &&
	((FactCtrlSet *) (factCtrlSetBuf->get_address()))->type ==
	FACT_CTRL_SET_TYPE) {
      factOK = 1;
      return 0;
    }
  }

  (void) cleanupNmlBuffers();
  factOK = 0;
  return 1;
}

#define rcs_stat_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

#define rcs_admin_state_to_string(s) (s) == ADMIN_INITIALIZED ? "Initialized" : (s) == ADMIN_UNINITIALIZED ? "Uninitialized" : (s) == ADMIN_SHUT_DOWN ? "Shut Down" : "?"

static void printAdmin(void)
{

  printf( "========================================\n" );
  printf( "************ Admin Status **************\n" );
  printf( "========================================\n" );
  if( factOK ) 
    {
      FactCtrlStat *fact = (FactCtrlStat*)factCtrlStatBuf->get_address();
      printf("   Admin State  : %s\n", 
	     rcs_admin_state_to_string(((FactCtrlStat *)fact)->admin_state));
      printf("   command_type : %s\n", 
	     factCtrl_symbol_lookup(((FactCtrlStat *)fact)->command_type));
      printf("   Status       : %s\n", 
	     rcs_stat_to_string(((FactCtrlStat *)fact)->status));
      printf("   HeartBeat    : %d\n", ((FactCtrlStat *)fact)->tt.count);
    }
}

static void printFactStat(void)
{
  FactCtrlStat * stat = (FactCtrlStat *) factCtrlStatBuf->get_address();

  if (! factOK) return;

  printf("command_type:       %s\n",
	 factCtrl_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("serial_number_sent: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", 
	 rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
  for( int count=0; count<stat->conveyorStat_length; count++ )
    printf( "Conveyor %s at speed %f\n", stat->conveyorStat[count].name,
	    stat->conveyorStat[count].speed );

}

static void printFactSet(void)
{
  FactCtrlSet *set = (FactCtrlSet *) factCtrlSetBuf->get_address();

  if (! factOK) return;

  printf("command_type:       %s\n",
	 factCtrl_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("line:               %d\n", (int) set->line);
  printf("source_line:        %d\n", (int) set->source_line);
  printf("source_file:        %s\n", (char *) set->source_file);
  printf("cycleTime:          %f\n", (double) set->cycleTime);
  printf("debug:              0x%X\n", (int) set->debug);
  for( int count=0; count<set->conveyorSet_length; count++ )
    printf( "Conveyor %s located at <%f %f %f> orient <%f %f %f> speed: <%f %f>\n",
	    set->conveyorSet[count].name,
	    set->conveyorSet[count].location.x,
	    set->conveyorSet[count].location.y,
	    set->conveyorSet[count].location.z,
	    set->conveyorSet[count].orientation.r,
	    set->conveyorSet[count].orientation.p,
	    set->conveyorSet[count].orientation.y,
	    set->conveyorSet[count].minSpeed,
	    set->conveyorSet[count].maxSpeed );
}

static int printIt(int see, int print)
{
  if (see == SEE_CMD) 
    {
      if (print == PRINT_STAT) printFactStat();
      else printFactSet();
    } 
  else if (see == SEE_ADMIN) 
    {
      printAdmin();
    } 
  else 
    {
      printf("nothing connected\n");
    }

  return 0;
}

static void printFactHelp(void)
{
  printf("admin             -- set admin mode\n" );
  printf("fact               -- set factory control mode\n" );
  printf("q | quit          -- quit\n");
  printf("? | help          -- print this help\n");
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
  printf("debug <0, 1, ...> -- configure debug level\n");
  printf("set               -- change to settings mode\n" );
  printf("stat              -- change to status mode\n" );
  printf("package           -- drop package\n" );
  printf("speed             -- set conveyor to a given speed\n" );
  printf("clear             -- clear all of the cargo\n" );
  printf("file              -- set a file to read package rfid tags from\n");
}

static void printAdminHelp(void)
{
  printf("admin             -- set admin mode\n" );
  printf("fact               -- set factory mode\n" );
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
}

static void printHelp(int see)
{
  if (see == SEE_CMD) 
    {
      printFactHelp();
    } 
  else if (see == SEE_ADMIN) 
    {
      printAdminHelp();
    } 
  else 
    {
      printf("nothing connected\n");
    }
  return;
}

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}


#define skipnonwhite(ptr) while(!isspace(*(ptr)) && *(ptr) != 0) (ptr)++
#define skipword(ptr) skipnonwhite(ptr); skipwhite(ptr)
#define arrayels(array) sizeof(array) / sizeof(*(array))

// scans a string of numbers into an array, returning number scanned
static int scanem(double array[], char string[], int n)
{
  double d;
  char * nptr;
  char * endptr;
  int i;

  for (i = 0, nptr = string; i < n; i++) 
    {
      d = strtod(nptr, &endptr);
      if (endptr == nptr) break;
      array[i] = d;
      nptr = endptr;
    }
  
  return i;
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  int bufnum = 1;
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char *line;
  char *ptr;
  int retval;
  int whichSee = SEE_NONE;
  int whichPrint = PRINT_STAT;

  FactCtrlStat *factCtrlStatPtr;
  FactCtrlSet *factCtrlSetPtr;
  FactCtrlCmdInit factCtrlCmdInit;
  FactCtrlCmdHalt factCtrlCmdHalt;
  FactCtrlCmdAbort factCtrlCmdAbort;
  FactCtrlCmdShutdown factCtrlCmdShutdown;
  FactCtrlCmdNop factCtrlCmdNop;
  FactCtrlCmdDropPackage factCtrlCmdDropPackage;
  FactCtrlCmdSetConveyor factCtrlCmdSetConveyor;
  FactCtrlCmdClearCargo  factCtrlCmdClearCargo;

  FactCtrlCfgCycleTime factCtrlCfgCycleTime;
  FactCtrlCfgDebug factCtrlCfgDebug;
  FactCtrlCfgNop factCtrlCfgNop;
  FactCtrlCfgSetPackFile factCtrlCfgSetPackFile;

#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  while (1) 
    {
      char option = getopt(argc, argv, "n:d:");
      if (option == -1)
	break;

      switch (option) 
	{
	case 'n':
	  NML_FILE = optarg;
	  break;
	case 'd':
	  if (1 != sscanf(optarg, "%i", &bufnum)) 
	    {
	      rcs_print_error("%s: bad value for buffer number: %s\n", 
			      argv[0], optarg);
	      return 1;
	    }
	  break;
	default:
	  rcs_print_error("%s: bad argument\n", argv[0]);
	  return 1;
	  break;
	}
    }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) 
    {
      nml_file_env = getenv("CONFIG_NML");
      if (NULL != nml_file_env) 
	{
	  NML_FILE = nml_file_env;
	} 
      else 
	{
	  NML_FILE = nml_file_default;
	}
    }

  // allocate all the factility buffers
  retval = initNmlBuffers(NML_FILE, bufnum);
  if (retval) 
    {
      fprintf(stderr, 
	      "%s: can't allocate factCtrl NML buffers, so ignoring them\n", 
	      argv[0]);
      factCtrlStatPtr = NULL;
      factCtrlSetPtr = NULL;
    }
  else 
    {
      factCtrlStatPtr = (FactCtrlStat *) factCtrlStatBuf->get_address();
      factCtrlSetPtr = (FactCtrlSet *) factCtrlSetBuf->get_address();
    }

  if ( factOK )
    whichSee = SEE_CMD;
  else 
    whichSee = SEE_NONE;

  while (!feof(stdin)) {
#ifndef HAVE_READLINE_READLINE_H
    printf(whichSee == SEE_ADMIN ? "FactAdmin> " :
	   whichSee == SEE_CMD ? "FactCmd> " : "FactNone> ");
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) 
      break;
    buffer[BUFFERSIZE - 1] = 0;
    line = buffer;
#else
    line = readline(whichSee == SEE_ADMIN ? "FactAdmin> " :
		    whichSee == SEE_CMD ? "FactCmd> " : "FactNone> ");
    if (NULL == line) {
      break;
    }
    if (*line) {
      add_history(line);
    }
#endif

    if (factOK) 
      {
	factCtrlStatBuf->read();
	factCtrlSetBuf->read();
      }

    // clean off any trailing whitespace
    ptr = line + strlen(line);	// at the null
    while (--ptr != line && isspace(*ptr)) *ptr = 0;

    ptr = line;
    while (isspace(*ptr))
      ptr++;			// skip leading white space
    if (0 == *ptr) 		// blank line
      printIt(whichSee, PRINT_STAT);
    else if (ptr[0] == 'q' || !strwcmp(ptr, "quit")) 
	break;
    else if (ptr[0] == '?' || !strwcmp(ptr, "help"))
      printHelp(whichSee);
    else if (!strwcmp(ptr, "admin")) 
      {
      (void) initNmlBuffers(NML_FILE, bufnum);
      if (factOK) 
	{
	// we need to do this again since the addresses have likely changed
	factCtrlStatPtr = (FactCtrlStat *) factCtrlStatBuf->get_address();
	factCtrlSetPtr = (FactCtrlSet *) factCtrlSetBuf->get_address();
	}
      printIt(whichSee, PRINT_STAT);
      }
    else if (!strwcmp(ptr, "fact")) 
      {
	(void) initNmlBuffers(NML_FILE, bufnum);
	if (factOK) 
	  {
	    // we need to do this again since the addresses have likely changed
	    factCtrlStatPtr = (FactCtrlStat *) factCtrlStatBuf->get_address();
	    factCtrlSetPtr = (FactCtrlSet *) factCtrlSetBuf->get_address();
	    whichSee = SEE_CMD;
	    printIt(whichSee, PRINT_STAT);
	  } 
	else 
	printf("fact not connected\n");
      }
    else if (!strwcmp(ptr, "stat")) 
      {
      whichPrint = PRINT_STAT;
      printIt(whichSee, PRINT_STAT);
      }
    else if (!strwcmp(ptr, "set")) 
      {
	whichPrint = PRINT_SET;
	printIt(whichSee, PRINT_SET);
      } 
    else if (!strwcmp(ptr, "init")) 
      {
	if (whichSee == SEE_CMD || (whichSee == SEE_ADMIN && factOK) ) 
	  FACT_WRITE(factCtrlCmdInit);
      } 
    else if (!strwcmp(ptr, "halt")) 
      {
      if (whichSee == SEE_CMD || (whichSee == SEE_ADMIN && factOK)) 
	  FACT_WRITE(factCtrlCmdHalt);
      if (!factOK)
	printf("nothing connected\n");
      }
    else if (!strwcmp(ptr, "abort")) 
      {
      if (whichSee == SEE_CMD || (whichSee == SEE_ADMIN && factOK)) 
	  FACT_WRITE(factCtrlCmdAbort);
      if (!factOK)
	printf("nothing connected\n");
      }
    else if (!strwcmp(ptr, "shutdown"))
      {
	if (whichSee == SEE_CMD || (whichSee == SEE_ADMIN && factOK)) 
	  FACT_WRITE(factCtrlCmdShutdown);
	if (!factOK)
	  printf("nothing connected\n");
      }
    else if (!strwcmp(ptr, "clear"))
      {
	if(!factOK)
	  continue;
	FACT_WRITE(factCtrlCmdClearCargo);
	printf( "cleared cargo\n" );
      }
    else if (!strwcmp(ptr, "file"))
      {
	if(!factOK)
	  continue;
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) 
	  {
	    printf("need file name\n");
	    continue;
	  }
	if( 1!= sscanf( ptr, "%s", factCtrlCfgSetPackFile.fileName ) )
	  printf( "bad file name\n");
	else
	  FACT_CFG_WRITE(factCtrlCfgSetPackFile);
      }
    else if (!strwcmp(ptr, "debug"))
      {
	if(!factOK)
	  continue;
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) 
	  {
	    printf("need debug level\n");
	    continue;
	  }
	if( 1!= sscanf( ptr, "%d", &factCtrlCfgDebug.debug ) )
	  printf( "bad debug level\n");
	else
	  FACT_CFG_WRITE(factCtrlCfgDebug);
      }

    else if (!strwcmp(ptr, "package"))
      {
	if(!factOK)
	  continue;
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) 
	  {
	    printf("need chute name\n");
	    continue;
	  }
	if( 1!= sscanf( ptr, "%s", factCtrlCmdDropPackage.chuteName ) )
	  printf( "bad chute name\n");
	else
	  FACT_WRITE(factCtrlCmdDropPackage);
      }
    else if (!strwcmp(ptr, "speed"))
      {
	if(!factOK)
	  continue;
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) 
	  {
	    printf("need conveyor name and speed\n");
	    continue;
	  }
	if( 2!= sscanf( ptr, "%s %f", 
			factCtrlCmdSetConveyor.conveyorName,
			&factCtrlCmdSetConveyor.speed) )
	  printf( "bad parameters name\n");
	else
	  FACT_WRITE(factCtrlCmdSetConveyor);
      }
	/*
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need x, y, tollerance, and speed\n");
	  continue;
	}
	if (4 != sscanf(ptr, "%lf %lf %lf %lf",
			&factCtrlCmdMoveArcSegment.arcList[0].end.x,
			&factCtrlCmdMoveArcSegment.arcList[0].end.y,
			&factCtrlCmdMoveArcSegment.arcList[0].annular_tol,
			&factCtrlCmdMoveArcSegment.arcList[0].speed)) {
	  printf("bad x, y, tollerance, speed  values: %s\n", ptr);
	} else {
	  factCtrlCmdMoveArcSegment.arcList[0].end.z = 0;
	  factCtrlCmdMoveArcSegment.arcList[0].isArc = 0;
	  factCtrlCmdMoveArcSegment.arcList_length = 1;
	  FACT_WRITE(factCtrlCmdMoveArcSegment);
	}
      }
    } else if (!strwcmp(ptr, "ct")) {
      if (whichSee == SEE_CMD) {
	if (1 !=
	    sscanf(ptr, "%*s %lf",
		   &factCtrlCfgCycleTime.cycleTime)) {
	  printf("need a cycleTime argument\n");
	} else {
	  factCtrlCfgCycleTime.serial_number =
	    factCtrlSetPtr->echo_serial_number + 1;
	  factCtrlCfgBuf->
	    write(&factCtrlCfgCycleTime);
	}
      */
    else 
      printf("?\n");

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  (void) cleanupNmlBuffers();

  return 0;
}

/*
  Modification history:

  $Log: factoryShell.cc,v $
  Revision 1.4  2009/03/30 18:24:58  dr_steveb
  Added ask/tell structure

  Revision 1.3  2009/02/21 18:29:39  dr_steveb
  Several minor bug fixes and enhanced readability

  Revision 1.2  2009/02/19 15:26:43  dr_steveb
  Added temporary settings and status messages.

  Revision 1.1  2009/02/17 18:42:18  dr_steveb
  Initial add of factory controller app and control shell


*/

