/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file moastNmlSvr.cc

  \brief NML server for all MOAST buffers.

  Buffers are named with number suffixes starting at 1, e.g.,
  navDataExt1, navDataExt2, the human way, but are associated with NML
  or RCS_CMD,STAT_CHANNEL buffer array indices starting at 0, the
  computer way.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.2 $
  $Date: 2005/12/20 15:19:51 $
  \endcode

  \author Fred Proctor
*/

#include <stddef.h>		// NULL
#include <stdlib.h>		// getenv()
#include <rcs.hh>
#include "moastNmlOffsets.hh"
#include "moastNmlGlobal.hh"	// NUM_NAV_DATA_BUFFERS, servoMobJACmdBuf, etc.
#include "primMisJA.hh"		// primMisJA_format()

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define THIS_PROCESS "moastNmlSvr"	// NML process name for the server

int main(int argc, char *argv[])
{
  char config_file_default[] = DEFAULT_NML_FILE;
  char *config_file_env;
  char *NML_FILE = config_file_default;
  int retval;

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (argc > 1) {
    NML_FILE = argv[1];
  } else {
    config_file_env = getenv("CONFIG_NML");
    if (NULL != config_file_env) {
      NML_FILE = config_file_env;
    }
  }

  // connect to the objective mob nav buffers
  moastZeroNmlBuffers();
  retval = moastInitNmlBuffers(NML_FILE, THIS_PROCESS);
  if (retval) {
    rcs_print_error("%s: can't allocate one or more NML buffers\n", argv[0]);
    return 1;
  }

  // start the nml servers
  run_nml_servers();

  // will never get here
  (void) moastCleanupNmlBuffers();
  return 0;
}

/*
  Modification history:

  $Log: moastNmlSvr.cc,v $
  Revision 1.2  2005/12/20 15:19:51  proctor
  Included primMisJA.hh

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.17  2005/10/05 17:38:49  proctor
  Referenced NML::error_type when deciding to fail; if it's a configuration
  error then continue, so we can step around commented-out buffers.

  Revision 1.16  2005/10/05 13:24:24  proctor
  moastNmlGlobal.cc, moastNmlSvr.cc: didn't fail when an NML buffer couldn't
  be created.

  Revision 1.15  2005/09/30 18:55:44  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.14  2005/06/29 14:31:39  proctor
  Took out annoying .nml, .ini file printing

  Revision 1.13  2005/06/07 21:45:32  proctor
  autogen

  Revision 1.12  2005/06/07 15:46:29  proctor
  Added primMis.cc to autogen list in 'makefile'; added handling of
  servo,primMis buffers in moastNmlGlobal.cc and moved out of moastNmlSvr.cc.

  Revision 1.11  2005/05/23 13:51:24  stephen
  Added primPLCfg channel to support teleport command

  Revision 1.10  2005/05/17 17:19:30  stephen
  Added vehPL buffers.

  Revision 1.9  2005/05/16 21:14:59  proctor
  First time consistently applying suffix numbers to buffers

  Revision 1.8  2005/05/16 12:33:34  stephen
  Fixed misplaced paren.

  Revision 1.7  2005/05/15 14:20:41  stephen
  added servoSP buffers

  Revision 1.6  2005/05/12 14:06:09  stephen
  Added additional navDataDataExt buffers and occupancyData buffers.

  Revision 1.5  2005/05/11 20:23:01  proctor
  Added primPL buffers

  Revision 1.4  2005/05/10 21:01:15  stephen
  Changed server to use NML channel for navDataExt.

  Revision 1.3  2005/05/10 17:34:52  proctor
  Changed vehicle to servoPL, sensor to servoSP

  Revision 1.2  2005/05/05 18:04:37  proctor
  Changed command to cmd, status to stat, etc.

  Revision 1.1  2005/05/05 15:56:37  proctor
  Changed naming convention

*/
