/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   rcsSectManfactEngine.hh
  \brief Provide an RCS engine for the section mob applicaiton

  Contains all of the state tables and nml methods

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.4 $
  $Date: 2008/02/15 20:20:35 $
  \endcode

  \author Stephen Balakirsky
  \date   03/14/2006
*/
#ifndef RCS__ENGINE
#define RCS__ENGINE
#include <iostream>
#include "sectMobPL.hh"
#include "vehMobPL.hh"
class rcsEngine
{
public:
  rcsEngine(double cycleTime, double resolution);
  ~rcsEngine();
  bool connect(const char *, const char*, const char*); // connect nml buffers
  void disconnect(int error); // disconnect nml buffers
  void disconnectSubordinates(); // disconnect nml buffers
  bool connectSubordinates(int *nmlPostfix,
			   int numSubordinates,
			   SectSubordinate subordinateClass );
  RCS_ADMIN_STATE getAdminState(); // return the admin state
  RCS_STATUS getStatus(); // return the module status
  const char *getCommandText(); // return the text for a given command
  bool runStateMachine(const char *systemNmlName); // runs the modules state machine

private:
  void doCmdInit();
  void doCmdAbort();
  void doCmdHalt();
  void doCmdShutdown();
  void doCmdNop();
  void doCmdConnect();
  void doCmdDisconnect();
  void doCmdExplore();
  void doCmdCrash();
  void doCmdLoad();
  void doCmdDeliver();
  void doCmdWaypoint();

  bool doCfgClearWM();
  bool doCfgCycleTime();
  bool doCfgDebug();
  bool doCfgDumpWM();
  bool doCfgLoadWM();
  bool doCfgNop();
  bool doCfgStoreWM();

  // user added routines
  bool checkVehCmdStatus(int which);
  bool loadWM( char *fileName );
  bool stopVehStateMachine(int vehID);
  bool setHashValue( char *key, int value );
  int getHashValue( char *key );
  // nml message types
  SectMobPLStat *nmlStatus; // the modules status buffer contents
  SectMobPLSet *nmlSettings; // the modules settings buffer
  RCS_STATUS vehicleCmdStatus[MAX_SECTION_VEHICLES];
  int vehicleCmdEchoSerialNumber[MAX_SECTION_VEHICLES];
  char vehiclePackageDest[MAX_SECTION_VEHICLES][MAX_WAYPOINT_NAME_LENGTH];
  char vehicleCachedDest[MAX_SECTION_VEHICLES][MAX_WAYPOINT_NAME_LENGTH];
  unsigned int debugFlags;

  // nml channels
  //! section planning channels
  RCS_CMD_CHANNEL  *sectMobPLCmdNml;
  RCS_CMD_CHANNEL  *sectMobPLCfgNml;
  RCS_STAT_CHANNEL *sectMobPLStatNml;
  RCS_STAT_CHANNEL *sectMobPLSetNml;
  //! veh planning channels
  RCS_CMD_CHANNEL  *vehicleCmdNml[MAX_SECTION_VEHICLES];
  RCS_STAT_CHANNEL *vehicleStatNml[MAX_SECTION_VEHICLES];
  RCS_CMD_CHANNEL  *vehicleCfgNml[MAX_SECTION_VEHICLES];
  RCS_STAT_CHANNEL *vehicleSetNml[MAX_SECTION_VEHICLES];

  // cycle timer
  RCS_TIMER *timer;
  // command helpers
  PM_CARTESIAN vehicleDestination[MAX_SECTION_VEHICLES];
  bool vehicleSuccess[MAX_SECTION_VEHICLES];
  //! serial number for commands sent to vehicle
  int *vehicleSerialNumber;
  // world model variables
  //! my verbose level
  bool myVerbose;
  //! the nml file that is being used
  char *myNmlFileName;
  //! the name of this process
  char *mySystemNmlName;
  //! the goal hash table
  int goalHash;
  //! save the states for the individual vehicles
  RCS_STATE *vehStates;
};
#endif

/*
  Modification history:
  $Log: rcsSectManfactEngine.hh,v $
  Revision 1.4  2008/02/15 20:20:35  dr_steveb
  Adde waypoint command to the section echelon

  Revision 1.3  2008/01/07 21:49:00  dr_steveb
  Version of section controller for manufacturing competition. Added the ability to have two vehicles pick-up and deliver packages.

  Revision 1.2  2007/12/20 02:32:45  dr_steveb
  Now have two vehicles performing start of manufacturing job. More to come...

  Revision 1.1  2007/12/17 21:36:03  dr_steveb
  Initial check-in of section level manufacturing application

*/
