/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file matildaserver.c

  \brief Socket server interface MATILDA robot. This includes serial Smart Motors 
         and the Hokuyo laser.

  Principles of operation:

  This program serves a socket for USARsim messages, acting just like
  a real USARsim on behalf of a MATILDA robot running serial Smart
  Motors and sensors.  Rather than connecting to USARsim, MOAST connects 
  to this application and can then drive the MATILDA around.

  One thread for each motor, the "motor thread", is responsible for
  writing commands and reading replies between the SM on the serial port.

  One thread for each socket connection, the "message thread", is
  responsible for parsing USARsim messages and sending replies. 

  The motor- and message threads share a global data structure. The
  motor threads continually update it with motor position, and consult
  it for the commanded motor speeds. The message threads write new
  motor speeds as needed, and read out motor position as
  requested. The global data structure is mutexed.

  The Smart Motor (SM) interface uses the following commands:

  "MV" sent out once to the SM to put it in velocity mode;

  "A<accel> sent out once to set up some crisp acceleration; and

  "V<vel>" sent out continually to set the motor speed in response
  to USARsim messages.

  "RP" is sent out continually to request position, and the SM responds
  with an integer number of counts.

  The main thread runs motor threads for each serial port, then creates a socket server port and listens for
  connections. For each connection, it spawns a message thread and
  returns to await more socket connections from clients. Only one
  socket connection is expected, but others could happen.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.6 $
  $Date: 2008/09/08 14:24:39 $
  \endcode

  \author Fred Proctor & Stephen Balakirsky
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		/* f,printf, stderr, perror */
#include <stddef.h>		/* sizeof, ptrdiff_t */
#include <stdlib.h>		/* malloc, free, atoi */
#include <string.h>		/* memset, strerror */
#include <stdarg.h>		/* va_list */
#include <math.h>		/* M_PI */
#include <errno.h>		/* errno */
#include <unistd.h>		/* read, write */
#include <time.h>               /* struct timespec, nanosleep */
#include <termios.h>		/* struct termios */
#include <fcntl.h>		/* open, O_RDWR */
#include <sys/time.h>           /* gettimeofday, struct timeval */
#include <pthread.h>		/* pthread_attr_int, ... */
#include <sstream>              /* stringstream */
#include "sockutils.h"		/* get_socket,client_fd */
#include "ttycfg.h"		/* tty_raw */
#include "matilda.h"		/* USARSIM,SMART_MOTOR_MSG_LEN */
#include "inifile.h"
#include "../sim/simWare/swKBUtils.hh"

#include "../tools/sockutils.c"
#include "../tools/ttycfg.c"

/*************************
globals
************************/
INIFILE inifile;
string robotName;
string robotClass = "MATILDA";
double maxWheelVelocity = 0;
double theta=0, xPos=0, yPos=0;
double wheelSeparation, wheelRadius;

/*
  'dbprintf' is debug printf. Set 'debug' and you get some diagnostics
  on stderr that show what's going on. Fatal errors are printed
  regardless.
*/
static int debug = 1;
static void dbprintf(const char * fmt, ...)
{
  va_list ap;

  if (debug) {
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
  }
}

/* get the time in seconds */
static double etime(void)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);

  return ((double) tv.tv_sec) + ((double) tv.tv_usec) * 1.0e-6;
}

static void esleep(double secs)
{
  struct timespec ts;

  ts.tv_sec = (int) secs;
  ts.tv_nsec = (int) ((secs - ts.tv_sec) * 1.0e9);

  (void) nanosleep(&ts, NULL);

  return;
}

#define ROUND(x) (x) < 0.0 ? (int) ((x) - 0.5) : (int) ((x) + 0.5)

typedef struct {
  /*
    'counts_per_rad' is how many motor counts per radian.
    'rads_per_count' is its reciprocal. We have both so that we avoid
    division in both conversion directions.
  */
  double counts_per_rad;
  double rads_per_count;

  /* 'position' is updated by the motor threads as fast as practical,
     and read by the message threads when responding to position
     queries. Units are radians. */
  double position;

  /* 'speed' is the commanded speed of the motor. It is written by the
     message thread when it receives a command, and read by the motor
     thread as fast as practical for sending to the motors. Units are
     in rads/sec. */
  double speed;

  /* 'mutex' is locked/unlocked by the motor- and message threads
     when accessing the data structure. */
  pthread_mutex_t mutex;
} smart_motor_struct;

/* the global Smart Motor data structure */
static smart_motor_struct smart_motor_data[SMART_MOTOR_NUM];

typedef struct {
  pthread_mutex_t mutex;
} write_socket_struct;

/* the global structure for socket writing */
static write_socket_struct write_socket_data;

typedef struct {
  /* 'serial_fd' is the file descriptor to the serial port. It is
     opened by the main thread before the motor thread runs. */
  int serial_fd;
  /* 'index' is which index into the smart_motor_data[] array the
     motor thread is working on. */
  int index;
} motor_thread_arg;

static void * motor_thread_code(void * arg)
{
  double delta_t;
  int serial_fd;
  int index;
  int ival;
  int nchars;
  enum {BUFFERLEN = SMART_MOTOR_MSG_LEN};
  char buffer[BUFFERLEN];

#define LOCK_MOTOR pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK_MOTOR pthread_mutex_unlock(&smart_motor_data[index].mutex)

  /* copy out our heap args and free them */
  serial_fd = ((motor_thread_arg *) arg)->serial_fd;
  index = ((motor_thread_arg *) arg)->index;
  free(arg);

  dbprintf("running motor thread %d on serial fd %d\n", index, serial_fd);

  /* clamp index so we won't seg fault if it's bad */
  if (index < 0) index = 0;
  else if (index >= SMART_MOTOR_NUM) index = SMART_MOTOR_NUM - 1;

  /* set up our time period if we're simulating */
  delta_t = 0.1;

  /* set up our motor for velocity mode with nice accel */
  LOCK_MOTOR;
  ival = ROUND(SMART_MOTOR_ACC * ACCEL_ADJUST * smart_motor_data[index].counts_per_rad);
  UNLOCK_MOTOR;
  sprintf(buffer, "MV A=%d G\r", ival);
  write(serial_fd, buffer, strlen(buffer));

  for (;;) {
    if (serial_fd >= 0) {
      LOCK_MOTOR;
      ival = ROUND(smart_motor_data[index].speed * SPEED_ADJUST * smart_motor_data[index].counts_per_rad);
      UNLOCK_MOTOR;
      sprintf(buffer, "V=%d RP G\r", ival);
      write(serial_fd, buffer, strlen(buffer));
      nchars = read(serial_fd, buffer, sizeof(buffer));
      if (nchars > 0) {
	LOCK_MOTOR;
	smart_motor_data[index].position = atoi(buffer) * smart_motor_data[index].rads_per_count; /* could do a better parsing job than 'atoi' */
	UNLOCK_MOTOR;
      }
    } else {
      /* clock our simulated position by vee delta tee */
      LOCK_MOTOR;
      smart_motor_data[index].position += smart_motor_data[index].speed * delta_t;
      UNLOCK_MOTOR;
      esleep(delta_t);
    }
  } /* for (;;) */

  /* if we ever break out of the loop above, we'll get here */
  close(serial_fd);
  dbprintf("read thread %d done\n", index);

  return NULL;

#undef LOCK_MOTOR
#undef UNLOCK_MOTOR
}

/* parsing functions
 */
int parseInit( const vector <string>&toks, int client_fd)
{
  int i;
  const char *inifileText;

  inifileText = inifile.find( "PLATFORM_TYPE", "USARSIM" );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key PLATFORM_TYPE in section USARSIM of inifile\n");
      return 0;
    }

  //  dumpTokens (toks);
  if( -1 == (i=findToken(toks, "ClassName")))
    {
      fprintf( stderr, "Unable to find 'ClassName' in init message\n" );
      return 0;
    }
  if( nocase_cmp(toks[i+1], inifileText) )
    {
      fprintf( stderr, "Wrong robot type! Should be matilda\n" );
      return 0;
    }

  if( -1 == (i=findToken(toks, "Name")))
    {
      fprintf( stderr, "Unable to find 'Name' in init message\n" );
      return 0;
    }
  robotName = toks[i+1];
  dbprintf( "Found robot named %s\n", robotName.c_str() );
  return 1;
}
string parseRangeScannerConf()
{
  stringstream retMessage;
  const char *inifileText;
  int sensorCount = 0;
  int sensorNameCount=1;
  char inifileSearch[200];

  for(;;)
    {
      sensorCount++;
      sprintf( inifileSearch, "SENSOR_%d", sensorCount );
      inifileText = inifile.find(inifileSearch, robotClass.c_str() );
      if( inifileText == NULL )
	{
	  retMessage << ";\r\n";
	  fprintf( stdout, "matildaInterface sending: %s\n", retMessage.str().c_str() );
	  return retMessage.str();
	}
      else if( !strncmp(inifileText, "HOKUYO", 6) )
	{
	  retMessage << "CONF ";
	  retMessage << "{Type ";
	  retMessage << "RangeScanner";
	  retMessage << "} ";

	  inifileText = inifile.find( "MAX_RANGE", "HOKUYO" );
	  if( inifileText == NULL )
	    {
	      fprintf( stderr, "unable to find key MAX_RANGE in section %s of inifile\n",
		       "HOKUYO");
	      return 0;
	    }
	  retMessage << "{MaxRange ";
	  retMessage << inifileText;
	  retMessage << "}";

	  inifileText = inifile.find( "MIN_RANGE", "HOKUYO" );
	  if( inifileText == NULL )
	    {
	      fprintf( stderr, "unable to find key MIN_RANGE in section %s of inifile\n",
		       "HOKUYO");
	      return 0;
	    }
	  retMessage << "{MinRange ";
	  retMessage << inifileText;
	  retMessage << "}";

	  inifileText = inifile.find( "RESOLUTION", "HOKUYO" );
	  if( inifileText == NULL )
	    {
	      fprintf( stderr, "unable to find key RESOLUTION in section %s of inifile\n",
		       "HOKUYO");
	      return 0;
	    }
	  retMessage << "{Resolution ";
	  retMessage << inifileText;
	  retMessage << "}";

	  inifileText = inifile.find( "FOV", "HOKUYO" );
	  if( inifileText == NULL )
	    {
	      fprintf( stderr, "unable to find key FOV in section %s of inifile\n",
		       "HOKUYO");
	      return 0;
	    }
	  retMessage << "{Fov ";
	  retMessage << inifileText;
	  retMessage << "}";

	  inifileText = inifile.find( "PANING", "HOKUYO" );
	  if( inifileText != NULL )
	    {
	      retMessage << "{Paning ";
	      retMessage << inifileText;
	      retMessage << "}";
	    }

	  inifileText = inifile.find( "TILTING", "HOKUYO" );
	  if( inifileText != NULL )
	    {
	      retMessage << "{Tilting ";
	      retMessage << inifileText;
	      retMessage << "}";
	    }

	  retMessage << "{Name ";
	  retMessage << "Hokuyo";
	  retMessage <<sensorNameCount++;
	  retMessage << "}";
	}
    }
  fprintf( stderr, "matildaInterface: hokuyo should not get here!\n" );
  return 0;
}

string parseRobotConf()
{
  string retMessage;
  const char *inifileText;

  retMessage = "CONF ";
  inifileText = inifile.find( "TYPE", robotClass.c_str() );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key TYPE in section %s of inifile\n",
	       robotClass.c_str());
      return 0;
    }
  retMessage += "{Type ";
  retMessage += inifileText;
  retMessage += "} ";
  retMessage += "{Name " + robotName + "} ";

  inifileText = inifile.find( "STEERING_TYPE", robotClass.c_str() );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key STEERING_TYPE in section %s of inifile\n",
	       robotClass.c_str());
      return 0;
    }
  retMessage += "{SteeringType ";
  retMessage += inifileText;
  retMessage += "}";

  inifileText = inifile.find( "MASS", robotClass.c_str() );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key MASS in section %s of inifile\n",
	       robotClass.c_str());
      return 0;
    }
  retMessage += "{Mass ";
  retMessage += inifileText;
  retMessage += "}";

  inifileText = inifile.find( "MAX_SPEED", robotClass.c_str() );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key MAX_SPEED in section %s of inifile\n",
	       robotClass.c_str());
      return 0;
    }
  maxWheelVelocity = atof(inifileText);
  retMessage += "{MaxSpeed ";
  retMessage += inifileText;
  retMessage += "}";

  inifileText = inifile.find( "MAX_TORQUE", robotClass.c_str() );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key MAX_TORQUE in section %s of inifile\n",
	       robotClass.c_str());
      return 0;
    }
  retMessage += "{MaxTorque ";
  retMessage += inifileText;
  retMessage += "}";

  inifileText = inifile.find( "MAX_FRONT_STEER", robotClass.c_str() );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key MAX_FRONT_STEER in section %s of inifile\n",
	       robotClass.c_str());
      return 0;
    }
  retMessage += "{MaxFrontSteer ";
  retMessage += inifileText;
  retMessage += "}";

  inifileText = inifile.find( "MAX_REAR_STEER", robotClass.c_str() );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key MAX_REAR_STEER in section %s of inifile\n",
	       robotClass.c_str());
      return 0;
    }
  retMessage += "{MaxRearSteer ";
  retMessage += inifileText;
  retMessage += "}";

  retMessage += ";\r\n";
  return( retMessage );
}

int parseGetConf( const vector <string>&toks, int client_fd)
{
  int i;
  string retMessage;
#define LOCK_WRITE_THREAD pthread_mutex_lock(&write_socket_data.mutex)
#define UNLOCK_WRITE_THREAD pthread_mutex_unlock(&write_socket_data.mutex)

  if( -1 == (i=findToken(toks, "Type")))
    {
      fprintf( stderr, "Unable to find 'Type' in GetConf message\n" );
      return 0;
    }
  dbprintf( "Received Conf message: %s\n", toks[i+1].c_str() );
  if( !nocase_cmp(toks[i+1], "Robot") )
    {
      retMessage = parseRobotConf();
    }
  else if( !nocase_cmp(toks[i+1], "MisPkg") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "MisPkg";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "Effecter") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "Effecter";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "Camera") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "Camera";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "Encoder") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "Encoder";
      retMessage += "} ";
      retMessage += "{Name ECLeft} {Resolution 0.0174} ";
      retMessage += "{Name ECRight} {Resolution 0.0174} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "GroundTruth") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "GroundTruth";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "INS") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "INS";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "INU") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "INU";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "Odometry") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "Odometry";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "RFID") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "RFID";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "RangeScanner") )
    {
      retMessage = parseRangeScannerConf();
    }
  else if( !nocase_cmp(toks[i+1], "Sonar") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "Sonar";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "Touch") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "Touch";
      retMessage += "} ";
      retMessage += ";\r\n";
    }
  else if( !nocase_cmp(toks[i+1], "VictSensor") )
    {
      retMessage = "CONF ";
      retMessage += "{Type ";
      retMessage += "VictSensor";
      retMessage += "} ";
      retMessage += ";\r\n";
    }

  else
    {
      fprintf( stderr, "GETCONF: Type %s not supported yet\n",
	       toks[i+1].c_str() );
    }
  LOCK_WRITE_THREAD;
  write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
  UNLOCK_WRITE_THREAD;
  return 1;
#undef LOCK_WRITE_THREAD
#undef UNLOCK_WRITE_THREAD
}


int parseGetGeo( const vector <string>&toks, int client_fd)
{
  int i;
  string retMessage;
  const char *inifileText;
#define LOCK_WRITE_THREAD pthread_mutex_lock(&write_socket_data.mutex)
#define UNLOCK_WRITE_THREAD pthread_mutex_unlock(&write_socket_data.mutex)

  //    dumpTokens (toks);

  if( -1 == (i=findToken(toks, "Type")))
    {
      fprintf( stderr, "Unable to find 'Type' in GetGeo message\n" );
      return 0;
    }
  if( !nocase_cmp(toks[i+1], "Robot") )
    {
      retMessage = "GEO ";
      inifileText = inifile.find( "TYPE", robotClass.c_str() );
      if( inifileText == NULL )
	{
	  fprintf( stderr, "unable to find key TYPE in section %s of inifile\n",
		   robotClass.c_str());
	  return 0;
	}
      retMessage += "{Type ";
      retMessage += inifileText;
      retMessage += "} ";
      retMessage += "{Name " + robotName + "} ";

      inifileText = inifile.find( "DIMENSIONS", robotClass.c_str() );
      if( inifileText == NULL )
	{
	  fprintf( stderr, "unable to find key DIMENSIONS in section %s of inifile\n",
		   robotClass.c_str());
	  return 0;
	}
      retMessage += "{Dimensions ";
      retMessage += inifileText;
      retMessage += "}";

      inifileText = inifile.find( "COG", robotClass.c_str() );
      if( inifileText == NULL )
	{
	  fprintf( stderr, "unable to find key COG in section %s of inifile\n",
		   robotClass.c_str());
	  return 0;
	}
      retMessage += "{COG ";
      retMessage += inifileText;
      retMessage += "}";

      inifileText = inifile.find( "WHEEL_RADIUS", robotClass.c_str() );
      if( inifileText == NULL )
	{
	  fprintf( stderr, "unable to find key WHEEL_RADIUS in section %s of inifile\n",
		   robotClass.c_str());
	  return 0;
	}
      retMessage += "{WheelRadius ";
      retMessage += inifileText;
      retMessage += "}";

      inifileText = inifile.find( "WHEEL_SEPARATION", robotClass.c_str() );
      if( inifileText == NULL )
	{
	  fprintf( stderr, "unable to find key WHEEL_SEPARATION in section %s of inifile\n",
		   robotClass.c_str());
	  return 0;
	}
      retMessage += "{WheelSeparation ";
      retMessage += inifileText;
      retMessage += "}";

      inifileText = inifile.find( "WHEEL_BASE", robotClass.c_str() );
      if( inifileText == NULL )
	{
	  fprintf( stderr, "unable to find key WHEEL_BASE in section %s of inifile\n",
		   robotClass.c_str());
	  return 0;
	}
      retMessage += "{WheelBase ";
      retMessage += inifileText;
      retMessage += "}";

      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
      printf( "%s\n", retMessage.c_str() );
      return 1;
    }
  else if( !nocase_cmp(toks[i+1], "MisPkg") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "MisPkg";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "Effecter") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "Effecter";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "Camera") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "Camera";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "Encoder") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "Encoder";
      retMessage += "} ";
      /*
	GEO
	Type
	Encoder
	Name
	ECLeft
	Location
	0.0000
	0.0000
	0.0000
	Orientation
	0.0000
	0.0000
	1.5706
	Mount
	LeftFWheel
	Name
	ECRight
	Location
	0.0000
	0.0000
	0.0000
	Orientation
	0.0000
	0.0000
	1.5706
	Mount
	RightFWheel
      */
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "GroundTruth") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "GroundTruth";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "INS") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "INS";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "INU") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "INU";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "Odometry") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "Odometry";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "RFID") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "RFID";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "RangeScanner") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "RangeScanner";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "Sonar") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "Sonar";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "Touch") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "Touch";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }
  else if( !nocase_cmp(toks[i+1], "VictSensor") )
    {
      retMessage = "GEO ";
      retMessage += "{Type ";
      retMessage += "VictSensor";
      retMessage += "} ";
      retMessage += ";\r\n";
      LOCK_WRITE_THREAD;
      write(client_fd, retMessage.c_str(), strlen(retMessage.c_str()));
      UNLOCK_WRITE_THREAD;
    }

  else
    {
      fprintf( stderr, "GETGEO: Type %s not supported yet\n",
	       toks[i+1].c_str() );
    }

  return 1;
#undef LOCK_WRITE_THREAD
#undef UNLOCK_WRITE_THREAD
}

int parseDrive( const vector <string>&toks, int client_fd)
{
  int i;
  string retMessage;
  static bool normalized = false;
  double speed;

#define LOCK_MOTOR(index) pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK_MOTOR(index) pthread_mutex_unlock(&smart_motor_data[index].mutex)


  if( -1 == (i=findToken(toks, "DRIVE")))
    {
      fprintf( stderr, "Unable to find 'DRIVE' in DRIVE message\n" );
      return 0;
    }

  dumpTokens (toks);
  while( ++i < (int)toks.size() )
    {
      if( !nocase_cmp(toks[i], "Left") )
	{
	  i++;
	  speed = atof(toks[i].c_str());
	  if( normalized )
	    {
	      speed /= maxWheelVelocity;
	      if( speed > 1 )
		speed = 1;
	      if( speed < -1 )
		speed = -1;
	    }
	  else
	    {
	      if( speed > maxWheelVelocity )
		speed = maxWheelVelocity;
	      if( speed < -maxWheelVelocity )
		speed = -maxWheelVelocity;
	    }
	  LOCK_MOTOR(LEFT_MOTOR);
	  smart_motor_data[LEFT_MOTOR].speed = speed;
	  UNLOCK_MOTOR(LEFT_MOTOR);
	}
      else if( !nocase_cmp(toks[i], "Right") )
	{
	  i++;
	  speed = atof(toks[i].c_str());
	  if( normalized )
	    {
	      speed /= maxWheelVelocity;
	      if( speed > 1 )
		speed = 1;
	      if( speed < -1 )
		speed = -1;
	    }
	  else
	    {
	      if( speed > maxWheelVelocity )
		speed = maxWheelVelocity;
	      if( speed < -maxWheelVelocity )
		speed = -maxWheelVelocity;
	    }
	  LOCK_MOTOR(RIGHT_MOTOR);
	  smart_motor_data[RIGHT_MOTOR].speed = speed;
	  UNLOCK_MOTOR(RIGHT_MOTOR);
	}
      else if( !nocase_cmp(toks[i], "Normalized") )
	{
	  i++;
	  if( !nocase_cmp(toks[i], "true" ) )
	    normalized = true;
	  else
	    normalized = false;
	}
      else if( !nocase_cmp(toks[i], "Light") )
	{
	  i++;
	  printf( "Received light command %s\n", toks[i].c_str() );
	}
      else if( !nocase_cmp(toks[i], "Flip") )
	{
	  i++;
	  printf( "Received flip command %s\n", toks[i].c_str() );
	}

    }
  return 1;
#undef LOCK_MOTOR
#undef UNLOCK_MOTOR
}
/*
  'handle_usarsim_message' takes a null-terminated message 'msg',
  parses it and handles it. For motor commands, by acting on the global 'smart_motor_data'
  array, either writing speeds into it or reading position out of it.
  Results are written out on the mutex protected socket.
*/
static int handle_usarsim_message(char * msg, int client_fd)
{
  /*
  double thleft;
  double thright;
  double wleft;
  double wright;
  */
  int retval;
  enum {BUFFERLEN = USARSIM_MSG_LEN};
  //  char buffer[BUFFERLEN];	/* messages back to client go here */
  vector < string > inMsgs;
  vector < string > toks;
  // Message delimiters
  const static string mDelims = "\r\n";
  const static string tDelims = "}{, \r\n";
  string inBuf = msg;

#define LOCK_MOTOR(index) pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK_MOTOR(index) pthread_mutex_unlock(&smart_motor_data[index].mutex)
#define LOCK_WRITE_THREAD pthread_mutex_lock(&write_socket_data.mutex)
#define UNLOCK_WRITE_THREAD pthread_mutex_unlock(&write_socket_data.mutex)

  retval = 0;		      /* default is OK, set to -1 if needed */

  /*
    Parse out the USARsim message and handle it. Some messages
    will require setting motor speeds by writing into the global
    data structure. Some will require querying position from the
    global data structure.
  */

  //Put valid tokens in 'inMsgs', remaining characters left in input buffer
  Tokenize2 (inBuf, inMsgs, mDelims);
      
  // Parse all valid mags recieved from socket
  while (!inMsgs.empty ())
    {
      toks.clear ();
      // Tokenize one message at a time
      Tokenize2 (inMsgs.back (), toks, tDelims);

      // discard parsed message from vector
      inMsgs.pop_back ();
	  
      // No tokens might indicate internal errors.
      if (toks.empty ())
	{
	  printf
	    ("** WARNING ** <matilda::parseMsg> : No tokens found in message\n");
	  break;
	}
	  
      // Send message to appropriate parser
      if (!nocase_cmp (toks[0], "NFO"))
	printf( "token: NFO\n");
      else if (!nocase_cmp (toks[0], "INIT"))
	{
	  if( !parseInit(toks, client_fd) )
	    retval = -1;
	}
      else if (!nocase_cmp (toks[0], "GETCONF"))
	{
	  if( !parseGetConf(toks, client_fd) )
	    retval = -1;
	}
      else if (!nocase_cmp (toks[0], "GETGEO"))
	{
	  if( !parseGetGeo(toks, client_fd) )
	    retval = -1;
	}
      else if (!nocase_cmp (toks[0], "DRIVE"))
	{
	  if( !parseDrive(toks, client_fd) )
	    retval = -1;
	}
      else
	{
	  printf
	    ("** WARNING ** <matilda::parseMsg> : Msg Type %s not handled\n",
	     toks[0].c_str ());
	  retval = -1;
	}
    }
  return retval;

#undef LOCK_MOTOR
#undef UNLOCK_MOTOR
#undef LOCK_WRITE_THREAD
#undef UNLOCK_WRITE_THREAD
}

typedef struct {
  /* 'client_fd' is the file descriptor to the client socket. It is
     opened by the main thread before the message thread runs. */
  int client_fd;
} message_thread_arg;


/* odometry code
 */
static void setOdometry(double leftRad, double rightRad)
{
  static double oldLeft=leftRad, oldRight=rightRad;
  double leftChange, rightChange, leftMeters, rightMeters, meters;

  // determine how man radians since last read
  leftChange = leftRad - oldLeft;
  rightChange = rightRad - oldRight;

  // update old values
  oldLeft = leftRad;
  oldRight = rightRad;

  // convert radians to meters
  leftMeters = leftChange * wheelRadius;
  rightMeters = rightChange * wheelRadius;

  // how far have we traveled since last time?
  meters = (leftMeters + rightMeters)/2.0;
  
  // accumulate total rotation around our center
  theta += (leftMeters - rightMeters)/wheelSeparation;
  //  fprintf( stderr, "Right: %f Left: %f theta: %f wheel: %f\n",
  //	   rightMeters, leftMeters, theta, wheelSeparation );

  // clip the rotation to plus or minus 360 degrees
  theta -= (double)((int)(theta/M_PI*2.))*M_PI*2.;

  // accumulate our position
  xPos += meters * cos(theta);
  yPos += meters * sin(theta);

  fprintf( stderr, "Our position is: %f <%f %f>\n", theta, xPos, yPos );
}

/* thread for writing out the encoder data
 */
typedef struct {
  /* 'client_fd' is the file descriptor to the client socket. It is
     opened by the main thread before the message thread runs. */
  int client_fd;
  double period;
} encoder_thread_arg;

static void * encoder_thread_code( void * arg)
{
  double period = ((encoder_thread_arg *) arg)->period;
  double myTime;
  double thleft;
  double thright;

#define LOCK_MOTOR(index) pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK_MOTOR(index) pthread_mutex_unlock(&smart_motor_data[index].mutex)
#define LOCK_WRITE_THREAD pthread_mutex_lock(&write_socket_data.mutex)
#define UNLOCK_WRITE_THREAD pthread_mutex_unlock(&write_socket_data.mutex)

  for(;;)
    {
      myTime = etime();
      LOCK_MOTOR(LEFT_MOTOR);
      thleft = smart_motor_data[LEFT_MOTOR].position;
      UNLOCK_MOTOR(LEFT_MOTOR);
      LOCK_MOTOR(RIGHT_MOTOR);
      thright = smart_motor_data[RIGHT_MOTOR].position;
      UNLOCK_MOTOR(RIGHT_MOTOR);
      //      fprintf(stdout, "LEFT=%f RIGHT=%f\n", thleft, thright);
      setOdometry( thleft, thright );

      // FIXME-- why lock then unlock without doing anything?
      LOCK_WRITE_THREAD;
      UNLOCK_WRITE_THREAD;

      myTime  = etime() - myTime;
      esleep( period - myTime );
    }
  return NULL;

#undef LOCK_MOTOR
#undef UNLOCK_MOTOR
#undef LOCK_WRITE_THREAD
#undef UNLOCK_WRITE_THREAD
}

static void * message_thread_code(void * arg)
{
  int client_fd;
  int nchars;
  enum {BUFFERLEN = USARSIM_MSG_LEN};
  char buffer[BUFFERLEN];
  char * buffer_ptr;
  char * buffer_end;
  int BUILDLEN = BUFFERLEN;
  char * build;
  char * build_ptr;
  char * build_end;
  ptrdiff_t offset;

#define LOCK_MOTOR(index) pthread_mutex_lock(&smart_motor_data[index].mutex)
#define UNLOCK_MOTOR(index) pthread_mutex_unlock(&smart_motor_data[index].mutex)

  /* copy out our heap args and free them */
  client_fd = ((message_thread_arg *) arg)->client_fd;
  free(arg);

  build = (char *) malloc(BUILDLEN * sizeof(char));
  build_ptr = build;
  build_end = build + BUILDLEN;

  for (;;) {
    nchars = read(client_fd, buffer, sizeof(buffer));
    if (nchars > 0) {
      /* compose a valid message, since we may have read a fragment */
      buffer_ptr = buffer;
      buffer_end = buffer + nchars;
      while (buffer_ptr != buffer_end) {
	if (build_ptr == build_end) {
	  /* grow build buffer to hold more data */
	  offset = build_ptr - build;
	  BUILDLEN *= 2;
	  build = (char *) realloc(build, BUILDLEN * sizeof(char));
	  build_ptr = build + offset;
	  build_end = build + BUILDLEN;
	}
	*build_ptr++ = *buffer_ptr;
	if (*buffer_ptr++ == USARSIM_MSG_DELIMITER) {
	  offset = build_ptr - build;
	  build[offset] = 0;
	  build_ptr = build;
	  /* now a full message is in 'build', with length 'offset',
	     including the delimiter kept and null-terminated for your
	     pleasure */
	  //	  dbprintf("got ``%s''\n", build);
	  (void) handle_usarsim_message(build, client_fd);
	} /* if (*buffer_ptr++ == USARSIM_MSG_DELIMITER */
      } /* while (buffer_ptr != buffer_end) */
    } else {
      break;
    }
  } /* for (;;) */

  /* if we ever break out of the loop above, we'll get here */
  free(build);
  close(client_fd);
  dbprintf("message thread %d done\n", client_fd);

  return NULL;

#undef LOCK_MOTOR
#undef UNLOCK_MOTOR
}

/* send an init message (by USARSim format) to the client
   return 1 on success, 0 on failure
 */
int sendInit( int client_fd )
{
  char iniString[255];
  const char *inifileText;

#define LOCK_WRITE_THREAD pthread_mutex_lock(&write_socket_data.mutex)
#define UNLOCK_WRITE_THREAD pthread_mutex_unlock(&write_socket_data.mutex)

  inifileText = inifile.find( "LEVEL", "USARSIM" );
  if( inifileText == NULL )
    {
      fprintf( stderr, "unable to find key LEVEL in section USARSIM of inifile\n");
      return 0;
    }
  
  sprintf( iniString, "NFO {Gametype RealRobot}{Level %s}{TimeLimit 0};\r\n",
	   inifileText );

  LOCK_WRITE_THREAD;
  write(client_fd, iniString, strlen(iniString));
  UNLOCK_WRITE_THREAD;
  dbprintf( "Wrote %s to client\n", iniString );
  return 1;
#undef LOCK_WRITE_THREAD
#undef UNLOCK_WRITE_THREAD
}

/*
  Syntax: matildaserver <TCP port number> {<serial ports ...>}
*/
int main(int argc, char * argv[])
{
  pthread_t thread;
  pthread_attr_t attr;
  struct termios save_termios;
  motor_thread_arg * motor_thread_arg_ptr;
  message_thread_arg * message_thread_arg_ptr;
  encoder_thread_arg * encoder_thread_arg_ptr;
  char * sports[SMART_MOTOR_NUM];
  int port;
  int index;
  int socket_fd;
  int client_fd;
  char *inifileName = "../etc/robots.cfg";
  const char *inifileText;
  char inifileSearch[200];
  int sensorCount;

  if (argc == 2) 
    {
      dbprintf( "%s using %s as ini file\n",
		argv[0], argv[1]);
      inifileName = argv[1];
    }

  if( (inifile.open(inifileName)) != 0 )
    {
      fprintf(stderr, "%s: Unable to open configuration file %s\n",
	      argv[0], inifileName );
      return 1;
    }

  inifileText = inifile.find("PORT_NUMBER", "USARSIM" );
  if( inifileText == NULL )
    {
      fprintf( stderr, "%s unable to find key PORT_NUMBER in section USARSIM of inifile %s\n",
	       argv[0], inifileText );
    }
  port = atoi(inifileText);

  // some things must be read into global memory, these are also read by the GetConf message
  // set max robot velocity
  inifileText = inifile.find( "MAX_SPEED", robotClass.c_str() );
  if( inifileText == NULL )
    maxWheelVelocity = 1;
  else
    maxWheelVelocity = atof(inifileText);
  inifileText = inifile.find( "WHEEL_RADIUS", robotClass.c_str() );
  if( inifileText == NULL )
    wheelRadius = 1;
  else
    wheelRadius = atof(inifileText);
  inifileText = inifile.find( "WHEEL_SEPARATION", robotClass.c_str() );
  if( inifileText == NULL )
    wheelSeparation = 1;
  else
    wheelSeparation = atof(inifileText);

#define BEFOREARGS 2		/* how many args before serial port args */
  for (index = 0; index < SMART_MOTOR_NUM; index++) {
    if (argc <= index + BEFOREARGS) sports[index] = NULL;
    else sports[index] = argv[index + BEFOREARGS];
  }

  /* get our socket server file descriptor */
  socket_fd = get_socket_fd(port);
  if (socket_fd < 0) {
    fprintf(stderr, "can't serve socket port %d\n", port);
    return 1;
  }

  /* initialize the Smart Motor data and start the read threads */
  for (index = 0; index < SMART_MOTOR_NUM; index++) {
    pthread_mutex_init(&smart_motor_data[index].mutex, NULL);
    /* insert actual values here, or inifile this */
    inifileText = inifile.find( "COUNTS_PER_RAD", "MOTOR_ENCODER" );
    if( inifileText == NULL )
      smart_motor_data[index].counts_per_rad = COUNTS_PER_RAD;
    else
      smart_motor_data[index].counts_per_rad = atof(inifileText);
    smart_motor_data[index].rads_per_count = 1.0 / smart_motor_data[index].counts_per_rad;
    /* in case we're simulating, start somewhere interesting */
    smart_motor_data[index].position = index * M_PI;
    /* but be nicely stopped */
    smart_motor_data[index].speed = 0.0;
    motor_thread_arg_ptr = (motor_thread_arg*)malloc(sizeof(*motor_thread_arg_ptr));
    if (NULL == motor_thread_arg_ptr) {
      fprintf(stderr, "can't allocate args for read thread %d\n", index);
      return 1;
    }
    if (sports[index] == NULL) {
      motor_thread_arg_ptr->serial_fd = -1;
    } else {
      motor_thread_arg_ptr->serial_fd = open(sports[index], O_RDWR);
      if (motor_thread_arg_ptr->serial_fd < 0) {
	fprintf(stderr, "can't open serial port %s\n", sports[index]);
	return 1;
      }
      tty_raw(motor_thread_arg_ptr->serial_fd, SMART_MOTOR_BAUD, &save_termios);
    }
    motor_thread_arg_ptr->index = index;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if (0 != pthread_create(&thread, &attr, motor_thread_code, (void *) motor_thread_arg_ptr)) {
      fprintf(stderr, "can't spawn motor thread %d\n", index);
      return 1;
    }
  }

  for (;;) {
    dbprintf("waiting for client connection...\n");
    client_fd = get_client_fd(socket_fd);
    if (client_fd < 0) {
      break;
    }
    dbprintf("got one on fd %d\n", client_fd);

    if( !sendInit( client_fd ) )
      {
	fprintf(stderr, "%s: unable to send init string to client\n", argv[0] );
      }
    message_thread_arg_ptr = (message_thread_arg*)malloc(sizeof(*message_thread_arg_ptr));
    if (NULL == message_thread_arg_ptr) {
      fprintf(stderr, "can't allocate args for message thread %d\n", index);
      return 1;
    }
    message_thread_arg_ptr->client_fd = client_fd;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if (0 != pthread_create(&thread, &attr, message_thread_code, (void *) message_thread_arg_ptr)) {
      fprintf(stderr, "can't spawn message thread %d\n", client_fd);
      return 1;
    }
    /* start a thread for each of the sensors in the .ini file */
    sensorCount=1;
    for(;;)
      {
	sprintf( inifileSearch, "SENSOR_%d", sensorCount );
	inifileText = inifile.find(inifileSearch, robotClass.c_str() );
	if( inifileText == NULL )
	  break;
	if (!nocase_cmp (inifileText, "MOTOR_ENCODER")) // start encoder thread
	  {
	    encoder_thread_arg_ptr = (encoder_thread_arg*)malloc(sizeof(*encoder_thread_arg_ptr));
	    if (NULL == encoder_thread_arg_ptr) {
	      fprintf(stderr, "can't allocate args for encoder thread\n" );
	      return 1;
	    }
	    inifileText = inifile.find("PERIOD", "MOTOR_ENCODER" );
	    if( inifileText == NULL )
	      {
		fprintf( stderr, "No period found for motor encoders, defaulting to 1\n" );
		encoder_thread_arg_ptr->period = 1;
	      }
	    else
	      encoder_thread_arg_ptr->period = atof(inifileText);
	    pthread_attr_init(&attr);
	    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	    if (0 != pthread_create(&thread, &attr, encoder_thread_code, (void *) encoder_thread_arg_ptr)) {
	      fprintf(stderr, "can't spawn encoder thread %d\n", client_fd);
	      return 1;
	    }
	  }
	else
	  fprintf( stderr, "Unknown sensor type: %s\n", inifileText );
	sensorCount++;
      }
  } /* while (! done) */

  dbprintf("server done\n");

  return 0;
}

/*
  Modification history:

  $Log: matildaInterface.cc,v $
  Revision 1.6  2008/09/08 14:24:39  dr_steveb
  Fixed lact of M_TWOPI on some linux systems

  Revision 1.5  2008/08/27 20:39:39  dr_steveb
  Added encoders

  Revision 1.4  2008/08/18 18:13:55  dr_steveb
  Commit before new version release. Includes new code for matilda real robot

  Revision 1.3  2008/05/15 19:56:56  proctor
  Took out the annoying debug printing

  Revision 1.2  2008/03/17 18:40:50  dr_steveb
  Fixed right wheel max velocity

  Revision 1.1  2008/03/17 18:36:20  dr_steveb
  Created server that will interface the matilda platform with moast thru simWare

  Revision 1.2  2008/03/07 19:22:47  proctor
  Fixed bug in debug print statement

  Revision 1.1  2008/03/07 17:45:42  proctor
  Added matilda code in servo, and socket, tty utils in tools

*/

