/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file wrap_servo.cc

  \brief Wrapper functions that provide C linkage for reading position
  and writing position setpoints to external servo systems, such as
  Go Motion.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.4 $
  $Date: 2009/05/04 20:55:03 $
  \endcode

  \author Fred Proctor
*/

#include <stdio.h>
#include <rcs.hh>
#include "moastTypes.hh"
#include "moastNmlOffsets.hh"
#include "servoMisJA.hh"

#define THIS_PROCESS "primMis"

static RCS_CMD_CHANNEL * servoMisJACmdBuf = NULL;
static RCS_STAT_CHANNEL * servoMisJAStatBuf = NULL;
static RCS_CMD_CHANNEL * servoMisJACfgBuf = NULL;
static RCS_STAT_CHANNEL * servoMisJASetBuf = NULL;

static ServoMisJACmdMove moveCmd;
static ServoMisJAStat * statBufPtr = NULL;

static void cleanupServoMisJANmlBuffers(void)
{
  if (NULL != servoMisJACmdBuf)
    delete servoMisJACmdBuf;
  if (NULL != servoMisJAStatBuf)
    delete servoMisJAStatBuf;
  if (NULL != servoMisJACfgBuf)
    delete servoMisJACfgBuf;
  if (NULL != servoMisJASetBuf)
    delete servoMisJASetBuf;

  servoMisJASetBuf = NULL;
  servoMisJACfgBuf = NULL;
  servoMisJAStatBuf = NULL;
  servoMisJACmdBuf = NULL;

  statBufPtr = NULL;

  return;
}

static int initServoMisJANmlBuffers(char * config_file, int vehid, int compid)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupServoMisJANmlBuffers();

  sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_CMD_NAME, vehid, compid);
  servoMisJACmdBuf =
    new RCS_CMD_CHANNEL(servoMisJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMisJACmdBuf->valid()) {
    delete servoMisJACmdBuf;
    servoMisJACmdBuf = 0;
    return -1;
  }

  sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_STAT_NAME, vehid, compid);
  servoMisJAStatBuf =
    new RCS_STAT_CHANNEL(servoMisJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMisJAStatBuf->valid()) {
    delete servoMisJAStatBuf;
    servoMisJAStatBuf = 0;
    statBufPtr = NULL;
    return -1;
  }
  statBufPtr = (ServoMisJAStat *) servoMisJAStatBuf->get_address();

  sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_CFG_NAME, vehid, compid);
  servoMisJACfgBuf =
    new RCS_CMD_CHANNEL(servoMisJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMisJACfgBuf->valid()) {
    delete servoMisJACfgBuf;
    servoMisJACfgBuf = 0;
    return -1;
  }

  sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_SET_NAME, vehid, compid);
  servoMisJASetBuf =
    new RCS_STAT_CHANNEL(servoMisJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMisJASetBuf->valid()) {
    delete servoMisJASetBuf;
    servoMisJASetBuf = 0;
    return -1;
  }

  return 0;
}

// these are the functions with C linkage called by ext_init,quit

extern "C" int c_init(char * init_string)
{
  int t;
  ServoMisJACmdInit initMsg;

  rcs_print("c_init %s\n", init_string);

  if (0 != initServoMisJANmlBuffers(init_string, 1, 1)) {
    rcs_print("c_init: can't initialize servo NML buffers\n");
    return -1;
  }
  servoMisJACmdBuf->write(initMsg);

  for (t = 0; t < SERVO_MIS_JA_LINK_MAX; t++) {
    moveCmd.linkCmd[t].linkID = t + 1;
    moveCmd.linkCmd[t].value = 0;
    moveCmd.linkCmd[t].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;
  }
  moveCmd.serial_number = 17;
  moveCmd.linkCmd_length = SERVO_MIS_JA_LINK_MAX;

  return 0;
}

extern "C" void c_quit(void)
{
  rcs_print("cpp_quit\n");

  cleanupServoMisJANmlBuffers();

  return;
}

#define INDEX(j) ((j) + 0)

extern "C" double c_read_pos(int joint)
{
  int index;

  if (joint >= 0 && joint < SERVO_MIS_JA_LINK_MAX) {
    index = INDEX(joint);
    // only query NML on the first one
    if (joint == 0) {
      servoMisJAStatBuf->read();
    }
    return statBufPtr == NULL ? 0.0 : statBufPtr->linkStat[index].jointVal;
  }

  return 0.0;
}

extern "C" int c_write_pos(int joint, double pos)
{
  int index;

  if (joint >= 0 && joint < SERVO_MIS_JA_LINK_MAX) {
    index = INDEX(joint);
    moveCmd.linkCmd[index].value = pos;
    // only write once per set, on the first one
    if (joint == 0) {
      moveCmd.serial_number++;
      servoMisJACmdBuf->write(moveCmd);
    }
    return 0;
  }

  return -1;
}

/*
  Modification history:

  $Log: wrap_servo.cc,v $
  Revision 1.4  2009/05/04 20:55:03  proctor
  Disabled shared library _so builds for Cygwin; added disclaimer and brief
  comments for servo wrapper code

*/
