#define PLAYER2_0
/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   playerProxies.cc
  \brief  Provides an interface to player for simWare.

Based on code from the PlayerViewer that is protected by the GNU public license 
as shown below:

 *  Copyright (C) Andrew Howard 2002
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.5 $
  $Date: 2009/04/30 16:55:34 $
  \endcode

  \author Stephen Balakirsky
  \date   10/02/2008
*/
#include "playerInfSkin.hh"

/////////////// ClientConnection Class routines
ClientConnection::ClientConnection( playerc_client_t *clientIn, int which )
{
  char section[256];

  client = clientIn;
  device.addr = client->devinfos[which].addr;
  device.drivername = strdup(client->devinfos[which].drivername);
  device.subscribe = 0;
/*
  snprintf(section, sizeof(section), "%s:%d:%s",
#ifdef PLAYER2_0
	   playerc_lookup_name(device.addr.interf),
#else // use for player 2.1.1	
	   interf_to_str(device.addr.interf),
#endif
	   device.addr.index,
	   device.drivername);*/
  uniqueName =  section;
}

///*************** Sonar routines **************///
// Destroy a sonar device
void sonar_destroy(ClientConnection *connectionIn)
{
  playerc_sonar_t *self = (playerc_sonar_t*)connectionIn->device.proxy;
  if (self->info.subscribed)
    playerc_sonar_unsubscribe(self);
  playerc_sonar_destroy(self);

  delete(connectionIn);
}

// Update a sonar device
void sonar_update(ClientConnection *connectionIn)
{
  playerc_sonar_t *self = (playerc_sonar_t*)connectionIn->device.proxy;
  SIMwareKBSenSonar *pSonar;
  int count;
  char sonarName[255];

  if( !self->info.fresh )
      return;
  self->info.fresh = 0;

  // lock sensor mutex
  kb->lockSenMutex();
  for( count=0; count<self->pose_count; count++ )
    {
      snprintf( sonarName, sizeof(sonarName), "%s:%d",
		connectionIn->uniqueName.c_str(), count );
      pSonar = (SIMwareKBSenSonar *)kb->getSensor(sonarName, true, SW_KB_SONAR_SENSOR_TYPE);
      if( pSonar == NULL )
	{
	  printf( "<PlayerInfSkin::configureKBSonar> ** ERROR ** Could not create sonar %s in KB\n",
		  sonarName );
	  return;
	}
      pSonar->setTimeStamp(self->info.datatime);
      // put the return into the KB
      pSonar->setRangeReturn( self->scan[count] );
    }
  // unlock sensor mutex
  kb->unlockSenMutex();
}

void ClientConnection::sonar_create()
{  
  SIMwareKBSenSonar *kbSonar;
  playerc_sonar_t *self;
  int count;
  char sonarName[255];

  kb->lockSenMutex();

  device.proxy = (void*)playerc_sonar_create(client, device.addr.index);
  device.fndestroy = (fndestroy_t) sonar_destroy;
  device.fnupdate = (fnupdate_t) sonar_update;

  self = (playerc_sonar_t*)device.proxy;
  if (playerc_sonar_subscribe(self, PLAYER_OPEN_MODE) != 0)
    printf("libplayerc error: %s", playerc_error_str());
  // Get the sonar geometry
  if (playerc_sonar_get_geom(self) != 0)
    printf("libplayerc error: %s", playerc_error_str());

  printf( "creating %d sonars named with root: %s\n", self->pose_count, uniqueName.c_str() );
  for( count=0; count<self->pose_count; count++ )
    {
      snprintf( sonarName, sizeof(sonarName), "%s:%d",
		uniqueName.c_str(), count );
      kbSonar = (SIMwareKBSenSonar *)kb->getSensor(sonarName, true, SW_KB_SONAR_SENSOR_TYPE);
      if( kbSonar == NULL )
	{
	  printf( "<PlayerInfSkin::configureKBSonar> ** ERROR ** Could not create sonar %s in KB\n",
		  sonarName );
	  return;
	}
#ifdef PLAYER2_0
      kbSonar->setSen2MountXYZ(self->poses[count].px, self->poses[count].py, 0 );
#else //use for player 2.2.1 
      kbSonar->setSen2MountRPY(0, 0, self->poses[count].pyaw );
#endif
      //kbSonar->setSen2MountRPY(0, 0, self->poses[count].pa );
    }
  kb->unlockSenMutex();
  return;
}

///*************** Laser routines **************///
// Destroy a laser device
void laser_destroy(ClientConnection *connectionIn)
{
  playerc_laser_t *laser = (playerc_laser_t*)connectionIn->device.proxy;
  if (laser->info.subscribed)
    playerc_laser_unsubscribe(laser);
  playerc_laser_destroy(laser);

  delete(connectionIn);
}

// Update a laser device
void laser_update(ClientConnection *connectionIn)
{
  playerc_laser_t *self = (playerc_laser_t*)connectionIn->device.proxy;
  SIMwareKBSenLinescan *pLine;
  static vector< double >retVals;
  int i;

  pLine = (SIMwareKBSenLinescan *)kb->getSensor(connectionIn->uniqueName);

  if( !self->info.fresh )
      return;
  self->info.fresh = 0;

  // check for kbLine initialization
  if( pLine->getFov() == 0 )
    {
      if (playerc_laser_get_geom(self) != 0)
	printf("laser_create:libplayerc error: %s", playerc_error_str());
      pLine->setBeamSpacing( self->scan_res );
      pLine->setFov( self->scan_count * self->scan_res );
      pLine->setBeamDispersion(self->scan_res);
      //NOW HARD CODED!
      //  kbLine->setMaxRange( laser->max_range ); 
      pLine->setMaxRange( 4.0 );
      pLine->setMinRange( 0 );
      printf( "Laser scan count: %d BeamSpacing: %f fov: %f minR: %f maxR: %f\n", 
	      self->scan_count, pLine->getBeamSpacing(),
	      pLine->getFov(), pLine->getMinRange(), pLine->getMaxRange() );
      kb->setSettingsFresh(true);
    }

  
  if ( pLine == NULL )   
    {
      printf(" <PlayerInfSkin: ** WARNING ** Could not find laserscanner %s in KB\n", 
	     connectionIn->uniqueName.c_str());
      return;
    }
  // lock sensor mutex
  kb->lockSenMutex();

  retVals.clear();
  pLine->setTimeStamp(self->info.datatime);
  // copy data into the vector.
  for( i=0; i < self->scan_count; i++)
    retVals.push_back(self->ranges[i]);
  // put the vector into the KB
  pLine->setRangeReturn( retVals );
  //  printf( "Processed %d laser return\n", retVals.size() );

  // unlock sensor mutex
  kb->unlockSenMutex();

}

void ClientConnection::laser_create()
{  
  ClientConnection *newConnection;
  SIMwareKBSenLinescan *kbLine;
  playerc_laser_t *laser;

  device.fndestroy = (fndestroy_t) laser_destroy;
  device.fnupdate = (fnupdate_t) laser_update;
  device.proxy = (void*)playerc_laser_create(client, device.addr.index);

  printf( "creating sensor %s for config laser\n", uniqueName.c_str() );
  kbLine = (SIMwareKBSenLinescan *)kb->getSensor(uniqueName, true, SW_KB_LINESCAN_SENSOR_TYPE);
  if( kbLine == NULL )
    {
      printf( "<PlayerInfSkin::configureKBLineScan> ** ERROR ** Could not create linescanner %s in KB\n",
	      uniqueName.c_str() );
      return;
    }

  laser = (playerc_laser_t*)device.proxy;

  if (playerc_laser_subscribe(laser, PLAYER_OPEN_MODE) != 0)
    printf("laser_create:libplayerc error: %s", playerc_error_str());

  // Get the laser geometry
  if (playerc_laser_get_geom(laser) != 0)
    printf("laser_create:libplayerc error: %s", playerc_error_str());


  kbLine->setSen2MountXYZ(laser->pose[0], laser->pose[1], -.3 );
  kbLine->setSen2MountRPY(0, 0, laser->pose[2]);
  kbLine->setFov(0);
  /* for some reason, this does not work here...
  kbLine->setBeamSpacing( laser->scan_res );
  kbLine->setFov( laser->scan_count * laser->scan_res );
  kbLine->setBeamDispersion(laser->scan_res);
  */
  //NOW HARD CODED!
  //  kbLine->setMaxRange( laser->max_range ); 
  kbLine->setMaxRange( 4.0 );
  kbLine->setMinRange( 0 );


  printf( "Laser geometry: <%f %f %f> \n",
	  laser->pose[0],
	  laser->pose[1],
	  laser->pose[2] );
  //	  laser->pose[3],
  //	  laser->pose[4],
  //	  laser->pose[5] );
  printf( "Laser min range: %f max range: %f beam spacing: %f fov: %f\n",
	  kbLine->getMinRange(), kbLine->getMaxRange(), kbLine->getBeamSpacing(), 
	  kbLine->getFov() );
}

///*************** Power routines **************///
// Destroy a power device
void power_destroy(ClientConnection *connectionIn)
{
  playerc_power_t *power = (playerc_power_t*)connectionIn->device.proxy;
  if (power->info.subscribed)
    playerc_power_unsubscribe(power);
  playerc_power_destroy(power);

  delete(connectionIn);
}

// Update a power device
void power_update(ClientConnection *connectionIn)
{
  playerc_power_t *self = (playerc_power_t*)connectionIn->device.proxy;

  kb->lockBotMutex();
  kb->bot->setBatteryLevel(self->percent);
  kb->unlockBotMutex();
}

void ClientConnection::power_create()
{  
  device.proxy = (void*)playerc_power_create(client, device.addr.index);
  device.fndestroy = (fndestroy_t) power_destroy;
  device.fnupdate = (fnupdate_t) power_update;

  if (playerc_power_subscribe((playerc_power_t*)device.proxy, PLAYER_OPEN_MODE) != 0)
    printf("libplayerc error: %s", playerc_error_str());

  return;
}

///*************** Position2d routines **************///
// Destroy a position2d device
void position2d_destroy(ClientConnection *connectionIn)
{
  playerc_position2d_t *position2d = (playerc_position2d_t*)connectionIn->device.proxy;
  if (position2d->info.subscribed)
    playerc_position2d_unsubscribe(position2d);
  playerc_position2d_destroy(position2d);

  delete(connectionIn);
}

// Update a position2d device
void position2d_update(ClientConnection *connectionIn)
//void position2d_update( playerc_position2d_t *self )
{
  playerc_position2d_t *self = (playerc_position2d_t*)connectionIn->device.proxy;
  SIMwareKBSenOdometry *pOdom;
  SIMwareKBSenGrdTruth *pGrdTruth;

  /*
  if( !self->info.fresh )
    {
      //      printf("No fresh data\n");
      return;
    }
  self->info.fresh = 0;
  */
  
  // lock sensor mutex
  kb->lockSenMutex();

  //  printf( "In position2d_update() for %s %f %f %f\n",connectionIn->uniqueName.c_str(),  
  //          self->px, self->py, self->pa );

  if( !strcmp(connectionIn->device.drivername, "rflex") 
      || !strcmp(connectionIn->device.drivername, "p2os") 
      || !strcmp(connectionIn->device.drivername, "stage") )
    {
      //	  p2d->RequestGeom(); 
      pOdom = (SIMwareKBSenOdometry *)kb->getSensor(connectionIn->uniqueName);
      if( pOdom == NULL )   
	{
	  printf(" <PlayerInfSkin: ** WARNING ** Could not find odometry %s in KB\n", 
		 connectionIn->uniqueName.c_str());
	  // unlock sensor mutex
	  kb->unlockSenMutex();
	  return;
	}
      pOdom->setX(self->px);
      if( !strcmp(connectionIn->device.drivername, "p2os") )
	{
	  pOdom->setOrient(-self->pa);
	  pOdom->setY(-self->py);
	}
      else
	{
	  pOdom->setOrient(self->pa);
	  pOdom->setY(self->py);
	}

      pOdom->setTimeStamp(self->info.datatime);

      //      printf("PlayerInfSkin: reported %s position: %f %f %f at time %f\n", connectionIn->device.drivername,
      //     pOdom->getX(), pOdom->getY(), pOdom->getOrient(), pOdom->getTimeStamp() );

    }
  else if( !strcmp(connectionIn->device.drivername, "sicknav200") )
    {
      pGrdTruth = (SIMwareKBSenGrdTruth *)kb->getSensor(connectionIn->uniqueName);
      if( pGrdTruth == NULL )
	{
	  printf( "<PlayerInfSkin: ** WARNING ** Could not find grdtruth %s in KB\n",
		  connectionIn->uniqueName.c_str());
	  // unlock sensor mutex
	  kb->unlockSenMutex();
	  return;
	}
      pGrdTruth->setLocation( self->px, -self->py, 0 );
      pGrdTruth->setOrientation(0,0,-self->pa/32.768*M_PI );
      pGrdTruth->setTimeStamp(self->info.datatime);

      /*
      printf("PlayerInfSkin: reported nav200 position: <%f %f %f> <%f %f %f> at time %f\n",
	     self->px, self->py, self->pa,
	     pGrdTruth->getX(), pGrdTruth->getY(), pGrdTruth->getYaw(), 
	     pGrdTruth->getTimeStamp() );
      */

    }

  // unlock sensor mutex
  kb->unlockSenMutex();
}

void ClientConnection::position2d_create()
{  
  device.proxy = (void*)playerc_position2d_create(client, device.addr.index);
  device.fndestroy = (fndestroy_t) position2d_destroy;
  device.fnupdate = (fnupdate_t) position2d_update;

  if (playerc_position2d_subscribe((playerc_position2d_t*)device.proxy, PLAYER_OPEN_MODE) != 0)
    printf("libplayerc error: %s", playerc_error_str());

  // Get the position2d geometry
  if (playerc_position2d_get_geom((playerc_position2d_t*)device.proxy) != 0)
    printf("libplayerc error: %s", playerc_error_str());

#ifdef PLAYER2_0
  playerc_position2d_t *position2dProxy = (playerc_position2d_t*)device.proxy;
  printf( "Position2d geometry: Position:<%f %f %f> Pose:<%f %f %f> Size:<%f %f>\n",
	  position2dProxy->px,
	  position2dProxy->py,
	  position2dProxy->pa,
	  position2dProxy->pose[0],
	  position2dProxy->pose[1],
	  position2dProxy->pose[2],
	  position2dProxy->size[0],
	  position2dProxy->size[1]);
#endif
  return;
}

ClientConnection *create_proxy(playerc_client_t *client, int which)
{
  ClientConnection *newConnection = new ClientConnection(client, which);

  switch (client->devinfos[which].addr.interf)
  {
    /*
    case PLAYER_ACTARRAY_CODE:
      connection->device->proxy = actarray_create(mainwnd, opt, client,
                                    connection->device->addr.index,
                                    connection->device->drivername,
                                    connection->device->subscribe);
      connection->device->fndestroy = (fndestroy_t) actarray_destroy;
      connection->device->fnupdate = (fnupdate_t) actarray_update;
      break;

    case PLAYER_AIO_CODE:
      connection->device->proxy = aio_create(mainwnd, opt, client,
                                    connection->device->addr.index,
                                    connection->device->drivername,
                                    connection->device->subscribe);
      connection->device->fndestroy = (fndestroy_t) aio_destroy;
      connection->device->fnupdate = (fnupdate_t) aio_update;
      break;

     case PLAYER_BUMPER_CODE:
       connection->device->proxy = bumper_create(mainwnd, opt, client,
                                   connection->device->addr.index,
                                   connection->device->drivername,
                                   connection->device->subscribe);
       connection->device->fndestroy = (fndestroy_t) bumper_destroy;
       connection->device->fnupdate = (fnupdate_t) bumper_update;
       break;

     case PLAYER_DIO_CODE:
       connection->device->proxy = dio_create(mainwnd, opt, client,
                                  connection->device->addr.index,
                                  connection->device->drivername,
                                  connection->device->subscribe);
       connection->device->fndestroy = (fndestroy_t) dio_destroy;
       connection->device->fnupdate = (fnupdate_t) dio_update;
       break;

    case PLAYER_IR_CODE:
      connection->device->proxy = ir_create(mainwnd, opt, client,
                                   connection->device->addr.index, connection->device->drivername, connection->device->subscribe);
      connection->device->fndestroy = (fndestroy_t) ir_destroy;
      connection->device->fnupdate = (fnupdate_t) ir_update;
      break;
    */
    case PLAYER_LASER_CODE:
      newConnection->laser_create();
      break;

    case PLAYER_POWER_CODE:
      newConnection->power_create();
      break;

    case PLAYER_POSITION2D_CODE:
      newConnection->position2d_create();
      break;

    case PLAYER_SONAR_CODE:
      newConnection->sonar_create();
      break;
      /*
    case PLAYER_PTZ_CODE:
      connection->device.proxy = ptz_create(mainwnd, opt, client,
                                        connection->device.addr.index,
                                        connection->device.drivername,
                                        connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) ptz_destroy;
      connection->device.fnupdate = (fnupdate_t) ptz_update;
      break;

    case PLAYER_RANGER_CODE:
      connection->device.proxy = ranger_create(mainwnd, opt, client,
                                    connection->device.addr.index,
                                    connection->device.drivername,
                                    connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) ranger_destroy;
      connection->device.fnupdate = (fnupdate_t) ranger_update;
      break;

    case PLAYER_WIFI_CODE:
      connection->device.proxy = wifi_create(mainwnd, opt, client,
                                        connection->device.addr.index,
                                        connection->device.drivername,
                                        connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) wifi_destroy;
      connection->device.fnupdate = (fnupdate_t) wifi_update;
      break;

    case PLAYER_BLOBFINDER_CODE:
      connection->device.proxy = blobfinder_create(mainwnd, opt, client,
					connection->device.addr.index,
                                        connection->device.drivername,
					connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) blobfinder_destroy;
      connection->device.fnupdate = (fnupdate_t) blobfinder_update;
      break;

    case PLAYER_CAMERA_CODE:
      connection->device.proxy = camera_create(mainwnd, opt, client,
					connection->device.addr.index,
                                        connection->device.drivername,
					connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) camera_destroy;
      connection->device.fnupdate = (fnupdate_t) camera_update;
      break;



    case PLAYER_FIDUCIAL_CODE:
      connection->device.proxy = fiducial_create(mainwnd, opt, client,
                                      connection->device.addr.index,
				      connection->device.drivername,
				      connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) fiducial_destroy;
      connection->device.fnupdate = (fnupdate_t) fiducial_update;
      break;

    case PLAYER_GRIPPER_CODE:
      connection->device.proxy = gripper_create(mainwnd, opt, client,
				     connection->device.addr.index,
				     connection->device.drivername,
				     connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) gripper_destroy;
      connection->device.fnupdate = (fnupdate_t) gripper_update;
      break;

    case PLAYER_MAP_CODE:
      connection->device.proxy = map_create(mainwnd, opt, client,
				 connection->device.addr.index,
				 connection->device.drivername,
				 connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) map_destroy;
      connection->device.fnupdate = (fnupdate_t) map_update;
      break;

    case PLAYER_VECTORMAP_CODE:
      connection->device.proxy = vectormap_create(mainwnd, opt, client,
         connection->device.addr.index,
         connection->device.drivername,
         connection->device.subscribe);
      connection->device.fndestroy = (fndestroy_t) vectormap_destroy;
      connection->device.fnupdate = (fnupdate_t) vectormap_update;
      break;
      */

    default:
      printf( "\n\nplayerProxies: unknown proxy %s:%d\n\n", 
	      client->devinfos[which].drivername,
	      client->devinfos[which].addr.interf );
      newConnection->device.proxy = NULL;
      newConnection->device.fndestroy = NULL;
      newConnection->device.fnupdate = NULL;
      break;
  }

  return newConnection;
}

/*
  Modification history:

  $Log: playerProxies.cc,v $
  Revision 1.5  2009/04/30 16:55:34  dr_steveb
  Fixed reporting of laser location

  Revision 1.4  2009/04/20 18:37:54  dr_steveb
  Fixed coordinate system for p2os player proxy

  Revision 1.3  2009/02/17 14:57:35  dr_steveb
  Fixes so that player will work with both 2.0 and 2.1. Also added fresh tag to database for settings.

  Revision 1.2  2009/02/04 15:22:21  wshackle
  allow moast to run on top of player/stage for testing-debug purposes.
  if player is not started or playerc_client_connect()  fails, shutdown more
  gracefully without the core dump.
  eliminate several compiler warnings about unused variables.

  Revision 1.1  2008/11/20 16:01:57  dr_steveb
  Fixed player skin to work with c interface of player 2.1.1

*/
