/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simpleSimInfSkin.cc

  \brief Implementation of the simple simulation inferior skin
  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.16 $
  $Date: 2007/09/04 21:06:10 $
  \endcode
  \author Fred Proctor
*/

/*
  WHEREWASI

  You just added the robot type stuff, for a ground robot. It compiles
  OK, but seg faults during the "myrun" script.
*/

#include <stdio.h>		// printf, vfprintf
#include <string.h>		// strerror
#include <stdarg.h>		// va_list, ...
#include <math.h>		// fmod, sqrt
#include <float.h>		// DBL_MAX
#include <pthread.h>		// pthread_mutex_init
#include "rcs.hh"		// esleep, PM_RPY,CARTESIAN
#include "swBaseSkin.hh"
#include "swKBTypes.hh"
#include "swKBSen.hh"		// SIMwareKBSenINU,GPS,GrdTruth,Linescan
#include "swKBBot.hh"		// SIMwareKBGroundVehicle
#include "simpleSimInfSkin.hh"	// these decls

#define DEBUG

static int
dprintf(const char * fmt, ...)
{
  va_list ap;

#ifdef DEBUG
  va_start(ap, fmt);
  fprintf(stderr, "simpleSimInfSkin: ");
  vfprintf(stderr, fmt, ap);
  fflush(stderr);
  va_end(ap);
#endif

  return 0;
}

#define DEFAULT_CYCLE_TIME 0.001

#define DEFAULT_X_JUMP_SIZE 1.0
#define DEFAULT_Y_JUMP_SIZE 1.0
#define DEFAULT_Z_JUMP_SIZE 1.0
#define DEFAULT_JUMP_TIME 10.0

#define DEFAULT_X_DRIFT_SPEED 0.001
#define DEFAULT_Y_DRIFT_SPEED 0.001
#define DEFAULT_Z_DRIFT_SPEED 0.001
#define DEFAULT_R_DRIFT_SPEED 0.001
#define DEFAULT_P_DRIFT_SPEED 0.001
#define DEFAULT_YAW_DRIFT_SPEED 0.001

#define DEFAULT_OBSTACLE_SPACING 10.0
#define DEFAULT_OBSTACLE_RADIUS 0.5
#define DEFAULT_SENSOR_HEIGHT 1.0
#define DEFAULT_SENSOR_PITCH 0.0

#define DEFAULT_ROBOT_TYPE SW_KB_GROUND_VEHICLE_TYPE
#define DEFAULT_STEER_TYPE SW_KB_SKID_STEER_TYPE
#define DEFAULT_SKID_WHEEL_RADIUS 1.0
#define DEFAULT_SKID_WHEEL_SEPARATION 1.0
#define DEFAULT_SKID_WMAX 1.0
#define DEFAULT_SKID_ALPHAMAX 1.0

//! how far away our sensor can see
#define MAX_SENSOR_RANGE 1.0

#define M_SQ(x) ((x)*(x))

SimpleWorld::SimpleWorld(void)
{
  int t;

  (void) pthread_mutex_init(&mutex, NULL);

  delta_t = DEFAULT_CYCLE_TIME;
  delta_t_inv = 1.0 / DEFAULT_CYCLE_TIME;
  acmRandomInit(&rnd);

  xTrue = 0.0;
  yTrue = 0.0;
  yawTrue = 0.0;
  vTrue = 0.0;
  wTrue = 0.0;

  xJump = 0.0;
  yJump = 0.0;
  zJump = 0.0;
  xJumpSize = DEFAULT_X_JUMP_SIZE;
  yJumpSize = DEFAULT_Y_JUMP_SIZE;
  zJumpSize = DEFAULT_Z_JUMP_SIZE;
  jumpTime = DEFAULT_JUMP_TIME;
  jumpClock = acmRandomReal(&rnd) * jumpTime;

  xDrift = 0.0;
  yDrift = 0.0;
  zDrift = 0.0;
  rDrift = 0.0;
  pDrift = 0.0;
  yawDrift = 0.0;
  xDriftSpeed = DEFAULT_X_DRIFT_SPEED;
  yDriftSpeed = DEFAULT_Y_DRIFT_SPEED;
  zDriftSpeed = DEFAULT_Z_DRIFT_SPEED;
  rDriftSpeed = DEFAULT_R_DRIFT_SPEED;
  pDriftSpeed = DEFAULT_P_DRIFT_SPEED;
  yawDriftSpeed = DEFAULT_YAW_DRIFT_SPEED;

  doObstacles = true;
  obstacleSpacing = DEFAULT_OBSTACLE_SPACING;
  obstacleRadius = DEFAULT_OBSTACLE_RADIUS;
  sensorHeight = DEFAULT_SENSOR_HEIGHT;
  sensorPitch = DEFAULT_SENSOR_PITCH;
  for (t = 0; t < SIMPLE_SIM_MAX_SENSOR_VALUES; t++) {
    sensorValues[t] = 0.0;
  }

  robotType = DEFAULT_ROBOT_TYPE;

  steerType = DEFAULT_STEER_TYPE;
  skid.wheel_radius = DEFAULT_SKID_WHEEL_RADIUS;
  skid.wheel_radius_inv = 1.0 / skid.wheel_radius;
  skid.wheel_separation = DEFAULT_SKID_WHEEL_SEPARATION;
  skid.wheel_separation_inv = 1.0 / skid.wheel_separation;
  skid.wmax = DEFAULT_SKID_WMAX;
  skid.alphamax = DEFAULT_SKID_ALPHAMAX;
  skid.wLeft = 0.0;
  skid.wRight = 0.0;
  skid.wLeftReq = 0.0;
  skid.wRightReq = 0.0;

  return;
}

SimpleWorld::~SimpleWorld(void)
{
  (void) pthread_mutex_unlock(&mutex);

  return;
}

bool
SimpleWorld::lock(void)
{
  return 0 == pthread_mutex_lock(&mutex) ? true : false;
}

bool
SimpleWorld::unlock(void)
{
  return 0 == pthread_mutex_unlock(&mutex) ? true : false;
}

PM_CARTESIAN
SimpleWorld::getGPS(void)
{
  PM_CARTESIAN gps;

  lock();

  gps.x = xTrue + xJump;
  gps.y = yTrue + yJump;
  gps.z = zJump;

  unlock();

  return gps;
}

PM_POSE
SimpleWorld::getINS(void)
{
  PM_POSE ins;
  PM_RPY rpy;
  
  lock();

  ins.tran.x = xTrue + xDrift;
  ins.tran.y = yTrue + yDrift;
  ins.tran.z = zDrift;
  
  rpy.r = rDrift;
  rpy.p = pDrift;
  rpy.y = yawTrue + yawDrift;

  unlock();

  ins.rot = rpy;

  return ins;
}

PM_POSE
SimpleWorld::getTruth(void)
{
  PM_RPY rpy;
  PM_POSE truth;
  
  lock();

  truth.tran.x = xTrue;
  truth.tran.y = yTrue;
  rpy.y = yawTrue;

  unlock();

  truth.tran.z = 0.0;
  rpy.r = 0.0;
  rpy.p = 0.0;
  truth.rot = rpy;

  return truth;
}

double
SimpleWorld::getSensorValue(int i)
{
  if (i < 0 || i >= SIMPLE_SIM_MAX_SENSOR_VALUES) return 0.0;

  return sensorValues[i];
}

bool
SimpleWorld::setRobotType(SIMwareRobotType type)
{
  bool retval;

  switch (type) {
  case SW_KB_GROUND_VEHICLE_TYPE:
  case SW_KB_LEGGED_ROBOT_TYPE:
  case SW_KB_NAUTIC_ROBOT_TYPE:
  case SW_KB_AERIAL_ROBOT_TYPE:
    robotType = type;
    retval = true;
    break;

  default:
    retval = false;
    break;
  }

  return retval;
}

bool
SimpleWorld::setSteerType(SIMwareSteerType type)
{
  bool retval;

  switch (type) {
  case SW_KB_ACKERMAN_STEER_TYPE: 
  case SW_KB_SKID_STEER_TYPE:
  case SW_KB_OMNI_STEER_TYPE:
  case SW_KB_SUBMARINE_STEER_TYPE: 
    steerType = type;
    retval = true;
    break;

  default:
    retval = false;
    break;
  }

  return retval;
}

bool
SimpleWorld::reqAckerman(double vLinear,
			 double steerAngle)
{
  if (steerType != SW_KB_ACKERMAN_STEER_TYPE) return false;

  if (vLinear < 0.0) ackerman.vLinearReq = 0.0;
  else if (vLinear > ackerman.vLinearMax) ackerman.vLinearReq = ackerman.vLinearMax;
  else ackerman.vLinearReq = vLinear;

  if (steerAngle < -ackerman.steerAngleMax) ackerman.steerAngleReq = -ackerman.steerAngleMax;
  else if (steerAngle > ackerman.steerAngleMax) ackerman.steerAngleReq = ackerman.steerAngleMax;
  else ackerman.steerAngleReq = steerAngle;

  return true;
}

// kinematics for skid steer vehicle

#if 0
// not used yet
static int
skidInvJacobian(double v, double w,
		double R_inv, double L,
		double * w1, double * w2)
{
  *w1 = R_inv * (v + 0.5 * w * L);
  *w2 = R_inv * (v - 0.5 * w * L);

  return 0;
}
#endif

/*
  Ackerman steering follows

  tan(s) = wL/v

  where s is steer angle, w is angular speed, L is wheel base, and v is 
  tangential speed.
 */
static int
ackermanFwdJacobian(double speed, double steer, double L_inv,
		    double * v, double * w)
{
  *v = speed;
  // minus sign necessary for vehicle conventions
  *w = -speed * tan(steer) * L_inv;

  return 0;
}

static int
skidFwdJacobian(double w1, double w2,
		double R, double L_inv,
		double * v, double * w)
{
  *v = 0.5 * R * (w1 + w2);
  *w = L_inv * R * (w1 - w2);

  return 0;
}

static int
computeDs(double v, double w, double th,
	      double delta_t,
	      double *dx, double *dy, double *dth)
{
  double r;
  double cth, sth;
  double dthp, dxp, dyp;
#define MIN_W 1.0e-6

  if (fabs(w) < MIN_W) {
    dthp = 0.0;
    dxp = v * delta_t;
    dyp = 0.0;
  } else {
    r = v / w;
    dthp = w * delta_t;
    dxp = r * sin(dthp);
    dyp = r * (1.0 - cos(dthp));
  }

  cth = cos(th);
  sth = sin(th);

  *dx = cth * dxp - sth * dyp;
  *dy = sth * dxp + cth * dyp;
  *dth = dthp;

#undef MIN_W
  return 0;
}

bool
SimpleWorld::reqSkid(double wLeft,
		     double wRight)
{
  if (steerType != SW_KB_SKID_STEER_TYPE) return false;

  if (wLeft < -skid.wmax) skid.wLeftReq = -skid.wmax;
  else if (wLeft > skid.wmax) skid.wLeftReq = skid.wmax;
  else skid.wLeftReq = wLeft;

  if (wRight < -skid.wmax) skid.wRightReq = -skid.wmax;
  else if (wRight > skid.wmax) skid.wRightReq = skid.wmax;
  else skid.wRightReq = wRight;

  return true;
}

bool
SimpleWorld::reqOmni(double vLateral,
		     double vLinear,
		     double vRot)
{
  if (steerType != SW_KB_OMNI_STEER_TYPE) return false;

  // FIXME-- implement

  return true;
}

bool
SimpleWorld::reqSubmarine(double propeller,
			  double rudder,
			  double sternPlane)
{
  if (steerType != SW_KB_SUBMARINE_STEER_TYPE) return false;

  // FIXME-- implement

  return true;
}

static int
checkObstacle(double x, double y,
	      double spacing, double radius,
	      double * xnew, double * ynew)
{
  double xmod, ymod;
  double xobst, yobst;
  double r2, R2, R_minus_r;
  PM_CARTESIAN vec;

  // translate arbitrary x,y into 0..SPACING square
  xmod = fmod(x, spacing);
  if (xmod < 0.0) xmod += spacing;
  ymod = fmod(y, spacing);
  if (ymod < 0.0) ymod += spacing;
  // R2 is the obstacle radius squared; saves a sqrt op
  R2 = M_SQ(radius);
  vec.z = 0.0;

  /*
    This snippet checks if xmod,ymod is within the obstacle radius
    from xobst,yobst. If so, it computes the unit vector from the
    obstacle center to the xmod,ymod point, stretches this by the
    radius and adds it to the obstacle center to get the point
    on the obstacle edge nearest xmod,ymod. If the point is at the
    center, it favors a point on the x axis.
  */
#define DOIT					\
  r2 = M_SQ(xmod - xobst) + M_SQ(ymod - yobst); \
  if (r2 <= R2) {				\
    R_minus_r = radius - sqrt(r2);		\
    vec.x = xmod - xobst, vec.y = ymod - yobst; \
    if (! isNorm(vec = unit(vec))) {		\
      vec.x = radius, vec.y = vec.z = 0.0;	\
    }						\
    vec = R_minus_r * vec;			\
    *xnew = x + vec.x, *ynew = y + vec.y;	\
    return 1;					\
  }
  xobst = 0.0, yobst = 0.0;	// obstacle 1
  DOIT;
  xobst = spacing;		// obstacle 2
  DOIT;
  yobst = spacing;		// obstacle 3
  DOIT;
  xobst = 0.0;			// obstacle 4
  DOIT;

#undef DOIT
  return 0;
}

void
SimpleWorld::runAckerman(void)
{
  double a_dt;
  double alpha_dt;
  double dx, dy, dth;

  // FIXME

  a_dt = ackerman.aLinearMax * delta_t;

  if (ackerman.vLinearReq - ackerman.vLinear < -a_dt)
    ackerman.vLinear -= a_dt;
  else if (ackerman.vLinearReq - ackerman.vLinear > a_dt)
    ackerman.vLinear += a_dt;
  else
    ackerman.vLinear = ackerman.vLinearReq;

  alpha_dt = ackerman.steerAlphaMax * delta_t;

  if (ackerman.steerAngleReq - ackerman.steerAngle < -alpha_dt)
    ackerman.steerAngle -= alpha_dt;
  else if (ackerman.steerAngleReq - ackerman.steerAngle > alpha_dt)
    ackerman.steerAngle += alpha_dt;
  else
    ackerman.steerAngle = ackerman.steerAngleReq;

  ackermanFwdJacobian(ackerman.vLinear, ackerman.steerAngle,
		      ackerman.wheelbase_inv,
		      &vTrue, &wTrue);

  computeDs(vTrue, wTrue, yawTrue,
	    delta_t,
	    &dx, &dy, &dth);

  xTrue += dx;
  yTrue += dy;
  yawTrue += dth;

  return;
}

void
SimpleWorld::runSkid(void)
{
  double alpha_dt;
  double dx, dy, dth;

  alpha_dt = skid.alphamax * delta_t;

  if (skid.wLeftReq - skid.wLeft < -alpha_dt)
    skid.wLeft -= alpha_dt;
  else if (skid.wLeftReq - skid.wLeft > alpha_dt)
    skid.wLeft += alpha_dt;
  else
    skid.wLeft = skid.wLeftReq;

  if (skid.wRightReq - skid.wRight < -alpha_dt)
    skid.wRight -= alpha_dt;
  else if (skid.wRightReq - skid.wRight > alpha_dt)
    skid.wRight += alpha_dt;
  else
    skid.wRight = skid.wRightReq;

  skidFwdJacobian(skid.wLeft, skid.wRight,
		  skid.wheel_radius, skid.wheel_separation_inv,
		  &vTrue, &wTrue);

  computeDs(vTrue, wTrue, yawTrue,
	    delta_t,
	    &dx, &dy, &dth);

  xTrue += dx;
  yTrue += dy;
  yawTrue += dth;

  return;
}

#define TWO_PI (M_PI + M_PI)
#define NORM(x) while ((x) > TWO_PI) (x) -= TWO_PI; while ((x) < 0.0) (x) += TWO_PI

enum {RANGE_OK = 0, RANGE_TOO_FAR, RANGE_INSIDE};
#define RANGE_INFINITE -1.0

/*
  angleWithin assumes that start, end and angle are all normalized 0..360. 
  "Within" means that moving clockwise from 'start' to 'end', you hit 'angle'.
*/
static int
angleWithin(double start, double end, double angle)
{
  if (start < end) return (angle > start) && (angle < end);
  return (angle < end) || (angle > start);
}

//! calcHorizontalRange assumes SENSOR_ANGLE is 0, or horizontal viewing.

static int
calcHorizontalRange(double R, double xp, double yp, double th, double *range)
{
  double r, ang, alpha, ang1, ang2, rcosb;

  r = sqrt(xp*xp + yp*yp);
  if (r <= R) return RANGE_INSIDE;
  ang = atan2(-yp, -xp);
  alpha = asin(R/r);
  ang1 = ang - alpha;
  ang2 = ang + alpha;
  NORM(ang1);
  NORM(ang2);
  NORM(th);
  if (angleWithin(ang1, ang2, th)) {
    rcosb = r*cos(ang - th);
    *range = rcosb - sqrt(rcosb*rcosb + R*R - r*r);
    return RANGE_OK;
  }
  return RANGE_TOO_FAR;
}

/*!
  \c calcRange takes the cylindrical obstacle radius \b R, the
  X and Y position of the vehicle center \b xp and \b yp relative
  to the obstacle center, and the viewing angle \b th, and computes
  the range to the front edge of the obstacle. The view is pointed
  from the horizontal by \c SENSOR_ANGLE, so if the ground is closer
  than the obstacle, that range is computed.
  
  Returns \c RANGE_OK and sets \b *range if the obstacle or ground are in view.
  Returns \c RANGE_TOO_FAR if the obstacle is out of view.
  Returns \c RANGE_INSIDE if the vehicle is inside the obstacle.
*/

static int 
calcRange(double R,	      //!< vehicle-to-obstacle center distance
	  double xp,		//!< x relative to obstacle center
	  double yp,		//!< y relative to obstacle center
	  double th,		//!< yaw angle you're looking
	  double height,	//!< height of sensor above ground
	  double pitch,		//!< pitch of sensor from horizontal
	  double * range)	//!< calculated range
{
  double sinpitch, cospitch;
  double secSensorAngle;
  double groundRange;
  double a2;
  double b2;
  double ab;
  double r, ang, d, alpha, ang1, ang2, t2, rr;

  // check for extreme cases: horizontal or vertical up, down
  sinpitch = sin(pitch), cospitch = cos(pitch);
  if (sinpitch == 0.0) {
    // horizontal
    return calcHorizontalRange(R, xp, yp, th, range);
  }
  if (cospitch == 0.0) {
    // vertical
    if (pitch < 0.0) {
      // straight down
      *range = height;
      return RANGE_OK;
    } else {
      return RANGE_TOO_FAR;
    }
  }

  secSensorAngle = 1.0 / cos(pitch);
  groundRange = pitch > 0.0 ? DBL_MAX : -height / sin(pitch);
  a2 = R * secSensorAngle;
  b2 = R;
  ab = M_SQ(R) * secSensorAngle;

  r = sqrt(xp*xp + yp*yp);	// veh center to obs center
  if (r <= R) return RANGE_INSIDE;
  ang = atan2(-yp, -xp);
  d = r * secSensorAngle;
  alpha = atan(sqrt(b2 / (M_SQ(d) - a2)));
  ang1 = ang - alpha;
  ang2 = ang + alpha;
  NORM(ang1);
  NORM(ang2);
  NORM(th);
  t2 = M_SQ(tan(ang - th));

  if (angleWithin(ang1, ang2, th)) {
    rr = (b2*d - ab*sqrt(a2*t2 - M_SQ(d)*t2 + b2))/(a2*t2 + b2) * secSensorAngle;
    // report closer of ground or obstacle
    if (rr < groundRange) {
      *range = rr;
    } else {
      *range = groundRange;
    }
    return RANGE_OK;
  }

  if (groundRange < MAX_SENSOR_RANGE) {
    *range = groundRange;
    return RANGE_OK;
  }

  return RANGE_TOO_FAR;
}

void
SimpleWorld::update(void)
{
  double xNew, yNew;
  double th, thStart, dth;
  double xmod, ymod, xp, yp;
  double range, groundRange;
  int t;
#define RND(x) (acmRandomReal(&rnd) * (x))
#define RNG(x) ((2.0 * acmRandomReal(&rnd) - 1.0) * (x))

  lock();

  // run vehicle equations
  switch (steerType) {
  case SW_KB_ACKERMAN_STEER_TYPE: 
    break;

  case SW_KB_SKID_STEER_TYPE:
    runSkid();
    break;

  case SW_KB_OMNI_STEER_TYPE:
    break;

  case SW_KB_SUBMARINE_STEER_TYPE: 
    break;

  default:
    break;
  }

  // keep vehicle outside obstacle
  if (doObstacles) {
    if (checkObstacle(xTrue, yTrue,
		      obstacleSpacing,
		      obstacleRadius,
		      &xNew, &yNew)) {
      xTrue = xNew, yTrue = yNew;
      vTrue = wTrue = 0.0;
    }
  }

  // jump GPS within range of 'jumpTime' seconds, in '?JumpSize' range
  jumpClock -= delta_t;
  if (jumpClock <= 0.0) {
    jumpClock = RND(jumpTime);
    xJump = RNG(xJumpSize);
    yJump = RNG(yJumpSize);
    zJump = RNG(zJumpSize);
  }
  // these will be added to true values by GPS access function above

  // drift INS cumulatively each cycle
  xDrift += RNG(xDriftSpeed * delta_t);
  yDrift += RNG(yDriftSpeed * delta_t);
  zDrift += RNG(zDriftSpeed * delta_t);
  rDrift += RNG(rDriftSpeed * delta_t);
  pDrift += RNG(pDriftSpeed * delta_t);
  yawDrift += RNG(yawDriftSpeed * delta_t);
  // these will be added to true values by INS access function above

  // update linescan sensor

  thStart = yawTrue - M_PI_4;
  dth = M_PI_2 / SIMPLE_SIM_MAX_SENSOR_VALUES;

  if (doObstacles) {
    xmod = fmod(xTrue, obstacleSpacing);
    if (xmod < 0.0) xmod += obstacleSpacing;
    ymod = fmod(yTrue, obstacleSpacing);
    if (ymod < 0.0) ymod += obstacleSpacing;
  } else {
    groundRange = sensorPitch >= 0.0 ? DBL_MAX : -sensorHeight / sin(sensorPitch);
  }

  for (t = 0, th = thStart;
       t < SIMPLE_SIM_MAX_SENSOR_VALUES;
       t++, th += dth) {
    if (doObstacles) {
#define DOIT if (RANGE_OK == \
		 calcRange(obstacleRadius, xp, yp, th, \
			   sensorHeight, sensorPitch, \
			   &range)) { \
	  sensorValues[t] = range; \
	  continue; \
	}
      xp = xmod, yp = ymod;	// first obstacle at 0,0
      DOIT;
      xp = xmod - obstacleSpacing; // second obstacle at L,0
      DOIT;
      yp = ymod - obstacleSpacing; // third obstacle at L,L
      DOIT;
      xp = xmod;		// fourth obstacle at 0,L
      DOIT;
#undef DOIT
      sensorValues[t] = MAX_SENSOR_RANGE;
    } else {
      sensorValues[t] = groundRange < MAX_SENSOR_RANGE ? groundRange : MAX_SENSOR_RANGE;
    }
  }

  unlock();

#undef RNG
#undef RND
  return;
}

// the command thread code
void
SimpleSimInfSkin::worldThrCode(void)
{
  PM_CARTESIAN gpsCart;
  //SIMWareKBSenGPSCoordinate gpsCoordinate;
  bool done;
  double timeIn, timeTaken, time;
  int t;

  dprintf("worldThrCode started\n");

  worldThrTimeMin = DBL_MAX;
  worldThrTimeMax = DBL_MIN;

  for (done = false; done == false; esleep(simpleWorld.delta_t)) {
    timeIn = etime();
    simpleWorld.update();
    time = 0.5 * (timeIn + etime());

    timeIn = etime();
    
    kb->bot->setTimeStamp(time);

    inu->setTimeStamp(time);
    inu->setOrientation((PM_RPY) simpleWorld.getINS().rot);

    gps->setTimeStamp(time);
    gpsCart = simpleWorld.getGPS();
    //gpsData.northing = gpsCart.x;
    //gpsData.easting = gpsCart.y;
    //gpsData.elevation = gpsCart.z;
    //gpsData.zone = 1;		// we're always in zone 1 north
    //gps->setPosition(gpsData);

    grd->setTimeStamp(time);
    grd->setPose(simpleWorld.getTruth());

    linescan->setTimeStamp(time);
    for (t = 0; t < SIMPLE_SIM_MAX_SENSOR_VALUES; t++) {
      linescan->setRangeReturn(t, simpleWorld.getSensorValue(t));
    }

    timeTaken = etime() - timeIn;
    if (timeTaken < worldThrTimeMin) worldThrTimeMin = timeTaken;
    else if (timeTaken > worldThrTimeMax) worldThrTimeMax = timeTaken;
  }

  dprintf("worldThrCode done\n");

  return;
}

static void *
worldThrCodeStartRoutine(void * arg)
{
  // arg is 'this'
  ((SimpleSimInfSkin *) arg)->worldThrCode();

  return NULL;
}

void
SimpleSimInfSkin::doMobCmd(SIMwareMobCmd * cmd)
{
  int cmdType; 
  SIMwareAckermanCmd * ackerCmd;
  SIMwareSkidCmd * skidCmd;
  SIMwareOmniCmd * omniCmd;
  SIMwareSubmarineCmd * submarineCmd;

  cmdType = cmd->getMobCmdType();
  switch (cmdType) {
  case SW_MOB_CMD_INIT_TYPE: 
  case SW_MOB_CMD_POWERDOWN_TYPE: 
    break;

  case SW_MOB_CMD_DRV_ACKERMAN_TYPE:
    ackerCmd = (SIMwareAckermanCmd *) cmd;
    simpleWorld.reqAckerman(ackerCmd->vLinear,
			    ackerCmd->steerAngle);
    break;

  case SW_MOB_CMD_DRV_SKID_TYPE:
    skidCmd = (SIMwareSkidCmd *) cmd;
    simpleWorld.reqSkid(skidCmd->wLeft,
			skidCmd->wRight);
    break;

  case SW_MOB_CMD_DRV_OMNI_TYPE: 
    omniCmd = (SIMwareOmniCmd *) cmd;
    simpleWorld.reqOmni(omniCmd->vLateral,
			omniCmd->vLinear,
			omniCmd->vRot);
    break;

  case SW_MOB_CMD_DRV_SUBMARINE_TYPE: 
    submarineCmd = (SIMwareSubmarineCmd *) cmd;
    simpleWorld.reqSubmarine(submarineCmd->propeller,
			     submarineCmd->rudder,
			     submarineCmd->sternPlane);
    break;

  default:
    break;
  } // end switch (cmdType)

  return;
}

void
SimpleSimInfSkin::doSenCmd(SIMwareSenCmd * cmd)
{
  map <const char *, SIMwareKBSen *, strCmp >::iterator senItr;
  int cmdType;
  int senType;

  cmdType = cmd->getSenCmdType();

  switch (cmdType) {
  case SW_SEN_CMD_INIT_TYPE:
    // init all the sensors
    for (senItr = kb->senMap.begin(); senItr != kb->senMap.end(); senItr++) {
      senType = senItr->second->getSensorType();
      switch (senType) {
      case SW_KB_ODOMETRY_SENSOR_TYPE:
      case SW_KB_ENCODER_SENSOR_TYPE:
      case SW_KB_INU_SENSOR_TYPE:
      case SW_KB_GPS_SENSOR_TYPE:
      case SW_KB_GRD_TRUTH_SENSOR_TYPE:
      case SW_KB_LINESCAN_SENSOR_TYPE:
	dprintf("doSenCmd: initing %s\n", senItr->first);
	break;

      default:
	dprintf("doSenCmd: unknown sensor type %d\n", senType);
	break;
      }
    }

  case SW_SEN_CMD_POWERDOWN_TYPE:
    break;

  default:
    dprintf("doSenCmd: unknown command %d\n", cmdType);
    break;
  }

  return;
}

#if 0
// awaiting decls
void
SimpleSimInfSkin::doMisCmd(SIMwareMisCmd * cmd)
{
  return;
}

void
SimpleSimInfSkin::doEffCmd(SIMwareEffCmd * cmd)
{
  return;
}

void
SimpleSimInfSkin::doSetCmd(SIMwareSetCmd * cmd)
{
  return;
}
#endif

// the command thread code
void
SimpleSimInfSkin::cmdThrCode(void)
{
  SIMwareCmdMsg * cmdMsg;
  int type;
  double timeIn, timeTaken;
  bool done;

  dprintf("cmdThrCode started\n");

  cmdThrTimeMin = DBL_MAX;
  cmdThrTimeMax = DBL_MIN;

  for (done = false; done == false; esleep(simpleWorld.delta_t)) {
    timeIn = etime();
    // check for new command
    cmdMsg = kb->popCmdFromQ();
    timeTaken = etime() - timeIn;
    if (timeTaken < cmdThrTimeMin) cmdThrTimeMin = timeTaken;
    else if (timeTaken > cmdThrTimeMax) cmdThrTimeMax = timeTaken;

    // handle command
    if (NULL != cmdMsg) {
      type = cmdMsg->getCmdType();
      switch (type) {
      case SW_MOB_CMD_TYPE:
	doMobCmd((SIMwareMobCmd *) cmdMsg);
	break;
      case SW_SEN_CMD_TYPE:
	doSenCmd((SIMwareSenCmd *) cmdMsg);
	break;
#if 0
	// awaiting decls
      case SW_MIS_CMD_TYPE:
	doMisCmd((SIMwareMisCmd *) cmdMsg);
	break;
      case SW_EFF_CMD_TYPE:
	doEffCmd((SIMwareEffCmd *) cmdMsg);
	break;
      case SW_SET_CMD_TYPE:
	doSetCmd((SIMwareSetCmd *) cmdMsg);
	break;
#endif
      default:
	dprintf("cmdThrCode error: command %d not recognized\n", type);
	break;
      }	// end of switch (cmdMsg->getCmdType())
    } 
    // else null command means no command, so sleep a bit
  } // end of for (; done == false; esleep(simpleWorld.delta_t))

  dprintf("cmdThrCode done\n");

  return;
}

static void *
cmdThrCodeStartRoutine(void * arg)
{
  // arg is 'this'
  ((SimpleSimInfSkin *) arg)->cmdThrCode();

  return NULL;
}

SimpleSimInfSkin::SimpleSimInfSkin(void)
{
  dprintf("ctor called\n");

  vehicleID = 0;
  platformTypeID = 0;
  startPoseID = 0;

  inu = NULL;
  gps = NULL;
  grd = NULL;
  linescan = NULL;
}

SimpleSimInfSkin::~SimpleSimInfSkin(void)
{
  dprintf("dtor called\n");
}

/*
  iniLoad() will load INI file params from 'filename', returning INI_OK
  if everything is fine, INI_INVALID if there is some fatal problem, or
  INI_DEFAULT if one or more default values were used.

  Values loaded:

  [SIMPLE_SIM.INFSKIN]

  ; cycle time for vehicle simulation, [s]
  CYCLE_TIME = cycle time for simulation, [s]

  ; GPS-related noise parameters, [m], [s]
  X_JUMP_SIZE = 1.0
  Y_JUMP_SIZE  = 1.0
  Z_JUMP_SIZE = 1.0
  JUMP_TIME = 10.0

  ; INS-related noise parameters [m/s], [deg/s]
  X_DRIFT_SPEED = 0.001
  Y_DRIFT_SPEED = 0.001
  Z_DRIFT_SPEED = 0.001
  R_DRIFT_SPEED = 0.001
  P_DRIFT_SPEED = 0.001
  YAW_DRIFT_SPEED = 0.001

  ; enable obstacles by setting non-zero
  OBSTACLES = 1
  ; how far apart obstacles are, center-to-center, [m]
  OBSTACLE_SPACING = 10.0
  ; obstacle radius, [m]
  OBSTACLE_RADIUS = 0.5

  ; how high above ground is the line scan sensor, [m]
  SENSOR_HEIGHT = 1.0
  ; where sensor is pointing from horizontal, negative down, [deg]
  SENSOR_PITCH = -30.0

  ; type of vehicle, one of GROUND_VEHICLE, LEGGED_ROBOT, 
  ; NAUTIC_ROBOT or AERIAL_ROBOT
  ROBOT_TYPE = GROUND_VEHICLE

  ; type of steering, one of ACKERMAN, SKID, OMNI or SUBMARINE
  STEER_TYPE = SKID

  ; for ACKERMAN steering
  ; wheelbase, [m]
  WHEELBASE = 1.0
  ; max linear acceleration, [m/s]
  AMAX = 1.0
  ; max steering acceleration, [deg/s/s]
  ALPHAMAX = 360.0

  ; for SKID steering
  ; wheel radius, [m]
  WHEEL_RADIUS = 1.0
  ; separation between wheels, aka axle length, [m]
  WHEEL_SEPARATION = 1.0
  ; max angular wheel speed, [deg/s];
  WMAX = 360.0
  ; max angular wheel acceleration, [deg/s/s]
  ALPHAMAX = 360.0
*/

static SIMwareRobotType
robotNameToNum(const char * name)
{
  if (! strcmp(name, "GROUND_VEHICLE")) return SW_KB_GROUND_VEHICLE_TYPE;
  if (! strcmp(name, "LEGGED_ROBOT")) return SW_KB_LEGGED_ROBOT_TYPE;
  if (! strcmp(name, "NAUTIC_ROBOT")) return SW_KB_NAUTIC_ROBOT_TYPE;
  if (! strcmp(name, "AERIAL_ROBOT")) return SW_KB_AERIAL_ROBOT_TYPE;
  return SW_KB_UNKNOWN_ROBOT_TYPE;
}

static const char *
robotNumToName(SIMwareRobotType num)
{
  if (num == SW_KB_GROUND_VEHICLE_TYPE) return "GROUND_VEHICLE";
  if (num == SW_KB_LEGGED_ROBOT_TYPE) return "LEGGED_ROBOT";
  if (num == SW_KB_NAUTIC_ROBOT_TYPE) return "NAUTIC_ROBOT";
  if (num == SW_KB_AERIAL_ROBOT_TYPE) return "AERIAL_ROBOT";
  return "UNKNOWN";
}

static SIMwareSteerType
steerNameToNum(const char * name)
{
  if (! strcmp(name, "ACKERMAN")) return SW_KB_ACKERMAN_STEER_TYPE;
  if (! strcmp(name, "SKID")) return SW_KB_SKID_STEER_TYPE;
  if (! strcmp(name, "OMNI")) return SW_KB_OMNI_STEER_TYPE;
  if (! strcmp(name, "SUBMARINE")) return SW_KB_SUBMARINE_STEER_TYPE;
  return SW_KB_UNKNOWN_STEER_TYPE;
}

static const char *
steerNumToName(SIMwareSteerType num)
{
  if (num == SW_KB_ACKERMAN_STEER_TYPE) return "ACKERMAN";
  if (num == SW_KB_SKID_STEER_TYPE) return "SKID";
  if (num == SW_KB_OMNI_STEER_TYPE) return "OMNI";
  if (num == SW_KB_SUBMARINE_STEER_TYPE) return "SUBMARINE";
  return "UNKNOWN";
}

int
SimpleSimInfSkin::iniLoad(const char * filename)
{
  INIFILE inifile;
  const char * key;
  const char * section;
  const char * inistring;
  double d;
  int i;
  int retval;

  if (0 != inifile.open(filename)) {
    dprintf("can't open ini file %s\n", filename);
    return INI_INVALID;
  }

  section = "SIMPLE_SIM.INFSKIN";
  retval = INI_OK;

  // reads in a double, checks for positive
#define READ_DOUBLE(KEY,VAR)					\
  key = KEY;							\
  if (NULL != (inistring = inifile.find(key, section))) {	\
    if (1 == sscanf(inistring, "%lf", &d) &&			\
	d > 0.0) {						\
      /* found, and valid */					\
      simpleWorld.VAR = d;					\
      dprintf("using [%s] %s = %f\n",				\
	      section, key, (double) simpleWorld.VAR);		\
    } else {							\
      /* found, but invalid */					\
      dprintf("invalid value '%s' for [%s] %s in file %s\n",	\
	      inistring, section, key, filename);		\
      inifile.close();						\
      return INI_INVALID;					\
    }								\
  } else {							\
    /* not found, using default */				\
    dprintf("[%s] %s not found in file %s, using default %f\n",	\
	    section, key, filename, simpleWorld.VAR);		\
    retval = INI_DEFAULT;					\
  }

  READ_DOUBLE("CYCLE_TIME", delta_t);

  READ_DOUBLE("X_JUMP_SIZE", xJumpSize);
  READ_DOUBLE("Y_JUMP_SIZE", yJumpSize);
  READ_DOUBLE("Z_JUMP_SIZE", zJumpSize);
  READ_DOUBLE("JUMP_TIME", jumpTime);

  READ_DOUBLE("X_DRIFT_SPEED", xDriftSpeed);
  READ_DOUBLE("Y_DRIFT_SPEED", yDriftSpeed);
  READ_DOUBLE("Z_DRIFT_SPEED", zDriftSpeed);
  READ_DOUBLE("R_DRIFT_SPEED", rDriftSpeed);
  simpleWorld.rDriftSpeed *= TO_RAD;
  READ_DOUBLE("P_DRIFT_SPEED", pDriftSpeed);
  simpleWorld.pDriftSpeed *= TO_RAD;
  READ_DOUBLE("YAW_DRIFT_SPEED", yawDriftSpeed);
  simpleWorld.yawDriftSpeed *= TO_RAD;

  key = "OBSTACLES";
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%i", &i)) {
      /* found, and valid */
      simpleWorld.doObstacles = i ? true : false;
      dprintf("using [%s] %s = %s\n",
	      section, key, simpleWorld.doObstacles ? "true" : "false");
    } else {
      /* found, but invalid */
      dprintf("invalid value '%s' for [%s] %s in file %s\n",
	      inistring, section, key, filename);
      inifile.close();
      return INI_INVALID;
    }
  } else {
    /* not found, using default */
    dprintf("[%s] %s not found in file %s, using default %s\n",
	    section, key, filename, simpleWorld.doObstacles ?  "true" : "false");
    retval = INI_DEFAULT;
  }

  READ_DOUBLE("OBSTACLE_SPACING", obstacleSpacing);
  READ_DOUBLE("OBSTACLE_RADIUS", obstacleRadius);
  READ_DOUBLE("SENSOR_HEIGHT", sensorHeight);

  key = "SENSOR_PITCH";
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &d)) {
      /* found, and valid */
      simpleWorld.sensorPitch = d;
      dprintf("using [%s] %s = %f\n",
	      section, key, (double) simpleWorld.sensorPitch);
    } else {
      /* found, but invalid */
      dprintf("invalid value '%s' for [%s] %s in file %s\n",
	      inistring, section, key, filename);
      inifile.close();
      return INI_INVALID;
    }
  } else {
    /* not found, using default */
    dprintf("[%s] %s not found in file %s, using default %f\n",
	    section, key, filename, simpleWorld.sensorPitch);
    retval = INI_DEFAULT;
  }
  simpleWorld.sensorPitch *= TO_RAD;

  key = "ROBOT_TYPE";
  if (NULL != (inistring = inifile.find(key, section))) {
    simpleWorld.robotType = robotNameToNum(inistring);
    dprintf("using [%s] %s = %s\n",
	    section, key, robotNumToName(simpleWorld.robotType));
  } else {
    // not found, using default
    dprintf("[%s] %s not found in file %s, using default %d\n",
	    section, key, filename, simpleWorld.robotType);
    retval = INI_DEFAULT;
  }

  key = "STEER_TYPE";
  if (NULL != (inistring = inifile.find(key, section))) {
    simpleWorld.steerType = steerNameToNum(inistring);
    dprintf("using [%s] %s = %s\n",
	    section, key, steerNumToName(simpleWorld.steerType));
  } else {
    // not found, using default
    dprintf("[%s] %s not found in file %s, using default %d\n",
	    section, key, filename, simpleWorld.steerType);
    retval = INI_DEFAULT;
  }

  switch (simpleWorld.steerType) {
  case SW_KB_ACKERMAN_STEER_TYPE:
    READ_DOUBLE("WHEELBASE", ackerman.wheelbase);
    // check for wheelbase > 0 already done
    simpleWorld.ackerman.wheelbase_inv = 1.0 / simpleWorld.ackerman.wheelbase;
    READ_DOUBLE("AMAX", ackerman.aLinearMax);
    READ_DOUBLE("ALPHAMAX", ackerman.steerAlphaMax);
    break;
  case SW_KB_SKID_STEER_TYPE:
    READ_DOUBLE("WHEEL_RADIUS", skid.wheel_radius);
    READ_DOUBLE("WHEEL_SEPARATION", skid.wheel_separation);
    READ_DOUBLE("WMAX", skid.wmax);
    READ_DOUBLE("ALPHAMAX", skid.alphamax);
    break;
  case SW_KB_OMNI_STEER_TYPE:
    break;
  case SW_KB_SUBMARINE_STEER_TYPE:
    break;
  default:
    break;
  }

  inifile.close();

#undef READ_DOUBLE

  return retval;
}

// FIXME-- should have a return value
void
SimpleSimInfSkin::configInfSkin(SIMwareKB * k, SIMwareSupSkin * s) 
{
  extern char *SIMWARE_INIFILE_PATH;
  SIMwareKBGroundVehicle gv;
  int t;

  dprintf("configInfSkin called\n");

  kb = k;
  supSkin = s;
  state = SW_KB_STATE_UNINITIALIZED;

  // fills out our SimpleWorld structure
  iniLoad(SIMWARE_INIFILE_PATH);

#define LARGE 1.0e6

  // set up our robot
  switch (simpleWorld.robotType) {
  case SW_KB_GROUND_VEHICLE_TYPE:
    gv.setPlatformName(robotNumToName(simpleWorld.robotType));
    gv.setName(robotNumToName(simpleWorld.robotType));
    gv.setBotID(1);
    gv.setBotType(simpleWorld.robotType);
    gv.setSteerType(simpleWorld.steerType);
    gv.setDimensions(PM_CARTESIAN(1.0, 1.0, 1.0));
    gv.setCentroid(PM_CARTESIAN(0.5, 0.5, 0.5));
    gv.setMass(1.0);
    gv.setLightsOn(false);
    gv.setBatteryLevel(1.0);
    gv.setFuelLevel(1.0);
    gv.setLightLevel(1.0);
    gv.setMaxWheelRot(LARGE);
    gv.setMaxTorque(LARGE);
    gv.setMaxSteerAngle(LARGE);
    gv.setMaxCrabAngle(LARGE);
    gv.setWheelSeparation(simpleWorld.skid.wheel_separation);
    gv.setWheelBase(1.0);
    gv.setWheelRadius(simpleWorld.skid.wheel_radius);
    gv.setSteerAngle(0.0);
    gv.setCrabAngle(0.0);
    if (kb->bot != NULL) delete kb->bot;
    kb->bot = new SIMwareKBGroundVehicle;
    if (kb->bot == NULL) {
      dprintf("configInfSkin: error instantiating bot\n");
    } else {
      *((SIMwareKBGroundVehicle *) kb->bot) = gv;
      dprintf("configInfSkin: instantiated bot\n");
    }
    break;

  case SW_KB_LEGGED_ROBOT_TYPE:
  case SW_KB_NAUTIC_ROBOT_TYPE:
  case SW_KB_AERIAL_ROBOT_TYPE:
    break;

  default:
    break;
  }

  inu = (SIMwareKBSenINU *)
    kb->getSensor("INU", // my name
		  true, // insert this new one
		  SW_KB_INU_SENSOR_TYPE);
  if (NULL == inu) {
    dprintf("configInfSkin: error instantiating inu\n");
  } else {
    inu->setCycleTime(simpleWorld.delta_t);
  }

  gps = (SIMwareKBSenGPS *)
    kb->getSensor("GPS", // my name
		  true, // insert this new one
		  SW_KB_GPS_SENSOR_TYPE);
  if (NULL == gps) {
    dprintf("configInfSkin: error instantiating gps\n");
  } else {
    gps->setCycleTime(simpleWorld.delta_t);
  }

  grd = (SIMwareKBSenGrdTruth *)
    kb->getSensor("GrdTruth", // my name
		  true, // insert this new one
		  SW_KB_GRD_TRUTH_SENSOR_TYPE);
  if (NULL == grd) {
    dprintf("configInfSkin: error instantiating grd\n");
  } else {
    grd->setCycleTime(simpleWorld.delta_t);
  }

  linescan = (SIMwareKBSenLinescan *)
    kb->getSensor("Linescan", // my name
		  true, // insert this new one
		  SW_KB_LINESCAN_SENSOR_TYPE);
  if (NULL == linescan) {
    dprintf("configInfSkin: error instantiating linescan\n");
  } else {
    linescan->setCycleTime(simpleWorld.delta_t);
    linescan->setMaxRange(MAX_SENSOR_RANGE);
    linescan->setMinRange(0.0);
    linescan->setBeamSpacing(M_PI_2 / ((double) SIMPLE_SIM_MAX_SENSOR_VALUES));
    linescan->setFov(M_PI_2);
    // create the storage by appending to the vector
    for (t = 0; t < SIMPLE_SIM_MAX_SENSOR_VALUES; t++) {
      linescan->appendRangeReturn(0.0);
    }
  }

  return;
}

bool
SimpleSimInfSkin::openSkin(void) 
{ 
  int retval;
  pthread_attr_t attr;

  dprintf("openSkin called\n");

  // init the common thread attributes, used for all threads below
  retval = pthread_attr_init(&attr);
  if (retval != 0) {
    dprintf("openSkin error: %s\n", strerror(retval));
    return (int) false;
  }

  // start the world simulation thread
  retval = pthread_create(&worldThr, &attr, worldThrCodeStartRoutine, (void *) this);
  if (retval != 0) {
    dprintf("openSkin error: %s\n", strerror(retval));
    return (int) false;
  }

  // start the command handler thread
  retval = pthread_create(&cmdThr, &attr, cmdThrCodeStartRoutine, (void *) this);
  if (retval != 0) {
    dprintf("openSkin error: %s\n", strerror(retval));
    return (int) false;
  }

  return true;
}

void
SimpleSimInfSkin::closeSkin() 
{
  dprintf("closeSkin called\n");
  dprintf("world min/max: %f %f\n", worldThrTimeMin, worldThrTimeMax);
  dprintf("cmd min/max: %f %f\n", cmdThrTimeMin, cmdThrTimeMax);
}

bool
SimpleSimInfSkin::startBot(int vehid, string platformName, string platformUniqueID, PM_CARTESIAN xyz, PM_RPY rpy )
{
  return startBot( vehid, 0, 0 );
}

bool
SimpleSimInfSkin::startBot(int vehid, int platid, int startid) 
{ 
  dprintf("startBot(%d, %d, %d) called\n", vehid, platid, startid);

  /*
    FIXME-- this is where we get the ids from the SimWare base class,
    but we already called configInfSkin where this information is
    needed. Give these ids to configInfSkin earlier.
  */
  vehicleID = vehid;
  platformTypeID = platid;
  startPoseID = startid;

  kb->setKBState(SW_KB_STATE_INITIALIZED);

  return true;
}

/*
  Modification history:

  $Log: simpleSimInfSkin.cc,v $
  Revision 1.16  2007/09/04 21:06:10  dr_steveb
  Added robot name to spawned robot in usarSim. The name is a combination of an id and the robot type.

  Revision 1.15  2007/07/25 20:01:42  dr_steveb
  Made .ini file a command line option.

  Revision 1.14  2007/06/07 20:13:46  cj_scrapper
  Added the ability to control flippers in SIMware.  seServoMob takes MOAST flipskid command and translates it to a skid drive cmd and a flipper cmd. Also added parsing utilities for setting and configuration information for flippers and the aerial robot

  Revision 1.13  2007/05/03 16:34:47  dr_steveb
  Changes to allow for multiple robots to be spawned from a single controller (i.e. trafficSim). TrafficSim is not done...

  Revision 1.12  2007/04/06 19:11:11  proctor
  In servoShell.cc, changed printfs with %lf to %f, since there is no %lf
  in printf; in simpleSimInfSkin.cc, cast the bot type to ground vehicle
  so that the extra fields got copied out; robojoy.cc changes sign and uses
  offsets to match puma kinematics more closely

  Revision 1.11  2007/02/26 18:55:43  proctor
  Added performance measures for min,max times in moast sup skin, simplesim
  inf skin; changed the kill signal to INT for simplesim in the run script

  Revision 1.10  2007/02/26 16:06:29  proctor
  Added esleep to simpleSimWare command thread code when receiving null
  command, since popCmdFromQueue doesn't block

  Revision 1.9  2007/02/02 18:15:08  proctor
  Added Ackerman driving to simpleSimInfSkin

  Revision 1.8  2007/02/01 22:02:58  proctor
  Used SIMPLE_SIM.INFSKIN section in simpleSimInfSkin.cc and changed
  moast.ini accordingly;changed 'int' return types to 'bool' in
  swBaseSkin.hh and related moastSup,usarInfSkin.* files; added a debug
  flag to slamStub.cc

  Revision 1.7  2007/02/01 18:46:36  proctor
  Fixed null pointer in call to setBotType

  Revision 1.6  2007/02/01 18:20:24  proctor
  Added normal and exponential variates to random number generator; updated
  simpleSimWare to use ground vehicle bot.

  Revision 1.5  2007/01/31 18:51:53  cj_scrapper
  Fixed compilation errors in SimpleSim infskin. Needs attention

  Revision 1.4  2007/01/29 20:53:25  proctor
  Added linescan sensor to simple simware.
  Changed joySDL.cc to use even command numbers.

  Revision 1.3  2007/01/29 19:22:35  proctor
  Changed GPS sensor to use northing, easting fields; added .ini file handling
  of all SIMPLE_SIM variables

  Revision 1.2  2007/01/29 15:33:20  proctor
  Added simpleSimInfSkin support, including GPS sensor in the SimWare KB

  Revision 1.1  2007/01/26 22:18:10  proctor
  Initial checkin

*/

