/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simpleSimInfSkin.hh

  \brief Declarations of the simple simulation inferior skin
  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.9 $
  $Date: 2007/09/04 21:06:10 $
  \endcode
  \author Fred Proctor
*/

#ifndef SIMPLE_SIM_INF_SKIN_HH
#define SIMPLE_SIM_INF_SKIN_HH

#include <pthread.h>		// pthread_t,mutex,attr_t
#include <posemath.h>		// PM_CARTESIAN,RPY
#include "acmRandom.h"		// AcmRandomStruct
#include "swBaseSkin.hh"
#include "swKBTypes.hh"

/*!
  The world is a plane with equally spaced, infinitely tall
  cylindrical obstacles. The vehicle is an oriented point with state
  [x y theta] that is prevented from moving into these
  obstacles.

  The vehicle in the world is a skid-steer vehicle, but this
  may change and Ackerman and other vehicles could be added, with
  associated kinematics difference equations. 

  A line scanning sensor returns scans that reflect the obstacles and
  ground in view.

  Noisy sensors for GPS (jump) and INS (drift) are maintained.
*/

// how many linescan values we set
#define SIMPLE_SIM_MAX_SENSOR_VALUES 100

struct AckermanStruct {
  // config params
  double wheelbase;
  double wheelbase_inv;
  double vLinearMax;		// max linear speed
  double aLinearMax;		// max linear acceleration
  double steerAngleMax;		// max steering angle
  double steerAlphaMax;		// max steering angle accel
  // current values
  double vLinear;
  double steerAngle;
  // requested values
  double vLinearReq;
  double steerAngleReq;
};

struct SkidStruct {
  // config params
  double wheel_radius;
  double wheel_radius_inv;
  double wheel_separation;
  double wheel_separation_inv;
  double wmax;			// max angular speed
  double alphamax;		// max angular acceleration
  // current values
  double wLeft;
  double wRight;
  // requested values
  double wLeftReq;
  double wRightReq;
};

class SimpleWorld {
public:
  SimpleWorld(void);
  ~SimpleWorld(void);
  void update(void);

  bool lock(void);
  bool unlock(void);
  PM_CARTESIAN getGPS(void);
  PM_POSE getINS(void);
  PM_POSE getTruth(void);
  double getSensorValue(int);

  bool setRobotType(SIMwareRobotType type);

  bool setSteerType(SIMwareSteerType type);
  bool reqAckerman(double vLinear, double steerAngle);
  bool reqSkid(double wLeft, double wRight);
  bool reqOmni(double vLateral, double vLinear, double vRot);
  bool reqSubmarine(double propeller, double rudder, double sternPlane);

  void runAckerman(void);
  void runSkid(void);

  pthread_mutex_t mutex;

  double delta_t;		//< simulation time increment
  double delta_t_inv;		//< inverse, for faster calcs
  AcmRandomStruct rnd;		//< random number generator

  // vehicle starts at with all these at 0
  double xTrue;			//< X position of the vehicle
  double yTrue;			//< Y position
  double yawTrue;		//< Yaw
  double vTrue;			//< tangential speed 
  double wTrue;			//< angular speed

  // GPS jumping, only for X-Y-Z since GPS is 3DOF
  double xJump;			//< added to true X
  double yJump;			//< added to true Y
  double zJump;			//< added to 0, since true Z is 0
  double xJumpSize;		//< range to jump X
  double yJumpSize;		//< range to jump Y
  double zJumpSize;		//< range to jump Z
  double jumpTime;		//< time window for random jump
  double jumpClock;		//< a stopwatch for setting jump intervals

  // INS drifting, full 6DOF
  double xDrift;		//< added to true X
  double yDrift;		//< added to true Y
  double zDrift;		//< added to 0, like zJump
  double rDrift;		//< is roll, since true roll is 0
  double pDrift;		//< is roll, since true roll is 0
  double yawDrift;		//< added to true Yaw
  double xDriftSpeed;		//< speed to drift X
  double yDriftSpeed;		//< speed to drift Y
  double zDriftSpeed;		//< speed to drift Z
  double rDriftSpeed;		//< speed to drift Roll
  double pDriftSpeed;		//< speed to drift Pitch
  double yawDriftSpeed;		//< speed to drift Yaw

  //! non-zero means do obstacle calculations
  bool doObstacles;
  //! the center-to-center spacing of the column obstacles
  double obstacleSpacing;
  //! the radius of the column obstacles
  double obstacleRadius;
  //! the height of the sensor above its base, [m]
  double sensorHeight;
  //! pitch of sensor from horizontal, negative means downward, [rad]
  double sensorPitch;
  double sensorValues[SIMPLE_SIM_MAX_SENSOR_VALUES];

  // vehicle type, one of SW_KB_GROUND_VEHICLE,...,_ROBOT_TYPE
  SIMwareRobotType robotType;
  // FIXME-- add a union here as is done for steer type below

  // steer type, one of SW_KB_ACKERMAN,...,_STEER_TYPE
  SIMwareSteerType steerType;
  union {
    AckermanStruct ackerman;
    SkidStruct skid;
  };
};

enum {NUM_PERFS = 100000};

class SimpleSimInfSkin : public SIMwareInfSkin {
public:
  // required methods
  SimpleSimInfSkin(void);
  ~SimpleSimInfSkin(void);
  void configInfSkin(SIMwareKB *, SIMwareSupSkin *);
  bool openSkin(void);
  void closeSkin(void);
  bool startBot(int, int, int);
  bool startBot(int, string, string, PM_CARTESIAN, PM_RPY);

  // how to load values from ini files
  int iniLoad(const char * filename);

  // this thread blocks on a timer, then updates the simulated world
  void worldThrCode(void);
  // this thread blocks on commands
  void cmdThrCode(void);

  // this function handles just mobility commands
  void doMobCmd(SIMwareMobCmd * cmd);
  // this function handles just sensor commands
  void doSenCmd(SIMwareSenCmd * cmd);
#if 0
  // awaiting decls
  void doMisCmd(SIMwareMisCmd * cmd);
  void doEffCmd(SIMwareEffCmd * cmd);
  void doSetCmd(SIMwareSetCmd * cmd);
#endif

  // the start parameters
  int vehicleID;		// the vehicle number
  int platformTypeID;		// platform X for ini PLATFORM_TYPE_X
  int startPoseID;		// pose X for ini UTM_START_POSE_X

  // the world
  SimpleWorld simpleWorld;

  // sensor KB
  SIMwareKBSenINU * inu;
  SIMwareKBSenGPS * gps;
  SIMwareKBSenGrdTruth * grd;
  SIMwareKBSenLinescan * linescan;

  // thread structures
  pthread_t worldThr, cmdThr;

  // perf measures
  double worldThrTimeMin, worldThrTimeMax;
  double cmdThrTimeMin, cmdThrTimeMax;
};

// thread code, called with 'this' pointer

#endif
