//#define COLLECT_TIME
/*
DISCLAIMER:
This software was produced by the National Institute of Standards
and Technology (NIST), an agency of the U.S. government, and by statute is
not subject to copyright in the United States.  Recipients of this software
assume all responsibility associated with its operation, modification,
maintenance, and subsequent redistribution.

See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
\file slamStub.cc

\brief Stub for a process that uses SLAM to adjust navData (and
probably other things).

\code CVS Status:
$Author: tomrkramer $
$Revision: 1.23 $
$Date: 2009/05/22 21:25:12 $
\endcode

\author Chris Scrapper
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv(), atoi(), strtol()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <math.h>
#include <string>
#include <iostream>
#include <vector>
#include <sstream>
#include <gridfastslam/gridslamprocessor.h>
using namespace std;

#include <rcs.hh>
#include "swKBTypes.hh"
#include "sensorData.hh"
#include "navDataExt.hh"
#include "servoSP.hh"	
#include "gmapper.hh"
#include <time.h>
#include <sys/shm.h>

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"
#define PROCESS_NAME          "slamStub"
#define DEG(x) (57.295779513082323 * (x))
#define ABS(x)  (((x) < 0) ? -(x) : (x))
#define SLOPEPROTECTION

#define PI 3.14159265

static PM_POSE initialPose;
static PM_CARTESIAN oldPose;
static PM_RPY initialRPY;
ServoSPCmdReset spCmdReset;
int sp_serial_number=0;

char robotType;
bool isAirRobot = false;

class ServoNodeNmlBuffers
{
public:
	RCS_CMD_CHANNEL  *cmdBuf;
	RCS_STAT_CHANNEL *statBuf;
	RCS_CMD_CHANNEL  *cfgBuf;
	RCS_STAT_CHANNEL *setBuf;
	RCS_STAT_MSG_V2 * statMsg;
	RCS_STAT_MSG * setMsg;
	ServoNodeNmlBuffers();
	bool openNML(string nmlCfgFile, 
		string cmdName, 
		string statName, 
		string cfgName,
		string setName,
		NML_FORMAT_PTR fPtr,
		int botNum,
		int compID=0);
	void cleanUpNMLBuf();
};

//! INS VARIABLES
static const float INSSLOPE_RANGE = 2.0; //+ or - from the ground reading
static const int INSSLOPE_RESETCOUNT = 4; //reset the slope counter after a # of non slope readings
static const int INSSLOPE_GROUNDCHANGE = 15; //change the ground reading after # of similar readings

static float groundReading = 0.0f;
static float potentialGround = 0.0f;
static int sameReadingCount = 0;
static int slopeCounter = 0;

/// Scan Flag Pointer
unsigned char* scanFlagPtr = NULL;

//! OTHER VARIABLES
ServoNodeNmlBuffers sp;
NML *navDataExtBuf = NULL;
NML *servoSPNavBuf = NULL;
NML *sensorData1DBuf = NULL;
NavDataExt navDataExt;
SensorDataArray * servoSPNav;
bool tranTrueOnly = true;
int robotNum=1;
int openNML(string nmlFile, int botNum)
{
	robotNum = botNum;
	stringstream str; 

	if( ! sp.openNML(nmlFile,
		SERVO_SP_CMD_NAME,SERVO_SP_STAT_NAME,
		SERVO_SP_CFG_NAME,SERVO_SP_SET_NAME,  
		servoSP_format,botNum))
	{
		printf("** ERROR **  Error Initializing SP NML Buffers\n");  
		return false;
	}

	str << NAV_DATA_EXT_NAME <<  botNum;
	if(0)
		printf("Opening NML Buffer : %s\n", str.str().c_str());
	navDataExtBuf = new NML(navDataExt_format, 
		str.str().c_str(),
		PROCESS_NAME,
		nmlFile.c_str());
	if (!navDataExtBuf->valid()) 
	{
		delete navDataExtBuf;
		navDataExtBuf = 0;
		return false;
	}
	str.str("");


	str << SERVO_SP_NAV_BUF_NAME <<  botNum << "_1";;
	if(0)
		printf("Opening NML Buffer : %s\n", str.str().c_str());
	servoSPNavBuf = new NML(sensorData_format, 
		str.str().c_str(),
		PROCESS_NAME,
		nmlFile.c_str());

	if (!servoSPNavBuf->valid()) 
	{
		delete servoSPNavBuf;
		servoSPNavBuf = 0;
		return false;
	}
	servoSPNav=CLASS_CAST(SensorDataArray *, servoSPNavBuf->get_address());
	str.str("");

	if(!isAirRobot)
	{
		str << SERVO_SP_LINESCAN_BUF_NAME <<  botNum << "_1";
		if(0)
			printf("Opening NML Buffer : %s\n", str.str().c_str());
		sensorData1DBuf = new NML(sensorData_format, 
			str.str().c_str(),
			PROCESS_NAME,
			nmlFile.c_str());

		if (!sensorData1DBuf->valid()) 
		{
			delete sensorData1DBuf;
			sensorData1DBuf = 0;
			return false;
		}
		str.str("");
	}
	else printf("slamstub: Platform detected as AirRobot; skipping linescan init.\n");
	return true;
}

void closeNML()
{
	sp.cleanUpNMLBuf();
	if(navDataExtBuf!=NULL)
		delete navDataExtBuf;
	navDataExtBuf=NULL;
	if(servoSPNavBuf!=NULL)
		delete servoSPNavBuf;
	servoSPNavBuf=NULL;
}

void doNavView()
{
	printf("\033[2J");
	printf("=================================\n");
	printf("*****    NAV DATA EXT       *****\n");
	printf("=================================\n");
	printf("time (sec):         %0.2lf\n", navDataExt.time);
	printf("UTM:                %d %c\n", (int) navDataExt.utmZone, navDataExt.utmLetter);
	printf("tranAbs:            %f %f %f\n", (double) navDataExt.tranAbs.x,
		(double) navDataExt.tranAbs.y, (double) navDataExt.tranAbs.z);
	printf("rpyAbs:             %f %f %f\n", (double) DEG(navDataExt.rpyAbs.r),
		(double) DEG(navDataExt.rpyAbs.p), (double) DEG(navDataExt.rpyAbs.y));
	printf("tranRel:            %f %f %f\n", (double) navDataExt.tranRel.x,
		(double) navDataExt.tranRel.y, (double) navDataExt.tranRel.z);
	printf("rpyRel:             %f %f %f\n", (double) DEG(navDataExt.rpyRel.r),
		(double) DEG(navDataExt.rpyRel.p), (double) DEG(navDataExt.rpyRel.y));
	printf("tranTrue:           %f %f %f\n", (double) navDataExt.tranTrue.x,
		(double) navDataExt.tranTrue.y, (double) navDataExt.tranTrue.z);
	printf("rpyTrue:            %f %f %f\n", (double) DEG(navDataExt.rpyTrue.r),
		(double) DEG(navDataExt.rpyTrue.p), (double) DEG(navDataExt.rpyTrue.y));
	printf("tranAbsRate:        %f %f %f\n", (double) navDataExt.tranAbsRate.x,
		(double) navDataExt.tranAbsRate.y, (double) navDataExt.tranAbsRate.z);
	printf("tranRelRate:        %f %f %f\n", (double) navDataExt.tranRelRate.x,
		(double) navDataExt.tranRelRate.y, (double) navDataExt.tranRelRate.z);
	printf("rpyAbsRate:         %f %f %f\n", (double) navDataExt.rpyAbsRate.r,
		(double) navDataExt.rpyAbsRate.p, (double) navDataExt.rpyAbsRate.y);
	printf("rpyRelRate:         %f %f %f\n", (double) navDataExt.rpyRelRate.r,
		(double) navDataExt.rpyRelRate.p, (double) navDataExt.rpyRelRate.y);

	printf("speed:              %f\n", (double) navDataExt.speed);
}



bool computePoints(vector<double> &sData)
{
	double range;
	double theta;
	int step;
	int step1;
	int navReadType;
	int sensorReadType;
	double start;
	int inputLength;
	double inc;
	double timeDiff;
	double timeInc;
	static int featLength = 0; // length of featArray
	SensorData1D * input = (SensorData1D *)(sensorData1DBuf->get_address());
	static double lastTime = 0;

	sensorReadType = sensorData1DBuf->read(); // Get data from the level below us
	if (sensorReadType != SENSOR_DATA_1D_TYPE)
		return false;
	start = input->start;
	inputLength = input->featArray_length;
	if (inputLength > MAX_SENSOR_DATA_1D)
	{
		rcs_print_error("Input too large for SensorData1D\n");
		return false;
	}

	for (step=0; step < inputLength; step++)
	{
		sData.push_back(input->featArray[step]);
	}
	return !sData.empty();
}

void run(int debug, GMapper *SLAM)
{
	RCS_TIMER timer(0.02);
	int type;
	PM_CARTESIAN utmOffset;
	double delta_t=0;

	PM_CARTESIAN true_xyz;
	PM_RPY true_rpy;
	PM_CARTESIAN old_true_xyz;
	PM_RPY old_true_rpy;
	double last_true_time=0;

	PM_CARTESIAN odom_xyz;
	PM_RPY odom_rpy;
	PM_CARTESIAN old_odom_xyz;
	PM_RPY old_odom_rpy;
	double last_odom_time=0;

#ifdef COLLECT_TIME
	time_t startTime, nowTime;
	double cycleCount = 1;
	startTime = time(NULL);
#endif

	vector<double> sData; // Saves the range scanner values

	double x = initialPose.tran.x;
	double y = initialPose.tran.y;
	double theta = initialRPY.y;
	double xChange = 0.0;
	double yChange = 0.0;

	PM_CARTESIAN newOdom(0,0,0);
	PM_CARTESIAN oldOdom(0,0,0);

	PM_CARTESIAN odomCorrection(0,0,0);

	ofstream GroundRobotPath;
	char path[255];
	sprintf(path,"/home/webenji/Desktop/GroundRobotPath-%d.txt",robotNum);
	GroundRobotPath.open(path, ios::out);
	if(!GroundRobotPath)
	{
		printf("ERROR Creating GroundRobotPath.txt file!!\n\n");
	}
	else
	{
		printf("Created GroundRobotPath.txt file!!\n\n");
	}
	bool onSlope = false;
	while(1)
	{
		sp.statBuf->read();
		type=sp.setBuf->read();
		switch (type)
		{
		case 0:
			break;
		case -1:
			rcs_print_error("NML cmd read error\n");
			break;
		case SERVO_SP_SET_TYPE:
			for(int i=0;i<CLASS_CAST(ServoSPSet *, sp.setMsg)->sSetElem_length;i++)
			{
				switch(CLASS_CAST(ServoSPSet *, sp.setMsg)->sSetElem[i].sType)
				{
				case SERVO_SP_GRD_TRUTH_SENSOR:
					navDataExt.utmZone=(int)CLASS_CAST(ServoSPSet *, sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_ZONE];
					navDataExt.utmLetter=(char)CLASS_CAST(ServoSPSet *, sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_LETTER];
					utmOffset=PM_CARTESIAN(CLASS_CAST(ServoSPSet *, sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_NORTHING],
						CLASS_CAST(ServoSPSet *, sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_EASTING],
						CLASS_CAST(ServoSPSet *, sp.setMsg)->sSetElem[i].parameters[SERVO_SP_GRD_TRUTH_UTM_OFFSET_Z_VALUE]);
					break;
				case SERVO_SP_ODOMETRY_SENSOR:
					break;
				default:
					break;
				}
			}
			break;
		default:
			break;
		}

		type = servoSPNavBuf->read();

		switch (type)
		{
		case 0:
			break;
		case -1:
			rcs_print_error("NML cmd read error\n");
			break;
		case SENSOR_DATA_ARRAY_TYPE:
			delta_t=0;
#ifdef SLOPEPROTECTION
			for(int i=0;i<servoSPNav->arrayElem_length;i++)
			{
				/*
				switch(servoSPNav->arrayElem[i].type)
				{
				case SERVO_SP_INU_SENSOR:
				case SERVO_SP_INS_SENSOR:
				//printf("r(%f) p(%f) y(%f)\n",servoSPNav->arrayElem[i].dataTuple[3],servoSPNav->arrayElem[i].dataTuple[4],servoSPNav->arrayElem[i].dataTuple[5]);
				float INSReading = DEG(servoSPNav->arrayElem[i].dataTuple[4]);
				if(ABS(potentialGround - INSReading) < 0.1){
				if(sameReadingCount >= INSSLOPE_GROUNDCHANGE){
				groundReading = INSReading;
				potentialGround = groundReading;
				printf("*** Set the ground value to <<< %f >>>\n", groundReading);
				}else sameReadingCount++;
				}
				else{
				potentialGround = INSReading;
				sameReadingCount = 0;
				}
				float lowerBound = groundReading + INSSLOPE_RANGE;
				float upperBound = groundReading - INSSLOPE_RANGE;
				if(lowerBound < 0) lowerBound += 360;
				else if(lowerBound > 360) lowerBound -= 360;
				if(upperBound < 0) upperBound += 360;
				else if(upperBound > 360) upperBound -= 360;	
				if( INSReading > lowerBound && INSReading < upperBound){
				printf("SLOPE!!!!!!!!!!!----------------------------------------- %f\n",INSReading);
				slopeCounter = INSSLOPE_RESETCOUNT;
				}else if(slopeCounter >0) slopeCounter--;
				//t=servoSPNav->arrayElem[i].collectionTime;
				//x=servoSPNav->arrayElem[i].dataTuple[0];
				//y=servoSPNav->arrayElem[i].dataTuple[1];
				//theta=servoSPNav->arrayElem[i].dataTuple[2];
				break;
				}
				*/
			}
#endif
			for(int i=0;i<servoSPNav->arrayElem_length;i++)
			{
				switch(servoSPNav->arrayElem[i].type)
				{
					/*case SERVO_SP_GRD_TRUTH_SENSOR:

					// Don't use this timestamp/position to compute velocities
					// if we are too close to the old time or using a 
					// non timestamped sensor.
					if(servoSPNav->arrayElem[i].collectionTime > 0 &&
					servoSPNav->arrayElem[i].collectionTime-last_true_time > 0.0001)
					{
					delta_t=servoSPNav->arrayElem[i].collectionTime-last_true_time;
					navDataExt.time=servoSPNav->arrayElem[i].collectionTime;
					last_true_time = navDataExt.time;
					}
					else
					{
					old_true_xyz=true_xyz;
					old_true_rpy=true_rpy;
					delta_t = 0;
					}

					true_xyz=PM_CARTESIAN(servoSPNav->arrayElem[i].dataTuple[0],
					servoSPNav->arrayElem[i].dataTuple[1],
					servoSPNav->arrayElem[i].dataTuple[2]);
					true_rpy=PM_RPY(servoSPNav->arrayElem[i].dataTuple[3],
					servoSPNav->arrayElem[i].dataTuple[4],
					servoSPNav->arrayElem[i].dataTuple[5]);
					#ifdef COLLECT_TIME
					if( fabs(true_xyz.x - navDataExt.tranTrue.x)>.00001 ||
					fabs(true_xyz.y - navDataExt.tranTrue.y)>.00001 ||
					fabs(true_rpy.y - navDataExt.rpyTrue.y) >.00001 )
					cycleCount++;
					#endif
					// In simulation delta_t could be zero. 
					// UsarSim seems to be producing relative timestamps.
					if(delta_t > 0.0001 && last_true_time > 0)
					{
					navDataExt.tranTrueRate=(true_xyz - navDataExt.tranTrue)/delta_t;
					navDataExt.rpyTrueRate=(true_rpy - navDataExt.rpyTrue)/delta_t;
					}
					navDataExt.tranTrue=true_xyz;
					navDataExt.rpyTrue=true_rpy;

					if( tranTrueOnly )
					{
					navDataExt.tranAbs=utmOffset + true_xyz;
					navDataExt.rpyAbs=true_rpy;

					navDataExt.tranAbsRate=navDataExt.tranTrueRate;
					navDataExt.rpyAbsRate=navDataExt.rpyTrueRate;
					navDataExt.tranRel=navDataExt.tranAbs;
					navDataExt.rpyRel=navDataExt.rpyAbs;
					navDataExt.tranRelRate=navDataExt.tranTrueRate;
					navDataExt.rpyRelRate=navDataExt.rpyTrueRate;
					navDataExt.absToRel = PM_POSE(navDataExt.tranAbs,navDataExt.rpyAbs) * inv(PM_POSE(navDataExt.tranRel,navDataExt.rpyRel));

					navDataExt.speed=sqrt(navDataExt.tranRelRate.x*navDataExt.tranRelRate.x +
					navDataExt.tranRelRate.y*navDataExt.tranRelRate.y +
					navDataExt.tranRelRate.z*navDataExt.tranRelRate.z );

					navDataExt.steeringCurvature=0;
					navDataExt.steeringCrabAngle=0;
					}
					if(delta_t <= 0.0001)
					{
					true_xyz = old_true_xyz;
					true_rpy = old_true_rpy;
					}
					navDataExt.isSLAMon = true;
					break;*/
				case SERVO_SP_ODOMETRY_SENSOR:
					//t=servoSPNav->arrayElem[i].collectionTime;
					//x=servoSPNav->arrayElem[i].dataTuple[0];
					//y=servoSPNav->arrayElem[i].dataTuple[1];
					//theta=servoSPNav->arrayElem[i].dataTuple[2];

					//if (tranTrueOnly )
					//break;

					// Don't use this timestamp/position to compute velocities
					// if we are too close to the old time or using a 
					// non timestamped sensor.
					if(servoSPNav->arrayElem[i].collectionTime > 0 &&
						servoSPNav->arrayElem[i].collectionTime-last_odom_time > 0.0001)
					{
						delta_t=servoSPNav->arrayElem[i].collectionTime-last_odom_time;
						navDataExt.time=servoSPNav->arrayElem[i].collectionTime;
						last_odom_time = navDataExt.time;
					}
					else 
					{
						old_odom_xyz=odom_xyz;
						old_odom_rpy=odom_rpy;
						delta_t = 0.0;
					}

					newOdom.x = servoSPNav->arrayElem[i].dataTuple[0];
					newOdom.y = servoSPNav->arrayElem[i].dataTuple[1];
					newOdom.z = servoSPNav->arrayElem[i].dataTuple[2];

					//printf("ODOMETRY: %f,%f,%f\n\n", newOdom.x, newOdom.y, newOdom.z);

					if(*(scanFlagPtr) == 'N')
					{
						newOdom.x += odomCorrection.x;
						newOdom.y += odomCorrection.y;
						newOdom.z += odomCorrection.z;
					}

					xChange = cos(initialRPY.y)*(newOdom.x-oldOdom.x) - sin(initialRPY.y)*(newOdom.y-oldOdom.y);
					yChange = sin(initialRPY.y)*(newOdom.x-oldOdom.x) + cos(initialRPY.y)*(newOdom.y-oldOdom.y);
					x += xChange;
					y += yChange;

					theta += (newOdom.z-oldOdom.z);


					if(computePoints(sData) && *(scanFlagPtr) == 'Y')
					{
						SLAM->UpdateMap(x, y, theta, sData,0.0, 0.0, 0.0); // Run one iteration of SLAM 0.372

						SLAM->getPose(x,y,theta); // Get best pose from SLAM

						if(hypot(x-oldPose.x,y-oldPose.y) > 0.3)
						{
							x = oldPose.x + (newOdom.x - oldOdom.x);
							y = oldPose.y + (newOdom.y - oldOdom.y);
							theta = oldPose.z + (newOdom.z - oldOdom.z);
						}

						odomCorrection.x = x - (servoSPNav->arrayElem[i].dataTuple[0] + initialPose.tran.x);
						odomCorrection.y = y - (servoSPNav->arrayElem[i].dataTuple[1] + initialPose.tran.y);
						odomCorrection.z = theta - (servoSPNav->arrayElem[i].dataTuple[2] + initialRPY.y);

					}else if( slopeCounter > 0 ){
						
					}
					sData.clear();

					oldPose.x = x;
					oldPose.y = y;
					oldPose.z = theta; 
					//remember the last odometry reading
					oldOdom.x = newOdom.x;
					oldOdom.y = newOdom.y;
					oldOdom.z = newOdom.z; 

					odom_xyz=PM_CARTESIAN(x, y, 0);
					odom_rpy=PM_RPY(0, 0, theta);

#ifdef COLLECT_TIME
					if(fabs(odom_xyz.x - navDataExt.tranTrue.x)>.00001 ||
						fabs(odom_xyz.y - navDataExt.tranTrue.y)>.00001 ||
						fabs(odom_rpy.y - navDataExt.rpyTrue.y) >.00001 )
						cycleCount++;
#endif

					// In simulation delta_t could be zero. 
					// UsarSim seems to be producing relative timestamps.
					if(delta_t > 0.0001 && last_odom_time > 0.0)
					{
						navDataExt.tranTrueRate=(odom_xyz - navDataExt.tranTrue)/delta_t;
						navDataExt.rpyTrueRate=(odom_rpy - navDataExt.rpyTrue)/delta_t;
						navDataExt.tranRelRate=(odom_xyz - navDataExt.tranRel)/delta_t;
						navDataExt.rpyRelRate=(odom_rpy - navDataExt.rpyRel)/delta_t;
					}
					navDataExt.tranTrue=odom_xyz;
					navDataExt.rpyTrue=odom_rpy;
					navDataExt.tranRel=odom_xyz;
					navDataExt.rpyRel=odom_rpy;

					navDataExt.tranAbs=odom_xyz;
					navDataExt.rpyAbs=odom_rpy;

					navDataExt.tranAbsRate=navDataExt.tranRelRate;
					navDataExt.rpyAbsRate=navDataExt.rpyRelRate;

					navDataExt.absToRel = PM_POSE(navDataExt.tranAbs,navDataExt.rpyAbs) * inv(PM_POSE(navDataExt.tranRel,navDataExt.rpyRel));

					navDataExt.speed=sqrt(navDataExt.tranRelRate.x*navDataExt.tranRelRate.x +
						navDataExt.tranRelRate.y*navDataExt.tranRelRate.y +
						navDataExt.tranRelRate.z*navDataExt.tranRelRate.z );

					navDataExt.steeringCurvature=0;
					navDataExt.steeringCrabAngle=0;
					if(delta_t <= 0.0001)
					{
						odom_xyz = old_odom_xyz;
						odom_rpy = old_odom_rpy;
					}

					navDataExt.isSLAMon = true;
					break;
				default:
					break;
				}
				/* write navData Here for now */
				navDataExtBuf->write(navDataExt); 
			}


			if (debug & 1) {
				// print nav data
#if 0	 
				printf("%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf \n", 
					navDataExt.time, 
					delta_t, 
					navDataExt.tranRel.x, 
					navDataExt.tranRel.y,
					navDataExt.rpyRel.y,
					navDataExt.rpyRelRate.y,
					navDataExt.speed,
					t,
					x,
					y,
					theta);
				//#else
				printf("%lf %lf %lf %lf %lf %lf \n", 
					navDataExt.time, 
					navDataExt.tranRel.x, 
					navDataExt.tranRel.y,
					t,
					x,
					y);
#endif
			}

			if (debug == ~0) {
				// only if verbose
				doNavView();
			}

			break;
		default:
			break;
		}
		timer.wait();
#ifdef COLLECT_TIME
		if( fmod(cycleCount, 100) == 0 )
		{
			nowTime = time(NULL);
			printf( "slamStub at: %f hz\r", cycleCount/(nowTime-startTime) );
			fflush(stdout);
		}
#endif
	}
	GroundRobotPath.close();
}

PM_POSE getInitialPose()
{
	sp.statBuf->read();
	int type=sp.setBuf->read();
	switch (type)
	{
	case SERVO_SP_SET_TYPE:
		return ((ServoSPSet *)sp.setMsg)->startPose;
	}

	return PM_POSE(0,0,0,0,0,0,0);
}


int main(int argc, char *argv[])
{

	string nmlFile=DEFAULT_NML_FILE;
	string iniFile=DEFAULT_INI_FILE;
	int botNum=1;
	long int debug=1;

	char *configNML = getenv("CONFIG_NML");
	if(configNML)
	{
		nmlFile = configNML;
	}


	while (1) {
		char option = getopt(argc, argv, "p:i:g:vt");
		if (option == -1)
			break;

		switch (option) {
case 'i':
	botNum = atoi(optarg);
	break;
case 'g':
	// set some debug bits, see 'debug' lines below for bit meanings
	debug = strtol(optarg, NULL, 0);
	break;
case 'v':
	// verbose means set all debug bits
	debug = ~0;
	break;
case 't':
	tranTrueOnly = false;
	break;
case 'p':
	if(strcmp(optarg, "Talon") == 0)
		robotType = 'T';
	else if(strcmp(optarg, "Matilda") == 0)
		robotType = 'M';
	else if(strcmp(optarg, "P2AT") == 0)
		robotType = 'P';
	else if(strcmp(optarg, "AirRobot") == 0)
	{
		robotType == 'A';
		isAirRobot = true;
	}
	else
	{
		printf("slamStub: Unable to determine robot type. Unexpected behavior may result.\n");
		robotType = '\0';
	}
	break;
default:
	fprintf(stderr, "%s: usage: \n", argv[0]);
	fprintf(stderr, "           i <botNumber>\tthe id of the robot\n" );
	fprintf(stderr, "           g <debug bits>\t set individual debug bits\n" );
	fprintf(stderr, "           v            \t set all debug bits on\n");
	fprintf(stderr, "           t            \t do not set all navData values to ground truth\n" );
	return 1;
	break;
		}
	}

	/* Scan Flag shared mem -> Determine whether to update the map or not.*/
	// Info for the byte-sized shared memory used to turn off scanning.
	key_t scanKey;
	int scanShmID;

	// make the scanning key:
	char scanTok[33];
	sprintf(scanTok, "Robot%dScanFlag",botNum);
	if ((scanKey = ftok(scanTok, 'R')) == -1)
	{
		perror("ftok(scanTok)");
		exit(1);
	}

	// connect to and possibly create the scanFlag segment
	if ((scanShmID = shmget(scanKey, 1, 0644 | IPC_CREAT)) == -1)
	{
		perror("shmget(scanKey)");
		exit(1);
	}

	// attach to the scanFlag segment and get a pointer to it
	scanFlagPtr = (unsigned char*)shmat(scanShmID, (void*)0, 0);
	if (scanFlagPtr == (unsigned char*)(-1))
	{
		perror("shmat(scanShmID)");
		exit(1);
	}
	/*-----------Done----------------------*/

	if(! openNML(nmlFile,botNum) )
	{
		printf("** ERROR ** <slamStub::main> error while opening NML buffers\n");
		exit(-1);
	}

	// Set up GMaping SLAM
	GrisettiParameters gParams;
	gParams.numberOfBeams = 181; //was 181
	gParams.angResRad = M_PI/gParams.numberOfBeams;
	gParams.maxRangeM = 19.0; //was 19.0
	gParams.numberOfParticles = 20;//was 20
	gParams.minX = -100;
	gParams.maxX = 100;
	gParams.minY = -100;
	gParams.maxY = 100;
	gParams.delta = 0.1;
	gParams.skipBeams = 1;
	gParams.queued=false;
	gParams.minDist=500;
	gParams.maxQueueSize=5;
	gParams.tooBigHandler=0;
	gParams.handlerParameter=1.0;


	initialPose = getInitialPose();
	initialRPY = initialPose.rot;
	printf("\n\nStarting SLAM at: %lf,%lf,%lf\n", initialPose.tran.x, initialPose.tran.y, initialRPY.y);  

	OrientedPoint initPos = OrientedPoint(initialPose.tran.x, initialPose.tran.y, initialRPY.y);
	oldPose.x = initialPose.tran.x;
	oldPose.y = initialPose.tran.y;
	oldPose.z = initialRPY.y;
	GMapper* SLAM = new GMapper(gParams, initPos);
	//GMapper* SLAM = NULL;
	nml_start();
	run(debug, SLAM);
	closeNML();

	// Clean up shared mem.
	shmdt(scanFlagPtr);
	shmctl(scanShmID, IPC_RMID, NULL);

	return 0;
}

ServoNodeNmlBuffers::ServoNodeNmlBuffers() 
{
	cmdBuf=NULL;
	statBuf=NULL;
	cfgBuf=NULL;
	setBuf=NULL;
}
bool ServoNodeNmlBuffers::openNML(
	string nmlFile,
	string cmdName, 
	string statName, 
	string cfgName,
	string setName,
	NML_FORMAT_PTR fPtr,
	int botNum,
	int compID) {
	  stringstream str; 
	  char suffix[10];

	  // Check to see if there is a component id, this changes naming convention
	  if(compID==0)
		  sprintf(suffix,"%d",botNum);
	  else
		  sprintf(suffix,"%d_%d",botNum,compID);

	  str << cmdName << suffix;

	  if(0)
		  printf("Opening NML CMD Buffer  : %s\n", str.str().c_str());

	  cmdBuf = new RCS_CMD_CHANNEL(fPtr, 
		  str.str().c_str(), 
		  PROCESS_NAME, 
		  nmlFile.c_str());
	  if (!cmdBuf->valid()) 
	  {
		  delete cmdBuf;
		  cmdBuf = NULL;
		  return false;
	  }
	  str.str("");


	  str << cfgName << suffix;

	  if(0)
		  printf("Opening NML CFG Buffer  : %s\n", str.str().c_str());

	  cfgBuf = new RCS_CMD_CHANNEL(fPtr, 
		  str.str().c_str(), 
		  PROCESS_NAME,
		  nmlFile.c_str());
	  if (!cfgBuf->valid()) 
	  {
		  delete cfgBuf;
		  cfgBuf = NULL;
		  return false;
	  }
	  str.str("");

	  str << statName << suffix;

	  if(0)
		  printf("Opening NML STAT Buffer  : %s\n", str.str().c_str());

	  statBuf = new RCS_STAT_CHANNEL(fPtr, 
		  str.str().c_str(), 
		  PROCESS_NAME,
		  nmlFile.c_str());
	  if (!statBuf->valid()) 
	  {
		  delete statBuf;
		  statBuf = NULL;
		  return false;
	  }
	  statMsg= CLASS_CAST(RCS_STAT_MSG_V2 *, statBuf->get_address());
	  str.str("");

	  str << setName << suffix;

	  if(0)
		  printf("Opening NML SET Buffer  : %s\n", str.str().c_str());

	  setBuf = new RCS_STAT_CHANNEL(fPtr, 
		  str.str().c_str(), 
		  PROCESS_NAME,
		  nmlFile.c_str());
	  if (!setBuf->valid()) 
	  {
		  delete setBuf;
		  setBuf = NULL;
		  return false;
	  }
	  setMsg= setBuf->get_address();
	  str.str("");
	  return true;
}

void ServoNodeNmlBuffers::cleanUpNMLBuf()
{
	if (NULL != cmdBuf)
		delete cmdBuf;
	cmdBuf=NULL;

	if (NULL != statBuf)
		delete statBuf;
	statBuf=NULL;

	if (NULL != setBuf)
		delete setBuf;
	setBuf=NULL;

	if (NULL != cfgBuf)
		delete cfgBuf;
	cfgBuf=NULL;
} 

/*
Modification history:

$Log: slamStub.cc,v $
Revision 1.23  2009/05/22 21:25:12  tomrkramer
Corrected CVS data at top, added CVS Log at bottom


*/
