/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file 

  \brief 
  \code CVS Status:
  $Author: cj_scrapper $
  $Revision: 1.13 $
  $Date: 2007/12/11 15:03:42 $
  \endcode

  \author Chris Scrapper
*/

#ifndef SIMWARE_KB_BOT_HH
#define SIMWARE_KB_BOT_HH

//! C++ STL Headers
#include <string>
using namespace std;
//! RCS Headers
#include <posemath.h>
#include <inifile.h>
#include <math.h>
//! SIMware Headers
#include "swKBTypes.hh"

class SIMwareKBFlipper
{
protected:
  SIMwareFlipperType fType;
  double width;
  double length;
  double minAng;
  double maxAng;
  PM_CARTESIAN loc;
  PM_RPY rpy;
  double fAng;
public:
  SIMwareKBFlipper(SIMwareFlipperType);
  ~SIMwareKBFlipper();
  double getWidth();
  double getLength();
  double getMinAng();
  double getMaxAng();
  PM_CARTESIAN getLocation();
  PM_RPY getOrientation();
  double getFlipperAngle();
  SIMwareFlipperType getFlipperType();
  void setWidth(double);
  void setLength(double);
  void setMinAng(double);
  void setMaxAng(double);
  void setLocation(PM_CARTESIAN);
  void setLocation(double,double,double);
  void setOrientation(PM_RPY);
  void setOrientation(double,double,double);
  void setFlipperAngle(double);
};


class  SIMwareKBBot {
protected: 
  string pName;
  string bName;
  int bID;
  SIMwareRobotType rType;
  SIMwareSteerType sType;
  //! Center of gravity
  PM_CARTESIAN cg;
  //! Bounding box of robotic platform
  PM_CARTESIAN dim;
  //! Weight striped robotic platform
  double mass;  
  //! Headlights on/off
  bool lightsOn;
  //! Headlight Lumensiences
  double lLvl; /* light level */
  //! Battery level
  double bLvl;
  //! Fuel Level
  double fLevel;
  //! Timestamp
  double time;
public:
  //! Constructor
  SIMwareKBBot();
  //! Destructor
  ~SIMwareKBBot();
  //! Gets the robotic platform name
  string getPlatformName();
  //! Sets the robotic platform name
  void setPlatformName(string);
  //! Sets the robot's name 
  void setName(string);
  //! Returns robot's name 
  string getName();
  //! Sets the robot's id
  void setBotID(int);
  //! Returns robotic id
  int getBotID();
  //! Sets robot's type
  bool setBotType(SIMwareRobotType);
  //! Returns robot's type
  SIMwareRobotType getBotType();
  //! Sets the steering type of robotic platform
  void setSteerType(SIMwareSteerType);
  //! Gets steering typeof robotic platform
  SIMwareSteerType getSteerType();
  //! Sets robotic platforms center of gravity
  void setCentroid(PM_CARTESIAN);
  void setCentroid(double, double, double);
  //! Gets robotic platform center of gravity
  PM_CARTESIAN getCentroid();
  //! Sets robotic platforms dimensions
  void setDimensions(PM_CARTESIAN);
  void setDimensions(double, double, double);
  //! Gets robotic platforms dimensions
  PM_CARTESIAN getDimensions();
  //! Sets robotic platform mass
  void setMass(double);
  //! Gets robotic platform mass
  double getMass();
  //! Sets whether headlights are on
  void setLightsOn(bool);
  //! Gets if the headlights are on
  bool getLightsOn();
  //! Sets the level of battery power
  void setBatteryLevel(double);
  //! Gets the level of battery
  double getBatteryLevel();
  //! updates the fuel level
  void setFuelLevel(double);
  //! Gets the level of battery
  double getFuelLevel();
  void setLightLevel(double);
  double getLightLevel();
  void setTimeStamp(double);
  double getTimeStamp();
};

class SIMwareKBGroundVehicle : public  SIMwareKBBot 
{
protected:
  double maxWheelRot;
  double maxSteerAng;
  double maxCrabAng;
  double wRadius;
  double wSep;
  double wBase;
  double maxTorque;
  double steerAng;
  double crabAng;  
  SIMwareKBFlipper * flippers[SIMWARE_FLIPPER_MAX];
public:
  SIMwareKBGroundVehicle();
  ~SIMwareKBGroundVehicle();
  //! Set maximum angular velocity of axle rads/s
  void setMaxWheelRot(double);
  //! Get maximum angular velocity of axle rads/s
  double getMaxWheelRot();
  //! Sets maximum torque
  void setMaxTorque(double);
  //! Gets maximum torque
  double getMaxTorque();
  //! Sets the maximum steering angle of the front tire
  void setMaxSteerAngle(double);
  //! Gets the maximum steering angle of the front tire
  double getMaxSteerAngle();
  //! Updates the maximum steering angle of rear tires
  void setMaxCrabAngle(double);
  //! Retrieves the maximum steering angle of rear tires
  double getMaxCrabAngle();
  //! Update the wheel separation of right and left wheels
  void setWheelSeparation(double);
  //! RetrievesUpdate the wheel separation of right and left wheels
  double getWheelSeparation();
  //! Update the chasis length, length from the front wheels to back wheels
  void setWheelBase(double);
  //! Retrieve the chasis length, length from the front wheels to back wheels
  double getWheelBase();
  //! Updates the wheel radius 
  void setWheelRadius(double);
  //! Retrieves the wheel radius
  double getWheelRadius();
  double getMinTurnRadius();
  void setSteerAngle(double);
  void setCrabAngle(double);
  double getSteerAngle();
  double getCrabAngle();
  SIMwareKBFlipper * getFlipper(SIMwareFlipperType);
  bool addFlipper(SIMwareKBFlipper *);
};

class SIMwareKBNauticRobot : public  SIMwareKBBot 
{
protected:
  double sternPlaneAngle;
  double rudderAngle;
public:
  SIMwareKBNauticRobot();
  ~SIMwareKBNauticRobot();
  double getSternPlaneAngle();
  double getRudderAngle();
  void setSternPlaneAngle(double);
  void setRudderAngle(double);
};

class SIMwareKBAerialRobot : public  SIMwareKBBot 
{
protected:
  double maxRotVel;
  double maxLatVel;
  double maxLinVel;
  double maxVertVel;
  double vRot;
  double vLat;
  double vLin;
  double vVert;
public:
  SIMwareKBAerialRobot();
  ~SIMwareKBAerialRobot();
  double getMaxRotVel();
  double getMaxLatVel();
  double getMaxLinVel();
  double getMaxVertVel();
  void setMaxRotVel(double);
  void setMaxLatVel(double);
  void setMaxLinVel(double);
  void setMaxVertVel(double);
  double getRotVel();
  double getLatVel();
  double getLinVel();
  double getVertVel();
  void setRotVel(double);
  void setLatVel(double);
  void setLinVel(double);
  void setVertVel(double);
};
#endif
