/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file swKBMsg.hh

  \brief 
  \code CVS Status:
  $Author: cj_scrapper $
  $Revision: 1.13 $
  $Date: 2007/12/10 16:58:26 $
  \endcode

  \author Chris Scrapper
*/


#include <string>
#include <vector>
using namespace std;

#include <posemath.h>
#include "swKBTypes.hh"

////////////// SIMWARE COMMAND //////////////
class SIMwareCmdMsg 
{
private:
  static int count;
protected:
  SIMwareCmdType cmdType;
  int msgID;
public:
  SIMwareCmdMsg (SIMwareCmdType);
  virtual ~SIMwareCmdMsg() = 0;
  int getCmdType();
  int getCmdID();
};

////////////// SIMWARE MOBILITY COMMANDS //////////////
class SIMwareMobCmd : public SIMwareCmdMsg 
{
protected:
  SIMwareMobCmdType mobCmdType;
public:
  SIMwareMobCmd(SIMwareMobCmdType);
  SIMwareMobCmdType getMobCmdType();
};

typedef struct SIMWARE_FLIPPER_PAIR
{
  SIMwareFlipperType fType;
  double flipperAngle;
}SIMwareFlipperPair;

class SIMwareFlipperCmd : public SIMwareMobCmd 
{
public:
  SIMwareFlipperCmd();
  vector< SIMwareFlipperPair> fCmd;
};


class SIMwareAckermanCmd : public SIMwareMobCmd 
{
public:
  double vLinear;
  double steerAngle;
  double crabAngle;

  SIMwareAckermanCmd();
  SIMwareAckermanCmd(double,double,double);
};

class SIMwareSkidCmd : public SIMwareMobCmd 
{
public: 
  double wRight;
  double wLeft;
  
  SIMwareSkidCmd();
  SIMwareSkidCmd(double,double);
};

class SIMwareOmniCmd : public SIMwareMobCmd {
public:
  double vLateral;
  double vLinear;
  double vRot;

  SIMwareOmniCmd();
  SIMwareOmniCmd(double,double,double);
};


class SIMwareSubmarineCmd : public SIMwareMobCmd {
public:
  double propeller;
  double rudder;
  double sternPlane;
  
  SIMwareSubmarineCmd();
  SIMwareSubmarineCmd(double,double,double);
};

class SIMwareRotaryWingCmd : public SIMwareMobCmd {
public:
  double vLinear;     //linear velocity
  double vLateral;    //lateral velocity 
  double vVertical;   //Vertical Velocity
  double vRotation;   //Rotational Velocity
  SIMwareRotaryWingCmd();
  SIMwareRotaryWingCmd(double,double,double,double);
};

////////////// SIMWARE MISPKG   COMMANDS //////////////
class SIMwareMisLinkCmd
{
public:
  int linkID;
  double value;
  SIMwareMisLinkCmdType type;
  
  SIMwareMisLinkCmd();
  SIMwareMisLinkCmd(int,double, SIMwareMisLinkCmdType);  
};


class SIMwareMisCmd : public SIMwareCmdMsg 
{
protected:
  SIMwareMisCmdType misCmdType;
public:
  SIMwareMisCmd(SIMwareMisCmdType);
  SIMwareMisCmdType getMisCmdType();
};

class SIMwareMisMoveCmd : public SIMwareMisCmd {
public:
  string misPkgName;
  vector<SIMwareMisLinkCmd>linkCmd;
  SIMwareMisMoveCmd();
};


////////////// SIMWARE EFFECTER COMMANDS //////////////

class SIMwareEffCmd : public SIMwareCmdMsg 
{
protected:
  SIMwareEffCmdType effCmdType;
public:
  SIMwareEffCmd(SIMwareEffCmdType);
  SIMwareEffCmdType getEffCmdType();
};

class SIMwareEffOpcodeCmd : public SIMwareEffCmd {
public:
  string effName;
  SIMwareEffOpcodeType opcode;
  double param;
  SIMwareEffOpcodeCmd();
  SIMwareEffOpcodeCmd(string, SIMwareEffOpcodeType, double);
};

////////////// SIMWARE SENSOR COMMANDS //////////////
class SIMwareSenCmd : public SIMwareCmdMsg 
{
protected:
  SIMwareSenCmdType senCmdType;
public:
  SIMwareSenCmd(SIMwareSenCmdType);
  SIMwareSenCmdType getSenCmdType();
};

////////////// SIMWARE DATA MESSAGES //////////////
class SIMwareDataMsg 
{
protected:
  double time;
  
public:
  SIMwareDataMsg();
  double getMsgTime();
  void setMsgTime(double);
};

//! Sensor Messages
class SIMwareRfidMsg : public SIMwareDataMsg{
public:
  //! RFID tag id, defaults to -1
  int rfidID;
  //! Data read from rfid memory
  string rfidData;
  
  //! Constructor
  SIMwareRfidMsg();
  SIMwareRfidMsg(int, string);
};


class SIMwareVictimMsg  : public SIMwareDataMsg {
public:
  SIMwareVictimPartType partType;
  PM_CARTESIAN partLoc;
  

  //! Constructor
  SIMwareVictimMsg();
  SIMwareVictimMsg(SIMwareVictimPartType, PM_CARTESIAN);
};

