/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simpleSPSim.cc

  \brief Simple simulator for servo SP data.

  simpleSPSim is one of three threads that make up simpleSim. It simulates
  a 1D horizontal ladar with a 90-degree field of view and range returns
  every degree. The simulated world is flat, with cylindrical column
  obstacles spaced regularly. simpleSPSim fills in the 1D ladar data with
  simulated returns for the one or two of the four closest obstacles
  within the field of view. The range values are nominally the range to
  the closest edge of the obstacle. Infinite returns are set to max range.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.6 $
  $Date: 2006/12/12 19:20:40 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		// sscanf()
#include <stddef.h>		// NULL
#include <stdlib.h>
#include <math.h>		// sqrt(), asin(), atan2(), fmod()
#include <float.h>		// DBL_MAX
#include <pthread.h>		// pthread_exit()
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>		// RCS_STAT_CHANNEL, NML, ...
#include <posemath.h>		// PM_POSE, ...
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_EXT_BASE
#include "servoSP.hh"		// ServoSPSet
#include "sensorData.hh"	// SensorData1D
#include "simpleSim.hh"		// simpleSPSimInit,Step,Quit

static RCS_CMD_CHANNEL *servoSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *servoSPStatBuf = NULL;
static RCS_STAT_CHANNEL *servoSPSetBuf = NULL;
static RCS_CMD_CHANNEL *servoSPCfgBuf = NULL;
static NML *servoSPLinescanBuf = NULL;

static ServoSPStat servoSPStat;
static ServoSPSet servoSPSet;
static SensorData1D sensorData1D;

#define THIS_PROCESS "simware"	// NML process name for the server

#define DEFAULT_CYCLE_TIME 0.010
static double cycle_time = DEFAULT_CYCLE_TIME;

#define SQ(x) ((x)*(x))

//! how far away our sensor can see
static double maxRange = 1.0;

static int cleanupSensorNmlBuffers(void)
{
  if (NULL != servoSPCmdBuf)
    delete servoSPCmdBuf;
  if (NULL != servoSPStatBuf)
    delete servoSPStatBuf;
  if (NULL != servoSPSetBuf)
    delete servoSPSetBuf;
  if (NULL != servoSPCfgBuf)
    delete servoSPCfgBuf;
  if (NULL != servoSPLinescanBuf)
    delete servoSPLinescanBuf;

  servoSPStatBuf = NULL;
  servoSPSetBuf = NULL;
  servoSPCmdBuf = NULL;
  servoSPCfgBuf = NULL;
  servoSPLinescanBuf = NULL;

  return 0;
}

static int initSensorNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupSensorNmlBuffers();

  sprintf(chanName, "%s%d", SERVO_SP_CMD_NAME, bufnum);
  servoSPCmdBuf =
    new RCS_CMD_CHANNEL(servoSP_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoSPCmdBuf->valid()) {
    delete servoSPCmdBuf;
    servoSPCmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_SP_STAT_NAME, bufnum);
  servoSPStatBuf =
    new RCS_STAT_CHANNEL(servoSP_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoSPStatBuf->valid()) {
    delete servoSPStatBuf;
    servoSPStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_SP_SET_NAME, bufnum);
  servoSPSetBuf =
    new RCS_STAT_CHANNEL(servoSP_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoSPSetBuf->valid()) {
    delete servoSPSetBuf;
    servoSPSetBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SERVO_SP_CFG_NAME, bufnum);
  servoSPCfgBuf =
    new RCS_CMD_CHANNEL(servoSP_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoSPCfgBuf->valid()) {
    delete servoSPCfgBuf;
    servoSPCfgBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d_1", SERVO_SP_LINESCAN_BUF_NAME, bufnum);
  servoSPLinescanBuf =
    new NML(sensorData_format, chanName,
	    THIS_PROCESS, config_file);
  if (!servoSPLinescanBuf->valid()) {
    delete servoSPLinescanBuf;
    servoSPLinescanBuf = NULL;
    return 1;
  }

  return 0;
}

static void doCmdInit(ServoSPStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdHalt(ServoSPStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdAbort(ServoSPStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdShutdown(ServoSPStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    (void) simpleSPSimQuit();
    state_default(stat);
  }
}

/*
  FIXME: doCmdGo should be what generates and outputs the sensor data.
  Here it does nothing, since generation and output of sensor data
  happens all the time in the main loop.
 */
static void doCmdGo(ServoSPStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_EXEC);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCfgCycleTime(ServoSPCfgCycleTime * cfg,
			   ServoSPSet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    if (cfg->cycleTime <= 0.0) {
      status_next(set, RCS_ERROR);
    } else {
      set->cycleTime = cfg->cycleTime;
      status_next(set, RCS_DONE);
    }
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

static void doCfgDebug(ServoSPCfgDebug * cfg,
		  ServoSPSet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->debug = cfg->debug;
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

#define TWO_PI (M_PI + M_PI)
#define NORM(x) while ((x) > TWO_PI) (x) -= TWO_PI; while ((x) < 0.0) (x) += TWO_PI

enum {RANGE_OK = 0, RANGE_TOO_FAR, RANGE_INSIDE};
#define RANGE_INFINITE -1.0

/*
  angleWithin assumes that start, end and angle are all normalized 0..360. 
  "Within" means that moving clockwise from 'start' to 'end', you hit 'angle'.
 */
static int angleWithin(double start, double end, double angle)
{
  if (start < end) return (angle > start) && (angle < end);
  return (angle < end) || (angle > start);
}

//! calcHorizontalRange assumes SENSOR_ANGLE is 0, or horizontal viewing.

int calcHorizontalRange(double R, double xp, double yp, double th, double *range)
{
  double r, ang, alpha, ang1, ang2, rcosb;

  r = sqrt(xp*xp + yp*yp);
  if (r <= R) return RANGE_INSIDE;
  ang = atan2(-yp, -xp);
  alpha = asin(R/r);
  ang1 = ang - alpha;
  ang2 = ang + alpha;
  NORM(ang1);
  NORM(ang2);
  NORM(th);
  if (angleWithin(ang1, ang2, th)) {
    rcosb = r*cos(ang - th);
    *range = rcosb - sqrt(rcosb*rcosb + R*R - r*r);
    return RANGE_OK;
  }
  return RANGE_TOO_FAR;
}

/*!
  \c calcRange takes the cylindrical obstacle radius \b R, the
  X and Y position of the vehicle center \b xp and \b yp relative
  to the obstacle center, and the viewing angle \b th, and computes
  the range to the front edge of the obstacle. The view is pointed
  from the horizontal by \c SENSOR_ANGLE, so if the ground is closer
  than the obstacle, that range is computed.
  
  Returns \c RANGE_OK and sets \b *range if the obstacle or ground are in view.
  Returns \c RANGE_TOO_FAR if the obstacle is out of view.
  Returns \c RANGE_INSIDE if the vehicle is inside the obstacle.

  This figure shows how the computations are derived:
  \image html fig1.jpg
*/

int calcRange(double R,		//!< vehicle-to-obstacle center distance
	      double xp,	//!< x relative to obstacle center
	      double yp,	//!< y relative to obstacle center
	      double th,	//!< yaw angle you're looking
	      double height,	//!< height of sensor above ground
	      double pitch,	//!< pitch of sensor from horizontal
	      double *range)	//!< calculated range
{
  double sinpitch, cospitch;
  double secSensorAngle;
  double groundRange;
  double a2;
  double b2;
  double ab;
  double r, ang, d, alpha, ang1, ang2, t2, rr;

  // check for extreme cases: horizontal or vertical up, down
  sinpitch = sin(pitch), cospitch = cos(pitch);
  if (sinpitch == 0.0) {
    // horizontal
    return calcHorizontalRange(R, xp, yp, th, range);
  }
  if (cospitch == 0.0) {
    // vertical
    if (pitch < 0.0) {
      // straight down
      *range = height;
      return RANGE_OK;
    } else {
      return RANGE_TOO_FAR;
    }
  }

  secSensorAngle = 1.0 / cos(pitch);
  groundRange = pitch > 0.0 ? DBL_MAX : -height / sin(pitch);
  a2 = R * secSensorAngle;
  b2 = R;
  ab = SQ(R) * secSensorAngle;

  r = sqrt(xp*xp + yp*yp);	// veh center to obs center
  if (r <= R) return RANGE_INSIDE;
  ang = atan2(-yp, -xp);
  d = r * secSensorAngle;
  alpha = atan(sqrt(b2 / (SQ(d) - a2)));
  ang1 = ang - alpha;
  ang2 = ang + alpha;
  NORM(ang1);
  NORM(ang2);
  NORM(th);
  t2 = SQ(tan(ang - th));

  if (angleWithin(ang1, ang2, th)) {
    rr = (b2*d - ab*sqrt(a2*t2 - SQ(d)*t2 + b2))/(a2*t2 + b2) * secSensorAngle;
    // report closer of ground or obstacle
    if (rr < groundRange) {
      *range = rr;
    } else {
      *range = groundRange;
    }
    return RANGE_OK;
  }

  if (groundRange < maxRange) {
    *range = groundRange;
    return RANGE_OK;
  }

  return RANGE_TOO_FAR;
}

/*
  iniLoad() will load INI file params from 'filename', returning 0
  if everything is fine, 1 if the file can't be opened or there are
  errors in the values.

  Values loaded:

  [SERVO_SP]
  CYCLE_TIME = <positive number>
*/

#if ! defined(INI_OK)
#define INI_OK      0
#define INI_DEFAULT 1
#define INI_INVALID 2
#endif

static int iniLoad(const char * prog, const char * filename, ServoSPSet * set)
{
  INIFILE inifile;
  const char * inistring;
  PM_POSE IDENTITY_POSE = PM_POSE(PM_CARTESIAN(0,0,0),PM_QUATERNION(1,0,0,0));
  PM_CARTESIAN tran;
  PM_RPY rpy;
  double d;
  int retval = INI_OK;

  if (0 != inifile.open(filename)) {
    rcs_print_error("%s: can't open INI file %s, using defaults for servoSPSet\n", prog, filename);
    set->cycleTime = DEFAULT_CYCLE_TIME;
    return INI_INVALID;
  }

  if (NULL != (inistring = inifile.find("CYCLE_TIME", "SERVO_SP"))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->cycleTime = d;
    }
    else {
      // found, but invalid
      set->cycleTime = DEFAULT_CYCLE_TIME;
      rcs_print_error("%s: invalid value '%s' for [SERVO_SP] CYCLE_TIME in '%s'\n", prog, inistring, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [SERVO_SP] CYCLE_TIME in '%s'\n", prog, filename);
    set->cycleTime = DEFAULT_CYCLE_TIME;
    rcs_print_error("%s: using %f\n", prog, set->cycleTime);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  inifile.close();

  simpleWorldLock(&simpleWorld);
  simpleWorld.sensorHeight = tran.z;
  simpleWorld.sensorPitch = rpy.p;
  simpleWorldUnlock(&simpleWorld);

  return retval;
}

int simpleSPSimInit(int argc, char *argv[])
{
  int retval;
  ServoSPCmdInit servoSPCmdInit;
  ServoSPCfgCycleTime servoSPCfgCycleTime;

  // connect to the NML buffers
  retval = initSensorNmlBuffers(NML_FILE, bufnum);
  if (retval) {
    rcs_print_error("%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }
  
  // read in the INI file params into the settings struct
  if (INI_INVALID == iniLoad(argv[0], INI_FILE, &servoSPSet)) {
    rcs_print_error("%s: error loading INI file '%s'\n", argv[0], INI_FILE);
    return 1;
  }

  // just write it out once, since it doesn't change
  servoSPSetBuf->write(&servoSPSet);

  // clear out the sensor
  sensorData1D.isGlobal = 0;	// it's relative to sensor origin
  sensorData1D.start = -M_PI_4;
  sensorData1D.increment = M_PI_2 / 90.0;
  sensorData1D.featArray_length = 90;
  for (int t = 0; t < MAX_SENSOR_DATA_1D; t++) {
    sensorData1D.featArray[t] = 0.0;
  }

  simpleWorldLock(&simpleWorld);
  maxRange = simpleWorld.obstacleSpacing - simpleWorld.obstacleRadius;
  simpleWorldUnlock(&simpleWorld);

  sensorData1D.maxRange = maxRange;
  sensorData1D.minRange = 0.0;

  // initialize and report our status
  servoSPStat.echo_serial_number = 0;
  servoSPStat.heartbeat = 0;
  servoSPStatBuf->write(&servoSPStat);

  // initialize and report our settings
  servoSPSet.echo_serial_number = 0;
  servoSPSetBuf->write(&servoSPSet);

  // stuff an init cmd
  servoSPCmdInit.serial_number = 1;
  servoSPCmdBuf->write(&servoSPCmdInit);

  // stuff a cycle time config
  servoSPCfgCycleTime.serial_number = 1;
  servoSPCfgCycleTime.cycleTime = servoSPSet.cycleTime;
  servoSPCfgBuf->write(&servoSPCfgCycleTime);

  return 0;
}

static int done = 0;

int simpleSPSimQuit(void)
{
  done = 1;

  return 0;
}

void * simpleSPSimStep(void * arg)
{
  NMLTYPE cmd_type, cfg_type;
  int serial_number;
  double xmod, ymod;
  double xp, yp;
  double th_start, th_end, th, dth;
  double range;
  double obstacleSpacing;
  double obstacleRadius;
  double sensorHeight;
  double sensorPitch;
  double groundRange;
  int doObstacles;
  int index;

  dth = sensorData1D.increment;

  done = 0;
  while (! done) {
    // read cmd
    cmd_type = servoSPCmdBuf->read();
    serial_number = servoSPCmdBuf->get_address()->serial_number;
    switch (cmd_type) {

    case 0:
      // no new cmd
      break;

    case -1:
      // comm error
      rcs_print_error("NML cmd read error\n");
      break;

    case SERVO_SP_CMD_INIT_TYPE:
    case SERVO_SP_CMD_HALT_TYPE:
    case SERVO_SP_CMD_ABORT_TYPE:
    case SERVO_SP_CMD_SHUTDOWN_TYPE:
    case SERVO_SP_CMD_GO_TYPE:
      servoSPStat.command_type = cmd_type;
      if (serial_number != servoSPStat.echo_serial_number) {
	servoSPStat.echo_serial_number = serial_number;
	servoSPStat.state = NEW_COMMAND;
	servoSPStat.status = RCS_EXEC;
      }
      break;

    default:
      rcs_print_error("unknown cmd: %d\n", (int) cmd_type);
      break;
    }				// switch (cmd_type)

    // read cfg
    cfg_type = servoSPCfgBuf->read();
    serial_number = servoSPCfgBuf->get_address()->serial_number;
    switch (cfg_type) {
    case 0:
      // no new cmd
      break;

    case -1:
      // comm error
      rcs_print_error("NML cfg read error\n");
      break;

    case SERVO_SP_CFG_CYCLE_TIME_TYPE:
      servoSPSet.command_type = cfg_type;
      if( serial_number != servoSPSet.echo_serial_number) {
	servoSPSet.echo_serial_number = serial_number;
	servoSPSet.state = NEW_COMMAND;
	servoSPSet.status = RCS_EXEC;
      }
      break;

    case SERVO_SP_CFG_DEBUG_TYPE:
      servoSPSet.command_type = cfg_type;
      if( serial_number != servoSPSet.echo_serial_number) {
	servoSPSet.echo_serial_number = serial_number;
	servoSPSet.state = NEW_COMMAND;
	servoSPSet.status = RCS_EXEC;
      }
      break;

    default:
      break;
    }

    // handle cmd state tables
    switch (servoSPStat.command_type) {

    case SERVO_SP_CMD_INIT_TYPE:
      doCmdInit(&servoSPStat);
      break;

    case SERVO_SP_CMD_HALT_TYPE:
      doCmdHalt(&servoSPStat);
      break;

    case SERVO_SP_CMD_ABORT_TYPE:
      doCmdAbort(&servoSPStat);
      break;

    case SERVO_SP_CMD_SHUTDOWN_TYPE:
      doCmdShutdown(&servoSPStat);
      break;

    case SERVO_SP_CMD_GO_TYPE:
      doCmdGo(&servoSPStat);
      break;

    default:
      // no cmd to handle
      break;
    }				// switch (cmd)

    // handle config state tables
    switch (servoSPSet.command_type) {
    case SERVO_SP_CFG_CYCLE_TIME_TYPE:
      doCfgCycleTime((ServoSPCfgCycleTime *) servoSPCfgBuf->get_address(), &servoSPSet);
      break;

    case SERVO_SP_CFG_DEBUG_TYPE:
      doCfgDebug((ServoSPCfgDebug *) servoSPCfgBuf->get_address(), &servoSPSet);
      break;

    default:
      // no config to handle
      break;
    }

    // update sensor processing data

    simpleWorldLock(&simpleWorld);
    th_start = simpleWorld.yaw - M_PI_4;
    th_end = simpleWorld.yaw + M_PI_4;
    obstacleSpacing = simpleWorld.obstacleSpacing;
    obstacleRadius = simpleWorld.obstacleRadius;
    sensorHeight = simpleWorld.sensorHeight;
    sensorPitch = simpleWorld.sensorPitch;
    doObstacles = simpleWorld.doObstacles;
    simpleWorldUnlock(&simpleWorld);

    if (doObstacles) {
      xmod = fmod(simpleWorld.x, obstacleSpacing);
      if (xmod < 0.0) xmod += obstacleSpacing;
      ymod = fmod(simpleWorld.y, obstacleSpacing);
      if (ymod < 0.0) ymod += obstacleSpacing;
    } else {
      groundRange = sensorPitch > 0.0 ? DBL_MAX : -sensorHeight / sin(sensorPitch);
    }

    for (index = 0, th = th_start;
	 index < 90 && index < MAX_SENSOR_DATA_1D;
	 index++, th += dth) {
      sensorData1D.collectTimeStart = etime();
      if (doObstacles) {
#define DOIT if (RANGE_OK == \
		 calcRange(obstacleRadius, xp, yp, th, \
			   sensorHeight, sensorPitch, \
			   &range)) { \
	  sensorData1D.featArray[index] = range; \
	  continue; \
	}
	xp = xmod, yp = ymod;	// first obstacle at 0,0
	DOIT;
	xp = xmod-obstacleSpacing; // second obstacle at L,0
	DOIT;
	yp = ymod-obstacleSpacing; // third obstacle at L,L
	DOIT;
	xp = xmod;		// fourth obstacle at 0,L
	DOIT;
#undef DOIT
	//sensorData1D.featArray[index] = RANGE_INFINITE;
	sensorData1D.featArray[index] = maxRange; // TK changed 9/26/2005
      } else {
	//sensorData1D.featArray[index] = groundRange < maxRange ? groundRange : RANGE_INFINITE;
	sensorData1D.featArray[index] = groundRange < maxRange ? groundRange : maxRange;
	// TK changed 9/26/2005
      }
      sensorData1D.collectTimeDuration = etime() - sensorData1D.collectTimeStart;
    }

    // update stat and set

    servoSPStat.heartbeat++;
    servoSPStatBuf->write(&servoSPStat);
    servoSPSetBuf->write(&servoSPSet);
    sensorData1D.maxRange = maxRange;    //TK added 9/22/5
    servoSPLinescanBuf->write(&sensorData1D);

    esleep(cycle_time);
  }

  (void) cleanupSensorNmlBuffers();

  pthread_exit(0);

  return (void *) 0;
}

/*
  Modification history:

  $Log: simpleSPSim.cc,v $
  Revision 1.6  2006/12/12 19:20:40  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.5  2006/07/26 15:44:22  dr_steveb
  Changes to new buffer and process names

  Revision 1.4  2006/04/13 14:30:10  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.3  2006/04/07 13:12:12  proctor
  Took out getopt() from the threads, and made NML_FILE, INI_FILE and bufnum
  globals. Apparently getopt() doesn't work from within threads.

  Revision 1.2  2005/12/16 21:59:12  proctor
  simpleMobSim.cc: set navDataExt.time to etime().
  simpleSPSim.cc: ditto, plus filled in the sensorData1D.collectTimes.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.21  2005/10/03 18:41:33  proctor
  Changed range of view from twice the obstacle spacing to the obstacle spacing
  minus the obstacle radius, to prevent far-off obstacles from being incorrectly
  reported as absent.

  Revision 1.20  2005/09/30 19:21:35  proctor
  Fixed bug with range computations in negative quadrants.

  Revision 1.19  2005/09/26 21:30:10  kramer
  replaced infinite range with maxRange in two places

  Revision 1.18  2005/09/23 11:49:54  kramer
  added one line putting maxRange into sensorData1D

  Revision 1.17  2005/08/04 14:47:39  proctor
  Removed baseToVeh, changed misToVeh to mountToVeh

  Revision 1.16  2005/07/20 21:54:46  proctor
  simpleMis,Mob,SPSim.cc: added quitting upon shutdown.
  simpleSPSim.cc: also added cmd and cfg interface to make it a full RCS module.

  Revision 1.15  2005/07/19 18:56:58  proctor
  Added .ini file setting of obstacle params

  Revision 1.14  2005/07/18 21:16:52  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.13  2005/07/18 20:34:45  proctor
  Took out sincos.h

  Revision 1.12  2005/07/18 20:34:20  proctor
  Changed sincos() for sin(), cos()

  Revision 1.11  2005/07/18 20:27:18  proctor
  Added sincos.h

  Revision 1.10  2005/07/14 17:50:48  proctor
  simpleMisSim.cc: incremented heartbeat
  simpleSPSim.cc: handled horizontal, vertical pitch better

  Revision 1.9  2005/07/13 19:48:20  proctor
  Fixed handling of sensorHeight, sensorPitch globals

  Revision 1.8  2005/07/12 21:43:35  proctor
  Added tilted ladar simulation

  Revision 1.7  2005/06/30 19:44:20  proctor
  Changed SENSOR_DATA_1D 'lowerLeft' PM_CARTESIAN to a 'start' double.

  Revision 1.6  2005/06/29 21:48:57  proctor
  Set the 'isGlobal', 'lowerLeft' and 'increment' fields of SensorData1D.

  Revision 1.5  2005/06/29 18:04:41  proctor
  Added more accurate obstacle range.

  Revision 1.4  2005/06/29 14:32:05  proctor
  Took out annoying .nml, .ini file printing; added obstacle collisions
  in simpleMobSim.cc

  Revision 1.3  2005/06/28 20:49:16  proctor
  Added simulated 1D sensor data from column obstacles

  Revision 1.2  2005/06/23 22:03:39  proctor
  Changed 1D sensor data to the style declared in sensorData.hh

  Revision 1.1  2005/06/23 17:53:27  proctor
  Initial checkin
*/
