/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simware.h

  \brief Declarations of simware data structures exchanged between
  the superior and inferior skins.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.3 $
  $Date: 2009/04/24 19:37:14 $
  \endcode

  \author Fred Proctor
*/

#ifndef SIMWARE_H
#define SIMWARE_H

#include <string.h>

#define SW_NAME_MAX 80

#define SW_NONE 0

/*
  The resources, such as vehicles, machines and sensors.
*/  
enum {
  SW_EFFECTOR = 1,
  SW_SEN_ENCODER,
  SW_SEN_SONAR,
  SW_SEN_RANGESCANNER,
  SW_SEN_TOUCH,
  SW_SEN_CO2,
  SW_SEN_GROUNDTRUTH,
  SW_SEN_INS,
  SW_SEN_GPS,
  SW_SEN_ODOMETRY,
  SW_SEN_VICTIM,
  SW_MISPKG,
  SW_ROBOT_FIXED,
  SW_ROBOT_GROUNDVEHICLE,
  SW_ROBOT_AIRBOT
};
typedef int sw_type;

#define sw_type_to_string(TYPE)					\
  (TYPE) == SW_EFFECTOR ? "SW_EFFECTOR" :		\
  (TYPE) == SW_SEN_ENCODER ? "SW_SEN_ENCODER" :			\
  (TYPE) == SW_SEN_SONAR ? "SW_SEN_SONAR" :			\
  (TYPE) == SW_SEN_RANGESCANNER ? "SW_SEN_RANGESCANNER" :	\
  (TYPE) == SW_SEN_TOUCH ? "SW_SEN_TOUCH" :			\
  (TYPE) == SW_SEN_CO2 ? "SW_SEN_CO2" :				\
  (TYPE) == SW_SEN_GROUNDTRUTH ? "SW_SEN_GROUNDTRUTH" :		\
  (TYPE) == SW_SEN_INS ? "SW_SEN_INS" :				\
  (TYPE) == SW_SEN_GPS ? "SW_SEN_GPS" :				\
  (TYPE) == SW_SEN_ODOMETRY ? "SW_SEN_ODOMETRY" :		\
  (TYPE) == SW_SEN_VICTIM ? "SW_SEN_VICTIM" :			\
  (TYPE) == SW_MISPKG ? "SW_MISPKG" :				\
  (TYPE) == SW_ROBOT_FIXED ? "SW_ROBOT_FIXED" :	\
  (TYPE) == SW_ROBOT_GROUNDVEHICLE ? "SW_ROBOT_GROUNDVEHICLE" :	\
  (TYPE) == SW_ROBOT_AIRBOT ? "SW_ROBOT_AIRBOT" :		\
  "Unknown"

/*
  The operation to perform.
*/
typedef int sw_op;

typedef struct {
  double x;
  double y;
  double z;
  double roll;
  double pitch;
  double yaw;
} sw_pose;

/*
  Each resource has an associated structure that contains everything
  neeed to command it, configure it and report settings and status.
  The \a sw_op determines which fields are actually used.
*/

enum {
  SW_ROBOT_SKID_MOVE = 1,	/*!< move the vehicle the skid way */
  SW_ROBOT_ACKERMAN_MOVE,    /*!< move the vehicle the Ackerman way */
  SW_ROBOT_STAT,
  SW_ROBOT_SET
};

typedef enum {
  SW_STEER_UNKNOWN = 0,
  SW_STEER_SKID,
  SW_STEER_ACKERMAN,
  SW_STEER_OMNI
} sw_steer_type;

typedef struct {
  sw_pose mount;
} sw_robot_fixed_struct;

typedef struct {
  sw_steer_type steertype;
  double speed;
  double heading;
  double leftspeed;
  double rightspeed;
} sw_robot_groundvehicle_struct;

typedef struct {
  double speed;
  double heading;
} sw_robot_airbot_struct;

enum {
  SW_EFFECTOR_STAT = 1,
  SW_EFFECTOR_SET
};
typedef struct {
  char parent[SW_NAME_MAX];
  double value;			/*!< commanded value */
  sw_pose mount;
  double minvalue;		/*!< minimum value */
  double maxvalue;		/*!< maximum value */
} sw_effector_struct;

enum {
  SW_SEN_ENCODER_STAT = 1,
  SW_SEN_ENCODER_SET
};
typedef struct {
  /* status */
  int tick;
  /* settings */
  double resolution;
  sw_pose mount;
} sw_sen_encoder_struct;

enum {
  SW_SEN_SONAR_STAT = 1,
  SW_SEN_SONAR_SET
};
typedef struct {
  double range;
  double minrange;
  double maxrange;
  double beamangle;
  sw_pose mount;
} sw_sen_sonar_struct;

enum {
  SW_SEN_RANGESCANNER_STAT = 1,
  SW_SEN_RANGESCANNER_SET
};
#define SW_SEN_RANGESCANNER_MAX 192	/*!< how many ranges we can have  */
typedef struct {
  double range[SW_SEN_RANGESCANNER_MAX];
  double maxrange;
  double minrange;
  double resolution;
  double fov;
  sw_pose mount;
  int number;			/*!< how many there are  */
} sw_sen_rangescanner_struct;

enum {
  SW_SEN_TOUCH_STAT = 1,
  SW_SEN_TOUCH_SET
};
typedef struct {
  int touched;			/*!< non-zero means touching  */
} sw_sen_touch_struct;

enum {
  SW_SEN_CO2_STAT = 1,
  SW_SEN_CO2_SET
};
typedef struct {
  double density;		/*!< density of gas */
} sw_sen_co2_struct;

enum {
  SW_SEN_GROUNDTRUTH_STAT = 1,
  SW_SEN_GROUNDTRUTH_SET
};
typedef struct {
  /* status */
  sw_pose position;
  /* settings */
  sw_pose mount;
  double period;
} sw_sen_groundtruth_struct;

enum {
  SW_SEN_INS_STAT = 1,
  SW_SEN_INS_SET
};
typedef struct {
  sw_pose mount;
  sw_pose position;
  double period;
} sw_sen_ins_struct;

enum {
  SW_SEN_GPS_STAT = 1,
  SW_SEN_GPS_SET
};
typedef struct {
  double latitude;
  double longitude;
  int fix;
  int satellites;
  sw_pose mount;
  sw_pose position;
  double period;
} sw_sen_gps_struct;

enum {
  SW_SEN_ODOMETRY_STAT = 1,
  SW_SEN_ODOMETRY_SET
};
typedef struct {
  sw_pose mount;
  sw_pose position;
  double period;
  double resolution;
} sw_sen_odometry_struct;

enum {
  SW_SEN_VICTIM_STAT = 1,
  SW_SEN_VICTIM_SET
};
typedef struct {
  char parent[SW_NAME_MAX];
  sw_pose mount;
  double maxrange;
  double hfov;
  double vfov;
  int victims;
} sw_sen_victim_struct;

#define SW_MISPKG_LINK_MAX 16	/*!< how many links a mission package can have */

enum {
  SW_LINK_REVOLUTE = 1,
  SW_LINK_PRISMATIC,
  SW_LINK_SCISSOR
};
typedef int link_type;

#define sw_link_to_string(TYPE) ((TYPE) == SW_LINK_REVOLUTE ? "Revolute" : (TYPE) == SW_LINK_PRISMATIC ? "Prismatic" : (TYPE) == SW_LINK_SCISSOR ? "Scissor" : "Unknown")

typedef struct {
  int parent;			/*!< Index of parent, 0 for the base. */
  link_type type;
  sw_pose mount;
  double minrange;
  double maxrange;
  double maxspeed;
  double maxtorque;
  double position;		/*!< Linear or angular position. */
  double speed;
  double torque;
} sw_link_struct;

enum {
  SW_MISPKG_POSITION = 1,	/*!< Move the joints to a position. */
  SW_MISPKG_SPEED,		/*!< Move the joints at a speed. */
  SW_MISPKG_TORQUE,		/*!< Move the joints by force or torque. */
  SW_MISPKG_STAT,
  SW_MISPKG_SET
};
typedef struct {
  sw_link_struct link[SW_MISPKG_LINK_MAX];
  int number;			/*!< how many links */
} sw_mispkg_struct;

typedef struct {
  double time;
  sw_type type;			/*!< Which resource is selected. */
  sw_op op;			/*!< What operation to do on the resource.  */
  char name[SW_NAME_MAX]; /*!< Resources are identified by string names. */
  union {
    /* robots */
    sw_robot_groundvehicle_struct groundvehicle;
    sw_robot_airbot_struct airbot;
    /* machines */
    sw_effector_struct effector;
    /* sensors */
    sw_sen_encoder_struct encoder;
    sw_sen_sonar_struct sonar;
    sw_sen_rangescanner_struct rangescanner;
    sw_sen_touch_struct touch;
    sw_sen_co2_struct co2sensor;
    sw_sen_groundtruth_struct groundtruth;
    sw_sen_ins_struct ins;
    sw_sen_gps_struct gps;
    sw_sen_odometry_struct odometry;
    sw_sen_victim_struct victim;
    /* mission packages */
    sw_mispkg_struct mispkg;
  } data;
} sw_struct;

#define sw_set_time(PTR,TIME) (PTR)->time = (TIME)
#define sw_set_op(PTR,OP) (PTR)->op = (OP)
#define sw_set_type(PTR,TYPE) (PTR)->type = (TYPE)
#define sw_set_name(PTR,STR) strncpy((PTR)->name,(STR),SW_NAME_MAX); (PTR)->name[SW_NAME_MAX - 1] = 0
#define sw_clear_name(PTR) ((PTR)->name[0] = 0)
#define sw_no_name(PTR) (0 == (PTR)->name[0])
#define sw_match_name(PTR,STR) (! strcmp((PTR)->name,(STR)))

/* the sw_get_xxx macros need to resolve to compile-time constants,
   since we switch on them */
#define sw_get_time(PTR) (PTR)->time
#define sw_get_op(PTR) (PTR)->op
#define sw_get_type(PTR) (PTR)->type
#define sw_get_name(PTR) (PTR)->name

#define sw_init(PTR) sw_set_time(PTR,0); sw_set_op(PTR,SW_NONE); sw_set_type(PTR,SW_NONE); sw_clear_name(PTR)

#endif

/*
  Modification history:

  $Log: simware.h,v $
  Revision 1.3  2009/04/24 19:37:14  proctor
  Added handling of effectors

  Revision 1.2  2009/04/20 20:46:23  proctor
  Added src/simware to CVS; added support to scissors-type links; added
  named start locations

  Revision 1.1  2009/04/20 14:42:49  proctor
  Initial checking of new simware code in src/simware

*/

