/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*!
  \file exampleDijkTabs.cc
  \brief provides a driver for testing dijkTabs search

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Tom Kramer

See the documentation of dijkTabs.hh for an overview of this search routine.

This is a Dijkstra search routine driver for testing Dijstra search
with incremental graph building. The driver reads an entire graph from
a file picked by the user (the static graph), but drip-feeds the graph
to the Searcher, which builds its graph while searching.

The indexes of the nodes in the Searcher (since they are assigned on
the fly according to how the search goes) will differ from the indexes
of the nodes in the test file read by the driver.  To implement
insertKids and to report the path correctly, a map (driveMap) from the
indexes of driver's graph to the indexes of the Searcher's graph is
needed. A map the other way (searchMap) is also needed. The maps are
simply arrays of integers declared here as global variables. If the
index of a node in the driver's nodes is n, then the index of the
corresponding node in the Searcher's nodes is driveMap[n]. The maps
have 200000 entries, so graphs with more nodes than that should not be
used.

The routines here use stdio for reading and writing.

*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "dijkTabs.hh" 

#define DIJK_MAP_SIZE 200000
#define DIJK_PATH_SIZE 1000

DijkTabsGraph driverGraph;
DijkTabsGraph searcherGraph;
int driveMap[DIJK_MAP_SIZE];
int searchMap[DIJK_MAP_SIZE];

static void getData(int argc, char * argv[],
	     int * start, int * finish, DijkTabsGraph * aGraph);
static void initSearcherGraph(int start, int finish);
static void insertKids(int searcherDad, int startIndex);
int main(int argc, char * argv[]);
static void printAnswer(int length, int * path, double cost);
static void readGraph(char * fileName, DijkTabsGraph * aGraph);
static void suppress(char * costString);
static void writeGraph(char * fileName, DijkTabsGraph * aGraph);

/***********************************************************************/

//! getData reads the graph, the start index, and the finish index
/*
Returned Value: none

Called By:
  main

This reads the graph, the start index, and the finish index given in
the command arguments and stores them.

*/

void getData(            /* ARGUMENTS                          */
 int argc,               /*!< number of command line arguments */
 char * argv[],          /*!< array of command line arguments  */
 int * start,            /*!< index of start node, set here    */
 int * finish,           /*!< index of end node, set here      */
 DijkTabsGraph * aGraph) /*!< graph to build                   */
{
  if (argc == 4)
    {
      readGraph(argv[1], aGraph);
      if (sscanf(argv[2], "%d", start) != 1)
	{
	  fprintf(stderr, "cannot read start index\n");
	  exit(1);
	}
      if (sscanf(argv[3], "%d", finish) != 1)
	{
	  fprintf(stderr, "cannot read finish index\n");
	  exit(1);
	}
      if ((*start < 0) || (*start >= aGraph->size))
	{
	  fprintf(stderr, "start index out of range\n");
	  exit(1);
	}
      if ((*finish < 0) || (*finish >= aGraph->size))
	{
	  fprintf(stderr, "finish index out of range\n");
	  exit(1);
	}
      if (*start == *finish)
	{
	  fprintf(stderr, "start must differ from finish\n");
	  exit(1);
	}
    }
  else
    {
      fprintf(stderr, "Usage: %s <fileName> <start> <finish>\n", argv[0]);
      exit(1);
    }
}

/***********************************************************************/

//! initSearcherGraph puts 20 nodes in the searcherGraph and fills two
/*
Returned Value: none

Called By:
  main

This puts 20 nodes in the searcherGraph and fills two of them, the
first (index 0) is equivalent to the goal node from the driverGraph,
the second (index 1) is equivalent to the start node from the
driverGraph.

It also updates the driveMap and the searchMap.

This uses the global variables searcherGraph, driveMap, and searchMap.

*/

void initSearcherGraph( /* ARGUMENTS                                 */
 int start,             /*!< index in driver's graph of start node   */
 int finish)            /*!< index in driver's graph of finish node  */
{
  searcherGraph.nodes = new DijkTabsNode[20];
  searcherGraph.capacity = 20;
  searcherGraph.size = 2;
  searcherGraph.nodes[0].moreInfo = new NodeInfo(0);
  searcherGraph.nodes[0].kidsSize = 0;
  searcherGraph.nodes[1].kidsSize = 0;
  searcherGraph.nodes[1].moreInfo = new NodeInfo(1);
  searchMap[0] = finish;
  searchMap[1] = start;
  driveMap[finish] = 0;
  driveMap[start] = 1;
}

/***********************************************************************/

//! insertKids adds kids to the dad node and creates new kids
/*
Returned Value: none 

Called By: main
  
The startIndex argument is used to set the index attribute of the
nodeInfos and to update the driveMap and searchMap.

*/

static void insertKids( /* ARGUMENTS                                        */
 int searcherDad,  /*!< index in Searcher nodes of node to insert kids for  */
 int startIndex)   /*!< starting index in Searcher nodes for new added kids */
{
  DijkTabsNode * driverDad; // driver node from which to get data
  int numberAdded;          // total number of kids added
  int numberNew;            // number of new added kids
  int numberOld;            // number of old added kids
  NodeInfo ** newInfos;     // NodeInfos for new kids added
  double * newCosts;        // costs of new kids
  int * oldAdded;           // indexes in nodes of old kids added
  double * oldCosts;        // costs of old kids added
  int n;                    // index for kids of driverDad
  int j;                    // index for new kids of searcherDad node
  int k;                    // index for old kids of searcherDad node
  int driverN;              // index in driver nodes of a kid

  numberNew = 0;
  numberOld = 0;
  j = 0;
  k = 0;
  driverDad = (driverGraph.nodes + searchMap[searcherDad]);
  numberAdded = driverDad->kidsSize;
  for (n = 0; n < numberAdded; n++)
    {
      driverN = driverDad->kidsIndexes[n];
      if (driveMap[driverN] >= searcherGraph.size)
	fprintf(stderr, "bug in maintaining driveMap\n");
      else if (driveMap[driverN] >= 0)
	numberOld++;
      else
	numberNew++;
    }
  if (numberNew > 0)
    {
      newInfos = new NodeInfo * [numberNew];
      newCosts = new double[numberNew];
    }
  else
    {
      newInfos = 0;
      newCosts = 0;
    }
  if (numberOld > 0)
    {
      oldAdded = new int[numberOld];
      oldCosts = new double[numberOld];
    }
  else
    {
      oldAdded = 0;
      oldCosts = 0;
    }
  for (n = 0; n < numberAdded; n++)
    {
      driverN = driverDad->kidsIndexes[n];
      if (driveMap[driverN] >= 0)
	{
	  oldAdded[k] = driveMap[driverN];
	  oldCosts[k] = driverDad->kidsCosts[n];
	  k++;
	}
      else
	{
	  newInfos[j] = new NodeInfo(startIndex + j);
	  newCosts[j] = driverDad->kidsCosts[n];
	  driveMap[driverN] = (startIndex + j);
	  searchMap[startIndex + j] = driverN;
	  j++;
	}
    }
  searcherGraph.addKids(searcherDad, numberAdded, numberNew, newInfos,
			newCosts, oldAdded, oldCosts);
  if (numberNew > 0)
    {
      delete [] newInfos;
      delete [] newCosts;
    }
  if (numberOld > 0)
    {
      delete [] oldAdded;
      delete [] oldCosts;
    }
}

/***********************************************************************/

/* main

This:
1. initializes the maps between the driverGraph and the searcherGraph.
2. calls getData to read the graph named in the command arguments into
   the driverGraph and to read the indexes of the start and finish nodes.
3. calls initSearcherGraph to initialize the searcherGraph.
4. calls DijkTabsSearcher::initSearch to initialize the Searcher.
5. runs a loop in which:
   A. By calling getNextNode, the next node to be opened is identified
      (call it Noe).
   B. By calling insertKids, data about kids of Noe is inserted in Noe,
      and the kid nodes of Noe are built in the searcherGraph if they do
      not already exist there.
   C. CycleSearch is called.
   The loop runs until one of the following occurs:
   A. It is determined that there is no path to the goal.
   B. The cheapest path to the goal has been found.
6. prints the path, if one has been found.
7. calls DijkTabsSearcher::closeSearch to close the Searcher.

This uses all 4 global variables: driveMap, searchMap, driverGraph,
and searcherGraph.

*/

int main(       /* ARGUMENTS                                     */
 int argc,      /*!< one more than number of arguments           */
 char * argv[]) /*!< array of executable name and argument names */
{	
  int start;
  int finish;
  DijkTabsSearcher theSearcher;
  int opens;
  int status[1];
  int path[DIJK_PATH_SIZE];
  double cost[1];
  int length;
  int n;
  
  for (n = 0; n < DIJK_MAP_SIZE; n++)
    {
      driveMap[n] = -1;
      searchMap[n] = -1;
    }
  getData(argc, argv, &start, &finish, &driverGraph);
  initSearcherGraph(start, finish);
  theSearcher.initSearch(&searcherGraph, 1, 0, &opens, status, cost);
  while (opens && (status[0] != 2))
    {
      n = theSearcher.getNextNode();
      insertKids(n, searcherGraph.size);
      theSearcher.cycleSearch();
    }
  if (status[0] == 2)
    {
      length = theSearcher.reportSearch(DIJK_PATH_SIZE, path);
      printAnswer(length, path, cost[0]);
    }
  else
    fprintf(stderr, "No path found\n");
  theSearcher.closeSearch();
  return 0;
}

/***********************************************************************/

//! printAnswer prints a search path and its cost
/*
Returned Value: none

Called By:
  main

This prints a search path and its cost.

*/

void printAnswer( /* ARGUMENTS                                */
 int length,      /*!< length of the path                     */
 int * path,      /*!< array of index numbers giving the path */
 double cost)     /*!< cost of the path                       */
{
  int n;
  
  printf("Path from start to finish:");
  for (n = 0; n < length; n++)
    {
      printf(" %d", searchMap[path[n]]);
    }
  printf("\n");
  printf("With cost: %.4f\n\n", cost);
}

/***********************************************************************/

//! readGraph reads the file named fileName into aGraph
/*
Returned Value: none

Called by:
  getData

This reads the file named fileName into aGraph. The size and capacity
of the graph are set to the size given on the first line of the file.

The input is expected to be a file representing a directed graph.
The input does not have to be acyclic.
The expected format of an input file is:

1. one line with the number of nodes
2. one line for each node of the form:
      index  numberOfChildren  child  child  child ... 
   where a child is of the form:
      (childIndex, connectionCost)
   connectionCost is read as a double.
   If there are more lines of nodes than specified, the extras will
     be ignored.
   If a node line has more children than specified, an error will occur.
   A node should not be its own child.

EXAMPLE file

3
0       1       (1,12)
1       2       (2,19.2) (0,14.5)
2       2       (0,20) (1,15)

In the last fscanf format string, the space before the left parenthesis
is needed. This seems contrary to what C manuals say: leading white
space is skipped automatically. Perhaps leading white space is skipped
only before numbers.

*/

void readGraph(          /* ARGUMENTS                  */
 char * fileName,        /*!< name of file to read     */
 DijkTabsGraph * aGraph) /*!< graph to build from file */
{ 
  int i;
  int j;
  DijkTabsNode * nodes;
  int index;
  int kidSize;
  FILE * inport;

  inport = fopen(fileName, "r");
  if (inport == 0)
    {
      fprintf(stderr, "cannot open file for reading\n");
      exit(1);
    }
  if (fscanf(inport, "%d \n", &(aGraph->size)) != 1)
    {
      fprintf(stderr, "cannot read graph size\n");
      exit(1);
    }
  aGraph->capacity = aGraph->size;
  nodes = new DijkTabsNode[aGraph->size];
  aGraph->nodes = nodes;
  for (i = 0; i < aGraph->size; i++)
    {
      if (fscanf(inport, "%d", &index) != 1)
	{
	  fprintf(stderr, "cannot read index\n");
	  exit(1);
	}
      if (i != index)
	{
	  fprintf(stderr, "bad index\n");
	  exit(1);
	}
      nodes[i].moreInfo = new NodeInfo(i);
      if (fscanf(inport, "%d", &kidSize) != 1)
	{
	  fprintf(stderr, "cannot read kids size\n");
	  exit(1);
	}
      nodes[i].kidsSize = kidSize;
      nodes[i].kidsIndexes = new int[kidSize];
      nodes[i].kidsCosts = new double[kidSize];
      for (j = 0; j < kidSize; j++)
	{
	  if (fscanf(inport, " ( %d , %lf )", &(nodes[i].kidsIndexes[j]),
		     &(nodes[i].kidsCosts[j])) != 2)
	    {
	      fprintf(stderr, "cannot read kids data\n");
	      exit(1);
	    }
	}
      fscanf(inport, "\n");
    }
}

/***********************************************************************/

//! suppress removes trailing zeros for printing numbers
/*
Returned Value: none

Called By: writeGraph

This removes trailing zeros from a string with a decimal point in it.
If the number is an integer, the decimal point is also removed.
This is assuming the costString is not empty.

*/

static void suppress( /* ARGUMENTS                            */
 char * costString)   /*!< number string in which to suppress */
{
  int n;
  
  for (n = 0; costString[n] != 0; n++);
  for (n--; costString[n] == '0'; n--);
  if (costString[n] == '.')
    costString[n] = 0;
  else
    costString[n+1] = 0;
}

/***********************************************************************/

//! writeGraph prints a graph to a file
/*
Returned Value: none

Called By: Not normally called.

This is suppressing trailing zeros in costs, and suppressing
the decimal point for integer-valued costs. This is for two reasons:
1. Shortens the output (by 6 megabytes for graph64000_16).
2. Allows using diff to compare output files with input files
   with integer costs.

This writes graphs in the same format as the automatic graph generator.

*/

void writeGraph(         /* ARGUMENTS                  */
 char * fileName,        /*!< name of file to write    */
 DijkTabsGraph * aGraph) /*!< graph to write into file */
{
  DijkTabsNode * nodes;
  int i;
  int j;
  FILE * outport;
  char costString[40];

  nodes = aGraph->nodes;
  outport = fopen(fileName, "w");
  if (outport == 0)
    {
      fprintf(stderr, "cannot open file for writing\n");
      exit(1);
    }
  fprintf(outport, "%d\n", aGraph->size);
  for (i = 0; i < aGraph->size; i++)
    {
      fprintf(outport, "%d%c%d%c", i, '\t', nodes[i].kidsSize, '\t');
      for (j = 0; j < nodes[i].kidsSize; j++)
	{
	  sprintf(costString, "%.8f", nodes[i].kidsCosts[j]);
	  suppress(costString);
	  fprintf(outport, "(%d,%s) ", nodes[i].kidsIndexes[j], costString);
	}
      fprintf(outport, "\n");
    }
}

/***********************************************************************/

/*
  $Log: exampleDijkTabs.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.10  2005/07/18 15:07:28  kramer
  added check that start is not goal

  Revision 1.9  2005/07/15 21:32:31  kramer
  changed interface, improved documentation

  Revision 1.8  2005/06/29 21:08:14  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.7  2005/06/08 22:20:17  kramer
  added logging
*/
