/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file gpsToSerial.cc

  \brief outputs navDataExt message as standard serial gps message

  This program has the ability to read multiple navDataExt channels and encode them
  over a single serial port. When multiple channels are used, the vehid is encoded
  as the number of satellites. The options supported by the program include:
  -p <serial port i.e. "/dev/ttyS1">       opens this serial port for output
  -i <veh id> reads only this navDataExt buffer.
  -c <num> try to connect to navDataExt1 - navDataExtNum
  -n <nml config file>
  -t use true position instead of gps position

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.2 $
  $Date: 2006/02/08 18:31:13 $
  \endcode

  \author Stephen Balakirsky
*/
/*! \ingroup Tools 

*/
//
#include <rcs.hh>
#include "moastNmlOffsets.hh"
#include "navDataExt.hh"
#include "utmConvert.hh"
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <sys/select.h>
#include <time.h>    /* time functions for gmt */
#include "gpsToSerial.hh"



#define MY_CONFIG_FILE "../etc/moast.nml"
#define MY_NML_NAME "gpsToSerial"
#define MY_COM_PORT "/dev/ttyS1"

bool myVerbose;

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! Compute the check sum for the gps message
/*!
  \param stringIn The string to compute the check on.
  \return Returns the checksum value of the string
*/
unsigned int computeCheck( char *stringIn )
{
  int count, length;
  unsigned int retValue=0;

  length = (int)strlen(stringIn);
  for( count=1; count<length; count++ )
    {
      retValue ^= stringIn[count];
    }
  return retValue;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! Compute the string to send out over the serial port
/*!
  \param vehid     The id of the vehicle that this string is for
  \param outString The location to put the string
  \param channel   The nml channel to read the navDataExt buffer from
  \param useTruth  Should the truth data or gps data be sent out
  \return Returns true on success and false on failure
*/
bool fillString( int vehid, char *outString, NML *channel, bool useTruth )
{
  NavDataExt *navData;
  int type;
  char utmLetter[4];
  libUtmConvert::utmConvert converter;

  char latChar, lonChar;
  time_t ltime;
  struct tm *gmt;
  double lat, lon;
  double tmpValue;
  char utmZone[80];
  unsigned int check;

  type = channel->read();
  switch(type)
    {
    case 0:
      if( myVerbose )
	{
	  fprintf( stderr, "." );
	  fflush(stderr);
	}
      return false;
    case NAV_DATA_EXT_TYPE:
      navData = (NavDataExt*)channel->get_address();
      sprintf(utmLetter, "%d%c", navData->utmZone, navData->utmLetter);
      if( useTruth )
	converter.UTMtoLL(23, navData->tranTrue.x, navData->tranTrue.y, utmLetter,
			  lat, lon );
      else
	converter.UTMtoLL(23, navData->tranAbs.x, navData->tranAbs.y, utmLetter,
			  lat, lon );
      if( lat > 0 )
	latChar = 'N';
      else
	{
	  lat *= -1;
	  latChar = 'S';
	}
      if( lon > 0 )
	lonChar = 'E';
      else
	{
	  lon *= -1;
	  lonChar = 'W';
	}
      /* Display UTC. */
      time( &ltime );                 //Get time as long integer. 
      gmt = gmtime( &ltime );
      sprintf( outString, "$GPGGA," );
      if( gmt->tm_hour < 10 )
	sprintf( utmZone, "0%d", gmt->tm_hour );
      else
	sprintf( utmZone, "%d", gmt->tm_hour );
      strcat(outString, utmZone);
      if( gmt->tm_min < 10 )
	sprintf( utmZone, "0%d", gmt->tm_min );
      else
	sprintf( utmZone, "%d", gmt->tm_min );
      strcat(outString, utmZone);
      if( gmt->tm_sec < 10 )
	sprintf( utmZone, "0%d.00,", gmt->tm_sec );
      else
	sprintf( utmZone, "%d.00,", gmt->tm_sec );
      strcat(outString, utmZone);

      tmpValue = floor(lat);
      if( tmpValue < 10 )
	sprintf(utmZone, "0%d", (int)tmpValue);
      else
	sprintf(utmZone, "%d", (int)tmpValue);
      strcat(outString, utmZone);
      tmpValue = (lat - tmpValue) * 60.;
      if( tmpValue < 10 )
	sprintf(utmZone, "0%7.5lf,%c,", tmpValue, latChar);
      else
	sprintf(utmZone, "%7.5lf,%c,", tmpValue, latChar);
      strcat(outString, utmZone);

      tmpValue = floor(lon);
      if( tmpValue < 10 )
	sprintf(utmZone, "00%d", (int)tmpValue);
      else if( tmpValue < 100 )
	sprintf( utmZone, "0%d", (int)tmpValue);
      else
	sprintf( utmZone, "%d", (int)tmpValue);
      strcat(outString, utmZone);
      tmpValue = (lon - tmpValue) * 60.;
      if( tmpValue < 10 )
	sprintf(utmZone, "0%7.5lf,%c,", tmpValue, lonChar);
      else
	sprintf(utmZone, "%7.5lf,%c,", tmpValue, lonChar);
      strcat(outString, utmZone);

      sprintf( utmZone, "2,0%d,0.0,0.0,M,0.0,M,0.0,0001", vehid );
      strcat(outString, utmZone);
      check = computeCheck(outString);
      sprintf(utmZone, "*%x\r\n", check);
      strcat(outString, utmZone);

      //sprintf( outString, "Loc: %lf %lf\n", observedStatus->tranAbs.x,
      //		observedStatus->tranAbs.y );
      break;
    case -1:
      if( myVerbose )
	{
	  fprintf( stderr, "=" );
	  fflush( stderr );
	}
      return false;
    default:
      if( myVerbose )
	{
	  fprintf( stderr, "%d", type );
	  fflush(stderr);
	}
    }
  return true;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! Open a navdata buffer with for the given vehID
/*!
  If vehID is -1, then use no extension
  \param vehID The id of the vehicle whose navDataExt buffer should be opened
  \return Returns the channel
*/
NML *openNavDataExt( int vehID, char *myConfigFile, char *myNmlName )
{
  char chanName[80];
  NML *channel;
  sprintf( chanName, "%s%d", NAV_DATA_EXT_NAME, vehID );

  channel = new NML( navDataExt_format, chanName, myNmlName, myConfigFile );
  if( channel==NULL || !channel->valid() )
    {
      return(NULL);
    }
  return(channel);
}


// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! Main program loop
/*!
  \param argc The number of arguments received
  \param argv The actual arguments
*/
int main(int argc, char *argv[])
{
  char c;
  int nRetCode = 0;
  int count, numConnected;
  NML *navDataExt[5];
  //  CSerialComm port;
  char outString[255];
  RCS_TIMER timer(.250);
  char *serialPort = NULL;
  char *nmlConfig = NULL;
  int vehID = 1;
  int numConnect = 1;
  SerialConnect port;
  bool useTruth = false;
  
  myVerbose = false;

  while((c=getopt(argc,argv,"c:i:n:p:tv"))!=EOF)
    {
      switch(c)
	{
	case 'c':
	  numConnect = atoi(optarg);
	  fprintf( stderr, "%s connecting to %d channels.\n", argv[0], numConnect );
	  break;
	case 'i':
	  vehID = atoi(optarg);
	  fprintf( stderr, "%s connecting to vehicle id %d\n", argv[0], vehID );
	  break;
	case 'n':
	  nmlConfig = optarg;
	  fprintf( stderr, "%s using nml config file %s\n", argv[0], nmlConfig );
	  break;
	case 'p':
	  serialPort = optarg;
	  fprintf( stderr, "%s using serial port %s for output\n", argv[0], serialPort );
	  break;
	case 't':
	  useTruth = true;
	  fprintf( stderr, "%s using true position for reports\n", argv[0] );
	  break;
	case 'v':
	  myVerbose = true;
	  break;
	}
    }
  if( nmlConfig == NULL )
    {
      nmlConfig = new char[80];
      sprintf( nmlConfig, "%s", MY_CONFIG_FILE );
    }
  if( serialPort == NULL )
    {
      serialPort = new char[80];
      sprintf( serialPort, "%s", MY_COM_PORT );
    }

  if( !connectSerial( &port, serialPort ) )
    {
      fprintf( stderr, "%s error on opening of serial port %s\n",
	       argv[0], serialPort );
    }

  numConnected = 0;
  if( numConnect == 1 )
    {
      navDataExt[0] = openNavDataExt(vehID, nmlConfig, MY_NML_NAME);
      if( navDataExt[0] != NULL )
	numConnected++;
    }
  else
    {
      for( count=0; count<numConnect; count++ )
	{
	  navDataExt[numConnected] = openNavDataExt(count+1, nmlConfig, MY_NML_NAME);
	  if( navDataExt[numConnected] != NULL )
	    numConnected++;
	}
    }
  if( numConnected == 0 )
    return(-1);
  count=0;
  if( myVerbose )
    fprintf( stderr, "Going to main loop\n" );
  for(;;)
    {
      if( fillString(count+1, outString, navDataExt[count], useTruth) )
	write( port.fd, outString, strlen(outString) );
      if( myVerbose )
	fprintf( stderr, "%s\n", outString );
      timer.wait();
      count++;
      if( count>=numConnected )
	count=0;
    }
  return nRetCode;
}

int connectSerial(SerialConnect *s, char *portName) 
{
  struct termios options;

  s->connected=0; /* reset in case we fail to connect */

  s->fd = open(portName, O_RDWR | O_NOCTTY);
  if (s->fd == -1) return 0;
  else fcntl(s->fd, F_SETFL, 0);

  tcgetattr(s->fd, &options);

  /* go to 9600 baud */
  cfsetispeed(&options, B9600);
  cfsetospeed(&options, B9600);

  options.c_cflag |= (CLOCAL | CREAD); /* enable */

  options.c_cflag &= ~PARENB; /* 8N1 */
  options.c_cflag &= ~CSTOPB;
  options.c_cflag &= ~CSIZE;
  options.c_cflag |= CS8;

  /* set all of the options */
  tcsetattr(s->fd, TCSANOW, &options);

  s->connected=1;
  return 1;
}


/*
  Modification history:

  $Log: gpsToSerial.cc,v $
  Revision 1.2  2006/02/08 18:31:13  dr_steveb
  Added debug information

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.8  2005/09/30 18:55:56  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.7  2005/07/13 18:20:51  stephen
  Changed inclusion of getline

  Revision 1.6  2005/05/23 17:08:20  stephen
  Fixed name error in opennml

  Revision 1.5  2005/05/23 16:51:08  stephen
  Added command line input to exampleUtmonvert, changed buffer names on gpsToSerial

  Revision 1.4  2005/05/12 16:58:56  stephen
  Added 't' option to getopt

  Revision 1.3  2005/05/12 16:57:35  stephen
  Added ability to read true navigation solution.

  Revision 1.2  2005/05/12 12:22:25  stephen
  First working version

  Revision 1.1  2005/05/11 17:02:42  stephen
  First check-in of gpsToSerial tool.


*/
