/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   laserLog.cc
  \brief Log laser data for performance testing

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.4 $
  $Date: 2009/03/13 12:30:38 $
  \endcode

  \author Stephen Balakisky
  \date   12/16/2008
*/

#include <stddef.h>		// NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#include <stdio.h>              // fopen
#include <time.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>		// getopt_long()
#else
#include "getopt.h"
#endif

#include "rcs.hh"		// RCS_CMD,STAT_CHANNEL, rcs_print()
#include "sensorData.hh"

static RCS_CMD_CHANNEL * skel_cmd_buf= NULL;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_NML_BUFFER "servoSPLinescan"
#define THIS_PROCESS "nmlPrint" // NML process name 

bool done = false;

static void quit(int sig)
{
  done = true;
}

static void printHelp(void)
{
  printf( "options:\n-b <buffer name>\n-f <output file name>\n-p <process name>\n" );
  printf( "-c print entire laser array\n-n <buffer number>\n-i <mission number>\n");
}

#define BIG_SIZE 10000000
int main(int argc, char *argv[])
{
  char * envptr = 0;
  char nmlFile[MOAST_FILE_NAME_LEN] = {0};
  char bufferName[MOAST_NML_BUFFER_NAME_LEN] = {0};
  char processName[MOAST_NML_BUFFER_NAME_LEN] = {0};
  char bufNumString[MOAST_NML_BUFFER_NAME_LEN] = {0};
  int bufNum = 1;
  int retval;
  NMLTYPE cmdType;
  RCS_TIMER timer(.05);
  char c;
  NML * nmlBuf = 0;
  int mission_id = 1;
  FILE *fp = NULL;
  int count;
  bool centerOnly = true;
  char *fileName = "dataLog.dat";
  time_t currentTime;
  bool firstTime = true;

  // get input options
  while ((c = getopt(argc, argv, "b:cf:i:n:p:h")) != EOF) {
    switch (c) 
      {
      case 'b':
	strncpy(bufferName, optarg, MOAST_NML_BUFFER_NAME_LEN - 1);
	break;
      case 'c':
	centerOnly = false;
	break;
      case 'f':
	fileName = optarg;
	break;
      case 'i':
	mission_id = atoi(optarg);
	break;
      case 'p':
	strncpy(processName, optarg, MOAST_NML_BUFFER_NAME_LEN - 1);
	break;
      case 'n':
	if (1 != sscanf(optarg, "%i", &bufNum)) 
	  {
	    printf("%s: bad number parameter: ``%s''\n", argv[0],  optarg);
	    return 1;
	  }
      break;
      case 'h':
	printHelp();
	return 0;
	break;
      default:
	fprintf(stderr, "%s: bad argument `%s'\n", argv[0], optarg);
	return 1;
	break;
      }
  }

  // OK, we made it through the arguments

  if(0 == nmlFile[0]) 
    {
      if(NULL == (envptr = getenv("CONFIG_NML"))) 
	{
	  strncpy(nmlFile, DEFAULT_NML_FILE, MOAST_FILE_NAME_LEN);
	  nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
	} 
      else 
	{
	  strncpy(nmlFile, envptr, MOAST_FILE_NAME_LEN);
	  nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
	}
    }
  if( 0 == bufferName[0] )
    sprintf(bufferName, "%s", DEFAULT_NML_BUFFER );
  sprintf(bufNumString, "%d", bufNum);
  strcat(bufferName, bufNumString);
  if( mission_id != 0 )
    {
      sprintf(bufNumString, "_%d", mission_id );
      strcat(bufferName, bufNumString);
    }

  if( 0 == processName[0] )
    sprintf(processName, "%s", THIS_PROCESS );

  nmlBuf = new NML(sensorData_format, bufferName, processName, nmlFile);
  if (!nmlBuf->valid()) {
    delete nmlBuf;
    fprintf(stderr, "%s can't connect to NML buffer %s\n",
	    processName, bufferName);
    return 1;
  }

  signal(SIGINT, quit);

  // read a value from the buffer
  /****************************************************************** 
   * READ COMMAND - Command information is received, and the system
   *                is prepared to execute the command. When a new
   *		    command is received, the old command is immediately
   *		    replaced by this command. If the command is not 
   *		    able to be interrupted, error flags are set and
   *		    will be reported during the writng of status
   */

    // read command
  fp = fopen( fileName, "w" );
  if( fp <=0 )
    {
      printf( "Unable to open output file\n" );
      return -1;
    }
  currentTime = time(NULL);
  fprintf( fp, "# Date of datarun: %s", ctime(&currentTime));
  while( !done )
    {
      cmdType = nmlBuf->read();
      switch (cmdType) {
      case 0:
	// no new command
	break;
      case -1:
	// comm error
	rcs_print_error("%s: NML command read error\n", argv[0]);
	break;
      case SENSOR_DATA_1D_TYPE:
	  SensorData1D *sensorData1D = (SensorData1D *) nmlBuf->get_address();
	  if( firstTime )
	    {
	      firstTime = false;
	      fprintf( fp, "# Start: %lf Increment: %lf\n", 
		       sensorData1D->start, sensorData1D->increment );
	    }
	  if( centerOnly )
	    fprintf( fp, "%lf %4.2f\n", sensorData1D->collectTimeStart, 
		     sensorData1D->featArray[sensorData1D->featArray_length/2] );
	  else
	    {
	      fprintf( fp, "%lf ", sensorData1D->collectTimeStart );
	      for( count=0; count<sensorData1D->featArray_length; count++ )
		fprintf( fp, "%4.2f ", sensorData1D->featArray[count] );
	      fprintf( fp, "\n" );
	    }
	  break;
      }
      timer.wait();
    }
  fclose( fp );
  return 0;
}

/*
  Modification history:

  $Log: laserLog.cc,v $
  Revision 1.4  2009/03/13 12:30:38  proctor
  Fixed a char v. char* printf format bug

  Revision 1.3  2009/01/07 20:21:35  dr_steveb
  Added header information to file

  Revision 1.2  2008/12/18 21:59:37  dr_steveb
  Added ability to log all data, not just center beam

  Revision 1.1  2008/12/17 21:44:18  dr_steveb
  File to log laser data (center point only)

*/
