/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file linkMon.cc

  \brief Provide a measure of the link strength between a base station 
  and the robots

  In reality, to have connectivity between a robot and the base station,
  each robot could communicate directly or determine a route through other
  robots to get to the base station. The computation of this route may be
  very complex. Sice we are in simulation, we can cheat and compute these
  routes very simply. This program computes these routes and provides them
  over a status channel.

  This program only works with USARSim!

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.5 $
  $Date: 2007/09/20 20:14:57 $
  \endcode

  \author Stephen Balakirsky
*/
#include <unistd.h> // getopt
#include <stdlib.h> // getenv
#include <signal.h> // signal
#include <errno.h> // errno
//#include "linkMon.hh"

// socket stuff
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "rcs.hh"
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "linkMonCtrl.hh"
#include "trafficCtrl.hh"
#include "dijkHeapN.hh" 

static RCS_CMD_CHANNEL *linkMonCmdNml = NULL;
static RCS_STAT_CHANNEL *linkMonStatNml = NULL;
static RCS_STAT_CHANNEL *trafficCtrlSetNml = NULL;

static LinkMonStat linkMonStat;
static int done = 0;
static int server_fd = -1; // connection to unreal
static int *server_listen = NULL; // connection to message sent over comm server
static int *server_listenCon = NULL; // connection to message sent over comm server

static float radioCutOff = 60;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/traffic.ini"

#define THIS_PROCESS "linkMon"

static char *USARSimHostName;
static int USARSimPort = 5874;
static int USARSimListenPortBase = 7795;

// global variables and defines for searcher
#define DIJK_MAP_SIZE MAX_TRAFFIC_VEHICLES * MAX_TRAFFIC_VEHICLES
#define DIJK_PATH_SIZE MAX_TRAFFIC_VEHICLES + 1
#define MAX_GOALS MAX_TRAFFIC_VEHICLES
#define MAX_CYCLES 5000

//int searchMap[DIJK_MAP_SIZE];
DijkHeapNGraph searcherGraph;
int searcherGoals[MAX_TRAFFIC_VEHICLES];
DijkHeapNSearcher theSearcher;



/////////////////////////////////////////////////////////////////////////////////////////
//// socket code from fred
/////////////////////////////////////////////////////////////////////////////////////////
/* sockclient returns the file descriptor of the socket to the server */
static int sockclient(int port, char *hostname)
{
  int sockfd;			/* this process' socket fd */
  struct sockaddr_in server_addr, local_addr;
  struct hostent *ent;
  struct in_addr *in_a;

  //*! if (-1 == (sockfd = socket(AF_INET, SOCK_STREAM, 0)))
  if (-1 == (sockfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)))
   {
      fprintf(stderr, "can't create socket file descriptor\n");
      return -1;
    }

  if (NULL == (ent = gethostbyname(hostname)))
    {
      perror("gethostbyname");
      close(sockfd);
      sockfd = -1;
      return -1;
    }
  in_a = (struct in_addr *) ent->h_addr_list[0];

  /* fill in server's address */
  memset(&server_addr, 0, sizeof(struct sockaddr_in));
  //*! server_addr.sin_family = AF_INET;
  //server_addr.sin_family = PF_INET;
  server_addr.sin_family = ent->h_addrtype;
  server_addr.sin_addr.s_addr = in_a->s_addr;
  server_addr.sin_port = htons(port);

  /* bind any port number */
  //  localAddr.sin_family = AF_INET;
  local_addr.sin_family = PF_INET;
  local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  local_addr.sin_port = htons(0);

  int rc;
  rc = bind(sockfd, (struct sockaddr *) &local_addr, sizeof(local_addr));
  if(rc<0) {
    //printf("%s: cannot bind port TCP %u\n",argv[0],SERVER_PORT);
    perror("error ");
    return -1;
  }
  
  
  /* connect to the server */
  if (-1 == connect(sockfd, (struct sockaddr *) &server_addr,
		    sizeof(struct sockaddr_in)))
    {
      perror("can't connect to server");
      close(sockfd);
      sockfd = -1;
      return -1;
    }

  return sockfd;
}

/*
 * Accept a new TCP connection and write a message about who was
 * accepted to the system log.
 */

static int tcp_accept(int listen)
{
    struct sockaddr_storage ss;
    socklen_t ss_len = sizeof(ss);
    char host[NI_MAXHOST];
    char serv[NI_MAXSERV];
    int n, fd;

    fd = accept(listen, (struct sockaddr *) &ss, &ss_len);
    if (fd == -1) 
      {
        printf("accept failed: %s\n", strerror(errno));
        return -1;
      }

    n = getnameinfo((struct sockaddr *) &ss, ss_len,
                    host, sizeof(host), serv, sizeof(serv),
                    NI_NUMERICHOST);
    if (n) 
      {
        printf("getnameinfo failed: %s\n", gai_strerror(n));
      } 
    else 
      {
        printf("connection from %s:%s\n", host, serv);
      }
    return fd;
}

static int tcp_listen(char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, on = 1;
    
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(NULL, port, &hints, &ai_list);
    if (n) 
      {
        fprintf(stderr, "%s: getaddrinfo failed: %s\n",
                THIS_PROCESS, gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) 
      {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) 
	  {
	    continue;
	  }

        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
        if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0) 
	  {
            break;
	  }
        close(fd);
      }
    
    freeaddrinfo(ai_list);

    if (ai == NULL) 
      {
        fprintf(stderr, "%s: bind failed for port %s\n",
                THIS_PROCESS, port);
        return -1;
      }

    if (listen(fd, 42) < 0) 
      {
        fprintf(stderr, "%s: listen failed: %s\n",
                THIS_PROCESS, strerror(errno));
	close(fd);
        return -1;
      }
    
    return fd;
}

/////////////////////////////////////////////////////////////////////////////////////////
//// graph search routines from Tom Kramer and me
/////////////////////////////////////////////////////////////////////////////////////////
bool  computeCosts(int node)
{
  DijkHeapNNode * nodes;
  int i, nn;
  char msg[128];
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();
  char *msgPt;
  float cost;

  nodes = searcherGraph.nodes;
  for( i=0; i<nodes[node].kidsSize; i++ )
    {
      if( nodes[node].kidsCosts[i] == -2 ) // need to get information
	{
	  sprintf( msg, "getss %s %s;", set->robotInfo[node].robotName,
		   set->robotInfo[nodes[node].kidsIndexes[i]].robotName );
	  //	  printf("%s: sending %d length message: %s\n", THIS_PROCESS, strlen(msg), msg);
	  nn = send( server_fd, msg, strlen(msg)+1, 0);
	  //      send( server_fd, "\r\n", 2, 0);
	  if (nn != (int)strlen(msg)+1) 
	    {
	      fprintf(stderr, "Write Failed\n");
	      return false;
	    }
	  //Wait for response
	  nn = recv(server_fd, msg, sizeof(msg)-1, 0);
	  if( nn == -1 )
	    {
	      return false;
	    }
	  msg[nn] = '\0';
	  msgPt = &msg[4];
	  cost = atof(msgPt);
	  if( cost > radioCutOff )
	    cost = -1;
	  nodes[node].kidsCosts[i] = cost;
	  nodes[i].kidsCosts[node] = cost;
	  //	  printf("%s: message received:%s cost:%f\n", THIS_PROCESS, msg,
	  //		 nodes[node].kidsCosts[i]);
	  
	}
    }
  return true;
}

void initSearcherGraph()
{
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();
  int n;
  int numGoals = set->robotInfo_length;

  searcherGraph.nodes = new DijkHeapNNode[numGoals + 1];
  searcherGraph.capacity = (numGoals + 1);
  searcherGraph.size = (numGoals + 1);
  for (n = 0; n < numGoals; n++)
    {
      searcherGraph.nodes[n].moreInfo = new NodeInfo(n);
      searcherGraph.nodes[n].kidsSize = 0;
      //      searchMap[n] = searcherGoals[n];
      //      driveMap[driverGoals[n]] = n;
    }
  //  searchMap[numGoals] = start;
  //  driveMap[start] = numGoals;
}


void setGraph()
{
  DijkHeapNNode * nodes;
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();
  int i,j;

  searcherGraph.size = set->robotInfo_length;
  searcherGraph.capacity = searcherGraph.size;
  nodes = new DijkHeapNNode[searcherGraph.size];
  searcherGraph.nodes = nodes;

  for( i=0; i<searcherGraph.size; i++ )
    {
      nodes[i].moreInfo = new NodeInfo(i);
      nodes[i].kidsSize = set->robotInfo_length;
      nodes[i].kidsIndexes = new int[nodes[i].kidsSize];
      nodes[i].kidsCosts = new double[nodes[i].kidsSize];
      for( j=0; j<searcherGraph.size; j++ )
	{
	  nodes[i].kidsIndexes[j] = j;
	  if( i == j )
	    nodes[i].kidsCosts[j] = -1; // negative costs invalidates arc
	  else
	    nodes[i].kidsCosts[j] = -2; // my flag that we need a cost
	}
    }
  for( i=1; i<searcherGraph.size; i++ ) // set the goals
    searcherGoals[i-1] = i;
}

static void printAnswer( /* ARGUMENTS                                */
 int length,             /*!< length of the path                     */
 int * path,             /*!< array of index numbers giving the path */
 double cost,            /*!< cost of the path                       */
 LinkMonStat * stat)     /*!< status structure                       */
{
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();
  int n;
  
  for (n = 0; n < length; n++)
    {
      //  printf(" %s", set->robotInfo[path[n]].robotName);
      strcpy(stat->route[stat->route_length].name[n], set->robotInfo[path[n]].robotName);
    }
  stat->route[stat->route_length].length = length;
  stat->route[stat->route_length].cost = cost;
  stat->route_length++;
  /*
  printf("\n");
  printf("With cost: %.4f\n\n", cost);
  */
}

/////////////////////////////////////////////////////////////////////////////////////////
//// wss code from Chris and I
/////////////////////////////////////////////////////////////////////////////////////////
bool wssWrite( char *robotName, char *msg )
{
  int nn;
  char mess[128];

  printf( "Need to send message %s to robot %s\n", msg, robotName );
  sprintf( mess, "SEND %d %s;", strlen(msg)+2, msg );
  printf( "Sending message: %s\n", mess );
  nn = write(server_listenCon[atoi(robotName)], mess, strlen(mess+1));
  //  nn = write(server_listenCon[0], mess, strlen(mess+1));
  if( nn == -1 )
    {
      printf("Error form write\n");
      return false;
    }
  // close the connection ?

  return true;
}

char *wssRead()
{
  int nn;
  char msg[128];
  int trash = 0;

  if( trash == 0 )
    nn = recv(server_listenCon[1], msg, sizeof(msg)-1, MSG_DONTWAIT);
  else
    nn = read(server_listenCon[1], msg, sizeof(msg)-1);
  if( nn == -1 )
    {
      return NULL;
    }
  msg[nn] = '\0';
  printf( "Message received: %s\n", msg );
  return NULL;
}

/////////////////////////////////////////////////////////////////////////////////////////
//// state tables
/////////////////////////////////////////////////////////////////////////////////////////
void doCmdNop(LinkMonStat * stat)
{
  //  ServoMobJACmdNop nopMsg;

  if(state_match(stat, NEW_COMMAND)) 
    {
      state_new(stat);
      status_next(stat, RCS_DONE);
      state_next(stat, S0);
    } 
  else // SO
    state_default(stat);
}

void doCmdInit(LinkMonStat * stat)
{
  char msg[128];
  int nn;
  int i;
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();

  if(state_match(stat, NEW_COMMAND)) 
    {
      printf( "Received init command\n" );
      state_new(stat);
      state_next(stat, S0);
    }
  if(state_match(stat, S0))
    {
      //Set up a socket to connect to the server
      if (-1 == (server_fd = sockclient(USARSimPort, USARSimHostName)))
	{
	  fprintf(stderr, "can't connect to server\n");
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      server_listen = new int[set->robotInfo_length];
      server_listenCon = new int[set->robotInfo_length];
      status_next(stat, RCS_EXEC);
      state_next(stat, S1);
    }
  if(state_match(stat, S1))
    {
      //Register with the Server
      for( i=0; i<set->robotInfo_length; i++ )
	{
	  sprintf( msg, "REGISTER %s %s;", set->robotInfo[i].robotName, "129.6.33.142" );
	  printf("%s: sending %d length message: %s\n", THIS_PROCESS, strlen(msg), msg);
	  nn = send( server_fd, msg, strlen(msg)+1, 0);
	  //	  send( server_fd, "\r\n", 2, 0);
	  if (nn != (int)strlen(msg)+1) 
	    {
	      fprintf(stderr, "Write Failed\n");
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  //Wait for OK response
	  nn = recv(server_fd, msg, sizeof(msg)-1, 0);
	  if( nn == -1 )
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  msg[nn] = '\0';
	  printf("%s: message received:%s\n", THIS_PROCESS, msg);
	  if(msg[0]!='O' || msg[1]!='K')
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  else
	    {
	      state_next(stat, S2);

	    }
	}
    }
  if(state_match(stat, S2))
    {
      //set robots to listen REMEMBER that robot 0 is the comm station
      for( i=0; i<set->robotInfo_length; i++ )
	//for( i=0; i<2; i++ )
	{
	  sprintf( msg, "listen %s %d;", set->robotInfo[i].robotName, USARSimListenPortBase + 3*i);
	  printf("%s: sending %d length message: %s\n", THIS_PROCESS, strlen(msg), msg);
	  nn = send( server_fd, msg, strlen(msg)+1, 0);
	  //	  send( server_fd, "\r\n", 2, 0);
	  if (nn != (int)strlen(msg)+1) 
	    {
	      fprintf(stderr, "Write Failed\n");
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  //Wait for OK response
	  nn = recv(server_fd, msg, sizeof(msg)-1, 0);
	  if( nn == -1 )
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  msg[nn] = '\0';
	  printf("%s: message received:%s\n", THIS_PROCESS, msg);
	  if(msg[0]!='O' || msg[1]!='K')
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  else
	    {
	      state_next(stat, S3);
	      stat->admin_state = ADMIN_INITIALIZED;
	    }
	}
    }

  if(state_match(stat, S3))
    {
      //open connections between robot 0 and the other robots
      for( i=1; i<set->robotInfo_length; i++ )
	//      for( i=1; i<2; i++ )
	{
	  sprintf( msg, "OPEN %s %d %s %d;", set->robotInfo[0].robotName, USARSimListenPortBase,
		   set->robotInfo[i].robotName, USARSimListenPortBase+3*i);
	  printf("%s: sending %d length message: %s\n", THIS_PROCESS, strlen(msg), msg);
	  nn = send( server_fd, msg, strlen(msg)+1, 0);
	  //	  send( server_fd, "\r\n", 2, 0);
	  if (nn != (int)strlen(msg)+1) 
	    {
	      fprintf(stderr, "Write Failed\n");
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  //Wait for OK response
	  nn = recv(server_fd, msg, sizeof(msg)-1, 0);
	  if( nn == -1 )
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  msg[nn] = '\0';
	  printf("%s: message received:%s\n", THIS_PROCESS, msg);
	  if(msg[0]!='O' || msg[1]!='K')
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	      return;
	    }
	  else
	    {
	      state_next(stat, S4);
	    }
	}
    }
  if(state_match(stat, S4))
    {
      for( i=0; i<set->robotInfo_length; i++ )
	{
	  printf( "Listening to listen port %d\n", USARSimListenPortBase+i*3 );
	  sprintf(msg, "%d", USARSimListenPortBase+i*3 );
	  if( -1 == (server_listen[i] = tcp_listen(msg)))
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	    }
	  else if (-1 == (server_listenCon[i] = tcp_accept(server_listen[i])))
	    {
	      status_next(stat, RCS_ERROR);
	      state_next(stat, S10);
	    }
	  state_next(stat, S5);
	}
      stat->admin_state = ADMIN_INITIALIZED;
      printf( "Done with listening\n" );
    }
  // else wait here
  else
    state_default(stat);
}

void doCmdShutdown(LinkMonStat * stat)
{
  if(state_match(stat, NEW_COMMAND)) 
    {
      done = 1;
      stat->admin_state = ADMIN_UNINITIALIZED;
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
    }
  // else wait here
  else 			// S1
    state_default(stat);
}

void doCmdHalt(LinkMonStat * stat)
{
  if(state_match(stat, NEW_COMMAND)) 
    {
      done = 1;
      stat->admin_state = ADMIN_UNINITIALIZED;
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
    }
  // else wait here
  else 			// S1
    state_default(stat);
}

void doCmdAbort(LinkMonStat * stat)
{
  if(state_match(stat, NEW_COMMAND)) 
    {
      done = 1;
      stat->admin_state = ADMIN_UNINITIALIZED;
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
    }
  // else wait here
  else 			// S1
    state_default(stat);
}

void doCmdWrite( LinkMonStat * stat, LinkMonCmdWrite *cmdIn )
{
  int nn;
  char msg[128];
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();

  if(state_match(stat, NEW_COMMAND)) 
    {
      state_new(stat);
      if(stat->admin_state != ADMIN_INITIALIZED) 
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S2);
	} 
      else 
	{
	  state_next(stat, S1);
	  status_next(stat, RCS_EXEC);
	}
    }

  // state NOT USED at this time
  if(state_match(stat, S0))
    {
      //open connections between robot 0 and the other robots
      sprintf( msg, "open %s %d %s %d;", 
	       set->robotInfo[atoi(cmdIn->robotToWrite)].robotName, 
	       USARSimListenPortBase+3*atoi(cmdIn->robotToWrite),
	       set->robotInfo[0].robotName, USARSimListenPortBase);
      printf("%s: sending %d length message: %s\n", THIS_PROCESS, strlen(msg), msg);
      nn = send( server_fd, msg, strlen(msg)+1, 0);
      //	  send( server_fd, "\r\n", 2, 0);
      if (nn != (int)strlen(msg)+1) 
	{
	  fprintf(stderr, "Write Failed\n");
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      //Wait for OK response
      nn = recv(server_fd, msg, sizeof(msg)-1, 0);
      if( nn == -1 )
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      msg[nn] = '\0';
      printf("%s: message received:%s\n", THIS_PROCESS, msg);
      if(msg[0]!='O' || msg[1]!='K')
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      else
	{
	  state_next(stat, S1);
	}
    }

  if(state_match(stat, S1))
    {

      if( !wssWrite(cmdIn->robotToWrite, cmdIn->msg ) )
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S2);
	}
      else
	{
	  status_next(stat, RCS_DONE);
	  state_next(stat, S2);
	}
    }
  if(state_match(stat, S2))
    {
      //open connections between robot 0 and the other robots
      sprintf( msg, "CLOSE;");
      printf("%s: sending %d length message: %s\n", THIS_PROCESS, strlen(msg), msg);
      nn = send( server_fd, msg, strlen(msg)+1, 0);
      //	  send( server_fd, "\r\n", 2, 0);
      if (nn != (int)strlen(msg)+1) 
	{
	  fprintf(stderr, "Write Failed\n");
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      //Wait for OK response
      nn = recv(server_fd, msg, sizeof(msg)-1, 0);
      if( nn == -1 )
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      msg[nn] = '\0';
      printf("%s: message received:%s\n", THIS_PROCESS, msg);
      if(msg[0]!='O' || msg[1]!='K')
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      else
	{
	  state_next(stat, S4);
	}
    }

}

void doCmdMonitor( LinkMonStat * stat )
{
  int n;
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetNml->get_address();
  int numGoals = set->robotInfo_length -1;
  int statuses[MAX_TRAFFIC_VEHICLES];
  double costs[MAX_TRAFFIC_VEHICLES];
  int searcherGoals[MAX_TRAFFIC_VEHICLES];
  int goalsFound;
  int optimaFound;
  int cycles = 0;
  int length;
  int opens;                      // number of open nodes 
  int path[MAX_TRAFFIC_VEHICLES];
  int start = 0;

  if(state_match(stat, NEW_COMMAND)) 
    {
      state_new(stat);
      if(stat->admin_state != ADMIN_INITIALIZED) 
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S2);
	} 
      else 
	{
	  state_next(stat, S0);
	  status_next(stat, RCS_EXEC);
	}
    }

  if(state_match(stat, S0))
    {
      setGraph(); // based on readGraph
      status_next(stat, RCS_EXEC);
      state_next(stat, S1);
    }
  if(state_match(stat, S1)) 
    {
      for( n=0; n<numGoals; n++ )
	{
	  searcherGoals[n] = n+1;
	}
      if(theSearcher.initSearch(&searcherGraph, start, numGoals, 
				searcherGoals, &opens, statuses, costs,
				&goalsFound, &optimaFound))
	{ // initialization error occurred
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S10);
	  return;
	}
      while (opens &&
	     ((statuses[0] != 2) || (cycles < MAX_CYCLES)) &&
	     (optimaFound < numGoals))
	{
	  n = theSearcher.getNextNode();
	  computeCosts(n);
	  //	  insertKids(n, searcherGraph.size);
	  theSearcher.cycleSearch();
	  cycles++;
	}
      //      printf("%d cycles\n", cycles);
      stat->route_length = 0;
      for (n = 0; n < numGoals; n++)
	{
	  if (statuses[n] > 0)
	    {
	      length = theSearcher.reportSearch(searcherGoals[n],
						DIJK_PATH_SIZE, path);
	      /*
	      printf("%s", ((statuses[n] == 2) ? "Optimal path " : "Path "));
	      printf("from start to node %d:", searcherGoals[n]);
	      */
	      printAnswer(length, path, costs[n], stat);
	    }
	}
      theSearcher.closeSearch();
      status_next(stat, RCS_DONE);
      state_next(stat, S0);
    }
  else 
    {
      state_default(stat);
    }
  return;
}

//////////////////////////////////////////////////////////////////
//// end of state tables
//////////////////////////////////////////////////////////////////

static int cleanupNmlBuffers(void)
{
  if(NULL != linkMonCmdNml)
    delete linkMonCmdNml;
  if(NULL != linkMonStatNml)
    delete linkMonStatNml;
  if (NULL != trafficCtrlSetNml)
    delete trafficCtrlSetNml;

  linkMonCmdNml = NULL;
  linkMonStatNml = NULL;
  trafficCtrlSetNml = NULL;
  return 0;
}

static int initNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupNmlBuffers();

  sprintf(chanName, "%s%d", LINK_MON_CMD_NAME, bufnum);
  linkMonCmdNml = new RCS_CMD_CHANNEL(linkMonCtrl_format, chanName,
				      THIS_PROCESS, config_file);
  if(!linkMonCmdNml->valid())
    {
      delete linkMonCmdNml;
      linkMonCmdNml = 0;
      return 1;
    }

  sprintf(chanName, "%s%d", LINK_MON_STAT_NAME, bufnum);
  linkMonStatNml = new RCS_STAT_CHANNEL(linkMonCtrl_format, chanName,
					THIS_PROCESS, config_file);
  if(!linkMonStatNml->valid()) 
    {
      delete linkMonStatNml;
      linkMonStatNml = 0;
      return 1;
    }

  sprintf(chanName, "%s%d", TRAFFIC_CTRL_SET_NAME, bufnum);
  trafficCtrlSetNml =
    new RCS_STAT_CHANNEL(trafficCtrl_format, chanName,
			 THIS_PROCESS, config_file);
  if (!trafficCtrlSetNml->valid()) {
    delete trafficCtrlSetNml;
    trafficCtrlSetNml = 0;
    return 1;
  }

  nml_start();
  return 0;
}

void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  char *msgIn;
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  int bufnum = 1;
  int retval;
  NMLTYPE cmd_type;
  int serial_number;
  LinkMonCmdNop nopCmd;
  RCS_TIMER *timer;
  float cycleTime = 1;
  //  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetBuf->get_address();
  INIFILE if_d;
  char *INI_FILE = NULL;
  char *ini_file_env;
  char ini_file_default[] = DEFAULT_INI_FILE;
  const char *strPtr;

  timer = new RCS_TIMER(cycleTime);
  while (1) 
    {
      char option = getopt(argc, argv, "n:d:i:");
      if (option == -1)
	break;
      switch (option) 
	{
	case 'i':
	  INI_FILE = optarg;
	  break;
	case 'n':
	  NML_FILE = optarg;
	  break;
	case 'd':
	  if(1 != sscanf(optarg, "%i", &bufnum))
	    {
	      rcs_print_error("%s: bad value for buf number: %s\n", argv[0], 
			      optarg);
	      return 1;
	    }
	  break;
	default:
	  rcs_print_error("%s: bad argument\n", argv[0]);
	  return 1;
	  break;
	}
    }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if(NULL == NML_FILE) 
    {
      nml_file_env = getenv("CONFIG_NML");
      if(NULL != nml_file_env)
	{
	  NML_FILE = nml_file_env;
	} 
      else 
	{
	  NML_FILE = nml_file_default;
	}
    }

  // read in the INI file params
  if (NULL == INI_FILE) 
    {
      ini_file_env = getenv("TRAFFIC_INI");
      if (NULL != ini_file_env) 
	{
	  INI_FILE = ini_file_env;
	} 
      else 
	{
	  INI_FILE = ini_file_default;
	}
    }
  if (if_d.open (INI_FILE))
    {
      printf("** ERROR ** %s: invalid inifile path %s\n",
	     THIS_PROCESS, INI_FILE );
      if_d.close ();
      return false;
    }

  strPtr = if_d.find ("HOST_NAME", "USARSIM.INFSKIN");
  if ( strPtr == 0 )
    {
      strPtr = getenv("USARSimHost");
    }
  if (NULL == strPtr)
    {
      printf("** ERROR ** %s: 'HOST_NAME' not defined in section 'USARSIM.INFSKIN' in %s\n",
	     THIS_PROCESS, INI_FILE );
      if_d.close ();
      return false;
    }
  USARSimHostName = (char*)strPtr;
  strPtr = if_d.find ("CUTOFF", "LINKMON");
  if(strPtr != 0 )
    radioCutOff = -atof(strPtr);
  if_d.close ();

  // connect to my buffers
  retval = initNmlBuffers(NML_FILE, bufnum);
  if(retval) 
    {
      rcs_print_error("%s: can't allocate linkMon NML buffers\n", argv[0]);
      (void) cleanupNmlBuffers();
      return 1;
    }

  // initialize and set the status up so that init command goes through
  linkMonStat.echo_serial_number = 0;
  reset_time_tracker(&linkMonStat.tt);

  // stuff a NOP command
  nopCmd.serial_number = 1;
  linkMonCmdNml->write(&nopCmd);

  done = 0;
  signal(SIGINT, quit);
  while (!done) 
    {
      // clear stat buffer of routes
      linkMonStat.route_length = 0;

      // read buffers
      cmd_type = linkMonCmdNml->read();
      trafficCtrlSetNml->read();
      if( linkMonStat.admin_state == ADMIN_INITIALIZED )
	{
	  if( (msgIn=wssRead()) != NULL )
	    printf( "Message received: %s\n", msgIn );
	}

      // listen to all robots over USARSim commServer

      serial_number = linkMonCmdNml->get_address()->serial_number;
      switch (cmd_type) 
	{
	case 0:
	  // no new command
	  break;
	case -1:
	  // comm error
	  rcs_print_error("%s: NML command read error\n", argv[0]);
	  break;
	case LINK_MON_CMD_NOP_TYPE:
	case LINK_MON_CMD_INIT_TYPE:
	case LINK_MON_CMD_HALT_TYPE:
	case LINK_MON_CMD_ABORT_TYPE:
	case LINK_MON_CMD_MONITOR_TYPE:
	case LINK_MON_CMD_SHUTDOWN_TYPE:
	case LINK_MON_CMD_WRITE_TYPE:
	  linkMonStat.command_type = cmd_type;
	  if(serial_number != linkMonStat.echo_serial_number) 
	    {
	      linkMonStat.echo_serial_number = serial_number;
	      linkMonStat.state = NEW_COMMAND;
	      linkMonStat.status = RCS_EXEC;
	    }
	  break;
	default:
	  rcs_print_error("%s: unknown cmd: %d\n", argv[0], (int) cmd_type);
	  break;
	}				// switch (cmd_type)

      // handle command state tables
      switch (linkMonStat.command_type) 
	{
	case LINK_MON_CMD_NOP_TYPE:
	  doCmdNop(&linkMonStat);
	  break;
	case LINK_MON_CMD_INIT_TYPE:
	  doCmdInit(&linkMonStat);
	  break;
	case LINK_MON_CMD_HALT_TYPE:
	  doCmdHalt(&linkMonStat);
	  break;
	case LINK_MON_CMD_ABORT_TYPE:
	  doCmdAbort(&linkMonStat);
	  break;
	case LINK_MON_CMD_SHUTDOWN_TYPE:
	  doCmdShutdown(&linkMonStat);
	  break;
	case LINK_MON_CMD_MONITOR_TYPE:
	  doCmdMonitor(&linkMonStat);
	  break;
	case LINK_MON_CMD_WRITE_TYPE:
	  doCmdWrite(&linkMonStat, (LinkMonCmdWrite*)linkMonCmdNml->get_address());
	default:
	  // no command to handle
	  break;
	}				// switch (cmd)
      cycle_time_tracker(&linkMonStat.tt);

      // update status and settings
      linkMonStatNml->write(&linkMonStat);
      timer->wait();
    }
  while (!done);

  (void) cleanupNmlBuffers();

  rcs_print("linkMon done\n");

  return 0;
}

/*
  Modification history:

  $Log: linkMon.cc,v $
  Revision 1.5  2007/09/20 20:14:57  dr_steveb
  Better, but still not working

  Revision 1.4  2007/09/20 19:26:37  dr_steveb
  Added additional link connection code for USARSim. Not working yet!

  Revision 1.3  2007/09/18 20:03:49  dr_steveb
  Added write command

  Revision 1.2  2007/09/11 20:52:00  dr_steveb
  Changed default cuttoff to read a negative value

  Revision 1.1  2007/09/11 20:26:53  dr_steveb
  Initial check-in

*/

