/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*!
  \file nmlPrint.cc
  \brief Reads and prints the specified NML buffer

  Synopsis: nmlPrint {--nml \<file\>} \n
                     {--period \<secs\>} \n
		     {--block} \n
		     --buffer,-b \<buffer name\> \n
		     --process,-p \<process name\>
		     --forever
  
  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.24 $
  $Date: 2008/11/20 15:59:31 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		// sscanf()
#include <string.h>		// strcmp()
#include <stddef.h>		// NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#include <sys/time.h>               // gettimeofday()
#ifdef HAVE_GETOPT_H
#include <getopt.h>		// getopt_long()
#else
#include "getopt.h"
#endif
#include <rcs.hh>		// NMTYPE, CMS
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_EXT_BASE, ...
#include "moastError.hh"	// moastError_format()
#include "navDataExt.hh"	// navDataExt_format(), NavDataExt, ...
#include "sensorData.hh"	// sensorData_format(), SensorData1D, ...
#include "servoMobJA.hh"	// servoMobJA_format(), ServoMobJAStat, ...
#include "servoSP.hh"   	// servoMobJA_format(), ServoMobJAStat, ...
#include "servoMisJA.hh"        // servoMisJA_format(), ServoMisJAStat, ...
#include "amMobJA.hh"           // amMobJA_format(), amMobJACmd, ...
#include "primMobJA.hh"         // primMobJA_format(), primMobJACmd, ..

#define DEFAULT_NML_FILE "../etc/moast.nml"

static int all_format(NMLTYPE type, void *buffer, CMS *cms)
{
  if (moastError_format(type, buffer, cms)) return 1;
  if (navDataExt_format(type, buffer, cms)) return 1;
  if (sensorData_format(type, buffer, cms)) return 1;
  if (servoMobJA_format(type, buffer, cms)) return 1;
  if (servoMisJA_format(type, buffer, cms)) return 1;
  if (amMobJA_format(type, buffer, cms)) return 1;
  if (primMobJA_format(type, buffer, cms)) return 1;
  return 0;
}

static void printHelp(void)
{
  printf("options:\n{--nml <file>}\n{--period <secs>}\n{--block}\n--buffer,-b <buffer name>\n--process,-p <process name>\n{--number,-n <buffer number>}\n{--help,-h}\n {--forever - keep going forever}\n");
}

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  int c;
  int this_option_optind;
  int option_index;
  struct option long_options[] = {
    /* --option, needs_arg, 0, 'shortcut char' */
    {"nml", required_argument, 0, 0},
    {"period", required_argument, 0, 0},
    {"block", no_argument, 0, 0},
    {"buffer", required_argument, 0, 'b'},
    {"process", required_argument, 0, 'p'},
    {"number", required_argument, 0, 'n'},
    {"help", no_argument, 0, 'h'},
    {"id", required_argument, 0, 'i'},
    {"notime", no_argument, 0, 0},
    {"forever", no_argument, 0, 0},
    {0, 0, 0, 0}
  };
  NML * nmlBuf = 0;
  NMLTYPE type;
  NMLTYPE baseType;
  char * envptr = 0;
  char nmlFile[MOAST_FILE_NAME_LEN] = {0};
  char bufferName[MOAST_NML_BUFFER_NAME_LEN] = {0};
  char processName[MOAST_NML_BUFFER_NAME_LEN] = {0};
  char bufNumString[MOAST_NML_BUFFER_NAME_LEN] = {0};
  int bufNum = 1;
  double period = .1;		
  int doBlock = 0;
  int printTimestamp = 1;
  int t, u;
  double th;
  timeval tv;
  double timeOut;
  int mission_id = 0;
  
  opterr = 0;			// inhibit getopt_long's error printing
  done = 1;
  while (1) {
    this_option_optind = optind ? optind : 1;
    option_index = 0;

    // The leading colon means generate a ':' for missing parameters.
    // Colons mean a required argument for the preceding option.
    c = getopt_long(argc, argv, ":b:p:n:hi:",
		    long_options, &option_index);
    if (c == -1)
      break;

    switch (c) {
      // this handles --long args
    case 0:		
      if (!strcmp("nml", long_options[option_index].name)) {
	strncpy(nmlFile, optarg, MOAST_FILE_NAME_LEN);
	nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
      } else if (!strcmp("period", long_options[option_index].name)) {
	if (1 != sscanf(optarg, "%lf", &period)) {
	  fprintf(stderr, "%s: bad paramter for --period: `%s'\n",
		  argv[0], optarg);
	}
      } else if (!strcmp("block", long_options[option_index].name)) {
	doBlock = 1;
      } else if (!strcmp("notime", long_options[option_index].name)) {
	printTimestamp = 0;
      } else if (!strcmp("id", long_options[option_index].name)) {
	mission_id = atoi(optarg);
      } else if (!strcmp("forever", long_options[option_index].name)){
	done = 0;
      } else {
	// should never get here, since 0 above means OK
	fprintf(stderr, "%s: unknown option `%s'",
		argv[0],
		long_options[option_index].name);
	if (optarg)
	  fprintf(stderr, " with arg %s", optarg);
	fprintf(stderr, "\n");
      }
      break;

      // the rest handle -short args

    case 'i':
	mission_id = atoi(optarg);
	break;

    case 'b':
      strncpy(bufferName, optarg, MOAST_NML_BUFFER_NAME_LEN - 1);
      break;

    case 'p':
      strncpy(processName, optarg, MOAST_NML_BUFFER_NAME_LEN - 1);
      break;

    case 'n':
      if (1 != sscanf(optarg, "%i", &bufNum)) {
	printf("%s: bad number parameter: ``%s''\n", argv[0],  optarg);
	return 1;
      }
      break;

    case 'h':
      printHelp();
      return 0;
      break;

    case '?':
      fprintf(stderr, "%s: unknown option `%s'\n",
	      argv[0], argv[this_option_optind]);
      return 1;
      break;

    case ':':
      fprintf(stderr, "%s: missing parameter to `%s'\n",
	      argv[0], argv[this_option_optind]);
      return 1;
      break;

    default:
      fprintf(stderr, "%s: bad argument `%s'\n",
	      argv[0], argv[this_option_optind]);
      return 1;
      break;
    }
  }
  if (optind < argc) {
    fprintf(stderr, "%s: extra arguments provided:", argv[0]);
    while (optind < argc)
      fprintf(stderr, " %s", argv[optind++]);
    fprintf(stderr, "\n");
    return 1;
  }
  if (bufferName[0] == 0) {
    fprintf(stderr, "%s: missing the --buffer,-b option\n", argv[0]);
    return 1;
  }
  if (processName[0] == 0) {
    fprintf(stderr, "%s: missing the --process,-p option\n", argv[0]);
    return 1;
  }

  // OK, we made it through the arguments

  if (0 == nmlFile[0]) {
    if (NULL == (envptr = getenv("CONFIG_NML"))) {
      strncpy(nmlFile, DEFAULT_NML_FILE, MOAST_FILE_NAME_LEN);
      nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
    } else {
      strncpy(nmlFile, envptr, MOAST_FILE_NAME_LEN);
      nmlFile[MOAST_FILE_NAME_LEN - 1] = 0;
    }
  }

  if (0 != strcmp(bufferName, "moastError")) {
    sprintf(bufNumString, "%d", bufNum);
    strcat(bufferName, bufNumString);
    if( mission_id != 0 )
      {
	sprintf(bufNumString, "_%d", mission_id );
	strcat(bufferName, bufNumString);
      }
    
  }

  nmlBuf = new NML(all_format, bufferName, processName, nmlFile);
  if (!nmlBuf->valid()) {
    delete nmlBuf;
    fprintf(stderr, "%s can't connect to NML buffer %s\n",
	    processName, bufferName);
    return 1;
  }

  signal(SIGINT, quit);
  //  while (!done) {
  for(;;){
    if (doBlock)
      {
	type = nmlBuf->blocking_read(-1.0);
      }
    else
      {
	type = nmlBuf->read();
      }
    baseType = ((type / 1000) * 1000); // relies on 1000 gap between bases
    gettimeofday( &tv, NULL );
    timeOut = (double)tv.tv_sec + (double)tv.tv_usec/1000000; 
    if (printTimestamp)
      printf( "Timestamp: %lf\n", timeOut );
    switch (baseType) {
/*****************************************/
    case 0:
      switch (type) {
      case 0:
	break;
	
      case MOAST_ERROR_TYPE:
	{
	  MoastError *moastError = (MoastError *) nmlBuf->get_address();
	  printf("[%d] %s\n", moastError->error, moastError->message);
	}
	break;

      default:
	printf( "Unrecognized message type %d\n", (int) type );
	break;
      }
      break;

/*****************************************/
    case NAV_DATA_EXT_BASE:
      switch (type) {
	
      case NAV_DATA_EXT_TYPE:
	{
	  NavDataExt *navDataExt = (NavDataExt *) nmlBuf->get_address();
	  printf("%f %f %f %f %f %f %f ",
		 navDataExt->time,
		 navDataExt->tranAbs.x, 
		 navDataExt->tranAbs.y,
		 navDataExt->tranAbs.z,
		 navDataExt->tranRel.x,
		 navDataExt->tranRel.y,
		 navDataExt->tranRel.z);
	  printf("%f %f %f %f %f %f ",
		 navDataExt->rpyAbs.r,
		 navDataExt->rpyAbs.p,
		 navDataExt->rpyAbs.y,
		 navDataExt->rpyRel.r,
		 navDataExt->rpyRel.p,
		 navDataExt->rpyRel.y);
	  printf("%f %f %f\n",
		 navDataExt->tranRelRate.x, 
		 navDataExt->tranRelRate.y,
		 navDataExt->tranRelRate.z);
	}
	break;

      default:
	printf( "Unrecognized nav data message type %d\n", (int) type );
	break;
      }
      break;
      
/*****************************************/
    case SENSOR_DATA_BASE:
      switch (type) {
	
      case SENSOR_DATA_ARRAY_TYPE:
	{
	  SensorDataArray *sensorDataArray =
	    (SensorDataArray *) nmlBuf->get_address();
	  for(int i=0; i < sensorDataArray->arrayElem_length; i++)
	    {
	      if( sensorDataArray->arrayElem[i].type == SERVO_SP_SONAR_SENSOR )
		printf( "%d %f\n", i, sensorDataArray->arrayElem[i].dataTuple[0] );
	      else
		{
		  printf("type: %s id:%d cTime:%lf\n", 
			 servoSP_enum_ServoSensorType_symbol_lookup
			 (sensorDataArray->arrayElem[i].type),
			 sensorDataArray->arrayElem[i].id,
			 sensorDataArray->arrayElem[i].collectionTime);
		  for(int j=0; j < MAX_SENSOR_DATA_ARRAY_TUPLE; j++)
		    {
		      printf("%lf ", sensorDataArray->arrayElem[i].dataTuple[j]);
		    }
		  printf("\n");
		}
	    }
	}
	break;


      case SENSOR_DATA_TEXT_ARRAY_TYPE:
	{
	  SensorDataTextArray *sensorDataTextArray =
	    (SensorDataTextArray *) nmlBuf->get_address();
	  for(int i=0; i < sensorDataTextArray->dataElem_length; i++)
	    {
	      printf("type: %s id:%d cTime:%lf\n", 
		     servoSP_enum_ServoSensorType_symbol_lookup
		     (sensorDataTextArray->dataElem[i].sensorType),
		     sensorDataTextArray->dataElem[i].sensorID,
		     sensorDataTextArray->dataElem[i].collectionTime);
	      for(int j=0; j < sensorDataTextArray->dataElem[i].dataTuple_length; j++)
		{
		  printf("%lf ", sensorDataTextArray->dataElem[i].dataTuple[j]);
		}
	      printf("\n");
	      if( sensorDataTextArray->dataElem[i].dataStr_length > 0 )
		printf("%s : %d\n", sensorDataTextArray->dataElem[i].dataStr, sensorDataTextArray->dataElem[i].dataStr_length);
	      else 
		printf("No Text Data Present\n");
	    }
	}
	break;
	
      case SENSOR_DATA_TXT_TYPE:
	{
	  SensorDataText * sensorDataTxt =
	    (SensorDataText *) nmlBuf->get_address();
	  char * sPtr;
	  for (t=0; t < sensorDataTxt->strPtr_length; t++)
	    {
	      sPtr = &sensorDataTxt->txtBuffer[sensorDataTxt->strPtr[t]];
	      printf("%s %lf %lf %lf %lf\n", sPtr, 
		     sensorDataTxt->dVal[t*4],
		     sensorDataTxt->dVal[t*4+1],
		     sensorDataTxt->dVal[t*4+2],
		     sensorDataTxt->dVal[t*4+3]); 
	    }
	  printf("\n");
	}
	break;
	
      case SENSOR_DATA_1D_TYPE:
	{
	  SensorData1D *sensorData1D = (SensorData1D *) nmlBuf->get_address();
	  if( printTimestamp )
	    printf( "%lf\n", sensorData1D->collectTimeStart );
	  for (t = 0, th = sensorData1D->start;
	       t < sensorData1D->featArray_length;
	       t++, th += sensorData1D->increment)
	    {
	      printf("%f %f\n", th, sensorData1D->featArray[t]);
	    }
	}
	break;
	
      case SENSOR_DATA_LIST_TYPE:
	{
	  SensorDataList *sensorDataList =
	    (SensorDataList *) nmlBuf->get_address();
	  if(printTimestamp)
	    {
	      printf("%lf %d %lf %lf %d\n", 
		     sensorDataList->collectTime,
		     sensorDataList->isGlobal,
		     sensorDataList->maxRange,
		     sensorDataList->minRange,
		     sensorDataList->featList_length);
	      for (t = 0; t < sensorDataList->featList_length; t++ )
		printf("%lf %lf %lf %lf %lf %lf %lf\n", 
		       sensorDataList->featList[t].dataLoc.x,
		       sensorDataList->featList[t].dataLoc.y,
		       sensorDataList->featList[t].dataLoc.z,
		       sensorDataList->featList[t].sensorLoc.x,
		       sensorDataList->featList[t].sensorLoc.y,
		       sensorDataList->featList[t].sensorLoc.z,
		       sensorDataList->featList[t].param1);
	    }
	  else
	    {
	      for (t = 0; t < sensorDataList->featList_length; t++ )
		printf("%lf %lf\n%lf %lf \n\n", 
		       sensorDataList->featList[t].dataLoc.x,
		       sensorDataList->featList[t].dataLoc.y,
		       sensorDataList->featList[t].sensorLoc.x,
		       sensorDataList->featList[t].sensorLoc.y);
	    }
	}
	break;

      case SENSOR_DATA_2D_TYPE:
	{
	  SensorData2D *sensorData2D = (SensorData2D *) nmlBuf->get_address();
	  printf( "P3\n#%lf %lf %lf %lf %lf\n%d %d\n255\n", 
		  sensorData2D->collectTime,
		  sensorData2D->lowerLeft.x, 
		  sensorData2D->lowerLeft.y,
		  sensorData2D->incrementX, sensorData2D->incrementY,
		  sensorData2D->x, sensorData2D->y);
	  for (t=0; t < sensorData2D->x; t++ )
	    {
	      for (u=0; u < sensorData2D->y; u++ )
		{
		  /*
		    printf("%lf ",
		           sensorData2D->featArray[t*sensorData2D->x+u]);
		  */
		  if( sensorData2D->featArray[t*sensorData2D->x+u] == 0 )
		    printf( "0 0 0 " );
		  else if(sensorData2D->featArray[t*sensorData2D->x+u] <= .21)
		    printf( "128 128 128" );
		  else if(sensorData2D->featArray[t*sensorData2D->x+u] <= .41)
		    printf( "254 254 0" );
		  else 
		    printf( "254 0 0" );
		  printf( " " );
		}
	      printf( "\n" );
	    }
	  // now print data as float
	  printf("# now for the raw data\n" );
	  for (t=0; t<sensorData2D->x; t++ )
	    {
	      for (u=0; u<sensorData2D->y; u++ )
		{
		  printf("%lf ", sensorData2D->featArray[t*sensorData2D->x+u]);
		}
	    }
	  printf( "\n" );
	  fflush(stdout);
	}
	break;

      default:
	printf( "Unrecognized sensor data message type %d\n", (int) type );
	break;
      }
      break;

/*****************************************/
    case SYMBOLIC_DATA_BASE:
      switch (type) {

      case SYMBOLIC_DATA_FEATURE_TYPE:
	{
	  SymbolicDataFeature *symbolicData = 
	    (SymbolicDataFeature *) nmlBuf->get_address();
	  for (t=0; t<symbolicData->featList_length; t++ )
	    {
	      for( u=0; u<symbolicData->featList[t].vertices_length; u++ )
		{
		  printf( "%f %f\n", 
			  symbolicData->featList[t].vertices[u].north,
			  symbolicData->featList[t].vertices[u].east );
		}
	      printf( "\n" );
	    }
	}
	break;
	
      default:
	printf( "Unrecognized symbolic data message type %d\n", (int) type );
	break;
      }
      break;
      
/*****************************************/
    case SERVO_MOB_JA_BASE:
      switch (type) {

      case SERVO_MOB_JA_STAT_TYPE:
	{
	  ServoMobJAStat *servoMobJAStat =
	    (ServoMobJAStat *) nmlBuf->get_address();
	  printf("%d\n", servoMobJAStat->echo_serial_number);
	  printf("%d\n", servoMobJAStat->heartbeat);
	}
	break;
	
      case SERVO_MIS_JA_STAT_TYPE:
	break;

      default:
	printf( "Unrecognized servo mob message type %d\n", (int) type );
	break;
      }
      break;

/*****************************************/
    case AM_MOB_JA_BASE:
      switch (type) {
      
      case AM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
	{
	  AmMobJACmdMoveWaypoint *amMoveWp = 
	    (AmMobJACmdMoveWaypoint *)nmlBuf->get_address();
	  printf( "x y z neighborhood speed quality %d waypoints\n",
		  amMoveWp->waypoint_length);
	  for (t=0; t<amMoveWp->waypoint_length; t++ )
	    {
	      printf("%lf %lf %lf %lf %lf %d\n", 
		     amMoveWp->waypoint[t].p.x,
		     amMoveWp->waypoint[t].p.y,
		     amMoveWp->waypoint[t].p.z,
		     amMoveWp->waypoint[t].neighborhood,
		     amMoveWp->waypoint[t].speed,
		     amMoveWp->waypoint[t].quality);
	    }
	}
	break;

      default:
	printf( "Unrecognized AM mob message type %d\n", (int) type );
	break;
      }
      break;

/*****************************************/
    case PRIM_MOB_JA_BASE:
      switch (type) {
      
      case PRIM_MOB_JA_CMD_NOP_TYPE:
	printf( "nop\n");
	break;

      case PRIM_MOB_JA_CMD_INIT_TYPE:
	printf( "init\n");
	break;

      case PRIM_MOB_JA_CMD_ABORT_TYPE:
	printf( "abort\n");
	break;

      case PRIM_MOB_JA_CMD_HALT_TYPE:
	printf( "halt\n");
	break;

      case PRIM_MOB_JA_CMD_SHUTDOWN_TYPE:
	printf( "shutdown\n");
	break;

      case PRIM_MOB_JA_CMD_MOVE_ARC_SEGMENT_TYPE:
	{
	  PrimMobJACmdMoveArcSegment *primMoveArcSeg = 
	    (PrimMobJACmdMoveArcSegment *)nmlBuf->get_address();
	  if( printTimestamp )
	    {
	      printf("for %d arc segments\n", primMoveArcSeg->arcList_length);
	      printf("  (center)      (end)       angle ann_tol  speed isArc\n");
	    }
	  for (t = 0; t < primMoveArcSeg->arcList_length; t++ )
	    {
	      if( printTimestamp )
		printf("(%5.2lf %5.2lf) (%5.2lf %5.2lf) %.2lf  %.2lf     "
		       "%.2lf  %s\n",
		       primMoveArcSeg->arcList[t].center.x,
		       primMoveArcSeg->arcList[t].center.y,
		       primMoveArcSeg->arcList[t].end.x,
		       primMoveArcSeg->arcList[t].end.y,
		       primMoveArcSeg->arcList[t].theta,
		       primMoveArcSeg->arcList[t].annular_tol,
		       primMoveArcSeg->arcList[t].speed,
		       (primMoveArcSeg->arcList[t].isArc ? "arc" : "line"));
	      else
		{
		  if( primMoveArcSeg->arcList[t].isArc )
		    printf("%5.2lf %5.2lf  %5.2lf %5.2lf  %.2lf  %.2lf  %.2lf\n",
			   primMoveArcSeg->arcList[t].center.x,
			   primMoveArcSeg->arcList[t].center.y,
			   primMoveArcSeg->arcList[t].end.x,
			   primMoveArcSeg->arcList[t].end.y,
			   primMoveArcSeg->arcList[t].theta,
			   primMoveArcSeg->arcList[t].annular_tol,
			   primMoveArcSeg->arcList[t].speed );
		  else
		    printf("%5.2lf %5.2lf  %.2lf  %.2lf\n",
			   primMoveArcSeg->arcList[t].end.x,
			   primMoveArcSeg->arcList[t].end.y,
			   primMoveArcSeg->arcList[t].annular_tol,
			   primMoveArcSeg->arcList[t].speed );
		}
	    }
	}
	break;

      case PRIM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
	{
	  PrimMobJACmdMoveWaypoint *primMoveWp = 
	    (PrimMobJACmdMoveWaypoint *)nmlBuf->get_address();
	  printf( "x y z neighborhood speed quality %d waypoints\n",
		  primMoveWp->waypoint_length);
	  for (t=0; t<primMoveWp->waypoint_length; t++ )
	    {
	      printf("%lf %lf %lf %lf %lf %d\n", 
		     primMoveWp->waypoint[t].p.x,
		     primMoveWp->waypoint[t].p.y,
		     primMoveWp->waypoint[t].p.z,
		     primMoveWp->waypoint[t].neighborhood,
		     primMoveWp->waypoint[t].speed,
		     primMoveWp->waypoint[t].quality);
	    }
	}
	break;

      case PRIM_MOB_JA_CMD_ROTATE_TYPE:
	{
	  PrimMobJACmdRotate *primRotate =
	    (PrimMobJACmdRotate *)nmlBuf->get_address();
	  printf("PrimRotate angle=%.2lf tolerance=%.2lf\n",
		 primRotate->theta, primRotate->thetaTol);
	}
	break;

      case PRIM_MOB_JA_CMD_VEL_TYPE:
	{
	  PrimMobJACmdVel *primVel =
	    (PrimMobJACmdVel *)nmlBuf->get_address();
	  printf("PrimVel v=%.2lf  w=%.2lf\n", primVel->v, primVel->w);
	}
	break;

      case PRIM_MOB_JA_CMD_STOP_TYPE:
	printf("stop\n");
	break;

      default:
	printf( "Unrecognized prim mob message type %d\n", (int) type );
	break;
      }
      break;
      
/*****************************************/
    default:
      printf( "Unrecognized message type %d\n", (int) type );
      break;
    }

/*****************************************/
    
    if( done )
      {
	//	printf("Exiting...\n\n\n" );
	delete nmlBuf;
	return 0;
      }

    if (period < 0.0)
      {
	continue;
      }
    else
      {
	esleep(period);
      }
  }
}

/*
  Modification history:

  $Log: nmlPrint.cc,v $
  Revision 1.24  2008/11/20 15:59:31  dr_steveb
  Added ability to plot sonar positions with nmlPrint

  Revision 1.23  2008/09/08 14:45:00  dr_steveb
  Got rid of "Exiting..." printout that was killing automatic scripts

  Revision 1.22  2008/03/11 17:24:37  dr_steveb
  Fixed some syntax errors

  Revision 1.21  2008/03/11 13:17:26  dr_steveb
  Added code to plot prim commanded arcs in gnuplot.

  Revision 1.20  2007/12/10 16:57:32  cj_scrapper
  Added case for new SensorDataTextArray message

  Revision 1.19  2007/11/29 19:19:36  tomrkramer
  Added printing more prim commands. Modularized the function.

  Revision 1.18  2007/11/28 23:00:05  tomrkramer
  Added printing information about the Prim commands for (1) stop and
  (2) move arc segment. For move arc segment, only the speed is currently
  being printed.

  Revision 1.17  2007/03/12 20:14:49  cj_scrapper
  Deleted servoMis stuff

  Revision 1.16  2007/02/27 14:47:55  dr_steveb
  *** empty log message ***

  Revision 1.15  2007/01/29 15:33:20  proctor
  Added simpleSimInfSkin support, including GPS sensor in the SimWare KB

  Revision 1.14  2006/12/14 16:16:53  dr_steveb
  More changes for new interface

  Revision 1.13  2006/12/12 17:54:58  cj_scrapper
  added case for SensorDataArrayType

  Revision 1.12  2006/11/27 19:41:52  dr_steveb
  Added plan horizon to am settings message.

  Revision 1.11  2006/09/19 18:05:45  cj_scrapper
  Added SENSOR_DATA_TEXT to handle VictRFIDs

  Revision 1.10  2006/08/02 21:43:02  dr_steveb
  Added option of -i (or id) to set the mission package number in NML channels.

  Revision 1.9  2006/04/25 20:30:18  proctor
  Added support for moastError to nmlPrint, and added error printing for
  bad configurations to primMobJA

  Revision 1.8  2006/03/15 21:54:08  dr_steveb
  Files changed to allow for pretty pictures to be made for ICRA tutorial.

  Revision 1.7  2006/02/17 14:33:59  dr_steveb
  Added printing of symbolic data buffer.

  Revision 1.6  2006/02/07 18:04:25  dr_steveb
  Fixed print format function.

  Revision 1.5  2006/02/07 17:56:00  dr_steveb
  *** empty log message ***

  Revision 1.4  2005/12/14 14:52:55  cj_scrapper
  MAde changes to reflect changes in sensorData

  Revision 1.3  2005/12/07 21:43:58  dr_steveb
  Added the printing of the new time stamp for sensor messages.

  Revision 1.2  2005/10/20 20:59:14  dr_steveb
  Changed sensor2D buffer to output ppm file.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.13  2005/10/03 15:05:58  stephen
  Added printing of prim waypoint command

  Revision 1.12  2005/09/30 19:42:42  stephen
  Added printing of amMobJACmd

  Revision 1.11  2005/09/30 18:55:56  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.10  2005/09/30 14:31:19  stephen
  Added additional fields to sensor_2D printing

  Revision 1.9  2005/09/30 13:58:13  proctor
  Added --notime flag to inhibit printing of timestamp

  Revision 1.8  2005/09/29 13:16:37  stephen
  Made changes to work with sensorPoint changes.

  Revision 1.7  2005/09/21 00:15:20  stephen
  Added sensor2D printing.

  Revision 1.6  2005/09/15 21:17:16  stephen
  Fixed gettimeofday to have NULL parameter.

  Revision 1.5  2005/09/14 14:10:03  stephen
  Added additional channels for printing

  Revision 1.4  2005/07/26 21:15:55  proctor
  Changed some printing formats

  Revision 1.3  2005/07/25 22:41:19  proctor
  Printed tranRelRate

  Revision 1.2  2005/07/07 19:04:41  proctor
  Removed navPrint.cc, it's now nmlPrint.cc

  Revision 1.1  2005/07/06 22:27:36  proctor
  Initial checkin

*/
