/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   sweep.cc
  \brief  provides an implementation of a line sweep algorithm for polygon intersection

  This algorithm may be found in the CS 373 Non-Lecture F: Line Segment 
  Intersecion document from the Fall 2002 course taught by Jeff Erickson 
  of the University of Illinois at Urbana-Champaign 
  (http://compgeom.cs.uiuc.edu/~jeffe/).

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.3 $
  $Date: 2007/12/18 21:34:46 $
  \endcode

  \author Stephen Balakirsky
  \date   06/01/2007
*/

#include "sweep.hh"

//////////////////////////////////////////////////////////////////////////////
// supporting classes
/////////////////////////////////////////////////////////////////////////////
LinePoint::LinePoint()
{
  myLine = 0;
}

LinePoint::LinePoint(const LinePoint &copyIn) // copy constructor to handle pass by value
{
  isStart = copyIn.isStart;
  myLine = copyIn.myLine;
}

LinePoint& LinePoint::operator=(const LinePoint &rhs)
{
  this->isStart = rhs.isStart;
  this->myLine = rhs.myLine;
  return *this;
}

int LinePoint::operator==(const LinePoint &rhsIn) const
{
   Location2D *lhs, *rhs;
   LineSegment *myLine;

   myLine = (LineSegment*)this->myLine;
  if( this->isStart )
    lhs = &myLine->start;
  else
    lhs = &myLine->end;

  myLine = (LineSegment*)rhsIn.myLine;
  if( rhsIn.isStart )
    rhs = &myLine->start;
  else
    rhs = &myLine->end;

  if( lhs->north != rhs->north ) return 0;
  if( lhs->east != rhs->east ) return 0;

  return 1;
}

// sort by x and then y (north, east)
int LinePoint::operator<(const LinePoint &rhsIn) const
{
  Location2D *lhs, *rhs;
  LineSegment *myLine;

  myLine = (LineSegment*)this->myLine;
  if( this->isStart )
    lhs = &myLine->start;
  else
    lhs = &myLine->end;

  myLine = (LineSegment*)rhsIn.myLine;
  if( rhsIn.isStart )
    rhs = &myLine->start;
  else
    rhs = &myLine->end;

  if( lhs->north == rhs->north && lhs->east < rhs->east ) return 1; 
  if( lhs->north < rhs->north ) return 1;
  return 0;
}

LabelSeq::LabelSeq()
{
}

bool LabelSeq::CCW( Location2D &p1, Location2D &p2, Location2D &p3 )
{
  if( (p3.east-p1.east)*(p2.north-p1.north)>(p2.east-p1.east)*(p3.north-p1.north))
    return true;
  else
    return false;
}

void LabelSeq::clear()
{
  labels.clear();
}

bool LabelSeq::isInterior( LineSegment *line, Location2D &point )
{
  //  float a;
  
  //  a = pointLineDistance( line, point ); // debug

  if(pointLineDistance(line, point) < .01) // intersects now check for interior
    {
      if( fabs(line->start.north-line->end.north) < .01 ) // verticle
	{
	  if( line->start.east>point.east && line->end.east<point.east )
	    return true;
	}
      else  // not verticle
	{
	  if( line->start.north<point.north && line->end.north>point.north)
	    return true;
	}
    }
  return false;
}

// returns true if insertion results in an intersection
void LabelSeq::seqInsert( LineSegment *line, Location2D &sweepValue )
{
  list<LineSegment*>::iterator i;
  Location2D toCheckIn, toCheckArray;
  Location2D pointInsert;
  bool found = false;
  LineSegment sweepLine;

  // are we verticle?
  if( fabs(line->start.north-line->end.north)<.01 )
    {
      for( i=labels.begin(); i!=labels.end(); i++ )
	{
	  if( isInterior( (*i), sweepValue ) )
	    {
	      found = true;
	    }
	  else if( found )
	    {
	      break;
	    }
	}
      if( !found )
	i = labels.begin();
    }
  else
    {
      // move sweep a little to the right
      sweepValue.north += .01;
      sweepLine.start.east = line->start.east - 10;
      sweepLine.end.east = line->end.east + 10;
      sweepLine.start.north = sweepValue.north;
      sweepLine.end.north = sweepValue.north;
      
      pointInsert = line->start;
      toCheckIn = findIntersection( &sweepLine, line );
      
      for( i=labels.begin(); i!=labels.end(); i++ )
	{
	  toCheckArray = findIntersection( &sweepLine, (*i) );
	  if( toCheckIn.east > toCheckArray.east ) // goes above this array element
	    break;
	}
    }
  labels.insert( i, line );


  /*
  printf( "current line sequence at sweep %f:\n", sweepValue.north );
  for( i=labels.begin(); i!=labels.end(); i++ )
    {
      printf( "line: 0x%x <%f %f> <%f %f>\n",
	      (int)(*i), (*i)->start.north, (*i)->start.east,
	      (*i)->end.north, (*i)->end.east);

    }
  */

}

void LabelSeq::seqDelete( LineSegment *line )
{
  labels.remove(line);
}

bool LabelSeq::getNeighbors( LineSegment **pointIn, LineSegment **top, LineSegment **bottom )
{
  list<LineSegment*>::iterator i;
  bool found = false;

  *top = NULL;
  *bottom = NULL;
  for( i=labels.begin(); i!=labels.end(); i++ )
    {
      if(found)
	{
	  *bottom = (*i);
	  break;
	}
      if( (*i) == *pointIn ) // found it
	{
	  found = true;
	}
      if(!found)
	*top = (*i);
    }
  return found;
}

//////////////////////////////////////////////////////////////////////////////
// end of supporting classes
/////////////////////////////////////////////////////////////////////////////


Sweep::Sweep()
{
}

Sweep::~Sweep()
{
}

void Sweep::anyIntersections()
{
  vector<LineSegment>::iterator k;
  Location2D currentPoint;
  //  int label;
  //  Location2D intLocation;
  //  LineSegment *myLine;
  //  bool sharePoint;

  labelseq.clear();
  pointList.clear();

  // fill in point list
  for( k=segments.begin(); k!=segments.end(); k++ )
    {
      currentPoint = (*k).start;
      pointList.push_back(currentPoint);
      currentPoint = (*k).end;
      pointList.push_back(currentPoint);
    }
  pointList.sort();
  pointList.unique();

  while( !pointList.empty() )
    {
      currentPoint = (*pointList.begin());
      pointList.pop_front();
      handleEventPoint( currentPoint );
    }
}

void Sweep::handleEventPoint( Location2D &currentPoint )
{
  list<LineSegment*>leftPoint;
  list<LineSegment*>rightPoint;
  list<LineSegment*>labelPoint;
  list<LineSegment*>::iterator labelIt;
  list<LineSegment*>::iterator unionIt;
  list<LineSegment*>::reverse_iterator labelRIt;
  vector<LineSegment>::iterator segIt;
  LineSegment *successor, *predecessor, *current;
  bool found;
  long int firstFound;

  // load up all lines that have a matching end point  NEED TO FIX ME!
  for( segIt=segments.begin(); segIt!=segments.end(); segIt++ )
    {
      if( fabs((*segIt).start.north-currentPoint.north)<.1 && fabs((*segIt).start.east-currentPoint.east)<.1 )
	leftPoint.push_back(&(*segIt));
      if( fabs((*segIt).end.north-currentPoint.north)<.1 && fabs((*segIt).end.east-currentPoint.east)<.1 )
	rightPoint.push_back(&(*segIt));
    }

  //search for all segments that contain p in their interior
  for( labelIt=labelseq.labels.begin(); labelIt!=labelseq.labels.end(); labelIt++ )
    {
      if( labelseq.isInterior( (*labelIt), currentPoint ) )
	  labelPoint.push_back((*labelIt));
    }

  // if have more then one segment, report intersection
  if((labelPoint.size() + rightPoint.size() + leftPoint.size()) > 1 )
    {
      firstFound = 0;
      //      printf( "line numbers intersect: " ); //debug
      for( labelIt=labelPoint.begin(); labelIt!=labelPoint.end(); labelIt++ )
	{
	  //	  printf( "%ld ", (*labelIt)->properties ); //debug
	  if( firstFound == 0 )
	    firstFound = (*labelIt)->properties;
	  else
	    if( firstFound != (*labelIt)->properties )
	      printf( "%f %f\n", currentPoint.north, currentPoint.east );
	}
      for( labelIt=rightPoint.begin(); labelIt!=rightPoint.end(); labelIt++ )
	{
	  //	  printf( "%ld ", (*labelIt)->properties ); //debug
	  if( firstFound == 0 )
	    firstFound = (*labelIt)->properties;
	  else
	    if( firstFound != (*labelIt)->properties )
	      printf( "%f %f\n", currentPoint.north, currentPoint.east );
	}
      for( labelIt=leftPoint.begin(); labelIt!=leftPoint.end(); labelIt++ )
	{
	  //	  printf( "%ld ", firstFound ); //debug
	  if( firstFound == 0 )
	    firstFound = (*labelIt)->properties;
	  else
	    if( firstFound != (*labelIt)->properties )
	      printf( "%f %f\n", currentPoint.north, currentPoint.east );
	}
      //      printf( " at %f %f\n", currentPoint.north, currentPoint.east ); //debug
    }

  // delete segments 
 for( labelIt=rightPoint.begin(); labelIt!=rightPoint.end(); labelIt++ )
    labelseq.seqDelete((*labelIt));
  for( labelIt=labelPoint.begin(); labelIt!=labelPoint.end(); labelIt++ )
    labelseq.seqDelete((*labelIt));

  // add segments
  for( labelIt=leftPoint.begin(); labelIt!=leftPoint.end(); labelIt++ )
    labelseq.seqInsert((*labelIt), currentPoint);
  for( labelIt=labelPoint.begin(); labelIt!=labelPoint.end(); labelIt++ )
    labelseq.seqInsert((*labelIt), currentPoint);

  // not yet working for vertical segments!!!!

  if( (leftPoint.size() + labelPoint.size()) == 0 )
    {
      findNewEvent( predecessor, successor, currentPoint );
    }
  else
    {
      for( labelIt=leftPoint.begin(); labelIt!=leftPoint.end(); labelIt++ )
	labelPoint.push_back((*labelIt));
      labelPoint.sort();

      // need to search for label in labelSeq
      found = false;
      for( labelIt=labelseq.labels.begin(); labelIt!=labelseq.labels.end(); labelIt++ )
	{
	  predecessor = (*labelIt);
	  for( unionIt=labelPoint.begin(); unionIt!=labelPoint.end(); unionIt++ )
	    {
	      current = (*unionIt);
	      if( current == predecessor )
		{
		  found = true;
		  break;
		}
	    }
	  if( found )
	    break;
	}
      labelseq.getNeighbors( &current, &predecessor, &successor );
      findNewEvent( predecessor, current, currentPoint );
      
      // need to search for label in labelSeq
      found = false;
      for( labelRIt=labelseq.labels.rbegin(); labelRIt!=labelseq.labels.rend(); labelRIt++ )
	{
	  predecessor = (*labelRIt);
	  for( unionIt=labelPoint.begin(); unionIt!=labelPoint.end(); unionIt++ )
	    {
	      current = (*unionIt);
	      if( current == predecessor )
		{
		  found = true;
		  break;
		}
	    }
	  if( found )
	    break;
	}

      labelseq.getNeighbors( &current, &predecessor, &successor );
      findNewEvent( successor, current, currentPoint );
    }
}

void Sweep::findNewEvent( LineSegment *upper, LineSegment *lower, Location2D &sweep )
{
  Location2D intersectPoint;
  list<Location2D>::iterator it;

  if( upper == NULL || lower == NULL )
    return;
  if( intersect(upper, lower) )
    {
      intersectPoint = labelseq.findIntersection( upper, lower );
      if( intersectPoint.north>sweep.north || (intersectPoint.north==sweep.north && intersectPoint.east<sweep.east) )
	{
	  //	  printf( "potential new point: %f %f\n", intersectPoint.north, intersectPoint.east );
	  pointList.push_front(intersectPoint);
	  pointList.sort();
	  pointList.unique();

	  /*
	  printf( "sorted point list:\n" );
	  for( it=pointList.begin(); it!=pointList.end(); it++ )
	    {
	      printf( "<%f %f>\n", (*it).north, (*it).east );
	    }
	  */

	}
    }
}



/*
  for( i=pointList.begin(); i!=pointList.end(); i++ )
    {
      label = (*i).myLine;
      myLine = (LineSegment*)label;
      if( (*i).isStart )
	{
	  sharePoint = labelseq.seqInsert(label); // labels intersect
	  labelseq.getNeighbors( label, predecessor, successor );
	  if( sharePoint )
	    {
	      intLocation = findIntersection( label, successor );
	      printf( "%f %f\n", intLocation.north, intLocation.east );
	    }
	  else if( intersect((LineSegment*)label, (LineSegment*)successor) )
	    {
	      intLocation = findIntersection( label, successor );
	      printf( "%f %f\n", intLocation.north, intLocation.east );
	    }
	  //	    return true;
	  else if( intersect((LineSegment*)label, (LineSegment*)predecessor) )
	    {
	      intLocation = findIntersection( label, predecessor );
	      printf( "%f %f\n", intLocation.north, intLocation.east );
	    }

	      //	    return true;
	}
      else
	{
	  if( intersect((LineSegment*)successor, (LineSegment*)predecessor) )
	    {
	      intLocation = findIntersection( successor, predecessor );
	      printf( "%f %f\n", intLocation.north, intLocation.east );
	    }

	    //	    return true;
	  labelseq.seqDelete(label);
	}
    }
  return false;
}
*/

Location2D LabelSeq::findIntersection( LineSegment *line1, LineSegment *line2 )
{
  double denominator, slope1; //, slope2;
  Location2D retValue;

  denominator = (line2->end.east - line2->start.east) *
    (line1->end.north - line1->start.north) -
    (line2->end.north - line2->start.north) *
    (line1->end.east - line1->start.east);
  slope1 = ( (line2->end.north - line2->start.north) *
	     (line1->start.east - line2->start.east) -
	     (line2->end.east - line2->start.east) *
	     (line1->start.north - line2->start.north) )/denominator;
  /*
  slope2 = ( (line1.end.north - line1.start.north) *
	     (line1.start.east - line2.start.east) -
	     (line1.end.east - line1.start.east) *
	     (line1.start.north - line2.start.north) );
  */

  // could use either slope1 or slope 2
  retValue.north = line1->start.north + slope1 * (line1->end.north - line1->start.north);
  retValue.east = line1->start.east + slope1 * (line1->end.east - line1->start.east);
  return retValue;
}

double LabelSeq::pointLineDistance( LineSegment *line, Location2D &point )
{
  return ( fabs((line->end.north - line->start.north)*(line->start.east - point.east) -
		(line->start.north - point.north)*(line->end.east - line->start.east))/
	   sqrt((line->end.north - line->start.north)*(line->end.north - line->start.north) +
		(line->end.east - line->start.east)*(line->end.east - line->start.east)) );
}


bool Sweep::intersect( LineSegment *l1, LineSegment *l2 )
{
  //  bool a, b, c, d;

  if( l1 == NULL || l2 == NULL )
    return false;

  /*
  a = labelseq.CCW(l1->start, l2->start, l2->end);
  b = labelseq.CCW(l1->end, l2->start, l2->end);
  c = labelseq.CCW(l1->start, l1->end, l2->start);
  d = labelseq.CCW(l1->start, l1->end, l2->end);
  */

  return( (labelseq.CCW(l1->start, l2->start, l2->end)!=labelseq.CCW(l1->end, l2->start, l2->end)) &&
	  (labelseq.CCW(l1->start, l1->end, l2->start)!=labelseq.CCW(l1->start, l1->end, l2->end)) );
}

/*
  Modification history:

  $Log: sweep.cc,v $
  Revision 1.3  2007/12/18 21:34:46  proctor
  In prim mob, read out the platform name from servo mob and used it to
  get platform-specific parameters V_CUTOFF_ANGLE et al; changed moast.ini
  accordingly; fixed a return value error in the segment sweep code, where
  anyIntersections doesn't really return anything

  Revision 1.2  2007/11/08 13:22:30  dr_steveb
  Not quite ready, but premature checking-in of other files means that these need to go to!

  Revision 1.1  2007/09/12 15:23:37  dr_steveb
  Added tools for generating sweep line for pologon world model

*/
