/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   virtSensorRead.cc
  \brief  Provide an example of reading from a sensor

  This task reads from a sensor channel and outputs the information.

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Stephen Balakirsky
  \date   May 13, 2005
*/
//! my name for connecting to the network
#define MY_NAME "sensorTest"
//! default nml config file
#define MY_CONFIG_FILE "../etc/moast.nml"
//! default sensor channel name
#define MY_SENSOR_CHAN_NAME "occupancyData1"
//! the navDataExt buffer to connect to
#define NAV_DATA_EXT_NAME "navDataExt"

#include <stdio.h> // sprintf
#include <math.h> // floor
#include <sys/types.h> // open
#include <sys/stat.h> // open
#include <fcntl.h> // open
#include <unistd.h> // write
#include <rcs.hh>
#include "moastNmlOffsets.hh"
#include "occupancyData.hh"
#include "navDataExt.hh"
#include "utmConvert.hh"

using namespace std;
using namespace libUtmConvert;

// predefines
bool growWorld( int newGraphWidth, int newGraphHeight, PM_CARTESIAN &newOffset );
bool growWorld( PM_CARTESIAN &includeLoc );

// globals
//! the world model
unsigned char **worldModel;
//! the height of the world model
int height;
//! the width of the world model
int  width;
//! The center of the world model
PM_CARTESIAN graphOffset;
//! scale factor of world model
double multFactor;
//! place to store data that we will write
char *datapt;
//! only output to file
bool outputFileOnly;

/***********************************************************************/
void dump( PM_CARTESIAN &vehLoc, char *fileName, bool closeOut )
{
  static int fileID = 0;
  int count, i, j, trash;
  float x,y;
  unsigned char red, blue, green;
  //  time_t timeOfDay;

  if( fileName == NULL )
    return;
  if( fileID == 0 )
    fileID = open( fileName, O_CREAT|O_WRONLY, 
		   S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  if( fileID <=0 )
    {
      fprintf( stderr, "Error %d from opening output file %s\n",
	       fileID, fileName );
      return;
    }
  if( write( fileID, "P6\n", strlen("P6\n") ) == -1 )
    {
      fileID = 0;
      return;
    }
  //  timeOfDay = time(NULL);
  //  sprintf( datapt, "# %s\n", ctime(&timeOfDay));
  /*
  if( write( fileID, datapt, strlen(datapt) ) == -1 )
    {
      fileID = 0;
      return;
    }
  */
  sprintf( datapt, "%d %d\n255\n", height, width );
  if( write( fileID, datapt, strlen(datapt) ) == -1 )
    {
      fileID = 0;
      return;
    }
  for( j=0; j<width; j++ )
    {
      for( i=0; i<height; i++ )
	{
	  count = (j*height+i)*3;
	  red = worldModel[i][j];
	  blue = worldModel[i][j];
	  green = worldModel[i][j];
	  //	      count = ((height - i - 1 )*width+j)*3;
	  if( worldModel[i][j] >= 254 )
	    {
	      red = 255;
	      green = 0;
	      blue = 0;
	    }
	  datapt[count] = red;
	  datapt[count+1] = green;
	  datapt[count+2] = blue;
	}
    }
  // draw vehicle location
  x = (float)vehLoc.x;
  y = (float)vehLoc.y;
  count = (int)((y*height +x)*3);
  if( count < width*height*3 )
    {
      datapt[count+1] = 255;
      datapt[count+2] = 0;
    }

  if( write( fileID, datapt, height*width*3 ) == -1)
    {
      fileID = 0;
      return;
    }
  if( closeOut )
    {
      close( fileID );
      fileID = 0;
    }
  else
    fsync( fileID );
}

/***********************************************************************/
// Routine to grow world model
// This routine will create a new larger world model and then copy the old one into it
// It returns true on success and false on failure
bool growWorld( PM_CARTESIAN &includeLoc )
{
  PM_CARTESIAN newOffset;
  int newGraphWidth, newGraphHeight;

  // determine the new lower left corner of the world 
  // (need to offset it if it is the includeLoc)
  if( includeLoc.x <= graphOffset.x  ) // below world
    {
      newOffset.x = (includeLoc.x - ((height / 10) /multFactor));
      newGraphHeight = height + (int)floor((graphOffset.x-newOffset.x)*multFactor);
    }
  else if( includeLoc.x >= graphOffset.x + height/multFactor ) // above world
    {
      newOffset.x = graphOffset.x;
      newGraphHeight = (int)(height*1.1) + 
	(int)floor((includeLoc.x - (graphOffset.x + height/multFactor))*multFactor);
    }
  else // everything OK
    {
      newOffset.x = graphOffset.x;
      newGraphHeight = height;
    }

  if( includeLoc.y <= graphOffset.y ) // left of world
    {
      newOffset.y = (includeLoc.y - ((width/2) /multFactor));
      newGraphWidth = width + (int)floor((graphOffset.y-newOffset.y)*multFactor);
    }
  else if( includeLoc.y >= graphOffset.y +width/multFactor ) // right of world
    {
      newOffset.y = graphOffset.y;
      newGraphWidth = (int)(width*1.1) + 
	(int)floor((includeLoc.y - (graphOffset.y + width/multFactor))*multFactor);
    }
  else // everything OK
    {
      newOffset.y = graphOffset.y;
      newGraphWidth = width;
    }
  if( newGraphWidth/multFactor > 1000 || newGraphHeight > 1000 ) // map wants to grow larger then 1000 meters!
    {
      fprintf( stderr, "Map wants to grow too large! <%d %d>\n", newGraphWidth/multFactor,
	       newGraphHeight/multFactor );
      return(false);
    }
  return( growWorld( newGraphWidth, newGraphHeight, newOffset ) );
}

bool growWorld( int newGraphWidth, int newGraphHeight, PM_CARTESIAN &newOffset )
{
  int count, count2;
  unsigned char **newWorldModel;
  int widthDiff, heightDiff;
  int oldWidth, oldHeight;

  //**************
  // check new bounds
  // first, new offset must be positioned to allow new map to contain old map
  if( graphOffset.x < newOffset.x || graphOffset.y < newOffset.y )
    {
      fprintf( stderr, "growWorld error on new offset position\n" );
      return( false );
    }
  // next, new world must be large enough to contain old world plus offset
  if( ((int)floor((graphOffset.x - newOffset.x)*multFactor + height)) > newGraphHeight ||
      ((int)floor((graphOffset.y - newOffset.y)*multFactor + width)) > newGraphWidth )
    {
      fprintf( stderr, "growWorld error on new map size\n" );
      return( false );
    }
  
  //*********************
  // allocate new world model
  newWorldModel = new (unsigned char*)[newGraphHeight];
  for( count=0; count<newGraphHeight; count++)
    {
      newWorldModel[count] = new (unsigned char)[newGraphWidth];
      if( newWorldModel[count]==NULL )
	return(false);
      for( count2=0; count2<newGraphWidth; count2++ )
	{
	  newWorldModel[count][count2] = 0;
	}
    }

  //******************
  // copy in old world model and delete space
  widthDiff = (int)floor((graphOffset.y - newOffset.y) * multFactor);
  heightDiff = (int)floor((graphOffset.x - newOffset.x) * multFactor);
  for( count=0; count<height; count++)
    {
      for( count2=0; count2<width; count2++ )
	{
	  newWorldModel[count+heightDiff][count2+widthDiff] =
	    worldModel[count][count2];
	}
      delete [] worldModel[count];
    }
  delete [] worldModel;

  //******************
  // reset offsets and widths
  delete [] datapt;

  oldWidth = width;
  oldHeight = height;
  width = newGraphWidth;
  height = newGraphHeight;
  graphOffset = newOffset;
  worldModel = newWorldModel;
  datapt = new char[height*width*3+500];
  return( true );
}


/***********************************************************************/
//! This returns the world model north index of any point whose map X coordinate is mapX.
/*!
  This is NOT checking that the returned value is in bounds.
  \param mapX Coordinate to convert
  \returns Converted location
*/
int mapXToWmNorth(double mapX)
{
  return ((int)floor((mapX - graphOffset.x) * multFactor));
}

/***********************************************************************/
//! This returns the world model east index of any point whose map Y coordinate is mapY.
/*!
  The world model array is best thought of as a grid lying on top
  of the offset local map. The first index is north and the second
  is east.

  This is NOT checking that the returned value is in bounds.
  \param mapX Coordinate to convert
  \returns Converted location
*/
int mapYToWmEast(double mapY)
{
  return ((int)floor((mapY - graphOffset.y) * multFactor));
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! update the world model with new information
/*!
  \param dataList The new information
*/
void updateWorld( OccupancyDataList *dataList, bool myVerbose )
{
  int count;
  int northLoc, eastLoc;
  utmConvert converter;
  double Lat, Long;
  char UTMZone[4] = "18S";
  int RefEllipsoid = 23;

  for( count=0; count<dataList->featCell_length; count++ )
    {
      if( myVerbose || outputFileOnly )
	{
	  converter.UTMtoLL( RefEllipsoid, dataList->featCell[count].x,
			     dataList->featCell[count].y, UTMZone, Lat, Long );
	  fprintf( stdout, "%lf, %lf\n", Long, Lat );
	  fflush( stdout );
	}
      //		 dataList->featCell[count].z );
      // decode cell with regard to location
      if( outputFileOnly )
	continue;
      northLoc = mapXToWmNorth(dataList->featCell[count].x); 
      eastLoc = mapYToWmEast(dataList->featCell[count].y); 

      // if necessary, expand world to accomodate new obstacle
      if( northLoc<0 || northLoc>=height || 
	  eastLoc<0 || eastLoc>=width )
	{
	  if( !growWorld( dataList->featCell[count] ) )
	    continue;
	  northLoc = mapXToWmNorth(dataList->featCell[count].x); 
	  eastLoc = mapYToWmEast(dataList->featCell[count].y); 
	}
      worldModel[northLoc][eastLoc] = (unsigned char)dataList->featCell[count].z;
    }
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! Place the vehicle in the center of the map
/*!
  \param vehLocation The location of the vehicle
 */
void setGraphOffset( PM_CARTESIAN &vehLocation )
{
  graphOffset.x = (vehLocation.x - ((height / 2) / multFactor));
  graphOffset.y = (vehLocation.y - ((width / 2) / multFactor));
  graphOffset.z = vehLocation.z;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! update the position of the sensor via navDataExt buffer
/*!
  This routine reads the navigation solution from the tranTrue and returns it.
  If it is unable to read the channel, then the currentLoc location is returned.
  \param chan The channel to read navDataExt from
  \param currentLoc The robot's current location
  \return Returns the robot location in world model coordinates
*/
PM_CARTESIAN updatePosition( NML *chan, PM_CARTESIAN &currentLoc )
{
  int type;
  NavDataExt *navdata;
  PM_CARTESIAN retValue;
  static bool firstTime = true;

  type = chan->read();
  switch( type )
    {
    case 0:
      retValue = currentLoc;
      break;
    case NAV_DATA_EXT_TYPE:
      navdata = (NavDataExt*)chan->get_address();
      // set graph offset
      if( firstTime )
	{
	  setGraphOffset( navdata->tranTrue );
	  firstTime = false;
	}
      // decode cell with regard to location
      retValue.x = mapXToWmNorth(navdata->tranTrue.x); 
      retValue.y = mapYToWmEast(navdata->tranTrue.y); 
      retValue.z = navdata->tranTrue.z;

      // if necessary, expand world to accomodate new obstacle
      if( retValue.x<0 || retValue.x>=height || 
	  retValue.y<0 || retValue.y>=width )
	{
	  if( !growWorld( navdata->tranTrue ) )
	    {
	      fprintf(stderr, "Unable to grow world!\n" );
	      retValue.x = 0;
	      retValue.y = 0;
	      return retValue;
	    }
	  retValue.x = mapXToWmNorth(navdata->tranTrue.x); 
	  retValue.y = mapYToWmEast(navdata->tranTrue.y); 
	}
      break;
    default:
      retValue = currentLoc;
      break;
    }
  return retValue;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//! Main program loop
/*!
  \param argc The number of arguments received
  \param argv The actual arguments
*/
int main(int argc, char *argv[])
{
  char c;
  char *nmlConfig = NULL;
  char *sensorChanName = NULL;
  char *myNmlName = NULL;
  char *navDataName = NULL;
  NML *sensorChan, *navDataChan;
  int type;
  OccupancyDataList *dataList;
  OccupancyDataMap  *dataMap;
  RCS_TIMER timer(.250);
  PM_CARTESIAN vehLoc;
  int count, count2;

  // initialize stuff
  outputFileOnly = false;
  graphOffset.z = 0;
  multFactor = 1/.1; // where .1 is cell resolution
  vehLoc.x = 0;
  vehLoc.y = 0;
  vehLoc.z = 0;
  height = 200;
  width = 200;
  graphOffset.x = height/2;
  graphOffset.y = width/2;
  datapt = new char[height*width*3+500];
  worldModel = new (unsigned char*)[height];
  for( count=0; count<height; count++ )
    {
      worldModel[count] = new (unsigned char)[width];
      for( count2=0; count2<width; count2++ )
	worldModel[count][count2] = 0;
    }


  if( nmlConfig == NULL )
    {
      nmlConfig = new char[80];
      sprintf( nmlConfig, "%s", MY_CONFIG_FILE );
    }

  if( sensorChanName == NULL )
    {
      sensorChanName = new char[80];
      sprintf( sensorChanName, "%s", MY_SENSOR_CHAN_NAME );
    }
  if( myNmlName == NULL )
    {
      myNmlName = new char[80];
      sprintf( myNmlName, "%s", MY_NAME );
    }
  if( navDataName == NULL )
    {
      navDataName = new char[80];
      sprintf( navDataName, "%s", NAV_DATA_EXT_NAME );
    }
  sensorChan = new NML( occupancyData_format, sensorChanName, myNmlName, nmlConfig );
  if( sensorChan==NULL || !sensorChan->valid() )
    {
      fprintf(stderr, "Unable to open sensor channel\n" );
      return(-1);
    }
  navDataChan = new NML( navDataExt_format, navDataName, myNmlName, nmlConfig );
  if( navDataChan==NULL || !navDataChan->valid() )
    {
      fprintf(stderr, "Unable to open navDataExt channel\n" );
      return(-1);
    }

  for(;;)
    {
      if( !outputFileOnly )
	vehLoc = updatePosition( navDataChan, vehLoc );
      type = sensorChan->read();
      switch(type)
	{
	case 0:
	  break;
	case OCCUPANCY_DATA_LIST_TYPE:
	  dataList = (OccupancyDataList*)sensorChan->get_address();
	  fprintf( stderr, "Received %d ", dataList->featCell_length );
	  if( dataList->isGlobal )
	    fprintf( stderr, "global " );
	  else
	    fprintf( stderr, "local " );
	  fprintf( stderr, "cells of resolution %lf\n", dataList->cellResolution );
	  updateWorld( dataList, true );
	  /*
	  for( type=0; type<dataList->featCell_length; type++ )
	    {
	      fprintf(stderr, " <%lf %lf> %lf\n",
			 dataList->featCell[type].x,
			 dataList->featCell[type].y,
			 dataList->featCell[type].z );
	    }
	  */
	  break;
	case OCCUPANCY_DATA_MAP_TYPE:
	  dataMap = (OccupancyDataMap*)sensorChan->get_address();
	  break;
	}
      if( !outputFileOnly )
	dump( vehLoc, "testOut.ppm", false );
      timer.wait();
    }
}

/*
  $Log: virtSensorRead.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.2  2005/06/29 21:08:14  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.1  2005/05/24 13:32:14  stephen
  First check-in of code
*/
