/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehMobMap.cc
  \brief  scrolling map for vehicle planner
  This map is based on the scrolling map contained in scrollingMap.hh


  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.6 $
  $Date: 2007/05/16 15:19:04 $
  \endcode

  \author Stephen Balakirsky
  \date   09/20/2006
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>	   // round()
#include <string.h>	   // strlen()
#include <fcntl.h>	   // open()
#include <sys/stat.h>      // S_IWUSR, etc.
#include <unistd.h>        // close()
#include "config.h"        // make HAVE_CYGWIN flag work
#include "vehMobMap.hh"    // header file
#include "../common/scrollingMap.cc" // code for scrolling map using VehMobCell


/***********************************************************************/

/*! VehMobMap::VehMobMap constructor taking two arguments creates a map
  based on an initial range, resolution. Total size is rangeIn*2 +1 per side
*/
VehMobMap::VehMobMap( int rangeIn, double resolutionIn ):ScrollingMap(rangeIn, resolutionIn)
{
  clearMap();
}

/*! return the location of a cell
 */
PM_CARTESIAN VehMobMap::graphToMap( int nodeIn )
{
  PM_CARTESIAN returnValue;

  returnValue.x = getXCoordAtId(nodeIn);
  returnValue.y = getYCoordAtId(nodeIn);
  return( returnValue );
}

void VehMobMap::clearMap()
{
  int count;
  VehMobCell *cellPt;
  for( count=0; count<getRows()*getCols(); count++ )
    {
      cellPt = getCellWithId(count);
      cellPt->init();
    }
}

VehMobMap *VehMobMap::copy( int newRange )
{
  double xCount, yCount;
  VehMobMap *mapReturn;
  VehMobCell *cellNew, *cellOld;
  int xDiff, yDiff;

  if( newRange < getRange() )
    {
      fprintf(stderr, "VehMobMap: trying to copy a map to too small of a map\n");
      return NULL;
    }
  mapReturn = new VehMobMap( newRange, getResolution() );
  if( mapReturn == NULL )
    {
      fprintf(stderr, "VehMobMap: unable to copy map\n" );
      return NULL;
    }
  
  mapReturn->setCenter( getCenterX(), getCenterY(), &xDiff, &yDiff );
  //  mapReturn->setCenter( vehLoc.x, vehLoc.y );
  for( xCount = getXLow()-getResolution()/2.; xCount<getXHi()+getResolution()/2.; xCount+=getResolution() )
    {
      for( yCount = getYLow()-getResolution()/2.; yCount<getYHi()+getResolution()/2.; yCount+=getResolution() )
	{
	  cellNew = mapReturn->getCellAtCoords( xCount, yCount );
	  cellOld = getCellAtCoords( xCount, yCount );
	  if( cellNew!=NULL && cellOld!=NULL )
	    cellNew->setCellData(cellOld->obstacleMax, cellOld->obstacleMin,
				 cellOld->grow, 
				 cellOld->hazard, cellOld->lastSeen,
				 cellOld->mark, cellOld->isCleared);
	}
    }
  return( mapReturn );
}

VehMobMap::~VehMobMap()
{
}

/***********************************************************************/
unsigned char VehMobMap::getMapObstacleMin( /* ARGUMENTS                */
 int row,                 /*!< row index of cell      */
 int col)                 /*!< column index of cell   */
{
  VehMobCell *cellPt;
  if ((row > -1) && (row < getRows()) && (col > -1) && (col < getCols()))
    {
      cellPt = getCellAtIndexes(row, col);
      return cellPt->obstacleMin;
    }
  return 0;
}

/*! CVS Log Information
 * $Log: vehMobMap.cc,v $
 * Revision 1.6  2007/05/16 15:19:04  dr_steveb
 * Trying to fix graph wrapping. Not fixed yet, however, fixed compile issue with rotate command.
 *
 * Revision 1.5  2006/10/18 00:18:21  dr_steveb
 * Added victim channel to vehicle level. It now hunts for victims. Also added CImg library from sourceforge to perform annotation. Modified make files to include this library.
 *
 * Revision 1.4  2006/10/13 16:50:19  tomrkramer
 * Changed two lines to call new version of scrollingMap::setCenter correctly.
 *
 * Revision 1.3  2006/10/10 13:30:18  dr_steveb
 * New vehicle level exploration based on camera FOV implemented.
 *
 * Revision 1.2  2006/10/03 20:24:25  dr_steveb
 * Exploration command added to vehicle level
 *
 * Revision 1.1  2006/09/30 16:05:30  dr_steveb
 * New world model files
 *
 * 
 * 
 */
