/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehMobPlWm.cc
  \brief Provides a world model for use by the vehicle level planner

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Stephen Balakirsky
  \date   05/25/2005
*/
#include <stdio.h> // sprintf, fprintf, stdout
#include "vehMobPlWm.hh"
#include "search.h"

VehMobWmNode::VehMobWmNode( int numVerts, Location2D *verticesIn, int idIn )
{
  numVertices = numVerts;
  vertices = new Location2D[numVertices];
  for( int count=0; count<numVertices; count++ )
    vertices[count] = verticesIn[count];
  next = NULL;
  id = idIn;
}

VehMobWmNode::~VehMobWmNode()
{
  if( numVertices != 0 )
    delete [] vertices;
}

LineSegment VehMobWmNode::getLine(int id)
{
  if( id >= numVertices )
    return LineSegment(0,0,0,0);
  if( id < numVertices-1 )
    return LineSegment(vertices[id], vertices[id+1]);
  return LineSegment(vertices[id], vertices[0]);
}

//**************************************************************************************
VehMobPlWm::VehMobPlWm( bool myVerboseIn )
{
  myVerbose = myVerboseIn;
  wmNodes = NULL;
  hcreate( 100 ); // default hash table to start with 100 entries
  wmNodes = new VehMobWmNode(0, NULL, -1 );
  tail = wmNodes;
}

VehMobPlWm::~VehMobPlWm()
{
  VehMobWmNode *next = wmNodes;
  VehMobWmNode *temp;
  
  while( next != NULL )
    {
      temp = next->getNext();
      delete next;
      next = temp;
    }
  hdestroy();
}

/*! Adds a node to the world model
   /param numVerts      Number of vertices in object
   /param verticesIn    The location of each vertice
   /param idIn          The id of this object

   /return              true on sucess
                        false on failure
*/
bool VehMobPlWm::add(int numVerts, Location2D *verticesIn, int idIn)
{
  char key[10];
  ENTRY entry;

  tail->setNext( new VehMobWmNode(numVerts, verticesIn, idIn) );
  tail = tail->getNext();
  sprintf( key, "%d", tail->getId());
  entry.key = key;
  entry.data = (char *) tail;
  hsearch( entry, ENTER );

  return true;
}

void VehMobPlWm::dump()
{
  int count;
  VehMobWmNode *next = wmNodes;
  LineSegment line;
  
  while( next != NULL )
    {
      for( count=0; count<next->getNumVertices()-1; count++ )
	{
	  line = next->getLine(count);
	  fprintf( stdout, "L %d %d %d %d\n",
		   (int)line.start.north, 
		   (int)line.start.east,
		   (int)line.end.north,
		   (int)line.end.east );
	}
      next = next->getNext();
    }
  
}

/*
  Modification history:

  $Log: vehMobPlWm.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.5  2005/09/30 19:26:22  proctor
  Fixed warnings with %lf used with printf, it's for scanf only

  Revision 1.4  2005/09/30 18:55:58  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.3  2005/07/27 12:34:59  stephen
  Fixes to get code to compile.

  Revision 1.2  2005/07/26 20:30:37  stephen
  Debug in progress.

  Revision 1.1  2005/07/26 13:25:12  stephen
  Added visibility graph planner (not yet working)


*/
