#!/usr/bin/python

## ****************************************************************************
##  DISCLAIMER:
##  This software was produced by the National Institute of Standards
##  and Technology (NIST), an agency of the U.S. government, and by statute is
##  not subject to copyright in the United States.  Recipients of this software
##  assume all responsibility associated with its operation, modification,
##  maintenance, and subsequent redistribution.
##
##  See NIST Administration Manual 4.09.07 b and Appendix I. 
## ****************************************************************************
## ****************************************************************************
## \file   USARSimDataCapture.py
## \brief  This application is a tool to help you make referenced data
##         sets by capturing 3D scans of the environment at specified poses
##
## \code CVS Status:
## $Author: cj_scrapper $
## $Revision: 1.2 $
## $Date: 2008/01/18 01:19:01 $
## \endcode
## \author Chris Scrapper
## \date   01/17/2008
## ****************************************************************************

from socket import *
from Numeric import *
import sys,string,os

## ****************************************************************************
## Configuration Parameters -
## See documentation in MOAST's reference manual.
## ****************************************************************************
## String - name of machine running USARSim
HOST_NAME="localhost"
## Integer - port number for socket connecting to USARSim
PORT_NUM=3000
## String - input file name
INPUT_FILE="groundTruth"
## String - format for names of output file
FILENAME_FORMAT="scan_%3d"
## String - path where data files will be written
DATA_DIR="./data"
## double - Vertical field-of-view, [degs]
VERT_FOV=20.
## double - angular separation between scans, [degs]
VERT_INC=5.
## ****************************************************************************






## ****************************************************************************
## Application Source -
## Don't change anything below this line
## ****************************************************************************
def DEG_TO_RADS (x): return x*pi/180.

def compute_fov_intervals(fov,inc):
    return DEG_TO_RADS(arange(-fov/2.,fov/2.+1,inc))
        
class USARSimDataCollection:
    
    def __init__(self):
       
        try:
            self.host       = HOST_NAME
            self.port       = PORT_NUM
            self.input_file = open( INPUT_FILE )
            self.output_format = FILENAME_FORMAT
            self.data_dir   = DATA_DIR
            self.scan_elev  = compute_fov_intervals(VERT_FOV, VERT_INC)
        except:
            print "*** ERROR ***"
            print "<USARSimDataCollection:__init__>"
            print sys.exc_info()[0]
            print sys.exc_info()[1]
            sys.exit();
        
    
    def get_message(self):
        msg_end = False
        data=""
        while not msg_end:
            data += self.sock.recv(512)
            if(string.find(data,"\r\n") != -1):
                msg_end=True
        messages = string.replace(data,"\r","").split("\n")
        return messages

    
    def find_data_message(self):
        while True:
            messages = self.get_message()
            for msg in messages:
                if(string.find(msg,"RangeScanner") != -1):
                    return msg

    def pTuple(self,key,data):
        sPtr=string.find(data,key)+len(key)
        ePtr=string.find(data[sPtr:],"}")+sPtr
        return string.replace(data[sPtr:ePtr],","," ")

    def parseSickData(self,data):
        data_str=""
        #data_str = self.pTuple("Time",data)
        data_str += self.pTuple("MinRange",data)
        data_str += self.pTuple("MaxRange",data)
        data_str += self.pTuple("Location",data)
        data_str += self.pTuple("Rotation",data)
        data_str += self.pTuple("FOV",data)
        data_str += self.pTuple("Resolution",data)
        data_str += self.pTuple("ScanCount",data)
        data_str += self.pTuple("{Range",data)
        return string.lstrip(data_str)

    def do_single_step(self,pose):

        #create a TCP socket
        self.sock = socket(AF_INET, SOCK_STREAM)
        # connect to server on the port
        self.sock.connect((self.host, self.port))
        # read socket for NFO message
        data = self.sock.recv(512)
        
        # Initialize data collection object at pose
        str = "INIT {ClassName USARBot.DataCollectionBase}"
        str = str + ("{Location %lf,%lf,%lf}")%(pose[0],pose[1],pose[2])
        str = str + ("{Rotation %lf,%lf,%lf}")%(pose[3],pose[4],pose[5])
        self.sock.send(str)
        self.sock.send("\r\n")

        # Get message and parse it
        msg = self.find_data_message()
        self.sock.close()
        output_str = self.parseSickData(msg)

        # return data set string
        return output_str
          
        
        
    def do_data_collection(self):
        # check to make sure the output directory is there
        if not os.path.isdir(self.data_dir):
            os.mkdir(self.data_dir)
        
            
        ## WHERE AM I : add coutn for file intext
        file_count = 1 
        for line in self.input_file:

            # Ignore comments
            if line[0] == "#" : continue
            
            #parse data
            input_data = map(string.atof,line.split())

            # Verify format -> [ x, y, z, r, p, y]
            if len(input_data) != 6: continue

            #Separate timestamp and pose
            pose = array(input_data)
            
            # Open output file that will contain one 3D scan
            outfile_name =((self.output_format)%(file_count)).replace(" ","0")
            print "Writing output file: "+ outfile_name
            outfile = open( self.data_dir+"/"+outfile_name,"w")
            outfile.write("## DATA FORMAT\n")
            outfile.write("## MinRangeMaxRange Location<x y z> Rotation<r p y> FOV<rads> Resolution<rads> ScanCount RangeData\n")

            print self.scan_elev
            # simulate 3D scan
            for x in self.scan_elev:
                scan_pose = pose + array([0, 0 ,0 ,0, 0, 1]) * x
                output_str = self.do_single_step(scan_pose)
                outfile.write(output_str)

            file_count = file_count + 1
            

if __name__ == "__main__":
    dc = USARSimDataCollection()
    dc.do_data_collection()
