/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file amMisJA.hh

  \brief NML declarations for autonomous mobility (AM) level mission package
  that generates sequences of point-to-point moves for the mission package,
  where points are the ending angular position of mission package.

  Synopsis:
  #include <rcs.hh>
  #include <posemath.h>
  #include "moastNmlOffsets.hh"
  
  \code CVS Status:
  $Author: proctor $
  $Revision: 1.2 $
  $Date: 2006/06/21 19:38:13 $
  \endcode

  \author Fred Proctor
*/

#ifndef AM_MIS_JA_HH
#define AM_MIS_JA_HH

#include "moastNmlOffsets.hh"

#define AM_MIS_JA_CMD_NAME "amMisJACmd"
#define AM_MIS_JA_STAT_NAME "amMisJAStat"
#define AM_MIS_JA_CFG_NAME "amMisJACfg"
#define AM_MIS_JA_SET_NAME "amMisJASet"

#define AM_MIS_JA_CMD_BASE        (AM_MIS_JA_BASE)
#define AM_MIS_JA_STAT_BASE       (AM_MIS_JA_BASE + 100)
#define AM_MIS_JA_CFG_BASE        (AM_MIS_JA_BASE + 200)
#define AM_MIS_JA_SET_BASE        (AM_MIS_JA_BASE + 300)

#define AM_MIS_JA_CMD_INIT_TYPE     (AM_MIS_JA_CMD_BASE + 1)
#define AM_MIS_JA_CMD_ABORT_TYPE    (AM_MIS_JA_CMD_BASE + 2)
#define AM_MIS_JA_CMD_HALT_TYPE     (AM_MIS_JA_CMD_BASE + 3)
#define AM_MIS_JA_CMD_SHUTDOWN_TYPE (AM_MIS_JA_CMD_BASE + 4)

#define AM_MIS_JA_STAT_TYPE       (AM_MIS_JA_STAT_BASE + 1)

#define AM_MIS_JA_CFG_CYCLE_TIME_TYPE (AM_MIS_JA_CFG_BASE + 1)
#define AM_MIS_JA_CFG_DEBUG_TYPE      (AM_MIS_JA_CFG_BASE + 2)

#define AM_MIS_JA_SET_TYPE        (AM_MIS_JA_SET_BASE + 1)

// AmMisJACmdXXX

class AmMisJACmdInit:public RCS_CMD_MSG {
public:
  AmMisJACmdInit():RCS_CMD_MSG(AM_MIS_JA_CMD_INIT_TYPE,
			       sizeof(AmMisJACmdInit)) {};
  void update(CMS *);
};

class AmMisJACmdAbort:public RCS_CMD_MSG {
public:
  AmMisJACmdAbort():RCS_CMD_MSG(AM_MIS_JA_CMD_ABORT_TYPE,
				sizeof(AmMisJACmdAbort)) {};
  void update(CMS *);
};

class AmMisJACmdHalt:public RCS_CMD_MSG {
public:
  AmMisJACmdHalt():RCS_CMD_MSG(AM_MIS_JA_CMD_HALT_TYPE,
			       sizeof(AmMisJACmdHalt)) {};
  void update(CMS *);
};

class AmMisJACmdShutdown:public RCS_CMD_MSG {
public:
  AmMisJACmdShutdown():RCS_CMD_MSG(AM_MIS_JA_CMD_SHUTDOWN_TYPE,
				   sizeof(AmMisJACmdShutdown)) {};
  void update(CMS *);
};

/*
  AmMisJAStat is the class that holds dynamic mission package status
*/
class AmMisJAStat : public RCS_STAT_MSG {
public:
  AmMisJAStat() : RCS_STAT_MSG(AM_MIS_JA_STAT_TYPE, sizeof(AmMisJAStat)) {};
  void update(CMS *);

  unsigned int heartbeat;
};

/*
  AmMisJACfgCycleTime sets the controller's cycle time
*/
class AmMisJACfgCycleTime:public RCS_CMD_MSG {
public:
  AmMisJACfgCycleTime():RCS_CMD_MSG(AM_MIS_JA_CFG_CYCLE_TIME_TYPE,
			       sizeof(AmMisJACfgCycleTime)) {};
  void update(CMS *);

  //! period in seconds
  double cycleTime;
};

/*
  AmMisJACfgDebug sets the controller's debug level
*/
class AmMisJACfgDebug:public RCS_CMD_MSG {
public:
  AmMisJACfgDebug():RCS_CMD_MSG(AM_MIS_JA_CFG_DEBUG_TYPE,
			       sizeof(AmMisJACfgDebug)) {};
  void update(CMS *);

  //! debug level
  int debug;
};

/*
  AmMisJASet is the class that holds the quasi-static mission package
  settings, those that change infrequently and in response to sensor 
  processing configuration messages.
*/
class AmMisJASet : public RCS_STAT_MSG {
public:
  AmMisJASet():RCS_STAT_MSG(AM_MIS_JA_SET_TYPE, sizeof(AmMisJASet)) {};
  void update(CMS *);

  double cycleTime;
  unsigned int debug;
};

extern int amMisJA_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * amMisJA_symbol_lookup(long type);

#endif // AM_MIS_JA_HH

/*
  Modification history:

  $Log: amMisJA.hh,v $
  Revision 1.2  2006/06/21 19:38:13  proctor
  Added support for Java, which involved including some left-out headers
  like moastNmlOffsets.hh and moastTypes.hh

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.2  2005/06/08 14:29:53  proctor
  Added vehMobPLMain name change

  Revision 1.1  2005/06/08 12:47:39  proctor
  Name change per Mis,MobJA naming convention

  Revision 1.2  2005/06/07 21:45:11  proctor
  Changed servoPL to servoMobJA

  Revision 1.1  2005/06/07 15:54:01  proctor
  Added amMis buffers

*/
