/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*!
  \file dijkTabs.hh
  \brief defines classes used by dijkTabs graph searcher

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Tom Kramer

This is used by dijkTabs.cc.

The dijkTabs.hh and dijkTabs.cc files define an abstract graph class
(DijkTabsGraph) and a Dijkstra graph search class (DijkTabsSearcher)
with the following characteristics:

1. The graph may be built or changed dynamically during the search
   in several ways using the addKids, removeKids, and/or useKids
   functions.
2. The user can add information to graph nodes by making a
   child class of nodeInfo.
3. The open nodes list is managed using the tabs method.
4. Arcs with a negative cost between nodes are allowed and treated
   as though there were no arc. This is so that graphs can be handled in
   which every node has the same apparent number of children, but some
   are not actually children. Such graphs are often convenient to use.

This searcher does NOT implement searching for multiple goals simultaneously.

The start must not be the same as the goal. It is up to the caller to
check this.

A driver, exampleDijkTabs.cc, for making a stand-alone executable is
in src/tools.  Test files for testing the executable exist but are not
currently part of MOAST.

*/

#ifndef DIJK_TABS_HH
#define DIJK_TABS_HH

#include "nodeInfo.hh"

class DijkTabsSearcher;
class DijkTab;
class DijkTabsGraph;
class DijkTabsNode;

/***********************************************************************/

/*! \class DijkTabsSearcher
    \brief provides a searcher class used in Dijkstra search with tabs

The DijkTabsSearcher uses a doubly linked list of nodes as the list of
open nodes. The links are formed by the "more" and "less" attributes
of DijkTabsNode, which identify the index in nodes of the next more
expensive and next cheapest node, respectively. The index in nodes of
the cheap end of the list is kept in the cheap attribute of the
DijkTabsSearcher. The index in nodes of the expensive end of the list
is kept in the expensive attribute of the DijkTabsSearcher. When the
list is empty, cheap and expensive are set to -1.

*/

class DijkTabsSearcher
{
public:
  DijkTabsSearcher(){ cheap = -1; expensive = -1; }
  ~DijkTabsSearcher(){}

  int closeSearch(); /*! closes the searcher */

  int cycleSearch(); /*! finds cheapest node, removes it, processes it */

  int getNextNode(); /*! returns the index of cheapest open node remaining */

  int initSearch(          /*!  initializes the searcher                   */
   DijkTabsGraph * aGraph, /*!< the graph to search in                     */
   int start,              /*!< index in nodes of aGraph of start node     */
   int finish,             /*!< index in nodes of aGraph of finish node    */
   int * opensOut,         /*!< 0=open list empty, non-zero=list not empty */
   int * statusOut,        /*!< status of reaching goal                    */
   double * costOut);      /*!< cost of reaching goal                      */

  int reportSearch( /*! writes the path to the goal into an array         */
   int maxLength,   /*!< maximum length of path (capacity of path array)  */
   int * path);

private:
  int bestSize;             /* best size for number of tabs                  */
  int cheap;                /* index of node at cheap end of open nodes list */
  DijkTab * cheapTab;       /* first tab (at cheap end of tab list)          */
  double * cost;            /* array (size 1) for cost of reaching goal      */
  int expensive;            /* index of node at other end of open nodes list */
  int goal;                 /* index in nodes of goal (finish) node          */
  DijkTabsNode * nodes;     /* array of nodes from theGraph                  */
  int numOpen;              /* length of open nodes list                     */
  int * numOpenOut;         /* length of open nodes list for caller          */
  int * status;             /* array (size 1) for status flag                */
  DijkTabsGraph * theGraph; /* the graph to be searched                      */

  void addTab(DijkTab * nextTab);
  void deleTabDown(DijkTab * aTab, DijkTab * downTab);
  void deleTabUp(DijkTab * aTab, DijkTab * upTab, DijkTab * downTab);
  void insertNode(int inNode, double cost);
  void insertNode0(int inNode, double cost);
  void insertNode1(int inNode, double cost);
  void insertNode2(int inNode, double cost);
  void moveOrDeleTab(DijkTab * aTab);
  void moveTabDown(DijkTab * aTab, DijkTab * downTab);
  void moveTabUp(DijkTab * aTab, DijkTab * upTab);
  void processNode(int numNode, int from, double cost);
  void removeCheap();
  void removeNode(int outNode);
};

/***********************************************************************/

/*! \class DijkTab
    \brief provides the tab class used in Dijkstra search with tabs

1. GENERAL IDEA

The idea of a tab is best explained by thinking of the open list of
nodes as a card file and the tabs as tabs attached to some cards. Each
node is represented by one card. The card file is kept in increasing
order of cost. Each tab has written on it the cost of the card to
which it is attached. Each card is attached by a wire to the
following tab.

For ideal searching, there are fewer tabs than cards. There is always
a tab on the last card. There is no tab on the first card unless there
is only one card. The DijkTabsSearcher keeps a pointer to the first card and a
separate pointer to the first tab. Tabs are added and removed as
necessary occasionally in connection with adding, moving, or removing
cards. A count is kept of the total number of cards in the file; this
is used in deciding when to add or delete tabs.

Without the tabs, it is possible only to go through the card file one card
at a time. With the tabs, it is possible to skip from tab to tab and to
skip from a card to the next tab. Both the node list and the tab list are
doubly linked, so it is possible to go backwards and forwards in both
the cards and the tabs.


2. OPERATIONS ON CARD AND TABS

There are three operations on cards:
a. Insert a card not already in the file.
b. Remove a card from the file and reinsert it towards the cheap end.
c. Remove the first card in the file.

Operations on tabs are done in conjunction with operations on cards.

a. Insert New Card
------------------

If there are two or more tabs, a card not already in the file is
inserted into the file by looking through the tabs for the first tab
that is of higher or equal cost, jumping to the preceding tab (or
to the beginning of the file if the tab is the first tab), and
looking through the cards after that until the first card of
higher or equal cost is found. The new card is inserted just before
that (including attaching the wire from the card to the tab). When
this is done, the card count of the tab is increased by one, and a
check is made of whether a new tab should be inserted between the
preceding tab and this one; if so, a new tab is inserted. If there is
no tab of higher or equal cost, the last tab is removed from the last
card, the last tab is attached to the new card, and the new card is
inserted at the end of the file. The total card count is increased by
one.

If there is only one tab, inserting a new card is done as described
above, starting on the fourth line.

If there is no tab, there is also no card, so insert the card at
the beginning and attach a tab to it.

Because this method for inserting a card is used, the number of tabs
for N cards that will minimize the average number of comparisons
required to insert a card is sqrt(N). With this number of tabs, the
average number of cards between tabs is also sqrt(N), so the average
number of comparisons that must be made to insert a card is half the
number of tabs plus half the number of cards between tabs, which is
((sqrt(N)/2) + (sqrt(N)/2)) = sqrt(N). Without the tabs, the average
number of comparisons that must be made is N/2. If, for example, there
are 10000 cards, 100 comparisons are needed with tabs, 5000 without,
a difference of a factor of 50. However, there is more overhead
required for using tabs, so the difference is not so great in practice.
The methods described here for handling the tabs are intended to
minimize this overhead.

b. Remove and Reinsert Card
---------------------------

It is not necessary to search for cards already in the file. Such
cards may be encountered as children of a node being processed,
however, and may have to be relocated toward the cheap end of the
file. A card is relocated toward the cheap end of the file as follows.

First the card is removed. If the card is the first card, this is
done by the removeCheap function. Otherwise, it is done by the
removeNode function. See the documentation of those functions for
details.

Second, the card is reinserted as described above, except that the
total card count is not changed.

c. Remove First Card
--------------------

This is done in the removeCheap function. See the documentation of
removeCheap for how it is done. If removeCheap is called from dijk,
after removeCheap runs, dijk reduces numOpen (the total card count) by
one.


3. MAINTAINING OPTIMAL NUMBER OF TABS

As mentioned above, the optimal number of tabs is sqrt(N), where N is
the total number of cards. Optimally, the tabs should be equally spaced.
The optimal number changes as the number of cards changes (typically it
increases then decreases). The spacing is affected by the three operations.
It has to be decided when to add and delete tabs.

A tab is added whenever the number of cards between two tabs is twice
(or more than twice) the optimum. This is done by making a new tab,
going to the preceding tab of the one that has too many cards before
it, counting off half the in-between cards (while resetting the wires
on those cards to the new tab) and attaching the new tab to the last
of these cards. The tab counts on the new tab and old tab are set. It
is not clear whether this should be done at 2 times the optimum or
something between 1 and 2 times the optimum.

A tab will be deleted or moved whenever the number of cards between
two tabs is half the optimum or less. See the documentation of
moveOrDeleTab for how this is done.

4. NOTES

The visitCost attribute of a tab is not logically needed, since it is
the cost of the tab's myNode, but search time is saved by copying the
cost from myNode into the visitCost of the tab.

The wire from the card to the next tab is needed so that the size of
the next tab can be decremented when a card is relocated toward the
cheap end of the file. Without the wire, it would be necessary to
go through cards until the next one with a tab attached was found.

*/

class DijkTab
{
public:
  DijkTab(){}
  ~DijkTab(){}

  int myNode;       /*!< index in nodes of node to which tab is attached    */
  DijkTab * down;   /*!< next cheaper tab                                   */
  DijkTab * up;     /*!< next more expensive tab                            */
  double visitCost; /*!< cost of myNode                                     */
  int numBefore;    /*!< number of nodes between the preceding tab and this */
};

/***********************************************************************/

/*! \class DijkTabsNode
    \brief provides a node class used in Dijkstra search with tabs

See dijkTabs.cc for descriptions of what the other functions do.

The destructor (~DijkTabsNode) does not delete kidsIndexes or kidsCosts.
This is so that copying a node can be done faster; kidsIndexes and
kidsCosts are reused by the copy.  If it is desired to delete
kidsIndexes or kidsCosts in order to free memory, that must be done
explicitly.

Note that everything required of a DijkTabsNode (other than NodeInfo)
is here. There is no graphNode base class -- not worth the trouble. 

*/

class DijkTabsNode
{
public:
  enum status {UNOPENED, OPEN, CLOSED};
  DijkTabsNode() { kidsSize= 0; open = UNOPENED; }
  ~DijkTabsNode();

  int copyFrom           /*!<  returns 0 if from is 0, otherwise returns 1 */
  (DijkTabsNode * from); /*!<  node to copy from                           */

  NodeInfo * moreInfo; //!< additional information about node
  int kidsSize;        //!< number of kids of node
  int * kidsIndexes;   //!< array of kid indexes of node
  double * kidsCosts;  //!< array of kid costs of node
  int open;            //!< flag for UNOPENED, OPEN, CLOSED
  int arrivedFrom;     //!< index in nodes of node from which arrived
  double visitCost;    //!< current total cost to visit node
  int less;            //!< index in nodes of next less expensive node
  int more;            //!< index in nodes of next more expensive node
  DijkTab * myTab;     //!< tab for node
};

/***********************************************************************/

/*! \class DijkTabsGraph
    \brief provides a graph class used in Dijkstra search with tabs

Three public methods are provided to allow an application to change the
graph being searched during the search.

In addition, the application may change the graph during the search
without using any of these methods.

In the following discussions, keep in mind that a DijkTabsGraph has
both a "capacity", which is the number of nodes currently available,
and a "size", which is the number of available nodes currently being
used.

************************
THE addKids FUNCTION
************************

AddKids may activate additional nodes of the graph. If it does, it also
adds arcs from dad to those new nodes. This may also add arcs from dad
to existing nodes.  The memory for dad's old kidsIndexes and kidsCosts
is freed. Memory for dad's revised kidsIndexes and kidsCosts is
allocated.

If there are not enough nodes in the graph to accommodate the number
to be activated, addKids calls increaseCapacity to make more nodes.

The new kids are built in nodes whose indexes are assigned in
numerical order, starting with the current size of the graph. The size
is increased as the new kids are added. The index in the graph should
already be recorded as the value of "index" in the nodeInfo. It is
checked that this number is correct.

The additional kids in oldAdded must not already be kids of dad.
AddKids is not currently checking for that. A check could be added.

************************
THE useKids FUNCTION
************************

The useKids function removes the current kids of dad (if any) and
replaces them with the ones given by the arguments. The arrays
specified in the arguments are used as is, so they must be created
with "new" in the implementing system and not changed or deleted by
the implementing system.

The kids themselves must already exist and are not affected by useKids.

**********************
OTHER MEMBER FUNCTIONS
**********************

The increaseCapacity function makes a new and larger array of nodes
and copies the old nodes into it. Since all the references in the
DijkTabsSearcher to nodes in the graph are done by index number, not
by pointer, the DijkTabsSearcher does not notice the switch.

The destructor deletes the contents of the nodes and the nodes.

The removeKids function removes all the kids of the dad node.

*/

class DijkTabsGraph
{
public:
  DijkTabsGraph(){}
  ~DijkTabsGraph();

  int addKids            /*!  adds kids to dad                              */
  (int dad,              /*!< index in nodes of node for which to add kids  */
   int numberAdded,      /*!< total number of kids added                    */
   int numberNew,        /*!< number of new kids added                      */
   NodeInfo ** newInfos, /*!< array of pointers to NodeInfos for new kids   */
   double * newCosts,    /*!< array of arc costs from dad for new kids      */
   int * oldAdded,       /*!< number of existing kids added                 */
   double * oldCosts);   /*!< array of arc costs from dad for existing kids */

  int removeKids  /* removes kidIndexes & kidCosts from dad, not kids  */
  (int dad);      /* index in nodes of node from which to remove kids  */

  int useKids           /* specifies new kids of dad                        */
  (int dad,             /* index in nodes of node from which to remove kids */
   int kidsSize,        /* new number of kids                               */
   int * kidsIndexes,   /* new array of kids indexes in nodes               */
   double * kidsCosts); /* new array of kids costs                          */

  int capacity;         /*!< total number of nodes in nodes array         */
  int size;             /*!< number of active nodes in nodes array        */
  DijkTabsNode * nodes; /*!< array of nodes                               */

private:

  int increaseCapacity  /*! makes larger array, copies old nodes into it   */
  (int numberAdded);    /* minimum number by which to increase capacity    */

};

/***********************************************************************/

#endif // DIJK_TABS_HH

/*
  $Log: dijkTabs.hh,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.9  2005/07/18 14:07:05  kramer
  added goal not start warning to documentation

  Revision 1.8  2005/07/15 21:24:44  kramer
  changed interface, improved documentation

  Revision 1.7  2005/06/29 21:08:05  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.6  2005/06/09 12:16:45  kramer
  added CVS logging
*/
