/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primMisJA.hh

  \brief NML declarations prim-level mission package that generates
  one-dimension point-to-point moves for the mission package, where
  points are the ending angular position of mission package.

  Synopsis:
  #include <rcs.hh>
  #include <posemath.h>
  #include "moastNmlOffsets.hh"
  
  \code CVS Status:
  $Author: proctor $
  $Revision: 1.9 $
  $Date: 2008/09/25 20:49:34 $
  \endcode

  \author Fred Proctor
*/

#ifndef PRIM_MIS_JA_HH
#define PRIM_MIS_JA_HH

#include "moastNmlOffsets.hh"
#include "servoMisJA.hh"	// SERVO_MIS_JA_LINK_MAX

#define PRIM_MIS_JA_CMD_NAME "primMisJACmd"
#define PRIM_MIS_JA_STAT_NAME "primMisJAStat"
#define PRIM_MIS_JA_CFG_NAME "primMisJACfg"
#define PRIM_MIS_JA_SET_NAME "primMisJASet"

#define PRIM_MIS_JA_CMD_BASE        (PRIM_MIS_JA_BASE)
#define PRIM_MIS_JA_STAT_BASE       (PRIM_MIS_JA_BASE + 100)
#define PRIM_MIS_JA_CFG_BASE        (PRIM_MIS_JA_BASE + 200)
#define PRIM_MIS_JA_SET_BASE        (PRIM_MIS_JA_BASE + 300)

#define PRIM_MIS_JA_CMD_NOP_TYPE        (PRIM_MIS_JA_CMD_BASE + 1)
#define PRIM_MIS_JA_CMD_INIT_TYPE       (PRIM_MIS_JA_CMD_BASE + 2)
#define PRIM_MIS_JA_CMD_ABORT_TYPE      (PRIM_MIS_JA_CMD_BASE + 3)
#define PRIM_MIS_JA_CMD_HALT_TYPE       (PRIM_MIS_JA_CMD_BASE + 4)
#define PRIM_MIS_JA_CMD_SHUTDOWN_TYPE   (PRIM_MIS_JA_CMD_BASE + 5)
#define PRIM_MIS_JA_CMD_MOVE_JOINT_TYPE (PRIM_MIS_JA_CMD_BASE + 6)
#define PRIM_MIS_JA_CMD_MOVE_WORLD_TYPE (PRIM_MIS_JA_CMD_BASE + 7)
#define PRIM_MIS_JA_CMD_STOP_TYPE       (PRIM_MIS_JA_CMD_BASE + 8)

#define PRIM_MIS_JA_STAT_TYPE       (PRIM_MIS_JA_STAT_BASE + 1)

#define PRIM_MIS_JA_CFG_NOP_TYPE        (PRIM_MIS_JA_CFG_BASE + 1)
#define PRIM_MIS_JA_CFG_CYCLE_TIME_TYPE (PRIM_MIS_JA_CFG_BASE + 2)
#define PRIM_MIS_JA_CFG_DEBUG_TYPE      (PRIM_MIS_JA_CFG_BASE + 3)

#define PRIM_MIS_JA_SET_TYPE        (PRIM_MIS_JA_SET_BASE + 1)

// PrimMisJACmdXXX

class PrimMisJACmdNop:public RCS_CMD_MSG {
public:
  PrimMisJACmdNop():RCS_CMD_MSG
    (PRIM_MIS_JA_CMD_NOP_TYPE,
     sizeof(PrimMisJACmdNop)) {};
  void update(CMS *);
};

class PrimMisJACmdInit:public RCS_CMD_MSG {
public:
  PrimMisJACmdInit():RCS_CMD_MSG(PRIM_MIS_JA_CMD_INIT_TYPE,
			       sizeof(PrimMisJACmdInit)) {};
  void update(CMS *);
};

class PrimMisJACmdAbort:public RCS_CMD_MSG {
public:
  PrimMisJACmdAbort():RCS_CMD_MSG(PRIM_MIS_JA_CMD_ABORT_TYPE,
				sizeof(PrimMisJACmdAbort)) {};
  void update(CMS *);
};

class PrimMisJACmdHalt:public RCS_CMD_MSG {
public:
  PrimMisJACmdHalt():RCS_CMD_MSG(PRIM_MIS_JA_CMD_HALT_TYPE,
			       sizeof(PrimMisJACmdHalt)) {};
  void update(CMS *);
};

class PrimMisJACmdShutdown:public RCS_CMD_MSG {
public:
  PrimMisJACmdShutdown():RCS_CMD_MSG(PRIM_MIS_JA_CMD_SHUTDOWN_TYPE,
				   sizeof(PrimMisJACmdShutdown)) {};
  void update(CMS *);
};

class PrimMisJACmdMoveJoint:public RCS_CMD_MSG {
public:
  PrimMisJACmdMoveJoint():RCS_CMD_MSG(PRIM_MIS_JA_CMD_MOVE_JOINT_TYPE,
			       sizeof(PrimMisJACmdMoveJoint)) {};
  void update(CMS *);

  //! values of the joint variables, in their SI units
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(double, joints, SERVO_MIS_JA_LINK_MAX);
  //! elapsed time to get there, [s]
  double time;
};

class PrimMisJACmdMoveWorld:public RCS_CMD_MSG {
public:
  PrimMisJACmdMoveWorld():RCS_CMD_MSG(PRIM_MIS_JA_CMD_MOVE_WORLD_TYPE,
			       sizeof(PrimMisJACmdMoveWorld)) {};
  void update(CMS *);

  //! end pose of mission package
  PM_POSE end;
  //! elapsed time to get there, [s]
  double time;
};

class PrimMisJACmdStop:public RCS_CMD_MSG {
public:
  PrimMisJACmdStop():RCS_CMD_MSG(PRIM_MIS_JA_CMD_STOP_TYPE,
			       sizeof(PrimMisJACmdStop)) {};
  void update(CMS *);
};

/*
  PrimMisJAStat is the class that holds dynamic mission package status
*/
class PrimMisJAStat : public RCS_STAT_MSG_V2 {
public:
  PrimMisJAStat() : RCS_STAT_MSG_V2(PRIM_MIS_JA_STAT_TYPE, sizeof(PrimMisJAStat)) {};
  void update(CMS *);

  //! time the position was acquired
  double timestamp;

  //! current joint position
  double joints[SERVO_MIS_JA_LINK_MAX];

  //! current pose relative to mounting origin
  PM_POSE pose;

  //! is the mission package in position
  int inpos;
};

class PrimMisJACfgNop:public RCS_CMD_MSG {
public:
  PrimMisJACfgNop():RCS_CMD_MSG
    (PRIM_MIS_JA_CFG_NOP_TYPE,
     sizeof(PrimMisJACfgNop)) {};
  void update(CMS *);
};

/*
  PrimMisJACfgCycleTime sets the controller's cycle time
*/
class PrimMisJACfgCycleTime:public RCS_CMD_MSG {
public:
  PrimMisJACfgCycleTime():RCS_CMD_MSG(PRIM_MIS_JA_CFG_CYCLE_TIME_TYPE,
			       sizeof(PrimMisJACfgCycleTime)) {};
  void update(CMS *);

  //! period in seconds
  double cycleTime;
};

/*
  PrimMisJACfgDebug sets the controller's debug level
*/
class PrimMisJACfgDebug:public RCS_CMD_MSG {
public:
  PrimMisJACfgDebug():RCS_CMD_MSG(PRIM_MIS_JA_CFG_DEBUG_TYPE,
			       sizeof(PrimMisJACfgDebug)) {};
  void update(CMS *);

  //! debug level
  int debug;
};

/*
  PrimMisJASet is the class that holds the quasi-static mission package
  settings, those that change infrequently and in response to sensor 
  processing configuration messages.
*/
class PrimMisJASet : public RCS_STAT_MSG {
public:
  PrimMisJASet():RCS_STAT_MSG(PRIM_MIS_JA_SET_TYPE, sizeof(PrimMisJASet)) {};
  void update(CMS *);

  //! time for trajectory planning calculations, [s]
  double cycleTime;
  //! debug level
  int debug;

  //! pose of mission package origin relative to vehicle
  PM_POSE pose;
};

extern int primMisJA_format(NMLTYPE type, void *buf, CMS * cms);
extern const char * primMisJA_symbol_lookup(long type);

#endif // PRIM_MIS_JA_HH

/*
  Modification history:

  $Log: primMisJA.hh,v $
  Revision 1.9  2008/09/25 20:49:34  proctor
  Completely rewrote primMisMain,the prim mission package, to interface to
  a Go Motion controller for doing all the real work. Autotools was changed
  to look for Go Motion. The prim mission interface was changed to eliminate
  the pan and tilt commands in favor of joint and world moves.

  Revision 1.8  2006/06/21 19:38:13  proctor
  Added support for Java, which involved including some left-out headers
  like moastNmlOffsets.hh and moastTypes.hh

  Revision 1.7  2006/04/13 14:30:07  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.6  2006/01/06 16:04:48  proctor
  Added pan and tilt messages

  Revision 1.5  2005/12/22 19:46:08  proctor
  Added fraction to scan command

  Revision 1.4  2005/12/20 15:20:57  proctor
  Added scan

  Revision 1.3  2005/12/19 21:23:31  proctor
  Changed from pose to pan, tilt due to non-intuitive conversions between
  representations like RPY or ZYZ Eulers.

  Revision 1.2  2005/12/16 20:55:02  proctor
  Changed the primMisJACmdMove field to a PM_POSE, and made related changes.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.2  2005/06/08 14:29:54  proctor
  Added vehMobPLMain name change

  Revision 1.1  2005/06/08 12:47:40  proctor
  Name change per Mis,MobJA naming convention

  Revision 1.2  2005/06/07 18:01:26  proctor
  Added primMisJA.hh; changed servoMIS to servoMis

  Revision 1.1  2005/06/07 15:50:28  proctor
  Initial checkin

*/
