/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   scrollingBox.hh
  \brief  Contains class declaration for ScrollingBox

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.1 $
  $Date: 2008/01/24 13:48:09 $
  \endcode
  \par CVS Log:
  \ref scrollingBox_hh "More..."

  \author Tom Kramer
  \date   2007-08-09

*****************************************************************************

HOW TO USE THIS FILE

To use this file in an application:
1.  The user must write an application header file (call it app.hh).
2.  The user must write an application implementation file (call it app.cc).
3.  The user must write a file (call it voxel.hh) that defines a voxel class.
4.  app.hh must #define VOXEL_NAME.
5.  app.hh must #include voxel.hh.
6.  app.hh must #include "scrollingBox.hh" (AFTER voxel.hh!).
7.  The name of the voxel class in voxel.hh must be the value of VOXEL_NAME.
8.  The voxel class in voxel.hh must be derived from the
    ScrollingBoxVoxel class.
9.  voxel.hh must #include "scrollingBoxVoxel.hh"
10. voxel.hh must have a public "init()" method taking no arguments.
11. app.hh may define a box class that is a derived class of the
    ScrollingBox class defined here (or it may use the ScrollingBox class).
12. app.cc must #include app.hh.
13. app.cc must #include "scrollingBox.cc".

EXAMPLE 1 (detailed)

See

*************************************************************************

Conceptually, a scrolling box is a box-shaped 3D volume that lies in
unbounded cartesian XYZ space and can move around (scroll) so that
different parts of space can be seen in the box. The edges of the
box stay aligned with the XYZ axes at all times. The box is
divided into a 3D grid of cubical voxels called the "box grid".
When the box moves, it moves an integer number of grid voxels in
the X and/or Y and/or Z directions. The user can tell the box to
scroll so that its center voxel contains a specific cartesian point
called the center point. Note that although the center point is
always in the center voxel, except immediately after a scrollingBox
has been constructed, the center point is not usually at the center
of the center voxel.

The number of voxels on each side of the box grid is odd to make it
possible to keep the center point in the voxel at the center of the
box grid. To ensure the number is odd, the size of the box is specified
by giving xRange, yRange, and zRange, which are the numbers of voxels
in the X, Y, and Z directions, respectively between the center voxel
and a face of the box.

Four methods dealing with the center point are provided:
1. getCenterX (takes no arguments, returns the X coordinate used to locate
   the center voxel of the box grid).
2. getCenterY (takes no arguments, returns the Y coordinate used to locate
   the center voxel of the box grid).
3. getCenterZ (takes no arguments, returns the Z coordinate used to locate
   the center voxel of the box grid).
4. setCenter (takes doubles x, y, and z, and pointers to ints xDiff, yDiff,
   and zDiff scrolls the box so that the point (x, y, z) is in the center
   voxel of the box grid; also sets xDiff, yDiff, and zDiff to the number
   of voxels the center voxel has moved).

The constructor of a scrollingBox, takes as arguments:
1. resolution (a double representing the length of a side of a voxel).
2. xRangeIn (a positive int representing the number of voxels in the
   X direction between the center voxel of the box grid and a face of the
   box).
3. yRangeIn (a positive int representing the number of voxels in the
   Y direction between the center voxel of the box grid and a face of the
   box).
4. zRangeIn (a positive int representing the number of voxels in the
   Z direction between the center voxel of the box grid and a face of the
   box).

The number of voxels in the X direction along a face of the box grid
parallel to the X-axis is called "xSide". xSide is one more than twice
the xRange. In equation form, xSide = ((2 * xRange) + 1). ySide and
zSide are defined similarly.

No length units are specified. Units are left to the user and may be
implicit.

A "xRow" of a grid is a set of voxels with the same X index.
A "yRow" of a grid is a set of voxels with the same Y index.
A "zCol" of a grid is a set of voxels with the same Z index.

Several access methods are provided to get data about a ScrollingBox.
1.  getXRange returns the box xRange.
2.  getYRange returns the box yRange.
3.  getZRange returns the box zRange.
4.  getXSide returns the box xSide.
5.  getYSide returns the box ySide.
6.  getZSide returns the box zSide.
7.  getResolution returns the box resolution.
8.  getXHi returns the maximum value of X of any point in the box. This is
    the value of X at all points on the box face lying in the +X direction.
9.  getYHi returns the maximum value of Y of any point in the box. This is
    the value of Y at all points on the box face lying in the +Y direction.
10. getZHi returns the maximum value of Z of any point in the box. This is
    the value of Z at all points on the box face lying in the +Z direction.
11. getXLow returns the minimum value of X of any point in the box. This is
    the value of X at all points on the box face lying in the -X direction.
11. getYLow returns the minimum value of Y of any point in the box. This is
    the value of Y at all points on the box face lying in the -Y direction.
12. getZLow returns the minimum value of Z of any point in the box. This is
    the value of Z at all points on the box face lying in the -Z direction.

Three access methods are provided to get a pointer to a voxel:
1. getVoxelAtIndexes (which takes X, Y, and Z box grid indexes as arguments).
2. getVoxelAtCoords (which takes X, Y, and Z coordinates as arguments).
3. getVoxelWithId (which takes an id number as an argument).

Several access methods are provided for getting coordinates and indexes.
1.  getXCoordAtVoxel returns the X coordinate of the center of a given voxel.
2.  getYCoordAtVoxel returns the Y coordinate of the center of a given voxel.
3.  getZCoordAtVoxel returns the Z coordinate of the center of a given voxel.
4.  getXCoordAtId returns the X coordinate of the center of the voxel
    with a given id.
5.  getYCoordAtId returns the Y coordinate of the center of the voxel
    with a given id.
6.  getZCoordAtId returns the Z coordinate of the center of the voxel
    with a given id.
7.  getXIndexAtVoxel returns the box grid X index of a given voxel.
8.  getYIndexAtVoxel returns the box grid Y index of a given voxel.
9.  getZIndexAtVoxel returns the box grid Z index of a given voxel.
10. getXIndexAtCoord returns the box grid X index of an X coordinate value.
11. getYIndexAtCoord returns the box grid Y index of a  Y coordinate value.
11. getZIndexAtCoord returns the box grid Z index of a  Z coordinate value.

Two methods are provided to make it possible to get all the voxels that
are crossed by a line between two points.
1. initLine takes the coordinates of the two points and returns a pointer
   to the voxel containing the first point. It also gets the system ready
   for calls to getNextVoxel.
2. getNextVoxel returns a pointer to the next voxel on the line defined by
   the most recent call to initLine.

To make it dynamic and efficient, the implementation of the ScrollingBox
class differs from the conceptual view.  The implementation uses what is
called the "actual grid", a fixed grid of the same size as the box grid.
The voxel in the corner of the box grid with minimum X, Y, and Z is located
on the actual grid by the xLowIndex, yLowIndex, and zLowIndex, which are
indexes on the actual grid of that voxel.  The actual grid is divided into
eight sections by three planes (parallel to the XY-plane, the YZ-plane,
and the ZX-plane) through the -X, -Y, and -Z faces of the voxel whose
indexes are xLowIndex, yLowIndex, and zLowIndex. The eight sections of
the actual grid represent eight parts of the box grid.

Voxels are expected to be instances of a subclass of the
ScrollingBoxVoxel class, which is defined in scrollingBoxVoxel.hh. The
subclass should be defined in some other file by the user. The
ScrollingBoxVoxel class has only one data member, an integer id
number. Id numbers are assigned to voxels by the ScrollingBox class
constructor. There is no method of changing them.

It is required that that an "init" method taking no arguments and
returning void be defined in the voxel subclass defined by the
user. The init method is invoked by methods of the ScrollingBox class
to initialize voxels as desired by the user. The ScrollingBox class
knows the locations and id numbers of the voxels but does not know
anything about the user's voxel class except that it has an init
method.

The actual grid is implemented as a dynamic 1D array of voxels called
"voxels". The private method getVoxel1D gets a pointer to a voxel
given an argument that is the voxel's index in the one-dimensional
array.  This is identical to the public getVoxelWithId method except
that it does not check whether the id argument is in bounds.

The getVoxel3D private access method gets a pointer to a voxel as
though the actual grid were a 3D array. That method takes the voxel's
indexes on the actual grid as arguments. For the getVoxel3D method,
the 1D array is conceptually mapped to the 3D actual grid as follows.
Chop the 1D array into sections xSide voxels long.  Align the sections
parallel to the X-axis.  Stack subsequent sections in the Y direction
to make sheets ySide wide. Stack sheets in the Z direction to make a
box zSide high. Note that in the XY plane, this differs from the way
it is done in (2D) scrollingMap.

When a scrolling box is first created, the box grid and the actual
grid are the same, and the origin of space is at the center of the
voxel in the center of the grid. Also, at that time, the X, Y, and Z
box grid indexes of the voxel at the center of the actual grid are
equal to the xRange, yRange, and zRange.

Conceptually, when the box scrolls, empty voxels enter the box grid
from the direction in which the box moved, and voxels leave the box
grid in the opposite direction.  For implementation efficiency, when
scrolling takes place, rather than copying from one place to another data
for the voxels on the actual grid that represent parts of the box grid
that did not enter or leave the box grid, the data for those voxels is
not touched. Rather, the data in the voxels that conceptually left the
box grid is wiped out, and those same voxels are reused to represent the
voxels that entered the box grid. As mentioned earlier, the user must
define an init method that reinitializes the contents of a voxel. That
method is invoked by the box during scrolling.

The best way to visualize the relationship between the box and the
actual grid is to imagine that space is filled with copies of the
actual grid. The edges of the copies divide the box into one, two,
four, or eight parts. The location of each of the parts relative to
the copy of the actual grid in which the part lies is where that part
lies on the actual grid.

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.1 $
  $Date: 2008/01/24 13:48:09 $
  \endcode
  \par CVS Log:
  \ref amSPBox_hh "More..."

  \author Tom Kramer
  \date   2007-07-08
*/

/***********************************************************************/

#ifndef __scrollingBox_hh
#define __scrollingBox_hh

#include <math.h> // floor

/***********************************************************************/
/***********************************************************************/

class ScrollingBox {

/**PUBLIC***************************************************************/

public:

/***********************************************************************/

//! Public enumeration dir is a choice of one of six principal directions.

  enum dir {X_PLUS_DIR, X_MINUS_DIR,
	    Y_PLUS_DIR, Y_MINUS_DIR,
	    Z_PLUS_DIR, Z_MINUS_DIR};

/***********************************************************************/

/*! The constructor taking no arguments creates a ScrollingBox but sets
  nothing. */

  ScrollingBox(){};

/***********************************************************************/

/*! The constructor taking four arguments creates a ScrollingBox based
  on three ranges and resolution. */

  ScrollingBox(int xRangeIn, int yRangeIn, int zRangeIn, double resIn);

/***********************************************************************/

//!  The destructor deletes allocated memory.

  ~ScrollingBox();

/***********************************************************************/

/*! Public method checkEnd takes the coordinates of a start point in
  the box and pointers to the coordinates of an end point that may or
  may not be in the box. If the end point is not in the box, it is
  moved to the point on a face of the box where the line enters the box. */
  
  void checkEnd(double startX, double startY, double startZ,
		double * endX, double * endY, double * endZ);

/***********************************************************************/

/*! Public method getCenterX returns the cartesian X coordinate of the
  point used to set the center of the box. */

  double getCenterX() {return centerX;}

/***********************************************************************/

/*! Public method getCenterY returns the cartesian Y coordinate of the
  point used to set the center of the box. */

  double getCenterY() {return centerY;}

/***********************************************************************/

/*! Public method getCenterZ returns the cartesian Z coordinate of the
  point used to set the center of the box. */

  double getCenterZ() {return centerZ;}

/***********************************************************************/

//! Public method getEndInside gets the value of endInside

  bool getEndInside() {return endInside;}

/***********************************************************************/

/*! Public method getNextVoxel returns a pointer to the next voxel on
  the line defined by the most recent call to initLine. If the end of
  the line has been reached, a null pointer is returned. If setCenter
  has been called since the most recent call to initLine, a null
  pointer is returned. */

  VOXEL_NAME * getNextVoxel();
  
/***********************************************************************/

//! Public method getResolution returns the resolution of the ScrollingBox.

  double getResolution() {return res;}

/***********************************************************************/

  /*! Public method getPrincipalDir returns the principal direction
  of the ScrollingBox. This is useful only when a line is being
  followed through the box, so that inLine is true. */

  dir getPrincipalDir(){return principalDir;}

/***********************************************************************/

/*! Public method getVoxelAtCoords returns a pointer to a voxel whose
  cartesian X, Y, and Z coordinates are given. If the point with those
  coordinates is not in the box, the method returns NULL. */

  VOXEL_NAME * getVoxelAtCoords(double x, double y, double z);

/***********************************************************************/

/*! Public method getVoxelAtIndexes returns a pointer to the voxel whose
  X, Y, and Z indexes in the box grid are given. */

VOXEL_NAME * getVoxelAtIndexes(int xIndex, int yIndex, int zIndex);

/***********************************************************************/

/*! Public method getVoxelWithId returns a pointer to a voxel whose id
  is given. If the id is not the id of a voxel, this returns NULL. */

  VOXEL_NAME * getVoxelWithId(int id)
  {
    if ((id > -1) && (id < numberOfVoxels))
      return (voxels + id);
    else
      return 0;
  }

/***********************************************************************/

/*! Public method getXCoordAtId returns the X coordinate of the center
  of the voxel with the given id. If the id is not the id of a voxel,
  this returns xLow. */

  double getXCoordAtId(int id);

/***********************************************************************/

/*! Public method getXCoordAtIndex returns the X coordinate of the
  center of the voxel with the given X box index. If the index is out of
  bounds, this returns a value outside the box. */

  double getXCoordAtIndex(int xIndex){return(xLow + ((xIndex + 0.5) * res));}

/***********************************************************************/

/*! Public method getXCoordAtVoxel returns the X coordinate of the
  center of the given voxel. If the voxel is NULL or its id is not the
  id of a voxel, this returns xLow. */

  double getXCoordAtVoxel(VOXEL_NAME * aVoxel);

/***********************************************************************/

/*! Public method getXHi returns the cartesian X coordinate of a point
  of the box with the greatest X value (i.e., any point on the +X
  face of the box). */

  double getXHi() {return xHi;}

/***********************************************************************/

/*! Public method getXIndexAtVoxel returns the box grid X index of the
  given voxel. If the voxel is NULL or its id is not the id of a voxel,
  this returns -1. */

  int getXIndexAtVoxel(VOXEL_NAME * aVoxel);

/***********************************************************************/

/*! Public method getXIndexAtCoord returns the box grid X index of
  the X coordinate value x, if x is inside the box. Otherwise, it
  returns -1. */

  int getXIndexAtCoord(double x);

/***********************************************************************/

/*! Public method getXLow returns the cartesian X coordinate of a
  point of the box with the least X value (i.e., any point on the -X
  face of the box). */

  double getXLow() {return xLow;}

/***********************************************************************/

//! Public method getXRange returns the xRange of the ScrollingBox.

  int getXRange() {return xRange;}

/***********************************************************************/

//! Public method getXSide returns the xSide of the ScrollingBox.

  int getXSide() {return xSide;}

/***********************************************************************/

/*! Public method getYCoordAtId returns the Y coordinate of the center
  of the voxel with the given id. If the id is not the id of a voxel,
  this returns yLow. */

  double getYCoordAtId(int id);

/***********************************************************************/

/*! Public method getYCoordAtIndex returns the Y coordinate of the
  center of the voxel with the given Y box index. If the index is out of
  bounds, this returns a value outside the box. */

  double getYCoordAtIndex(int yIndex){return(yLow + ((yIndex + 0.5) * res));}

/***********************************************************************/

/*! Public method getYCoordAtVoxel returns the Y coordinate of the
  center of the given voxel. If the voxel is NULL or its id is not the
  id of a voxel, this returns yLow. */

  double getYCoordAtVoxel(VOXEL_NAME * aVoxel);

/***********************************************************************/

/*! Public method getYHi returns the cartesian Y coordinate of a point
  of the box with the greatest Y value (i.e., any point on the +Y
  face of the box). */

  double getYHi() {return yHi;}

/***********************************************************************/

/*! Public method getYIndexAtVoxel returns the box grid Y index of the
  given voxel. If the voxel is NULL or its id is not the id of a voxel,
  this returns -1. */

  int getYIndexAtVoxel(VOXEL_NAME * aVoxel);

/***********************************************************************/

/*! Public method getYIndexAtCoord returns the box grid Y index of
  the Y coordinate value y, if y is inside the box. Otherwise, it
  returns -1. */

  int getYIndexAtCoord(double y);

/***********************************************************************/

/*! Public method getYLow returns the cartesian Y coordinate of a
  point of the box with the least Y value (i.e., any point on the -Y
  face of the box). */

  double getYLow() {return yLow;}

/***********************************************************************/

//! Public method getYRange returns the yRange of the ScrollingBox.

  int getYRange() {return yRange;}

/***********************************************************************/

//! Public method getYSide returns the ySide of the ScrollingBox.

  int getYSide() {return ySide;}

/***********************************************************************/

/*! Public method getZCoordAtId returns the Z coordinate of the center
  of the voxel with the given id. If the id is not the id of a voxel,
  this returns zLow. */

  double getZCoordAtId(int id);

/***********************************************************************/

/*! Public method getZCoordAtIndex returns the Z coordinate of the
  center of the voxel with the given Z box index. If the index is out of
  bounds, this returns a value outside the box. */

  double getZCoordAtIndex(int zIndex){return(zLow + ((zIndex + 0.5) * res));}

/***********************************************************************/

/*! Public method getZCoordAtVoxel returns the Z coordinate of the
  center of the given voxel. If the voxel is NULL or its id is not the
  id of a voxel, this returns zLow. */

  double getZCoordAtVoxel(VOXEL_NAME * aVoxel);

/***********************************************************************/

/*! Public method getZHi returns the cartesian Z coordinate of a point
  of the box with the greatest Z value (i.e., any point on the +Z
  face of the box). */

  double getZHi() {return zHi;}

/***********************************************************************/

/*! Public method getZIndexAtVoxel returns the box grid Z index of the
  given voxel. If the voxel is NULL or its id is not the id of a voxel,
  this returns -1. */

  int getZIndexAtVoxel(VOXEL_NAME * aVoxel);

/***********************************************************************/

/*! Public method getZIndexAtCoord returns the box grid Z index of
  the Z coordinate value z, if z is inside the box. Otherwise, it
  returns -1. */

  int getZIndexAtCoord(double z);

/***********************************************************************/

/*! Public method getZLow returns the cartesian Z coordinate of a
  point of the box with the least Z value (i.e., any point on the -Z
  face of the box). */

  double getZLow() {return zLow;}

/***********************************************************************/

//! Public method getZRange returns the zRange of the ScrollingBox.

  int getZRange() {return zRange;}

/***********************************************************************/

//! Public method getZSide returns the zSide of the ScrollingBox.

  int getZSide() {return zSide;}

/***********************************************************************/

/*! Public method initLine initializes following a line from a start
  point to an end point. If the start point is in the box, a
  pointer to the voxel containing the start point is returned; if not,
  a null pointer is returned.  The end point does not need to be in
  the box. The line will be followed from the start point up to the
  edge of the map. After this method is called, calls may be made
  repeatedly to getNextVoxel to get the voxels through which the line
  passes. See the documentation of getNextVoxel. */

/*

The following discussion is relevant to initLine and getNextVoxel.

There is a problem with deciding that voxels that contain obstacles do
not contain obstacles (i.e. false negatives). This occurs because
lines may pass through the empty part of voxels that contain
obstacles.

Modifications to line processing that improve the situation for
some configurations may make the situation worse for other
configurations. For example, if a solid surface is made of voxels in a
plane parallel to one of the principal planes, then moving each hit
point to the point on the hit voxel surface nearest the sensor point
will reduce the problem.  If the same thing is done when the surface
is made of voxels lying on a plane that is parallel to [a principal
plane rotated 45 degrees about one of the two axes defining the
plane], moving hit points as described may well make the problem
worse.

Processing hit points from farther away to closer will always improve
the situation when data more recently processed is weighted more
heavily, but sorting ladar rays by distance is computationally
expensive.

Processing hit points from straight ahead to one side and then from
straight ahead to the other side (as opposed to from one side to the
other) will usually make an improvement when the sensor point is
being controlled so that it stays away from obstacles.

Thickening obstacles in the direction of rays will help, although
it will produce sensed obstacles that are thicker than the actual
obtacles really are.

*/

  VOXEL_NAME * initLine(double startX, double startY, double startZ,
			double endX, double endY, double endZ);

/***********************************************************************/

/*! Public method setCenter causes the box to scroll so that the
  cartesian point (x, y, z) lies inside the voxel at the center of the
  box grid.  The xLowIndex, yLowIndex, zLowIndex, xLow, yLow, zLow,
  xHi, yHi, zHi, centerX, centerY, and centerZ of the map change when
  the center moves. Also, the voxels of the actual grid that represent
  the voxels entering (and leaving) the box grid are reinitialized.
  xDiff, yDiff, and zDiff are set to the number of voxels the center
  has moved in the X, Y, and Z directions; the values may be positive,
  zero, or negative. */

  void setCenter(double x, double y, double z,
		 int * xDiff, int * yDiff, int * zDiff);

/***********************************************************************/
/***********************************************************************/
/**PRIVATE**************************************************************/

private:

/*! Method getActualRawXIndex returns the X index on the actual grid of
  a cartesian X coordinate as though the grid were unbounded. */

  int getActualRawXIndex(double x)
  {
    return (int)floor((x / res) + xRange + 0.5);
  }
  
/***********************************************************************/

/*! Method getActualRawYIndex returns the Y index on the actual grid of
  a cartesian Y coordinate as though the grid were unbounded. */

  int getActualRawYIndex(double y)
  {
    return (int)floor((y / res) + yRange + 0.5);
  }

/***********************************************************************/

/*! Method getActualRawZIndex returns the Z index on the actual grid of
  a cartesian Z coordinate as though the grid were unbounded. */

  int getActualRawZIndex(double z)
  {
    return (int)floor((z / res) + zRange + 0.5);
  }

/***********************************************************************/

/*! Method getVoxel1D returns a pointer to the voxel of the 1D array
  with the given 1D index. */

/*

For speed, this is not checking index bounds. The index should satisfy
(0 <= index < numberOfVoxels).

*/

  VOXEL_NAME * getVoxel1D(int index) {return (voxels + index);}

/***********************************************************************/

/*! Method getVoxel3D returns a pointer to the voxel of the actual grid
  with the given 3D indexes. */

/* 

For speed, this is not checking index bounds. The indexes should satisfy
(0 <= xIndex < xSide), (0 <= yIndex < ySide), and (0 <= zIndex < zSide).

*/

VOXEL_NAME * getVoxel3D(int xIndex, int yIndex, int zIndex)
  {
    return (voxels + (zIndex * xSide * ySide) + (yIndex * xSide) + xIndex);
  }

/***********************************************************************/


  double centerX;       //! cartesian X coordinate of center point
  double centerY;       //! cartesian Y coordinate of center point
  double centerZ;       //! cartesian Y coordinate of center point  
  int numberOfVoxels;   //! number of voxels
  double res;           //! length of a side of a voxel, abbr for resolution
  VOXEL_NAME * voxels;  //! 1D array of voxels
  double xHi;           //! maximum cartesian value of X in box
  double xLow;          //! minimum cartesian value of X in box
  int xLowIndex;        //! X index on actual grid of corner of box grid
  int xRange;           //! num voxels between center and face of box X dir
  int xSide;            //! number of voxels along face of box in X direction
  double yHi;           //! maximum cartesian value of Y in box
  double yLow;          //! minimum cartesian value of Y in box
  int yLowIndex;        //! Y index on actual grid of corner of box grid
  int yRange;           //! num voxels between center and face of box Y dir
  int ySide;            //! number of voxels along face of box in Y direction
  double zHi;           //! maximum cartesian value of Z in box
  double zLow;          //! minimum cartesian value of Z in box
  int zLowIndex;        //! Z index on actual grid of corner of box grid
  int zRange;           //! num voxels between center and side of box Z dir
  int zSide;            //! number of voxels along face of box in Z direction

  /* The following are used by initLine and getNextVoxel */

  double deltaX;        //! change in cartesian X from start to end
  double deltaY;        //! change in cartesian Y from start to end
  double deltaZ;        //! change in cartesian Z from start to end
  int di;               //! +1 or -1, according to sign of deltaX
  int dI;               //! amount by which i has changed (positive)
  int dj;               //! +1 or -1, according to sign of deltaY
  int dJ;               //! amount by which j has changed (positive)
  int dk;               //! +1 or -1, according to sign of deltaZ
  int dK;               //! amount by which k has changed (positive)
  bool endInside;       //! true if original end point inside box, else false
  bool inLine;          //! whether a line is being followed
  int i;                //! box grid X index of voxel
  int j;                //! box grid Y index of voxel
  int k;                //! box grid Z index of voxel
  double lack1;         //! distance needed in 1st secondary dir (positive)
  double lack2;         //! distance needed in 2nd secondary dir (positive)
  VOXEL_NAME * next1;   //! immediate next voxel
  VOXEL_NAME * next2;   //! first cached next voxel (may be void)
  VOXEL_NAME * next3;   //! second cached next voxel (may be void)
  dir principalDir;     //! principal direction of line
  double slope1;        //! positive slope in 1st secondary direction
  double slope2;        //! positive slope in 2nd secondary direction
  double step1;         //! step size in 1st secondary direction (positive)
  double step2;         //! step size in 2nd secondary direction (positive)
  int totalI;           //! total number of voxels in X direction
  int totalJ;           //! total number of voxels in Y direction
  int totalK;           //! total number of voxels in Z direction
};

#endif

/*! \defgroup amSPMap_hh CVS Log Information
 * $Log: scrollingBox.hh,v $
 * Revision 1.1  2008/01/24 13:48:09  tomrkramer
 * initial entry
 *
 */
