/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file trafficData.hh

  \brief  NML declarations for traffic data channel. 

  This channel provides
  an array of navigation information. One element for each
  traffic vehicle.


  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.2 $
  $Date: 2007/05/03 16:36:26 $
  \endcode

  \author Stephen Balakirsky
*/

#ifndef trafficData_HH
#define trafficData_HH

#include <rcs.hh>		// NMLmsg, PM_CARTESIAN,RPY,POSE
#include "moastNmlOffsets.hh"	// NAV_DATA_EXT_BASE
#include "moastTypes.hh"        // max traffic vehicles

#define TRAFFIC_DATA_TYPE (TRAFFIC_DATA_BASE + 1)
#define TRAFFIC_DATA_NAME "trafficData"

typedef struct {
  /*!
    \a tranAbs holds absolute position, in meters, where \a x is UTM northing,
    \a y is UTM easting, and \a z points positive downward. 
  */
  PM_CARTESIAN tranAbs;
  /*!
    \a rpyAbs holds absolute orientation, in radians, where \a roll
    is rotation about the UTM northing (X) direction, \a pitch is rotation
    about the UTM easting (Y) direction, and \a yaw is rotation about
    the downward-pointing Z direction. Positive changes in yaw correspond to
    clockwise rotation as viewed looking down.
  */
  PM_RPY rpyAbs;
  //! absolute velocity, time deriv of tranAbs, [m/s]
  PM_CARTESIAN tranAbsRate;
  //! time deriv of rpyAbs, [rad/s]
  PM_RPY rpyAbsRate;

  //! relative position, [m]
  PM_CARTESIAN tranRel;
  //! relative orientation, roll-pitch-yaw, [rad]
  PM_RPY rpyRel;
  //! relative velocity, time deriv of tranAbs, [m/s]
  PM_CARTESIAN tranRelRate;
  //! time deriv of rpyRel, [rad/s]
  PM_RPY rpyRelRate;

  /*!
    'truth' values of position and orientation, in some global coordinate
    frame, obtainable in simulation but never in real life 
  */
  PM_CARTESIAN tranTrue;
  PM_RPY rpyTrue;
  PM_CARTESIAN tranTrueRate;
  PM_RPY rpyTrueRate;
  //! [m/s] speed of vehicle, pos=forward, neg=reverse
  double speed;
  double steeringCurvature;
}TrafficVeh;

class TrafficData : public NMLmsg {
public:
  TrafficData() : NMLmsg(TRAFFIC_DATA_TYPE, sizeof(TrafficData)) {};
  void update(CMS *);

  //! seconds relative to common origin
  double time;			
  // will automatically create the variable vehicles_length for the length of the array
  DECLARE_NML_DYNAMIC_LENGTH_ARRAY(TrafficVeh, vehicles, MAX_TRAFFIC_VEHICLES);
};

extern int trafficData_format(NMLTYPE type, void *buf, CMS * cms);
extern const char *trafficData_symbol_lookup(long type);

#endif
/*
  Modification history:

  $Log: trafficData.hh,v $
  Revision 1.2  2007/05/03 16:36:26  dr_steveb
  Fixed minor issues with interface

  Revision 1.1  2007/04/25 15:08:57  dr_steveb
  Added traffic simulation files

*/
