/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file ulapi.h

  \brief Declarations of the user-level application programming interface,
  ULAPI, for portable communication between user-level processes and 
  real-time tasks conforming to the \ref RTAPI real-time API.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.6 $
  $Date: 2009/04/16 19:17:38 $
  \endcode

  \author Fred Proctor
*/

/*!
  \defgroup ULAPI User-Level API

  The User-Level Application Programming Interface (ULAPI) provides a
  portable interface for communicating with real-time tasks conforming
  to the \ref RTAPI real-time API.
*/

#ifndef ULAPI_H
#define ULAPI_H

#ifdef __cplusplus
extern "C" {
#if 0
} /* just to match one above, for indenters */
#endif
#endif

#include <stdio.h>		/* printf */

/* make sure we have enough string space to print numbers as strings */
#define DIGITS_IN(x) (sizeof(x) * 3)

#if defined(ULAPI_RESULT_CHAR)
typedef char ulapi_result;
#else
typedef int ulapi_result;
#endif

#if defined(ULAPI_INTEGER_LONG)
typedef long int ulapi_integer;
#else
typedef int ulapi_integer;
#endif

#if defined(ULAPI_ID_LONG)
typedef long int ulapi_id;
#else
typedef int ulapi_id;
#endif

#if defined(ULAPI_FLAG_INT)
typedef int ulapi_flag;
#else
typedef char ulapi_flag;
#endif

#if defined(ULAPI_REAL_FLOAT)
typedef float ulapi_real;
#elif defined(ULAPI_REAL_LONG_DOUBLE)
typedef long double ulapi_real;
#else
typedef double ulapi_real;
#endif

enum {
  UL_USE_DEFAULT = 0,
  UL_USE_UNIX,
  UL_USE_RTAI
};

extern ulapi_real ulapi_time(void);

extern void ulapi_sleep(ulapi_real secs);

enum {
  ULAPI_OK = 0,
  ULAPI_ERROR,
  ULAPI_IMPL_ERROR,
  ULAPI_BAD_ARGS
};

/*!
  Call this once before any other ULAPI functions, passing
  UL_USE_DEFAULT for the default ULAPI implementation, or one of the
  platform-specific values like UL_USE_UNIX or UL_USE_RTAI to force an
  implementation, if available. Returns ULAPI_OK if successful,
  otherwise ULAPI_RESULT_IMPL_ERROR if the implementation specified by
  \a sel is not available, or ULAPI_RESULT_BAD_ARGS if \a sel is not
  recognized.
*/
extern ulapi_result ulapi_init(ulapi_integer sel);
extern ulapi_result ulapi_exit(void);

extern ulapi_integer ulapi_to_argv(char *str, char ***argv);
extern void ulapi_free_argv(ulapi_integer argc, char **argv);

/*!
  Set and get the debug level.
*/
enum {ULAPI_DEBUG_WARN = 0x01,
      ULAPI_DEBUG_ERROR = 0x02,
      ULAPI_DEBUG_ALL = 0xFF};

extern void ulapi_set_debug(ulapi_integer mask);

/*!
  As with RTAPI, priorities range from ulapi_prio_lowest() to
  ulapi_prio_highest(), inclusive. To use this API, use one of two
  methods:

  Set your lowest priority task to ulapi_prio_lowest(), and for each
  task of the next lowest priority, set their priorities to
  ulapi_prio_next_higher(previous).

  Or,

  Set your highest priority task to ulapi_prio_highest(), and for each
  task of the next highest priority, set their priorities to
  ulapi_prio_next_lower(previous).
*/

#if defined(ULAPI_PRIO_CHAR)
typedef char ulapi_prio;
#else
typedef int ulapi_prio;
#endif

extern ulapi_prio ulapi_prio_highest(void);
extern ulapi_prio ulapi_prio_lowest(void);
extern ulapi_prio ulapi_prio_next_higher(ulapi_prio prio);
extern ulapi_prio ulapi_prio_next_lower(ulapi_prio prio);

/*!
  Allocates space for a platform-specific data structure that holds
  the task information. Pass this to the \a ulapi_task_ functions.
*/
extern void *ulapi_task_new(void);
extern ulapi_result ulapi_task_delete(void *task);

typedef void (*ulapi_task_code)(void *);

extern ulapi_result ulapi_task_start(void *task,
				     void (*taskcode)(void *),
				     void *taskarg,
				     ulapi_prio prio,
				     ulapi_integer period_nsec);

extern ulapi_result ulapi_task_stop(void * task);
extern ulapi_result ulapi_task_pause(void * task);
extern ulapi_result ulapi_task_resume(void * task);
extern ulapi_result ulapi_task_set_period(void * task,
					  ulapi_integer period_nsec);
extern ulapi_result ulapi_task_init(void);
extern ulapi_result ulapi_self_set_period(ulapi_integer period_nsec);
extern ulapi_result ulapi_wait(ulapi_integer period_nsec);
extern ulapi_result ulapi_task_exit(void);
extern ulapi_result ulapi_task_join(void * task);
extern ulapi_integer ulapi_task_id(void);

/*!
  Returns a pointer to an implementation-defined structure that
  is passed to the other mutex functions, or NULL if no mutex can
  be created.
*/
extern void * ulapi_mutex_new(ulapi_id key);

/*! Deletes the mutex. */
extern ulapi_result ulapi_mutex_delete(void * mutex);

/*| Releases the mutex, signifying that the associated shared resource 
  is now free for another task to take. */
extern ulapi_result ulapi_mutex_give(void * mutex);

/*! Takes the mutex, signifying that the associated shared resource
  will now be used by the task. If the mutex is already taken, this
  blocks the caller until the mutex is given. */
extern ulapi_result ulapi_mutex_take(void * mutex);

/*!
  Returns a pointer to an implementation-defined structure that is
  passed to the other condition variable functions, or NULL if no
  condition variable can be created.
*/
extern void * ulapi_cond_new(ulapi_id key);

/*! Deletes the condition variable. */
extern ulapi_result ulapi_cond_delete(void * cond);

/*| Signals that the condition variable has reached its release value */
extern ulapi_result ulapi_cond_signal(void * cond);

/*| Broadcasts that the condition variable has reached its release value */
extern ulapi_result ulapi_cond_broadcast(void * cond);

/*! Waits until the condition variable has reached its release value */
extern ulapi_result ulapi_cond_wait(void * cond, void * mutex);

/*!
  Allocates space for a platform-specific data structure that holds
  the shared memory configuration. Pass this to \a ulapi_shm_addr
  to get a pointer to the actual shared memory.
*/
extern void * ulapi_shm_new(ulapi_id key, ulapi_integer size);
/*!
  Returns a pointer to the actual shared memory, given a shared memory
  data structure previously created with \a ulapi_shm_new.
 */
extern void * ulapi_shm_addr(void * shm);
/*!
  Deletes shared memory previously allocated with \a ulapi_shm_new.
 */
extern ulapi_result ulapi_shm_delete(void * shm);

extern ulapi_result ulapi_fifo_new(ulapi_integer key, ulapi_integer * fd,
				   ulapi_integer size);
extern ulapi_result ulapi_fifo_delete(ulapi_integer key, ulapi_integer fd,
				      ulapi_integer size);
extern ulapi_integer ulapi_fifo_write(ulapi_integer fd, char * buf,
				    ulapi_integer size);
extern ulapi_integer ulapi_fifo_read(ulapi_integer fd, char * buf,
				   ulapi_integer size);

/*!
  Connects as a client to the socket server on \a port and \a host.
  Returns the integer socket descriptor for later sends and receives.
 */
extern ulapi_integer ulapi_socket_client(ulapi_integer port, char * host);

/*!
  Creates a server connection to the \a port. Returns the integer
  socket descriptor for later use in \a
  ulapi_socket_get_client_connection.
*/
extern ulapi_integer ulapi_socket_server(ulapi_integer port);

/*!
  Gets a connection from a client. Returns the integer socket descriptor
  for later sends and receives.
*/
extern ulapi_integer ulapi_socket_get_client(ulapi_integer id);

/*!
  Sets the socket to be blocking or nonblocking.
*/
extern ulapi_result ulapi_socket_set_nonblocking(ulapi_integer id);
extern ulapi_result ulapi_socket_set_blocking(ulapi_integer id);

/*!
  Reads up to \a len bytes from socket \a id into \a buf. Returns the
  number of bytes read, or -1 on error.
 */
extern ulapi_integer ulapi_socket_read(ulapi_integer id, char * buf, ulapi_integer len);

/*!
  Writes \a len bytes from \a buf to socket \a id. Returns the
  number of bytes written, or -1 on error.
 */
extern ulapi_integer ulapi_socket_write(ulapi_integer id, char * buf, ulapi_integer len);

/*!
  Closes the socket id, whether that for a client, for a server, or
  to a client.
 */
extern ulapi_result ulapi_socket_close(ulapi_integer id);

/*
  File descriptor (fd) API
*/

typedef enum {
  ULAPI_STDIN,
  ULAPI_STDOUT,
  ULAPI_STDERR
} ulapi_stdio;

/*! Allocates space for a file identifier, and returns its pointer. */
extern void *
ulapi_fd_new(void);

/*! Deallocates a previously allocated file identifier. */
extern ulapi_result
ulapi_fd_delete(void * id);

/*! Copies the stdio file descriptor into \a id. */
extern ulapi_result
ulapi_std_open(ulapi_stdio io,	/*!< one of ULAPI_STDIN,STDOUT,STDERR */
	       void * id);	/*!< pointer to fd where it will be copied */

/*!
  Opens a connection to a file. Fills in the identifier and
  returns ULAPI_OK if it worked, otherwise returns ULAPI_ERROR.
 */
extern ulapi_result
ulapi_fd_open(char * port, void * id);

/*!
  Sets the file descriptor \a id to be blocking or nonblocking.
*/
extern ulapi_result
ulapi_fd_set_nonblocking(void * id);

extern ulapi_result
ulapi_fd_set_blocking(void * id);

/*!
  Reads up to \a len bytes from file descriptor \a id into \a
  buf. Returns the number of bytes read, or -1 on error.
 */
extern ulapi_integer
ulapi_fd_read(void * id, char * buf, ulapi_integer len);

/*!
  Writes \a len bytes from \a buf to file descriptor \a
  id. Returns the number of bytes written, or -1 on error.
 */
extern ulapi_integer
ulapi_fd_write(void * id, char * buf, ulapi_integer len);

/*!
  Closes the file descriptor.
 */
extern ulapi_result
ulapi_fd_close(void * id);

/*
  Serial communication API, similar to file descriptor API
*/

/*! Allocates space for a serial port identifier, and returns its pointer. */
extern void *
ulapi_serial_new(void);

/*! Deallocates a previously allocated serial port identifier. */
extern ulapi_result
ulapi_serial_delete(void * id);

/*!
  Opens a connection to a serial port. Fills in the identifier and
  returns ULAPI_OK if it worked, otherwise returns ULAPI_ERROR.
 */
extern ulapi_result
ulapi_serial_open(char * port, void * id);

/*!
  Sets the serial port descriptor \a id to be blocking or nonblocking.
*/
extern ulapi_result
ulapi_serial_set_nonblocking(void * id);

extern ulapi_result
ulapi_serial_set_blocking(void * id);

/*!
  Reads up to \a len bytes from serial port descriptor \a id into \a
  buf. Returns the number of bytes read, or -1 on error.
 */
extern ulapi_integer
ulapi_serial_read(void * id, char * buf, ulapi_integer len);

/*!
  Writes \a len bytes from \a buf to serial port descriptor \a
  id. Returns the number of bytes written, or -1 on error.
 */
extern ulapi_integer
ulapi_serial_write(void * id, char * buf, ulapi_integer len);

/*!
  Closes the serial port descriptor.
 */
extern ulapi_result
ulapi_serial_close(void * id);

#define ULAPI_DECL_SHARED
#ifdef WIN32
#undef ULAPI_DECL_SHARED
#define ULAPI_DECL_SHARED __declspec(dllexport)
#endif

/*!
  Opens a shared library \a objname and returns a pointer to it, for
  subsequent reference by ulapi_dl_sym and ulapi_dl_close. If the
  object cannot be opened, NULL is returned. If \a errstr is not NULL,
  up to \a errlen characters of a descriptive error string is copied
  into it.
*/
extern void *ulapi_dl_open(char *objname, char *errstr, int errlen);

/*!
  Closes a shared library handle previously returned by ulapi_dl_open.
*/
extern void ulapi_dl_close(void *handle);

/*!
  Returns a pointer to a shared object (e.g., a function) with symbol name
  \a name. The shared object could be NULL, so to detect an error, pass
  a pointer to an error string \a errstr, and its length \a errlen, and
  check the string against an empty string.
*/
extern void *ulapi_dl_sym(void *handle, char *name, char *errstr, int errlen);

#define ulapi_print printf
#ifdef WIN32
#define ulapi_snprintf sprintf_s
#else
#define ulapi_snprintf snprintf
#endif

#ifdef __cplusplus
#if 0
{			  /* just to match one below, for indenters */
#endif
}
#endif

#endif /* ULAPI_H */

/*
  Modification history:

  $Log: ulapi.h,v $
  Revision 1.6  2009/04/16 19:17:38  proctor
  Updated the NML config files to reflect the simware changes; updated ulapi
  to be consistent with the master files in gomotion; added targets for the
  new simware

  Revision 1.23  2009/03/16 18:59:38  proctor
  Added ulapi_snprintf

  Revision 1.22  2009/03/13 21:29:24  proctor
  Added to_argv utility

  Revision 1.21  2009/02/24 15:59:54  proctor
  Factored out dl_xxx funcs into ulapi, and added dltest,funcs to the Win32
  project

  Revision 1.20  2009/02/24 13:53:07  proctor
  Added ulapi_dl_XXX functions

  Revision 1.19  2009/01/13 16:23:11  proctor
  Fixed problem with casting function pointers to void pointers

  Revision 1.18  2008/12/08 17:40:50  proctor
  Added ultest.c and ulapi_cond_xxx condition variables

  Revision 1.17  2008/12/02 15:55:09  proctor
  Added ulapi_task_join

  Revision 1.16  2008/02/22 18:39:42  proctor
  Reworked serial interface to allocate a pointer to the id

  Revision 1.15  2008/02/21 16:35:01  proctor
  Added rt/ulapi_serial functions, and smartmotor build

  Revision 1.14  2008/01/09 20:31:59  proctor
  Added xinv logging, threads in igps client

  Revision 1.13  2007/12/21 19:47:26  proctor
  Added nonblocking functions to ulapi; made igpsserver really get position
  from a controller

  Revision 1.12  2007/12/07 14:34:55  proctor
  Added socket code to ulapi, and stubbed out IGPS application

  Revision 1.11  2007/10/02 20:22:42  proctor
  Fixed bug in trajloop's walk_in where zero vectors were converted into
  unit vectors; added ulapi_mutex functions

  Revision 1.10  2007/09/21 19:28:33  proctor
  Pulled the SDL joystick code stubs into gotk.c, and reworked ulapi
  so that task/thread functions were available.

  Revision 1.9  2007/02/08 18:50:58  proctor
  Added getopt-style arg handling, and a select interface to ulapi

  Revision 1.8  2006/12/13 18:42:31  proctor
  Changed etime,esleep to ulapi_time,ulapi_sleep to avoid conflicts with rcslib

  Revision 1.7  2006/12/11 21:30:17  proctor
  Fleshed out win32 apis

  Revision 1.6  2006/12/06 18:10:34  proctor
  Reworked shared memory interface to use a malloc'ed structure

  Revision 1.5  2006/10/20 21:27:48  proctor
  Added a bunch of doxygen documentation.

  Revision 1.4  2006/10/05 13:56:12  proctor
  Reworked rtapi mutex, sem and fifo to conform to pointer argument convention

  Revision 1.3  2006/09/27 15:29:47  proctor
  Changed rtapi_count to rtapi_int

  Revision 1.2  2006/08/15 18:30:59  proctor
  Changed conflicting go_rcs.h types such as RCS_ERROR to non-conflicting
  types such as GO_RCS_STATUS_ERROR, and recoded all the affected files
  of which there were many. Also put a global_ prefix on the traj_comm_ptr,
  servo_comm_ptr structures to make this obvious. Added extern "C" in
  ulapi.h so that C++ code can connect to ulapi.

  Revision 1.1  2006/05/19 13:04:21  proctor
  Initial checkin

  Revision 1.1.1.1  2006-02-10 21:15:15  proctor
  Initial import

*/
