/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file dijkHeapN.cc
    \brief implements Dijkstra graph search with open list managed by heap

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.2 $
  $Date: 2006/10/30 16:31:36 $
  \endcode

  \author Tom Kramer

See the documentation of dijkHeapN.hh for an overview of this search routine.

*/

#include "dijkHeapN.hh"
#include <stdio.h>

/***********************************************************************/
/***********************************************************************/

/* DijkHeapNGraph functions */

/***********************************************************************/

DijkHeapNGraph::~DijkHeapNGraph()
{
  int n;

  if (deleat == true)
    {
      for (n = 0; n < capacity; n++)
	{
	  if (nodes[n].kidsSize > 0)
	    {
	      delete [] nodes[n].kidsIndexes;
	      delete [] nodes[n].kidsCosts;
	    }
	}
      delete [] nodes;
    }
}

/***********************************************************************/

//! DijkHeapNGraph::addKids adds kids to the dad node
/*
Returned Value: If an error occurs, this returns 0. Otherwise, this
  returns 1. An error occurs if:
  1. Any of numberAdded, numberNew, and numberOld is negative.
  2. In any of the newInfos, the index is incorrect.

Called By:
  external functions (such as main in exampleDijkHeapN.cc)

This may activate additional nodes of the graph. If it does, it also
adds arcs from dad to those new nodes. This may also add arcs from dad
to existing nodes. The memory for dad's old kidsIndexes and kidsCosts
is freed. Memory for dad's revised kidsIndexes and kidsCosts is
allocated.

If there are not enough nodes in the graph to accommodate the number
to be activated, increaseCapacity is called to make more nodes.

The new kids are built in nodes whose indexes are assigned in
numerical order, starting with the current size of the graph. The size
is increased in the next-to-last "for" loop. This number should
already be recorded as the value of "index" in the NodeInfo. It is
checked that this number is correct.

The additional kids in oldAdded must not already be kids of dad.
This is not currently checking for that. A check could be added.

*/

int DijkHeapNGraph::addKids( /* ARGUMENTS                                */
 int dad,               /* index in nodes of node to which to add kids   */
 int numberAdded,       /* total number of kids added                    */
 int numberNew,         /* number of new kids added                      */
 NodeInfo ** newInfos,  /* array of pointers to NodeInfos for new kids   */
 double * newCosts,     /* array of arc costs from dad for new kids      */
 int * oldAdded,        /* number of existing kids added                 */
 double * oldCosts)     /* array of arc costs from dad for existing kids */
{
  int numberOld;
  int saveKidsSize;
  int * saveIndexes;
  double * saveCosts;
  int n;
  int k;

  numberOld = (numberAdded - numberNew);
  if ((numberAdded <= 0) || (numberNew < 0) || (numberOld < 0))
    {
      fprintf(stderr, "negative number of nodes found in addKids\n");
      return 1;
    }
  if ((size + numberNew) > capacity)
    increaseCapacity(numberNew);
  if ((saveKidsSize = nodes[dad].kidsSize)) // set and test if non-zero
    {
      saveIndexes = nodes[dad].kidsIndexes;
      saveCosts = nodes[dad].kidsCosts;
    }
  for (n = 0; n < numberNew; n++)
    {
      if (newInfos[n]->index != (size + n))
	{
	  fprintf(stderr, "incorrect node index found in addKids\n");
	  return 1;
	}
      nodes[size + n].moreInfo = newInfos[n];
      nodes[size + n].open = DijkHeapNNode::UNOPENED;
    }
  nodes[dad].kidsSize = (saveKidsSize + numberAdded);
  nodes[dad].kidsIndexes = new int[nodes[dad].kidsSize];
  nodes[dad].kidsCosts = new double[nodes[dad].kidsSize];
  for (n = 0; n < saveKidsSize; n++)
    {
      nodes[dad].kidsIndexes[n] = saveIndexes[n];
      nodes[dad].kidsCosts[n] = saveCosts[n];
    }
  for (k = 0; k < numberNew; k++, n++)
    {
      nodes[dad].kidsIndexes[n] = size++;
      nodes[dad].kidsCosts[n] = newCosts[k];
    }
  for (k = 0; k < numberOld; k++, n++)
    {
      nodes[dad].kidsIndexes[n] = oldAdded[k];
      nodes[dad].kidsCosts[n] = oldCosts[k];
    }
  if (saveKidsSize)
    {
      delete [] saveIndexes;
      delete [] saveCosts;
    }
  return 0;
}

/***********************************************************************/

//!DijkHeapNGraph::increaseCapacity increases capacity of a graph's nodes array
/*
Returned Value: int (0)

Called By:
  DijkHeapNGraph::addKids

This function makes a new and larger array of nodes and copies the old
nodes into the front part of it. Since all array referencing is done
by index rather than by pointer, it does not matter that new memory is
being used.

The increase in capacity is the maximum of (1) the numberAdded and
(2) the minimum of the old capacity and 100000. The idea is to double
in capacity (or more, if necessary) any time an increase is needed,
until there are 100000 nodes, after which only 100000 nodes are added
each time (so that a significant amount of memory is not wasted). The
reason for doubling is to minimize the number of increases in capacity.

*/

int DijkHeapNGraph::increaseCapacity( /* ARGUMENTS                      */
 int numberAdded)     /*!< minimum number by which to increase capacity */
{
  int increase;
  DijkHeapNNode * oldNodes;
  int n;

  oldNodes = nodes;
  increase = ((capacity > 100000) ? 100000 : capacity);
  if (numberAdded > increase)
    increase = numberAdded;
  capacity = (capacity + increase);
  nodes = new DijkHeapNNode[capacity];
  for (n = 0; n < size; n++)
    nodes[n].copyFrom(oldNodes + n);
  delete [] oldNodes;
  return 0;
}

/***********************************************************************/

/* DijkHeapNGraph::removeKids

Returned Value: int (1)

Called By: external functions

This function removes all arcs from nodes[dad] by setting kidsSize to
0 and deleting kidsIndexes and kidsCosts.

kidsIndexes and kidsCosts must have been made using "new", since they
are deleted here.

The kids themselves are not affected by this function.

*/

int DijkHeapNGraph::removeKids( /* ARGUMENTS                                */
 int dad)             /*!< index in nodes of node from which to remove kids */
{
  nodes[dad].kidsSize = 0;
  delete [] nodes[dad].kidsIndexes;
  delete [] nodes[dad].kidsCosts;
  nodes[dad].kidsIndexes = 0;
  nodes[dad].kidsCosts = 0;
  return 1;
}

/***********************************************************************/

/* DijkHeapNGraph::useKids

Returned Value: int (1)

Called By: external functions

This removes the current kids of dad (if any) and replaces them with
the ones given by the arguments. The arrays specified in the arguments
are used as is, so they must be created with "new" in the implementing
system and not changed or deleted by the implementing system.

The kids themselves must already exist and are not affected by this
function.

*/

int DijkHeapNGraph::useKids( /* ARGUMENTS                                 */
 int dad,                    /* index in nodes of node to change kids for */
 int kidsSize,               /* new number of kids                        */
 int * kidsIndexes,          /* new array of kids indexes in nodes        */
 double * kidsCosts)         /* new array of kids costs                   */
{
  if (nodes[dad].kidsSize > 0)
    {
      delete [] nodes[dad].kidsIndexes;
      delete [] nodes[dad].kidsCosts;
    }
  nodes[dad].kidsSize = kidsSize;
  nodes[dad].kidsIndexes = kidsIndexes;
  nodes[dad].kidsCosts = kidsCosts;
  return 1;
}

/***********************************************************************/
/***********************************************************************/

/* DijkHeapNNode functions */

/***********************************************************************/

/* DijkHeapNNode::~DijkHeapNNode

This destructor (which does nothing in addition to default destruction)
is defined here only because objectcenter thought it was undefined
when it was in dijkHeapN.hh. I

*/

DijkHeapNNode::~DijkHeapNNode() {}

/***********************************************************************/

/* DijkHeapNNode::copyFrom

Returned Value: int
  If the "from" argument is 0, this returns 0. Otherwise, it returns 1.

Called By:
  DijkHeapNGraph::increaseCapacity

This copies all of the data of the "from" node into "this" node.  The
kidsIndexes and kidsCosts arrays are used as given in the "from" node,
not made with "new".

*/

int DijkHeapNNode::copyFrom( /*   ARGUMENTS         */
 DijkHeapNNode * from)       /*!< node to copy from */
{
  if (from != 0)
    {
      moreInfo = from->moreInfo;
      kidsSize = from->kidsSize;
      kidsIndexes = from->kidsIndexes;
      kidsCosts = from->kidsCosts;
      open = from->open;
      arrivedFrom = from->arrivedFrom;
      visitCost = from->visitCost;
      heapIndex = from->heapIndex;
    }
  return ((from == 0) ? 0 : 1);
}

/***********************************************************************/
/***********************************************************************/

/* DijkHeapNSearcher functions */

/***********************************************************************/

//! DijkHeapNSearcher::buildGoalsMap builds and populates the goals map
/*
Returned Value: none

Called By: DijkHeapNSearcher::initSearch

This: finds the size of the goalsMap, builds the goalsMap, and populates it.

*/

void DijkHeapNSearcher::buildGoalsMap() /* NO ARGUMENTS */
{
  int n;

  goalsMapSize = 0;
  for (n = 0; n < numGoals; n++)
    {
      if (goals[n] > goalsMapSize)
	goalsMapSize = goals[n];
    }
  goalsMapSize++;
  goalsMap = new int[goalsMapSize];
  for (n = 0; n < goalsMapSize; n++)
    goalsMap[n] = -1;
  for (n = 0; n < numGoals; n++)
    goalsMap[goals[n]] = n;
}

/***********************************************************************/

/* DijkHeapNSearcher::closeSearch

Returned Value: int (0)

Called By: external functions (e.g., main in exampleDijkHeapN.cc)

This closes the searcher. All it does is set the heap size to zero,
and even this is not really necessary.

*/

int DijkHeapNSearcher::closeSearch() /* NO ARGUMENTS */
{
  heapSize = 0;
  return 0;
}

/***********************************************************************/

/*  DijkHeapNSearcher::cycleSearch

Returned Value: int (0)

Called By: external functions (e.g., main in exampleDijkHeapN.cc)

This removes the first (index 1) node on the heap (which is the
cheapest one) and processes the children of that node.

1. If a path to goals[n] known to be the cheapest path has been found
for some n (this may happen for only one n), this sets statuses[n] to 2,
sets costs[n], adds 1 to optimaFound, and, if statuses[n] was 0,
adds 1 to goalsFound.

2. Otherwise, if a path to goals[n] not known to be the cheapest has been
found for some n (this may happen for several n):
  a. If statuses[n] is 0, this sets statuses[n] to 1, sets costs[n],
     and adds 1 to goalsFound.
  b. Otherwise, if statuses[n] is 1, this sets costs[n], which may
     or may not be less than previously (but will not be more).

3. Otherwise, this does not change the output data.

Once all cheapest paths to the goals have been found or the heap
is empty, this does nothing.

*/

int DijkHeapNSearcher::cycleSearch() /* NO ARGUMENTS */
{
  int i;           // index for kids of oldCheap
  double kidCost;  // total cost to visit i'th kid of oldCheap
  int kidNum;      // index in nodes of i'th kid of oldCheap
  int numKids;     // number of kids of oldCheap
  int oldCheap;    // index of cheap node when this starts
  int goalsIndex;  // index in goals of a node, or -1
  
  if ((numGoals != *optimaFound) && (heapSize != 0))
    {
      if (heapCapacity < theGraph->capacity)
	{
	  nodes = theGraph->nodes;         // has been changed by addKids
	  increaseHeap(theGraph->capacity);
        }
      oldCheap = heap[1];
      removeCheap();
      nodes[oldCheap].open = DijkHeapNNode::CLOSED; // old cheap node closed
      goalsIndex = ((oldCheap < goalsMapSize) ? goalsMap[oldCheap] : -1);
      if (goalsIndex > -1)
	{
	  if (statuses[goalsIndex] == 0)
	    (*goalsFound)++;
	  statuses[goalsIndex] = 2;
	  costs[goalsIndex] = nodes[oldCheap].visitCost;
	  (*optimaFound)++;
	}
      numKids = nodes[oldCheap].kidsSize;
      for (i = 0; i < numKids; i++)
	{
	  if ((nodes[oldCheap].kidsCosts)[i] < 0)
	    continue;
	  kidNum = (nodes[oldCheap].kidsIndexes)[i];
	  kidCost = (nodes[oldCheap].visitCost +
		     (nodes[oldCheap].kidsCosts)[i]);
	  processNode(kidNum, oldCheap, kidCost);
	  goalsIndex = ((kidNum < goalsMapSize) ? goalsMap[kidNum] : -1);
	  if ((goalsIndex > -1) && 
	      (nodes[kidNum].open != DijkHeapNNode::CLOSED))
	    {
	      costs[goalsIndex] = nodes[kidNum].visitCost;
	      if (statuses[goalsIndex] == 0) // status not 2 since open
		{
		  statuses[goalsIndex] = 1;
		  (*goalsFound)++;
		}
	    }
	}
    }
  return 0;
}

/***********************************************************************/

/* DijkHeapNSearcher::getNextNode

Returned Value: int
This returns the index in nodes of the cheapest open node remaining.
If there are no open nodes remaining, this returns -1.

Called By: external functions

*/

int DijkHeapNSearcher::getNextNode() /* NO ARGUMENTS */
{
  return ((heapSize == 0) ? -1 : heap[1]);
}

/***********************************************************************/

//! DijkHeapNSearcher::heapDown causes misplaced node to sink to correct place
/* 
Returned Value: None

Called By:  DijkHeapNSearcher::removeCheap

Conceptually, the heap consists of nodes partially ordered by cost and
linked in a binary tree. In the (standard) representation used here,
however, the nodes are kept in an array called Nodes, and the heap is
a linear array of indexes in Nodes. The binary tree is implicitly
linked by having the kids of the node pointed to by heap[N] be pointed
to by heap[2N] and heap[2N+1].

The conceptual idea of the algorithm is simple: the node that was at
the top of the heap sinks by changing positions with the cheaper of
its kids repeatedly until neither kid is cheaper than it is. Because
the representation differs greatly from the concept, understanding
how the implementation works is tricky.

In order to find a node in the heap easily, the heap index which
stores a node's index is recorded in the heapIndex attribute of the
node. This must be updated whenever the node's position in the heap
changes.

The three loop variables are heap indexes:
hex (a heap index currently being examined) - conceptually a dad node
kid1 (2*hex) - conceptually one kid node
kid2 (2*hex+1) - conceptually the other kid node

The initial value of hex is 1, the heap index of the top of the heap.

Each time around the loop:

1. If the cost of the original top node is larger than the smaller of
the costs of the two kids (let kidC be the heap index of the cheaper
kid):
  A. The value in heap[hex] is changed to the value in heap[kidC].
     This represents the kid node moving to where the dad was.
  B. The heapIndex recorded in nodes[heap[kidC]] is changed to hex.
     This updates the node's knowledge of where it is in the heap.
  C. hex is changed to kidC to get ready for the next go around the loop.
2. Otherwise, the original top node has stopped sinking, so:
  A. The value in heap[hex] is set to the Nodes index of the original top node.
  B. The heapIndex recorded in the original top node is set to hex.
  C. The function returns.

The loop terminates if there are no kids or only one kid.  After the
loop exits, a check is made for only one kid existing and a swap is
made if appropriate.

Although the representation of a kid's position is changed each time a
dad sinks, the representation of the dad's new position is not changed
each time. We know the dad will be where the kid was, and we know the
value recorded in the kid's former position would be the Nodes index
of the node that started at the top of the heap, so we don't have to
record it. When the dad stops sinking, its final position is recorded.
Not recording the intermediate positions saves time but makes the code
a little more complex.

*/

void DijkHeapNSearcher::heapDown() /* NO ARGUMENTS */
{
  int hex;  // index in heap of upper node (changed in loop)
  int kid1; // (2*hex) = heap index of left lower node (changed in loop)
  int kid2; // (kid1+1) = heap index of right lower node (changed in loop)
  double cost0; // visit_cost of node at top of heap (fixed)
  double cost1; // visit_cost of kid1 (changed in loop)
  double cost2; // visit_cost of kid2 (changed in loop)
  int topNodeIndex; // Nodes index of node at top of heap (fixed)

  topNodeIndex = heap[1];
  cost0 = nodes[topNodeIndex].visitCost;

  for (hex = 1; ((kid1 = (2 * hex)) < heapSize); ) // always 2 kids
    {
      kid2 = (kid1 + 1);
      cost1 = nodes[heap[kid1]].visitCost;
      cost2 = nodes[heap[kid2]].visitCost;
      if (cost1 < cost2) // kid1 cheaper
	{
	  if (cost0 > cost1)
	    { // original top node must sink more
	      heap[hex] = heap[kid1];
	      nodes[heap[kid1]].heapIndex = hex;
	      hex = kid1;
	    }
	  else
	    { // original top node goes in hex position
	      heap[hex] = topNodeIndex;
	      nodes[topNodeIndex].heapIndex = hex;
	      return;
	    }
	}
      else // if (cost2 <= cost1) kid2 cheaper
	{
	  if (cost0 > cost2)
	    { // original top node must sink more
	      heap[hex] = heap[kid2];
	      nodes[heap[kid2]].heapIndex = hex;
	      hex = kid2;
	    }
	  else
	    { // original top node goes in hex position
	      heap[hex] = topNodeIndex;
	      nodes[topNodeIndex].heapIndex = hex;
	      return;
	    }
	}
    }
  if (kid1 == heapSize)    // only one kid
    {
      if (cost0 > nodes[heap[kid1]].visitCost) // top falls to bottom
	{ // swap hex and kid1
	  nodes[heap[kid1]].heapIndex = hex;
	  nodes[topNodeIndex].heapIndex = kid1;
	  heap[hex] = heap[kid1];
	  heap[kid1] = topNodeIndex;
	}
      else // no swap, but record dad position
	{
	  heap[hex] = topNodeIndex;
	  nodes[topNodeIndex].heapIndex = hex;
	}
    }
  else // no kids, top has fallen to bottom, record dad position
    {
      heap[hex] = topNodeIndex;
      nodes[topNodeIndex].heapIndex = hex;
    }
}

/***********************************************************************/

//! DijkHeapNSearcher::heapUp causes misplaced node to rise to correct place
/*
Returned Value:  None

Called By:  DijkHeapNSearcher::processNode

See the documentation of heapDown. This function is simpler.
Conceptually, the kid0 node keeps changing places with its dad node
until the kid0 node is cheaper than the dad node. As in heapUp, the
position of the target node index (heap[kid0] in this case) is not
recorded until it stops moving.

*/

void DijkHeapNSearcher::heapUp( /* ARGUMENTS                           */
 int kid0)                      /*!< heap index of element being upped */
{
  int kid;           // heap index of lower node (changed in loop)
  int dad;           // heap index of upper node (changed in loop)
  int kid0NodeIndex; // Nodes index of original kid (fixed)
  double cost0;      // cost of kid0 (fixed)
  
  kid0NodeIndex = heap[kid0];
  cost0 = nodes[kid0NodeIndex].visitCost;
  for (kid = kid0; ((dad = (kid / 2)) > 0); kid = dad)
    {
      if (cost0 < nodes[heap[dad]].visitCost)
	{ // dad sinks down
	  heap[kid] = heap[dad];
	  nodes[heap[dad]].heapIndex = kid;
	}
      else 
	{ // kid has stopped rising, record position
	  heap[kid] = kid0NodeIndex;
	  nodes[kid0NodeIndex].heapIndex = kid;
	  return;
	}
    }
  // if this point is reached, the node being upped is at the top
  heap[1] = kid0NodeIndex;
  nodes[kid0NodeIndex].heapIndex = 1;
}

/***********************************************************************/

//! DijkHeapNSearcher::increaseHeap increases the capacity of the heap
/*
Returned Value: None

Called By: DijkHeapNSearcher::cycleSearch

This makes a new heap with the given newCapacity and copies the old
heap into it. 

newCapacity must be larger than current capacity. This function is not
checking that.

This leaves garbage at the end of the new heap; that does not matter.

It will never happen that every node is on the heap. If it could
happen, the capacity of the heap would need to be one greater than the
number of nodes, since heap[0] is a dummy.

*/

void DijkHeapNSearcher::increaseHeap( /* ARGUMENTS              */
 int newCapacity)                     /*!< new capacity of heap */
{
  int * oldHeap;
  int n;

  oldHeap = heap;
  heapCapacity = newCapacity;
  heap = new int[newCapacity];
  for (n = 0; n <= heapSize; n++)
    heap[n] = oldHeap[n];
  delete [] oldHeap;
}

/***********************************************************************/

//! DijkHeapNSearcher::initSearch initializes the searcher
/*
Returned Value: int
If initialization occurs without error, this returns 0;
This prints an error message and returns 1 if:
1. The nodes of aGraph is NULL, or
2. the size of the kids of any node is negative, or
3. the size of aGraph is greater than its capacity.

Called By: external functions (e.g., main in exampleDijkHeapN.cc)

The arrivedFrom of the start node is set to -1 to serve as
a stop marker for reportSearch.

This:
1. Saves (as the values of data members of the Searcher) the locations
   where the user wants to have search data recorded. These are the
   four arguments: statusesOut, costsOut, goalsFoundOut, optimaFoundOut.
2. Sets the value of the other data members of the Searcher that need setting.
3. Builds the goalsMap.
4. Initializes the start node and places it on the open nodes list.

*/

int DijkHeapNSearcher::initSearch( /* ARGUMENTS                         */
 DijkHeapNGraph * aGraph, /* the graph to search in                     */
 int start,               /* index in nodes of aGraph of start node     */
 int numGoalsIn,          /* number of goals                            */
 int * goalsIn,           /* indexes in nodes of aGraph of goal nodes   */
 int * opensOut,          /* 0=open list empty, non-zero=list not empty */
 int * statusesOut,       /* statuses of reaching goals                 */
 double * costsOut,       /* costs of reaching goals                    */
 int * goalsFoundOut,     /* number of goals reached                    */
 int * optimaFoundOut)    /* number of goals reached optimally          */
{
  int n;

  numGoals = numGoalsIn;
  goals = goalsIn;
  heapCapacity = aGraph->capacity;
  heap = new int[heapCapacity];
  heap[0] = -1;
  heapSize = 0;
  heapSizeOut = opensOut;
  *heapSizeOut = heapSize;
  statuses = statusesOut;
  costs = costsOut;
  for (n = 0; n < numGoals; n++)
    {
      statuses[n] = 0;
      costs[n] = 0;
    }
  goalsFound = goalsFoundOut;
  *goalsFound = 0;
  optimaFound = optimaFoundOut;
  *optimaFound = 0;
  buildGoalsMap();
  theGraph = aGraph;
  nodes = theGraph->nodes;
  if (nodes == 0)
    {
      fprintf(stderr, "nodes array is null");
      return 1;
    }
  if (theGraph->size > theGraph->capacity)
    {
      fprintf(stderr, "size greater than capacity\n");
      return 1;
    }
  for (n = 0; n < theGraph->size; n++)
    {
      if (nodes[n].kidsSize < 0)
	{
	  fprintf(stderr, "bad kids size\n");
	  return 1;
	}
      nodes[n].open = DijkHeapNNode::UNOPENED;
    }
  processNode(start, -1, 0);
  return 0;
}

/***********************************************************************/

//! DijkHeapNSearcher::processNode processes a node (popped off the open list)
/*
Returned Value: None

Called By:
  DijkHeapNSearcher::cycleSearch
  DijkHeapNSearcher::initSearch

This function processes the node (call it Noe) whose index in the
nodes being searched is numNode. Outline numbers here match comment
numbers below.

A. If Noe is open:
  1. If Noe's currently recorded cost is greater than cost:
     a. Set Noe's visitCost to cost.
     b. Set Noe's arrivedFrom to from.
     c. HeapUp from Noe in the open nodes heap.
B. Else if Noe has not yet been opened:
  1. Mark Noe open.
  2. Set Noe's visitCost to cost.
  3. Set Noe's arrivedFrom to from.
  4. Insert Noe at the end of the open nodes heap.
  5. HeapUp from Noe in the open nodes heap.

If Noe is closed (so that nodes[numnode].open == CLOSED), nothing is done.

*/

void DijkHeapNSearcher::processNode( /* ARGUMENTS                           */
 int numNode, /*!< index in nodes of node processed                         */
 int from,    /*!< index in nodes of node from which node processed visited */
 double cost) /*!< total cost of getting to node processed via current path */
{
  if (nodes[numNode].open == DijkHeapNNode::OPEN)          // A
    {
      if (nodes[numNode].visitCost > cost)                 // A1
	{
	  nodes[numNode].visitCost = cost;                 // A1a
	  nodes[numNode].arrivedFrom = from;               // A1b
	  heapUp(nodes[numNode].heapIndex);                // A1c
	}
    }
  else if (nodes[numNode].open == DijkHeapNNode::UNOPENED) // B
    {
      nodes[numNode].open = DijkHeapNNode::OPEN;           // B1
      nodes[numNode].visitCost = cost;                     // B2
      nodes[numNode].arrivedFrom = from;                   // B3
      heapSize++;                                          // B4
      *heapSizeOut = heapSize;
      nodes[numNode].heapIndex = heapSize;
      heap[heapSize] = numNode;
      heapUp(heapSize);                                    // B5
    }
}

/***********************************************************************/

//! DijkHeapNSearcher::removeCheap removes the cheapest node from the heap
/*
Returned Value: None

Called By:
  DijkHeapNSearcher::cycleSearch

This is used to remove the cheapest member of the open nodes heap,
which is at the top of the heap (index 1).

The removal is done by putting the current last element of the heap
into the first place, reducing the size of the heap by one, and calling
heapDown.

*/

void DijkHeapNSearcher::removeCheap() /* NO ARGUMENTS */
{
  heap[1] = heap[heapSize];
  heapSize--;
  *heapSizeOut = heapSize;
  heapDown();
}

/***********************************************************************/

/* DijkHeapNSearcher::reportSearch

Returned Value: int
If the length of the path is less than maxLength, this returns the
length of the path. Otherwise, it returns -1.

Called By: external functions (e.g., main in exampleDijkHeapN.cc)

The arrivedFrom of the start node is set to -1 in initSearch, so
testing for -1 tests for the start node.

This is not reporting the cost because the caller will have direct
access to it.

*/

int DijkHeapNSearcher::reportSearch( /* ARGUMENTS                    */
 int index,      /* index in nodes of node to find path to           */
 int maxLength,  /* maximum length of the path                       */
 int * path)     /* array of index numbers giving the path, set here */
{
  int i;
  int n;
  int length;

  length = 1;
  for (i = index; nodes[i].arrivedFrom != -1; i = nodes[i].arrivedFrom)
    (length)++;
  i = index;
  if( length >= maxLength )
    return -1;
  for (n = (length - 1); n > -1; n--)
    {
      path[n] = i;
      i = nodes[i].arrivedFrom;
    }
  return length;
}

/***********************************************************************/

/*
  $Log: dijkHeapN.cc,v $
  Revision 1.2  2006/10/30 16:31:36  tomrkramer
  Added use of deleat in destructor. Fixed bug in setting heapSizeOut.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.3  2005/07/15 15:47:51  kramer
  removed unneeded include, improved documentation

  Revision 1.2  2005/07/15 12:43:55  kramer
  improved documentation, alphabetized
 
  Revision 1.1  2005/07/14 16:35:18  kramer
  first version
*/
