/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   scrollingMap.hh
  \brief  Contains class method definitions for ScrollingMap
*/

#include <iostream>

#include "scrollingMap.hh"
#define SMAP_TINY 0.00001

/***********************************************************************/

/* SEE scrollingMap.hh FOR AN EXPLANATION OF HOW TO USE THIS FILE. */
/* IT IS NOT INTENDED THAT THIS FILE BE EDITED.                    */

/***********************************************************************/

/* map constructor

This is only place the id of a cell is set.

*/

ScrollingMap::ScrollingMap( /*   ARGUMENTS                               */
 int rangeIn,               /*!< number of cells between center and edge */
 double resolutionIn)       /*!< length of the edge of a cell            */
{
  int n;

  resolution = resolutionIn;
  res = resolution;
  range = rangeIn;
  side = ((range * 2) + 1);
  halfSide = (side / 2.0);
  cols = side;
  rows = side;
  numberOfCells = (rows * cols);
  centerX = 0.0;
  centerY = 0.0;
  originX = 0;
  originY = 0;
  xHi = (resolution * halfSide);
  yHi = (resolution * halfSide);
  xLow = -xHi;
  yLow = -yHi;
  cells = new CELL_NAME[numberOfCells];
  for (n = 0; n < numberOfCells; n++)
    cells[n].id = n;
}

/***********************************************************************/

/* map destructor

This deletes the cells if they were created.

*/

ScrollingMap::~ScrollingMap() /* NO ARGUMENTS */
{
  if (cells)
    delete[] cells;
}

/***********************************************************************/

/*! Public method getCellAtCoords returns a pointer to the cell
  containing the point whose cartesian X and Y coordinates are x and
  y. If the point is not in the window, this returns NULL. */

CELL_NAME * ScrollingMap::getCellAtCoords( /*   ARGUMENTS             */
 double x,                                 /*!< X coordinate of point */
 double y)                                 /*!< Y coordinate of point */
{
  int row;
  int col;

  if ((x < xLow) || (x > xHi) || (y < yLow) || (y > yHi))
    return 0;
  row = ((int)floor((x / resolution) + halfSide) % rows);
  if (row < 0)
    row = (row + rows);
  col = ((int)floor((y / resolution) + halfSide) % cols);
  if (col < 0)
    col = (col + cols);
  return (cells + (row * cols) + col);
}

/***********************************************************************/

/*! Public method getCellAtIndexes returns a pointer to the cell whose
  X and Y indexes in the window grid are row and col. */
/* 

For speed, this is not checking index bounds. The indexes should satisfy
(0 <= row < cols) and (0 <= col < rows).

*/

CELL_NAME * ScrollingMap::getCellAtIndexes( /*   ARGUMENTS              */
 int row,                                   /*!> window X index of cell */
 int col)                                   /*!> window Y index of cell */
{
  return (((originX+row) >= rows) ?
	  (((originY+col) >= cols) ?
	   (cells + ((originX+row - rows) * cols) + originY+col - cols) :
	   (cells + ((originX+row - rows) * cols) + originY+col)) :
	  (((originY+col) >= cols) ?
	   (cells + ((originX+row) * cols) + originY+col - cols) :
	   (cells + ((originX+row) * cols) + originY+col)));
}

/***********************************************************************/

/*! Public method getNextCell returns a pointer to the next cell lying
  on the line whose start point and end point were given in the most
  recent call to initLine. If the end of the line was reached
  previously, this returns a NULL pointer. */

CELL_NAME * ScrollingMap::getNextCell() /* NO ARGUMENTS */
{
  if (!inLine)
    next1 = 0;
  else if (nextCell)
    {
      next1 = nextCell;
      nextCell = 0;
      if ((dI >= deltaI) && (dJ >= deltaJ))
	inLine = 0;
    }
  else if ((dI >= deltaI) && (dJ >= deltaJ))
    {
      inLine = 0;
      next1 = 0;
    }
  else if (fabs(deltaX) >= fabs(deltaY))
    { //step1 in X direction
      next1 = getCellAtIndexes(i, j);
      if ((step2 > lack) && (dJ < deltaJ))
	{ // line passes through cell to right or left
	  j = ((deltaY >= 0) ? (j + 1) : (j - 1));
	  dJ++;
	  nextCell = getCellAtIndexes(i, j);
	  lack  = (res - (step2 - lack));
	}
      else
	{ // line does not pass through cell to right or left
	  lack = (lack - step2);
	  nextCell = 0;
	}
      i = ((deltaX > 0) ? (i + 1) : (i - 1));
      dI++;
    }
  else // if (fabs(deltaX) < fabs(deltaY))
    { //step1 in Y direction
      next1 = getCellAtIndexes(i, j);
      if ((step2 > lack) && (dI < deltaI))
	{ // line passes through cell above or below
	  i = ((deltaX >= 0) ? (i + 1) : (i - 1));
	  dI++;
	  nextCell = getCellAtIndexes(i, j);
	  lack  = (res - (step2 - lack));
	}
      else
	{ // line does not pass through cell above or below
	  lack = (lack - step2);
	  nextCell = 0;
	}
      j = ((deltaY > 0) ? (j + 1) : (j - 1));
      dJ++;
    }  
  return next1;
}

/***********************************************************************/

/*! Public method getXCoordAtCell returns the X coordinate of the
  center of the given cell. If the cell is NULL or its id is not the
  id of a cell, this returns xLow. */

/* This works because the id (which is assigned by the map constructor
   and cannot be changed) is the index in the 1D array of the cell. */

double ScrollingMap::getXCoordAtCell( /*   ARGUMENTS                     */
 CELL_NAME * aCell)                   /*!> cell for which to get X coord */
{
  int ix;
  if ((!aCell) || (aCell->id < 0) || (aCell->id >= numberOfCells))
    return xLow;
  ix = ((aCell->id / side) - originX);
  if (ix < 0)
    ix = (ix + side);
  return (xLow + ((ix + 0.5) * res));
}

/***********************************************************************/

/*! Public method getXCoordAtId returns the X coordinate of the center
  of the cell with the given id. If the id is not the id of a cell,
  this returns xLow. */

/* This works because the id (which is assigned by the map constructor
   and cannot be changed) is the index in the 1D array of the cell. */
  
double ScrollingMap::getXCoordAtId( /*   ARGUMENTS                           */
 int id)                            /*!> id of cell for which to get X coord */
{
  int ix;
  if ((id < 0) || (id >= numberOfCells))
    return xLow;
  ix = ((id / side) - originX);
  if (ix < 0)
    ix = (ix + side);
  return (xLow + ((ix + 0.5) * res));
}

/***********************************************************************/

/*! Public method getXIndexAtCell returns the window grid X index of
  the given cell. If the cell is NULL or its id is not the id of a
  cell, this returns -1. */

/* This works because the id (which is assigned by the map constructor
   and cannot be changed) is the index in the 1D array of the cell. */
  
int ScrollingMap::getXIndexAtCell( /*   ARGUMENTS                     */
 CELL_NAME * aCell)                /*!> cell for which to get X index */
{
  int ix;

  if ((!aCell) || (aCell->id < 0) || (aCell->id >= numberOfCells))
    return -1;
  ix = ((aCell->id / side) - originX);
  return ((ix < 0) ? (ix + side) : ix);
}

/***********************************************************************/

/*! Public method getXIndexAtCoord returns the window grid X index of
  the X coordinate value x, if x is inside the window. Otherwise, it
  returns -1. */

int ScrollingMap::getXIndexAtCoord( /*   ARGUMENTS    */
 double x)                          /*!< X coordinate */
{
  if ((x < xLow) || (x > xHi))
    return -1;

  //printf("\n\nxLow: %lf, resolution: %lf\n", xLow, resolution);
  //printf("yLow: %lf, resolution: %lf\n\n", yLow, resolution);

  return (int)(floor((x - xLow) / resolution));
}

/***********************************************************************/

/*! Public method getYCoordAtCell returns the Y coordinate of the
  center of the given cell. If the cell is NULL or its id is not the
  id of a cell, this returns yLow. */

/* This works because the id (which is assigned by the map constructor
   and cannot be changed) is the index in the 1D array of the cell. */
  
double ScrollingMap::getYCoordAtCell( /*   ARGUMENTS                     */
 CELL_NAME * aCell)                   /*!> cell for which to get Y coord */
{
  int iy;
  if ((!aCell) || (aCell->id < 0) || (aCell->id >= numberOfCells))
    return yLow;
  iy = ((aCell->id % side) - originY);
  if (iy < 0)
    iy = (iy + side);
  return (yLow + ((iy + 0.5) * res));
}

/***********************************************************************/

/*! Public method getYCoordAtId returns the Y coordinate of the center
  of the cell with the given id. If the id is not the id of a cell,
  this returns yLow. */

/* This works because the id (which is assigned by the map constructor
   and cannot be changed) is the index in the 1D array of the cell. */
  
double ScrollingMap::getYCoordAtId( /*   ARGUMENTS                           */
 int id)                            /*!> id of cell for which to get Y coord */
{
  int iy;
  if ((id < 0) || (id >= numberOfCells))
    return yLow;
  iy = ((id % side) - originY);;
  if (iy < 0)
    iy = (iy + side);
  return (yLow + ((iy + 0.5) * res));
}

/***********************************************************************/

/*! Public method getYIndexAtCell returns the window grid Y index of
  the given cell. If the cell is NULL or its id is not the id of a
  cell, this returns -1. */

/* This works because the id (which is assigned by the map constructor
   and cannot be changed) is the index in the 1D array of the cell. */
  
int ScrollingMap::getYIndexAtCell( /*   ARGUMENTS                     */
 CELL_NAME * aCell)                /*!> cell for which to get Y index */
{
  int iy;

  if ((!aCell) || (aCell->id < 0) || (aCell->id >= numberOfCells))
    return -1;
  iy = ((aCell->id % side) - originY);
  return ((iy < 0) ? (iy + side) : iy);
}

/***********************************************************************/

/*! Public method getYIndexAtCoord returns the window grid Y index of
  the Y coordinate value y, if y is inside the window. Otherwise, it
  returns -1. */

int ScrollingMap::getYIndexAtCoord( /*   ARGUMENTS    */
 double y)                          /*!< Y coordinate */
{
  if ((y < yLow) || (y > yHi))
    return -1;
  return (int)(floor((y - yLow) / resolution));
}

/***********************************************************************/

/*! Public method initLine returns a pointer to the cell containing
   the point on the map whose coordinates are startX and startY.  If
   the start point is off the map, this returns a NULL pointer. If the
   start point is on the map, this also sets data in the map to
   support the work done by findNextCell, which finds the next cell on
   the line from the start point to the end point. */
/*

If (endX, endY) is off the map, this resets (endX, endY) to the point
where the line from (endX, endY) to (startX, startY) enters the map.

*/

CELL_NAME * ScrollingMap::initLine( /*   ARGUMENTS                   */
 double startX,                     /*!< X coordinate of start point */
 double startY,                     /*!< Y coordinate of start point */
 double endX,                       /*!< X coordinate of end point   */
 double endY)                       /*!< Y coordinate of end point   */
{
  double step1;   // size of step in primary direction (always positive)
  double slope;   // slope with respect to principal direction

  if ((startX < xLow) || (startX > xHi) || (startY < yLow) || (startY > yHi))
    {
      inLine = false;
      return 0;
    }
  else if ((startX == endX) && (startY == endY))
    {
      inLine = false;
      i = getXIndexAtCoord(startX);
      j = getYIndexAtCoord(startY);
      return getCellAtIndexes(i,j);
    }

  inLine = true;
  if (endY < yLow)
    {
      endX = (endX + ((startX - endX) * ((yLow - endY) / (startY - endY))));
      endY = (yLow + SMAP_TINY);
      if (endX < xLow)
	{
	  endY = (endY + ((startY - endY) * ((xLow - endX)/(startX - endX))));
	  endX = (xLow + SMAP_TINY);
	}
      else if (endX > xHi)
	{
	  endY = (endY + ((startY - endY) * ((endX - xHi)/(endX - startX))));
	  endX = (xHi - SMAP_TINY);
	}
      else if (endX == xLow)
	endX = (xLow + SMAP_TINY);
      else if (endX == xHi)
	endX = (xHi - SMAP_TINY);
    }
  else if (endY > yHi)
    {
      endX = (endX + ((startX - endX) * ((endY - yHi) / (endY - startY))));
      endY = (yHi - SMAP_TINY);
      if (endX < xLow)
	{
	  endY = (endY - ((endY - startY) * ((xLow - endX)/(startX - endX))));
	  endX = (xLow + SMAP_TINY);
	}
      else if (endX > xHi)
	{
	  endY = (endY - ((endY - startY) * ((endX - xHi)/(endX - startX))));
	  endX = (xHi - SMAP_TINY);
	}
      else if (endX == xLow)
	endX = (xLow + SMAP_TINY);
      else if (endX == xHi)
	endX = (xHi - SMAP_TINY);
    }
  else if (endX > xHi)
    {
      endY = (endY + ((startY - endY) * ((endX - xHi)/(endX - startX))));
      endX = (xHi - SMAP_TINY);
    }
  else if (endX < xLow)
    {
      endY = (endY + ((startY - endY) * ((xLow - endX)/(startX - endX))));
      endX = (xLow + SMAP_TINY);
    }
  i = getXIndexAtCoord(startX);
  j = getYIndexAtCoord(startY);
  dI = 0;
  dJ = 0;
  deltaX = (endX - startX);
  deltaY = (endY - startY);
  next1 = getCellAtIndexes(i, j);
  if (fabs(deltaX) >= fabs(deltaY))
    { //step1 in X direction
      slope = fabs(deltaY / deltaX);
      if (deltaX > 0)
	{ // step1 towards positive X
	  step1 = (xLow + ((getXIndexAtCoord(startX) + 1) * res) - startX);
	  deltaI = (int)((deltaX + res - step1 - SMAP_TINY) / res);
	}
      else // if (deltaX < 0)
	{ // step1 towards negative X
	  step1 = (startX - (xLow + (getXIndexAtCoord(startX) * res)));
	  deltaI = (int)((-deltaX + res - step1 - SMAP_TINY) / res);
	}
      step2 = (step1 * slope);
      if (deltaY >= 0)
	{ // Y increasing
	  lack = (yLow + ((getYIndexAtCoord(startY) + 1) * res) - startY);
	  deltaJ = (int)((deltaY + res - lack - SMAP_TINY) / res);
	}
      else // if (deltaY < 0)
	{ // Y decreasing
	  lack = (startY - (yLow + (getYIndexAtCoord(startY) * res)));
	  deltaJ = (int)((-deltaY + res - lack - SMAP_TINY) /res);
	}
      if (step2 > lack)
	{ // line passes through cell to right or left
	  j = ((deltaY >= 0) ? (j + 1) : (j - 1));
	  dJ++;
	  nextCell = getCellAtIndexes(i, j);
	  lack = (res - (step2 - lack));
	}
      else
	{  // line does not pass through cell to right or left
	  nextCell = 0;
	  lack = (lack - step2);
	}
      step2 = (res * slope);
      i = ((deltaX > 0) ? (i + 1) : (i - 1));
    }
  else // if (fabs(deltaX) < fabs(deltaY))
    { //step1 in Y direction
      slope = fabs(deltaX / deltaY);
      if (deltaY > 0)
	{ // step1 towards positive Y
	  step1 = (yLow + ((getYIndexAtCoord(startY) + 1) * res) - startY);
	  deltaJ = (int)((deltaY + res - step1 - SMAP_TINY) / res);
	}
      else // if (deltaY < 0)
	{ // step1 towards negative Y
	  step1 = (startY - (yLow + (getYIndexAtCoord(startY) * res)));
	  deltaJ = (int)((-deltaY + res - step1 - SMAP_TINY) / res);
	}
      step2 = (step1 * slope);
      if (deltaX >= 0)
	{ // X increasing
	  lack = (xLow + ((getXIndexAtCoord(startX) + 1) * res) - startX);
	  deltaI = (int)((deltaX + res - lack - SMAP_TINY) / res);
	}
      else // if (deltaX < 0)
	{ // X decreasing
	  lack = (startX - (xLow + (getXIndexAtCoord(startX) * res)));
	  deltaI = (int)((-deltaX + res - lack - SMAP_TINY) /res);
	}
      if (step2 > lack)
	{ // line passes through cell above or below
	  i = ((deltaX >= 0) ? (i + 1) : (i - 1));
	  dI++;
	  nextCell = getCellAtIndexes(i, j);
	  lack = (res - (step2 - lack));
	}
      else
	{  // line does not pass through cell above or below
	  nextCell = 0;
	  lack = (lack - step2);
	}
      step2 = (res * slope);
      j = ((deltaY > 0) ? (j + 1) : (j - 1));
    }
  return next1;
}

/***********************************************************************/

/*! Public method setCenter causes the window to scroll so that the
  cartesian point (x, y) lies inside the cell at the center of the
  window grid.  The originX, originY, xLow, xHi, yLow, yHi, centerX,
  and centerY of the map change when the center moves. Also, the cells
  of the actual grid that represent the cells entering (and leaving)
  the window grid are reinitialized.  xDiff and yDiff are set to the
  number of cells the center has moved in the X and Y directions; the
  values may be positive, zero, or negative. */
/*

If the center point has made a jump larger than rows or cols, this
reinitializes all the cells in the actual grid. Otherwise, it
reinitializes the cells in a portion of the actual grid consisting of
two rectangles, one extending the width of the grid and bounded above
and below by the old originX and the new originX, the other extending
the height of the grid and bounded right and left by the old originY
and the new originY. If originX or originY wraps around to get to its
new value, the corresponding rectangle divides into two rectangles.
If originX or originY does not change, there is no corresponding
rectangle.

*/

void ScrollingMap::setCenter( /*   ARGUMENTS                                */
 double x,                    /*!< X coordinate of new center point         */
 double y,                    /*!< Y coordinate of new center point         */
 int * xDiff,                 /*!< no. of cells center moved in X, set here */
 int * yDiff)                 /*!< no. of cells center moved in Y, set here */
{
  int n;
  int row;
  int col;
  int newOriginX;
  int newOriginY;
  int stop;
  int xRaw;
  int yRaw;

  xRaw = getActualRawXIndex(x);
  yRaw = getActualRawYIndex(y);
  *xDiff = (xRaw - getActualRawXIndex(centerX));
  *yDiff = (yRaw - getActualRawYIndex(centerY));
  newOriginX = ((originX + *xDiff) % rows);
  if (newOriginX < 0)
    newOriginX = (newOriginX + rows);
  newOriginY = ((originY + *yDiff) % cols);
  if (newOriginY < 0)
    newOriginY = (newOriginY + cols);
  if ((*xDiff >= rows) || (*xDiff <= -rows) ||
      (*yDiff >= cols) || (*yDiff <= -cols))
    { // clear the entire map
      for (row = 0; row < rows; row++)
	{
	  for (col = 0; col < cols; col++)
	    {
	      (cells + (row * cols) + col)->init();
	    }
	}
    }
  else
    {
      if (*xDiff > 0)
	{ // center cell moved in the positive x direction
	  stop = *xDiff;
	  row = originX;
	}
      else
	{ // center cell moved in the negative x direction or not at all
	  stop = -*xDiff;
	  row = newOriginX;
	}
      for (n = 0; n < stop; n++) // this does nothing if *xDiff is zero
	{ // initialize cells in horizontal rectangle
	  for (col = 0; col < cols; col++)
	    {
	      (cells + (row * cols) + col)->init();
	    }
	  row++;
	  if (row >= rows)
	    row = (row - rows);
	}
      if (*yDiff > 0)
	{ // center cell moved in the positive y direction
	  stop = *yDiff;
	  col = originY;
	}
      else
	{ // center cell moved in the negative y direction or not at all
	  stop = -*yDiff;
	  col = newOriginY;
	}
      for (n = 0; n < stop; n++) // this does nothing if *yDiff is zero
	{ // initialize cells in vertical rectangle
	  for (row = 0; row < rows; row++)
	    {
	      (cells + (row * cols) + col)->init();
	    }
	  col++;
	  if (col >= cols)
	    col = (col - cols);
	}
    }
  centerX = x;
  centerY = y;
  originX = newOriginX;
  originY = newOriginY;
  xLow = ((xRaw - (range + halfSide)) * resolution);
  yLow = ((yRaw - (range + halfSide)) * resolution);
  xHi  = (xLow + (side * resolution));
  yHi  = (yLow + (side * resolution));
}

/***********************************************************************/

