/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file ulselect.c

  \brief For those functions that differ between ULAPI interfaces to
  Unix or RTAI, this selects which particular ULAPI implementation to
  call, allowing dynamic switching between them. It also provides
  all the functions that are the same for each, so that code isn't
  duplicated.

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.5 $
  $Date: 2009/04/16 19:17:38 $
  \endcode

  \author Fred Proctor
*/

/*
  Some of these functions are implemented exactly the same way, with
  slightly different names, in unix_rtapi.c. If you change them here,
  change them there.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "ulapi.h"		/* these decls */
#include "unix_ulapi.h"		/* decls for Unix */
#ifdef HAVE_RTAI
#include "rtai_ulapi.h"		/* decls for RTAI */
#else
#include <stddef.h>		/* NULL */
#endif 
#include <stdlib.h>		/* malloc */
#include <string.h>		/* memset */
#include <ctype.h>		/* isspace */
#include <pthread.h>		/* pthread_create(), pthread_mutex_t */
#include <time.h>		/* struct timespec, nanosleep */
#include <sys/time.h>		/* gettimeofday(), struct timeval */
#include <unistd.h>		/* select(), write() */
#include <fcntl.h>		/* O_RDONLY, O_NONBLOCK */
#include <sys/types.h>		/* fd_set, FD_ISSET() */
#include <sys/socket.h>		/* PF_INET, socket(), listen(), bind(), etc. */
#include <netinet/in.h>		/* struct sockaddr_in */
#include <netdb.h>		/* gethostbyname */
#ifdef HAVE_DLFCN_H
#include <dlfcn.h>
#endif

static ulapi_integer ulapi_debug_level = 0;

#define PERROR(x) if (ulapi_debug_level & ULAPI_DEBUG_ERROR) perror(x)

static ulapi_integer selected = UL_USE_UNIX;
static ulapi_integer _ulapi_wait_offset_nsec = 1;

ulapi_real ulapi_time(void)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);

  return (ulapi_real) (((double) tv.tv_sec) + ((double) tv.tv_usec) * 1.0e-6);
}

void ulapi_sleep(ulapi_real secs)
{
  int isecs, insecs;
  struct timespec ts;

  isecs = (int) secs;
  insecs = (int) ((secs - isecs) * 1.0e9);

  ts.tv_sec = isecs;
  ts.tv_nsec = insecs;

  (void) nanosleep(&ts, NULL);
}

ulapi_result ulapi_init(ulapi_integer sel)
{
  if (UL_USE_UNIX == sel ||
      UL_USE_DEFAULT == sel) {
    selected = UL_USE_UNIX;
    return unix_ulapi_init(sel);
  }

  if (UL_USE_RTAI == sel) {
#ifdef HAVE_RTAI
    selected = sel;
    return rtai_ulapi_init(sel);
#else
    return ULAPI_IMPL_ERROR;
#endif
  }

  return ULAPI_BAD_ARGS;
}

ulapi_result ulapi_exit(void)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_exit() : 
#ifdef HAVE_RTAI
    rtai_ulapi_exit();
#else
  ULAPI_IMPL_ERROR;
#endif
}

ulapi_integer ulapi_to_argv(char *src, char ***argv)
{
  char *cpy;
  char *ptr;
  ulapi_integer count;
  ulapi_flag inquote;

  *argv = NULL;
  cpy = malloc(strlen(src) + 1);
  count = 0;
  inquote = 0;

  for (; ; count++) {
    while (isspace(*src)) src++;
    if (0 == *src) return count;
    ptr = cpy;
    while (0 != *src) {
      if ('"' == *src) {
	inquote = !inquote;
      } else if ((!inquote) && isspace(*src)) {
	break;
      } else {
	*ptr++ = *src;
      }
      src++;
    }
    *ptr++ = 0;
    *argv = realloc(*argv, (count+1)*sizeof(**argv));
    (*argv)[count] = malloc(strlen(cpy) + 1);
    strcpy((*argv)[count], cpy);
  }
   
  free(cpy);

  return count;
}

void ulapi_free_argv(ulapi_integer argc, char **argv)
{
  ulapi_integer t;

  for (t = 0; t < argc; t++) {
    free(argv[t]);
  }
  free(argv);

  return;
}

void ulapi_set_debug(ulapi_integer mask)
{
  ulapi_debug_level = mask;
}

ulapi_integer ulapi_get_debug(void)
{
  return ulapi_debug_level;
}

ulapi_prio ulapi_prio_highest(void)
{
  return 1;
}

ulapi_prio ulapi_prio_lowest(void)
{
  return 31;
}

ulapi_prio ulapi_prio_next_higher(ulapi_prio prio)
{
  if (prio == ulapi_prio_highest())
    return prio;

  return prio - 1;
}

ulapi_prio ulapi_prio_next_lower(ulapi_prio prio)
{
  if (prio == ulapi_prio_lowest())
    return prio;

  return prio + 1;
}

void * ulapi_task_new(void)
{
  return malloc(sizeof(pthread_t));
}

ulapi_result ulapi_task_delete(void *task)
{
  if (NULL != task) free(task);

  return ULAPI_OK;
}

typedef void *(*pthread_task_code)(void *);

ulapi_result
ulapi_task_start(void *task,
		 void (*taskcode)(void *),
		 void *taskarg,
		 ulapi_prio prio,
		 ulapi_integer period_nsec)
{
  pthread_attr_t attr;
  struct sched_param sched_param;

  pthread_attr_init(&attr);
  sched_param.sched_priority = prio;
  pthread_attr_setschedparam(&attr, &sched_param);
  pthread_create((pthread_t *) task, &attr, (pthread_task_code) taskcode, taskarg);
  pthread_setschedparam(*((pthread_t *) task), SCHED_OTHER, &sched_param);
  /* ignore period_nsec, since we can't handle it in pthreads, so the
     application must call ulapi_wait(period_nsec) */

  return ULAPI_OK;
}

ulapi_result ulapi_task_stop(void *task)
{
  return (pthread_cancel(*((pthread_t *) task)) == 0 ? ULAPI_OK : ULAPI_ERROR);
}

ulapi_result ulapi_task_pause(void *task)
{
  return ULAPI_OK;
}

ulapi_result ulapi_task_resume(void *task)
{
  return ULAPI_OK;
}

ulapi_result ulapi_task_set_period(void *task, ulapi_integer period_nsec)
{
  return ULAPI_OK;
}

ulapi_result ulapi_task_init(void)
{
  if (0 != pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL)) return ULAPI_ERROR;
  if (0 != pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL)) return ULAPI_ERROR;

  return ULAPI_OK;
}

ulapi_result ulapi_self_set_period(ulapi_integer period_nsec)
{
  return ULAPI_OK;
}

ulapi_result ulapi_wait(ulapi_integer period_nsec)
{
  struct timespec ts;

  if (period_nsec < _ulapi_wait_offset_nsec + 1) period_nsec = 1;
  else period_nsec -= _ulapi_wait_offset_nsec;

  ts.tv_sec = 0;
  ts.tv_nsec = period_nsec;

  (void) nanosleep(&ts, NULL);

  return ULAPI_OK;
}

ulapi_result ulapi_task_exit(void)
{
  return ULAPI_OK;
}

ulapi_result ulapi_task_join(void * task)
{
  return (pthread_join(*((pthread_t *) task), NULL) == 0 ? ULAPI_OK : ULAPI_ERROR);
}

ulapi_integer ulapi_task_id(void)
{
  return (ulapi_integer) pthread_self();
}

void * ulapi_mutex_new(ulapi_id key)
{
  pthread_mutex_t * mutex;

  mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  if (NULL == (void *) mutex) return NULL;

  /* initialize mutex to default attributes, and give it */
  if (0 == pthread_mutex_init(mutex, NULL)) {
    (void) pthread_mutex_unlock(mutex);
    return (void *) mutex;
  }
  /* else got an error, so free the mutex and return null */

  free(mutex);
  return NULL;
}

ulapi_result ulapi_mutex_delete(void * mutex)
{
  if (NULL == (void *) mutex) return ULAPI_ERROR;

  (void) pthread_mutex_destroy((pthread_mutex_t *) mutex);
  free(mutex);

  return ULAPI_OK;
}

ulapi_result ulapi_mutex_give(void * mutex)
{
  return (0 == pthread_mutex_unlock((pthread_mutex_t *) mutex) ? ULAPI_OK : ULAPI_ERROR);
}

ulapi_result ulapi_mutex_take(void * mutex)
{
  return (0 == pthread_mutex_lock((pthread_mutex_t *) mutex) ? ULAPI_OK : ULAPI_ERROR);
}

void * ulapi_cond_new(ulapi_id key)
{
  pthread_cond_t * cond;

  cond = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));
  if (NULL == (void *) cond) return NULL;

  if (0 == pthread_cond_init(cond, NULL)) {
    return (void *) cond;
  }
  /* else got an error, so free the condition variable and return null */

  free(cond);
  return NULL;
}

ulapi_result ulapi_cond_delete(void * cond)
{
  if (NULL == (void *) cond) return ULAPI_ERROR;

  (void) pthread_cond_destroy((pthread_cond_t *) cond);
  free(cond);

  return ULAPI_OK;
}

ulapi_result ulapi_cond_signal(void * cond)
{
  return (0 == pthread_cond_signal((pthread_cond_t *) cond) ? ULAPI_OK : ULAPI_ERROR);
}

ulapi_result ulapi_cond_broadcast(void * cond)
{
  return (0 == pthread_cond_broadcast((pthread_cond_t *) cond) ? ULAPI_OK : ULAPI_ERROR);
}

ulapi_result ulapi_cond_wait(void * cond, void * mutex)
{
  return (0 == pthread_cond_wait((pthread_cond_t *) cond, (pthread_mutex_t *) mutex) ? ULAPI_OK : ULAPI_ERROR);
}

void * ulapi_shm_new(ulapi_id key, ulapi_integer size)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_shm_new(key, size) : 
#ifdef HAVE_RTAI
    rtai_ulapi_shm_new(key, size);
#else
  NULL;
#endif
}

void * ulapi_shm_addr(void * shm)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_shm_addr(shm) : 
#ifdef HAVE_RTAI
    rtai_ulapi_shm_addr(shm);
#else
  NULL;
#endif
}

ulapi_result ulapi_shm_delete(void * shm)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_shm_delete(shm) : 
#ifdef HAVE_RTAI
    rtai_ulapi_shm_delete(shm);
#else
  ULAPI_IMPL_ERROR;
#endif
}

ulapi_result ulapi_fifo_new(ulapi_integer key, ulapi_integer * fd,
			    ulapi_integer size)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_fifo_new(key, fd, size) : 
#ifdef HAVE_RTAI
    rtai_ulapi_fifo_new(key, fd, size);
#else
  ULAPI_IMPL_ERROR;
#endif
}

ulapi_result ulapi_fifo_delete(ulapi_integer key, ulapi_integer fd,
			       ulapi_integer size)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_fifo_delete(key, fd, size) : 
#ifdef HAVE_RTAI
    rtai_ulapi_fifo_delete(key, fd, size);
#else
  ULAPI_IMPL_ERROR;
#endif
}

ulapi_integer ulapi_fifo_write(ulapi_integer fd, char *buf,
			       ulapi_integer size)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_fifo_write(fd, buf, size) : 
#ifdef HAVE_RTAI
    rtai_ulapi_fifo_write(fd, buf, size);
#else
  -1;
#endif
}

ulapi_integer ulapi_fifo_read(ulapi_integer fd, char *buf,
			      ulapi_integer size)
{
  return selected == UL_USE_UNIX ?
    unix_ulapi_fifo_read(fd, buf, size) : 
#ifdef HAVE_RTAI
    rtai_ulapi_fifo_read(fd, buf, size);
#else
  -1;
#endif
}

ulapi_integer
ulapi_socket_client(ulapi_integer port,
		    char * hostname)
{
  int socket_fd;
  struct sockaddr_in server_addr;
  struct hostent *ent;
  struct in_addr *in_a;

  if (NULL == (ent = gethostbyname(hostname))) {
    PERROR("gethostbyname");
    return -1;
  }
  in_a = (struct in_addr *) ent->h_addr_list[0];

  memset(&server_addr, 0, sizeof(struct sockaddr_in));
  server_addr.sin_family = PF_INET;
  server_addr.sin_addr.s_addr = in_a->s_addr;
  server_addr.sin_port = htons(port);

  if (-1 == (socket_fd = socket(PF_INET, SOCK_STREAM, 0))) {
    PERROR("socket");
    return -1;
  }

  if (-1 == connect(socket_fd, 
		    (struct sockaddr *) &server_addr,
		    sizeof(struct sockaddr_in))) {
    PERROR("connect");
    close(socket_fd);
    return -1;
  }

  return socket_fd;
}

ulapi_integer
ulapi_socket_server(ulapi_integer port)
{
  int socket_fd;
  struct sockaddr_in myaddr;
  int on;
  struct linger mylinger = { 0 };
  enum {BACKLOG = 5};

  if (-1 == (socket_fd = socket(PF_INET, SOCK_STREAM, 0))) {
    PERROR("socket");
    return -1;
  }

  /*
    Turn off bind address checking, and allow
    port numbers to be reused - otherwise the
    TIME_WAIT phenomenon will prevent binding
    to these address.port combinations for
    (2 * MSL) seconds.
  */
  on = 1;
  if (-1 == 
      setsockopt(socket_fd,
		 SOL_SOCKET,
		 SO_REUSEADDR,
		 (const char *) &on,
		 sizeof(on))) {
    PERROR("setsockopt");
    close(socket_fd);
    return -1;
  }

  /*
    When connection is closed, there is a need
    to linger to ensure all data is transmitted.
  */
  mylinger.l_onoff = 1;
  mylinger.l_linger = 30;
  if (-1 ==
      setsockopt(socket_fd,
		 SOL_SOCKET,
		 SO_LINGER,
		 (const char *) &mylinger,
		 sizeof(struct linger))) {
    PERROR("setsockopt");
    close(socket_fd);
    return -1;
  }

  memset(&myaddr, 0, sizeof(struct sockaddr_in));
  myaddr.sin_family = PF_INET;
  myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  myaddr.sin_port = htons(port);

  if (-1 == bind(socket_fd, (struct sockaddr *) &myaddr,
		 sizeof(struct sockaddr_in))) {
    PERROR("bind");
    close(socket_fd);
    return -1;
  }

  if (-1 == listen(socket_fd, BACKLOG)) {
    PERROR("listen");
    close(socket_fd);
    return -1;
  }

  return socket_fd;
}

ulapi_integer
ulapi_socket_get_client(ulapi_integer socket_fd)
{
  fd_set rfds;
  struct sockaddr_in client_addr;
  int client_len;
  int client_fd;
  int retval;

  do {
    FD_ZERO(&rfds);
    FD_SET(socket_fd, &rfds);
    retval = select(socket_fd + 1, &rfds, NULL, NULL, NULL);
  } while (0 == retval ||
	   ! FD_ISSET(socket_fd, &rfds));

  if (retval == -1) {
    PERROR("select");
    return -1;
  }

  memset(&client_addr, 0, sizeof(struct sockaddr_in));
  client_len = sizeof(struct sockaddr_in);
  client_fd = 
    accept(socket_fd,
	   (struct sockaddr *) &client_addr, 
	   &client_len);
  if (-1 == client_fd) {
    PERROR("accept");
    return -1;
  }
  
  return client_fd;
}

ulapi_result
ulapi_socket_set_nonblocking(ulapi_integer fd)
{
  int flags;

  flags = fcntl(fd, F_GETFL);
  if (-1 == fcntl(fd, F_SETFL, flags | O_NONBLOCK)) {
    return ULAPI_ERROR;
  }

  return ULAPI_OK;
}

ulapi_result
ulapi_socket_set_blocking(ulapi_integer fd)
{
  int flags;

  flags = fcntl(fd, F_GETFL);
  if (-1 == fcntl(fd, F_SETFL, flags & ~O_NONBLOCK)) {
    return ULAPI_ERROR;
  }

  return ULAPI_OK;
}

ulapi_integer
ulapi_socket_read(ulapi_integer id,
		  char * buf,
		  ulapi_integer len)
{
  return read(id, buf, len);
}

ulapi_integer
ulapi_socket_write(ulapi_integer id,
		   char * buf,
		   ulapi_integer len)
{
  return write(id, buf, len);
}

ulapi_result
ulapi_socket_close(ulapi_integer id)
{
  return 0 == close((int) id) ? ULAPI_OK : ULAPI_ERROR;
}

/* File descriptor interface */

void * 
ulapi_fd_new(void)
{
  return malloc(sizeof(int));
}

ulapi_result
ulapi_fd_delete(void * id)
{
  if (NULL != id) free(id);

  return ULAPI_OK;
}

ulapi_result
ulapi_std_open(ulapi_stdio io, void * id)
{
  if (ULAPI_STDIN == io) {
    *((int *) id) = STDIN_FILENO;
    return ULAPI_OK;
  }
  if (ULAPI_STDOUT == io) {
    *((int *) id) = STDOUT_FILENO;
    return ULAPI_OK;
  }
  if (ULAPI_STDERR == io) {
    *((int *) id) = STDERR_FILENO;
    return ULAPI_OK;
  }
  return ULAPI_ERROR;
}

ulapi_result
ulapi_fd_open(char * port, void * id)
{
  int fd;

  fd = open(port, O_RDWR);

  if (fd < 0) return ULAPI_ERROR;
  
  *((int *) id) = fd;
  return ULAPI_OK;
}

ulapi_result
ulapi_fd_set_nonblocking(void * id)
{
  int fd;
  int flags;

  fd = *((int *) id);
  flags = fcntl(fd, F_GETFL);
  if (-1 == fcntl(fd, F_SETFL, flags | O_NONBLOCK)) {
    return ULAPI_ERROR;
  }

  return ULAPI_OK;
}

ulapi_result
ulapi_fd_set_blocking(void * id)
{
  int fd;
  int flags;

  fd = *((int *) id);
  flags = fcntl(fd, F_GETFL);
  if (-1 == fcntl(fd, F_SETFL, flags & ~O_NONBLOCK)) {
    return ULAPI_ERROR;
  }

  return ULAPI_OK;
}

ulapi_integer
ulapi_fd_read(void * id,
		  char * buf,
		  ulapi_integer len)
{
  return read(*((int *) id), buf, len);
}

ulapi_integer
ulapi_fd_write(void * id,
		   char * buf,
		   ulapi_integer len)
{
  return write(*((int *) id), buf, len);
}

ulapi_result
ulapi_fd_close(void * id)
{
  return 0 == close(*((int *) id)) ? ULAPI_OK : ULAPI_ERROR;
}

/* Serial interface is implemented using the file descriptor interface,
   which is the same */

void * 
ulapi_serial_new(void)
{
  return ulapi_fd_new();
}

ulapi_result
ulapi_serial_delete(void * id)
{
  return ulapi_fd_delete(id);
}

ulapi_result
ulapi_serial_open(char * port, void * id)
{
  return ulapi_fd_open(port, id);
}

ulapi_result
ulapi_serial_set_nonblocking(void * id)
{
  return ulapi_fd_set_nonblocking(id);
}

ulapi_result
ulapi_serial_set_blocking(void * id)
{
  return ulapi_fd_set_blocking(id);
}

ulapi_integer
ulapi_serial_read(void * id,
		  char * buf,
		  ulapi_integer len)
{
  return ulapi_fd_read(id, buf, len);
}

ulapi_integer
ulapi_serial_write(void * id,
		   char * buf,
		   ulapi_integer len)
{
  return ulapi_fd_write(id, buf, len);
}

ulapi_result
ulapi_serial_close(void * id)
{
  return ulapi_fd_close(id);
}

#ifdef HAVE_DLFCN_H

void *ulapi_dl_open(char *objname, char *errstr, int errlen)
{
  void *handle;
  void *error;

  dlerror();
  handle = dlopen(objname, RTLD_LAZY);
  error = dlerror();

  if (NULL == handle) {
    if (NULL != errstr && NULL != error) {
      strncpy(errstr, error, errlen);
      errstr[errlen - 1] = 0;
    }
    return NULL;
  }

  if (NULL != errstr) {
    errstr[0] = 0;
  }
  return handle;
}

void ulapi_dl_close(void *handle)
{
  if (NULL != handle) {
    dlclose(handle);
  }
}

void *ulapi_dl_sym(void *handle, char *name, char *errstr, int errlen)
{
  void *sym;
  void *error;

  dlerror();
  sym = dlsym(handle, name);
  error = dlerror();

  if (error != NULL) {
    if (NULL != errstr) {
      strncpy(errstr, error, errlen);
      errstr[errlen - 1] = 0;
    }
    return NULL;
  }

  if (NULL != errstr) {
    errstr[0] = 0;
  }
  return sym;
}

#endif

/*
  Modification history:

  $Log: ulselect.c,v $
  Revision 1.5  2009/04/16 19:17:38  proctor
  Updated the NML config files to reflect the simware changes; updated ulapi
  to be consistent with the master files in gomotion; added targets for the
  new simware

  Revision 1.16  2009/03/15 13:39:37  proctor
  Added quote handling in ulapi_to_argv

  Revision 1.15  2009/03/13 21:29:24  proctor
  Added to_argv utility

  Revision 1.14  2009/02/26 14:11:19  proctor
  Fixed bug where error string was not zeroed upon success

  Revision 1.13  2009/02/24 13:53:07  proctor
  Added ulapi_dl_XXX functions

  Revision 1.12  2009/01/13 16:23:11  proctor
  Fixed problem with casting function pointers to void pointers

  Revision 1.11  2008/12/08 17:40:50  proctor
  Added ultest.c and ulapi_cond_xxx condition variables

  Revision 1.10  2008/12/02 15:55:09  proctor
  Added ulapi_task_join

  Revision 1.9  2008/02/22 18:39:42  proctor
  Reworked serial interface to allocate a pointer to the id

  Revision 1.8  2008/02/21 16:35:01  proctor
  Added rt/ulapi_serial functions, and smartmotor build

  Revision 1.7  2008/01/18 15:35:29  proctor
  Fixed getopt error printing bugs in igps; inited mutexes to defaults
  and gave them in ulselect.c; took out GO_STEPPER entries in genhex.ini
  since PIDs were used and starting the stepper module wasn't needed

  Revision 1.6  2007/12/21 19:47:26  proctor
  Added nonblocking functions to ulapi; made igpsserver really get position
  from a controller

  Revision 1.5  2007/12/07 14:34:55  proctor
  Added socket code to ulapi, and stubbed out IGPS application

  Revision 1.4  2007/10/02 20:22:42  proctor
  Fixed bug in trajloop's walk_in where zero vectors were converted into
  unit vectors; added ulapi_mutex functions

  Revision 1.3  2007/09/21 19:28:33  proctor
  Pulled the SDL joystick code stubs into gotk.c, and reworked ulapi
  so that task/thread functions were available.

  Revision 1.2  2007/05/16 12:36:12  proctor
  Fixed genserkins iterative inverse, adding timing printouts to genkintest.

  Revision 1.1  2007/02/08 18:51:59  proctor
  Initial checkin

*/


