/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simware.h

  \brief Declarations of simware data structures exchanged between
  the superior and inferior skins.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.1 $
  $Date: 2009/03/30 18:30:44 $
  \endcode

  \author Fred Proctor
*/

#ifndef SIMWARE_H
#define SIMWARE_H

#define SW_NAME_MAX 80

/* from the INIT string */
typedef struct {
  char name[SW_NAME_MAX];
  char robot_type[SW_NAME_MAX];
  double north;		/* right hand coordinate system with z-down */
  double east;
  double elevation;
  double roll;
  double pitch;
  double yaw;
} init_struct;

typedef struct {
  char type[SW_NAME_MAX];
  char errorString[SW_NAME_MAX];
  int status;                    /*!< -1 error, 0 ok */
} response_struct;

typedef struct {
  double left;			/*!< left speed  */
  double right;			/*!< right speed */
} skid_struct;

typedef struct {
  double RFID_value;            /*!< value to place in RFID tag of package */
}boxchute_struct;

typedef struct {
  double speed;			/*!< commanded speed  */
  double minSpeed;		/*!< minimum speed */
} conveyor_struct;

#define MAX_CARGO 20
typedef struct {
  int numCargo;
  int id[MAX_CARGO];
  int memory[MAX_CARGO];
  double x[MAX_CARGO];
  double y[MAX_CARGO];
  double z[MAX_CARGO];
  double roll[MAX_CARGO];
  double pitch[MAX_CARGO];
  double yaw[MAX_CARGO];
} cargo_struct;

#define SW_SENSOR_ENCODER_MAX 8	/*!< how many encoder sensors we can have  */

typedef struct {
  double resolution;
  double x;
  double y;
  double z;
  double roll;
  double pitch;
  double yaw;
  int tick[SW_SENSOR_ENCODER_MAX];
  int number;
} encoder_struct;

#define SW_SENSOR_SONAR_MAX 32	/*!< how many sonar sensors we can have  */

typedef struct {
  double range[SW_SENSOR_SONAR_MAX];
  double minrange;
  double maxrange;
  double beamangle;
  double x;
  double y;
  double z;
  double roll;
  double pitch;
  double yaw;
  int number;			/*!< how many there are  */
} sonar_struct;

#define SW_SENSOR_RANGESCANNER_MAX 192	/*!< how many ranges we can have  */

typedef struct {
  double range[SW_SENSOR_RANGESCANNER_MAX];
  double maxrange;
  double minrange;
  double resolution;
  double fov;
  double x;
  double y;
  double z;
  double roll;
  double pitch;
  double yaw;
  int number;			/*!< how many there are  */
} rangescanner_struct;

typedef enum {
  /* leave 0 as the 'not assigned' convention */
  SW_NONE = 0,
  /* commands */
  SW_COMMAND_BOXCHUTE_RELEASE,
  SW_COMMAND_CLEAR_CARGO,
  SW_COMMAND_CONVEYOR_SPEED,
  SW_COMMAND_INIT,
  SW_COMMAND_RESPONSE,
  SW_COMMAND_SKID,
  /* sensor messages */
  SW_SENSOR_ENCODER,
  SW_SENSOR_SONAR,
  SW_SENSOR_RANGESCANNER,
  SW_SENSOR_TOUCH,
  SW_SENSOR_CO2,
  SW_SENSOR_GROUNDTRUTH,
  SW_SENSOR_INS,
  SW_SENSOR_GPS,
  SW_SENSOR_ODOMETRY,
  SW_SENSOR_VICTIM,
  SW_SENSOR_CARGO,
  /* config messages */
  SW_CONF_ENCODER,
  SW_CONF_SONAR,
  SW_CONF_RANGESCANNER,
  SW_CONF_TOUCH,
  SW_CONF_CO2,
  SW_CONF_GROUNDTRUTH,
  SW_CONF_INS,
  SW_CONF_GPS,
  SW_CONF_ODOMETRY,
  SW_CONF_VICTIM,
  /* geometry messages */
  SW_GEO_ENCODER,
  SW_GEO_SONAR,
  SW_GEO_RANGESCANNER,
  SW_GEO_TOUCH,
  SW_GEO_CO2,
  SW_GEO_GROUNDTRUTH,
  SW_GEO_INS,
  SW_GEO_GPS,
  SW_GEO_ODOMETRY,
  SW_GEO_VICTIM,
  /* last one, for diags */
  SW_TYPE_LAST
} sw_type;

typedef struct {
  double time;
  sw_type type;			/*!< one of 'sw_type' enumerations  */
  /*!
    Simware resources are identified by string names.
   */
  char name[SW_NAME_MAX];
  union {
    /* house keeping */
    init_struct initRobot;
    response_struct response;
    /* actuators */
    boxchute_struct boxchute;
    conveyor_struct conveyor;
    skid_struct skid;
    /* sensors */
    cargo_struct cargo;
    encoder_struct encoder;
    sonar_struct sonar;
    rangescanner_struct rangescanner;
  } data;
} simware_struct;

#define simware_set_name(ptr,str) strncpy((ptr)->name,(str),SW_NAME_MAX); (ptr)->name[SW_NAME_MAX - 1] = 0
#define simware_clear_name(ptr) ((ptr)->name[0] = 0)
#define simware_no_name(ptr) (0 == (ptr)->name[0])
#define simware_match_name(ptr,str) (! strcmp((ptr->name),(str)))
#define simware_init(ptr) simware_clear_name(ptr); (ptr)->time = 0; (ptr)->type = SW_NONE
#endif

/*
  Modification history:

  $Log: simware.h,v $
  Revision 1.1  2009/03/30 18:30:44  dr_steveb
  These files are redundant and will be deleted soon!

*/

