/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file primShell.cc

  \brief Command line shell to vehicle prim level NML buffers.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.34 $
  $Date: 2009/02/27 21:08:13 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <string.h>		// strncmp()
#include <stdlib.h>		// getenv()
#include <ctype.h>		// isspace
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "navDataExt.hh"
#include "primMobJA.hh"		// primMobJA_format()
#include "primMisJA.hh"
#include "primSP.hh"

static NML *navDataExtBuf = NULL;

static RCS_CMD_CHANNEL *primMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *primMobJAStatBuf = NULL;
static RCS_CMD_CHANNEL *primMobJACfgBuf = NULL;
static RCS_STAT_CHANNEL *primMobJASetBuf = NULL;

static RCS_CMD_CHANNEL *primMisJACmdBuf = NULL;
static RCS_STAT_CHANNEL *primMisJAStatBuf = NULL;
static RCS_CMD_CHANNEL *primMisJACfgBuf = NULL;
static RCS_STAT_CHANNEL *primMisJASetBuf = NULL;

static RCS_CMD_CHANNEL *primSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *primSPStatBuf = NULL;
static RCS_CMD_CHANNEL *primSPCfgBuf = NULL;
static RCS_STAT_CHANNEL *primSPSetBuf = NULL;
static int mobSerialNumber = 1;	// odd for shells
static int misSerialNumber = 1;
static int spSerialNumber = 1;	

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "primShell"
#define MOB_WRITE(msg) mobSerialNumber += 2; (msg).serial_number = mobSerialNumber; primMobJACmdBuf->write(&(msg))

#define MIS_WRITE(msg) misSerialNumber += 2; (msg).serial_number = misSerialNumber; primMisJACmdBuf->write(&(msg))

#define SP_WRITE(msg) spSerialNumber += 2; (msg).serial_number = spSerialNumber; primSPCmdBuf->write(&(msg))


enum {SEE_NONE, SEE_ADMIN, SEE_MOB, SEE_MIS, SEE_SP};
enum {PRINT_STAT, PRINT_SET};

static int mobOK = 0;
static int misOK = 0;
static int spOK = 0;

static int cleanupPrimMobJANmlBuffers(void)
{
  if (NULL != navDataExtBuf)
    delete navDataExtBuf;
  if (NULL != primMobJACmdBuf)
    delete primMobJACmdBuf;
  if (NULL != primMobJAStatBuf)
    delete primMobJAStatBuf;
  if (NULL != primMobJASetBuf)
    delete primMobJASetBuf;
  if (NULL != primMobJACfgBuf)
    delete primMobJACfgBuf;

  navDataExtBuf = NULL;
  primMobJAStatBuf = NULL;
  primMobJASetBuf = NULL;
  primMobJACmdBuf = NULL;
  primMobJACfgBuf = NULL;

  mobOK = 0;

  return 0;
}

static int initPrimMobJANmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupPrimMobJANmlBuffers();

  sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
  navDataExtBuf =
    new NML(navDataExt_format, chanName, THIS_PROCESS, config_file);
  if (!navDataExtBuf->valid()) {
    (void) cleanupPrimMobJANmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MOB_JA_CMD_NAME, bufnum);
  primMobJACmdBuf =
    new RCS_CMD_CHANNEL(primMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!primMobJACmdBuf->valid()) {
    (void) cleanupPrimMobJANmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MOB_JA_STAT_NAME, bufnum);
  primMobJAStatBuf =
    new RCS_STAT_CHANNEL(primMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primMobJAStatBuf->valid()) {
    (void) cleanupPrimMobJANmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MOB_JA_CFG_NAME, bufnum);
  primMobJACfgBuf =
    new RCS_CMD_CHANNEL(primMobJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!primMobJACfgBuf->valid()) {
    (void) cleanupPrimMobJANmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MOB_JA_SET_NAME, bufnum);
  primMobJASetBuf =
    new RCS_STAT_CHANNEL(primMobJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primMobJASetBuf->valid()) {
    (void) cleanupPrimMobJANmlBuffers();
    return 1;
  }

  for (double end = etime() + 2.0; etime() < end; esleep(0.1)) {
    navDataExtBuf->read();
    primMobJAStatBuf->read();
    primMobJASetBuf->read();
    // nav data does not start until after an init command so change this...
    //    if (((NavDataExt *) (navDataExtBuf->get_address()))->type ==
    //	NAV_DATA_EXT_TYPE &&
    if( ((PrimMobJAStat *) (primMobJAStatBuf->get_address()))->type ==
	PRIM_MOB_JA_STAT_TYPE &&
	((PrimMobJASet *) (primMobJASetBuf->get_address()))->type ==
	PRIM_MOB_JA_SET_TYPE) {
      mobOK = 1;
      return 0;
    }
  }

  (void) cleanupPrimMobJANmlBuffers();
  mobOK = 0;
  return 1;
}

static int cleanupPrimMisJANmlBuffers(void)
{
  if (NULL != primMisJACmdBuf)
    delete primMisJACmdBuf;
  if (NULL != primMisJAStatBuf)
    delete primMisJAStatBuf;
  if (NULL != primMisJASetBuf)
    delete primMisJASetBuf;
  if (NULL != primMisJACfgBuf)
    delete primMisJACfgBuf;

  primMisJAStatBuf = NULL;
  primMisJASetBuf = NULL;
  primMisJACmdBuf = NULL;
  primMisJACfgBuf = NULL;

  misOK = 0;

  return 0;
}

static int initPrimMisJANmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupPrimMisJANmlBuffers();

  sprintf(chanName, "%s%d", PRIM_MIS_JA_CMD_NAME, bufnum);
  primMisJACmdBuf =
    new RCS_CMD_CHANNEL(primMisJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!primMisJACmdBuf->valid()) {
    (void) cleanupPrimMisJANmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MIS_JA_STAT_NAME, bufnum);
  primMisJAStatBuf =
    new RCS_STAT_CHANNEL(primMisJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primMisJAStatBuf->valid()) {
    (void) cleanupPrimMisJANmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MIS_JA_CFG_NAME, bufnum);
  primMisJACfgBuf =
    new RCS_CMD_CHANNEL(primMisJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!primMisJACfgBuf->valid()) {
    (void) cleanupPrimMisJANmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_MIS_JA_SET_NAME, bufnum);
  primMisJASetBuf =
    new RCS_STAT_CHANNEL(primMisJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primMisJASetBuf->valid()) {
    (void) cleanupPrimMisJANmlBuffers();
    return 1;
  }

  for (double end = etime() + 1.0; etime() < end; esleep(0.1)) {
    primMisJAStatBuf->read();
    primMisJASetBuf->read();
    if (((PrimMisJAStat *) (primMisJAStatBuf->get_address()))->type ==
	PRIM_MIS_JA_STAT_TYPE &&
	((PrimMisJASet *) (primMisJASetBuf->get_address()))->type ==
	PRIM_MIS_JA_SET_TYPE) {
      misOK = 1;
      return 0;
    }
  }

  (void) cleanupPrimMisJANmlBuffers();
  misOK = 0;
  return 1;
}

static int cleanupPrimSPNmlBuffers(void)
{
  if (NULL != primSPCmdBuf)
    delete primSPCmdBuf;
  if (NULL != primSPStatBuf)
    delete primSPStatBuf;
  if (NULL != primSPSetBuf)
    delete primSPSetBuf;
  if (NULL != primSPCfgBuf)
    delete primSPCfgBuf;

  primSPStatBuf = NULL;
  primSPSetBuf = NULL;
  primSPCmdBuf = NULL;
  primSPCfgBuf = NULL;

  spOK = 0;

  return 0;
}

static int initPrimSPNmlBuffers(char *config_file, int bufnum)
{
  printf("\nInside initPrimSPNmlBuffers\n");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupPrimSPNmlBuffers();

  sprintf(chanName, "%s%d", PRIM_SP_CMD_NAME, bufnum);
  primSPCmdBuf =
    new RCS_CMD_CHANNEL(primSP_format, chanName,
			THIS_PROCESS, config_file);
  if (!primSPCmdBuf->valid()) {
    (void) cleanupPrimSPNmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_SP_STAT_NAME, bufnum);
  primSPStatBuf =
    new RCS_STAT_CHANNEL(primSP_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primSPStatBuf->valid()) {
    (void) cleanupPrimSPNmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_SP_CFG_NAME, bufnum);
  primSPCfgBuf =
    new RCS_CMD_CHANNEL(primSP_format, chanName,
			THIS_PROCESS, config_file);
  if (!primSPCfgBuf->valid()) {
    (void) cleanupPrimSPNmlBuffers();
    return 1;
  }

  sprintf(chanName, "%s%d", PRIM_SP_SET_NAME, bufnum);
  primSPSetBuf =
    new RCS_STAT_CHANNEL(primSP_format, chanName,
			 THIS_PROCESS, config_file);
  if (!primSPSetBuf->valid()) {
    (void) cleanupPrimSPNmlBuffers();
    return 1;
  }
  
  for (double end = etime() + 1.0; etime() < end; esleep(0.1)) {
    primSPStatBuf->read();
    primSPSetBuf->read();
    if (((PrimSPStat *) (primSPStatBuf->get_address()))->type ==
	PRIM_SP_STAT_TYPE &&
	((PrimSPSet *) (primSPSetBuf->get_address()))->type ==
	PRIM_SP_SET_TYPE) {
      spOK = 1;
      return 0;
    }
  }
  
  (void) cleanupPrimSPNmlBuffers();
  spOK = 0;
  return 1;
}

#define rcs_stat_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

#define rcs_admin_state_to_string(s) (s) == ADMIN_INITIALIZED ? "Initialized" : (s) == ADMIN_UNINITIALIZED ? "Uninitialized" : (s) == ADMIN_SHUT_DOWN ? "Shut Down" : "?"

#define DEG(x) (57.295779513082323 * (x))
#define RAD(x)  (0.017453292519943 * (x))

static void printAdmin(void)
{

  printf( "========================================\n" );
  printf( "************ Admin Status **************\n" );
  printf( "========================================\n" );
  printf("Mobility\n");
  if( mobOK ) 
    {
      PrimMobJAStat *mob = (PrimMobJAStat*)primMobJAStatBuf->get_address();
      printf("   Admin State  : %s\n", rcs_admin_state_to_string(((PrimMobJAStat *)mob)->admin_state));
      printf("   command_type : %s\n", primMobJA_symbol_lookup(((PrimMobJAStat *)mob)->command_type));
      printf("   Status       : %s\n", rcs_stat_to_string(((PrimMobJAStat *)mob)->status));
      printf("   HeartBeat    : %d\n", ((PrimMobJAStat *)mob)->tt.count);
    }
 
  printf("Sensor Processing\n");
  if( spOK )
    {
      PrimSPStat *sp = (PrimSPStat*)primSPStatBuf->get_address();
      printf("   Admin State  : %s\n", rcs_admin_state_to_string(((PrimSPStat *)sp)->admin_state));
      printf("   command_type : %s\n", primSP_symbol_lookup(((PrimSPStat *)sp)->command_type));
      printf("   Status       : %s\n", rcs_stat_to_string(((PrimSPStat *)sp)->status));
      printf("   HeartBeat    : %d\n", ((PrimSPStat *)sp)->tt.count);
    }
  printf( "Mission Package\n" );
  if( misOK )
    {
      PrimMisJAStat *mis = (PrimMisJAStat*)primMisJAStatBuf->get_address();
      printf("   Admin State  : %s\n", rcs_admin_state_to_string(((PrimMisJAStat *)mis)->admin_state));
      printf("   command_type : %s\n", primSP_symbol_lookup(((PrimMisJAStat *)mis)->command_type));
      printf("   Status       : %s\n", rcs_stat_to_string(((PrimMisJAStat *)mis)->status));
      printf("   HeartBeat    : %d\n", ((PrimMisJAStat *)mis)->tt.count);
    }

}

static void printMobStat(void)
{
  NavDataExt * nav = (NavDataExt *) navDataExtBuf->get_address();
  PrimMobJAStat * stat = (PrimMobJAStat *) primMobJAStatBuf->get_address();

  if (! mobOK) return;

  printf("command_type:       %s\n",
	 primMobJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("serial_number_sent: %d\n", (int) stat->serial_number_sent);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
  printf("pathIndex:          %d\n", (int) stat->pathIndex);
  printf("UTMZone:            %d%c\n", (int)nav->utmZone,nav->utmLetter);
  printf("tranGndTruth:       %f %f %f\n", (double) nav->tranAbs.x,
	 (double) nav->tranTrue.y, (double) nav->tranTrue.z);
  printf("rpyGndTruth:        %f %f %f\n", (double) DEG(nav->rpyAbs.r),
	 (double) DEG(nav->rpyTrue.p), (double) DEG(nav->rpyTrue.y));
  printf("tranAbs:            %f %f %f\n", (double) nav->tranAbs.x,
	 (double) nav->tranAbs.y, (double) nav->tranAbs.z);
  printf("rpyAbs:             %f %f %f\n", (double) DEG(nav->rpyAbs.r),
	 (double) DEG(nav->rpyAbs.p), (double) DEG(nav->rpyAbs.y));
  printf("tranRel:            %f %f %f\n", (double) nav->tranRel.x,
	 (double) nav->tranRel.y, (double) nav->tranRel.z);
  printf("rpyRel:             %f %f %f\n", (double) DEG(nav->rpyRel.r),
	 (double) DEG(nav->rpyRel.p), (double) DEG(nav->rpyRel.y));
  printf("Time:               %f\n", nav->time );
}

static void printMisStat(void)
{
  PrimMisJAStat * stat = (PrimMisJAStat *) primMisJAStatBuf->get_address();

  if (! misOK) return;

  printf("command_type:       %s\n",
	 primMisJA_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
  printf("pose:               %f %f %f\n", (double) stat->pose.tran.x, (double) stat->pose.tran.y, (double) stat->pose.tran.z);
  printf("inpos:              %d\n", (int) stat->inpos);
}

static void printSPStat(void)
{
  PrimSPStat * stat = (PrimSPStat *) primSPStatBuf->get_address();

  if (! spOK) return;

  printf("command_type:       %s\n",
	 primSP_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("admin_state:        %s\n", rcs_admin_state_to_string(stat->admin_state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("timetracker:        %d\n", (int) stat->tt.count);
}

static void printMobSet(void)
{
  PrimMobJASet *set = (PrimMobJASet *) primMobJASetBuf->get_address();
  PM_RPY rpy;

  if (! mobOK) return;

  rpy = set->controlPoint.rot;

  printf("command_type:       %s\n",
	 primMobJA_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("line:               %d\n", (int) set->line);
  printf("source_line:        %d\n", (int) set->source_line);
  printf("source_file:        %s\n", (char *) set->source_file);
  printf("cycleTime:          %f\n", (double) set->cycleTime);
  printf("debug:              0x%X\n", (int) set->debug);
  printf("vmax:               %f\n", (double) set->vmax);
  printf("amax:               %f\n", (double) set->amax);
  printf("wmax:               %f\n", (double) TO_DEG * set->wmax);
  printf("alphamax:           %f\n", (double) TO_DEG * set->alphamax);
  printf("vCutoffAngle:       %f\n", (double) TO_DEG * set->vCutoffAngle);
  printf("wCutoffAngle:       %f\n", (double) TO_DEG * set->wCutoffAngle);
  printf("controlPoint:       %f %f %f %f %f %f\n",
	 (double) set->controlPoint.tran.x,
	 (double) set->controlPoint.tran.y,
	 (double) set->controlPoint.tran.z,
	 (double) TO_DEG * rpy.r,
	 (double) TO_DEG * rpy.p,
	 (double) TO_DEG * rpy.y);
}

static void printMisSet(void)
{
  PrimMisJASet *set = (PrimMisJASet *) primMisJASetBuf->get_address();

  if (!misOK) return;

  printf("command_type:       %s\n",
	 primMisJA_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("line:               %d\n", (int) set->line);
  printf("source_line:        %d\n", (int) set->source_line);
  printf("source_file:        %s\n", (char *) set->source_file);
  printf("cycleTime:          %f\n", (double) set->cycleTime);
  printf("debug:              0x%X\n", (int) set->debug);
}

static void printSPSet(void)
{
  PrimSPSet *set = (PrimSPSet *) primSPSetBuf->get_address();

  if (!spOK) return;

  printf("command_type:       %s\n",
	 primSP_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("line:               %d\n", (int) set->line);
  printf("source_line:        %d\n", (int) set->source_line);
  printf("source_file:        %s\n", (char *) set->source_file);
  printf("cycleTime:          %f\n", (float) set->cycleTime);
  printf("debug:              0x%X\n", (int) set->debug);
}

static int printIt(int see, int print)
{
  if (see == SEE_MOB) {
    if (print == PRINT_STAT) printMobStat();
    else printMobSet();
  } else if (see == SEE_MIS) {
    if (print == PRINT_STAT) printMisStat();
    else printMisSet();
  } else if (see == SEE_SP) {
    if (print == PRINT_STAT) printSPStat();
    else printSPSet();
  } else if (see == SEE_ADMIN) {
    printAdmin();
  } else {
    printf("nothing connected\n");
  }

  return 0;
}

static void printMobHelp(void)
{
  printf("admin             -- set admin mode\n" );
  printf("mob               -- set mob mode\n" );
  printf("mis               -- set mis mode\n" );
  printf("sp                -- set sp mode\n" );
  printf("q | quit          -- quit\n");
  printf("? | help          -- print this help\n");
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
  printf("point <x> <y> <tol> <speed> -- drive to a single point\n");
  printf("curve <cent x> <cent y> <end x> <end y> <theta> <tol> <speed> -- drive along the given arc\n");
  printf("arc <file>        -- send arcs in <file>\n");
  printf("wp <file>         -- send waypoints in <file>\n");
  printf("rotate <theta> <theta tol> cw | ccw -- rotate to angle\n");
  printf("vel <v> <w>       -- set speeds to v, w\n");
  printf("stop              -- stop now\n");
  printf("dock <station name> <vector file> -- dock with the given station that is specified in the given file\n");
  printf("ct <secs>         -- set cycle time to <secs>\n");
  printf("pars <6, vmax...> -- configure vmax, amax, wmax, alphamax, vcut, wcut\n");
  printf("debug <0, 1, ...> -- configure debug level\n");
}

static void printMisHelp(void)
{
  printf("admin             -- set admin mode\n" );
  printf("mob               -- set mob mode\n" );
  printf("mis               -- set mis mode\n" );
  printf("sp                -- set sp mode\n" );
  printf("q | quit                -- quit\n");
  printf("? | help                -- print this help\n");
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
}

static void printSPHelp(void)
{
  printf("admin             -- set admin mode\n" );
  printf("mob               -- set mob mode\n" );
  printf("mis               -- set mis mode\n" );
  printf("sp                -- set sp mode\n" );
  printf("q | quit                -- quit\n");
  printf("? | help                -- print this help\n");
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
  printf("go                      -- start SP subsystem\n");
}

static void printAdminHelp(void)
{
  printf("admin             -- set admin mode\n" );
  printf("mob               -- set mob mode\n" );
  printf("mis               -- set mis mode\n" );
  printf("sp                -- set sp mode\n" );
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
}

static void printHelp(int see)
{
  if (see == SEE_MOB) {
    printMobHelp();
  } else if (see == SEE_MIS) {
    printMisHelp();
  } else if (see == SEE_SP) {
    printSPHelp();
  } else if (see == SEE_ADMIN) {
    printAdminHelp();
  } else {
    printf("nothing connected\n");
  }

  return;
}

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}

static void convertAbsToRel(double x, double y, double z,
			    double * xx, double * yy, double * zz)
{
  PM_CARTESIAN tran;

  tran = ((NavDataExt *) (navDataExtBuf->get_address()))->absToRel * PM_CARTESIAN(x, y, z);
  *xx = tran.x, *yy = tran.y, *zz = tran.z;

  return;
}

static int loadPath(char *filename, ArcSegment * path, int *path_length)
{
  FILE *fp;
#define BUFFERLEN 256
  char buffer[BUFFERLEN];
  char *ptr;
  ArcSegment *path_ptr;
  int line;
  int abs;
  int retval;

  if (NULL == (fp = fopen(filename, "r")))
    return 1;

  abs = 1;			// default is absolute coords
  *path_length = 0;
  path_ptr = &path[0];
  line = 0;
  retval = 0;
  while (!feof(fp)) {
    if (NULL == fgets(buffer, BUFFERLEN, fp))
      break;
    line++;
    ptr = buffer;
    while (isspace(*ptr))
      ptr++;
    if (0 == *ptr)
      continue;			// blank line
    if ('#' == *ptr)
      continue;			// comment
    if (! strncmp(ptr, "abs", 3)) abs = 1;
    else if (! strncmp(ptr, "rel", 3)) abs = 0;
    else if (7 == sscanf(ptr, "%lf %lf %lf %lf %lf %lf %lf",
		    &path_ptr->center.x,
		    &path_ptr->center.y,
		    &path_ptr->end.x,
		    &path_ptr->end.y,
		    &path_ptr->theta,
		    &path_ptr->annular_tol,
		    &path_ptr->speed)) {
      /* FIXME-- we are in the process of adding 3D waypoints, but deferring
	 it here, so set the unused Z to 0 */
      path_ptr->center.z = path_ptr->end.z = 0.0;
      if (abs) {
	convertAbsToRel(path_ptr->center.x,
			path_ptr->center.y,
			path_ptr->center.z,
			&path_ptr->center.x,
			&path_ptr->center.y,
			&path_ptr->center.z);
	convertAbsToRel(path_ptr->end.x,
			path_ptr->end.y,
			path_ptr->end.z,
			&path_ptr->end.x,
			&path_ptr->end.y,
			&path_ptr->end.z);
      }
      path_ptr->isArc = 1;
      (*path_length)++;
      path_ptr++;
      if (*path_length >= MAX_ARC_SEGMENTS) break;
    } else if (4 == sscanf(ptr, "%lf %lf %lf %lf",
			   &path_ptr->end.x,
			   &path_ptr->end.y,
			   &path_ptr->annular_tol,
			   &path_ptr->speed)) {
      path_ptr->center.z = path_ptr->end.z = 0.0; // FIXME per comment above
      if (abs) {
	convertAbsToRel(path_ptr->end.x, 
			path_ptr->end.y,
			path_ptr->end.z,
			&path_ptr->end.x,
			&path_ptr->end.y,
			&path_ptr->end.z);
      }
      path_ptr->isArc = 0;
      (*path_length)++;
      path_ptr++;
      if (*path_length >= MAX_ARC_SEGMENTS) break;
    } else {
      fprintf(stderr, "syntax error in %s, line %d: %s\n",
	      filename, line, buffer);
      retval = 1;
      break;
    }
    // else got a good line
  }

  fclose(fp);
  return retval;
#undef BUFFERLEN
}

static int loadWaypoints(char * filename,
			 Waypoint * waypoint,
			 int * waypoint_length)
{
  FILE *fp;
#define BUFFERLEN 256
  char buffer[BUFFERLEN];
  char *ptr;
  Waypoint *wptr;
  int line;
  int abs;
  int retval;

  if (NULL == (fp = fopen(filename, "r")))
    return 1;

  abs = 1;			// default is absolute coords
  *waypoint_length = 0;
  wptr= &waypoint[0];
  line = 0;
  retval = 0;
  while (!feof(fp)) {
    if (NULL == fgets(buffer, BUFFERLEN, fp))
      break;
    line++;
    ptr = buffer;
    while (isspace(*ptr))
      ptr++;
    if (0 == *ptr)
      continue;			// blank line
    if ('#' == *ptr)
      continue;			// comment
    if (! strncmp(ptr, "abs", 3)) abs = 1;
    else if (! strncmp(ptr, "rel", 3)) abs = 0;
    else if (5 == sscanf(ptr, "%lf %lf %lf %lf %lf",
			 &wptr->p.x,
			 &wptr->p.y,
			 &wptr->p.z,
			 &wptr->neighborhood,
			 &wptr->speed)) {
      if (abs) {
	convertAbsToRel(wptr->p.x, wptr->p.y, wptr->p.z,
			&wptr->p.x, &wptr->p.y, &wptr->p.z);
      }
      (*waypoint_length)++;
      wptr++;
      if (*waypoint_length >= MAX_WAYPOINTS) break;
    } else if (4 == sscanf(ptr, "%lf %lf %lf %lf",
			   &wptr->p.x,
			   &wptr->p.y,
			   &wptr->neighborhood,
			   &wptr->speed)) {
      wptr->p.z = 0.0;
      if (abs) {
	convertAbsToRel(wptr->p.x, wptr->p.y, wptr->p.z,
			&wptr->p.x, &wptr->p.y, &wptr->p.z);
      }
      wptr->p.z = 0.0;
      (*waypoint_length)++;
      wptr++;
      if (*waypoint_length >= MAX_WAYPOINTS) break;
    } else {
      fprintf(stderr, "syntax error in %s, line %d: %s\n",
	      filename, line, buffer);
      retval = 1;
      break;
    }
  }

  fclose(fp);
  return retval;
#undef BUFFERLEN
}

static int doCmdDock(char *ptr, Waypoint * waypoint,
                         int * waypoint_length)
{
  char *stationName, *fileName;
  PM_CARTESIAN penultimate, goal;
  int retval=0;
  Waypoint *wptr;
  PrimMobJACmdDock primMobJACmdDock;
  bool loadAtBottom = true;

  *waypoint_length = 0;
  wptr= &waypoint[0];

  /*
  while (!isspace(*ptr) && 0 != *ptr)
    ptr++;			// skip over word
  while (isspace(*ptr))
    ptr++;
  if (0 == *ptr) {
    printf("need <station name> <vector file name>\n");
    return 1;
  }
  stationName = ptr;
  while (!isspace(*ptr) && 0 != *ptr)
    ptr++;			// skip over word
  fileName = ptr;
  printf( "station: %s file: %s\n", stationName, fileName );
  */
  
  /*
  primMobJACmdDock.freeSpaceUpperRight.x = 5.12;
  primMobJACmdDock.freeSpaceUpperRight.y = 5.45;
  primMobJACmdDock.freeSpaceLowerLeft.x = -4.88;
  primMobJACmdDock.freeSpaceLowerLeft.y = -4.57;
  primMobJACmdDock.stationUpperRight.x = .366;
  primMobJACmdDock.stationUpperRight.y = -2.93;
  primMobJACmdDock.stationLowerLeft.x = -.48;
  primMobJACmdDock.stationLowerLeft.y = -4.55;
  primMobJACmdDock.dockPoint.x = -.06;
  primMobJACmdDock.dockPoint.y = -2.93;

  primMobJACmdDock.stationUpperRight.x = 10.61;
  primMobJACmdDock.stationUpperRight.y = -2.32;
  primMobJACmdDock.stationLowerLeft.x = 9.76;
  primMobJACmdDock.stationLowerLeft.y = -3.95;
  */
/*
  primMobJACmdDock.stationUpperRight.x = 18.83;
  primMobJACmdDock.stationUpperRight.y = -1.30;
  primMobJACmdDock.stationLowerLeft.x = 17.96;
  primMobJACmdDock.stationLowerLeft.y = -2.91;
*/
/*
  primMobJACmdDock.stationUpperRight.x = 25.05;
  primMobJACmdDock.stationUpperRight.y = -0.37;
  primMobJACmdDock.stationLowerLeft.x = 24.19;
  primMobJACmdDock.stationLowerLeft.y = -1.96;
*/

/* values for VMAC1 world loading station
 */
  loadAtBottom = false;
  primMobJACmdDock.stationLowerLeft.x = 2843. / 250.;
  primMobJACmdDock.stationLowerLeft.y = 6968. / 250.;
  primMobJACmdDock.stationUpperRight.x = 3033. / 250.;
  primMobJACmdDock.stationUpperRight.y = 7341. / 250.;

  if( loadAtBottom ) // load at bottom of station (as seen in editor)
    {
      penultimate.x = primMobJACmdDock.stationLowerLeft.x  -.49;
      penultimate.y = primMobJACmdDock.stationUpperRight.y +.54;
      
      goal.x = primMobJACmdDock.stationLowerLeft.x         +.26;
      goal.y = primMobJACmdDock.stationUpperRight.y        +.54;
    }
  else // load from top of sation (as seen in editor), nothing given for side loading
    {
      penultimate.x = primMobJACmdDock.stationUpperRight.x  +.49;
      penultimate.y = primMobJACmdDock.stationLowerLeft.y -.54;
      
      goal.x = primMobJACmdDock.stationLowerLeft.x         +.39;
      goal.y = primMobJACmdDock.stationLowerLeft.y         -.64;
    }
  printf( "Penultimate points: <%f %f> goal <%f %f>\n",
	  penultimate.x, penultimate.y, goal.x, goal.y );
  wptr->p.x = penultimate.x;
  wptr->p.y = penultimate.y;
  wptr->p.z = 0.0;
  wptr->neighborhood = .3;
  wptr->speed = 5;

  convertAbsToRel(wptr->p.x, wptr->p.y, wptr->p.z,
	&wptr->p.x, &wptr->p.y, &wptr->p.z);

  (*waypoint_length)++;
  wptr++;

  wptr->p.x = goal.x;
  wptr->p.y = goal.y;
  wptr->p.z = 0.0;
  wptr->neighborhood = .3;
  wptr->speed = -1;

  convertAbsToRel(wptr->p.x, wptr->p.y, wptr->p.z,
	&wptr->p.x, &wptr->p.y, &wptr->p.z);

  (*waypoint_length)++;
  wptr++;

  return retval;
}

#define skipwhite(ptr) while(isspace(*(ptr))) (ptr)++
#define skipnonwhite(ptr) while(!isspace(*(ptr)) && *(ptr) != 0) (ptr)++
#define skipword(ptr) skipnonwhite(ptr); skipwhite(ptr)
#define arrayels(array) sizeof(array) / sizeof(*(array))

// scans a string of numbers into an array, returning number scanned
static int scanem(double array[], char string[], int n)
{
  double d;
  char * nptr;
  char * endptr;
  int i;

  for (i = 0, nptr = string; i < n; i++) {
    d = strtod(nptr, &endptr);
    if (endptr == nptr) break;
    array[i] = d;
    nptr = endptr;
  }

  return i;
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char *line;
  char *ptr;
  int retval;
  int whichSee = SEE_NONE;
  int whichPrint = PRINT_STAT;
  double d1, d2, d3, d4, d5, d6;
  double darray[SERVO_MIS_JA_LINK_MAX];
  double movetime = -1.0;
  int i1;

  PrimMobJAStat *primMobJAStatPtr;
  PrimMobJASet *primMobJASetPtr;
  PrimMobJACmdInit primMobJACmdInit;
  PrimMobJACmdHalt primMobJACmdHalt;
  PrimMobJACmdAbort primMobJACmdAbort;
  PrimMobJACmdShutdown primMobJACmdShutdown;
  PrimMobJACmdMoveArcSegment primMobJACmdMoveArcSegment;
  PrimMobJACmdMoveWaypoint primMobJACmdMoveWaypoint;
  PrimMobJACmdRotate primMobJACmdRotate;
  PrimMobJACmdVel primMobJACmdVel;
  PrimMobJACmdStop primMobJACmdStop;
  PrimMobJACmdDock primMobJACmdDock;
  PrimMobJACfgCycleTime primMobJACfgCycleTime;
  PrimMobJACfgDebug primMobJACfgDebug;
  PrimMobJACfgMove primMobJACfgMove;

  PrimMisJAStat *primMisJAStatPtr;
  PrimMisJASet *primMisJASetPtr;
  PrimMisJACmdInit primMisJACmdInit;
  PrimMisJACmdHalt primMisJACmdHalt;
  PrimMisJACmdAbort primMisJACmdAbort;
  PrimMisJACmdShutdown primMisJACmdShutdown;
  PrimMisJACmdMoveJoint primMisJACmdMoveJoint;
  PrimMisJACmdMoveWorld primMisJACmdMoveWorld;
  PrimMisJACmdStop primMisJACmdStop;
  PrimMisJACfgCycleTime primMisJACfgCycleTime;
  PrimMisJACfgDebug primMisJACfgDebug;

  PrimSPStat *primSPStatPtr;
  PrimSPSet *primSPSetPtr;
  PrimSPCmdInit primSPCmdInit;
  PrimSPCmdAbort primSPCmdAbort;
  PrimSPCmdHalt primSPCmdHalt;
  PrimSPCmdShutdown primSPCmdShutdown;
  PrimSPCmdGo primSPCmdGo;
  PrimSPCfgDebug primSPCfgDebug;

#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      break;
    case 'i':
      INI_FILE = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // allocate all the mobility buffers
  retval = initPrimMobJANmlBuffers(NML_FILE, bufnum);
  if (retval) {
    fprintf(stderr, "%s: can't allocate primMobJA NML buffers, so ignoring them\n", argv[0]);
    primMobJAStatPtr = NULL;
    primMobJASetPtr = NULL;

  }
  else {
    primMobJAStatPtr = (PrimMobJAStat *) primMobJAStatBuf->get_address();
    primMobJASetPtr = (PrimMobJASet *) primMobJASetBuf->get_address();
  }

  // allocate all the mission buffers
  retval = initPrimMisJANmlBuffers(NML_FILE, bufnum);
  if (retval) {
    fprintf(stderr, "%s: can't allocate primMisJA NML buffers, so ignoring them\n", argv[0]);
    primMisJAStatPtr = NULL;
    primMisJASetPtr = NULL;
  }
  else {
    primMisJAStatPtr = (PrimMisJAStat *) primMisJAStatBuf->get_address();
    primMisJASetPtr = (PrimMisJASet *) primMisJASetBuf->get_address();
  }

  // allocate all the sensory processing buffers
  retval = initPrimSPNmlBuffers(NML_FILE, bufnum);
  if (retval) {
    fprintf(stderr, "%s: can't allocate primSP NML buffers, so ignoring them\n", argv[0]);
    primSPStatPtr = NULL;
    primSPSetPtr = NULL;
  }
  else {
    primSPStatPtr = (PrimSPStat *) primSPStatBuf->get_address();
    primSPSetPtr = (PrimSPSet *) primSPSetBuf->get_address();
  }

  if ( mobOK || misOK || spOK ){
  whichSee = SEE_ADMIN;
/*
  if (mobOK) {
    whichSee = SEE_MOB;
  } else if (misOK) {
    whichSee = SEE_MIS;
  } else if (spOK) {
    whichSee = SEE_SP;
*/
  } else {
    whichSee = SEE_NONE;
  }

  while (!feof(stdin)) {
#ifndef HAVE_READLINE_READLINE_H
    printf(whichSee == SEE_ADMIN ? "PrimAdmin> " :
	   whichSee == SEE_MOB ? "PrimMob> " : 
	   whichSee == SEE_MIS ? "PrimMis> " :
	   whichSee == SEE_SP ? "PrimSP> " : "PrimNone> ");
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) {
      break;
    }
    buffer[BUFFERSIZE - 1] = 0;
    line = buffer;
#else
    line = readline(whichSee == SEE_ADMIN ? "PrimAdmin> " :
		    whichSee == SEE_MOB ? "PrimMob> " : 
		    whichSee == SEE_MIS ? "PrimMis> " :
		    whichSee == SEE_SP ? "PrimSP> " : "PrimNone> ");
    if (NULL == line) {
      break;
    }
    if (*line) {
      add_history(line);
    }
#endif

    if (mobOK) {
      navDataExtBuf->read();
      primMobJAStatBuf->read();
      primMobJASetBuf->read();
    }
    if (misOK) {
      primMisJAStatBuf->read();
      primMisJASetBuf->read();
    }
    if (spOK) {
      primSPStatBuf->read();
      primSPSetBuf->read();
    }

    // clean off any trailing whitespace
    ptr = line + strlen(line);	// at the null
    while (--ptr != line && isspace(*ptr)) *ptr = 0;

    ptr = line;
    while (isspace(*ptr))
      ptr++;			// skip leading white space
    if (0 == *ptr) {		// blank line
      printIt(whichSee, whichPrint);
    } else if (ptr[0] == 'q' || !strwcmp(ptr, "quit")) {
      break;
    } else if (ptr[0] == '?' || !strwcmp(ptr, "help")) {
      printHelp(whichSee);
    } else if (!strwcmp(ptr, "admin")) {
      (void) initPrimMobJANmlBuffers(NML_FILE, bufnum);
      if (mobOK) {
	// we need to do this again since the addresses have likely changed
	primMobJAStatPtr = (PrimMobJAStat *) primMobJAStatBuf->get_address();
	primMobJASetPtr = (PrimMobJASet *) primMobJASetBuf->get_address();
      }
      (void) initPrimMisJANmlBuffers(NML_FILE, bufnum);
      if (misOK) {
	// we need to do this again since the addresses have likely changed
	primMisJAStatPtr = (PrimMisJAStat *) primMisJAStatBuf->get_address();
	primMisJASetPtr = (PrimMisJASet *) primMisJASetBuf->get_address();
	whichSee = SEE_ADMIN;
	printIt(whichSee, whichPrint);
      }
      (void) initPrimSPNmlBuffers(NML_FILE, bufnum);
      if (spOK) {
	// we need to do this again since the addresses have likely changed
	primSPStatPtr = (PrimSPStat *) primSPStatBuf->get_address();
	primSPSetPtr = (PrimSPSet *) primSPSetBuf->get_address();
      }
      printIt(whichSee, whichPrint);
    } else if (!strwcmp(ptr, "mob")) {
      (void) initPrimMobJANmlBuffers(NML_FILE, bufnum);
      if (mobOK) {
	// we need to do this again since the addresses have likely changed
	primMobJAStatPtr = (PrimMobJAStat *) primMobJAStatBuf->get_address();
	primMobJASetPtr = (PrimMobJASet *) primMobJASetBuf->get_address();
	whichSee = SEE_MOB;
	printIt(whichSee, whichPrint);
      } else {
	printf("mob not connected\n");
      }
    } else if (!strwcmp(ptr, "mis")) {
      (void) initPrimMisJANmlBuffers(NML_FILE, bufnum);
      if (misOK) {
	// we need to do this again since the addresses have likely changed
	primMisJAStatPtr = (PrimMisJAStat *) primMisJAStatBuf->get_address();
	primMisJASetPtr = (PrimMisJASet *) primMisJASetBuf->get_address();
	whichSee = SEE_MIS;
	printIt(whichSee, whichPrint);
      } else {
	printf("mis not connected\n");
      }
    } else if (!strwcmp(ptr, "sp")) {
      (void) initPrimSPNmlBuffers(NML_FILE, bufnum);
      if (spOK) {
	// we need to do this again since the addresses have likely changed
	primSPStatPtr = (PrimSPStat *) primSPStatBuf->get_address();
	primSPSetPtr = (PrimSPSet *) primSPSetBuf->get_address();
	whichSee = SEE_SP;
	printIt(whichSee, whichPrint);
      } else {
	printf("sp not connected\n");
      }
    } else if (!strwcmp(ptr, "stat")) {
      whichPrint = PRINT_STAT;
      printIt(whichSee, whichPrint);
    } else if (!strwcmp(ptr, "set")) {
      whichPrint = PRINT_SET;
      printIt(whichSee, whichPrint);
    } else if (!strwcmp(ptr, "init")) {
      if (whichSee == SEE_MOB || (whichSee == SEE_ADMIN && mobOK) ) {
	MOB_WRITE(primMobJACmdInit);
      } if (whichSee == SEE_MIS || (whichSee == SEE_ADMIN && misOK)) {
	MIS_WRITE(primMisJACmdInit);
      } if (whichSee == SEE_SP || (whichSee == SEE_ADMIN && spOK)) {
	SP_WRITE(primSPCmdInit);
      } if (!mobOK && !misOK && !spOK ) {
	printf("nothing connected\n");
      }
    } else if (!strwcmp(ptr, "halt")) {
      if (whichSee == SEE_MOB || (whichSee == SEE_ADMIN && mobOK)) {
	MOB_WRITE(primMobJACmdHalt);
      } if (whichSee == SEE_MIS || (whichSee == SEE_ADMIN && misOK)) {
	MIS_WRITE(primMisJACmdHalt);
      } if (whichSee == SEE_SP || (whichSee == SEE_ADMIN && spOK)) {
	SP_WRITE(primSPCmdHalt);
      } if (!mobOK && !misOK && !spOK ) {
	printf("nothing connected\n");
      }
    } else if (!strwcmp(ptr, "abort")) {
      if (whichSee == SEE_MOB || (whichSee == SEE_ADMIN && mobOK)) {
	MOB_WRITE(primMobJACmdAbort);
      } if (whichSee == SEE_MIS || (whichSee == SEE_ADMIN && misOK)) {
	MIS_WRITE(primMisJACmdAbort);
      } if (whichSee == SEE_SP || (whichSee == SEE_ADMIN && spOK) ) {
	SP_WRITE(primSPCmdAbort); 
      } if (!mobOK && !misOK && !spOK ) {
	printf("nothing connected\n");
      }
    } else if (!strwcmp(ptr, "shutdown")) {
      if (whichSee == SEE_MOB || (whichSee == SEE_ADMIN && mobOK)) {
	MOB_WRITE(primMobJACmdShutdown);
      } if (whichSee == SEE_MIS || (whichSee == SEE_ADMIN && misOK)) {
	MIS_WRITE(primMisJACmdShutdown);
      } if (whichSee == SEE_SP || (whichSee == SEE_ADMIN && spOK)) {
	SP_WRITE(primSPCmdShutdown);
      } if (!mobOK && !misOK && !spOK ) {
	printf("nothing connected\n");
      }
    } else if (!strwcmp(ptr, "point")) {
      if (mobOK) {
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need x, y, tollerance, and speed\n");
	  continue;
	}
	if (4 != sscanf(ptr, "%lf %lf %lf %lf",
			&primMobJACmdMoveArcSegment.arcList[0].end.x,
			&primMobJACmdMoveArcSegment.arcList[0].end.y,
			&primMobJACmdMoveArcSegment.arcList[0].annular_tol,
			&primMobJACmdMoveArcSegment.arcList[0].speed)) {
	  printf("bad x, y, tollerance, speed  values: %s\n", ptr);
	} else {
	  primMobJACmdMoveArcSegment.arcList[0].end.z = 0;
	  primMobJACmdMoveArcSegment.arcList[0].isArc = 0;
	  primMobJACmdMoveArcSegment.arcList_length = 1;
	  MOB_WRITE(primMobJACmdMoveArcSegment);
	}
      }
    } else if (!strwcmp(ptr, "curve")) {
      if (mobOK) {
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need cent_x, cent_y, end_x, end_y, theta, tol, and speed\n");
	  continue;
	}
	if (7 != sscanf(ptr, "%lf %lf %lf %lf %lf %lf %lf",
			&primMobJACmdMoveArcSegment.arcList[0].center.x,
			&primMobJACmdMoveArcSegment.arcList[0].center.y,
			&primMobJACmdMoveArcSegment.arcList[0].end.x,
			&primMobJACmdMoveArcSegment.arcList[0].end.y,
			&primMobJACmdMoveArcSegment.arcList[0].theta,
			&primMobJACmdMoveArcSegment.arcList[0].annular_tol,
			&primMobJACmdMoveArcSegment.arcList[0].speed)) {
	  printf("bad cent_x, cent_y, end_x, end_y, theta, tol, speed  values: %s\n", ptr);
	} else {
	  primMobJACmdMoveArcSegment.arcList[0].end.z = 0;
	  primMobJACmdMoveArcSegment.arcList[0].isArc = 1;
	  primMobJACmdMoveArcSegment.arcList_length = 1;
	  MOB_WRITE(primMobJACmdMoveArcSegment);
	}
      }
    } else if (!strwcmp(ptr, "arc")) {
      if (mobOK) {
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need a data file\n");
	  continue;
	}
	if (0 != loadPath(ptr, &primMobJACmdMoveArcSegment.arcList[0],
			  &primMobJACmdMoveArcSegment.arcList_length)) {
	  printf("can't load ``%s''\n", ptr);
	} else {
	  MOB_WRITE(primMobJACmdMoveArcSegment);
	}
      } else {
	printf("mob not connected\n");
      }
    } else if (!strwcmp(ptr, "wp")) {
      if (mobOK) {
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need a data file\n");
	  continue;
	}
	if (0 !=
	    loadWaypoints(ptr,
			  &primMobJACmdMoveWaypoint.waypoint[0],
			  &primMobJACmdMoveWaypoint.waypoint_length)) {
	  printf("can't load ``%s''\n", ptr);
	} else {
	  primMobJACmdMoveWaypoint.prim_path = true;
	  MOB_WRITE(primMobJACmdMoveWaypoint);
	}
      } else {
	printf("mob not connected\n");
      }
    } else if (!strwcmp(ptr, "rotate")) {
      if (mobOK) {
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need <theta> <thetaTol>, optional cw, ccw\n");
	  continue;
	}
	if (2 != sscanf(ptr, "%lf %lf",
			&d1,
			&d2)) {
	  printf("bad theta, thetaTol values: %s\n", ptr);
	} else {
	  while (!isspace(*ptr) && 0 != *ptr) ptr++; // skip theta
	  while (isspace(*ptr)) ptr++; 
	  while (!isspace(*ptr) && 0 != *ptr) ptr++; // skip thetaTol
	  while (isspace(*ptr)) ptr++; 
	  if (0 == *ptr) {
	    // go shortest route
	    primMobJACmdRotate.direction = MOAST_ROTATE_SHORTEST;
	  } else if (! strwcmp(ptr, "cw")) {
	    primMobJACmdRotate.direction = MOAST_ROTATE_CW;
	  } else if (! strwcmp(ptr, "ccw")) {
	    primMobJACmdRotate.direction = MOAST_ROTATE_CCW;
	  } else {
	    printf("need cw or ccw for direction\n");
	    continue;
	  }
	  primMobJACmdRotate.theta = RAD(d1);
	  primMobJACmdRotate.thetaTol = RAD(d2);
	  MOB_WRITE(primMobJACmdRotate);
	}
      } else {
	printf("mob not connected\n");
      }
    } else if (!strwcmp(ptr, "vel")) {
      if (mobOK) {
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need v and w\n");
	  continue;
	}
	if (2 != sscanf(ptr, "%lf %lf",
			&primMobJACmdVel.v,
			&primMobJACmdVel.w)) {
	  printf("bad v, w values: %s\n", ptr);
	} else {
	  MOB_WRITE(primMobJACmdVel);
	}
      } else {
	printf("mob not connected\n");
      }
    } else if (!strwcmp(ptr, "stop")) {
      if (mobOK) {
	MOB_WRITE(primMobJACmdStop);
      } else {
	printf("mob not connected\n");
      }
    } else if (!strwcmp(ptr, "dock")) {
      if (mobOK) {
	//	char *stationName, *fileName;
	/*
	  while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	  while (isspace(*ptr))
	  ptr++;
	  if (0 == *ptr) {
	  printf("need <station name> <vector file name>\n");
	  }
	  stationName = ptr;
	  while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	  fileName = ptr;
	  printf( "station: %s file: %s\n", stationName, fileName );
	*/
	if (0 != doCmdDock(ptr, 
		&primMobJACmdMoveWaypoint.waypoint[0],
		&primMobJACmdMoveWaypoint.waypoint_length)) {
	  printf("can't load make dock info\n");
	} else {
	  MOB_WRITE(primMobJACmdMoveWaypoint);
	}
      } else {
	printf("mob not connected\n");
      }
    } else if (!strwcmp(ptr, "stop")) {
      if (misOK) {
	MIS_WRITE(primMisJACmdStop);
      } else {
	printf("mis not connected\n");
      }
    } else if (!strwcmp(ptr, "time")) {
      if (misOK) {
	if (1 != sscanf(ptr, "%*s %lf", &d1)) {
	  printf("need a time for subsequent moves\n");
	} else {
	  movetime = d1;
	}
      } else {
	printf("mis not connected\n");
      }
    } else if (!strwcmp(ptr, "movej")) {
      if (misOK) {
	skipword(ptr);
	i1 = scanem(darray, ptr, arrayels(darray));
	if (i1 > 0) {
	  primMisJACmdMoveJoint.joints_length = i1;
	  for (int i = 0; i < i1; i++) {
	    primMisJACmdMoveJoint.joints[i] = darray[i];
	  }
	  primMisJACmdMoveJoint.time = movetime;
	  MIS_WRITE(primMisJACmdMoveJoint);
	} else {
	  printf("need <joint values>\n");
	}
      } else {
	printf("mis not connected\n");
      }
    } else if (!strwcmp(ptr, "movew")) {
      if (misOK) {
	if (6 !=
	    sscanf(ptr, "%*s %lf %lf %lf %lf %lf %lf",
		   &d1, &d2, &d3, &d4, &d5, &d6)) {
	  printf("need <x y z r p y>\n");
	} else {
	  PM_RPY rpy;
	  primMisJACmdMoveWorld.end.tran.x = d1;
	  primMisJACmdMoveWorld.end.tran.y = d2;
	  primMisJACmdMoveWorld.end.tran.z = d3;
	  rpy.r = d4, rpy.p = d5, rpy.y = d6;
	  primMisJACmdMoveWorld.end.rot = rpy;
	  primMisJACmdMoveWorld.time = movetime;
	  MIS_WRITE(primMisJACmdMoveWorld);
	}
      } else {
	printf("mis not connected\n");
      }
    } else if (!strwcmp(ptr, "ct")) {
      if (whichSee == SEE_MOB) {
	if (1 !=
	    sscanf(ptr, "%*s %lf",
		   &primMobJACfgCycleTime.cycleTime)) {
	  printf("need a cycleTime argument\n");
	} else {
	  primMobJACfgCycleTime.serial_number =
	    primMobJASetPtr->echo_serial_number + 1;
	  primMobJACfgBuf->
	    write(&primMobJACfgCycleTime);
	}
      } else if (whichSee == SEE_MIS) {
	if (1 !=
	    sscanf(ptr, "%*s %lf",
		   &primMisJACfgCycleTime.cycleTime)) {
	  printf("need a cycleTime argument\n");
	} else {
	  primMisJACfgCycleTime.serial_number =
	    primMisJASetPtr->echo_serial_number + 1;
	  primMisJACfgBuf->
	    write(&primMisJACfgCycleTime);
	}
      } else {
	printf("nothing connected\n");
      }
    } else if (!strwcmp(ptr, "debug")) {
      if (whichSee == SEE_MOB) {
	if (1 != sscanf(ptr, "%*s %i",
			&primMobJACfgDebug.debug)) {
	  printf("need debug mask\n");
	} else {
	  primMobJACfgDebug.serial_number =
	    primMobJASetPtr->echo_serial_number + 1;
	  primMobJACfgBuf->write(&primMobJACfgDebug);
	}
      } else if (whichSee == SEE_MIS) {
	if (1 != sscanf(ptr, "%*s %i",
			&primMisJACfgDebug.debug)) {
	  printf("need debug mask\n");
	} else {
	  primMisJACfgDebug.serial_number =
	    primMisJASetPtr->echo_serial_number + 1;
	  primMisJACfgBuf->write(&primMisJACfgDebug);
	}
      } else if (whichSee == SEE_SP) {
	if (1 != sscanf(ptr, "%*s %i",
			&primSPCfgDebug.debug)) {
	  printf("need debug mask\n");
	} else {
	  primSPCfgDebug.serial_number =
	    primSPSetPtr->echo_serial_number + 1;
	  primSPCfgBuf->write(&primSPCfgDebug);
	}
      } else {
	printf("nothing connected\n");
      }
    } else if (!strwcmp(ptr, "pars")) {
      if (mobOK) {
	if (6 != sscanf(ptr, "%*s %lf %lf %lf %lf %lf %lf",
			&primMobJACfgMove.vmax,
			&primMobJACfgMove.amax,
			&d1,
			&d2,
			&d3,
			&d4)) {
	  printf("need 6 parameters, vmax amax wmax alphamax vcut wcut\n");
	} else {
	  primMobJACfgMove.wmax = RAD(d1);
	  primMobJACfgMove.alphamax = RAD(d2);
	  primMobJACfgMove.vCutoffAngle = RAD(d3);
	  primMobJACfgMove.wCutoffAngle = RAD(d4);
	  primMobJACfgMove.serial_number =
	    primMobJASetPtr->echo_serial_number + 1;
	  primMobJACfgBuf->write(&primMobJACfgMove);
	}
      } else {
	printf("nothing connected\n");
      }
    } else if (!strwcmp(ptr, "go")) {
      if (spOK) {
	SP_WRITE(primSPCmdGo);
	printf( "Go command with serial %d send\n", primSPCmdGo.serial_number);
      }
    } else {
      printf("?\n");
    }

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  (void) cleanupPrimMobJANmlBuffers();
  (void) cleanupPrimMisJANmlBuffers();
  (void) cleanupPrimSPNmlBuffers();

  return 0;
}

/*
  Modification history:

  $Log: primShell.cc,v $
  Revision 1.34  2009/02/27 21:08:13  dr_steveb
  Fixed docking command and prim now issues "done"

  Revision 1.33  2008/12/17 15:24:37  dr_steveb
  Added printing of ground truth to status print

  Revision 1.32  2008/12/05 20:34:12  dr_steveb
  Fixed shell to have admin mode. Now consistant with other shells.

  Revision 1.31  2008/10/19 21:50:21  dr_steveb
  Added point and curve commands

  Revision 1.30  2008/10/13 18:26:39  ghenson
  Update to prim. Only sent the robot into reverse if prim generated the path

  Revision 1.29  2008/09/25 20:49:35  proctor
  Completely rewrote primMisMain,the prim mission package, to interface to
  a Go Motion controller for doing all the real work. Autotools was changed
  to look for Go Motion. The prim mission interface was changed to eliminate
  the pan and tilt commands in favor of joint and world moves.

  Revision 1.28  2008/05/22 18:42:11  ghenson
  added code to perform docking with fixed point docking stations

  Revision 1.27  2008/04/15 17:05:22  dr_steveb
  Sample docking command in prim

  Revision 1.26  2008/01/08 20:18:16  proctor
  Added control point transform that lets the prim mobility controller act
  on the rotational center of the vehicle, which in general won't coincide
  with the nav data frame

  Revision 1.25  2007/12/24 14:37:56  tomrkramer
  Increased time for opening buffers in order to avoid failing.

  Revision 1.24  2007/05/10 13:18:30  dr_steveb
  Separated the prim mobility planning engine from the interface wrappers. This will allow a controller to be created (the traffic controller) that controls a large number of vehicles without the overhead of individual NML channels for each vehicle. No changes in the algorithms were made.

  Revision 1.23  2007/01/25 18:09:47  proctor
  Used local serial numbers, odd only

  Revision 1.22  2007/01/25 17:44:50  dr_steveb
  Changed state to be explicit in the waypoint movement command

  Revision 1.21  2006/12/12 20:15:58  proctor
  Changed bad flag in prim shell to not bomb when wheel parameters aren't set

  Revision 1.20  2006/11/30 17:32:56  proctor
  Tweaked the prim mobility controller for Nautic driving; stern plane gain
  now looks good and it tracks depths nicely

  Revision 1.19  2006/11/07 22:10:01  proctor
  Fixed help print for "pars" to reflect true usage

  Revision 1.18  2006/10/02 13:59:10  proctor
  Changed PrimMobJACmdRotate to take a 'direction' value, either
  PRIM_MOB_JA_CMD_ROTATE_CW,CCW,SHORTEST. Changed primShell accordingly.

  Revision 1.17  2006/09/28 20:59:49  proctor
  Added Nautic steering

  Revision 1.16  2006/09/28 19:56:37  proctor
  Added Z field to waypoints and handled them backward-compatible in primShell.
  Computed pitch correction in doCmdMoveWaypoint, but didn't do anything with
  them yet pending print debugging.

  Revision 1.15  2006/09/20 21:28:07  proctor
  Added PrimMobJACmdStop and related code for sending it from the prim shell
  and handling it in primMobMain.

  Revision 1.14  2006/09/20 12:47:47  dr_steveb
  Added clockwise flag to rotate command

  Revision 1.13  2006/09/14 20:49:24  dr_steveb
  Set unopened buffers to null to avoid crashes

  Revision 1.12  2006/09/01 19:55:57  proctor
  Fixed bug where the get_address() wasn't called after an NML re-init

  Revision 1.11  2006/07/25 13:41:37  dr_steveb
  Added SP commands

  Revision 1.10  2006/07/21 20:06:58  dr_steveb
  Changed buffers to conform to new standard. Removed check on navData before "go" performed.

  Revision 1.9  2006/07/18 14:46:29  proctor
  Added a few lines to primShell status printing

  Revision 1.8  2006/04/13 14:30:09  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.7  2006/03/15 21:54:07  dr_steveb
  Files changed to allow for pretty pictures to be made for ICRA tutorial.

  Revision 1.6  2006/03/08 21:08:11  proctor
  Fixed scan format bug in primShell.cc

  Revision 1.5  2006/03/08 21:04:25  proctor
  Added time to "move" command

  Revision 1.4  2006/02/17 14:35:54  dr_steveb
  Added printing of UTM zone with status.

  Revision 1.3  2006/01/06 16:05:44  proctor
  Added timed loop when checking for validity of NML buffers, and printed
  which NML buffers failed as a helpful diagnostic

  Revision 1.2  2005/12/23 16:26:09  proctor
  Added pan, tilt printing

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.18  2005/09/30 18:55:46  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.17  2005/08/10 21:05:44  proctor
  Changed "arc" to be like "wp" and increment serial number each time

  Revision 1.16  2005/08/10 16:37:28  proctor
  primMobMain.cc: handled vCutoffAngle, wCutoffAngle in iniLoad, waypoint move.
  primShell.cc: handled vCutoffAngle, wCutoffAngle; incremented serial number
  for "wp" command each time.

  Revision 1.15  2005/08/02 20:40:32  proctor
  Handled speed element of Waypoint type

  Revision 1.14  2005/08/02 19:09:00  proctor
  Added "abs", "rel" selection of coords for wp, arc commands

  Revision 1.13  2005/07/28 21:18:58  proctor
  Fixed pars printing to use degrees

  Revision 1.12  2005/07/20 21:53:39  proctor
  primMobMain.cc: added 10-second timeout for reading in servo mob settings.
  primShell.cc: added handling of lines via isArc in loadPath().

  Revision 1.11  2005/07/18 21:16:46  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.10  2005/07/15 15:01:40  proctor
  Added handling of cycle time, debug settings

  Revision 1.9  2005/07/01 16:01:23  proctor
  Took out annoying .nml and .ini config printing

  Revision 1.8  2005/06/27 14:27:20  proctor
  Added arc segment move command

  Revision 1.7  2005/06/08 14:30:55  proctor
  Added VehMobJA name changes

  Revision 1.6  2005/05/20 15:18:45  proctor
  In primMobJAMain.cc, fixed a bug where the command number wasn't set properly
  for new commands; changed INI file angle units to degrees; set the pathIndex
  in status. In primShell.cc, changed angle units to degrees.

  Revision 1.5  2005/05/17 21:30:50  proctor
  Referenced Waypoint structure instead of individual fields that were used
  as an interim workaround for the code gen bug

  Revision 1.4  2005/05/16 21:15:02  proctor
  First time consistently applying suffix numbers to buffers

  Revision 1.3  2005/05/13 20:19:38  proctor
  Added [PRIM_PL] MAX_TRAN_VEL; more shell features

  Revision 1.2  2005/05/13 19:02:55  proctor
  Added waypoint driving

  Revision 1.1  2005/05/12 13:27:36  proctor
  Initial checkin

*/

