/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   rcsEngine.cc
  \brief Provide an RCS engine template

  Contains all of the state table and nml methods

  \code CVS Status:
  $Author: ghenson $
  $Revision: 1.10 $
  $Date: 2009/01/17 16:45:12 $
  \endcode

  \author Stephen Balakirsky
  \date   03/14/2006
*/

#include <unistd.h>		// close(), write()
#include <math.h>
#include <sys/types.h>          // open()
#include <sys/stat.h>           // open()
#include <fcntl.h>              // open()
#include <stdlib.h>		// rand()
#include "rcsSectMobPlEngine.hh"
#include "amSP.hh"              // AMSP_OBSTACLE, etc.
#include "vehMobPL.hh"
#include "vehMobPL.hh"
#include "../vehicle/vehMobCell.hh"
//#include "myDefaults.hh"

/***********************************************************************/
rcsEngine::rcsEngine(double cycleTime, double resolution)
{
  int count;

  timer = new RCS_TIMER( cycleTime );
  nmlStatus = new SectMobPLStat();
  nmlSettings = new SectMobPLSet();

  nmlStatus->heartbeat = 0;
  nmlStatus->admin_state = ADMIN_SHUT_DOWN;
  nmlStatus->command_type = SECT_MOB_PL_CMD_NOP_TYPE;
  nmlStatus->status = RCS_DONE;
  nmlStatus->echo_serial_number = 0;
  reset_time_tracker( &nmlStatus->tt );
  // user status initialization
  nmlStatus->numVehicles = 0;

  nmlSettings->echo_serial_number = 0;
  nmlSettings->command_type = SECT_MOB_PL_CFG_NOP_TYPE;
  nmlSettings->status = RCS_DONE;
  nmlSettings->echo_serial_number = 0;
  nmlSettings->cycleTime = cycleTime;
  strcpy( nmlStatus->message, "" );

  // default cmd, stat, cfg, set buffers
  sectMobPLCmdNml = NULL;
  sectMobPLCfgNml = NULL;
  sectMobPLSetNml = NULL;
  sectMobPLStatNml = NULL;

  // user included buffers and command pointers
  for( count=0; count<MAX_SECTION_VEHICLES; count++ )
    {
      navDataExtNml[count] = NULL;
      sensorDataNml[count] = NULL;
      vehicleStatNml[count] = NULL;
      vehicleCmdNml[count] = NULL;
      vehicleCfgNml[count] = NULL;
      vehicleSetNml[count] = NULL;
      primMobCmdNml[count] = NULL;
      primMobStatNml[count] = NULL;
      primSPCmdNml[count] = NULL;
      primSPStatNml[count] = NULL;
    }
  vehicleSerialNumber = 0;
  nmlSettings->worldResolution = resolution;
  myNmlFileName = NULL;
  mySystemNmlName = NULL;
}

/***********************************************************************/
rcsEngine::~rcsEngine()
{
  if( myNmlFileName != NULL )
    delete [] myNmlFileName;
  if( mySystemNmlName != NULL )
    delete [] mySystemNmlName;
}

/***********************************************************************/
/*! read the map being generated by one of the subordinates
   \param which The vehicle to be updated from
   \return True if settings channel should be written
 */
bool rcsEngine::readVehicleMap(int which)
{
  int type;
  SensorDataList *dataList;
  SensorData2D   *data2D;
  bool retValue = false;
  PM_CARTESIAN obsLoc;

  type = sensorDataNml[which]->read();
  switch(type)
    {
    case 0: // no new data
      return false;
    case SENSOR_DATA_LIST_TYPE:
      dataList = (SensorDataList*)sensorDataNml[which]->get_address();
      break;
    case SENSOR_DATA_2D_TYPE:
      data2D = (SensorData2D*)sensorDataNml[which]->get_address();
      break;
    default:
      fprintf( stderr, "obstacle channel error\n" );
      fprintf( stdout, "Command %s not yet supported\n", 
	       sensorData_symbol_lookup(type) );
      return false;
    }
  if( type == SENSOR_DATA_2D_TYPE )
    {
      //      printf("New obstacles from vehicle %d\n", which );
      if( !data2D->isGlobal ) // set local coordinates to global
	{
	  obsLoc = data2D->lowerLeft;
	  obsLoc = transformPoint(which, obsLoc, false);
	  data2D->lowerLeft = obsLoc;
	}
      if( updateWorld( data2D ) )
	{
	  retValue = true;
	}
    }
  return retValue;
}

/***********************************************************************/
bool rcsEngine::runStateMachine(const char *systemNMLName)
{
  NMLTYPE cmdType, cfgType;
  int cmdSerialNumber, cfgSerialNumber;
  //  char newText[100];
  int count;
  bool writeSettings;

  /* status housekeeping */
  timer->wait(); // wait for cycle start
  nmlStatus->heartbeat++;
  cycle_time_tracker( &nmlStatus->tt );

  // read command
  if( sectMobPLCmdNml != NULL )
    cmdType = sectMobPLCmdNml->read();
  else
    cmdType = 0;
  if( cmdType > 0 )
    cmdSerialNumber = sectMobPLCmdNml->get_address()->serial_number;

  // read configuration
  if( sectMobPLCfgNml != NULL )
    cfgType = sectMobPLCfgNml->read();
  else
    cfgType = 0;
  if( cfgType > 0 )
    cfgSerialNumber = sectMobPLCfgNml->get_address()->serial_number;

  // read data and status buffers
  // read each vehicle's location, sensor data, and status
  for( count=0; count<nmlStatus->numVehicles; count++ )
    {
      if( navDataExtNml[count] != NULL )
	updatePosition(count);
      if( sensorDataNml[count] != NULL )
	writeSettings |= readVehicleMap(count);
      if( vehicleStatNml[count] != NULL )
	checkVehCmdStatus(count);
      if( primMobStatNml[count] != NULL )
	checkPrimMobCmdStatus(count);
      if( primSPStatNml[count] != NULL )
	checkPrimSPCmdStatus(count);
    }

  // perform housekeeping on command and configuration
  switch(cmdType)
    {
    case 0:
      // no new command
      break;
    case -1:
      // comm error
      disconnect(1);
      break;
    case SECT_MOB_PL_CMD_INIT_TYPE:
    case SECT_MOB_PL_CMD_ABORT_TYPE:
    case SECT_MOB_PL_CMD_HALT_TYPE:
    case SECT_MOB_PL_CMD_SHUTDOWN_TYPE:
    case SECT_MOB_PL_CMD_NOP_TYPE:
    case SECT_MOB_PL_CMD_CONNECT_TYPE:
    case SECT_MOB_PL_CMD_DISCONNECT_TYPE:
    case SECT_MOB_PL_CMD_EXPLORE_TYPE:
    case SECT_MOB_PL_CMD_CRASH_TYPE:
      nmlStatus->command_type = cmdType;
      if(cmdSerialNumber != nmlStatus->echo_serial_number) 
	{
	  nmlStatus->echo_serial_number = cmdSerialNumber;
	  nmlStatus->state = NEW_COMMAND;
	  nmlStatus->status = RCS_EXEC;
	}
      // clear out status message
      sprintf(nmlStatus->message, "%s", "All monitored functions OK\n" );
      break;
    default:
      fprintf(stderr, "%s: unknown command %s\n", 
	      systemNMLName, sectMobPL_symbol_lookup(cmdType));
      break;
    } // switch (cmdType)

  switch(cfgType)
    {
    case 0:
      // no new command
      break;
    case -1:
      // comm error
      disconnect(1);
      break;
    case SECT_MOB_PL_CFG_CYCLE_TIME_TYPE:
    case SECT_MOB_PL_CFG_DEBUG_TYPE:
    case SECT_MOB_PL_CFG_NOP_TYPE:
    case SECT_MOB_PL_CFG_DUMP_WM_TYPE:
      nmlSettings->command_type = cfgType;
      if(cfgSerialNumber != nmlSettings->echo_serial_number) 
	{
	  nmlSettings->echo_serial_number = cfgSerialNumber;
	  nmlSettings->state = NEW_COMMAND;
	  nmlSettings->status = RCS_EXEC;
	}
      break;
    default:
      fprintf(stderr, "%s: unknown configuration %s\n", 
	      systemNMLName, sectMobPL_symbol_lookup(cfgType));
      break;
    }

  // perform command
  switch( nmlStatus->command_type )
    {
    case SECT_MOB_PL_CMD_INIT_TYPE:
      doCmdInit();
      break;
    case SECT_MOB_PL_CMD_ABORT_TYPE:
      doCmdAbort();
      break;
    case SECT_MOB_PL_CMD_HALT_TYPE:
      doCmdHalt();
      break;
    case SECT_MOB_PL_CMD_SHUTDOWN_TYPE:
      doCmdShutdown();
      break;
    case SECT_MOB_PL_CMD_NOP_TYPE:
      doCmdNop();
      break;
    case SECT_MOB_PL_CMD_CONNECT_TYPE:
      doCmdConnect();
      break;
    case SECT_MOB_PL_CMD_DISCONNECT_TYPE:
      doCmdDisconnect();
      break;
    case SECT_MOB_PL_CMD_EXPLORE_TYPE:
      doCmdExplore();
      break;
    case SECT_MOB_PL_CMD_CRASH_TYPE:
      doCmdCrash();
      break;
    }

  // perform configuration
  switch( nmlSettings->command_type )
    {
    case SECT_MOB_PL_CFG_CYCLE_TIME_TYPE:
      writeSettings |= doCfgCycleTime();
      break;
    case SECT_MOB_PL_CFG_DEBUG_TYPE:
      writeSettings |= doCfgDebug();
      break;
    case SECT_MOB_PL_CFG_NOP_TYPE:
      writeSettings |= doCfgNop();
      break;
    case SECT_MOB_PL_CFG_DUMP_WM_TYPE:
      writeSettings |= doCfgDumpWM();
      break;
    default:
      break;
    }

  // write out buffers
  if( sectMobPLStatNml != NULL )
    {
      sectMobPLStatNml->write(nmlStatus);
    }
  if( sectMobPLSetNml != NULL && writeSettings )
    {
      sectMobPLSetNml->write(nmlSettings);
      writeSettings = false;
    }
  return true;
}

/***********************************************************************/
RCS_ADMIN_STATE rcsEngine::getAdminState()
{
  return( nmlStatus->admin_state );
}

/***********************************************************************/
RCS_STATUS rcsEngine::getStatus()
{
  return( nmlStatus->status );
}

/***********************************************************************/
bool rcsEngine::connect(const char *nmlFileName,const char *systemNMLName,
			     const char *nmlPostfix )
{
  SectMobPLCmdNop *nopCmd;
  SectMobPLCfgNop *nopCfg;
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  if( myNmlFileName != NULL )
    delete [] myNmlFileName;
  if( mySystemNmlName != NULL )
    delete [] mySystemNmlName;
  
  myNmlFileName = new char[strlen(nmlFileName)+1];
  mySystemNmlName = new char[strlen(systemNMLName)+1];
  strcpy( myNmlFileName, nmlFileName );
  strcpy(mySystemNmlName, systemNMLName);

  sprintf( chanName, "%s%s", SECT_MOB_PL_CMD_NAME, nmlPostfix );
  sectMobPLCmdNml = new RCS_CMD_CHANNEL( sectMobPL_format, chanName,
					  (char*)systemNMLName, 
					  (char*)nmlFileName );
  if( NULL == sectMobPLCmdNml || !sectMobPLCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  sprintf( chanName, "%s%s",SECT_MOB_PL_STAT_NAME, nmlPostfix );
  sectMobPLStatNml = new RCS_STAT_CHANNEL( sectMobPL_format, chanName,
					    (char*)systemNMLName,
					    (char*)nmlFileName );
  if( NULL == sectMobPLStatNml || !sectMobPLStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  sprintf( chanName, "%s%s", SECT_MOB_PL_SET_NAME, nmlPostfix );
  sectMobPLSetNml = new RCS_STAT_CHANNEL( sectMobPL_format, chanName,
					   (char*)systemNMLName, 
					   (char*)nmlFileName );
  if( NULL == sectMobPLSetNml || !sectMobPLSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  sprintf( chanName, "%s%s", SECT_MOB_PL_CFG_NAME, nmlPostfix );
  sectMobPLCfgNml = new RCS_CMD_CHANNEL( sectMobPL_format, chanName,
					  (char*)systemNMLName, 
					  (char*)nmlFileName );
  if( NULL == sectMobPLCfgNml || !sectMobPLCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
		       systemNMLName, chanName );
      disconnect(1);
      return(false);
    }

  nmlStatus->admin_state = ADMIN_UNINITIALIZED;
  // stuff a nop cmd
  nopCmd = new SectMobPLCmdNop();
  nopCfg = new SectMobPLCfgNop();

  if( nopCmd <= 0 || nopCfg <= 0 )
    {
      printf( "Allocation error!!!\n");
      return false;
    }

  nopCmd->serial_number = nmlStatus->echo_serial_number+1;
  nopCfg->serial_number = nmlSettings->echo_serial_number+1;

  sectMobPLCmdNml->write(nopCmd);
  sectMobPLCfgNml->write(nopCfg);
  delete nopCmd;
  delete nopCfg;

  return true;
}

/***********************************************************************/
bool rcsEngine::connectSubordinates(int *nmlPostfix,
				    int numSubordinates,
				    SectSubordinate subordinateClass)
{
  int count;
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  if( subordinateClass == VEHICLE )
    {
      for( count=0; count<numSubordinates; count++ )
	{
	  sprintf( chanName, "%s%d", VEH_MOB_PL_CMD_NAME, nmlPostfix[count] );
	  vehicleCmdNml[count] = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
						      mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleCmdNml[count] || !vehicleCmdNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", VEH_MOB_PL_STAT_NAME, nmlPostfix[count] );
	  vehicleStatNml[count] = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
							mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleStatNml[count] || !vehicleStatNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", VEH_MOB_PL_SET_NAME, nmlPostfix[count] );
	  vehicleSetNml[count] = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
						       mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleSetNml[count] || !vehicleSetNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", VEH_MOB_PL_CFG_NAME, nmlPostfix[count] );
	  vehicleCfgNml[count] = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
						      mySystemNmlName, myNmlFileName );
	  if( NULL == vehicleCfgNml[count] || !vehicleCfgNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }
	  /*
	    sprintf( chanName, "%s%d", AM_SP_CMD_NAME, nmlPostfix[count] );
	    amSPCmdNml[count] = new RCS_CMD_CHANNEL( amSP_format, chanName,
	    mySystemNmlName, myNmlFileName );
	    if( NULL == amSPCmdNml[count] || !amSPCmdNml[count]->valid() )
	    {
	    rcs_print_error( "%s: Error - Can't open %s buffer\n",
	    mySystemNmlName, chanName );
	    return(false);
	    }
	  */

	  //  sprintf( chanName, "%s%d", PRIM_SP_LINESCAN_NAME, nmlPostfix[count] );
	  //	  sprintf( chanName, "%s%d", AM_SP_MAP_NAME, nmlPostfix[count] );
	  sprintf( chanName, "%s%d", VEH_MOB_MAP_NAME, nmlPostfix[count] );
	  sensorDataNml[count] = new NML( sensorData_format, chanName,
					  mySystemNmlName, myNmlFileName );
	  if( NULL == sensorDataNml[count] || !sensorDataNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
			       mySystemNmlName, chanName );
	      return(false);
	    }
	  sprintf( chanName, "%s%d", NAV_DATA_EXT_NAME, nmlPostfix[count] );
	  navDataExtNml[count] = new NML( navDataExt_format, chanName,
					  mySystemNmlName, 
					  myNmlFileName );
	  if( NULL == navDataExtNml[count] || !navDataExtNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
			       mySystemNmlName, chanName );
	      disconnect(1);
	      return false;
	    }
	}
    }
  else if( subordinateClass == PRIM )
    {
      for( count=0; count<numSubordinates; count++ )
	{
	  sprintf( chanName, "%s%d", PRIM_MOB_JA_CMD_NAME, nmlPostfix[count] );
	  primMobCmdNml[count] = new RCS_CMD_CHANNEL( primMobJA_format, chanName,
						      mySystemNmlName, myNmlFileName );
	  if( NULL == primMobCmdNml[count] || !primMobCmdNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", PRIM_MOB_JA_STAT_NAME, nmlPostfix[count] );
	  primMobStatNml[count] = new RCS_STAT_CHANNEL( primMobJA_format, chanName,
							mySystemNmlName, myNmlFileName );
	  if( NULL == primMobStatNml[count] || !primMobStatNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", PRIM_SP_CMD_NAME, nmlPostfix[count] );
	  primSPCmdNml[count] = new RCS_CMD_CHANNEL( primSP_format, chanName,
						      mySystemNmlName, myNmlFileName );
	  if( NULL == primSPCmdNml[count] || !primSPCmdNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }

	  sprintf( chanName, "%s%d", PRIM_SP_STAT_NAME, nmlPostfix[count] );
	  primSPStatNml[count] = new RCS_STAT_CHANNEL( primSP_format, chanName,
							mySystemNmlName, myNmlFileName );
	  if( NULL == primSPStatNml[count] || !primSPStatNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n",
			       mySystemNmlName, chanName );
	      return(false);
	    }
	  sprintf( chanName, "%s%d", NAV_DATA_EXT_NAME, nmlPostfix[count] );
	  navDataExtNml[count] = new NML( navDataExt_format, chanName,
					  mySystemNmlName, 
					  myNmlFileName );
	  if( NULL == navDataExtNml[count] || !navDataExtNml[count]->valid() )
	    {
	      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
			       mySystemNmlName, chanName );
	      disconnect(1);
	      return false;
	    }
	}
    }
  nmlStatus->numVehicles = numSubordinates;
  return true;
}

/***********************************************************************/
void rcsEngine::disconnectSubordinates()
{
  int count;

  // disconnect from buffers
  for( count=0; count<nmlStatus->numVehicles; count++ )
    {
      if( sensorDataNml[count] != NULL )
	{
	  delete sensorDataNml[count];
	  sensorDataNml[count] = NULL;
	}
      if( navDataExtNml[count] != NULL )
	{
	  delete navDataExtNml[count];
	  navDataExtNml[count] = NULL;
	}
      if( vehicleCmdNml[count] != NULL )
	{
	  delete vehicleCmdNml[count];
	  vehicleCmdNml[count] = NULL;
	}
      if( vehicleStatNml[count] != NULL )
	{
	  delete vehicleStatNml[count];
	  vehicleStatNml[count] = NULL;
	}
      if( vehicleCfgNml[count] != NULL )
	{
	  delete vehicleCfgNml[count];
	  vehicleCfgNml[count] = NULL;
	}
      if( vehicleSetNml[count] != NULL )
	{
	  delete vehicleSetNml[count];
	  vehicleSetNml[count] = NULL;
	}
      if( primMobCmdNml[count] != NULL )
	{
	  delete primMobCmdNml[count];
	  primMobCmdNml[count] = NULL;
	}
      if( primMobStatNml[count] != NULL )
	{
	  delete primMobStatNml[count];
	  primMobStatNml[count] = NULL;
	}
      if( primSPCmdNml[count] != NULL )
	{
	  delete primSPCmdNml[count];
	  primSPCmdNml[count] = NULL;
	}
      if( primSPStatNml[count] != NULL )
	{
	  delete primSPStatNml[count];
	  primSPStatNml[count] = NULL;
	}
    }
}

/***********************************************************************/
void rcsEngine::disconnect(int error)
{
  // disconnect from subordinates
  disconnectSubordinates();

  nmlStatus->admin_state = ADMIN_SHUT_DOWN;
  if( error )
    nmlStatus->status = RCS_ERROR;
  else
    nmlStatus->status = RCS_DONE;
  nmlStatus->command_type = SECT_MOB_PL_CMD_NOP_TYPE;
  // send last status message out 
  if( sectMobPLCmdNml != NULL )
    sectMobPLCmdNml->write(nmlStatus);

  if( sectMobPLCmdNml != NULL )
    delete sectMobPLCmdNml;
  sectMobPLCmdNml = NULL;

  if( sectMobPLCfgNml != NULL )
    delete sectMobPLCfgNml;
  sectMobPLCfgNml = NULL;

  if( sectMobPLStatNml != NULL )
    delete sectMobPLStatNml;
  sectMobPLStatNml = NULL;

  if( sectMobPLSetNml != NULL )
    delete sectMobPLSetNml;
  sectMobPLSetNml = NULL;
}

/***********************************************************************/
const char *rcsEngine::getCommandText()
{
  return sectMobPL_symbol_lookup(nmlStatus->command_type);
}


/***********************************************************************/
/*!
  Read in new vehicle position

  \param statIn Channel to read from

  \return true on success
  \return false on failure
*/
bool rcsEngine::updatePosition(int which)
{
  int type;
  NavDataExt *navData;

  type = navDataExtNml[which]->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case NAV_DATA_EXT_TYPE:
      navData = (NavDataExt*)navDataExtNml[which]->get_address();
      vehicleNav[which] = *navData;
      homLocal[which].tran = navData->tranRel;
      homLocal[which].rot = navData->rpyRel;
      homGlobal[which].tran = navData->tranAbs;
      homGlobal[which].rot = navData->rpyAbs;
      return(true);
    default:
      return(false);
    }
  return(false);
}


/***********************************************************************/
/*!
  Perform connect state machine
*/
void rcsEngine::doCmdConnect()
{

}

/***********************************************************************/
/*!
  Perform disconnect state machine
*/
void rcsEngine::doCmdDisconnect()
{

}

/***********************************************************************/
/*!
  Perform init state machine

  \return true finished with initialization
  \return false still working
*/
void rcsEngine::doCmdInit()
{
  VehMobPLCmdInit  *initCmd;
  SectMobPLCmdInit *sectInitCmd;
  PrimSPCmdGo      *primSPCmdGo;
  PrimSPCmdInit    *primSPCmdInit;
  PrimMobJACmdInit *primMobCmdInit;

  int count;


  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state == ADMIN_SHUT_DOWN )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
      // allocate world model
      // current default size is 20x20 meters
      nmlSettings->worldHeight = 20;
      nmlSettings->worldWidth = 20;
      dataPt = new char[nmlSettings->worldHeight*nmlSettings->worldWidth*3+500];
      clearMap();
    }
  if( state_match(nmlStatus, S0) ) // connect to subordinates
    {
      sectInitCmd = (SectMobPLCmdInit*)sectMobPLCmdNml->get_address();
      nmlStatus->subordinate = sectInitCmd->subordinate;
      connectSubordinates(sectInitCmd->vehList, sectInitCmd->vehList_length, 
			  sectInitCmd->subordinate);
      if( sectInitCmd->subordinate == VEHICLE )
	state_next(nmlStatus, S1);
      else if( sectInitCmd->subordinate == PRIM )
	state_next(nmlStatus, S3);
      else
	{
	  status_next(nmlStatus, RCS_ERROR);
	  state_next(nmlStatus,S10 );
	}
    }
  if( state_match(nmlStatus, S1 ))  // send vehicle init commands if necessary
    {
      initCmd = new VehMobPLCmdInit();
      initCmd->serial_number = ++vehicleSerialNumber;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	vehicleCmdNml[count]->write(initCmd);
      delete initCmd;
      state_next(nmlStatus, S2);
      return;
    }
  if( state_match(nmlStatus, S2) ) //wait for vehicle done
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  /*
	  printf( "In doCmdInit vehicle %d with serial nubmers: %d %d",
		  count,
		  vehicleCmdEchoSerialNumber[count], vehicleSerialNumber);
	  printf( " with status %d\n", vehicleCmdStatus[count] );
	  */
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S10);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S9);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S3 )) // send prim sp init commands if necessary
    {
      primSPCmdInit = new PrimSPCmdInit();
      primSPCmdInit->serial_number = ++vehicleSerialNumber;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	primSPCmdNml[count]->write(primSPCmdInit);
      delete primSPCmdInit;
      //      printf( "Done with state S3\n" );
      state_next(nmlStatus, S4);
      return;
    }
  if( state_match(nmlStatus, S4 )) // wait for prim sp to finish
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( primSPCmdEchoSerialNumber[count] != vehicleSerialNumber )
	    return;
	  if( primSPCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S10);
	    }
	  else
	    if( primSPCmdStatus[count] == RCS_EXEC)
	      return;
	}
      //      printf( "Done with state S4\n" );
      state_next(nmlStatus, S5);
    }
  if( state_match(nmlStatus, S5 )) // send prim sp go commands; this command never sends done
    {
      primSPCmdGo = new PrimSPCmdGo();
      primSPCmdGo->serial_number = ++vehicleSerialNumber;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	primSPCmdNml[count]->write(primSPCmdGo);
      delete primSPCmdGo;
      //      printf( "Done with state S5\n" );
      state_next(nmlStatus, S7);
      return;
    }
  /*
  if( state_match(nmlStatus, S6 )) // wait for prim sp to finish
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( primSPCmdEchoSerialNumber[count] != vehicleSerialNumber )
	    return;
	  if( primSPCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S10);
	    }
	  else
	    if( primSPCmdStatus[count] == RCS_EXEC)
	      return;
	}
      printf( "Done with state S6\n" );
      state_next(nmlStatus, S7);
    }
  */
  if( state_match(nmlStatus, S7 )) // send prim mob init commands if necessary
    {
      primMobCmdInit = new PrimMobJACmdInit();
      primMobCmdInit->serial_number = ++vehicleSerialNumber;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	primMobCmdNml[count]->write(primMobCmdInit);
      delete primMobCmdInit;
      //      printf( "Done with state S7\n" );
      state_next(nmlStatus, S8);
      return;
    }
  if( state_match(nmlStatus, S8 )) // wait for prim mob to finish
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( primMobCmdEchoSerialNumber[count] != vehicleSerialNumber )
	    return;
	  if( primMobCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S10);
	    }
	  else
	    if( primMobCmdStatus[count] == RCS_EXEC)
	      return;
	}
      //      printf( "Done with state S8\n" );
      state_next(nmlStatus, S9);
    }
  if( state_match(nmlStatus, S9) )
    {
      nmlStatus->admin_state = ADMIN_INITIALIZED;
      state_next(nmlStatus, S10);
    }
}

/***********************************************************************/
/*!
  Perform abort command state machine

  \param stat status message
  \return true need to update and send set message
  \return false no message needs to be sent
*/
void rcsEngine::doCmdAbort()
{
  VehMobPLCmdAbort *abortCmd;
  int count;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
    }
  if( state_match(nmlStatus, S0) )
    {
      abortCmd = new VehMobPLCmdAbort();
      abortCmd->serial_number = ++vehicleSerialNumber;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	vehicleCmdNml[count]->write(abortCmd);
      delete abortCmd;
      state_next(nmlStatus, S1);
    }
  else if( state_match(nmlStatus, S1) )
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S2);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S2);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S2) )
    {
      nmlStatus->admin_state = ADMIN_UNINITIALIZED;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
/*!
  Perform  halt command state machine

  \param stat status message
  \return true done with command
  \return false still working
*/
void rcsEngine::doCmdHalt()
{
  VehMobPLCmdHalt *haltCmd;
  int count;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
    }
  if( state_match(nmlStatus, S0) )
    {
      haltCmd = new VehMobPLCmdHalt();
      haltCmd->serial_number = ++vehicleSerialNumber;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	vehicleCmdNml[count]->write(haltCmd);
      delete haltCmd;
      state_next(nmlStatus, S1);
    }
  else if( state_match(nmlStatus, S1) )
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S2);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S2);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S2) )
    {
      nmlStatus->admin_state = ADMIN_UNINITIALIZED;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
/*!
  Perform shutdown command to subordinate planning system

  \param stat vehicle status
  \return true done with shutdown
  \return false not done yet
*/
void rcsEngine::doCmdShutdown()
{
  VehMobPLCmdShutdown *shutdownCmd;
  int count;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S0);
      if( nmlStatus->admin_state == ADMIN_INITIALIZED )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
    }
  if( state_match(nmlStatus, S0) )
    {
      shutdownCmd = new VehMobPLCmdShutdown();
      shutdownCmd->serial_number = ++vehicleSerialNumber;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	vehicleCmdNml[count]->write(shutdownCmd);
      delete shutdownCmd;
      state_next(nmlStatus, S1);
    }
  else if( state_match(nmlStatus, S1) )
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleCmdEchoSerialNumber[count] != vehicleSerialNumber )
	    return;
	  if( vehicleCmdStatus[count] == RCS_ERROR )
	    {
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S2);
	    }
	  else
	    if( vehicleCmdStatus[count] == RCS_EXEC)
	      return;
	}
      state_next(nmlStatus, S2);
      status_next(nmlStatus, RCS_DONE);
    }
  if( state_match(nmlStatus, S2) )
    {
      disconnect(0);
      nmlStatus->admin_state = ADMIN_SHUT_DOWN;
      state_next(nmlStatus, S3);
    }
}

/***********************************************************************/
bool rcsEngine:: computeDestination(int whichVeh, double exploreDist) 
{
  double temp;
  double dblNorth, dblEast;
  int intNorth, intEast;
  PM_CARTESIAN growLoc;

  temp = (float)rand()/(float)RAND_MAX * 15.;
  if( (float)rand()/(float)RAND_MAX > .5 )
    temp *= -1;
  dblNorth = vehicleNav[whichVeh].tranAbs.x + temp;
  temp = (float)rand()/(float)RAND_MAX * 15.;
  if( (float)rand()/(float)RAND_MAX > .5 )
    temp *= -1;
  dblEast = vehicleNav[whichVeh].tranAbs.y + temp;
  intNorth = mapXToWmNorth(dblNorth); 
  intEast = mapYToWmEast(dblEast); 
  // if necessary, expand world to accomodate new goal
  if( intNorth<0 || intNorth>=nmlSettings->worldHeight || 
      intEast<0 || intEast>=nmlSettings->worldWidth )
    {
      growLoc.x = dblNorth;
      growLoc.y = dblEast;
      growLoc.z = 0;
      if( !growWorld( growLoc ) )
	return(false);
      intNorth = mapXToWmNorth(dblNorth);
      intEast = mapYToWmEast(dblEast);
    }
  fprintf( stdout, "\nnew points computed %lf %lf\n",
	   dblNorth, dblEast);
  fflush(stdout);
  if( worldModel[intNorth][intEast] == 0 )
    {
      vehicleDestination[whichVeh].x = dblNorth;
      vehicleDestination[whichVeh].y = dblEast;
      return true;
    }
  return false;
}

/***********************************************************************/
void rcsEngine::doCmdNop()
{
  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      state_next(nmlStatus, S1);
      status_next(nmlStatus, RCS_DONE);
    }
}

/***********************************************************************/
void rcsEngine::doCmdCrash()
{
  char fileName[256], fileBase[256];
  int count;
  float distance;
  static PrimMobJACmdMoveWaypoint primMobJACmdMoveWaypoint;
  SectMobPLCmdCrash *sectCrashCmd =
    (SectMobPLCmdCrash*)sectMobPLCmdNml->get_address();

  if( state_match(nmlStatus, NEW_COMMAND) ) // check for init state and vehicles
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      // check if we are allowed to execute this command
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  sprintf(nmlStatus->message, "%s", "System not initialized\n" );
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
      if( nmlStatus->numVehicles != 2 )
	{
	  sprintf(nmlStatus->message, "%d%s", nmlStatus->numVehicles, "Connected, need two\n" );
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
      if( nmlStatus->subordinate != PRIM )
	{
	  sprintf(nmlStatus->message, "%d%s", nmlStatus->subordinate, "Connected, need prim echelon\n" );
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
      sprintf( fileBase, "%s", "ardaPrim" );
      state_next(nmlStatus, S1);
    }
  if( state_match(nmlStatus, S1) ) // send vehicle commands out
    {
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  sprintf( fileName, "%s%d.dat", fileBase, count+1 );
	  primMobJACmdMoveWaypoint.waypoint_length = 0;
	  if(0 !=
	     loadWaypoints(fileName, count,
			   &primMobJACmdMoveWaypoint.waypoint[0],
			   &primMobJACmdMoveWaypoint.waypoint_length)) 
	    {
	      sprintf(nmlStatus->message, "can't load waypoint file ``%s''\n", fileName );
	      status_next(nmlStatus, RCS_ERROR);
	      state_next(nmlStatus, S10);
	    } 
	  else 
	    {
	      // send out command
	      primMobJACmdMoveWaypoint.serial_number =
		++vehicleSerialNumber;
	      primMobCmdNml[count]->write(&primMobJACmdMoveWaypoint);
	      printf( "sending points from %s to vehicle %d\n", fileName, count );
	    }
	}
      // set up command for slow driving
      primMobJACmdMoveWaypoint.serial_number =
	++vehicleSerialNumber;
      primMobJACmdMoveWaypoint.waypoint[0] = 
	primMobJACmdMoveWaypoint.waypoint[primMobJACmdMoveWaypoint.waypoint_length-1];
      primMobJACmdMoveWaypoint.waypoint_length = 1;
      primMobJACmdMoveWaypoint.waypoint[0].speed = 1;
      state_next(nmlStatus, S2);
    }
  if( state_match(nmlStatus, S2) ) // watch closing distance and wait for done
    {
      distance = disp( vehicleNav[0].tranAbs, vehicleNav[1].tranAbs );
      if( distance < sectCrashCmd->warnDist )
	{
	  primMobCmdNml[1]->write(&primMobJACmdMoveWaypoint);
	  status_next(nmlStatus, RCS_DONE);
	  state_next(nmlStatus,S3);
	}
      printf( "Closing distance: %f            \r", distance );
      fflush(stdout);
    }
}

/***********************************************************************/
void rcsEngine::doCmdExplore()
{
  static int waitCounter;
  int count;
  bool fail;
  VehMobPLCmdMoveWaypoint vehicleMoveCmd;
  SectMobPLCmdExplore *sectExploreCmd = 
    (SectMobPLCmdExplore*)sectMobPLCmdNml->get_address();

  if( state_match(nmlStatus, NEW_COMMAND) ) // init
    {
      state_new(nmlStatus);
      state_default(nmlStatus);
      // check if we are allowed to execute this command
      if( nmlStatus->admin_state != ADMIN_INITIALIZED )
	{
	  status_next(nmlStatus, RCS_ERROR);
	  return;
	}
      state_next(nmlStatus, S1);
    }
  if( state_match(nmlStatus, S1) ) // reset all flags
    {
      // set wait counter to 0
      waitCounter = 0;
      // zero success flags
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  // reset success flag
	  vehicleSuccess[count] = false;
	}
      state_next(nmlStatus, S2);
    } // this state did not take long, so fall through
  if( state_match(nmlStatus, S2) ) // compute destination
    {
      fail = false;
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  if( vehicleSuccess[count] == false )
	    {
	      // computeDestination puts location in vehicleDestination[]
	      if( computeDestination(count, sectExploreCmd->exploreDist) )
		vehicleSuccess[count] = true;
	      else
		fail = true;
	    }
	}
      if( !fail )
	state_next(nmlStatus, S3); // set all destinations
    } // this still did not take very long so fall through
  if( state_match(nmlStatus, S3) )
    {
      // set up movement command
      vehicleMoveCmd.serial_number = ++vehicleSerialNumber;
      vehicleMoveCmd.neighborhood = 1;
      vehicleMoveCmd.isGlobal = 1;
      // send out command
      for( count=0; count<nmlStatus->numVehicles; count++ )
	{
	  vehicleMoveCmd.p = vehicleDestination[count];
	  vehicleCmdNml[count]->write(vehicleMoveCmd);
	}
      state_next(nmlStatus, S4);
    } // that's enough time, so return until next time
  else if( state_match(nmlStatus, S4) )
    {
      if( ++waitCounter > 30 )
	{
	  state_next(nmlStatus, S1);
	}
    }
}

/***********************************************************************/
/*!
  Perform clear wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgNop()
{
  bool retValue = false;

  if( state_match(nmlSettings, NEW_COMMAND) )
    {
      state_new(nmlSettings);
      status_next(nmlSettings, RCS_DONE);
      state_next(nmlSettings, S0);
      retValue = true;
    } // fall through to checking the status
  return retValue;
}

/***********************************************************************/
/*!
  Perform cfgcycletime  state machine
*/
bool rcsEngine::doCfgCycleTime()
{
  SectMobPLCfgCycleTime *sectCfgCycleTime;

  sectCfgCycleTime = (SectMobPLCfgCycleTime*)sectMobPLCfgNml->get_address();
  delete timer;
  timer = new RCS_TIMER( sectCfgCycleTime->cycleTime );
  return false;
}

/***********************************************************************/
/*!
  Perform cfgDebug  state machine
*/
bool rcsEngine::doCfgDebug()
{
  return false;
}

/***********************************************************************/
/*!
  Perform clear wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgClearWM()
{
  bool retValue = false;

  if( state_match(nmlSettings, NEW_COMMAND) )
    {
      state_new(nmlSettings);
      status_next(nmlSettings, RCS_EXEC);
      state_next(nmlSettings, S0);
      retValue = true;
    } // fall through to checking the status

  if( state_match(nmlSettings, S0) )
    {
      clearMap();
      status_next(nmlSettings, RCS_DONE);
      state_next(nmlSettings, S1);
      retValue = true;
    }
  return retValue;
}


/***********************************************************************/
/*!
  Perform load wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgLoadWM()
{
  bool retValue = false;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      status_next(nmlStatus, RCS_EXEC);
      state_next(nmlStatus, S0);
      retValue = true;
    } // fall through to loading the file
      
  if( state_match(nmlStatus, S1) )
    {
      loadWM( nmlSettings->wmFileName );
      status_next(nmlStatus, RCS_DONE);
      state_next(nmlStatus, S1);
      retValue = true;
    }
  return retValue;
}

/***********************************************************************/
/*!
  Perform store wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgStoreWM()
{
  bool retValue = false;

  if( state_match(nmlStatus, NEW_COMMAND) )
    {
      state_new(nmlStatus);
      status_next(nmlStatus, RCS_EXEC);
      state_next(nmlStatus, S0);
      retValue = true;
    } // fall through to storing the file

  if( state_match(nmlStatus, S0) )
    {
      storeWM( nmlSettings->wmFileName );
      status_next(nmlStatus, RCS_DONE);
      state_next(nmlStatus, S1);
      retValue = true;
    }
  return retValue;
}


/***********************************************************************/
/*!
  Perform dump wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool rcsEngine::doCfgDumpWM()
{
  bool retValue = false;
  static int fileID = 0;
  static int skip = 0;
  char sysStr[200];
  SectMobPLCfgDumpWM *dumpCfg;

  if( state_match(nmlSettings, NEW_COMMAND) )
    {
      dumpCfg = (SectMobPLCfgDumpWM*)sectMobPLCfgNml->get_address();
      if( dumpCfg->dumpContinuous )
	nmlSettings->fileSkip = dumpCfg->skip;
      else
	nmlSettings->fileSkip = -1;
      strcpy(nmlSettings->wmFileName, dumpCfg->fileName);
      state_new(nmlSettings);
      sprintf( sysStr, "rm -rf %s", nmlSettings->wmFileName );
      if( system(sysStr) == -1)
	{
	  status_next(nmlSettings, RCS_ERROR);
	  state_next(nmlSettings, S4);
	}
      else
	{
	  sprintf(sysStr,"./imageview -f%s&\n", 
		  nmlSettings->wmFileName);
	  if( system(sysStr) == -1 )
	    {
	      status_next(nmlSettings, RCS_ERROR);
	      state_next(nmlSettings, S4);
	    }
	  else
	    {
	      skip = nmlSettings->fileSkip;
	      status_next(nmlSettings, RCS_EXEC);
	      state_next(nmlSettings, S0);
	    }
	}
	  retValue = true;
    } // if no error, fall through to seeing if file should be written

  if( state_match(nmlSettings, S0) )
    {
      if( skip++ >= nmlSettings->fileSkip )
	{
	  status_next(nmlSettings, RCS_EXEC);
	  state_next(nmlSettings, S1);
	}
      // no set settings have changed, so do not write
    } // possibly fall through to writing file
  if( state_match(nmlSettings, S1) )
    {
      fileID = open(nmlSettings->wmFileName, O_CREAT|O_WRONLY,
		    S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
      if( fileID <=0 )
	{
	  fprintf( stderr, 
		   "SectMobPLEngine: Error %d from opening output file %s\n",
		   fileID, nmlSettings->wmFileName );
	  status_next(nmlSettings, RCS_ERROR);
	  state_next(nmlSettings, S4);
	  retValue = true;
	}
      else
	{
	  dump( fileID );
	  close( fileID );
	}
      if( nmlSettings->fileSkip == -1 ) // done
	{
	  status_next(nmlSettings, RCS_DONE);
	  state_next(nmlSettings, S4);
	  retValue = true;
	}
      else
	{
	  skip = 0;
	  status_next(nmlSettings, RCS_EXEC);
	  state_next(nmlSettings, S0);
	}
    }
  return retValue;
}

/***********************************************************************/
/*!
  Transform a point from/to global coordinates to/from local coordinates

\param pointIn The point to be converted
\param transformGlobalToLocal True- transform to local, false- transform to global
\return Transformed point
*/
PM_CARTESIAN rcsEngine::transformPoint(int which, PM_CARTESIAN &pointIn, 
				       bool transformGlobalToLocal)
{
  PM_HOMOGENEOUS homTransform;
  PM_HOMOGENEOUS homIn, homOut;
  PM_CARTESIAN pointReturn;

  homIn.tran.x = pointIn.x;
  homIn.tran.y = pointIn.y;
  homIn.tran.z = pointIn.z;
  homIn.rot = PM_RPY(0, 0, 0);

  if( transformGlobalToLocal )
    {
      homTransform = homLocal[which] * inv(homGlobal[which]);
    }
  else
    {
      homTransform = homGlobal[which] * inv(homLocal[which]);
    }
  homOut = homTransform * homIn;
  pointReturn.x = homOut.tran.x;
  pointReturn.y = homOut.tran.y;
  pointReturn.z = homOut.tran.z;

  return(pointReturn);
}


/***********************************************************************/

//! worldModelToMapX

/*!
The addition of straightDist puts the location in the middle of
the cell.
\param index The cell to convert
\return The X coordinate of the middle of the cell on which the world model 
        whose index is "index" is built.
*/

double rcsEngine::worldModelToMapX(int index)
{
  return (((index / nmlSettings->worldWidth) * nmlSettings->worldResolution) + worldOffset.x + 
	  straightDist);
}

/***********************************************************************/

//! worldModelToMapY
/*!
The addition of straightDist puts the location in the middle of
the cell.

\param index The cell to convert
\return This returns the Y coordinate of the middle of the cell on which 
        the world model whose index is "index" is built.
*/

double rcsEngine::worldModelToMapY(int index)
{
  return (((index % nmlSettings->worldWidth) * nmlSettings->worldResolution) + worldOffset.y + straightDist);
}

/***********************************************************************/

//! mapToWorldModel

/*!
This returns the index in the world model of the node built on the cell in
which the "site" lies. If the site lies outside the area covered by
the world model, this returns -1.

The "floor" is used to force the round-off toward the more negative,
so that, for example, -0.01 becomes -1.0 and not 0.0.
Without this, small negative values of x and y would be in bounds,
which they should not be. 

\param site The location to convert x-north, y-east
\return the world location
*/

int rcsEngine::mapToWorldModel(PM_CARTESIAN site)
{
  int north;
  int east;
  
  north = ((int)floor((site.x - worldOffset.x) / (double)nmlSettings->worldResolution));
  east  = ((int)floor((site.y - worldOffset.y) / (double)nmlSettings->worldResolution));
  return ((north >= nmlSettings->worldHeight) ? -1 :
          (north < 0)            ? -1 :
          (east >= nmlSettings->worldWidth)   ? -1 :
          (east < 0)             ? -1 :
          ((north * nmlSettings->worldWidth) + east));
}

/***********************************************************************/

//! mapXToWmNorth

/*!
This returns the world model north index of any point whose map X
coordinate is mapX.

This is NOT checking that the returned value is in bounds.
\param mapX The location to convert
\return The world model north index
*/

int rcsEngine::mapXToWmNorth(double mapX)
{
  return ((int)floor((mapX - worldOffset.x) / (double)nmlSettings->worldResolution));
}

/***********************************************************************/
//! mapYToWmEast
/*!
The world model array is best thought of as a grid lying on top
of the offset local map. The first index is north and the second
is east.

This returns the world model east index of any point whose map Y
coordinate is mapY.

This is NOT checking that the returned value is in bounds.
\param mapX The location to convert
\return The world model east index

*/

int rcsEngine::mapYToWmEast(double mapY)
{
  return ((int)floor((mapY - worldOffset.y) / (double)nmlSettings->worldResolution));
}

/***********************************************************************/
//! Routine to grow world model
/*!
  This routine will create a new larger world model and then copy the old one into it.
  \param includeLoc Location to include in new world model
  \return true on success and false on failure
*/
bool rcsEngine::growWorld( PM_CARTESIAN &includeLoc )
{
  PM_CARTESIAN newOffset;
  int newWorldWidth, newWorldHeight;
  double fudge = nmlSettings->worldResolution; // account for double rounding errors 
                                  // in comparisions

  // determine the new lower left corner of the world 
  // (need to offset it if it is the includeLoc)
  if( includeLoc.x <= worldOffset.x + fudge  ) // below world
    {
      newOffset.x = (includeLoc.x - ((nmlSettings->worldHeight / 10) * nmlSettings->worldResolution));
      // 1 added for rounding errors.
      newWorldHeight = nmlSettings->worldHeight + 
	(int)floor((worldOffset.x-newOffset.x)/nmlSettings->worldResolution) + 1;
    }
  else if( includeLoc.x >= 
	   worldOffset.x + nmlSettings->worldHeight*nmlSettings->worldResolution - fudge ) // above world
    {
      newOffset.x = worldOffset.x;
      newWorldHeight = (int)(nmlSettings->worldHeight*1.1) + 
	(int)floor((includeLoc.x - (worldOffset.x + nmlSettings->worldHeight*nmlSettings->worldResolution))/nmlSettings->worldResolution);
    }
  else // everything OK
    {
      newOffset.x = worldOffset.x;
      newWorldHeight = nmlSettings->worldHeight;
    }

  if( includeLoc.y <= worldOffset.y + fudge ) // left of world
    {
      newOffset.y = (includeLoc.y - ((nmlSettings->worldWidth/2) * nmlSettings->worldResolution));
      // 1 added for rounding errors
      newWorldWidth = nmlSettings->worldWidth + (int)floor((worldOffset.y-newOffset.y)/nmlSettings->worldResolution) + 1;
    }
  else if( includeLoc.y >= worldOffset.y + nmlSettings->worldWidth*nmlSettings->worldResolution - fudge ) // right of world
    {
      newOffset.y = worldOffset.y;
      newWorldWidth = (int)(nmlSettings->worldWidth*1.1) + 
	(int)floor((includeLoc.y - (worldOffset.y + nmlSettings->worldWidth*nmlSettings->worldResolution))/nmlSettings->worldResolution);
    }
  else // everything OK
    {
      newOffset.y = worldOffset.y;
      newWorldWidth = nmlSettings->worldWidth;
    }
  if( newWorldWidth*nmlSettings->worldResolution > 1000 || newWorldHeight > 1000 ) // map wants to grow larger then 1000 meters!
    {
      fprintf( stderr, "Map wants to grow too large! <%d %d>\n", (int) (newWorldWidth*nmlSettings->worldResolution),
	       (int) (newWorldHeight*nmlSettings->worldResolution) );
      return(false);
    }
  return( growWorld( newWorldWidth, newWorldHeight, newOffset ) );
}


/***********************************************************************/
//! Routine to grow world model
/*!
  This routine will create a new larger world model and then copy the old one into it.
  \param newWorldWidth new width of world model
  \param newWorldHeight new height of world model
  \param newOffset The new offset of (0,0) of the world model
  \return true on success and false on failure
*/
bool rcsEngine::growWorld( int newWorldWidth, int newWorldHeight, 
			   PM_CARTESIAN &newOffset )
{
  int count, count2;
  char **newWorldModel;
  int widthDiff, heightDiff;
  int oldWidth, oldHeight;

  //**************
  // check new bounds
  // first, new offset must be positioned to allow new map to contain old map
  if( worldOffset.x < newOffset.x || worldOffset.y < newOffset.y )
    {
      fprintf( stderr, "SectMobPL::growWorld error on new offset position\n" );
      return( false );
    }
  // next, new world must be large enough to contain old world plus offset
  if( ((int)floor((worldOffset.x - newOffset.x)/nmlSettings->worldResolution + nmlSettings->worldHeight)) > newWorldHeight ||
      ((int)floor((worldOffset.y - newOffset.y)/nmlSettings->worldResolution + nmlSettings->worldWidth)) > newWorldWidth )
    {
      fprintf( stderr, "SectMobPL:: growWorld error on new map size:\n" \
	       " New h,w: (%d %d) needed: (%d %d)\n)",
	       newWorldHeight, newWorldWidth,
	       ((int)floor((worldOffset.x - newOffset.x)/nmlSettings->worldResolution + nmlSettings->worldHeight)),
	       ((int)floor((worldOffset.y - newOffset.y)/nmlSettings->worldResolution + nmlSettings->worldWidth)));
      return( false );
    }
  
  //*********************
  // allocate new world model
  newWorldModel = new char*[newWorldHeight];
  for( count=0; count<newWorldHeight; count++)
    {
      newWorldModel[count] = new char[newWorldWidth];
      if( newWorldModel[count]==NULL )
	return(false);
      for( count2=0; count2<newWorldWidth; count2++ )
	{
	  newWorldModel[count][count2] = 0;
	}
    }

  //******************
  // copy in old world model and delete space
  widthDiff = (int)floor((worldOffset.y - newOffset.y) / nmlSettings->worldResolution);
  heightDiff = (int)floor((worldOffset.x - newOffset.x) / nmlSettings->worldResolution);
  for( count=0; count<nmlSettings->worldHeight; count++)
    {
      for( count2=0; count2<nmlSettings->worldWidth; count2++ )
	{
	  newWorldModel[count+heightDiff][count2+widthDiff] =
	    worldModel[count][count2];
	}
      delete [] worldModel[count];
    }
  delete [] worldModel;

  //******************
  // reset offsets and widths
  delete [] dataPt;

  oldWidth = nmlSettings->worldWidth;
  oldHeight = nmlSettings->worldHeight;
  nmlSettings->worldWidth = newWorldWidth;
  nmlSettings->worldHeight = newWorldHeight;
  worldOffset = newOffset;
  worldModel = newWorldModel;
  dataPt = new char[nmlSettings->worldHeight*nmlSettings->worldWidth*3+500];
  return( true );
}

/***********************************************************************/
/*! Update the world model the the given map updates

  \param updates The updates to be applied
  \param length The number of updates to apply
  \return true if the world grows, false otherwise
*/
bool rcsEngine::updateWorld( SensorPoint *updates, int length )
{
  int count;
  int northLoc, eastLoc;
  PM_CARTESIAN growLoc;
  bool retValue = false;

  for( count=0; count<length; count++ )
    {
      if( myVerbose )
	fprintf( stdout, "<%f %f %f> %f\n", updates[count].dataLoc.x, updates[count].dataLoc.y,
		 updates[count].dataLoc.z,  updates[count].param1 );
      // decode cell with regard to location
      northLoc = mapXToWmNorth(updates[count].dataLoc.x); 
      eastLoc = mapYToWmEast(updates[count].dataLoc.y); 

      // if necessary, expand world to accomodate new obstacle
      if( northLoc<0 || northLoc>=nmlSettings->worldHeight || 
	  eastLoc<0 || eastLoc>=nmlSettings->worldWidth )
	{
	  growLoc.x = updates[count].dataLoc.x;
	  growLoc.y = updates[count].dataLoc.y;
	  growLoc.z = 0;
	  if( !growWorld( growLoc ) )
	    continue;
	  northLoc = mapXToWmNorth(updates[count].dataLoc.x); 
	  eastLoc = mapYToWmEast(updates[count].dataLoc.y); 
	  retValue = true;
	}
      /* reading from am
      if( updates[count].param1 == AMSP_OBSTACLE || 
	  updates[count].param1 == AMSP_MAYBE_OBSTACLE )
	worldModel[northLoc][eastLoc] = 2;
      else if( updates[count].param1 != AMSP_UNSEEN &&
	       worldModel[northLoc][eastLoc] != 2 )
	worldModel[northLoc][eastLoc] = 1;
      */
      // reading from vehicle
      if( updates[count].param1 > VEH_WM_EXTREAM_MAX )
	worldModel[northLoc][eastLoc] = 2;
      else if( updates[count].param1 != VEH_WM_UNKNOWN &&
	       worldModel[northLoc][eastLoc] != 2 )
	worldModel[northLoc][eastLoc] = 1;
      /* never want already seen area to become unseen
      else
	worldModel[northLoc][eastLoc] = 0;
      */
    }
  return(retValue);
}

/***********************************************************************/
/*! Update the world model the the given map updates

  This routine calls the point version for each element of the map.
  This could be made more efficient.

  \param updates The updates to be applied
  \return true if the world grows, false otherwise
*/
bool rcsEngine::updateWorld( SensorData2D *updates )
{
  SensorPoint updatePoint;
  int xCount, yCount;
  PM_CARTESIAN location;
  int featLoc;
  bool retValue = 0;

  location.x = updates->lowerLeft.x;
  for( xCount = 0; xCount < updates->x; xCount++ )
    {
      location.y = updates->lowerLeft.y;
      for( yCount = 0; yCount < updates->y; yCount++ )
	{
	  updatePoint.dataLoc = location;
	  featLoc = xCount*updates->y + yCount;
	  updatePoint.param1 = updates->featArray[featLoc];
	  retValue |= updateWorld( &updatePoint, 1 );
	  location.y += updates->incrementY;
	}
      location.x += updates->incrementX;
    }
  return(retValue);
}

/***********************************************************************/
//! Store the world model to a file
/*!
  This is not really a planner function, but it is implemented here for
  convience.

\param fileName The file to store to
\return True on success
\return False on failure
*/
bool rcsEngine::storeWM( char *fileName )
{
  FILE *fid;
  int count, count2;
  int nodeID;

  fid = fopen( fileName, "w" );
  if( fid == NULL )
    return false;

  for( count=0; count<nmlSettings->worldHeight; count++)
    {
      for( count2=0; count2<nmlSettings->worldWidth; count2++ )
	{
	  if( worldModel[count][count2] != 0 )
	    {
	      nodeID = count*nmlSettings->worldWidth+count2;
	      fprintf( fid, "%f %f\n", worldModelToMapX(nodeID),
		       worldModelToMapY(nodeID) );
	    }
	}
    }
  fclose( fid );
  return true;
}

/***********************************************************************/
//! Load the world model to a file
/*!
  This is not really a planner function, but it is implemented here for
  convience.

\param fileName The file to load from
\return True on success
\return False on failure
*/
bool rcsEngine::loadWM( char *fileName )
{
  SensorPoint *updates;
  int length;
  FILE *fid;

  fid = fopen( fileName, "r" );
  if( fid == NULL )
    return false;

  updates = new SensorPoint [100];
  length = 0;
  while( fscanf( fid, "%lf %lf", &updates[length].dataLoc.x,
		 &updates[length].dataLoc.y ) == 2 )
    {
      length++;
      if( length == 100 )
	{
	  updateWorld( updates, length );
	  length = 0;
	}
    }
  if( length > 0 )
    updateWorld( updates, length );
  length = 0;
  delete [] updates;
  return true;
}

/***********************************************************************/
//!  Dump the contents of the world model to a file
/*!
  The file may be opened once and left open to permit repeated output 
  (for example to a pipe or history file). If a file is open, then changing
  the filename will not affect the location of the output.

  \param fileName The name of the file to dump to
  \param vehLocation The location of the vehicle (drawn in a different color)
  \param closeOut Should the file be closed? 1-yes
*/
void rcsEngine::dump( int fileID )
{
  int count, i, j;
  //  float x,y;
  unsigned char red, blue, green;
  //  time_t timeOfDay;

  if( fileID <= 0 || dataPt <= 0 )
    return;

  if( write( fileID, "P6\n", strlen("P6\n") ) == -1 )
    {
      fileID = 0;
      return;
    }
  /*
  timeOfDay = time(NULL);
  sprintf( dataPt, "# %s\n", ctime(&timeOfDay));
  if( write( fileID, dataPt, strlen(dataPt) ) == -1 )
    {
      fileID = 0;
      return;
    }
  */
  sprintf( dataPt, "%d %d\n255\n", nmlSettings->worldWidth, nmlSettings->worldHeight );
  if( write( fileID, dataPt, strlen(dataPt) ) == -1 )
    {
      fileID = 0;
      return;
    }
  count = 0;
  for( i=nmlSettings->worldHeight-1; i>=0; i-- )
    {
      for( j=0; j<nmlSettings->worldWidth; j++ )
	{
	  if( worldModel[i][j] == 1 )
	    {
	      red = 255;
	      green = 255;
	      blue = 255;
	    }
	  else if( worldModel[i][j] == 2 )
	    {
	      red = 255;
	      green = 0;
	      blue = 0;
	    }
	  else
	    {
	      red = 0;
	      blue = 0;
	      green = 0;
	    }
	  dataPt[count++] = red;
	  dataPt[count++] = green;
	  dataPt[count++] = blue;
	}
    }
  // draw vehicle location
  /*
  x = ((int)floor((vehLocation.x - worldOffset.x) / nmlSettings->worldResolution));
  y  = ((int)floor((vehLocation.y - worldOffset.y) / nmlSettings->worldResolution));
  count = (int)(( (nmlSettings->worldHeight-1-x)*nmlSettings->worldWidth + y )*3);
  if( count < nmlSettings->worldWidth*nmlSettings->worldHeight*3 )
    {
      dataPt[count]   = 0;
      dataPt[count+1] = 255;
      dataPt[count+2] = 0;
    }
  */
  if( write( fileID, dataPt, nmlSettings->worldHeight*nmlSettings->worldWidth*3 ) == -1)
    {
      fileID = 0;
      return;
    }
}

/***********************************************************************/
/*! 
  clear the world model and allocate "clean" memory

\return 0 error
        1 no error
*/
bool rcsEngine::clearMap()
{
  int count, count2;

  if( worldModel != NULL )
    {
      for( count=0; count<nmlSettings->worldHeight; count++ )
	{
	  for( count2=0; count2<nmlSettings->worldWidth; count2++ )
	    {
	      delete [] worldModel[count];
	    }
	  delete [] worldModel;
	}
    }

  worldModel = new char*[nmlSettings->worldHeight];
  if( worldModel <= 0 )
    return(0);
  for( count=0; count<nmlSettings->worldHeight; count++)
    {
      worldModel[count] = new char [nmlSettings->worldWidth];
      if( worldModel[count] <= 0 )
	return(0);
      for( count2=0; count2<nmlSettings->worldWidth; count2++ )
	{
	  worldModel[count][count2] = 0;
	}
    }
  return(1);
}

/***********************************************************************/
/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool rcsEngine::checkVehCmdStatus(int which)
{
  int type;

  type = vehicleStatNml[which]->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_STAT_TYPE:
      vehicleCmdStatus[which] = vehicleStatNml[which]->get_address()->status;
      vehicleCmdEchoSerialNumber[which] = 
	vehicleStatNml[which]->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

/***********************************************************************/
/*!
  Check the status back from prim mobility
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool rcsEngine::checkPrimMobCmdStatus(int which)
{
  int type;

  type = primMobStatNml[which]->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case PRIM_MOB_JA_STAT_TYPE:
      primMobCmdStatus[which] = primMobStatNml[which]->get_address()->status;
      primMobCmdEchoSerialNumber[which] = 
	primMobStatNml[which]->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

/***********************************************************************/
/*!
  Check the status back from prim sp
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool rcsEngine::checkPrimSPCmdStatus(int which)
{
  int type;

  type = primSPStatNml[which]->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case PRIM_SP_STAT_TYPE:
      primSPCmdStatus[which] = primSPStatNml[which]->get_address()->status;
      primSPCmdEchoSerialNumber[which] = 
	primSPStatNml[which]->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

// routine borrowed from primShell
int rcsEngine::loadWaypoints(char * filename,
			     int vehicleID,
			     Waypoint * waypoint,
			     int * waypoint_length)
{
  FILE *fp;
#define BUFFERLEN 256
  char buffer[BUFFERLEN];
  char *ptr;
  Waypoint *wptr;
  int line;
  int abs;
  int retval;

  if (NULL == (fp = fopen(filename, "r")))
    return 1;
  
  printf( "opened file %s\n", filename );
  abs = 1;			// default is absolute coords
  *waypoint_length = 0;
  wptr= &waypoint[0];
  line = 0;
  retval = 0;
  while (!feof(fp)) {
    if (NULL == fgets(buffer, BUFFERLEN, fp))
      break;
    line++;
    ptr = buffer;
    while (isspace(*ptr))
      ptr++;
    if (0 == *ptr)
      continue;			// blank line
    if ('#' == *ptr)
      continue;			// comment
    if (! strncmp(ptr, "abs", 3)) abs = 1;
    else if (! strncmp(ptr, "rel", 3)) abs = 0;
    else if (4 == sscanf(ptr, "%lf %lf %lf %lf",
			 &wptr->p.x,
			 &wptr->p.y,
			 &wptr->neighborhood,
			 &wptr->speed)) {
      if (abs) {
	wptr->p = transformPoint( vehicleID, wptr->p, true );
      }
      wptr->p.z = 0.0;
      printf( "Set point: %f %f %f\n", wptr->p.x, wptr->p.y, wptr->p.z );
      (*waypoint_length)++;
      wptr++;
      if (*waypoint_length >= MAX_WAYPOINTS) break;
    } else {
      fprintf(stderr, "syntax error in %s, line %d: %s\n",
	      filename, line, buffer);
      retval = 1;
      break;
    }
  }

  fclose(fp);
  return retval;
#undef BUFFERLEN
}

/*
  Modification history:
  $Log: rcsSectMobPlEngine.cc,v $
  Revision 1.10  2009/01/17 16:45:12  ghenson
  included stdlib.h for system() and rand()
  Could just be a gcc 4.3.x thing

  Revision 1.9  2006/10/31 09:30:42  dr_steveb
  Changes for SPQR integration

  Revision 1.8  2006/09/26 14:52:31  tomrkramer
  Added #include "amSP.hh" so AMSP_OBSTACLE, etc. would be defined.

  Revision 1.7  2006/09/18 16:59:27  dr_steveb
  Added crash behavior

  Revision 1.6  2006/05/04 17:42:00  dr_steveb
  Changes to make multi-vehicle section exploration work.

  Revision 1.5  2006/04/14 23:25:27  proctor
  Added some missing ANSI headers

  Revision 1.4  2006/04/12 13:02:16  dr_steveb
  Section level exploration now works for a single vehicle

  Revision 1.3  2006/04/07 18:33:57  dr_steveb
  Check-in so compiles work for Fred. Still needs more work.

  Revision 1.2  2006/04/05 20:48:09  dr_steveb
  Fixed case sensitive issue with file

  Revision 1.1  2006/04/05 20:25:43  dr_steveb
  Generic rcs implementation for section level

*/

