/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file sectShell.cc

  \brief Command line shell to section level NML buffers.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.12 $
  $Date: 2009/03/24 17:15:44 $
  \endcode

  \author Stephen Balakirsky & Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "sectMobPL.hh"		// secMobPL_format()

static RCS_CMD_CHANNEL *sectMobPLCmdBuf = NULL;
static RCS_STAT_CHANNEL *sectMobPLStatBuf = NULL;
static RCS_CMD_CHANNEL *sectMobPLCfgBuf = NULL;
static RCS_STAT_CHANNEL *sectMobPLSetBuf = NULL;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "sectShell"

static int cleanupSectMobPLNmlBuffers(void)
{
  if (NULL != sectMobPLCmdBuf)
    delete sectMobPLCmdBuf;
  if (NULL != sectMobPLStatBuf)
    delete sectMobPLStatBuf;
  if (NULL != sectMobPLSetBuf)
    delete sectMobPLSetBuf;
  if (NULL != sectMobPLCfgBuf)
    delete sectMobPLCfgBuf;

  sectMobPLStatBuf = NULL;
  sectMobPLSetBuf = NULL;
  sectMobPLCmdBuf = NULL;
  sectMobPLCfgBuf = NULL;

  return 0;
}

static int initSectMobPLNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupSectMobPLNmlBuffers();

  sprintf(chanName, "%s%d", SECT_MOB_PL_CMD_NAME, bufnum);
  sectMobPLCmdBuf =
    new RCS_CMD_CHANNEL(sectMobPL_format, chanName,
			THIS_PROCESS, config_file);
  if (!sectMobPLCmdBuf->valid()) {
    delete sectMobPLCmdBuf;
    sectMobPLCmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SECT_MOB_PL_STAT_NAME, bufnum);
  sectMobPLStatBuf =
    new RCS_STAT_CHANNEL(sectMobPL_format, chanName,
			 THIS_PROCESS, config_file);
  if (!sectMobPLStatBuf->valid()) {
    delete sectMobPLStatBuf;
    sectMobPLStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SECT_MOB_PL_CFG_NAME, bufnum);
  sectMobPLCfgBuf =
    new RCS_CMD_CHANNEL(sectMobPL_format, chanName,
			THIS_PROCESS, config_file);
  if (!sectMobPLCfgBuf->valid()) {
    delete sectMobPLCfgBuf;
    sectMobPLCfgBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", SECT_MOB_PL_SET_NAME, bufnum);
  sectMobPLSetBuf =
    new RCS_STAT_CHANNEL(sectMobPL_format, chanName,
			 THIS_PROCESS, config_file);
  if (!sectMobPLSetBuf->valid()) {
    delete sectMobPLSetBuf;
    sectMobPLSetBuf = 0;
    return 1;
  }

  return 0;
}

#define rcs_stat_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

static void printStat(void)
{
  SectMobPLStat * stat = (SectMobPLStat *) sectMobPLStatBuf->get_address();
  SectMobPLSet * set = (SectMobPLSet *) sectMobPLSetBuf->get_address();

  printf("\nset command_type:   %s\n",
	 sectMobPL_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("line:               %d\n", (int) set->line);
  printf("source_line:        %d\n", (int) set->source_line);
  printf("source_file:        %s\n", (char *) set->source_file);
  printf("cycleTime:          %f\n", (double) set->cycleTime);
  printf("debug:              %X\n", set->debug);
  //  printf("fileSkip:           %i\n", (int)set->fileSkip);
  printf("adminState   :      %d\n", (int) set->admin_state);
  //  printf("wmFileName:         %s\n", set->wmFileName);
  //  printf("worldHeight:        %d\n", (int) set->worldHeight);
  //  printf("worldRes:           %f\n", (double) set->worldResolution);
  //  printf("worldWidth:         %d\n", (int) set->worldWidth);

  printf("command_type:       %s\n",
	 sectMobPL_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char*) stat->source_file);
  printf("heartbeat:          %d\n", (int) stat->heartbeat);
  printf("last cycleTime:     %f\n", (float)stat->tt.last);
  printf("avg cycleTime:      %f\n", (float)stat->tt.avg);
  printf("status message:     %s\n", stat->message );
  printf("num vehicles:       %d\n", stat->numVehicles );
}

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}

static void printHelp(void)
{
  printf("q | quit                    -- quit\n");
  printf("? | help                    -- print this help\n");
  printf("commands:\n" );
  printf("abort                       -- abort\n");
  printf("clear                       -- clear world model\n");
  printf("crash <reactTime> <warnDist>-- crash two vehicles for DOT\n");
  printf("deliver                     -- deliver packages in a factory\n");
  printf("exp <dist> <maxcost>        -- explore environment w/ at <dist> length legs\n");
  printf("file <file name> <repeat>     -- execute the commands in the waypoint file, repeat( 1 yes, 0 no)\n" );
  printf("halt                        -- halt\n");
  printf("init <ctrl level> <num veh> -- initialize\n");
  printf("shutdown                    -- shutdown\n");
  printf("wp <file name>              -- read file of vehicle id's and target locations" );
  printf("config commands:\n" );
  printf("ct <secs>                   -- set cycle time to <secs>\n");
  printf("debug <0, 1, ...>           -- configure debug level\n");
  printf("dump  <0, 1>                -- dump world model and display with ppmsdisplay (1 on, 0 off)\n");
  printf("load <file name>            -- load file as basis for world model\n");
  printf("store <file name>           -- write world model out to file\n");
}

SectMobPLCmdWaypoint *readWaypoints( char *fileName )
{
  FILE *fp;
#define BUFFERLEN 256
  char buffer[BUFFERLEN];
  char *ptr;
  int line;
  SectMobPLCmdWaypoint *retCmd;
  int numVehicles;
  bool error = false;

  if (NULL == (fp = fopen(fileName, "r")))
    return NULL;
  retCmd = new SectMobPLCmdWaypoint();
  
  numVehicles = 0;
  line = 0;
  while (!feof(fp)) {
    if (NULL == fgets(buffer, BUFFERLEN, fp))
      break;
    line++;
    ptr = buffer;
    while (isspace(*ptr))
      ptr++;
    if(0 == *ptr)
      continue;			// blank line
    if('#' == *ptr)
      continue;			// comment
    if(7 == sscanf(ptr, "%d %lf %lf %lf %lf %lf %c",
		   &retCmd->vehPoints[numVehicles].vehID,
		   &retCmd->vehPoints[numVehicles].waypoint.p.x,
		   &retCmd->vehPoints[numVehicles].waypoint.p.y,
		   &retCmd->vehPoints[numVehicles].waypoint.p.z,
		   &retCmd->vehPoints[numVehicles].waypoint.neighborhood,
		   &retCmd->vehPoints[numVehicles].waypoint.speed,
		   &retCmd->vehPoints[numVehicles].waypoint.quality))

      {
	numVehicles++;
	if( numVehicles >= MAX_SECTION_VEHICLES ) break;
      }
    else 
      {
	fprintf(stderr, "syntax error in %s, line %d: %s\n",
		fileName, line, buffer);
	error = true;
	break;
      }
    // else got a good line
  }

  if( error )
    delete retCmd;
  else
    retCmd->vehPoints_length = numVehicles;
  fclose(fp);
  return retCmd;
#undef BUFFERLEN
  
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
  char controlLevel[80];
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char fileName[MOAST_FILE_NAME_LEN];
  char *line;
  char *ptr;
  int i1;
  int retval;
  SectMobPLStat *sectMobPLStatPtr;
  SectMobPLSet *sectMobPLSetPtr;
  SectMobPLCmdInit sectMobPLCmdInit;
  SectMobPLCmdHalt sectMobPLCmdHalt;
  SectMobPLCmdAbort sectMobPLCmdAbort;
  SectMobPLCmdShutdown sectMobPLCmdShutdown;
  SectMobPLCmdCrash sectMobPLCmdCrash;
  SectMobPLCmdLoadWM sectMobPLCmdLoadWM;
  SectMobPLCmdDeliver sectMobPLCmdDeliver;
  SectMobPLCmdNop sectMobPLCmdNop;
  SectMobPLCmdFile sectMobPLCmdFile;
  SectMobPLCmdWaypoint *sectMobPLCmdWaypoint;
  SectMobPLCfgCycleTime sectMobPLCfgCycleTime;
  SectMobPLCfgDebug sectMobPLCfgDebug;
  SectMobPLCfgDumpWM sectMobPLCfgDump;
  SectMobPLCfgClearWM sectMobPLCfgClearWM;
  SectMobPLCfgStoreWM sectMobPLCfgStoreWM;
  
  SectMobPLCmdExplore sectMobPLCmdExplore;

#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      break;
    case 'i':
      INI_FILE = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // initial allocate all the buffers
  retval = initSectMobPLNmlBuffers(NML_FILE, bufnum);

  if (retval) {
    fprintf(stderr, "%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }
  // set up our convenient pointer to status and settings
  sectMobPLStatPtr = (SectMobPLStat *) sectMobPLStatBuf->get_address();
  sectMobPLSetPtr = (SectMobPLSet *) sectMobPLSetBuf->get_address();

  while (!feof(stdin)) {
#ifndef HAVE_READLINE_READLINE_H
    printf("> ");
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) {
      break;
    }
    line = buffer;
#else
    line = readline("> ");
    if (NULL == line) {
      break;
    }
    if (*line) {
      add_history(line);
    }
#endif

    sectMobPLStatBuf->read();
    sectMobPLSetBuf->read();

    ptr = line;
    while (isspace(*ptr))
      ptr++;			// skip leading white space
    if (0 == *ptr) {		// blank line
      printStat();
    } else if (ptr[0] == 'q' || !strcmp(ptr, "quit")) {
      break;
    } else if (ptr[0] == '?' || !strcmp(ptr, "help")) {
      printHelp();

      // begin commands
    } else if (!strwcmp(ptr, "init")) {
      if( 2 != sscanf(ptr, "%*s %s %d",
		      controlLevel,
		      &sectMobPLCmdInit.vehList_length)){
	printf("need control level and number of vehicles to control\n");
      }else{
	sectMobPLCmdInit.serial_number =
	  sectMobPLStatPtr->echo_serial_number + 1;
	// set up init command
	for( retval = 0; retval<sectMobPLCmdInit.vehList_length; retval++ )
	  {
	    sectMobPLCmdInit.vehList[retval] = retval+1;
	  }
	if( !strwcmp(controlLevel, "prim" ))
	  sectMobPLCmdInit.subordinate = PRIM;
	else if( !strwcmp(controlLevel, "veh" ))
	  sectMobPLCmdInit.subordinate = VEHICLE;
	else
	  {
	    printf("illegal control level %s\n", controlLevel );
	    break;
	  }
	sectMobPLCmdBuf->write(&sectMobPLCmdInit);
      }
    } else if (!strwcmp(ptr, "crash")) {
      if( 2!= sscanf(ptr, "%*s %f %f",
		     &sectMobPLCmdCrash.driverReact,
		     &sectMobPLCmdCrash.warnDist )) {
	printf("need reaction time in sec. and warn distance in m\n");
      } else {
	sectMobPLCmdCrash.serial_number = 
	  sectMobPLStatPtr->echo_serial_number + 1;
	sectMobPLCmdBuf->write(&sectMobPLCmdCrash);
      }
    } else if (!strwcmp(ptr, "abort")) {
      sectMobPLCmdAbort.serial_number =
	sectMobPLStatPtr->echo_serial_number + 1;
      sectMobPLCmdBuf->write(&sectMobPLCmdAbort);
    } else if (!strwcmp(ptr, "halt")) {
      sectMobPLCmdHalt.serial_number =
	sectMobPLStatPtr->echo_serial_number + 1;
      sectMobPLCmdBuf->write(&sectMobPLCmdHalt);
    } else if (!strwcmp(ptr, "shutdown")) {
      sectMobPLCmdShutdown.serial_number =
	sectMobPLStatPtr->echo_serial_number + 1;
      sectMobPLCmdBuf->write(&sectMobPLCmdShutdown);
    } else if (!strwcmp(ptr, "exp")) {
      if (2 != sscanf(ptr, "%*s %lf %lf",
		      &sectMobPLCmdExplore.exploreDist,
		      &sectMobPLCmdExplore.maxCost)) {
	printf("need max explore dist and max cost\n");
      } else {
	sectMobPLCmdExplore.serial_number =
	  sectMobPLStatPtr->echo_serial_number + 1;
	sectMobPLCmdBuf->write(&sectMobPLCmdExplore);
      }
    } else if (!strwcmp(ptr, "load")) {
      if( 1 != sscanf(ptr, "%*s %s", sectMobPLCmdLoadWM.fileName) ){
	printf("need file name\n");
      } else {
	sectMobPLCmdLoadWM.serial_number =
	  sectMobPLStatPtr->echo_serial_number + 1;
	sectMobPLCmdBuf->write(&sectMobPLCmdLoadWM);
      }
    } else if (!strwcmp(ptr, "file")) {
      if( 2 != sscanf(ptr, "%*s %s %d", sectMobPLCmdFile.fileName, &i1) ){
	printf("need file name and repeat (0 or 1)\n");
      } else {
	if( i1 )
	  sectMobPLCmdFile.repeat = true;
	else
	  sectMobPLCmdFile.repeat = false;
	sectMobPLCmdFile.serial_number =
	  sectMobPLStatPtr->echo_serial_number + 1;
	sectMobPLCmdBuf->write(&sectMobPLCmdFile);
      }
    } else if (!strwcmp(ptr, "wp")) {
      if( 1 != sscanf(ptr, "%*s %s", fileName) ){
	printf("need file name\n");
      } else {
	sectMobPLCmdWaypoint = readWaypoints(fileName);
	if( sectMobPLCmdWaypoint != NULL )
	  {
	    for( int count=0; count<sectMobPLCmdWaypoint->vehPoints_length; count++ )
	      {
		printf( "%d %lf %lf %lf %lf %lf %c\n",
			sectMobPLCmdWaypoint->vehPoints[count].vehID,
			sectMobPLCmdWaypoint->vehPoints[count].waypoint.p.x,
			sectMobPLCmdWaypoint->vehPoints[count].waypoint.p.y,
			sectMobPLCmdWaypoint->vehPoints[count].waypoint.p.z,
			sectMobPLCmdWaypoint->vehPoints[count].waypoint.neighborhood,
			sectMobPLCmdWaypoint->vehPoints[count].waypoint.speed,
			sectMobPLCmdWaypoint->vehPoints[count].waypoint.quality);
	      }
	    sectMobPLCmdBuf->write(sectMobPLCmdWaypoint);
	    delete sectMobPLCmdWaypoint;
	  }
      }
    } else if (!strwcmp(ptr, "deliver")) {
      sectMobPLCmdDeliver.serial_number =
	sectMobPLStatPtr->echo_serial_number + 1;
      sectMobPLCmdBuf->write(&sectMobPLCmdDeliver);
    } else if (!strwcmp(ptr, "nop")) {
      sectMobPLCmdNop.serial_number =
	sectMobPLStatPtr->echo_serial_number + 1;
      sectMobPLCmdBuf->write(&sectMobPLCmdNop);
      // begin config channel commands
    } else if (!strwcmp(ptr, "ct")) {
      if (1 !=
	  sscanf(ptr, "%*s %lf",
		 &sectMobPLCfgCycleTime.cycleTime)) {
	printf("need a cycleTime argument\n");
      } else {
	sectMobPLCfgCycleTime.serial_number =
	  sectMobPLSetPtr->echo_serial_number + 1;
	sectMobPLCfgBuf->
	  write(&sectMobPLCfgCycleTime);
      }
    } else if (!strwcmp(ptr, "debug")) {
      if (1 != sscanf(ptr, "%*s %i",
		      &sectMobPLCfgDebug.debug)) {
	printf("need debug mask\n");
      } else {
	sectMobPLCfgDebug.serial_number =
	  sectMobPLSetPtr->echo_serial_number + 1;
	sectMobPLCfgBuf->write(&sectMobPLCfgDebug);
      }
    } else if (!strwcmp(ptr, "clear")) {
      sectMobPLCfgClearWM.serial_number =
	sectMobPLSetPtr->echo_serial_number + 1;
      sectMobPLCfgBuf->write(&sectMobPLCfgClearWM);
    } else if (!strwcmp(ptr, "store")) {
      if (1 != sscanf(ptr, "%*s %s",
		      sectMobPLCfgStoreWM.fileName)) {
	printf("need file name\n");
      } else {
	sectMobPLCfgStoreWM.serial_number =
	  sectMobPLSetPtr->echo_serial_number + 1;
	sectMobPLCfgBuf->write(&sectMobPLCfgStoreWM);
      }
    } else if (!strwcmp(ptr, "dump")) {
      if( 1 != sscanf(ptr, "%*s %d", &i1)) {
	printf("need 0 for off or 1 for on\n" );
      }else{
	sectMobPLCfgDump.turnOn = i1;
	sectMobPLCfgDump.serial_number =
	  sectMobPLSetPtr->echo_serial_number + 1;
	sectMobPLCfgDump.skip = 5;
	sectMobPLCfgDump.dumpContinuous = true;
	strcpy(sectMobPLCfgDump.fileName, "SectMobPLDispOutput.ppm");
	sectMobPLCfgBuf->write(&sectMobPLCfgDump);
      }
    } else {
      printf("?\n");
    }

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  return cleanupSectMobPLNmlBuffers();
}

/*
  Modification history:

  $Log: sectShell.cc,v $
  Revision 1.12  2009/03/24 17:15:44  dr_steveb
  Added waypoint command

  Revision 1.11  2009/02/25 21:08:30  dr_steveb
  Fixed files to work with renamed section commands.

  Revision 1.10  2008/02/15 20:20:35  dr_steveb
  Adde waypoint command to the section echelon

  Revision 1.9  2007/12/20 02:32:45  dr_steveb
  Now have two vehicles performing start of manufacturing job. More to come...

  Revision 1.8  2007/12/17 21:36:03  dr_steveb
  Initial check-in of section level manufacturing application

  Revision 1.7  2006/09/18 16:59:28  dr_steveb
  Added crash behavior

  Revision 1.6  2006/04/13 14:30:09  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.5  2006/04/07 18:33:57  dr_steveb
  Check-in so compiles work for Fred. Still needs more work.

  Revision 1.4  2006/04/05 20:24:21  dr_steveb
  Start of new section algorithm for multi-vehicle control

  Revision 1.3  2005/11/14 18:06:58  dr_steveb
  Simple exloration behavior added to section level.

  Revision 1.2  2005/10/21 14:44:31  dr_steveb
  Added world dimensions to settings message.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.1  2005/10/11 12:06:58  stephen
  Initial check-in

*/

