/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file trafficShell.cc

  \brief Command line shell to traffic simulator

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.4 $
  $Date: 2007/09/11 19:50:26 $
  \endcode

  \author Stephen Balakirsky
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_BASE, etc.
#include "trafficCtrl.hh"
#include "trafficData.hh"
#include "primMobJA.hh"

static RCS_CMD_CHANNEL *trafficCtrlCmdBuf = NULL;
static RCS_STAT_CHANNEL *trafficCtrlStatBuf = NULL;
static RCS_CMD_CHANNEL *trafficCtrlCfgBuf = NULL;
static RCS_STAT_CHANNEL *trafficCtrlSetBuf = NULL;
static NML *trafficDataBuf = NULL;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define THIS_PROCESS "trafficShell"

static char* rcsPrimStateToString(int s) {
  static char str[40];
  switch (s) {
  case S0: return "S0";
  case S1: return "S1";
  case S2: return "S2";
  case S3: return "S3";
  case S4: return "S4";
  case S5: return "S5";
  case S6: return "S6";
  default:
    sprintf(str, "S%d?", s);
  }
  return str;
}


static int loadWaypoints(char * filename,
			 Waypoint * waypoint,
			 int * waypoint_length)
{
  FILE *fp;
#define BUFFERLEN 256
  char buffer[BUFFERLEN];
  char *ptr;
  Waypoint *wptr;
  int line;
  int retval;
  int abs;

  if (NULL == (fp = fopen(filename, "r")))
    return 1;

  *waypoint_length = 0;
  wptr= &waypoint[0];
  line = 0;
  retval = 0;
  while (!feof(fp)) {
    if (NULL == fgets(buffer, BUFFERLEN, fp))
      break;
    line++;
    ptr = buffer;
    while (isspace(*ptr))
      ptr++;
    if (0 == *ptr)
      continue;			// blank line
    if ('#' == *ptr)
      continue;			// comment

    if (! strncmp(ptr, "abs", 3)) abs = 1;
    else if (! strncmp(ptr, "rel", 3)) abs = 0;
    else if (5 == sscanf(ptr, "%lf %lf %lf %lf %lf",
		    &wptr->p.x,
		    &wptr->p.y,
		    &wptr->p.z,
		    &wptr->neighborhood,
		    &wptr->speed)) {
      (*waypoint_length)++;
      wptr++;
      if (*waypoint_length >= MAX_WAYPOINTS) break;
    } else if (4 == sscanf(ptr, "%lf %lf %lf %lf",
			   &wptr->p.x,
			   &wptr->p.y,
			   &wptr->neighborhood,
			   &wptr->speed)) {
      wptr->p.z = 0.0;
      (*waypoint_length)++;
      wptr++;
      if (*waypoint_length >= MAX_WAYPOINTS) break;
    } else {
      fprintf(stderr, "syntax error in %s, line %d: %s\n",
	      filename, line, buffer);
      retval = 1;
      break;
    }
  }

  fclose(fp);
  return retval;
#undef BUFFERLEN
}

static int cleanupTrafficNmlBuffers(void)
{
  if (NULL != trafficCtrlCmdBuf)
    delete trafficCtrlCmdBuf;
  if (NULL != trafficCtrlStatBuf)
    delete trafficCtrlStatBuf;
  if (NULL != trafficCtrlSetBuf)
    delete trafficCtrlSetBuf;
  if (NULL != trafficCtrlCfgBuf)
    delete trafficCtrlCfgBuf;
  if (NULL != trafficDataBuf)
    delete trafficDataBuf;

  trafficCtrlStatBuf = NULL;
  trafficCtrlSetBuf = NULL;
  trafficCtrlCmdBuf = NULL;
  trafficCtrlCfgBuf = NULL;
  trafficDataBuf = NULL;

  return 0;
}

static int initTrafficNmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupTrafficNmlBuffers();

  sprintf(chanName, "%s%d", TRAFFIC_CTRL_CMD_NAME, bufnum);
  trafficCtrlCmdBuf =
    new RCS_CMD_CHANNEL(trafficCtrl_format, chanName,
			THIS_PROCESS, config_file);
  if (!trafficCtrlCmdBuf->valid()) {
    delete trafficCtrlCmdBuf;
    trafficCtrlCmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", TRAFFIC_CTRL_STAT_NAME, bufnum);
  trafficCtrlStatBuf =
    new RCS_STAT_CHANNEL(trafficCtrl_format, chanName,
			 THIS_PROCESS, config_file);
  if (!trafficCtrlStatBuf->valid()) {
    delete trafficCtrlStatBuf;
    trafficCtrlStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", TRAFFIC_CTRL_CFG_NAME, bufnum);
  trafficCtrlCfgBuf =
    new RCS_CMD_CHANNEL(trafficCtrl_format, chanName,
			THIS_PROCESS, config_file);
  if (!trafficCtrlCfgBuf->valid()) {
    delete trafficCtrlCfgBuf;
    trafficCtrlCfgBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", TRAFFIC_CTRL_SET_NAME, bufnum);
  trafficCtrlSetBuf =
    new RCS_STAT_CHANNEL(trafficCtrl_format, chanName,
			 THIS_PROCESS, config_file);
  if (!trafficCtrlSetBuf->valid()) {
    delete trafficCtrlSetBuf;
    trafficCtrlSetBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d", TRAFFIC_DATA_NAME, bufnum);
  trafficDataBuf = 
    new NML(trafficData_format, chanName,
	    THIS_PROCESS, config_file);
  if (!trafficDataBuf->valid()) {
    delete trafficDataBuf;
    trafficDataBuf = 0;
    return 1;
  }
  return 0;
}

#define rcs_stat_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

static void printStat(void)
{
  int count;
  TrafficCtrlStat * stat = (TrafficCtrlStat *) trafficCtrlStatBuf->get_address();
  TrafficCtrlSet * set = (TrafficCtrlSet *) trafficCtrlSetBuf->get_address();
  TrafficData *data = (TrafficData *) trafficDataBuf->get_address();

  printf("command_type:       %s\n",
	 trafficCtrl_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_stat_to_string(stat->status));
  //  printf("state:              %s\n", rcsVehStateToString(stat->state));
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("heartbeat:          %d\n", (int) stat->tt.count);
  printf("set cycleTime:      %f\n", (double) set->cycleTime);
  printf("actual cycleTime:   %lf\n",stat->tt.avg);
  printf("debug:              %X\n", set->debug);
  printf("message:            %s\n", stat->message);
  printf("error:              %d\n", stat->error);
  printf("time:               %f\n", data->time);
  printf("data for %d vehicles\n",   data->vehicles_length );
  printf("xLoc:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].tranTrue.x );
    }
  printf("\nyLoc:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].tranTrue.y );
    }
  printf("\nzLoc:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].tranTrue.z );
    }
  printf("\nroll:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].rpyTrue.r );
    }
  printf("\npitch:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].rpyTrue.p );
    }
  printf("\nyall:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].rpyTrue.y );
    }
  printf("\nspeed:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].speed );
    }
  printf("\nsteer:");
  for( count=0; count<data->vehicles_length; count++ )
    {
      printf( "\t%f", data->vehicles[count].steeringCurvature );
    }
  printf("\npIndex:");
  for( count=0; count<stat->pathIndex_length; count++ )
    {
      printf( "\t%d", stat->pathIndex[count] );
    }
  printf("\npState:");
  for( count=0; count<stat->primState_length; count++ )
    {
      printf( "\t%s", rcsPrimStateToString(stat->primState[count]) );
    }
  printf("\npStatus:");
  for( count=0; count<stat->primStatus_length; count++ )
    {
      printf( "\t%s", rcs_stat_to_string(stat->primStatus[count]) );
    }
  printf("\nbotName:");
  for( count=0; count<set->robotInfo_length; count++ )
    {
      printf( "\t%s", set->robotInfo[count].robotName );
    }

  printf( "\n" );
}

static int strwcmp(char *test, char *match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match)
      return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test)))
    return 0;
  return 1;
}

static void printHelp(void)
{
  printf("q | quit          -- quit\n");
  printf("? | help          -- print this help\n");
  printf("init              -- initialize\n");
  printf("halt              -- halt\n");
  printf("abort             -- abort\n");
  printf("shutdown          -- shutdown\n");
  /*
  printf("ct <secs>         -- set cycle time to <secs>\n");
  printf("mvl <x> <y>       -- move to <x> <y> in local coordinates\n");
  printf("mvg <x> <y>       -- move to <x> <y> in global coordinates\n");
  printf("exp <x> <y>       -- explore while moving to local waypoint\n");
  printf("dump              -- dump world model and display with imageview\n");
  */
  printf("debug <0, 1, ...> -- configure debug level\n");
}

int main(int argc, char *argv[])
{
  int retVal;
  int serialNumber = 1;
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  int bufnum = 1;
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char *line;
  char *ptr;
  int retval;
  int count, count2;
  TrafficCtrlCmdNop trafficCtrlCmdNop;
  TrafficCtrlCmdInit trafficCtrlCmdInit;
  TrafficCtrlCmdAbort trafficCtrlCmdAbort;
  TrafficCtrlCmdHalt trafficCtrlCmdHalt;
  TrafficCtrlCmdShutdown trafficCtrlCmdShutdown;
  TrafficCtrlCmdMoveWaypoint trafficCtrlCmdWaypoint;

  TrafficCtrlCfgNop trafficCtrlCfgNop;
  TrafficCtrlCfgCycleTime trafficCtrlCfgCycleTime;
  TrafficCtrlCfgDebug trafficCtrlCfgDebug;
  TrafficCtrlCfgMove trafficCtrlCfgMove;


#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      break;
    case 'i':
      INI_FILE = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // initial allocate all the buffers
  retval = initTrafficNmlBuffers(NML_FILE, bufnum);

  if (retval) {
    fprintf(stderr, "%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }

  while (!feof(stdin)) {
#ifndef HAVE_READLINE_READLINE_H
    printf("> ");
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) {
      break;
    }
    line = buffer;
#else
    line = readline("> ");
    if (NULL == line) {
      break;
    }
    if (*line) {
      add_history(line);
    }
#endif

    trafficCtrlStatBuf->read();
    trafficCtrlSetBuf->read();
    trafficDataBuf->read();

    ptr = line;
    while (isspace(*ptr))
      ptr++;			// skip leading white space
    if (0 == *ptr) {		// blank line
      printStat();
    } else if (ptr[0] == 'q' || !strcmp(ptr, "quit")) {
      break;
    } else if (ptr[0] == '?' || !strcmp(ptr, "help")) {
      printHelp();
    } else if (!strwcmp(ptr, "init")) {
      trafficCtrlCmdInit.serial_number = serialNumber;
      serialNumber+=2;
      trafficCtrlCmdBuf->write(&trafficCtrlCmdInit);
    } else if (!strwcmp(ptr, "nop")) {
      trafficCtrlCmdNop.serial_number = serialNumber;
      serialNumber+=2;
      trafficCtrlCmdBuf->write(&trafficCtrlCmdNop);
    } else if (!strwcmp(ptr, "halt")) {
      trafficCtrlCmdHalt.serial_number = serialNumber;
      serialNumber+=2;
      trafficCtrlCmdBuf->write(&trafficCtrlCmdHalt);
    } else if (!strwcmp(ptr, "abort")) {
      trafficCtrlCmdAbort.serial_number = serialNumber;
      serialNumber+=2;
      trafficCtrlCmdBuf->write(&trafficCtrlCmdAbort);
    } else if (!strwcmp(ptr, "shutdown")) {
      trafficCtrlCmdShutdown.serial_number = serialNumber;
      serialNumber+=2;
      trafficCtrlCmdBuf->write(&trafficCtrlCmdShutdown);
    } else if (!strwcmp(ptr, "ct")) {
      if (1 !=
	  sscanf(ptr, "%*s %lf",
		 &trafficCtrlCfgCycleTime.cycleTime)) {
	printf("need a cycleTime argument\n");
      } else {
	trafficCtrlCfgCycleTime.serial_number = serialNumber;
	serialNumber+=2;
	trafficCtrlCfgBuf->
	  write(&trafficCtrlCfgCycleTime);
      }
    } else if (!strwcmp(ptr, "debug")) {
      if (1 != sscanf(ptr, "%*s %i",
		      &trafficCtrlCfgDebug.debug)) {
	printf("need debug mask\n");
      } else {
	trafficCtrlCfgDebug.serial_number = serialNumber;
	serialNumber+=2;
	trafficCtrlCfgBuf->write(&trafficCtrlCfgDebug);
     }
    } else if (!strwcmp(ptr, "wp")) {
      if (1 != sscanf(ptr, "%*s %i",
		      &trafficCtrlCmdWaypoint.numVehicles)) {
	printf("need number of vehicles and file\n");
      }else{
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over word
	while (isspace(*ptr))
	  ptr++;                        // skip over space
	while (!isspace(*ptr) && 0 != *ptr)
	  ptr++;			// skip over num vehicles
	while (isspace(*ptr))
	  ptr++;
	if (0 == *ptr) {
	  printf("need a data file\n");
	  continue;
	}
	retVal = loadWaypoints(ptr,
			       &trafficCtrlCmdWaypoint.waypoint[0][0],
			       &trafficCtrlCmdWaypoint.waypoint_length[0]);
	if( retVal != 0 ){
	  printf("can't load ``%s''\n", ptr);
	}else{
	  for( count=1; count<trafficCtrlCmdWaypoint.numVehicles; count++ )
	    {
	      for( count2=0;count2<trafficCtrlCmdWaypoint.waypoint_length[0]; count2++ )
		{
		  trafficCtrlCmdWaypoint.waypoint[count][count2] =
		    trafficCtrlCmdWaypoint.waypoint[0][count2];
		}
	      trafficCtrlCmdWaypoint.waypoint_length[count] = 
		trafficCtrlCmdWaypoint.waypoint_length[0];
	    }

	  trafficCtrlCmdBuf->write(&trafficCtrlCmdWaypoint);
	}
      }
    } else {
      printf("?\n");
    }

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  return cleanupTrafficNmlBuffers();
}

/*
  Modification history:

  $Log: trafficShell.cc,v $
  Revision 1.4  2007/09/11 19:50:26  dr_steveb
  Updates to get name of vehicle into structure.

  Revision 1.3  2007/05/16 15:17:19  dr_steveb
  Traffic simulation controls now working

  Revision 1.2  2007/05/03 16:34:49  dr_steveb
  Changes to allow for multiple robots to be spawned from a single controller (i.e. trafficSim). TrafficSim is not done...

  Revision 1.1  2007/04/25 15:09:03  dr_steveb
  Added traffic simulation files

*/

