/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file   simpleMisSim.cc

  \brief  Simple simulator for mission package servo level.

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.3 $
  $Date: 2006/07/26 15:44:22 $
  \endcode

  \ref simpleMisSim

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		// sscanf()
#include <stddef.h>		// NULL
#include <stdlib.h>
#include <pthread.h>		// pthread_exit()
#include <rcs.hh>
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"	// NAV_DATA_EXT_BASE
#include "servoMisJA.hh"	// servoMisJA_format()
#include "simpleSim.hh"		// simpleMisSimInit,Step,Quit, simpleWorld

static RCS_CMD_CHANNEL *servoMisJACmdBuf = NULL;
static RCS_STAT_CHANNEL *servoMisJAStatBuf = NULL;
static RCS_STAT_CHANNEL *servoMisJASetBuf = NULL;
static RCS_CMD_CHANNEL *servoMisJACfgBuf = NULL;

static ServoMisJAStat servoMisJAStat;
static ServoMisJASet servoMisJASet;

#define THIS_PROCESS "simware"	// NML process name for the server

#define DEFAULT_CYCLE_TIME 0.010

static int cleanupServoMisJANmlBuffers(void)
{
  if (NULL != servoMisJACmdBuf)
    delete servoMisJACmdBuf;
  if (NULL != servoMisJAStatBuf)
    delete servoMisJAStatBuf;
  if (NULL != servoMisJASetBuf)
    delete servoMisJASetBuf;
  if (NULL != servoMisJACfgBuf)
    delete servoMisJACfgBuf;

  servoMisJAStatBuf = NULL;
  servoMisJASetBuf = NULL;
  servoMisJACmdBuf = NULL;
  servoMisJACfgBuf = NULL;

  return 0;
}

static int initServoMisJANmlBuffers(char *config_file, int bufnum)
{
  char chanName[MOAST_NML_BUFFER_NAME_LEN];

  (void) cleanupServoMisJANmlBuffers();

  sprintf(chanName, "%s%d_1", SERVO_MIS_JA_CMD_NAME, bufnum);
  servoMisJACmdBuf =
    new RCS_CMD_CHANNEL(servoMisJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMisJACmdBuf->valid()) {
    delete servoMisJACmdBuf;
    servoMisJACmdBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d_1", SERVO_MIS_JA_STAT_NAME, bufnum);
  servoMisJAStatBuf =
    new RCS_STAT_CHANNEL(servoMisJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMisJAStatBuf->valid()) {
    delete servoMisJAStatBuf;
    servoMisJAStatBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d_1", SERVO_MIS_JA_SET_NAME, bufnum);
  servoMisJASetBuf =
    new RCS_STAT_CHANNEL(servoMisJA_format, chanName,
			 THIS_PROCESS, config_file);
  if (!servoMisJASetBuf->valid()) {
    delete servoMisJASetBuf;
    servoMisJASetBuf = 0;
    return 1;
  }

  sprintf(chanName, "%s%d_1", SERVO_MIS_JA_CFG_NAME, bufnum);
  servoMisJACfgBuf =
    new RCS_CMD_CHANNEL(servoMisJA_format, chanName,
			THIS_PROCESS, config_file);
  if (!servoMisJACfgBuf->valid()) {
    delete servoMisJACfgBuf;
    servoMisJACfgBuf = 0;
    return 1;
  }

  return 0;
}

static void doCmdInit(ServoMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdHalt(ServoMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdAbort(ServoMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    state_default(stat);
  }
}

static void doCmdShutdown(ServoMisJAStat * stat)
{
  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
  } else {
    (void) simpleMisSimQuit();
    state_default(stat);
  }
}

static void doCmdMove(ServoMisJACmdMove * cmd, ServoMisJAStat * stat)
{
  PM_RPY rpy = cmd->goal.rot;

  if (state_match(stat, NEW_COMMAND)) {
    state_new(stat);
    status_next(stat, RCS_DONE);
    state_next(stat, S0);
    simpleWorldLock(&simpleWorld);
    simpleWorld.sensorHeight = cmd->goal.tran.z;
    simpleWorld.sensorPitch = rpy.p;
    simpleWorldUnlock(&simpleWorld);
  } else {
    state_default(stat);
  }
}

static void doCfgCycleTime(ServoMisJACfgCycleTime * cfg,
			   ServoMisJASet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    if (cfg->cycleTime <= 0.0) {
      status_next(set, RCS_ERROR);
    } else {
      set->cycleTime = cfg->cycleTime;
      status_next(set, RCS_DONE);
    }
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

static void doCfgDebug(ServoMisJACfgDebug * cfg,
		  ServoMisJASet * set)
{
  if (state_match(set, NEW_COMMAND)) {
    state_new(set);
    set->debug = cfg->debug;
    status_next(set, RCS_DONE);
    state_next(set, S0);
  } else {
    state_default(set);
  }
}

/*
  iniLoad() will load INI file params from 'filename', returning 0
  if everything is fine, 1 if the file can't be opened or there are
  errors in the values.

  Values loaded:

  [SERVO_MIS_JA]
  CYCLE_TIME = <positive number>
  MOUNT_TO_VEH = <X> <Y> <Z> <Roll> <Pitch> <Yaw>
*/

#if ! defined(INI_OK)
#define INI_OK      0
#define INI_DEFAULT 1
#define INI_INVALID 2
#endif

static int iniLoad(const char * prog, const char * filename, ServoMisJASet * set, ServoMisJAStat * stat)
{
  INIFILE inifile;
  const char * inistring;
  double d;
  PM_POSE IDENTITY_POSE = PM_POSE(PM_CARTESIAN(0,0,0),PM_QUATERNION(1,0,0,0));
  PM_CARTESIAN tran;
  PM_RPY rpy;
  int retval = INI_OK;

  if (0 != inifile.open(filename)) {
    rcs_print_error("%s: can't open INI file %s, using defaults for servoMisJASet\n", prog, filename);
    set->cycleTime = DEFAULT_CYCLE_TIME;
    return INI_INVALID;
  }

  if (NULL != (inistring = inifile.find("CYCLE_TIME", "SERVO_MIS_JA"))) {
    if (1 == sscanf(inistring, "%lf", &d) &&
	d > 0.0) {
      // found, and valid
      set->cycleTime = d;
    }
    else {
      // found, but invalid
      set->cycleTime = DEFAULT_CYCLE_TIME;
      rcs_print_error("%s: invalid value '%s' for [SERVO_MIS_JA] CYCLE_TIME in '%s'\n", prog, inistring, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    rcs_print_error("%s: can't find [SERVO_MIS_JA] CYCLE_TIME in '%s'\n", prog, filename);
    set->cycleTime = DEFAULT_CYCLE_TIME;
    rcs_print_error("%s: using %f\n", prog, set->cycleTime);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  if (NULL != (inistring = inifile.find("MOUNT_TO_VEH", "SERVO_MIS_JA"))) {
    if (6 == sscanf(inistring, "%lf %lf %lf %lf %lf %lf", 
		    &tran.x, &tran.y, &tran.z,
		    &rpy.r, &rpy.p, &rpy.y)) {
      // found, and valid
      rpy.r *= TO_RAD;		// convert to radians, since ini is degrees
      rpy.p *= TO_RAD;
      rpy.y *= TO_RAD;
      stat->mountToVeh.tran = tran;
      stat->mountToVeh.rot = rpy;
    }
    else {
      // found, but invalid
      tran.z = 0.0;
      rpy.p = 0.0;
      stat->mountToVeh = IDENTITY_POSE;
      rcs_print_error("%s: invalid value '%s' for [SERVO_MIS_JA] MOUNT_TO_VEH in '%s', setting to I[]\n", prog, inistring, filename);
      retval = INI_INVALID;
    }
  }
  else {
    // not found at all
    tran.z = 0.0;
    rpy.p = 0.0;
    stat->mountToVeh = IDENTITY_POSE;
    rcs_print_error("%s: can't find [SERVO_MIS_JA] MOUNT_TO_VEH in '%s', setting to I[]\n", prog, filename);
    retval = (retval == INI_OK ? INI_DEFAULT : retval);
  }

  inifile.close();

  return retval;
}

int simpleMisSimInit(int argc, char *argv[])
{
  int retval;
  ServoMisJACmdInit servoMisJACmdInit;
  ServoMisJACfgCycleTime servoMisJACfgCycleTime;

  // connect to the NML buffers
  retval = initServoMisJANmlBuffers(NML_FILE, bufnum);
  if (retval) {
    rcs_print_error("%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }
  
  // read in the INI file params into the settings struct
  if (INI_INVALID == iniLoad(argv[0], INI_FILE, &servoMisJASet, &servoMisJAStat)) {
    rcs_print_error("%s: error loading INI file '%s'\n", argv[0], INI_FILE);
    return 1;
  }

  // initialize and report our status
  servoMisJAStat.echo_serial_number = 0;
  servoMisJAStat.heartbeat = 0;
  servoMisJAStatBuf->write(&servoMisJAStat);

  // initialize and report our settings
  servoMisJASet.echo_serial_number = 0;
  servoMisJASetBuf->write(&servoMisJASet);

  // stuff an init cmd
  servoMisJACmdInit.serial_number = 1;
  servoMisJACmdBuf->write(&servoMisJACmdInit);

  // stuff a cycle time config
  servoMisJACfgCycleTime.serial_number = 1;
  servoMisJACfgCycleTime.cycleTime = servoMisJASet.cycleTime;
  servoMisJACfgBuf->write(&servoMisJACfgCycleTime);

  return 0;
}

static int done = 0;

int simpleMisSimQuit(void)
{
  done = 1;

  return 0;
}

void * simpleMisSimStep(void * arg)
{
  NMLTYPE cmd_type, cfg_type;
  int serial_number;
  PM_RPY rpy(0, 0, 0);

  done = 0;
  while (! done) {
    ServoMisJACfgCycleTime *cfgCycleTimePtr;
    ServoMisJACfgDebug *cfgDebugPtr;
    
    // read cmd
    cmd_type = servoMisJACmdBuf->read();
    serial_number = servoMisJACmdBuf->get_address()->serial_number;
    switch (cmd_type) {

    case 0:
      // no new cmd
      break;

    case -1:
      // comm error
      rcs_print_error("NML cmd read error\n");
      break;

    case SERVO_MIS_JA_CMD_INIT_TYPE:
    case SERVO_MIS_JA_CMD_HALT_TYPE:
    case SERVO_MIS_JA_CMD_ABORT_TYPE:
    case SERVO_MIS_JA_CMD_SHUTDOWN_TYPE:
    case SERVO_MIS_JA_CMD_MOVE_TYPE:
      servoMisJAStat.command_type = cmd_type;
      if (serial_number != servoMisJAStat.echo_serial_number) {
	servoMisJAStat.echo_serial_number = serial_number;
	servoMisJAStat.state = NEW_COMMAND;
	servoMisJAStat.status = RCS_EXEC;
      }
      break;

    default:
      rcs_print_error("unknown cmd: %d\n", (int) cmd_type);
      break;
    }				// switch (cmd_type)

    // read cfg
    cfg_type = servoMisJACfgBuf->read();
    serial_number = servoMisJACfgBuf->get_address()->serial_number;
    switch (cfg_type) {
    case 0:
      // no new cmd
      break;

    case -1:
      // comm error
      rcs_print_error("NML cfg read error\n");
      break;

    case SERVO_MIS_JA_CFG_CYCLE_TIME_TYPE:
      servoMisJASet.command_type = cfg_type;
      if( serial_number != servoMisJASet.echo_serial_number) {
	servoMisJASet.echo_serial_number = serial_number;
	servoMisJASet.state = NEW_COMMAND;
	servoMisJASet.status = RCS_EXEC;
      }
      cfgCycleTimePtr = 
	(ServoMisJACfgCycleTime *) servoMisJACfgBuf->get_address();
      break;

    case SERVO_MIS_JA_CFG_DEBUG_TYPE:
      servoMisJASet.command_type = cfg_type;
      if( serial_number != servoMisJASet.echo_serial_number) {
	servoMisJASet.echo_serial_number = serial_number;
	servoMisJASet.state = NEW_COMMAND;
	servoMisJASet.status = RCS_EXEC;
      }
      cfgDebugPtr = 
	(ServoMisJACfgDebug *) servoMisJACfgBuf->get_address();
      break;

    default:
      break;
    }

    // handle cmd state tables
    switch (servoMisJAStat.command_type) {

    case SERVO_MIS_JA_CMD_INIT_TYPE:
      doCmdInit(&servoMisJAStat);
      break;

    case SERVO_MIS_JA_CMD_HALT_TYPE:
      doCmdHalt(&servoMisJAStat);
      break;

    case SERVO_MIS_JA_CMD_ABORT_TYPE:
      doCmdAbort(&servoMisJAStat);
      break;

    case SERVO_MIS_JA_CMD_SHUTDOWN_TYPE:
      doCmdShutdown(&servoMisJAStat);
      break;

    case SERVO_MIS_JA_CMD_MOVE_TYPE:
      doCmdMove((ServoMisJACmdMove *) servoMisJACmdBuf->get_address(),
		&servoMisJAStat);
      break;

    default:
      // no cmd to handle
      break;
    }				// switch (cmd)

    // handle config state tables
    switch (servoMisJASet.command_type) {
    case SERVO_MIS_JA_CFG_CYCLE_TIME_TYPE:
      doCfgCycleTime(cfgCycleTimePtr, &servoMisJASet);
      break;

    case SERVO_MIS_JA_CFG_DEBUG_TYPE:
      doCfgDebug(cfgDebugPtr, &servoMisJASet);
      break;

    default:
      // no config to handle
      break;
    }

    // update stat and set

    simpleWorldLock(&simpleWorld);
    servoMisJAStat.mountToVeh.tran.z = simpleWorld.sensorHeight;
    rpy.p = simpleWorld.sensorPitch;
    simpleWorldUnlock(&simpleWorld);

    servoMisJAStat.heartbeat++;
    servoMisJAStat.mountToVeh.rot = rpy;

    servoMisJAStatBuf->write(&servoMisJAStat);
    servoMisJASetBuf->write(&servoMisJASet);

    esleep(servoMisJASet.cycleTime);
  }

  (void) cleanupServoMisJANmlBuffers();

  pthread_exit(0);

  return (void *) 0;
}

/*! \defgroup simpleMisSim
  This simulation reads and writes the servoMisJA NML buffers,
  and simulates the response of the pointing of the mission package.
*/

/*
  Modification history:

  $Log: simpleMisSim.cc,v $
  Revision 1.3  2006/07/26 15:44:22  dr_steveb
  Changes to new buffer and process names

  Revision 1.2  2006/04/07 13:12:12  proctor
  Took out getopt() from the threads, and made NML_FILE, INI_FILE and bufnum
  globals. Apparently getopt() doesn't work from within threads.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.13  2005/08/04 14:47:38  proctor
  Removed baseToVeh, changed misToVeh to mountToVeh

  Revision 1.12  2005/07/20 21:54:46  proctor
  simpleMis,Mob,SPSim.cc: added quitting upon shutdown.
  simpleSPSim.cc: also added cmd and cfg interface to make it a full RCS module.

  Revision 1.11  2005/07/14 18:51:01  proctor
  Handled height in the move command

  Revision 1.10  2005/07/14 17:50:48  proctor
  simpleMisSim.cc: incremented heartbeat
  simpleSPSim.cc: handled horizontal, vertical pitch better

  Revision 1.9  2005/07/13 21:01:12  proctor
  Handled move command

  Revision 1.8  2005/07/13 20:36:29  proctor
  Handled cycle time configuration

  Revision 1.7  2005/07/13 19:48:20  proctor
  Fixed handling of sensorHeight, sensorPitch globals

  Revision 1.6  2005/06/29 14:32:05  proctor
  Took out annoying .nml, .ini file printing; added obstacle collisions
  in simpleMobSim.cc

  Revision 1.5  2005/06/23 17:54:28  proctor
  Cleaned up, no new features

 */
