/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simpleSim.cc

  \brief Main loop that calls mobility, mission and sensor processing
  simulations, and maintains the simulated world.

  Syntax: simpleSim {-n \<nml file\>} {-i \<ini file\>} {-d \<buffer digit\>},
  where

  \<nml file\> is the name of the NML file to use, e.g., moast.nml

  \<ini file\> is the name of the INI file to use, e.g., moast.ini

  \<buffer number\> is the suffix number of the NML buffers to connect to,
  e.g., '1' for navDataExt1.

  \ref simpleSim

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.3 $
  $Date: 2006/12/12 19:20:40 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <stdio.h>		// perror()
#include <signal.h>		// signal(), ...
#include <stdlib.h>		// getenv()
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <pthread.h>		// pthread_XXX
#include <rcs.hh>		// INIFILE
#include <posemath.h>		// PM_POSE, for simpleSim.hh
#include "simpleSim.hh"		// simpleMobSimStep(), ...

#define DEFAULT_INI_FILE "../etc/moast.ini"

//! the global structure that holds the simulated world
SimpleWorld simpleWorld;

int simpleWorldLock(SimpleWorld * world)
{
  return pthread_mutex_lock((pthread_mutex_t *) world->mutexPtr);
}

int simpleWorldUnlock(SimpleWorld * world)
{
  return pthread_mutex_unlock((pthread_mutex_t *) world->mutexPtr);
}

static int simpleWorldInit(SimpleWorld * world, void * mutex)
{
  world->x = 0;
  world->y = 0;
  world->yaw = 0;
  world->v = 0;
  world->w = 0;
  world->wLeft = 0;
  world->wRight = 0;
  world->obstacleSpacing = 0;
  world->obstacleRadius = 0;
  world->sensorHeight = 0;
  world->sensorPitch = 0;
  world->doObstacles = 0;
  world->mutexPtr = mutex;

  return 0;
}

/*
  iniLoad() will load INI file params from 'filename', returning 0
  if everything is fine, 1 if the file can't be opened or there are
  errors in the values.

  Values loaded:

  [SIMPLE_SIM]
  OBSTACLE_SPACING = <positive number>
  OBSTACLE_RADIUS = <positive number>
*/
static int iniLoad(const char *prog, const char *filename)
{
  INIFILE inifile;
  const char SIMPLE_SIM[] = "SIMPLE_SIM";
  const char OBSTACLE_SPACING[] = "OBSTACLE_SPACING";
  const char OBSTACLE_RADIUS[] = "OBSTACLE_RADIUS";
  const char *section;
  const char *key;
  const char *inistring;
  double obstacleSpacing;
  double obstacleRadius;
  int retval = 0;

  if (0 != inifile.open(filename)) {
    rcs_print_error("%s: can't open INI file %s\n", prog, filename);
    return 1;
  }

  section = SIMPLE_SIM;

  key = OBSTACLE_SPACING;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &obstacleSpacing)) {
      // found, and valid
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      rcs_print_error("%s: using zero value\n", prog);
      obstacleSpacing = -1.0;
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    rcs_print_error("%s: using zero value\n", prog);
    obstacleSpacing = -1.0;
    retval = 1;
  }

  key = OBSTACLE_RADIUS;
  if (NULL != (inistring = inifile.find(key, section))) {
    if (1 == sscanf(inistring, "%lf", &obstacleRadius)) {
      // found, and valid
    } else {
      // found, but invalid
      rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n", prog,
		      inistring, section, key, filename);
      rcs_print_error("%s: using zero value\n", prog);
      obstacleRadius = -1.0;
      retval = 1;
    }
  } else {
    // not found at all
    rcs_print_error("%s: can't find [%s] %s in '%s'\n", prog, section, key,
		    filename);
    rcs_print_error("%s: using zero value\n", prog);
    obstacleRadius = -1.0;
    retval = 1;
  }

  inifile.close();

  if (obstacleSpacing <= 0.0 ||
      obstacleRadius <= 0.0) {
    simpleWorld.obstacleSpacing = 0.0;
    simpleWorld.obstacleRadius = 0.0;
    simpleWorld.doObstacles = 0;
  } else {
    simpleWorld.obstacleSpacing = obstacleSpacing;
    simpleWorld.obstacleRadius = obstacleRadius;
    simpleWorld.doObstacles = 1;
  }

  return retval;
}

static void quit(int sig)
{
  (void) simpleSPSimQuit();
  (void) simpleMisSimQuit();
  (void) simpleMobSimQuit();
}

/* globals for use by other threads */
char * NML_FILE = NULL;
char * INI_FILE = NULL;
int bufnum = 1;

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  pthread_mutex_t simpleWorldMutex = PTHREAD_MUTEX_INITIALIZER;
  pthread_t mobThr, misThr, spThr;
  pthread_attr_t mobAttr, misAttr, spAttr;
  int retval;

  printf("WARNING! I DON'T WORK ANYMORE! FIX ME TO MATCH SIMWARE!\n");
  printf("Let's see what happens...\n");

  (void) pthread_mutex_unlock(&simpleWorldMutex);

  simpleWorldInit(&simpleWorld, (void *) &simpleWorldMutex);

  while (1) {
    char option = getopt(argc, argv, "n:i:d:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      break;
    case 'i':
      INI_FILE = optarg;
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
    } else {
      NML_FILE = nml_file_default;
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }

  // get the name of the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
    } else {
      INI_FILE = ini_file_default;
    }
  }
  
  // read in the INI file params
  if (INI_INVALID == iniLoad(argv[0], INI_FILE)) {
    rcs_print_error("%s: error loading INI file '%s'\n", argv[0], INI_FILE);
    return 1;
  }

  retval = simpleMobSimInit(argc, argv);
  if (retval != 0) {
    fprintf(stderr, "simpleSim: can't initialize simpleMobSim\n");
    return 1;
  }
  retval = pthread_attr_init(&mobAttr);
  if (retval != 0) {
    perror("simpleSim: pthread_attr_init");
    return 1;
  }
  retval = pthread_create(&mobThr, &mobAttr, simpleMobSimStep, 0);
  if (retval != 0) {
    perror("simpleSim: pthread_create");
    return 1;
  }

  retval = simpleMisSimInit(argc, argv);
  if (retval != 0) {
    fprintf(stderr, "simpleSim: can't initialize simpleMisSim\n");
    return 1;
  }
  retval = pthread_attr_init(&misAttr);
  if (retval != 0) {
    perror("simpleSim: pthread_attr_init");
    return 1;
  }
  retval = pthread_create(&misThr, &misAttr, simpleMisSimStep, 0);
  if (retval != 0) {
    perror("simpleSim: pthread_create");
    return 1;
  }

  retval = simpleSPSimInit(argc, argv);
  if (retval != 0) {
    fprintf(stderr, "simpleSim: can't initialize simpleSPSim\n");
    return 1;
  }
  retval = pthread_attr_init(&spAttr);
  if (retval != 0) {
    perror("simpleSim: pthread_attr_init");
    return 1;
  }
  retval = pthread_create(&spThr, &spAttr, simpleSPSimStep, 0);
  if (retval != 0) {
    perror("simpleSim: pthread_create");
    return 1;
  }

  // set up SIGINT to call simpleMob,Mis,SPQuit(), which will
  // release the pthread_joins below
  signal(SIGINT, quit);

  // hang out here until all three threads have exited
  pthread_join(spThr, NULL);
  pthread_join(misThr, NULL);
  pthread_join(mobThr, NULL);

  return 0;
}

/*!
  \defgroup simpleSim
  simpleSim encompasses three threads of simulation, all referencing
  a shared world model that includes the position and orientation of
  the simulated vehicle.
*/

/*
  Modification history:

  $Log: simpleSim.cc,v $
  Revision 1.3  2006/12/12 19:20:40  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.2  2006/04/07 13:12:12  proctor
  Took out getopt() from the threads, and made NML_FILE, INI_FILE and bufnum
  globals. Apparently getopt() doesn't work from within threads.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.23  2005/09/30 18:55:51  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.22  2005/07/21 21:44:55  proctor
  Handled shutdown so that SIGINT sets flags, and main always waits at end
  with pthread_join(). This means that if all the threads exit, main exits,
  and if SIGINT is received, everything exits.

  Revision 1.21  2005/07/19 19:13:39  proctor
  Didn't error out on -n, -d args that are passed to other threads

  Revision 1.20  2005/07/19 18:56:58  proctor
  Added .ini file setting of obstacle params

  Revision 1.19  2005/07/13 19:48:21  proctor
  Fixed handling of sensorHeight, sensorPitch globals

  Revision 1.18  2005/06/29 14:32:05  proctor
  Took out annoying .nml, .ini file printing; added obstacle collisions
  in simpleMobSim.cc

  Revision 1.17  2005/06/28 20:47:39  proctor
  Added a \ref, \defgroup comment

*/
