/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file simpleSim.hh

  \brief Local decls for shared globals between simpleSim, simpleMobSim,
  simpleMisSim and simpleSPSim.

  Synopsis:
  #include "navDataExt.hh"
  #include "servoMobJA.hh"

  \code CVS Status:
  $Author: proctor $
  $Revision: 1.2 $
  $Date: 2006/04/07 13:12:12 $
  \endcode

  \author Fred Proctor
*/

#ifndef SIMPLE_SIM_HH
#define SIMPLE_SIM_HH

/*!
  SimpleWorld is the structure that holds the state of the simulated world.
  In this world, the vehicle has a position relative to the simulated
  coordinate system. This simulated coordinate system starts at 0,0,...
  and the mobility simulator in simpleMobSim.cc maps this to Moast's
  absolute (e.g., GPS-based) and relative coordinate systems. 
*/
typedef struct {
  double x;
  double y;
  double yaw;
  double v;
  double w;
  double wLeft;
  double wRight;
  //! the center-to-center spacing of the column obstacles
  double obstacleSpacing;
  //! the radius of the column obstacles
  double obstacleRadius;
  //! the height of the sensor above its base, [m]
  double sensorHeight;
  //! pitch of sensor from horizontal, negative means downward, [rad]
  double sensorPitch;
  //! non-zero means do obstacle calculations
  int doObstacles;
  void *mutexPtr;
} SimpleWorld;

extern int simpleWorldLock(SimpleWorld * world);
extern int simpleWorldUnlock(SimpleWorld * world);

//! globals values for NML, INI and buffer number for use by other threads
extern char * NML_FILE;
extern char * INI_FILE;
extern int bufnum;

//! the global structure that holds the simulated world
extern SimpleWorld simpleWorld;

//! initialize mobility simulation, e.g., open NML buffers
extern int simpleMobSimInit(int argc, char *argv[]);

//! step through one cycle of mobility simulation
extern void *simpleMobSimStep(void *);

//! quit mobility simulation, e.g., close NML buffers
extern int simpleMobSimQuit(void);

//! initialize mission simulation, e.g., open NML buffers
extern int simpleMisSimInit(int argc, char *argv[]);

//! step through one cycle of mission simulation
extern void *simpleMisSimStep(void *);

//! quit mission simulation, e.g., close NML buffers
extern int simpleMisSimQuit(void);

//! initialize sensor processing simulation, e.g., open NML buffers
extern int simpleSPSimInit(int argc, char *argv[]);

//! step through one cycle of sensor processing simulation
extern void *simpleSPSimStep(void *);

//! quit sensor processing simulation, e.g., close NML buffers
extern int simpleSPSimQuit(void);

/*
  Modification history:

  $Log: simpleSim.hh,v $
  Revision 1.2  2006/04/07 13:12:12  proctor
  Took out getopt() from the threads, and made NML_FILE, INI_FILE and bufnum
  globals. Apparently getopt() doesn't work from within threads.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.6  2005/07/19 18:56:58  proctor
  Added .ini file setting of obstacle params

  Revision 1.5  2005/07/13 19:48:21  proctor
  Fixed handling of sensorHeight, sensorPitch globals

  Revision 1.4  2005/06/28 20:47:18  proctor
  Added SIMPLE_SIM_OBSTACLE_SPACING,RADIUS

  Revision 1.3  2005/06/27 21:04:33  proctor
  Took out extra "true" transform, adding SimpleWorld structure.

  Revision 1.2  2005/06/23 17:55:22  proctor
  Added simpleSPSim

  Revision 1.1  2005/06/21 20:18:28  proctor
  Initial checkin of declarations for simpleSimMob,Mis,SP thread interface

*/

#endif /* SIMPLE_SIM_HH */
