/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I.
*****************************************************************************/

import java.io.*;
import skel.*;

//import rcs.*;
import rcs.nml.*;


/**
 *
 *  brief example of reading an NML buffer
 *
 *  @author Dr. Stephen Balakisky
 *  @author Steve Tran
 *
 *  @version 1.1   2009/01/10 04:59:52 Added java examples of nmlRead and
 *      nmlWrite. Fixed makefile to make these examples and added run scripts
 *      to run them.
 *  @version 1.2   2009/01/18 04:06:43  Another try at the java code
 *  @version 1.3   2009/01/18 13:42:49  Fixed nmlRead and nmlWrite
 *  @version 2.0   2009/01/25 20:00:00  Refactored code to follow Java Coding
 *      Standards.  Fixed indentation and spacing of code to improve
 *      readability.  Reduced the amount of logic in the main method by
 *      creating several helper methods to make the code more modular. 
 */
public class nmlRead {

    private static String DEFAULT_NML_FILE = "skel.nml";
    private static String SKEL_CMD_NAME = "skelCmd"; // NML cmd buffer name
    private static String THIS_PROCESS = "nmlRead"; // NML server process name
    private static NMLConnection skel_cmd_buf;

    /*
     *  Initialization of the NML buffers.  The code in here was provided
     *   to me.  The only modification I made was I changed its signature to
     *   return a boolean instead or 1 or 0.
     *
     *  @param config_file The specific configuration file to use
     *  @return true if successful, false if errored
     */
    private static boolean init_nml_buffers(String config_file) {

        boolean success = false;

        try {
            skel_cmd_buf = new NMLConnection(
                new skel_MsgDict(), SKEL_CMD_NAME, THIS_PROCESS, config_file );

            success = true;
        } catch(NMLException e) {
            System.out.println("Cannot open nml buffer:" + SKEL_CMD_NAME);
            System.out.println("Problem Reading");
            System.out.println("Buffer Line: " + e.buffer_line);
            System.out.println("Buffer Name:" + e.buffer_name);
            System.out.println("Config file: " + e.config_file);
            System.out.println("Internal Exception: " + e.internal_exception);
        }

        return success;
    }


    /*
     * Sets up the configurations for the application.  If the user specified
     *  their own NML file, then use that if and only if there is one parameter
     *  on the command line.  Else, use the default file.
     *
     * If the NML file does not exist, then an exit code of 1 is returned.
     *
     * @param params The command line arguments
     */
    private static void config(String[] params) {

        String NML_FILE = DEFAULT_NML_FILE;

        if (params.length == 1) {
            NML_FILE = params[0];
            System.out.println("using specified config file " + NML_FILE);
        }
        else {
            System.out.println("using default config file " + NML_FILE);
        }


        if(!new File(NML_FILE).exists()) {
            System.out.println("file " + NML_FILE + " does not exist");
            System.exit(1);
        }


        if (!init_nml_buffers(NML_FILE)) {
            System.out.println("can't allocate NML buffers");
            return;
        }
    }

    /*
     * Command information is received, and the system is prepared to execute
     *  the command. When a new command is received, the old command is
     *  immediately replaced by this command. If the command is not able to be
     *  interrupted, error flags are set and will be reported during the writng
     *  of status.
     *
     *  @throws InterruptedException When this thread gets interrupted by
     *      another thread when sleeping.
     */
    private static void process() throws InterruptedException {
        RCS_CMD_MSG msgIn;
        for(;;) {

            try {
                msgIn = (RCS_CMD_MSG) skel_cmd_buf.read();
            } catch(NMLException nml_except) {
                System.out.println("Cannot read from nml buffer.");
                return;
            }

            if( msgIn != null ) {

                switch (msgIn.type) {
                    case -1:    //  communication error
                                System.out.println("NML command read error");
                    case 0:     break;    // no new command
                    case skel.skel_MsgDict.SkelCmdGo_TYPE:
                                handleGo(msgIn);
                                break;
                    case skel.skel_MsgDict.SkelCmdShutdown_TYPE:
                                System.out.println("Exiting...");
                                return;
                }
            }

            Thread.sleep(100);
        }// end for
    }



    /*
     * Handles the Go command by performing custom actions.
     *
     * Right now, it only prints out the "where" of the
     *  command object to standard out.
     */
    private static void handleGo(RCS_CMD_MSG cmd) {
        SkelCmdGo goCmd = (SkelCmdGo) cmd;
        System.out.println("new Go command received");
        System.out.println("goCmd.where = " + goCmd.where +
			   "\n");
    }


    /**
     *  Main Driver for the Read program.  Reads in whatever's in the NML
     *   buffer and handles each type of action appropriately.  This method
     *   has been refactored so that all the logic has been modularized.
     *
     *  @param args The arguments passed in through the command line.
     */
    public static void main(String[] args) {

        System.out.println("nmlRead Java Version!\n");

        config(args);

        try {
            process();
        } catch (InterruptedException ie) {
            ie.printStackTrace();
            System.exit(2);
        }
    }


}//end class

/*
  Modification history:

  $Log: nmlRead.java,v $
  Revision 1.4  2009/02/24 19:40:31  dr_steveb
  Java code should now work!

  Revision 1.3  2009/01/18 13:42:49  dr_steveb
  Fixed nmlRead and nmlWrite

  Revision 1.2  2009/01/18 04:06:43  dr_steveb
  Another try at the java code

  Revision 1.1  2009/01/10 04:59:52  dr_steveb
  Added java examples of nmlRead and nmlWrite. Fixed makefile to make these examples and added run scripts to run them.

*/
