/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   skelShell.cc
  \brief skeleton RCS shell

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.7 $
  $Date: 2009/01/19 03:59:11 $
  \endcode

  \author Fred Proctor
  \date   03/02/2005
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <ctype.h>		// isspace()
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include "rcs.hh"
#include "skelCommon.hh"
#include "skel.hh"

static RCS_CMD_CHANNEL * skelCmdBuf= NULL;
static RCS_STAT_CHANNEL * skelStatBuf = NULL;
static RCS_CMD_CHANNEL * skelCfgBuf= NULL;
static RCS_STAT_CHANNEL * skelSetBuf = NULL;

#define DEFAULT_NML_FILE "./skel.nml"
#define SKEL_CMD_NAME "skelCmd" // NML cmd buffer name
#define SKEL_STAT_NAME "skelStat" // NLM stat buffer name
#define SKEL_CFG_NAME "skelCfg" // NML stat buffer name
#define SKEL_SET_NAME "skelSet" // NML stat buffer name
#define THIS_PROCESS "skelShell" // NML process name for the server

static int initNmlBuffers(char * configFile)
{
  if (NULL != skelCmdBuf) delete skelCmdBuf;
  skelCmdBuf = new RCS_CMD_CHANNEL(skel_format, SKEL_CMD_NAME, THIS_PROCESS, configFile);
  if (! skelCmdBuf->valid()) {
    delete skelCmdBuf;
    skelCmdBuf = 0;
    return 1;
  }

  if (NULL != skelStatBuf) delete skelStatBuf;
  skelStatBuf = new RCS_STAT_CHANNEL(skel_format, SKEL_STAT_NAME, THIS_PROCESS, configFile);
  if (! skelStatBuf->valid()) {
    delete skelStatBuf;
    skelStatBuf = 0;
    return 1;
  }

  if (NULL != skelCfgBuf) delete skelCfgBuf;
  skelCfgBuf = new RCS_CMD_CHANNEL(skel_format, SKEL_CFG_NAME, THIS_PROCESS, configFile);
  if (! skelCfgBuf->valid()) {
    delete skelCfgBuf;
    skelCfgBuf = 0;
    return 1;
  }

  if (NULL != skelSetBuf) delete skelSetBuf;
  skelSetBuf = new RCS_STAT_CHANNEL(skel_format, SKEL_SET_NAME, THIS_PROCESS, configFile);
  if (! skelSetBuf->valid()) {
    delete skelSetBuf;
    skelSetBuf = 0;
    return 1;
  }

  return 0;
}

int cleanUpNmlBuffers(void)
{
  if (NULL != skelCmdBuf) delete skelCmdBuf;
  if (NULL != skelStatBuf) delete skelStatBuf;
  if (NULL != skelCfgBuf) delete skelCfgBuf;
  if (NULL != skelSetBuf) delete skelSetBuf;

  skelCmdBuf = NULL;
  skelStatBuf = NULL;
  skelCfgBuf = NULL;
  skelSetBuf = NULL;

  return 0;
}

#define rcs_status_to_string(s) (s) == RCS_DONE ? "RCS_DONE" : (s) == RCS_EXEC ? "RCS_EXEC" : (s) == RCS_ERROR ? "RCS_ERROR" : "?"

void printHelp(void)
{
  printf( "Available commands:\n" );
  printf( "abort\n" );
  printf( "cycle <floating point arg>\n" );
  printf( "debug <floating point arg>\n" );
  printf( "go <floating point arg>\n" );
  printf( "halt\n" );
  printf( "init\n" );
  printf( "length <floating point arg>\n" );
  printf( "shutdown\n" );
  printf( "width <floating point arg>\n" );
  printf( "q or quit: exit this shell\n" );
}

void printStatus(void)
{
  SkelStat * stat = (SkelStat *) skelStatBuf->get_address();
  SkelSet * set = (SkelSet *) skelSetBuf->get_address();

  printf("STATUS\n");
  if( stat->admin_state == ADMIN_UNINITIALIZED )
    printf("Admin:              ADMIN_UNINITIALIZED\n");
  if( stat->admin_state == ADMIN_INITIALIZED )
    printf("Admin:              ADMIN_INITIALIZED\n");
  if( stat->admin_state == ADMIN_SHUT_DOWN )
    printf("Admin:              ADMIN_SHUT_DOWN\n");
  printf("command_type:       %s\n", skel_symbol_lookup(stat->command_type));
  printf("echo_serial_number: %d\n", (int) stat->echo_serial_number);
  printf("status:             %s\n", rcs_status_to_string(stat->status));
  printf("state:              %d\n", (int) stat->state);
  printf("line:               %d\n", (int) stat->line);
  printf("source_line:        %d\n", (int) stat->source_line);
  printf("source_file:        %s\n", (char *) stat->source_file);
  printf("where:              %f\n", (double) stat->where);
  printf("here:               %f\n", (double) stat->here);
  printf("Heartbeat:          %d\n", (int) stat->tt.count);
  printf("Actual cycle:       %f\n", (double) stat->tt.last);
  printf("Avg. cycle:         %f\n", (double) stat->tt.avg);

  printf("SETTINGS\n");
  printf("command_type:       %s\n", skel_symbol_lookup(set->command_type));
  printf("echo_serial_number: %d\n", (int) set->echo_serial_number);
  printf("status:             %s\n", rcs_status_to_string(set->status));
  printf("state:              %d\n", (int) set->state);
  printf("line:               %d\n", (int) set->line);
  printf("source_line:        %d\n", (int) set->source_line);
  printf("source_file:        %s\n", (char *) set->source_file);
  printf("length, req:        %f %f\n", (double) set->length, (double) set->lengthReq);
  printf("width, req:         %f %f\n", (double) set->width, (double) set->widthReq);
  printf("debug:              %d\n", set->debug);
  printf("cycleTime:          %lf\n", set->cycleTime);
}

int strwcmp(char * test, char * match)
{
  while (0 != *test && 0 != *match) {
    if (*test != *match) return 1;
    test++, match++;
  }

  if (0 == *match && (0 == *test || isspace(*test))) return 0;
  return 1;
}

int main(int argc, char *argv[])
{
  char configFileDefault[] = DEFAULT_NML_FILE;
  char * configFileEnv;
  char * NMLFile = configFileDefault;
#ifndef HAVE_READLINE_READLINE_H
  enum {BUFFERSIZE = 256};
  char buffer[BUFFERSIZE];
#endif
  char * line;
  char * ptr;
  int retval;
  SkelStat * skelStatPtr;
  SkelSet * skelSetPtr;
  SkelCmdInit skelCmdInit;
  SkelCmdGo skelCmdGo;
  SkelCmdAbort skelCmdAbort;
  SkelCmdHalt skelCmdHalt;
  SkelCmdShutdown skelCmdShutdown;
  SkelCfgLength skelCfgLength;
  SkelCfgWidth skelCfgWidth;
  SkelCfgCycleTime skelCfgCycleTime;
  SkelCfgDebug skelCfgDebug;

#ifdef HAVE_READLINE_READLINE_H
  using_history();
#endif

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (argc > 1) {
    NMLFile = argv[1];
    printf("%s: using specified config file %s\n", argv[0], NMLFile);
  } else {
    configFileEnv = getenv("CONFIG_NML");
    if (NULL != configFileEnv) {
      NMLFile = configFileEnv;
      printf("%s: using environment config file %s\n", argv[0], NMLFile);
    } else {
      printf("%s: using default config file %s\n", argv[0], NMLFile);
    }
  }

  // initial allocate all the buffers
  retval = initNmlBuffers(NMLFile);

  if (retval) {
    fprintf(stderr, "%s: can't allocate NML buffers\n", argv[0]);
    return 1;
  }

  // set up our convenient pointers to status and settings
  skelStatPtr = (SkelStat *) skelStatBuf->get_address();
  skelSetPtr = (SkelSet *) skelSetBuf->get_address();

  while (! feof(stdin)) {
#ifndef HAVE_READLINE_READLINE_H
    printf("> ");
    fflush(stdout);
    if (NULL == fgets(buffer, BUFFERSIZE, stdin)) {
      break;
    }
    buffer[BUFFERSIZE - 1] = 0;
    line = buffer;
#else
    line = readline("> ");
    if (NULL == line) {
      break;
    }
    if (*line) {
      add_history(line);
    }
#endif

    skelStatBuf->read();
    skelSetBuf->read();

    ptr = line;
    while (isspace(*ptr)) ptr++; // skip leading white space
    if (0 == *ptr) { // blank line
      printStatus();
    } else if (ptr[0] == 'q' ||	! strwcmp(ptr, "quit")) {
      break;
    } else if (! strwcmp(ptr, "init")) {
      skelCmdInit.serial_number = skelStatPtr->echo_serial_number + 1;
      skelCmdBuf->write(&skelCmdInit);
    } else if (! strwcmp(ptr, "go")) {
      if (1 == sscanf(ptr, "%*s %lf", &skelCmdGo.where)) {
	  skelCmdGo.serial_number = skelStatPtr->echo_serial_number + 1;
	  skelCmdBuf->write(&skelCmdGo);
      //	printf( "Writing goCmd serial number %d value %f\n", skelCmdGo.serial_number, skelCmdGo.where );
      } else {
	printf("'go' needs an argument\n");
      }
    } else if (! strwcmp(ptr, "halt")) {
      skelCmdHalt.serial_number = skelStatPtr->echo_serial_number + 1;
      skelCmdBuf->write(&skelCmdHalt);
    } else if (! strwcmp(ptr, "abort")) {
      skelCmdAbort.serial_number = skelStatPtr->echo_serial_number + 1;
      skelCmdBuf->write(&skelCmdAbort);
    } else if (! strwcmp(ptr, "shutdown")) {
      skelCmdShutdown.serial_number = skelStatPtr->echo_serial_number + 1;
      skelCmdBuf->write(&skelCmdShutdown);
    } else if (! strwcmp(ptr, "length")) {
      if (1 == sscanf(ptr, "%*s %lf", &skelCfgLength.length)) {
	skelCfgLength.serial_number = skelSetPtr->echo_serial_number + 1;
	skelCfgBuf->write(&skelCfgLength);
      } else {
	printf("'length' needs an argument\n");
      }
    } else if (! strwcmp(ptr, "width")) {
      if (1 == sscanf(ptr, "%*s %lf", &skelCfgWidth.width)) {
	skelCfgWidth.serial_number = skelSetPtr->echo_serial_number + 1;
	skelCfgBuf->write(&skelCfgWidth);
      } else {
	printf("'width' needs an argument\n");
      }
    } else if (! strwcmp(ptr, "debug")) {
      if (1 == sscanf(ptr, "%*s %d", &skelCfgDebug.debug)) {
	skelCfgDebug.serial_number = skelSetPtr->echo_serial_number + 1;
	skelCfgBuf->write(&skelCfgDebug);
      } else {
	printf("'debug' needs an argument\n");
      }
    } else if (! strwcmp(ptr, "cycle")) {
      if (1 == sscanf(ptr, "%*s %lf", &skelCfgCycleTime.cycleTime)) {
	skelCfgCycleTime.serial_number = skelSetPtr->echo_serial_number + 1;
	skelCfgBuf->write(&skelCfgCycleTime);
      } else {
	printf("'cycle' needs an argument\n");
      }
    } else {
      printHelp();
    }

#ifdef HAVE_READLINE_READLINE_H
    free(line);
#endif
  }

  return cleanUpNmlBuffers();
}

/*
  Modification history:

  $Log: skelShell.cc,v $
  Revision 1.7  2009/01/19 03:59:11  dr_steveb
  Fixed skelApp program so that it works!

  Revision 1.6  2008/09/19 15:22:53  dr_steveb
  Final changes to skelApp before tutorial

  Revision 1.5  2008/09/19 02:53:51  dr_steveb
  Made Skel app more readable and fixed makefile for cygwin

  Revision 1.4  2006/03/15 15:24:14  proctor
  Added RCSLIB support libraries to configure scripts

  Revision 1.3  2006/03/03 20:59:01  dr_steveb
  Added printing of available commands.

  Revision 1.2  2005/10/18 21:20:30  dr_steveb
  Added cfgCycleTime and cfgDebug.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.1  2005/10/11 18:45:03  stephen
  Initial addition of skeleton rcs controller.

*/
