/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file displayGraph.cc

  \brief Gtk+ based program for displaying planning graphs
  Borrowed Gtk+ basis code from Fred

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.2 $
  $Date: 2007/11/08 13:22:29 $
  \endcode

  \author Stephen Balakirsky
*/

/*! \defgroup Tools Tools that are provided with the MOAST Distribution

*/
#include <stdio.h>
#include <stdlib.h> // exit
#include <gtk/gtk.h>

#define DEFAULT_FILE "../etc/graph.dat"
#define WINDOW_WIDTH 400
#define WINDOW_HEIGHT 400

#define DRAWING_WIDTH 1500
#define DRAWING_HEIGHT 1500

/*
  The timeout callback will update the labels, so they need to be global.
  Only the labels need to be global, but we'll keep them all together
  so we can see what's related to what.
*/

static GtkWidget *window;
static GtkWidget *hbox;

static GtkWidget *vbox1;
static GtkWidget *frame_plot;
static GtkWidget *scrolled_window;
static GtkWidget *drawing_area;

static GdkPixmap *pixmap = NULL;

static gboolean expose_event(GtkWidget * widget, GdkEventExpose * event)
{
  gdk_draw_drawable(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		    pixmap,
		    event->area.x, event->area.y,
		    event->area.x, event->area.y,
		    event->area.width, event->area.height);

  return FALSE;
}

#define min(a,b) ((a) < (b) ? (a) : (b))

static gboolean configure_event(GtkWidget * widget, GdkEventConfigure * event)
{
  GdkPixmap *old;
  gint old_width;
  gint old_height;

  old = pixmap;

  pixmap = gdk_pixmap_new(widget->window,
			  widget->allocation.width,
			  widget->allocation.height, -1);
  gdk_draw_rectangle(pixmap,
		     widget->style->white_gc,
		     TRUE,
		     0, 0,
		     widget->allocation.width, widget->allocation.height);

  if (NULL != old) {
    gdk_drawable_get_size(old, &old_width, &old_height);
    gdk_draw_drawable(pixmap, widget->style->white_gc, old,	/* src */
		      0, 0,	/* xsrc, ysrc */
		      0, 0,	/* xdst, ydst */
		      min(old_width, widget->allocation.width),
		      min(old_height, widget->allocation.height));
    g_object_unref(old);
  }

  return TRUE;
}

static void draw_point(GtkWidget * widget, gint x, gint y)
{
  GdkRectangle update_rect;

  update_rect.x = x - 1;
  update_rect.y = y - 1;
  update_rect.width = 2;
  update_rect.height = 2;
  gdk_draw_rectangle(pixmap,
		     widget->style->bg_gc[3],
		     TRUE,
		     update_rect.x, update_rect.y,
		     update_rect.width, update_rect.height);
  gtk_widget_queue_draw_area(widget,
			     update_rect.x, update_rect.y,
			     update_rect.width, update_rect.height);
}

static void draw_line(GtkWidget * widget, gint x1, gint y1, gint x2, gint y2)
{
  GdkRectangle update_rect;

  update_rect.x = x1 - 1;
  update_rect.y = y1 - 1;
  update_rect.width = x2-x1+2;
  update_rect.height = y2-y1+2;
  gdk_draw_line(pixmap,
		widget->style->black_gc,
		x1, y1,
		x2, y2 );
  gtk_widget_queue_draw_area(widget,
			     update_rect.x, update_rect.y,
			     update_rect.width, update_rect.height);
  draw_point( widget, x1, y1 );
  draw_point( widget, x2, y2 );
}

#define LINE_LENGTH 100
static void readGraph(GtkWidget * widget, char *fileName)
{
  FILE *fp;
  char *line = new char[LINE_LENGTH];
  char type;
  int x1Loc, y1Loc, x2Loc, y2Loc;

  if( fileName == NULL )
    fp = fopen(DEFAULT_FILE, "r");
  else
    fp = fopen( fileName, "r" );
  if( fp == NULL )
    exit(-1);
  while(fgets( line, LINE_LENGTH, fp ) != NULL)
    {
      //      sscanf( line, "%c", &type );
      sscanf( line, "%c %d %d %d %d", &type, &x1Loc, &y1Loc, 
	      &x2Loc, &y2Loc );
      x1Loc *= 10;
      x2Loc *= 10;
      y1Loc *= 10;
      y2Loc *= 10;
      switch( type )
	{
	case 'P':
	  //	  sscanf( line, "%*c %d %d", &x1Loc, &y1Loc ); 
	  draw_point( widget, x1Loc, y1Loc );
	  break;
	case 'L':
	  draw_line( widget, x1Loc, y1Loc, x2Loc, y2Loc );
	  break;
	}
    } 
  if( line != NULL )
    delete line;
  fclose( fp );
}
  


static gint timeout_callback(gpointer data)
{
  readGraph( drawing_area, NULL );
  /*
  draw_point(drawing_area,
	     (DRAWING_WIDTH / 2) + (gint) (10.0 * nav->tranRel.x),
	     (DRAWING_HEIGHT / 2) + (gint) (-10.0 * nav->tranRel.y));
  */

  return FALSE;			/* FALSE means stop timer */
}

/* Another callback */
static void destroy_callback(GtkWidget * widget, gpointer data)
{
  gtk_main_quit();
}

#define BORDER 5
#define SPACING 5
#define PADDING 5

int main(int argc, char *argv[])
{
  char *fileName = NULL;

  if( argc == 2 )
    {
      fileName = argv[1];
      printf( "Opening file: %s\n", fileName );
    }
  gtk_init(&argc, &argv);

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER(window), BORDER);

  hbox = gtk_hbox_new(FALSE,	/* homogeneous child width */
		      SPACING);	/* spacing between objects */
  gtk_container_add(GTK_CONTAINER(window),	/* parent */
		    hbox);	/* child */

  vbox1 = gtk_vbox_new(TRUE, SPACING);
  gtk_box_pack_start(GTK_BOX(hbox), vbox1, TRUE, TRUE, PADDING);

  frame_plot = gtk_frame_new("Plot");
  gtk_box_pack_start(GTK_BOX(vbox1), frame_plot, TRUE, TRUE, PADDING);

  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_set_size_request(GTK_WIDGET(scrolled_window), 
			      WINDOW_WIDTH,
			      WINDOW_HEIGHT);
  gtk_container_set_border_width(GTK_CONTAINER(scrolled_window), 10);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), 
				 GTK_POLICY_ALWAYS,	// horizontal
				 GTK_POLICY_ALWAYS);	// vertical
  gtk_container_add(GTK_CONTAINER(frame_plot), scrolled_window);

  drawing_area = gtk_drawing_area_new();
  gtk_widget_set_size_request(GTK_WIDGET(drawing_area),
			      DRAWING_WIDTH,
			      DRAWING_HEIGHT);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),
					drawing_area);

  gtk_widget_show_all(window);

  pixmap = gdk_pixmap_new(drawing_area->window,
			  drawing_area->allocation.width,
			  drawing_area->allocation.height, -1);

  gdk_draw_rectangle(pixmap,
		     drawing_area->style->white_gc,
		     TRUE,
		     0, 0,
		     drawing_area->allocation.width,
		     drawing_area->allocation.height);

  /* put the scrollbars in the middle */
  GtkAdjustment * adj = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  adj->value = 0.5 * (adj->lower + adj->upper - adj->page_size);
  gtk_scrolled_window_set_hadjustment(GTK_SCROLLED_WINDOW(scrolled_window), adj);
  gtk_adjustment_value_changed(adj);
  adj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window));
  adj->value = 0.5 * (adj->lower + adj->upper - adj->page_size);
  gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(scrolled_window), adj);
  gtk_adjustment_value_changed(adj);

  /* start the timer */
  /*
  (void) g_timeout_add(100,	// millisecs
		       timeout_callback, 0);
*/
  g_signal_connect(G_OBJECT(drawing_area), "expose_event",
		   G_CALLBACK(expose_event), NULL);

  g_signal_connect(G_OBJECT(drawing_area), "configure_event",
		   G_CALLBACK(configure_event), NULL);

  g_signal_connect(G_OBJECT(window), "delete_event",
		   G_CALLBACK(destroy_callback), NULL);

  readGraph( drawing_area, fileName );
  gtk_main();

  return 0;
}

/*
  Modification history:

  $Log: displayGraph.cc,v $
  Revision 1.2  2007/11/08 13:22:29  dr_steveb
  Not quite ready, but premature checking-in of other files means that these need to go to!

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.5  2005/07/13 18:28:08  stephen
  Fixed use of fgets

  Revision 1.4  2005/07/13 18:24:52  stephen
  Removed use of getline (now using fgets) for portability

  Revision 1.3  2005/07/13 18:20:51  stephen
  Changed inclusion of getline

  Revision 1.2  2005/07/01 19:53:24  stephen
  Rotation tree now working.

  Revision 1.1  2005/06/29 19:31:23  stephen
  Added displayGraph program.

*/
