/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*! \file exampleAStarHeapNA.cc
    \brief provides a driver for testing AStarHeapN search

  \code CVS Status:
  $Author: tomrkramer $
  $Revision: 1.2 $
  $Date: 2006/11/17 15:24:24 $
  \endcode

  \author Tom Kramer

This is an aStar search routine driver for testing aStar search with
incremental updating of the heuristic for the minimum cost of reaching
the goal from a node. This routine allows searching simultaneously for
several other goals in addition to the finish (the target). The
heuristic is the Euclidean distance to the goal. The graph is built
on a square grid of points whose coordinates are integers, with one
corner at the origin, and sides on the X and Y axes. A node is built
on each point of the grid. Each node is connected to the eight surrounding
nodes (or (on the edges) to however many of them exist). The cost
of each arc is the Euclidean distance between the points underlying
the nodes it connects.

This driver also illustrates using a child class of NodeInfo for
recording the location of the point with which a node is associated.

The routines here use stdio for reading and writing.

*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "aStarHeapN.hh"

#define A_STAR_PATH_SIZE 1000
#define MAX_GOALS 100
#define MAX_CYCLES 5000

/***********************************************************************/

/*! \class NodeData
    \brief provides a child class of NodeInfo with coordinates

This class is defined as a child class of NodeInfo. This class has
the coordinates of the point on the grid occupied by the node.

*/

class NodeData : public NodeInfo
{
public:
  NodeData(){}
  NodeData(int index_in, int x_in, int y_in) : NodeInfo(index_in)
  {
    x = x_in;
    y = y_in;
  }
  ~NodeData(){}
  int x;        /*!< X coordinate of point */
  int y;        /*!< Y coordinate of point */
};

/***********************************************************************/

static int changeKidsGoalCost(AStarHeapNGraph * theGraph, int dad,
			      double finishX, double finishY);
static void getGoals(int start, int finish, int * numGoals,
		     int * goals, int size);
static void getGraphData(int argc, char * argv[], int * side,
			 AStarHeapNGraph * aGraph);
static void getStartAndFinish(int * start, int * finish, int size);
static void makeGraph(AStarHeapNGraph * aGraph, int side);
static void printAnswer(int length, int * path, double cost);

/***********************************************************************/

//! changeKidsGoalCost changes the goal costs of dad's kids
/*
Returned Value: int (0)

Called By:
  main

This sets the goalCost of each kid of dad to the Euclidean distance
from the kid to the goal.

*/

static int changeKidsGoalCost( /* ARGUMENTS                                 */
 AStarHeapNGraph * theGraph,/*!< the graph to change                        */
 int dad,                   /*!< index of node whose kids have goalCost set */
 double finishX,            /*!< X coordinate of goal                       */
 double finishY)            /*!< Y coordinate of goal                       */
{
  int n;
  int numberChanged;
  double * newGoalCosts;
  NodeData * data;
  AStarHeapNNode * nodes;

  nodes = theGraph->nodes;
  numberChanged = nodes[dad].kidsSize;
  newGoalCosts = new double[numberChanged];
  for (n = 0; n < numberChanged; n++)
    {
      data = (NodeData *)(nodes[nodes[dad].kidsIndexes[n]].moreInfo);
      newGoalCosts[n] = hypot((data->x - finishX), (data->y - finishY));
    }
  theGraph->changeKidsGoalCosts(dad, numberChanged, newGoalCosts);
  return 0;
}

/***********************************************************************/

//! getGoals records goals entered by the user
/*
This records the finish as goals[0]. Then, it repeatedly asks the user
to enter the index of a goal or to enter s to search. It records the
goals given by the user and it sets numGoals. numGoals includes the
finish, so that, for example, if the user gives 2 goals, numGoals will
be 3.

This prevents the user from entering the same goal twice, or putting
a goal at the finish, or setting a goal out of range.

*/

static void getGoals( /* ARGUMENTS                        */
 int start,           /*!< index in nodes of start node   */
 int finish,          /*!< index in nodes of finish node  */
 int * numGoals,      /*!< number of goal nodes, set here */
 int * goals,         /*!< array of goals, filled here    */
 int size)            /*!< number of nodes in the graph   */
{
  int n;
  char buffer[128];

  goals[0] = finish;
  *numGoals = 1;
  for ( *numGoals = 1; *numGoals < MAX_GOALS; )
    {
      printf("enter goal index, or s to search => ");
      fflush(stdout);
      fgets(buffer, 128, stdin);
      if (buffer[0] == 's')
	break;
      else if (sscanf(buffer, "%d", (goals + *numGoals)) != 1)
	continue;
      else if ((goals[*numGoals] < 0) || (goals[*numGoals] >= size))
	fprintf(stderr, "goal out of range\n");
      else if (goals[*numGoals] == start)
	fprintf(stderr, "start must not be goal\n");
      else
	{
	  for (n = 0; n < *numGoals; n++)
	    {
	      if (goals[n] == goals[*numGoals])
		{
		  printf("cannot use same goal twice\n");
		  break;
		}
	    }
	  if (n == *numGoals) // everything OK
	    (*numGoals)++;
	}
    }
}

/***********************************************************************/

//! getGraphData reads the size provided by the user and makes a graph
/*
Returned Value: none

Called By:  main

This reads the number of grid squares (along the side of a square
grid) given in the command arguments. Then it calls makeGraph to
make a graph whose nodes correspond to the grid squares.

If the number of command arguments is not 1, this prints an error
message and exits.

*/

static void getGraphData(    /* ARGUMENTS                          */
 int argc,                   /*!< number of command line arguments */
 char * argv[],              /*!< array of command line arguments  */
 int * side,                 /*!< size of side of grid, set here   */
 AStarHeapNGraph * theGraph) /*!< graph to build, built here       */
{
  if (argc == 2)
    {
      sscanf(argv[1], "%d", side);
      makeGraph(theGraph, *side);
    }
  else
    {
      fprintf(stderr, "Usage: %s <length of side>\n", argv[0]);
      exit(1);
    }
}

/***********************************************************************/

/*! getStartAndFinish gets the indexes of the start and finish nodes
  from the user and allows the user to quit. */
/*
Returned Value: none

Called by: main

This asks the user to give the indexes of the start and finish nodes or
to enter q to quit. If the user enters node indexes, *start and *finish
are set. This checks that the values are in range and the start and
finish differ.

*/

static void getStartAndFinish( /* ARGUMENTS                        */
 int * start,                  /*!< index of start node, set here  */
 int * finish,                 /*!< index of finish node, set here */
 int size)                     /*!< number of nodes in graph       */
{
  char buffer[128];

  *start = -1;
  *finish = -1;
  while (1)
    {
      printf("Enter start and finish or q to quit ==> ");
      fflush(stdout);
      fgets(buffer, 128, stdin);
      if (buffer[0] == 'q')
	exit(0);
      sscanf(buffer, "%d %d", start, finish);
      if ((*start < 0) || (*start >= size))
	printf("start out of range\n");
      else if ((*finish < 0) || (*finish >= size))
	printf("finish out of range\n");
      else if (*start == *finish)
	fprintf(stderr, "start must differ from finish\n");
      else // everything OK
	break;
    }
}

/***********************************************************************/

//! main runs the driver
/*
This:
1. calls getGraphData to set the size of a side of the (square) grid
   and make the graph.
2. calls getStartAndFinish to ask the user to pick start and finish nodes
   or to quit.
3. Sets finishX and finishY.
4. calls getGoals to ask the user to pick goal nodes.
5. initializes the searcher.
6. runs a loop in which:
   A. By calling getNextNode, the next node to be opened is identified
      (call it Noe).
   B. By calling changeKidsGoalCost, new goalCosts are inserted for the
      kids of Noe.
   C. CycleSearch is called.
   The loop runs until any one of the following occurs:
   A. It is determined there is no path to the finish.
   B. The cheapest path to the finish has been found and MAX_CYCLES have run.
   C. The cheapest paths to the finish and all the goals have been found.
7. prints the path and its cost, if a path from start to finish has been
   found, and prints the number of cycles (calls to cycleSearch) used
   during the search.
8. prints the path and its cost to each of the goals to which a path
   was found, and prints whether the cost is known to be optimal.
9. closes the searcher.
10. goes back to step 2.

*/

int main(       /* ARGUMENTS                                     */
 int argc,      /*!< one more than number of arguments           */
 char * argv[]) /*!< array of executable name and argument names */
{	
  int start;
  int finish;
  double finishX;
  double finishY;
  int side;
  AStarHeapNGraph theGraph;
  AStarHeapNSearcher theSearcher;
  int numGoals;
  int goals[MAX_GOALS];
  int opens;
  int statuses[MAX_GOALS];
  double costs[MAX_GOALS];
  int goalsFound;
  int optimaFound;
  int path[A_STAR_PATH_SIZE];
  int length;
  int cycles;
  int n;
  
  getGraphData(argc, argv, &side, &theGraph);
  for (;;)
    {
      cycles = 0;
      getStartAndFinish(&start, &finish, theGraph.size);
      theGraph.nodes[start].goalCost = 0; // used when start node processed
      getGoals(start, finish, &numGoals, goals, theGraph.size);
      finishX = (double)(finish % side);
      finishY = (double)(finish / side);
      if (theSearcher.initSearch(&theGraph, start, numGoals, goals, &opens,
				 statuses, costs, &goalsFound, &optimaFound))
	return 1; // initialization error occurred
      while (opens &&
	     ((statuses[0] != 2) || (cycles < MAX_CYCLES)) &&
	     (optimaFound < numGoals))
	{
	  n = theSearcher.getNextNode();
	  changeKidsGoalCost(&theGraph, n, finishX, finishY);
	  theSearcher.cycleSearch();
	  cycles++;
	}
      if (statuses[0] == 2)
	{
	  length = theSearcher.reportSearch(goals[0], A_STAR_PATH_SIZE, path);
	  printf("Path from start to finish:");
	  printAnswer(length, path, costs[0]);
	  printf("%d cycles\n", cycles);
	}
      else
	printf("No path found from start to finish\n");
      for (n = 1; n < numGoals; n++)
	{
	  if (statuses[n] > 0)
	    {
	      length = theSearcher.reportSearch(goals[n],
						A_STAR_PATH_SIZE, path);
	      printf("%s", ((statuses[n] == 2) ? "Optimal path " : "Path "));
	      printf("from start to node %d:", goals[n]);
	      printAnswer(length, path, costs[n]);
	    }
	}
      theSearcher.closeSearch();
    }
  return 0;
}

/***********************************************************************/

//! makeGraph makes an AStarHeapNGraph on a square grid
/*
Returned Value: none

Called by:  getGraphData

This makes an AStarHeapNGraph on a square grid. For purposes of assigning
neighbors, the square is made into a torus by wrapping both ways.
Where a neighbor is on the other side of the graph, the cost is -1.

The moreInfo of each node is an instance of the NodeData class defined
above and contains the x and y coordinates of the grid square occupied
by the node.

The kids of node O are indexed as shown below.

   3   2   1
    \  |  /
     \ | /
 4 --- O --- 0
     / | \
    /  |  \
   5   6   7

Where a kid is on the other side of the grid, the arc cost is -1.0.
Otherwise, the cost of arcs to kids 0, 2, 4, and 6 is 1.0 and the
cost of arcs to kids 1, 3, 5, and 7 is sqrt(2.0).

*/

static void makeGraph(     /* ARGUMENTS          */
 AStarHeapNGraph * aGraph, /*!< graph to build   */
 int side)                 /*!< length of a side */
{ 
  int n;
  AStarHeapNNode * nodes;
  int length;
  int x;
  int y;
  int xRt;
  int yUp;
  int xLt;
  int yDn;
  
  length = (side * side);
  aGraph->capacity = length;
  aGraph->size = length;
  nodes = new AStarHeapNNode[length];
  aGraph->nodes = nodes;
  for (n = 0; n < length; n++)
    {
      x = (n % side);
      y = (n / side);
      nodes[n].moreInfo = new NodeData(n, x, y);
      nodes[n].kidsSize = 8;
      nodes[n].kidsIndexes = new int[8];
      nodes[n].kidsCosts = new double[8];
      xRt = (((x + 1) == side) ? 0 : (x + 1));
      xLt = (((x - 1) < 0) ? (side - 1) : (x - 1));
      yUp = (((y + 1) == side) ? 0 : (y + 1));
      yDn = (((y - 1) < 0) ? (side - 1) : (y - 1));
      nodes[n].kidsIndexes[0] = ((y * side) + xRt);
      nodes[n].kidsIndexes[1] = ((yUp * side) + xRt);
      nodes[n].kidsIndexes[2] = ((yUp * side) + x);
      nodes[n].kidsIndexes[3] = ((yUp * side) + xLt);
      nodes[n].kidsIndexes[4] = ((y * side) + xLt);
      nodes[n].kidsIndexes[5] = ((yDn * side) + xLt);
      nodes[n].kidsIndexes[6] = ((yDn * side) + x);
      nodes[n].kidsIndexes[7] = ((yDn * side) + xRt);
      nodes[n].kidsCosts[0] = ((xRt > x) ? 1.0 : -1.0);
      nodes[n].kidsCosts[1] = (((yUp > y) && (xRt > x)) ? sqrt(2.0) : -1.0);
      nodes[n].kidsCosts[2] = ((yUp > y) ? 1.0 : -1.0);
      nodes[n].kidsCosts[3] = (((yUp > y) && (xLt < x)) ? sqrt(2.0) : -1.0);
      nodes[n].kidsCosts[4] = ((xLt < x) ? 1.0 : -1.0);
      nodes[n].kidsCosts[5] = (((yDn < y) && (xLt < x)) ? sqrt(2.0) : -1.0);
      nodes[n].kidsCosts[6] = ((yDn < y) ? 1.0 : -1.0);
      nodes[n].kidsCosts[7] = (((yDn < y) && (xRt > x)) ? sqrt(2.0) : -1.0);
    }
}

/***********************************************************************/

//! printAnswer prints a search path and its cost
/*
Returned Value: none

Called By:  main

This prints a search path and its cost.

*/

static void printAnswer( /* ARGUMENTS                                */
 int length,             /*!< length of the path                     */
 int * path,             /*!< array of index numbers giving the path */
 double cost)            /*!< cost of the path                       */
{
  int n;
  
  for (n = 0; n < length; n++)
    {
      printf(" %d", path[n]);
    }
  printf("\n");
  printf("With cost: %.4f\n\n", cost);
}

/***********************************************************************/

