/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   joystick.cc
  \brief Provides a joystick controller that interfaces to servo

  Interfaces to servo with the ll command. Provides automatic
  decelleration.

  \code CVS Status:
  $Author: cj_scrapper $
  $Revision: 1.12 $
  $Date: 2007/01/31 18:36:12 $
  \endcode

  \author Stephen Balakirsky
  \date   07/28/2005
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h> // getenv
#include <rcs.hh>
//#include <fcntl.h> // fcntl call
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "servoMobJA.hh"
#include "servoSP.hh"
#include <signal.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
// #include <ncurses.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
//#include </sys/time.h>


#define MY_NAME_BASE "joystick"

#define OUTER_WINDOW_MIN_WIDTH (80)
#define OUTER_WINDOW_MIN_HEIGHT (80)
#define OUTER_WINDOW_DEF_WIDTH (OUTER_WINDOW_MIN_WIDTH + 100)
#define OUTER_WINDOW_DEF_HEIGHT (OUTER_WINDOW_MIN_HEIGHT + 100)
#define OUTER_WINDOW_DEF_X 100
#define OUTER_WINDOW_DEF_Y 100

/////////////////////////////////////////////
///////////////  classes  ///////////////////
/////////////////////////////////////////////


/////////////////////////////////////////////
//////////////   globals   //////////////////
/////////////////////////////////////////////
//! Serial number to servo level
int myServoSerial;
// NML Channels
//! servo Mob channels
RCS_CMD_CHANNEL   *servoMobJACmdNml;
RCS_STAT_CHANNEL  *servoMobJAStatNml;
RCS_CMD_CHANNEL   *servoMobJACfgNml;
RCS_STAT_CHANNEL  *servoMobJASetNml;
RCS_CMD_CHANNEL   *servoSPCmdNml;

ServoMobJACmdSkid   servoSkid;
ServoMobJACmdAckerman ackerMsg;
ServoMobJACmdSubmarine submarineMsg;

//! X Stuff
Display *dpy;
int screen;

//! control keys
bool cmdChanged;
bool leftAccPressed;
bool leftDecPressed;
bool rightAccPressed;
bool rightDecPressed;
bool accPressed;
bool decPressed;
bool upPressed;
bool downPressed;
bool leftTurnPressed;
bool rightTurnPressed;
int mode;

double accInc;
double decInc;
double maxSpeed;
double minSpeed;
bool myVerbose;

/* code for costing
	  if(  0 )
	    {
	      beenReleased = true;
	      // drift to zero
	      servoSkid.wLeft *= 1-accInc;
	      servoSkid.wRight *= 1-accInc;
	      if( fabs(servoSkid.wLeft) < 1 )
		servoSkid.wLeft = 0;
	      if( fabs(servoSkid.wRight) < 1 )
		servoSkid.wRight = 0;
	      servoMobJACmdNml->write( servoSkid );
	      continue;
	    }
*/
#define LEFTACC "r"
#define LEFTDEC "f"
#define RIGHTACC "Up"
#define RIGHTDEC "Down"

void do_KeyPress (XEvent *eventp)
{
    XKeyEvent *e = (XKeyEvent *) eventp;
    KeySym ks;
    char *ksname;
    int nbytes;
    char str[256+1];

    nbytes = XLookupString (e, str, 256, &ks, NULL);
    if (ks == NoSymbol)
	ksname = "NoSymbol";
    else if (!(ksname = XKeysymToString (ks)))
	ksname = "(no name)";
    // switch on mode
    switch( mode )
      {
      case 0: // left/right wheel drive
	if( !strcmp( ksname, LEFTACC ))
	  leftAccPressed = true;
	else if( !strcmp( ksname, LEFTDEC ))
	  leftDecPressed = true;
	else if( !strcmp( ksname, RIGHTACC ))
	  rightAccPressed = true;
	else if( !strcmp( ksname, RIGHTDEC ))
	  rightDecPressed = true;
	else
	  printf( "Key: %s\n", ksname );
	break;
      case 1:
      case 2:
	if( !strcmp( ksname, "Up" ) )
	  accPressed = true;
	else if( !strcmp(ksname, "Down") )
	  decPressed = true;
	if( !strcmp(ksname, "Left") )
	  leftTurnPressed = true;
	else if( !strcmp(ksname, "Right") )
	  rightTurnPressed = true;
	else if( !strcmp(ksname, "s") )
	  ackerMsg.steerAngle = 0;
	else if( !strcmp(ksname, "S") )
	  ackerMsg.velocity = 0;
	break;
      case 3: // 3-D Driving
	if( !strcmp( ksname, "Up" ) )
	  upPressed = true;
	else if( !strcmp(ksname, "Down") )
	  downPressed = true;
	if( !strcmp(ksname, "Left") )
	  leftTurnPressed = true;
	else if( !strcmp(ksname, "Right") )
	  rightTurnPressed = true;
	else if( !strcmp(ksname, "s") )
	  {
	    submarineMsg.rudder = 0;
	    submarineMsg.sternPlane = 0;
	  }
	else if( !strcmp(ksname, "S") )
	  submarineMsg.propeller = 0;
	else if( !strcmp(ksname, "a") )
	  accPressed = true;
	else if( !strcmp(ksname, "d") )
	  decPressed = true;
	break;
      default:
	printf( "unknown mode %d\n", mode );
	break;
      }
    return;
}

void do_KeyRelease (XEvent *eventp)
{
    XKeyEvent *e = (XKeyEvent *) eventp;
    KeySym ks;
    char *ksname;
    int nbytes;
    char str[256+1];

    nbytes = XLookupString (e, str, 256, &ks, NULL);
    if (ks == NoSymbol)
	ksname = "NoSymbol";
    else if (!(ksname = XKeysymToString (ks)))
	ksname = "(no name)";

    switch( mode )
      {
      case 0:
	if( !strcmp( ksname, LEFTACC ))
	  leftAccPressed = false;
	else if( !strcmp( ksname, LEFTDEC ))
	  leftDecPressed = false;
	else if( !strcmp( ksname, RIGHTACC ))
	  rightAccPressed = false;
	else if( !strcmp( ksname, RIGHTDEC ))
	  rightDecPressed = false;
	else
	  printf( "Key: %s\n", ksname );
	break;
      case 1:
      case 2:
	if( !strcmp( ksname, "Up" ) )
	  accPressed = false;
	else if( !strcmp(ksname, "Down") )
	  decPressed = false;
	if( !strcmp(ksname, "Left") )
	  leftTurnPressed = false;
	else if( !strcmp(ksname, "Right") )
	  rightTurnPressed = false;
	break;
      case 3: // 3-D Driving
	if( !strcmp( ksname, "Up" ) )
	  upPressed = false;
	else if( !strcmp(ksname, "Down") )
	  downPressed = false;
	if( !strcmp(ksname, "Left") )
	  leftTurnPressed = false;
	else if( !strcmp(ksname, "Right") )
	  rightTurnPressed = false;
	else if( !strcmp(ksname, "a") )
	  accPressed = false;
	else if( !strcmp(ksname, "d") )
	  decPressed = false;
	break;
      default:
	printf( "unknown mode %d\n", mode );
	break;
      }
    return;
}


/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool checkServoCmdStatus( RCS_STATUS &stat, int &echo_serial_number )
{
  int type;

  type = servoMobJAStatNml->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case SERVO_MOB_JA_STAT_TYPE:
      stat = servoMobJAStatNml->get_address()->status;
      echo_serial_number = servoMobJAStatNml->get_address()->echo_serial_number;
      return(true);
    default:
      fprintf( stderr, "Unknown cmd type received: %s %d\n", servoMobJA_symbol_lookup(type), type );
      break;
    }
  return false;
}

/*!
  Open all of the NML channels
  
  \param vehID The number to add to all of the channel names
  \return true on success
  \return false on failure
*/
bool openNML( int vehID )
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];

  sprintf( myName, "%s%c", MY_NAME_BASE,(char)NULL );
  
  if(!configNml)
    {
      fprintf(stderr,"Please set CONFIG_NML environment variable.\n");
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, vehID );
  servoMobJACmdNml = new RCS_CMD_CHANNEL( servoMobJA_format, chanName,
					  myName, configNml );
  if( NULL == servoMobJACmdNml || !servoMobJACmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, vehID );
  servoMobJAStatNml = new RCS_STAT_CHANNEL( servoMobJA_format, chanName,
					    myName, configNml );
  if( NULL == servoMobJAStatNml || !servoMobJAStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_SET_NAME, vehID );
  servoMobJASetNml = new RCS_STAT_CHANNEL( servoMobJA_format, chanName,
					   myName, configNml );
  if( NULL == servoMobJASetNml || !servoMobJASetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_MOB_JA_CFG_NAME, vehID );
  servoMobJACfgNml = new RCS_CMD_CHANNEL( servoMobJA_format, chanName,
					  myName, configNml );
  if( NULL == servoMobJACfgNml || !servoMobJACfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_SP_CMD_NAME, vehID );
  servoSPCmdNml = new RCS_CMD_CHANNEL( servoSP_format, chanName,
					  myName, configNml );
  if( NULL == servoSPCmdNml || !servoSPCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }
  return(true);
}

/*!
  free all memory and prepare to exit
*/
void closeAll()
{
  ServoMobJACmdSkid skidDrive;
  ServoMobJACmdAckerman ackermanDrive;
  ServoMobJACmdSubmarine nauticDrive;

  skidDrive.wLeft = 0;
  skidDrive.wRight = 0;
  ackermanDrive.velocity = 0;
  ackermanDrive.crabAngle = 0;
  ackermanDrive.steerAngle = 0;
  nauticDrive.propeller = 0;
  nauticDrive.rudder = 0;
  nauticDrive.sternPlane = 0;

  switch( mode )
    {
    case 0:
    case 1:
      skidDrive.serial_number = ++myServoSerial;
      servoMobJACmdNml->write(skidDrive);
      break;
    case 2:
      ackermanDrive.serial_number = ++myServoSerial;
      servoMobJACmdNml->write(ackermanDrive);
      break;
    case 3:
      nauticDrive.serial_number = ++myServoSerial;
      servoMobJACmdNml->write(nauticDrive);
      break;
    }
  return;

  // deletes seem to cause seg fault?
  if( servoMobJACmdNml != NULL )
    delete servoMobJACmdNml;
  servoMobJACmdNml = NULL;
  if( servoMobJAStatNml != NULL )
    delete servoMobJAStatNml;
  servoMobJAStatNml = NULL;
  if( servoMobJASetNml != NULL )
    delete servoMobJASetNml;
  servoMobJASetNml = NULL;
  if( servoMobJACfgNml != NULL )
    delete servoMobJACfgNml;
  servoMobJACfgNml = NULL;
  if( servoSPCmdNml != NULL )
    delete servoSPCmdNml;
  servoSPCmdNml = NULL;
}


/*!
  Signal handler
  \param signum Signal that has been caught
*/
void sigHandler( int signum )
{
  switch( signum )
    {
    case SIGINT:
      fprintf( stderr, "Shutting down joystick program\n" );
      //      closeAll();
      exit(1);
      break;
    case SIGALRM:
      switch( mode )
	{
	case 0:
	  if( leftAccPressed )
	    {
	      if( servoSkid.wLeft == 0 )
		{
		  servoSkid.wLeft = minSpeed;
		}
	      else if( servoSkid.wLeft > 0 )
		{
		  servoSkid.wLeft *= (1+accInc );
		  if( servoSkid.wLeft > maxSpeed )
		    servoSkid.wLeft = maxSpeed;
		}
	      else
		{
		  servoSkid.wLeft *= (1-decInc );
		  if( servoSkid.wLeft > -minSpeed )
		    servoSkid.wLeft = 0;
		}
	    }
	  else if( leftDecPressed )
	    {
	      if( servoSkid.wLeft == 0 )
		servoSkid.wLeft = -minSpeed;
	      else if( servoSkid.wLeft < 0 )
		{
		  servoSkid.wLeft *= (1+accInc );
		  if( servoSkid.wLeft < -maxSpeed )
		    servoSkid.wLeft = -maxSpeed;
		}
	      else
		{
		  servoSkid.wLeft *= (1-decInc );
		  if( servoSkid.wLeft < minSpeed )
		    servoSkid.wLeft = 0;
		}
	    }
	  else
	    {
	      servoSkid.wLeft *= 1-accInc;
	      if( fabs(servoSkid.wLeft) < minSpeed )
		servoSkid.wLeft = 0;
	    }

	  if( rightAccPressed)
	    {
	      if( servoSkid.wRight == 0 )
		{
		  servoSkid.wRight = minSpeed;
		}
	      else if( servoSkid.wRight > 0 )
		{
		  servoSkid.wRight *= (1+accInc);
		  if( servoSkid.wRight > maxSpeed )
		    servoSkid.wRight = maxSpeed;
		}
	      else
		{
		  servoSkid.wRight *= (1-decInc);
		  if( servoSkid.wRight > -minSpeed )
		    servoSkid.wRight = 0;
		}
	    }
	  else if( rightDecPressed )
	    {
	      if( servoSkid.wRight == 0 )
		{
		  servoSkid.wRight = -minSpeed;
		}
	      else if( servoSkid.wRight < 0 )
		{
		  servoSkid.wRight *= (1+accInc);
		  if( servoSkid.wRight < -maxSpeed )
		    servoSkid.wRight = -maxSpeed;
		}
	      else
		{
		  servoSkid.wRight *= (1-decInc);
		  if( servoSkid.wRight < minSpeed )
		    servoSkid.wRight = 0;
		}
	    }
	  else
	    {
	      servoSkid.wRight *= 1-accInc;
	      if( fabs(servoSkid.wRight) < minSpeed )
		servoSkid.wRight = 0;
	    }
	  break;
	case 1:
	  if(accPressed)
	    {
	      servoSkid.wLeft  += accInc;
	      servoSkid.wRight += accInc;
	    }
	  if(decPressed)
	    {
	      servoSkid.wLeft  -= accInc;
	      servoSkid.wRight -= accInc;
	    }
	  if(leftTurnPressed)
	    {
	      servoSkid.wLeft  -= accInc;
	      servoSkid.wRight += accInc;
	    }
	  if(rightTurnPressed)
	    {
	      servoSkid.wLeft  += accInc;
	      servoSkid.wRight -= accInc;
	    }
	  break;
	case 2:
	  if(accPressed)
	    {
	      ackerMsg.velocity += accInc;
	    }
	  if(decPressed)
	    {
	      ackerMsg.velocity -= accInc;
	    }
	  if(leftTurnPressed)
	    {
	      ackerMsg.steerAngle += accInc *.1;
	    }
	  if(rightTurnPressed)
	    {
	      ackerMsg.steerAngle -= accInc* .1;
	    }
	  break;
	case 3:
	  if(accPressed)
	    {
	      submarineMsg.propeller += accInc;
	    }
	  if(decPressed)
	    {
	      submarineMsg.propeller -= accInc;
	    }
	  if(leftTurnPressed)
	    {
	      submarineMsg.rudder += accInc *.1;
	    }
	  if(rightTurnPressed)
	    {
	      submarineMsg.rudder -= accInc* .1;
	    }
	  if(upPressed)
	    {
	      submarineMsg.sternPlane += accInc *.1;
	    }
	  if(downPressed)
	    {
	      submarineMsg.sternPlane -= accInc *.1;
	    }
	  break;

	default:
	  break;
	}
      if( myVerbose )
	{
	  switch(mode)
	    {
	    case 0:
	    case 1:
	      fprintf(stderr, "Left: %f Right: %f:                        \r", 
		      servoSkid.wLeft, servoSkid.wRight );
	      break;
	    case 2:
	      fprintf(stderr, "Vel: %f Steer: %f                         \r",
		      ackerMsg.velocity, ackerMsg.steerAngle );
	      break;
	    case 3:
	      fprintf(stderr, "Prop: %f Rudder: %f Stern: %f           \r",
		      submarineMsg.propeller, submarineMsg.rudder, submarineMsg.sternPlane );
	      break;
	    }
	  fflush(stderr);
	}
      switch(mode)
	{
	case 0:
	case 1:
	  /*
	  if( accPressed || decPressed || leftTurnPressed || rightTurnPressed ||
	      leftAccPressed || leftDecPressed || rightAccPressed || rightDecPressed )
	    {
	  */
	      servoSkid.serial_number++;
	      servoMobJACmdNml->write( servoSkid );
	      //	    }
	  break;
	case 2:
	  /*
	  if( accPressed || decPressed || leftTurnPressed || rightTurnPressed )
	    {
	  */
	      ackerMsg.serial_number++;
	      servoMobJACmdNml->write( ackerMsg );
	      //	    }
	  break;
	case 3:
	  /*
	  if( accPressed || decPressed || leftTurnPressed || rightTurnPressed || upPressed || downPressed )
	    {
	  */
	      submarineMsg.serial_number++;
	      servoMobJACmdNml->write( submarineMsg );
	      //	    }
	}
      break;
    default:
      fprintf( stderr, "Unknown signal caught\n" );
      break;
    }
}

void set_sizehints (XSizeHints *hintp, int min_width, int min_height,
	       int defwidth, int defheight, int defx, int defy, char *geom)
{
    int geom_result;

    /* set the size hints, algorithm from xlib xbiff */

    hintp->width = hintp->min_width = min_width;
    hintp->height = hintp->min_height = min_height;
    hintp->flags = PMinSize;
    hintp->x = hintp->y = 0;
    geom_result = NoValue;
    if (geom != NULL) {
        geom_result = XParseGeometry (geom, &hintp->x, &hintp->y,
				      (unsigned int *)&hintp->width,
				      (unsigned int *)&hintp->height);
	if ((geom_result & WidthValue) && (geom_result & HeightValue)) {
#define max(a,b) ((a) > (b) ? (a) : (b))
	    hintp->width = max (hintp->width, hintp->min_width);
	    hintp->height = max (hintp->height, hintp->min_height);
	    hintp->flags |= USSize;
	}
	if ((geom_result & XValue) && (geom_result & YValue)) {
	    hintp->flags += USPosition;
	}
    }
    if (!(hintp->flags & USSize)) {
	hintp->width = defwidth;
	hintp->height = defheight;
	hintp->flags |= PSize;
    }
/*
    if (!(hintp->flags & USPosition)) {
	hintp->x = defx;
	hintp->y = defy;
	hintp->flags |= PPosition;
    }
 */
    if (geom_result & XNegative) {
	hintp->x = DisplayWidth (dpy, DefaultScreen (dpy)) + hintp->x -
		    hintp->width;
    }
    if (geom_result & YNegative) {
	hintp->y = DisplayHeight (dpy, DefaultScreen (dpy)) + hintp->y -
		    hintp->height;
    }
    return;
}

int main(int argc, char **argv)
{
  int done;
  char *name = "Joy Stick";
  int borderwidth = 2;
  unsigned long back, fore;
  unsigned long mask = 0L;
  char *displayname = NULL;
  char *geom = NULL;
  Window w = 0;
  XSizeHints hints;
  XSetWindowAttributes attr;
  XWindowAttributes wattr;
  bool skipInit;

  struct itimerval timerVal;

  RCS_STAT_MSG *status = new RCS_STAT_MSG( 100, sizeof(RCS_STAT_MSG) );
  RCS_TIMER *timer;
  float cycleTime = .25;
  char c;
  ServoMobJACmdInit servoInit;
  ServoSPCmdGo servoGo;
  int echo_serial_number;
  RCS_STATUS servoStatus;
  bool isServoInitialized=false;

  //  int joyInput, charIn;
  mode     = 0;
  myVerbose = false;
  maxSpeed = 4;
  skipInit = false;
  
  int vehID = 1;
  myServoSerial = 18;
  while((c=getopt(argc,argv,"c:d:m:s:vn"))!=EOF) 
    {
      switch( c )
	{
	case 'c':
	  // set cycle time
	  cycleTime = atof(optarg);
	  break;
	case 'd':
	  // set vehicle id
	  vehID = atoi(optarg);
	  break;
	case 'm':
	  // set operational mode
	  mode = atoi(optarg);
	  break;
	case 'n':
	  // skip initialization
	  skipInit = true;
	  break;
	case 's':
	  // set max spped
	  maxSpeed = atof(optarg);
	  break;
	case 'v':
	  // turn on verbose printouts
	  myVerbose = true;
	  break;
	}
    }

  // set up parameters
  switch( mode )
    {
    case 0:
      accInc   = 0.05;
      decInc   = 0.04;
      //      maxSpeed = 4.0;
      minSpeed = 0.1;
      break;
    case 1:
      accInc   = 0.01;
      decInc   = 0.01;
      //      maxSpeed = 4.0;
      minSpeed = 0.1;
      break;
    case 2:
      accInc = 0.01;
      decInc = 0.01;
      break;
    case 3:
      accInc = 0.01;
      decInc = 0.01;
      break;
    }

  // initialize run
  signal( SIGINT, sigHandler );
  signal( SIGALRM, sigHandler);
  status->state = NEW_COMMAND;

  timer = new RCS_TIMER(cycleTime);
  for(;;)
    {
      timer->wait();
      if( state_match(status, NEW_COMMAND) ) // NEW_COMMAND: open nml buffers
	{
	  if( myVerbose )
	    fprintf(stderr, "In state NEW_COMMAND:\r" );
	  state_new(status);
	  status_next(status, RCS_EXEC);
	  if(!openNML( vehID ))
	    {
	      fprintf( stderr, "Error opening nml buffers\n" );
	      return(-1);
	    }
	  // set up X Stuff
	  dpy = XOpenDisplay (displayname);
	  if (!dpy) 
	    {
	      fprintf (stderr, "%s:  unable to open display '%s'\n",
		       argv[0], XDisplayName (displayname));
	      exit (1);
	    }

	  screen = DefaultScreen (dpy);

	  /* select for events */
	  attr.event_mask = KeyPressMask | KeyReleaseMask;

	  if (w) 
	    {
	      XGetWindowAttributes(dpy, w, &wattr);
	      if (wattr.all_event_masks & ButtonPressMask)
		attr.event_mask &= ~ButtonPressMask;
	      attr.event_mask &= ~SubstructureRedirectMask;
	      XSelectInput(dpy, w, attr.event_mask);
	    }
	  else 
	    {
	      set_sizehints (&hints, OUTER_WINDOW_MIN_WIDTH, 
			     OUTER_WINDOW_MIN_HEIGHT,
			     OUTER_WINDOW_DEF_WIDTH, OUTER_WINDOW_DEF_HEIGHT, 
			     OUTER_WINDOW_DEF_X, OUTER_WINDOW_DEF_Y, geom);
	      back = BlackPixel(dpy,screen);
	      fore = WhitePixel(dpy,screen);
	      attr.background_pixel = back;
	      attr.border_pixel = fore;
	      mask |= (CWBackPixel | CWBorderPixel | CWEventMask);

	      w = XCreateWindow (dpy, RootWindow (dpy, screen), hints.x, 
				 hints.y,
				 hints.width, hints.height, borderwidth, 0,
				 InputOutput, (Visual *)CopyFromParent,
				 mask, &attr);

	      XSetStandardProperties (dpy, w, name, NULL, (Pixmap) 0,
				      argv, argc, &hints);
	      XMapWindow (dpy, w);

	      if( !skipInit )
		state_next(status, S0);
	      else
		{
		  state_next(status, S2);
		  servoSkid.wLeft = 0;
		  servoSkid.wRight = 0;
		  ackerMsg.velocity = 0;
		  ackerMsg.steerAngle = 0;
		  ackerMsg.crabAngle = 0;
		  submarineMsg.propeller = 0;
		  submarineMsg.rudder = 0;
		  submarineMsg.sternPlane = 0;
		  isServoInitialized = true;
		}
	    }
	}
      else if( state_match(status, S0 ) ) // S0: Initialization: send init to servo
	{
	  if( myVerbose )
	    fprintf(stderr, "In state S0:\r" );
	  if( !isServoInitialized )
	    {
	      servoSkid.wLeft = 0;
	      servoSkid.wRight = 0;
	      ackerMsg.velocity = 0;
	      ackerMsg.steerAngle = 0;
	      ackerMsg.crabAngle = 0;
	      submarineMsg.propeller = 0;
	      submarineMsg.rudder = 0;
	      submarineMsg.sternPlane = 0;

	      status_next(status, RCS_EXEC);
	      servoInit.serial_number = myServoSerial++;
	      servoMobJACmdNml->write( servoInit );
	      isServoInitialized = true;
	    }
	  else
	    {
	      checkServoCmdStatus( servoStatus, echo_serial_number );
	      if( servoStatus==RCS_DONE && echo_serial_number==myServoSerial-1 )
		{
		  state_next(status, S1 );
		  if( myVerbose )
		    fprintf( stderr, "Initialized servo vehicle\n" );
		}
	    }
	}
      else if( state_match( status, S1) ) // S1: send go command to sp
	{
	  servoGo.serial_number = myServoSerial++;
	  servoSPCmdNml->write( servoGo );
	  state_next(status, S2);
	  if( myVerbose )
	    fprintf( stderr, "Sent go command to SP\n" );
	}
      else if( state_match( status, S2) ) // S2: stay here and drive
	{
	  status_next(status, RCS_EXEC);
	  leftAccPressed = false;
	  leftDecPressed = false;
	  rightAccPressed = false;;
	  rightDecPressed = false;;

	  timerVal.it_interval.tv_usec = 250;
	  timerVal.it_interval.tv_sec = 0;
	  timerVal.it_value.tv_usec = 250;
	  timerVal.it_value.tv_sec = 0;
	  setitimer(ITIMER_REAL, &timerVal, 0);
	  for (done = 0; !done; ) 
	    {
	      XEvent event;
	      
	      XNextEvent (dpy, &event);
	      switch (event.type) {
	      case KeyPress:
		do_KeyPress (&event);
		break;
	      case KeyRelease:
		do_KeyRelease (&event);
		break;
	      }
	    }
	}
    }
}

/*
  Modification history:

  $Log: joystick.cc,v $
  Revision 1.12  2007/01/31 18:36:12  cj_scrapper
  - Added robot types to SIMware knowledge base (ground vehicles, nautic vehicles, etc. )
  - Added robot type specific stat and set messages to ServoMobJA.hh
  - Update USARSim infskin to work with USARSim version 3.01
  - Modified PrimMobMain, joystick, and joySDL to comply with new ServoMobJA messages

  Revision 1.11  2006/12/12 17:55:57  cj_scrapper
  modified for new servo msgs

  Revision 1.10  2006/10/10 13:30:17  dr_steveb
  New vehicle level exploration based on camera FOV implemented.

  Revision 1.9  2006/09/19 14:21:54  dr_steveb
  Added control mode 3 for controlling a submarine

  Revision 1.8  2006/09/13 14:44:42  dr_steveb
  Fixed segfault on exit.

  Revision 1.7  2006/07/25 14:14:11  dr_steveb
  Added "go" command to prim initialization.

  Revision 1.6  2006/07/11 16:47:25  dr_steveb
  Added the ability to control the maximum speed

  Revision 1.5  2006/04/12 13:02:16  dr_steveb
  Section level exploration now works for a single vehicle

  Revision 1.4  2006/04/05 01:59:37  dr_steveb
  Added new mode for control (using 4 arrow keys).

  Revision 1.3  2006/01/06 19:22:44  dr_steveb
  Fixed joystick to work with scaled speeds and acc/dec. Max and Min are now parameters.

  Revision 1.2  2005/10/27 15:02:55  dr_steveb
  Changed max speeds for better control.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.10  2005/09/30 19:33:41  proctor
  Took out unused variables, added forgotten return values

  Revision 1.9  2005/09/30 18:55:56  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.8  2005/08/09 21:08:30  stephen
  Small changes to make work better in demo

  Revision 1.7  2005/08/05 13:33:24  stephen
  Modified interface to respond to x events rather then the terminal io.

  Revision 1.6  2005/08/03 13:46:55  proctor
  Restored terminal attributes upon exit

  Revision 1.5  2005/08/01 15:03:07  stephen
  Added max speed and fred's changes for file io

  Revision 1.4  2005/07/28 15:57:23  stephen
  Fixed left vs. right

  Revision 1.3  2005/07/28 15:41:26  stephen
  Joystick now works

  Revision 1.2  2005/07/28 14:39:30  stephen
  Added math.h for fabs

  Revision 1.1  2005/07/28 14:38:13  stephen
  initial check-in

*/
