/*
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*/

/*!
  \file robojoy.cc

  \brief Joystick interface using Simple DirectMedia Layer for mission package

  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.34 $
  $Date: 2008/12/17 15:28:07 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>		// atoi,f
#include <string.h>		// memset
#include <math.h>
#include <rcs.hh>		// esleep
#include "SDL.h"
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "servoMobJA.hh"
#include "servoMisJA.hh"
#include "trafficCtrl.hh"
#include "joystickData.hh"
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
extern "C" {
#include "gotypes.h"
#include "gomath.h"
}

#define DEBUG

#define VERBOSE
/*
  define INHIBIT_PRISMATIC to inhibit the motion of prismatic joints
  during Cartesian motion
*/
#define INHIBIT_PRISMATIC

/*
  define CLAMP_FLIPPERS if you want the commanded positions of the
  flippers to be clamped between their min and max limits.  Testing
  showed that these limits didn't allow the flippers to move through
  their full ranges, and USARsim should be fixed.
*/
#undef CLAMP_FLIPPERS
// make FLIPPER_SPEED larger for higher speeds
#define FLIPPER_SPEED 0.01

/*
  RCS command writers get their own number pair for commands,
  with 0 reserved to indicate nobody, and pairs being an odd
  number and the next even number.
*/
#define WHONUM 9091		// nobody else should have this

// Serial number to servo level
static int servoMobSerial = WHONUM;
static int servoMisSerial = WHONUM;

// just add two each time and stay odd
#define NEXTS(s) (s += 2)

// NML Channels

static RCS_CMD_CHANNEL *servoMobJACmdNml = NULL;
static RCS_STAT_CHANNEL *servoMobJAStatNml = NULL;
static RCS_STAT_CHANNEL *servoMobJASetNml = NULL;

static RCS_CMD_CHANNEL * servoMisJACmdNml = NULL;
static RCS_STAT_CHANNEL * servoMisJAStatNml = NULL;
static RCS_STAT_CHANNEL * servoMisJASetNml = NULL;

static RCS_CMD_CHANNEL   *trafficCmdNml = NULL;

static NML * joystickDataNml = NULL;
static JoystickData joystickData;

static ServoMisJAStat * servoMisJAStatPtr = NULL;
static ServoMisJASet * servoMisJASetPtr = NULL;

static ServoMobJAStat * servoMobJAStatPtr = NULL;
static ServoMobJASet * servoMobJASetPtr = NULL;

static ServoMobJACmdSkid skidMsg;
static ServoMobJACmdFlipSkid flipSkidMsg;
static ServoMobJACmdAckerman ackerMsg;
static ServoMobJACmdSubmarine submarineMsg;
static ServoMobJACmdRotaryWing rotaryWingMsg;
static ServoMisJACmdMove misCmdMove;

static TrafficCtrlCmdSkid trafSkidMsg;
static TrafficCtrlCmdFlipSkid trafFlipSkidMsg;
static TrafficCtrlCmdAckerman trafAckerMsg;
static TrafficCtrlCmdSubmarine trafSubmarineMsg;
static TrafficCtrlCmdRotaryWing trafRotaryWingMsg;

static int mode;
static int vehID;
static int compID;
static int doInit;
static double scale[4];
static int trafVehID = 0;

static enum {COORD_JOINT = 1,
	     COORD_WORLD}
  coordMode = COORD_JOINT;

static int whichJoint = 0;

static int flipperCount = 0;
static int frontFlipperFlag = 0;
static int rearFlipperFlag = 0;
static double flipperPos[SERVO_MOB_JA_FLIPPER_TYPE_MAX] = {0.0};
static double flipperIncr[SERVO_MOB_JA_FLIPPER_TYPE_MAX];
static ServoMobJAFlipperSet flipperSet[SERVO_MOB_JA_FLIPPER_TYPE_MAX];
bool needToWrite;
bool userScale;
bool button8Down;
bool invertAxis;

#define MY_NAME_BASE "joystick"


bool openMission( int vehID )
{
  bool retval = true;
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];
  ServoMisJACmdInit servoMisInit;
  int gotIt;
  double end;

  sprintf(myName, "%s", MY_NAME_BASE);

  if (NULL == configNml) {
    configNml = "../etc/moast.nml";
  }

  printf( "robojoy: opening nml file: %s\n", configNml );
  if( servoMisJACmdNml != NULL )
    {
      delete servoMisJACmdNml;
      servoMisJACmdNml = NULL;
    }
  if (NULL == servoMisJACmdNml) {
    sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_CMD_NAME, vehID, compID);
    servoMisJACmdNml = new RCS_CMD_CHANNEL(servoMisJA_format, chanName, myName, configNml);
    if (! servoMisJACmdNml->valid()) {
      delete servoMisJACmdNml;
      servoMisJACmdNml = NULL;
      printf("can't open %s\n", chanName);
      retval = false;
    }
  }

  if( servoMisJAStatNml != NULL )
    {
      delete servoMisJAStatNml;
      servoMisJAStatNml = NULL;
    }
  if (NULL == servoMisJAStatNml) {
    sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_STAT_NAME, vehID, compID);
    servoMisJAStatNml = new RCS_STAT_CHANNEL(servoMisJA_format, chanName, myName, configNml);
    if (! servoMisJAStatNml->valid()) {
      delete servoMisJAStatNml;
      servoMisJAStatNml = NULL;
      printf("can't open %s\n", chanName);
      retval = false;
    } else {
      servoMisJAStatNml->read();
      servoMisJAStatPtr = (ServoMisJAStat *) (servoMisJAStatNml->get_address());
    }
  }

  if( servoMisJASetNml != NULL )
    {
      delete servoMisJASetNml;
      servoMisJASetNml = NULL;
    }
  if (NULL == servoMisJASetNml) {
    sprintf(chanName, "%s%d_%d", SERVO_MIS_JA_SET_NAME, vehID, compID);
    servoMisJASetNml = new RCS_STAT_CHANNEL(servoMisJA_format, chanName, myName, configNml);
    if (! servoMisJASetNml->valid()) {
      delete servoMisJASetNml;
      servoMisJASetNml = NULL;
      printf("can't open %s\n", chanName);
      retval = false;
    } else {
      servoMisJASetNml->read();
      servoMisJASetPtr = (ServoMisJASet *) (servoMisJASetNml->get_address());
    }
  }
  if( retval == false )
    return true;
  // send init command to servo mis
  servoMisInit.serial_number = NEXTS(servoMisSerial);
  servoMisJACmdNml->write(servoMisInit);
  for (end = etime() + 3.0, gotIt = 0; etime() < end; esleep(0.1)) {
    servoMisJAStatNml->read();
    if ( servoMisJAStatPtr->echo_serial_number == servoMisSerial &&
	 servoMisJAStatPtr->status == RCS_DONE) {
      gotIt = 1;
    }
  }
  if (!gotIt) {
    fprintf(stderr, "can't initialize servo mis\n");
    return 1;
  }
  else
    printf( "Initialized mission package %d\n", compID );
  
  return true;
}

/*!
  Open all of the NML channels
  
  \param vehID The number to add to all of the channel names
  \return true on success
  \return false on failure
*/
bool openNML(int vehID)
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];
  bool retval = true;

  sprintf(myName, "%s", MY_NAME_BASE);

  if (NULL == configNml) {
    configNml = "../etc/moast.nml";
  }

  if( vehID != -1 )
    {
      if (NULL == servoMobJACmdNml) {
	sprintf(chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, vehID);
	servoMobJACmdNml = new RCS_CMD_CHANNEL(servoMobJA_format, chanName, myName, configNml);
	if (! servoMobJACmdNml->valid()) {
	  delete servoMobJACmdNml;
	  servoMobJACmdNml = NULL;
	  printf("can't open %s\n", chanName);
	  retval = false;
	}
      }

      if (NULL == servoMobJAStatNml) {
	sprintf(chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, vehID);
	servoMobJAStatNml = new RCS_STAT_CHANNEL(servoMobJA_format, chanName, myName, configNml);
	if (! servoMobJAStatNml->valid()) {
	  delete servoMobJAStatNml;
	  servoMobJAStatNml = NULL;
	  printf("can't open %s\n", chanName);
	  retval = false;
	} else {
	  servoMobJAStatPtr = (ServoMobJAStat *) (servoMobJAStatNml->get_address());
	}
      }

      if (NULL == servoMobJASetNml) {
	sprintf(chanName, "%s%d", SERVO_MOB_JA_SET_NAME, vehID);
	servoMobJASetNml = new RCS_STAT_CHANNEL(servoMobJA_format, chanName, myName, configNml);
	if (! servoMobJASetNml->valid()) {
	  delete servoMobJASetNml;
	  servoMobJASetNml = NULL;
	  printf("can't open %s\n", chanName);
	  retval = false;
	} else {
	  servoMobJASetPtr = (ServoMobJASet *) (servoMobJASetNml->get_address());
	}
      }

      retval &= openMission( vehID );
    }
  else
    {
      sprintf(chanName, "%s%d", TRAFFIC_CTRL_CMD_NAME, 1);
      trafficCmdNml =  new RCS_CMD_CHANNEL(trafficCtrl_format, chanName,
					   myName, configNml );
      if(NULL == trafficCmdNml || !trafficCmdNml->valid() )
	{
	  rcs_print_error( "%s: Error - Can't open %s buffer\n",
			   myName, chanName );
	  retval = false;
	}
    }

  if (NULL == joystickDataNml) {
    sprintf(chanName, "joystickData");
    joystickDataNml = new NML(joystickData_format, chanName, myName, configNml);
    if (! joystickDataNml->valid()) {
      delete joystickDataNml;
      joystickDataNml = NULL;
      printf("can't open %s; this is NOT a fatal error\n", chanName);
      //      retval = false;
    }
  }
  
  return retval;
}

/*!
  free all memory and prepare to exit
*/
void closeNML()
{
  skidMsg.wLeft = 0;
  skidMsg.wRight = 0;
  ackerMsg.velocity = 0;
  ackerMsg.crabAngle = 0;
  ackerMsg.steerAngle = 0;
  submarineMsg.propeller = 0;
  submarineMsg.rudder = 0;
  submarineMsg.sternPlane = 0;
  rotaryWingMsg.vVertical = 0;
  rotaryWingMsg.vLinear = 0;
  rotaryWingMsg.vLateral = 0;
  rotaryWingMsg.vRotation = 0;

  trafSkidMsg.cmd[0].wLeft = 0;
  trafSkidMsg.cmd[0].wRight = 0;
  trafSkidMsg.cmd[0].vehID = trafVehID;
  trafSkidMsg.cmd_length = 1;

  trafAckerMsg.cmd[0].velocity = 0;
  trafAckerMsg.cmd[0].crabAngle = 0;
  trafAckerMsg.cmd[0].steerAngle = 0;
  trafAckerMsg.cmd[0].vehID = trafVehID;
  trafAckerMsg.cmd_length = 1;

  trafSubmarineMsg.cmd[0].propeller = 0;
  trafSubmarineMsg.cmd[0].rudder = 0;
  trafSubmarineMsg.cmd[0].sternPlane = 0;
  trafSubmarineMsg.cmd[0].vehID = trafVehID;
  trafSubmarineMsg.cmd_length = 1;

  trafRotaryWingMsg.cmd[0].vVertical = 0;
  trafRotaryWingMsg.cmd[0].vLinear = 0;
  trafRotaryWingMsg.cmd[0].vLateral = 0;
  trafRotaryWingMsg.cmd[0].vRotation = 0;
  trafRotaryWingMsg.cmd[0].vehID = trafVehID;
  trafRotaryWingMsg.cmd_length = 1;

  int t;
  
  switch (mode) {
  case 0:
  case 1:
    if( vehID != -1 )
      {
	skidMsg.serial_number = NEXTS(servoMobSerial);
	servoMobJACmdNml->write(skidMsg);
	//	printf( "From 0:%f %f\n", skidMsg.wLeft, skidMsg.wRight);
      }
    else
      {
	trafSkidMsg.serial_number = NEXTS(servoMobSerial);
	trafficCmdNml->write(trafSkidMsg);
      }
    break;
  case 2:
    if( vehID != -1 )
      {
	ackerMsg.serial_number = NEXTS(servoMobSerial);
	servoMobJACmdNml->write(ackerMsg);
      }
    else
      {
	trafAckerMsg.serial_number = NEXTS(servoMobSerial);
	trafficCmdNml->write(trafAckerMsg);
      }
    break;
  case 3:
    if( vehID != -1 )
      {
	submarineMsg.serial_number = NEXTS(servoMobSerial);
	servoMobJACmdNml->write(submarineMsg);
      }
    else
      {
	trafSubmarineMsg.serial_number = NEXTS(servoMobSerial);
	trafficCmdNml->write(trafSubmarineMsg);
      }
    break;
  case 4:
    if( vehID != -1 )
      {
	rotaryWingMsg.serial_number = NEXTS(servoMobSerial);
	servoMobJACmdNml->write(rotaryWingMsg);
      }
    else
      {
	trafRotaryWingMsg.serial_number = NEXTS(servoMobSerial);
	trafficCmdNml->write(trafRotaryWingMsg);
      }
    break;
  }

  misCmdMove.linkCmd_length = SERVO_MIS_JA_LINK_MAX;
  misCmdMove.serial_number = NEXTS(servoMisSerial);
  for (t = 0; t < SERVO_MIS_JA_LINK_MAX; t++) {
    misCmdMove.linkCmd[t].linkID = t + 1;
    misCmdMove.linkCmd[t].value = 0.0;
    misCmdMove.linkCmd[t].type = SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE;
  }
  if( servoMisJACmdNml!= NULL )
    servoMisJACmdNml->write(misCmdMove);

  if (servoMobJACmdNml != NULL)
    delete servoMobJACmdNml;
  servoMobJACmdNml = NULL;

  if (servoMobJAStatNml != NULL)
    delete servoMobJAStatNml;
  servoMobJAStatNml = NULL;

  if (servoMobJASetNml != NULL)
    delete servoMobJASetNml;
  servoMobJASetNml = NULL;

  if (servoMisJACmdNml != NULL)
    delete servoMisJACmdNml;
  servoMisJACmdNml = NULL;

  if (servoMisJAStatNml != NULL)
    delete servoMisJAStatNml;
  servoMisJAStatNml = NULL;

  if (servoMisJASetNml != NULL)
    delete servoMisJASetNml;
  servoMisJASetNml = NULL;

  if (joystickDataNml != NULL)
    delete joystickDataNml;
  joystickDataNml = NULL;

  if (trafficCmdNml != NULL)
    delete trafficCmdNml;
  trafficCmdNml = NULL;
}

enum {JOINTS = 8};

int compute_jfwd(go_link * link_params, int link_number, go_matrix * Jfwd, go_pose * T_L_0)
{
  GO_MATRIX_DECLARE(Jv, Jvstg, 3, JOINTS);
  GO_MATRIX_DECLARE(Jw, Jwstg, 3, JOINTS);
  GO_MATRIX_DECLARE(R_i_ip1, R_i_ip1stg, 3, 3);
  GO_MATRIX_DECLARE(scratch, scratchstg, 3, JOINTS);
  GO_MATRIX_DECLARE(R_inv, R_invstg, 3, 3);
  go_pose pose;
  go_quat quat;
  go_vector P_ip1_i[3];
  int row, col;

  /* init matrices to possibly smaller size */
  go_matrix_init(Jv, Jvstg, 3, link_number);
  go_matrix_init(Jw, Jwstg, 3, link_number);
  go_matrix_init(R_i_ip1, R_i_ip1stg, 3, 3);
  go_matrix_init(scratch, scratchstg, 3, link_number);
  go_matrix_init(R_inv, R_invstg, 3, 3);

  Jv.el[0][0] = 0, Jv.el[1][0] = 0, Jv.el[2][0] = (GO_QUANTITY_LENGTH == link_params[0].quantity ? 1 : 0);
  Jw.el[0][0] = 0, Jw.el[1][0] = 0, Jw.el[2][0] = (GO_QUANTITY_ANGLE == link_params[0].quantity ? 1 : 0);

  /* initialize inverse rotational transform */
  if (GO_LINK_DH == link_params[0].type) {
    go_dh_pose_convert(&link_params[0].u.dh, &pose);
  } else if (GO_LINK_PP == link_params[0].type) {
    pose = link_params[0].u.pp.pose;
  } else {
    return 1;
  }

  *T_L_0 = pose;

  for (col = 1; col < link_number; col++) {
    /* T_ip1_i */
    if (GO_LINK_DH == link_params[col].type) {
      go_dh_pose_convert(&link_params[col].u.dh, &pose);
    } else if (GO_LINK_PP == link_params[col].type) {
      pose = link_params[col].u.pp.pose;
    } else {
      return 1;
    }

    go_cart_vector_convert(&pose.tran, P_ip1_i);
    go_quat_inv(&pose.rot, &quat);
    go_quat_matrix_convert(&quat, &R_i_ip1);

    /* Jv */
    go_matrix_vector_cross(&Jw, P_ip1_i, &scratch);
    go_matrix_matrix_add(&Jv, &scratch, &scratch);
    go_matrix_matrix_mult(&R_i_ip1, &scratch, &Jv);
    Jv.el[0][col] = 0, Jv.el[1][col] = 0, Jv.el[2][col] = (GO_QUANTITY_LENGTH == link_params[col].quantity ? 1 : 0);
    /* Jw */
    go_matrix_matrix_mult(&R_i_ip1, &Jw, &Jw);
    Jw.el[0][col] = 0, Jw.el[1][col] = 0, Jw.el[2][col] = (GO_QUANTITY_ANGLE == link_params[col].quantity ? 1 : 0);
    pose = link_params[col].u.pp.pose;
    go_pose_pose_mult(T_L_0, &pose, T_L_0);
  }

#if 0
  for (row = 0; row < 6; row++) {
    for (col = 0; col < link_number; col++) {
      if (row < 3) {
	printf("%f\t", Jv.el[row][col]);
      } else {
	printf("%f\t", Jw.el[row - 3][col]);
      }
    }
    printf("\n");
  }
  printf("---\n");
#endif

  /* rotate back into {0} frame */
  go_quat_matrix_convert(&T_L_0->rot, &R_inv);
  go_matrix_matrix_mult(&R_inv, &Jv, &Jv);
  go_matrix_matrix_mult(&R_inv, &Jw, &Jw);

  /* put Jv atop Jw in J */
  for (row = 0; row < 6; row++) {
    for (col = 0; col < link_number; col++) {
      if (row < 3) {
	Jfwd->el[row][col] = Jv.el[row][col];
      } else {
	Jfwd->el[row][col] = Jw.el[row - 3][col];
      }
    }
  }

#if 0
  for (row = 0; row < 6; row++) {
    for (col = 0; col < link_number; col++) {
      printf("%f\t", Jfwd->el[row][col]);
    }
    printf("\n");
  }
#endif

  return 0;
}

int compute_jinv(go_matrix * Jfwd, go_matrix * Jinv)
{
  go_result retval;
  int row, col;

  /* compute inverse, or pseudo-inverse */
  if (Jfwd->rows == Jfwd->cols) {
    if (GO_RESULT_OK != go_matrix_inv(Jfwd, Jinv)) {
      fprintf(stderr, "can't invert Jacobian\n");
#if 1
      printf("---\n");
      for (row = 0; row < Jfwd->rows; row++) {
	for (col = 0; col < Jfwd->cols; col++) {
	  printf("%f\t", Jfwd->el[row][col]);
	}
	printf("\n");
      }
#endif
      return 1;
    }
  } else if (Jfwd->rows < Jfwd->cols) {
    /* underdetermined, optimize on smallest sum of square of speeds */
    /* JT(JJT)inv */
    GO_MATRIX_DECLARE(JT, JTstg, SERVO_MIS_JA_LINK_MAX, 6);
    GO_MATRIX_DECLARE(JJT, JJTstg, 6, 6);
    
    go_matrix_init(JT, JTstg, Jfwd->cols, Jfwd->rows);
    go_matrix_init(JJT, JJTstg, Jfwd->rows, Jfwd->rows);
    go_matrix_transpose(Jfwd, &JT);
    go_matrix_matrix_mult(Jfwd, &JT, &JJT);
    retval = go_matrix_inv(&JJT, &JJT);
    if (GO_RESULT_OK != retval) {
      fprintf(stderr, "can't invert Jacobian\n");
#if 1
      printf("---\n");
      for (row = 0; row < JJT.rows; row++) {
	for (col = 0; col < JJT.cols; col++) {
	  printf("%f\t", JJT.el[row][col]);
	}
	printf("\n");
      }
#endif
      return 1;
    }
    go_matrix_matrix_mult(&JT, &JJT, Jinv);
  } else {
    /* overdetermined, do least-squares best fit */
    /* (JTJ)invJT */
    GO_MATRIX_DECLARE(JT, JTstg, SERVO_MIS_JA_LINK_MAX, 6);
    GO_MATRIX_DECLARE(JTJ, JTJstg, SERVO_MIS_JA_LINK_MAX, SERVO_MIS_JA_LINK_MAX);
    
    go_matrix_init(JT, JTstg, Jfwd->cols, Jfwd->rows);
    go_matrix_init(JTJ, JTJstg, Jfwd->cols, Jfwd->cols);
    go_matrix_transpose(Jfwd, &JT);
    go_matrix_matrix_mult(&JT, Jfwd, &JTJ);
    retval = go_matrix_inv(&JTJ, &JTJ);
    if (GO_RESULT_OK != retval) {
      fprintf(stderr, "can't invert Jacobian\n");
#if 1
      printf("---\n");
      for (row = 0; row < JTJ.rows; row++) {
	for (col = 0; col < JTJ.cols; col++) {
	  printf("%f\t", JTJ.el[row][col]);
	}
	printf("\n");
      }
#endif
      return 1;
    }
    go_matrix_matrix_mult(&JTJ, &JT, Jinv);
  }

#if 0
  printf("---\n");
  for (row = 0; row < Jinv->rows; row++) {
    for (col = 0; col < 6; col++) {
      printf("%f\t", Jinv->el[row][col]);
    }
    printf("\n");
  }
#endif

  return 0;
}

static go_quat QWorld = {0.0, 0.707107, 0.707107, 0.0};
static go_quat QTool = {0.5, 0.5, 0.5, 0.5};

void WatchJoystick(SDL_Joystick * joystick)
{
  int axis, t, tt;
  double values[4] = {0.0};
  int done;
  SDL_Event event;
  NMLTYPE type;
  // 
  int link_number;
  go_link link_params[JOINTS];
  go_link linkout[JOINTS];
  go_real maxVelocity[JOINTS];
  GO_MATRIX_DECLARE(RWorld, RWorldstg, 3, 3);
  GO_MATRIX_DECLARE(RTool, RToolstg, 3, 3);
  GO_MATRIX_DECLARE(Jfwd, Jfwdstg, 6, JOINTS);
  GO_MATRIX_DECLARE(Jinv, Jinvstg, JOINTS, 6);
  go_vector jointvels[JOINTS];
  go_pose T_L_0;
  int toolMode = 0;
  int doRot = 0;
  int doMob = 1;
  int last8 = 0;
  int last9 = 0;
  ServoMobJASetGrdVeh *servoMobJASetGrdVeh = (ServoMobJASetGrdVeh*)servoMobJASetPtr;
  double angularVel, vehRate;

#define MAX_ANG_VEL 2.5

#ifdef VERBOSE
  const char * name = SDL_JoystickName(SDL_JoystickIndex(joystick));
  printf("Watching joystick %d: (%s)\n", SDL_JoystickIndex(joystick),
	 NULL != name ? name : "Unknown Joystick");
  printf("Joystick has %d axes, %d hats, %d balls, and %d buttons\n",
	 SDL_JoystickNumAxes(joystick), SDL_JoystickNumHats(joystick),
	 SDL_JoystickNumBalls(joystick), SDL_JoystickNumButtons(joystick));
#endif

  // Get manipulator info out of NML
  if (doInit && servoMisJASetNml != NULL) {
    ServoMisJALinkSet link;
    go_pose pose, last_pose;
    go_rpy rpy;
    int real_link_number;

    if( servoMisJASetNml != NULL )
      {
	type = servoMisJASetNml->read();
	link_number = servoMisJASetPtr->linkSet_length;
      }
    else 
      {
	type = 0;
	link_number = 0;
      }
    /*
      Force the link number to be 7 for the Telemax
    */
    if (link_number > 7) link_number = 7;
    last_pose = go_pose_identity();
    real_link_number = 0;
    for (t = 0; t < link_number; t++) {
      pose.tran.x = servoMisJASetPtr->linkSet[t].loc.x;
      pose.tran.y = servoMisJASetPtr->linkSet[t].loc.y;
      pose.tran.z = servoMisJASetPtr->linkSet[t].loc.z;
      rpy.r = servoMisJASetPtr->linkSet[t].rot.r;
      rpy.p = servoMisJASetPtr->linkSet[t].rot.p;
      rpy.y = servoMisJASetPtr->linkSet[t].rot.y;
#ifdef VERBOSE
      printf("link %d has settings %f %f %f %f %f %f\n", t,
	     (double) pose.tran.x, (double) pose.tran.y, (double) pose.tran.z,
	     (double) rpy.r, (double) rpy.p, (double) rpy.y);
#endif
      go_rpy_quat_convert(&rpy, &pose.rot);
      go_pose_pose_mult(&last_pose, &pose, &pose);
#ifdef INHIBIT_PRISMATIC
      if (SERVO_MIS_JA_PRISMATIC_JOINT_TYPE == servoMisJASetPtr->linkSet[t].jType) {
	// don't make this a real link, just save its pose to prepend
	// to the next one
#ifdef VERBOSE
	printf("accumulating link %d\n", t);
#endif
	last_pose = pose;
      } else {
#endif
	last_pose = go_pose_identity();
	link_params[real_link_number].u.pp.pose = pose;
	link_params[real_link_number].type = GO_LINK_PP;
	link_params[real_link_number].quantity = (SERVO_MIS_JA_PRISMATIC_JOINT_TYPE == servoMisJASetPtr->linkSet[t].jType ? GO_QUANTITY_LENGTH : GO_QUANTITY_ANGLE);
	maxVelocity[real_link_number] = servoMisJASetPtr->linkSet[t].maxVelocity;
	real_link_number++;
      }
#ifdef INHIBIT_PRISMATIC
    }
#endif
#ifdef VERBOSE
    printf("%d physical links in manipulator\n", link_number);
    printf("%d  handled links in manipulator\n", real_link_number);
#endif
    link_number = real_link_number;

    for (t = 0; t < link_number; t++) {
      go_quat_rpy_convert(&link_params[t].u.pp.pose.rot, &rpy);
#ifdef VERBOSE
      printf("real link %d has settings %f %f %f %f %f %f\n", t,
	     (double) link_params[t].u.pp.pose.tran.x,
	     (double) link_params[t].u.pp.pose.tran.y,
	     (double) link_params[t].u.pp.pose.tran.z,
	     (double) rpy.r, (double) rpy.p, (double) rpy.y);
#endif
    }

    go_matrix_init(RWorld, RWorldstg, 3, 3);
    go_matrix_init(RTool, RToolstg, 3, 3);
    go_matrix_init(Jfwd, Jfwdstg, 6, link_number);
    go_matrix_init(Jinv, Jinvstg, link_number, 6);

    go_quat_matrix_convert(&QWorld, &RWorld);
    go_quat_matrix_convert(&QTool, &RTool);

    if( vehID != -1 )
      {
	type = servoMobJAStatNml->read();
	type = servoMobJASetNml->read();
	flipperCount = (int) ((ServoMobJAStatGrdVeh *) servoMobJAStatPtr)->fStat_length;
      }
    else
      {
	flipperCount = 0; // !!FIXME - should read flipper count from traffic sim
      }
    for (t = 0; t < flipperCount; t++) {
      flipperSet[t] = ((ServoMobJASetGrdVeh *) servoMobJASetPtr)->fSet[t];
      // FIXME-- min/max appear to be reversed in settings
      if (flipperSet[t].minAngle > flipperSet[t].maxAngle) {
	double tmp = flipperSet[t].minAngle;
	flipperSet[t].minAngle = flipperSet[t].maxAngle;
	flipperSet[t].maxAngle = tmp;
      }
      flipperIncr[t] = FLIPPER_SPEED * (flipperSet[t].maxAngle - flipperSet[t].minAngle);
    }
  } // doInit

  printf("Joystick active\n");

  /* loop, getting joystick events */
  for (done = 0; done == 0; esleep(0.1)) {
    while (SDL_PollEvent(&event)) {
      switch (event.type) {

      case SDL_JOYAXISMOTION:
	axis = event.jaxis.axis;
	if (axis < 0) axis = 0;
	else if (axis > 3) axis = 3;
#define INV_RANGE (1.0 / 32768.0)
	values[axis] = scale[axis] * ((double) event.jaxis.value) * INV_RANGE;
	if (fabs(values[axis]) < 0.1 * fabs(scale[axis])) values[axis] = 0.0;
	if( invertAxis )
	  values[axis] *= -1;
	if (! doInit)
	  printf("Axis: %d value %d\n",axis, event.jaxis.value );
	break;

      case SDL_JOYHATMOTION:
	printf("Joystick %d hat %d value:", event.jhat.which, event.jhat.hat);
	if (event.jhat.value == SDL_HAT_CENTERED)
	  printf(" centered");
	if (event.jhat.value & SDL_HAT_UP)
	  printf(" up");
	if (event.jhat.value & SDL_HAT_RIGHT)
	  printf(" right");
	if (event.jhat.value & SDL_HAT_DOWN)
	  printf(" down");
	if (event.jhat.value & SDL_HAT_LEFT)
	  printf(" left");
	printf("\n");
	break;

      case SDL_JOYBALLMOTION:
	printf("Joystick %d ball %d delta: (%d,%d)\n",
	       event.jball.which,
	       event.jball.ball, event.jball.xrel, event.jball.yrel);
	break;

      case SDL_JOYBUTTONDOWN:
	if (! doInit) {
	  printf("Joystick %d button %d down\n",
		 event.jbutton.which, event.jbutton.button);
	}
	switch (event.jbutton.button) {
	case 0:
	  if (doMob) {
	    trafVehID = event.jbutton.button;
	  } else {
	    if (button8Down) {
	      button8Down = false;
	      compID = event.jbutton.button + 1;
	      openMission( vehID );
	    }else {
	    if (COORD_JOINT == coordMode)
	      {
		whichJoint = 0;
		if( whichJoint < servoMisJASetPtr->linkSet_length && !userScale )
		  scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
	      }
	    }
	  }
	  break;
	case 1:
	  if (doMob) {
	    trafVehID = event.jbutton.button;
	  }else{
	    if (button8Down) {
	      button8Down = false;
	      compID = event.jbutton.button + 1;
	      openMission( vehID );
	    }else {
	    if (COORD_JOINT == coordMode) 
	      {
		whichJoint = 1;
		if( whichJoint < servoMisJASetPtr->linkSet_length && !userScale )
		  scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
	      }
	    }
	  }
	  break;
	case 2:
	  if (doMob) {
	    trafVehID = event.jbutton.button;
	  }else{
	    if (button8Down) {
	      button8Down = false;
	      compID = event.jbutton.button + 1;
	      openMission( vehID );
	    }else {
	    if (COORD_JOINT == coordMode)
	      {
		whichJoint = 2;
		if( whichJoint < servoMisJASetPtr->linkSet_length && !userScale )
		  scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
	      }
	    }
	  }
	  break;
	case 3:
	  if (doMob) {
	    trafVehID = event.jbutton.button;
	  }else{
	    if (button8Down) {
	      button8Down = false;
	      compID = event.jbutton.button + 1;
	      openMission( vehID );
	    }else {
	    if (COORD_JOINT == coordMode)
	      {
		whichJoint = 3;
		if( whichJoint < servoMisJASetPtr->linkSet_length && !userScale )
		  scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
	      }
	    }
	  }
	  break;
	case 4:
	  if (! doMob) {
	    if (button8Down) {
	      button8Down = false;
	      compID = event.jbutton.button + 1;
	      openMission( vehID );
	    }else {
	    if (COORD_JOINT == coordMode) 
	      {
		whichJoint = 4;
		if( whichJoint < servoMisJASetPtr->linkSet_length && !userScale )
		  scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
	      }
	    }
	  } else {
	    frontFlipperFlag = -1;
	  }
	  break;
	case 5:
	  if (! doMob) {
	    if (button8Down) {
	      button8Down = false;
	      compID = event.jbutton.button + 1;
	      openMission( vehID );
	    }else {
	    if (COORD_JOINT == coordMode)
	      {
		whichJoint = 5;
		if( whichJoint < servoMisJASetPtr->linkSet_length && !userScale )
		  scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
	      }
	    }
	  } else {
	    rearFlipperFlag = 1;
	  }
	  break;
	case 6:
	  if (! doMob) {
	    if (button8Down) {
	      button8Down = false;
	      compID = event.jbutton.button + 1;
	      openMission( vehID );
	    }else {
	    if (COORD_JOINT == coordMode)
	      {
		whichJoint = 6;
		if( whichJoint < servoMisJASetPtr->linkSet_length && !userScale )
		  scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
	      }
	    }
	  } else {
	    frontFlipperFlag = 1;
	  }
	  break;
	case 7:
	  if (! doMob) {
	    doRot = 1;
	  } else {
	    rearFlipperFlag = -1;
	  }
	  break;
	case 8:
	  button8Down = true;
	  break;
	case 9:
	  coordMode = COORD_WORLD;
	  doMob = 0;
	  if (last9) {
	    if (toolMode) {
	      toolMode = 0;
	    } else {
	      toolMode = 1;
	    }
	  }
	  last9 = 1;
	  last8 = 0;
	  printf("%s mode\n", toolMode ? "Tool" : "World");
	  break;
	case 10:
	  break;
	case 11:
	  break;
	  } // switch (event.jbutton.button)
	break;

      case SDL_JOYBUTTONUP:
	if (! doInit) {
	  printf("Joystick %d button %d up\n",
		 event.jbutton.which, event.jbutton.button);
	}else{
	switch( event.jbutton.button )
	  {
	  case 4:
	  case 6:
	    if(doMob) {
	      frontFlipperFlag = 0;
	    }
	    break;
	  case 5:
	  case 7:
	    if(doMob) {
	      rearFlipperFlag = 0;
	    }
	    else
	      doRot = 0;
	    break;
	  case 8:
	    if( button8Down == false )
	      break;
	    button8Down = false;
	    coordMode = COORD_JOINT;
	    toolMode = 1;
	    if (last8) {
	      if (doMob) {
		doMob = 0;
	      } else {
		doMob = 1;
	      }
	    }
	    last8 = 1;
	    last9 = 0;
	    printf("%s mode\n", doMob ? "Mobility" : "Joint");
	    if( doMob )
	      {
		if( !userScale )
		  {
		    switch(mode)
		      {
		      case 1:
			scale[1] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxWheelRot;      
			scale[3] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxWheelRot;
			break;
		      case 2:
			scale[1] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxWheelRot;
			scale[0] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxSteerAngle;
			scale[2] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxCrabAngle;
			break;
		      case 4:
			scale[1] = -((ServoMobJASetAirBot *)servoMobJASetPtr)->maxVertVel;      
			scale[3] = -((ServoMobJASetAirBot *)servoMobJASetPtr)->maxLinVel;	
			scale[2] = ((ServoMobJASetAirBot *)servoMobJASetPtr)->maxLatVel;
			scale[0] = ((ServoMobJASetAirBot *)servoMobJASetPtr)->maxRotVel;
			break;
		      }
		  }
	      }
	    else // joint
	      {
		if( !userScale )
		  {
		    if( whichJoint < servoMisJASetPtr->linkSet_length )
		      scale[0] = servoMisJASetPtr->linkSet[whichJoint].maxVelocity;
		  }
	      }
	  }
	}
	break;
      case SDL_KEYDOWN:
	if (event.key.keysym.sym != SDLK_ESCAPE) {
	  break;
	}
	/* Fall through to signal quit */
	
      case SDL_QUIT:
	done = 1;
	break;
	
      default:
	break;
      }	// switch (event.type)
    } // while (SDLPollEvent)

    if (doInit) {
      if( servoMisJAStatNml != NULL )
	type = servoMisJAStatNml->read();
      else
	type = 0;
      if (doMob) {
	switch (mode) {
	case 0:
	case 1:
	  if (flipperCount > 0) {

	    if (frontFlipperFlag > 0) {
	      flipperPos[0] += flipperIncr[0];
	      flipperPos[1] += flipperIncr[1];
#ifdef CLAMP_FLIPPERS
	      if (flipperPos[0] > flipperSet[0].maxAngle) flipperPos[0] = flipperSet[0].maxAngle;
	      if (flipperPos[1] > flipperSet[1].maxAngle) flipperPos[1] = flipperSet[1].maxAngle;
#endif
	    } else if (frontFlipperFlag < 0) {
	      flipperPos[0] -= flipperIncr[0];
	      flipperPos[1] -= flipperIncr[1];
#ifdef CLAMP_FLIPPERS
	      if (flipperPos[0] < flipperSet[0].minAngle) flipperPos[0] = flipperSet[0].minAngle;
	      if (flipperPos[1] < flipperSet[1].minAngle) flipperPos[1] = flipperSet[1].minAngle;
#endif
	    }

	    if (rearFlipperFlag > 0) {
	      flipperPos[2] += flipperIncr[2];
	      flipperPos[3] += flipperIncr[3];
#ifdef CLAMP_FLIPPERS
	      if (flipperPos[2] > flipperSet[2].maxAngle) flipperPos[2] = flipperSet[2].maxAngle;
	      if (flipperPos[3] > flipperSet[3].maxAngle) flipperPos[3] = flipperSet[3].maxAngle;
#endif
	    } else if (rearFlipperFlag < 0) {
	      flipperPos[2] -= flipperIncr[2];
	      flipperPos[3] -= flipperIncr[3];
#ifdef CLAMP_FLIPPERS
	      if (flipperPos[2] < flipperSet[2].minAngle) flipperPos[2] = flipperSet[2].minAngle;
	      if (flipperPos[3] < flipperSet[3].minAngle) flipperPos[3] = flipperSet[3].minAngle;
#endif
	    }

	    if( vehID != -1 )
	      {
		flipSkidMsg.wLeft = values[1];
		flipSkidMsg.wRight = values[3];
		flipSkidMsg.fCmd[0].fType = MOAST_FR_FLIPPER_TYPE;
		flipSkidMsg.fCmd[1].fType = MOAST_FL_FLIPPER_TYPE;
		flipSkidMsg.fCmd[2].fType = MOAST_RR_FLIPPER_TYPE;
		flipSkidMsg.fCmd[3].fType = MOAST_RL_FLIPPER_TYPE;
		flipSkidMsg.fCmd[0].flipperAngle = flipperPos[0];
		flipSkidMsg.fCmd[1].flipperAngle = flipperPos[1];
		flipSkidMsg.fCmd[2].flipperAngle = flipperPos[2];
		flipSkidMsg.fCmd[3].flipperAngle = flipperPos[3];
		flipSkidMsg.fCmd_length = flipperCount;
		flipSkidMsg.serial_number = NEXTS(servoMobSerial);
		servoMobJACmdNml->write(flipSkidMsg);
	      }
	    else
	      {
		trafFlipSkidMsg.cmd[0].wLeft = values[1];
		trafFlipSkidMsg.cmd[0].wRight = values[3];
		trafFlipSkidMsg.cmd[0].vehID = trafVehID;
		trafFlipSkidMsg.cmd_length = 1;
		trafFlipSkidMsg.fCmd[0].fType = MOAST_FR_FLIPPER_TYPE;
		trafFlipSkidMsg.fCmd[1].fType = MOAST_FL_FLIPPER_TYPE;
		trafFlipSkidMsg.fCmd[2].fType = MOAST_RR_FLIPPER_TYPE;
		trafFlipSkidMsg.fCmd[3].fType = MOAST_RL_FLIPPER_TYPE;
		trafFlipSkidMsg.fCmd[0].flipperAngle = flipperPos[0];
		trafFlipSkidMsg.fCmd[1].flipperAngle = flipperPos[1];
		trafFlipSkidMsg.fCmd[2].flipperAngle = flipperPos[2];
		trafFlipSkidMsg.fCmd[3].flipperAngle = flipperPos[3];
		trafFlipSkidMsg.fCmd[0].vehID = trafVehID;
		trafFlipSkidMsg.fCmd[1].vehID = trafVehID;
		trafFlipSkidMsg.fCmd[2].vehID = trafVehID;
		trafFlipSkidMsg.fCmd[3].vehID = trafVehID;
		trafFlipSkidMsg.fCmd_length = flipperCount;
		trafFlipSkidMsg.serial_number = NEXTS(servoMobSerial);
		servoMobJACmdNml->write(trafFlipSkidMsg);
	      }

	  } else {
	    if( vehID != -1 )
	      {
		angularVel = (values[3]-values[1])/servoMobJASetGrdVeh->wheelSeparation;
		if( fabs(angularVel) > MAX_ANG_VEL )
		  {
		    if(angularVel > 0 )
		      angularVel = MAX_ANG_VEL;
		    else
		      angularVel = -MAX_ANG_VEL;
		    vehRate = servoMobJASetGrdVeh->wheelSeparation/2.*(values[1]+values[3])/
		      (values[3]-values[1]);
		    skidMsg.wLeft = angularVel * (vehRate - servoMobJASetGrdVeh->wheelSeparation/2.);
		    skidMsg.wRight = angularVel * (vehRate + servoMobJASetGrdVeh->wheelSeparation/2.);
		  }
		else
		  {
		    skidMsg.wLeft = values[1];
		    skidMsg.wRight = values[3];
		  }
		skidMsg.serial_number = NEXTS(servoMobSerial);
		servoMobJACmdNml->write(skidMsg);
		/*
		printf( "Wheel sep: %f Ang. Veh: %f\n",
			servoMobJASetGrdVeh->wheelSeparation,
			(skidMsg.wLeft-skidMsg.wRight)/servoMobJASetGrdVeh->wheelSeparation );
		*/
		//		printf( "From 1:%f %f\n", skidMsg.wLeft, skidMsg.wRight);
	      }
	    else
	      {
		trafSkidMsg.cmd[0].wLeft = values[1];
		trafSkidMsg.cmd[0].wRight = values[3];
		trafSkidMsg.cmd[0].vehID = trafVehID;
		trafSkidMsg.cmd_length = 1;
		trafSkidMsg.serial_number = NEXTS(servoMobSerial);
		trafficCmdNml->write(trafSkidMsg);
	      }
	  }
	  break;
	case 2:
	  if( vehID != -1 )
	    {
	      ackerMsg.velocity = values[1];
	      ackerMsg.steerAngle = values[0];
	      ackerMsg.crabAngle = values[2];
	      ackerMsg.serial_number = NEXTS(servoMobSerial);
	      /*
		printf( "Velocity: %f steer: %f crab: %f\n", 
		ackerMsg.velocity,
		ackerMsg.steerAngle,
		ackerMsg.crabAngle);
	      */
	      servoMobJACmdNml->write(ackerMsg);
	    }
	  else
	    {
	      trafAckerMsg.cmd[0].velocity = values[1];
	      trafAckerMsg.cmd[0].steerAngle = values[2];
	      trafAckerMsg.cmd[0].crabAngle = values[3];
	      trafAckerMsg.cmd[0].vehID = trafVehID;
	      trafAckerMsg.cmd_length = 1;
	      trafAckerMsg.serial_number = NEXTS(servoMobSerial);
	      trafficCmdNml->write(trafAckerMsg);
	    }
	  break;
	case 3:
	  if( vehID != -1 )
	    {
	      submarineMsg.rudder = values[2];
	      submarineMsg.sternPlane = values[3];
	      submarineMsg.propeller = values[1];
	      submarineMsg.serial_number = NEXTS(servoMobSerial);
	      servoMobJACmdNml->write(submarineMsg);
	    }
	  else
	    {
	      trafSubmarineMsg.cmd[0].rudder = values[2];
	      trafSubmarineMsg.cmd[0].sternPlane = values[3];
	      trafSubmarineMsg.cmd[0].propeller = values[1];
	      trafSubmarineMsg.cmd[0].vehID = trafVehID;
	      trafSubmarineMsg.cmd_length = 1;
	      trafSubmarineMsg.serial_number = NEXTS(servoMobSerial);
	      trafficCmdNml->write(trafSubmarineMsg);
	    }
	  break;
	case 4:
	  if( vehID != -1 )
	    {
	      rotaryWingMsg.vVertical = values[1];
	      rotaryWingMsg.vLinear = values[3];
	      rotaryWingMsg.vLateral = values[2];
	      rotaryWingMsg.vRotation = values[0];
	      rotaryWingMsg.serial_number = NEXTS(servoMobSerial);
	      servoMobJACmdNml->write(rotaryWingMsg);
	    }
	  else
	    {
	      trafRotaryWingMsg.cmd[0].vVertical = values[1];
	      trafRotaryWingMsg.cmd[0].vLinear = values[3];
	      trafRotaryWingMsg.cmd[0].vLateral = values[2];
	      trafRotaryWingMsg.cmd[0].vRotation = values[0];
	      trafRotaryWingMsg.cmd[0].vehID = trafVehID;
	      trafRotaryWingMsg.cmd_length = 1;
	      trafRotaryWingMsg.serial_number = NEXTS(servoMobSerial);
	      trafficCmdNml->write(trafRotaryWingMsg);
	    }
	  break;
	} // switch (mode)
      } else {
	if (2 == axis) {
	  // this makes the right X joystick open and close the gripper
	  // linkIDs are 8,9 for one side, 10,11 for the other
	  if (values[axis] < -0.9) {

	    misCmdMove.linkCmd[0].linkID = 8;
	    misCmdMove.linkCmd[0].value = 1000;
	    misCmdMove.linkCmd[0].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd[1].linkID = 9;
	    misCmdMove.linkCmd[1].value = -1000;
	    misCmdMove.linkCmd[1].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd[2].linkID = 10;
	    misCmdMove.linkCmd[2].value = 1000;
	    misCmdMove.linkCmd[2].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd[3].linkID = 11;
	    misCmdMove.linkCmd[3].value = -1000;
	    misCmdMove.linkCmd[3].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd_length = 4;
	    misCmdMove.serial_number = NEXTS(servoMisSerial);
	    if( servoMisJACmdNml != NULL )
	      servoMisJACmdNml->write(misCmdMove);

	    esleep(0.1);
	  } else if (values[axis] > 0.9){

	    misCmdMove.linkCmd[0].linkID = 8;
	    misCmdMove.linkCmd[0].value = 0;
	    misCmdMove.linkCmd[0].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd[1].linkID = 9;
	    misCmdMove.linkCmd[1].value = 0;
	    misCmdMove.linkCmd[1].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd[2].linkID = 10;
	    misCmdMove.linkCmd[2].value = 0;
	    misCmdMove.linkCmd[2].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd[3].linkID = 11;
	    misCmdMove.linkCmd[3].value = 0;
	    misCmdMove.linkCmd[3].type = SERVO_MIS_JA_LINK_CMD_ABS_VALUE_TYPE;

	    misCmdMove.linkCmd_length = 4;
	    misCmdMove.serial_number = NEXTS(servoMisSerial);
	    if( servoMisJACmdNml != NULL )
	      servoMisJACmdNml->write(misCmdMove);

	    esleep(0.1);
	  }
	}

	if (COORD_JOINT == coordMode) {
	  if (whichJoint < 0) whichJoint = 0;
	  else if (whichJoint >= SERVO_MIS_JA_LINK_MAX) whichJoint = SERVO_MIS_JA_LINK_MAX - 1;
	  needToWrite = false;
	  // use only axis 0, the left joystick L/R, for joint moves
	  if (0 == axis) {
	    // zero all other joints in case they were moving and we switched
	    for (t = 0; t < link_number; t++) {
	      misCmdMove.linkCmd[t].linkID = t + 1;
	      if( misCmdMove.linkCmd[t].value != 0.0 )
		needToWrite = true;
	      misCmdMove.linkCmd[t].value = 0.0;
	      misCmdMove.linkCmd[t].type = SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE;
	    }
	    // now overwrite the selected joint's 0 with the speed
	    if( misCmdMove.linkCmd[whichJoint].value != values[0] )
	      needToWrite = true;
	    misCmdMove.linkCmd[whichJoint].value = values[0];
	    misCmdMove.linkCmd_length = t;
	    misCmdMove.serial_number = NEXTS(servoMisSerial);
	    if( needToWrite && servoMisJACmdNml != NULL )
	      servoMisJACmdNml->write(misCmdMove);
	  }
	} else {		// WORLD COORDINATE MOTION
	  go_vector vw[6];
	  go_vector joints[JOINTS];
	  // fill in the commanded vel
	  if (doRot) {
	    vw[3] = values[0];
	    vw[4] = values[1];
	    vw[5] = values[3];
	    vw[0] = vw[1] = vw[2] = 0.0;
	  } else {
	    vw[0] = values[0];
	    vw[1] = values[1];
	    vw[2] = values[3];
	    vw[3] = vw[4] = vw[5] = 0.0;
	  }

	  // run the inverse Jacobian
	  for (t = 0, tt = 0; t < link_number; t++, tt++) {
#ifdef INHIBIT_PRISMATIC
	    while (SERVO_MIS_JA_PRISMATIC_JOINT_TYPE == servoMisJASetPtr->linkSet[tt].jType) tt++;		// skip the prismatic joint
#endif
	    joints[t] = (go_real) servoMisJAStatPtr->linkStat[tt].jointVal;
	    go_link_joint_set(&link_params[t], joints[t], &linkout[t]);
	  }
	  compute_jfwd(linkout, link_number, &Jfwd, &T_L_0);
#ifdef DEBUG
	  go_rpy rpy;
	  go_quat_rpy_convert(&T_L_0.rot, &rpy);
	  printf("%f %f %f %f %f %f\n",
		 T_L_0.tran.x, T_L_0.tran.y, T_L_0.tran.z,
		 rpy.r, rpy.p, rpy.y);
#endif
	  if (toolMode) {
	    go_cart cart;
	    // rotate v
	    go_matrix_vector_mult(&RTool, &vw[0], &vw[0]);
	    go_vector_cart_convert(&vw[0], &cart);
	    go_quat_cart_mult(&T_L_0.rot, &cart, &cart);
	    go_cart_vector_convert(&cart, &vw[0]);
	    // rotate w
	    go_matrix_vector_mult(&RTool, &vw[3], &vw[3]);
	    go_vector_cart_convert(&vw[3], &cart);
	    go_quat_cart_mult(&T_L_0.rot, &cart, &cart);
	    go_cart_vector_convert(&cart, &vw[3]);
	  } else {
	    go_matrix_vector_mult(&RWorld, &vw[0], &vw[0]);
	    go_matrix_vector_mult(&RWorld, &vw[3], &vw[3]);
	  }
	  if (0 == compute_jinv(&Jfwd, &Jinv)) {
	    double vscale, vs;
	    go_matrix_vector_mult(&Jinv, vw, jointvels);
	    // scale back the jointvels so none exceeds its max
	    // and all are scaled equally
	    vscale = 1.0;
	    for (t = 0; t < link_number; t++) {
	      if (fabs(jointvels[t]) > maxVelocity[t]) {
		vs = maxVelocity[t] / fabs(jointvels[t]);
		if (vs < vscale) vscale = vs;
	      }
	    }
	    for (t = 0; t < link_number; t++) {
	      jointvels[t] *= vscale;
	    }
	    needToWrite = false;
	    for (t = 0, tt = 0; t < link_number; t++, tt++) {
#ifdef INHIBIT_PRISMATIC
	      while (SERVO_MIS_JA_PRISMATIC_JOINT_TYPE == servoMisJASetPtr->linkSet[tt].jType) {
		misCmdMove.linkCmd[tt].linkID = tt + 1;
		if( misCmdMove.linkCmd[tt].value != 0.0 )
		  needToWrite = true;
		misCmdMove.linkCmd[tt].value = 0.0;
		misCmdMove.linkCmd[tt].type = SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE;
		tt++;
	      }
#endif
	      misCmdMove.linkCmd[tt].linkID = tt + 1;
	      if( misCmdMove.linkCmd[tt].value != jointvels[t] )
		needToWrite = true;
	      misCmdMove.linkCmd[tt].value = jointvels[t];
	      misCmdMove.linkCmd[tt].type = SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE;
	    }
	    misCmdMove.linkCmd_length = tt;
	    misCmdMove.serial_number = NEXTS(servoMisSerial);
	    if( needToWrite == true && servoMisJACmdNml != NULL)
	      servoMisJACmdNml->write(misCmdMove);
	  } else {
	    printf("can't invert Jacobian\n");
	  }
	}
      }
      joystickData.mode = 
	doMob ? JOYSTICK_MODE_MOBILITY :
	coordMode == COORD_JOINT ? JOYSTICK_MODE_JOINT :
	toolMode ? JOYSTICK_MODE_TOOL : JOYSTICK_MODE_WORLD;
      if( joystickDataNml != NULL )
	joystickDataNml->write(joystickData);
    } // if (doInit)
  } // for (done)
}

int main(int argc, char *argv[])
{
  int joystickNum;
  double end;
  int gotIt;
  ServoMobJACmdInit servoMobInit;
  TrafficCtrlCmdInit trafficInit;
  char c;
  int i;
  SDL_Joystick *joystick;
  double d1, d2, d3;
  go_rpy rpy;
  

  userScale = false;
  button8Down = false;
  invertAxis = false;

  joystickNum = 0;

  mode = 1;
  vehID = 1;
  compID = 1;
  doInit = 1;

  skidMsg.wLeft = 0;
  skidMsg.wRight = 0;
  ackerMsg.velocity = 0;
  ackerMsg.steerAngle = 0;
  ackerMsg.crabAngle = 0;
  submarineMsg.propeller = 0;
  submarineMsg.rudder = 0;
  submarineMsg.sternPlane = 0;
  rotaryWingMsg.vVertical = 0;
  rotaryWingMsg.vLinear = 0;
  rotaryWingMsg.vLateral = 0;
  rotaryWingMsg.vRotation = 0;

  // If scale factors are still 0 after reading the parameters,
  // roboJoy will read the setting channel
  scale[0] = 0;
  scale[1] = 0;
  scale[2] = 0;
  scale[3] = 0;

  while ((c = getopt(argc, argv, ":j:d:m:c:x:y:z:w:W:T:ni")) != EOF) {
    switch (c) {
    case 'j':
      // set joystick number
      joystickNum = atoi(optarg);
      break;
    case 'd':
      // set vehicle id
      vehID = atoi(optarg);
      break;

      /* Now, the setting's channel is read to determine the mode of operation

      case 'm':
      // set operational mode
      mode = atoi(optarg);
      break;
      */

    case 'c':
      // set component id
      compID = atoi(optarg);
      break;
    case 'n':
      // skip initialization
      doInit = 0;
      break;
    case 'i':
      invertAxis = true;
      break;
    case 'x':
      // set scale[0]
      userScale = true;
      scale[0] = atof(optarg);
      break;
    case 'y':
      // set scale[1]
      userScale = true;
      scale[1] = atof(optarg);
      break;
    case 'z':
      // set scale[2]
      userScale = true;
      scale[2] = atof(optarg);
      break;
    case 'w':
      // set scale[3]
      userScale = true;
      scale[3] = atof(optarg);
      break;
    case 'W':
      // set world frame orientation
      if (3 != sscanf(optarg, "%lf %lf %lf", &d1, &d2, &d3)) {
	fprintf(stderr, "need roll pitch and yaw\n");
	return 1;
      } else {
	rpy.r = GO_TO_RAD(d1);
	rpy.p = GO_TO_RAD(d2);
	rpy.y = GO_TO_RAD(d3);
	go_rpy_quat_convert(&rpy, &QWorld);
      }
      break;
    case 'T':
      // set tool frame orientation
      if (3 != sscanf(optarg, "%lf %lf %lf", &d1, &d2, &d3)) {
	fprintf(stderr, "need roll pitch and yaw\n");
	return 1;
      } else {
	rpy.r = GO_TO_RAD(d1);
	rpy.p = GO_TO_RAD(d2);
	rpy.y = GO_TO_RAD(d3);
	go_rpy_quat_convert(&rpy, &QTool);
      }
      break;
    case ':':
      fprintf(stderr, "missing value for -%c\n", optopt);
      return 1;
      break;
    default:			/* '?' */
      fprintf (stderr, "unrecognized option -%c\n", optopt);
      return 1;
      break;
    }
  }
  if (optind < argc) {
    fprintf(stderr, "extra non-option characters: %s\n", argv[optind]);
    return 1;
  }

  misCmdMove.linkCmd_length = SERVO_MIS_JA_LINK_MAX;
  for (i = 0; i < SERVO_MIS_JA_LINK_MAX; i++) {
    misCmdMove.linkCmd[i].linkID = i + 1;
    misCmdMove.linkCmd[i].type = SERVO_MIS_JA_LINK_CMD_VELOCITY_TYPE;
  }

  if (doInit) {
    while (! openNML(vehID)) {
      printf("waiting for NML...\n");
      esleep(1.0);
    }

    // send init command to servo mob
    if( vehID != -1 )
      {
	servoMobInit.serial_number = NEXTS(servoMobSerial);
	servoMobJACmdNml->write(servoMobInit);
	for (end = etime() + 3.0, gotIt = 0; etime() < end; esleep(0.1)) {
	  servoMobJAStatNml->read();
	  if ( servoMobJAStatPtr->echo_serial_number == servoMobSerial &&
	       servoMobJAStatPtr->status == RCS_DONE) {
	    gotIt = 1;
	  }
	}
	if (!gotIt) {
	  fprintf(stderr, "can't initialize servo mob\n");
	  closeNML();
	  return 1;
	}

	servoMobJASetNml->read();
	switch (((ServoMobJASet *)servoMobJASetPtr)->steerType) {
	case MOAST_SKID_STEER_TYPE:
	  mode = 1;

	  if(scale[1]==0) // If scale factor 1 has not been set
	    scale[1] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxWheelRot;	// Scale for the left wheel spin speed
      
	  if(scale[3]==0) // If scale factor 3 has not been set
	    scale[3] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxWheelRot; // Scale for the right wheel spin speed
	  printf( "\n\nScales: x: %f y: %f z: %f w: %f\n\n",
		  scale[0], scale[1], scale[2], scale[3] );

	  break;

	case MOAST_ACKERMAN_STEER_TYPE:
	  mode = 2;

	  if(scale[1]==0) // If scale factor 1 has not been set
            scale[1] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxWheelRot; // Scale for the wheel spin speed

	  if(scale[0]==0) // If scale factor 0 has not been set
            scale[0] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxSteerAngle; // Scale for the front steer angle

	  if(scale[2]==0) // If scale factor 2 has not been set
            scale[2] = -((ServoMobJASetGrdVeh *)servoMobJASetPtr)->maxCrabAngle; // Scale for the rear steer angle
	  printf( "Scales: x: %f y: %f z: %f w: %f\n",
		  scale[0], scale[1], scale[2], scale[3] );
	  break;

	case MOAST_ROTARY_WING_STEER_TYPE:
	  mode = 4;

	  if(scale[1]==0)
            scale[1] = -((ServoMobJASetAirBot *)servoMobJASetPtr)->maxVertVel; // Scale for Altitude (up/down) Velocity
      
	  if(scale[3]==0)
	    scale[3] = -((ServoMobJASetAirBot *)servoMobJASetPtr)->maxLinVel;  // Scale for Linear (forward/backward) Velocity
	
	  if(scale[2]==0)
            scale[2] = ((ServoMobJASetAirBot *)servoMobJASetPtr)->maxLatVel; // Scale for Lateral (left/right) Velocity

	  if(scale[0]==0)
	    scale[0] = ((ServoMobJASetAirBot *)servoMobJASetPtr)->maxRotVel; // Scale for Rotational Velocity
	  break;

	default:
	  if(scale[0]==0) scale[0] = 1;
	  if(scale[1]==0) scale[1] = -1;
	  if(scale[2]==0) scale[2] = 1;
	  if(scale[3]==0) scale[3] = -1;
	  break;
	} 
      }
    else
      {
	trafficInit.serial_number = NEXTS(servoMobSerial);
	trafficCmdNml->write(trafficInit);
	
      }
  }				// doInit

  /* Initialize SDL (Note: video is required to start event loop) */
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0) {
    fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
    exit(1);
  }

#ifdef VERBOSE
  /* Print information about the joysticks */
  printf("There are %d joysticks attached\n", SDL_NumJoysticks());
  for (i = 0; i < SDL_NumJoysticks(); ++i) {
    const char * name = SDL_JoystickName(i);
    printf("Joystick %d: %s\n", i, name ? name : "Unknown Joystick");
  }
#endif

  joystick = SDL_JoystickOpen(joystickNum);
  if (joystick == NULL) {
    printf("Couldn't open joystick %d: %s\n", joystickNum, SDL_GetError());
  } else {
    WatchJoystick(joystick);	// the main loop
    SDL_JoystickClose(joystick);
  }

  SDL_QuitSubSystem(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK);

  return 0;
}

/*
  Modification history:

  $Log: robojoy.cc,v $
  Revision 1.34  2008/12/17 15:28:07  dr_steveb
  Made the max rotational velocity a variable

  Revision 1.33  2008/12/02 16:27:48  proctor
  Added ulapi to servoShell and the gocfg, wrap stuff for go motion at the
  mission servo level

  Revision 1.32  2008/05/27 14:42:17  proctor
  Added an ini file section for Matilda platform; changed a robotjoy typo

  Revision 1.31  2008/02/19 16:10:52  dr_steveb
  Limited rotational velocity for skid steered vehicles.

  Revision 1.30  2008/02/19 14:56:15  dr_steveb
  Fixed minor bugs

  Revision 1.29  2007/11/14 22:02:08  dr_steveb
  Added "i" flag to invert joystick values

  Revision 1.28  2007/10/17 17:18:43  dr_steveb
  Added ability to control multiple mission packages

  Revision 1.27  2007/08/29 18:28:34  dr_steveb
  Added ability for the joystick to control traffic vehicles

  Revision 1.26  2007/08/09 18:15:47  proctor
  Added define for INHIBIT_PRISMATIC to allow compile-time selection of
  whether you want to inhibit the movement of prismatic joints or not.
  Tested with prismatic joints enabled, with the Telemax, and it worked
  except for the problem in USARsim where the prismatic joint goes back to 0.
  Left the compile-time flag enabled so the default behavior is to inhibit
  prismatic joints, as before. We'll need to fix the prismatic joint
  behavior in USARsim.

  Revision 1.25  2007/08/09 17:54:05  proctor
  Fixed problem with singular pseudoinverse; now a 7-DOF robot should move
  all joints in Cartesian space based on minimum sum of joint speeds

  Revision 1.24  2007/08/03 18:24:33  dr_steveb
  Joystick no longer waits on mission package.

  Revision 1.23  2007/08/02 18:55:16  proctor
  Clamped joystick speeds that were below 0.1 times the axis scale to be zero,
  to stop drifting around non-centered joysticks.
  CVSCVS: ----------------------------------------------------------------------

  Revision 1.22  2007/07/26 17:56:14  proctor
  Took out nav data and servo SP, and added both mob and mis init, and added
  NML wait loop so you can bring this up before the controller

  Revision 1.21  2007/07/24 18:35:43  proctor
  Added joystickData NML buffer to robojoy.cc

  Revision 1.20  2007/07/18 18:53:07  proctor
  Made ../etc/moast.nml the default NML file

  Revision 1.19  2007/06/28 13:21:44  proctor
  Changed usleep to esleep, switching from Unix-specific to the one in RCS lib

  Revision 1.18  2007/06/14 23:35:27  webenji
  Added control for the AirRobot.
  robojoy.cc now reads in the setting's channel to determine the mode of
  operation as well as the necessary scale factors (if they have not been
  defined in the parameters).

  Revision 1.17  2007/06/12 20:31:18  dr_steveb
  Fixed arm drift

  Revision 1.16  2007/06/08 16:14:00  proctor
  Added flipper control to robojoy.

  Revision 1.15  2007/05/30 13:45:34  proctor
  Handled the dynamic array of joints better, limiting the joints to only
  those moving, and stopping any moving joints when switching joints.

  Revision 1.14  2007/05/29 13:42:36  proctor
  Kept gripper fingers parallel

  Revision 1.13  2007/05/25 20:02:17  proctor
  Added gripper open/close

  Revision 1.12  2007/05/09 15:19:08  proctor
  Added the math source files from go motion, and changed the Makefile.am
  so that robojoy links these in.

  Revision 1.11  2007/05/08 21:23:09  proctor
  Added entries for navDataExt in NML config files, for joystick; set some
  defaults in robojoy.cc so you don't have to pass arcane command line options
  in the usual case.

  Revision 1.10  2007/05/07 19:18:47  proctor
  Combined mobility joystick joySDL into this one, so we only need one joystick program.
  Still need to add a gravity vector world mode.

  Revision 1.9  2007/05/07 15:38:32  proctor
  Moved motion computations and NML reading, writing out of joystick event
  loop so they are sent continually using the up-to-date Jacobian; added
  world and tool motion, and orientation control

  Revision 1.8  2007/05/03 20:26:11  proctor
  Fixed index bug where prismatic joints bypassed real joints, and added better
  setting of pose parameters when filling in joints

  Revision 1.7  2007/05/03 15:35:55  proctor
  Added go_pp parameters type instead of go_dh, since DH params don't follow same
  convention as arbitrary pose representation of link-to-link transforms

  Revision 1.6  2007/05/02 21:17:14  proctor
  Scaled speeds to max speed, equally

  Revision 1.5  2007/05/01 12:53:52  proctor
  Added jacobian-based teleop

  Revision 1.4  2007/04/27 12:57:18  proctor
  Stripped out old test code that prevented compile

  Revision 1.3  2007/04/06 19:11:11  proctor
  In servoShell.cc, changed printfs with %lf to %f, since there is no %lf
  in printf; in simpleSimInfSkin.cc, cast the bot type to ground vehicle
  so that the extra fields got copied out; robojoy.cc changes sign and uses
  offsets to match puma kinematics more closely

  Revision 1.2  2007/04/04 20:39:26  proctor
  Split out world and joint mode, still need to put in inverse jacobian

  Revision 1.1  2007/04/04 19:01:25  proctor
  Added robojoy.cc, robot joystick via SDL, cloned from joySDL.cc and some
  of servoShell.cc.
*/
