/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*!
  \file spPlay.cc
  \brief Reads data log file in "TCHANG" format and pushes into the
  1-D SP, NavDataExt buffers.

  The TCHANG format goes:

  id
  tranrel_x tranrel_y tranrel_z
  tranabs_x tranabs_y tranabs_z
  roll pitch yaw [abs?]
  time
  rot.s     rot.x     rot.y     rot.z [rel?]
  rotVel.s  rotVel.x  rotVel.y  rotVel.z [rel?]
  rpyRate.x rpyRate.y rpyRate.z [abs?]
  speed
  leftCurb_idx rightCurb_idx
  x1 y1 z1 x2 y2 z2 ... x361 y361 z361

  where xyz are in hmmwv or local frame (INS, 0,0,0 is start of
  engine).  Curb indexes start with 1, 0 means invalid no curb
  detected. Output range data > 60 [m] are considered invalid.


  \code CVS Status:
  $Author: proctor $
  $Revision: 1.5 $
  $Date: 2006/12/12 19:20:40 $
  \endcode

  \author Fred Proctor
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>		// sprintf()
#include <stdlib.h>		// getenv(), atoi,f()
#include <stddef.h>		// NULL
#include <signal.h>		// SIGINT, signal()
#include <limits.h>		// INT_MAX
#include <time.h>		// struct timespec, nanosleep()
#include <math.h>		// sqrt
#include <ctype.h>		// isspace()
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>		// RCS_CMD,STAT_CHANNEL, rcs_print()
#include "moastTypes.hh"	// MOAST_NML_BUFFER_NAME_LEN
#include "moastNmlOffsets.hh"
#include "navDataExt.hh"	// NavDataExt etc
#include "sensorData.hh"	// SensorData1D

#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_INI_FILE "../etc/moast.ini"

#define SP_PROCESS_NAME "servoSP"
#define MOB_PROCESS_NAME "servoMob"

typedef struct {
  NavDataExt nav;
  SensorData1D sensor;
} SPData;

double getOriginalRange(SPData * data, double x, double y, double z)
{
  // We don't yet know about the sensor's location, so we'll set it to 0
  // The sensor's location on the vehicle
  PM_CARTESIAN senTran(0,0,0);

  // For rotations we convert rpy to quaternion for ease of processing
  // The sensor's default rotation on the vehicle (at '0 degrees')
  PM_QUATERNION senQuat = PM_RPY(0,0,0);
  // The vehicle's rotation in the world
  PM_QUATERNION vehQuat = data->nav.rpyRel;

  // Tommy pre-translated coordinates in centimeters
  PM_CARTESIAN objTran(x,y,z);
  // Convert them to meters to match our processing units
  objTran = objTran * 0.01;

  // Convert the coordinates to vehicle-relative
  objTran = (objTran - data->nav.tranRel) * (-1 * vehQuat);
  // Convert the coordinates to sensor-relative
  objTran = (objTran - senTran) * (-1 * senQuat);

  // Return the cartesian distance as the real range
  return sqrt((objTran.x)*(objTran.x) +
	      (objTran.y)*(objTran.y) +
	      (objTran.z)*(objTran.z));
}

static int parseSPData(char * buffer, SPData * data)
{
  char * ptr = buffer;
  double d;
  PM_QUATERNION quat;

#define SKIPWHITE while (isspace(*ptr)) ptr++; if (0 == *ptr) return 1
#define SKIPBLACK while (!isspace(*ptr) && 0 != *ptr) ptr++; if (0 == *ptr) return 1

  SKIPWHITE;
  d = atof(ptr);		// id

  SKIPBLACK; SKIPWHITE;
  data->nav.tranRel.x = atof(ptr); // tranrel_x
  SKIPBLACK; SKIPWHITE;
  data->nav.tranRel.y = atof(ptr); // tranrel_y
  SKIPBLACK; SKIPWHITE;
  data->nav.tranRel.z = atof(ptr); // tranrel_z

  SKIPBLACK; SKIPWHITE;
  data->nav.tranAbs.x = atof(ptr); // tranabs_x
  SKIPBLACK; SKIPWHITE;		
  data->nav.tranAbs.y = atof(ptr); // tranabs_y
  SKIPBLACK; SKIPWHITE;
  data->nav.tranAbs.z = atof(ptr); // tranabs_z

  SKIPBLACK; SKIPWHITE;
  data->nav.rpyAbs.r = atof(ptr); // roll [abs]
  SKIPBLACK; SKIPWHITE;
  data->nav.rpyAbs.p = atof(ptr); // pitch
  SKIPBLACK; SKIPWHITE;
  data->nav.rpyAbs.y = atof(ptr); // yaw

  SKIPBLACK; SKIPWHITE;
  data->nav.time = atoi(ptr);	// time

  SKIPBLACK; SKIPWHITE;
  quat.s = atof(ptr);		// rot.s
  SKIPBLACK; SKIPWHITE;
  quat.x = atof(ptr);		// rot.x
  SKIPBLACK; SKIPWHITE;
  quat.y = atof(ptr);		// rot.y
  SKIPBLACK; SKIPWHITE;
  quat.z = atof(ptr);		// rot.z
  data->nav.rpyRel = quat;	// rpy, [rel]

  SKIPBLACK; SKIPWHITE;
  quat.s = atof(ptr);		// rotVel.s
  SKIPBLACK; SKIPWHITE;
  quat.x = atof(ptr);		// rotVel.x
  SKIPBLACK; SKIPWHITE;
  quat.y = atof(ptr);		// rotVel.y
  SKIPBLACK; SKIPWHITE;
  quat.z = atof(ptr);		// rotVel.z
  data->nav.rpyRelRate = quat;	// rpy rate, [rel]

  SKIPBLACK; SKIPWHITE;
  data->nav.rpyAbsRate.r = atof(ptr); // rpyRate.x, [abs]
  SKIPBLACK; SKIPWHITE;
  data->nav.rpyAbsRate.p = atof(ptr); // rpyRate.y
  SKIPBLACK; SKIPWHITE;
  data->nav.rpyAbsRate.y = atof(ptr); // rpyRate.z

  SKIPBLACK; SKIPWHITE;
  d = atof(ptr);		// speed
  SKIPBLACK; SKIPWHITE;
  d = atof(ptr);		// leftCurbIdX
  SKIPBLACK; SKIPWHITE;
  d = atof(ptr);		// rightCurbIdX

  // now for the 3 X 361 sensor entries, XYZ, but we'll only do
  // do up to MAX_SENSOR_DATA_1D
  for (int t = 0; t < MAX_SENSOR_DATA_1D; t++) {
    double x, y, z;
    SKIPBLACK; SKIPWHITE;
    x = atof(ptr);
    SKIPBLACK; SKIPWHITE;
    y = atof(ptr);
    SKIPBLACK; SKIPWHITE;
    z = atof(ptr);
    data->sensor.featArray[t] = getOriginalRange(data, x, y, z);
  }
  data->sensor.featArray_length = MAX_SENSOR_DATA_1D;

  return 0;

#undef SKIPWHITE
#undef SKIPBLACK
}

static int done = 0;
static void quit(int sig)
{
  done = 1;
}

int main(int argc, char *argv[])
{
  char nml_file_default[] = DEFAULT_NML_FILE;
  char *nml_file_env;
  char *NML_FILE = NULL;
  char ini_file_default[] = DEFAULT_INI_FILE;
  char *ini_file_env;
  char *INI_FILE = NULL;
  char *DATA_FILE = NULL;
  int bufnum = 1;
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  NML * navDataExtBuf;
  NML * sensorDataBuf;
  FILE * fp;
  enum {BUFFERLEN = 16384};
  char buffer[BUFFERLEN];
  SPData data1;			// ping-ping buffers
  SPData data2;
  SPData *dataNow;		// and their pointers
  SPData *dataNext;
  int delta_t_msec;		// time between now and next, in msec
  struct timespec ts;

  // read command line args
  while (1) {
    char option = getopt(argc, argv, "n:i:d:f:");
    if (option == -1)
      break;

    switch (option) {
    case 'n':
      NML_FILE = optarg;
      rcs_print("%s: using specified NML file %s\n", argv[0], NML_FILE);
      break;
    case 'i':
      INI_FILE = optarg;
      rcs_print("%s: using specified INI file %s\n", argv[0], INI_FILE);
      break;
    case 'd':
      if (1 != sscanf(optarg, "%i", &bufnum)) {
	rcs_print_error("%s: bad value for buffer number: %s\n", argv[0], optarg);
	return 1;
      }
      rcs_print("%s: using specified buffer suffix %d\n", argv[0], bufnum);
      break;
    case 'f':
      DATA_FILE = optarg;
      rcs_print("%s: using specified data file %s\n", argv[0], DATA_FILE);
      break;
    default:
      rcs_print_error("%s: bad argument\n", argv[0]);
      return 1;
      break;
    }
  }

  // get the config file name from arg, if provided, next environment,
  // if there, else default
  if (NULL == NML_FILE) {
    nml_file_env = getenv("CONFIG_NML");
    if (NULL != nml_file_env) {
      NML_FILE = nml_file_env;
      rcs_print("%s: using environment NML file %s\n", argv[0], NML_FILE);
    } else {
      NML_FILE = nml_file_default;
      rcs_print("%s: using default NML file %s\n", argv[0], NML_FILE);
    }
  }

  // ditto for the .ini file
  if (NULL == INI_FILE) {
    ini_file_env = getenv("CONFIG_INI");
    if (NULL != ini_file_env) {
      INI_FILE = ini_file_env;
      rcs_print("%s: using environment INI file %s\n", argv[0], INI_FILE);
    } else {
      INI_FILE = ini_file_default;
      rcs_print("%s: using default INI file %s\n", argv[0], INI_FILE);
    }
  }

  // set file pointer to argument, if provided, or stdin, if not
  if (NULL == DATA_FILE) {
    fp = stdin;
  } else {
    if (NULL == (fp = fopen(DATA_FILE, "r"))) {
      rcs_print("%s: can't open data file %s\n", argv[0], DATA_FILE);
      return 1;
    }
  } 

  // open NavDataExt
  sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
  navDataExtBuf =
    new NML(navDataExt_format, chanName, MOB_PROCESS_NAME, NML_FILE);
  if (!navDataExtBuf->valid()) {
    delete navDataExtBuf;
    navDataExtBuf = 0;
    return 1;
  }

  // open sensor
  //sprintf(chanName, "%s%d", SERVO_SP_LINESCAN_BUF_NAME, bufnum);
  sprintf(chanName, "%s%d", SERVO_SP_LINESCAN_BUF_NAME, bufnum);
  sensorDataBuf =
    new NML(sensorData_format, chanName, SP_PROCESS_NAME, NML_FILE);
  if (!sensorDataBuf->valid()) {
    delete sensorDataBuf;
    sensorDataBuf = 0;
    return 1;
  }

  dataNow = &data1;
  dataNext = &data2;

  // read the first line
  if (NULL == fgets(buffer, BUFFERLEN, fp) ||
      0 != parseSPData(buffer, dataNow)) {
    rcs_print_error("bad input line\n");
  } else {
    signal(SIGINT, quit);
    done = 0;
    while (! feof(fp) &&
	   ! done) {
      if (NULL == fgets(buffer, BUFFERLEN, fp)) {
	break;
      }

      if (0 != parseSPData(buffer, dataNext)) {
	rcs_print_error("bad input line\n");
	break;
      }

      // write out the NML buffers
      navDataExtBuf->write(dataNow->nav);
      sensorDataBuf->write(dataNow->sensor);

      // figure out how long we need to sleep
      delta_t_msec = (int) (dataNext->nav.time - dataNow->nav.time);
      if (delta_t_msec < 0) {
	delta_t_msec += INT_MAX;
      }
      ts.tv_sec = delta_t_msec / 1000;
      ts.tv_nsec = (delta_t_msec % 1000) * 1000000;

      // swap ping pong pointers
      dataNow = (dataNow == &data1 ? &data2 : &data1);
      dataNext = (dataNext == &data1 ? &data2 : &data1);

      // and sleep
      nanosleep(&ts, NULL);
    }
  }

  // we're done
  delete(navDataExtBuf);
  delete(sensorDataBuf);
  fclose(fp);

  return 0;
}

/*
  Modification history:

  $Log: spPlay.cc,v $
  Revision 1.5  2006/12/12 19:20:40  proctor
  Fixed miscellaneous outdated code to reflect simware

  Revision 1.4  2006/11/19 20:30:55  cj_scrapper
  Changed servo linescan macro back, until simWare committed

  Revision 1.3  2006/11/19 20:28:30  cj_scrapper
  Changed Servo linescan buffer macro

  Revision 1.2  2006/04/13 14:30:10  proctor
  Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
  primMisMain.cc.
  Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
  Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
  Added handling of servoSPCmdGo in simpleSPSim.cc.
  Fixed omitted return value in timeout_callback in callbacks.c, and added
  a prototype for setTimerCallback.

  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.5  2005/07/18 21:16:56  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.4  2005/06/24 17:46:10  proctor
  Changed a .data to a .featArray to match the new sensorData.hh decls

  Revision 1.3  2005/06/24 12:57:54  beytin
  Fixed the translation of Tommy's xyz data to conform with
  sensor-local coordinates

  Revision 1.2  2005/06/23 22:04:14  proctor
  Changed 1D sensor data to the style declared in sensorData.hh

  Revision 1.1  2005/06/22 19:05:17  proctor
  Initial checkin

*/
