/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/* program to convert spqr data channels to standard moast channels for use
   by the section level
*/
/*!
  \code CVS Status:
  $Author: dr_steveb $
  $Revision: 1.2 $
  $Date: 2007/12/17 21:55:48 $
  \endcode

  \author Stephen Balakirsky
  \date   10/2006
*/
#include <stdlib.h> // getenv
#include <stdio.h> // sprintf, stderr, ...
#include <unistd.h> // getopt
#include "vehMobPL.hh"
#include "nmlMap.hh"
#include "nmlPosition.hh"
#include "navDataExt.hh"
#include "sensorData.hh"
#include "../vehicle/vehMobCell.hh"
#define MY_NAME_BASE "spqrConvert"

//! SPQR Mappipng channel
NML              *spqrMapInNml;
NML              *spqrPositionOutNml;
NML              *spqrPositionInNml;
//! veh planning channels
RCS_CMD_CHANNEL  *vehMobPLCmdNml;
RCS_STAT_CHANNEL *vehMobPLStatNml;
RCS_CMD_CHANNEL  *vehMobPLCfgNml;
RCS_STAT_CHANNEL *vehMobPLSetNml;
NML              *vehMapOutNml;
//! navigation data channel
NML              *navDataExtNml;
//! goal point
PM_CARTESIAN spqrGoal;

///////////////////////////////////////////////////////////////////////////////////
void readCmd( RCS_CMD_CHANNEL *cmdIn, VehMobPLStat *currentStat )
{
  VehMobPLCmdMoveWaypoint *waypointCmd;
  VehMobPLCmdExplore *exploreCmd;
  int type, serial_number;
  static int  noPrint = 0;


  type = cmdIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_CMD_ABORT_TYPE:
    case VEH_MOB_PL_CMD_HALT_TYPE:
    case VEH_MOB_PL_CMD_INIT_TYPE:
    case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
    case VEH_MOB_PL_CMD_NOP_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
      waypointCmd = (VehMobPLCmdMoveWaypoint*)cmdIn->get_address();
      serial_number = waypointCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  spqrGoal = waypointCmd->p;
	}
      break;
    case VEH_MOB_PL_CMD_EXPLORE_TYPE:
      exploreCmd = (VehMobPLCmdExplore*)cmdIn->get_address();
      serial_number = exploreCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  spqrGoal = exploreCmd->p;
	}
      break;
    default:
      if( noPrint < 10 )
	{
	  fprintf( stdout, "Command %s not yet supported\n", vehMobPL_symbol_lookup(type) );
	  noPrint++;
	}
      break;
    }
  return;
}


///////////////////////////////////////////////////////////////////////////////////
/*!
  Read in spqr obstacles and populate world model

\param obsChan The channel to read from

\return Obstacle message on success 
        NULL on failure
*/
MapMsg *readSPQRMap( NML *obsChan )
{
  int type;

  type = obsChan->read();
  switch( type )
    {
    case 0: // no new data
      return(NULL);
    case MAP_MSG_TYPE:
      return((MapMsg*)obsChan->get_address());
    default:
      fprintf( stderr, "SPQR map channel error\n" );
      return(NULL);
    }
}

///////////////////////////////////////////////////////////////////////////////////
bool openNML( int vehID )
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];
  VehMobPLCmdNop *nopCmd;
  VehMobPLCfgNop *nopCfg;

  sprintf( myName, "%s", MY_NAME_BASE );

  if(!configNml)
    {
      fprintf(stderr,"Please set CONFIG_NML environment variable.\n");
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CMD_NAME, vehID );
  vehMobPLCmdNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
				      myName, configNml );
  if( NULL == vehMobPLCmdNml || !vehMobPLCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_STAT_NAME, vehID );
  vehMobPLStatNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				       myName, configNml );
  if( NULL == vehMobPLStatNml || !vehMobPLStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_SET_NAME, vehID );
  vehMobPLSetNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				     myName, configNml );
  if( NULL == vehMobPLSetNml || !vehMobPLSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CFG_NAME, vehID );
  vehMobPLCfgNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
					myName, configNml );
  if( NULL == vehMobPLCfgNml || !vehMobPLCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  spqrMapInNml = new NML( nmlMap_format, "MapBufP2DX",
			  myName, configNml );
  if( NULL == spqrMapInNml || !spqrMapInNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, "MapBufP2DX" );
      // not a critical problem, so keep going
    }

  spqrPositionOutNml = new NML( nmlPosition_format, NML_POS_VICTIM_NAME,
			   myName, configNml );
  if( NULL == spqrPositionOutNml || !spqrPositionOutNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, "MapBufZerg" );
      // not a critical problem, so keep going
    }

  spqrPositionInNml = new NML( nmlPosition_format, NML_POS_MAPPER_NAME,
			   myName, configNml );
  if( NULL == spqrPositionInNml || !spqrPositionInNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, "MapBufZerg" );
      // not a critical problem, so keep going
    }

  sprintf( chanName, "%s%d", VEH_MOB_MAP_NAME, vehID );
  vehMapOutNml = new NML( vehMobPL_format, chanName,
				     myName, configNml );
  if( NULL == vehMapOutNml || !vehMapOutNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", NAV_DATA_EXT_NAME, vehID );
  navDataExtNml = new NML( navDataExt_format, chanName,
			   myName, configNml );
  if( NULL == navDataExtNml || !navDataExtNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, chanName );
      return(false);
    }

  nml_start();
  // stuff a nop
  nopCmd = new VehMobPLCmdNop();
  nopCfg = new VehMobPLCfgNop();
  nopCmd->serial_number = 100;
  nopCfg->serial_number = 100;

  vehMobPLCmdNml->write(nopCmd);
  vehMobPLCfgNml->write(nopCfg);
  delete nopCmd;
  delete nopCfg;
  return(true);

}
///////////////////////////////////////////////////////////////////////////////////
void outputMap( NML *channel, MapMsg *spqrMap )
{
  SensorData2D vehMapOut;
  int countRow, countCol, spqrLocation, moastLocation;

  vehMapOut.incrementX = spqrMap->mapRealWidth/spqrMap->nCols;
  vehMapOut.incrementY = vehMapOut.incrementX;
  vehMapOut.x = spqrMap->nRows;
  vehMapOut.y = spqrMap->nCols;
  vehMapOut.featArray_length = spqrMap->map_length;
  for(countRow=0; countRow<(int)spqrMap->nRows; countRow++ )
    {
      for(countCol=0; countCol<(int)spqrMap->nCols; countCol++ )
	{
	  spqrLocation = countRow*spqrMap->nCols + countCol;
	  moastLocation = countCol*spqrMap->nRows + countRow;
	  if( spqrMap->map[spqrLocation] == C8Black )
	    vehMapOut.featArray[moastLocation] = VEH_WM_DANGEROUS_MAX;
	  else if( spqrMap->map[spqrLocation] == C8Blue )
	    vehMapOut.featArray[moastLocation] = VEH_WM_UNKNOWN;
	  else if( spqrMap->map[spqrLocation] == C8White )
	    vehMapOut.featArray[moastLocation] = 0; // not obstacle
	  else // should not happen
	    vehMapOut.featArray[moastLocation] = VEH_WM_DIFFICULT_MAX;
	}
    }
  channel->write(vehMapOut);
}

///////////////////////////////////////////////////////////////////////////////////
void readSPQRPos( NavDataExt *navData, NML *channelIn )
{
  int type;
  PositionMsg *posMsg;

  type = channelIn->read();
  switch( type )
    {
    case 0: // no new data
      return;
    case POSITION_MSG_TYPE:
      posMsg = (PositionMsg *)channelIn->get_address();
      navData->tranAbs.x = posMsg->x;
      navData->tranAbs.y = posMsg->y;
      navData->tranRel = navData->tranAbs;
      navData->rpyAbs.y = posMsg->theta;
      navData->rpyAbs.p = 0;
      navData->rpyAbs.r = 0;
      navData->rpyRel = navData->rpyAbs;
      break;
    default:
      fprintf( stderr, "SPQR obstacle channel error\n" );
      fprintf( stdout, "Command %s not yet supported\n", sensorData_symbol_lookup(type) );
      return;
    }
  return;
}


///////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
  char c;
  int vehID = 2;
  int myVerbose = 0;
  double cycleTime = .2;
  RCS_TIMER *timer;
  VehMobPLStat *vehMobPLStat;
  VehMobPLSet  *vehMobPLSet;
  MapMsg *spqrMap = NULL;
  NavDataExt *navData;
  PositionMsg posMsg;


  navData = new NavDataExt();
  while((c=getopt(argc,argv,"i:v"))!=EOF) 
    {
      switch(c) 
	{
	case 'v':
	  myVerbose = 1;
	  break;
	case 'i':
	  vehID = atoi(optarg);
	  break;
	case 'c':
	  cycleTime = atof(optarg);
	  break;
	}
    }

  if( !openNML(vehID) )
    {
      printf( "%s: Error on open, exiting!\n", argv[0] );
      exit(1);
    }
  timer = new RCS_TIMER(cycleTime);

  // send out first settings message
  vehMobPLSet = new VehMobPLSet();
  //  vehMobPLSet->command_type = VEH_MOB_PL_CFG_DUMP_WM_TYPE;
  vehMobPLSet->command_type = VEH_MOB_PL_CFG_NOP_TYPE;
  vehMobPLSet->state = NEW_COMMAND;
  vehMobPLSet->cycleTime = cycleTime;
  vehMobPLSet->obstaclePad = 0;
  vehMobPLSet->usePathDiscount = 0;
  vehMobPLSet->debug = myVerbose;
  vehMobPLSet->isInitialized = false;
  vehMobPLSet->isUsingSQL = false;
  vehMobPLSet->confDistance = 0;
  vehMobPLSet->cellResolution = 0;
  vehMobPLSet->zeroConf = 0;
  vehMobPLSet->fiveConf = 0;
  strcpy(vehMobPLSet->wmFileName, "VehMobPLDispOutput.ppm" );
  vehMobPLSet->fileSkip = 5;
  vehMobPLSet->echo_serial_number = -1;
  vehMobPLSetNml->write(vehMobPLSet);

  // send out first status message
  vehMobPLStat = new VehMobPLStat();
  vehMobPLStat->echo_serial_number = -1;
  vehMobPLStat->state = NEW_COMMAND;
  vehMobPLStat->status = RCS_EXEC;
  vehMobPLStat->command_type = VEH_MOB_PL_CMD_NOP_TYPE;
  vehMobPLStat->planStartHorizon = 4.5;
  vehMobPLStatNml->write(vehMobPLStat);

  for(;;)
    {
      // read vehicle command
      readCmd(vehMobPLCmdNml, vehMobPLStat);
      // read spqr map
      spqrMap = readSPQRMap(spqrMapInNml);
      // read spqr position
      readSPQRPos( navData, spqrPositionInNml );

      // process vehicle command
      switch( vehMobPLStat->command_type )
	{
	case VEH_MOB_PL_CMD_NOP_TYPE:
	  status_next(vehMobPLStat, RCS_DONE);
	  break;
	case VEH_MOB_PL_CMD_ABORT_TYPE:
	  status_next(vehMobPLStat, RCS_DONE);
	  break;
	case VEH_MOB_PL_CMD_HALT_TYPE:
	  status_next(vehMobPLStat, RCS_DONE);
	  break;
	case VEH_MOB_PL_CMD_INIT_TYPE:
	  vehMobPLSet->isInitialized = true;
	  status_next(vehMobPLStat, RCS_DONE);
	  vehMobPLSetNml->write(vehMobPLSet);
	  break;
	case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
	  vehMobPLSet->isInitialized = false;
	  status_next(vehMobPLStat, RCS_DONE);
	  vehMobPLStatNml->write(vehMobPLStat);
	  break;
	case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
	  if( state_match(vehMobPLStat, NEW_COMMAND) )
	    {
	      posMsg.x = spqrGoal.x;
	      posMsg.y = spqrGoal.y;
	      posMsg.theta = 0;
	      spqrPositionOutNml->write( posMsg );
	    }
	  status_next(vehMobPLStat, RCS_DONE);
	  break;
	case VEH_MOB_PL_CMD_EXPLORE_TYPE:
	  if( state_match(vehMobPLStat, NEW_COMMAND) )
	    {
	      posMsg.x = spqrGoal.x;
	      posMsg.y = spqrGoal.y;
	      posMsg.theta = 0;
	      spqrPositionOutNml->write( posMsg );
	    }
	  status_next(vehMobPLStat, RCS_DONE);
	  break;
	}

      // output moast veh map
      if( spqrMap != NULL )
	outputMap( vehMapOutNml, spqrMap );
      // output navdata
      navData->time = vehMobPLStat->tt.count;
      navDataExtNml->write(navData);
      // output status
      vehMobPLStatNml->write(vehMobPLStat);
      // output settings
      vehMobPLSetNml->write(vehMobPLSet);


      cycle_time_tracker(&vehMobPLStat->tt);
      timer->wait();
    }
}
