/*****************************************************************************
  DISCLAIMER:
  This software was produced by the National Institute of Standards
  and Technology (NIST), an agency of the U.S. government, and by statute is
  not subject to copyright in the United States.  Recipients of this software
  assume all responsibility associated with its operation, modification,
  maintenance, and subsequent redistribution.

  See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/
/*!
  \file   vehMobPLMain.cc
  \brief Provides a very basic grid based planner for use in the moast
  system.

  This planner operates on a grid-graph with the graph nodes occupying
  the center of each world model grid cell. The costs of world model
  grid are assumed to be the cost per unit travel through the
  area. Edge costs are computed as a distance traveled through each
  grid cell times the cost of the cell. The program may be configured
  to use various graph search techniques.

  Input parameters include:
  

  \code CVS Status:
  $Author: root $
  $Revision: 1.1.1.1 $
  $Date: 2005/10/12 20:50:32 $
  \endcode

  \author Stephen Balakirsky
  \date   05/17/2005
*/
#include <stdio.h>
#include <strings.h>
#include <iostream>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <rcs.hh>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "moastNmlOffsets.hh"
#include "moastTypes.hh"
#include "sensorData.hh"
#include "navDataExt.hh"
#include "servoSP.hh"
#include "primMobJA.hh"
#include "vehMobPL.hh"
#include "nodeInfo.hh"
#include "dijkTabs.hh"
#include "vehMobPLEngine.hh"
#if HAVE_MYSQL
#include <mysql.h>
#include "moastMySQL.hh"
#endif

#define MY_NAME_BASE "vehMobPL"

//! my verbose level
int myVerbose;
#if HAVE_MYSQL
//! connection to mySql. Only exists if we have  HAVE_MYSQL defined
MoastMySQL sqlConnect;
#endif
//! password to use for sql access
char sqlPassword[MOAST_FILE_NAME_LEN];

//! Should the map be output
bool drawMap;
//! How many planning cycles should be skipped between output cycles
int drawSkip;
//! is the goal global or local?
bool goalIsGlobal;
//! is the map global or local?
bool mapIsGlobal;
//! Transforms from local to global and back
PM_HOMOGENEOUS homLocal, homGlobal;
//! Serial number to prim level
int myPrimSerial;
// NML Channels
//! prim planning channels
RCS_CMD_CHANNEL  *primMobJACmdNml;
RCS_STAT_CHANNEL *primMobJAStatNml;
RCS_CMD_CHANNEL  *primMobJASetNml;
RCS_STAT_CHANNEL *primMobJACfgNml;
//! servo SP channels
RCS_CMD_CHANNEL  *servoSPCmdNml;
RCS_STAT_CHANNEL *servoSPStatNml;
RCS_CMD_CHANNEL  *servoSPSetNml;
RCS_STAT_CHANNEL *servoSPCfgNml;
//! servo sensor data channel
NML              *sensorDataNml;
//! navigation data channel
NML              *navDataExtNml;
//! veh planning channels
RCS_CMD_CHANNEL  *vehMobPLCmdNml;
RCS_STAT_CHANNEL *vehMobPLStatNml;
RCS_CMD_CHANNEL  *vehMobPLCfgNml;
RCS_STAT_CHANNEL *vehMobPLSetNml;

// global commands received or sent
PrimMobJACmdMoveWaypoint *moveWaypointCmd;

// important positions to remember
//! location of goal in local coordinates
PM_CARTESIAN localGoal;
//! location of goal in global coordinates
PM_CARTESIAN globalGoal;
//! where are we anyway?
PM_CARTESIAN vehLocation;

/*!
  Transform a point from/to global coordinates to/from local coordinates

\param pointIn The point to be converted
\param transformGlobalToLocal True- transform to local, false- transform to global
\return Transformed point
*/
PM_CARTESIAN transformPoint(PM_CARTESIAN &pointIn, bool transformGlobalToLocal)
{
  PM_HOMOGENEOUS homTransform;
  PM_HOMOGENEOUS homIn, homOut;
  PM_CARTESIAN pointReturn;

  homIn.tran.x = pointIn.x;
  homIn.tran.y = pointIn.y;
  homIn.tran.z = pointIn.z;
  homIn.rot = PM_RPY(0, 0, 0);

  if( transformGlobalToLocal )
    {
      homTransform = homLocal * inv(homGlobal);
    }
  else
    {
      homTransform = homGlobal * inv(homLocal);
    }
  homOut = homTransform * homIn;
  pointReturn.x = homOut.tran.x;
  pointReturn.y = homOut.tran.y;
  pointReturn.z = homOut.tran.z;

  return(pointReturn);
}

/*!
  Compute a simple cofidence measure on the the maximum cost of the path.
  If path goes through an obstacle, the quality is low. 
  If path goes near an obstacle, the quality is ok.
  If path avoids all obstacles, the quality is great

\param maxCost The maximum cost of the path
\param obsCost The cost of an obstacle
\param proxCost The cost of being near an obstacle

\return 0 - low quality
\return 5 - ok quality
\return 10 - great quality
*/
int computeConfidence( double maxCost, double obsCost, double proxCost )
{
  //  fprintf( stderr, "Max Cost is: %lf\n", maxCost );
  if( maxCost >= obsCost )
    return(0);
  else if( maxCost >= proxCost )
    return(5);
  else
    return(10);
}

/*!
  Check the status back from subordinate planning system
  
  \param stat The status that was returned
  \param serialNumber The serial number of the command this status is for.

  \return true on good read
  \return false on no new satus
*/
bool checkPrimCmdStatus( RCS_STATUS &stat, int &echo_serial_number )
{
  int type;

  type = primMobJAStatNml->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case PRIM_MOB_JA_STAT_TYPE:
      stat = primMobJAStatNml->get_address()->status;
      echo_serial_number = primMobJAStatNml->get_address()->echo_serial_number;
      return(true);
    default:
      break;
    }
  return false;
}

/*!
  Read in new vehicle position

  \param statIn Channel to read from

  \return true on success
  \return false on failure
*/
bool updatePosition( NML *statIn )
{
  int type;
  NavDataExt *navData;

  type = statIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case NAV_DATA_EXT_TYPE:
      navData = (NavDataExt*)statIn->get_address();
      if( mapIsGlobal )
	{
	  vehLocation.x = navData->tranAbs.x;
	  vehLocation.y = navData->tranAbs.y;
	}
      else
	{
	  vehLocation.x = navData->tranRel.x;
	  vehLocation.y = navData->tranRel.y;
	}
      homLocal.tran = navData->tranRel;
      homLocal.rot = navData->rpyRel;
      homGlobal.tran = navData->tranAbs;
      homGlobal.rot = navData->rpyAbs;
      return(true);
    default:
      return(false);
    }
  return(false);
}

/*!
  Send the planned path to the subordinate

  \param confidence Confidence in path
  \param path       Array of path points
  \param length     Number of path points

  \return true on success
  \return false on failure
*/
bool sendPath( unsigned char confidence, PM_CARTESIAN *path, int length )
{
  int i;
  if( primMobJACmdNml==NULL )
    return false;
  if( length>MAX_WAYPOINTS )
    length = MAX_WAYPOINTS;

  moveWaypointCmd->waypoint_length = length;
  for(i=0; i<length; i++ )
    {
      if( mapIsGlobal )
	{
	  moveWaypointCmd->waypoint[i].p = transformPoint( path[i], true );
	}
      else
	{
	  moveWaypointCmd->waypoint[i].p.x = path[i].x;
	  moveWaypointCmd->waypoint[i].p.y = path[i].y;
	  moveWaypointCmd->waypoint[i].p.z = path[i].z;
	}
      moveWaypointCmd->waypoint[i].neighborhood = 2;
      moveWaypointCmd->waypoint[i].quality = confidence;
    }
  moveWaypointCmd->serial_number = myPrimSerial++;
  primMobJACmdNml->write(moveWaypointCmd);
  return true;
}

/*!
  Perform abort command state machine

  \param stat status message
*/
void doCmdAbort(VehMobPLStat *stat)
{
  PrimMobJACmdAbort *abortCmd;
  int echo_serial_number;
  RCS_STATUS primStatus;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      abortCmd = new PrimMobJACmdAbort();
      abortCmd->serial_number = ++myPrimSerial;
      primMobJACmdNml->write(abortCmd);
      delete abortCmd;
    } // fall through to checking the satus
  if( state_match(stat, S0) )
    {
      if( checkPrimCmdStatus( primStatus, echo_serial_number ) )
	{
	  if( echo_serial_number == myPrimSerial )
	    {
	      status_next(stat, primStatus);
	      if( stat->status == RCS_DONE || stat->status == RCS_ERROR )
		state_default(stat);
	    }
	}
    }
}

/*!
  Perform  halt command state machine

  \param stat status message
  \return true need to update and send set message
  \return false no message needs to be sent
*/
bool doCmdHalt(VehMobPLStat *stat)
{
  PrimMobJACmdHalt *haltCmd;
  int echo_serial_number;
  RCS_STATUS primStatus;
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      haltCmd = new PrimMobJACmdHalt();
      haltCmd->serial_number = ++myPrimSerial;
      primMobJACmdNml->write(haltCmd);
      delete haltCmd;
      retValue = true;
    } // fall through to checking the satus
  if( state_match(stat, S0) )
    {
      if( checkPrimCmdStatus( primStatus, echo_serial_number ) )
	{
	  if( echo_serial_number == myPrimSerial )
	    {
	      status_next(stat, primStatus);
	      if( stat->status == RCS_DONE || stat->status == RCS_ERROR )
		state_default(stat);
	    }
	}
    }
  return retValue;
}

/*!
  Perform init state machine

  \param stat status message
  \return true need to update and send set message
  \return false no message needs to be sent
*/
bool doCmdInit(VehMobPLStat *stat, VehMobPLSet *settings, VehMobPLEngine *engine)
{
  PrimMobJACmdInit *initCmd;
  int echo_serial_number;
  RCS_STATUS primStatus;
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      settings->isInitialized = false;
      initCmd = new PrimMobJACmdInit();
      initCmd->serial_number = ++myPrimSerial;
      primMobJACmdNml->write(initCmd);
      delete initCmd;
    } // fall through to reading navigation data
  if( state_match(stat, S0) )
    {
      status_next(stat, RCS_EXEC);
      if( updatePosition(navDataExtNml) )
	  state_next(stat, S1);
    }// fall througgh to setting the graph offsets
  if( state_match(stat, S1) )
    {
      engine->setGraphLocation( vehLocation );
      status_next(stat, RCS_EXEC);
      state_next(stat, S2);
    } // fall through to checking the status
  if( state_match(stat, S2) )
    {
      if( checkPrimCmdStatus( primStatus, echo_serial_number ) )
	{
	  if( echo_serial_number == myPrimSerial )
	    {
	      status_next(stat, primStatus);
	      if( stat->status == RCS_DONE || stat->status == RCS_ERROR )
		{
		  if( stat->status == RCS_DONE )
		    retValue = true;
		  state_default(stat);
		  state_next(stat, S3);
		}
	    }
	}
    }
  else if( state_match(stat, S3) )
    retValue = true;
  return retValue;
}

/*!
  Perform shutdown command to subordinate planning system

  \param stat vehicle status
  \return true done with shutdown
  \return false not done yet
*/
bool doCmdShutdown(VehMobPLStat *stat)
{
  PrimMobJACmdShutdown *shutdownCmd;
  int echo_serial_number;
  RCS_STATUS primStatus;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);

      shutdownCmd = new PrimMobJACmdShutdown();
      shutdownCmd->serial_number = ++myPrimSerial;
      primMobJACmdNml->write(shutdownCmd);
      delete shutdownCmd;
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      if( checkPrimCmdStatus( primStatus, echo_serial_number ) )
	{
	  if( echo_serial_number == myPrimSerial )
	    {
	      status_next(stat, primStatus);
	      if( stat->status == RCS_DONE || stat->status == RCS_ERROR )
		state_next(stat, S1);
	    }
	}
    }
  else if( state_match(stat,S1) )
    return(true);
  return(false);
}

/*! 
  Perform waypoint planning state machine

  \param stat vehicle status
*/
void doCmdMoveWaypoint(VehMobPLStat *stat, VehMobPLSet *settings, VehMobPLEngine *engine)
{
  PM_CARTESIAN goalToSend;
  int confidence;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
    } // fall through to planning
  if( state_match(stat, S0) )
    {
      if( goalIsGlobal && !mapIsGlobal )
	localGoal = transformPoint(globalGoal, true);
      else if( !goalIsGlobal && mapIsGlobal )
	globalGoal = transformPoint(localGoal, false);
      if( mapIsGlobal )
	goalToSend = globalGoal;
      else
	goalToSend = localGoal;

      if( settings->usePathDiscount )
	engine->discount(true);
      if( engine->search( vehLocation, goalToSend ) != -1 )
	{
	  status_next(stat, RCS_EXEC);
	  confidence = computeConfidence( engine->maxCost((int)(settings->confDistance/settings->cellResolution)), 
					  settings->zeroConf, settings->fiveConf );
	  sendPath( confidence, engine->pathOut, engine->pathLength );
	  if( engine->pathLength <= 2 )
	    {
	      state_next(stat, S2);
	      status_next(stat, RCS_DONE);
	    }
	}
      else
	{
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S2);
	  //	  sendAbort();
	}
      if( settings->usePathDiscount )
	engine->discount(false);
    }
}


/*!
  Perform clear wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgClearWM(VehMobPLSet *stat, VehMobPLEngine *engine)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      status_next(stat, RCS_EXEC);
      state_next(stat, S0);
      retValue = true;
    } // fall through to checking the status

  if( state_match(stat, S0) )
    {
      engine->clearMap();
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}

/*!
  Perform load wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgLoadWM(VehMobPLSet *stat, VehMobPLSet *settings, VehMobPLEngine *engine)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      if( !mapIsGlobal )
	{
	  fprintf( stderr, "\n\n\n" );	
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "Map storage/retrieval not supported in local maps\n" );
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "\n\n\n" );	
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S1);
	}
      else
	{
	  status_next(stat, RCS_EXEC);
	  state_next(stat, S0);
	}
      retValue = true;
    } // fall through to loading the file
      
  if( state_match(stat, S1) )
    {
      engine->loadWM( settings->wmFileName );
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}

/*!
  Perform store wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgStoreWM(VehMobPLSet *stat, VehMobPLSet *settings, VehMobPLEngine *engine)
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      if( !mapIsGlobal )
	{
	  fprintf( stderr, "\n\n\n" );	
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "Map storage/retrieval not supported in local maps\n" );
	  fprintf( stderr, "*********************************************************\n" );
	  fprintf( stderr, "\n\n\n" );	
	  status_next(stat, RCS_ERROR);
	  state_next(stat, S1);
	}
      else
	{
	  status_next(stat, RCS_EXEC);
	  state_next(stat, S0);
	}
      retValue = true;
    } // fall through to storing the file

  if( state_match(stat, S0) )
    {
      engine->storeWM( settings->wmFileName );
      status_next(stat, RCS_DONE);
      state_next(stat, S1);
      retValue = true;
    }
  return retValue;
}


/*!
  Perform dump wm state machine

  \param stat    vehicle status
  \param engine  the planning engine

  \return        true if set channel needs to be updated
*/
bool doCfgDumpWM(VehMobPLSet *settings, VehMobPLEngine *engine)
{
  bool retValue = false;
  static int fileID = 0;
  static int skip = 0;

  if( state_match(settings, NEW_COMMAND) )
    {
      state_new(settings);
      if( fileID != 0 )
	close( fileID );
      fileID = open( settings->wmFileName, O_CREAT|O_WRONLY, 
		     S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
      if( fileID <=0 )
	{
	  fprintf( stderr, "VehMobPLEngine: Error %d from opening output file %s\n",
		   fileID, settings->wmFileName );
	  status_next(settings, RCS_ERROR);
	  state_next(settings, S4);
	  retValue = true;
	}
      else
	{
	  skip = settings->fileSkip;
	  status_next(settings, RCS_EXEC);
	  state_next(settings, S0);
	  retValue = true;
	}
    } // if no error, fall through to seeing if file should be written

  if( state_match(settings, S0) )
    {
      if( skip++ >= settings->fileSkip )
	{
	  status_next(settings, RCS_EXEC);
	  state_next(settings, S1);
	}
      // no set settings have changed, so do not write
    } // possibly fall through to writing file
  if( state_match(settings, S1) )
    {
      engine->dump( fileID, vehLocation );
      if( settings->fileSkip == -1 ) // close file
	{
	  close( fileID );
	  fileID = 0;
	  status_next(settings, RCS_DONE);
	  state_next(settings, S4);
	  retValue = true;
	}
      else
	{
	  fsync( fileID );
	  skip = 0;
	  status_next(settings, RCS_EXEC);
	  state_next(settings, S0);
	}
    }
  return retValue;
}

/*!
  Perform sql initialization

  \param stat    vehicle status
  \param sqlParams mysql parameter block

  \return        true if set channel needs to be updated
*/
bool doCfgSQL(VehMobPLSet *stat )
{
  bool retValue = false;

  if( state_match(stat, NEW_COMMAND) )
    {
      state_new(stat);
      state_next(stat, S1);

#ifdef HAVE_MYSQL
      if( sqlConnect.init( stat->sqlServer, stat->sqlUser,
			   sqlPassword,
			   stat->sqlDatabase ) != 1 )
	{
	  stat->sqlConnected = false;
	  status_next(stat, RCS_ERROR);
	}
      else
	{
	  stat->sqlConnected = true;
	  status_next(stat, RCS_DONE);
	}
      retValue = true;
#else
      // no sql compiled in, so command is an error
      fprintf( stderr, "mysql not supported\n" );
      status_next(stat, RCS_ERROR);
      retValue = true;
#endif
    }
  return retValue;
}


/*!
  Read in command and act on it

\param cmdIn         The handle of the channel for commands to be read from
\param engine        The handle of the planner engine
\param stat          The status return

*/
RCS_STATUS readCfg( RCS_CMD_CHANNEL *cmdIn, VehMobPLEngine *engine, VehMobPLSet *currentStat )
{
  VehMobPLCfgStoreWM      *storeWMCmd;
  VehMobPLCfgLoadWM       *loadWMCmd;
  VehMobPLCfgDumpWM       *dumpWMCmd;
  VehMobPLCfgSQL          *sqlCmd;
  int type;
  int serial_number;

  type = cmdIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_CFG_CLEAR_WM_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CFG_STORE_WM_TYPE:
      storeWMCmd = (VehMobPLCfgStoreWM*)cmdIn->get_address();
      serial_number = storeWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  strncpy( currentStat->wmFileName, storeWMCmd->fileName, MOAST_FILE_NAME_LEN );
	}
      break;
    case VEH_MOB_PL_CFG_LOAD_WM_TYPE:
      loadWMCmd = (VehMobPLCfgLoadWM*)cmdIn->get_address();
      serial_number = loadWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  strncpy( currentStat->wmFileName, loadWMCmd->fileName, MOAST_FILE_NAME_LEN );
	}
      break;
    case VEH_MOB_PL_CFG_DUMP_WM_TYPE:
      dumpWMCmd = (VehMobPLCfgDumpWM*)cmdIn->get_address();
      serial_number = dumpWMCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  strncpy( currentStat->wmFileName, dumpWMCmd->fileName, MOAST_FILE_NAME_LEN );
	  if( dumpWMCmd->dumpContinuous )
	    currentStat->fileSkip = dumpWMCmd->skip;
	  else
	    currentStat->fileSkip = -1;	    
	}
      break;
    case VEH_MOB_PL_CFG_SQL_TYPE:
      sqlCmd = (VehMobPLCfgSQL*)cmdIn->get_address();
      serial_number = sqlCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;

	  strncpy( currentStat->sqlServer, sqlCmd->server, MOAST_FILE_NAME_LEN );
	  strncpy( currentStat->sqlUser, sqlCmd->user, MOAST_FILE_NAME_LEN );
	  strncpy( currentStat->sqlDatabase, sqlCmd->database, MOAST_FILE_NAME_LEN );
	  strncpy( sqlPassword, sqlCmd->password, MOAST_FILE_NAME_LEN );
	}
      break;
    default:
      fprintf( stdout, "Command %s not yet supported\n", vehMobPL_symbol_lookup(type) );
      currentStat->command_type = type;
      return RCS_ERROR;
    }

  return RCS_DONE;
}

/*!
  Read in command and act on it

\param cmdIn         The handle of the channel for commands to be read from
\param engine        The handle of the planner engine
\param currentStat   The current status of the module

*/
void readCmd( RCS_CMD_CHANNEL *cmdIn, VehMobPLEngine *engine, VehMobPLStat *currentStat )
{
  VehMobPLCmdMoveWaypoint *waypointCmd;
  int type, serial_number;
  static int  noPrint = 0;

  type = cmdIn->read();
  switch( type )
    {
    case 0: // no new data
      break;
    case VEH_MOB_PL_CMD_ABORT_TYPE:
    case VEH_MOB_PL_CMD_HALT_TYPE:
    case VEH_MOB_PL_CMD_INIT_TYPE:
    case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
      serial_number = cmdIn->get_address()->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	}
      break;
    case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
      waypointCmd = (VehMobPLCmdMoveWaypoint*)cmdIn->get_address();
      serial_number = waypointCmd->serial_number;
      currentStat->command_type = type;
      if( serial_number != currentStat->echo_serial_number )
	{
	  currentStat->echo_serial_number = serial_number;
	  currentStat->state = NEW_COMMAND;
	  currentStat->status = RCS_EXEC;
	  
	  if( waypointCmd->isGlobal )
	    {
	      goalIsGlobal = true;
	      globalGoal.x = waypointCmd->p.x;
	      globalGoal.y = waypointCmd->p.y;
	      globalGoal.z = waypointCmd->p.z;
	    }
	  else
	    {
	      goalIsGlobal = false;
	      localGoal.x = waypointCmd->p.x;
	      localGoal.y = waypointCmd->p.y;
	      localGoal.z = waypointCmd->p.z;
	    }
	}
      break;
    default:
      if( noPrint < 10 )
	{
	  fprintf( stdout, "Command %s not yet supported\n", vehMobPL_symbol_lookup(type) );
	  noPrint++;
	}
      break;
    }
  return;
}

/*!
  Read in obstacles and populate world model

\param obsChan The channel to read from
\param engine  A handle to the planning engine

\return Obstacle message on success 
\return NULL on failure
*/
SensorDataList *readObstacles( NML *obsChan, VehMobPLEngine *engine )
{
  int type;
  SensorDataList *listReturn;


  type = obsChan->read();
  switch( type )
    {
    case 0: // no new data
      return(NULL);
    case SENSOR_DATA_LIST_TYPE:
      listReturn = (SensorDataList*)obsChan->get_address();
      return(listReturn);
    default:
      fprintf( stderr, "obstacle channel error\n" );
      fprintf( stdout, "Command %s not yet supported\n", sensorData_symbol_lookup(type) );
      return(NULL);
    }
}

/*!
  free all memory and prepare to exit
*/
void closeAll()
{
  if( primMobJACmdNml != NULL )
    delete primMobJACmdNml;
  primMobJACmdNml = NULL;
  if( primMobJAStatNml != NULL )
    delete primMobJAStatNml;
  primMobJAStatNml = NULL;
  if( primMobJASetNml != NULL )
    delete primMobJASetNml;
  primMobJASetNml = NULL;
  if( primMobJACfgNml != NULL )
    delete primMobJACfgNml;
  primMobJACfgNml = NULL;
  if( servoSPCmdNml != NULL )
    delete servoSPCmdNml;
  servoSPCmdNml = NULL;
  if( servoSPStatNml != NULL )
    delete servoSPStatNml;
  servoSPStatNml = NULL;
  if( servoSPSetNml != NULL )
    delete servoSPSetNml;
  servoSPSetNml = NULL;
  if( servoSPCfgNml != NULL )
    delete servoSPCfgNml;
  servoSPCfgNml = NULL;
  if( vehMobPLCmdNml != NULL )
    delete vehMobPLCmdNml;
  vehMobPLCmdNml = NULL;
  if( vehMobPLStatNml != NULL )
    delete vehMobPLStatNml;
  vehMobPLStatNml = NULL;
  if( vehMobPLSetNml != NULL )
    delete vehMobPLSetNml;
  vehMobPLSetNml = NULL;
  if( vehMobPLCfgNml != NULL )
    delete vehMobPLCfgNml;
  vehMobPLCfgNml = NULL;
  if( sensorDataNml != NULL )
    delete sensorDataNml;
  sensorDataNml = NULL;
  if( navDataExtNml != NULL )
    delete navDataExtNml;
  navDataExtNml = NULL;

  if( moveWaypointCmd != NULL )
    delete moveWaypointCmd;
}

/*!
  Open all of the NML channels
  
  \param vehID The number to add to all of the channel names
  \return true on success
  \return false on failure
*/
bool openNML( int vehID )
{
  char *configNml = getenv("CONFIG_NML");
  char chanName[MOAST_NML_BUFFER_NAME_LEN];
  char myName[MOAST_NML_BUFFER_NAME_LEN];

  sprintf( myName, "%s%d", MY_NAME_BASE, vehID );

  if(!configNml)
    {
      fprintf(stderr,"Please set CONFIG_NML environment variable.\n");
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_CMD_NAME, vehID );
  primMobJACmdNml = new RCS_CMD_CHANNEL( primMobJA_format, chanName,
				      myName, configNml );
  if( NULL == primMobJACmdNml || !primMobJACmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_STAT_NAME, vehID );
  primMobJAStatNml = new RCS_STAT_CHANNEL( primMobJA_format, chanName,
					myName, configNml );
  if( NULL == primMobJAStatNml || !primMobJAStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_SET_NAME, vehID );
  primMobJASetNml = new RCS_CMD_CHANNEL( primMobJA_format, chanName,
				      myName, configNml );
  if( NULL == primMobJASetNml || !primMobJASetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", PRIM_MOB_JA_CFG_NAME, vehID );
  primMobJACfgNml = new RCS_STAT_CHANNEL( primMobJA_format, chanName,
				       myName, configNml );
  if( NULL == primMobJACfgNml || !primMobJACfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_SP_CMD_NAME, vehID );
  servoSPCmdNml = new RCS_CMD_CHANNEL( servoSP_format, chanName,
				       myName, configNml );
  if( NULL == servoSPCmdNml || !servoSPCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_SP_STAT_NAME, vehID );
  servoSPStatNml = new RCS_STAT_CHANNEL( servoSP_format, chanName,
					 myName, configNml );
  if( NULL == servoSPStatNml || !servoSPStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_SP_SET_NAME, vehID );
  servoSPSetNml = new RCS_CMD_CHANNEL( servoSP_format, chanName,
				       myName, configNml );
  if( NULL == servoSPSetNml || !servoSPSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", SERVO_SP_CFG_NAME, vehID );
  servoSPCfgNml = new RCS_STAT_CHANNEL( servoSP_format, chanName,
					myName, configNml );
  if( NULL == servoSPCfgNml || !servoSPCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CMD_NAME, vehID );
  vehMobPLCmdNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
				      myName, configNml );
  if( NULL == vehMobPLCmdNml || !vehMobPLCmdNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_STAT_NAME, vehID );
  vehMobPLStatNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				       myName, configNml );
  if( NULL == vehMobPLStatNml || !vehMobPLStatNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_SET_NAME, vehID );
  vehMobPLSetNml = new RCS_STAT_CHANNEL( vehMobPL_format, chanName,
				     myName, configNml );
  if( NULL == vehMobPLSetNml || !vehMobPLSetNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", VEH_MOB_PL_CFG_NAME, vehID );
  vehMobPLCfgNml = new RCS_CMD_CHANNEL( vehMobPL_format, chanName,
				      myName, configNml );
  if( NULL == vehMobPLCfgNml || !vehMobPLCfgNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n",
                       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", AM_SP_MAP_NAME, vehID );
  sensorDataNml = new NML( sensorData_format, chanName,
			      myName, configNml );
  if( NULL == sensorDataNml || !sensorDataNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, chanName );
      return(false);
    }

  sprintf( chanName, "%s%d", NAV_DATA_EXT_NAME, vehID );
  navDataExtNml = new NML( navDataExt_format, chanName,
			   myName, configNml );
  if( NULL == navDataExtNml || !navDataExtNml->valid() )
    {
      rcs_print_error( "%s: Error - Can't open %s buffer\n", 
		       myName, chanName );
      return(false);
    }

  return(true);
}

/*!
  Signal handler
  \param signum Signal that has been caught
*/
void sigHandler( int signum )
{
  switch( signum )
    {
    case SIGPIPE:
      fprintf( stderr, "Output pipe closed\n" );
      drawMap = false;
      break;
    case SIGINT:
      fprintf( stderr, "Shutting down vehMobPL program\n" );
      closeAll();
      exit(1);
      break;
    default:
      fprintf( stderr, "Unknown signal caught\n" );
      break;
    }
}

int main(int argc, char **argv)
{
  int count;
  PM_CARTESIAN obsLoc;
  char c;
  VehMobPLEngine *engine;
  SensorDataList *newObs;
  double runTime, accumTime;
  int cycleCount, skip;
  RCS_TIMER *timer;
  VehMobPLStat *vehMobPLStat;
  VehMobPLSet  *vehMobPLSet;
  double obsProx = 250;
  double obsCost = 5000;
  bool done, writeSet;
  VehMobPLCfgSQL *vehMobPLCfgSQL;
  VehMobPLCmdHalt *vehMobPLCmdHalt;

  // initial settings of variables that may be set with getopt
  //! initial map extents
  int extents = 50;
  //! The re-planning cycle time
  double cycleTime = .2;
  //! The initial resolution of theh planning map
  float cellResolution = .5;
  //! Is the planner actively planning
  bool isIdle = true;;
  //! The number of cells to grow obstacles
  int obsGrow = 1;
  //! Use path discounting to avoid rapid path switching
  bool useDiscount = false;
  //! How far to look down the path for determining the confidence
  int confDist = 10;

  // variables that are initialized here or by getopt
  // Set drawing to off, with a skip of 5 
  drawMap = false;
  drawSkip = 5;
  // map is global by default
  mapIsGlobal = true;
  // set verbose off
  myVerbose = false;

  myPrimSerial = 1;
  goalIsGlobal = false;
  moveWaypointCmd = new PrimMobJACmdMoveWaypoint();
  skip = 0;
  
  signal( SIGPIPE, sigHandler );
  signal( SIGINT, sigHandler );
  while((c=getopt(argc,argv,"c:d:e:g:hil:n:r:s:v"))!=EOF) 
    {
      switch(c) 
	{
	case 'c':
	  // set cycle time
	  cycleTime = atof(optarg);
	  fprintf( stderr, "Cycle time set to %f\n", cycleTime );
	  break;
	  /*
	case 'd':
	  // set map drawing on
	  drawSkip = atoi(optarg);
	  drawMap = true;
	  fprintf( stderr, "Drawing map on with skip of %d\n", drawSkip );
	  break;
	  */
	case 'e':
	  // get goal y value
	  isIdle = false;
	  localGoal.y = atof(optarg);
	  fprintf( stderr, "East goal set to %f\n", localGoal.y );
	  break;
	case 'g':
	  // grow obstacles
	  obsGrow = atoi(optarg);
	  fprintf( stderr, "Obstacles growing by %d cells in all directions\n", obsGrow );
	  break;
	case 'h':
	  // turn on discounting
	  useDiscount = true;
	  fprintf( stderr, "Using path discount to reduce hysterisis\n" );
	  break;
	case 'i':
	  // use local imu solution instead of global
	  mapIsGlobal = false;
	  fprintf( stderr, "Using local imu (local coord) for planning map\n" );
	  break;
	case 'l':
	  confDist = atoi(optarg);
	  fprintf( stderr, "Confidence look-ahead is %dm\n", confDist );
	  break;
	case 'n':
	  // get goal x value
	  localGoal.x = atof(optarg);
	  isIdle = false;
	  fprintf( stderr, "North goal set to %f\n", localGoal.x );
	  break;
	case 'r':
	  // set cell resolution
	  cellResolution = atof(optarg);
	  fprintf( stderr, "Cell resolution set to %f\n", cellResolution );
	  break;
	case 's':
	  extents = atoi(optarg);
	  if( extents < 50 )
	    {
	      extents = 50;
	      fprintf( stderr, "Extents must be at least 50.\n" );
	    }
	  fprintf( stderr, "Graph extents set to %d\n", extents );
	  break;
	case 'v':
	  myVerbose = true;
	  break;
	default:
	  fprintf( stderr, "usage: %s -c<cycle time> -g<grow obstacle num> -h "
		   "(turn hystrisis on) -i (use local imu coord for map) "
		   "-l<confidance look ahead> "
		   "-r<cell resolution> -n<goal north> -e<goal east> "
		   "-s<worldmodel initial size> -v (verbose)\n", argv[0] );
	  exit(1);
	}
    }
  timer = new RCS_TIMER(cycleTime);
  engine = new VehMobPLEngine( extents, extents, cellResolution, obsGrow, myVerbose );
  engine->setCostLookup();
  engine->buildGraph();

  if( !openNML(1) )
    exit(1);
  vehMobPLSet = new VehMobPLSet();
  vehMobPLSet->cycleTime = cycleTime;
  vehMobPLSet->obstaclePad = obsGrow;
  vehMobPLSet->usePathDiscount = useDiscount;
  vehMobPLSet->debug = myVerbose;
  vehMobPLSet->isInitialized = false;
  vehMobPLSet->isUsingSQL = false;
  vehMobPLSet->confDistance = confDist;
  vehMobPLSet->cellResolution = cellResolution;
  vehMobPLSet->zeroConf = obsCost;
  vehMobPLSet->fiveConf = obsProx;
  strcpy(vehMobPLSet->wmFileName, "unitialized" );
  vehMobPLSet->fileSkip = -1;
  vehMobPLSet->echo_serial_number = -1;

  // send out first set message
  vehMobPLCfgSQL = new VehMobPLCfgSQL();
  vehMobPLCfgSQL->serial_number = 0;
  strncpy(vehMobPLCfgSQL->server, "sylvester.mxnet.mel.nist.gov", MOAST_FILE_NAME_LEN);
  strncpy(vehMobPLCfgSQL->user, "otbsaf", MOAST_FILE_NAME_LEN);
  strncpy(vehMobPLCfgSQL->database, "DemoWorldModel", MOAST_FILE_NAME_LEN);
  strncpy(vehMobPLCfgSQL->password, "simulate", MOAST_FILE_NAME_LEN);
  vehMobPLCfgNml->write( vehMobPLCfgSQL );
  delete vehMobPLCfgSQL;

  vehMobPLStat = new VehMobPLStat();
  vehMobPLStat->echo_serial_number = -1;
  // write out first command
  vehMobPLCmdHalt = new VehMobPLCmdHalt();
  vehMobPLCmdHalt->serial_number = 0;
  vehMobPLCmdNml->write( vehMobPLCmdHalt );
  delete vehMobPLCmdHalt;

  cycleCount = 0;
  accumTime = 0;
  runTime = etime();
  for(;;)
    {
      /////////////////////////////////////////////
      // update vehicle information
      /////////////////////////////////////////////
      // update vehicle position
      if( vehMobPLSet->isInitialized )
	updatePosition(navDataExtNml);
      // read configuration channel
      readCfg(vehMobPLCfgNml, engine, vehMobPLSet);
      // read command channel
      readCmd(vehMobPLCmdNml, engine, vehMobPLStat);
      // read obstacle channel
      if( vehMobPLSet->isInitialized == true )
	newObs = readObstacles(sensorDataNml, engine);
      else
	newObs = NULL;
      /////////////////////////////////////////////
      // process new informatio
      /////////////////////////////////////////////
      // process configuration
      switch( vehMobPLSet->command_type )
	{
	case VEH_MOB_PL_CFG_CLEAR_WM_TYPE:
	  writeSet = doCfgClearWM( vehMobPLSet, engine );
	  break;
	case VEH_MOB_PL_CFG_STORE_WM_TYPE:
	  writeSet = doCfgStoreWM( vehMobPLSet, vehMobPLSet, engine );
	  break;
	case VEH_MOB_PL_CFG_LOAD_WM_TYPE:
	  writeSet = doCfgLoadWM( vehMobPLSet, vehMobPLSet, engine );
	  break;
	case VEH_MOB_PL_CFG_DUMP_WM_TYPE:
	  writeSet = doCfgDumpWM( vehMobPLSet, engine );
	  break;
	case VEH_MOB_PL_CFG_SQL_TYPE:
	  writeSet = doCfgSQL( vehMobPLSet );
	default:
	  writeSet = false;
	  break;
	}
      if( writeSet )
	vehMobPLSetNml->write(vehMobPLSet);
      // process obstacles
      if( newObs != NULL )
	{
	  if( mapIsGlobal && !newObs->isGlobal ) // set local coordinates to global
	    {
	      for( count=0; count<newObs->featList_length; count++ )
		{
		  obsLoc.x = newObs->featList[count].dataLoc.x;
		  obsLoc.y = newObs->featList[count].dataLoc.y;
		  obsLoc.z = newObs->featList[count].dataLoc.z;
		  obsLoc = transformPoint(obsLoc, false );
		  newObs->featList[count].dataLoc.x = obsLoc.x;
		  newObs->featList[count].dataLoc.y = obsLoc.y;
		}
	    }
	  else if( !mapIsGlobal && newObs->isGlobal ) // set global coordinates to local
	    {
	      for( count=0; count<newObs->featList_length; count++ )
		{
		  obsLoc.x = newObs->featList[count].dataLoc.x;
		  obsLoc.y = newObs->featList[count].dataLoc.y;
		  obsLoc.z = newObs->featList[count].dataLoc.z;
		  obsLoc = transformPoint(obsLoc, true );
		  newObs->featList[count].dataLoc.x = obsLoc.x;
		  newObs->featList[count].dataLoc.y = obsLoc.y;
		}
	    }

	  engine->updateWorld( newObs->featList, newObs->featList_length );
	}
      // process commands
      switch( vehMobPLStat->command_type )
	{
	case VEH_MOB_PL_CMD_ABORT_TYPE:
	  doCmdAbort( vehMobPLStat );
	  break;
	case VEH_MOB_PL_CMD_HALT_TYPE:
	  if( doCmdHalt( vehMobPLStat ) )
	    {
	      vehMobPLSet->isInitialized = false;
	      vehMobPLSetNml->write(vehMobPLSet);
	    }
	  break;
	case VEH_MOB_PL_CMD_INIT_TYPE:
	  if( doCmdInit( vehMobPLStat, vehMobPLSet, engine ))
	    {
	      vehMobPLSet->isInitialized = true;
	      vehMobPLSetNml->write(vehMobPLSet);
	    }
	  break;
	case VEH_MOB_PL_CMD_SHUTDOWN_TYPE:
	  done = doCmdShutdown( vehMobPLStat );
	  vehMobPLStatNml->write(vehMobPLStat);
	  return(1);
	  break;
	case VEH_MOB_PL_CMD_MOVE_WAYPOINT_TYPE:
	  if( !vehMobPLSet->isInitialized )
	    {
	      status_next(vehMobPLStat, RCS_ERROR);
	      break;
	    }
	  doCmdMoveWaypoint( vehMobPLStat, vehMobPLSet, engine );
	  break;
	}
      vehMobPLStat->cycleTime = etime()-runTime;
      accumTime += vehMobPLStat->cycleTime;
      //      vehMobPLStat->worldX = engine->getGraphHeight();
      //      vehMobPLStat->worldY = engine->getGraphWidth();
      vehMobPLStatNml->write(vehMobPLStat);
      if( ++cycleCount == 1000 )
	{
	  fprintf( stderr, "Time per cycle: %f\n", accumTime/cycleCount );
	  cycleCount = 0;
	  accumTime = 0;
	}
      runTime = etime();
      timer->wait();
    }
  // clock_gettime
}

/*
  Modification history:

  $Log: vehMobPLMainOld.cc,v $
  Revision 1.1.1.1  2005/10/12 20:50:32  root
  initial import

  Revision 1.4  2005/09/30 19:26:22  proctor
  Fixed warnings with %lf used with printf, it's for scanf only

  Revision 1.3  2005/09/30 18:55:58  proctor
  Cleaned up unused variables, forgotten return values

  Revision 1.2  2005/09/29 13:18:10  stephen
  Beginning of changes to work with Tom's new prim and am sp.

  Revision 1.1  2005/07/26 13:25:11  stephen
  Added visibility graph planner (not yet working)

  Revision 1.11  2005/07/18 21:16:59  proctor
  Replaced sensorData with servoSPLinescan, ...

  Revision 1.10  2005/07/18 19:56:12  stephen
  Changed interface to confrom to dijkHeapN

  Revision 1.9  2005/06/29 21:08:17  proctor
  Took out \defgroup comment for CVS history, since that's not what \defgroup
  is used for.

  Revision 1.8  2005/06/15 13:39:53  stephen
  Added use of mysql class for opening channel to sql and fixed writing of set channel.

  Revision 1.7  2005/06/14 19:19:14  stephen
  Changed to use SensorPoint in sensor data message.

  Revision 1.6  2005/06/13 17:26:29  stephen
  Continuing work on adding commands and creating true state-machine.

  Revision 1.5  2005/06/09 16:43:46  stephen
  Fixed initialization problem.

  Revision 1.4  2005/06/09 12:32:58  stephen
  Removal of sqrtf and start of turning cfg channel into state machine.

  Revision 1.3  2005/06/08 20:49:37  stephen
  Changed occupancyData to sensorData

  Revision 1.2  2005/06/08 19:29:14  proctor
  Moved state_match, etc. macros into moastTypes.hh

  Revision 1.1  2005/06/08 14:31:15  proctor
  Added MobJA name changes

  Revision 1.10  2005/06/08 13:40:10  stephen
  Beginnings of changes to have true cfg cycle.

  Revision 1.9  2005/06/06 12:45:06  stephen
  Made true state machine.

  Revision 1.8  2005/06/02 12:56:35  stephen
  Changed waypoint to use only a single waypoint.

  Revision 1.7  2005/06/01 13:21:46  proctor
  Changed code that handles waypoints to use the Waypoint structure now
  declared in moastTypes.cc

  Revision 1.6  2005/05/31 21:19:33  proctor
  Changed tabsGraph, searcher to DijkTabsGraph, DijkTabsSearcher, etc. to
  reflect Tom's changes to the naming convention.

  Revision 1.5  2005/05/27 23:22:48  proctor
  Fixed floorf nonportability

  Revision 1.4  2005/05/25 17:48:16  stephen
  Fixed waitForStatus to wait between channel reads.

  Revision 1.3  2005/05/24 17:05:07  stephen
  Changed to Tom's new Dijk algorithm.

  Revision 1.2  2005/05/24 12:39:18  stephen
  Added signal handler for SIGINT signal to properly exit.

  Revision 1.1  2005/05/23 18:03:31  stephen
  Added vehicle level planner.
*/
