/******************************************************************************
DISCLAIMER:
This software was produced by the National Institute of Standards
and Technology (NIST), an agency of the U.S. government, and by statute is
not subject to copyright in the United States.  Recipients of this software
assume all responsibility associated with its operation, modification,
maintenance, and subsequent redistribution.

See NIST Administration Manual 4.09.07 b and Appendix I. 
*****************************************************************************/

/*!
\file amMobMain.cc
\brief Vehicle autonomous mobility level mobility controller.

\code CVS Status:
$Author: dr_steveb $
$Revision: 1.59 $
$Date: 2009/04/27 16:09:00 $
\endcode

\author Tom Kramer
\date 9 June 2005
*/

/*

The intended method of terminating this process is to use "kill -INT"
to send it SIGINT, which (because of the "signal(SIGINT, quit);" line
in the source code) will cause the "quit" function to run, setting
the "done" global variable to 1. That will be caught by one of the
two "while (!done)" loops and result in an orderly exit.

The side of the obstacle map is currently one cell larger than the
side of the planning area. The planning area is a square 10 meters on
a side.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <fstream>
#include <stdio.h>
#include <math.h>               // floor, M_PI, cos
#include <stddef.h>		// sizeof(), NULL
#include <stdlib.h>		// getenv()
#include <signal.h>		// SIGINT, signal()
#include <fcntl.h>	        // open()
#include <sys/stat.h>           // S_IWUSR, etc.
#include <unistd.h>	        // close()
#include <string>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <rcs.hh>               // INIFILE, etc.
#include "moastTypes.hh"
#include "moastNmlOffsets.hh"
#include "navDataExt.hh"	// navDataExt_format(), etc.
#include "primMobJA.hh"         // prim commands and enums
#include "amMobJA.hh"
#include "amMobMain.hh"         // classes used in execution
#include "sensorData.hh"        // sensorData_format(), etc.
#include "servoMobJA.hh"        // SERVO_MOB_JA_SET_NAME, servoMobJA_format
#include "servoSP.hh"           // servoSP_format
#include "amSP.hh"              // AMSP_NOT_OBSTACLE, etc.
#include "aStarHeap.hh"         // nodes, graphs, searcher for aStar
#include "dijkHeapN.hh"         // nodes, graphs, searcher for dijkstra
#include "RoboCupGUIState.hh"

// For WSS
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/shm.h>

#include "httpnml.hh"

// Communications class
#include "communications.h"

// Map class for the pathfinding algorithm
#include "robocupmap.h"

// Deployment code for path planning on the voronoi
#include "deploymentplan.h"

using namespace std;

/***********************************************************************/

/*

AMOB_MAX_CENTRIPETAL is the maximum centripetal acceleration allowed.
It is currently set at half a G (which is 9.8 m/s/s).

AMOB_MAX_LEG is the largest amount that may be rounded off a straight
line in order to round corners.

The following values were taken from vehMobCell.hh. This was done
so that AM costs may be combined Vehicle costs for cooperative planning.

AM_EASY2 = VEH_WM_EASY_MAX/2 = 59/2
AM_UNKNOWN = VEH_WM_UNKNOWN = 100
AM_DIFFICULT = VEH_WM_DIFFICULT_MAX = 179
AM_DANGEROUS = VEH_WM_DANGEROUS_MAX = 254

*/

#define AMOB_MAX_CENTRIPETAL 4.9
#define AMOB_MAX_LEG 3.0
#define AMOB_TINY_DIST 0.00001
#define AMOB_TINY_ANGLE 0.01
#define AMSP_NEAR_OBSTACLE 0.3
#define AM_EASY2 29
#define AM_UNKNOWN 100
#define AM_DIFFICULT 179
#define AM_DANGEROUS 254
#define DEFAULT_CELL_RESOLUTION 0.100
#define DEFAULT_CYCLE_TIME 0.100
#define DEFAULT_INI_FILE "../etc/moast.ini"
#define DEFAULT_NML_FILE "../etc/moast.nml"
#define DEFAULT_MAP_IMAGE_FILE_NAME "AmMobMapImage.ppm"
#define DEFAULT_MAX_SPEED 1.0
#define DEFAULT_MIN_SPEED 0.1
#define DEFAULT_OBS_MAP_SIDE_LENGTH 60
#define DEFAULT_VEHICLE_MIN_TURN_RAD 0.0
#define DEFAULT_VEHICLE_WIDTH 0.9
#define LOG_FILE  "amOut.log"
#define MAX_MAP_SIDE MAX_SENSOR_DATA_1D
#define MAX_MAP_CELLS MAX_MAP_SIDE * MAX_MAP_SIDE
#define THIS_PROCESS "amMob"
#define UNSET_CELL_RESOLUTION 0.0
#define UNSET_OBS_MAP_SIDE_LENGTH 0
#define VERY_BIG_COST 1000000.0

#ifndef INI_OK
#define INI_OK      0
#define INI_DEFAULT 1
#define INI_INVALID 2
#endif

#define ABS(x)  (((x) < 0) ? -(x) : (x))
#define MIN(x,y)  (((x) < (y)) ? (x) : (y))
#define MAX(x,y)  (((x) > (y)) ? (x) : (y))

#define DEPLOYMENT_DEFAULT_NEIGHBORHOOD 0.3
#define DEPLOYMENT_DEFAULT_SPEED 0.2

/***********************************************************************/

// NML Channels

//! navigation data channel
static NML *navDataExtBuf = NULL;

//! prim planning channels
static RCS_CMD_CHANNEL *primMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *primMobJAStatBuf = NULL;
static RCS_CMD_CHANNEL *primMobJACfgBuf = NULL;
static RCS_STAT_CHANNEL *primMobJASetBuf = NULL;

//! am planning channels
static RCS_CMD_CHANNEL *amMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *amMobJAStatBuf = NULL;
static RCS_CMD_CHANNEL *amMobJACfgBuf = NULL;
static RCS_STAT_CHANNEL *amMobJASetBuf = NULL;

//! am sp data channel
static NML *sensorData2DBuf = NULL;

//! servo settings channel (for vehicle width)
static RCS_STAT_CHANNEL *servoMobJASetBuf = NULL;

/// servo sp channel
static RCS_CMD_CHANNEL* servoSPCmdBuf = NULL;
static RCS_STAT_CHANNEL* servoSPStatBuf = NULL;
static RCS_STAT_CHANNEL* servoSPSetBuf = NULL;
static ServoSPStat* servoSPStat;

static AmMobJAStat * amMobJAStat;

//! am sp cmd/stat channel/commands
static RCS_CMD_CHANNEL *amSPCmdBuf = NULL;
static RCS_STAT_CHANNEL *amSPStatBuf = NULL;
static AmSPCmdInit amSPCmdInit;
static AmSPCmdGo amSPCmdGo;
static AmSPStat *amSPStat;

//! AMMobJA init command
static AmMobJACmdInit amMobJACmdInit;

/// Servo Mobility channels
static RCS_CMD_CHANNEL *servoMobJACmdBuf = NULL;
static RCS_STAT_CHANNEL *servoMobJAStatBuf = NULL;
static ServoMobJAStat* servoMobJAStat;

//! ServoMobJA commands
ServoMobJACmdSkid servoMobJACmdSkid;
ServoMobJACmdRotaryWing servoMobJACmdRotaryWing;
ServoMobJACmdInit servoMobJACmdInit;

/// ServoSP commands
ServoSPCmdInit servoSPCmdInit;
ServoSPCmdGo servoSPCmdGo;

/// Commands used for interrupting a waypoint command
PrimMobJACmdStop primMobJACmdStop;
AmMobJACmdNop amMobJACmdNop;

/***********************************************************************/

// Other global variables

/*

processName is the actual name of the process, while THIS_PROCESS is
the name used in the NML config file and, hence, required by NML.

Items that are treated as global variables but change infrequently and
might be set by configuration commands are included in the "set"
configuration status message class. This includes set.cellResolution 
and set.vehicleWidth. However, the global variable "res" is a copy of
set.cellResolution.

The primCmdNumber is the number of the latest command to prim from
amMobMain. It is expected that primShell might also be sending
commands to prim. To deal with this, amMobMain sends only even
command numbers. For sending a command to prim, amMobMain uses
the smallest even number that is larger than the current value
of primCmdNumber and larger than the most recent command
number echoed by prim. If the primShell sends odd command
numbers using a similar rule, no duplicate command numbers
will be sent to prim. The primCmdNumber is initialized to 2.

When a response is expected from prim because a command has been sent
to prim and its execution is being monitored, this checks the command
number echoed by prim. If it is not the number of the command most
recently sent by AM Mob, AM Mob does not do anything until the next
cycle. This will make it appear in the AM Mob status message that the
last command received by amMob is still executing (even though it
isn't making any progress). This is probably OK behavior most of the
time.

*/
// lookup table, table[arc]=waypoint 
static int          arcToWaypointTable[MAX_ARC_SEGMENTS];
// obsMapSideLength in command args
static int          argObsMapSideLength = UNSET_OBS_MAP_SIDE_LENGTH;
// cell resolution in command args
static double       argResolution = UNSET_CELL_RESOLUTION;
static int          cellPixels;          // pixels per cell in obsMap display
static int          cellsPerMeter;       // number of cells per meter
static double       diagDist;            // half the diagonal of a cell
static int          done = 0;            // flag for exiting
static double       extraDist;           // (resolution * (sqrt(2) - 1))
static int *        goals = NULL;        // array of goal graph node ids
static char *       iniFileName = NULL;  // name of ini file
static FILE *       logFp = NULL;        // log file pointer
// name of ppm image file or pipe
static char *       mapImageFileName = DEFAULT_MAP_IMAGE_FILE_NAME;
static double       maxSpeed;            // m/s maximum vehicle speed
static double       minSpeed;            // m/s minimum vehicle speed
static char *       nmlFile = NULL;      // name of NML file
static AStarHeapNode nodesAStar[MAX_MAP_CELLS]; // nodes for A* graph search
static DijkHeapNNode nodesDijk[MAX_MAP_CELLS]; // nodes for Dijk graph search
static SensorData2D * obsData;           // data in SensorData2DBuf
static int          obsGrow;             // number of cells to grow obstacles
static double       obsMap[MAX_MAP_SIDE][MAX_MAP_SIDE]; // obstacle map
static int          obsMapSide = -1;     // number of cells on side of obsMap
static int          obsMapSi = -2;       // one less than obsMapSide
static int          obsMapSideLength=0;  // length of map side in meters
static int          primCmdNumber = 2;   // number of latest command to prim
static char *       processName;         // name of this process
static double       res;                 // copy of set.cellResolution
static AmMobJASet   set;                 // config command status message
static int *        skips = NULL;        // array of skips for growing obs
static double       straightDist;        // half the resolution of a cell
static AmMobJAStat  stts;                // task command status message
static RCS_TIMER *  timer;               // timer for cycle time
static double       vehLength = 0;       // vehicle length (temporary)
static double       vehSpeed;            // m/s XY speed of vehicle
static double       vehSpin;             // rad/s rate of change of vehYaw
static PM_CARTESIAN vehVelocity;         // m/s vector in local coordinates
static double       vehX;                // vehicle local X coordinate
static double       vehY;                // vehicle local Y coordinate
static double       vehYaw;  // rad angle CCW (seen from +Z) from local +X
static int          verbose = 1;         // 0=no messages, 1=more, 2=most
static double       xMax;                // maximum value of x on obsMap
static double       xMin;                // minimum value of x on obsMap
static double       yMax;                // maximum value of y on obsMap
static double       yMin;                // minimum value of y on obsMap
static PM_POSE      localToGlobal;       // transform local to global coords
//static int          LatitudeDegrees;
//static float        LatitudeMinutes;
//static int          LongitudeDegrees;
//static float        LongitudeMinutes;
//static int          Fix;
//static int          Satellites;
//static RoboCupGUIState initGPS;
//static RoboCupGUIState currentGP
// Set up sS;
static RoboCupGUIState GPSWaypoint;
static bool unprocessedGPSWaypoint = false;
static char         *xmlHostName;
//static int          initLatitudeDegrees;
//static float        initLatitudeMinutes;
//static int          initLongitudeDegrees;
//static float        initLongitudeMinutes;
//static int          initFix;
//static int          initSatellites;
//static double       initX;
//static double       initY;
//static bool         gotInitialPose = false;
static AmMobJACmdMoveWaypoint amMobJACmdMoveWaypoint;
static int socket_descriptor;

// Config Variables
static char wssIP[33];

//communications variables
static Communications* comm;
static int commID;

// Robot Type
char robotType;

/***********************************************************************/

static int badMove(AmMobJACmdMoveWaypoint * amMoveCmd);
static double calcMaxSpeed(double radius);
static double calcMinRadius(double speed);
static void calcSkips();
static int checkDitherWaypoints(AmMobJACmdMoveWaypoint * amMoveCmd,
								int start, int * index, int * savedLength, int * savedIndexes,
								PM_CARTESIAN * savedData);
static int checkObsArc(
					   int ccw, double centerX,  double centerY, double startX, double startY,
					   double endX, double endY);
static int checkObsLineThick(
							 double startX, double startY, double endX, double endY);
static int checkObsLineThickXNN(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static int checkObsLineThickXNP(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static int checkObsLineThickXPN(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static int checkObsLineThickXPP(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static int checkObsLineThickYNN(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static int checkObsLineThickYNP(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static int checkObsLineThickYPN(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static int checkObsLineThickYPP(
								double startX, double startY, double deltaX, double deltaY, int i, int j,
								int dI, int dJ);
static void cleanupAmMobJANmlBuffers();
static void cleanupNmlBuffers();
static void cleanupPrimMobJANmlBuffers();
static void closeLinearizedPath(int * aStarPath, int size, int * newSize);
static void doCfg();
static void doCfgCellResolution();
static void doCfgCycleTime();
static void doCfgDebug();
static void doCfgDumpWM();
static void doCfgNop();
static void doCfgPlanHorizon();
static void doCfgVehicleMinTurnRad();
static void doCfgVehicleWidth();
static void doCmd();
static void doCmdAbort();
static void doCmdHalt();
static void doCmdInit();
static void doCmdMoveWaypoint();
static void doCmdNop();
static void doCmdPrepareMove();
static void doCmdShutdown();
static void doCmdSpin();
static void doXML();
//static PM_CARTESIAN GPStoLocalWaypoint();
//static void drawObsData();
static void drawObsMapBuf(char * buf);
static void drawObsMapFile(PrimMobJACmdMoveArcSegment * primMoveCmd);
static void drawPath(char * buf, PrimMobJACmdMoveArcSegment * primMoveCmd);
static void drawPathLine(char * buf, int i0, int j0, int i1, int j1);
static void drawVehicle(double vehX, double vehY, char * buf);
static void findBestGoal(
						 PM_CARTESIAN * data, int dataIndex, int gcsIndex, int * statuses,
						 double * costs, int gcsIndexStop);
static int findEdgePoint(
						 double x0, double y0, double x1, double y1, double * x, double * y);
int findIdAtCoords(double x, double y);
static double findMapXCoord(int i);
static int findMapXIndex(double x);
static double findMapYCoord(int j);
static int findMapYIndex(double y);
static int findMaxEnd(int m, int n, int * path, int * j);
static int findMinStart(int m, int n, int * path, int * k);
double findPathXCoord(int * path, int index);
double findPathYCoord(int * path, int index);
//static void findPlanTimeCost(
//  PrimMobJACmdMoveArcSegment * plan, double * cost,
//  double penalty, double startX, double startY);
static void getFileNames();
static int getOptions(int argc, char * argv[], int * bufNum);
static void handleDisplay(bool turnOn);
static int iniLoad(char * platformName);
static void iniLoadControlPoint(
								INIFILE * iniFile,  const char * parameter, const char * heading,
								PM_CARTESIAN * tran, PM_QUATERNION * rot, int * retval);
static void iniLoadOneDouble(
							 INIFILE * iniFile, const char * parameter, const char * heading,
							 double * setMe, int * retval);
static void iniLoadOneInt(
						  INIFILE * iniFile, const char * parameter, const char * heading,
						  int * setMe, int * retval);
static void iniLoadOneString(
							 INIFILE * iniFile, const char * parameter, const char * heading,
							 char ** setMe, int * retval);
static int iniSetVehicleParameters(double * width, double * turnRad,
								   char ** platformName); 
static int initAmMobJANmlBuffers(int bufnum);
static int initBuffers(int bufnum);
static void initGoals();
static void initKidsAStar(
						  int dadIndex, int dadRow, int dadCol, int endRow, int endCol,
						  double startX, double startY, double dX, double dY, double dist);
static void initKidsDijk(int dadIndex, int dadRow, int dadCol);
static int initNavDataBuffer(int bufnum);
static int initNodesAStar();
static int initNodesDijk();
static int initParameters();
static int initPrimMobJANmlBuffers(int bufnum);
static int initSensorData2DBuffer(int bufnum);
static int initServoMobJASetBuffer(int bufnum);
//static int intersectLines(
//  double x1, double y1, double a1, double x2, double y2, double a2,
//  double * ix, double * iy);
static int linearizeAStarPath(int m, int n, int * path);
//static int locateCells(PM_CARTESIAN * cells);
int main(int argc, char * argv[]);
static void mergeColinear(int * path, int size);
static double minusPiToPi(double angle);
static int planAll(double tol, double speed);
static void planPathAStar(
						  double startX, double startY, double endX, double endY, int * aStarPath,
						  int * length);
static void planPathsDijk(double startX, double startY, PM_CARTESIAN * cells);
static int planPrimMove(
						AmMobJACmdMoveWaypoint * amMoveCmd,
						PrimMobJACmdMoveArcSegment * primMoveCmd,
						int * primEndsOnWaypoint, int index);
static int planPrimMoveOne(
						   double startX, double startY, double endX, double endY,
						   int * path, int * pathLength);
static int planPrimMoveOneEdge(
							   double xStart, double yStart, double xEdge, double yEdge,
							   int * path, int * pathLength);
static int planPrimMoveSome(
							AmMobJACmdMoveWaypoint * amMoveCmd,
							PrimMobJACmdMoveArcSegment * primMoveCmd,
							int * primEndsOnWaypoint, int index);
static void quit(int sig);
static void readCfg();
static void readCmd();
static void readPrimMobStat();
static void readXML();
static void restoreWaypoints(AmMobJACmdMoveWaypoint * amMoveCmd,
							 int savedLength, int * savedIndexes, PM_CARTESIAN * savedData);
static int roundCorners(
						PrimMobJACmdMoveArcSegment * primMoveCmd, int * path, int length,
						AmMobJACmdMoveWaypoint * amMoveCmd, int * lineToWaypointTable);
static int roundSide(
					 PrimMobJACmdMoveArcSegment * primMoveCmd, int * path, int index,
					 double speedArc, double speedLine, double tol, int wayIndex);
static int roundSideAux(
						PrimMobJACmdMoveArcSegment * primMoveCmd, int ccw, double startX,
						double startY, double endX, double endY, double * tanX, double * tanY,
						double mainDir, double centerDir, double speedArc, double speedLine,
						double tol, int wayIndex);
static void setEvenSerial(int echoed, int * serial);
static void setObsCells(int x, int y, double obsVal);
static int setRelatedValues(double cellRes);
static int updateObsMap();
static int updatePosition();
static int vehicleOffObsMap();
static int writeArc(
					PrimMobJACmdMoveArcSegment * primMoveCmd, int ccw, double centerX,
					double centerY, double startX, double startY, double endX, double endY,
					double speed, double tol, int wayIndex);
static int writeLine(
					 PrimMobJACmdMoveArcSegment * primMoveCmd, double endX, double endY,
					 double speed, double tol, int wayIndex);
static void writeObsMapData(char * fileName);
static void writePrimCmd(RCS_CMD_MSG * cmd, int echoed);
static void writeMobCmd(RCS_CMD_MSG * cmd);
static void writeSPCmd(RCS_CMD_MSG * cmd);
static void writeServoMobCmd(RCS_CMD_MSG* cmd);
static void setOddSerial(int echoed, int * serial);

// WSS function declarations
static bool SendInit(string RobotName, int Port);
static float GetSignalStrength(string RobotName);
static string INITResponse();
static float SignalStrengthResponse(string msg);
static int DNSResponse(string msg);

static bool readConfigFile();

// ServoSP supporting functions
static int initServoSPBuffers(int bufnum);
static void writeServoSPCmd(RCS_CMD_MSG* cmd);

/***********************************************************************/
/***********************************************************************/

//! badMove checks whether an AmMobJACmdMoveWaypoint command is bad.
/*
Returned Value: int
If the number of waypoints is not positive, or there are more than
MAX_ARC_SEGMENTS waypoints, or two adjacent waypoints are closer than
0.1 meter, this returns 1.
Otherwise, it returns 0.

Called By: doCmdMoveWaypoint

This is requiring that successive waypoints be at least 0.1 meter
apart. This is so that arcs with tiny radii will not be forced to be
planned. The planner will report failure of movePoints if the waypoints
are too close together. It might be useful to make this minimum
separation larger so that it will be possible to plan correctly for
vehicles with a minimum turning radius.

The error code for the errors found here is AM_MOB_JA_ERROR_WAYPOINT.
The error code is set in doCmdMoveWaypoint.

If the verbose flag is 2 or more, this will print an error message if
there is an error.

*/

static int badMove(                  /*   ARGUMENTS             */
				   AmMobJACmdMoveWaypoint * amMoveCmd) /*!< move command to check */
{
	int n;
	Waypoint * waypoints;

	if (amMoveCmd->waypoint_length < 1)
	{
		stts.seed_length = 2;
		if (verbose > 1)
			rcs_print_error("%s: number of waypoints is %d but must be positive\n",
			processName, amMoveCmd->waypoint_length);

		return 1;
	}
	if (amMoveCmd->waypoint_length > MAX_ARC_SEGMENTS)
	{
		stts.seed_length = 0;
		if (verbose > 1)
			rcs_print_error("%s: more than %d waypoints\n",
			processName, MAX_ARC_SEGMENTS);

		return 1;
	}
	waypoints = amMoveCmd->waypoint;
	for (n = 1; n < amMoveCmd->waypoint_length; n++)
	{
		if (hypot((waypoints[n].p.x - waypoints[n-1].p.x),
			(waypoints[n].p.y - waypoints[n-1].p.y)) < 0.1)
		{
			stts.seed_length = 3;
			// need to convert to global coordinates
			stts.seed[0] = waypoints[n-1].p;
			stts.seed[0] = localToGlobal * stts.seed[0];
			stts.seed[1] = waypoints[n].p;
			stts.seed[1] = localToGlobal * stts.seed[1];
			stts.seed[2].x = 0.1;
			if (verbose > 1)
				rcs_print_error("%s: waypoints too close together\n", processName);
			return 1;
		}
	}
	return 0;
}

/***********************************************************************/

/*! calcMaxSpeed calculates the speed for a given radius at which
centripetal acceleration is the maximum allowed. */
/*
Returned Value: double
This returns the value of the speed in meters per second

Called By:
roundSideAux
writeArc

The speed is calculated using speed = sqrt(radius * acceleration).

*/

static double calcMaxSpeed( /*   ARGUMENTS                      */
						   double radius)             /*!< radius for which to find speed */
{
	return sqrt(radius * AMOB_MAX_CENTRIPETAL);
}

/***********************************************************************/

/*! calcMinRadius calculates the minimum radius for a given speed at which
centripetal acceleration is the maximum allowed. */
/*
Returned Value: double
This returns the value of the radius in meters

Called By:
roundSideAux

The radius is calculated using radius = ((speed * speed) / acceleration)

*/

static double calcMinRadius( /*   ARGUMENTS                              */
							double speed)               /*!< speed for which to find minimum radius */
{
	return  ((speed * speed) / AMOB_MAX_CENTRIPETAL);
}

/***********************************************************************/

/*! calcSkips calculates obsGrow and how many cells to skip at the ends
of each row of a square containing a grown obstacle. */
/*
Returned Value: none

Called By:
doCfgVehicleWidth
setRelatedValues

Where obsGrow is set, the extra 1 is added to allow for truncating the
int.

The -0.00001 in the setting of obsGrow is to prevent obsGrow from
being too large when half the vehicle width is evenly divisible by
the cell resolution.

A cell is skipped if no point in the cell is within a given distance
(the radius argument) of the center cell of the square. The radius is
expected to be half the width of the vehicle. Results are stored in
the global skips array.

The radius is first converted to a multiple of the length of the side
of a cell. After this conversion, obsGrow is the smallest integer
greater than or equal to the radius; obsGrow is calculated in
setRelatedValues.

The square is ((2 * obsGrow) + 1) cells on a side.  The obstacle cell
being grown is the center cell of the square.  It is assumed this
center cell is entirely full of obstacle. In each corner of the
square, the cells in each row to skip are those lying entirely outside
of a quarter circle whose center is at the nearest corner of the
center cell and whose radius is the given radius.

An example is shown below. In the example, obsGrow is 5, so "side" is
11; radius (after conversion) is 4.9.  The center cell is represented
by M, the cells to skip are shown represented by a dot, and the other
cells are represented by X. The values of the skips array that would
be calculated are shown at the left of the example.

2   . . X X X X X X X . .
1   . X X X X X X X X X .
0   X X X X X X X X X X X
0   X X X X X X X X X X X
0   X X X X X X X X X X X
0   X X X X X M X X X X X
0   X X X X X X X X X X X
0   X X X X X X X X X X X
0   X X X X X X X X X X X
1   . X X X X X X X X X .
2   . . X X X X X X X . .

Since the figure is always symmetric about horizontal and vertical lines
through the center cell, the same skips value applies at both ends of
each row, and the skips array itself is symmetric about the middle
(so only half of it needs to be calculated).

In the code, the distances being compared are (1) the converted
radius, and (2) the distance (in numbers of cell sides) between the
lower left corner of the center cell and the upper right corner of
a cell at the lower left of the figure (since those are the closest
points of the two cells).

The middle three rows and middle three columns can never have a skip
cell, so they are not calculated, but they (along with the rest of the
skips array) are initialized to 0.

*/

static void calcSkips() /* NO ARGUMENTS */
{
	int side;
	double radius;
	int row;
	int col;
	double vehSize;

	vehSize = (MAX(set.vehicleWidth, vehLength) - 0.00001); // temporary
	obsGrow = (0 + (int)(vehSize / (4.5 * res)));
	radius = (vehSize / (2 * res));
	side = ((2 * obsGrow) + 1);
	if (skips)
		delete skips;
	skips = new int[side];
	for (row = 0; row < side; row++)
	{
		skips[row] = 0;
	}
	for (row = 1; row < obsGrow; row++)
	{
		for (col = 1; col < obsGrow; col++)
		{
			if (hypot((double)(obsGrow - row), (double)(obsGrow - col)) < radius)
				break;
		}
		skips[row - 1] = (col - 1);
		skips[side - row] = (col - 1);
	}
}

/***********************************************************************/

/* checkDitherWaypoints

Returned Value: int (see below)

Called By: planPrimMoveSome

The *index argument is the index of the next waypoint to visit from
the current position. Call the waypoint at that index P1. This checks
whether the waypoints in a certain set S do not lie in obstacle
cells. S consists of those waypoints starting with P1 and ending with
the waypoint preceding the first waypoint that is outside the planning
horizon.

If any waypoint in S is on an obstacle, this attempts to find a nearby
replacement point that is not on or near an obstacle. If such a point
can be found, the coordinates of the waypoint are saved (in the x and
y coordinates of a cartesian point) and replaced by the coordinates of
the nearby point. The neighborhood of the waypoint is also saved (in
the z coordinate of the cartesian point) and replaced.

A square area is checked in order to find a replacement point. The
square is originally the largest square of whole cells that will fit
inside a circle whose center is at the center of the cell containing
the waypoint and whose radius is the neighborhood of the waypoint.
But then the side of the square is reduced by four times the
resolution (i.e. a strip two cells wide is removed all around the
edges of the square) so that a neighborhood of at least twice the
resolution can be left around the replacement point without going
outside the original neighborhood. To work with the search algorithm
being used, the square must have an odd number of cells on a side, so
if the number of cells on a side is even, the side is decreased by one
more.

If all the waypoints (after some are moved, possibly) inside the
planning horizon are not on an obstacle, this returns 0.
Otherwise, it returns 1.

If there is a waypoint in S that is on an obstacle, and no nearby
replacement can be found, the function sets stts.amMobJAError to
AM_MOB_JA_ERROR_WAYPOINT, sets stts.seed_length to 1, sets stts.seed[0]
to the first waypoint in S that has the problem, and returns 1.

The saved data from any waypoints that have been moved are put back
into the amMoveCmd in planPrimMoveSome by restoreWaypoints, after the
Prim move command has been planned or planPrimMoveSome has failed.

This also resets *index to the index of the first waypoint whose index
is start or larger that lies outside the planning horizon.

The procedure here for visiting all the cells inside a square area
starts in the middle cell of the square (a middle cell always exists
since the number of cells on a side is odd) and then going around
expanding square annuluses of cells in a spiral-like pattern.

It might be useful to hunt for replacement points inside a circle
rather than a square, since that would test more points. The procedure
could be made fast by precalculating things similarly to the way
calcSkips precalculates data for growing obstacles.

*/

static int checkDitherWaypoints(     /*   ARGUMENTS                         */
								AmMobJACmdMoveWaypoint * amMoveCmd, /*!< AM move command to check          */
								int start,                          /*!< waypoint index at which to start  */
								int * index,                        /*!< waypoint index, reset here        */
								int * savedLength,                  /*!< number of saved waypoints         */
								int * savedIndexes,                 /*!< indexes of saved waypoints        */
								PM_CARTESIAN * savedData)           /*!< x, y, and nghd of saved waypoints */
{
	Waypoint * aWaypoint; // waypoint being checked
	double x;             // X coordinate of aWaypoint
	double y;             // Y coordinate of aWaypoint
	double nghd;          // neighborhood of aWaypoint
	int n;                // spiral side length
	int k;                // spiral side iterator
	int delta;            // +1 or -1 (amount by which to change i or j)
	int i;                // obsMap X index of point, reset in loop
	int j;                // obsMap Y index of point, reset in loop
	int side;             // number of cells on the side of the search square
	int found;            // 1=found a replacement cell, 0=not

	*savedLength = 0;
	for (*index = start; *index < amMoveCmd->waypoint_length; (*index)++)
	{ // check for obstacles at waypoints inside planning horizon
		aWaypoint = (amMoveCmd->waypoint + *index);
		x = aWaypoint->p.x;
		y = aWaypoint->p.y;
		if ((x > xMax) || (x < xMin) || (y < yMin) || (y > yMax))
			break;
		i = findMapXIndex(x);
		j = findMapYIndex(y);
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // try to find a nearby point not on an obstacle to replace waypoint
			nghd = aWaypoint->neighborhood;
			side = ((int)((nghd * 1.414) / res) - 4);
			side = ((side % 2) ? side : (side - 1)); // need odd side
			delta = 1;
			found = 0;
			for (n = 1; n < side; n++)
			{ // two sides of the spiral are checked each time around
				for (k = 0; k < n; k++)
				{ // check cells on side parallel to X axis
					i = (i + delta);
					if (obsMap[i][j] < AMSP_NEAR_OBSTACLE)
					{
						found = 1;
						break;
					}
				}
				if (found)
					break;
				for (k = 0; k < n; k++)
				{ // check cells on side parallel to Y axis
					j = (j + delta);
					if (obsMap[i][j] < AMSP_NEAR_OBSTACLE)
					{
						found = 1;
						break;
					}
				}
				if (found)
					break;
				delta = -delta;
			}
			if (!found)
			{ // did not find replacement point yet, check last side of square
				n--;
				for (k = 0; k < n; k++)
				{ // check cells on side parallel to X axis
					i = (i + delta);
					if (obsMap[i][j] < AMSP_NEAR_OBSTACLE)
					{
						found = 1;
						break;
					}
				}
			}
			if (found)
			{ // found a replacement point
				savedIndexes[*savedLength] = *index;
				savedData[*savedLength].x = x;
				savedData[*savedLength].y = y;
				savedData[*savedLength].z = nghd;
				aWaypoint->p.x = findMapXCoord(i);
				aWaypoint->p.y = findMapYCoord(j);
				aWaypoint->neighborhood = (2 * res);
				(*savedLength)++;
			}
			else
			{ // did not find a replacement point
				stts.amMobJAError = AM_MOB_JA_ERROR_WAYPOINT;
				stts.seed_length = 1;
				stts.seed[0] = aWaypoint->p;
				// convert to global
				stts.seed[0] = localToGlobal * stts.seed[0];
				if (verbose > 1)
					rcs_print("%s: waypoint (%.2f, %.2f) is on obstacle\n",
					processName, x, y);
				return 1;
			}
		}
	}
	return 0;
}

/***********************************************************************/

/*! checkObsArc checks if an arc segment hits an obstacle or passes outside
the obstacle map */
/*
Returned Value: int
If the arc passes outside the obsMap or passes through an obstacle,
this returns 1. Otherwise, it returns 0.

Called By:
roundSideAux

This divides the arc into equal pieces (pick a piece and call it A)
small enough that a rectangle with its long side on the chord C between
the ends of A and its short side equal to (2 * set.cellResolution) will
cover A entirely.  Then the line C' is found by offsetting C towards A
half the maximum distance from A to C.

Each of the C' lines is tested with checkObsLineThick, which is at
least (2 * set.cellResolution) thick. If an obstacle is hit during
testing this returns 1 immediately.  The C' lines are extended a very
small amount at the ends to intersect each other, since that is
simpler (half as many points are needed and the computation is
shorter) than generating disconnected line segments.

It is easily calculated that if a circle of radius R is divided into N
equal arcs, the maximum distance d from a chord on one of the arcs
to the arc is given by d = (R - (R * cos((Pi / N))). This assumes R > d.

Solving for N:  N = (Pi / acos((R - d) / R))

So for an arc of angle theta, to make the maximum distance d, we need
at least M = (N * (theta / (2 * Pi))) equal arcs.
Substituting for N:

M = ((Pi / acos((R - d) / R)) * (theta / (2 * Pi)))

M = (theta / (2 * acos((R - d) / R)))

Substituting (2 * set.cellResolution) for d:

M = (theta / (2 * acos((R - (2 * set.cellResolution)) / R)))

(which we round up to the nearest integer)

For an arc with included angle delta, with radii drawn to the ends
of the arc and extended by equal amounts, if a line is drawn between
the ends of the extended radii, the increase in radius required
to make that line be a distance b from the chord on the arc
is (b / cos(delta/2)). We are using b = set.cellResolution.

Arcs should not pass outside the obsMap; this returns 1 if an arc
passes outside. It is actually testing that the endpoints of [the
lines through the middles of the rectangles containing segments of the
arc] are on the obsMap. It is possible, therefore, that an arc may
pass one cell outside the obsMap and this function will return 0.

*/

static int checkObsArc( /*   ARGUMENTS                                  */
					   int ccw,               /*!< non-zero=counterclockwise, zero=clockwise  */
					   double centerX,        /*!< X coordinate of end point of arc           */
					   double centerY,        /*!< Y coordinate of end point of arc           */
					   double startX,         /*!< X coordinate of start point of arc         */
					   double startY,         /*!< Y coordinate of start point of arc         */
					   double endX,           /*!< X coordinate of end point of arc           */
					   double endY)           /*!< Y coordinate of end point of arc           */
{
	int M;            // number of subdivisions
	int n;            // loop counter
	double R;         // radius of circle
	double sX;        // X coordinate of start point of small arc
	double sY;        // Y coordinate of start point of small arc
	double eX;        // X coordinate of end point of small arc
	double eY;        // Y coordinate of end point of small arc
	double theta;     // total angle of arc
	double beta;      // angle of point, reset in loop
	double delta;     // increment of angle
	int i;            // obsMap X index of point, reset in loop
	int j;            // obsf Y index of point, reset in loop

	beta = atan2((startY - centerY), (startX - centerX));
	theta = (atan2((endY - centerY), (endX - centerX)) - beta);
	if (ccw && (theta < 0))
		theta = (theta + (2 * M_PI));
	else if ((!ccw) && (theta > 0))
		theta = (theta - (2 * M_PI));
	R = hypot((centerY - startY), (centerX - startX));
	if (R > (2 * res))
		M = (int)(1 + ABS(theta / (2 * acos((R - (2 * res)) / R))));
	else // very small radius
		M = 4;
	delta = (theta / M);
	R = (R + (res / cos(delta / 2.0)));
	eX = (centerX + (R * cos(beta))); // prepare for loop
	eY = (centerY + (R * sin(beta))); // prepare for loop
	for (n = 0; n < M; n++)
	{
		sX = eX;
		sY = eY;
		i = findMapXIndex(sX);
		j = findMapYIndex(sY);
		if ((i < 0) || (i > obsMapSi) || (j < 0) || (j > obsMapSi))
			return 1;
		beta = (beta + delta);
		eX = (centerX + (R * cos(beta)));
		eY = (centerY + (R * sin(beta)));
		i = findMapXIndex(eX);
		j = findMapYIndex(eY);
		if ((i < 0) || (i > obsMapSi) || (j < 0) || (j > obsMapSi))
			return 1;
		if (checkObsLineThick(sX, sY, eX, eY))
			return 1;
	}
	return 0;
}

/***********************************************************************/

//! checkObsLineThick checks if a thick line segment hits an obstacle.
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By:
checkObsArc
findMaxEnd
findMinStart
planPrimMoveOne
roundSideAux

This function thickens the line by one cell on each side when it tests
for obstacles. This is to provide for a little clearance from obstacles
(which are dynamic and may move).

The end points of the line being checked must lie inside the map.

*/

static int checkObsLineThick( /* ARGUMENTS                           */
							 double startX,     /*!< X coordinate of start point of line segment */
							 double startY,     /*!< Y coordinate of start point of line segment */
							 double endX,       /*!< X coordinate of end point of line segment   */
							 double endY)       /*!< Y coordinate of end point of line segment   */
{
	int i;          // X index in obsMap of cell
	int j;          // Y index in obsMap of cell
	int dI;         // amount by which i has changed
	int dJ;         // amount by which j has changed
	double deltaX;  // change in X from start to end
	double deltaY;  // change in Y from start to end

	i = findMapXIndex(startX);
	j = findMapYIndex(startY);
	dI = 0;
	dJ = 0;
	deltaX = (endX - startX);
	deltaY = (endY - startY);
	if (obsMap[i][j] == AMSP_OBSTACLE)
	{ // cell b
		return 1;
	}
	if (fabs(deltaX) > fabs(deltaY))
	{ //step1 in X direction
		if (deltaX > 0)
		{ // step1 towards positive X
			if (deltaY >= 0)
			{ // Y increasing
				return checkObsLineThickXPP(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
			else // if (deltaY < 0)
			{ // Y decreasing
				return checkObsLineThickXPN(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
		}
		else // if (deltaX < 0)
		{ // step1 towards negative X
			if (deltaY >= 0)
			{ // Y increasing
				return checkObsLineThickXNP(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
			else // if (deltaY < 0)
			{ // Y decreasing
				return checkObsLineThickXNN(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
		}
	}
	else // if (fabs(deltaY) > fabs(deltaX))
	{ // step1 in the Y direction
		if (deltaY > 0)
		{ // step1 towards positive Y
			if (deltaX >= 0)
			{ // X increasing
				return checkObsLineThickYPP(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
			else // if (deltaX < 0)
			{ // X decreasing
				return checkObsLineThickYPN(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
		}
		else // if (deltaY < 0)
		{ // step1 towards negative Y
			if (deltaX >= 0)
			{ // X increasing
				return checkObsLineThickYNP(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
			else // if (deltaX < 0)
			{ // X decreasing
				return checkObsLineThickYNN(startX, startY, deltaX, deltaY,
					i, j, dI, dJ);
			}
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickXNN carries out checkObsLineThick when
(1) deltaX is not less than deltaY in absolute value.
(2) deltaX is negative, so the primary step is towards -X.
(3) deltaY is negative, so the secondary step is towards -Y. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

See documentation of checkObsLineThickXPP. This is very similar.  In
that function, the primary step is towards +X and the secondary step
is towards +Y. In this function the primary step is towards -X and the
secondary step is towards -Y.

----------------
|c   |b / |a   |
|    | /  |    |
-----------/---------
|    |    /    |    |
|g   |f  /|e   |d   |
--------/------------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickXNN( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaY / deltaX);
	step1 = (startX - (obsData->lowerLeft.x + (i * res)));
	deltaI = (int)((-deltaX + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (step1 * slope);
	lack = (startY - (obsData->lowerLeft.y + (j * res)));
	deltaJ = (int)((-deltaY + res - lack - AMOB_TINY_DIST)/ res);
	if (step2 > lack)
	{ // line passes through cell c
		j--;
		dJ++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (step1 * slope);
	for (i-- ; ; i--)
	{ // do rest of line
		dI++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((j < obsMapSi) && (obsMap[i][j+1] == AMSP_OBSTACLE)) ||
			((j > 0) && (obsMap[i][j-1] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			j--;
			dJ++;
			if ((j > 0) && (obsMap[i][j-1] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickXNP carries out checkObsLineThick when
(1) deltaX is not less than deltaY in absolute value.
(2) deltaX is negative, so the primary step is towards -X.
(3) deltaY is positive, so the secondary step is towards +Y. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

See documentation of checkObsLineThickXPP. This is very similar.
In that function, the primary step is towards +X. In this function
the primary step is towards -X.

----------------
|a   |b\  |c   |
|    |  \ |    |
---------\-----------
|    |    \    |    |
|d   |e   |\  f|   g|
------------\--------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickXNP( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaY / deltaX);
	step1 = (startX - (obsData->lowerLeft.x + (i * res)));
	deltaI = (int)((-deltaX + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (-1 * step1 * slope);
	lack = ((obsData->lowerLeft.y + ((j + 1) * res)) - startY);
	deltaJ = (int)((deltaY + res - lack - AMOB_TINY_DIST) / res);
	if (step2 > lack)
	{ // line passes through cell c
		j++;
		dJ++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (-1 * step1 * slope);
	for (i-- ; ; i--)
	{ // do rest of line
		dI++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((j > 0) && (obsMap[i][j-1] == AMSP_OBSTACLE)) ||
			((j < obsMapSi) && (obsMap[i][j+1] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			j++;
			dJ++;
			if ((j < obsMapSi) && (obsMap[i][j+1] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickXPN carries out checkObsLineThick when
(1) deltaX is not less than deltaY in absolute value.
(2) deltaX is positive, so the primary step is towards +X.
(3) deltaY is negative, so the secondary step is towards -Y. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

See documentation of checkObsLineThickXPP. This is very similar.
In that function, the secondary step is towards +Y. In this function
the secondary step is towards -Y.

--------\------------
|    |   \|    |    |
|g   |f   |e   |d   |
-----------\---------
|c   |b\  |a   |
|    |  \ |    |
----------------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickXPN( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaY / deltaX);
	step1 = ((obsData->lowerLeft.x + ((i + 1) * res)) - startX);
	deltaI = (int)((deltaX + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (-1 * step1 * slope);
	lack = (startY - (obsData->lowerLeft.y + (j * res)));
	deltaJ = (int)((-deltaY + res - lack - AMOB_TINY_DIST)/ res);
	if (step2 > lack)
	{ // line passes through cell c
		j--;
		dJ++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (-1 * step1 * slope);
	for (i++ ; ; i++)
	{ // do rest of line
		dI++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((j < obsMapSi) && (obsMap[i][j+1] == AMSP_OBSTACLE)) ||
			((j > 0) && (obsMap[i][j-1] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			j--;
			dJ++;
			if ((j > 0) && (obsMap[i][j-1] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickXPP carries out checkObsLineThick when
(1) deltaX is not less than deltaY in absolute value.
(2) deltaX is positive, so the primary step is towards +X.
(3) deltaY is positive, so the secondary step is towards +Y. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

This function works by following along the line from startX to endX
from the top (i.e. the +X side) of one cell to the top either of the cell
above or the cell above and to the right (i.e. in the +Y direction).

step1 is the distance necessary to go in the +X direction to get to
the next top. Since the start point is usually not at the top of a
cell, step1 is originally set to the X distance from the start point
to the top of the cell containing the start point. After the first
step up has been taken, step1 is set to the height of the cell (res).

step2 is the distance moved in Y when X has changed by step1, so it
starts at (step1 * slope) and changes to (res * slope) after the first
step up.

lack is the distance from the current point to the right side of the
cell in which the current point lies. Each time the current point
moves up, lack decreases by step2. Each time the current point passes
into the cell to the right (and out through the top), lack also
increases by res.

Before moving up each time, it is checked whether the current cell,
the cell to its right, or the cell to its left is an obstacle.
In both figures below, when cell e is checked, so are cells d
and f. If the line goes cell f (as in the figure on the left),
it is also checked whether cell g is an obstacle.

Line does not pass through c        Line passes through c

------------/--------             ---------/-----------
|    |    |/   |    |             |    |  / |    |    |
|d   |e   |f   |g   |             |d   |e/  |f   |g   |
---------/-----------        -----------/---------
|a   |b / |c   |             |a   |b   /   c|
|    | /  |    |             |    |   /|    |
----------------              ----------------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickXPP( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaY / deltaX);
	step1 = ((obsData->lowerLeft.x + ((i + 1) * res)) - startX);
	deltaI = (int)((deltaX + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (step1 * slope);
	lack = ((obsData->lowerLeft.y + ((j + 1) * res)) - startY);
	deltaJ = (int)((deltaY + res - lack - AMOB_TINY_DIST) / res);
	if (step2 > lack)
	{ // line passes through cell c
		j++;
		dJ++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (step1 * slope);
	for (i++ ; ; i++)
	{ // do rest of line
		dI++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((j > 0) && (obsMap[i][j-1] == AMSP_OBSTACLE)) ||
			((j < obsMapSi) && (obsMap[i][j+1] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			j++;
			dJ++;
			if ((j < obsMapSi) && (obsMap[i][j+1] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickYNN carries out checkObsLineThick when
(1) deltaX is less than deltaY in absolute value.
(2) deltaY is negative, so the primary step is towards -Y.
(3) deltaX is negative, so the secondary step is towards -X. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

See documentation of checkObsLineThickXPP. This is very similar.  In
that function, the primary step is towards +X and the secondary step
is towards +Y. In this function the primary step is towards -Y and the
secondary step is towards -X.

The cells are shown stretched in the Y direction here for drawing
purposes. They are actually square. The diagonal line is actually
closer to horizontal than to vertical.

-----------
|    |    |
|    |    |
|d   |a   |
|----------
|    |/   |
|e   /b   |
|   /|    |
---/-------
|f/  |c   |
|/   |    |
/    |    |
/-----------
|    |
|    |
|g   |
------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickYNN( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaX / deltaY);
	step1 = (startY - (obsData->lowerLeft.y + (j * res)));
	deltaJ = (int)((-deltaY + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (step1 * slope);
	lack = (startX - (obsData->lowerLeft.x + (i * res)));
	deltaI = (int)((-deltaX + res - lack - AMOB_TINY_DIST)/ res);
	if (step2 > lack)
	{ // line passes through cell c
		i--;
		dI++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (step1 * slope);
	for (j-- ; ; j--)
	{ // do rest of line
		dJ++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((i < obsMapSi) && (obsMap[i+1][j] == AMSP_OBSTACLE)) ||
			((i > 0) && (obsMap[i-1][j] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			i--;
			dI++;
			if ((i > 0) && (obsMap[i-1][j] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickYNP carries out checkObsLineThick when
(1) deltaX is less than deltaY in absolute value.
(2) deltaY is negative, so the primary step is towards -Y.
(3) deltaX is positive, so the secondary step is towards +X. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

See documentation of checkObsLineThickXPP. This is very similar.  In
that function, the primary step is towards +X and the secondary step
is towards +Y. In this function the primary step is towards -Y and the
secondary step is towards +X.

The cells are shown stretched in the Y direction here for drawing
purposes. They are actually square. The diagonal line is actually
closer to horizontal than to vertical.

------
|    |
|    |
|g   |
\-----------
\    |    |
|\   |    |
|f\  |c   |
|--\-------
|   \|    |
|e   \b   |
|    |\   |
-----------
|d   |a   |
|    |    |
|    |    |
-----------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickYNP( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaX / deltaY);
	step1 = (startY - (obsData->lowerLeft.y + (j * res)));
	deltaJ = (int)((-deltaY + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (-1 * step1 * slope);
	lack = ((obsData->lowerLeft.x + ((i + 1) * res))- startX);
	deltaI = (int)((deltaX + res - lack - AMOB_TINY_DIST)/ res);
	if (step2 > lack)
	{ // line passes through cell c
		i++;
		dI++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (-1 * step1 * slope);
	for (j-- ; ; j--)
	{ // do rest of line
		dJ++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((i > 0) && (obsMap[i-1][j] == AMSP_OBSTACLE)) ||
			((i < obsMapSi) && (obsMap[i+1][j] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			i++;
			dI++;
			if ((i < obsMapSi) && (obsMap[i+1][j] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickYPN carries out checkObsLineThick when
(1) deltaX is less than deltaY in absolute value.
(2) deltaY is positive, so the primary step is towards +Y.
(3) deltaX is negative, so the secondary step is towards -X. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

See documentation of checkObsLineThickXPP. This is very similar.  In
that function, the primary step is towards +X and the secondary step
is towards +Y. In this function the primary step is towards +Y and the
secondary step is towards -X.

The cells are shown stretched in the Y direction here for drawing
purposes. They are actually square. The diagonal line is actually
closer to horizontal than to vertical.

-----------
|    |    |
|    |    |
|a   |d   |
-----------
|   \|    |
|b   \e   |
|    |\   |
-------\---
|c   |f \ |
|    |   \|
|    |    \
-----------\
|    |
|    |
|g   |
------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickYPN( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaX / deltaY);
	step1 = ((obsData->lowerLeft.y + ((j + 1) * res)) - startY);
	deltaJ = (int)((deltaY + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (-1 * step1 * slope);
	lack = (startX - (obsData->lowerLeft.x + (i * res)));
	deltaI = (int)((-deltaX + res - lack - AMOB_TINY_DIST)/ res);
	if (step2 > lack)
	{ // line passes through cell c
		i--;
		dI++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (-1 * step1 * slope);
	for (j++ ; ; j++)
	{ // do rest of line
		dJ++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((i < obsMapSi) && (obsMap[i+1][j] == AMSP_OBSTACLE)) ||
			((i > 0) && (obsMap[i-1][j] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			i--;
			dI++;
			if ((i > 0) && (obsMap[i-1][j] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! checkObsLineThickYPP carries out checkObsLineThick when
(1) deltaX is less than deltaY in absolute value.
(2) deltaY is positive, so the primary step is towards +Y.
(3) deltaX is positive, so the secondary step is towards +X. */
/*
Returned Value: int
If no obstacle is hit, this returns 0.
If an obstacle is hit, this returns 1.

Called By: checkObsLineThick

See documentation of checkObsLineThickXPP. This is very similar.  In
that function, the primary step is towards +X and the secondary step
is towards +Y. In this function the primary step is towards +Y and the
secondary step is towards +X.

The cells are shown stretched in the Y direction here for drawing
purposes. They are actually square. The diagonal line is actually
closer to horizontal than to vertical.

------
|    |
|    |
|g   |
-----------/
|    |    /
|    |   /|
|c   |f / |
-------/---
|    |/   |
|b   /e   |
|   /|    |
-----------
|a   |d   |
|    |    |
|    |    |
-----------

The line starts in cell b and may or may not pass through cell c. Cell
b is checked for being an obstacle by checkObsLineThick before this
function is called. This function is not called if cell b is an
obstacle.

*/

static int checkObsLineThickYPP( /* ARGUMENTS                            */
								double startX,     /*!< X coordinate of start point of line segment     */
								double startY,     /*!< Y coordinate of start point of line segment     */
								double deltaX,     /*!< change in X from start to end                   */
								double deltaY,     /*!< change in Y from start to end                   */
								int i,             /*!< X index in obsMap of cell (initially at startX) */
								int j,             /*!< Y index in obsMap of cell (initially at startY) */
								int dI,            /*!< amount by which i has changed (always positive) */
								int dJ)            /*!< amount by which j has changed (always positive) */
{
	int deltaI;     // amount by which i must change (always positive)
	int deltaJ;     // amount by which j must change (always positive)
	double slope;   // slope with respect to primary direction
	double lack;    // distance needed in secondary direction (always positive)
	double step1;   // size of step in primary direction (always positive)
	double step2;   // size of step in secondary direction (always positive)

	slope = (deltaX / deltaY);
	step1 = ((obsData->lowerLeft.y + ((j + 1) * res)) - startY);
	deltaJ = (int)((deltaY + res - step1 - AMOB_TINY_DIST) / res);
	step2 = (step1 * slope);
	lack = ((obsData->lowerLeft.x + ((i + 1) * res)) - startX);
	deltaI = (int)((deltaX + res - lack - AMOB_TINY_DIST) / res);
	if (step2 > lack)
	{ // line passes through cell c
		i++;
		dI++;
		if (obsMap[i][j] == AMSP_OBSTACLE)
		{ // cell c
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		lack = (res - (step2 - lack));
	}
	else
	{  // line does not pass through cell c
		lack = (lack - step2);
	}
	step1 = res;
	step2 = (step1 * slope);
	for (j++ ; ; j++)
	{ // do rest of line
		dJ++;
		if ((obsMap[i][j] == AMSP_OBSTACLE) ||
			((i > 0) && (obsMap[i-1][j] == AMSP_OBSTACLE)) ||
			((i < obsMapSi) && (obsMap[i+1][j] == AMSP_OBSTACLE)))
		{
			return 1;
		}
		if ((dI >= deltaI) && (dJ >= deltaJ))
			return 0;
		if (step2 > lack)
		{ // line passes through cell f
			i++;
			dI++;
			if ((i < obsMapSi) && (obsMap[i+1][j] == AMSP_OBSTACLE))
			{ // cell g
				return 1;
			}
			if ((dI >= deltaI) && (dJ >= deltaJ))
				return 0;
			lack  = (res - (step2 - lack));
		}
		else
		{ // line does not pass through cell f
			lack = (lack - step2);
		}
	}
	return 1; // should never get here
}

/***********************************************************************/

/*! cleanupAmMobJANmlBuffers cleans up NML buffers between AM mobility
and its superior (vehicle mobility or a command shell). */
/*
Returned Value: none

Called By:
cleanupNmlBuffers
initAmMobJANmlBuffers

*/

static void cleanupAmMobJANmlBuffers() /* NO ARGUMENTS */
{
	if (NULL != amMobJACmdBuf)
		delete amMobJACmdBuf;
	if (NULL != amMobJAStatBuf)
		delete amMobJAStatBuf;
	if (NULL != amMobJACfgBuf)
		delete amMobJACfgBuf;
	if (NULL != amMobJASetBuf)
		delete amMobJASetBuf;

	//! AM SP Cmd/Stat buffers
	if (NULL != amSPCmdBuf)
		delete amSPCmdBuf;
	if (NULL != amSPStatBuf)
		delete amSPStatBuf;

	//! ServoMobJA Cmd buffer 
	if (NULL != servoMobJACmdBuf)
		delete servoMobJACmdBuf;

	//! ServoMobJA Stat buffer
	if (NULL != servoMobJAStatBuf)
		delete servoMobJAStatBuf;

	amMobJACmdBuf = NULL;
	amMobJAStatBuf = NULL;
	amMobJACfgBuf = NULL;
	amMobJASetBuf = NULL;

	amSPCmdBuf = NULL;
	amSPStatBuf = NULL;

	servoMobJACmdBuf = NULL;
	servoMobJAStatBuf = NULL;
}

/***********************************************************************/

//! cleanupNmlBuffers cleans up all the NML buffers
/*
Returned Value: none

Called By:
initBuffers
main 

*/

static void cleanupNmlBuffers() /* NO ARGUMENTS */
{
	cleanupAmMobJANmlBuffers();
	cleanupPrimMobJANmlBuffers();
	if (NULL != navDataExtBuf)
		delete navDataExtBuf;
	navDataExtBuf = NULL;
	if (NULL != sensorData2DBuf)
		delete sensorData2DBuf;
	sensorData2DBuf = NULL;
	if (NULL != servoMobJASetBuf)
		delete servoMobJASetBuf;
	sensorData2DBuf = NULL;

	// Servo SP Buffers
	if (NULL != servoSPCmdBuf)
		delete servoSPCmdBuf;
	servoSPCmdBuf = NULL;
	if (NULL != servoSPStatBuf)
		delete servoSPStatBuf;
	servoSPStatBuf = NULL;
	if (NULL != servoSPSetBuf)
		delete servoSPSetBuf;
	servoSPSetBuf = NULL;
}

/***********************************************************************/

/*! cleanupPrimMobJANmlBuffers cleans up NML buffers between AM mobility
and Prim mobility. */
/*
Returned Value: none

Called By:
cleanupNmlBuffers
initBuffers
initPrimMobJANmlBuffers

*/

static void cleanupPrimMobJANmlBuffers() /* NO ARGUMENTS */
{
	if (NULL != primMobJACmdBuf)
		delete primMobJACmdBuf;
	if (NULL != primMobJAStatBuf)
		delete primMobJAStatBuf;
	if (NULL != primMobJACfgBuf)
		delete primMobJACfgBuf;
	if (NULL != primMobJASetBuf)
		delete primMobJASetBuf;

	primMobJACmdBuf = NULL;
	primMobJAStatBuf = NULL;
	primMobJACfgBuf = NULL;
	primMobJASetBuf = NULL;
}

/***********************************************************************/

//! closeLinearizedPath removes -1's from a path and resets size
/*
Returned Value: none

Called By: planPrimMoveOne

When this is called, the path is an array of node ids, representing
a sequence of straight lines from cell center to cell center. The cells
have the same ids as the nodes. Some of the node ids may be -1, indicating
that that array position should be ignored. The first and last node ids
are never -1.

The first point on the incoming path is the start point. The function
removes the first point so that calling functions will not have to do
so. Thus, the newSize always is reduced by at least 1 from the size.

This function:
removes the first point,
removes the -1 values,
closes up the remaining values,
sets newSize.

For example, if the starting size is 8 and the path is
{3, 4, -1, 2, -1, -1, 9, 7, ...}, then newSize will be set to 4,
and the path will become {4, 2, 9, 7, ...}. The ...'s mean that
the array may be larger than the path it contains. The values
in the extra slots of the array are irrelevant.

*/

static void closeLinearizedPath( /*   ARGUMENTS                        */
								int * path,                     /*!< path to close, array of node ids */
								int size,                       /*!< starting length of path          */
								int * newSize)                  /*!< length of path, reset here       */
{
	int gap;
	int n;

	gap = 1;
	*newSize = (size - 1);
	for (n = 1; n < size; n++)
	{
		if (path[n] == -1)
		{
			gap++;
			(*newSize)--;
		}
		else
			path[n - gap] = path[n];
	}
}

/***********************************************************************/

//! doCfg executes config commands
/*
Returned Value: none

Called By: main

If admin_state is ADMIN_SHUT_DOWN, this returns immediately.

*/

static void doCfg()  /* NO ARGUMENTS */
{
	if (stts.admin_state == ADMIN_SHUT_DOWN)
	{ // do nothing if shut down
		return;
	}
	switch (set.command_type)
	{
	case AM_MOB_JA_CFG_CELL_RESOLUTION_TYPE:
		doCfgCellResolution();
		break;
	case AM_MOB_JA_CFG_CYCLE_TIME_TYPE:
		doCfgCycleTime();
		break;
	case AM_MOB_JA_CFG_DEBUG_TYPE:
		doCfgDebug();
		break;
	case AM_MOB_JA_CFG_DUMP_WM_TYPE:
		doCfgDumpWM();
		break;
	case AM_MOB_JA_CFG_NOP_TYPE:
		doCfgNop();
		break;
	case AM_MOB_JA_CFG_PLAN_HORIZON_TYPE:
		doCfgPlanHorizon();
		break;
	case AM_MOB_JA_CFG_VEHICLE_MIN_TURN_RAD_TYPE:
		doCfgVehicleMinTurnRad();
		break;
	case AM_MOB_JA_CFG_VEHICLE_WIDTH_TYPE:
		doCfgVehicleWidth();
		break;
	default:
		// no config to handle
		break;
	}
}

/***********************************************************************/

//! doCfgCellResolution sets the configuration cell resolution
/* 
Returned Value: none

Called By: doCfg

The cell resolution cannot be changed. If the commanded resolution is
the same as the current resolution, this reports done immediately.  If
the commanded resolution differs from the current resolution, this
sends an error message.  To change the cell resolution, the AM Mob
process must be stopped and restarted.

*/

static void doCfgCellResolution() /* NO ARGUMENTS */
{
	AmMobJACfgCellResolution * cfg;

	cfg = (AmMobJACfgCellResolution *)(amMobJACfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		if (cfg->cellResolution != set.cellResolution)
		{
			rcs_print_error("%s: cannot change cell resolution while running\n",
				processName);
			status_next(&set, RCS_ERROR);
		}
		else
			status_next(&set, RCS_DONE);
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCfgCycleTime sets the configuration cycle time and the timer
/* 
Returned Value: none

Called By: doCfg

The old timer is deleted, and a new one is made.

*/

static void doCfgCycleTime() /* NO ARGUMENTS */
{
	AmMobJACfgCycleTime * cfg;

	cfg = (AmMobJACfgCycleTime *)(amMobJACfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		if (cfg->cycleTime <= 0.0)
		{
			status_next(&set, RCS_ERROR);
			rcs_print_error("%s: cycleTime must be positive\n", processName);
		}
		else
		{
			set.cycleTime = cfg->cycleTime;
			if (timer)
				delete timer;
			timer = new RCS_TIMER(set.cycleTime);
			status_next(&set, RCS_DONE);
			if (verbose)
				rcs_print("cycleTime set to %f \n", set.cycleTime);
		}
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCfgDebug sets debugging; when debugging is on, a log file is written
/*
Returned Value: none

Called By: doCfg

Currently, the debugging level (an int) is being used as a boolean. If
debug is set to zero and the log file is open, it is closed. If debug
is set to non-zero and the log file is open, it is closed and reopened
in write mode, so that the previous contents are lost. There is no
choice of what file to use. The name is always that specified by the
value of LOG_FILE.

logFp is a static global variable.

The function executes in a single cycle, reporting RCS_DONE (if the
file can be opened) or RCS_ERROR (if the file cannot be opened)
immediately.

*/

static void doCfgDebug() /* NO ARGUMENTS */
{
	AmMobJACfgDebug * cfg;

	cfg = (AmMobJACfgDebug *)(amMobJACfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		state_next(&set, S0);
		set.debug = cfg->debug;
		if (logFp != NULL)
		{
			fclose(logFp);
			logFp = NULL;
		}
		if (set.debug != 0)
		{
			logFp = fopen(LOG_FILE, "w");
			if (NULL == logFp)
			{
				status_next(&set, RCS_ERROR);
				rcs_print_error("%s: could not open log file for debug\n",
					processName);
			}
			else
			{
				status_next(&set, RCS_DONE);
				if (verbose)
					rcs_print("debug set to %d\n", set.debug);
			}
		}
		else
		{
			status_next(&set, RCS_DONE);
			if (verbose)
				rcs_print("debug set to %d\n", set.debug);
		}
	}
	else // S0
		state_default(&set);
}

/***********************************************************************/

/*! doCfgDumpWM sets drawing the world model (obstacle map, vehicle,
and planned path) on or off, according to the setting of turnOn
in the command being executed. */
/*
Returned Value: none

Called By: doCfg

*/

static void doCfgDumpWM() /* NO ARGUMENTS */
{
	AmMobJACfgDumpWM *cfg;

	cfg = (AmMobJACfgDumpWM *)(amMobJACfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		handleDisplay(cfg->turnOn);
		set.dumpWM = cfg->turnOn;
		status_next(&set, RCS_DONE);
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCfgNop executes a config Nop by doing nothing
/*
Returned Value: none

Called By: doCfg

*/

static void doCfgNop() /* NO ARGUMENTS */
{
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		status_next(&set, RCS_DONE);
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

//! doCfgPlanHorizon sets the configuration plan horizon
/* 
Returned Value: none

Called By: doCfg

The plan horizon cannot be changed. If the commanded horizon is
the same as the current horizon, this reports done immediately.  If
the commanded horizon differs from the current horizon, this
sends an error message.  To change the plan horizon, the AM Mob
process must be stopped and restarted.

*/

static void doCfgPlanHorizon() /* NO ARGUMENTS */
{
	AmMobJACfgPlanHorizon * cfg;

	cfg = (AmMobJACfgPlanHorizon *)(amMobJACfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		if (cfg->planHorizon != set.planHorizon)
		{
			rcs_print_error("%s: cannot change plan horizon while running\n",
				processName);
			status_next(&set, RCS_ERROR);
		}
		else
			status_next(&set, RCS_DONE);
		state_next(&set, S0);
	}
	else
	{
		state_default(&set);
	}
}

/***********************************************************************/

/*! doCfgMinTurnRad sets the vehicle minimum turning radius to use
during planning. */
/*

Returned Value: none

Called By: doCfg

The vehicleMinTurnRad must be a non-negative number.

If cfg->vehicleMinTurnRad (the value for the vehicleMinTurnRad in
the configuration command) is non-negative, this sets
set.vehicleMinTurnRad to that value.

The function executes in a single cycle, reporting RCS_DONE (if
cfg->vehicleMinTurnRad is non-negative) or RCS_ERROR (if
cfg->vehicleMinTurnRad is negative) immediately.

*/

static void doCfgVehicleMinTurnRad() /* NO ARGUMENTS */
{
	AmMobJACfgVehicleMinTurnRad * cfg;

	cfg = (AmMobJACfgVehicleMinTurnRad *)(amMobJACfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		if (cfg->vehicleMinTurnRad >= 0)
		{
			set.vehicleMinTurnRad = cfg->vehicleMinTurnRad;
			status_next(&set, RCS_DONE);
			if (verbose)
				rcs_print("vehicleMinTurnRad set to %f\n", set.vehicleMinTurnRad);
		}
		else
		{
			status_next(&set, RCS_ERROR);
			rcs_print_error("%s: vehicleMinTurnRad must not be negative\n",
				processName);
		}
		state_next(&set, S0);
	}
	else
		state_default(&set);
}

/***********************************************************************/

//! doCfgVehicleWidth sets the vehicle width to use during planning
/*

Returned Value: none

Called By: doCfg

The vehicleWidth must be a positive number.

If cfg->vehicleWidth (the value for the vehicleWidth given in the
configuration command) is positive, this sets set.vehicleWidth to that
value. This also resets obsGrow (by calling calcSkips).

The function executes in a single cycle, reporting RCS_DONE (if
cfg->vehicleWidth is positive) or RCS_ERROR (if cfg->vehicleWidth
is not positive) immediately.

*/

static void doCfgVehicleWidth() /* NO ARGUMENTS */
{
	AmMobJACfgVehicleWidth * cfg;

	cfg = (AmMobJACfgVehicleWidth *)(amMobJACfgBuf->get_address());
	if (state_match(&set, NEW_COMMAND))
	{
		state_new(&set);
		if (cfg->vehicleWidth > 0)
		{
			set.vehicleWidth = cfg->vehicleWidth;
			calcSkips();
			status_next(&set, RCS_DONE);
			if (verbose)
				rcs_print("vehicleWidth set to %f\n", set.vehicleWidth);
		}
		else
		{
			status_next(&set, RCS_ERROR);
			rcs_print_error("%s: vehicleWidth must be positive\n", processName);
		}
		state_next(&set, S0);
	}
	else
		state_default(&set);
}

/***********************************************************************/

//! doCmd executes task commands
/*
Returned Value: none

Called By: main

If admin_state is ADMIN_SHUT_DOWN, this returns immediately.
This sets stts.amMobJAError to AM_MOB_JA_ERROR_NONE before
calling the appropriate doCmdXXX function.

*/

static void doCmd()  /* NO ARGUMENTS */
{
	if (stts.admin_state == ADMIN_SHUT_DOWN)
	{ // do nothing if shut down
		return;
	}
	if (state_match(&stts, NEW_COMMAND))
	{
		stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
		stts.seed_length = 0;
	}
	switch (stts.command_type)
	{

	case AM_MOB_JA_CMD_ABORT_TYPE:
		doCmdAbort();
		break;

	case AM_MOB_JA_CMD_HALT_TYPE:
		doCmdHalt();
		break;

	case AM_MOB_JA_CMD_INIT_TYPE:
		doCmdInit();
		break;

	case AM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
		//printf("they told me to move!\n");
		doCmdMoveWaypoint();
		break;

	case AM_MOB_JA_CMD_NOP_TYPE:
		doCmdNop();
		break;

	case AM_MOB_JA_CMD_PREPARE_MOVE_TYPE:
		doCmdPrepareMove();
		break;

	case AM_MOB_JA_CMD_SHUTDOWN_TYPE:
		doCmdShutdown();
		break;

	case AM_MOB_JA_CMD_SPIN_TYPE:
		doCmdSpin();
		break;

	default:
		// no command to handle
		break;
	} // switch (cmd)
}

/***********************************************************************/

//! doCmdAbort aborts AM mobility
/*
Returned Value: none

Called By: doCmd

When the command is new, this send an abort command to Prim mobility.

If the command is not new and Prim is still executing the abort command
it was sent, this waits.

When the Prim abort command is finished executing successfully, this sets
stts.admin_state to UNINITIALIZED and reports the AM abort command as RCS_DONE.

This changes stts.admin_state to UNINITIALIZED, even if abort fails.

*/

static void doCmdAbort() /* NO ARGUMENTS */
{
	PrimMobJACmdAbort abortMsg;
	PrimMobJAStat * subStatPtr;

	subStatPtr = (PrimMobJAStat *) (primMobJAStatBuf->get_address());
	if (state_match(&stts, NEW_COMMAND))
	{ // command is new: send abort command to sub, go to state S0, status EXEC
		state_new(&stts);
		writePrimCmd(&abortMsg, subStatPtr->echo_serial_number);
		stts.admin_state = ADMIN_UNINITIALIZED;
		status_next(&stts, RCS_EXEC);
		state_next(&stts, S0);
	}
	else if (state_match(&stts, S0))
	{ // command is not new
		if (primCmdNumber != subStatPtr->echo_serial_number); // stay in state S0
		else if (subStatPtr->status == RCS_EXEC); // stay in state S0
		else if (subStatPtr->status == RCS_DONE)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_DONE);
			if (verbose)
				rcs_print("amMob aborted\n");
		}
		else // if (subStatPtr->status is RCS_ERROR or unknown)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_PRIM;
			if (verbose)
				rcs_print_error("%s: primMob reported %s\n", processName,
				((subStatPtr->status == RCS_ERROR) ?
				"abort error, amMob not aborted" :
			"unknown status, amMob not aborted"));
		}
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stts);
	}
}

/***********************************************************************/

//! doCmdHalt handles a halt command
/*
Returned Value: none

Called By: doCmd

When the command is new, this send a halt command to Prim mobility.

If the command is not new and prim is still executing the halt command
it was sent, this waits.

When the prim halt command is finished executing successfully, this sets
stts.admin_state to UNINITIALIZED and reports the AM halt command as RCS_DONE.

This changes stts.admin_state to UNINITIALIZED, even if halt fails.

*/

static void doCmdHalt() /* NO ARGUMENTS */
{
	PrimMobJACmdHalt haltMsg;
	PrimMobJAStat *subStatPtr;

	subStatPtr = (PrimMobJAStat *)(primMobJAStatBuf->get_address());
	if (state_match(&stts, NEW_COMMAND))
	{ // command is new: send halt command to sub, go to state S0, status EXEC
		state_new(&stts);
		writePrimCmd(&haltMsg, subStatPtr->echo_serial_number);
		stts.admin_state = ADMIN_UNINITIALIZED;
		status_next(&stts, RCS_EXEC);
		state_next(&stts, S0);
	}
	else if (state_match(&stts, S0))
	{ // command is not new
		if (primCmdNumber != subStatPtr->echo_serial_number); // stay in state S0
		else if (subStatPtr->status == RCS_EXEC); // stay in state S0
		else if (subStatPtr->status == RCS_DONE)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_DONE);
			if (verbose)
				rcs_print("amMob halted\n");
		}
		else // if (subStatPtr->status is RCS_ERROR or unknown)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_PRIM;
			if (verbose)
				rcs_print_error("%s: primMob reported %s\n", processName,
				((subStatPtr->status == RCS_ERROR) ?
				"halt error, amMob not halted" :
			"unknown status, amMob not halted"));
		}
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stts);
	}
}

/***********************************************************************/

//! doCmdInit handles an Init command
/*
Returned Value: none

Called By: doCmd

A. If the command is new:
A1.  Call initParameters.
A2.  Call iniSetVehicleParameters and iniLoad.
A2a. If either of these returns INI_INVALID:
Set admin_state to ADMIN_UNINITIALIZED.
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_UNSPECIFIED.
Set state to S1 (done executing).
A2b. Otherwise (everything is OK):
Send an init command to Prim Mob.
Set admin_state to ADMIN_UNINITIALIZED.
Set status to RCS_EXEC.
Set state to S0 (executing).

B. Otherwise, if execution is not finished (state S0):

B1. If the status report from prim is not for the command sent in
part A, or if prim is still executing the init command it was
sent, do not change anything.

B2. Otherwise, if prim is finished executing successfully the init
command it was sent:
Set admin_state to INITIALIZED.
Set status to RCS_DONE.
Set state to S1 (done with execution).

B3. Otherwise, prim has reported an error or unknown status:
Set admin_state to ADMIN_UNINITIALIZED.
Set state to S1 (done with execution).
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_PRIM.

C. Otherwise, the command is finished executing (state S1):
Do nothing.

The primMobJAStatBuf was read in main so does not need to be read here.

*/

static void doCmdInit() /* NO ARGUMENTS */
{
	PrimMobJACmdInit initMsg;
	PrimMobJAStat * subStatPtr;
	char * platformName;

	subStatPtr = (PrimMobJAStat *)(primMobJAStatBuf->get_address());
	if (state_match(&stts, NEW_COMMAND))
	{ // command is new
		if( stts.admin_state == ADMIN_INITIALIZED )
		{
			state_next(&stts, S1);
			return;
		}
		initParameters();
		if ((iniSetVehicleParameters(&(set.vehicleWidth),
			&(set.vehicleMinTurnRad),
			&platformName) == INI_INVALID) ||
			(iniLoad(platformName) == INI_INVALID))
		{
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_UNSPECIFIED;
			state_next(&stts, S1);
			rcs_print_error("%s: iniLoad failed, amMob not initialized\n",
				processName);
		}
		else
		{
			state_new(&stts);
			writePrimCmd(&initMsg, subStatPtr->echo_serial_number);
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_EXEC);        // go to status EXEC
			state_next(&stts, S0);               // go to state S0
		}
	}
	else if (state_match(&stts, S0))
	{ 
		// command is not new
		if (primCmdNumber != subStatPtr->echo_serial_number); // stay in state S0
		else if (subStatPtr->status == RCS_EXEC)
			printf("Hello3\n\n"); // stay in state S0
		else if (subStatPtr->status == RCS_DONE)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_INITIALIZED;
			status_next(&stts, RCS_DONE);
			if (verbose)
				rcs_print("amMob initialized\n");
		}
		else // if (subStatPtr->status is RCS_ERROR or unknown)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_PRIM;
			if (verbose)
				rcs_print_error("%s: primMob reported %s\n", processName,
				((subStatPtr->status == RCS_ERROR) ?
				"error initializing, amMob not initialized" :
			"unknown status, amMob not initialized"));
		}
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stts);
	}
}

/***********************************************************************/

//! doCmdMoveWaypoint handles a move waypoint command.
/*
Returned Value: none

Called By: doCmd

This command involves two types of path planning which may be done
simultaneously:
1. planning to many points on the planning horizon; this will be
called h-planning. This is done only if computeHorizonCosts is
true in the AmMobJACmdMoveWaypoint command being executed.
2. planning to move through a set of waypoints; this will be called
w-planning.
In addition, this command executes a w-plan. A move command sent to
Prim in order to execute a w-plan will be called a w-move.

If h-planning or w-planning fails, or if Prim fails while executing a
w-move, the command stops executing. If there is any error, status is
set to RCS_ERROR, and stts.amMobJAError is set as described in the details
below.

If a w-move fails for any reason or if w-planning fails after a
move command has been sent to Prim, a command is sent to Prim telling
it to stop. The execution of the stopping command is not monitored.

W-planning will fail whenever any unvisited waypoint inside the
planning horizon is on an obstacle. All unvisited waypoints inside the
planning horizon must be checked each time w-planning begins.

For some types of error, the settings of seed_length and seed are
meaningful; see amMobJA.hh. Except as provided there, the settings of
seed_length and seed have no meaning if an error has occurred.

If this function sets status to RCS_EXEC and sets stts.amMobJAError
to AM_MOB_JA_ERROR_NONE, that means it has finished h-planning, and
the data in seed and seed_length may be taken as the results of the
h-planning half of the command. If this function sets status to
RCS_EXEC and sets stts.amMobJAError to AM_MOB_JA_ERROR_POSITION,
however, that mean h-planning is not done and the data in seed and
seed_length should not be used.

The w-plans made by this function go only as far as the AM planning
horizon. If a waypoint is beyond the horizon, this plans to go towards
the distant waypoint as directly as possible up to the planning
horizon without going through an obstacle.

Each w-planned path is a smooth connected path consisting of straight
line segments and arcs of circles.

To keep track of the progress of executing a w-plan, the pathIndex
reported by Prim mobility is used to determine if a waypoint has been
reached and passed.

This makes a new Prim w-move command every cycle as long as the last
waypoint has not been reached, no error has occurred, and position
data is available.

If no errors occur, this function make one attempt to get position
data each time it is called. The function reports an error if position
data cannot be obtained 30 times in a row.

Normally, the AmMobJACmdMoveWaypoint command is superseded while AM is
still executing it.

The meaning of the states is:

S0 - Everything is OK so far but execution is not done. At least one
move command has been sent to Prim.
S1 - The command is not new, but position data has not been obtained
since the command was new, so no move command has been sent to Prim.
S2 - Command processing is finished (because of success or failure).

**********************************************************************

DETAILS

A. If AM Mob is not initialized, do the following and return:
Set status to RCS_ERROR.
Set state to S2.
Set amMobJAError to AM_MOB_JA_ERROR_NOT_INIT.

B. Otherwise, if the command is defective (which needs to be checked
on each cycle since obstacles may move), seed_length and possibly
seed will have been set by the badMove checking function. Do the
following and then return.
Send a stop command to Prim.
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_WAYPOINT.
Set state to S2.

C. Otherwise, if command is new or if in state S1 (meaning the
position has never been updated successfully since the command
was new):

C1. If the command is new, set updatePositionTries to 0 and go on to
the next step.

C2. Call updatePosition to get the position of the vehicle.

C2a. If updatePosition fails, increase updatePositionTries by 1,
set stts.amMobJAError to AM_MOB_JA_ERROR_POSITION, do one of
the following, and return:

C2a_i. If the command is new:
Send a stop command to Prim.
Set status to RCS_EXEC.
Set state to S1.

C2a_ii. Otherwise (in state S1) if updatePositionTries is greater than 30:
Set status to RCS_ERROR.
Set state to S2.

C2a_iii. Otherwise (in state S1 and updatePositionTries
not greater than 30) do nothing, status and state are correct as is.

C2b. If updatePosition succeeds, go on to the next step.

C3. Call planPrimMove to compose a PrimMobJACmdMoveArcSegment command
to start moving to waypoints.

C3a. If planPrimMove fails, do the following and return:
Send a stop command to Prim.
Set status to RCS_ERROR.
Set state to S2.
If amMobJAError is not already set to an error of some sort, set it to
AM_MOB_JA_ERROR_PLAN_PRIM.

C3b. If planPrimMove succeeds, go on to the next step

C4. Check the setting of computeHorizonCosts in the waypoint command.

C4a. If the waypoint command has computeHorizonCosts set to true,
call planAll.

C4a_i. If planAll fails, do the following and return:
Send a stop command to Prim.
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_PLAN_HORIZON.
Set state to S2.

C4b_ii. If planAll succeeds, set seed_length will have been set to a
positive number which is 40 if all points are planned to, and the
results of planning will be in the seed array. Do the following
and return.
Send the already composed Prim move command to Prim.
Set status to RCS_EXEC.
Set amMobJAError to AM_MOB_JA_ERROR_NONE.
Set state to S0.
Set updatePositionTries to 0.

C4b. If the waypoint command has computeHorizonCosts set to false,
go on to the next step.

C5. Do the following and return.
Send Prim the already composed move command.
Set status to RCS_EXEC.
Set amMobJAError to AM_MOB_JA_ERROR_NONE.
Set state to S0.
Set updatePositionTries to 0.

D. Otherwise, if state is S0 (implying status is RCS_EXEC):

D1. If the status message from Prim is for some other command, do
nothing and return. Otherwise, go on to the next step.

D2. If if updatePositionTries is greater than 30, do the following
and return.
Send a stop command to Prim.
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_POSITION.
Set state to S2.

D3. From the pathIndex in the status message, determine the index of
the first unvisited waypoint.

D4. If the index set in the preceding step is out of bounds, do
the following and return. Otherwise, go on to the next step.
Send a stop command to Prim.
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_PRIM.
Set state to S2.

D5. If a call to updatePosition fails, do the following and return.
Otherwise, go on to the next step.
Increase updatePositionTries by 1.
Set status to RCS_EXEC.
Set amMobJAError to AM_MOB_JA_ERROR_NONE.
Set state to S0.

D6. Otherwise, if Prim status is RCS_DONE:

D6a. If not all waypoints are planned for, call planPrimMove,
starting towards the first unvisited waypoint.

D6a_i. If planPrimMove fails, do the following and return:
Send a stop command to Prim.
Set status to RCS_ERROR
If amMobJAError is AM_MOB_JA_ERROR_NONE, set it to AM_MOB_JA_ERROR_MOVE. 
Set state to S2.

D6a_ii. If planPrimMove succeeds, do the following and return:
Send the move command to Prim.
Stay status RCS_EXEC.
Set amMobJAError to AM_MOB_JA_ERROR_NONE. 
Stay in state S0.

D6b. If all waypoints are planned for, do the following and return:
Set status to RCS_DONE.
Leave amMobJAError at AM_MOB_JA_ERROR_NONE. 
Set state to S2.

D7. Otherwise, if Prim status is RCS_EXEC:

D7a. Call planPrimMove, starting towards the current waypoint.

D7a_1. If planPrimMove fails, do the following and return:
Send a stop command to Prim.
Set status to RCS_ERROR
If amMobJAError is AM_MOB_JA_ERROR_NONE, set it to AM_MOB_JA_ERROR_MOVE. 
Set state to S2.

D7a_ii. If planPrimMove succeeds, do the following and return:
Send the move command to Prim.
Stay in status RCS_EXEC.
Set amMobJAError to AM_MOB_JA_ERROR_NONE. 
Stay in state S0.

D8. Otherwise, Prim status is RCS_ERROR or unknown (both of which are bad).
Do the following and return.
Send a stop command to Prim.
Set status to RCS_ERROR
Set amMobJAError to AM_MOB_JA_ERROR_PRIM.
Set state to S2.

E. Otherwise, state must be S2, indicating command is done executing
successfully or unsuccessfully. Do not do anything but return.

**********************************************************************

The arcToWaypointTable is a table giving the index in the waypoints of
each arc in the primMoveCmd (i.e., arcToWaypointTable[Prim arc index] =
waypoint index). This is needed so that when replanning starts, it
will be known (from Prim's report of what arc index it is working on)
what waypoint that index is for. Then replanning can start by heading
towards that waypoint.

*/

static void doCmdMoveWaypoint() /* NO ARGUMENTS */
{
	PrimMobJACmdMoveArcSegment primMoveCmd; // move command to send to Prim
	PrimMobJACmdStop primStopCmd;           // stop command to send to Prim
	AmMobJACmdMoveWaypoint * amMoveCmd; // command being executed
	PrimMobJAStat * subStatPtr;         // status message from Prim
	static int primEndsOnWaypoint = 0;  // 1=prim move cmd ends on waypoint
	int index;                          // index of next waypoint in amMoveCmd
	static int updatePositionTries = 0; // counter for calls to updatePosition

	amMoveCmd = (AmMobJACmdMoveWaypoint*)(amMobJACmdBuf->get_address());
	subStatPtr = (PrimMobJAStat *)(primMobJAStatBuf->get_address());

	if (stts.admin_state != ADMIN_INITIALIZED)
	{
		status_next(&stts, RCS_ERROR);
		stts.amMobJAError = AM_MOB_JA_ERROR_NOT_INIT;
		state_next(&stts, S2);
		if (verbose > 1)
			rcs_print_error("%s: amMob not initialized, "
			"cannot execute command\n", processName);
	}
	else if (badMove(amMoveCmd)) 
	{ // move command is defective, so stop the vehicle
		writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
		status_next(&stts, RCS_ERROR);
		stts.amMobJAError = AM_MOB_JA_ERROR_WAYPOINT;
		state_next(&stts, S2);
	}
	else if ((state_match(&stts, NEW_COMMAND)) || (state_match(&stts, S1)))
	{ // do not have position yet
		if (state_match(&stts, NEW_COMMAND))
		{ // command is new
			state_new(&stts);
			primEndsOnWaypoint = 0;
			updatePositionTries = 0;
		}
		if (updatePosition())
		{ // updatePosition failed
			updatePositionTries++;
			stts.amMobJAError = AM_MOB_JA_ERROR_POSITION;
			if (updatePositionTries == 1)
			{ // command is new
				writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
				status_next(&stts, RCS_EXEC);
				state_next(&stts, S1);
			}
			else if (updatePositionTries > 30)
			{ // in state S1 and too many attempts to get position
				status_next(&stts, RCS_ERROR);
				state_next(&stts, S2);
			}
			else; // status and state are already set correctly
		}
		else if (planPrimMove(amMoveCmd, &primMoveCmd, &primEndsOnWaypoint, 0))
		{ // attempt to make move command failed, so stop the vehicle
			writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
			status_next(&stts, RCS_ERROR);
			if (stts.amMobJAError == AM_MOB_JA_ERROR_NONE) // may already be set
				stts.amMobJAError = AM_MOB_JA_ERROR_PLAN_PRIM;
			state_next(&stts, S2);
		}
		else if (amMoveCmd->computeHorizonCosts == true)
		{ // try to plan to periphery
			if (planAll(amMoveCmd->neighborhood, amMoveCmd->speed))
			{ // planning to periphery failed, so stop the vehicle
				writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
				status_next(&stts, RCS_ERROR);
				stts.amMobJAError = AM_MOB_JA_ERROR_PLAN_HORIZON;
				state_next(&stts, S2);
			}
			else
			{ // planning to periphery done OK, so send prim move command
				writePrimCmd(&primMoveCmd, subStatPtr->echo_serial_number);
				status_next(&stts, RCS_EXEC);
				stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
				state_next(&stts, S0);
				updatePositionTries = 0;
			}
		}
		else
		{ // planning to periphery not attempted, so send prim move command
			writePrimCmd(&primMoveCmd, subStatPtr->echo_serial_number);
			status_next(&stts, RCS_EXEC);
			stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
			state_next(&stts, S0);
			updatePositionTries = 0;
		}
	}
	else if (state_match(&stts, S0))
	{ // everything OK so far, a move command has been sent to Prim
		if (primCmdNumber != subStatPtr->echo_serial_number)
			return; // do nothing; prim status is not for this command
		if (updatePositionTries > 30)
		{ // too many unsuccessful attempts to get position
			writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_POSITION;
			state_next(&stts, S2);
		}
		index = arcToWaypointTable[subStatPtr->pathIndex];
		if ((index < 0) || (index >= amMoveCmd->waypoint_length))
		{ // index out of range, so stop the vehicle
			writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_PRIM;
			state_next(&stts, S2);
			rcs_print_error("%s: prim pathIndex (%d) out of range\n ",
				processName, index);
		}
		else if (updatePosition())
		{ // updatePosition failed
			updatePositionTries++;
			status_next(&stts, RCS_EXEC);
			stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
			state_next(&stts, S0);
		}
		else if ((index == (amMoveCmd->waypoint_length - 1)) &&
			(hypot((vehX - amMoveCmd->waypoint[index].p.x),
			(vehY - amMoveCmd->waypoint[index].p.y))
			< amMoveCmd->waypoint[index].neighborhood))
		{ // all done (going towards last waypoint and close enough)
			writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
			status_next(&stts, RCS_DONE);
			stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
			state_next(&stts, S2);
		}
		else if (subStatPtr->status == RCS_DONE)
		{ // assumes prim returns index of arc just done
			updatePositionTries = 0;
			if ((index == (amMoveCmd->waypoint_length - 1)) &&
				primEndsOnWaypoint) // could get here if (prim nghd > am nghd)
			{ // all done (going towards last waypoint and prim cmd got there)
				status_next(&stts, RCS_DONE);
				stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
				state_next(&stts, S2);
			}
			else // either more waypoints or going towards last waypoint
			{
				if (primEndsOnWaypoint) // start towards next waypoint
					index++;
				else; // start towards current waypoint
				if (planPrimMove(amMoveCmd, &primMoveCmd,
					&primEndsOnWaypoint, index))
				{ // attempt to make command failed, so stop the vehicle
					writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
					status_next(&stts, RCS_ERROR);
					if (stts.amMobJAError == AM_MOB_JA_ERROR_NONE) // may be set
						stts.amMobJAError = AM_MOB_JA_ERROR_PLAN_PRIM;
					state_next(&stts, S2);
				}
				else
				{ // move command made OK; send it; status and state unchanged
					writePrimCmd(&primMoveCmd, subStatPtr->echo_serial_number);
					stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
				}
			}
		}
		else if (subStatPtr->status == RCS_EXEC)
		{ // start towards the waypoint we are already heading towards
			updatePositionTries = 0;
			if (planPrimMove(amMoveCmd, &primMoveCmd,
				&primEndsOnWaypoint, index))
			{ // attempt to make command failed, so stop the vehicle
				writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
				status_next(&stts, RCS_ERROR);
				if (stts.amMobJAError == AM_MOB_JA_ERROR_NONE) // may be set
					stts.amMobJAError = AM_MOB_JA_ERROR_PLAN_PRIM;
				state_next(&stts, S2);
			}
			else
			{ // move command made OK; send it; status and state unchanged
				writePrimCmd(&primMoveCmd, subStatPtr->echo_serial_number);
				stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
			}
		}
		else // if (subStatPtr->status == RCS_ERROR) or unknown
		{ // command is messed up, so stop the vehicle
			writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_PRIM;
			state_next(&stts, S2);
		}
	}
	else
	{ // should be in state S2; stay there
		state_default(&stts);
	}
}

/***********************************************************************/

//! doCmdNop executes a Nop command by doing nothing
/*
Returned Value: none

Called By: doCmd

A. If the command is new:
Set state to S0 (done executing).
Set status to RCS_DONE.

B. Otherwise, the command is done executing:
Do nothing.

Note that this does not check for any errors.

*/

static void doCmdNop() /* NO ARGUMENTS */
{
	if (state_match(&stts, NEW_COMMAND))
	{
		state_new(&stts);
		status_next(&stts, RCS_DONE);
		state_next(&stts, S0);
	}
	else
	{
		state_default(&stts);
	}
}

/***********************************************************************/

//! doCmdPrepareMove handles a prepare move command. 
/*
Returned Value: none

Called By: doCmd

A. If AM Mob is not initialized, set status to RCS ERROR and
set stts.amMobJAError to AM_MOB_JA_ERROR_NOT_INIT.

B. Otherwise, if the command is new:

B1. Send a stop command to Prim (because if the stop command is not
sent, prim will continue to execute any move command it received
previously).

B2. Set updatePositionTries to 0.

B3. Call updatePosition to get the vehicle position.

B3a. If updatePosition fails, set status to RCS_EXEC,
set state_next to S0, and increase updatePositionTries by 1.

B3b. If updatePosition succeeds, call planAll.

B3b1. If planAll executes without error, set status to RCS_DONE,
and set state_next to S1.

B3b2. If planAll fails, set status to RCS_ERROR, set stts.amMobJAError
to AM_MOB_JA_ERROR_PLAN_HORIZON, and set state_next to S1.

C. Otherwise, if the state is S0 (indicating the command is still
executing because updatePosition has not succeeded yet)

C1. If updatePositionTries exceeds 30, set status to RCS_ERROR,
set stts.amMobJAError to AM_MOB_JA_ERROR_POSITION,
and set state_next to S1.

C2. Otherwise, call updatePosition to get the vehicle position.

C2a. If updatePosition fails, leave status at RCS_EXEC,
leave state at S0, and increase updatePositionTries by 1.

C2b. If updatePosition succeeds, call planAll.

C2b1. If planAll executes without error, set status to RCS_DONE,
and set state_next to S1.

C2b2. If planAll fails, set status to RCS_ERROR,
and set state_next to S1.

D. Otherwise (state is S1), command execution is finished, so
do nothing.

This is not monitoring the status of the prim stop command after it is
sent.

This expects planAll to finish in less than the AM Mob cycle time.

*/

static void doCmdPrepareMove() /* NO ARGUMENTS */
{
	AmMobJACmdPrepareMove * amPrepareCmd;
	PrimMobJACmdStop primStopCmd;
	PrimMobJAStat * subStatPtr;
	static int updatePositionTries = 0;

	amPrepareCmd = (AmMobJACmdPrepareMove*)(amMobJACmdBuf->get_address());
	subStatPtr = (PrimMobJAStat *)(primMobJAStatBuf->get_address());
	if (stts.admin_state != ADMIN_INITIALIZED)
	{
		status_next(&stts, RCS_ERROR);
		stts.amMobJAError = AM_MOB_JA_ERROR_NOT_INIT;
		state_next(&stts, S1);
		if (verbose > 1)
			rcs_print_error("%s: amMob not initialized, "
			"cannot execute command\n", processName);
	}
	else if (state_match(&stts, NEW_COMMAND))
	{ // command is new
		state_new(&stts);
		updatePositionTries = 0;
		writePrimCmd(&primStopCmd, subStatPtr->echo_serial_number);
		if (updatePosition())
		{ // updatePosition failed
			status_next(&stts, RCS_EXEC);
			state_next(&stts, S0);
			updatePositionTries++;
			if (verbose > 1)
				rcs_print_error("%s: update position failed\n", processName);
		}
		else
		{ // updatePosition succeeded
			if (planAll(amPrepareCmd->neighborhood, amPrepareCmd->speed))
			{ // planAll failed
				status_next(&stts, RCS_ERROR);
				stts.amMobJAError = AM_MOB_JA_ERROR_PLAN_HORIZON;
			}
			else
			{ // planAll succeeded
				status_next(&stts, RCS_DONE);
			}
			state_next(&stts, S1);
		}
	}
	else if (state_match(&stts, S0))
	{ // command is not new, but updatePosition has not succeeded yet
		if (updatePositionTries > 30)
		{
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_POSITION;
			state_next(&stts, S1);
		}
		else if (updatePosition())
		{ // updatePosition failed
			//status_next(&stts, RCS_EXEC); already set
			//state_next(&stts, S0); already set
			updatePositionTries++;
			if (verbose > 1)
				rcs_print_error("%s: update position failed\n", processName);
		}
		else
		{ // updatePosition succeeded
			if (planAll(amPrepareCmd->neighborhood, amPrepareCmd->speed))
			{ // planAll failed
				status_next(&stts, RCS_ERROR);
				stts.amMobJAError = AM_MOB_JA_ERROR_PLAN_HORIZON;
			}
			else
			{ // planAll succeeded
				status_next(&stts, RCS_DONE);
			}
			state_next(&stts, S1);
		}
	}
	else
	{ // S1 command execution is finished. Do nothing.
		state_default(&stts);
	}
}

/***********************************************************************/

//! doCmdShutdown handles a shutdown command
/*
Returned Value: none

Called By: doCmd

A. If the admin_state is not UNINITIALIZED:
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_UNSPECIFIED.
Set state to S1 (done with execution).
Do not change admin_state.

B. Otherwise, if the command is new:
Send a shutdown command to Prim.
Set status to RCS_EXEC.
Set state to S0 (executing).

C. Otherwise, if execution is not finished (state S0):

C1. If the status report from prim is not for the command sent in
part B, or if prim is still executing the shutdown command it was
sent, do not change anything.

C2. Otherwise, if prim is finished executing successfully the shutdown
command it was sent:
Set admin_state to SHUT_DOWN.
Set status to RCS_DONE.
Set state to S1 (done with execution).

C3. Otherwise, prim has reported an error or unknown status:
Set admin_state to UNINITIALIZED.
Set state to S1 (done with execution).
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_PRIM.

D. Otherwise, the command is finished executing (state S1):
Do nothing.

*/

static void doCmdShutdown() /* NO ARGUMENTS */
{
	PrimMobJACmdShutdown shutdownMsg;
	PrimMobJAStat *subStatPtr;

	subStatPtr = (PrimMobJAStat *)(primMobJAStatBuf->get_address());
	if (stts.admin_state != ADMIN_UNINITIALIZED)
	{
		status_next(&stts, RCS_ERROR);
		stts.amMobJAError = AM_MOB_JA_ERROR_UNSPECIFIED;
		state_next(&stts, S1);
		rcs_print_error("%s: amMob not uninitialized, "
			"cannot execute shutdown command\n", processName);
	}
	else if (state_match(&stts, NEW_COMMAND))
	{// new command: send shutdown cmd to sub, go to state S0, stat EXEC
		state_new(&stts);
		writePrimCmd(&shutdownMsg, subStatPtr->echo_serial_number);
		status_next(&stts, RCS_EXEC);
		state_next(&stts, S0);
	}
	else if (state_match(&stts, S0))
	{ // command is not new
		if (primCmdNumber != subStatPtr->echo_serial_number); // stay in state S0
		else if (subStatPtr->status == RCS_EXEC); // stay in state S0
		else if (subStatPtr->status == RCS_DONE)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_SHUT_DOWN;
			status_next(&stts, RCS_DONE);
			if (verbose)
				rcs_print("amMob shut down\n");
		}
		else // if (subStatPtr->status is RCS_ERROR or unknown)
		{
			state_next(&stts, S1);
			stts.admin_state = ADMIN_UNINITIALIZED;
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_PRIM;
			if (verbose)
				rcs_print_error("%s: primMob reported %s\n", processName,
				((subStatPtr->status == RCS_ERROR) ?
				"shutdown error, amMob not shut down" :
			"unknown status, amMob not shut down"));
		}
	}
	else
	{ // command is finished executing: stay in state S1
		state_default(&stts);
	}
}

/***********************************************************************/

//! doCmdSpin handles a spin command
/*
Returned Value: none

Called By: doCmd

This command causes the vehicle to spin to a certain absolute angle at 
a given rotational velocity. The direction is determined by the direction
flag.

A. If AM Mob is not initialized:
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_NOT_INIT.
Set state to S2 (finished executing).

B. Otherwise, if the command is new:
Send prim a rotate command.
Set status to RCS_EXEC.
Set state to S1.

C. Otherwise if execution is not done (in state S0):

C1. If the status report from prim is not for the command sent in
part B, or if prim is still executing the rotate command it was
sent, do not change anything.

C2. Otherwise, if prim is finished executing successfully the rotate
command it was sent:
Set status to RCS_DONE.
Set state to S2 (done with execution).

C3. Otherwise, prim has reported an error or unknown status:
Set state to S2 (done with execution).
Set status to RCS_ERROR.
Set amMobJAError to AM_MOB_JA_ERROR_PRIM.

D. Otherwise, the command is finished executing (state S2):
Do nothing.

*/

static void doCmdSpin() /* NO ARGUMENTS */
{
	PrimMobJACmdRotate primRotateCmd;
	AmMobJACmdSpin * amSpinCmd;
	PrimMobJAStat * subStatPtr;

	amSpinCmd = (AmMobJACmdSpin*)(amMobJACmdBuf->get_address());
	subStatPtr = (PrimMobJAStat *)(primMobJAStatBuf->get_address());

	if (stts.admin_state != ADMIN_INITIALIZED)
	{
		status_next(&stts, RCS_ERROR);
		stts.amMobJAError = AM_MOB_JA_ERROR_NOT_INIT;
		state_next(&stts, S2);
		if (verbose > 1)
			rcs_print_error("%s: amMob not initialized, "
			"cannot execute command\n", processName);
	}
	else if (state_match(&stts, NEW_COMMAND))
	{
		state_new(&stts);
		// try to make and send a command to prim
		primRotateCmd.direction = amSpinCmd->direction;
		primRotateCmd.theta = amSpinCmd->absAngle;
		primRotateCmd.thetaTol = amSpinCmd->tolerance;
		// send command to prim
		writePrimCmd(&primRotateCmd, subStatPtr->echo_serial_number);
		status_next(&stts, RCS_EXEC);
		state_next(&stts, S0);
	}
	else if (state_match(&stts, S0))
	{ // everything OK so far
		if (primCmdNumber != subStatPtr->echo_serial_number); // stay in state S0
		else if (subStatPtr->status == RCS_EXEC); // stay in state S0
		else if (subStatPtr->status == RCS_DONE)
		{ // all done
			status_next(&stts, RCS_DONE);
			state_next(&stts, S2);
		}
		else // if (subStatPtr->status is RCS_ERROR or unknown)
		{
			state_next(&stts, S2);
			status_next(&stts, RCS_ERROR);
			stts.amMobJAError = AM_MOB_JA_ERROR_PRIM;
			if (verbose > 1)
				rcs_print_error("%s: primMob reported %s\n", processName,
				((subStatPtr->status == RCS_ERROR) ?
				"error executing rotate command" :
			"unknown status"));
		}
	}
	else
	{ // should be in state S2; stay there
		state_default(&stts);
	}
}

static void doXML()
{
	// If we have a waypoint that has yet to be processed
	if(unprocessedGPSWaypoint)
	{
		// Convert to local coordinates
		//PM_CARTESIAN globalPoint = GPStoLocalWaypoint();

		/*
		printf("\n\n=======================================\n");
		printf("Initial GPS Information:\n");
		printf("=======================================\n");
		printf("Latitude Degrees:%i\nLatitude Minutes:%f\n",initGPS.LatitudeDegrees,
		initGPS.LatitudeMinutes);
		printf("Longitude Degrees:%i\nLongitude Minutes:%f\n",initGPS.LongitudeDegrees,
		initGPS.LongitudeMinutes);
		printf("\n\n=======================================\n");
		printf("Initial CARTESIAN Information:\n");
		printf("=======================================\n");
		printf("Initial Coordinates:%lf,%lf\n",initX,
		initY);
		printf("\n\n=======================================\n");
		printf("Waypoint Information:\n");
		printf("=======================================\n");
		printf("Time Stamp:%f\n", GPSWaypoint.TimeStamp);
		printf("Latitude Degrees:%i\nLatitude Minutes:%f\n",GPSWaypoint.LatitudeDegrees,
		GPSWaypoint.LatitudeMinutes);
		printf("Longitude Degrees:%i\nLongitude Minutes:%f\n",GPSWaypoint.LongitudeDegrees,
		GPSWaypoint.LongitudeMinutes); 
		*/
		PM_CARTESIAN globalPoint;
		globalPoint.x = GPSWaypoint.xPos;
		globalPoint.y = GPSWaypoint.yPos;
		printf("\n\nGlobal Desired Coordinate: %f,%f\n\n", globalPoint.x, globalPoint.y);

		amMobJACmdMoveWaypoint.waypoint[0].p.x = globalPoint.x;
		amMobJACmdMoveWaypoint.waypoint[0].p.y = globalPoint.y;
		amMobJACmdMoveWaypoint.waypoint[0].p.z = 0;
		amMobJACmdMoveWaypoint.waypoint[0].neighborhood = 0.2;
		amMobJACmdMoveWaypoint.waypoint[0].speed = 1;
		amMobJACmdMoveWaypoint.waypoint_length = 1;
		amMobJACmdMoveWaypoint.computeHorizonCosts =  false;
		writeMobCmd(&amMobJACmdMoveWaypoint);

		unprocessedGPSWaypoint = false; // The GPS waypoint is now processed
	}
}
/*
static PM_CARTESIAN GPStoLocalWaypoint()
{
PM_CARTESIAN result;

float LatitudeMinutesDiff = GPSWaypoint.LatitudeMinutes - initGPS.LatitudeMinutes;
LatitudeMinutesDiff += ((float)GPSWaypoint.LatitudeDegrees - (float)initGPS.LatitudeDegrees)*60;
float LongitudeMinutesDiff = GPSWaypoint.LongitudeMinutes - initGPS.LongitudeMinutes;
LongitudeMinutesDiff += ((float)GPSWaypoint.LongitudeDegrees - (float)initGPS.LongitudeDegrees)*60;

result.x = (LatitudeMinutesDiff/6)*11120 + initX;
result.y = (LongitudeMinutesDiff/6)*7167 + initY;
return result;
}
*/
/***********************************************************************/

//! drawObsData draws the obstacle data received from AM SP.
/*
Returned Value: none

Called By: not currently called

This prints a file named "obsData" showing the featArray of the
obsData in ppm format. The file may be displayed with a ppm drawer
such as xv or ppm2.tcl or gimp or imageview. Obstacles are shown in
red, maybe obstacles in orange, clear areas in green, and unseen areas
in yellow. The vehicle is black.

*/
/* COMMENTED OUT
static void drawObsData() // NO ARGUMENTS
{
int n;              // buffer index
int i;              // row index (reused for buf index of vehicle)
int j;              // column index
FILE * outFile;
double obsVal;
static char buf[3 * MAX_MAP_CELLS];
size_t noWarn; // just to stop compiler warnings

// Write the output file header
outFile = fopen("obsData", "w");
if (outFile == 0)
{
rcs_print_error("%s: Could not open obsData file for writing\n",
processName);
}
fprintf(outFile, "P6\n");
// Write rows, cols, and max color intensity
fprintf(outFile, "%d %d\n255\n", obsData->x, obsData->y);

// Now buffer the data itself
n = 0;
for (i = (obsData->x - 1); i >= 0; i--)
{
for (j = 0; j < obsData->y; j++)
{
obsVal = obsData->featArray[(i * obsData->y)+j];
if (obsVal ==  AMSP_UNSEEN)
{ // if a cell has not been seen, color it yellow
buf[n]   = 255;  // Red
buf[n+1] = 255;  // Green
buf[n+2] = 0;    // Blue
}
else if (obsVal == AMSP_NOT_OBSTACLE)
{ // if a cell is seen and not an obstacle, color it green
buf[n]   = 0;  // Red
buf[n+1] = 255;  // Green
buf[n+2] = 0;    // Blue
}
else if (obsVal == AMSP_MAYBE_OBSTACLE)
{ // if a cell is seen and may be an obstacle, color it orange
buf[n]   = 255;  // Red
buf[n+1] = 127;  // Green
buf[n+2] = 0;    // Blue
}
else // if (obsVal == AMSP_OBSTACLE)
{ // if a cell is seen and is an obstacle, color it red
buf[n]   = 255;  // Red
buf[n+1] = 0;          // Green
buf[n+2] = 0;          // Blue
}
n = (n + 3);
}
}
// put the vehicle on the map in black
i = (3 * ((((obsData->x - 1) - findMapXIndex(vehX)) * obsData->y) +
(findMapYIndex(vehY))));
buf[i]   = 0;  // Red
buf[i+1] = 0;  // Green
buf[i+2] = 0;  // Blue
// Write the data to the output file
noWarn = fwrite(buf, sizeof(char), n, outFile);
fclose(outFile);
}
*/

/***********************************************************************/

//! drawObsMapBuf draws the obstacle map into a buffer.
/*
Returned Value: none

Called By: drawObsMapFile

See documentation of drawObsMapFile.

*/

static void drawObsMapBuf( /*   ARGUMENTS             */
						  char * buf)               /*!< buffer to draw map in */
{
	int n;              // buffer index
	int i;              // row index (reused for buf index of vehicle)
	int ii;             // row pixel index
	int j;              // column index
	int jj;             // column pixel index
	double obsVal;      // obstacle value of cell

	n = 0;
	for (i = obsMapSi; i >= 0; i--)
	{ // loop on rows
		for(ii = 0; ii < cellPixels; ii++)
		{ // loop on output pixel size
			for (j = 0; j < obsMapSide; j++)
			{ // loop on columns
				for (jj = 0; jj < cellPixels; jj++)
				{ // loop on output pixel size
					obsVal = obsMap[i][j];
					if (obsVal ==  AMSP_UNSEEN)
					{ // if a cell has not been seen, color it yellow
						buf[n]   = 255;  // Red
						buf[n+1] = 255;  // Green
						buf[n+2] = 0;    // Blue
					}
					else if (obsVal == AMSP_NOT_OBSTACLE)
					{ // if a cell is seen and not an obstacle, color it green
						buf[n]   = 0;    // Red
						buf[n+1] = 255;  // Green
						buf[n+2] = 0;    // Blue
					}
					else if (obsVal == AMSP_NEAR_OBSTACLE)
					{ // if cell seen and near obstacle, color it yellow-orange
						buf[n]   = 255;  // Red
						buf[n+1] = 150;  // Green
						buf[n+2] = 0;    // Blue
					}
					else if (obsVal == AMSP_MAYBE_OBSTACLE)
					{ // if cell seen and may be obstacle, color it red-orange
						buf[n]   = 255;  // Red
						buf[n+1] = 100;  // Green
						buf[n+2] = 0;    // Blue
					}
					else // if (obsVal == AMSP_OBSTACLE)
					{ // if a cell is seen and is an obstacle, color it red
						buf[n]   = 255;  // Red
						buf[n+1] = 0;    // Green
						buf[n+2] = 0;    // Blue
					}
					n = (n + 3);
				}
			}
		}
	}
}

/***********************************************************************/

/*! drawObsMapFile writes the obstacle map, the vehicle, and the planned
path (if there is one) in the file "obsMap". */
/*
Returned Value: none

Called By:  planPrimMove

This prints a file whose name is given in mapImageFileName showing the
obsMap in ppm format. The file may be displayed with a ppm drawer such
as imagview, xv or ppm2.tcl.  Obstacles are shown in red, maybe
obstacles in red-orange, near obstacles in yellow-orange, clear areas
in green, and unseen areas in yellow. The vehicle is black.  If the
primMoveCmd argument is not NULL, this also shows in white the path
given in the primMoveCmd.

Note that since +Z points into the ground, positive angles look
clockwise on the map.

The obstacles on the obsMap have been grown from the obsData.

The "buf" buffer is large enough for cellPixels up to 4.

The header has three lines of the following form, where the "P6" is
a format descriptor, and the "255" is the number of intensity levels.

P6
<number columns> <number rows>
255

The printData variable is intended to be used when amMobMain is run
inside gdb. printData may be set to 1 during a gdb break.  When
printData is non-zero, the data from the map will be printed to the
file amMobDataOut.  The function sets printData back to 0 so that the
map is only written once.

*/

static void drawObsMapFile(                /*   ARGUMENTS           */
						   PrimMobJACmdMoveArcSegment * primMoveCmd) /*!< primMoveCmd to draw */
{
	int n;              // number of pixels on each side of picture
	FILE * outFile;
	static char buf[4 * 4 * 3 * MAX_MAP_CELLS];
	size_t noWarn; // just to stop compiler warnings
	static int printData = 0;

	if (cellPixels > 4)
		cellPixels = 4;
	//put the obstacle map in the picture
	drawObsMapBuf(buf);
	//put the path in the primMoveCmd in the picture
	if (primMoveCmd)
		drawPath(buf, primMoveCmd);
	// put the vehicle in the picture
	drawVehicle(vehX, vehY, buf);
	outFile = fopen(mapImageFileName, "w");
	if (outFile == 0)
	{
		rcs_print_error("%s: Could not open file %s for writing\n",
			processName, mapImageFileName);
	}
	n = (obsMapSide * cellPixels);
	// Write the output file header
	fprintf(outFile, "P6\n%d %d\n255\n", n, n);
	// Write the data to the output file
	noWarn = fwrite(buf, sizeof(char), (3 * n * n), outFile);
	fclose(outFile);
	if (printData)
	{
		writeObsMapData("amMobDataOut");
		printData = 0;
	}
}

/***********************************************************************/

//! drawPath draws a planned path in the buffer as a white line
/*
Returned Value: none

Called By: drawObsMapFile

The path is inserted in the buffer in white.

This draws each segment of the path separately. If a segment is
a line, this calls drawPathLine once to draw it. If a segment is
an arc of a circle, this approximates the arc with several straight
line segments and calls drawPathLine once for each of them.

*/

static void drawPath(                      /*   ARGUMENTS         */
					 char * buf,                               /*!< buffer to draw in */
					 PrimMobJACmdMoveArcSegment * primMoveCmd) /*!< path to draw      */
{
	ArcSegment * anArc;
	int i0;
	int j0;
	int i1;
	int j1;
	int n;         // index of segments
	int k;         // index of parts of an an arc
	double beta;   // angle of part of an arc
	double alpha;  // angle of first point on arc
	int stop;
	double R;      // radius of an arc

	i0 = findMapXIndex(vehX);
	j0 = findMapYIndex(vehY);
	for (n = 0; n < primMoveCmd->arcList_length; n++)
	{
		anArc = (primMoveCmd->arcList + n);
		if (anArc->isArc)
		{
			R = hypot((anArc->end.x - anArc->center.x),
				(anArc->end.y - anArc->center.y));
			stop = (1 + (int)(ABS(anArc->theta) / (M_PI / 24)));
			alpha = (atan2((anArc->end.y - anArc->center.y),
				(anArc->end.x - anArc->center.x)) - anArc->theta);
			beta = (anArc->theta / stop);
			for (k = 0; k < stop; k++)
			{
				alpha = alpha + beta;
				i1 = findMapXIndex(anArc->center.x + (R * cos(alpha)));
				j1 = findMapYIndex(anArc->center.y + (R * sin(alpha)));
				drawPathLine(buf, i0, j0, i1, j1);
				i0 = i1;
				j0 = j1;
			}
		}
		else
		{
			i1 = findMapXIndex(anArc->end.x);
			j1 = findMapYIndex(anArc->end.y);
			drawPathLine(buf, i0, j0, i1, j1);
			i0 = i1;
			j0 = j1;
		}
	}
}

/***********************************************************************/

//! drawPathLine draws a white line in the buffer
/*
Returned Value: none

Called By: drawPath

The line is inserted in the buffer in white.

The method of finding the smaller change is not particularly
efficient and is slightly skewed. Could switch to the method used
elsewhere, but it takes more code.

*/

static void drawPathLine( /*   ARGUMENTS                     */
						 char * buf,              /*!< buffer to draw in             */
						 int i0,                  /*!< obsMap X index of start point */
						 int j0,                  /*!< obsMap Y index of start point */
						 int i1,                  /*!< obsMap X index of end point   */
						 int j1)                  /*!< obsMap Y index of end point   */
{
	int i;         // index for rows
	int ii;        // index for pixel rows
	int j;
	int jj;        // index for pixel columns
	int k;         // index of points
	int m;         // index in buffer
	int deltaI;    // change in X index on a line
	int deltaJ;    // change in Y index on a line
	int inc;       // increment of +1 or -1
	int stop;      // 

	deltaI = (i1 - i0);
	deltaJ = (j1 - j0);
	i = i0;
	j = j0;
	if (ABS(deltaI) >= ABS(deltaJ))
	{
		inc = ((deltaI < 0) ? -1 : 1);
		stop = ABS(deltaI);
		for (k = 0; k <= stop; k++)
		{
			m = (3 * 
				(((obsMapSi - i) * obsMapSide * cellPixels * cellPixels)
				+ (j * cellPixels)));
			for (ii = 0; ii < cellPixels; ii++)
			{
				for (jj = 0; jj < (3 * cellPixels); jj++)
				{
					buf[m + jj] = 255;
				}
				m = (m + (3 * obsMapSide * cellPixels));
			}
			i = (i + inc);
			j = (j0 + (int)(((i - i0) / (double)deltaI) * deltaJ));
		}
	}
	else // if (ABS(deltaI) < ABS(deltaJ))
	{
		inc = ((deltaJ < 0) ? -1 : 1);
		stop = ABS(deltaJ);
		for (k = 0; k <= stop; k++)
		{
			m = (3 * 
				(((obsMapSi - i) * obsMapSide * cellPixels * cellPixels)
				+ (j * cellPixels)));
			for (ii = 0; ii < cellPixels; ii++)
			{
				for (jj = 0; jj < (3 * cellPixels); jj++)
				{
					buf[m + jj] = 255;
				}
				m = (m + (3 * obsMapSide * cellPixels));
			}
			j = (j + inc);
			i = (i0 + (int)(((j - j0) / (double)deltaJ) * deltaI));
		}
	}
}

/***********************************************************************/

//! drawVehicle draws the vehicle in the buffer as a black square
/*
Returned Value: none

Called By: drawObsMapFile

The vehicle is drawn as a square cellPixels on a side.

*/

static void drawVehicle( /*   ARGUMENTS                        */
						double vehX,            /*!< X coordinate of vehicle location */
						double vehY,            /*!< Y coordinate of vehicle location */
						char * buf)             /*!< buffer to draw in                */
{
	int m;
	int i;
	int j;

	m = (3 * (((obsMapSi - findMapXIndex(vehX)) * obsMapSide *
		cellPixels * cellPixels) +
		(findMapYIndex(vehY) * cellPixels)));
	for (i = 0; i < cellPixels; i++)
	{
		for (j = 0; j < (3 * cellPixels); j++) // cellPixels, 3 chars per pixel
		{
			buf[m+j] = 0;
		}
		m = (m + (3 * obsMapSide * cellPixels)); // go to next row
	}
}

/***********************************************************************/

/*! findBestGoal finds the least cost goal in a meter of the planning
horizon */
/*
Returned Value: none

Called By: planPathsDijk

The goals and their costs have already been found. This examines the
goals in a meter of periphery whose indexes in the goals array start
at the gcsIndex argument. From these, it selects the one with least
cost and puts its location and cost in the data array at the given
dataIndex.

If none of the cells in the meter has been reached, this sets
data[dataIndex] to the point in the middle of the meter, with a cost of
-1.0.

In the last meter to be checked, the last goal to check is goals[0].
Thus, this checks when the gcsIndex has gone beyond its maximum value
and resets it to 0 if so. The loop stopper is reset to 1 when that
happens.

*/

static void findBestGoal( /*   ARGUMENTS                                   */
						 PM_CARTESIAN * data,     /*!< points to fill with data                    */
						 int dataIndex,           /*!< index of data to fill                       */
						 int gcsIndex,            /*!< index for goals, costs, and statuses arrays */
						 int * statuses,          /*!< array of search statuses                    */
						 double * costs,          /*!< array of costs                              */
						 int gcsIndexStop)        /*!< one more than maximum gcsIndex              */
{
	int bestGcsIndex; // value of gcsIndex for least cost
	double leastCost; // least cost of goal in meter
	int stop;         // gcsIndex to stop before
	int id;           // cell id at bestGcsIndex;

	bestGcsIndex = (gcsIndex + (cellsPerMeter / 2)); // default: middle of meter
	leastCost = -1.0;
	for (stop = (gcsIndex + cellsPerMeter); gcsIndex < stop; gcsIndex++)
	{
		if (gcsIndex == gcsIndexStop)
		{
			gcsIndex = 0;
			stop = 1;
		}
		if (statuses[gcsIndex])
		{
			bestGcsIndex = gcsIndex;
			leastCost = costs[gcsIndex];
			break;
		}
	}
	for (gcsIndex++; gcsIndex < stop; gcsIndex++)
	{
		if (gcsIndex == gcsIndexStop)
		{
			gcsIndex = 0;
			stop = 1;
		}
		if (statuses[gcsIndex] && (costs[gcsIndex] < leastCost))
		{
			bestGcsIndex = gcsIndex;
			leastCost = costs[gcsIndex];
		}
	}
	id = goals[bestGcsIndex];
	data[dataIndex].x = findMapXCoord(id / obsMapSide);
	data[dataIndex].y = findMapYCoord(id % obsMapSide);
	data[dataIndex].z = leastCost;
}

/***********************************************************************/

/*! findEdgePoint finds the point where the line between (x0,y0) and
(x1,y1) crosses the edge of the obstacle map. */
/*
Returned Value: int
If (x1, y1) is not outside the obstacle map, this returns 1.
Otherwise, it returns 0;

Called By: planPrimMoveSome

Side Effects: If (x1, y1) is outside the obstacle map, (*x, *y) is set
to the point where the line between (x0,y0) and (x1,y1) crosses the
edge of the obstacle map. If (x1, y1) is not outside the obstacle map,
*x and *y are not changed.

It is OK if (x1, y1) and (*x, *y) are originally the same point.

*/

static int findEdgePoint( /* ARGUMENTS                                  */
						 double x0,      /*!< X coordinate of start point of line crossing edge */
						 double y0,      /*!< Y coordinate of start point of line crossing edge */
						 double x1,      /*!< X coordinate of end point of line crossing edge   */
						 double y1,      /*!< Y coordinate of end point of line crossing edge   */
						 double * x,     /*!< X coordinate of edge point, set here              */
						 double * y)     /*!< Y coordinate of edge point, set here              */
{
	if (x1 > xMax)
	{ // maybe on north edge
		*y = (y0 + ((y1 - y0) * ((xMax - x0) / (x1 - x0))));
		if (*y > yMax)
		{ // nope, on east edge
			*y = (yMax - 0.0001);
			*x = (x0 + ((x1 - x0) * ((yMax - y0) / (y1 - y0))));
		}
		else if (*y < yMin)
		{ // nope, on west edge
			*y = (yMin + 0.0001);
			*x = (x0 + ((x1 - x0) * ((y0 - yMin) / (y0 - y1))));
		}
		else
		{ // yes, on north edge
			*x = (xMax - 0.0001);
		}
	}
	else if (x1 < xMin)
	{ // maybe on south edge
		*y = (y0 + ((y1 - y0) * ((x0 - xMin) / (x0 - x1))));
		if (*y > yMax)
		{ // nope, on east edge
			*y = (yMax - 0.0001);
			*x = (x0 + ((x1 - x0) * ((yMax - y0) / (y1 - y0))));
		}
		else if (*y < yMin)
		{ // nope, on west edge
			*y = (yMin + 0.0001);
			*x = (x0 + ((x1 - x0) * ((y0 - yMin) / (y0 - y1))));
		}
		else
		{ // yes, on south edge
			*x = (xMin + 0.0001);
		}
	}
	else if (y1 > yMax)
	{ // on east edge
		*x = (x0 + ((x1 - x0) * ((yMax - y0) / (y1 - y0))));
		*y = (yMax - 0.0001);
	}
	else if (y1 < yMin)
	{ // on west edge
		*x = (x0 + ((x1 - x0) * ((y0 - yMin) / (y0 - y1))));
		*y = (yMin + 0.0001);
	}
	else
		return 1;
	return 0;
}

/***********************************************************************/

//! findIdAtCoords finds the id of the cell containing a given point, (x,y)
/*
Returned Value: int (the id)

Called By:
planPathAStar
planPrimMoveOne

This is not checking that x and y are in bounds, so it should not be
called unless they are known to be in bounds.

*/

int findIdAtCoords( /*   ARGUMENTS                                  */
				   double x,          /*!< X coordinate of point for which to find id */
				   double y)          /*!< Y coordinate of point for which to find id */
{
	return ((findMapXIndex(x) * obsMapSide) + findMapYIndex(y));
}

/***********************************************************************/

/*! findMapXCoord finds the X coordinate of the center of a cell
whose X index on the obstacle map is given. */
/*
Returned Value: double (the X coordinate)

Called By:
findBestGoal
findPathXCoord
initKidsAStar
planPrimMoveOneEdge

The lowerLeft is the point at the lower left corner of the obstacle map.
It is also the point at the lower left corner of the first cell.

*/

static double findMapXCoord( /*   ARGUMENTS                     */
							int i)                      /*!< obstacle map X index of point */
{
	return (obsData->lowerLeft.x + ((i + 0.5) * res));
}

/***********************************************************************/

/*! findMapXIndex finds the X index on the obstacle map of a point
whose X-coordinate is given. */
/*
Returned Value: int (the X index)

Called By:
checkObsArc
checkObsLineThick
drawObsData
drawPath
drawVehicle
findIdAtCoords
planPathAStar
planPathDijk
planPrimMove
planPrimMoveOneEdge
planPrimMoveSome

This will return a value whether the point is on the obstacle map or
not, but if the point is not on the map, the value will not be valid.

*/

static int findMapXIndex( /*   ARGUMENTS                          */
						 double x)                /*!< obstacle map X coordinate of point */
{
	return (int)((x - obsData->lowerLeft.x) / res);
}

/***********************************************************************/

/*! findMapYCoord finds the Y coordinate of the center of a cell
whose Y index on the obstacle map is given. */
/*
Returned Value: double (the Y coordinate)

Called By:
findBestGoal
findPathYCoord
initKidsAStar
planPrimMoveOneEdge

The lowerLeft is the point at the lower left corner of the obstacle map.
It is also the point at the lower left corner of the first cell.

*/

static double findMapYCoord( /*   ARGUMENTS                     */
							int j)                      /*!< obstacle map Y index of point */
{
	return (obsData->lowerLeft.y + ((j + 0.5) * res));
}

/***********************************************************************/

/*! findMapYIndex finds the Y index on the obstacle map of a point
whose Y-coordinate is given. */
/*
Returned Value: int (the Y index)

Called By:
checkObsArc
checkObsLineThick
drawObsData
drawPath
drawVehicle
findIdAtCoords
planPathAStar
planPathsDijk
planPrimMove
planPrimMoveOneEdge
planPrimMoveSome

This will return a value whether the point is on the obstacle map or
not, but if the point is not on the map, the value will not be valid.

*/

static int findMapYIndex( /*   ARGUMENTS                          */
						 double y)                /*!< obstacle map Y coordinate of point */
{
	return (int)((y - obsData->lowerLeft.y) / res);
}

/***********************************************************************/

/*! findMaxEnd finds the largest value between m and n of a path index
*j such that the thick straight line from path[m] to path[*j] does
not hit an obstacle. */
/*
Returned Value: int (0)

Called By: linearizeAStarPath

n is at least 2 more than m when this starts.

It is possible that checkObsLineThick will fail for every end index
tried, even the index of the second cell. In this case, since the A*
algorithm selected the second cell, j is set to the index of the
second cell.

*/

static int findMaxEnd( /*   ARGUMENTS           */
					  int m,                /*!< first path index    */
					  int n,                /*!< last path index     */
					  int * path,           /*!< the path to check   */
					  int * j)              /*!< path index set here */
{
	int i;
	int dI;
	int obs;

	for (dI = 1; (dI * 2) < (n - m); dI = (dI * 2));
	obs = 0;
	for (i = m; dI; dI = (dI / 2))
	{
		i = ((obs == 0) ? (i + dI) : (i - dI));
		if (i > n)
			i = (i - dI);
		else
		{
			obs = checkObsLineThick(findPathXCoord(path, m),
				findPathYCoord(path, m),
				findPathXCoord(path, i),
				findPathYCoord(path, i));
		}
	}
	*j = ((obs == 0) ? ((i > n) ? n : i) : ((i - 1) == m) ? i : (i - 1));
	return 0;
}

/***********************************************************************/

/*! findMinStart finds the smallest value between m and n of a path index
*k such that the thick straight line from path[*k] to path[n] does
not hit an obstacle. */
/*
Returned Value: int (0)

Called By: linearizeAStarPath

n is at least 2 more than m when this starts.

It is possible that checkObsLineThick will fail for every end index
tried, even the index of the next-to-last cell. In this case, since
the A* algorithm selected the next-to-last cell, k is set to the index
of the next-to-last cell.

*/

static int findMinStart( /*   ARGUMENTS           */
						int m,                  /*!< first path index    */
						int n,                  /*!< last path index     */
						int * path,             /*!< the path to check   */
						int * k)                /*!< path index set here */
{
	int i;
	int dI;
	int obs;

	for (dI = 1; (dI * 2) < (n - m); dI = (dI * 2));
	obs = 0;
	for (i = n; dI; dI = (dI / 2))
	{
		i = ((obs == 0) ? (i - dI) : (i + dI));
		if (i < m)
			i = (i + dI);
		else
		{
			obs = checkObsLineThick(findPathXCoord(path, n),
				findPathYCoord(path, n),
				findPathXCoord(path, i),
				findPathYCoord(path, i));
		}
	}
	*k = ((obs == 0) ? ((i < m) ? m : i) : ((i + 1) == n) ? i : (i + 1));
	return 0;
}

/***********************************************************************/

/*! findPathXCoord finds the obsMap X coordinate of the center of the
cell whose index in a path (array of cell ids) is given. */
/*
Returned Value: double (the X coordinate)

Called By:
findMaxEnd
findMinStart
roundSide

*/

double findPathXCoord( /*   ARGUMENTS              */
					  int * path,           /*!< linear path            */
					  int index)            /*!< index in path of point */
{
	return findMapXCoord(path[index] / obsMapSide);
}

/***********************************************************************/

/*! findPathYCoord finds the obsMap Y coordinate of the center of the
cell whose index in a path (array of cell ids) is given. */
/*
Returned Value: double (the Y coordinate)

Called By:
findMaxEnd
findMinStart
roundSide

*/

double findPathYCoord( /*   ARGUMENTS              */
					  int * path,           /*!< linear path            */
					  int index)            /*!< index in path of point */
{
	return findMapYCoord(path[index] % obsMapSide);
}

/***********************************************************************/

/*! findPlanTimeCost finds the cost in terms of time of a
PrimMobJACmdMoveArcSegment plan. */
/*
Returned Value: none

Called By: not currently called

This sets *cost to the cost of the plan.

The cost is set to the extra cost plus the time taken to traverse
the planned path at the planned speeds. If the speed of any arc is less
than or equal to zero and the end of the arc is not the same as the
start, this sets *cost to VERY_BIG_COST, since the time taken will be
NaN if the speed is zero, and a negative speed is an error.

This is not using the obstacleness of the map cells through which
the path passes in computing the cost.

This will need to be changed to be compatible with the way the
vehicle level computes cost (which uses obstacleness and distance).

*/
/* COMMENTED OUT
static void findPlanTimeCost(       //   ARGUMENTS
PrimMobJACmdMoveArcSegment * plan, // plan for which to find cost
double * cost,                     // place to record cost found here
double extra,                      // extra cost
double startX,                     // X coord of start point of first arc
double startY)                     // Y coord of start point of first arc
{
int n;
ArcSegment * seg;
double d;

*cost = extra;
for (n = 0; n < plan->arcList_length; n++)
{
seg = (plan->arcList + n);
if ((seg->end.x == startX) && (seg->end.y == startY));
else if (seg->speed <= 0)
{
*cost = VERY_BIG_COST;
break;
}
else if (seg->isArc)
{
d = hypot((seg->center.x - seg->end.x),(seg->center.y - seg->end.y));
*cost = (*cost + ((d * fabs(seg->theta)) / seg->speed));
}
else
{
d = hypot((startX - seg->end.x), (startY - seg->end.y));
*cost = (*cost + (d / seg->speed));
}
startX = seg->end.x;
startY = seg->end.y;
}
}
*/
/***********************************************************************/

//! getFileNames gets the names of the nmlFile and the iniFile
/*
Returned Value: none

Called By: main

If the nmlFile (a global variable) is NULL (implying it was not in the
command arguments), this gets the nmlFile name (1) from the
environment if it is there or (2) from the default if not.

This handles the iniFileName (also a global variable) similarly.

*/

static void getFileNames() /* NO ARGUMENTS */
{
	char * nmlFileEnv;
	char * iniFileEnv;

	if (NULL == nmlFile)
	{
		nmlFileEnv = getenv("CONFIG_NML");
		if (NULL != nmlFileEnv)
		{
			nmlFile = nmlFileEnv;
		}
		else
		{
			nmlFile = DEFAULT_NML_FILE;
		}
	}
	if (NULL == iniFileName)
	{
		iniFileEnv = getenv("CONFIG_INI");
		if (NULL != iniFileEnv)
		{
			iniFileName = iniFileEnv;
		}
		else
		{
			iniFileName = DEFAULT_INI_FILE;
		}
	}
}

/***********************************************************************/

//! getOptions gets the command line options for starting the process
/*
Returned Value: int
If there is an error, this returns 1. Otherwise, it returns 0.

Called By: main

Five command line options are recognized:
1. -n <NML file name>  (a string)
2. -i <ini file name>  (a string)
3. -d <buffer number>  (int giving a suffix for using different buffers)
4. -o <map side cells> (odd int giving number of cells on side of output map)
5. -z <meters>         (double giving resolution, i.e. size of side of a cell)

The first three get used only at startup. The last two are saved and
reused whenever an init command is received.

This uses the global (?) variable optarg.

*/

static int getOptions( /*   ARGUMENTS                                       */
					  int argc,             /*!< number of command line arguments                */
					  char * argv[],        /*!< array of command and arguments                  */
					  int * bufNum)         /*!< buffer number, set here if specified with -d    */
{
	char option;

	while (1)
	{
		option = getopt(argc, argv, "n:i:d:o:z:c:p:");
		if (option == -1)
			break;
		switch (option)
		{
		case 'n':
			nmlFile = optarg;
			break;
		case 'i':
			iniFileName = optarg;
			break;
		case 'd':
			if (1 != sscanf(optarg, "%d", bufNum))
			{
				rcs_print_error("%s: bad value for buffer number: %s\n",
					processName, optarg);
				return 1;
			}
			if (verbose)
				rcs_print("%s: using specified buffer suffix %d\n",
				processName, *bufNum);
			break;
		case 'o':
			if (1 != sscanf(optarg, "%d", &argObsMapSideLength))
			{
				rcs_print_error("%s: side length of obstacle map must be an "
					"integer in meters, but %s is given\n",
					processName, optarg);
				return 1;
			}
			else if (argObsMapSideLength < 1)
			{
				rcs_print_error("%s: side length of obstacle map (in meters) "
					"must be positive, but %d is given\n",
					processName, argObsMapSideLength);
				return 1;
			}
			break;
		case 'z':
			if (1 != sscanf(optarg, "%lf", &argResolution))
			{
				rcs_print_error("%s: bad value for obstacle map "
					"resolution: %s\n", processName, optarg);
				return 1;
			}
			break;
		case 'c':
			if (1 != sscanf(optarg, "%d", &commID))
			{
				rcs_print_error("%s: bad value for commID: %s\n",
					processName, optarg);
				return 1;
			}
			break;
		case 'p':
			//robotType = optarg[0];
			robotType = '\0';
			if(strcmp(optarg,"Matilda")==0)
				robotType = 'M';
			else if(strcmp(optarg,"P2AT")==0)
				robotType = 'P';
			else if(strcmp(optarg,"AirRobot")==0)
				robotType = 'A';
			else if(strcmp(optarg,"Talon")==0)
				robotType = 'T';
			break;

		default:
			rcs_print_error("%s: bad argument\n", processName);
			return 1;
			break;
		}
	}
	return 0;
}

/***********************************************************************/

/*!  handleDisplay turns the map display on or off.
If the turnOn argument is false, handleDisplay only deletes the
map image file. If the turnOn argument is true, handleDisplay deletes
the map image file and starts imageview so that every 100 milliseconds,
imageview will redraw the map image file if it is new. */
/*
Returned Value: none

Called By:
doCfgDumpWM
main

Imageview expects a file in ppm format.

*/

static void handleDisplay( /*   ARGUMENTS                                   */
						  bool turnOn)              /*!< true=clean up & reopen, false=clean up only */
{
	char sysStr[200];
	int noWarn;  // just to stop compiler warnings

	if (turnOn == false)
	{
		sprintf(sysStr, "rm -rf %s\n", mapImageFileName);
		noWarn = system(sysStr);
	}
	else
	{
		sprintf(sysStr, "rm -rf %s\n", mapImageFileName);
		noWarn = system(sysStr);
		sprintf(sysStr, "./imageview -f %s -r 100 -s 1 &\n", mapImageFileName);
		noWarn = system(sysStr);
	}
}

/***********************************************************************/

//! iniLoad loads parameters from the named inifile
/*
Returned Value: int
If the inifile can't be opened or if any given value is invalid, this 
returns INI_INVALID. If an expected value is missing, so that the default
value is used, it returns INI_DEFAULT. Otherwise, it returns INI_OK.

Called By: doCmdInit

Throughout MOAST, the top-down order of precedence for the source of
most values is:
1. command line argument
2. iniFile
3. default value

For the mapImageFileName, the order is:
1. iniFile
2. default value

For the iniFile and the NML file name, the order is:
1. command line argument
2. environment variable (CONFIG_INI or CONFIG_NML)
3. default value

In this function, the values for Prim parameters must be available in
the iniFile.

Values loaded from inifile:

[PRIM_MOB_JA]
MAX_TRAN_VEL        <positive number>  meters per second

[AM_MOB_JA]
CELL_RESOLUTION     <positive number>  meters
CYCLE_TIME          <positive number>  seconds
MAP_IMAGE_FILE_NAME <char *> name of file into which to print map image
MAP_SIDE_LENGTH     <positive int>     length of side of map in meters

Before this is called, doCmdInit calls initParameters, which resets
"stts" and "set" and gives all other global variables their default
values. This function loads the inifile, which results in some values
being reset.

If none of the attempts to find a value returns INI_INVALID, this
resets the cycle timer and calls setRelatedValues, which sets several
values based on cell resolution and obsMapSideLength.

Note that iniLoadOneDouble and iniLoadOneInt require all parameters to be
positive, so that is not rechecked here.

*/

static int iniLoad(   /* ARGUMENTS          */
				   char * platformName) /*!< name of platform */
{
	INIFILE iniFile;
	int retval;
	static char platformAndName[200];

	retval = INI_OK;
	if (iniFile.open(iniFileName))
	{
		rcs_print_error("%s: can't open INI file %s\n",
			processName, iniFileName);
		return INI_INVALID;
	}

	if (argResolution == UNSET_CELL_RESOLUTION)
		iniLoadOneDouble(&iniFile, "CELL_RESOLUTION", "AM_MOB_JA",
		&(set.cellResolution), &retval);
	iniLoadOneDouble(&iniFile, "CYCLE_TIME", "AM_MOB_JA",
		&(set.cycleTime), &retval);
	iniLoadOneString(&iniFile, "MAP_IMAGE_FILE_NAME", "AM_MOB_JA",
		&mapImageFileName, &retval);
	if (argObsMapSideLength == UNSET_OBS_MAP_SIDE_LENGTH)
		iniLoadOneInt(&iniFile, "MAP_SIDE_LENGTH", "AM_MOB_JA",
		&obsMapSideLength, &retval);
	if (platformName)
		sprintf(platformAndName, "PLATFORM.%s", platformName);
	else
		sprintf(platformAndName, "PLATFORM");
	iniLoadOneDouble(&iniFile, "MAX_TRAN_VEL", platformAndName,
		&maxSpeed, &retval);
	iniLoadControlPoint(&iniFile, "CONTROL_POINT", platformAndName,
		&set.controlPoint.tran, &set.controlPoint.rot, &retval);
	// load ip address of XML server
	iniLoadOneString(&iniFile, "XML_HOST_NAME", "USARSIM.INFSKIN", &xmlHostName, &retval);
	iniFile.close();
	if (retval != INI_INVALID)
	{ // set several values based on cell resolution and obsMapSideLength
		if (setRelatedValues(set.cellResolution))
			retval = INI_INVALID;
		if (timer)
			delete timer;
		timer = new RCS_TIMER(set.cycleTime);
	}
	initNodesAStar();
	initNodesDijk();
	initGoals();
	return retval;
}

/***********************************************************************/

//! iniLoadControlPoint sets the control point from the iniFile.
/*

Returned Value: none

Called By: iniLoad

1. If the parameter (which is "CONTROL_POINT") is found in the iniFile:
1A. If six double are not read, retval is set to INI_INVALID.
1B. If six double are read, the double arguments are set to those values,
and retval keeps its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

This uses the global variable
iniFileName

*/

static void iniLoadControlPoint(INIFILE * iniFile,  const char * parameter,
								const char * heading, PM_CARTESIAN * tran,
								PM_QUATERNION * rot, int * retval)
{
	const char * iniString;
	double d1;
	double d2;
	double d3;
	double d4;
	double d5;
	double d6;
	PM_RPY rpy;

	iniString = iniFile->find(parameter, heading);
	if (iniString)
	{
		if (6 == sscanf(iniString, "%lf %lf %lf %lf %lf %lf", 
			&d1, &d2, &d3, &d4, &d5, &d6))
		{ // convert degrees to radians for rot
			tran->x = d1;
			tran->y = d2;
			tran->z = d3;
			rpy.r = RAD(d4);
			rpy.p = RAD(d5);
			rpy.y = RAD(d6);
			*rot = rpy;
		}
		else
		{ // found, but invalid
			rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
				processName, iniString, heading, parameter,
				iniFileName);
			*retval = INI_INVALID;
		}
	}
	else
	{ // not found at all
		rcs_print("can't find [%s] %s in '%s'. using default value\n",
			heading, parameter, iniFileName);
		*retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
	} 
}

/***********************************************************************/

//! iniLoadOneDouble sets one parameter value from the iniFile.
/*
Returned Value: none

Called By: iniLoad

1. If the parameter is found in the iniFile:
1A. If it has an invalid value, retval is set to INI_INVALID.
1B. If it has a valid value, setMe is set to that value, and retval keeps
its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

Note that this requires all parameters to be positive.

This uses the global variable
iniFileName

*/

static void iniLoadOneDouble( /*   ARGUMENTS                              */
							 INIFILE * iniFile,           /*!< INIFILE class instance to use          */
							 const char * parameter,      /*!< name of parameter in ini file          */
							 const char * heading,        /*!< heading under which to find parameter  */
							 double * setMe,              /*!< number to set                          */
							 int * retval)                /*!< iniLoad return value, maybe reset here */
{
	const char * iniString;
	double d;

	iniString = iniFile->find(parameter, heading);
	if (iniString)
	{
		if ((1 == sscanf(iniString, "%lf", &d)) && (d > 0.0))
			*setMe = d;
		else
		{ // found, but invalid
			rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
				processName, iniString, heading, parameter,
				iniFileName);
			*retval = INI_INVALID;
		}
	}
	else
	{ // not found at all
		rcs_print("can't find [%s] %s in '%s'. using %f\n",
			heading, parameter, iniFileName, *setMe);
		*retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
	}
}

/***********************************************************************/

//! iniLoadOneInt sets one integer parameter value from the iniFile.
/*
Returned Value: none

Called By: iniLoad

1. If the parameter is found in the iniFile:
1A. If it has an invalid value, retval is set to INI_INVALID.
1B. If it has a valid value, setMe is set to that value, and retval keeps
its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

Note that this requires all parameters to be positive.

*/

static void iniLoadOneInt( /*   ARGUMENTS                              */
						  INIFILE * iniFile,        /*!< INIFILE class instance to use          */
						  const char * parameter,   /*!< name of parameter in ini file          */
						  const char * heading,     /*!< heading under which to find parameter  */
						  int * setMe,              /*!< number to set                          */
						  int * retval)             /*!< iniLoad return value, maybe reset here */
{
	const char * iniString;
	int d;

	iniString = iniFile->find(parameter, heading);
	if (iniString)
	{
		if ((1 == sscanf(iniString, "%d", &d)) && (d > 0))
			*setMe = d;
		else
		{ // found, but invalid
			rcs_print_error("%s: invalid value '%s' for [%s] %s in '%s'\n",
				processName, iniString, heading, parameter,
				iniFileName);
			*retval = INI_INVALID;
		}
	}
	else
	{ // not found at all
		rcs_print("can't find [%s] %s in '%s'. using %d\n",
			heading, parameter, iniFileName, *setMe);
		*retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
	}
}

/***********************************************************************/

//! iniLoadOneString sets one char * parameter value from the iniFile.
/*
Returned Value: none

Called By: iniLoad

1. If the parameter is found in the iniFile, setMe is set to that value,
and retval keeps its previous value.

2. If the parameter is not found in the iniFile:
2A. If retval was previously INI_OK, retval is set to INI_DEFAULT.
2B. Otherwise, retval keeps its previous value.

*/

static void iniLoadOneString( /*  ARGUMENTS                               */
							 INIFILE * iniFile,           /*! INIFILE class instance to use           */
							 const char * parameter,      /*! name of parameter in ini file           */
							 const char * heading,        /*! heading under which to find parameter   */
							 char ** setMe,               /*! string to set                           */
							 int * retval)                /*! iniLoad return value, maybe reset here  */
{
	const char * iniString;

	iniString = iniFile->find(parameter, heading);
	if (iniString)
	{ // found
		*setMe = strdup(iniString);
	}
	else
	{ // not found
		rcs_print("can't find [%s] %s in '%s', using %s\n",
			heading, parameter, iniFileName, *setMe);
		*retval = ((*retval == INI_OK) ? INI_DEFAULT : *retval);
	}
}

/***********************************************************************/

/*! iniSetVehicleParameters sets the vehicle width, the vehicle
minimum turning radius, and the platform name from the servoMobJASet
buffer. */
/*
Returned Value: int
If the servoMobJASet buffer cannot be read or the width is not positive,
or the minimum turning radius is negative, this returns INI_INVALID.
Otherwise, it returns INI_OK.

Called By: doCmdInit

If the buffer can be read and the width is positive and the minimum
turning radius is non-negative, this sets the width and the minimum
turning radius.  It is expected that the width and the minimum turning
radius will already be set to their default values when this is
called.

*/

static int iniSetVehicleParameters( /*   ARGUMENTS                      */
								   double * width,                    /*!< width to set                   */
								   double * minTurnRad,               /*!< minimum turning radius to set  */
								   char ** platformName)              /*!< platform name to set           */
{
	NMLTYPE setType;
	ServoMobJASetGrdVeh * setMsg;
	ServoMobJASetAirBot * setAirMsg;
	int allOK;

	setType = servoMobJASetBuf->read();

	// servoMobJA only writes the set buffer once so read() will return
	// zero after the first init.
	// Fixit so we can run init more than once if we need to.
	if(setType == 0)
	{
		setType = servoMobJASetBuf->get_address()->type;
	}

	// Make the type available in general  
	allOK = 0;
	switch (setType)
	{
	case 0:
		// no new command
		break;
	case -1:
		// comm error
		rcs_print_error("%s: NML read error in iniSetVehicleWidth\n",
			processName);
		break;
	case SERVO_MOB_JA_SET_GRD_VEH_TYPE:
		setMsg = (ServoMobJASetGrdVeh *)(servoMobJASetBuf->get_address());
		if ((setMsg->width > 0) && (setMsg->minTurningRadius >= 0))
		{
			allOK = 1;
			*width = setMsg->width;
			*minTurnRad = setMsg->minTurningRadius;
			*platformName = setMsg->platformName;
			vehLength = setMsg->length; // temporary
		}
		break;
	case SERVO_MOB_JA_SET_AIR_BOT_TYPE:
		setAirMsg = (ServoMobJASetAirBot *)(servoMobJASetBuf->get_address());
		allOK = 1;
		*width = setAirMsg->width;
		*platformName = setAirMsg->platformName;
		vehLength = setAirMsg->length; // temporary
		//rcs_print_error("%lf,%lf,%s\n\n", *width, vehLength, *platformName);
		break;

	default:
		rcs_print_error("%s: unknown config: %d\n", processName, (int) setType);
		break;
	}  // switch (setType)
	return (allOK ? INI_OK : INI_INVALID);
}

/***********************************************************************/

//! initAmMobJANmlBuffers initializes the AM level mobility NML buffers
/*
Returned Value: int
If there is an error, this returns 1. Otherwise it returns 0.

Called By: initBuffers

If there is an error, any buffers that have been allocated will need
to be deleted. This is handled in initBuffers.

*/

static int initAmMobJANmlBuffers( /*   ARGUMENTS                            */
								 int bufnum)                      /*!< number to use as channel name suffix */
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	cleanupAmMobJANmlBuffers();

	sprintf(chanName, "%s%d", AM_MOB_JA_CMD_NAME, bufnum);
	amMobJACmdBuf =
		new RCS_CMD_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((amMobJACmdBuf == NULL) || (!amMobJACmdBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", AM_MOB_JA_STAT_NAME, bufnum);
	amMobJAStatBuf =
		new RCS_STAT_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((amMobJAStatBuf == NULL) || (!amMobJAStatBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", AM_MOB_JA_CFG_NAME, bufnum);
	amMobJACfgBuf =
		new RCS_CMD_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((amMobJACfgBuf == NULL) || (!amMobJACfgBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", AM_MOB_JA_SET_NAME, bufnum);
	amMobJASetBuf =
		new RCS_STAT_CHANNEL(amMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((amMobJASetBuf == NULL) || (!amMobJASetBuf->valid()))
		return 1;

	//! AM SP Cmd Channel
	sprintf(chanName, "%s%d", AM_SP_CMD_NAME, bufnum);
	amSPCmdBuf =
		new RCS_CMD_CHANNEL(amSP_format, chanName, THIS_PROCESS, nmlFile);
	if ((amSPCmdBuf == NULL) || (!amSPCmdBuf->valid()))
		return 1;

	//! AM SP Stat Channel
	sprintf(chanName, "%s%d", AM_SP_STAT_NAME, bufnum);
	amSPStatBuf =
		new RCS_STAT_CHANNEL(amSP_format, chanName, THIS_PROCESS, nmlFile);
	if((amSPStatBuf == NULL) || (!amSPStatBuf->valid()))
		return 1;

	//! ServoMobJACmd Channel
	sprintf(chanName, "%s%d", SERVO_MOB_JA_CMD_NAME, bufnum);
	servoMobJACmdBuf = 
		new RCS_CMD_CHANNEL(servoMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if (servoMobJACmdBuf == NULL || (!servoMobJACmdBuf->valid()))
		return 1;

	//! ServoMobJAStat Channel
	sprintf(chanName, "%s%d", SERVO_MOB_JA_STAT_NAME, bufnum);
	servoMobJAStatBuf = 
		new RCS_STAT_CHANNEL(servoMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if((servoMobJAStatBuf==NULL) || (!servoMobJAStatBuf->valid()))
		return 1;

	return 0;
}

/***********************************************************************/

//! initBuffers initializes all buffers
/*
Returned Value: int
If there is an error, this returns 1. Otherwise, it returns 0.

Called By: main

This initializes the NML buffers that connect AM mobility to:
1. prim mobility
2. vehicle mobility (or a command shell)
3. nav data
4. map data (provided by AM SP)
5. servo mobility (to get vehicle width and minimum turning radius)

If there is an error at any stage, previously allocated buffers are
cleaned up.

*/

static int initBuffers( /*   ARGUMENTS                            */
					   int bufnum)            /*!< number to use as channel name suffix */
{
	int retval;

	retval = initPrimMobJANmlBuffers(bufnum);
	if (retval)
	{
		rcs_print_error("%s: can't allocate primMobJA NML buffers\n",
			processName);
		cleanupPrimMobJANmlBuffers();
		return 1;
	}
	retval = initAmMobJANmlBuffers(bufnum);
	if (retval)
	{
		rcs_print_error("%s: can't allocate amMobJA NML buffers\n",
			processName);
		cleanupNmlBuffers();
		return 1;
	}
	retval = initNavDataBuffer(bufnum);
	if (retval)
	{
		rcs_print_error("%s: can't allocate navData NML buffer\n",
			processName);
		cleanupNmlBuffers();
		return 1;
	}
	retval = initSensorData2DBuffer(bufnum);
	if (retval)
	{
		rcs_print_error("%s: can't allocate sensorData2D NML buffer\n",
			processName);
		cleanupNmlBuffers();
		return 1;
	}
	retval = initServoMobJASetBuffer(bufnum);
	if (retval)
	{
		rcs_print_error("%s: can't allocate servoMobJASet NML buffer\n",
			processName);
		cleanupNmlBuffers();
		return 1;
	}

	retval = initServoSPBuffers(bufnum);
	if (retval)
	{
		rcs_print_error("%s: can't allocate servoSP buffers.\n", processName);
		cleanupNmlBuffers();
		return 1;
	}
	return 0;
}

/***********************************************************************/

/*! initGoals records in the goals array the node indexes of the cells
on the periphery of the planning horizon */
/*
Returned Value: none

Called By: iniLoad

The points start near the northwest (upper left) corner of the square,
and proceed clockwise around the periphery from there.

*/

static void initGoals() /* NO ARGUMENTS */
{
	int n; // goals index
	int k; // cell id of goal
	int stop;

	if (goals)
		delete goals;
	goals = new int[4 * obsMapSi];
	n = 0;
	k = (obsMapSi * obsMapSide);
	stop = obsMapSi;
	for (n = 0; n < stop; n++)
	{
		goals[n] = k;
		k++;
	}
	stop = (2 * obsMapSi);
	for (; n < stop; n++)
	{
		goals[n] = k;
		k = (k - obsMapSide);
	}
	stop = (3 * obsMapSi);
	for (; n < stop; n++)
	{
		goals[n] = k;
		k--;
	}
	stop = (4 * obsMapSi);
	for (; n < stop; n++)
	{
		goals[n] = k;
		k = (k + obsMapSide);
	}
}

/***********************************************************************/

/*! initKidsAStar sets the kidsCosts of a node. It also sets the goal
cost of each unopened kid. */
/*
Returned Value: none

Called By: planPathAStar

The goal cost of a kid is [the cheapest cost per unit distance
(AM_EASY2)] times [the actual shortest distance from the kid to the
end along lines between centers of adjacent cells]. If the kid lies
towards the end from the start, a small amount is added to the actual
shortest distance, namely [0.01 times (the distance from the kid to
the straight line from start to end)]. The added amount tends to make
the path be close to the line from start to end. The added amount may
make the heuristic inadmissable in a strict sense, but makes the path
closer to the actual shortest path (which is not limited to following
lines between centers of adjacent cells).

Without the added amount, all paths from start (S) to end (E) that are
in a parallelogram-shaped area and that continually get closer to the
end will have the same cost, as shown on the left below, and all those
nodes may be opened (the actual path taken may be determined by
rounding error!). With the added amount, there is a unique cheapest
path as shown on the right, and only those nodes will be opened.

Sxxxxxxxx               Sx
xxxxxxxxx                xxxx
xxxxxxxxx                   xxxx
xxxxxxxxE                      xE

The actual shortest distance is found by first computing dX and dY
(the number of cells difference in X and Y from start to end).  Next,
let maxD be (the maximum of dX and dY) and let minD be (the minimum of
dX and dY). Then the shortest distance is [(maxD minus minD) times
(the straight distance between cell centers)] plus [minD times (the
diagonal distance between cell centers)].

The [distance from the kid to the straight line from start to end] is
calculated by dividing [the cross product of (the line from start to
kid) with (the line from start to end)] by [the length of the line from
start to end].

Whether the kid lies towards the end from the start is found by
taking the dot product of the line from start to end with the line
from start to kid and seeing whether it is positive.

The cost of an arc from a dad node to a kid node is equal to (the cost
per unit length of crossing the dad node cell times the distance
traveled across the dad node cell) plus (the cost per unit length of
crossing the kid node cell times the distance traveled across the kid
node cell).  The cost per unit length is:
AM_EASY2 if the cell is marked AMSP_NOT_OBSTACLE
AM_UNKNOWN if the cell is marked AMSP_UNSEEN
AM_DIFFICULT if the cell is marked AMSP_NEAR_OBSTACLE
AM_DANGEROUS if the cell is marked AMSP_MAYBE_OBSTACLE

The distance traveled across a cell is straightDist (middle to edge)
if the neighbor index is even and diagDist (middle to corner) if the
neighbor index is odd.

If a kid does not exist because the cell for the node is on an edge or
corner of the map, which is indicated by kidsIndexes[n] being -1, the
kidsCosts for that location is set to -1.  Also, if the cell for a kid
node is marked AMSP_OBSTACLE, the cost of an arc to the node for that
cell is set to -1, indicating that there really is no arc to it.  The
searcher knows that a cost of -1 means there is no arc and does not
process any kid node whose cost is -1.

*/

static void initKidsAStar( /*   ARGUMENTS                       */
						  int dadIndex,             /*!< index of node at start of arcs  */
						  int dadRow,               /*!< row of node at start of arcs    */
						  int dadCol,               /*!< column of node at start of arcs */
						  int endRow,               /*!< row of end point                */
						  int endCol,               /*!< column of end point             */
						  double startX,            /*!< X coordinate of start point     */
						  double startY,            /*!< Y coordinate of start point     */
						  double dX,                /*!< change in X from start to end   */
						  double dY,                /*!< change in Y from start to end   */
						  double dist)              /*!< distance from start to end      */
{
	AStarHeapNode * node1;
	int kidIndex;
	int kidRow;
	int kidCol;
	int n;
	double dadFactor; // cost per unit distance of crossing dad node cell
	double kidFactor; // cost per unit distance of crossing kid node cell
	double kidX;
	double kidY;
	int dRow;         // absolute value of (kidRow - endRow)
	int dCol;         // absolute value of (kidCol - endCol)
	int minD;         // smaller of dRow and dCol
	int maxD;         // larger of dRow and dCol
	double offset;    // distance from kid to line from start to end
	double obsVal;

	node1 = (nodesAStar + dadIndex);
	obsVal = obsMap[dadRow][dadCol];
	dadFactor = ((obsVal == AMSP_NOT_OBSTACLE) ? AM_EASY2 :
		(obsVal == AMSP_UNSEEN) ? AM_UNKNOWN :
		(obsVal == AMSP_NEAR_OBSTACLE) ? AM_DIFFICULT :
		AM_DANGEROUS);
	for (n = 0; n < 8; n++)
	{
		kidIndex = node1->kidsIndexes[n];
		if (kidIndex == -1)
		{
			node1->kidsCosts[n] = -1;
			continue;
		}
		kidRow = (kidIndex / obsMapSide);
		kidCol = (kidIndex % obsMapSide);
		if (obsMap[kidRow][kidCol] == AMSP_OBSTACLE)
		{
			node1->kidsCosts[n] = -1;
			continue;
		}
		obsVal = obsMap[kidRow][kidCol];
		kidFactor = ((obsVal == AMSP_NOT_OBSTACLE)? AM_EASY2:
			(obsVal == AMSP_UNSEEN) ? AM_UNKNOWN :
			(obsVal == AMSP_NEAR_OBSTACLE) ? AM_DIFFICULT :
			AM_DANGEROUS);
		node1->kidsCosts[n] =
			((dadFactor + kidFactor) * ((n % 2) ? diagDist : straightDist));
		if (nodesAStar[kidIndex].open == AStarHeapNode::UNOPENED)
		{
			kidX = findMapXCoord(kidRow);
			kidY = findMapYCoord(kidCol);
			dRow = (endRow - kidRow);
			dRow = ((dRow < 0) ? -dRow : dRow);
			dCol = (endCol - kidCol);
			dCol = ((dCol < 0) ? -dCol : dCol);
			minD = ((dRow < dCol) ? dRow : dCol);
			maxD = ((dRow > dCol) ? dRow : dCol);
			if (((dX * (kidX - startX)) + (dY * (kidY - startY))) > 0)
			{
				offset = (fabs((dX * (kidY-startY)) - (dY * (kidX-startX)))
					/ dist);
				nodesAStar[kidIndex].goalCost =
					(AM_EASY2 * 
					(((minD * extraDist) + (maxD * res)) + (0.01 * offset)));
			}
			else
			{
				nodesAStar[kidIndex].goalCost =
					(AM_EASY2 * ((minD * extraDist) + (maxD * res)));
			}
		}
	}
}

/***********************************************************************/

//! initKidsDijk sets the kidsCosts of a node.
/*
Returned Value: none

Called By: planPathsDijk

This is identical to initKidsAStar, except that in Dijkstra search there
is no goalCost. Finding goalCost takes a lot of code, so this is much
simpler than initKidsAStar.

The cost of an arc from a dad node to a kid node is equal to (the cost
per unit length of crossing the dad node cell times the distance
traveled across the dad node cell) plus (the cost per unit length of
crossing the kid node cell times the distance traveled across the kid
node cell).  The cost per unit length is:
AM_EASY2 if the cell is marked AMSP_NOT_OBSTACLE
AM_UNKNOWN if the cell is marked AMSP_UNSEEN
AM_DIFFICULT if the cell is marked AMSP_NEAR_OBSTACLE
AM_DANGEROUS if the cell is marked AMSP_MAYBE_OBSTACLE

The distance traveled across a cell is straightDist (middle to edge)
if the neighbor index is even and diagDist (middle to corner) if the
neighbor index is odd.

If a kid does not exist because the cell for the node is on an edge or
corner of the map, which is indicated by kidsIndexes[n] being -1, the
kidsCosts for that location is set to -1.  Also, if the cell for a kid
node is marked AMSP_OBSTACLE, the cost of an arc to the node for that
cell is set to -1, indicating that there really is no arc to it.  The
searcher knows that a cost of -1 means there is no arc and does not
process any kid node whose cost is -1.

*/

static void initKidsDijk( /*   ARGUMENTS                       */
						 int dadIndex,            /*!< index of node at start of arcs  */
						 int dadRow,              /*!< row of node at start of arcs    */
						 int dadCol)              /*!< column of node at start of arcs */
{
	DijkHeapNNode * node1;
	int kidRow;
	int kidCol;
	int n;
	int kidIndex;
	double dadFactor;
	double kidFactor;
	double obsVal;

	node1 = (nodesDijk + dadIndex);
	obsVal = obsMap[dadRow][dadCol];
	dadFactor = ((obsVal == AMSP_NOT_OBSTACLE) ? AM_EASY2 :
		(obsVal == AMSP_UNSEEN) ? AM_UNKNOWN :
		(obsVal == AMSP_NEAR_OBSTACLE) ? AM_DIFFICULT :
		AM_DANGEROUS);
	for (n = 0; n < 8; n++)
	{
		kidIndex = node1->kidsIndexes[n];
		if (kidIndex == -1)
		{
			node1->kidsCosts[n] = -1;
			continue;
		}
		kidRow = (kidIndex / obsMapSide);
		kidCol = (kidIndex % obsMapSide);
		if (obsMap[kidRow][kidCol] == AMSP_OBSTACLE)
		{
			node1->kidsCosts[n] = -1;
			continue;
		}
		obsVal = obsMap[kidRow][kidCol];
		kidFactor = ((obsVal == AMSP_NOT_OBSTACLE)? AM_EASY2:
			(obsVal == AMSP_UNSEEN) ? AM_UNKNOWN :
			(obsVal == AMSP_NEAR_OBSTACLE) ? AM_DIFFICULT :
			AM_DANGEROUS);
		node1->kidsCosts[n] =
			((dadFactor + kidFactor) * ((n % 2) ? diagDist : straightDist));
	}
}

/***********************************************************************/

//! initNavDataBuffer initializes the navDataExtBuf
/*
Returned Value: int
If there is an error in creating the navDataExtBuf, this returns 1.
Otherwise, it returns 0.

Called By: initBuffers

*/

static int initNavDataBuffer( /*   ARGUMENTS                            */
							 int bufnum)                  /*!< number to use as channel name suffix */
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	if (NULL != navDataExtBuf)
		delete navDataExtBuf;
	navDataExtBuf = NULL;
	sprintf(chanName, "%s%d", NAV_DATA_EXT_NAME, bufnum);
	navDataExtBuf = new NML(navDataExt_format, chanName, THIS_PROCESS, nmlFile);
	if ((navDataExtBuf == NULL) || (!navDataExtBuf->valid()))
		return 1;
	return 0;
}

/***********************************************************************/

/*! initNodesAStar initializes the nodes of the graph on the obsMap and
builds the arcs between nodes by assigning kids to each node. */
/*
Returned Value: int (0)

Called By: iniLoad

For each graph node, this:
- initializes the moreInfo
- sets kidsSize to 8
- gets memory for the kidsIndexes and kidsCosts arrays
- assigns values to the kidsIndexes

This does not assign values to the kidsCosts array.

Each node is identified with a cell in the obsMap, so we can say
nodes adjoin if the cells to which they are connected adjoin.
The kids of a node are some or all of the 8 adjoining nodes.
If a node is on a corner of the map, it has only 3 adjoining
nodes. If a node is on the edge of the map but not on a corner,
it has 5 adjoining nodes.

The indexes of the 8 neighbors of a node are implicitly assigned as
follows (N is the node). The zero index is directly north of N, and
numbering proceeds clockwise.

7 0 1
6 N 2
5 4 3

Rather than setting kidsSize to the actual number of nodes that can be
reached, kidsSize is always set to 8. If a kid does not exist because
the node is on a corner or edge, the value of kidsIndexes for that
location is set to -1.

*/

static int initNodesAStar() /* NO ARGUMENTS */
{
	static const int rowIncs[] = {1, 1, 0, -1, -1, -1,  0,  1};
	static const int colIncs[] = {0, 1, 1,  1,  0, -1, -1, -1};
	int dadRow;
	int dadCol;
	int kidRow;
	int kidCol;
	int n;
	int k;
	int stop;
	AStarHeapNode * aNode;
	int * kids;

	stop = (obsMapSide * obsMapSide);
	for (n = 0; n < stop; n++)
	{
		aNode = (nodesAStar + n);
		aNode->moreInfo = new NodeInfo(n);
		aNode->kidsSize = 8;
		kids = new int[8];
		aNode->kidsIndexes = kids;
		aNode->kidsCosts = new double[8];
		dadRow = (n / obsMapSide);
		dadCol = (n % obsMapSide);
		for (k = 0; k < 8; k++)
		{
			kidRow = (dadRow + rowIncs[k]);
			kidCol = (dadCol + colIncs[k]);
			kids[k] = (((kidRow < 0) || (kidRow > obsMapSi) ||
				(kidCol < 0) || (kidCol > obsMapSi)) ? -1 :
				((kidRow * obsMapSide) + kidCol));
		}
	}
	return 0;
}

/***********************************************************************/

//! initNodesDijk initializes the nodes of the graph on the obsMap
/*
Returned Value: int (0)

Called By: iniLoad

This is identical to initNodesAStar except that it uses DijkHeapNNode
and nodesDijk.

*/

static int initNodesDijk() /* NO ARGUMENTS */
{
	static const int rowIncs[] = {1, 1, 0, -1, -1, -1,  0,  1};
	static const int colIncs[] = {0, 1, 1,  1,  0, -1, -1, -1};
	int dadRow;
	int dadCol;
	int kidRow;
	int kidCol;
	int n;
	int k;
	int stop;
	DijkHeapNNode * aNode;
	int * kids;

	stop = (obsMapSide * obsMapSide);
	for (n = 0; n < stop; n++)
	{
		aNode = (nodesDijk + n);
		aNode->moreInfo = new NodeInfo(n);
		aNode->kidsSize = 8;
		kids = new int[8];
		aNode->kidsIndexes = kids;
		aNode->kidsCosts = new double[8];
		dadRow = (n / obsMapSide);
		dadCol = (n % obsMapSide);
		for (k = 0; k < 8; k++)
		{
			kidRow = (dadRow + rowIncs[k]);
			kidCol = (dadCol + colIncs[k]);
			kids[k] = (((kidRow < 0) || (kidRow > obsMapSi) ||
				(kidCol < 0) || (kidCol > obsMapSi)) ? -1 :
				((kidRow * obsMapSide) + kidCol));
		}
	}
	return 0;
}

/***********************************************************************/

/*! initParameters puts data in key system parameters, command status,
and config status. */
/*
Returned Value: int (0)

Called By:
doCmdInit
main

Command status (stts), configuration status (set), and most other
global variables are given initial values here.

*/

static int initParameters() /* NO ARGUMENTS */
{
	// initialize task command status
	stts.admin_state = ADMIN_UNINITIALIZED;
	stts.amMobJAError = AM_MOB_JA_ERROR_NONE;
	stts.seed_length = 0;

	// initialize configuration command status
	if (argResolution != UNSET_CELL_RESOLUTION)
		set.cellResolution = argResolution; // reset in setRelatedValues
	else
		set.cellResolution = DEFAULT_CELL_RESOLUTION; // reset in setRelatedValues
	set.controlPoint = PM_POSE(PM_CARTESIAN(0,0,0),PM_QUATERNION(1,0,0,0));
	set.cycleTime = DEFAULT_CYCLE_TIME;
	set.debug = 0;
	set.dumpWM = 0;
	// set.planHorizon is set in setRelatedValues, called below
	set.vehicleMinTurnRad = DEFAULT_VEHICLE_MIN_TURN_RAD;
	set.vehicleWidth = DEFAULT_VEHICLE_WIDTH;

	if (argObsMapSideLength != UNSET_OBS_MAP_SIDE_LENGTH)
		obsMapSideLength = argObsMapSideLength;
	else 
		obsMapSideLength = DEFAULT_OBS_MAP_SIDE_LENGTH;
	if (setRelatedValues(set.cellResolution))
	{ // sets obsGrow. May reset set.cellResolution slightly.
		rcs_print_error("%s: bad DEFAULT_CELL_RESOLUTION\n", processName);
		return 1;
	}

	// initialize other global variables
	// arcToWaypointTable needs no initialization
	// argObsMapSideLength was set in getOptions
	// argResolution was set in getOptions
	cellPixels = 2;
	// cellsPerMeter was set in setRelatedValues
	// diagDist was set in setRelatedValues
	done = 0;
	// extraDist was set in setRelatedValues
	// goals is initialized in iniLoad by call to initGoals
	// iniFileName is not to be reset
	if (logFp)
		fclose(logFp);
	logFp = NULL;
	// mapImageFileName needs no initialization
	maxSpeed = DEFAULT_MAX_SPEED;
	minSpeed = DEFAULT_MIN_SPEED;
	// nmlFile is not to be reset
	// nodesAStar is initialized in iniLoad by call to initNodesAStar
	// nodesDijk is initialized in iniLoad by call to initNodesDijk
	obsData = NULL;
	// obsGrow was set in setRelatedValues
	// obsMap needs no initialization
	// obsMapSide was set in setRelatedValues
	// obsMapSi was set in setRelatedValues
	// obsMapSideLength was set above
	// primCmdNumber is reset only in writePrimCmd
	// processName is not to be reset
	// res was set in setRelatedValues
	// set was set above
	// skips was set in setRelatedValues
	// straightDist was set in setRelatedValues
	// stts was set above
	if (timer)
		delete timer;
	timer = new RCS_TIMER(set.cycleTime);
	// vehSpeed needs no initialization
	// vehSpin needs no initialization
	// vehVelocity needs no initialization
	// vehX needs no initialization
	// vehY needs no initialization
	// vehYaw needs no initialization
	// verbose is not currently resettable, probably should be
	// xMax needs no initialization
	// xMin needs no initialization
	// yMax needs no initialization
	// yMin needs no initialization
	return 0;
}

/***********************************************************************/

//! initPrimMobJANmlBuffers opens NML channels to Prim mobility
/*
Returned Value: int
If an error occurs, this returns 1. Otherwise, it returns 0.

Called By: initBuffers

*/

static int initPrimMobJANmlBuffers(/*   ARGUMENTS                            */
								   int bufnum)                       /*!< number to use as channel name suffix */
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	cleanupPrimMobJANmlBuffers();

	sprintf(chanName, "%s%d", PRIM_MOB_JA_CMD_NAME, bufnum);
	primMobJACmdBuf =
		new RCS_CMD_CHANNEL(primMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((primMobJACmdBuf == NULL) || (!primMobJACmdBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", PRIM_MOB_JA_STAT_NAME, bufnum);
	primMobJAStatBuf =
		new RCS_STAT_CHANNEL(primMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((primMobJAStatBuf == NULL) || (!primMobJAStatBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", PRIM_MOB_JA_CFG_NAME, bufnum);
	primMobJACfgBuf =
		new RCS_CMD_CHANNEL(primMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((primMobJACfgBuf == NULL) || (!primMobJACfgBuf->valid()))
		return 1;

	sprintf(chanName, "%s%d", PRIM_MOB_JA_SET_NAME, bufnum);
	primMobJASetBuf =
		new RCS_STAT_CHANNEL(primMobJA_format, chanName, THIS_PROCESS, nmlFile);
	if ((primMobJASetBuf == NULL) || (!primMobJASetBuf->valid()))
		return 1;

	return 0;
}

/***********************************************************************/

//! initSensorData2DBuffer initializes the sensorData2DBuf
/*
Returned Value: int
If there is an error in creating the sensorData2DBuf, this returns 1.
Otherwise, it returns 0.

Called By: initBuffers

*/

static int initSensorData2DBuffer( /*   ARGUMENTS                            */
								  int bufnum)                       /*!< number to use as channel name suffix */
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	if (NULL != sensorData2DBuf)
		delete sensorData2DBuf;
	sensorData2DBuf = NULL;
	sprintf(chanName, "%s%d", AM_SP_MAP_NAME, bufnum);
	sensorData2DBuf = new NML(sensorData_format, chanName, THIS_PROCESS,nmlFile);
	if ((sensorData2DBuf == NULL) || (!sensorData2DBuf->valid()))
		return 1;
	return 0;
}

/***********************************************************************/

//! initServoMobJASetBuffer initializes the servoSetBuf
/*
Returned Value: int
If there is an error in creating the sensorData2DBuf, this returns 1.
Otherwise, it returns 0.

Called By: initBuffers

The servoMobJASetBuf is needed in order to get the following information
about the vehicle:
1. width (for growing obstacles)
2. minimum turning radius (for generating paths for Prim Mob)
3. name (so that more info can be extracted from moast.ini).

*/

static int initServoMobJASetBuffer(/*   ARGUMENTS                            */
								   int bufnum)                       /*!< number to use as channel name suffix */
{
	char chanName[MOAST_NML_BUFFER_NAME_LEN];

	if (NULL != servoMobJASetBuf)
		delete servoMobJASetBuf;
	servoMobJASetBuf = NULL;
	sprintf(chanName, "%s%d", SERVO_MOB_JA_SET_NAME, bufnum);
	servoMobJASetBuf = new RCS_STAT_CHANNEL(servoMobJA_format, chanName,
		THIS_PROCESS, nmlFile);
	if ((servoMobJASetBuf == NULL) || (!servoMobJASetBuf->valid()))
		return 1;
	return 0;
}

/***********************************************************************/

/* intersectLines

Returned Value: int (0)

Called By: (not currently called)

Notes:

This finds the intersection point (ix, iy) of two lines, one going
through (x1, y1) at an angle of a1, the other going through (x2, y2)
at an angle of a2. 

Let d1 be the distance from (x1, y1) to the intersection point and d2
be the distance from (x2, y2) to the intersection point.

Using the usual equations for polar coordinates, there are two
equations for iy and two for ix. We eliminate ix and iy by setting the
two sets of right sides equal, then solve for d1. We can then
substitute d1 back into the original equations to find ix and iy.

iy = y1 + (d1 * sin(a1))
iy = y2 + (d2 * sin(a2))

ix = x1 + (d1 * cos(a1))
ix = x2 + (d2 * cos(a2))

y1 + (d1 * sin(a1)) = y2 + (d2 * sin(a2))
x1 + (d1 * cos(a1)) = x2 + (d2 * cos(a2))

Eliminate d2 by multiplying the top equation by cos(a2) and the
bottom equation by sin(a2), then subtracting.

(y1*cos(a2)) + (d1*sin(a1)*cos(a2)) = (y2*cos(a2)) + (d2*sin(a2)*cos(a2))
(x1*sin(a2)) + (d1*cos(a1)*sin(a2)) = (x2*sin(a2)) + (d2*cos(a2)*sin(a2))

d1*((sin(a1)*cos(a2)) - (cos(a1)*sin(a2))) =
(y2*cos(a2)) - (x2*sin(a2)) + (x1*sin(a2)) - (y1*cos(a2))

d1 * sin(a1-a2) = ((x1-x2)*sin(a2)) + ((y2-y1)*cos(a2))

d1 = (((x1-x2)*sin(a2)) + ((y2-y1)*cos(a2))) / sin(a1-a2)

Note that d1 may be negative.

The last equation fails (as it should) when sin(a1 - a2) is 0, since
in that case the lines are parallel or antiparallel and do not
intersect (unless the two lines are identical).

This equation works for all angles.

This is assuming the lines are not colinear, the points are distinct,
and the pointers are not NULL, which is the case for the use made of
this function.

*/
/* COMMENTED OUT
static int intersectLines( //   ARGUMENTS
double x1,                // X-coordinate of first point
double y1,                // Y-coordinate of first point
double a1,                // angle of first line
double x2,                // X-coordinate of second point
double y2,                // Y-coordinate of second point
double a2,                // angle of second line
double * ix,              // X-coordinate of intersection point, set here
double * iy)              // Y-coordinate of intersection point, set here
{
double d1;

d1 = ((((x1 - x2) * sin(a2)) + ((y2 - y1) * cos(a2))) / sin(a1 - a2));
*ix = (x1 + (d1 * cos(a1)));
*iy = (y1 + (d1 * sin(a1)));
return 0;
}
*/
/*************************************************************************/

/*! linearizeAStarPath linearizes the part P of the aStarPath between two
given path indexes m and n (with m < n). */
/*
Returned Value: int
If the second index is not greater than the first index, this returns 1.
If a recursive call is made to itself, this returns whatever the
recursive call returns. Otherwise, it returns 0.

Called By:
linearizeAStarPath (recursively)
planPrimMoveOne

The general approach is to find the longest straight lines from the
start and end of P extending to points on P and not hitting obstacles,
to use those lines as part of the linearized path, and then to call
the function recursively for the middle part of P. Straight lines are
defined by setting the elements of path between the end points to -1.

1. If n == (m+1) return. There is no obstacle between the two points.
Otherwise, continue to the next step.

2a. Find the smallest index j <= n for which the line from path[m] to
path[j] hits an obstacle.

2b. If there is no such j, set all the elements of the path between
m and n to -1 and return 0. Otherwise, continue to the next step.

2c. Subtract 1 from j.

2d. Find the smallest k >= m for which the line from path[k] to
path[n] hits an obstacle. There must be such a k.

2e. Add 1 to k.

3. If k <= j, set j = k = (j + k)/2.

4. Set all the elements of the path between m and j and between k and
n to -1.

5. If (k - j) > 1, call this function with j and k as end indexes
and return whatever it returns. Otherwise, return 0.

*/

static int linearizeAStarPath( /*   ARGUMENTS             */
							  int m,                        /*!< first index           */
							  int n,                        /*!< second index          */
							  int * path)                   /*!< copy of the aStarPath */
{
	int i;    // utility index
	int j;    // index for end of line from start
	int k;    // index for start of line to end

	if (n <= m)
	{
		rcs_print_error("%s: bug. n must be greater than m but isn't\n",
			processName);
		return 1;
	}
	if (n == (m + 1))
		return 0;
	findMaxEnd(m, n, path, &j);
	if (j == n)
	{
		for (i = (m+1); i < n; i++)
			path[i] = -1;
		return 0;
	}
	findMinStart(m, n, path, &k);
	if (j >= k)
	{
		j = ((j + k + 1) / 2);
		k = j;
	}
	for (i = (m+1); i < j; i++)
		path[i] = -1;
	for (i = (k+1); i < n; i++)
		path[i] = -1;
	return (((k - j) > 1) ? linearizeAStarPath(j, k, path) : 0);
}

/***********************************************************************/

//! locateCells locates points on the periphery of the planning area.
/*

Returned Value: int
If the obsMapSideLength is not an integer number of meters, this
returns 1. Otherwise, it returns 0.

Called By: planAll

Global Variables Used: set, obsData

When cellResolution is set, it is required to be such that there are
an integer number of cells per meter.

The planning area is the same area as the most recently built obstacle
map.

This locates points by picking the middle cell in each meter of the
edge if the number of cells per meter is odd, and the cell just to the
left of the middle (looking from inside the square) if the number of
cells per meter is even. This makes a total of (4 * obsMapSideLength)
points. Each point is at the middle of its cell. The coordinates of
the point associated with the cell are recorded in the x and y values
of a PM_CARTESIAN, respectively. The z value of each PM_CARTESIAN is
set to -1.0; this will serve as a flag that no plan has been made to
that point yet.

The results are recorded in the "cells" array of PM_CARTESIAN.

Recall that X is north (up) and Y is east (right).

The points start near the northwest (upper left) corner of the square,
and proceed clockwise around the periphery from there. On each side,
there are obsMapSideLength points, and successive points are 1 meter
apart. The diagram below shows the situation for a square 4 meters on
a side with 5 AM cells per meter. The points are marked with an X.

---------------------------------------
|    X    |    X    |    X    |    X    |
|         |         |         |         |
|X        |         |         |        X|
|         |         |         |         |
|         |         |         |         |
|---------------------------------------|
|         |         |         |         |
|         |         |         |         |
|X        |         |         |        X|
|         |         |         |         |
|         |         |         |         |
|---------------------------------------|
|         |         |         |         |
|         |         |         |         |
|X        |         |         |        X|
|         |         |         |         |
|         |         |         |         |
|---------------------------------------|
|         |         |         |         |
|         |         |         |         |
|X        |         |         |        X|
|         |         |         |         |
|    X    |    X    |    X    |    X    |
---------------------------------------


The diagram below shows the analogous situation for a square 4 meters
on a side with 4 AM cells per meter. The points being used are
marked with an X. Note that points in the same column on the top and bottom
have the same east value when the number of cells per meter is odd, but
have different east values when the number of cells per meter is even.
Similarly for the right and left sides.

-----------------------------------
|   X    |   X    |   X    |   X    |
|        |        |        |       X|
|X       |        |        |        |
|        |        |        |        |
|-----------------------------------|
|        |        |        |        |
|        |        |        |       X|
|X       |        |        |        |
|        |        |        |        |
|-----------------------------------|
|        |        |        |        |
|        |        |        |       X|
|X       |        |        |        |
|        |        |        |        |
|-----------------------------------|
|        |        |        |        |
|        |        |        |       X|
|X       |        |        |        |
|    X   |    X   |    X   |    X   |
-----------------------------------

The lowerLeft is the point at the southwest corner of the obstacle
map. It is also the point at the southwest corner of the cell in the
southwest corner of the obstacle map.

This fails if obsMapSideLength is not an integer number of meters.

*/
/* COMMENTED OUT
static int locateCells( //   ARGUMENTS
PM_CARTESIAN * cells)  // array of cells to write
{
int n;
double north;
double east;
int cellsPerMeter;
double inc1;
double inc2;

cellsPerMeter = (int)((1.0 / res) + 0.1);
if ((obsMapSide % cellsPerMeter) != 0)
{
rcs_print_error("%s: map side must be an integer number of meters\n",
processName);
return 1;
}
inc1 = (res * ((cellsPerMeter - 1) / 2));
inc2 = (res *  (cellsPerMeter      / 2));
east = (obsData->lowerLeft.y + inc1 + (res * 0.5));
north = (obsData->lowerLeft.x +
(res * ((obsMapSideLength * cellsPerMeter) - 0.5)));
for (n = 0; n < obsMapSideLength; n++)
{
cells[n].x = north;
cells[n].y = east;
cells[n].z = -1.0;
east = (east + 1.0);
}
east = ((east - 1.0) + inc2);
north = (north - inc1);
for ( ; n < (2 * obsMapSideLength); n++)
{
cells[n].x = north;
cells[n].y = east;
cells[n].z = -1.0;
north = (north - 1.0);
}
east = (east - inc1);
north = ((north + 1.0) - inc2);
for ( ; n < (3 * obsMapSideLength); n++)
{
cells[n].x = north;
cells[n].y = east;
cells[n].z = -1.0;
east = (east - 1.0);
}
east = ((east + 1.0) - inc2);
north = (north + inc1);
for ( ; n < (4 * obsMapSideLength); n++)
{
cells[n].x = north;
cells[n].y = east;
cells[n].z = -1.0;
north = (north + 1.0);
}
return 0;
}
*/

/******************************
WSS Functions
******************************/
static bool SendInit(string RobotName, int Port)
{
	char response[8192];
	char sPort[33];
	bool result = false;

	sprintf(sPort, "%d", Port);
	string cmd = "INIT {Robot ";
	cmd.append(RobotName);
	cmd.append("} {Port ");
	cmd.append(sPort);
	cmd.append("}\r\n");

	// Send Command
	send(socket_descriptor, cmd.c_str(), strlen(cmd.c_str()), 0);

	// Get Response
	//recv(socket_descriptor, response, 8192, 0);  
	bool done = false;
	for(int i=0; !done; i++)
	{
		read(socket_descriptor, response+i, 1);
		if((int)response[i] == 13)
		{
			read(socket_descriptor, response+i+1, 1);
			done = true;
		}
	}

	// Evaluate Response
	if(string(response).compare(INITResponse())==0)
		result = true;

	return result;
}

static float GetSignalStrength(string RobotName)
{
	char response[8192];
	string cmd = "GETSS {Robot ";//100;
	cmd.append(RobotName);
	cmd.append("}\r\n");

	printf("Sending Command: %s\n", cmd.c_str());
	// Send Command
	send(socket_descriptor, cmd.c_str(), strlen(cmd.c_str()), 0);
	printf("Done Sending...\n");

	printf("Receiving Response\n");

	//while(!done)
	// Get Response
	//recv(socket_descriptor, response, 8192, 0);  


	bool done = false;
	for(int i=0; !done; i++)
	{
		read(socket_descriptor, response+i, 1);
		if((int)response[i] == 13)
		{
			read(socket_descriptor, response+i+1, 1);
			done = true;
		}
	}

	printf("Done Receiving Response (Got: %s\n",response);

	return 5.0;//SignalStrengthResponse(string(response));
}

static string INITResponse()
{
	return "INITREPLY {Status OK}\r\n";
}

static float SignalStrengthResponse(string msg)
{
	float result;
	string tmp = msg.substr(msg.find("{Strength ") + 10, msg.length() - 1 - msg.find("{Strength ") - 10);

	if(tmp.compare("NaN")==0)
	{
		result = 100;
	}
	else if(tmp.compare("INF")==0)
	{
		result = 101;
	}
	else
	{
		sscanf(tmp.c_str(), "%f", &result);
	}

	return result;
}

static int DNSResponse(string msg)
{
	int result;

	string tmp = msg.substr(msg.find("{Port ") + 6, msg.length() - 1 - msg.find("{Port ") - 6);
	sscanf(tmp.c_str(), "%d", &result);

	return result;
}


static bool readConfigFile()
{
	INIFILE if_d;
	const char *strPtr;

	if (if_d.open (DEFAULT_INI_FILE))
	{
		printf
			("** ERROR ** <amMobMain::readConfigFile> Inifile %s not Found \n\t",
			DEFAULT_INI_FILE);
		return false;
	}

	if (if_d.isSection ("USARSIM.INFSKIN"))
	{
		printf
			("** ERROR ** <amMobMain::readConfigFile> : Section USARSIM.INFSKIN not found\n");
		if_d.close();
		return false;
	}

	if (NULL == (strPtr = if_d.find ("HOST_NAME", "USARSIM.INFSKIN")))
	{
		printf("** WARNING ** <amMobMain::readConfigFile> : Parameter HOST_NAME not found\n");
	}
	else
	{
		strcpy(wssIP, strPtr);
	}

	printf("Read wssIP from config file: %s\n", wssIP);

	return true;
}


/***********************************************************************/

//! main is the main routine for AM mobility
/*
Returned Value: int
If there is an error in getOptions, initBuffers, or initParameters
(which are called only once at start-up) this returns 1.
Otherwise, it returns 0.

Called By: not called (but runs automatically)

Global Variables Used:
amMobJASetBuf, amMobJAStatBuf, done, logFpset, stts, timer

This starts the AM controller in the uninitialized admin_state, but
default values are set for every parameter that would be set by
reading moast.ini.

This stuffs no ops in its own command and config buffers, so that
when cycling starts nothing happens.

*/

int main(       /*   ARGUMENTS                             */
		 int argc,      /*!< one more than the number of arguments */
		 char * argv[]) /*!< array of process name and arguments   */
{
	int bufnum = 1;          // default NML channel name suffix
	AmMobJACfgNop nopCfg;
	AmMobJACmdNop nopCmd;

	processName = argv[0];

	if (getOptions(argc, argv, &bufnum))
	{
		return 1;
	}
	getFileNames();
	if (initBuffers(bufnum))
	{
		return 1;
	}
	reset_time_tracker(&stts.tt);
	stts.echo_serial_number = 0;
	set.echo_serial_number = 0;
	if (initParameters())
	{
		return 1;
	}

	amMobJAStat = (AmMobJAStat *) amMobJAStatBuf->get_address();
	servoMobJAStat = (ServoMobJAStat*) servoMobJAStatBuf->get_address();
	amSPStat = (AmSPStat *) amSPStatBuf->get_address();

	servoSPStat = (ServoSPStat*) servoSPStatBuf->get_address();

	for(; ; timer->wait())
	{ // check repeatedly until no garbage status is read
		amSPStatBuf->read();
		if ((amSPStat->echo_serial_number >= 1) &&
			(amSPStat->command_type >= AM_SP_CMD_INIT_TYPE) &&
			(amSPStat->command_type <= AM_SP_CMD_NOP_TYPE))
			break;
	}

	nopCmd.serial_number = 1;
	amMobJACmdBuf->write(&nopCmd);   // stuff a no op in own command buffer
	nopCfg.serial_number = 1;
	amMobJACfgBuf->write(&nopCfg);   // stuff a no op in own config buffer
	signal(SIGINT, quit);


	// Set up some global variables
	readConfigFile();

	// Spawn an instance of the Communications class.
	printf("ROBOT TYPE: %c\n", robotType);
	comm = new Communications(commID,wssIP,robotType);

	float SS;

	key_t key;
	int shmid;
	static bool readOnce = false;

	// make the key:
	char tok[7];
	sprintf(tok, "Robot%d", bufnum);
	if ((key = ftok(tok, 'R')) == -1) 
	{
		perror("ftok");
		exit(1);
	}

	// connect to (and possibly create) the segment
	if ((shmid = shmget(key, 1024000, 0644 | IPC_CREAT)) == -1) 
	{
		perror("shmget");
		exit(1);
	}

	// attach to the segment to get a pointer to it: */
	unsigned char *dataPtr = (unsigned char *)shmat(shmid, (void *)0, 0);
	if (dataPtr == (unsigned char *)(-1)) 
	{
		perror("shmat");
		exit(1);
	}

	/* Scan Flag shared mem -> Determine whether to update the map or not.*/
	// Info for the char-sized shared memory used to turn off scanning.
	key_t scanKey;
	int scanShmID;

	// make the scanning key:
	char scanTok[33];
	sprintf(scanTok, "Robot%dScanFlag",bufnum);
	if ((scanKey = ftok(scanTok, 'R')) == -1)
	{
		perror("ftok(scanTok)");
		exit(1);
	}

	// connect to and possibly create the scanFlag segment
	if ((scanShmID = shmget(scanKey, 1, 0644 | IPC_CREAT)) == -1)
	{
		perror("shmget(scanKey)");
		exit(1);
	}

	// attach to the scanFlag segment and get a pointer to it
	unsigned char* scanFlagPtr = (unsigned char*)shmat(scanShmID, (void*)0, 0);
	if (scanFlagPtr == (unsigned char*)(-1))
	{
		perror("shmat(scanShmID)");
		exit(1);
	}

	*scanFlagPtr = 'Y';

	char lastCmd[4];
	bzero(lastCmd, 4);
	while (!done)
	{
		//if(!gotInitialPose)
		//updatePosition();
		readCmd();
		readCfg();
		readPrimMobStat();
		amSPStatBuf->read();

		// Read from shared memory
		if(!readOnce)
		{
			if((char)(dataPtr[5])=='M' && (char)(dataPtr[6])=='A' && (char)(dataPtr[7])=='P')
			{
				//printf("\n\nInside Read Once\n\n");
				readOnce = true;
			}
		}

		if(readOnce && (char)(dataPtr[0])=='R')
		{
			//printf("\n\nSending com message\n");

			// Read Size
			int msgSize = 0;
			double xLow;
			double yLow;
			double resolution;
			float xPos;
			float yPos;
			int numUpdates;

			memcpy(&(msgSize), &dataPtr[1], 4);
			memcpy(&xLow, &dataPtr[8],8);
			memcpy(&yLow, &dataPtr[16],8);
			memcpy(&resolution, &dataPtr[24],8);
			memcpy(&xPos, &dataPtr[32],4);
			memcpy(&yPos, &dataPtr[36],4);
			memcpy(&numUpdates, &dataPtr[40],4);

			printf("Sent MAP: xPos=%f yPos=%f numUpdates=%d\n\n", xPos, yPos, numUpdates);

			//updatePosition();

			comm->sendMessage(0, &dataPtr[5], msgSize);

			dataPtr[0] = (unsigned char)('W');
		}


		CommMsg * msg;
		char* curTok;
		for(int i=0;i<Communications::ROBOTCOUNT;i++)
		{
			while(comm->readIncomingMessage(i,&msg))
			{

				/// INIT Command received.
				if(strcmp(msg->type,"INI") == 0)
				{
					printf("Processing Initialize Command.\n");
					memcpy(lastCmd, msg->type, 4);

					if(robotType != 'A')
					{
						writeMobCmd(&amMobJACmdInit);
						writeSPCmd(&amSPCmdInit);
						doCmd();
					}
					else
					{
						writeServoSPCmd(&servoSPCmdInit);
						writeServoMobCmd(&servoMobJACmdInit);
					}
					//doCmdInit();

					delete msg; // Clean up is handled on this end for now.
					break;
				}
				/// SPGo Command received.
				else if(strcmp(msg->type,"SPG") == 0)
				{
					memcpy(lastCmd, msg->type, 4);
					if(robotType != 'A')
					{
						writeSPCmd(&amSPCmdGo);
						doCmd();
					}
					else
						writeServoSPCmd(&servoSPCmdGo);

					delete msg; // Clean up is handled on this end for now.
					break;
				}
				/// SKID Steer Command received
				else if(strcmp(msg->type,"SKD") == 0)
				{
					// If the last command was a waypoint command, we need to stop the robot first.
					if(strcmp(lastCmd,"WAY") == 0)
					{
						PrimMobJAStat* subStatPtr = (PrimMobJAStat *) (primMobJAStatBuf->get_address());
						writePrimCmd(&primMobJACmdStop, subStatPtr->echo_serial_number);
						writeMobCmd(&amMobJACmdNop);
					}
					// Update the lastCmd as SKD
					memcpy(lastCmd, msg->type, 4);

					//pthread_mutex_lock(&(comm.msgInBlock[i]));
					//pthread_mutex_unlock(&(comm.msgInBlock[i]));
					servoMobJACmdSkid.wLeft=msg->params[0];
					servoMobJACmdSkid.wRight=msg->params[1];

					//printf("Got SKD command, LEFT: %f RIGHT: %f\n", msg->params[0], msg->params[1]);
					writeServoMobCmd(&servoMobJACmdSkid);

					doCmd();

					delete msg; // Cleanup is handled on this end for now.
					//printf("Wrote SKD command to servoMobJACmdChannel.\n");
					break;
				}
				/// Aerial Drive Command received.timer->wait();
				else if(strcmp(msg->type,"ARD") == 0)
				{
					servoMobJACmdRotaryWing.vVertical=msg->params[0];
					servoMobJACmdRotaryWing.vLinear=msg->params[1];
					servoMobJACmdRotaryWing.vLateral=msg->params[2];
					servoMobJACmdRotaryWing.vRotation=msg->params[3];
					writeServoMobCmd(&servoMobJACmdRotaryWing);

					doCmd();

					delete msg;
					break;
				}
				/// Waypoint command recieved.
				else if(strcmp(msg->type,"WAY") == 0)
				{
					printf("\namMobMain: Got Waypoint command\n");
					memcpy(lastCmd, msg->type, 4);
					amMobJACmdMoveWaypoint.waypoint[0].p.x = msg->params[0];
					amMobJACmdMoveWaypoint.waypoint[0].p.y = msg->params[1];
					amMobJACmdMoveWaypoint.waypoint[0].p.z = 0;
					amMobJACmdMoveWaypoint.waypoint[0].neighborhood = DEPLOYMENT_DEFAULT_NEIGHBORHOOD;
					amMobJACmdMoveWaypoint.waypoint[0].speed = DEPLOYMENT_DEFAULT_SPEED;
					amMobJACmdMoveWaypoint.waypoint_length=1;
					amMobJACmdMoveWaypoint.computeHorizonCosts =  false;
					writeMobCmd(&amMobJACmdMoveWaypoint);

					doCmd();

					delete msg;
					break;
				}
				/// Mapping ON command (turn on scanmatching).
				else if(strcmp(msg->type,"MON") == 0)
				{
					printf("Got MON Command.\n");
					*scanFlagPtr = 'Y'; // Y for yes

					delete msg;
					break;
				}
				/// Mapping OFF command (turn off scanmatching).
				else if(strcmp(msg->type,"MOF") == 0)
				{
					printf("amMobMain: Got MOF Command.\n");
					*scanFlagPtr = 'N'; // N for no

					delete msg;
					break;
				}
			}
		}

		//printf("Got Signal Strength.\n");

		doCmd();
		doCfg();
		cycle_time_tracker(&stts.tt);
		amMobJAStatBuf->write(&stts);
		amMobJASetBuf->write(&set);
		//timer->wait();

		servoMobJAStatBuf->read();
		comm->batteryLife = servoMobJAStat->batteryVolt;
	}

	shmdt(dataPtr);
	shmctl(shmid, IPC_RMID, NULL);

	shmdt(scanFlagPtr);
	shmctl(scanShmID, IPC_RMID, NULL);

	//WSSLog.close();

	while (!done);
	if (logFp != NULL)
	{
		fclose(logFp);
		logFp = NULL;
	}
	handleDisplay(false);
	cleanupNmlBuffers();
	rcs_print("amMobJA done\n");
	return 0;
}

/***********************************************************************/

/* mergeColinear merges consecutive colinear path segments into a
single line */
/*
Returned Value: none

Called By: planPrimMoveOne

When a path has been linearized, it may still have one or more
instances of two or more consecutive colinear segments. This can
happen, in particular, if a path has consecutive colinear segments
near an obstacle (since linearizePath calls checkObsLineThick, which
may return 1 if the line it is checking passes through a cell next to
an obstacle).

Colinear path segments must be removed because roundCorners cannot
handle them. It is desirable to remove them even if roundCorners could
handle them, since their existence wastes time and degrades path
quality.

This function goes through the path looking for consecutive colinear
path segments. If it finds any, it removes all the points in the
middle of those segments and decreases the length of the path.
Removal is accomplished by setting the index to -1.

Since the path is in terms of cell indexes, it is feasible to check
for exact slope matches.

When this is called, some of the indexes may already be -1; these are
ignored. The first and last indexes are never -1.

The path includes the first point when this is called.

IT MIGHT BE SMARTER TO USE checkObsLineThin IN linearizePath SO THAT
THIS FUNCTION WOULD NOT BE NEEDED.

*/

static void mergeColinear( /*   ARGUMENTS                          */
						  int * path,               /*!< path to work in, array of node ids */
						  int size)                 /*!< length of path                     */
{
	int m;   // previous node index with path[m] not -1
	int n;   
	int x1;
	int y1;
	int x2;
	int y2;
	int dx1;
	int dy1;
	int dx2;
	int dy2;

	for (n = 1; path[n] == -1; n++);
	x2 = (path[n] / obsMapSide);
	y2 = (path[n] % obsMapSide);
	dx2 = (x2 - (path[0] / obsMapSide));
	dy2 = (y2 - (path[0] % obsMapSide));
	m = n;
	for (n++ ; n < size; n++)
	{
		if (path[n] == -1)
			continue;
		x1 = x2;
		y1 = y2;
		dx1 = dx2;
		dy1 = dy2;
		x2 = (path[n] / obsMapSide);
		y2 = (path[n] % obsMapSide);
		dx2 = (x2 - x1);
		dy2 = (y2 - y1);
		if ((dx1 * dy2) == (dx2 * dy1))
			path[m] = -1;
		m = n;
	} 
}

/***********************************************************************/

/*! minusPiToPi takes an angle and returns an angle in the range [-Pi, Pi)
that is equal to the original angle plus or minus a multiple of 2Pi. */
/*
Returned Value: double (see above)

Called By:
minusPiToPi (recursively)
roundSide
updatePosition

This is checking explicitly for equality to M_PI to avoid problems
(such as oscillation between less than -M_PI and M_PI) that might be
caused by rounding error.

*/

static double minusPiToPi( /*   ARGUMENTS                 */
						  double angle)             /*!< the angle to get in range */
{
	return ((angle < -M_PI) ? minusPiToPi(angle + (2 * M_PI)) :
		(angle >  M_PI) ? minusPiToPi(angle - (2 * M_PI)) :
		(angle == M_PI) ? -M_PI : angle);
}

/***********************************************************************/

/*! planAll plans how to get from the current location to 40 points
on the periphery of the planning area and calculates the cost
of each plan. */
/*
Returned Value: int
If any of the following errors occur, this returns 1.
Otherwise, it returns 0.
1. updateObsMap fails.
2. The vehicle is off the obsMap.

Called By:
doCmdMoveWaypoint
doCmdPrepareMove

Global Variables Used: obsMapSideLength, stts, vehX, vehY

The planning area is a square obsMapSideLength (an integer) meters on
a side, located so that the edges of the square are on cell boundaries
and the current location of the vehicle is near the center of the
square. This attempts to plan how to move from the current vehicle
location to one cell in each meter of the periphery of the square for
a total of (4 times the length of a side) cells. The xy location of
the center of the nth cell is recorded in the x and y values of
stts.seed[n]. The starting xy location of the vehicle for the plans is
recorded in stts.seed[m].

For each cell of index n to which a plan is made, the cost of getting
to the cell is recorded in stts.seed[n].z.  If planning to a point
fails, the cost is recorded in stts.seed[n].z as -1.0.

The value of stts.seed_length is set to (m+1). The extra 1 is there
because the value of stts.seed[m] is set to the current location.

The cells are placed in the planPathsDijk function.

The vehicle is located according to the most recent nav data read by
the caller.  If the most recently read nav data is the same as the nav
data read by AM SP when it built the most recent obstacle data, the
vehicle will be very near the center of the planning area. However,
the two nav data readings will probably differ, so that the vehicle
will not be at the center of the planning area.

*/

static int planAll(   /*   ARGUMENTS                         */
				   double tol,          /*!< half width of path tolerance zone */
				   double speed)        /*!< target speed, not to be exceeded  */
{
	int m;                 // number of goal cells reported

	if (updateObsMap() > 0) // error message already printed
		return 1;
	if (vehicleOffObsMap())
		return 1;
	m = (4 * obsMapSideLength);
	planPathsDijk(vehX, vehY, stts.seed);
	stts.seed[m].x = vehX;
	stts.seed[m].y = vehY;
	stts.seed[m].z = 0;
	// convert to global coordinates
	for( int i=0; i<m+1; i++ )
		stts.seed[i] = localToGlobal * stts.seed[i];
	stts.seed_length = (m+1);
	return 0;
}

/***********************************************************************/

/*! planPathAStar attempts to plan a path from the start point to the
end point using A* search. The heuristic for the goal cost of a
node is minimum actual distance to the goal along connections between
cells times the cheapest cost per unit distance. */
/*
Returned Value: none

Called By: planPrimMoveOne

If a path is found, it is stored in the aStarPath array, and
*length is set to the length of the path. If no path is found, *length
is set to 0. The path found by A* may be very jagged and require
smoothing to be suitable. Smoothing is not done here.

The version of initKidsAStar being used tends to smooth the path.

In timing tests on linus on 10/18/06 in simpleSim using a square grid
100 cells on a side, this took around 0.006 seconds (on a 2005 vintage
Dell Precision 670) when there were many unknown cells but around
0.0004 seconds after the area had been mapped and there were almost no
unknown cells (only clear or obstacle).  The planning algorithm has
been made somewhat faster since then.

*/

static void planPathAStar( /*   ARGUMENTS                                    */
						  double startX,            /*!< X coordinate of start point                  */
						  double startY,            /*!< Y coordinate of start point                  */
						  double endX,              /*!< X coordinate of end point                    */
						  double endY,              /*!< Y coordinate of end point                    */
						  int * aStarPath,          /*!< array of node indexes giving path, set here  */
						  int * length)             /*!< length of path found, or 0 if none, set here */
{
	AStarHeapGraph graph;
	AStarHeapSearcher searcher;
	int start;
	int end;
	int opens;           // 0=open list empty, non-zero=list not empty
	int status[1];       // status of reaching goal
	double cost[1];      // cost of reaching goal
	int endRow;
	int endCol;
	int dadRow;
	int dadCol;
	int cheapIndex;
	int aStarLength;
	double dX;
	double dY;
	double dist;

	start = findIdAtCoords(startX, startY);
	endRow = findMapXIndex(endX);
	endCol = findMapYIndex(endY);
	end = ((endRow * obsMapSide) + endCol);
	dX = (endX - startX);
	dY = (endY - startY);
	dist = hypot(dX, dY);
	graph.nodes = nodesAStar;
	graph.capacity = (obsMapSide * obsMapSide);
	graph.size = graph.capacity;
	graph.deleat = false;
	searcher.initSearch(&graph, start, end, &opens, status, cost);
	for (; ((status[0] < 2) && opens) ;)
	{
		cheapIndex = searcher.getNextNode();
		dadRow = (cheapIndex / obsMapSide);
		dadCol = (cheapIndex % obsMapSide);
		initKidsAStar(cheapIndex, dadRow, dadCol, endRow, endCol,
			startX, startY, dX, dY, dist);
		searcher.cycleSearch();
	}
	if (status[0] == 2)
	{
		aStarLength = searcher.reportSearch(MAX_MAP_CELLS, aStarPath);
		*length = ((aStarLength < 1) ? 0 : aStarLength);
	}
	else
	{
		*length = 0;
	}
	searcher.closeSearch();
}

/***********************************************************************/

/*! planPathsDijk attempts to plan paths from the start point to an
array of goal points using Dijkstra search. It also reports the
results in the cells array. */
/*
Returned Value: none

Called By: planAll

This attempts to find a path to every point on the periphery of the
planning region. The ids of those cells are always the same and have
been stored in the goals array.

Since the Dijkstra algorithm with multiple goals expands the search
area continually until all goals have been reached or there are no
more reachable nodes (think of an ink blot spreading), the computation
time required to search to all cells on the periphery is only slightly
greater than the time required to search to one cell in each meter of
the periphery. In fact, if even one meter of the periphery has no
reachable cells, the searches are identical in the two cases. Thus,
it is feasible (as done here) to search to all cells on the periphery
and then pick the cheapest one in each meter to report.

The cost of the optimal path to each point is stored in the z
coordinate of the point. If no path is found to a point, the cost is
set to -1.0.  The optimal paths are not saved. The paths are not
smoothed.

This reuses the nodesDijk for each graph. The kids of each node are
updated as the search proceeds.

It is possible for a corner cell to be the least cost cell for two
different meters of the periphery, so two different data points may
contain the same data.

*/

static void planPathsDijk( /*   ARGUMENTS                      */
						  double startX,            /*!< X coordinate of start point    */
						  double startY,            /*!< Y coordinate of start point    */
						  PM_CARTESIAN * data)      /*!< cartesian points for reporting */
{
	DijkHeapNGraph graph; // graph to search in
	DijkHeapNSearcher searcher; // searcher to do the searching
	int start;           // index of start node
	int opens;           // 0=open list empty, non-zero=list not empty
	int * statuses;      // array of statuses of reaching goal
	double * costs;      // array of costs of reaching goal
	int goalsFound;      // number of goals reached
	int optima;          // number of goals reached optimally
	int cheapIndex;      // index of current cheapest node
	int m;               // number of goals
	int dataIndex;       // index for data array
	int gcsIndex;        // index for goals, costs, and statuses arrays
	int stopDataIndex;   // limiter for dataIndex
	int sideIndex;       // index for sides of the planning area

	m = (4 * obsMapSi);  // not 4*obsMapSide, since corners used twice
	statuses = new int[m];
	costs = new double[m];
	start = ((findMapXIndex(startX) * obsMapSide) + findMapYIndex(startY));
	graph.nodes = nodesDijk;
	graph.capacity = (obsMapSide * obsMapSide);
	graph.size = graph.capacity;
	graph.deleat = false;
	searcher.initSearch(&graph, start, m, goals, &opens, statuses,
		costs, &goalsFound, &optima);
	for (; ((optima < m) && opens) ;)
	{
		cheapIndex = searcher.getNextNode();
		initKidsDijk(cheapIndex,
			(cheapIndex / obsMapSide), (cheapIndex % obsMapSide));
		searcher.cycleSearch();
	}
	searcher.closeSearch();
	// Now find the cheapest goal in each meter of periphery
	dataIndex = 0;
	gcsIndex = 0;
	for (sideIndex = 1; sideIndex < 5; sideIndex++)
	{
		stopDataIndex = (sideIndex * obsMapSideLength);
		for (; dataIndex < stopDataIndex; dataIndex++)
		{
			findBestGoal(data, dataIndex, gcsIndex, statuses, costs, m);
			gcsIndex = (gcsIndex + cellsPerMeter);
		}
		gcsIndex--; // back up 1 at corners since corner cells used twice
	}
}

/***********************************************************************/

//! planPrimMove prepares for and then tries planPrimMoveSome
/*
Returned Value: int
If any of the following functions fails, this returns 1.
Otherwise, it returns 0.
1. updateObsMap
2. vehicleOffObsMap
3. planPrimMoveSome

Called By: doCmdMoveWaypoint

This:
updates the vehicle position
updates the obstacle map
checks that the vehicle is on the map
clears a little space on the obstacle map if the vehicle is on an obstacle
calls planPrimMoveSome to find a path
draws the obstacle map (with the path if planPrimMoveSome succeeds or
with no path if planPrimMoveSome fails)

If the vehicle is on an obstacle, this marks the 9 cells immediately
around the vehicle as clear. These cells are probably on the periphery
of a grown obstacle, since the vehicle will not plan to go into an
obstacle and (if the relative velocities are low) will first be seen
to be in contact with the obstacle at its periphery.  Since obstacles
are grown more than half the width of the vehicle, when the vehicle
shrunk to a point is on the periphery of a grown obstacle, the actual
simulated vehicle is probably not in contact with an actual simulated
obstacle.  By clearing the cells immediately around the vehicle, the
planner has a chance to plan a way out. If the vehicle gets any
distance inside the periphery of a grown obstacle, this will just make
a little hole in the grown obstacle containing the vehicle, and the
planner will not be able to get out.

*/

static int planPrimMove(                   /*   ARGUMENTS                    */
						AmMobJACmdMoveWaypoint * amMoveCmd,       /*!< AM move command to plan from */
						PrimMobJACmdMoveArcSegment * primMoveCmd, /*!< prim move command to build   */
						int * primEndsOnWaypoint,                 /*!< 1=prim cmd ends on waypoint  */
						int index)                                /*!< index of 1st unvisited waypt */
{
	int i; // obsMap X index of vehicle
	int j; // obsMap Y index of vehicle

	if (updateObsMap() > 0) // error message already printed
		return 1;
	if (vehicleOffObsMap())
		return 1;
	i = findMapXIndex(vehX);
	j = findMapYIndex(vehY);
	if (obsMap[findMapXIndex(vehX)][findMapYIndex(vehY)] == AMSP_OBSTACLE)
	{ // make a local clear spot (probably on edge of grown obstacle)
		obsMap[i][j] = 0; 
		if (i > 0)
		{
			obsMap[i-1][j] = 0;
			if (j > 0)
				obsMap[i-1][j-1] = 0;
			if (j < obsMapSi)
				obsMap[i-1][j+1] = 0;
		}
		if (i < obsMapSi)
		{
			obsMap[i+1][j] = 0;
			if (j > 0)
				obsMap[i+1][j-1] = 0;
			if (j < obsMapSi)
				obsMap[i+1][j+1] = 0;
		}
		if (j > 0)
			obsMap[i][j-1] = 0;
		if (j < obsMapSi)
			obsMap[i][j+1] = 0;
	}
	if (planPrimMoveSome(amMoveCmd, primMoveCmd, primEndsOnWaypoint, index))
	{ // attempt to plan failed.
		if (set.dumpWM == true)
			drawObsMapFile(NULL);
		return 1;
	}
	else if (set.dumpWM == true)
	{
		drawObsMapFile(primMoveCmd);
	}
	return 0;
}

/***********************************************************************/

//! planPrimMoveOne plans moving through one waypoint in a series
/*
Returned Value: int
If no path can be found, or if there are too many arc segments, this
returns 1. Otherwise, this returns 0.

Called By:
planPrimMoveOneEdge
planPrimMoveSome

Side Effects: This adds one or more nodes to the path being built.
This also always sets pathLength.

Notes:

When this is called, it is anticipated that it might fail. Therefore,
this does not print an error message when it fails (except when the
cause of the error is a bug).

The general approach here is to go directly from the start
point to the end point if there is no obstacle on a thick straight
line between them. If there is an obstacle, this finds a path
(if there is one) using A* search and then linearizes and closes
up the path.

The end point is known to be on the obstacle map and known not to be
on a cell marked as an obstacle on the obstacle map.

The start point is known to be on the obstacle map but may be on a cell
marked as an obstacle on the obstacle map. This function signals an
error if that occurs.

Obstacles are already grown by half the vehicle width in the obsMap.

The path that that is constructed does not include the start point.
If planPathAStar is called and succeeds, the start point is initially
included on the path, but it is removed in closeLinearizedPath.

*/

static int planPrimMoveOne( /*   ARGUMENTS                   */
						   double startX,             /*!< X coordinate of start point */
						   double startY,             /*!< Y coordinate of start point */
						   double endX,               /*!< X coordinate of end point   */
						   double endY,               /*!< Y coordinate of end point   */
						   int * path,                /*!< array of node indexes       */
						   int * pathLength)          /*!< length of path, set here    */
{
	if (!checkObsLineThick(startX, startY, endX, endY))
	{ // no obstacles on path straight from start to end, so go straight
		path[0] = findIdAtCoords(endX, endY);
		*pathLength = 1;
	}
	else
	{
		planPathAStar(startX, startY, endX, endY, path, pathLength);
		if (*pathLength)
		{ // planPathAStar found a path
			if (*pathLength > 2) //path is more than just start and end
			{
				linearizeAStarPath(0, (*pathLength - 1), path);
				mergeColinear(path, *pathLength);
				closeLinearizedPath(path, *pathLength, pathLength);
			}
			else if (*pathLength == 2) // unlikely, but possible
			{ // still need to remove first point
				path[0] = path[1];
				*pathLength = 1;
			}
		}
	}
	return (*pathLength ? 0 : 1);
}

/***********************************************************************/

/*! planPrimMoveOneEdge attempts to plan a path from the start point
to an end point on the edge of the planning area. The end point
may be moved. */
/*

Returned Value: int
If no plan can be found or there is a bug in selecting a side, this
returns 1. Otherwise, it returns 0.

Called By: planPrimMoveSome

This is called if a moveWaypoint command includes a point W outside
the planning horizon. In that case, a line has been drawn from the
start point S to W and found to intersect the edge at E = (xEdge,
yEdge).  This function tries to make a plan to reach E. In some cases,
as described next, the point E will be moved to a different place on
the same edge. It is OK to move E since it is not a waypoint.

A. If planPrimMoveOne can make a plan to move from S to E, that plan
is used.

B. Otherwise, different locations for E are tried on an edge of the
planning horizon on which E lies (if E is at a corner, E is on two
edges, but only one edge is tried). The trial points are chosen at
gradually increasing distances from E on the right of E, and then (if
no point on the right of E works) at gradually increasing distances on
the left of E. As soon as a location for E is found for which a plan
can be made by planPrimMoveOne, the function returns zero. If no
location for E is found, the function returns one.

In the diagram:
The box is the planning horizon.
The o's are obstacle cells.
S is the start point.
W is the waypoint outside the planning horizon.
E is the edge point where line SW crosses the planning horizon.
e is the alternative point to E that the function will choose.

W

----oo-E-ooe----
|    oo oo     |
|     ooo      |
|              |
|              |
|      S       |
|              |
---------------


*/

static int planPrimMoveOneEdge(   /*   ARGUMENTS                   */
							   double xStart,                   /*!< X coordinate of start point */
							   double yStart,                   /*!< Y coordinate of start point */
							   double xEdge,                    /*!< X coordinate of end point   */
							   double yEdge,                    /*!< Y coordinate of end point   */
							   int * path,                      /*!< array of node indexes       */
							   int * pathLength)                /*!< length of path, set here    */
{
	int edgeI;
	int edgeJ;
	double x;
	double y;
	int n;

	edgeI = findMapXIndex(xEdge);
	edgeJ = findMapYIndex(yEdge);

	if (edgeI == obsMapSi)
	{ // E is on the north edge (parallel to Y axis)
		x = findMapXCoord(obsMapSi);
		for (n = edgeJ; n < obsMapSi; n++)
		{
			if ((obsMap[obsMapSi][n] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, x, findMapYCoord(n),
				path, pathLength)))
				return 0;
		}
		for (n = (edgeJ - 1); n > 0; n--)
		{
			if ((obsMap[obsMapSi][n] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, x, findMapYCoord(n),
				path, pathLength)))
				return 0;
		}
	}
	else if (edgeI == 0)
	{ // E is on the south edge (parallel to Y axis)
		x = findMapXCoord(0);
		for (n = edgeJ; n > 0; n--)
		{
			if ((obsMap[0][n] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, x, findMapYCoord(n),
				path, pathLength)))
				return 0;
		}
		for (n = (edgeJ + 1); n < obsMapSi; n++)
		{
			if ((obsMap[0][n] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, x, findMapYCoord(n),
				path, pathLength)))
				return 0;
		}
	}
	else if (edgeJ == obsMapSi)
	{ // E is on east edge (parallel to X axis)
		y = findMapYCoord(obsMapSi);
		for (n = edgeI; n > 0; n--)
		{
			if ((obsMap[n][obsMapSi] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, findMapXCoord(n), y,
				path, pathLength)))
				return 0;
		}
		for (n = (edgeI + 1); n < obsMapSi; n++)
		{
			if ((obsMap[n][obsMapSi] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, findMapXCoord(n), y,
				path, pathLength)))
				return 0;
		}
	}
	else if (edgeJ == 0)
	{ // E is on west edge (parallel to X axis)
		y = findMapYCoord(0);
		for (n = edgeI; n < obsMapSi; n++)
		{
			if ((obsMap[n][0] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, findMapXCoord(n), y,
				path, pathLength)))
				return 0;
		}
		for (n = (edgeI - 1); n > 0; n--)
		{
			if ((obsMap[n][0] != AMSP_OBSTACLE) &&
				(!planPrimMoveOne(xStart, yStart, findMapXCoord(n), y,
				path, pathLength)))
				return 0;
		}
	}
	else
		rcs_print_error("%s: bug in planPrimMoveOneEdge\n", processName);
	return 1; // did not find plan
}

/***********************************************************************/

//! planPrimMoveSome plans a smooth path for prim for moving through waypoints
/*
Returned Value: int
If any of following functions returns 1, this returns 1.
Otherwise, it returns 0.
1. checkDitherWaypoints.
2. planPrimMoveOne.
3. planPrimMoveOneEdge.
4. findEdgePoint.
5. roundCorners.

Called By: planPrimMove

When this is called, the position and the obsMap have just been
updated, and it is known that the vehicle is on the obsMap and not
on an obstacle. The index argument is the index of the next waypoint
to visit from the current position. Call the waypoint at that index
P1.

This calls checkDitherWaypoints to (1) check whether any unvisited
waypoints inside the planning horizon lie on obstacles, and (2)
find replacement points for those waypoints that do lie on obstacles.
If checkDitherWaypoints has moved any waypoints, they are put back
where they were by restoreWaypoints before this function exits.

If all the waypoints starting with P1 are inside the planning horizon,
this tries to make a prim command to pass through all those waypoints.

If one or more waypoints are outside the planning horizon, this tries
to make a prim command to pass through all the waypoints before the
first one outside plus a point on the edge of the planning horizon, as
selected by planPrimMoveOneEdge.

If the last waypoint inside the planning horizon (call it W) is on the
edge of the planning horizon and the line from that waypoint to the
first waypoint outside the planning horizon crosses the planning
horizon on the same edge, then no plan is made to the crossing point.
Note that in this case, the cell in which the crossing point lies
could be the one in which W lies, which would mess up corner rounding.

It is assumed here that the vehicle is moving in the direction it is
headed.

Values are set in the arcToWaypointTable so that for each index, segIndex,
of the arcList in the primMoveCmd, arcToWaypointTable[segIndex] is the
index in the waypoints of the amMoveCmd of the waypoint towards which
the ArcSegment arcList[segIndex] is moving.  Entries are needed in the
arcToWaypointTable by doCmdMoveWaypoint. After each time more arcs are
added to the primMoveCmd by planPrimMoveOne or planPrimMoveOneEdge,
new entries are added to the arcToWaypointTable here. segIndex is set to 0
near the beginning of this function so that the arcToWaypointTable gets
rewritten each time the function is called.

If planning has started and then fails, when the next attempt is made,
the entries in the arcToWaypointTable and the primMoveCmd made during the
failed attempt are overwritten.

The wholePath does not include the start point at (vehX, vehY), since
that is not necessarily at the center of a cell. The amMoveCmd does
not include the start point, and the partPaths found by planPrimMoveOne
and planPrimMoveOneEdge do not include their start points.

*/

static int planPrimMoveSome(      /*   ARGUMENTS                             */
							AmMobJACmdMoveWaypoint * amMoveCmd,       /*!< AM move command to plan from */
							PrimMobJACmdMoveArcSegment * primMoveCmd, /*!< prim move command to build   */
							int * primEndsOnWaypoint,        /*!< 1=prim cmd ends on waypoint, set here */
							int index)                       /*!< index of 1st unvisited waypoint       */
{
	int n;                   // index for Waypoints and utility int
	int m;                   // utility int
	int stop;                // one more than index of last waypoint to plan to
	double x;                // X coordinate of point
	double y;                // Y coordinate of point
	int xI;                  // X index of point
	int yJ;                  // Y index of point
	double xEdge;            // X coordinate of point on edge of obsMap
	double yEdge;            // Y coordinate of point on edge of obsMap
	int edgeI;               // X index of point on edge of obsMap
	int edgeJ;               // Y index of point on edge of obsMap
	Waypoint * aWaypoint;    // current waypoint
	static int wholePath[MAX_ARC_SEGMENTS]; // complete path of node indexes
	int wholeLength;         // index of line in wholePath
	static int partPath[MAX_ARC_SEGMENTS]; // partial path of node indexes
	int partLength;          // number of entries in partPath
	static int lineToWaypointTable[MAX_ARC_SEGMENTS];  // table[line]=waypoint
	static int savedLength;
	static int savedIndexes[MAX_ARC_SEGMENTS];
	static PM_CARTESIAN savedData[MAX_ARC_SEGMENTS];

	if (checkDitherWaypoints(amMoveCmd, index, &n, &savedLength,
		savedIndexes, savedData))
	{
		restoreWaypoints(amMoveCmd, savedLength, savedIndexes, savedData);
		return 1;
	}
	stop = n; // either amMoveCmd->waypoint_length or index of 1st pt outside
	wholeLength = 0;
	primMoveCmd->arcList_length = 0;
	x = vehX;
	y = vehY;
	for (n = index; n < stop; n++)
	{ // Plan to all points inside planning horizon.
		aWaypoint = (amMoveCmd->waypoint + n);
		if (planPrimMoveOne(x, y, aWaypoint->p.x, aWaypoint->p.y, 
			partPath, &partLength))
		{ // planPrimMoveOne failed
			restoreWaypoints(amMoveCmd, savedLength, savedIndexes, savedData);
			return 1;
		}
		for (m = 0; m < partLength; m++, wholeLength++)
		{ // add path to waypoint to whole path
			wholePath[wholeLength] = partPath[m];
			lineToWaypointTable[wholeLength] = n;
		}
		x = aWaypoint->p.x;
		y = aWaypoint->p.y;
	}
	if (stop < amMoveCmd->waypoint_length)
	{ // Are waypoints outside. Find point on horizon and try to plan to it.
		aWaypoint = (amMoveCmd->waypoint + stop);
		if (findEdgePoint(x, y, aWaypoint->p.x, aWaypoint->p.y, &xEdge, &yEdge))
		{ // findEdgePoint failed
			if (wholeLength == 0)
			{ // fail if cannot plan to edge and no path planned
				primMoveCmd->arcList_length = 0;
				restoreWaypoints(amMoveCmd, savedLength,
					savedIndexes,savedData);
				return 1;
			}
			else
			{ // OK if cannot plan to edge but some path planned
				*primEndsOnWaypoint = 1;
			}
		}
		else
		{ // findEdgePoint succeeded
			xI = findMapXIndex(x);
			yJ = findMapYIndex(y);
			edgeI = findMapXIndex(xEdge);
			edgeJ = findMapYIndex(yEdge);
			if (((xI == edgeI) && ((xI == 0) || (xI == obsMapSi))) ||
				((yJ == edgeJ) && ((yJ == 0) || (yJ == obsMapSi))))
			{ // edge pt on same edge as last planned waypoint, don't go there
				*primEndsOnWaypoint = 1;
			}
			else if (planPrimMoveOneEdge(x, y, xEdge, yEdge,
				partPath, &partLength))
			{ // planPrimMoveOneEdge failed
				if (wholeLength == 0)
				{ // fail if cannot plan to edge and no path planned
					primMoveCmd->arcList_length = 0;
					restoreWaypoints(amMoveCmd, savedLength,
						savedIndexes,savedData);
					return 1;
				}
				else
				{ // OK if cannot plan to edge but some path planned
					*primEndsOnWaypoint = 1;
				}
			}
			else
			{ // made plan to all inside waypoints plus edge point
				for (m = 0; m < partLength; m++, wholeLength++)
				{ // add path to edge point to whole path
					wholePath[wholeLength] = partPath[m];
					lineToWaypointTable[wholeLength] = stop;
				}
				*primEndsOnWaypoint = 0;
			}
		}
	}
	else // if (stop == amMoveCmd->waypoint_length)
	{ // All waypoints are inside horizon and made plan to all waypoints.
		*primEndsOnWaypoint = 1;
	}
	if (roundCorners(primMoveCmd, wholePath, wholeLength,
		amMoveCmd, lineToWaypointTable))
	{ // roundCorners failed; this should not happen
		rcs_print_error("%s: roundCorners failed\n", processName);
		restoreWaypoints(amMoveCmd, savedLength, savedIndexes, savedData);
		return 1;
	}
	restoreWaypoints(amMoveCmd, savedLength, savedIndexes, savedData);
	return 0;
}

/***********************************************************************/

//! quit sets done to 1 so that the while loop in main will terminate
/*
Returned Value: none

Called By: pointed at (not called) by signal, called if SIGINT occurs.

See documentation at the beginning of this file regarding termination.
The signal function requires a pointer to a function with one argument
which is an int.

*/

static void quit( /*   ARGUMENTS                                          */
				 int sig)         /*!< argument required by the signal function, not used */
{
	printf("closing\n");
	comm->closeAllSockets();
	delete comm;
	done = 1;
	exit(0);
}

/***********************************************************************/

//! readCfg reads the config command and modifies the config status buffer.
/*
Returned Value: none

Called By: main

*/

static void readCfg()  /* NO ARGUMENTS */
{
	NMLTYPE cfgType;
	int serialNumber;

	cfgType = amMobJACfgBuf->read();
	serialNumber = amMobJACfgBuf->get_address()->serial_number;
	switch (cfgType)
	{
	case 0:
		// no new command
		break;

	case -1:
		// comm error
		rcs_print_error("%s: NML config read error\n", processName);
		break;

	case AM_MOB_JA_CFG_CELL_RESOLUTION_TYPE:
	case AM_MOB_JA_CFG_CYCLE_TIME_TYPE:
	case AM_MOB_JA_CFG_DEBUG_TYPE:
	case AM_MOB_JA_CFG_DUMP_WM_TYPE:
	case AM_MOB_JA_CFG_NOP_TYPE:
	case AM_MOB_JA_CFG_PLAN_HORIZON_TYPE:
	case AM_MOB_JA_CFG_VEHICLE_MIN_TURN_RAD_TYPE:
	case AM_MOB_JA_CFG_VEHICLE_WIDTH_TYPE:
		set.command_type = cfgType;
		if (serialNumber != set.echo_serial_number)
		{
			set.echo_serial_number = serialNumber;
			set.state = NEW_COMMAND;
			set.status = RCS_EXEC;
		}
		break;

	default:
		rcs_print_error("%s: unknown config: %d\n", processName, (int) cfgType);
		break;
	}  // switch (cfgType)
}

/***********************************************************************/

//! readCmd reads in a task command and modifies the status buffer
/*
Returned Value: none

Called By: main

This is checking when the command serial number equals the
stts.echo_serial_number that the command type has not changed.  It
prints a warning if so. It might be better to do something more
drastic, since the command will not be treated as new, even though
processing it will not have started.

*/

static void readCmd()  /* NO ARGUMENTS */
{
	NMLTYPE cmdType;
	int serialNumber;

	cmdType = amMobJACmdBuf->read();
	switch (cmdType)
	{
	case 0:
		// no new command
		break;

	case -1:
		// comm error
		rcs_print_error("%s: NML command read error\n", processName);
		break;

	case AM_MOB_JA_CMD_INIT_TYPE:
	case AM_MOB_JA_CMD_HALT_TYPE:
	case AM_MOB_JA_CMD_ABORT_TYPE:
	case AM_MOB_JA_CMD_SHUTDOWN_TYPE:
	case AM_MOB_JA_CMD_MOVE_WAYPOINT_TYPE:
	case AM_MOB_JA_CMD_NOP_TYPE:
	case AM_MOB_JA_CMD_SPIN_TYPE:
	case AM_MOB_JA_CMD_PREPARE_MOVE_TYPE:
		serialNumber = amMobJACmdBuf->get_address()->serial_number;
		if (serialNumber == stts.echo_serial_number)
		{
			if (stts.command_type != cmdType)
				rcs_print("Warning: same serial number, different command type\n");
		}
		else
		{
			stts.echo_serial_number = serialNumber;
			stts.state = NEW_COMMAND;
			stts.status = RCS_EXEC;
		}
		stts.command_type = cmdType;
		break;

	default:
		rcs_print_error("%s: unknown command: %d\n", processName, (int)cmdType);
		break;
	} // switch (cmdType)
}

/***********************************************************************/

//! readPrimMobStat reads the prim mobility status buffer
/*
Returned Value: none

Called By: main

*/

static void readPrimMobStat()  /* NO ARGUMENTS */
{
	NMLTYPE statType;

	statType = primMobJAStatBuf->read();
	switch (statType)
	{
	case 0:
		// no new status
		break;

	case -1:
		// comm error
		rcs_print_error("%s: NML prim mobility status read error\n",
			processName);
		break;

	case PRIM_MOB_JA_STAT_TYPE:
		// new status message
		break;

	default:
		rcs_print_error("%s: unknown prim mobility status: %d\n",
			processName, (int)statType);
		break;
	}
}

static void readXML()
{
	char hostString[80];

	sprintf( hostString, "http://%s:50000/robocupgui\n", xmlHostName);
	HttpXmlNml hxn(RoboCupGUIState_format, hostString );

	RoboCupGUIState *cf = (RoboCupGUIState *) hxn.readMsg();
	if(cf)
	{
		if(cf->TimeStamp != GPSWaypoint.TimeStamp)
		{
			unprocessedGPSWaypoint = true;
			GPSWaypoint.TimeStamp = cf->TimeStamp;
			GPSWaypoint.xPos = cf->xPos;
			GPSWaypoint.yPos = cf->yPos; 
			/*
			if(cf->LatitudeCardinal=='N')
			{
			GPSWaypoint.LatitudeDegrees = cf->LatitudeDegrees;
			GPSWaypoint.LatitudeMinutes = cf->LatitudeMinutes;
			}
			else if(cf->LatitudeCardinal=='S')
			{
			GPSWaypoint.LatitudeDegrees = -cf->LatitudeDegrees;
			GPSWaypoint.LatitudeMinutes = -cf->LatitudeMinutes;
			}
			GPSWaypoint.LatitudeCardinal = cf->LatitudeCardinal;

			if(cf->LongitudeCardinal=='E')
			{
			GPSWaypoint.LongitudeDegrees = cf->LongitudeDegrees;
			GPSWaypoint.LongitudeMinutes = cf->LongitudeMinutes;
			}
			else if(cf->LongitudeCardinal=='W')
			{
			GPSWaypoint.LongitudeDegrees = -cf->LongitudeDegrees;
			GPSWaypoint.LongitudeMinutes = -cf->LongitudeMinutes;
			}
			GPSWaypoint.LongitudeCardinal = cf->LongitudeCardinal;
			*/
		}
	}
}

/***********************************************************************/

/*! restoreWaypoints puts the original x, y, and neighborhood values back
into waypoints that have had those values changed. */
/*
The original values have been saved in the savedData. In each
Cartesian C point of savedData, p.x and p.y of the waypoint are in x
and y of C, and the neighborhood of the waypoint is in z of C.

*/

static void restoreWaypoints(        /*   ARGUMENTS                          */
							 AmMobJACmdMoveWaypoint * amMoveCmd, /*!< move waypoint cmd being restored   */
							 int savedLength,                    /*!< number of saved waypoint data sets */
							 int * savedIndexes,                 /*!< indexes of waypoints w. saved data */
							 PM_CARTESIAN * savedData)           /*!< saved data sets                    */
{
	int n;

	Waypoint * aWaypoint;
	for (n = 0; n < savedLength; n++)
	{
		aWaypoint = (amMoveCmd->waypoint + savedIndexes[n]);
		aWaypoint->p.x = savedData[n].x;
		aWaypoint->p.y = savedData[n].y;
		aWaypoint->neighborhood = savedData[n].z;
	}
}

/***********************************************************************/

/*! roundCorners rounds the corners of a linearized path, writes the 
results into a prim move command, and creates the arcToWaypointTable
from the incoming lineToWaypointTable. */
/*
Returned Value: int
If any of the following errors occur, this returns 1.
Otherwise, it returns 0.
1. The length argument is less than 1.
2. roundSide fails. 

Called By: planPrimMoveSome

This:
- calls roundSide with a NULL path argument to initialize loop variables
in roundSide
- runs a loop in which roundSide is called repeatedly to round the
corners of the path.

The path does not include the first point (which is at (vehX, vehY)),
so length must be only at least 1. The primMoveCmd that is written
does not include the first point, either.

When this is called, there may be consecutive colinear segments of the
path that join at waypoints, but consecutive colinear segments have been
eliminated elsewhere (by linearizeAStarPath and mergeColinear). The
subordinates of this function all look for zero angle turns.

The tol (tolerance) argument here is used in roundSide and
roundSideAux as the annular tolerance put into the prim move
command. Tol is set here to the minimum of (1) the neighborhood
specified for a particular waypoint in the AM move command, and (2)
twice the side of a cell. Twice the side of a cell is used because (1)
usually the path planned by AM Mob will be at least two cells from the
nearest obstacle (so it is usually safe), and (2) Prim has trouble
doing better than that and tends to miss waypoints and circle back to
try again if the annular tolerance is less than that.

*/

static int roundCorners(             /*   ARGUMENTS                        */
						PrimMobJACmdMoveArcSegment * primMoveCmd, /*!< prim move command to write */
						int * path,                         /*!< linear path                      */
						int length,                         /*!< length of path                   */
						AmMobJACmdMoveWaypoint * amMoveCmd, /*!< AM move command with speed, tol  */
						int * lineToWaypointTable)          /*!< gives waypoint for line          */
{
	int n;             // index for path
	int wayIndex;      // index in amMoveCmd waypoints of waypoint
	double speed;      // desired speed to use
	double tol;        // neighborhood specified for waypoint

	primMoveCmd->arcList_length = 0;
	if (length < 1)
	{
		rcs_print_error("%s: bug: length must be at least 1, but is %d\n",
			processName, length);
		return 1;
	}
	roundSide(NULL, NULL, 0,0,0,0,0);
	wayIndex = lineToWaypointTable[0];
	speed = amMoveCmd->waypoint[wayIndex].speed;
	tol = MIN(amMoveCmd->waypoint[wayIndex].neighborhood, (res * 2));
	if (roundSide(primMoveCmd, path, 0, MAX(vehSpeed, speed), speed,
		tol, wayIndex))
		return 1;
	for (n = 1; n < length; n++)
	{
		wayIndex = lineToWaypointTable[n];
		speed = amMoveCmd->waypoint[wayIndex].speed;
		tol = MIN(amMoveCmd->waypoint[wayIndex].neighborhood, (res * 2));
		if (roundSide(primMoveCmd, path, n, speed, speed, tol, wayIndex))
			return 1;
	}
	return 0;
}

/***********************************************************************/

/*! roundSide generates a smooth path corresponding to one segment of
a linearized path and puts the smooth path into a prim command. The
items inserted in the prim command are either a straight line or an
arc followed by a straight line tangent to the arc. */
/* 
Returned Value int:
If roundSideAux or writeLine returns 1 (an error), this returns 1.
Otherwise, it returns 0.

Called By: roundCorners

It is expected that this function will be called several times in a
row to round a continuous path consisting of straight line
segments. The path is expected to start where the vehicle is. Three
static variables of the function (startX, startY, and heading) are
initialized by calling it with a NULL path argument.

Let S be the input straight line. If S is the first segment of the
path (indicated by the NULL path argument), heading is the heading of
the vehicle and startX and startY are at the location of the vehicle.
Otherwise, heading is the direction of the straight line found by the
previous call to this function, and (startX, startY) is the end point
of that line.

The end of S is the middle of the cell whose id is given by the entry
in the path array with the given index.

If heading is about the same as the direction of S, this just transcribes
S into the prim command. 

If heading differs appreciably from the direction of S, two items are
put into the prim command:
1. an arc starting at the start of S, tangent to the
heading, and ending at the beginning of item 2.
2. a straight line starting tangent to the arc and ending at the
end of S.

We are assuming here that the vehicle is heading the way it is going,
so that the yaw angle is the same as atan2(vehVelocity.y, vehVelocity.x).

This function is designed to be called from a loop using the path
index as the primary loop variable. The static variables in this
function make it unnecessary to pass loop variables back and forth
between this function and the caller.

*/

static int roundSide(         /*   ARGUMENTS                                */
					 PrimMobJACmdMoveArcSegment * primMoveCmd, /*!< prim move command to write  */
					 int * path,                  /*!< linear path                              */
					 int index,                   /*!< index in path of end of line             */
					 double speedArc,             /*!< best speed to use on arc                 */
					 double speedLine,            /*!< speed to use on line                     */
					 double tol,                  /*!< neighborhood specified for waypoint      */
					 int wayIndex)                /*!< index in amMoveCmd waypoints of waypoint */
{
	static double startX;     // X coordinate of start
	static double startY;     // Y coordinate of start
	static double heading;    // heading of vehicle at start
	double endX;              // X coordinate of end
	double endY;              // Y coordinate of end
	double mainDir;           // direction of line from start to end
	double turn;              // amount of turn at start
	double tanX;              // X coordinate of tangent point
	double tanY;              // Y coordinate of tangent point

	if (!path)
	{ // initialize variables
		startX = vehX;
		startY = vehY;
		heading = vehYaw;
	}
	else
	{
		endX = findPathXCoord(path, index);
		endY = findPathYCoord(path, index);
		mainDir = minusPiToPi(atan2((endY - startY), (endX - startX)));
		turn = minusPiToPi(mainDir - heading);
		if (turn > AMOB_TINY_ANGLE)
		{
			if (roundSideAux(primMoveCmd, 1, startX, startY, endX, endY,
				&tanX, &tanY, mainDir, (heading + M_PI_2),
				speedArc, speedLine, tol, wayIndex))
				return 1;
			heading = minusPiToPi(atan2((endY - tanY), (endX - tanX)));
		}
		else if (turn < -AMOB_TINY_ANGLE)
		{
			if (roundSideAux(primMoveCmd, 0, startX, startY, endX, endY,
				&tanX, &tanY, mainDir, (heading - M_PI_2),
				speedArc, speedLine, tol, wayIndex))
				return 1;
			heading = minusPiToPi(atan2((endY - tanY), (endX - tanX)));
		}
		else // if (ABS(turn) < AMOB_TINY_ANGLE)
		{ // pointing at end point already; use one straight line
			if (writeLine(primMoveCmd, endX, endY, speedLine, tol, wayIndex))
				return 1;
			heading = mainDir;
		}
		startX = endX;
		startY = endY;
	}
	return 0;
}

/***********************************************************************/

/*! roundSideAux generates a smooth path corresponding to one segment of
a linearized path and puts the smooth path into a prim command. The
items inserted in the prim command are either a straight line or an
arc followed by a straight line tangent to the arc. */
/* 
Returned Value: int
If writeArc fails, this returns 1.
Otherwise, it returns what the call to writeLine returns (0 or 1).

Called By: roundSide

This function runs a loop in which it tries the arc with the most
desirable radius first. If that radius is too big or using the arc
produces a path that hits an obstacle, the radius is cut in half and
the loop repeats. The process stops when either (1) the radius is not
too big and no obstacle is hit, or (2) the radius is so small that the
smoothed path is essentially identical to the line segment it is
replacing.

The speed to use for finding the most desirable value for the radius
of the arc is given in the speedArc argument. If the minimum radius at
speedArc is less than the resolution, then that radius is set to the
resolution, and then the maximum speed for the radius will be larger
than speedArc. Since we do not want to exceed speedArc on an arc,
"MIN(speedArc, speedArcMax)" is used for the speed in the call to writeArc.

If the most desirable value for the radius based on speedArc is less
than the minimum turning radius of the vehicle, then the most
desirable radius is reset to the minimum turning radius of the vehicle.
If the most desirable radius is too big, then the radius is made
smaller than that. Thus, a plan may be produced that cannot be
executed. This is bad and should be fixed. The alternative would be to
produce no plan or figure out how to include backing up and making
three-point turns.

A radius is too big if the end of the straight line is inside the
circle from which an arc is chosen.

*/

static int roundSideAux(     /*   ARGUMENTS                                */
						PrimMobJACmdMoveArcSegment * primMoveCmd, /*!< prim move command to write */
						int ccw,                    /*!< 1=counterclockwise, 0=clockwise          */
						double startX,              /*!< X coordinate of start                    */
						double startY,              /*!< Y coordinate of start                    */
						double endX,                /*!< X coordinate of end                      */
						double endY,                /*!< Y coordinate of end                      */
						double * tanX,              /*!< X coordinate of tangent point, set here  */
						double * tanY,              /*!< Y coordinate of tangent point, set here  */
						double mainDir,             /*!< direction of line from start to end      */
						double centerDir,           /*!< direction of line from start to center   */
						double speedArc,            /*!< best speed to use on arc                 */
						double speedLine,           /*!< speed to use on line                     */
						double tol,                 /*!< neighborhood specified for waypoint      */
						int wayIndex)               /*!< index in amMoveCmd waypoints of waypoint */
{
	double centerX;     // X coordinate of center of circle
	double centerY;     // Y coordinate of center of circle
	double radius;      // radius of arc
	double tanDir;      // direction of line from center to tangent point
	double d;           // distance from center of circle to end point
	double speedArcMax; // maximum speed on arc

	for (radius = MAX(calcMinRadius(speedArc), res),
		radius = MAX(radius, set.vehicleMinTurnRad); ; radius = (radius / 2))
	{
		centerX = (startX + (radius * cos(centerDir)));
		centerY = (startY + (radius * sin(centerDir)));
		d = hypot((endX - centerX), (endY - centerY));
		if (d <= radius)
			continue;
		tanDir = (atan2((endY - centerY), (endX - centerX)) -
			(ccw ? acos(radius / d) : -acos(radius / d)));
		*tanX = (centerX + (radius * cos(tanDir)));
		*tanY = (centerY + (radius * sin(tanDir)));
		if ((!checkObsLineThick(*tanX, *tanY, endX, endY)) &&
			(!checkObsArc(ccw, centerX, centerY, startX, startY, *tanX, *tanY)))
			break; // no obstacle on path
		if (radius < (res /2 ))
			break; // minimum radius limit reached
	}
	speedArcMax = calcMaxSpeed(radius);
	if (writeArc(primMoveCmd, ccw, centerX, centerY, startX, startY,
		*tanX, *tanY, MIN(speedArc, speedArcMax), tol, wayIndex))
		return 1;
	return writeLine(primMoveCmd, endX, endY, speedLine, tol, wayIndex);
}

/***********************************************************************/

/*! setEvenSerial sets a serial number to the smallest even number larger
than the echoed number and the incoming value of the serial number */
/*

Returned Value: none

Called By: writePrimCmd

Even numbers larger than the echoed number are used as serial numbers
to avoid reusing serial numbers when another process (such as primShell)
is also sending commands to Prim Mob. It is expected that the other
process will use odd numbers larger than the echoed number.

*/

static void setEvenSerial( /*   ARGUMENTS            */
						  int echoed,               /*!< echoed serial number */
						  int * serial)             /*!< serial number to set */
{
	*serial = ((echoed <= *serial) ?
		(*serial + ((*serial % 2) ? 1 : 2)) :
		(echoed  + ((echoed  % 2) ? 1 : 2)));
}

/***********************************************************************/

//! setObsCells sets cells of the obstacle map around an obstacle cell.
/*
Returned Value: none

Called By: updateObsMap

This is not called with obsVal = AMSP_NOT_OBSTACLE

This marks all cells within obsGrow of the given cell.  The rules are:
if a cell to be marked already has a value equal to or larger than
obsVal, the cell is not changed, but if the value is less than obsVal,
the value is changed to obsVal.
The marks in increasing order are:
AMSP_UNSEEN
AMSP_MAYBE_OBSTACLE
AMSP_OBSTACLE

In addition, if obsVal is AMSP_MAYBE_OBSTACLE or AMSP_OBSTACLE, this
adds two annular rings to the outside of the grown obstacle, the inner
of which has value AMSP_MAYBE_OBSTACLE, and the outer of which has
value AMSP_NEAR_OBSTACLE. The rule for the inner ring is: if a cell to
be marked already has a value equal to or larger than
AMSP_MAYBE_OBSTACLE, the cell is not changed, but if the value is less
than AMSP_MAYBE_OBSTACLE, the value is changed to AMSP_MAYBE_OBSTACLE.
The rule for the outer ring is similar.

The rings are formed as follows.
a. To each side of each row of the grown obstacle, add two cells with the
nearer one marked AMSP_MAYBE_OBSTACLE and the farther one marked
AMSP_NEAR_OBSTACLE.
b. Two rows above and below the grown obstacle, add a row of cells
the same length as the top (or bottom) row of the grown obstacle.
Mark all those cells AMSP_NEAR_OBSTACLE.
c. One row above and below the grown obstacle, add a row of cells two
cells longer than the length of the top (or bottom) row of the grown
obstacle. Mark all those cells AMSP_MAYBE_OBSTACLE. Add a cell at each
end of the row marked AMSP_NEAR_OBSTACLE.

This method of forming rings makes slightly smaller rings than
might be formed by other methods.

This works by marking every cell in a square (with rounded corners so
that it is close to a circle) around the given cell as an
obstacle. The length of a side of the square is ((2 * obsGrow) + 1).

The numbers in the skips array tell how many places not to mark at
the end of each row of the square in order to round the corners of the
square.

*********************************************************************

DISCUSSION

1. Two factors tend to make obstacles grow too much.  First, it is
assumed a cell marked as obstacle is entirely full of obstacle. Second
in this function, if any part of a cell C is within half a vehicle
radius of an obstacle cell, C is marked as obstacle. At one extreme,
this leaves no extra clearance. At the other extreme it leaves
(2 * (cell width) * sqrt(2)) extra clearance, i.e. two cell diagonals.
On the average, there is probably a little over one cell cell width
extra. This excessive growth is good because it tends to prevent
collisions with obstacles when the vehicle gets off the intended
path. The excessive growth is bad because it prevents the vehicle from
going through tight spaces where it could fit.

Previously, skips was calculated based on rounding half the
vehicle width up to the nearest integer number of cell widths.
This was a third factor tending to make obstacles grow too much.
skips is now calculated based on the actual vehicle width.

An interesting effect of this type of obstacle growing is that the
marking of cells near the edge of the map may change when the map
moves, even though the un-grown obstacles in those areas have not
changed. For example, if there is a single obstacle cell on the edge
of the map, when it is grown it will appear to be roughly a semicircle
of radius obsGrow on the edge of the map. When the map moves so that
the obstacle cell is just one cell off the map, the entire semicircle
disappears. This effect makes the boundaries of grown obstacles near
the edges of the map tend to curve to be perpendicular to the edge.

*/

static void setObsCells( /*   ARGUMENTS                         */
						int x,                  /*!< X index on obsMap of cell to grow */
						int y,                  /*!< Y index on obsMap of cell to grow */
						double obsVal)          /*!< value to insert                   */
{
	int i;
	int j;
	int yy;
	int grow;
	int stop;
	int length;
	int skip;

	if (obsVal == AMSP_UNSEEN)
	{ // this does not add two extra rings
		length = ((2 * obsGrow) + 1);
		x = (x - obsGrow);
		y = (y - obsGrow);
		for(i = 0; x < 0; i++, x++); // skip cells below the bottom of the map
		for ( ; ((i < length) && (x < obsMapSide)); i++, x++)
		{ // process a row
			skip = skips[i];
			stop = (length - skip);
			for (j = skip; j < stop; j++)
			{ // skip cells to the left of the map
				if ((y + j) > -1)
					break;
			}
			for (; j < stop; j++)
			{
				yy = (y + j);
				if (yy > obsMapSi)
					break; // skip cells to the right of the map
				else if (obsMap[x][yy] < AMSP_UNSEEN)
					obsMap[x][yy] = AMSP_UNSEEN;
			}
		}
	}
	else // if ((obsVal == AMSP_OBSTACLE) || (obsVal == AMSP_MAYBE_OBSTACLE))
	{ // this adds two extra rings
		grow = (obsGrow + 2);
		length = ((2 * grow) + 1);
		x = (x - grow);
		y = (y - grow);
		for(i = 0; x < 0; i++, x++); // skip cells below the bottom of the map
		if (i == 0)
		{ //Do bottom row. It is all AMSP_NEAR_OBSTACLE
			skip = (skips[0] + 2);
			j = skip;
			stop = (length - skip);
			for (j = skip; j < stop; j++)
			{ // skip cells to the left of the map
				if ((y + j) > -1)
					break;
			}
			for (; j < stop; j++)
			{
				yy = (y + j);
				if (yy > obsMapSi)
					break; // skip cells to the right of the map
				else if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
					obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
			}
			i++;
			x++;
		}
		if (i == 1)
		{ // Do second row from bottom. All AMSP_MAYBE_OBSTACLE, except ends
			skip = (skips[0] + 1);
			j = skip;
			yy = (y + j);
			if (yy > -1) // skip cells to the left of the map
			{ // left of outer ring
				if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
					obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
			}
			yy++;
			stop = (length - (skip + 1));
			for (j = (skip + 1); j < stop; j++, yy++)
			{ // skip cells to the left of the map
				if (yy > -1)
					break;
			}
			for (; j < stop; j++, yy++)
			{
				if (yy > obsMapSi)
					break; // skip cells to the right of the map
				else if (obsMap[x][yy] < obsVal)
					obsMap[x][yy] = AMSP_MAYBE_OBSTACLE;
			}
			if (yy < obsMapSide) // skip cells to the right of the map
			{ // right of outer ring
				if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
					obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
			}
			i++;
			x++;
		}
		for ( ; ((i < (length - 2)) && (x < obsMapSide)); i++, x++)
		{ // process rest of rows except last two
			skip = skips[i-2];
			j = skip;
			yy = (y + j);
			if (yy > -1) // skip cells to the left of the map
			{ // left of outer ring
				if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
					obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
			}
			yy++;
			if (yy > -1) // skip cells to the left of the map
			{ // left of inner ring
				if (obsMap[x][yy] < AMSP_MAYBE_OBSTACLE)
					obsMap[x][yy] = AMSP_MAYBE_OBSTACLE;
			}
			yy++;
			stop = (length - (skip + 2));
			for (j = (skip + 2); j < stop; j++, yy++)
			{ // skip cells to the left of the map
				if (yy > -1)
					break;
			}
			for (; j < stop; j++, yy++)
			{
				if (yy > obsMapSi)
					break; // skip cells to the right of the map
				else if (obsMap[x][yy] < obsVal)
					obsMap[x][yy] = obsVal;
			}
			if (yy < obsMapSide) // skip cells to the right of the map
			{ //right of inner ring
				if (obsMap[x][yy] < AMSP_MAYBE_OBSTACLE)
					obsMap[x][yy] = AMSP_MAYBE_OBSTACLE;
				yy++;
				if (yy < obsMapSide) // skip cells to the right of the map
				{ // right of outer ring
					if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
						obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
				}
			}
		}
		if (x < obsMapSide)
		{ // Do second row from top. All AMSP_MAYBE_OBSTACLE, except ends
			skip = (skips[0] + 1);
			j = skip;
			yy = (y + j);
			if (yy > -1) // skip cells to the left of the map
			{ // left of outer ring
				if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
					obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
			}
			yy++;
			stop = (length - (skip + 1));
			for (j = (skip + 1); j < stop; j++, yy++)
			{ // skip cells to the left of the map
				if (yy > -1)
					break;
			}
			for (; j < stop; j++, yy++)
			{
				if (yy > obsMapSi)
					break; // skip cells to the right of the map
				else if (obsMap[x][yy] < obsVal)
					obsMap[x][yy] = AMSP_MAYBE_OBSTACLE;
			}
			if (yy < obsMapSide) // skip cells to the right of the map
			{ // right of outer ring
				if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
					obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
			}
			i++;
			x++;
		}
		if (x < obsMapSide)
		{ //Do top row. It is all AMSP_NEAR_OBSTACLE
			skip = (skips[0] + 2);
			j = skip;
			stop = (length - skip);
			for (j = skip; j < stop; j++)
			{ // skip cells to the left of the map
				if ((y + j) > -1)
					break;
			}
			for (; j < stop; j++)
			{
				yy = (y + j);
				if (yy > obsMapSi)
					break; // skip cells to the right of the map
				else if (obsMap[x][yy] < AMSP_NEAR_OBSTACLE)
					obsMap[x][yy] = AMSP_NEAR_OBSTACLE;
			}
		}
	}
}

/***********************************************************************/

/*! setRelatedValues sets parameters whose values are interdependent,
based on the obsMapSideLength and cell resolution. */
/*
Returned Value: int
If any of the following errors occurs, this returns 1.
Otherwise, it returns 0.
1. The cellRes is not positive.
2. The cellRes or does not divide 1 meter almost evenly.
3. The number of cells on a side of the map is more than MAX_MAP_SIDE.

Called By:
iniLoad
initParameters

This sets set.cellResolution, set.planHorizon, and the following global
variables:

cellsPerMeter
diagDist
extraDist
obsGrow (by calling calcSkips)
obsMapSide
obsMapSide
res
skips
straightDist

The cell resolution must be such that there are an integer number of
cells in a meter. The expression resetting res recalculates the
resolution so that if the given resolution is a little inexact
(e.g. 0.3333) it gets made as exact as possible using a double
(e.g. 0.333333333333333).

It would make more sense to have cellsPerMeter rather than cell
resolution be specified in the iniFile or arguments, because
cellsPerMeter must be an integer.  Since cell resolution is a double
but must equal 1/cellsPerMeter, it is easy to give an illegal value
for cell resolution.

*/

static int setRelatedValues( /*   ARGUMENTS                */
							double cellRes)             /*!< proposed cell resolution */
{
	if (cellRes <= 0)
	{
		rcs_print_error("%s: cell resolution must not be negative, but is %f\n",
			processName, cellRes);
		return 1;
	}
	cellsPerMeter = (int)((1.0 / cellRes) + 0.1);
	res = (1.0 / cellsPerMeter);
	if (fabs(cellRes - res) > 0.001)
	{
		rcs_print_error("%s: cell resolution must divide 1 evenly, but is %f\n",
			processName, cellRes);
		return 1;
	}
	set.cellResolution = res;
	if (verbose > 1)
		rcs_print("cellResolution set to %f\n", res);
	set.planHorizon = (obsMapSideLength / 2.0);
	obsMapSide = (obsMapSideLength * cellsPerMeter);
	obsMapSi = (obsMapSide - 1);
	if (obsMapSide > MAX_MAP_SIDE)
	{
		rcs_print_error("%s: number of cells on side of obstacle map (%d) "
			"more than maximum allowed (%d)\n",
			processName, obsMapSide, MAX_MAP_SIDE);
		return 1;
	}
	calcSkips();
	straightDist = (res / 2.0);
	diagDist = (straightDist * sqrt(2.0));
	extraDist = (res * (sqrt(2.0) - 1));
	return 0;
}

/***********************************************************************/

//! updateObsMap updates the obstacle map
/*
Returned Value: int
If there is no new data, this returns -1.
If there is an error, this returns 1,
If there is new data and no error, this returns 0.
An error occurs if:
1. A communications error occurs while trying to read the data buffer.
2. The cells in the data being read are not square.
3. The cells in the data being read are not the same size as the cells
of the obstacle map.
4. A side of the data rectangle is smaller than the corresponding
side of the obstacle map.

Called By:
planAll
planPrimMove

If there is new data and no error, a new obstacle map will be read into
the obsDataBuf buffer and the obsMap will be updated.

The sensor data map may be larger than the obsMap, but not smaller.
Any extra data is ignored.

The obstacle map is growing things by obsGrow, which is set elsewhere
to half the width of the vehicle plus the allowed path error.

Three types of cell are grown: AMSP_UNSEEN, AMSP_MAYBE_OBSTACLE, and
AMSP_OBSTACLE. The most severe of these takes precedence. For example,
if a cell has already been marked AMSP_OBSTACLE and it could be
marked AMSP_UNSEEN, it stays AMSP_OBSTACLE. But if it has already been
marked AMSP_UNSEEN and it could be marked AMSP_OBSTACLE, it is so marked.

For each item of obsData, if the obsData cell represented by that item
is marked anything but AMSP_NOT_OBSTACLE, all the cells of obsMap
within obsGrow of that cell could be marked that way.

*/

static int updateObsMap() /* NO ARGUMENTS */
{
	int dataType;
	int retval;    // value to return
	int i;         // row index
	int j;         // column index
	double obsVal; // obstacle value (AMSP_NOT_OBSTACLE, etc.)

	dataType = (int)sensorData2DBuf->read();
	switch(dataType)
	{
	case 0: // no new data
		retval = -1;
		break;
	case -1:
		// comm error
		rcs_print_error("%s: NML command read error\n", processName);
		retval = 1;
		break;
	case SENSOR_DATA_2D_TYPE:
		obsData = (SensorData2D *)sensorData2DBuf->get_address();
		// TODO: Remove debug statement:
		writeObsMapData("./ObsMapDataDump.txt");

		if (obsData->incrementX != obsData->incrementY)
		{
			rcs_print_error("%s: sensor data map cells not square\n",
				processName);
			retval = 1;
		}
		else if ((obsData->incrementX > (1.000001 * res)) ||
			(obsData->incrementX < (0.999999 * res)))
		{
			rcs_print_error("%s: sensor data map and obstacle map "
				"resolutions differ\n", processName);
			retval = 1;
		}
		else if (obsData->x < obsMapSide)
		{
			rcs_print_error("%s: number of cells on X side is less for sensor "
				"data map (%d) than for obstacle map (%d)\n",
				processName, obsData->x, obsMapSide);
			retval = 1;
		}
		else if (obsData->y < obsMapSide)
		{
			rcs_print_error("%s: number of cells on Y side is less for sensor "
				"data map (%d) than for obstacle map (%d)\n",
				processName, obsData->y, obsMapSide);
			retval = 1;
		}
		else
		{ // all OK so far
			for (i = 0; i < obsMapSide; i++) // clear map
			{
				for (j = 0; j < obsMapSide; j++)
				{
					obsMap[i][j] = AMSP_NOT_OBSTACLE;
				}     
			}
			for (i = 0; i < obsMapSide; i++) // put obstacle data on map
			{
				for (j = 0; j < obsMapSide; j++)
				{
					obsVal = obsData->featArray[(i * obsData->y)+j];
					if (obsVal != AMSP_NOT_OBSTACLE)
						setObsCells(i, j, obsVal);
				}
			}
			xMin = (obsData->lowerLeft.x);
			xMax = (xMin + (obsMapSide * res));
			yMin = (obsData->lowerLeft.y);
			yMax = (yMin + (obsMapSide * res));
			retval = 0;
		}
		break;
	default: // bad type
		rcs_print_error("%s: bad obstacle data type: %d\n",
			processName, dataType);
		retval = 1;
		break;
	}
	return retval;
}

/***********************************************************************/

//! updatePosition sets the vehicle position from nav data
/*
Returned Value: int
If there is no new data, this returns -1.
If there is an error, this returns 1,
If there is new data and no error, this returns 0.

Called By:
doCmdPrepareMove
doCmdMoveWaypoint

This tries up to get new location data.

If there is new data and no error, this sets several global variables
pertaining to the vehicle to the values given by the data.  Local
coordinates are used for this data.

If there is an error, this prints an error message.

The vehicle XY position is being transformed (using the set.controlPoint
transform) from the position of the location point reported in navData
to the position of the controlled point. For the unitloader (AGV),
which has steerable rear wheels, the controlled point is the midpoint
of the front axle (if there were an axle). For vehicles with no
steerable wheels, the controlled point is usually the same as the
location point.

The vehicle yaw, velocity, and speed are the yaw, velocity, and speed
of the location point, not of the controlled point.

*/

static int updatePosition() /* NO ARGUMENTS */
{
	int dataType;
	NavDataExt * navData;
	int retVal;
	PM_POSE pose;

	dataType = (int)navDataExtBuf->read();
	switch(dataType)
	{
	case 0: // no new data
		retVal = -1;
		break;
	case -1: // comm error
		rcs_print_error("%s: NML command read error\n", processName);
		retVal = 1;
		break;
	case NAV_DATA_EXT_TYPE: // got position
		navData = (NavDataExt *)navDataExtBuf->get_address();
		pose.tran = navData->tranRel;
		pose.rot = navData->rpyRel;
		pose = pose * set.controlPoint;
		vehX = pose.tran.x;
		vehY = pose.tran.y;
		vehYaw = minusPiToPi(navData->rpyRel.y);
		vehSpin = navData->rpyRelRate.y;
		vehVelocity.x = navData->tranRelRate.x;
		vehVelocity.y = navData->tranRelRate.y;
		vehSpeed = hypot(vehVelocity.x, vehVelocity.y);
		localToGlobal = inv(navData->absToRel);
		retVal = 0;

		// Store GPS information
		/*if(navData->Fix==1)
		{
		currentGPS.LatitudeDegrees = navData->LatitudeDegrees;
		currentGPS.LatitudeMinutes = navData->LatitudeMinutes;
		currentGPS.LongitudeDegrees = navData->LongitudeDegrees;
		currentGPS.LongitudeMinutes = navData->LongitudeMinutes;
		//currentGPS.Fix = navData->Fix;
		//currentGPS.Satellites = navData->Satellites;
		}

		if(!gotInitialPose)
		{
		initX = vehX;
		initY = vehY;

		// Store GPS information
		initGPS.LatitudeDegrees = currentGPS.LatitudeDegrees;
		initGPS.LatitudeMinutes = currentGPS.LatitudeMinutes;
		initGPS.LongitudeDegrees = currentGPS.LongitudeDegrees;
		initGPS.LongitudeMinutes = currentGPS.LongitudeMinutes;
		//initGPS.Fix = currentGPS.Fix;
		//initGPS.Satellites = currentGPS.Satellites;

		gotInitialPose = true;

		printf("\n\n\n\n\n!!=======================================\n");
		printf("Initial GPS Information:\n");
		printf("!!=======================================\n");
		printf("Latitude Degrees:%i\nLatitude Minutes:%f\n",initGPS.LatitudeDegrees,
		initGPS.LatitudeMinutes);
		printf("Longitude Degrees:%i\nLongitude Minutes:%f\n",initGPS.LongitudeDegrees,
		initGPS.LongitudeMinutes);
		printf("\n\n!!=======================================\n");
		printf("Initial CARTESIAN Information:\n");
		printf("!!=======================================\n");
		printf("Initial Coordinates:%lf,%lf\n\n\n",initX,
		initY);

		}*/
		break;
	default: // bad type
		rcs_print_error("%s: bad nav data type: %d\n", processName, dataType);
		retVal = 1;
		break;
	}
	return retVal;
}

/***********************************************************************/

//! vehicleOffObsMap determines if the vehicle is off the obstacle map
/*
Returned Value: int
If the vehicle is off the map, this returns 1. Otherwise, it returns 0.

Called By:
planAll
planPrimMove

This is assuming obsData (a global variable pointing at the most
recent obstacle data in the sensorData2DBuf) is not NULL. That is
checked in doCmdMoveWaypoint.

The obsData->lowerLeft point is at the lower left corner of the
obstacle map and at the lower left corner of the first cell.

*/

static int vehicleOffObsMap() /* NO ARGUMENTS */
{
	double extent;

	extent = (obsMapSide * res);
	if ((vehX <= (obsData->lowerLeft.x)) ||
		(vehX >= (obsData->lowerLeft.x + extent)) ||
		(vehY <= (obsData->lowerLeft.y)) ||
		(vehY >= (obsData->lowerLeft.y + extent)))
	{
		rcs_print_error("%s: vehicle is not on the obstacle map\n", processName);
		return 1;
	}
	return 0;
}

/***********************************************************************/

//! writeArc writes a circular arc move in a prim move command.
/*
Returned Value: int
If there are more than MAX_ARC_SEGMENTS, this returns 1.
Otherwise, it returns 0.

Called By: roundSideAux

*/

static int writeArc(        /*   ARGUMENTS                               */
					PrimMobJACmdMoveArcSegment * primMoveCmd, /*!< move command to write in */
					int ccw,                   /*!< non-zero=arc is counterclockwise        */
					double centerX,            /*!< X coordinate of arc center              */
					double centerY,            /*!< Y coordinate of arc center              */
					double startX,             /*!< X coordinate of start point             */
					double startY,             /*!< Y coordinate of start point             */
					double endX,               /*!< X coordinate of end point               */
					double endY,               /*!< Y coordinate of end point               */
					double speed,              /*!< speed to put into primMoveCmd           */
					double tol,                /*!< tolerance to put into primMoveCmd       */
					int wayIndex)              /*!< index in waypoints for this arc         */
{
	ArcSegment * anArc;
	int length;
	double theta;
	double maxSpeed;

	length = primMoveCmd->arcList_length;
	if (length >= MAX_ARC_SEGMENTS)
	{
		rcs_print_error("%s: too many arcs for command\n", processName);
		return 1;
	}
	theta = (atan2((endY - centerY), (endX - centerX)) - 
		atan2((startY - centerY), (startX - centerX)));
	if (ccw && (theta < 0))
		theta = (theta + (2 * M_PI));
	else if ((!ccw) && (theta > 0))
		theta = (theta - (2 * M_PI));
	anArc = (primMoveCmd->arcList + length);
	maxSpeed = calcMaxSpeed(hypot((startX - centerX), (startY - centerY)));
	anArc->isArc = 1;
	anArc->center.x = centerX;
	anArc->center.y = centerY;
	anArc->center.z = 0.0;
	anArc->end.x = endX;
	anArc->end.y = endY;
	anArc->end.z = 0.0;
	anArc->theta = theta;
	anArc->speed = MIN(speed, maxSpeed);
	anArc->annular_tol = tol;
	arcToWaypointTable[length] = wayIndex;
	primMoveCmd->arcList_length++;
	return 0;
}

/***********************************************************************/

/*! writeLine writes a straight line move in a prim move command. It
also adds one to the arcList_length in the command */
/*
Returned Value: int
If the arcList_length is more than MAX_ARC_SEGMENTS, this returns 1.
Otherwise, it returns 0.

Called By:
roundSide
roundSideAux

*/

static int writeLine(       /*   ARGUMENTS                               */
					 PrimMobJACmdMoveArcSegment * primMoveCmd, /*!< move command to write in */
					 double endX,               /*!< X coordinate of end point               */
					 double endY,               /*!< Y coordinate of end point               */
					 double speed,              /*!< speed to put into primMoveCmd           */
					 double tol,                /*!< tolerance to put into primMoveCmd       */
					 int wayIndex)              /*!< index in waypoints for this line        */
{
	ArcSegment * anArc;
	int length;

	length = primMoveCmd->arcList_length;
	if (length >= MAX_ARC_SEGMENTS)
	{
		rcs_print_error("%s: too many arcs for command\n", processName);
		return 1;
	}
	anArc = (primMoveCmd->arcList + length);
	anArc->isArc = 0;
	anArc->end.x = endX;
	anArc->end.y = endY;
	anArc->end.z = 0.0;
	anArc->speed = speed;
	anArc->annular_tol = tol;
	arcToWaypointTable[length] = wayIndex;
	primMoveCmd->arcList_length++;
	return 0;
}

/***********************************************************************/

/*! writeObsMapData writes a file representation of the map centered on
the vehicle and having a side length of outSideElts. This is the same
area as built by drawObsMapFile. After the header, the file contains
one line for each cell of the map. Each such line has the probability
in the cell. */
/*
Returned Value: none

Called By: drawObsMapFile

This is writing into the "fileName" argument.

*/

static void writeObsMapData( /*   ARGUMENTS             */
							char * fileName)            /*!< name of file to write */
{
	int i;   // X index for output square
	int j;   // Y index for output square
	FILE * outFile;

	outFile = fopen(fileName, "w");
	fprintf(outFile, "resolution = %.4f\n", res);
	fprintf(outFile, "number of cells on a side = %d\n", obsMapSide);
	fprintf(outFile, "total number of cells = %d\n", (obsMapSide * obsMapSide));
	fprintf(outFile, "lowerLeft.x = %f\n", obsData->lowerLeft.x);
	fprintf(outFile, "lowerLeft.y = %f\n", obsData->lowerLeft.y);
	for (i = 0; i < obsMapSide; i++)
	{
		for (j = 0; j < obsMapSide; j++)
		{
			fprintf(outFile, "%.4f\n", obsMap[i][j]);
		}
	}
	fclose(outFile);
}

/***********************************************************************/

//! writePrimCmd writes a prim command.
/*
Returned Value: none

Called By:
doCmdAbort
doCmdHalt
doCmdInit
doCmdMoveWaypoint
doCmdPrepareMove
doCmdShutdown
doCmdSpin

*/

static void writePrimCmd( /*   ARGUMENTS            */
						 RCS_CMD_MSG * cmd,       /*!< command to write     */
						 int echoed)              /*!< echoed serial number */
{
	setEvenSerial(echoed, &primCmdNumber);
	cmd->serial_number = primCmdNumber;
	primMobJACmdBuf->write(cmd);
}

static void writeMobCmd( /*   ARGUMENTS         */
						RCS_CMD_MSG * cmd)      /*!< command to write  */
{
	static int serial = 1; // mob command serial number

	setOddSerial(amMobJAStat->echo_serial_number, &serial);
	cmd->serial_number = serial;
	amMobJACmdBuf->write(cmd);
}

static void writeSPCmd( /*   ARGUMENTS        */
					   RCS_CMD_MSG * cmd)     /*!< command to write */
{
	static int serial = 1; // SP command serial number

	setOddSerial(amSPStat->echo_serial_number, &serial);
	cmd->serial_number = serial;
	amSPCmdBuf->write(cmd);
}

static void writeServoMobCmd(RCS_CMD_MSG* cmd)
{
	static int serial = 1; // Servo Mob Cmd serial number

	//setOddSerial(servoMobJAStat->echo_serial_number, &serial);

	cmd->serial_number = serial;
	serial+=2;
	servoMobJACmdBuf->write(cmd);
}

static void writeServoSPCmd(RCS_CMD_MSG* cmd)
{
	static int serial = 1; // ServoSPCmd serial number

	cmd->serial_number = serial;
	serial += 2;
	servoSPCmdBuf->write(cmd);
}

static void setOddSerial( /*   ARGUMENTS            */
						 int echoed,              /*!< echoed serial number */
						 int * serial)            /*!< serial number to set */
{
	*serial = ((echoed <= *serial) ?
		(*serial + ((*serial % 2) ? 2 : 1)) :
		(echoed  + ((echoed  % 2) ? 2 : 1)));
}


static int initServoSPBuffers(int bufNum)
{
	char bufName[MOAST_NML_BUFFER_NAME_LEN]; 

	/// ServoSPCmd Channel
	sprintf(bufName,"%s%d",SERVO_SP_CMD_NAME,bufNum);
	servoSPCmdBuf = new RCS_CMD_CHANNEL(servoSP_format, bufName, 
		THIS_PROCESS, nmlFile);

	if (servoSPCmdBuf == NULL || !servoSPCmdBuf->valid()) 
	{
		delete servoSPCmdBuf;
		servoSPCmdBuf = NULL;
		return 1;
	}
	bzero(bufName, MOAST_NML_BUFFER_NAME_LEN);

	/// ServoSPStat Channel
	sprintf(bufName,"%s%d",SERVO_SP_STAT_NAME,bufNum);
	servoSPStatBuf = new RCS_STAT_CHANNEL(servoSP_format, bufName, 
		THIS_PROCESS, nmlFile);

	if (servoSPStatBuf == NULL || !servoSPStatBuf->valid()) 
	{
		delete servoSPStatBuf;
		servoSPStatBuf = NULL;
		return 1;
	}
	bzero(bufName, MOAST_NML_BUFFER_NAME_LEN);

	/// ServoSPSet Channel
	sprintf(bufName,"%s%d",SERVO_SP_SET_NAME,bufNum);
	servoSPSetBuf = new RCS_STAT_CHANNEL(servoSP_format, bufName, 
		THIS_PROCESS, nmlFile);

	if (servoSPSetBuf == NULL || !servoSPSetBuf->valid()) 
	{
		delete servoSPSetBuf;
		servoSPSetBuf = NULL;
		return 1;
	}

	printf("amMob: ServoSP Buffers opened successfully.\n");
	return 0;
}


/***********************************************************************/
/***********************************************************************/

/*
Modification history:

$Log: amMobMain.cc,v $
Revision 1.59  2009/04/27 16:09:00  dr_steveb
Made init return if the system is already in an initialized state

Revision 1.58  2009/03/04 18:19:06  wshackle
fix only init's once problem

Revision 1.57  2008/02/12 00:15:59  tomrkramer
Removed unused plans array. Added iniLoadControlPoint. Fixed bug in
calcSkips which was causing odd shapes at corners of grown obstacles.
Added setting of set.controlPoint. Changed method of relocating
control point from a hard-coded offset of 0.3 to using the offset
data in the inifile.

Revision 1.56  2008/02/01 20:15:57  dr_steveb
Added transform to make AM send information to vehicle in global coordinates

Revision 1.55  2008/01/09 16:32:50  tomrkramer
Moved controlled point to 0.3 meter in front of the navData point to work
with AGV the same way as Prim Mobility.

Revision 1.54  2008/01/05 15:26:40  tomrkramer
Changed computation of obstacle growing to use maximum of length and
width of vehicle. Previously, only width was used.

Revision 1.53  2007/12/19 19:30:25  tomrkramer
Changed method of getting MAX_TRAN_VEL to work with new moast.ini

Revision 1.52  2007/12/14 23:36:31  tomrkramer
1. Added getting minimum turning radius from Servo
2. Added using minimum turning radius in path planning.
3. Added executing configuration commands for planHorizon and
minimum turning radius.
4. Added finding and using points near waypoints on obstacle when
executing a moveWaypoint command
5. Made it OK if planning to the edge of the planning horizon from
the last waypoint inside the horizon fails.
6. Improved documentation.

Revision 1.51  2007/12/08 13:34:08  tomrkramer
Changed code so AM Mob does not block while waiting for nav data.
Improved documentation.

Revision 1.50  2007/12/03 21:20:40  tomrkramer
Removed debugging message.

Revision 1.49  2007/12/01 17:42:28  tomrkramer
Changed method of getting vehicle width from using ini file (old) to
using servo mob settings (new).

Revision 1.48  2007/07/30 20:08:43  dr_steveb
Made am wait longer for navData

Revision 1.47  2007/04/25 15:28:48  dr_steveb
Backed out accidental mods.

Revision 1.45  2007/02/06 22:48:05  tomrkramer
Changed where drawing the obstacle map is called so that the map is
drawn every time an attempt to plan to move is made in response to
a moveWaypoint command. Also made a lot of minor fixes to documentation
and moved the findBestGoal function to be in alphabetical order.

Revision 1.44  2007/02/06 13:50:27  tomrkramer
Changed planning to horizon so it chooses the least cost cell in each
meter of the periphery.

Revision 1.43  2007/01/26 17:19:57  tomrkramer
Improved sending of commands to Prim.

Revision 1.42  2007/01/25 18:47:55  tomrkramer
Changed handling of prim command numbers.

Revision 1.41  2007/01/25 16:19:29  tomrkramer
Revised handling of position update.

Revision 1.40  2007/01/24 22:15:03  tomrkramer
Implemented more informative error codes.

Revision 1.39  2007/01/24 16:26:04  tomrkramer
Updated getting data from moast.ini and added more AM_MOB_JA_ERROR types.

Revision 1.38  2007/01/05 15:38:10  tomrkramer
Implemented corner rounding.

Revision 1.37  2006/12/01 16:11:52  tomrkramer
Removed boxing of obstacles, which was causing planning to fail
unnecessarily. Also removed the "corner" argument in dozens of
places in other code, since it was needed only for boxing.

Revision 1.35  2006/11/16 16:16:47  tomrkramer
Changed setting of stts.seed_length by adding 1 (to include the
array element in which the current location is set).

Revision 1.34  2006/11/16 13:49:04  tomrkramer
Added setting stts.seed_length when planning to points on the planning horizon.

Revision 1.33  2006/11/13 14:09:47  tomrkramer
Revised documentation of calcSkips (the code for which is mysterious).

Revision 1.32  2006/11/10 18:49:18  tomrkramer
Reformatted slightly. Remove comment about fix to calcSkips being temporary.

Revision 1.30  2006/10/30 16:29:42  tomrkramer
Changed planning to points on periphery to use Dijkstra search.

Revision 1.29  2006/10/26 20:04:18  tomrkramer
Revised method of growing obstacles to use half the vehicle width rather than
the smallest integer multiple of the resolution larger than that.
Changed code so that resolution is printed only when an init command
is executed.

Revision 1.28  2006/10/23 16:25:36  tomrkramer
Implemented path planning by A* search followed by linearizing the path.
Removed planning by rubberbanding around boxes and planning by ray casting.

Revision 1.26  2006/10/13 16:26:32  tomrkramer
Changed path finding routine planPrimMoveSome3 to try only once each cycle
(rather than three times), since three times takes much too long.
Put in conditional compilation #define and #ifdef's to run timing tests
on how long it takes to get the vehicle position.
Changed use of moast.ini so vehicle width is not taken from the [AM_MOB_JA]
section. Rather, the width is taken from the DIMENSIONS of the vehicle.
being used.

Revision 1.25  2006/10/05 20:45:03  tomrkramer
Fixed bug in planPathLines.

Revision 1.24  2006/10/04 21:32:05  tomrkramer
Improved planner by (1) rebuilding function that plans from one
waypoint to the next so that it tries planPathLines first and planPathBox
second, (2) extending planPathLines so it tries paths with three
parts as wells as paths with two parts, (3) fixing serious bug
in planPathLines, (4) fixing minor bug in planPathBox.

Revision 1.23  2006/10/02 21:24:33  tomrkramer
Updated to implement new forms of AM spin and prim rotate commands.

Revision 1.22  2006/09/26 15:01:23  tomrkramer
Changed #include "amSPMap.hh" to #include "amSP.hh", so
AMSP_NOT_OBSTACLE, etc. would be defined.

Revision 1.21  2006/09/25 21:54:06  tomrkramer
Changed target point selector to tend to stay right. Removed code for
lineHop class, since lineHop is no longer used.

Revision 1.20  2006/09/25 16:54:04  tomrkramer
Fixed bugs in status reporting. changed "tollerance" to "tolerance".
Removed printing of error message if moveWaypoint fails (since detailed
error status is now being reported and amShell prints it).

Revision 1.19  2006/09/24 21:29:52  dr_steveb
Added spin command

Revision 1.18  2006/09/22 13:11:45  tomrkramer
Fixed two serious bugs in planPath2Lines.

Revision 1.17  2006/09/20 21:07:01  tomrkramer
Added more detailed status information in case of some errors.
Largely rewrote doCmdMoveWaypoint; included stopping vehicle if
planning fails.
Made AM planning map size independent of SP map size (as long as AM
map is inside SP map).
Adjusted calculations using lowerLeft, so it means the lower left of the
first cell (and hence, of the whole map), not the center of the first cell.

Revision 1.16  2006/09/18 21:11:32  tomrkramer
Added behavior that if planning fails, AM mobility will send a command to
Prim telling it to stop the vehicle.

Revision 1.15  2006/09/18 18:44:18  tomrkramer
Modified to add an additional method of rubberbanding around obstacles and
to work with new amSPMain that uses scrolling map tool. Also fixed a bug.

Revision 1.14  2006/04/20 12:49:30  tomrkramer
Removed heartbeat and replaced it with functions that work equivalently
by updating stat.tt.count (reset_time_tracker(&stat.tt) and
cycle_time_tracker(&stat.tt)).

Revision 1.13  2006/04/13 14:30:09  proctor
Handled NOP in primMobMain.cc, added NOP to primMisJA.hh and handled in
primMisMain.cc.
Added STAT_MSG_V2 to primMob,Mis, with new admin_state and timetracker.
Added "go" to servoShell.cc, and fixed an error setting the MOVE_TYPE enum.
Added handling of servoSPCmdGo in simpleSPSim.cc.
Fixed omitted return value in timeout_callback in callbacks.c, and added
a prototype for setTimerCallback.

Revision 1.12  2006/04/12 22:09:03  tomrkramer
Made minor improvements in fault resistance and documentation.

Revision 1.11  2006/04/11 22:16:45  tomrkramer
Added the geometry routines I thought I had already copied in that
are not yet being called (but have been tested).

Revision 1.10  2006/04/10 21:21:50  tomrkramer
Minor changes to implementation of administrative states.

Revision 1.9  2006/04/07 17:03:36  tomrkramer
Implemented using amMobJAError in status messages in move waypoint command.
Changed NO_OP to NOP and noOp to nop. Made use of "admin_state" consistent
in documentation.

Revision 1.8  2006/04/05 15:18:08  tomrkramer
Removed inappropriate checks.

Revision 1.7  2006/04/05 13:26:30  tomrkramer
Switched to using new admin_state and values for it. Added a lot of
geometry functions not yet being called.

Revision 1.6  2006/03/21 14:50:31  tomrkramer
moved setting stts.echo_serial_number = 0 and set.echo_serial_number = 0
to main from initParameters so that they are never again set to 0

Revision 1.5  2006/02/22 23:28:25  tomrkramer
finished implementing all planned functionality, but with
straight line paths and rubberbanding only.

Revision 1.4  2006/02/16 21:20:47  tomrkramer
added execution of dumpWm command; fixed planner bug;
put simple preference for staying right into path planning.

Revision 1.3  2006/02/15 15:51:22  tomrkramer
implemented straight line rubberbanding using boxing of obstacles

Revision 1.12  2005/10/06 22:58:56  kramer
minor fixes

Revision 1.11  2005/10/06 21:46:09  kramer
changed two comments

Revision 1.10  2005/10/06 14:17:04  kramer
improved handling of timer plus other minor improvements

Revision 1.9  2005/10/05 23:09:49  kramer
added documentation

Revision 1.8  2005/10/03 15:50:03  kramer
fixed exceeding array bounds bug

Revision 1.6  2005/09/19 14:59:34  kramer
major revision, added convert waypoints to lines

Revision 1.5  2005/07/01 16:01:14  proctor
Took out annoying .nml and .ini config printing in amMobMain.cc; added
amShell.cc

Revision 1.4  2005/06/29 21:08:08  proctor
Took out \defgroup comment for CVS history, since that's not what \defgroup
is used for.

Revision 1.3  2005/06/17 15:31:07  kramer
modularized, added move waypoint pass-through

Revision 1.2  2005/06/08 19:28:44  proctor
Moved state_match, etc. macros into moastTypes.hh

Revision 1.1  2005/06/08 15:51:52  proctor
Initial checkin
*/

